/* 
 * tkMain.c --
 *
 *	This file contains a generic main program for Tk-based applications.
 *	It can be used as-is for many applications, just by supplying a
 *	different appInitProc procedure for each specific application.
 *	Or, it can be used as a template for creating new main programs
 *	for Tk applications.
 *
 * Copyright (c) 1990-1994 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

static char sccsid[] = "@(#) tkMain.c 1.114 95/06/08 11:43:00";

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#    include "wndproc.h"
#    include "util.h"
#    include <io.h>
#    include <fcntl.h>
#    include <errno.h>
#    include <direct.h>
#endif /* _WINDOWS */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <signal.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif

/*
 * Declarations for various library procedures and variables (don't want
 * to include tkInt.h or tkPort.h here, because people might copy this
 * file out of the Tk source directory to make their own modified versions).
 * Note: don't declare "exit" here even though a declaration is really
 * needed, because it will conflict with a declaration elsewhere on
 * some systems.
 */

extern int		isatty _ANSI_ARGS_((int fd));
extern int		read _ANSI_ARGS_((int fd, char *buf, size_t size));
extern char *		strrchr _ANSI_ARGS_((CONST char *string, int c));

/*
 * Global variables used by the main program:
 */

static Tk_Window mainWindow;	/* The main window for the application.  If
				 * NULL then the application no longer
				 * exists. */
static Tcl_Interp *interp;	/* Interpreter for this application. */
static Tcl_DString command;	/* Used to assemble lines of terminal input
				 * into Tcl commands. */
static int tty;			/* Non-zero means standard input is a
				 * terminal-like device.  Zero means it's
				 * a file. */
static char errorExitCmd[] = "exit 1";

/*
 * Command-line options:
 */

static int synchronize = 0;
static int noMainWindow = 0;
static char *fileName = NULL;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;

static Tk_ArgvInfo argTable[] = {
    {"-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
	"Initial geometry for window"},
    {"-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
	"Display to use"},
    {"-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
	"Name to use for application"},
    {"-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
	"Use synchronous mode for display server"},
    {"-nomain", TK_ARGV_CONSTANT, (char *) 1, (char *) &noMainWindow,
	"Don't create main window"},
    {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
	(char *) NULL}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

static void		Prompt _ANSI_ARGS_((Tcl_Interp *interp, int partial));
static void		StdinProc _ANSI_ARGS_((ClientData clientData,
			    int mask));

/*
 * Called for a graceful exit on bus errors, etc.
 */

void
bye(int sig)
{
    static int called;		/* To prevent recursion */
    char temp[64];
    char *str;

    switch (sig) {
    /* Catch SIGINT, and keep going */
    case SIGINT:
	str = "interupt";
	break;
#ifdef SIGBUS
    case SIGBUS:
	str = "bus error";
	break;
#endif
    case SIGSEGV:
	str = "segmentation violation";
	break;
    case SIGTERM:
	str = "terminate";
	break;
    default:
	sprintf(temp, "unknown (%d).", sig);
	str = temp;
	break;
    }
    if (!called) {
	called = 1;
	fprintf(stderr, "Caught %s signal. Cleaning up and exiting.\n",
		str);
#ifdef COSMO
	clClose(interp);
#endif
	(void) Tcl_Eval(interp, "exit");
	
	exit(0);
    } else {
	fprintf(stderr, "Caught %s signal while cleaning up. Aborting.\n",
		str);
	abort();
    }
}

/*
 *--------------------------------------------------------------
 *
 * SplitCommandLine --
 *
 *	When cmwish is called as a shell script, the command line
 *	arguments are set a bit funny.  For example, suppose a script
 *	named "foo" has the following top line
 *
 *		#!/usr/local/bin/cmwish -nomain -f
 *
 *	And the command line is "./foo a b c"
 *	Then main() gets called with
 *		argc=6
 *		argv[1] = "-nomain -f"
 *		argv[2] = "./foo"
 *		argv[3] = "a"
 *		argv[4] = "b"
 *		argv[5] = "c"
 *	This function rearranges argc/argv so that the elements of
 *	argv[1] are separated.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The caller should free up the memory associated with argvPtr
 *	when it's done by calling free((char *)*argvPtr).
 *
 * Parameters:
 *	Tcl_Interp *interp	Interpreter to  use  for error reporting
 *	int *argcPtr		Pointer to argc from main. It will be
 *				modified
 *	char ***argvPtr		Pointer to argv from main. It will be
 *				modified.
 *
 *--------------------------------------------------------------
 */
static void
SplitCommandLine(Tcl_Interp *interp, int *argcPtr, char ***argvPtr)
{
    int newArgc;
    char **newArgv;
    int i, count, resultSize;
    char **result;

    if ((*argvPtr)[1] == NULL) return;

    Tcl_SplitList(interp, (*argvPtr)[1], &newArgc, &newArgv);

    /*
     * If there's no change, just free and return
     */
    if (newArgc == 1) {
	free((char *)newArgv);
	return;
    }

    /*
     * Merge lists
     */
    resultSize = newArgc+*argcPtr-1;
    result = (char **)ckalloc(resultSize*sizeof(char *));
    count = 0;
    result[count++] = (*argvPtr)[0];
    for (i=0; i<newArgc; i++) {
	result[count++] = newArgv[i];
    }
    free((char *)newArgv);
    for (i=2; i<*argcPtr; i++) {
	result[count++] = (*argvPtr)[i];
    }
    *argcPtr = count;
    *argvPtr = result;
}

/*
 *----------------------------------------------------------------------
 *
 * Tk_Main --
 *
 *	Main program for Wish.
 *
 * Results:
 *	None. This procedure never returns (it exits the process when
 *	it's done
 *
 * Side effects:
 *	This procedure initializes the wish world and then starts
 *	interpreting commands;  almost anything could happen, depending
 *	on the script being interpreted.
 *
 *----------------------------------------------------------------------
 */

void
Tk_Main(argc, argv, appInitProc)
    int argc;				/* Number of arguments. */
    char **argv;			/* Array of argument strings. */
    Tcl_AppInitProc *appInitProc;	/* Application-specific initialization
					 * procedure to call after most
					 * initialization but before starting
					 * to execute commands. */
{
    char *args, *p, *msg, *argv0;
    char buf[20];
    int code;
    size_t length;
#ifndef _WINDOWS
    int sock = 0;
#else
    SOCKET sockPair[2];
    SOCKET sock;
#endif

    interp = Tcl_CreateInterp();
#ifdef _WINDOWS
    Tcl_SetVar(interp, "tcl_msWindows", "1", TCL_GLOBAL_ONLY);
#else
    Tcl_SetVar(interp, "tcl_msWindows", "0", TCL_GLOBAL_ONLY);
#endif
#ifdef TCL_MEM_DEBUG
    Tcl_InitMemory(interp);
#endif

    SplitCommandLine (interp, &argc, &argv);

    /*
     * Parse command-line arguments.  A leading "-file" argument is
     * ignored (a historical relic from the distant past).  If the
     * next argument doesn't start with a "-" then strip it off and
     * use it as the name of a script file to process.  Also check
     * for other standard arguments, such as "-geometry", anywhere
     * in the argument list.
     */

    argv0 = argv[0];
    if (argc > 1) {
	length = strlen(argv[1]);
	if ((length >= 2) && (strncmp(argv[1], "-file", length) == 0)) {
	    argc--;
	    argv++;
	}
    }
    if ((argc > 1) && (argv[1][0] != '-')) {
	fileName = argv[1];
#ifdef _WINDOWS
        {
	    char *newf, *bp;
	    newf = ckalloc(strlen(fileName) + 1);
	    strcpy(newf, fileName);
	    fileName = newf;
	    for (bp = fileName; *bp; bp++) {
		if (*bp == '\\') {
		    *bp = '/';
		}
	    }
	}
#endif
	argc--;
	argv++;
    }
    if (Tk_ParseArgv(interp, (Tk_Window) NULL, &argc, argv, argTable, 0)
	    != TCL_OK) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }
    if (name == NULL) {
	if (fileName != NULL) {
	    p = fileName;
	} else {
	    p = argv[0];
	}
	name = strrchr(p, '/');
	if (name != NULL) {
	    name++;
	} else {
	    name = p;
	}
#ifdef _WINDOWS
	if (fileName == NULL) {
	    char *p1;
	    int len;

	    p1 = strrchr(name, '.');
	    if (p1 != NULL) {
		len = p1 - name;
		p1 = ckalloc(len+1);
		strncpy(p1, name, len);
		p1[len] = 0;
		name = p1;
	    }
	}
#endif
    }

    /*
     * If a display was specified, put it into the DISPLAY
     * environment variable so that it will be available for
     * any sub-processes created by us.
     */

    if (display != NULL) {
	Tcl_SetVar2(interp, "env", "DISPLAY", display, TCL_GLOBAL_ONLY);
    }

    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".  Also set the "geometry" variable from the geometry
     * specified on the command line.
     */

    args = Tcl_Merge(argc-1, argv+1);
    Tcl_SetVar(interp, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    sprintf(buf, "%d", argc-1);
    Tcl_SetVar(interp, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "argv0", (fileName != NULL) ? fileName : argv[0],
	    TCL_GLOBAL_ONLY);
    if (geometry != NULL) {
	Tcl_SetVar(interp, "geometry", geometry, TCL_GLOBAL_ONLY);
    }

    /*
     * Set the "tcl_interactive" variable.
     */
    tty = isatty(0);
    Tcl_SetVar(interp, "tcl_interactive",
	    ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /*
     * Initialize the Tk application.
     */
    if (!noMainWindow) {
	mainWindow = Tk_CreateMainWindow(interp, display, name, "Tk");
	if (mainWindow == NULL) {
	    fprintf(stderr, "%s\n", interp->result);
	    exit(1);
	}
	Tcm_SetDisplay(interp, Tk_Display(mainWindow));
	if (synchronize) {
	    XSynchronize(Tk_Display(mainWindow), True);
	}
	Tk_GeometryRequest(mainWindow, 200, 200);
    } else {
	Tcm_MakeDisplay(interp);
    }

    /*
     * Invoke application-specific initialization.
     */
    if (Tcl_Init(interp) != TCL_OK) {
	fprintf(stderr, "Tcl_Init failed: %s\n", interp->result);
	exit(1);
    }

    if (mainWindow && Tk_Init(interp) != TCL_OK) {
	fprintf(stderr, "Tk_Init failed: %s\n", interp->result);
	exit(1);
    }

    if (Tdp_Init(interp) != TCL_OK) {
	fprintf(stderr, "Tdp_Init failed: %s\n", interp->result);
	exit(1);
    }

#ifndef _WINDOWS
    if (Tns_Init(interp) != TCL_OK) {
	fprintf(stderr, "Tns_Init failed: %s\n", interp->result);
	exit(1);
    }
#endif

    if (CM_Init(interp,"") != TCL_OK) {
	fprintf(stderr, "CM_Init failed: %s\n", interp->result);
	exit(1);
    }

    if (VideoWidget_Init(interp) != TCL_OK) {
	fprintf(stderr,"## VideoWidget_Init failed!\n");
	exit(1);
    }

    /*
     * Specify a user-specific startup file to invoke if the application
     * is run interactively.  Typically the startup file is "~/.apprc"
     * where "app" is the name of the application.  If this line is deleted
     * then no user-specific startup file will be run under any conditions.
     */
    tcl_RcFileName = "~/.cmrc";

    /*
     * Set the geometry of the main window, if requested.
     */
    if (!noMainWindow && geometry != NULL) {
	code = Tcl_VarEval(interp, "wm geometry . ", geometry, (char *) NULL);
	if (code != TCL_OK) {
	    fprintf(stderr, "%s\n", interp->result);
	}
    }

    /*
     * Invoke the script specified on the command line, if any.
     */

    if (Tcl_Eval(interp, "CMP_InitCMWish") != TCL_OK) {
	fprintf(stderr, "CMP_InitCMWish failed: %s\n", interp->result);
	exit(1);
    }

    if (fileName != NULL) {
	code = Tcl_VarEval(interp, "source ", fileName, (char *) NULL);
	if (code != TCL_OK) {
	    goto error;
	}
	tty = 0;
    } else {
	/*
	 * Commands will come from standard input, so set up an event
	 * handler for standard input.  If the input device is aEvaluate the
	 * .rc file, if one has been specified, set up an event handler
	 * for standard input, and print a prompt if the input
	 * device is a terminal.
	 */

	if (tcl_RcFileName != NULL) {
	    Tcl_DString buffer;
	    char *fullName;
	    FILE *f;
    
	    fullName = Tcl_TildeSubst(interp, tcl_RcFileName, &buffer);
	    if (fullName == NULL) {
		fprintf(stderr, "%s\n", interp->result);
	    } else {
		f = fopen(fullName, "r");
		if (f != NULL) {
		    code = Tcl_EvalFile(interp, fullName);
		    if (code != TCL_OK) {
			fprintf(stderr, "%s\n", interp->result);
		    }
		    fclose(f);
		}
	    }
	    Tcl_DStringFree(&buffer);
	}
#ifdef _WINDOWS
	if (Tknt_GetHasConsole() && Tknt_GetUseSocket()) {
	    extern void panic();
	    int err;

	    err = Tknt_CreateSocketPair(sockPair);
	    if (err != 0) {
		char str[500];
		sprintf(str, "Unable to create connected sockets: %s\n",
			wsa_strerror(err));
		panic(str);
	    }
	    sock = sockPair[1];
	} else {
	    sock = INVALID_SOCKET;
	}
	CreateTclTerminal((Tknt_GetUseSocket() ?
			   (HANDLE) sockPair[0] : Tknt_GetPostWindow()),
			  Tknt_GetUseSocket());
#endif
	Tk_CreateFileHandler(sock, TK_READABLE, StdinProc, (ClientData) sock);
	if (tty) {
	    Prompt(interp, 0);
	}
    }
    fflush(stdout);
    Tcl_DStringInit(&command);
    Tcl_ResetResult(interp);

    /* Install signal handlers */
    signal(SIGINT, bye);
    signal(SIGTERM, bye);
#ifdef SIGBUS
    signal(SIGBUS, bye);
#endif
#ifdef SIGSEGV
    signal(SIGSEGV, bye);
#endif


    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we exit.
     */
    if (noMainWindow) {
	while (1) {
	    (void)Tk_DoOneEvent(0);
	}
    } else {
	Tk_MainLoop();
    }

    /*
     * Don't exit directly, but rather invoke the Tcl "exit" command.
     * This gives the application the opportunity to redefine "exit"
     * to do additional cleanup.
     */

    Tcl_Eval(interp, "exit");
    exit(1);

error:
    msg = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if (msg == NULL) {
	msg = interp->result;
    }
    fprintf(stderr, "%s\n", msg);
    Tcl_Eval(interp, errorExitCmd);
}

/*
 *----------------------------------------------------------------------
 *
 * StdinProc --
 *
 *	This procedure is invoked by the event dispatcher whenever
 *	standard input becomes readable.  It grabs the next line of
 *	input characters, adds them to a command being assembled, and
 *	executes the command if it's complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Could be almost arbitrary, depending on the command that's
 *	typed.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
static void
StdinProc(clientData, mask)
    ClientData clientData;		/* Not used. */
    int mask;				/* Not used. */
{
#define BUFFER_SIZE 4000
    char input[BUFFER_SIZE+1];
    static int gotPartial = 0;
    char *cmd;
    int code, count;
    int useSocket;
    int err;
#ifdef _WINDOWS
    SOCKET sock = (SOCKET) clientData;
    extern void TerminalThreadReenable();

    useSocket = (sock != INVALID_SOCKET);
    if (! useSocket) {
	char * pGlobal;
	HGLOBAL hGlobal;

	hGlobal = (HGLOBAL) GetWindowLong(Tknt_GetPostWindow(), 0);
	pGlobal = GlobalLock(hGlobal);
	strcpy(input, pGlobal);
	strcpy(pGlobal, "");
	GlobalUnlock(hGlobal);

	count = strlen(input);
    } else {
	int opt = 1;
	ioctlsocket(sock, FIONBIO, &opt);
	count = recv(sock, input, BUFFER_SIZE, 0);
	if (count < 0) {
	    err = WSAGetLastError();
	}
    }
#else
    int sock = 0;

    count = read(fileno(stdin), input, BUFFER_SIZE);
#endif /* _WINDOWS */
    if (count <= 0) {
	if (!gotPartial) {
	    if (tty) {
		Tcl_Eval(interp, "exit");
		exit(1);
	    } else {
		Tk_DeleteFileHandler(0);
	    }
	    return;
	} else {
	    count = 0;
	}
    }
    cmd = Tcl_DStringAppend(&command, input, count);
    if (count != 0) {
	if ((input[count-1] != '\n') && (input[count-1] != ';')) {
	    gotPartial = 1;
	    goto prompt;
	}
	if (!Tcl_CommandComplete(cmd)) {
	    gotPartial = 1;
	    goto prompt;
	}
    }
    gotPartial = 0;

    /*
     * Disable the stdin file handler while evaluating the command;
     * otherwise if the command re-enters the event loop we might
     * process commands from stdin before the current command is
     * finished.  Among other things, this will trash the text of the
     * command being evaluated.
     */

    Tk_CreateFileHandler(sock, 0, StdinProc, (ClientData) sock);
    code = Tcl_RecordAndEval(interp, cmd, TCL_EVAL_GLOBAL);
    Tk_CreateFileHandler(sock, TK_READABLE, StdinProc, (ClientData) sock);
    Tcl_DStringFree(&command);
    if (*interp->result != 0) {
	if ((code != TCL_OK) || (tty)) {
	    /*
	     * The statement below used to call "printf", but that resulted
	     * in core dumps under Solaris 2.3 if the result was very long.
	     */

	    puts(interp->result);
	}
    }

    /*
     * Output a prompt.
     */

    prompt:
#ifdef _WINDOWS
    TerminalThreadReenable();
#endif
    if (tty) {
	Prompt(interp, gotPartial);
    }
    Tcl_ResetResult(interp);
}

/*
 *----------------------------------------------------------------------
 *
 * Prompt --
 *
 *	Issue a prompt on standard output, or invoke a script
 *	to issue the prompt.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A prompt gets output, and a Tcl script may be evaluated
 *	in interp.
 *
 *----------------------------------------------------------------------
 */

static void
Prompt(interp, partial)
    Tcl_Interp *interp;			/* Interpreter to use for prompting. */
    int partial;			/* Non-zero means there already
					 * exists a partial command, so use
					 * the secondary prompt. */
{
    char *promptCmd;
    int code;

    promptCmd = Tcl_GetVar(interp,
	partial ? "tcl_prompt2" : "tcl_prompt1", TCL_GLOBAL_ONLY);
    if (promptCmd == NULL) {
	defaultPrompt:
	if (!partial) {
	    fputs("% ", stdout);
	}
    } else {
	code = Tcl_Eval(interp, promptCmd);
	if (code != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
		    "\n    (script that generates prompt)");
	    fprintf(stderr, "%s\n", interp->result);
	    goto defaultPrompt;
	}
    }
    fflush(stdout);
}

#ifndef _WINDOWS
void
main(argc, argv)
    int argc;
    char **argv;
{
    Tk_Main(argc, argv, NULL);
}
#else _WINDOWS
/*
 *----------------------------------------------------------------------
 * Formerly in wish/common_sh.c
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 *----------------------------------------------------------------------
 */

#ifndef MAXPATHLEN
#   ifdef PATH_MAX
#       define MAXPATHLEN PATH_MAX
#   else
#       define MAXPATHLEN 2048
#   endif
#endif
#define MAX_TITLE 1024

extern HWND CreateTerminal(HANDLE);
extern void compat_exit(int);
extern void compat_setisatty(int);

static HWND hConsoleWnd;
static int newconsole = 0;
static int iconic = 0;
static int tty = 0;		/* If this is 1, force a new console */
static CONSOLE_SCREEN_BUFFER_INFO consoleInfo;
static char oldTitle[MAX_TITLE];
static void Exit(void);

static void
CreateNewConsole() {
    FreeConsole();
    AllocConsole();
}

int
common_wish(int argc, char **argv, int hasWindows, int useSocket,
	    char *consoleTitle, int hasConsole, Tcl_AppInitProc *appInitProc)
{
    HANDLE hInstance = NULL, hPrevInstance = NULL;
    HWND hTerm, hwnd;
    HINSTANCE hLib;
    HHOOK hHook;
    HANDLE hStdout;
    COORD bufSize = {80, 200};
    char tmpName[MAX_TITLE];
    char title[MAX_TITLE];
    char *s;
    int match, prefixlen, suffix;
    WORD wVersionRequested;
    WSADATA wsaData;
    DWORD dwVersion;
    int version;
    int err;
    int i, j, len;
    char c, save;
    char *bp;
    char currdir[MAXPATHLEN+1];
    static char argv0_dir[MAXPATHLEN+1];
    extern void forkexec_setargv0_dir(char *);
    int hideconsole = 0;
    int showconsole = 0;

    Tknt_SetHasConsole(hasConsole);

    dwVersion = GetVersion();
    if ((dwVersion & 0x80000000) && (dwVersion & 0xff == 3)) {
	version = TKNT_WIN32S;
    } else {
	version = TKNT_WIN32;
    }

    Tknt_SetHasWindows(hasWindows);
    /* Convert backslashes to slashes */
    for (bp = argv[0]; *bp != '\0'; bp++) {
	if (*bp == '\\') {
	    *bp = '/';
	}
    }
	
    /* Extract the directory from argv[0] */
    len = strlen(argv[0]);
    for (i = len; i >= 0; i--) {
	c = argv[0][i];
	if (c == '/') {
	    break;
	}
    }
    i++;
    if (i == 0) {
	if (len >= 3 && argv[0][1] == ':') {
	    i = 2;
	}
    }

    if (i > 0) {
	save = argv[0][i];
	argv[0][i] = '\0';

	if (compat_getcwd(currdir, MAXPATHLEN+1) == NULL) {
	    panic("Unable to get working directory.");
	}
	if (chdir(argv[0]) == -1) {
	    panic("Unable to cd to %s\n", argv[0]);
	}
	if (compat_getcwd(argv0_dir, MAXPATHLEN+1) == NULL) {
	    panic("Unable to get working directory.");
	}
	if (chdir(currdir) == -1) {
	    panic("Unable to cd to %s\n", currdir);
	}
	argv[0][i] = save;
	len = strlen(argv0_dir);
	argv0_dir[len] = '/';
	argv0_dir[len+1] = '\0';
    } else {
	argv0_dir[0] = '\0';
    }

    forkexec_setargv0_dir(argv0_dir);

    for (i = 0; i < argc; i++) {
	if (strcmp(argv[i], "-console") == 0) {
	    newconsole = 1;
	    for (j = i + 1; j < argc; j++) {
		argv[j-1] = argv[j];
	    }
	    argv[j-1] = NULL;
	    i--; argc--;
	} else if (strcmp(argv[i], "-iconic") == 0) {
	    iconic = 1;
	    for (j = i + 1; j < argc; j++) {
		argv[j-1] = argv[j];
	    }
	    argv[j-1] = NULL;
	    i--; argc--;
	} else if (strcmp(argv[i], "-showconsole") == 0) {
	    showconsole = 1;
	    for (j = i + 1; j < argc; j++) {
		argv[j-1] = argv[j];
	    }
	    argv[j-1] = NULL;
	    i--; argc--;
	} else if (strcmp(argv[i], "-hideconsole") == 0) {
	    hideconsole = 1;
	    for (j = i + 1; j < argc; j++) {
		argv[j-1] = argv[j];
	    }
	    argv[j-1] = NULL;
	    i--; argc--;
	} else if (strcmp(argv[i], "-tty") == 0) {
	    compat_setisatty(1);
	    for (j = i + 1; j < argc; j++) {
		argv[j-1] = argv[j];
	    }
	    argv[j-1] = NULL;
	    i--; argc--;
	}
    }

    if (newconsole && hasConsole) {
	CreateNewConsole();
    }

    if (useSocket) {
	wVersionRequested = MAKEWORD(1, 1);
	err = WSAStartup(wVersionRequested, &wsaData);
	if (err != 0) {
	    panic("Could not find a useable winsock.dll\n");
	}
    }

    Tknt_SetUseSocket(useSocket);

    if (hasWindows) {
	Tknt_SetAppInstance(hInstance);

	if (!hPrevInstance) {
	    RegisterWidgetClass(hInstance);	
	    RegisterTopLevelClass(hInstance);
	    RegisterTerminalClass(hInstance);
	}

	hTerm = CreateTerminal(hInstance);  
	Tknt_SetPostWindow(hTerm);
    }

    if (hasWindows) {
#ifdef KANJI
	hLib = LoadLibrary("kxlib.dll");
#else
	hLib = LoadLibrary("lxlib.dll");
#endif

	hHook = SetWindowsHookEx(WH_MOUSE, 
				 (HOOKPROC)GetProcAddress(hLib, (LPCSTR)"DeliverEnterLeave"), 
				 hLib, (DWORD) GetCurrentThreadId() /*NULL*/); 
	SetHookDeliverEL(hHook);
	FreeLibrary(hLib);
    }

    if (hasConsole && hasWindows && version == TKNT_WIN32) {
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hStdout, &consoleInfo);
#if 0
	SetConsoleScreenBufferSize(hStdout, bufSize);
#endif
	GetConsoleTitle(oldTitle, MAX_TITLE);

	/* Get a Window handle for our console */
	s = "lr&gc";
	strcpy(tmpName, s);
	prefixlen = strlen(tmpName);

	for (suffix = 0; ; suffix++) {
	    match = 0;
	    itoa(suffix, &tmpName[prefixlen], 10);

	    len = strlen(tmpName);
	    SetConsoleTitle(tmpName);
	    for (hwnd = GetWindow(hTerm, GW_HWNDFIRST); IsWindow(hwnd);
		 hwnd = GetWindow(hwnd, GW_HWNDNEXT))
	    {
		if (GetWindowText(hwnd, title, MAX_TITLE) == len) {
		    if (strcmp(title, tmpName) == 0) {
			match++;
			hConsoleWnd = hwnd;
		    }
		}
	    }

	    if (match == 1) {
		break;
	    }
	}

	SetConsoleTitle(consoleTitle);
	Tknt_SetConsoleWindow(hConsoleWnd);
	if (hideconsole) {
	    ShowWindow(hConsoleWnd, SW_HIDE);
	} else if (showconsole) {
	    ShowWindow(hConsoleWnd, SW_SHOW);
	}
	if (iconic) {
	    SendMessage(hConsoleWnd, WM_SYSCOMMAND, SC_MINIMIZE, 0L);
	}
    } else if (!hasConsole) {
	if (iconic) {
	    ShowWindow(hTerm, SW_MINIMIZE);
	} else {
	    ShowWindow(hTerm, SW_SHOW);
	}
	UpdateWindow(hTerm);
    }

    atexit(Exit);
    Tk_Main(argc, argv, appInitProc);
    compat_exit(0);
    /* NOTREACHED */
    return 0;
}

int
common_tclsh(int argc, char **argv, int hasWindows, int useSocket,
	    char *consoleTitle, int hasConsole, Tcl_AppInitProc *appInitProc)
{
    return common_wish(argc, argv, hasWindows, useSocket,
		       consoleTitle, hasConsole, appInitProc);
}

void
Exit(void)
{
    HANDLE hStdout;

    fflush(stdout);
    fflush(stderr);

    if (Tknt_GetHasWindows()) {
	DestroyWindow(hConsoleWnd);
	DestroyTclTerminal();
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	SetConsoleScreenBufferSize(hStdout, consoleInfo.dwSize);
	SetConsoleTitle(oldTitle);
    }

    return;
}

void
main(argc, argv)
    int argc;
    char **argv;
{
    common_wish(argc, argv, 1, 1, "Wish", 1, NULL);
}
#endif
