/* 
 * main.c --
 *
 *	This file contains the main program for "wish", a windowing
 *	shell based on Tk and Tcl.  It also provides a template that
 *	can be used as the basis for main programs for other Tk
 *	applications.
 *
 * Copyright (c) 1990-1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char rcsid[] = "/usr/u/multimed/cmt/CVSROOT/cmt/cmx/tkMain.c,v 1.1.1.1 1995/02/17 23:17:47 bsmith Exp SPRITE (Berkeley)";
#endif

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#    include "wndproc.h"
#    include "util.h"
#    include <io.h>
#    include <fcntl.h>
#    include <errno.h>
#    include <direct.h>
#endif /* _WINDOWS */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <signal.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif

/*
 * Declarations for various library procedures and variables (don't want
 * to include tkInt.h or tkConfig.h here, because people might copy this
 * file out of the Tk source directory to make their own modified versions).
 */

extern int		read _ANSI_ARGS_((int fd, char *buf, size_t size));
extern int		isatty _ANSI_ARGS_((int fd));
extern char *		strrchr _ANSI_ARGS_((CONST char *string, int c));

/*
 * Global variables used by the main program:
 */

static char *myName;		/* argv[0]. */

static Tcl_Interp *interp;	/* Interpreter for this application. */
static Tcl_DString command;	/* Used to assemble lines of terminal input
				 * into Tcl commands. */
static int tty;			/* Non-zero means standard input is a
				 * terminal-like device.  Zero means it's
				 * a file. */
static char errorExitCmd[] = "exit";

/*
 * Command-line options:
 */

static int synchronize = 0;
static char *fileName = NULL;
static char *name = NULL;
static char *display = NULL;

static Tk_ArgvInfo argTable[] = {
    {"-file", TK_ARGV_STRING, (char *) NULL, (char *) &fileName,
	"File from which to read commands"},
    {"-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
	"Name to use for application"},
    {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
	(char *) NULL}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

static void		Prompt _ANSI_ARGS_((Tcl_Interp *interp, int partial));
static void		StdinProc _ANSI_ARGS_((ClientData clientData,
			    int mask));

/*
 * Called for a graceful exit on bus errors, etc.
 */

void
bye(int sig)
{
    static int called;		/* To prevent recursion */
    char temp[64];
    char *str;

    switch (sig) {
    /* Catch SIGINT, and keep going */
    case SIGINT:
	str = "interupt";
	return;
#ifdef SIGBUS
    case SIGBUS:
	str = "bus error";
	break;
#endif
    case SIGSEGV:
	str = "segmentation violation";
	break;
    case SIGTERM:
	str = "terminate";
	break;
    default:
	sprintf(temp, "unknown (%d).", sig);
	str = temp;
	break;
    }
    if (!called) {
	called = 1;
	fprintf(stderr, "%s: Caught %s signal. Cleaning up and exiting.\n",
		myName, str);
	(void) Tcl_Eval(interp, "exit");
	
	exit(0);
    } else {
	fprintf(stderr, "%s: Caught %s signal while cleaning up. Aborting.\n",
		myName, str);
	abort();
    }
}


/*
 *----------------------------------------------------------------------
 *
 * Tk_Main --
 *
 *	Main program for cmx.
 *
 * Results:
 *	None. This procedure never returns (it exits the process when
 *	it's done
 *
 * Side effects:
 *	This procedure initializes the wish world and then starts
 *	interpreting commands;  almost anything could happen, depending
 *	on the script being interpreted.
 *
 *----------------------------------------------------------------------
 */

void
Tk_Main(argc, argv, appInitProc)
    int argc;				/* Number of arguments. */
    char **argv;			/* Array of argument strings. */
    Tcl_AppInitProc *appInitProc;	/* Application-specific initialization
					 * procedure to call after most
					 * initialization but before starting
					 * to execute commands. */
{
    char *args, *p, *msg;
    char buf[20];
    int code;
#ifndef _WINDOWS
    int sock = 0;
#else
    SOCKET sockPair[2];
    SOCKET sock;
#endif

    myName = argv[0];

    interp = Tcl_CreateInterp();

#ifdef _WINDOWS
    Tcl_SetVar(interp, "tcl_msWindows", "1", TCL_GLOBAL_ONLY);
#else
    Tcl_SetVar(interp, "tcl_msWindows", "0", TCL_GLOBAL_ONLY);
#endif
#ifdef TCL_MEM_DEBUG
    Tcl_InitMemory(interp);
#endif

    /*
     * Parse command-line arguments.
     */
    if (Tk_ParseArgv(interp, (Tk_Window) NULL, &argc, argv, argTable, 0)
	    != TCL_OK) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }
    if (name == NULL) {
	if (fileName != NULL) {
	    p = fileName;
	} else {
	    p = argv[0];
	}
	name = strrchr(p, '/');
	if (name != NULL) {
	    name++;
	} else {
	    name = p;
	}
#ifdef _WINDOWS
	if (fileName == NULL) {
	    char *p1;
	    int len;

	    p1 = strrchr(name, '.');
	    if (p1 != NULL) {
		len = p1 - name;
		p1 = ckalloc(len+1);
		strncpy(p1, name, len);
		p1[len] = 0;
		name = p1;
	    }
	}
#endif
    }

    /*
     * If a display was specified, put it into the DISPLAY
     * environment variable so that it will be available for
     * any sub-processes created by us.
     */

    if (display != NULL) {
	Tcl_SetVar2(interp, "env", "DISPLAY", display, TCL_GLOBAL_ONLY);
    }

    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".  
     */

    args = Tcl_Merge(argc-1, argv+1);
    Tcl_SetVar(interp, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    sprintf(buf, "%d", argc-1);
    Tcl_SetVar(interp, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "argv0", (fileName != NULL) ? fileName : argv[0],
	    TCL_GLOBAL_ONLY);

    /*
     * Set the "tcl_interactive" variable.
     */

    tty = isatty(0);
    Tcl_SetVar(interp, "tcl_interactive",
	    ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /*
     * Initialize all the modules
     */
    if (Tcl_Init(interp) != TCL_OK) {
	fprintf(stderr, "Tcl_AppInit failed: %s\n", interp->result);
	exit(1);
    }

    if (Tkx_Init(interp) != TCL_OK) {
	fprintf(stderr, "Tcl_AppInit failed: %s\n", interp->result);
	exit(1);
    }

    if (Tdp_Init(interp) != TCL_OK) {
	fprintf(stderr, "Tdp_Init failed: %s\n", interp->result);
	exit(1);
    }

    if (Tns_Init(interp) != TCL_OK) {
	fprintf(stderr, "Tns_Init failed: %s\n", interp->result);
	exit(1);
    }

    if (CM_Init(interp,"CMP_InitCMX") != TCL_OK) {
	fprintf(stderr, "CM_Init failed: %s\n", interp->result);
	exit(1);
    }

    /*
     * Specify the user-specific startup file to invoke if the application
     * is run interactively.
     */
    tcl_RcFileName = "~/.cmrc";

    /*
     * Invoke the script specified on the command line, if any.
     */
    if (fileName != NULL) {
	code = Tcl_VarEval(interp, "source ", fileName, (char *) NULL);
	if (code != TCL_OK) {
	    goto error;
	}
	tty = 0;
    } else {
	/*
	 * Commands will come from standard input, so set up an event
	 * handler for standard input.  If the input device is aEvaluate the
	 * .rc file, if one has been specified, set up an event handler
	 * for standard input, and print a prompt if the input
	 * device is a terminal.
	 */

	if (tcl_RcFileName != NULL) {
	    Tcl_DString buffer;
	    char *fullName;
	    FILE *f;
    
	    fullName = Tcl_TildeSubst(interp, tcl_RcFileName, &buffer);
	    if (fullName == NULL) {
		fprintf(stderr, "%s\n", interp->result);
	    } else {
		f = fopen(fullName, "r");
		if (f != NULL) {
		    code = Tcl_EvalFile(interp, fullName);
		    if (code != TCL_OK) {
			fprintf(stderr, "%s\n", interp->result);
		    }
		    fclose(f);
		}
	    }
	    Tcl_DStringFree(&buffer);
	}
#ifdef _WINDOWS
	if (Tknt_GetHasConsole() && Tknt_GetUseSocket()) {
	    extern void panic();
	    int err;

	    err = Tknt_CreateSocketPair(sockPair);
	    if (err != 0) {
		char str[500];
		sprintf(str, "Unable to create connected sockets: %s\n",
			wsa_strerror(err));
		panic(str);
	    }
	    sock = sockPair[1];
	} else {
	    sock = INVALID_SOCKET;
	}
	CreateTclTerminal((Tknt_GetUseSocket() ?
			   (HANDLE) sockPair[0] : Tknt_GetPostWindow()),
			  Tknt_GetUseSocket());
#endif
	Tk_CreateFileHandler(sock, TK_READABLE, StdinProc, (ClientData) sock);
	if (tty) {
	    Prompt(interp, 0);
	}
    }
    fflush(stdout);
    Tcl_DStringInit(&command);

    /*
     * Install signal handlers.  These have to be installed if we
     * want application-specific cleanup actions to occur (e.g., freeing
     * shared memory buffers).
     */
    signal(SIGINT, bye);
    signal(SIGTERM, bye);
#ifdef SIGBUS
    signal(SIGBUS, bye);
#endif
#ifdef SIGSEGV
    signal(SIGSEGV, bye);
#endif

    /*
     * Loop infinitely, waiting for commands to execute. 
     */
    while (1) {
        Tk_DoOneEvent(0);
    }

error:
    msg = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if (msg == NULL) {
	msg = interp->result;
    }
    fprintf(stderr, "%s\n", msg);
    Tcl_Eval(interp, errorExitCmd);
    bye(SIGTERM);
}

/*
 *----------------------------------------------------------------------
 *
 * StdinProc --
 *
 *	This procedure is invoked by the event dispatcher whenever
 *	standard input becomes readable.  It grabs the next line of
 *	input characters, adds them to a command being assembled, and
 *	executes the command if it's complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Could be almost arbitrary, depending on the command that's
 *	typed.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
static void
StdinProc(clientData, mask)
    ClientData clientData;		/* Not used. */
    int mask;				/* Not used. */
{
#define BUFFER_SIZE 4000
    char input[BUFFER_SIZE+1];
    static int gotPartial = 0;
    char *cmd;
    int code, count;
    int useSocket;
    int err;
#ifdef _WINDOWS
    SOCKET sock = (SOCKET) clientData;
    extern void TerminalThreadReenable();

    useSocket = (sock != INVALID_SOCKET);
    if (! useSocket) {
	char * pGlobal;
	HGLOBAL hGlobal;

	hGlobal = (HGLOBAL) GetWindowLong(Tknt_GetPostWindow(), 0);
	pGlobal = GlobalLock(hGlobal);
	strcpy(input, pGlobal);
	strcpy(pGlobal, "");
	GlobalUnlock(hGlobal);

	count = strlen(input);
    } else {
	int opt = 1;
	ioctlsocket(sock, FIONBIO, &opt);
	count = recv(sock, input, BUFFER_SIZE, 0);
	if (count < 0) {
	    err = WSAGetLastError();
	}
    }
#else
    int sock = 0;

    count = read(fileno(stdin), input, BUFFER_SIZE);
#endif /* _WINDOWS */
    if (count <= 0) {
	if (!gotPartial) {
	    if (tty) {
		Tcl_Eval(interp, "exit");
		exit(1);
	    } else {
		Tk_DeleteFileHandler(0);
	    }
	    return;
	} else {
	    count = 0;
	}
    }
    cmd = Tcl_DStringAppend(&command, input, count);
    if (count != 0) {
	if ((input[count-1] != '\n') && (input[count-1] != ';')) {
	    gotPartial = 1;
	    goto prompt;
	}
	if (!Tcl_CommandComplete(cmd)) {
	    gotPartial = 1;
	    goto prompt;
	}
    }
    gotPartial = 0;

    /*
     * Disable the stdin file handler while evaluating the command;
     * otherwise if the command re-enters the event loop we might
     * process commands from stdin before the current command is
     * finished.  Among other things, this will trash the text of the
     * command being evaluated.
     */

    Tk_CreateFileHandler(sock, 0, StdinProc, (ClientData) sock);
    code = Tcl_RecordAndEval(interp, cmd, TCL_EVAL_GLOBAL);
    Tk_CreateFileHandler(sock, TK_READABLE, StdinProc, (ClientData) sock);
    Tcl_DStringFree(&command);
    if (*interp->result != 0) {
	if ((code != TCL_OK) || (tty)) {
	    /*
	     * The statement below used to call "printf", but that resulted
	     * in core dumps under Solaris 2.3 if the result was very long.
	     */

	    puts(interp->result);
	}
    }

    /*
     * Output a prompt.
     */

    prompt:
#ifdef _WINDOWS
    TerminalThreadReenable();
#endif
    if (tty) {
	Prompt(interp, gotPartial);
    }
    Tcl_ResetResult(interp);
}

/*
 *----------------------------------------------------------------------
 *
 * Prompt --
 *
 *	Issue a prompt on standard output, or invoke a script
 *	to issue the prompt.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A prompt gets output, and a Tcl script may be evaluated
 *	in interp.
 *
 *----------------------------------------------------------------------
 */

static void
Prompt(interp, partial)
    Tcl_Interp *interp;			/* Interpreter to use for prompting. */
    int partial;			/* Non-zero means there already
					 * exists a partial command, so use
					 * the secondary prompt. */
{
    char *promptCmd;
    int code;

    promptCmd = Tcl_GetVar(interp,
	partial ? "tcl_prompt2" : "tcl_prompt1", TCL_GLOBAL_ONLY);
    if (promptCmd == NULL) {
	defaultPrompt:
	if (!partial) {
	    fputs("cmx: ", stdout);
	}
    } else {
	code = Tcl_Eval(interp, promptCmd);
	if (code != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
		    "\n    (script that generates prompt)");
	    fprintf(stderr, "%s\n", interp->result);
	    goto defaultPrompt;
	}
    }
    fflush(stdout);
}

#ifndef _WINDOWS
void
main(argc, argv)
    int argc;
    char **argv;
{
    Tk_Main(argc, argv, NULL);
}
#else _WINDOWS
/*
 *----------------------------------------------------------------------
 * Formerly in wish/common_sh.c
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 *----------------------------------------------------------------------
 */

#ifndef MAXPATHLEN
#   ifdef PATH_MAX
#       define MAXPATHLEN PATH_MAX
#   else
#       define MAXPATHLEN 2048
#   endif
#endif
#define MAX_TITLE 1024

extern HWND CreateTerminal(HANDLE);
extern void compat_setisatty(int);

static HWND hConsoleWnd;
static int newconsole = 0;
static int iconic = 0;
static int tty = 0;		/* If this is 1, force a new console */
static CONSOLE_SCREEN_BUFFER_INFO consoleInfo;
static char oldTitle[MAX_TITLE];
static void Exit(void);

static void
CreateNewConsole() {
    FreeConsole();
    AllocConsole();
}

int
common_wish(int argc, char **argv, int hasWindows, int useSocket,
	    char *consoleTitle, int hasConsole, Tcl_AppInitProc *appInitProc)
{
    HANDLE hInstance = NULL, hPrevInstance = NULL;
    HWND hTerm, hwnd;
    HINSTANCE hLib;
    HHOOK hHook;
    HANDLE hStdout;
    COORD bufSize = {80, 200};
    char tmpName[MAX_TITLE];
    char title[MAX_TITLE];
    char *s;
    int match, prefixlen, suffix;
    WORD wVersionRequested;
    WSADATA wsaData;
    DWORD dwVersion;
    int version;
    int err;
    int i, j, len;
    char c, save;
    char *bp;
    char currdir[MAXPATHLEN+1];
    static char argv0_dir[MAXPATHLEN+1];
    extern void forkexec_setargv0_dir(char *);

    Tknt_SetHasConsole(hasConsole);

    dwVersion = GetVersion();
    if ((dwVersion & 0x80000000) && (dwVersion & 0xff == 3)) {
	version = TKNT_WIN32S;
    } else {
	version = TKNT_WIN32;
    }

    Tknt_SetHasWindows(hasWindows);
    /* Convert backslashes to slashes */
    for (bp = argv[0]; *bp != '\0'; bp++) {
	if (*bp == '\\') {
	    *bp = '/';
	}
    }
	
    /* Extract the directory from argv[0] */
    len = strlen(argv[0]);
    for (i = len; i >= 0; i--) {
	c = argv[0][i];
	if (c == '/') {
	    break;
	}
    }
    i++;
    if (i == 0) {
	if (len >= 3 && argv[0][1] == ':') {
	    i = 2;
	}
    }

    if (i > 0) {
	save = argv[0][i];
	argv[0][i] = '\0';

	if (compat_getcwd(currdir, MAXPATHLEN+1) == NULL) {
	    panic("Unable to get working directory.");
	}
	if (chdir(argv[0]) == -1) {
	    panic("Unable to cd to %s\n", argv[0]);
	}
	if (compat_getcwd(argv0_dir, MAXPATHLEN+1) == NULL) {
	    panic("Unable to get working directory.");
	}
	if (chdir(currdir) == -1) {
	    panic("Unable to cd to %s\n", currdir);
	}
	argv[0][i] = save;
	len = strlen(argv0_dir);
	argv0_dir[len] = '/';
	argv0_dir[len+1] = '\0';
    } else {
	argv0_dir[0] = '\0';
    }

    forkexec_setargv0_dir(argv0_dir);

    for (i = 0; i < argc; i++) {
	if (strcmp(argv[i], "-console") == 0) {
	    newconsole = 1;
	    for (j = i + 1; j < argc; j++) {
		argv[j-1] = argv[j];
	    }
	    argv[j-1] = NULL;
	    i--; argc--;
	} else if (strcmp(argv[i], "-iconic") == 0) {
	    iconic = 1;
	    for (j = i + 1; j < argc; j++) {
		argv[j-1] = argv[j];
	    }
	    argv[j-1] = NULL;
	    i--; argc--;
	} else if (strcmp(argv[i], "-tty") == 0) {
	    compat_setisatty(1);
	    for (j = i + 1; j < argc; j++) {
		argv[j-1] = argv[j];
	    }
	    argv[j-1] = NULL;
	    i--; argc--;
	}
    }

    if (newconsole && hasConsole) {
	CreateNewConsole();
    }

    if (useSocket) {
	wVersionRequested = MAKEWORD(1, 1);
	err = WSAStartup(wVersionRequested, &wsaData);
	if (err != 0) {
	    panic("Could not find a useable winsock.dll\n");
	}
    }

    Tknt_SetUseSocket(useSocket);

    if (hasWindows) {
	Tknt_SetAppInstance(hInstance);

	if (!hPrevInstance) {
	    RegisterWidgetClass(hInstance);	
	    RegisterTopLevelClass(hInstance);
	    RegisterTerminalClass(hInstance);
	}

	hTerm = CreateTerminal(hInstance);  
	Tknt_SetPostWindow(hTerm);
    }

    if (hasWindows) {
#ifdef KANJI
	hLib = LoadLibrary("kxlib.dll");
#else
	hLib = LoadLibrary("lxlib.dll");
#endif

	hHook = SetWindowsHookEx(WH_MOUSE, 
				 (HOOKPROC)GetProcAddress(hLib, (LPCSTR)"DeliverEnterLeave"), 
				 hLib, (DWORD) GetCurrentThreadId() /*NULL*/); 
	SetHookDeliverEL(hHook);
	FreeLibrary(hLib);
    }

    if (hasConsole && hasWindows && version == TKNT_WIN32) {
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hStdout, &consoleInfo);
	SetConsoleScreenBufferSize(hStdout, bufSize);

	GetConsoleTitle(oldTitle, MAX_TITLE);

	/* Get a Window handle for our console */
	s = "lr&gc";
	strcpy(tmpName, s);
	prefixlen = strlen(tmpName);

	for (suffix = 0; ; suffix++) {
	    match = 0;
	    itoa(suffix, &tmpName[prefixlen], 10);

	    len = strlen(tmpName);
	    SetConsoleTitle(tmpName);
	    for (hwnd = GetWindow(hTerm, GW_HWNDFIRST); IsWindow(hwnd);
		 hwnd = GetWindow(hwnd, GW_HWNDNEXT))
	    {
		if (GetWindowText(hwnd, title, MAX_TITLE) == len) {
		    if (strcmp(title, tmpName) == 0) {
			match++;
			hConsoleWnd = hwnd;
		    }
		}
	    }

	    if (match == 1) {
		break;
	    }
	}

	SetConsoleTitle(consoleTitle);
	if (iconic) {
	    SendMessage(hConsoleWnd, WM_SYSCOMMAND, SC_MINIMIZE, 0L);
	}
    } else if (!hasConsole) {
	if (iconic) {
	    ShowWindow(hTerm, SW_MINIMIZE);
	} else {
	    ShowWindow(hTerm, SW_SHOW);
	}
	UpdateWindow(hTerm);
    }

    atexit(Exit);
    Tk_Main(argc, argv, appInitProc);
    compat_exit(0);
    /* NOTREACHED */
    return 0;
}

int
common_tclsh(int argc, char **argv, int hasWindows, int useSocket,
	    char *consoleTitle, int hasConsole, Tcl_AppInitProc *appInitProc)
{
    return common_wish(argc, argv, hasWindows, useSocket,
		       consoleTitle, hasConsole, appInitProc);
}

void
Exit(void)
{
    HANDLE hStdout;

    fflush(stdout);
    fflush(stderr);

    if (Tknt_GetHasWindows()) {
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	DestroyWindow(hConsoleWnd);
	DestroyTclTerminal();
	SetConsoleScreenBufferSize(hStdout, consoleInfo.dwSize);
	SetConsoleTitle(oldTitle);
    }

    ExitProcess(0);
    return;
}

void
main(argc, argv)
    int argc;
    char **argv;
{
    common_wish(argc, argv, 1, 1, "Wish", 1, NULL);
}
#endif
