/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>
#include "cmAuPlay.h"

#ifdef AUDIO_FILE
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmAuFile.h"
#include "cmLTS.h"
#include <AF/AFlib.h>

typedef struct AFDevice {
    char *conn_name;            /* Name of AF server to connect to */
    AFAudioConn *connection;    /* Connection to the AF server */
    AC ac;                      /* audio context in AF server */
    int device;                 /* The device number to use */
    int silence;		/* Do we need to send silence? */
    int mingain;
    int maxgain;
} AFDevice;

/*
 * Interface routines
 */
static int AFDeviceInit(SpeakerRsrc *objPtr);
static int AFDeviceReinit(SpeakerRsrc *objPtr);
static int AFDeviceShutdown(SpeakerRsrc *objPtr);
static int AFDeviceUnready(SpeakerRsrc *objPtr);
static int AFDeviceLTSChanged(SpeakerRsrc *objPtr);
static int AFDevicePlay(CBInfo *cbPtr);
static char *AFDeviceError(SpeakerRsrc *objPtr, int errNum);
    
/*
 * Internal routines
 */
static int FindDefaultDevice(AFAudioConn *aud);
static void AF_SendSilence(AFDevice *afdev);
static void AFDeviceSetGain(AFDevice *afdev, double gain);

/*
 * Error codes
 */
#define NO_ERROR	0
#define CANT_OPEN	1
#define BAD_AC		2
#define BAD_DEVICE	3
#define	BAD_SAMPLE_RATE	4

/*
 * -------------------------------------------------------
 *
 * GetAFDevice 
 *
 *	Allocates a new SpeakerDevice structure, filled
 *	in to handle AF and return it.
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetAFDevice(SpeakerRsrc *objPtr)
{
    SpeakerDevice *sdev;
    AFDevice *afdev;
    
    afdev = (AFDevice *) ckalloc(sizeof(AFDevice));
    afdev->conn_name = NULL;
    afdev->connection = NULL;
    afdev->ac = NULL;
    afdev->device = -1;
    afdev->mingain = afdev->maxgain = 0;
    
    sdev = (SpeakerDevice *) ckalloc(sizeof(SpeakerDevice));
    sdev->deviceData =	(ClientData) afdev;
    sdev->initProc =	AFDeviceInit;
    sdev->reinitProc =	AFDeviceReinit;
    sdev->shutdownProc =	AFDeviceShutdown;
    sdev->unreadyProc =	AFDeviceUnready;
    sdev->ltsProc =	AFDeviceLTSChanged;
    sdev->playProc =	AFDevicePlay;
    sdev->errorProc =	AFDeviceError;
    
    return sdev;
}

/*
 * -------------------------------------------------------
 *
 * AFDeviceInit
 *
 * Gets the AFDevice ready to play sound -- opens a
 * connection to the AF server if necessary.
 *
 * -------------------------------------------------------
 */

static int
AFDeviceInit(SpeakerRsrc *objPtr) {
    AFDevice *afdev;
    AFSetACAttributes atts;
    unsigned int device;
    AFAudioConn *connection;
    
    afdev = (AFDevice *) objPtr->devicePtr->deviceData;
    
    if (afdev->connection != NULL) {
	AFCloseAudioConn(afdev->connection);
    }
    
    connection = afdev->connection = AFOpenAudioConn(afdev->conn_name);
    if (connection == NULL) {
	return CANT_OPEN;
    }
    
    /* Need to set anything in the AC?  Endian-ness?  Encoding Type? */

    if (afdev->device == -1) {
	device = FindDefaultDevice(connection);
    } else if (afdev->device >= ANumberOfAudioDevices(connection)) {
	return BAD_DEVICE;
    } else {
	device = afdev->device;
    }
    afdev->ac = AFCreateAC(afdev->connection, device, 0, &atts);
    if (afdev->ac == NULL) {
	return BAD_AC;
    }
    
    AFQueryOutputGain(afdev->ac, &(afdev->mingain), &(afdev->maxgain));

    AFDeviceSetGain(afdev, objPtr->gain);
    
    return NO_ERROR;
}

/*
 * -------------------------------------------------------
 *
 * AFDeviceReinit
 *
 * Re-initialize the device
 *
 * -------------------------------------------------------
 */

static int
AFDeviceReinit(SpeakerRsrc *objPtr) {
    AFDeviceSetGain((AFDevice *) objPtr->devicePtr->deviceData, objPtr->gain);
    return NO_ERROR;
}

/*
 * -------------------------------------------------------
 *
 * AFDeviceShutdown
 *
 * Sends silence to the speaker and closes the connection
 * to the AudioFile server to shutdown the device
 *
 * -------------------------------------------------------
 */

static int
AFDeviceShutdown(SpeakerRsrc *objPtr)
{
    AFDevice *afdev;
    
    afdev = (AFDevice *) objPtr->devicePtr->deviceData;
    
    AF_SendSilence(afdev);
    
    if (afdev->connection != NULL) {
	AFCloseAudioConn(afdev->connection);
    }
    
    return NO_ERROR;
}

/*
 * -------------------------------------------------------
 *
 * AFDeviceUnready
 *
 * Sends silence to the speaker when the device is unreadied
 *
 * -------------------------------------------------------
 */

static int
AFDeviceUnready(SpeakerRsrc *objPtr)
{
    AF_SendSilence((AFDevice *) objPtr->devicePtr->deviceData);
    return NO_ERROR;
}

/*
 * -------------------------------------------------------
 *
 * AFDeviceLTSChanged
 *
 *	Notify the AFDevice that the LTS has changed
 *
 * -------------------------------------------------------
 */

static int
AFDeviceLTSChanged (SpeakerRsrc *objPtr)
{
    AFDevice *afdev;
    
    afdev = (AFDevice *) objPtr->devicePtr->deviceData;
    
    if (afdev->silence == 1) {
	AF_SendSilence((AFDevice *) objPtr->devicePtr->deviceData);
    }
    afdev->silence = 0;
    return NO_ERROR;
}


/*
 * -------------------------------------------------------
 *
 * AFDevicePlay
 *
 * Play the data in the given structure
 *
 * -------------------------------------------------------
 */

static int
AFDevicePlay(CBInfo *cbPtr)
{
    SpeakerRsrc *objPtr;
    AFDevice *afdev;
    unsigned int amask, playTime;
    int format;
    AFSetACAttributes attrs;
    double now, start;

    objPtr = cbPtr->objPtr;
    afdev = (AFDevice *) objPtr->devicePtr->deviceData;
    
    /*
     * Check to make sure this sampling rate is
     * reasonably close to what the device expects
     */
    if ((abs(afdev->ac->device->playSampleFreq
	     - cbPtr->hdr.samplingRate)) > 200) {
	/* Not much we can do at this point */
	return BAD_SAMPLE_RATE;
    }

    /*
     * Set the format.  AF will take care of
     * necessary format conversion
     */
    amask = 0;
    format = cbPtr->hdr.dataFormat;
    if ((format == SUN_ULAW)
	&& (afdev->ac->attributes.type != MU255)) {
	attrs.type = MU255;
	amask |= ACEncodingType;
    } else if ((format == SUN_LIN_8)
	       && (afdev->ac->attributes.type != LIN8)) {
	attrs.type = LIN8;
	amask |= ACEncodingType;
    } else if ((format == SUN_LIN_16)
	       && (afdev->ac->attributes.type != LIN16)) {
	attrs.type = LIN16;
	amask |= ACEncodingType;
    }
    
    if (amask != 0) {
	AFChangeACAttributes(afdev->ac, amask, &attrs);
    }
    
    /* AFSync? */

    /*
     * Hand it off to the AF server
     */
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));
    start = cbPtr->start;
    playTime = AFGetTime(afdev->ac);

    if (now <= start) {
	playTime += ((int) ((start-now) * ((double) cbPtr->hdr.samplingRate)));
    }
    AFPlaySamples(afdev->ac, playTime, cbPtr->len, cbPtr->data);
    afdev->silence = 1;
    
    return NO_ERROR;
}

/*
 * -------------------------------------------------------
 *
 * FindDefaultDevice
 *
 * (Taken from AF source code)
 * Finds a default device to use if one is not specified
 *
 * -------------------------------------------------------
 */
static int
FindDefaultDevice(AFAudioConn *aud)
{
    AFDeviceDescriptor *aDev;
    int     i;

    for(i=0; i<ANumberOfAudioDevices(aud); i++) {
	aDev = AAudioDeviceDescriptor(aud, i);
	if(aDev->inputsFromPhone == 0 && aDev->outputsToPhone == 0)
	    return i;
    }
    return -1;
}

/*
 * -------------------------------------------------------
 *
 * AF_SendSilence
 *
 * Sends silence to the speaker, overriding any
 * previous data that had been sent
 *
 * -------------------------------------------------------
 */
static void
AF_SendSilence(AFDevice *afdev)
{
    int i;
    AFSetACAttributes attributes;
    unsigned char tmp[2048];
    ATime tm;
    
    /*
     * Send silence to the speaker
     */
    if (afdev->ac != NULL) {
	attributes = afdev->ac->attributes;
	AFSilence(attributes.type, tmp, 2048);
	attributes.preempt = 1;
	AFChangeACAttributes(afdev->ac, ACPreemption, &attributes);
	tm = AFGetTime(afdev->ac);
	for (i=0; i<16; i++) {
	    AFPlaySamples(afdev->ac, tm + (i*2048), 2048, tmp);
	}
    }
}

/*
 * -------------------------------------------------------
 *
 * AFDeviceSetGain
 *
 * Sets the gain on the device
 *
 * -------------------------------------------------------
 */
static void
AFDeviceSetGain(AFDevice *afdev, double gain)
{
    double newgain;
    
    if (gain > 1.0) {
	newgain = afdev->maxgain;
    } else if (gain < 0.0) {
	newgain = afdev->mingain;
    } else {
	newgain = afdev->mingain + gain * (afdev->maxgain - afdev->mingain);
    }

    AFSetOutputGain(afdev->ac, (int) newgain);
}
/*
 * -------------------------------------------------------
 *
 * AFDeviceError
 *
 * Map error numbers to strings describing the error
 *
 * -------------------------------------------------------
 */

static char *
AFDeviceError(SpeakerRsrc *objPtr, int errNum)
{
    switch (errNum) {
      case CANT_OPEN:
	return "Cannot open connection to AF Server";
	
      case BAD_AC:
	return "Cannot create AC";
	
      case BAD_DEVICE:
	return "Invalid AF device number";
	
      case BAD_SAMPLE_RATE:
	return "Sampling rate incompatible with audio device";
	
      default:
	return "Unknown AudioFile error";
    }
}

#else /* no AUDIO_FILE */

/*
 * -------------------------------------------------------
 *
 * GetAFDevice
 *
 *      Stub routine for non sparc machines
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetAFDevice(SpeakerRsrc *objPtr)
{
    return (SpeakerDevice *)NULL;
}

#endif /* AUDIO_FILE */
