/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 *  atom.c
 *
 *	To implement atom to ClientData mapping.
 *	Atoms are also sometimes also called handles.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include "tcl.h"
#include "cmAtom.h"

static Tcl_HashTable atoms;
static int atomsInited = 0;

static Tcl_Interp *interpFirst;

typedef struct {
    ClientData funcPtr;
    ClientData funcCD;
} FunctionRegisterEntry;


/*
 *--------------------------------------------------------------
 *
 * Tutil_AtomInit --
 *
 *	This procedure is invoked to initialize the atoms module;
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

 /* ARGSUSED */
int
Tutil_AtomInit(Tcl_Interp *interp)
                       		/* Current interpreter. */
{
    interpFirst = interp;

    if (!atomsInited)
	Tcl_InitHashTable (&atoms, TCL_STRING_KEYS);

    atomsInited = 1;

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tutil_FindAtom --
 *
 *	Find the ClientData associated with a given atom.
 *
 * Results:
 *	The ClientData of the atom if found or NULL, otherwise.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

ClientData
Tutil_FindAtom(Tcl_Interp *interp, char *handle)
{
    Tcl_HashEntry *entryPtr;
    FILE *filePtr;

    if (!handle)
	return NULL;

    if ((interp) &&
	((int)strlen(handle) > 4) &&
	(handle[0] == 'f') &&
	(handle[1] == 'i') &&
	(handle[2] == 'l') &&
	(handle[3] == 'e')) {
	if (Tcl_GetOpenFile(interp, handle, 0, 0, &filePtr) == TCL_OK)
	    return (ClientData) filePtr;
    }
    if (!atomsInited) {
	Tcl_InitHashTable(&atoms, TCL_STRING_KEYS);
	atomsInited = 1;
    }
    if (!(entryPtr = Tcl_FindHashEntry(&atoms, handle)))
	return NULL;

    return (Tcl_GetHashValue(entryPtr));
}

/*
 *--------------------------------------------------------------
 *
 * Tutil_RegisterAtom --
 *
 *	Associate ClientData with a given atom.
 *
 * Results:
 *	Returns -1 on failure, 0 if the atom is newly
 *	registered, and 1 if we're overwriting a
 *	previous registration.
 *
 * Side effects:
 *	An atom is registered in a hash table;
 *
 *--------------------------------------------------------------
 */

 /* ARGSUSED */
int
Tutil_RegisterAtom(Tcl_Interp *interp, char *handle, ClientData clientData)
{
    Tcl_HashEntry *entryPtr;
    int new;

    if (!handle)
	return -1;

    if (!atomsInited) {
	Tcl_InitHashTable(&atoms, TCL_STRING_KEYS);
	atomsInited = 1;
    }
    if (!(entryPtr = Tcl_CreateHashEntry(&atoms, handle, &new)))
	return -1;

    Tcl_SetHashValue(entryPtr, clientData);

    return new;
}

/*
 *--------------------------------------------------------------
 *
 * Tutil_UnegisterAtom --
 *
 *	Unassociate an atom.
 *
 * Results:
 *
 * Side effects:
 *	An atom is unregistered in a hash table;
 *
 *--------------------------------------------------------------
 */

 /* ARGSUSED */ 
void
Tutil_UnregisterAtom(Tcl_Interp *interp, char *handle)
{
    Tcl_HashEntry *entryPtr;

    if (!handle)
	return;

    if (!atomsInited) {
	Tcl_InitHashTable(&atoms, TCL_STRING_KEYS);
	atomsInited = 1;

	return;
    }
    if (!(entryPtr = Tcl_FindHashEntry(&atoms, handle)))
	return;

    Tcl_DeleteHashEntry(entryPtr);
}

/* ============================================================= */
/* ============================================================= */
/* ============================================================= */


/*
 *--------------------------------------------------------------
 *
 * Tcm_FindAtom --
 *
 *	Backwards compatability.
 *
 *--------------------------------------------------------------
 */

ClientData
Tcm_FindAtom(Tcl_Interp *interp, char *handle)
                       		/* Current interpreter. */
                 
{
  return (Tutil_FindAtom(interp, handle));
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_RegisterAtom --
 *
 *	Backwards compatability.
 *
 *--------------------------------------------------------------
 */

int
Tcm_RegisterAtom(char *handle, ClientData clientData)
{
  return (Tutil_RegisterAtom(interpFirst, handle, clientData));
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_UnegisterAtom --
 *
 *	Backwards compatability.
 *
 *--------------------------------------------------------------
 */

void
Tcm_UnregisterAtom(char *handle)
{
  Tutil_UnregisterAtom(interpFirst, handle);
}


/*
 *--------------------------------------------------------------
 *
 * Tutil_RegisterCommand --
 *
 *      Register a C funtion pointer and associated ClientData
 *      under a name (usally of the form <mod>.<cmd>
 *
 * Results:
 *     A standard Tcl result.
 *
 * Side effects:
 *     None.
 *
 * Parameters:
 *     interp           (In) Tcl interpreter for reporting errors.
 *     name             (In) Name to register under.
 *     funcPtr          (In) Pointer to C function.
 *     fundCD           (In) ClientData for C function.
 *
 *--------------------------------------------------------------
 */
int
Tutil_RegisterCommand(Tcl_Interp *interp,
		      char *name,
		      ClientData funcPtr,
		      ClientData funcCD)
{
    FunctionRegisterEntry *new_entry;

    new_entry = (FunctionRegisterEntry *) ckalloc(sizeof(FunctionRegisterEntry));
    
    new_entry->funcPtr = funcPtr;
    new_entry->funcCD = funcCD;

    Tutil_RegisterAtom(interp, name, (ClientData) new_entry);
   
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tutil_UnregisterCommand --
 *
 *      Unregisters a C funtion pointer and associated ClientData
 *      under a name (usally of the form <mod>.<cmd>
 *
 * Results:
 *     A standard Tcl result.
 *
 * Side effects:
 *     None.
 *
 * Parameters:
 *     interp           (In) Tcl interpreter for reporting errors.
 *     name             (In) Name to unregister.
 *
 *--------------------------------------------------------------
 */
int
Tutil_UnregisterCommand(Tcl_Interp *interp,
		      char *name)
{
    FunctionRegisterEntry *entry;

    entry = (FunctionRegisterEntry *) Tutil_FindAtom(interp, name);
    if (entry != NULL) {
	Tutil_UnregisterAtom(interp, name);
	ckfree((char *) entry);
	return TCL_OK;
    } else {
	return TCL_ERROR;
    }
}


/*
 *--------------------------------------------------------------
 *
 * Tutil_GetRegisteredCommand --
 *
 *	Swizzle a command string of the form "@<mod>.<cmd> into a
 *	C function and associated ClientData
 *
 * Results:
 *	A standard Tcl result
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	interp		(In) Tcl interpreter for reporting errors
 *	cmdStr		(In) Command to swizzle
 *	funcPtr		(Out) Pointer to swizzled function
 *	cdPtr		(Out) Arbitrary pointer for function
 *
 *--------------------------------------------------------------
 */
int
Tutil_GetRegisteredCommand(Tcl_Interp *interp, char *cmdStr,
			ClientData *cdPtr, ClientData *funcPtr)
{
    FunctionRegisterEntry *entry;

    /*
     * Lookup C function for method of object
     */
    if ((entry = (FunctionRegisterEntry *) Tutil_FindAtom(interp, (cmdStr+1))) == NULL) {
	Tcl_AppendResult (interp,
			"Couldn't find C command for \"", cmdStr, "\"", NULL);
	return TCL_ERROR;
    }
    *cdPtr = entry->funcCD;
    *funcPtr = entry->funcPtr;

    return TCL_OK;
}

