/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * tcmSparcAuCF.c --
 *
 *  This module provides for the reading of audio "clipfiles".
 *  
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include <string.h>
#include "cmAuUtil.h"
#include "cmAtom.h"
#include "cmAuFile.h"
#include "cmCookies.h"
#include "cmChunk.h"
#include "cmBufferlist.h"
#include "cmLTS.h"

#define ONE_BILLION    (1000000000.0)
#define DEF_CYCLE_TIME (1.0)
#define DEF_SEND_AHEAD (0.2)

/*
 * One of the following structures is created for each segment in the
 * clip file that is mapped onto the logical time system
 */
typedef struct Segment {
    double clipStart;		/* Start time of segment (file units) */
    double clipEnd;		/* End time of segment (file units) */
    double logicalStart;	/* Start time of segment (LTS units) */
    double logicalEnd;		/* End time of segment (LTS units) */
    struct Segment *next;	/* Link to previous segment */
    struct Segment *prev;	/* Link to next segment */
    int dataStart;		/* Start offset of data */
    int segId;			/* Handle for segment */
    double scale;		/* Ratio of LTS speed to natural speed */
    int reverse;		/* +1 = forward, -1 = reverse. */
    int numFrames;		/* Number of frames in the segment */
    int *FOT;			/* Frame Offset Table */
    double *FTT;		/* Frame Time Table  */
} Segment;

/*
 * The following structure is used when frames are actually read from
 * disk.  An array of the structures is created, one for each frame in
 * this round, and the values are filled in.
 */
typedef struct {
    Segment *segPtr;		/* Pointer to segment with data */
    double sTime;		/* Logical start time for this piece */
    double eTime;		/* Logical end time for this piece */
    Buffer *bufPtr;		/* Pointer to buffer with data */
    int offset;			/* Offset within file */
    int size;			/* Size (in bytes) of frame */
    int index;			/* Frame index */
    int sendFlag;
} FramePick;

/*
 * On eof the following structures is created for each clipfile object.
 */
typedef struct SparcAudioClipFile {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */

    double sendAhead;		/* Delay for sending ahead data */
    double cycleTime;		/* Cycle Period */

    ClientData lts;		/* LTS for this clipfile */
    char *ltsName;		/* Its tcl name */
    int ltsTriggerNum;		/* LTS trigger for changed lts. */

    InFunc outCCmd;		/* Command called to pass along data */
    ClientData outCD;		/* Object data is sent to */
    char *outCmd;		/* Command used for sending out data */

    Segment *segments;		/* List of clipfile segments */
    Segment *currentSeg;	/* current segment */
    int nextSegId;		/* Next segment ID */
    int segmentListChangeFlag;  /* Indicates seg. list has changed and overlaps
				   should be checked on the next ready. */

    tChunk cfState;		/* pointer to chunk_file state */
    SparcAuHeader *hdr;		/* Audio Header from clipfile */
    MediaIndepHeader *mediaHdr;	/* Media Independent Header */
    int ready;			/* Ready flag */
    char *filename;		/* Full pathname for clipfile */
    int bandwidth;
    int sampleSize;		/* Size of each audio sample, in bytes */
    double clipLength;		/* Duration of clipfile, in seconds  */
    int dataStart;		/* Offset to start of data. */
    int numFrames;		/* Number of frames in clipfile */

    int atToken;		/* Token for at event. */
    double atWhen;		/* When that event was scheduled */
    
    int fd;
    RawSparcAuHeader *rawHdr;
    int alignSize;		/* Alignment of data units.  1 for byte
				 * aligned, 2 for 16 bit aligned, etc. */
    int dataRate;
    int raw;
} SparcAudioClipFile;

/*
 * Globals used by this object module.
 */
static int nextSparcAuCFNum = 0;/* Variable for next unique object name. */

/*
 * Forward declarations
 */
static int SparcAuCFObjCmd	(ClientData clientData, Tcl_Interp *interp,
				 int argc, char **argv);
static int SparcAuAddSeg	(SparcAudioClipFile *objPtr,
				 double clipStart, double clipEnd,
				 double logStart, double logEnd,
				 int logStartFlag, int logEndFlag);
static int SparcAuDelSeg	(SparcAudioClipFile *objPtr, int segID);
static int SparcAuClearSegs	(SparcAudioClipFile *objPtr);
static int SparcAuSegInfo	(SparcAudioClipFile *objPtr, int segID);
static int SparcAuInfo		(SparcAudioClipFile *objPtr);
static int SparcAuReady		(SparcAudioClipFile *objPtr);
static int SparcAuUnready	(SparcAudioClipFile *objPtr);
static int SparcAuChangeOutCmd	(SparcAudioClipFile *objPtr, char *newOut);
static int SparcAuOpenClipFile	(SparcAudioClipFile *objPtr, char *filename);
static void SparcAuLTSChanged	(ClientData clientdata);
static void SparcAuSendData	(ClientData clientData);

static int SparcAuRawAddSeg	(SparcAudioClipFile *objPtr,
				 double clipStart, double clipEnd,
				 double logStart, double logEnd,
				 int logStartFlag, int logEndFlag);
static int SparcAuOpenRawFile	(SparcAudioClipFile *objPtr, char *filename);
static int ReadAudioHeader	(Tcl_Interp *interp, int fd,
				 RawSparcAuHeader **hdrPtr);
static int SparcAuSegmentList   (SparcAudioClipFile *objPtr);
static int SparcAuConfigureSegment (SparcAudioClipFile *objPtr, int segmentId,
				    double clipStart, double clipEnd, 
				    double logStart, double logEnd, 
				    int logStartFlag, int logEndFlag);


#ifndef _WINDOWS
#    define max(a,b)      ((a)>(b)?(a):(b))
#    define min(a,b)      ((a)<(b)?(a):(b))
#endif
#define rnd(a)        (floor(a + .5))

/************************************************************
 *
 * CreateNewSparcAuCF
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/
#if 0
int AudioLogFD;
#endif /* debug */
SparcAudioClipFile *
CreateNewSparcAuCF(Tcl_Interp * interp, char *name)
{
    SparcAudioClipFile *new;

    new = (SparcAudioClipFile *) ckalloc(sizeof(SparcAudioClipFile));

    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);
    new->filename = NULL;
    new->bandwidth = 0.0;
    new->lts = NULL;
    new->ltsName = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->hdr = NULL;
    new->mediaHdr = NULL;
    new->sendAhead = DEF_SEND_AHEAD;
    new->outCmd = NULL;
    new->segments = NULL;
    new->currentSeg = NULL;
    new->cfState = NULL;
    new->sampleSize = 1;
    new->cycleTime = DEF_CYCLE_TIME;
    new->nextSegId = 0;
    new->segmentListChangeFlag = 0;
    new->ready = 0;
    new->clipLength = 0;
    new->numFrames = 0;
    new->atToken = -1;
    new->atWhen = HUGE_VAL;
    new->ltsTriggerNum = -1;
    new->dataStart = 0;

    new->fd = -1;
    new->rawHdr = NULL;
    new->dataRate = 0;
    new->alignSize = 0;
    new->raw = 0;
#if 0
    AudioLogFD = creat("./audioLog",0666);
#endif /* debug */
    return new;
}

/************************************************************
 *
 * DestroySparcAudioClipFile
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/
int
DestroySparcAudioClipFile(SparcAudioClipFile *objPtr)
{
    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->atToken != -1) {
	Tcm_DeleteAtHandler(objPtr->atToken);
    }
    objPtr->atToken = -1;
    objPtr->atWhen = HUGE_VAL;

    if (objPtr->name != NULL) {
	ckfree((char *) objPtr->name);
	objPtr->name = NULL;
    }

    if (objPtr->filename != NULL) {
	ckfree((char *) objPtr->filename);
	objPtr->filename = NULL;
    }

    if (objPtr->cfState != NULL) {
	Chunk_CloseFile(objPtr->cfState);
	objPtr->cfState = NULL;
    }

    if (objPtr->mediaHdr != NULL) {
	ckfree((char *) objPtr->mediaHdr);
	objPtr->mediaHdr = NULL;
    }

    if (objPtr->hdr != NULL) {
	ckfree((char *) objPtr->hdr);
	objPtr->hdr = NULL;
    }

    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
	objPtr->outCmd = NULL;
    }

    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    if (objPtr->ltsName != NULL) {
	ckfree((char *) objPtr->ltsName);
	objPtr->ltsName = NULL;
    }
    
    /*
     * Free all segments.
     */
    SparcAuClearSegs(objPtr);
#if 0
    close(AudioLogFD);
#endif /* debug */

    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureSparcAudioClipFile
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

int
ConfigureSparcAudioClipFile(SparcAudioClipFile *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];

    interp = objPtr->interp;

    /*
     * Case 1: "<object> configure"
     */
    if (argc == 0) {
	if (objPtr->filename != NULL) {
	    Tcl_AppendResult(interp, "{-filename {} {", objPtr->filename,
			     "}}", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-filename {} {}}", NULL);
	}

	Tcl_PrintDouble(interp, objPtr->bandwidth, str);
	Tcl_AppendResult(interp, "{-bandwidth {0.0} ", str, "}", NULL);

	if (objPtr->lts != NULL) {
	    Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName, "}}", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {} {}}", NULL);
	}

	Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	Tcl_AppendResult(interp, "{-sendAhead {0.2} ", str, "}", NULL);

	if (objPtr->outCmd != NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}}", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {}}", NULL);
	}

	Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	Tcl_AppendResult(interp, "{-cycleTime {1.0} ", str, "}", NULL);
	return TCL_OK;
    }
    /*
     * Case 2: "<object> configure -<slot>"
     */
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 'f') && (strncmp("-filename", argv[0], len) == 0)) {
	    if (objPtr->filename != NULL) {
		Tcl_AppendResult(interp, "{-filename {} {", objPtr->filename,
				 "}}", NULL);
	    } else {
		Tcl_AppendResult(interp, "{-filename {} {}}", NULL);
	    }
	    return TCL_OK;

	} else if ((c == 'b') && (strncmp("-bandwidth", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->bandwidth, str);
	    Tcl_AppendResult(interp, "-bandwidth {0.0} ", str, NULL);
	    return TCL_OK;

	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->lts != NULL) {
		Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName, "}}",
				 NULL);
	    } else {
		Tcl_AppendResult(interp, "{-lts {} {}}", NULL);
	    }
	    return TCL_OK;

	} else if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	    Tcl_AppendResult(interp, "-sendAhead {0.2} ", str, NULL);
	    return TCL_OK;

	} else if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd != NULL) {
		Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
				 "}}", NULL);
	    } else {
		Tcl_AppendResult(interp, "{-outCmd {} {}}", NULL);
	    }
	    return TCL_OK;

	} else if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
	    char str[TCL_DOUBLE_SPACE];

	    Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	    Tcl_AppendResult(interp, "-cycleTime {1.0} ", str, NULL);
	    return TCL_OK;
	}
	Tcl_AppendResult(interp, "bad configuration option: ", argv[0], NULL);
	return TCL_ERROR;
    }
    /*
     * Case 3: "<object> configure -<slot>"
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;

	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	/* ------------------------ FILENAME ------------------------ */
	if ((c == 'f') && (strncmp("-filename", argv[cntr], len) == 0)) {
	  cntr++;
	  if (cntr == argc) {
	  usageError:
	    Tcl_AppendResult(interp,
			     "Must specify value with configuration option ",
			     argv[cntr - 1], NULL);
	    return TCL_ERROR;
	  }
	  if (objPtr->filename != NULL) {
	    Tcl_AppendResult(interp, "Cannot change clipfile name.\n",
			     "Create a new object and/or delete this one.",
			     NULL);
	    return TCL_ERROR;
	  }
	  objPtr->raw = 0;
	  if (SparcAuOpenClipFile(objPtr, argv[cntr]) != TCL_OK) {
	    objPtr->raw = 1;
	    if (SparcAuOpenRawFile(objPtr, argv[cntr]) != TCL_OK) {
	      if (objPtr->filename != NULL) {
		ckfree((char *) objPtr->filename);
		objPtr->filename = NULL;
	      }
	      return TCL_ERROR;
	    }
	  } 
	      
	  /* ------------------------ BANDWIDTH ------------------------ */
	} else if ((c == 'b') &&
		   (strncmp("-bandwidth", argv[cntr], len) == 0)) {
	  double newBandwidth;

	  cntr++;
	  if (cntr == argc) {
	    goto usageError;
	  }
	  if (Tcl_GetDouble(interp, argv[cntr], &newBandwidth) != TCL_OK) {
	  badValueError:
	    Tcl_AppendResult(interp, "bad value \"", argv[cntr],
			     "\" given for configuration option ",
			     argv[cntr - 1], NULL);
	    return TCL_ERROR;
	  }
	  objPtr->bandwidth = newBandwidth;

	  /* ------------------------ LTS ------------------------ */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	  ClientData newLts;

	  cntr++;
	  if (cntr == argc) {
	    goto usageError;
	  }
	  if ((newLts = Tcm_FindAtom(interp, argv[cntr])) == NULL) {
	    goto badValueError;
	  }

	  /*
	   * Release our lock and trigger on the old lts.
	   */
	  if (objPtr->lts != NULL) {
	    if (objPtr->ltsTriggerNum != -1) {
	      LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	      objPtr->ltsTriggerNum = -1;
	    }
	    Tk_Release(objPtr->lts);
	    objPtr->lts = NULL;
	  }
	  if (objPtr->ltsName != NULL) {
	    ckfree((char *) objPtr->ltsName);
	    objPtr->ltsName = NULL;
	  }

	  /*
	   * Set up new lts
	   */
	  objPtr->ltsName = ckalloc(strlen(argv[cntr]) + 1);
	  strcpy(objPtr->ltsName, argv[cntr]);
	  objPtr->lts = newLts;
	  Tk_Preserve (objPtr->lts);
	  LTS_SetTrigger(objPtr->lts,
			 LTS_SPEED_FLAG | LTS_OFFSET_FLAG | LTS_VALUE_FLAG,
			 SparcAuLTSChanged, (ClientData) objPtr);

	  /* ------------------------ SEND AHEAD ------------------------ */
	} else if ((c == 's') &&
		   (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	  double newSendAhead;

	  cntr++;
	  if (cntr == argc) {
	    goto usageError;
	  }
	  if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
	    goto badValueError;
	  }
	  objPtr->sendAhead = newSendAhead;

	  /* ------------------------ OUT CMD ------------------------ */
	} else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	  cntr++;
	  if (cntr == argc) {
	    goto usageError;
	  }
	  if (objPtr->outCmd != NULL) {
	    ckfree((char *) objPtr->outCmd);
	    objPtr->outCmd = NULL;
	  }
	  if (SparcAuChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
	    return TCL_ERROR;
	  }
	  /* ------------------------ CYCLE TIME ------------------------ */
	} else if ((c == 'c') &&
		   (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	  double newCycleTime;

	  cntr++;
	  if (cntr == argc) {
	    goto usageError;
	  }
	  if (Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) {
	    goto badValueError;
	  }
	  objPtr->cycleTime = newCycleTime;
	}
	  cntr++;
	}
    return SparcAuUnready(objPtr);
}

/************************************************************
 *
 * SparcAuCFCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the sparcAudioClipFile command.
 *     A new object of type SparcAudioClipFile is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/
int
SparcAuCFCreateCmd (ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
    char newName[256];
    SparcAudioClipFile *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", NULL);
	return TCL_ERROR;
    }
    nextSparcAuCFNum++;
    sprintf(newName, "%sauFile%d", argv[1], nextSparcAuCFNum);

    new = CreateNewSparcAuCF(interp, newName);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, newName, SparcAuCFObjCmd,
		      (ClientData) new, (void (*)())NULL);

    if (argc > 2) {
	ConfigureSparcAudioClipFile(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, newName, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * SparcAuCFObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     SparcAudioClipFile object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/
static int
SparcAuCFObjCmd (ClientData clientData, Tcl_Interp *interp,
		 int argc, char **argv)
{
    char c;
    int len, i;
    SparcAudioClipFile *objPtr = (SparcAudioClipFile *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " <command> ?arguments?", NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    /*---------------------------- CONFIGURE ----------------------------*/
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureSparcAudioClipFile(objPtr, argc - 2, argv + 2);

    /*---------------------------- DESTROY ------------------------------*/
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroySparcAudioClipFile(objPtr);

    /*---------------------------- ADD_SEGMENT -----------------------*/
    } else if ((c == 'a') && (strncmp("addSegment", argv[1], len) == 0)) {
	double clipStart;
	double clipEnd;
	double logStart = 0.0;
	double logEnd = 0.0;
	int logEndFlag = 0;
	int logStartFlag = 0;

	if (argc < 4) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			     " addSegment <timeValue> <timeValue> ",
			     "?-start <timeValue>? ?-end <timeValue>? ",
			     NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetDouble(interp, argv[2], &clipStart) != TCL_OK) {
	    Tcl_AppendResult (interp, "Bad clip start time \"",
			      argv[2], "\" given to: ", argv[0], " ",
			      argv[1], NULL);
	    return TCL_ERROR;
	}
	if (strcmp(argv[3], "end") == 0) {
	    clipEnd = HUGE_VAL;
	} else if (Tcl_GetDouble(interp, argv[3], &clipEnd) != TCL_OK) {
	    Tcl_AppendResult (interp, "Bad clip end time \"",
			      argv[3], "\" given to: ", argv[0], " ",
			      argv[1], NULL);
	    return TCL_ERROR;
	}

	/*
	 * Parse off the start and end time...
	 */
	for (i = 4; i < argc; i++) {
	    char ac;
	    int alen;

	    ac = argv[i][1];
	    alen = strlen(argv[i]);

	    if ((ac == 's') && (strncmp("-start", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Start value missing", NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logStart) != TCL_OK) {
		    Tcl_AppendResult (interp, "Bad start value \"", argv[i],
				      "\" given to: ", argv[0], " ", argv[1],
				      NULL);
		    return TCL_ERROR;
		}
		logStartFlag = 1;
	    } else if ((ac == 'e') && (strncmp("-end", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "End value missing", NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logEnd) != TCL_OK) {
		    Tcl_AppendResult (interp, "Bad end value \"", argv[i],
				      "\" given to: ", argv[0], " ", argv[1],
				      NULL);
		    return TCL_ERROR;
		}
		logEndFlag = 1;
	    }
	}
	if (objPtr->raw == 0) {
	  return SparcAuAddSeg(objPtr, clipStart, clipEnd, logStart, logEnd,
			       logStartFlag, logEndFlag);
	} else {
	  return SparcAuRawAddSeg(objPtr, clipStart, clipEnd, logStart, logEnd,
				  logStartFlag, logEndFlag);
	}

    /*---------------------------- CONFIGURE_SEGMENT -----------------------*/
    } else if ((c == 'c') && (strncmp("configureSegment", argv[1], len) == 0)) {
	double clipStart;
	double clipEnd;
	double logStart = 0.0;
	double logEnd = 0.0;
	int logEndFlag = 0;
	int logStartFlag = 0;
	int segmentId;

	if (argc < 5) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			     " configureSegment <segmentId> <timeValue> <timeValue> ",
			     "?-start <timeValue>? ?-end <timeValue>? ",
			     NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[2], &segmentId) != TCL_OK) {
	    Tcl_AppendResult (interp, "Bad segment id \"",
			      argv[2], "\" given to: ", argv[0], " ",
			      argv[1], (char *) NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetDouble(interp, argv[3], &clipStart) != TCL_OK) {
	    Tcl_AppendResult (interp, "Bad clip start time \"",
			      argv[3], "\" given to: ", argv[0], " ",
			      argv[1], NULL);
	    return TCL_ERROR;
	}
	if (strcmp(argv[4], "end") == 0) {
	    clipEnd = HUGE_VAL;
	} else if (Tcl_GetDouble(interp, argv[4], &clipEnd) != TCL_OK) {
	    Tcl_AppendResult (interp, "Bad clip end time \"",
			      argv[4], "\" given to: ", argv[0], " ",
			      argv[1], NULL);
	    return TCL_ERROR;
	}

	/*
	 * Parse off the start and end time...
	 */
	for (i = 5; i < argc; i++) {
	    char ac;
	    int alen;

	    ac = argv[i][1];
	    alen = strlen(argv[i]);

	    if ((ac == 's') && (strncmp("-start", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Start value missing", NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logStart) != TCL_OK) {
		    Tcl_AppendResult (interp, "Bad start value \"", argv[i],
				      "\" given to: ", argv[0], " ", argv[1],
				      NULL);
		    return TCL_ERROR;
		}
		logStartFlag = 1;
	    } else if ((ac == 'e') && (strncmp("-end", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "End value missing", NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logEnd) != TCL_OK) {
		    Tcl_AppendResult (interp, "Bad end value \"", argv[i],
				      "\" given to: ", argv[0], " ", argv[1],
				      NULL);
		    return TCL_ERROR;
		}
		logEndFlag = 1;
	    }
	}

	return SparcAuConfigureSegment(objPtr, segmentId, clipStart,
				       clipEnd, logStart, logEnd, 
				       logStartFlag, logEndFlag);

    /*---------------------------- REMOVE_SEGMENT -----------------------*/
    } else if ((c == 'r') && (strncmp("removeSegment", argv[1], len) == 0)) {
	int segID;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			     " removeSegment <segment-Identifier> ",
			     NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &segID) != TCL_OK) {
	    Tcl_AppendResult(interp, "Bad integer parameter value \"",
			     argv[2], "\" given to: ", argv[0], " ",
			     argv[1], NULL);
	    return TCL_ERROR;
	}
	return SparcAuDelSeg(objPtr, segID);

    /*------------------------ SEGMENT LIST ----------------------------*/
    } else if ((c == 's') && (strncmp("segmentList", argv[1], len) == 0)) {
	return SparcAuSegmentList(objPtr);

    /*---------------------------- CLEAR_SEGMENTS -----------------------*/
    } else if ((c == 'c') && (strncmp("clearSegments", argv[1], len) == 0)) {
	return SparcAuClearSegs(objPtr);

    /*---------------------------- SEGMENT_INFO ------------------------*/
    } else if ((c == 's') && (strncmp("segmentInfo", argv[1], len) == 0)) {
	int segID;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			     " segmentInfo <segment_ID> ", NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &segID) != TCL_OK) {
	    Tcl_AppendResult(interp, "Bad segment ID", NULL);
	    return TCL_ERROR;
	}
	return SparcAuSegInfo(objPtr, segID);

    /*---------------------------- INFO --------------------------------*/
    } else if ((c == 'i') && (strncmp("info", argv[1], len) == 0)) {
        return SparcAuInfo(objPtr);

    /*---------------------------- READY -------------------------------*/
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
        return SparcAuReady(objPtr);

    /*---------------------------- UNREADY -----------------------------*/
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
        return SparcAuUnready(objPtr);
    }
    Tcl_AppendResult(interp, "Bad <command> option: ", argv[0], argv[1],
		     NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  addSegment <clipStart> <clipEnd>",
			" ?-start <logicalStart>? ?-end <logicalEnd>?\n",
                "  removeSegment <segmentId>\n",
                "  clearSegments\n",
		"  configureSegment <segmentId> <clipStart> <clipEnd>",
			" ?-start <logicalStart>? ?-end <logicalEnd>?\n",
		"  segmentList\n",
                "  segmentInfo <segmentId>\n",
                "  ready\n",
                "  unready\n",
                "  info\n",
                NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * SparcAudioClipFileModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
SparcAudioClipFileModuleInit(Tcl_Interp *interp)
{
    Tcl_CreateCommand (interp, "auFile", SparcAuCFCreateCmd,
		       NULL, (void (*)()) NULL);
    Tcl_VarEval (interp, "cminfo add object_info auFile",
		 "{sparc audiosparc audio file}", NULL);
    return TCL_OK;
}

/************************************************************
 *
 * GetTimeFromClipFile
 *
 * Parameters:
 *     fd - file to read
 *
 * Return:
 *     time value as a double
 *
 * Action:
 *	Read a time value from a clip file and return it as a double.
 *	The value is stored internally as a 4 byte integer indicating
 *	seconds and a 4 byte integer indicating nanoseconds, in
 *	network byte order.
 *
 ************************************************************/
static double
GetTimeFromClipFile(int fd)
{
    int sec, nsec;
    double res;

    read (fd, &sec, 4);
    read (fd, &nsec, 4);
    
    sec = ntohl(sec);
    nsec = ntohl(nsec);

    res = sec + ((double)nsec/ONE_BILLION);
    return res;
}


/*
 *--------------------------------------------------------------
 *
 * GetOffsetFromClipFile --
 *
 *	Read a and return start and length pair from a chunk file.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	File is read
 *
 * Parameters:
 *	int fd		(in)	File to read
 *	int *start	(out)	start value (read from file)
 *	int *length	(out)	length value (read from file)
 *
 *--------------------------------------------------------------
 */
static void
GetOffsetFromClipFile (int fd, int *start, int *length)
{
    int tstart, tlength;

    read(fd, &tstart, 4);
    read(fd, &tlength, 4);

    (*start) = ntohl(tstart);
    (*length) = ntohl(tlength);

    return;
}

/************************************************************
 *
 * TimeToOffset
 *
 * This function returns the data offset and size of a 
 * "subframe" which exists inside an audio frame.
 *
 * Returns:  the offset for a frame
 *
 * Parameters:
 *	double time	(in)	Logical time we are interested in
 *	double ta	(in)	Beginning time of frame
 *	double tb	(in)	Beginning time of frame
 *	int da 		(in)	Frame offset
 *	int delta	(in)	Size of frame
 *	int size	(out)	Size of new frame
 *	int start	(in)	Are we finding the start of a subfraem
 *                              or the end
 *      int alignSize   (in)    Size of each sample (in bytes)
 *
 *************************************************************/
static int
TimeToOffset (double time, double ta, double tb, int da, int delta,
	      int *size, int start, int alignSize )
{
    int offset;

    offset = da + rnd(((delta) / (tb - ta)) * (time - ta));
    offset = alignSize * (offset/alignSize);

    if (start) {
	*size = da + delta - offset;
	return offset;
    } else {
	*size = offset - da;
	return da;
    }
}

/*
 *--------------------------------------------------------------
 *
 * SparcAuAddSeg --
 *
 *	Map a new segment of the clipfile into the logical time system
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	From now on, whenever the logical time system is inside the
 *	time value passed in, the appropriate segment of the clipfile
 *	will be played.
 *
 * Parameters:
 *     SparcAudioClipFile *objPtr	pointer to object
 *     double clipStart, clipEnd	Start and end of segment in file
 *     double logStart, logEnd		Destination section in LTS
 *     int logStartFlag			Has logStart been set?
 *     int logEndFlag			Has logEnd been set?
 *
 *--------------------------------------------------------------
 */
static int
SparcAuAddSeg(SparcAudioClipFile *objPtr, double clipStart, double clipEnd, 
	      double logStart, double logEnd, int logStartFlag,
	      int logEndFlag)
     
{
    Segment *newSeg, *nextSeg, *prevSeg;
    char retStr[32];
    double clipDur;
    tChunkHeader chunkHdr;
    int begFrame, endFrame;
    double thisTime, nextTime, firstStart, firstEnd, lastStart, lastEnd;
    int count, i, idx, lastOffset, lastSize, size;
    int *fot, n, t;
    double *ftt;

    SparcAuUnready(objPtr);

    if (clipStart > clipEnd) {
	Tcl_AppendResult(objPtr->interp,
			 "Clip start can not be after clip end.", NULL);
	return TCL_ERROR;
    }
    if (objPtr->filename == NULL) {
	Tcl_AppendResult(objPtr->interp, "Must specify clipfile first.",
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Check boundary conditions. 
     */
    if (clipStart < 0.0)
	clipStart = 0.0;
    if (clipStart > objPtr->clipLength) {
	Tcl_AppendResult(objPtr->interp, "Clip start too large.", NULL);
	return TCL_ERROR;
    }
    if (clipEnd > objPtr->clipLength)
	clipEnd = objPtr->clipLength;
    if (clipEnd < 0.0) {
	Tcl_AppendResult(objPtr->interp, "Clip end too small.", NULL);
	return TCL_ERROR;
    }
    if (clipStart >= clipEnd) {
	Tcl_AppendResult(objPtr->interp,
			 "ClipEnd must be greater than ClipStart.", NULL);
	return TCL_ERROR;
    }

    /*
     * Boundary conditions OK, create a new segment structure.
     */
    newSeg = (Segment *) ckalloc(sizeof(Segment));
    newSeg->clipStart = clipStart;
    newSeg->clipEnd = clipEnd;

    clipDur = clipEnd - clipStart;
    if (clipDur <= 0.0) {
	ckfree((char *) newSeg);
	Tcl_AppendResult(objPtr->interp, "Clip segment duration too small.",
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Handle the various cases of start and end in LTS system being
     * specified.  Set logicalStart, logicalEnd, and reverse fields of
     * newSeg
     */
    newSeg->reverse = 1;
    if (logStartFlag && logEndFlag) {
	/*
	 * Set logStart and logEnd. 
	 */
	if (logStart > logEnd) {
	    newSeg->reverse = -1;
	    newSeg->logicalStart = logEnd;
	    newSeg->logicalEnd = logStart;
	} else {
	    newSeg->logicalStart = logStart;
	    newSeg->logicalEnd = logEnd;
	}
    } else if (logStartFlag) {
	newSeg->logicalStart = logStart;
	newSeg->logicalEnd = logStart + clipDur;
    } else if (logEndFlag) {
	newSeg->logicalEnd = logEnd;
	newSeg->logicalStart = logEnd - clipDur;
    } else {
	newSeg->logicalStart = 0.0;
	newSeg->logicalEnd = clipDur;
    }

    /* 
     * Find insertion point in segment list. Sort by start times.
     */

    nextSeg = objPtr->segments;
    prevSeg = NULL;
    while (nextSeg != NULL) {
	if (nextSeg->logicalStart < newSeg->logicalStart) {
	    prevSeg = nextSeg;
	    nextSeg = nextSeg->next;
	} else {
	    break;
	}
    }

    /*
     * Set scale for non-uniform playback speed.
     */
    newSeg->scale = (newSeg->logicalEnd - newSeg->logicalStart) / clipDur;

    /*
     * Create Frame offset table for this segment, requires reading 
     * chunk file. 
     */
    if (objPtr->cfState != NULL) {
	Chunk_ReadRewind(objPtr->cfState);
    } else {
	objPtr->cfState = Chunk_OpenFile(objPtr->filename, "r");
    }
    if (Chunk_FindType(objPtr->cfState, CM_FTT, &chunkHdr)) {
	ckfree((char *) newSeg);
	Tcl_AppendResult(objPtr->interp,
			 "Problem reading clipfile frame time table.",
			 NULL);
	return TCL_ERROR;
    }
    thisTime = GetTimeFromClipFile(objPtr->cfState->fd);
    nextTime = GetTimeFromClipFile(objPtr->cfState->fd);

    /*
     * Search for first frame in segment (begFrame)
     */
    count = 0;
    while ((thisTime < clipStart) && (nextTime <= clipStart)) {
	thisTime = nextTime;
	nextTime = GetTimeFromClipFile(objPtr->cfState->fd);
	count++;
	if (count == objPtr->numFrames) {
	    Tcl_AppendResult(objPtr->interp,
			     "Error constructing segment frame time table.",
			     NULL);
	    return TCL_ERROR;
	}
    }
    begFrame = count;


    /*
     * Search for last frame in segment (endFrame)
     */
    while ((thisTime < clipEnd) && (nextTime < clipEnd)) {
	thisTime = nextTime;
	nextTime = GetTimeFromClipFile(objPtr->cfState->fd);
	count++;
	if (count == objPtr->numFrames) {
	    Tcl_AppendResult(objPtr->interp,
			     "Error constructing segment frame time table.",
			     NULL);
	    return TCL_ERROR;
	}
    }
    endFrame = count;


    /*
     * Construct Frame Time Table (FTT) 
     */
    newSeg->numFrames = endFrame - begFrame + 1;
    newSeg->FTT = (double *) ckalloc((newSeg->numFrames + 1) * sizeof(double));
    Chunk_ReadRewind(objPtr->cfState);
    Chunk_FindType(objPtr->cfState, CM_FTT, &chunkHdr);

    if (begFrame != 0) {
	lseek(objPtr->cfState->fd, begFrame * 2 * sizeof(int), SEEK_CUR);
    }
    for (i = 0; i <= newSeg->numFrames; i++) {
	newSeg->FTT[i] = GetTimeFromClipFile(objPtr->cfState->fd);
    }
    lastStart = newSeg->FTT[i - 2];
    lastEnd = newSeg->FTT[i - 1];
    firstStart = newSeg->FTT[0];
    firstEnd = newSeg->FTT[1];

    for (i = 1; i < newSeg->numFrames; i++) {
	newSeg->FTT[i] = ((newSeg->FTT[i] - newSeg->clipStart) * newSeg->scale)
	    + newSeg->logicalStart;
    }
    newSeg->FTT[0] = newSeg->logicalStart;
    newSeg->FTT[i] = newSeg->logicalEnd;

    /*
     * Construct Frame Offset Table (FOT) 
     */
    Chunk_ReadRewind(objPtr->cfState);
    if (Chunk_FindType(objPtr->cfState, CM_FOT, &chunkHdr)) {
	if (newSeg->FTT != NULL) {
	    ckfree((char *) newSeg->FTT);
	}
	ckfree((char *) newSeg);
	Tcl_AppendResult(objPtr->interp,
			 "Error constructing segment frame offset table.",
			 NULL);
	return TCL_ERROR;
    }
    newSeg->FOT = (int *) ckalloc(sizeof(int) * 2 * (newSeg->numFrames));

    if (begFrame != 0) {
	lseek(objPtr->cfState->fd, begFrame * 8, SEEK_CUR);
    }
    for (i = 0; i < newSeg->numFrames; i++) {
	GetOffsetFromClipFile(objPtr->cfState->fd, &(newSeg->FOT[i * 2]),
			  &(newSeg->FOT[(i * 2) + 1]));
    }

    /*
     * Adjust beginning and ending offsets since we can take
     * data from the middle of a frame. 
     */
    idx = (i - 1) * 2;

    /*
     * Get offsets from last frame just in case last frame = first frame
     */
    lastOffset = newSeg->FOT[idx];
    lastSize = newSeg->FOT[idx + 1];
    newSeg->FOT[0] = TimeToOffset(newSeg->clipStart, firstStart, firstEnd,
				  newSeg->FOT[0], newSeg->FOT[1],
				  &(newSeg->FOT[1]), 1, objPtr->alignSize);
    TimeToOffset( clipEnd, lastStart, lastEnd, lastOffset, 
		  lastSize, &size, 0, objPtr->alignSize );
    newSeg->FOT[idx + 1] = newSeg->FOT[idx + 1] - (lastSize - size);

    /*
     * If reverse is negative, we need to reverse the FOT and FTT 
     */
    if (newSeg->reverse < 0) {
        fot = (int *) ckalloc(sizeof(int) * 2 * (newSeg->numFrames));
        ftt = (double *) ckalloc((newSeg->numFrames + 1) * sizeof(double));

        n = newSeg->numFrames - 1;
        t = newSeg->FTT[0] + newSeg->FTT[newSeg->numFrames];

	fot[0] = newSeg->FOT[n * 2] + newSeg->FOT[n * 2 + 1];
	fot[1] = -newSeg->FOT[n * 2 + 1];
	ftt[0] = t - newSeg->FTT[n + 1];

	for (i = 1; i < newSeg->numFrames; i++) {
	    fot[2 * i] = newSeg->FOT[2 * (n - i + 1)];
	    fot[2 * i + 1] = -newSeg->FOT[2 * (n - i) + 1];
	    ftt[i] = t - newSeg->FTT[n - i + 1];
	}
	ftt[i] = t - newSeg->FTT[0];

	ckfree((char *) newSeg->FTT);
	ckfree((char *) newSeg->FOT);
	newSeg->FTT = ftt;
	newSeg->FOT = fot;
    }

    /*
     * Finish setting the segment (id, next and prev)
     */
    newSeg->segId = objPtr->nextSegId++;
    newSeg->next = nextSeg;
    newSeg->prev = prevSeg;

    if (prevSeg == NULL) {
	objPtr->segments = newSeg;
    } else {
	prevSeg->next = newSeg;
    }

    if (nextSeg != NULL) {
	nextSeg->prev = newSeg;
    }
    sprintf(retStr, "%d", newSeg->segId);

    Tcl_SetResult(objPtr->interp, retStr, TCL_VOLATILE);

    objPtr->segmentListChangeFlag = 1;

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * SparcAuDelSeg --
 *
 *	Delete a segment from those to be read from the clipfile.
 *	This is a C level interface to the removeSegment function.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	The segment must sever be referenced again.
 *
 * Parameters:
 *	objPtr		(in)	Pointer to data structure for the object
 *	int segId	(in)	Id of segment to delete
 *
 *--------------------------------------------------------------
 */
static int
SparcAuDelSeg(SparcAudioClipFile *objPtr, int segId)
{
    Segment *segPtr;

    /*
     * Find the segment
     */
    segPtr = objPtr->segments;
    while (segPtr != NULL) {
	if (segPtr->segId == segId) {
	    break;
	}
	segPtr = segPtr->next;
    }

    if (segPtr == NULL) {
	return TCL_OK;
    }

    /*
     * Remove it from the linked list
     */
    if (segPtr->prev == NULL) {
	objPtr->segments = segPtr->next;
    } else {
	segPtr->prev->next = segPtr->next;
    }
    if (segPtr->next != NULL) {
	segPtr->next->prev = segPtr->prev;
    }

    /*
     * Free up the frame offset table and frame time table.
     */
    if (segPtr->FOT != NULL)
	ckfree((char *) segPtr->FOT);
    if (segPtr->FTT != NULL)
	ckfree((char *) segPtr->FTT);

    ckfree((char *) segPtr);

    return TCL_OK;
}

/************************************************************
 *
 * SparcAuClearSegs
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *	Remove all the segments
 *
 ************************************************************/

static int
SparcAuClearSegs(SparcAudioClipFile *objPtr)
{
    while (objPtr->segments != NULL) {
	SparcAuDelSeg(objPtr, objPtr->segments->segId);
    }
    return TCL_OK;
}

/************************************************************
 *
 * SparcAuSegInfo
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segID - id of segment to query
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sets the Tcl result string to
 *		{clipStart clipEnd logStart logEnd numFrames} 
 *
 ************************************************************/

static int
SparcAuSegInfo(SparcAudioClipFile *objPtr, int segId)
{
    Segment *segPtr;
    char retStr[256];

    segPtr = objPtr->segments;

    while (segPtr != NULL) {
	if ((segPtr->segId == segId))
	    break;
	segPtr = segPtr->next;
    }

    if (segPtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "Segment doesn't exist.", NULL);
	return TCL_ERROR;
    }
    sprintf(retStr, "%g %g %g %g %d", segPtr->clipStart, segPtr->clipEnd,
	    segPtr->logicalStart, segPtr->logicalEnd, segPtr->numFrames);

    Tcl_AppendResult(objPtr->interp, retStr, NULL);
    return TCL_OK;
}

/************************************************************
 *
 * SparcAuInfo
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Return Tcl String "{numFrames timeLength dataFormat
 *                         samplingRate numChannels}"
 *
 ************************************************************/

static int
SparcAuInfo(SparcAudioClipFile *objPtr)
{
    char retStr[256];
    char format[256];

    if (objPtr->filename == NULL) {
	Tcl_AppendResult(objPtr->interp, "Must specify a clipfile first",
			 NULL);
	return TCL_ERROR;
    }
    switch (objPtr->hdr->dataFormat) {
    case SUN_ULAW:
	strcpy(format, "Sun 8bit Ulaw");
	break;
    case SUN_LIN_8:
	strcpy(format, "Sun 8bit Linear");
	break;
    case SUN_LIN_16:
	strcpy(format, "Sun 16bit Linear");
	break;
    default:
	strcpy(format, "Unkown format");
	break;
    }

    sprintf(retStr, "%d %g {%s} %d %d",
	    objPtr->numFrames, objPtr->clipLength, format,
	    objPtr->hdr->samplingRate, objPtr->hdr->channelCount);

    Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);

    return TCL_OK;
}

/************************************************************
 *
 * SparcAuReady
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
SparcAuReady(SparcAudioClipFile *objPtr)
{
    if (objPtr->atToken != -1) {
	Tcm_DeleteAtHandler(objPtr->atToken);
    }
    objPtr->atToken = -1;
    objPtr->atWhen = HUGE_VAL;

    if (objPtr->segmentListChangeFlag) {
      Segment *seg_ptr;

      seg_ptr = objPtr->segments;

      while (seg_ptr != NULL) {
	if (seg_ptr->next != NULL) {
	  if (seg_ptr->next->logicalStart < seg_ptr->logicalEnd) {
	    
	    /* Overlap detected. Definitely not ready. Return error. */

	    objPtr->ready = 0;
	    Tcl_AppendResult(objPtr->interp, "Segment overlap detected.", (char *) NULL);
	    return TCL_ERROR;
	  }
	}
	seg_ptr = seg_ptr->next;
      }

      objPtr->segmentListChangeFlag = 0;
    }

    objPtr->ready = 1;

    if (objPtr->lts != NULL) {
	SparcAuSendData((ClientData) objPtr);
    }
    return TCL_OK;
}

/************************************************************
 *
 * SparcAuUnready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
SparcAuUnready(SparcAudioClipFile *objPtr)
{
    if (objPtr->atToken != -1) {
	Tcm_DeleteAtHandler(objPtr->atToken);
    }
    objPtr->atToken = -1;
    objPtr->atWhen = HUGE_VAL;

    objPtr->ready = 0;
    return TCL_OK;
}

/************************************************************
 *
 * SparcAuOpenClipfile
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Opens the clipfile, reads basic info data.
 *     Closes file when done.
 *
 ************************************************************/

static int
SparcAuOpenClipFile (SparcAudioClipFile *objPtr, char *filename )
{
    tChunkFileState *cf;
    tChunkHeader ch;
    SparcAuHeader *sparcHdr;
    MediaIndepHeader *mediaHdr;

    if (objPtr->cfState != NULL) {
	Tcl_AppendResult(objPtr->interp, "Problem: chunk file already open",
			 (char *) NULL);
	return TCL_ERROR;
    }
    cf = objPtr->cfState = Chunk_OpenFile(filename, "r");

    /*
     *  Check to see if this is a chunk file.  If not, process raw
     *  audio file.  Otherwise, load media indep. and sparcAu headers.
     */
    if (cf == NULL) {
      return TCL_ERROR; 
    } else {
	objPtr->filename = (char *) ckalloc(strlen(filename) + 1);
	strcpy(objPtr->filename, filename);
    }

    /*
     * Load Media Indep. header 
     */
    if (Chunk_FindType(cf, CM_MEDIA_INDEP_CF, &ch) == 0) {
	if (ch.chunkSize != sizeof(MediaIndepHeader)) {
	    Tcl_AppendResult(objPtr->interp, "Problem reading header.",
			     (char *) NULL);
	    return TCL_ERROR;
	} else {
	    mediaHdr = (MediaIndepHeader *) ckalloc(sizeof(MediaIndepHeader));
	    read(cf->fd, (char *) mediaHdr, ch.chunkSize);
	}
    } else {
	Tcl_AppendResult(objPtr->interp,"Can't find media independent header.",
			 (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Convert integer data to host machine's Endian format
     */

    mediaHdr->media = ntohl(mediaHdr->media);
    mediaHdr->format = ntohl(mediaHdr->format);
    mediaHdr->device = ntohl(mediaHdr->device);
    mediaHdr->version = ntohl(mediaHdr->version);
    mediaHdr->num_frames = ntohl(mediaHdr->num_frames);
    mediaHdr->length_s = ntohl(mediaHdr->length_s);
    mediaHdr->length_ns = ntohl(mediaHdr->length_ns);

    /*
     * Sanity check.
     */

    if (mediaHdr->media != AUDIO_MEDIA_CODE) {
      Tcl_AppendResult(objPtr->interp, "Not a audio clipfile.",
		       (char *) NULL);
      return TCL_ERROR;
    }
	
    /*
     * Load Sparc Audio header 
     */
    Chunk_ReadRewind(cf);
    if (Chunk_FindType(cf, CM_SPARC_AU_CF, &ch) == 0) {
	if (ch.chunkSize != sizeof(SparcAuHeader)) {
	    Tcl_AppendResult(objPtr->interp, "Problem reading header.",
			     (char *) NULL);
	    return TCL_ERROR;
	} else {
	    sparcHdr = (SparcAuHeader *) ckalloc(sizeof(SparcAuHeader));
	    read(cf->fd, (char *) sparcHdr, ch.chunkSize);
	}
    } else {
	Tcl_AppendResult(objPtr->interp,"Can't find sparc header in clipfile.",
			 (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Convert integer data to host machine's Endian format
     */
    sparcHdr->dataFormat = ntohl(sparcHdr->dataFormat);
    sparcHdr->samplingRate = ntohl(sparcHdr->samplingRate);
    sparcHdr->channelCount = ntohl(sparcHdr->channelCount);

    objPtr->sampleSize = Tcm_AuComputeSampleSize(sparcHdr->dataFormat);
    objPtr->alignSize = sparcHdr->channelCount * objPtr->sampleSize;
    objPtr->dataRate = objPtr->alignSize * sparcHdr->samplingRate;

    objPtr->numFrames = mediaHdr->num_frames;
    objPtr->clipLength = ((double) mediaHdr->length_s) +
	(((double) mediaHdr->length_ns) / 1000000000.0);

    /*
     * Determine the starting point of the audio data. 
     */
    if (Chunk_FindType(objPtr->cfState, CM_DATA_SEGMENT, &ch) != 0) {
	Chunk_ReadRewind(objPtr->cfState);
	if (Chunk_FindType(objPtr->cfState, CM_DATA_SEGMENT, &ch) != 0) {
	    Tcl_AppendResult(objPtr->interp, "Couldn't find data segment.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
    }
    objPtr->dataStart = lseek(objPtr->cfState->fd, 0, SEEK_CUR);

    if (objPtr->hdr != NULL) {
	ckfree((char *) objPtr->hdr);
    }
    if (objPtr->mediaHdr != NULL) {
	ckfree((char *) objPtr->mediaHdr);
    }

    objPtr->mediaHdr = mediaHdr;
    objPtr->hdr = sparcHdr;

    Chunk_CloseFile(objPtr->cfState);
    objPtr->cfState = NULL;
    return TCL_OK;
}


/************************************************************
 *
 * SparcAuChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
SparcAuChangeOutCmd(SparcAudioClipFile *objPtr, char *newOut)
{
    /*
     * Clear the previous entry. 
     */
    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
	objPtr->outCmd = NULL;
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return. 
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set 
     * everything accordingly. 
     */
    if (newOut[0] == '@') {

	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
	       &(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			     "Could not find registered C procedure: ",
			     newOut, (char *) NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string. 
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut) + 1);
    strcpy(objPtr->outCmd, newOut);

    return TCL_OK;
}


/************************************************************
 *
 * SparcAuLTSChanged
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Procedure called whenever the LTS changes.
 *
 ************************************************************/

static void
SparcAuLTSChanged(ClientData clientdata)
{
    SparcAudioClipFile *objPtr = (SparcAudioClipFile *) clientdata;

    if (objPtr->ready == 1) {
	SparcAuReady(objPtr);
    }
    return;
}


/************************************************************
 *
 * SparcAuSendData
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sends data according to current value of LTS, queues
 *     itself on the at queue for next send time.
 *
 ************************************************************/
static void
SparcAuSendData(ClientData clientData) 
{
    SparcAudioClipFile *objPtr = (SparcAudioClipFile *) clientData;
    double now, speed, when, end;
    double eTime, sTime, lastETime;
    Segment *segPtr;
    int i;
    SparcAuBufferHeader bufHdr;
    Buffer *bufPtr;
    int idx;
    int dataSegCode, dataLength;
    SB_List *sbList;
    char *sbStr;
    int numToSend, bwToSend;
    int curIndex;
    int sbIdx;
    static int iter = 0;
    static int fPickNum = 0;
    static FramePick *fPickList = NULL;
    FramePick *f;
    int fd;

  retry:

    /*
     * If not ready, return. 
     */
    if (!objPtr->ready)
	goto done;

    /*
     * If no LTS, return. 
     */
    if (objPtr->lts == NULL)
	goto done;

    /*
     * If no out command, return. 
     */
    if (objPtr->outCmd == NULL)
	goto done;

    /*
     * The variable 'now' represents the time when this event
     * was scheduled.
     */
    if (objPtr->atToken != -1) {
	now = objPtr->atWhen;
    } 
    else {
	now = LTS_Value(objPtr->lts);
    }
    
    objPtr->atToken = -1;
    speed = LTS_Speed(objPtr->lts);
    now += objPtr->sendAhead * speed;
    iter++;
    segPtr = objPtr->segments;

    /*
     * Find segment we are in.  If speed is backwards so go to end of
     * segment list and search backwards.
     */
    if (speed >= 0.0) {
	while (segPtr != NULL) {
	    if (segPtr->logicalEnd > now)
		break;
	    segPtr = segPtr->next;
	}
    } else {
	if (segPtr != NULL) {
	    while (segPtr->next != NULL)
		segPtr = segPtr->next;
	    while (segPtr != NULL) {
		if (segPtr->logicalStart < now)
		    break;
		segPtr = segPtr->prev;
	    }
	}
    }

    /*
     * If segment pointer is NULL, means we will never be
     * needed (i.e. time has passed all segments or  is going 
     * backwards before all segments).
     */
    if (segPtr == NULL)
	goto done;

    /*
     * If we are not in a segment, calculate when we will be and
     * then create an event in the 'at queue'.
     */
    if (speed > 0.0) {
	if (now < segPtr->logicalStart) {
	    when = LTS_SystemTimeOf(objPtr->lts, (segPtr->logicalStart -
					   (objPtr->sendAhead * speed)));
	    objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					   "audio", SparcAuSendData, objPtr,
					   NULL, NULL, NULL, NULL, NULL);
	    objPtr->atWhen = segPtr->logicalStart - (objPtr->sendAhead * speed);
	    return;
	}
    } else if (speed < 0.0) {
	if (now > segPtr->logicalEnd) {
	    when = LTS_SystemTimeOf(objPtr->lts, (segPtr->logicalEnd +
					   (objPtr->sendAhead * speed)));
	    objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					   "audio", SparcAuSendData, objPtr,
					   NULL, NULL, NULL, NULL, NULL);
	    objPtr->atWhen = segPtr->logicalEnd - (objPtr->sendAhead * speed);
	    return;
	}
    } else {
	if ((now < segPtr->logicalStart) || (now >= segPtr->logicalEnd)) {
	    return;
	}
    }

    /*
     * If file isn't open, open it now. 
     */

    fd = -1;
    if (objPtr->raw == 0) {
      if (objPtr->cfState == NULL) {
	objPtr->cfState = Chunk_OpenFile(objPtr->filename, "r");
      }
      fd = objPtr->cfState->fd;
    } else {
      if (objPtr->fd == -1) {
	objPtr->fd = open(objPtr->filename, O_RDONLY, 0);
	if (objPtr->fd == -1) {
	  goto done;
	}
	fd = objPtr->fd;
      }
    }


    /*
     * If speed or cycle time are 0.0, there is nothing to do.  
     */
    if ((objPtr->cycleTime == 0.0) || (speed == 0.0)) {
	goto done;
    }

    /*
     * Find pieces of segments to be sent.  This is done by creating an
     * array of FramePick structures and filling them in.
     * We update the bwToSend and numToSend variables as we go.
     */
    if (fPickNum == 0) {
	fPickNum = 20;
	fPickList = (FramePick *) ckalloc(sizeof(FramePick) * fPickNum);
    }
    numToSend = 0;
    bwToSend = 0;

    /*
     * Find first frame (i) and set the data in fPickList
     */
    if (speed >= 0.0) {
	int offset, size;

	for (i = 0; i < segPtr->numFrames - 1; i++) {
	    if (segPtr->FTT[i + 1] > now)
		break;
	}
	offset = TimeToOffset( now, segPtr->FTT[i], segPtr->FTT[i + 1],
			       segPtr->FOT[2*i], segPtr->FOT[2*i+1], 
                               &size, 1, objPtr->alignSize );

	fPickList[0].sTime = now;
	fPickList[0].eTime = segPtr->FTT[i + 1];
	fPickList[0].index = i;
	fPickList[0].segPtr = segPtr;
	fPickList[0].offset = offset;
	fPickList[0].size = size;
	fPickList[0].sendFlag = 1;
    } else {
	int offset, size;

	for (i = segPtr->numFrames - 1; i > 0; i--) {
	    if (segPtr->FTT[i] < now) {
		break;
	    }
	}
	offset = TimeToOffset( now, segPtr->FTT[i], segPtr->FTT[i + 1],
			       segPtr->FOT[2*i], segPtr->FOT[2*i+1], 
                               &size, 0, objPtr->alignSize );

	fPickList[0].sTime = now;
	fPickList[0].eTime = segPtr->FTT[i];
	fPickList[0].index = i;
	fPickList[0].segPtr = segPtr;
	fPickList[0].offset = offset;
	fPickList[0].size = size;
	fPickList[0].sendFlag = 1;
    }

    /* 
     * Sometimes we ares sufficiently close to the beginning or end of
     * a segment that we think there is data to send, but in reality the
     * difference between now and the end (if forward) or beginning (if backward)
     * is smaller than even one sample size. This will show up as size = 0.
     * If this is the case, go back to beginning and try again. By now, we should
     * have be into the next segment or beyond all segments and the right thing
     * will happen.
     */

    if (fPickList[0].size == 0) {
      goto retry;
    }

    /*
     * Found the first frame.  Now, figure out the rest, updating the
     * bwToSend and numToSend variables as we go.
     */
    numToSend = 1;
    bwToSend = fPickList[0].size;
    curIndex = fPickList[0].index;
    end = now + (objPtr->cycleTime * speed);
    
    if (speed > 0.0) {
	int offset;
	Segment *lastSegPtr;

	/*
	 * Continue adding data until we reach the ending time
	 * for this cycle.
	 */
	while (1) {
	    curIndex++;

	    /*
	     * If we're past the end of this segment, move on to next. 
	     */
	    lastSegPtr = NULL;
	    if (curIndex >= segPtr->numFrames) {
		lastSegPtr = segPtr;
		segPtr = segPtr->next;
		if (segPtr == NULL) {
		    break;
		}
		curIndex = 0;
	    }

	    /*
	     * If we've got all the frames we need, end the construction.
	     */
	    if (segPtr->FTT[curIndex] >= end) {
		break;
	    }

	    /*
	     * Add this frame to the list of frames to send.
	     */
	    if (numToSend >= fPickNum) {
		fPickNum += 20;
		fPickList = (FramePick *) ckrealloc((char *) fPickList,
				    sizeof(FramePick) * fPickNum);
	    }
	    fPickList[numToSend].sTime = segPtr->FTT[curIndex];
	    fPickList[numToSend].eTime = segPtr->FTT[curIndex + 1];
	    fPickList[numToSend].index = curIndex;
	    fPickList[numToSend].segPtr = segPtr;
	    fPickList[numToSend].offset = segPtr->FOT[curIndex * 2];
	    fPickList[numToSend].size = segPtr->FOT[(curIndex * 2) + 1];
	    fPickList[numToSend].sendFlag = 1;
	    bwToSend += fPickList[numToSend].size;
	    numToSend++;
	}

	/*
	 * If we ended at the end of some segment, then reset 'end'
	 * to the actual ending time.  We ran out of data to send
	 * because we have reached the end of the clip or a silent
	 * period between segments.
	 */
	if (lastSegPtr != NULL) {
	    end = min(lastSegPtr->logicalEnd, end);
	}

	/*
	 * Update the size and time of the last frame
	 */
	f = &(fPickList[numToSend - 1]);
	offset = TimeToOffset( end, f->sTime, f->eTime, f->offset, f->size,
			       &(f->size), 0, objPtr->alignSize );
	f->eTime = end;

    } else {
	Segment *lastSegPtr;

        /*
	 * Speed is in reverse:  build the frame list by going backwards.
	 */
	lastSegPtr = NULL;
	while (1) {
	    curIndex--;
	    if (curIndex < 0) {
		lastSegPtr = segPtr;
		segPtr = segPtr->prev;
		if (segPtr == NULL) {
		    break;
		}
		curIndex = segPtr->numFrames - 1;
	    }

	    /*
	     * If we've got all the frames we need, end the construction.
	     */
	    if (segPtr->FTT[curIndex + 1] < end) {
		break;
	    }

	    /*
	     * Add this frame to the list of frames to send.
	     */
	    if (numToSend >= fPickNum) {
		fPickNum += 20;
		fPickList = (FramePick *) ckrealloc((char *) fPickList,
					sizeof(FramePick) * fPickNum);
	    }
	    fPickList[numToSend].sTime = segPtr->FTT[curIndex + 1];
	    fPickList[numToSend].eTime = segPtr->FTT[curIndex];
	    fPickList[numToSend].index = curIndex;
	    fPickList[numToSend].segPtr = segPtr;
	    fPickList[numToSend].offset = segPtr->FOT[curIndex * 2];
	    fPickList[numToSend].size = segPtr->FOT[(curIndex * 2) + 1];
	    fPickList[numToSend].sendFlag = 1;
	    bwToSend += fPickList[numToSend].size;
	    numToSend++;
	}

	/*
	 * If we ended at the end of some segment, then reset 'end'
	 * to the actual ending time.  We ran out of data to send
	 * because we have reached the end of the clip or a silent
	 * period between segments.
	 */
	if (lastSegPtr != NULL) {
	    end = max(lastSegPtr->logicalStart, end);
	}

	/*
	 * Update the time, size, and offset of the last frame
	 */
	f = &(fPickList[numToSend - 1]);
	f->offset = TimeToOffset( end, f->eTime, f->sTime, f->offset, f->size,
				  &(f->size), 1, objPtr->alignSize );
	f->eTime = end;
    }

    /*
     * For each frame chosen, create buffers with headers and data.
     */
    sbList = (SB_List *) ckalloc(sizeof(SB_List));
    sbList->sbs =
	(ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *) * numToSend);
    sbList->count = numToSend;
    sbIdx = 0;
    for (i = 0; i < numToSend; i++) {
	if (fPickList[i].sendFlag == 1) {
	    double lsS, lsNs, leS, leNs;

	    segPtr = fPickList[i].segPtr;

	    sTime = fPickList[i].sTime;
	    eTime = fPickList[i].eTime;

	    /* Note: subtract 8 (two int's) from the size
	     * of the buffer header for the chunkFile support fields.
	     */

	    bufHdr.hdrType = htonl(CM_SPARC_AU_BUF_HDR);
		bufHdr.device = htonl(SPARC_AU_DEV_CODE);
	    bufHdr.length = htonl(sizeof(SparcAuBufferHeader) - 8);
	    bufHdr.dataFormat = htonl(objPtr->hdr->dataFormat);
	    bufHdr.samplingRate = htonl(objPtr->hdr->samplingRate);
	    bufHdr.channelCount = htonl(objPtr->hdr->channelCount);
	    lsS = floor(sTime);
	    lsNs = (sTime - floor(sTime)) * ONE_BILLION;
	    leS = floor(eTime);
	    leNs = (eTime - floor(eTime)) * ONE_BILLION;
	    bufHdr.logStart_s = htonl((int) lsS);
	    bufHdr.logStart_ns = htonl((int) lsNs);
	    bufHdr.logEnd_s = htonl((int) leS);
	    bufHdr.logEnd_ns = htonl((int) leNs);

	    /*
	     * Load data -- Reversed segments need to be handled specially
	     */
	    if (fPickList[i].segPtr->reverse == -1) {
		fPickList[i].offset = fPickList[i].offset + fPickList[i].size;
		fPickList[i].size = -fPickList[i].size;
	    }
	    
	    lseek(fd, objPtr->dataStart + fPickList[i].offset,
		  SEEK_SET);
	      

	    bufPtr = MakeBuffer(objPtr->interp,
			   sizeof(SparcAuBufferHeader) + 8 + fPickList[i].size,
			   "tcmSparcAuCF:SparcAuSendData");
	    fPickList[i].bufPtr = bufPtr;
	    idx = 0;

	    memcpy(bufPtr->data, &bufHdr, sizeof(SparcAuBufferHeader));
	    idx += sizeof(SparcAuBufferHeader);

	    dataSegCode = htonl(CM_DATA_SEGMENT);
	    dataLength = htonl(fPickList[i].size);

	    memcpy(bufPtr->data + idx, &dataSegCode, 4);
	    idx += 4;

	    memcpy(bufPtr->data + idx, &dataLength, 4);
	    idx += 4;

	    read(fd, bufPtr->data + idx, fPickList[i].size);

	    fPickList[i].bufPtr = bufPtr =
		Tcm_AudioResample(bufPtr, fPickList[i].size,
			      fPickList[i].eTime - fPickList[i].sTime,
			      objPtr->alignSize, objPtr->hdr->samplingRate,
			      segPtr->reverse, objPtr->interp, speed,
			      &dataLength);

	    dataLength += sizeof(SparcAuBufferHeader) + 8;
	    sbList->sbs[sbIdx] = MakeScatterBuffer(bufPtr, dataLength, NULL);
#if 0
	    AudioLog( sbList->sbs[sbIdx] );
#endif /* debug */
	    sbIdx++;
	}
    }

    /*
     * If out command is C, create scatter buffer list and call it.
     */
    if (objPtr->outCCmd != NULL) {

	/*
	 * Need to check if it still exists since object may have been
	 * destroyed.
	 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd + 1) != NULL) {
			objPtr->outCCmd(objPtr->outCD, sbList);
		} else {
			objPtr->outCCmd = NULL;
			ckfree((char *) objPtr->outCmd);
			objPtr->outCmd = NULL;
		}

    } else {
	/*
	 * Out command is tcl, make sb list into a string and call using
	 * Tcl_Eval. 
	 */
	sbStr = MakeSBListString(sbList);
	Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL);
	ckfree((char *) sbStr);
    }
    FreeSBList(sbList);

    /*
     * We're done with the buffers, so release them.
     */
    lastETime = 0;
    for (i = 0; i < numToSend; i++) {
	if (fPickList[i].sendFlag == 1) {
	    lastETime = fPickList[i].eTime;
	    DetachBuffer(objPtr->interp, fPickList[i].bufPtr,
				     "tcmSparcAuCF:SparcAuSendData");
	}
    }

    /* 
     * Find next wake up time and set up 'at' event. 
     */
    when = LTS_SystemTimeOf(objPtr->lts, (lastETime - (objPtr->sendAhead * speed)));
    objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL, "audio", SparcAuSendData, 
					  objPtr, NULL, NULL, NULL, NULL, NULL);

    objPtr->atWhen = lastETime - (objPtr->sendAhead * speed);

    if (objPtr->fd != -1) {
	close (objPtr->fd);
	objPtr->fd = -1;
    }
    return;

done:
    objPtr->atToken = -1;
    if (objPtr->cfState != NULL) {
	Chunk_CloseFile(objPtr->cfState);
	objPtr->cfState = NULL;
    }
    if (objPtr->fd != -1) {
	close (objPtr->fd);
	objPtr->fd = -1;
    }
    return;
}


/*
 *--------------------------------------------------------------
 *
 * ReadAudioHeader --
 *
 *	Read a sparc/next audio file header from the file descriptor
 *	passed in.  Convert it to native format and make sure it's
 *	valid.  If so, alloc a new RawSparcAuHeader structure, fill it
 *	in, and return it to the caller.  The structure is completely
 *	filled in, including the dataSize field (which is sometimes
 *	left empty in sparc audio files).  The new memory belongs to
 *	the caller, and the caller should free it using ckfree() when
 *	they are finished.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	Memory is allocated and handed to the caller.
 *	The file pointer is reset to the point directly after the
 *	file header.
 *
 * Parameters:
 *	interp	(in)	Used to report errors
 *	fd	(in)	File descriptor to read
 *	hdrPtr	(out)	Plaace to write filled in structure.
 *
 *--------------------------------------------------------------
 */
static int
ReadAudioHeader(Tcl_Interp *interp, int fd, RawSparcAuHeader **hdrPtr)
{

    int n;
    RawSparcAuHeader srcHdr;
    RawSparcAuHeader *new;

    lseek (fd, 0, SEEK_SET);
    n = read (fd, &srcHdr, sizeof(srcHdr));
    if (n != sizeof(srcHdr)) {
	Tcl_AppendResult(interp, "File is not a sparc audio file", NULL);
	return TCL_ERROR;
    }

    srcHdr.magic        = ntohl(srcHdr.magic);
    srcHdr.dataLocation = ntohl(srcHdr.dataLocation);
    srcHdr.dataSize     = ntohl(srcHdr.dataSize);
    srcHdr.dataFormat   = ntohl(srcHdr.dataFormat);
    srcHdr.samplingRate = ntohl(srcHdr.samplingRate);
    srcHdr.channelCount = ntohl(srcHdr.channelCount);

    if (srcHdr.magic != SUN_MAGIC) {
	Tcl_AppendResult(interp, "File is not a sparc audio file", NULL);
	return TCL_ERROR;
    }

    /*
     * Figure out the size of the file, if it's not specified
     */
    if (srcHdr.dataSize == SUN_UNSPEC) {
	struct stat buf;
	fstat(fd, &buf);
	srcHdr.dataSize = buf.st_size - srcHdr.dataLocation;
    }

    if (srcHdr.dataFormat != SUN_ULAW &&
	srcHdr.dataFormat != SUN_LIN_8 &&
	srcHdr.dataFormat != SUN_LIN_16) {
	Tcl_AppendResult(interp, "Source is in an unsupported format.\n");
	return TCL_ERROR;
    }

    /*
     * Fill in the return value
     */
    new = (RawSparcAuHeader *)ckalloc(sizeof(RawSparcAuHeader));
    memcpy (new, &srcHdr, sizeof(RawSparcAuHeader));
    *hdrPtr = new;
    return TCL_OK;
}

/************************************************************
 *
 * SparcAuOpenRawFile
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Opens the clipfile, reads basic info data.
 *     Closes file when done.
 *
 ************************************************************/

static int
SparcAuOpenRawFile (SparcAudioClipFile *objPtr, char *filename)
{
    /*
     * Can only deal with one file at a time
     */
    if (objPtr->fd != -1) {
	Tcl_AppendResult(objPtr->interp, "Problem: file already open",
			 (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Open the file
     */
    objPtr->fd = open(filename, O_RDONLY, 0);
    if (objPtr->fd == -1) {
	Tcl_AppendResult(objPtr->interp, "Error opening sparc audio file \"",
			 filename, "\": ", Tcl_PosixError(objPtr->interp),
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Read in the audio header.
     */
    if (objPtr->rawHdr != NULL) {
	ckfree ((char *) objPtr->rawHdr);
	objPtr->rawHdr = NULL;
    }
    if (ReadAudioHeader(objPtr->interp,objPtr->fd,&objPtr->rawHdr) != TCL_OK) {
	close (objPtr->fd);
	objPtr->fd = -1;
	return TCL_ERROR;
    }

    /*
     * Configure SparcAuHeader
     */

    objPtr->hdr = (SparcAuHeader *) ckalloc(sizeof(SparcAuHeader));

    objPtr->hdr->dataFormat = objPtr->rawHdr->dataFormat;
    objPtr->hdr->samplingRate = objPtr->rawHdr->samplingRate;
    objPtr->hdr->channelCount = objPtr->rawHdr->channelCount;


    /*
     * Fill in misc fields
     */
    objPtr->dataStart = objPtr->rawHdr->dataLocation;
    objPtr->filename = (char *) ckalloc(strlen(filename) + 1);
    strcpy(objPtr->filename, filename);


    /*
     * Compute the clip file's length.  This is the total data size 
     * over the data rate of the file.
     */

    objPtr->sampleSize = Tcm_AuComputeSampleSize(objPtr->hdr->dataFormat);
    objPtr->alignSize = objPtr->rawHdr->channelCount * objPtr->sampleSize;
    objPtr->dataRate = objPtr->alignSize * objPtr->rawHdr->samplingRate;
    objPtr->clipLength = (double)objPtr->rawHdr->dataSize / (double)objPtr->dataRate;

    /*
     * Close and return
     */
    close (objPtr->fd);
    objPtr->fd = -1;
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * SparcAuRawAddSeg --
 *
 *	Map a new segment of the clipfile into the logical time system
 *
 * Results:
 *	A standard tcl result
 *
 * Side effects:
 *      From now on, whenever the logical time system is inside the
 *      time value passed in, the appropriate segment of the audio file
 *      will be sent on to the next object.
 *
 * Parameters:
 *     SparcAudioClipFile *objPtr	pointer to object
 *     double clipStart, clipEnd	Start and end of segment in file
 *     double logStart, logEnd		Destination section in LTS
 *     int logStartFlag			Has logStart been set?
 *     int logEndFlag			Has logEnd been set?
 *
 *--------------------------------------------------------------
 */
static int
SparcAuRawAddSeg(SparcAudioClipFile *objPtr, double clipStart, double clipEnd, 
		 double logStart, double logEnd, int logStartFlag,
		 int logEndFlag)
     
{
  Segment *newSeg, *nextSeg, *prevSeg;
  char retStr[32];
  double clipDur;

  SparcAuUnready(objPtr);

  if (clipStart > clipEnd) {
    Tcl_AppendResult(objPtr->interp,
		     "Clip start can not be after clip end.", NULL);
    return TCL_ERROR;
  }
  if (objPtr->filename == NULL) {
    Tcl_AppendResult(objPtr->interp, "Must specify clipfile first.",
		     NULL);
    return TCL_ERROR;
  }

  /*
   * Check boundary conditions. 
   */
  if (clipStart < 0.0)
    clipStart = 0.0;
  if (clipStart > objPtr->clipLength) {
    Tcl_AppendResult(objPtr->interp, "Clip start too large.", NULL);
    return TCL_ERROR;
  }
  if (clipEnd > objPtr->clipLength)
    clipEnd = objPtr->clipLength;
  if (clipEnd < 0.0) {
    Tcl_AppendResult(objPtr->interp, "Clip end too small.", NULL);
    return TCL_ERROR;
  }
  if (clipStart >= clipEnd) {
    Tcl_AppendResult(objPtr->interp,
		     "ClipEnd must be greater than ClipStart.", NULL);
    return TCL_ERROR;
  }

  /*
   * Boundary conditions OK, create a new segment structure.
   */
  newSeg = (Segment *) ckalloc(sizeof(Segment));
  newSeg->clipStart = clipStart;
  newSeg->clipEnd = clipEnd;

  clipDur = clipEnd - clipStart;
  if (clipDur <= 0.0) {
    ckfree((char *) newSeg);
    Tcl_AppendResult(objPtr->interp, "Clip segment duration too small.",
		     NULL);
    return TCL_ERROR;
  }

  /*
   * Handle the various cases of start and end in LTS system being
   * specified.  Set logicalStart, logicalEnd, and reverse fields of
   * newSeg
   */
  newSeg->reverse = 1;
  if (logStartFlag && logEndFlag) {
    /*
     * Set logStart and logEnd. 
     */
    if (logStart > logEnd) {
      newSeg->reverse = -1;
      newSeg->logicalStart = logEnd;
      newSeg->logicalEnd = logStart;
    } else {
      newSeg->logicalStart = logStart;
      newSeg->logicalEnd = logEnd;
    }
  } else if (logStartFlag) {
    newSeg->logicalStart = logStart;
    newSeg->logicalEnd = logStart + clipDur;
  } else if (logEndFlag) {
    newSeg->logicalEnd = logEnd;
    newSeg->logicalStart = logEnd - clipDur;
  } else {
    newSeg->logicalStart = 0.0;
    newSeg->logicalEnd = clipDur;
  }

  /*
   * Find insertion point in list. Sort by log. start times.
   */
  nextSeg = objPtr->segments;
  prevSeg = NULL;
  while (nextSeg != NULL) {
    if (nextSeg->logicalStart < newSeg->logicalStart) {
      prevSeg = nextSeg;
      nextSeg = nextSeg->next;
    } else {
      break;
    }
  }

  /*
   * Set scale for non-uniform playback speed.
   */
  newSeg->scale = clipDur / (newSeg->logicalEnd - newSeg->logicalStart);

  /*
   * We only have one frame.
   */

  newSeg->numFrames = 1;

  newSeg->FOT = (int *) ckalloc(sizeof(int) * 2);
  newSeg->FTT = (double *) ckalloc(sizeof(double) * 2);

  newSeg->FOT[0] = objPtr->dataStart + clipStart*objPtr->dataRate + 0.5;
  newSeg->FOT[1] = clipDur * objPtr->dataRate;
  newSeg->FTT[0] = newSeg->logicalStart;
  newSeg->FTT[1] = newSeg->logicalEnd;

  /*
   * Finish setting up segment:
   *	1. segment id
   *	2. next and prev pointers
   */
  newSeg->segId = objPtr->nextSegId++;
  newSeg->next = nextSeg;
  newSeg->prev = prevSeg;
  if (prevSeg == NULL) {
    objPtr->segments = newSeg;
  } else {
    prevSeg->next = newSeg;
  }
  if (nextSeg != NULL) {
    nextSeg->prev = newSeg;
  }
  sprintf(retStr, "%d", newSeg->segId);

  Tcl_SetResult(objPtr->interp, retStr, TCL_VOLATILE);

  objPtr->segmentListChangeFlag = 1;

  return TCL_OK;
}

/************************************************************
 *
 * SparcAuSegmentList
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Returns list of current segment id's. If no segments,
 *     list is empty.
 ************************************************************/

static int
SparcAuSegmentList(SparcAudioClipFile *objPtr)
{
    Tcl_DString res_list;
    Segment *seg_ptr;
    char temp_str[100];

    seg_ptr = objPtr->segments;

    Tcl_DStringInit(&res_list);

    while (seg_ptr != NULL) {
	sprintf(temp_str, "%d", seg_ptr->segId);
    
	Tcl_DStringAppendElement(&res_list, temp_str);
	
	seg_ptr = seg_ptr->next;
    }

    Tcl_SetResult(objPtr->interp, Tcl_DStringValue(&res_list), TCL_VOLATILE);
    
    Tcl_DStringFree(&res_list);

    return TCL_OK;
}

/************************************************************
 *
 * SparcAuConfigureSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segmentId - id of segment to be reconfigured
 *     clipStart - new clip start.
 *     clipEnd - new clip end.
 *     logStart - new logical start.
 *     logEnd - new logical end.
 *     logStartFlag - indicates whetehr logStart value set or not.
 *     logEndFlag - indicates whether logEnd value set or not.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Reconfigures a previously created segment. Allows all
 *     aspects of the segment to be changed (i.e. clip start/end
 *     and logical start/end). Comparisons made with current
 *     segment info to decide whether change requires rebuilding
 *     tables or if current tables can be slightly altered.
 *     Segment removed from list and re-inserted to ensure proper
 *     ordering.
 *
 ************************************************************/

static int
SparcAuConfigureSegment(SparcAudioClipFile *objPtr, int segmentId,
			double clipStart, double clipEnd, 
			double logStart, double logEnd, 
			int logStartFlag, int logEndFlag)
{
  int new_seg_id;
  int result;
  Segment *segment_ptr;

  /*
   * This function could be optimized to recognize cases where
   * the clip only moves in logical time or the new clip is a 
   * proper subset of the original. In interest of getting this
   * function working as soon as possible, however, this code
   * currently just calls SparcAuAddSeg or SparcAuRawAddSeg
   * to create a new segment with the new clip parameters, 
   * deletes the old segment and changes the segmentId of the new 
   * segment to be the same as the old segment.
   */

  /* Unready the object. */

  SparcAuUnready(objPtr);

  /* 
   * Find the original segment to be sure that it
   * existed.
   */

  segment_ptr = objPtr->segments;

  while (segment_ptr != NULL) {
    if (segment_ptr->segId == segmentId) 
      break;
    segment_ptr = segment_ptr->next;
  }

  if (segment_ptr == NULL) {
    char tempStr[40];

    sprintf(tempStr, "%d", segmentId);
    Tcl_AppendResult(objPtr->interp, "Segment ", tempStr, " does not exist.",
		     (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Remove the original segment.
   */

  SparcAuDelSeg(objPtr, segmentId);

  /* 
   * Create a segment with new parameters.
   */

  if (objPtr->raw == 0) {
      result = SparcAuAddSeg(objPtr, clipStart, clipEnd, logStart,
			     logEnd, logStartFlag, logEndFlag);
  } else {
      result = SparcAuRawAddSeg(objPtr, clipStart, clipEnd, logStart,
				logEnd, logStartFlag, logEndFlag);
  }

  /*
   * Change segment id of new segment to be segment id of original segment.
   */

  if (result == TCL_OK) {
    char resStr[40];

    if (Tcl_GetInt(objPtr->interp, objPtr->interp->result, &new_seg_id) != TCL_OK) {
      Tcl_AppendResult(objPtr->interp, "Couldn't reconfigure segment.", (char *) NULL);
      return TCL_ERROR;
    }

    segment_ptr = objPtr->segments;
    while (segment_ptr != NULL) {
      if (segment_ptr->segId == new_seg_id) {
	break;
      }
      segment_ptr = segment_ptr->next;
    }

    if (segment_ptr == NULL) {
      Tcl_AppendResult(objPtr->interp, "Couldn't reconfigure segment.", (char *) NULL);
      return TCL_ERROR;
    }

    segment_ptr->segId = segmentId;

    sprintf(resStr, "%d", segmentId);
    Tcl_SetResult(objPtr->interp, resStr, TCL_VOLATILE);

    /* 
     * Set segment list change flag so that next ready will check for overlaps.
     */

    objPtr->segmentListChangeFlag = 1;

    return TCL_OK;
  }

  return TCL_ERROR;
}




#if 0
static int
AudioLog ( ScatterBuffer *sb)
{
    Buffer *buf;
    char *data;
    int size;

    data = GetFromSB(sb, CM_DATA_SEGMENT, &buf, &size);

    write(AudioLogFD, data, size);
    return TCL_OK;
}
#endif /* debug */


