/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */



/*
 * $Id: cmAuFileInterface.c,v 1.8 1996/02/20 23:00:35 aswan Exp $
 *
 * Implements the au audio file reading routines, the real
 * implementation of auSegment functionality
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <tcl.h>
#include <math.h>
#include "cmBufferlist.h"
#include "cmAuFileInterface.h"

#ifndef O_BINARY
#    define O_BINARY 0
#endif

#define ONE_BILLION 1000000000.0

/*
 * Module global variables.
 */

static int moduleInitFlag = 0;		/* Indicates that file interface
					   module has been initialized. */
static Tcl_HashTable fileHashTable;	/* Hash table used to relat file names
					   to file interface data structures. */

/*
 * Declarations for static procedures defined in this file. 
 */

static int Au_FIP_ParseFile(AuDataFile *dataFile);
static void Au_FIP_ModuleInit();
static int Au_FIP_CalculateOffset(AuDataFile *dataFile, int frameNum,
				  double time);

/*************************************************************************
 * Au_FIP_ModuleInit
 *
 * Parameters: 
 *     None.
 * 
 * Results:
 *     None.
 *
 * Actions:
 *     Initializes module. Sets global init flag and prepares hash table.
 *
 *************************************************************************/
static void
Au_FIP_ModuleInit()
{
    if (moduleInitFlag) return;

    moduleInitFlag = 1;

    Tcl_InitHashTable(&fileHashTable, TCL_STRING_KEYS);
}

/*************************************************************************
 * Au_FIP_Open
 *
 * Parameters:
 *     fname - name of file to be accessed.
 *
 * Results:
 *     Pointer to AuFileInterface structure to be used for this file.
 *     If error occurs, returns NULL.
 * 
 * Action:
 *     Looks up filename in hash table to find previously created data file
 *     structure if it exists. If so, creates new AuFileInterface structure
 *     pointing to this data file structure. If not, creates a new data file
 *     structure, parses timing and audio headers as necessary, enters into
 *     hash table, and then creates new AuFileInterface with this new data
 *     file structure.
 *
 *************************************************************************/

AuFileInterface *
Au_FIP_Open(char *fname)
{
    AuFileInterface *new_fip;
    AuDataFile *dataFile;
    Tcl_HashEntry *entry_ptr;
    int newFile;

    if (!moduleInitFlag) {
	Au_FIP_ModuleInit();
    }

    entry_ptr = Tcl_CreateHashEntry(&fileHashTable, fname, &newFile);

    if (newFile) {
	
	dataFile = (AuDataFile *) ckalloc(sizeof(AuDataFile));
	dataFile->filename = (char *) ckalloc(strlen(fname)+1);
	strcpy(dataFile->filename, fname);
	dataFile->cfState = NULL;
	dataFile->hdr = NULL;
	dataFile->mediaHdr = NULL;
	dataFile->sampleSize = 1;
	dataFile->clipLength = 0.0;
	dataFile->dataStart = 0;
	dataFile->dataRate = 0;
	dataFile->numFrames = 0;
	dataFile->fd = -1;
	dataFile->file_refCount = 1;
	dataFile->open_refCount = 0;
	dataFile->FOT = NULL;
	dataFile->FTT = NULL;
	dataFile->raw_flag = 0;

	if (Au_FIP_ParseFile(dataFile) != TCL_OK) {
	    Tcl_DeleteHashEntry(entry_ptr);
	    ckfree(dataFile->filename);
	    ckfree((char *) dataFile);
	    return NULL;
	}

	Tcl_SetHashValue(entry_ptr, (ClientData) dataFile);
    } else {

	dataFile = (AuDataFile *) Tcl_GetHashValue(entry_ptr);
	dataFile->file_refCount++;
    }

    new_fip = (AuFileInterface *) ckalloc(sizeof(AuFileInterface));
    
    new_fip->data_file = dataFile;
    new_fip->is_open = 0;

    return new_fip;
}

/*************************************************************************
 * Au_FIP_Close 
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *
 * Results: 
 *     None.
 * 
 * Actions:
 *     Destroys AuFileInterface structure, decrementing file ref count in
 *     underlying data file structure. If ref count falls to 0, underlying
 *     data file structure is destroyed as well.
 *
 *************************************************************************/

void
Au_FIP_Close(AuFileInterface *fip)
{
    Tcl_HashEntry *entry_ptr;
    AuDataFile *dataFile;

    Au_FIP_FDClose(fip);

    dataFile = fip->data_file;

    dataFile->file_refCount--;
    ckfree((char *) fip);

    if (dataFile->file_refCount <= 0) {
	entry_ptr = Tcl_FindHashEntry(&fileHashTable, dataFile->filename);

	if (entry_ptr != NULL) {
	    Tcl_DeleteHashEntry(entry_ptr);
	}

	if (dataFile->filename != NULL) {
	    ckfree(dataFile->filename);
	}
	if (dataFile->hdr != NULL) {
	    ckfree((char *) dataFile->hdr);
	}
	if (dataFile->mediaHdr != NULL) {
	    ckfree((char *) dataFile->mediaHdr);
	}
	if (dataFile->FOT != NULL) {
	    ckfree((char *) dataFile->FOT);
	}
	if (dataFile->FTT != NULL) {
	    ckfree((char *) dataFile->FTT);
	}
	ckfree((char *) dataFile);
    }
}

/*************************************************************************
 * Au_FIP_FDOpen 
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *
 * Results:
 *     None.
 * 
 * Actions:
 *     Used to open file descriptor for file accessed through the AuFileInterface
 *     structure. Ref. count kept in underlying data file structure. Open status
 *     cached in AuFileInterface structure.
 *
 *************************************************************************/

void
Au_FIP_FDOpen(AuFileInterface *fip) 
{
    AuDataFile *dataFile;
    
    if (fip->is_open == 1) {
	return;
    }

    fip->is_open = 1;
    
    dataFile = fip->data_file;

    dataFile->open_refCount++;

    if (!dataFile->raw_flag) {
	if (dataFile->cfState == NULL) {
	    dataFile->cfState = Chunk_OpenFile(dataFile->filename, "r");
	}
    } else {
	if (dataFile->fd < 0) {
	    dataFile->fd = open(dataFile->filename, O_RDONLY|O_BINARY);
	}
    }
}

/*************************************************************************
 * Au_FIP_FDClose 
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *
 * Results:
 *     None.
 *
 * Actions:
 *     Used to close file descriptor accessed through the AuFileInterface 
 *     structure passed into this procedure. Ref. count maintained in underlying
 *     data file structure. File descriptor not actually closed until ref count
 *     falls to 0. Status of file descriptor relative to the AuFileInterface
 *     is cached in that structure.
 *
 *************************************************************************/

void
Au_FIP_FDClose(AuFileInterface *fip)
{
    AuDataFile *dataFile;

    if (fip->is_open == 0) {
	return;
    }

    fip->is_open = 0;
    dataFile = fip->data_file;

    dataFile->file_refCount--;

    if (dataFile->open_refCount <= 0) {
	if (dataFile->cfState != NULL) {
	    Chunk_CloseFile(dataFile->cfState);
	    dataFile->cfState = NULL;
	}
	if (dataFile->fd != -1) {
	    close(dataFile->fd);
	    dataFile->fd = -1;
	}
	dataFile->open_refCount = 0;
    }
}

/*************************************************************************
 * Au_FIP_ParseFile
 * 
 * Parameters:
 *     dataFile - pointer to underlying data file structure for a file.
 *
 * Results:
 *     Standard Tcl result code.
 * 
 * Actions:
 *     Parses timing and header information out of a file and fills the
 *     corresponding fields in the dataFile structure. Returns TCL_ERROR
 *     if parse error occurs or if file does not exist or if file is not
 *     in an understandable format (i.e., file is not Sparc Audio).
 *
 *************************************************************************/

static int
Au_FIP_ParseFile(AuDataFile *dataFile)
{
    int i, n;
    int *fileFTT;
    double sec, nsec;

    dataFile->cfState = Chunk_OpenFile(dataFile->filename, "r");

    if (dataFile->cfState == NULL) {
	
	dataFile->raw_flag = 1;
	dataFile->fd = open(dataFile->filename, O_RDONLY|O_BINARY);
	
	if (dataFile->fd < 0) {
	    return TCL_ERROR;
	}

	/* 
	 * Raw case here.
	 */

	/*
	 * Read in the audio header.
	 */
	dataFile->rawHdr = (RawSparcAuHeader *) ckalloc(sizeof(RawSparcAuHeader));
	n = read(dataFile->fd, dataFile->rawHdr, sizeof(RawSparcAuHeader));
	if (n != sizeof(RawSparcAuHeader)) goto error;
	
	dataFile->rawHdr->magic        = ntohl(dataFile->rawHdr->magic);
	dataFile->rawHdr->dataLocation = ntohl(dataFile->rawHdr->dataLocation);
	dataFile->rawHdr->dataSize     = ntohl(dataFile->rawHdr->dataSize);
	dataFile->rawHdr->dataFormat   = ntohl(dataFile->rawHdr->dataFormat);
	dataFile->rawHdr->samplingRate = ntohl(dataFile->rawHdr->samplingRate);
	dataFile->rawHdr->channelCount = ntohl(dataFile->rawHdr->channelCount);

	if (dataFile->rawHdr->magic != SUN_MAGIC) {
	    goto error;
	}

	/*
	 * Figure out the size of the file, if it's not specified
	 */

	if (dataFile->rawHdr->dataSize == SUN_UNSPEC) {
	    struct stat buf;
	    fstat(dataFile->fd, &buf);
	    dataFile->rawHdr->dataSize = buf.st_size - dataFile->rawHdr->dataLocation;
	}

	if (dataFile->rawHdr->dataFormat != SUN_ULAW &&
	    dataFile->rawHdr->dataFormat != SUN_LIN_8 &&
	    dataFile->rawHdr->dataFormat != SUN_LIN_16) {
	    goto error;
	}

	/*
	 * Configure SparcAuHeader
	 */

	dataFile->hdr = (SparcAuHeader *) ckalloc(sizeof(SparcAuHeader));

	dataFile->hdr->dataFormat = dataFile->rawHdr->dataFormat;
	dataFile->hdr->samplingRate = dataFile->rawHdr->samplingRate;
	dataFile->hdr->channelCount = dataFile->rawHdr->channelCount;


	/*
	 * Fill in misc fields
	 */
	dataFile->dataStart = dataFile->rawHdr->dataLocation;

	/*
	 * Compute the clip file's length.  This is the total data size 
	 * over the data rate of the file.
	 */

	dataFile->sampleSize = Tcm_AuComputeSampleSize(dataFile->hdr->dataFormat);
	dataFile->alignSize = dataFile->rawHdr->channelCount * dataFile->sampleSize;
	dataFile->dataRate = dataFile->alignSize * dataFile->rawHdr->samplingRate;
	dataFile->clipLength = (double)dataFile->rawHdr->dataSize / (double)dataFile->dataRate;

	dataFile->numFrames = 1;

	dataFile->FTT = (double *) ckalloc(sizeof(double) * 2);
	dataFile->FTT[0] = 0.0;
	dataFile->FTT[1] = dataFile->clipLength;

	dataFile->FOT = (int *) ckalloc(sizeof(int) * 2);
	dataFile->FOT[0] = 0;
	dataFile->FOT[1] = dataFile->rawHdr->dataSize;
	
    } else {
	
	/*
	 * Chunk file case here.
	 */

	tChunkHeader ch;
	

	dataFile->raw_flag = 0;

	if (Chunk_FindType(dataFile->cfState, CM_MEDIA_INDEP_CF, &ch) == 0) {
	    dataFile->mediaHdr = (MediaIndepHeader *) ckalloc(sizeof(MediaIndepHeader));
	    read(dataFile->cfState->fd, (char *) dataFile->mediaHdr, sizeof(MediaIndepHeader));
	} else {
	    goto error;
	}

	/*
	 * Convert integer data to host machine's Endian format
	 */

	dataFile->mediaHdr->media = ntohl(dataFile->mediaHdr->media);
	dataFile->mediaHdr->format = ntohl(dataFile->mediaHdr->format);
	dataFile->mediaHdr->device = ntohl(dataFile->mediaHdr->device);
	dataFile->mediaHdr->version = ntohl(dataFile->mediaHdr->version);
	dataFile->mediaHdr->num_frames = ntohl(dataFile->mediaHdr->num_frames);
	dataFile->mediaHdr->length_s = ntohl(dataFile->mediaHdr->length_s);
	dataFile->mediaHdr->length_ns = ntohl(dataFile->mediaHdr->length_ns);
	
	/*
	 * Sanity check.
	 */
	
	if (dataFile->mediaHdr->media != AUDIO_MEDIA_CODE) {
	    goto error;
	}
	
	/*
	 * Load Sparc Audio header 
	 */
	Chunk_ReadRewind(dataFile->cfState);
	if (Chunk_FindType(dataFile->cfState, CM_SPARC_AU_CF, &ch) == 0) {
	    if (ch.chunkSize != sizeof(SparcAuHeader)) {
		goto error;
	    } else {
		dataFile->hdr = (SparcAuHeader *) ckalloc(sizeof(SparcAuHeader));
		read(dataFile->cfState->fd, (char *) dataFile->hdr, ch.chunkSize);
	    }
	} else {
	    goto error;
	}

	/*
	 * Convert integer data to host machine's Endian format
	 */
	dataFile->hdr->dataFormat = ntohl(dataFile->hdr->dataFormat);
	dataFile->hdr->samplingRate = ntohl(dataFile->hdr->samplingRate);
	dataFile->hdr->channelCount = ntohl(dataFile->hdr->channelCount);

	dataFile->sampleSize = Tcm_AuComputeSampleSize(dataFile->hdr->dataFormat);
	dataFile->alignSize = dataFile->hdr->channelCount * dataFile->sampleSize;
	dataFile->dataRate = dataFile->alignSize * dataFile->hdr->samplingRate;

	dataFile->numFrames = dataFile->mediaHdr->num_frames;
	dataFile->clipLength = ((double) dataFile->mediaHdr->length_s) +
	  (((double) dataFile->mediaHdr->length_ns) / 1000000000.0);

	/*
	 * Determine the starting point of the audio data. 
	 */

	if (Chunk_FindType(dataFile->cfState, CM_DATA_SEGMENT, &ch) != 0) {
	    Chunk_ReadRewind(dataFile->cfState);
	    if (Chunk_FindType(dataFile->cfState, CM_DATA_SEGMENT, &ch) != 0) {
		goto error;
	    }
	}

	dataFile->dataStart = lseek(dataFile->cfState->fd, 0, SEEK_CUR);


	/*
	 * Construct FTT
	 */
	dataFile->FTT = (double *) ckalloc((dataFile->numFrames+1)*sizeof(double));

	Chunk_ReadRewind(dataFile->cfState);
	if (Chunk_FindType(dataFile->cfState, CM_FTT, &ch) != 0) {
	    goto error;
	}

	fileFTT = (int *) ckalloc((dataFile->numFrames+1)*2*sizeof(int));
	read(dataFile->cfState->fd, (char *) fileFTT, (dataFile->numFrames+1)*2*sizeof(int));

	for (i=0; i<dataFile->numFrames+1; i++) {
	    
	    sec = (double) (ntohl(fileFTT[i*2]));
	    nsec = (double) (ntohl(fileFTT[(i*2)+1]));

	    dataFile->FTT[i] = sec + (nsec / ONE_BILLION);
	}
	ckfree((char *) fileFTT);

	dataFile->FTT[dataFile->numFrames] = dataFile->clipLength;

	/*
	 * Construct FOT
	 */
	dataFile->FOT = (int *) ckalloc(dataFile->numFrames*2*sizeof(int));

	Chunk_ReadRewind(dataFile->cfState);
	if (Chunk_FindType(dataFile->cfState, CM_FOT, &ch) != 0) {
	    goto error;
	}
	
	read(dataFile->cfState->fd, (char *) dataFile->FOT, dataFile->numFrames*2*sizeof(int));
	
	for (i=0; i<dataFile->numFrames; i++) {
	    dataFile->FOT[i*2] = ntohl(dataFile->FOT[i*2]);
	    dataFile->FOT[(i*2)+1] = ntohl(dataFile->FOT[(i*2)+1]);
	}	
    }	

    if (dataFile->raw_flag) {
	close(dataFile->fd);
	dataFile->fd = -1;
    } else {
	Chunk_CloseFile(dataFile->cfState);
	dataFile->cfState = NULL;
    }

    return TCL_OK;

  error:
    
    if (dataFile->fd > 0) {
	close(dataFile->fd);
	dataFile->fd = -1;
    }

    if (dataFile->cfState != NULL) {
	Chunk_CloseFile(dataFile->cfState);
	dataFile->cfState = NULL;
    }
    if (dataFile->hdr != NULL) {
	ckfree((char *) dataFile->hdr);
	dataFile->hdr = NULL;
    }
    if (dataFile->mediaHdr != NULL) {
	ckfree((char *) dataFile->mediaHdr);
	dataFile->mediaHdr = NULL;
    }
    if (dataFile->rawHdr != NULL) {
	ckfree((char *) dataFile->rawHdr);
	dataFile->rawHdr = NULL;
    }
    if (dataFile->FTT != NULL) {
	ckfree((char *) dataFile->FTT);
	dataFile->FTT = NULL;
    }
    if (dataFile->FOT != NULL) {
	ckfree((char *) dataFile->FOT);
	dataFile->FOT = NULL;
    }

    return TCL_ERROR;
}

/*************************************************************************
 * Au_FIP_NumFrames 
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *     start - start time of region in question.
 *     end - end time of region in question.
 *
 * Results:
 *     Number of 'frames' in region of time specified by start and end 
 *     parameters.
 * 
 * Actions:
 *     Queries underly data file structure for number of frames in
 *     region specified and returns the value.
 *
 *************************************************************************/

int 
Au_FIP_NumFrames(AuFileInterface *fip, double start, double end)
{
    AuDataFile *dataFile;
    int start_frame, end_frame, i;

    if (fip != NULL) {
	dataFile = fip->data_file;
	
	if (dataFile != NULL) {

	    if (dataFile->FTT != NULL) {

		start_frame = -1;
		end_frame = -1;


		for (i=0; i<dataFile->numFrames; i++) {
		    if ((dataFile->FTT[i] <= start) && (dataFile->FTT[i+1] > start)) {
			start_frame = i;
			if (end_frame != -1) break;
		    }
		    if ((dataFile->FTT[i] < end) && (dataFile->FTT[i+1] >= end)) {
			end_frame = i;
			if (start_frame != -1) break;
		    }
		}
		if ((start_frame == -1) || (end_frame == -1)) {
		    return 0;
		}
		
		return (end_frame - start_frame + 1);
	    }
	}
    }
    
    return 0;
}

/*************************************************************************
 * Au_FIP_ChannelCount 
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *
 * Results:
 *     Number of channels in the audio file.
 * 
 * Actions:
 *     Returns number of channels defined by the file.
 *
 *************************************************************************/

int 
Au_FIP_ChannelCount(AuFileInterface *fip)
{
    AuDataFile *dataFile;
    
    if (fip != NULL) {
	dataFile = fip->data_file;
	
	if (dataFile != NULL) {
	    if (dataFile->hdr != NULL) {
		return dataFile->hdr->channelCount;
	    }
	}
    }
    return 0;
}

/*************************************************************************
 * Au_FIP_DataFormat
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *
 * Results:
 *     Integer data format code for data in the audio file.
 * 
 * Actions:
 *     Returns data format code for data in the file.
 *
 *************************************************************************/

int 
Au_FIP_DataFormat(AuFileInterface *fip)
{
    AuDataFile *dataFile;
    
    if (fip != NULL) {
	dataFile = fip->data_file;
	
	if (dataFile != NULL) {
	    if (dataFile->hdr != NULL) {
		return dataFile->hdr->dataFormat;
	    }
	}
    }

    return 0;
}

/*************************************************************************
 * Au_FIP_ClipLength
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *
 * Results:
 *     Length of audio clip in seconds. 
 * 
 * Actions:
 *     Returns length of audio clip in seconds.
 *
 *************************************************************************/

double 
Au_FIP_ClipLength(AuFileInterface *fip)
{
    AuDataFile *dataFile;
    
    if (fip != NULL) {
	dataFile = fip->data_file;
	
	if (dataFile != NULL) {
	    return dataFile->clipLength;
	}
    }
    
    return 0.0;
}

/*************************************************************************
 * Au_FIP_SamplingRate
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *
 * Results:
 *     Sampling rate of data in the audio file.
 * 
 * Actions:
 *     Returns sampling rate defined by data in the file.
 *
 *************************************************************************/

int 
Au_FIP_SamplingRate(AuFileInterface *fip)
{
    AuDataFile *dataFile;
    
    if (fip != NULL) {
	dataFile = fip->data_file;
	
	if (dataFile != NULL) {
	    if (dataFile->hdr != NULL) {
		return dataFile->hdr->samplingRate;
	    }
	}
    }

    return 0;
}

/*************************************************************************
 * Au_FIP_TimeFromFrame
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *     framenum - frame number in question.
 *     start - pointer to double used to return start time.
 *     end - pointer to double used to return end time.
 *
 * Results:
 *     None.
 * 
 * Actions:
 *     Start and end time for frame number passed in is returned in pointers
 *     also passed in. Times are clip times (i.e. beginning of file = 0.0).
 *
 *************************************************************************/

void
Au_FIP_TimeFromFrame(AuFileInterface *fip, int framenum, double *start, double *end) 
{
    AuDataFile *dataFile;

    if (fip == NULL)
      return;

    dataFile = fip->data_file;

    if (dataFile == NULL)
      return;

    if (dataFile->FTT == NULL)
      return;

    if (framenum < 0) framenum = 0;
    if (framenum >= dataFile->numFrames) framenum = dataFile->numFrames-1;

    *start = dataFile->FTT[framenum];
    *end = dataFile->FTT[framenum+1];

    return;
}

/*************************************************************************
 * Au_FIP_FrameListFromTime
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *     start - double indicating start time of interval in question.
 *     end - double indicating end time of interval in question.
 *     framelistPtr - pointer to integer pointer. Used to return list
 *                    of frames in the defined interval.
 *
 * Results:
 *     Number of frames in framelist constructed a passed back through pointer.
 * 
 * Actions:
 *     Constructs a list of integer frame numbers for frames included either
 *     wholly or partially by interval defined by start and end parameters.
 *     Pointer to this integer list is returned through framelistPtr parameter
 *     and length of list returned as result of procedure.
 *
 *************************************************************************/


int
Au_FIP_FrameListFromTime(AuFileInterface *fip, double start, double end, 
			 int **framelistPtr)
{
    int *framelist;
    AuDataFile *dataFile;
    int start_frame, end_frame;
    int index, cntr;

    *framelistPtr = NULL;

    if (fip == NULL) return 0;
    
    dataFile = fip->data_file;

    if (dataFile == NULL) return 0;

    if (start < 0.0) start = 0.0;
    if (end > dataFile->clipLength) end = dataFile->clipLength;

    start_frame = Au_FIP_FrameFromTime(fip, start);
    end_frame = Au_FIP_FrameFromTime(fip, end);

    if (start_frame == end_frame) {
	*framelistPtr = (int *) ckalloc(sizeof(int));
	(*framelistPtr)[0] = start_frame;
	return 1;
    }
    
    if (end_frame < start_frame) {
	*framelistPtr = (int *) ckalloc(sizeof(int)*(start_frame-end_frame));
	for (index=0,cntr = start_frame-1; cntr >= end_frame; cntr--,index++) {
	    (*framelistPtr)[index] = cntr;
	}
	return index;
    } else {
	*framelistPtr = (int *) ckalloc(sizeof(int*)*(end_frame-start_frame));
	for (index=0,cntr=start_frame; cntr < end_frame; cntr++,index++) {
	    (*framelistPtr)[index] = cntr;
	}
	return index;
    }
}

/*************************************************************************
 * Au_FIP_FrameFromTime
 * 
 * Parameters:
 *     fip - pointer to AuFileInterface structure.
 *     time - double value indicating time in question.
 *
 * Results:
 *     Integer frame number.
 * 
 * Actions:
 *     Returns frame number of frame that is active during time specified.
 *
 *************************************************************************/

int
Au_FIP_FrameFromTime(AuFileInterface *fip, double time) {

    AuDataFile *dataFile;
    int low_frame, high_frame, mid_frame;

    if (fip == NULL) 
      return 0;

    dataFile = fip->data_file;

    if (dataFile == NULL)
      return 0;

    high_frame = dataFile->numFrames;
    low_frame = 0;

    if (time <= dataFile->FTT[low_frame]) {
	return low_frame;
    }

    if (time >= dataFile->FTT[high_frame]) {
	return high_frame-1;
    }

    while (high_frame-low_frame > 1) {
	mid_frame = (high_frame+low_frame) / 2;
	
	if (time < dataFile->FTT[mid_frame]) {
	    high_frame = mid_frame;
	} else if (time > dataFile->FTT[mid_frame]) {
	    low_frame = mid_frame;
	} else {
	    return mid_frame;
	}
    }
    return low_frame;
}

/*************************************************************************
 * Au_FIP_ReadAndResample
 * 
 * Parameters:
 *     interp - pointer to Tcl interpreter.
 *     fip - pointer to AuFileInterface structure.
 *     frameNum - frame number that holds data in question.
 *     start - start time of data in question.
 *     end - end time of data in question.
 *     delta - length of time during which data is to be played.
 *     speed - speed at which data is to be played.
 *     dataLengthPtr - pointer to length of data in returned buffer.
 *
 * Results:
 *     Pointer to buffer structure containing resampled data read from
 *     file.
 * 
 * Actions:
 *     Reads data from file being accessed through AuFileInterface pointer
 *     corresponding to start and end times. Resamples the data according
 *     to delta and speed. Creates a buffer structure through buffer manager
 *     and formats data with all the proper headers and fills the buffer.
 *     Returns the buffer as the result and returns the length of data in
 *     the buffer through pointer passed as parameter.
 *
 *************************************************************************/

Buffer *
Au_FIP_ReadAndResample(Tcl_Interp *interp, AuFileInterface *fip,
		       int frameNum, double start, double end, 
		       double delta, double speed, 
		       int *dataLengthPtr)
{
    double ratio;
    int i, j, idx;
    int fd;
    int start_offset, end_offset;
    int norm_data_length, norm_num_samples;
    int data_length, num_samples;
    int chunk_type, chunk_length;
    Buffer *bufPtr;
    int norm_sample;
    AuDataFile *dataFile;
    int reverse;
    char *tmp_array;

    dataFile = fip->data_file;

    start_offset = Au_FIP_CalculateOffset(dataFile, frameNum, start);
    end_offset = Au_FIP_CalculateOffset(dataFile, frameNum, end);

    if (end_offset > start_offset) {
	norm_data_length = end_offset - start_offset;
    } else {
	norm_data_length = start_offset - end_offset;
    }

    data_length = dataFile->sampleSize * ((dataFile->hdr->samplingRate * delta / speed) + 0.5);

    norm_num_samples = norm_data_length / dataFile->sampleSize;
    num_samples = data_length / dataFile->sampleSize;
    
    if (data_length > norm_data_length) {
	bufPtr = MakeBuffer(interp, data_length + 8, "auFileInterface:ReadAndResample");
    } else  {
	bufPtr = MakeBuffer(interp, norm_data_length + 8, "auFileInterface:ReadAndResample");
    }

    /* Set return data length pointer to length of data + 8 for chunk type and length. */

    *dataLengthPtr = data_length+8;

    if (dataFile->raw_flag) {
	fd = dataFile->fd;
    } else {
	fd = dataFile->cfState->fd;
    }

    chunk_type = htonl(CM_DATA_SEGMENT);
    chunk_length = htonl(data_length);

    idx = 0;
    memcpy(bufPtr->data, &chunk_type, 4);
    idx += 4;
    memcpy(bufPtr->data+idx, &chunk_length, 4);
    idx += 4;

    if (start_offset < end_offset) {
	lseek(fd, start_offset+dataFile->dataStart, SEEK_SET);
	read(fd, bufPtr->data+idx, norm_data_length);
	reverse = 1;
    } else {
	lseek(fd, end_offset+dataFile->dataStart, SEEK_SET);
	read(fd, bufPtr->data+idx, norm_data_length);
	reverse = -1;
    }

    ratio = (double) norm_num_samples / (double) num_samples;

    if ((ratio == 1.0) && (reverse == 1)) {
	return bufPtr;
    }

    if (norm_data_length > data_length) {
	if (reverse == 1) {
	    char *data_src;
	    char *data_dest;

	    data_dest = bufPtr->data;
	    data_src = &(bufPtr->data[8]);
	    if (dataFile->sampleSize == 1) {
		for (i=0; i<num_samples; i++) {
		    norm_sample = (int) (ratio * i);
		    data_dest[idx++] = data_src[norm_sample];
		}
	    } else {		
		for (i=0; i<num_samples; i++) {
		    norm_sample = (int) (ratio * i);
		    data_src = &(bufPtr->data[8+(norm_sample*dataFile->sampleSize)]);
		    for (j=0; j<dataFile->sampleSize; j++) {
			data_dest[idx++] = data_src[j];
		    }
		}
	    }
	} else {
	    char *data_src;

	    tmp_array = (char *) ckalloc(data_length);
	    idx = 0;

	    if (dataFile->sampleSize == 1) {
		data_src = &(bufPtr->data[8]);
		for (i=0; i<num_samples; i++) {
		    norm_sample = norm_num_samples - ((int) (ratio * i));
		    tmp_array[idx++] = data_src[norm_sample];
		}
	    } else {
		for (i=0; i<num_samples; i++) {
		    norm_sample = norm_num_samples - ((int) (ratio * i));
		    data_src = &(bufPtr->data[8+(norm_sample*dataFile->sampleSize)]);
		    for (j=0; j<dataFile->sampleSize; j++) {
			tmp_array[idx++] = data_src[j];
		    }
		}
	    }
	    memcpy(bufPtr->data+8, tmp_array, data_length);
	    ckfree(tmp_array);
	}
    } else {
	if (reverse == 1) {
	    char *data_src;
	    char *data_dest;
	    
	    idx = data_length + 8 - 1;
	    
	    data_dest = bufPtr->data;
	    if (dataFile->sampleSize == 1) {
		data_src = &(bufPtr->data[8]);
		for (i=0; i<num_samples; i++) {
		    norm_sample = norm_num_samples - ((int) (ratio * i));
		    data_dest[idx--] = data_src[norm_sample];
		}
	    } else {
		for (i=0; i<num_samples; i++) {
		    norm_sample = norm_num_samples - ((int) (ratio * i));
		    data_src = &(bufPtr->data[8+(norm_sample*dataFile->sampleSize)]);
		    for (j=dataFile->sampleSize-1; j>=0; j--) {
			data_dest[idx--] = data_src[j];
		    }
		}
	    }
	} else {
	    char *data_src;

	    tmp_array = (char *) ckalloc(data_length);
	    idx = 0;

	    if (dataFile->sampleSize == 1) {
		data_src = &(bufPtr->data[8]);
		for (i=0; i<num_samples; i++) {
		    norm_sample = norm_num_samples - ((int) (ratio * i));
		    tmp_array[idx++] = data_src[norm_sample];
		}
	    } else {
		for (i=0; i<num_samples; i++) {
		    norm_sample = norm_num_samples - ((int) (ratio * i));
		    data_src = &(bufPtr->data[8+(norm_sample*dataFile->sampleSize)]);
		    for (j=0; j<dataFile->sampleSize; j++) {
			tmp_array[idx++] = data_src[j];
		    }
		}
	    }
	    memcpy(bufPtr->data+8, tmp_array, data_length);
	    ckfree(tmp_array);
	}
    }

    return bufPtr;
}

static int
Au_FIP_CalculateOffset(AuDataFile *dataFile, int frameNum, double time)
{
    int frame_num_samples;
    int sample_num;
    double frame_start, frame_end;

    if (frameNum < 0) {
	frameNum = 0;
    }

    if (frameNum >= dataFile->numFrames) {
	frameNum = dataFile->numFrames-1;
    }

    frame_num_samples = dataFile->FOT[(frameNum*2)+1] / dataFile->sampleSize;
    
    frame_start = dataFile->FTT[frameNum];
    frame_end = dataFile->FTT[frameNum+1];

    sample_num = frame_num_samples * ((time - frame_start) / (frame_end - frame_start));

    if (sample_num < 0) {
	sample_num = 0;
    }

    if (sample_num > frame_num_samples) {
	sample_num = frame_num_samples;
    }

    return (sample_num * dataFile->sampleSize);
}

