/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmAuFileSegment.c,v 1.4 1996/02/06 22:40:02 kpatel Exp $
 *
 *  This module provides for the reading of audio "clipfiles".
 *  
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include <string.h>
#include "cmAuUtil.h"
#include "cmAtom.h"
#include "cmAuFileInterface.h"
#include "cmCookies.h"
#include "cmBufferlist.h"
#include "cmLTS.h"

#define ONE_BILLION    (1000000000.0)
#define DEF_CYCLE_TIME (1.0)
#define DEF_SEND_AHEAD (0.2)

/*
 * The following structure is used when frames are actually read from
 * disk.  An array of the structures is created, one for each frame in
 * this round, and the values are filled in.
 */
typedef struct {
    double sTime;		/* Logical start time for this piece */
    double eTime;		/* Logical end time for this piece */
    Buffer *bufPtr;		/* Pointer to buffer with data */
    Buffer *hdrBufPtr;          /* Pointer to buffer with header. */
    int index;			/* Frame index */
    int sendFlag;
} FramePick;

/*
 * On of the following structures is created for each au segment
object.
 */

typedef struct AuFileSegment {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */

    double sendAhead;		/* Delay for sending ahead data */
    double cycleTime;		/* Cycle Period */

    ClientData lts;		/* LTS for this clipfile */
    char *ltsName;		/* Its tcl name */
    int ltsTriggerNum;		/* LTS trigger for changed lts. */

    InFunc outCCmd;		/* Command called to pass along data */
    ClientData outCD;		/* Object data is sent to */
    char *outCmd;		/* Command used for sending out data */

    int ready;			/* Ready flag */

    char *filename;		/* Full pathname for clipfile */
    AuFileInterface *fip;       /* File inteface pointer. */

    int bandwidth;

    int atToken;		/* Token for at event. */
    double atWhen;		/* When that event was scheduled */
    
    int reverse;                /* Set to -1 if clipStart > clipEnd. 1 otherwise. */

    double clipStart;           /* Start of segment relative to clip beginning. */
    double clipEnd;             /* End of segment relative to clip beginning. */
    double logicalStart;        /* When to begin segment in logical time. */
    double logicalEnd;          /* When to end segment in logical time. */

} AuFileSegment;

/*
 * Globals used by this object module.
 */
static int nextAuFileSegmentNum = 0;/* Variable for next unique object name. */

#ifndef _WINDOWS
#    define max(a,b)      ((a)>(b)?(a):(b))
#    define min(a,b)      ((a)<(b)?(a):(b))
#endif
#define rnd(a)        (floor(a + .5))

#if 0
int AudioLogFD;
static int AudioLog ( ScatterBuffer *sb);
#endif /* debug */

/*
 * Time conversion macros.
 */

#define LOG_TIME_FROM_CLIP_TIME(objPtr,time)	\
    ((((time - objPtr->clipStart) *		\
       (objPtr->logicalEnd -            	\
	objPtr->logicalStart)) /             	\
      (objPtr->clipEnd - objPtr->clipStart)) 	\
     + objPtr->logicalStart)

#define CLIP_TIME_FROM_LOG_TIME(objPtr,time)	\
    ((((time - objPtr->logicalStart) *		\
       (objPtr->clipEnd -			\
	objPtr->clipStart)) /			\
      (objPtr->logicalEnd -			\
       objPtr->logicalStart))			\
     + objPtr->clipStart)


/*
 * Forward Declarations.
 */
static AuFileSegment *CreateNewAuFileSegment(Tcl_Interp * interp, char *name);
static int DestroyAuFileSegment(AuFileSegment *objPtr);
static int ConfigureAuFileSegment(AuFileSegment *objPtr, int argc, char **argv);
static int AuFileSegmentObjCmd (ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
static int AuFileSegmentInfo(AuFileSegment *objPtr, double start, double end);
static int AuFileSegmentReady(AuFileSegment *objPtr);
static int AuFileSegmentUnready(AuFileSegment *objPtr);
static AuFileInterface *AuFileSegmentOpenClipfile (AuFileSegment *objPtr, char *filename );
static int AuFileSegmentChangeOutCmd(AuFileSegment *objPtr, char *newOut);
static void AuFileSegmentLTSChanged(ClientData clientdata);
static void AuFileSegmentSendData(ClientData clientData) ;


/************************************************************
 *
 * CreateNewAuFileSegment
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static AuFileSegment *
CreateNewAuFileSegment(Tcl_Interp * interp, char *name)
{
    AuFileSegment *new;

    new = (AuFileSegment *) ckalloc(sizeof(AuFileSegment));
 
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);
    new->sendAhead = DEF_SEND_AHEAD;
    new->cycleTime = DEF_CYCLE_TIME;
    new->lts = NULL;
    new->ltsName = NULL;
    new->ltsTriggerNum = -1;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->outCmd = NULL;
    new->ready = 0;
    new->filename = NULL;
    new->fip = NULL;
    new->bandwidth = 0.0;
    new->atToken = -1;
    new->atWhen = HUGE_VAL;
    new->clipStart = 0.0;
    new->clipEnd = 0.0;
    new->logicalStart = 0.0;
    new->logicalEnd = 0.0;
    
#if 0
    AudioLogFD = creat("./audioLog",0666);
#endif /* debug */
    return new;
}

/************************************************************
 *
 * DestroyAuFileSegment
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/
static int
DestroyAuFileSegment(AuFileSegment *objPtr)
{
    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->atToken != -1) {
	Tcm_DeleteAtHandler(objPtr->atToken);
    }
    objPtr->atToken = -1;
    objPtr->atWhen = HUGE_VAL;

    if (objPtr->name != NULL) {
	ckfree((char *) objPtr->name);
	objPtr->name = NULL;
    }

    if (objPtr->filename != NULL) {
	ckfree((char *) objPtr->filename);
	objPtr->filename = NULL;
    }

    if (objPtr->fip != NULL) {
	Au_FIP_Close(objPtr->fip);
	objPtr->fip = NULL;
    }

    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
	objPtr->outCmd = NULL;
    }

    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    if (objPtr->ltsName != NULL) {
	ckfree((char *) objPtr->ltsName);
	objPtr->ltsName = NULL;
    }
    
#if 0
    close(AudioLogFD);
#endif /* debug */

    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureAuFileSegment
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

static int
ConfigureAuFileSegment(AuFileSegment *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];

    interp = objPtr->interp;

    /*
     * Case 1: "<object> configure"
     */
    if (argc == 0) {
	if (objPtr->filename != NULL) {
	    Tcl_AppendResult(interp, "{-filename {} {", objPtr->filename,
			     "}} ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-filename {} {}} ", NULL);
	}

	Tcl_PrintDouble(interp, objPtr->bandwidth, str);
	Tcl_AppendResult(interp, "{-bandwidth {0.0} ", str, "} ", NULL);

	if (objPtr->lts != NULL) {
	    Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName, "}} ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {} {}} ", NULL);
	}

	Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	Tcl_AppendResult(interp, "{-sendAhead {0.2} ", str, "} ", NULL);

	if (objPtr->outCmd != NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", NULL);
	}

	Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	Tcl_AppendResult(interp, "{-cycleTime {1.0} ", str, "} ", NULL);
	
	Tcl_PrintDouble(interp, objPtr->clipStart, str);
	Tcl_AppendResult(interp, "{-clipStart 0.0 ", str, "} ", NULL);

	Tcl_PrintDouble(interp, objPtr->clipEnd, str);
	Tcl_AppendResult(interp, "{-clipEnd 0.0 ", str, "} ", NULL);

	Tcl_PrintDouble(interp, objPtr->logicalStart, str);
	Tcl_AppendResult(interp, "{-logicalStart 0.0 ", str, "} ", NULL);

	Tcl_PrintDouble(interp, objPtr->logicalEnd, str);
	Tcl_AppendResult(interp, "{-logicalEnd 0.0 ", str, "} ", NULL);
	return TCL_OK;

    }
    /*
     * Case 2: "<object> configure -<slot>"
     */
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 'f') && (strncmp("-filename", argv[0], len) == 0)) {
	    if (objPtr->filename != NULL) {
		Tcl_AppendResult(interp, "-filename {} {", objPtr->filename,
				 "}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-filename {} {}", NULL);
	    }
	    return TCL_OK;

	} else if ((c == 'b') && (strncmp("-bandwidth", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->bandwidth, str);
	    Tcl_AppendResult(interp, "-bandwidth {0.0} ", str, NULL);
	    return TCL_OK;

	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->lts != NULL) {
		Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName, "}",
				 NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {} {}", NULL);
	    }
	    return TCL_OK;

	} else if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	    Tcl_AppendResult(interp, "-sendAhead {0.2} ", str, NULL);
	    return TCL_OK;

	} else if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd != NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
				 "}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	    }
	    return TCL_OK;

	} else if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
	    char str[TCL_DOUBLE_SPACE];

	    Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	    Tcl_AppendResult(interp, "-cycleTime {1.0} ", str, NULL);
	    return TCL_OK;
	} else if ((c == 'c') && (strncmp("-clipStart", argv[0], len) == 0)) {
	    char str[TCL_DOUBLE_SPACE];

	    Tcl_PrintDouble(interp, objPtr->clipStart, str);
	    Tcl_AppendResult(interp, "-clipStart 0.0 ", str, NULL);
	    return TCL_OK;
	} else if ((c == 'c') && (strncmp("-clipEnd", argv[0], len) == 0)) {
	    char str[TCL_DOUBLE_SPACE];

	    Tcl_PrintDouble(interp, objPtr->clipEnd, str);
	    Tcl_AppendResult(interp, "-clipEnd 0.0 ", str, NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-logicalStart", argv[0], len) == 0)) {
	    char str[TCL_DOUBLE_SPACE];

	    Tcl_PrintDouble(interp, objPtr->logicalStart, str);
	    Tcl_AppendResult(interp, "-logicalStart 0.0 ", str, NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-logicalEnd", argv[0], len) == 0)) {
	    char str[TCL_DOUBLE_SPACE];

	    Tcl_PrintDouble(interp, objPtr->logicalEnd, str);
	    Tcl_AppendResult(interp, "-logicalEnd 0.0 ", str, NULL);
	    return TCL_OK;
	}
	
	Tcl_AppendResult(interp, "bad configuration option: ", argv[0], NULL);
	return TCL_ERROR;
    }
    /*
     * Case 3: "<object> configure -<slot>"
     */
    cntr = 0;
    AuFileSegmentUnready(objPtr);

    while (cntr < argc) {
	char c;
	int len;

	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	/* ------------------------ FILENAME ------------------------ */
	if ((c == 'f') && (strncmp("-filename", argv[cntr], len) == 0)) {
	    char *fname;
	    Tcl_DString fname_dstring;

	    cntr++;
	    if (cntr == argc) {
	      usageError:
		Tcl_AppendResult(interp,
				 "Must specify value with configuration option ",
				 argv[cntr - 1], NULL);
		return TCL_ERROR;
	    }
	    if (objPtr->filename != NULL) {
		Tcl_AppendResult(interp, "Cannot change clipfile name.\n",
				 "Create a new object and/or delete this one.",
				 NULL);
		return TCL_ERROR;
	    }
	    
	    fname = Tcl_TildeSubst(objPtr->interp, argv[cntr], &fname_dstring);
	    objPtr->fip = AuFileSegmentOpenClipfile(objPtr, fname);

	    if (objPtr->fip == NULL) {
		Tcl_DStringFree(&fname_dstring);
		Tcl_AppendResult(interp, "Couldn't open file \"",
				 fname, "\" as au file", 
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    objPtr->filename = (char *) ckalloc(strlen(fname)+1);
	    strcpy(objPtr->filename, fname);
	    Tcl_DStringFree(&fname_dstring);

	  /* ------------------------ BANDWIDTH ------------------------ */
	} else if ((c == 'b') &&
		   (strncmp("-bandwidth", argv[cntr], len) == 0)) {
	  double newBandwidth;

	  cntr++;
	  if (cntr == argc) {
	    goto usageError;
	  }
	  if (Tcl_GetDouble(interp, argv[cntr], &newBandwidth) != TCL_OK) {
	  badValueError:
	    Tcl_AppendResult(interp, "bad value \"", argv[cntr],
			     "\" given for configuration option ",
			     argv[cntr - 1], NULL);
	    return TCL_ERROR;
	  }
	  objPtr->bandwidth = newBandwidth;

	  /* ------------------------ LTS ------------------------ */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts = NULL;
	    
	    cntr++;
	    if (cntr == argc) {
		goto usageError;
	    }
	    if ((strlen(argv[cntr]) != 0) &&
		(newLts = Tcm_FindAtom(interp, argv[cntr])) == NULL) {
		
		goto badValueError;
	    }
	    
	    /*
	     * Release our lock and trigger on the old lts.
	     */
	    if (objPtr->lts != NULL) {
		if (objPtr->ltsTriggerNum != -1) {
		    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
		    objPtr->ltsTriggerNum = -1;
		}
		Tk_Release(objPtr->lts);
		objPtr->lts = NULL;
	    }
	    if (objPtr->ltsName != NULL) {
		ckfree((char *) objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }
	    
	    if (strlen(argv[cntr]) != 0) {
		
		/*
		 * Set up new lts
		 */
		objPtr->ltsName = ckalloc(strlen(argv[cntr]) + 1);
		strcpy(objPtr->ltsName, argv[cntr]);
		objPtr->lts = newLts;
		Tk_Preserve (objPtr->lts);
		LTS_SetTrigger(objPtr->lts,
			       LTS_SPEED_FLAG | LTS_OFFSET_FLAG | LTS_VALUE_FLAG,
			       AuFileSegmentLTSChanged, (ClientData) objPtr);
	    }
	    
	    /* ------------------------ SEND AHEAD ------------------------ */
	} else if ((c == 's') &&
		   (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	    double newSendAhead;
	    
	    cntr++;
	    if (cntr == argc) {
		goto usageError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->sendAhead = newSendAhead;
	    
	    /* ------------------------ OUT CMD ------------------------ */
	} else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto usageError;
	    }
	    if (objPtr->outCmd != NULL) {
		ckfree((char *) objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	    if (AuFileSegmentChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	    /* ------------------------ CYCLE TIME ------------------------ */
	} else if ((c == 'c') &&
		   (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	    double newCycleTime;
	    
	    cntr++;
	    if (cntr == argc) {
		goto usageError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->cycleTime = newCycleTime;
 	    /* ------------------------ CLIP START ------------------------ */
	} else if ((c == 'c') &&
		   (strncmp("-clipStart", argv[cntr], len) == 0)) {
	    double newClipStart;
	    
	    cntr++;
	    if (cntr == argc) {
		goto usageError;
	    }

	    if (argv[cntr][0] == 'e') {
		if (objPtr->fip != NULL) {
		    objPtr->clipStart = Au_FIP_ClipLength(objPtr->fip);
		}
	    } else if (argv[cntr][0] == '?') {
		objPtr->clipStart = objPtr->clipEnd - (objPtr->logicalEnd - objPtr->logicalStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newClipStart) != TCL_OK) {
		    goto badValueError;
		}
		objPtr->clipStart = newClipStart;
	    }
 	    /* ------------------------ CLIP END ------------------------ */
	} else if ((c == 'c') &&
		   (strncmp("-clipEnd", argv[cntr], len) == 0)) {
	    double newClipEnd;
	    
	    cntr++;
	    if (cntr == argc) {
		goto usageError;
	    }

	    if (argv[cntr][0] == 'e') {
		if (objPtr->fip != NULL) {
		    objPtr->clipEnd = Au_FIP_ClipLength(objPtr->fip);
		}
	    } else if (argv[cntr][0] == '?') {
		objPtr->clipEnd = objPtr->clipStart + (objPtr->logicalEnd - objPtr->logicalStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newClipEnd) != TCL_OK) {
		    goto badValueError;
		}
		objPtr->clipEnd = newClipEnd;
	    }
 	    /* ------------------------ LOGICAL START ------------------------ */
	} else if ((c == 'l') &&
		   (strncmp("-logicalStart", argv[cntr], len) == 0)) {
	    double newLogicalStart;
	    
	    cntr++;
	    if (cntr == argc) {
		goto usageError;
	    }

	    if (argv[cntr][0] == '?') {
		objPtr->logicalStart = objPtr->logicalEnd - (objPtr->clipEnd - objPtr->clipStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newLogicalStart) != TCL_OK) {
		    goto badValueError;
		}
		objPtr->logicalStart = newLogicalStart;
	    }
 	    /* ------------------------ LOGICAL END ------------------------ */
	} else if ((c == 'l') &&
		   (strncmp("-logicalEnd", argv[cntr], len) == 0)) {
	    double newLogicalEnd;
	    
	    cntr++;
	    if (cntr == argc) {
		goto usageError;
	    }

	    if (argv[cntr][0] == '?') {
		objPtr->logicalEnd = objPtr->logicalStart + (objPtr->clipEnd - objPtr->clipStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newLogicalEnd) != TCL_OK) {
		    goto badValueError;
		}
		objPtr->logicalEnd = newLogicalEnd;
	    }
	} else {
	    Tcl_AppendResult(interp, "Bad configuration option: ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }

    return TCL_OK;
}

/************************************************************
 *
 * AuFileSegmentCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the auSegment command.
 *     A new object of type AuFileSegment is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/
int
AuFileSegmentCreateCmd (ClientData clientData, Tcl_Interp *interp,
			int argc, char **argv)
{
    char newName[256];
    AuFileSegment *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", NULL);
	return TCL_ERROR;
    }
    nextAuFileSegmentNum++;
    sprintf(newName, "%sauSeg%d", argv[1], nextAuFileSegmentNum);

    new = CreateNewAuFileSegment(interp, newName);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, newName, AuFileSegmentObjCmd,
		      (ClientData) new, (void (*)())NULL);

    if (argc > 2) {
	ConfigureAuFileSegment(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, newName, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * AuFileSegmentObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     AuFileSegment object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/
static int
AuFileSegmentObjCmd (ClientData clientData, Tcl_Interp *interp,
		  int argc, char **argv)
{
    char c;
    int len, i;
    AuFileSegment *objPtr = (AuFileSegment *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " <command> ?arguments?", NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    /*---------------------------- CONFIGURE ----------------------------*/
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureAuFileSegment(objPtr, argc - 2, argv + 2);

    /*---------------------------- DESTROY ------------------------------*/
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyAuFileSegment(objPtr);

    /*---------------------------- INFO --------------------------------*/
    } else if ((c == 'i') && (strncmp("info", argv[1], len) == 0)) {
	double start_param, end_param;

	start_param = 0.0;
	end_param = 0.0;

	if (argc > 2) {
	    if (Tcl_GetDouble(interp, argv[2], &start_param) != TCL_OK) {
		Tcl_AppendResult(interp, "Bad value: \"", argv[2], "\" given",
				 " as argument to: ", argv[0], " ", argv[1],
				 (char *) NULL);
	    }
	    if (argc > 3) {
		if (Tcl_GetDouble(interp, argv[3], &end_param) != TCL_OK) {
		    Tcl_AppendResult(interp, "Bad value: \"", argv[3], 
				     "\" given as argument to: ", 
				     argv[0], " ", argv[1],
				     (char *) NULL);
		}
	    } else {
		end_param = start_param;
	    }
	} else {
	    start_param = objPtr->clipStart;
	    end_param = objPtr->clipEnd;
	}

        return AuFileSegmentInfo(objPtr, start_param, end_param);
	
	/*---------------------------- READY -------------------------------*/
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
        return AuFileSegmentReady(objPtr);
	
	/*---------------------------- UNREADY -----------------------------*/
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
        return AuFileSegmentUnready(objPtr);
    }
    Tcl_AppendResult(interp, "Bad <command> option: ", argv[0], argv[1],
		     NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  ready\n",
                "  unready\n",
                "  info ?clip_time ?clip_time??\n",
                NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * AuFileSegmentModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
AuFileSegmentModuleInit(Tcl_Interp *interp)
{
    Tcl_CreateCommand (interp, "auSegment", AuFileSegmentCreateCmd,
		       NULL, (void (*)()) NULL);
    return TCL_OK;
}


/************************************************************
 *
 * AuFileSegmentInfo
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Return Tcl String "{numFrames dataFormat
 *                         samplingRate numChannels}"
 *
 ************************************************************/

static int
AuFileSegmentInfo(AuFileSegment *objPtr, double start, double end)
{
    char retStr[256];
    char format[256];

    if (objPtr->filename == NULL) {
	Tcl_AppendResult(objPtr->interp, "Must specify a clipfile first",
			 NULL);
	return TCL_ERROR;
    }
    switch (Au_FIP_DataFormat(objPtr->fip)) {
    case SUN_ULAW:
	strcpy(format, "Sun 8bit Ulaw");
	break;
    case SUN_LIN_8:
	strcpy(format, "Sun 8bit Linear");
	break;
    case SUN_LIN_16:
	strcpy(format, "Sun 16bit Linear");
	break;
    default:
	strcpy(format, "Unkown format");
	break;
    }

    sprintf(retStr, "%d %g {%s} %d %d",
	    Au_FIP_NumFrames(objPtr->fip, start, end), 
	    Au_FIP_ClipLength(objPtr->fip),
	    format,
	    Au_FIP_SamplingRate(objPtr->fip),
	    Au_FIP_ChannelCount(objPtr->fip));
	  
    Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);

    return TCL_OK;
}

/************************************************************
 *
 * AuFileSegmentReady
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
AuFileSegmentReady(AuFileSegment *objPtr)
{
    double clip_length;

    if (objPtr->ready) return TCL_OK;

    if (objPtr->atToken != -1) {
	Tcm_DeleteAtHandler(objPtr->atToken);
    }
    objPtr->atToken = -1;
    objPtr->atWhen = HUGE_VAL;

    if (objPtr->fip == NULL) {
	Tcl_AppendResult(objPtr->interp, "Must specify a filename before readying this object", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Check for clip start, end and logical start, end consistency. 
     */

    if (objPtr->logicalEnd - objPtr->logicalStart <= 0.0) {
	Tcl_AppendResult(objPtr->interp, "Logical end time must be greater than ",
			 "logical start time.", (char *) NULL);
	return TCL_ERROR;
    }

    if (objPtr->clipStart < 0.0) {
	objPtr->clipStart = 0.0;
    }
    if (objPtr->clipEnd < 0.0) {
	objPtr->clipEnd = 0.0;
    }
    
    clip_length = Au_FIP_ClipLength(objPtr->fip);

    if (objPtr->clipStart > clip_length) {
	objPtr->clipStart = clip_length;
    }
    if (objPtr->clipEnd > clip_length) {
	objPtr->clipEnd = clip_length;
    }

    if (objPtr->clipEnd - objPtr->clipStart == 0.0) {
	Tcl_AppendResult(objPtr->interp, "Clip start and end times specified create ",
			 "empty segment.", (char *) NULL);
	return TCL_ERROR;
    }

    if (objPtr->clipStart > objPtr->clipEnd) {
	objPtr->reverse = -1;
    } else {
	objPtr->reverse = 1;
    }

    if (objPtr->lts == NULL) {
	Tcl_AppendResult(objPtr->interp, "Must specify lts for auSegment object ",
			 objPtr->name, NULL), " before doing ready";
	return TCL_ERROR;
    }

    objPtr->ready = 1;

    if (objPtr->lts != NULL) {
	AuFileSegmentSendData((ClientData) objPtr);
    }
    return TCL_OK;
}

/************************************************************
 *
 * AuFileSegmentUnready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
AuFileSegmentUnready(AuFileSegment *objPtr)
{
    if (objPtr->atToken != -1) {
	Tcm_DeleteAtHandler(objPtr->atToken);
    }
    objPtr->atToken = -1;
    objPtr->atWhen = HUGE_VAL;

    objPtr->ready = 0;
    return TCL_OK;
}

/************************************************************
 *
 * AuFileSegmentOpenClipfile
 *
 * Parameters:
 *     objPtr - pointer to object
 *     filename - name of file to open.
 *
 * Return:
 *     Returns a AuFileInterface po9inter valid for filename on
 *     success, NULL on failure.
 *
 * Action:
 *     Intefaces with Au File Interface module to get access to file.
 *
 ************************************************************/

static AuFileInterface *
AuFileSegmentOpenClipfile (AuFileSegment *objPtr, char *filename )
{
    AuFileInterface *fip;

    fip = Au_FIP_Open(filename);

    return fip;
}

/************************************************************
 *
 * AuFileSegmentChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *     newOut - new out command.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
AuFileSegmentChangeOutCmd(AuFileSegment *objPtr, char *newOut)
{
    /*
     * Clear the previous entry. 
     */
    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
	objPtr->outCmd = NULL;
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return. 
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set 
     * everything accordingly. 
     */
    if (newOut[0] == '@') {

	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
	       &(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			     "Could not find registered C procedure: ",
			     newOut, (char *) NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string. 
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut) + 1);
    strcpy(objPtr->outCmd, newOut);

    return TCL_OK;
}

/************************************************************
 *
 * AuFileSegmentLTSChanged
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Procedure called whenever the LTS changes.
 *
 ************************************************************/

static void
AuFileSegmentLTSChanged(ClientData clientdata)
{
    AuFileSegment *objPtr = (AuFileSegment *) clientdata;

    if (objPtr->ready == 1) {
	AuFileSegmentUnready(objPtr);
	AuFileSegmentReady(objPtr);
    }
    return;
}


/************************************************************
 *
 * AuFileSegmentSendData
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sends data according to current value of LTS, queues
 *     itself on the at queue for next send time.
 *
 ************************************************************/
static void
AuFileSegmentSendData(ClientData clientData) 
{
    AuFileSegment *objPtr = (AuFileSegment *) clientData;
    double now, speed, when, end;
    double eTime, sTime, lastETime, tmpTime;
    double clip_now, clip_cycle_end;
    int *framelist;
    int i;
    SparcAuBufferHeader bufHdr;
    int idx;
    int dataSegCode, dataLength;
    SB_List *sbList;
    char *sbStr;
    int numToSend, bwToSend;
    int curIndex;
    int sbIdx;
    static int iter = 0;
    static int fPickNum = 0;
    static FramePick *fPickList = NULL;
    FramePick *f;
    int fd;

  retry:

    /*
     * If not ready, return. 
     */
    if (!objPtr->ready)
	goto done;

    /*
     * If no LTS, return. 
     */
    if (objPtr->lts == NULL)
	goto done;

    /*
     * If no out command, return. 
     */
    if (objPtr->outCmd == NULL)
	goto done;

    /*
     * The variable 'now' represents the time when this event
     * was scheduled.
     */
    if (objPtr->atToken != -1) {
	now = objPtr->atWhen;
    } 
    else {
	now = LTS_Value(objPtr->lts);
    }
    
    objPtr->atToken = -1;
    speed = LTS_Speed(objPtr->lts);
    now += objPtr->sendAhead * speed;
    iter++;

    /*
     * If we are not in the segment, calculate when we will be and
     * then create an event in the 'at queue'.
     */
    if (speed > 0.0) {
	if (now < objPtr->logicalStart) {
	    when = LTS_SystemTimeOf(objPtr->lts, (objPtr->logicalStart -
					   (objPtr->sendAhead * speed)));
	    objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					   "audio", AuFileSegmentSendData, objPtr,
					   NULL, NULL, NULL, NULL, NULL);
	    objPtr->atWhen = objPtr->logicalStart - (objPtr->sendAhead * speed);
	    return;
	} else if (now >= objPtr->logicalEnd) {
	    /* Sleep forever case. */
	    Au_FIP_FDClose(objPtr->fip);
	    return;
	}
    } else if (speed < 0.0) {
	if (now > objPtr->logicalEnd) {
	    when = LTS_SystemTimeOf(objPtr->lts, (objPtr->logicalEnd +
					   (objPtr->sendAhead * speed)));
	    objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					   "audio", AuFileSegmentSendData, objPtr,
					   NULL, NULL, NULL, NULL, NULL);
	    objPtr->atWhen = objPtr->logicalEnd - (objPtr->sendAhead * speed);
	    return;
	} else if (now <= objPtr->logicalStart) {
	    /* Sleep forever case. */
	    Au_FIP_FDClose(objPtr->fip);
	    return;
	}

    } else {
	/* Speed is zero, sleep forever. */
	Au_FIP_FDClose(objPtr->fip);
	return;
    }

    /*
     * If file isn't open, open it now. 
     */

    Au_FIP_FDOpen(objPtr->fip);

    /*
     * If speed or cycle time are 0.0, there is nothing to do.  
     */
    if ((objPtr->cycleTime == 0.0) || (speed == 0.0)) {
	goto done;
    }

    /*
     * Find pieces of segments to be sent.  This is done by creating an
     * array of FramePick structures and filling them in.
     * We update the bwToSend and numToSend variables as we go.
     */
    if (fPickNum == 0) {
	fPickNum = 20;
	fPickList = (FramePick *) ckalloc(sizeof(FramePick) * fPickNum);
    }
    numToSend = 0;
    bwToSend = 0;

    /*
     * Calculate time in terms of clip time.
     */

    clip_now = CLIP_TIME_FROM_LOG_TIME(objPtr, now);
    end = now + objPtr->cycleTime*speed;
    clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, end);
    

    /*
     * Get list of frame numbers to be sent from file interface.
     * Note that the data that we want to send may only be a portion
     * of one frame of audio data since the framing of audio data is
     * arbitrary and generally consists of just one frame for the whole
     * file.
     */

    numToSend = Au_FIP_FrameListFromTime(objPtr->fip, clip_now, 
					 clip_cycle_end, &framelist);

    if (numToSend > fPickNum) {
	fPickNum = numToSend+1;
	fPickList = (FramePick *) ckrealloc((char *) fPickList,
					    sizeof(FramePick)*fPickNum);
    }

    for (i=0; i<numToSend; i++) {
	Au_FIP_TimeFromFrame(objPtr->fip, framelist[i], &sTime, &eTime);
	/* 
	 * Flip start and end times if lts running backwards.
	 */
	if (speed < 0) {
	    tmpTime = eTime;
	    eTime = sTime;
	    sTime = tmpTime;
	}

	/*
	 * Flip start and end times if clip is mapped in reverse.
	 */
	if (objPtr->reverse == -1) {
	    tmpTime = eTime;
	    eTime = sTime;
	    sTime = tmpTime;
	}

	if (i == 0) {
	    fPickList[i].sTime = clip_now;
	} else {
	    fPickList[i].sTime = sTime;
	}
	if (i == numToSend-1) {
	    if (objPtr->reverse != -1) {
		if (speed > 0.0) {
		    clip_cycle_end = min(clip_cycle_end, eTime);
		} else {
		    clip_cycle_end = max(clip_cycle_end, eTime);
		}
	    } else {
		if (speed > 0.0) {
		    clip_cycle_end = max(clip_cycle_end, eTime);
		} else {
		    clip_cycle_end = min(clip_cycle_end, eTime);
		}
	    }
	    fPickList[i].eTime = clip_cycle_end;
	} else {
	    fPickList[i].eTime = eTime;
	}
	fPickList[i].index = framelist[i];
	fPickList[i].sendFlag = 1;
	fPickList[i].sTime = LOG_TIME_FROM_CLIP_TIME(objPtr, fPickList[i].sTime);
	fPickList[i].eTime = LOG_TIME_FROM_CLIP_TIME(objPtr, fPickList[i].eTime);
    }
    ckfree((char *) framelist);

    /*
     * For each frame chosen, create buffers with headers and data.
     */
    sbList = (SB_List *) ckalloc(sizeof(SB_List));
    sbList->sbs =
      (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *) * numToSend);
    sbList->count = numToSend;
    sbIdx = 0;
    for (i = 0; i < numToSend; i++) {
	if (fPickList[i].sendFlag == 1) {
	    double lsS, lsNs, leS, leNs;
	    
	    sTime = fPickList[i].sTime;
	    eTime = fPickList[i].eTime;
	    
	    /* Note: subtract 8 (two int's) from the size
	     * of the buffer header for the chunkFile support fields.
	     */
	    
	    bufHdr.hdrType = htonl(CM_SPARC_AU_BUF_HDR);
	    bufHdr.device = htonl(SPARC_AU_DEV_CODE);
	    bufHdr.length = htonl(sizeof(SparcAuBufferHeader) - 8);
	    bufHdr.dataFormat = htonl(Au_FIP_DataFormat(objPtr->fip));
	    bufHdr.samplingRate = htonl(Au_FIP_SamplingRate(objPtr->fip));
	    bufHdr.channelCount = htonl(Au_FIP_ChannelCount(objPtr->fip));
	    lsS = floor(sTime);
	    lsNs = (sTime - floor(sTime)) * ONE_BILLION;
	    leS = floor(eTime);
	    leNs = (eTime - floor(eTime)) * ONE_BILLION;
	    bufHdr.logStart_s = htonl((int) lsS);
	    bufHdr.logStart_ns = htonl((int) lsNs);
	    bufHdr.logEnd_s = htonl((int) leS);
	    bufHdr.logEnd_ns = htonl((int) leNs);
	    
	    fPickList[i].hdrBufPtr = MakeBuffer(objPtr->interp,
						sizeof(SparcAuBufferHeader),
						"tcmSparcAuCF:SparcAuSendData");
	    memcpy(fPickList[i].hdrBufPtr->data, &bufHdr, sizeof(SparcAuBufferHeader));
	    
	    
	    fPickList[i].bufPtr =
	      Au_FIP_ReadAndResample(objPtr->interp, objPtr->fip, fPickList[i].index,
				     CLIP_TIME_FROM_LOG_TIME(objPtr, fPickList[i].sTime),
				     CLIP_TIME_FROM_LOG_TIME(objPtr, fPickList[i].eTime),
				     fPickList[i].eTime - fPickList[i].sTime, speed,
				     &dataLength);

	    sbList->sbs[sbIdx] = MakeScatterBuffer(fPickList[i].hdrBufPtr, sizeof(SparcAuBufferHeader),
						   fPickList[i].bufPtr, dataLength, 
						   NULL);
#if 0
	    AudioLog( sbList->sbs[sbIdx] );
#endif /* debug */
	    sbIdx++;
	}
    }

    /*
     * If out command is C, create scatter buffer list and call it.
     */
    if (objPtr->outCCmd != NULL) {

	/*
	 * Need to check if it still exists since object may have been
	 * destroyed.
	 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd + 1) != NULL) {
			objPtr->outCCmd(objPtr->outCD, sbList);
		} else {
			objPtr->outCCmd = NULL;
			ckfree((char *) objPtr->outCmd);
			objPtr->outCmd = NULL;
		}

    } else {
	/*
	 * Out command is tcl, make sb list into a string and call using
	 * Tcl_Eval. 
	 */
	sbStr = MakeSBListString(sbList);
	Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL);
	ckfree((char *) sbStr);
    }
    FreeSBList(sbList);

    /*
     * We're done with the buffers, so release them.
     */
    lastETime = 0;
    for (i = 0; i < numToSend; i++) {
	if (fPickList[i].sendFlag == 1) {
	    lastETime = fPickList[i].eTime;
	    DetachBuffer(objPtr->interp, fPickList[i].hdrBufPtr,
				     "tcmSparcAuCF:SparcAuSendData");
	    DetachBuffer(objPtr->interp, fPickList[i].bufPtr,
				     "tcmSparcAuCF:SparcAuSendData");
	}
    }

    /* 
     * Find next wake up time and set up 'at' event. 
     */
    when = LTS_SystemTimeOf(objPtr->lts, (lastETime - (objPtr->sendAhead * speed)));
    objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL, "audio", AuFileSegmentSendData, 
					  objPtr, NULL, NULL, NULL, NULL, NULL);

    objPtr->atWhen = lastETime - (objPtr->sendAhead * speed);
    return;

done:
    Au_FIP_FDClose(objPtr->fip);
    return;
}




#if 0
static int
AudioLog ( ScatterBuffer *sb)
{
    Buffer *buf;
    char *data;
    int size;

    data = GetFromSB(sb, CM_DATA_SEGMENT, &buf, &size);

    write(AudioLogFD, data, size);
    return TCL_OK;
}
#endif /* debug */


