/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif

#include <math.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "cmAuPlay.h"
#include "cmAuFile.h"
#include "tcmEvent.h"
#include "cmLTS.h"
#include "cmAtom.h"

#define ONE_BILLION 1000000000.0
#define DEF_SENDAHEAD 0.2

static int nextSpeakerRsrcNum = 0;	/* Variable for next unique object
					 * name. */

/*
 * Prototypes for routines used internally
 */
int SpeakerRsrcModuleInit			(Tcl_Interp *interp);
int SpeakerRsrcCreateCmd			(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc, char **argv);
int SpeakerRsrcObjCmd				(ClientData clientData,
						 Tcl_Interp *interp,
		       				 int argc, char **argv);
static SpeakerRsrc *CreateNewSpeakerRsrc	(Tcl_Interp *interp,
						 char *name);
static void ChangeDeviceType			(SpeakerRsrc *objPtr,
						 char *devtype);
static int DestroySpeakerRsrc			(SpeakerRsrc *objPtr);
static int ConfigureSpeakerRsrc			(SpeakerRsrc *objPtr,
						 int argc, char **argv);
static int SpeakerRsrcReady			(SpeakerRsrc *objPtr);
static int SpeakerRsrcUnready			(SpeakerRsrc *objPtr);
static void SpeakerRsrc_LTSChange		(ClientData clientData);
static int SpeakerRsrcAccept			(SpeakerRsrc *objPtr,
						 SB_List *sbl);
static int SpeakerRsrcStats			(SpeakerRsrc *objPtr);
static int SpeakerResourcePlayData		(CBInfo *cbPtr);
static void SpeakerResourceFreeFrame		(CBInfo *cbPtr);
static int SpeakerResourceDeviceList            (SpeakerRsrc *objPtr);

/*
 * -------------------------------------------------------
 *
 * SpeakerRsrcModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 * -------------------------------------------------------
 */

int
SpeakerRsrcModuleInit (Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "auPlay", SpeakerRsrcCreateCmd,
		      NULL, (void (*)()) NULL);
    Tcl_Eval(interp, "cminfo add media {audio}");

#ifdef AUDIO_FILE
    Tcl_Eval(interp, "cminfo add formats audio {AF}");
#endif
#ifdef SPARC_AUDIO
    Tcl_Eval(interp, "cminfo add formats audio {Sparc8}");
#endif

    Tcl_VarEval(interp, "cminfo add objects audio {",
#ifdef SPARC_AUDIO
			"Sparc8 ",
#endif
#ifdef AUDIO_FILE
			"AF ",
#endif
			"} auPlay", NULL);
    Tcl_VarEval(interp, "cminfo add object_info auPlay {",
#ifdef SPARC_AUDIO
			"Sparc8 ",
#endif
#ifdef AUDIO_FILE
			"AF ",
#endif
			"}", NULL);
    
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * SpeakerRsrcCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the "speakerRsrc" Tcl command.
 *     A new object of type SpeakerRsrc is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 * -------------------------------------------------------
 */

int
SpeakerRsrcCreateCmd (ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv)
{
    char new_name[256];
    SpeakerRsrc *new;
    
    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    
    nextSpeakerRsrcNum++;
    sprintf(new_name, "%sauPlay%d", argv[1], nextSpeakerRsrcNum);

    new = CreateNewSpeakerRsrc(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    
    Tcl_CreateCommand(interp, new_name, SpeakerRsrcObjCmd,
		      (ClientData) new, (void (*)()) NULL);
    
    if (argc > 2) {
	ConfigureSpeakerRsrc(new, argc-2, argv+2);
    }
    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/*
 * -------------------------------------------------------
 *
 * SpeakerRsrcObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     SpeakerRsrc object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 * -------------------------------------------------------
 */

int
SpeakerRsrcObjCmd (ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv)
{
    char c;
    int len;
    SpeakerRsrc *objPtr = (SpeakerRsrc *) clientData;
    
    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    
    c = argv[1][0];
    len = strlen(argv[1]);
    
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureSpeakerRsrc(objPtr, argc-2, argv+2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroySpeakerRsrc(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl;

	if (argc < 3) {
	    Tcl_AppendResult (interp, "wrong # of args, should be: ",
		argv[0], " accept <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}
	
	/*
	 * Parse the ScatterBufferList
	 */
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	    ckfree( (char *)sbl);
	    Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list ",
			     argv[2], NULL);
	    return TCL_ERROR;
	}

	return SpeakerRsrcAccept(objPtr, sbl);

    } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
	return SpeakerRsrcStats(objPtr);
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return SpeakerRsrcReady(objPtr);
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return SpeakerRsrcUnready(objPtr);
    } else if ((c == 'd') && (strncmp("deviceList", argv[1], len) == 0)) {
	return SpeakerResourceDeviceList(objPtr);
    }

    Tcl_AppendResult(interp, "Bad command option: ", argv[0], argv[1],
		     (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  accept <ScatterBufferList>\n",
                "  ready\n",
                "  unready\n",
                "  stats\n",
		"  deviceList\n",      
                (char *) NULL);

    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ChangeDeviceType --
 *
 *	Sets the device type for the SpeakerRsrc.  Called whenever
 *	the "devtype" slot is changed, and on initialization of the
 *	resource.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The old device is shut down, if appropriate, and the new
 *	device is created.
 *
 * Parameters:
 *	SpeakerRsrc *objPtr	Pointer to resource object.
 *	char *devtype		String giving name of new device
 *
 *--------------------------------------------------------------
 */
static void
ChangeDeviceType (SpeakerRsrc *objPtr, char *devtype)
{
    int len;
    char c;

    len = strlen(devtype);
    c = devtype[0];

    /*
     * If there's no change, return
     */
    if ((objPtr->devtype != NULL) &&
	(strncmp (objPtr->devtype, devtype, len) == 0)) {
	return;
    }

    /*
     * Shut down old device, if necessary, ignoring any errors it
     * returns.
     */
    SpeakerRsrcUnready(objPtr);
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
	objPtr->devtype = NULL;;
    }
    if (objPtr->devicePtr != NULL) {
        objPtr->devicePtr->shutdownProc(objPtr);
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
	objPtr->inited = 0;
    }

    /*
     * Create new device.
     */
    if ((c == 's') && (strncmp("sparc", devtype, len) == 0)) {
	objPtr->devicePtr = GetSparcDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("sparc")+1);
	strcpy(objPtr->devtype, "sparc");
    } else if ((c == 'a') && (strncmp("af", devtype, len) == 0)) {
	objPtr->devicePtr = GetAFDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("af")+1);
	strcpy(objPtr->devtype, "af");
    } else if ((c == 'w') && (strncmp("winaudio", devtype, len) == 0)) {
	objPtr->devicePtr = GetWinAudioDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("winaudio")+1);
	strcpy(objPtr->devtype, "winaudio");
    } else if ((c == 'l') && (strncmp("linux", devtype, len) == 0)) {
	objPtr->devicePtr = GetLinuxAuDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("linux")+1);
	strcpy(objPtr->devtype, "linux");
    } else if ((c == 'i') && (strncmp("irix", devtype, len) == 0)) {
	objPtr->devicePtr = GetIrixAuDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("irix")+1);
	strcpy(objPtr->devtype, "irix");
    } else {
	objPtr->devicePtr = GetDefaultSpeakerDevice(objPtr);
	objPtr->devtype = ckalloc(strlen(DEFAULT_AU_PLAY_DEV_TYPE)+1);
	strcpy(objPtr->devtype, DEFAULT_AU_PLAY_DEV_TYPE);
    }
}

/*
 * -------------------------------------------------------
 *
 * CreateNewSpeakerRsrc
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 * -------------------------------------------------------
 */

static SpeakerRsrc *
CreateNewSpeakerRsrc (Tcl_Interp *interp, char *name)
{
    SpeakerRsrc *new;
    char *atomName;

    new = (SpeakerRsrc *) ckalloc(sizeof(SpeakerRsrc));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);

    new->inited = 0;
    new->ready = 0;
    new->sampleRate = 0;
    new->channels = 0;
    new->format = 0;
    new->gain = 0.3;
    new->devtype = NULL;
    new->devicePtr = NULL;
    new->headphone = 0;
    new->lts = NULL;
    new->ltsTriggerNum = -1;
    new->ltsName = NULL;
    new->at_tokens = NULL;
    new->num_at_tokens = 0;
	new->sendAhead = DEF_SENDAHEAD;

    ChangeDeviceType (new, "default");

    /*
     * Register accept function
     */
    atomName = ckalloc(strlen(name) + 10);
    sprintf(atomName, "%s.accept", name);
    Tutil_RegisterCommand(interp, atomName, (ClientData)SpeakerRsrcAccept, 
			  (ClientData)new);
    new->acceptHandle = atomName;

    return new;
}


/*
 * -------------------------------------------------------
 *
 * DestroySpeakerRsrc
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 * -------------------------------------------------------
 */

static int 
DestroySpeakerRsrc (SpeakerRsrc *objPtr)
{
    int result = TCL_OK;
    int i, devReturn;

    /*
     * If not being called by DecodeDone first delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
    }
    /*
     * Shut down the device.
     */
    if (objPtr->devicePtr != NULL) {
	devReturn = 0;
	if (objPtr->devicePtr->shutdownProc != NULL) {
	    devReturn = objPtr->devicePtr->shutdownProc(objPtr);
	}
	if (devReturn != 0) {
	    result = TCL_ERROR;
	    if (objPtr->devicePtr->errorProc != NULL) {
		Tcl_AppendResult(objPtr->interp,
			 objPtr->devicePtr->errorProc(objPtr, devReturn),
				 (char *) NULL);
	    } else {
		Tcl_AppendResult(objPtr->interp,
				 "Error sutting down audio device", NULL);
	    }
	}
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
    }

    /*
     * Clean up lts
     */
    if (objPtr->ltsName != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	}
	ckfree(objPtr->ltsName);
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs and such.
     */
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    /*
     * Clean up previously scheduled display requests
     */
    if (objPtr->at_tokens != NULL) {
	for (i = 0; i < objPtr->num_at_tokens; i++) {
	    if (objPtr->at_tokens[i] > 0) {
		AtEvent *at_event;
		int token;

		token = objPtr->at_tokens[i];
		at_event = Tcm_FindAtEvent(token);
		if (token && (at_event != NULL)) {
		    SpeakerResourceFreeFrame((CBInfo *) at_event->fireClientData);
		    Tcm_DeleteAtHandler(token);
		    objPtr->at_tokens[i] = 0;
		}
	    }
	}
	ckfree((char *) objPtr->at_tokens);
    }

    ckfree((char *) objPtr);
    return result;
}


/*
 * -------------------------------------------------------
 *
 * ConfigureSpeakerRsrc
 *
 * Parameters: 
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 * 
 * Return: 
 *     Standard TCL return value
 * 
 * Action: 
 *     Does the usual configuration thing
 *
 * -------------------------------------------------------
 */

static int
ConfigureSpeakerRsrc (SpeakerRsrc *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
    
    /*
     * Case 1: <obj> configure
     */
    if (argc == 0) {
	sprintf(str, "{%1.2lf}", objPtr->gain);
	Tcl_AppendResult(interp, "{-gain {0.3} ", str, "} ", NULL);

	sprintf(str, "{%d}", objPtr->sampleRate);
	Tcl_AppendResult(interp, "{-sampleRate {8000} ", str, "} ", NULL);

	sprintf(str, "{%d}", objPtr->channels);
	Tcl_AppendResult(interp, "{-channels {1} ", str, "} ", (char *) NULL);

	if (objPtr->headphone) {
	    Tcl_AppendResult(interp, "{-output {speaker} {headphone} } ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-output {speaker} {speaker} } ", NULL);
	}

	Tcl_AppendResult(interp, "{", "-device ", "{", DEFAULT_AU_PLAY_DEV_TYPE, "} ",
			 objPtr->devtype, "} ", (char *)NULL);

	sprintf(str, "{%1.2lf} {%1.2lf}", DEF_SENDAHEAD, objPtr->sendAhead);
	Tcl_AppendResult(interp, "{-sendAhead ", str, " } ", NULL);

	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {NULL} {NULL} } ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {NULL} {",
			     objPtr->ltsName, "} } ", (char *) NULL);
	}
	return TCL_OK;
    }

    /*
     * Case 2: <obj> configure <-slot>
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 's') && (strncmp("-sampleRate", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->sampleRate);
	    Tcl_AppendResult(interp,
			     "-sampleRate {8000} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'g') && (strncmp("-gain", argv[0], len) == 0)) {
	    sprintf(str, "{%1.2lf}", objPtr->gain);
	    Tcl_AppendResult(interp, "-gain {0.3} ", str, NULL);
		return TCL_OK;
	} else if ((c == 'c') && (strncmp("-channels", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->channels);
	    Tcl_AppendResult(interp, "-channels {1} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'f') && (strncmp("-format", argv[0], len) == 0)) {
		Tcl_AppendResult(interp, "-format {mu-law} ", (char *)NULL);
		if (objPtr->format == SUN_ULAW)
			Tcl_AppendResult(interp, "{mu-law} ", (char *)NULL);
		else
			Tcl_AppendResult(interp, "{linear16} ", (char *)NULL);
		return TCL_OK;
	} else if ((c == 'o') && (strncmp("-output", argv[0], len) == 0)) {
	    if (objPtr->headphone) {
		Tcl_AppendResult(interp, "-output {speaker} {headphone}",
				 NULL);
	    } else {
		Tcl_AppendResult(interp, "-output {speaker} {speaker}",
				 NULL);
	    }
		return TCL_OK;
	} else if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
		sprintf(str, "{%1.2lf} {%1.2lf}", DEF_SENDAHEAD, objPtr->sendAhead);
		Tcl_AppendResult(interp, "-sendAhead ", str, NULL);
		return TCL_OK;
	} else if ((c == 'd') && (strncmp("-device", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-device ", "{", DEFAULT_AU_PLAY_DEV_TYPE, "} ",
			     objPtr->devtype, " ", (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->ltsName == NULL) {
		Tcl_AppendResult(interp, "-lts {NULL} {NULL}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {NULL} {",
				 objPtr->ltsName, "}",  (char *) NULL);
	    }
	    return TCL_OK;
	}

	Tcl_AppendResult(interp, "bad configuration option: ",
			 argv[0], (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: <obj> configure <-slot> <new value>
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	
	if ((c == 's') && (strncmp("-sampleRate", argv[cntr], len) == 0)) {
	    int newSampleRate;

	    cntr++;
	    if (cntr == argc) {
noValueError:
		Tcl_AppendResult(interp,
	    "Must specify value when trying to set configuration option ",
				argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newSampleRate) != TCL_OK) {
badValueError:
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->sampleRate = newSampleRate;

	/* --------------------------- GAIN ------------------------------- */
	} else if ((c == 'g') && (strncmp("-gain", argv[cntr], len) == 0)) {
	    double newGain;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (Tcl_GetDouble(interp, argv[cntr], &newGain) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->gain = newGain;

	    /*
	     * Call the device reinit procedure if we're ready
	     */
	    if (objPtr->ready) {
		if (objPtr->devicePtr->reinitProc != NULL) {
		    objPtr->devicePtr->reinitProc(objPtr);
		}
	    }
	/* --------------------------- FORMAT  ----------------------------- */
	} else if ((c == 'f') && (strncmp("-format", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    len = strlen(argv[cntr]);
	    if (strncmp("mu-law", argv[cntr], len) == 0) {
			objPtr->format = SUN_ULAW;
	    } else if (strncmp("linear16", argv[cntr], len) == 0) {
			objPtr->format = SUN_LIN_16;
	    } else {
		Tcl_AppendResult(interp,
			"Illegal selection of audio format to \"",
			argv[cntr],
			"\" must be one of \"mu-law\" or \"linear16\"",
			(char *) NULL);
		return TCL_ERROR;
	    }

	    /*
	     * Call the device reinit procedure if we're ready
	     */
	    if (objPtr->ready) {
		if (objPtr->devicePtr->reinitProc != NULL) {
		    objPtr->devicePtr->reinitProc(objPtr);
		}
	    }

	/* --------------------------- CHANNELS --------------------------- */
	} else if ((c == 'c') && (strncmp("-channels", argv[cntr], len) == 0)) {
	    int newChannels;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newChannels) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->channels = newChannels;

	/* --------------------------- DEVICE ----------------------------- */
	} else if ((c == 'd') && (strncmp("-device", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    ChangeDeviceType(objPtr, argv[cntr]);
	/* --------------------------- SENDAHEAD -------------------------- */
	} else if ((c == 's') && (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	    double newSendAhead;

	    cntr++;
	    if (cntr == argc) {
			goto noValueError;
	    }

	    if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->sendAhead = newSendAhead;
		
	/* --------------------------- OUTPUT ----------------------------- */
	} else if ((c == 'o') && (strncmp("-output", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    len = strlen(argv[cntr]);
	    if (strncmp("headphone", argv[cntr], len) == 0) {
		objPtr->headphone = 1;
	    } else if (strncmp("speaker", argv[cntr], len) == 0) {
		objPtr->headphone = 0;
	    } else {
		Tcl_AppendResult(interp,
			"Illegal selection of speaker output to \"",
			argv[cntr],
			"\" must be one of \"headphone\" or \"speaker\"",
			(char *) NULL);
		return TCL_ERROR;
	    }

	    /*
	     * Call the device reinit procedure if we're ready
	     */
	    if (objPtr->ready) {
		if (objPtr->devicePtr->reinitProc != NULL) {
		    objPtr->devicePtr->reinitProc(objPtr);
		}
	    }

	/* --------------------------- LTS -------------------------------- */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    
	    if (strlen(argv[cntr]) == 0) {
		newLts = NULL;
	    } else {
		newLts = (ClientData) Tcm_FindAtom(interp, argv[cntr]);
		if (newLts == NULL) {
		    Tcl_AppendResult(interp, argv[cntr],
				     " is not a valid lts.", NULL);
		    return TCL_ERROR;
		}
	    }

	    /*
	     * Release triggers and name of old LTS.
	     */
	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }
	    
	    /*
	     * Release our lock on the old lts...
	     */
	    if (objPtr->lts != NULL) {
			if (objPtr->ltsTriggerNum != -1) {
				LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
				objPtr->ltsTriggerNum = -1;
			}
			Tk_Release(objPtr->lts);
			objPtr->lts = NULL;
	    }
		
	    /*
	     * And set lock and arrange triggers for new one.
	     */
	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
		objPtr->ltsTriggerNum =
			LTS_SetTrigger(objPtr->lts,
					LTS_SPEED_FLAG|LTS_VALUE_FLAG|LTS_OFFSET_FLAG,
				        SpeakerRsrc_LTSChange,
				       (ClientData) objPtr);
		objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr])+1);
		strcpy(objPtr->ltsName,argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }

	/* --------------------------- ERROR ------------------------------ */
	} else {
	    Tcl_AppendResult(interp, "Invalid configuration parameter ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}

	cntr++;
    }
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * SpeakerRsrcReady
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *	Precomputes any values needed to get the object ready,
 *	and calls the device resady method, which does the same.
 *
 * -------------------------------------------------------
 */
static int
SpeakerRsrcReady (SpeakerRsrc *objPtr)
{
    int result;
    
    /*
     * Can only be made ready if
     *	1. We have a device.
     *	2. We have an lts
     */
    if (objPtr->devicePtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "Can't make speaker resource \"",
		objPtr->name, "\" ready: device type is not set.", (char *)NULL);
        return TCL_ERROR;
    }
    if (objPtr->lts == NULL) {
	Tcl_AppendResult(objPtr->interp, "Can't make speaker resource \"",
		objPtr->name, "\" ready: no lts has been attached.", (char *)NULL);
        return TCL_ERROR;
    }
    result = 0;

    /*
     * Call the device init/reinit procedure
     */
    if (!objPtr->inited) {
	if (objPtr->devicePtr->initProc != NULL) {
	    result = objPtr->devicePtr->initProc(objPtr);
	}
	if (result == 0)  {
	    objPtr->inited = 1;
	}
    } else {
	if (objPtr->devicePtr->reinitProc != NULL) {
	    result = objPtr->devicePtr->reinitProc(objPtr);
	}
    }

    /*
     * If bad result, get error string, return error
     */
    if (result != 0) {
	char *errStr;

	if (objPtr->devicePtr->errorProc != NULL) {
	    errStr = objPtr->devicePtr->errorProc(objPtr, result);
	} else {
	    errStr = "Error with device initialization.";
	}

	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
    }

    objPtr->ready = 1;
    
    return result;
}


/*
 * -------------------------------------------------------
 *
 * SpeakerRsrcUnready
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *	Performs any operations needed to make the object unready,
 *	and calls the device unready method, which does the same.
 *
 * -------------------------------------------------------
 */

static int
SpeakerRsrcUnready (SpeakerRsrc *objPtr)
{
    int result = 0;
    
    if (objPtr->devicePtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "No device was found.", (char *)
			 NULL);
        return TCL_ERROR;
    }

    if (objPtr->devicePtr->unreadyProc != NULL) {
	result = objPtr->devicePtr->unreadyProc(objPtr);
    }
    
    if (result != 0) {
	char *errStr;
	if (objPtr->devicePtr->errorProc != NULL) {
	    errStr = objPtr->devicePtr->errorProc(objPtr, result);
	} else {
	    errStr = "Error in device unready routine.";
	}
	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
    }

    objPtr->ready = 0;
    
    return result;
}


/*
 * -------------------------------------------------------
 *
 * SpeakerRsrc_LTSChange
 *
 * Parameters: 
 *	clientData - Really the object pointer
 * 
 * Return: 
 *	Standard Tcl return value
 * 
 * Action: 
 *	Called when the LTS changes value.  Dequeues any previously
 *	queued play requests, frees any memory we're holding, and
 *	returns.
 *
 * -------------------------------------------------------
 */
static void
SpeakerRsrc_LTSChange(ClientData clientData)
{
    SpeakerRsrc *objPtr = (SpeakerRsrc *) clientData;
    int i;
    
    /*
     * If not ready, do nothing.
     */
    if (objPtr->ready == 0) {
	return;
    }

    /*
     * Blow away "at" queue.
     */
    for (i=0; i<objPtr->num_at_tokens; i++) {
	if (objPtr->at_tokens[i] > 0) {
	    AtEvent *at_event;
	    int token;

	    token = objPtr->at_tokens[i];
	    at_event = Tcm_FindAtEvent(token);
	    if (token && (at_event != NULL)) {
		SpeakerResourceFreeFrame((CBInfo *) at_event->fireClientData);
		Tcm_DeleteAtHandler(token);
		objPtr->at_tokens[i] = 0;
	    }
	}
    }

    /*
     * Notify the device that the lts has changed.
     */
    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->ltsProc != NULL)) {
	objPtr->devicePtr->ltsProc(objPtr);
    }
}


/*
 * -------------------------------------------------------
 *
 * SpeakerRsrcAccept
 *
 * Parameters: 
 *     objPtr - pointer to object
 *     sbl -    ScatterBuffer list
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *
 * -------------------------------------------------------
 */

static int
SpeakerRsrcAccept (SpeakerRsrc *objPtr, SB_List *sbl)
{
    int i, tokId;
    double now, start, end;
    ScatterBuffer *sb;
    SparcAuBufferHeader *hdr;
    Buffer *buf;
    CBInfo *cbPtr;
    char *data, *tmp;
    int length;

    if (!objPtr->ready) {
		Tcl_AppendResult(objPtr->interp, "Cannot send data to ", objPtr->name,
						 " since it has not been readied", NULL);
		return TCL_ERROR;
    }
    
    /* 
     * Increment reference count on buffers passed in
     */
    AttachToAllBuffers(objPtr->interp, sbl, "speakerRsrc:SpeakerRsrcAccept");

	/* Don't copy sbl since don't need to pass it on */

    /*
     * Parse each buffer and create an AtEvent to play it.
     */
    for (i=0; i<sbl->count; i++) {
	sb = sbl->sbs[i];
	tmp = GetFromSB(sb, CM_SPARC_AU_BUF_HDR, NULL, NULL);
	if (tmp == NULL) {
	    DetachScatterBuffer (objPtr->interp, sb,
				"speakerRsrc:SpeakerRsrcAccept");
	    continue;
	}

	hdr = (SparcAuBufferHeader *) (tmp-8);
	start = ntohl(hdr->logStart_s) + ntohl(hdr->logStart_ns)/ONE_BILLION;
	end = ntohl(hdr->logEnd_s) + ntohl(hdr->logEnd_ns)/ONE_BILLION;

	now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));
	start = LTS_SystemTimeOf(objPtr->lts, start);
	end = LTS_SystemTimeOf(objPtr->lts, end);

	/*
	 * Check if we're too late to play
	 */
	if (now > end) {
	    DetachScatterBuffer (objPtr->interp, sb,
			         "speakerRsrc:SpeakerRsrcAccept");
	    continue;
	}

	/*
	 * Make a playback object data structure
	 */
	data = GetFromSB(sb, CM_DATA_SEGMENT, &buf, &length);
	if (data == NULL) {
	    DetachScatterBuffer (objPtr->interp, sb,
				"speakerRsrc:SpeakerRsrcAccept");
	    continue;
	}
	cbPtr = (CBInfo *)ckalloc(sizeof(CBInfo));
	cbPtr->objPtr = objPtr;
	cbPtr->data = data;
	cbPtr->len = length;
	cbPtr->buffer = buf;
	cbPtr->at_token = 0;
	cbPtr->hdr.hdrType = ntohl(hdr->hdrType);
	cbPtr->hdr.length = ntohl(hdr->length);
	cbPtr->hdr.device = ntohl(hdr->device);
	cbPtr->hdr.dataFormat = ntohl(hdr->dataFormat);
	cbPtr->hdr.samplingRate = ntohl(hdr->samplingRate);
	cbPtr->hdr.channelCount = ntohl(hdr->channelCount);
	cbPtr->hdr.logStart_s = ntohl(hdr->logStart_s);
	cbPtr->hdr.logStart_ns = ntohl(hdr->logStart_ns);
	cbPtr->hdr.logEnd_s = ntohl(hdr->logEnd_s);
	cbPtr->hdr.logEnd_ns = ntohl(hdr->logEnd_ns);

	{
	    char logStr[2000];
	    
	    sprintf(logStr, "SampleRate: %d Length: %d Start: %d %d End: %d %d", cbPtr->hdr.samplingRate,
		    cbPtr->len, cbPtr->hdr.logStart_s, cbPtr->hdr.logStart_ns,
		    cbPtr->hdr.logEnd_s, cbPtr->hdr.logEnd_ns);
	    Tcm_Log(logStr);
	}

	/*
	 * Set the cbPtr's start and end to be the correct system
	 * time, with end > start
	 */
	cbPtr->start = start;
	cbPtr->end = end;

	/*
	 * Search for a free at_token.  If we can't find
	 * one, to increase size of at_token array
	 */
	if (objPtr->num_at_tokens == 0) {
	    objPtr->num_at_tokens = 30;
	    objPtr->at_tokens = (int *) ckalloc(sizeof(int) * 30);
	    for (tokId=0; tokId<objPtr->num_at_tokens; tokId++) {
		objPtr->at_tokens[tokId] = 0;
	    }
	} 
	for (tokId=0; tokId<objPtr->num_at_tokens; tokId++) {
	    if (objPtr->at_tokens[tokId] == 0) {
		break;
	    }
	}
	if (tokId == objPtr->num_at_tokens) {
	    int newNum = objPtr->num_at_tokens + 10;
	    objPtr->at_tokens = (int *) realloc((char *) objPtr->at_tokens,
						      sizeof(int)*newNum);
	    for (tokId=objPtr->num_at_tokens; tokId < newNum; tokId++) {
		objPtr->at_tokens[tokId] = 0;
	    }
	    tokId = objPtr->num_at_tokens;
	    objPtr->num_at_tokens = newNum;
	}

	objPtr->at_tokens[tokId] = 
		Tcm_CreateAtHandler(start-objPtr->sendAhead, HUGE_VAL,
			    "audio", (Tk_TimerProc *) SpeakerResourcePlayData,
			    (ClientData) cbPtr, NULL,
			    NULL, (Tk_IdleProc *) NULL,
			    (ClientData) NULL, NULL);
	cbPtr->at_token = tokId;
    }
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * SpeakerResourceFreeFrame --
 *
 *	Detach a frame that has been played or will never be played
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The frame is recycled
 *
 * Parameters:
 *	cbPtr		Pointer to data to be played.
 *
 *--------------------------------------------------------------
 */
void
SpeakerResourceFreeFrame (CBInfo *cbPtr)
{
    SpeakerRsrc *objPtr;

    objPtr = cbPtr->objPtr;
    DetachBuffer (objPtr->interp, cbPtr->buffer,
		"speakerRsrc:SpeakerResourceFreeFrame");

    /*
     * The device may have put a lock on this frame structure
     * using Tk_Preserve(), so we have to free it using Tk_EventuallyFree
     */
    Tk_EventuallyFree (cbPtr, (Tk_FreeProc *)free);
}

/*
 *--------------------------------------------------------------
 *
 * SpeakerResourcePlayData --
 *
 *	Given a pointer to a ScatterBuffer containing the audio header 
 *	(converted to host byte order) and the audio data, call the
 *	device play routine.
 *
 * Results:
 *	A standard Tcl return value
 *
 * Side effects:
 *	The frame will be sent to the audio device.
 *
 * Parameters:
 *	cbPtr		Pointer to data to be played.
 *
 *--------------------------------------------------------------
 */
static int
SpeakerResourcePlayData (CBInfo *cbPtr)
{
    SpeakerRsrc *objPtr;
    double now;
    int res;

    objPtr = cbPtr->objPtr;

    /*
     * If we're in the window, call the play procedure.
     */
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));
    if ((cbPtr->start-objPtr->sendAhead <= now) && (now <= cbPtr->end)) {
	/*
	 * Call device's display procedure.
	 */
	if (objPtr->devicePtr != NULL) {
	    if (objPtr->devicePtr->playProc != NULL) {
		res = objPtr->devicePtr->playProc(cbPtr);
		if (res != 0) {
		    if (objPtr->devicePtr->errorProc != NULL) {
			Tcl_AppendResult (objPtr->interp,
				"Error playing data in Speaker Resource ",
				objPtr->name, ": ", 
				objPtr->devicePtr->errorProc(objPtr,res), NULL);
		    } else {
			Tcl_AppendResult (objPtr->interp,
				"Error playing data in Speaker Resource ",
				objPtr->name, NULL);
		    }
		    Tk_BackgroundError (objPtr->interp);
		    SpeakerRsrcUnready(objPtr);
		}
	    }
	}
    }

    /*
     * Recycle buffer and return
     */
    SpeakerResourceFreeFrame (cbPtr);
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * SpeakerRsrcStats
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *
 * -------------------------------------------------------
 */

static int
SpeakerRsrcStats (SpeakerRsrc *objPtr)
{
    Tcl_AppendResult(objPtr->interp, "This command is not yet implemented.",
		     NULL);
    return TCL_OK;
}

/******************************************************************
 *
 * SpeakerResourceDeviceList
 *
 * Parameters:
 *     objPtr - pointer to mpeg play object.
 *
 * Results:
 *     Standard Tcl Result.
 *
 * Actions:
 *     Sets result to list of devices supported. Returns TCL_OK;
 *
 ******************************************************************/
static int
SpeakerResourceDeviceList(SpeakerRsrc *objPtr)
{
    Tcl_Interp *interp = objPtr->interp;
    Tcl_DString resStr;

    Tcl_DStringInit(&resStr);

#ifdef AUDIO_FILE
    Tcl_DStringAppendElement(&resStr, "af");
#endif
#ifdef SPARC_AU
    Tcl_DStringAppendElement(&resStr, "sparc");
#endif
#ifdef WIN_AUDIO
    Tcl_DStringAppendElement(&resStr, "winaudio");
#endif
#ifdef LINUX_AU
    Tcl_DStringAppendElement(&resStr, "linux");
#endif
#ifdef IRIX_AU
    Tcl_DStringAppendElement(&resStr, "irix");
#endif

    Tcl_SetResult(interp, Tcl_DStringValue(&resStr), TCL_VOLATILE);
    Tcl_DStringFree(&resStr);

    return TCL_OK;
}
