/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * cmAuSegment.c
 *   author: Jonathan Logan, Cornell University
 *
 * HISTORY:
       6/22/95  Initial version.
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <netinet/in.h>
#include "cmAtom.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmChunk.h"
#include "cmLTS.h"
#include "cmAuFile.h"
#include "cmAuUtil.h"
#include "cmUtil.h"

#include "libaudio.h"
#include "audio_device.h"

#define	ONE_BILLION	1000000000.0

#define DEF_PACKET_FACTOR 4
#define MAX_PACKET_FACTOR 16

/*
 * One of these structures is created for each AuPriority object
 */

typedef struct AuPriority {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    char *outCmd;		/* Command to send buffers to. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;		/* ClientData for outCCmd. */
    char *acceptHandle;
    ClientData lts;		/* LTS for object. */
    char *ltsName;		/* Tcl name for lts */
    int ready;
    int pktFactor;		/* current packetization factor */
} AuPriority;

static int nextAuPriorityNum = 0;  /* Variable for next unique object name */
static unsigned int nextSegID = 0; /* used to identify groups of packets */

/*
 * Procedures defined in this module
 */
int			AuPriorityModuleInit(Tcl_Interp *interp);

static AuPriority *	CreateNewAuPriority(Tcl_Interp *interp, char *name);
static int		DestroyAuPriority(AuPriority *objPtr);
static int		ConfigureAuPriority(AuPriority *objPtr,
			    int argc, char **argv);
static int		AuPriorityCreateCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		AuPriorityObjCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		AuPriority_Accept(AuPriority *objPtr, SB_List *sbl);
static int		AuPriority_ChangeOutCmd(AuPriority *objPtr,
			    char *newOut);
static int		AuPriorityReady(AuPriority *objPtr);
static int		AuPriorityUnready(AuPriority *objPtr);

static void		PrintSparcAuHdr(char *ptr);


/************************************************************
 *
 * CreateNewAuPriority
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static AuPriority *
CreateNewAuPriority(Tcl_Interp *interp, char *name)
{

    AuPriority *new;
    char *acceptName;

    new = (AuPriority *) ckalloc(sizeof(AuPriority));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy (new->name, name);
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->lts = NULL;
    new->ltsName = NULL;
    new->pktFactor = DEF_PACKET_FACTOR;
    new->ready = 0;


    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+9);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) AuPriority_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    return new;
}

/************************************************************
 *
 * DestroyAuPriority
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/
static int
DestroyAuPriority(AuPriority *objPtr)
{
    /*
     * Remove command name
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
        Tk_Release(objPtr->lts);
        objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
	
    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureAuPriority
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

static int
ConfigureAuPriority(AuPriority * objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
  
    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}
		
	sprintf(str, "{%d} {%d}", DEF_PACKET_FACTOR, objPtr->pktFactor);
	Tcl_AppendResult(interp, "{-pktFactor ", str, "} ", NULL);
		
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {NULL} {NULL} } ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {NULL} {",
			     objPtr->ltsName, "} } ", (char *) NULL);
	}
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
				 "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->ltsName == NULL) {
		Tcl_AppendResult(interp, "-lts {NULL} {NULL}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {NULL} {",
				 objPtr->ltsName, "}",  (char *) NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'p') && (strncmp("-pktFactor", argv[0], len) == 0)) {
	    sprintf(str, "{%d} {%d}", DEF_PACKET_FACTOR, objPtr->pktFactor);
	    Tcl_AppendResult(interp, "-pktFactor ", str, NULL);
	    return TCL_OK;
	}


	Tcl_AppendResult(interp, "bad configuration option: ",
			 argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
	    noValueError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (AuPriority_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }

        /* --------------------------- LTS -------------------------------- */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    
	    if (strlen(argv[cntr]) == 0) {
		newLts = NULL;
	    } else {
		newLts = Tcm_FindAtom(interp, argv[cntr]);
		if (newLts == NULL) {
		    Tcl_AppendResult(interp, argv[cntr],
				     " is not a valid lts.", NULL);
		    return TCL_ERROR;
		}
	    }

	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }

            /*
             * Release our lock on the old lts...
             */
            if (objPtr->lts != NULL) {
                Tk_Release(objPtr->lts);
                objPtr->lts = NULL;
            }
	    
            /*
             * And set lock and arrange triggers for new one.
             */
	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
		objPtr->ltsName = (char *) ckalloc(sizeof(argv[cntr])+1);
		strcpy(objPtr->ltsName,argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }
        /* ----------------------- PKTFACTOR ---------------------------- */
	} else if ((c == 'p') && (strncmp("-pktFactor", argv[cntr], len) == 0)) {
	    int newPktFactor;

	    cntr++;
	    if (cntr == argc) goto noValueError;
		
	    if (Tcl_GetInt(interp, argv[cntr], &newPktFactor) != TCL_OK) {
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }

	    /* check for valid pktFactor */
	    if ((newPktFactor < 1) || (newPktFactor > MAX_PACKET_FACTOR)) {
		Tcl_AppendResult(interp, "Packet factor ", argv[cntr],
				 " out of range\n", (char *)NULL);
		return TCL_ERROR;
	    }
		
	    objPtr->pktFactor = newPktFactor;
	/* --------------------------- ERROR ------------------------------ */
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			     argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}


/************************************************************
 *
 * AuPriorityCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     A new object of type AuPriority is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

static int
AuPriorityCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    AuPriority *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextAuPriorityNum++;
    sprintf(new_name, "%sAuPriority%d", argv[1], nextAuPriorityNum);

    new = CreateNewAuPriority(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, AuPriorityObjCmd,
			(ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureAuPriority(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * AuPriorityObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     AuPriority object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

static int
AuPriorityObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    AuPriority *objPtr = (AuPriority *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureAuPriority(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyAuPriority(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl;
	int rv;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			     argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}
		
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	    ckfree((char *) sbl);
	    Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
		
	rv = AuPriority_Accept(objPtr, sbl);
	FreeSBList(sbl);
	return rv;

    /*---------------------------- READY -------------------------------*/
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
        return AuPriorityReady(objPtr);

    /*---------------------------- UNREADY -----------------------------*/
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
        return AuPriorityUnready(objPtr);
    }
	
    Tcl_AppendResult(interp, "Bad command option: ", argv[0],
		     argv[1], (char *) NULL);
  usage:
    Tcl_AppendResult(interp,
		     "\n<command> can be one of:\n",
		     "  configure ?-slot ?value??\n",
		     "  destroy\n",
		     "  ready\n", 
		     "  unready\n",
		     "  accept <ScatterBufferList>\n",
		     (char *) NULL);

    return TCL_ERROR;
}
	
/************************************************************
 *
 * AuPriorityModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
AuPriorityModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "auPriority", AuPriorityCreateCmd,
			NULL, (void (*)()) NULL);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Au_Prioritize --
 *
 *	A prioritization algorithm that computes the
 *	ordering for the au data packets.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters 
 *	int n		Number of frames to prioritize
 *	int m		Starting frame number.
 *	int k		Starting priority offset.
 *	int *pl		Output priority array
 *
 *--------------------------------------------------------------
 */
static void
Au_Prioritize (int n, int m, int k, int *pl)
{
    int s,t;
	
    if (n==1) {
	*pl = m;
    } else {
	s = n/2;
	t = n-s;
	Au_Prioritize(t,m,k*2,pl);
	Au_Prioritize(s,m+t,k*2,pl+k);
    }
}


/************************************************************
 *
 * Au_PrioritizeSBL
 *
 * Parameters:
 *     sbl - scatter buffer list to be prioritized.
 *   
 * Results:
 *     None.
 *
 * Actions:
 *     Reorders sbl according to priority algorithm.
 *
 ************************************************************/

static void
Au_PrioritizeSBL(SB_List *sbl)
{
    int numSBs, i;
    ScatterBuffer **tempSBs;
    int *orderVec = NULL;

    if (sbl == NULL) return;
    if ((sbl->count == 0) || (sbl->count == 1)) return;

    /*
     * Re-order sbl in priority order.
     */
    numSBs = sbl->count;
    
    orderVec = (int *)ckalloc(sizeof(int)*numSBs);
    tempSBs = (ScatterBuffer **)ckalloc(sizeof(ScatterBuffer *)*numSBs);
    Au_Prioritize(numSBs, 0, 1, orderVec);
    for (i=0; i<numSBs; i++) {
	tempSBs[i] = sbl->sbs[orderVec[i]];
    }
    for (i=0; i<numSBs; i++) {
	sbl->sbs[i] = tempSBs[i];
    }
    ckfree ((char *)tempSBs);
    ckfree ((char *)orderVec);
}


/************************************************************
 *
 * AuPriority_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -	scatter buffer list containing unsegmented audio
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Expects input list to contain buffers with two sections:
 *         (1) SparcAuBufferHeader and (2) CM Data segment w/audio data.
 *     Packetizes audio, prioritizes and passes it on. 
 ************************************************************/
static int
AuPriority_Accept(AuPriority *objPtr, SB_List *sbl)
{
    int i, j, index, which_sb, pktID;
    ScatterBuffer *destSBs[MAX_PACKET_FACTOR], *hdrSB;
    Buffer *bufPtr, *hdrBuffer;
    Buffer *destBufs[MAX_PACKET_FACTOR], destBufHdrs[MAX_PACKET_FACTOR];
    char *destData[MAX_PACKET_FACTOR];
    SparcAuPktHeader segHdr;
    SparcAuBufferHeader bufHdr;
    char *data, *src;
    double now;
    int sampleSize, destBufSize, destSBSize;
    int unitSize, numUnits, totDataLen;
    int dataCookie, dataLen;
    int dataFormat, channelCount;
    SB_List *out_sbl;
	
    if (!objPtr->ready)
	return TCL_OK;

    AttachToAllBuffers(objPtr->interp, sbl, "AuPriority:Accept");

    if (objPtr->lts != NULL) {
	now = LTS_Value(objPtr->lts);
    } else {
	now = Tcm_ReadSysClock();
    }
    /*	printf ("%s.accept (@%1.3lf)\n", objPtr->name, now);*/

    for (which_sb = 0; which_sb<sbl->count; which_sb++) {
	data = GetFromSB(sbl->sbs[which_sb], CM_SPARC_AU_BUF_HDR, &bufPtr, NULL);
	if (data == NULL) {
	    Tcl_AppendResult(objPtr->interp, "Missing audio buffer header.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	memcpy((char *)(&bufHdr)+8, data, sizeof(SparcAuBufferHeader)-8);

	/* fields in bufHdr should already be in network byte order */
	segHdr.hdrType = htonl(CM_SPARC_AU_PKT_HDR);
	segHdr.length = htonl(sizeof(SparcAuPktHeader)) - 8;
	segHdr.device = bufHdr.device;
	segHdr.dataFormat = bufHdr.dataFormat;
	segHdr.samplingRate = bufHdr.samplingRate;
	segHdr.channelCount = bufHdr.channelCount;
	segHdr.logStart_s = bufHdr.logStart_s;
	segHdr.logStart_ns = bufHdr.logStart_ns;
	segHdr.logEnd_s = bufHdr.logEnd_s;
	segHdr.logEnd_ns = bufHdr.logEnd_ns;

	nextSegID++;
	segHdr.segID = htonl(nextSegID);
	segHdr.pktFactor = htonl(objPtr->pktFactor);

	dataFormat = ntohl(bufHdr.dataFormat);
	channelCount = ntohl(bufHdr.channelCount);
		
	data = GetFromSB(sbl->sbs[which_sb], CM_DATA_SEGMENT, &bufPtr, &totDataLen);
	if (data == NULL) {
	    Tcl_AppendResult(objPtr->interp, "Missing audio data segment.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
		
	segHdr.totDataLen = htonl(totDataLen);

	sampleSize = Tcm_AuComputeSampleSize(dataFormat);
	unitSize = (sampleSize * channelCount);
	numUnits = totDataLen/unitSize;

	destBufSize = totDataLen / objPtr->pktFactor + 1;
	destSBSize = sizeof(SparcAuPktHeader) + 8 + destBufSize;

	dataCookie = htonl(CM_DATA_SEGMENT);

	/* reserve destination buffers */
	for (i=0; i<objPtr->pktFactor; i++) {
	    destBufs[i] = MakeBuffer(objPtr->interp, destSBSize,
				     "AuPriority:DestBufs");
	    destData[i] = destBufs[i]->data + sizeof(SparcAuPktHeader) + 8;

	    segHdr.pktID = htonl(i);

	    memcpy(destBufs[i]->data, (char *)&segHdr,
		   sizeof(SparcAuPktHeader));
	    memcpy(destBufs[i]->data + sizeof(SparcAuPktHeader), 
		   (char *)&dataCookie,sizeof(int));
			
	    /* the data length of a particular buffer may be shorter if 
	     *   there are an odd number of samples to go around.
	     */
	    dataLen = destBufSize - 
		((i>=(numUnits % objPtr->pktFactor) ) ?  1 : 0);
	    destSBs[i] = MakeScatterBuffer(destBufs[i],
					   dataLen+8+sizeof(SparcAuPktHeader),
					   NULL);
	    dataLen = htonl(dataLen);
	    memcpy(destBufs[i]->data +sizeof(SparcAuPktHeader)+sizeof(int),
		   (char *)&dataLen, sizeof(int));
	}

	out_sbl = (SB_List *)ckalloc(sizeof(SB_List));
	out_sbl->count = objPtr->pktFactor;
	out_sbl->sbs = (ScatterBuffer **)
	    ckalloc(sizeof(ScatterBuffer *) * out_sbl->count);

	for (i = 0; i<objPtr->pktFactor; i++) {
	    out_sbl->sbs[i] = destSBs[i];
	}

	/* fill in destination buffers */
	for (src=data, pktID = 0; src < (data+totDataLen); 
	     src += unitSize, pktID = ((pktID + 1) % objPtr->pktFactor)) {
	    memcpy(destData[pktID], src, unitSize);
	    destData[pktID] += unitSize;
	}

	/* Prioritize packets */
	Au_PrioritizeSBL(out_sbl);

	/*
	 * Pass on buffers to outCmd
	 */
	if (objPtr->outCmd != NULL) {
	    if (objPtr->outCCmd != NULL) {
		/*
		 * Need to check if it still exists since object may have
		 * been destroyed.
		 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		    objPtr->outCCmd(objPtr->outCD, out_sbl);
		} else {
		    /*
		     * Out C command has unexpectedly disappeared.
		     */
		    objPtr->outCCmd = NULL;
		    ckfree(objPtr->outCmd);
		    objPtr->outCmd = NULL;
		}
	    } else {
		/*
		 * Out command is tcl, make sb list into a string and 
		 * call using Tcl_Eval. 
		 */
		char *sbStr = MakeSBListString(out_sbl);
		Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ", 
			    sbStr, (char *) NULL); 
		ckfree(sbStr);
	    }
	}
	DetachToAllBuffers(objPtr->interp, out_sbl, "AuPriority:oneBuf");

	FreeSBList(out_sbl);
    }
    DetachToAllBuffers(objPtr->interp, sbl, "AuPriority:AcceptComplete");
	
    return TCL_OK;
}


/******************************************************************
 * AuPriorityReady
 *
 * 
 ******************************************************************/
static int
AuPriorityReady(AuPriority *objPtr) {
    
    objPtr->ready = 1;
    return TCL_OK;
}


/******************************************************************
 * AuPriorityUnready
 *
 * 
 ******************************************************************/
static int
AuPriorityUnready(AuPriority *objPtr) {
    objPtr->ready = 0;
    return TCL_OK;
}


/************************************************************
 *
 * AuPriority_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
AuPriority_ChangeOutCmd(AuPriority *objPtr, char *newOut) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
	    &(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Couldn't find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    

/*
 *--------------------------------------------------------------
 *
 * PrintSparcAuHdr --
 *
 *	Pretty print the contents of a SparcAuBufferHeader structure.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *ptr	Pointer to buffer with SparcAuBufferHeader
 *	int len		Size of buffer (ignored)
 *
 *--------------------------------------------------------------
 */
static void
PrintSparcAuHdr (char *ptr)
{
    SparcAuBufferHeader hdr;
    char *dataFormat;
    char *device;

    memcpy (&hdr, ptr, sizeof(SparcAuBufferHeader));

    hdr.device = ntohl(hdr.device);
    hdr.dataFormat = ntohl(hdr.dataFormat);
    hdr.samplingRate = ntohl(hdr.samplingRate);
    hdr.channelCount = ntohl(hdr.channelCount);
    hdr.logStart_s = ntohl(hdr.logStart_s);
    hdr.logStart_ns = ntohl(hdr.logStart_ns);
    hdr.logEnd_s = ntohl(hdr.logEnd_s);
    hdr.logEnd_ns = ntohl(hdr.logEnd_ns);

    if (hdr.dataFormat == SUN_ULAW) {
	dataFormat = "Mu Law";
    } else if (hdr.dataFormat == SUN_LIN_8) {
	dataFormat = "Linear 8 bit";
    } else if (hdr.dataFormat == SUN_LIN_16) {
	dataFormat = "Linear 16 bit";
    } else {
	dataFormat = "UNKNOWN";
    }

    if (hdr.device == PARALLAX_DEV_CODE) {
	device = "Parallax";
    } else if (hdr.device == J300_DEV_CODE) {
	device = "J300";
    } else if (hdr.device == SPARC_AU_DEV_CODE) {
	device = "Sparc Audio";
    } else {
	device = "UNKNOWN";
    }

    printf ("Sparc Audio Buffer Header:\n");
    printf ("  device = %s\n", device);
    printf ("  dataFormat = %s\n", dataFormat);
    printf ("  samplingRate = %d\n", hdr.samplingRate);
    printf ("  channelCount = %d\n", hdr.channelCount);
    printf ("  logStart = %1.3lf\n", 
		hdr.logStart_s+hdr.logStart_ns/ONE_BILLION);
    printf ("  logEnd = %1.3lf\n", 
		hdr.logEnd_s+hdr.logEnd_ns/ONE_BILLION);
}
