/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * cmAuReassem.c
 *
 *  original author:  Jon Logan, logan@acm.org, Cornell University
 *
 * ToDo:
 *  ?? Modify to keep go off slightly before sendAhead time in order to
 * allow more time to recombine if necessary? Make configurable parameter?
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <netinet/in.h>
#include <math.h>
#include <tcl.h>
#include <tk.h>
#include "tcmEvent.h"
#include "cmAtom.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmChunk.h"
#include "cmLTS.h"
#include "cmAuFile.h"
#include "cmAuUtil.h"
#include "cmUtil.h"

#include "libaudio.h"
#include "audio_device.h"

#define	ONE_BILLION	1000000000.0

#define DEF_SENDAHEAD     (0.150)
#define DEF_MODE           Fill
#define MAX_PACKET_FACTOR (16)

/* max number of SB_Lists that can be under construction */
#define MAX_WORKQ         (100) 

/* incoming segments are basically hashed into the circular work queue by their
 * segment ID modulo MAX_WORKQ
 *
 * Relies on the fact that segIDs are monotonically increasing.
 */

typedef struct WorkQSlot {
    SB_List *sbl;      /* the output buffer under construction */
    /* if NULL, then no buffer is currently in this slot */

    int segID;            /* segment ID of buffer under construction */
    int rcvFlag[MAX_PACKET_FACTOR];  /* 1 if packet[i] has been received */
    /* duplicates some information from pkt headers for easy reference */
    int pktFactor;
    int samplingRate;
    int channelCount;
    int unitSize;      /* unit = channelCount * sampleSize */
    int totBufLen;
    int format;
    int atToken;
} WorkQSlot;
 
/*
 * One of these structures is created for each AuReassem object
 */

typedef struct AuReassem {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;		    	/* Tcl string name of this object. */
    char *outCmd;		    /* Command to send buffers to. */
    InFunc outCCmd;		    /* C pointer to out command */
    ClientData outCD;		/* ClientData for outCCmd. */
    char *acceptHandle;
    ClientData lts;		    /* LTS for object. */
    char *ltsName;		    /* Tcl name for lts */
    int ltsTriggerNum;
    int ready;

    double sendAhead;       /* how far in advance of playback time to send */
	
    enum { Fill, Interpolate } mode;

    WorkQSlot workQ[MAX_WORKQ];  /* output buffers under construction */
} AuReassem;

/* Used as clientData for Tcm_AtHandler events */
typedef struct TimerSendEvent {
    AuReassem *objPtr;
    int segID;
} TimerSendEvent;

static int nextAuReassemNum = 0;	/* Variable for next unique object name */

/*
 * Procedures defined in this module
 */
int			AuReassemModuleInit(Tcl_Interp *interp);

static AuReassem *	CreateNewAuReassem(Tcl_Interp *interp, char *name);
static int		DestroyAuReassem(AuReassem *objPtr);
static int		ConfigureAuReassem(AuReassem *objPtr,
			    int argc, char **argv);
static int		AuReassemCreateCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		AuReassemObjCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		AuReassem_Accept(AuReassem *objPtr, SB_List *sbl);
static int		AuReassem_ChangeOutCmd(AuReassem *objPtr,
			    char *newOut);
static void		ReleaseWQSlot(AuReassem *objPtr, WorkQSlot *wqs,
			    int freeFlg);
static void		AuReassem_LTSChange(ClientData clientData);
static void		AuReassemSend(ClientData clientData);
static int		AuReassemResample(AuReassem *objPtr, WorkQSlot *wqs);
static int		AuReassemReady(AuReassem *objPtr);
static int		AuReassemUnready(AuReassem *objPtr);

/************************************************************
 *
 * CreateNewAuReassem
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static AuReassem *
CreateNewAuReassem(Tcl_Interp *interp, char *name)
{

    AuReassem *new;
    char *acceptName;
	int i, j;

    new = (AuReassem *) ckalloc(sizeof(AuReassem));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy (new->name, name);
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->lts = NULL;
    new->ltsName = NULL;
    new->sendAhead = DEF_SENDAHEAD;
    new->mode = DEF_MODE;
    new->ready = 0;

    for (i=0; i<MAX_WORKQ; i++) {
	new->workQ[i].sbl = NULL;

	new->workQ[i].segID = -1;
	for (j=0; j<MAX_PACKET_FACTOR; j++) {
	    new->workQ[i].rcvFlag[j] = 0;
	}
    }

    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+9);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) AuReassem_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    return new;
}

/************************************************************
 *
 * DestroyAuReassem
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/
static int
DestroyAuReassem(AuReassem *objPtr)
{
    int i;

    /*
     * Remove command name
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }
	
    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
        Tk_Release(objPtr->lts);
        objPtr->lts = NULL;
    }

    for (i=0; i<MAX_WORKQ; i++) {
	if (objPtr->workQ[i].sbl)
	    FreeSBList(objPtr->workQ[i].sbl);

	if (objPtr->workQ[i].atToken != -1) {
	    Tcm_DeleteAtHandler(objPtr->workQ[i].atToken);
	}
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }

    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureAuReassem
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

static int
ConfigureAuReassem(AuReassem * objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
  
    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}

	sprintf(str, "{%1.2lf} {%1.2lf}", DEF_SENDAHEAD, objPtr->sendAhead);
	Tcl_AppendResult(interp, "{-sendAhead ", str, "} ", NULL);
		
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {NULL} {NULL} } ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {NULL} {",
			     objPtr->ltsName, "} } ", (char *) NULL);
	}
    }


    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
				 "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->ltsName == NULL) {
		Tcl_AppendResult(interp, "-lts {NULL} {NULL}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {NULL} {",
				 objPtr->ltsName, "}",  (char *) NULL);
	    }
	    return TCL_OK;
	} else if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
	    sprintf(str, "{%1.2lf} {%1.2lf}", DEF_SENDAHEAD, objPtr->sendAhead);
	    Tcl_AppendResult(interp, "-sendAhead ", str, NULL);
	    return TCL_OK;
	}

	Tcl_AppendResult(interp, "bad configuration option: ",
			 argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
	      noValueError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (AuReassem_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }

        /* --------------------------- LTS -------------------------------- */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    
	    if (strlen(argv[cntr]) == 0) {
		newLts = NULL;
	    } else {
		newLts = Tcm_FindAtom(interp, argv[cntr]);
		if (newLts == NULL) {
		    Tcl_AppendResult(interp, argv[cntr],
				     " is not a valid lts.", NULL);
		    return TCL_ERROR;
		}
	    }

	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }
		
	    /*
	     * Release our lock on the old lts...
	     */
	    if (objPtr->lts != NULL) {
		if (objPtr->ltsTriggerNum != -1) {
		    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
		    objPtr->ltsTriggerNum = -1;
		}
		Tk_Release(objPtr->lts);
		objPtr->lts = NULL;
	    }
	    
	    /*
	     * And set lock and arrange triggers for new one.
	     */
	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
		objPtr->ltsTriggerNum =
		    LTS_SetTrigger(objPtr->lts,
				   LTS_SPEED_FLAG|LTS_VALUE_FLAG|LTS_OFFSET_FLAG,
				   AuReassem_LTSChange, (ClientData) objPtr);
		objPtr->ltsName = (char *) ckalloc(sizeof(argv[cntr])+1);
		strcpy(objPtr->ltsName,argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }
        /* ----------------------- SENDAHEAD ---------------------------- */
	} else if ((c == 's') && (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	    double newSendAhead;

	    cntr++;
	    if (cntr == argc) goto noValueError;
		
	    if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
		
	    objPtr->sendAhead = newSendAhead;
	/* --------------------------- ERROR ------------------------------ */
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			     argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}


/************************************************************
 *
 * AuReassemCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     A new object of type AuReassem is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

static int
AuReassemCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    AuReassem *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextAuReassemNum++;
    sprintf(new_name, "%sAuReassem%d", argv[1], nextAuReassemNum);

    new = CreateNewAuReassem(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, AuReassemObjCmd,
			(ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureAuReassem(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * AuReassemObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     AuReassem object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

static int
AuReassemObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    AuReassem *objPtr = (AuReassem *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureAuReassem(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyAuReassem(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl;
	int rv;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			     argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}
		
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	    ckfree((char *) sbl);
	    Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
		
	rv = AuReassem_Accept(objPtr, sbl);
	FreeSBList(sbl);
	return rv;

    /*---------------------------- READY -------------------------------*/
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
        return AuReassemReady(objPtr);

    /*---------------------------- UNREADY -----------------------------*/
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
        return AuReassemUnready(objPtr);
    }

    Tcl_AppendResult(interp, "Bad command option: ", argv[0],
		     argv[1], (char *) NULL);
  usage:
    Tcl_AppendResult(interp,
		     "\n<command> can be one of:\n",
		     "  configure ?-slot ?value??\n",
		     "  destroy\n",
		     "  ready\n", 
		     "  unready\n",
		     "  accept <ScatterBufferList>\n",
		     (char *) NULL);
    return TCL_ERROR;
}
	
/************************************************************
 *
 * AuReassemModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
AuReassemModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "auReassem", AuReassemCreateCmd,
			NULL, (void (*)()) NULL);
    return TCL_OK;
}


/******************************************************************
 * FindWQSlot
 *
 *   Given a segment ID, find the slot in our private hash table.
 * Guaranteed to return an address within our workQ, although the
 * slot itself may be unitialized.
 ******************************************************************/
static WorkQSlot *
FindWQSlot(AuReassem *objPtr, int segID)
{
    return &(objPtr->workQ[segID % MAX_WORKQ]);
}

/******************************************************************
 * ReleaseWQSlot
 *
 *   Cleanup elements of WQ slot.
 * if freeFlg=1 then free the sb_list. 
 ******************************************************************/
static void
ReleaseWQSlot(AuReassem *objPtr, WorkQSlot *wqs, int freeFlg)
{
    int j;

    if (wqs->sbl) {
	DetachToAllBuffers(objPtr->interp, wqs->sbl, "AuReassem:ReleaseWQSlot");
	if (freeFlg) FreeSBList(wqs->sbl);
	wqs->sbl = NULL;
    }
    wqs->segID = -1;
    for (j=0; j<MAX_PACKET_FACTOR; j++) {
	wqs->rcvFlag[j] = 0;
    }
	
    if (wqs->atToken != -1) {
	Tcm_DeleteAtHandler(wqs->atToken);
	wqs->atToken = -1;
    }
}


/************************************************************
 *
 * AuReassem_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -	scatter buffer list containing packetized, prioritized audio
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Expects input list to contain buffers with two sections:
 *         (1) SparcAuPktHeader and (2) CM Data segment w/audio data.
 *     Packetizes audio, prioritizes and passes it on. 
 ************************************************************/
static int
AuReassem_Accept(AuReassem *objPtr, SB_List *sbl)
{
    int i, j, index, which_sb, pktID;
    Buffer *destDataBufPtr, *srcBufPtr;
    ScatterBuffer *sbPtr;
    SparcAuPktHeader pktHdr;
    SparcAuBufferHeader auBufHdr;
    char *data, *dest, *stop;
    TimerSendEvent *timEvent;
    double when, now;
    int sampleSize, destBufSize, destSBSize;
    int numUnits;
    int incr;
    int dataSegLen, dataSegCode; 
    int pktDataLen;        /* length of data segment of incoming packet */
    int totBufLen;         /* total length of destination buffer */
    WorkQSlot *wqs;

    if ((objPtr->lts == NULL) || (!objPtr->ready)) {
	return TCL_OK;
    }

    AttachToAllBuffers(objPtr->interp, sbl, "AuReassem:Accept");
    sbl = CopySBList(sbl);

    if (objPtr->lts != NULL) {
	now = LTS_Value(objPtr->lts);
    } else {
	now = Tcm_ReadSysClock();
    }
    /*	printf ("%s.accept (@%1.3lf)\n", objPtr->name, now);*/

    for (which_sb = 0; which_sb<sbl->count; which_sb++) {
	data = GetFromSB(sbl->sbs[which_sb], CM_SPARC_AU_PKT_HDR, &srcBufPtr, NULL);
	if (data == NULL) {
	    Tcl_AppendResult(objPtr->interp, "Missing audio packet header.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
		
	memcpy((char *)&pktHdr, data-8, sizeof(SparcAuPktHeader));

	pktHdr.hdrType = ntohl(pktHdr.hdrType);
	pktHdr.length = ntohl(pktHdr.length);
	pktHdr.device = ntohl(pktHdr.device);
	pktHdr.dataFormat = ntohl(pktHdr.dataFormat);
	pktHdr.samplingRate = ntohl(pktHdr.samplingRate);
	pktHdr.channelCount = ntohl(pktHdr.channelCount);
	pktHdr.logStart_s = ntohl(pktHdr.logStart_s);
	pktHdr.logStart_ns = ntohl(pktHdr.logStart_ns);
	pktHdr.logEnd_s = ntohl(pktHdr.logEnd_s);
	pktHdr.logEnd_ns = ntohl(pktHdr.logEnd_ns);
		
	pktHdr.segID = ntohl(pktHdr.segID);
	pktHdr.pktFactor = ntohl(pktHdr.pktFactor);
	pktHdr.pktID = ntohl(pktHdr.pktID);
	pktHdr.totDataLen = ntohl(pktHdr.totDataLen);

	sampleSize = Tcm_AuComputeSampleSize(pktHdr.dataFormat);

	now = LTS_Value(objPtr->lts);
	when = pktHdr.logStart_s + 	pktHdr.logStart_ns/ONE_BILLION;

	/* See if we received the packet in time. */
	if (now > when - (objPtr->sendAhead * LTS_Speed(objPtr->lts))) {
	    printf("aiee - missed window:  now = %1.2lf  when = %1.2lf  sa = %1.2lf\n", now, when, objPtr->sendAhead);
	    printf("which_sb %i\n", which_sb); 
	    break;
	}
	/* Find or create entry for this segment in our work queue. */
	wqs = FindWQSlot(objPtr, pktHdr.segID);
	if (wqs->segID != pktHdr.segID) {
	    if (wqs->segID != -1) {  /* oops, we must have wrapped around */
		/* just silently destroy buffer */
		ReleaseWQSlot(objPtr, wqs, 1);		
	    }
	    /* Have not received segment before, create dest buffer */
	    totBufLen = pktHdr.totDataLen + 8 + sizeof(SparcAuBufferHeader);
	    destDataBufPtr = MakeBuffer(objPtr->interp, totBufLen, 
					"AuReassem:CreateDest");
	    /* Fill in entire au buf hdr & info for data segment */
	    auBufHdr.hdrType = htonl(CM_SPARC_AU_BUF_HDR);
	    auBufHdr.device = htonl(pktHdr.device);
	    auBufHdr.length = htonl(sizeof(SparcAuBufferHeader) - 8);
	    auBufHdr.dataFormat = htonl(pktHdr.dataFormat);
	    auBufHdr.samplingRate = htonl(pktHdr.samplingRate);
	    auBufHdr.channelCount = htonl(pktHdr.channelCount);
	    auBufHdr.logStart_s = htonl(pktHdr.logStart_s);
	    auBufHdr.logStart_ns = htonl(pktHdr.logStart_ns);
	    auBufHdr.logEnd_s = htonl(pktHdr.logEnd_s);
	    auBufHdr.logEnd_ns = htonl(pktHdr.logEnd_ns);

	    memcpy(destDataBufPtr->data, (char *)&auBufHdr, 
		   sizeof(SparcAuBufferHeader));

	    dataSegCode = htonl(CM_DATA_SEGMENT);
	    memcpy(destDataBufPtr->data + sizeof(SparcAuBufferHeader),
		   (char *)&dataSegCode, sizeof(int));

	    dataSegLen = htonl(pktHdr.totDataLen);
	    memcpy(destDataBufPtr->data
		   + sizeof(SparcAuBufferHeader) + sizeof(int),
		   (char *)&dataSegLen, sizeof(int));

	    sbPtr = MakeScatterBuffer(destDataBufPtr, totBufLen, NULL);
	    wqs->sbl = MakeSBList(sbPtr, NULL);

	    wqs->pktFactor = pktHdr.pktFactor;
	    wqs->samplingRate = pktHdr.samplingRate;
	    wqs->channelCount = pktHdr.channelCount;
	    wqs->format = pktHdr.dataFormat;
	    wqs->totBufLen = totBufLen;
	    wqs->segID = pktHdr.segID;

	    wqs->unitSize = wqs->channelCount * sampleSize;

	    /* and schedule output event */
	    timEvent = (TimerSendEvent *)ckalloc(sizeof(TimerSendEvent));
	    if (!timEvent) {
		Tcl_AppendResult(objPtr->interp, "out of memory\n", 
				 (char *)NULL);
		return TCL_ERROR;
	    }
	    timEvent->objPtr = objPtr;
	    timEvent->segID = pktHdr.segID;

	    when = LTS_SystemTimeOf(objPtr->lts, when - 
				    (objPtr->sendAhead * LTS_Speed(objPtr->lts)));
	    wqs->atToken = Tcm_CreateAtHandler(when, HUGE_VAL, "audio",
					       AuReassemSend, timEvent, NULL,
					       NULL, NULL, NULL, NULL);
	}

	/* assert: wqs points to valid slot */

	if (wqs->rcvFlag[pktHdr.pktID] == 1) {
	    Tcl_AppendResult(objPtr->interp, "Received duplicate packet.",
			     (char *) NULL);
	    return TCL_ERROR;
	}			

	/* Snarf the data segment */
	data = GetFromSB(sbl->sbs[which_sb], CM_DATA_SEGMENT, (Buffer **)NULL, 
			 &pktDataLen);

	if (data == NULL) {
	    Tcl_AppendResult(objPtr->interp, "Missing audio data segment.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
		
	dest = GetFromSB(wqs->sbl->sbs[0], CM_DATA_SEGMENT, (Buffer **)NULL, NULL);

	/* copy into dest buf - this loop needs to be optimized! */
	stop = dest + pktHdr.totDataLen;
	incr = pktHdr.pktFactor * wqs->unitSize;

	/* uses memcpy to avoid having to worry about network byte order
	 *  for now. This may be a performance problem though
         */
	for (dest= dest + (pktHdr.pktID * wqs->unitSize); 
	     dest < stop; 
	     dest+= incr, data += wqs->unitSize)
	{
	    memcpy(dest, data, wqs->unitSize);
	}
		
	/* Mark pkt received */
	wqs->rcvFlag[pktHdr.pktID] = 1;
    }

    /* XXX: If we have run out of time, (or if packet is complete???) */
    /*   then ship what we have - 1st case essentially covered by
	 other timer, but this might save some time. Have to think
	 bout this one */
    
    DetachToAllBuffers(objPtr->interp, sbl, "AuReassem:AcceptComplete");
    FreeSBList(sbl);

#if  0
    if (objPtr->lts != NULL) {
	now = LTS_Value(objPtr->lts);
    } else {
	now = Tcm_ReadSysClock();
    }
    printf ("%s DONE (@%1.3lf)\n", objPtr->name, now);
#endif

    return TCL_OK;
}

/******************************************************************
 * AuReassemResample
 *
 * Action:
 *   Fills in gaps where packets were not received according to the
 * interpolation mode selected. 
 *  Fill mode: simply duplicate the last valid sample received in place
 *             of missing samples. Advantage: simple, and avoids network
 *             byte order considerations. 
 *  Interpolate mode: if implemented, would average surrounding samples.
 *
 * Notes: 
 *   Strictly speaking this is not resampling, but the name is 
 * close enough for our purposes.
 ******************************************************************/
static int
AuReassemResample(AuReassem *objPtr, WorkQSlot *wqs) {
    int prevValidOffset[MAX_PACKET_FACTOR]; /* offset (in bytes) 
					     * of previously valid sample */
    int i, j, curPkt;
    char *dest, *baseAddr;
    int offset;
    Buffer *dummyBuf;
    double now;

    if (objPtr->lts != NULL) {
	now = LTS_Value(objPtr->lts);
    } else {
	now = Tcm_ReadSysClock();
    }
    /*	printf ("%s Resample (@%1.3lf)\n", objPtr->name, now);*/

    /* Calculate byte offsets to previous valid samples */
    for (i=wqs->pktFactor-1; i>=0; i--) {
	j=i;
	while (!wqs->rcvFlag[j]) 
	    j = ( (j==0) ? wqs->pktFactor : j - 1);
	if (j <= i)
	    prevValidOffset[i] = (j-i) * wqs->unitSize;
	else
	    prevValidOffset[i] = ((j - i) - wqs->pktFactor) * wqs->unitSize;
    }

    switch(objPtr->mode) {
        case Fill: /* simply copies sample from previously valid sample */
	/*
	 * for each unreceived packet, go through buffer duplicating 
	 *  valid previous samples
	 */
	baseAddr = GetFromSB(wqs->sbl->sbs[0], CM_DATA_SEGMENT, &dummyBuf, NULL);
	if (!baseAddr) {
	    Tcl_AppendResult(objPtr->interp, "AuReassem: lost data buffer",
			     (char *)NULL);
	    return TCL_ERROR;
	}
	curPkt = 0;
	while (curPkt < wqs->pktFactor) {
	    if (!(wqs->rcvFlag[curPkt])) {
		int incr;

		incr = wqs->unitSize * wqs->pktFactor;
		for (dest = baseAddr + (curPkt * wqs->unitSize),
		     offset = prevValidOffset[curPkt];  
		     dest < baseAddr + wqs->totBufLen; 
		     dest += incr) {

		    memcpy(dest, dest - offset, wqs->unitSize);
		}
	    }
	    curPkt++;
	}

	/* if packet 0 is missing, the first few samples will be bogus.
	 *  This happens so rarely that it would be a waste to put a check 
	 *  in the loop above.
	 */

	if (!wqs->rcvFlag[0]) {
	    for (i=0; i<wqs->pktFactor; i++) {
		dest=baseAddr + (i * wqs->unitSize);
		offset = prevValidOffset[i] + (wqs->unitSize * wqs->pktFactor);
		memcpy(dest, dest + offset, wqs->unitSize);
	    }
	}

	break;
      case Interpolate:
	/* currently unimplemented */
	break;
      default:
	break;
    }

    if (objPtr->lts != NULL) {
	now = LTS_Value(objPtr->lts);
    } else {
	now = Tcm_ReadSysClock();
    }
    /*	printf ("%s Resample DONE (@%1.3lf)\n", objPtr->name, now);*/

    return TCL_OK;
}

/******************************************************************
 * AuReassemSend
 *
 *  Invoked in response to TimerSendEvent going off.
 ******************************************************************/
static void
AuReassemSend(ClientData clientData) {
    TimerSendEvent *evt = (TimerSendEvent *)clientData;
    WorkQSlot *wqs = FindWQSlot(evt->objPtr, evt->segID);
    AuReassem *objPtr = evt->objPtr;
    int i, pktComplete;

    if (wqs->sbl) {
	pktComplete = 1;
	for (i=0; (i < wqs->pktFactor) && pktComplete; i++)
	    pktComplete = wqs->rcvFlag[i];

	if (!pktComplete) {
	    AuReassemResample(objPtr, wqs);
	}

	/*
	 * If out command is C, call it directly
	 */
	if (objPtr->outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have been
	     * destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd + 1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, wqs->sbl);
	    } else {
		objPtr->outCCmd = NULL;
		ckfree(objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and call using
	     * Tcl_Eval. 
	     */
	    char *sbStr = MakeSBListString(wqs->sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL);
	    ckfree(sbStr);
	}
	DetachToAllBuffers(objPtr->interp, wqs->sbl,"AuReassem:Send" );
    }

    ReleaseWQSlot(evt->objPtr, wqs, 1);
    ckfree((char *)evt);
}

/******************************************************************
 * AuReassemReady
 *
 * 
 ******************************************************************/
static int
AuReassemReady(AuReassem *objPtr)
{
    if (objPtr->lts == NULL) {
	Tcl_AppendResult(objPtr->interp, "Can't ready au reassembly object ",
			 objPtr->name, ": need to attach lts", (char *) NULL);
	return TCL_ERROR;
    }

    objPtr->ready = 1;
    return TCL_OK;
}


/******************************************************************
 * AuReassemUnready
 *
 *  Unschedules and frees everything in work queue. 
 ******************************************************************/
static int
AuReassemUnready(AuReassem *objPtr)
{
    int i;

    for (i=0; i<MAX_WORKQ; i++) {
	if (objPtr->workQ[i].sbl)
	    FreeSBList(objPtr->workQ[i].sbl);

	if (objPtr->workQ[i].atToken != -1) {
	    Tcm_DeleteAtHandler(objPtr->workQ[i].atToken);
	}
	objPtr->workQ[i].atToken = -1;
    }

    objPtr->ready = 0;
    return TCL_OK;
}

/******************************************************************
 * AuReassem_LTSChange
 *
 * Parameters: 
 *	clientData - Really the object pointer
 * 
 * Return: 
 *	Standard Tcl return value
 * 
 * Action: 
 *	Called when the LTS changes value. Reschedules any active events.
 *  
 ******************************************************************/
static void
AuReassem_LTSChange(ClientData clientData)
{
    AuReassem *objPtr = (AuReassem *)clientData;
    double when;
    int i;
    TimerSendEvent *timEvent;

    if (objPtr->ready == 0)
	return;

    for (i=0; i<MAX_WORKQ; i++) {
	if (objPtr->workQ[i].atToken != -1) {
	    Tcm_DeleteAtHandler(objPtr->workQ[i].atToken);

	    timEvent = (TimerSendEvent *)ckalloc(sizeof(TimerSendEvent));

	    when = LTS_SystemTimeOf(objPtr->lts, when - 
				    (objPtr->sendAhead * LTS_Speed(objPtr->lts)));
	    timEvent->segID = objPtr->workQ[i].segID;
	    timEvent->objPtr = objPtr;
	    objPtr->workQ[i].atToken = 
		Tcm_CreateAtHandler(when, HUGE_VAL, "audio",
				    AuReassemSend, timEvent, NULL,
				    NULL, NULL, NULL, NULL);
	}
    }
    return;
}


/************************************************************
 *
 * AuReassem_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
AuReassem_ChangeOutCmd(AuReassem *objPtr, char *newOut) 
{
    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
	    &(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Couldn't find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    
