/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif

#include <math.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "cmAuRecord.h"
#include "tcmEvent.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmAuFile.h"
#include "cmCookies.h"

#define ONE_BILLION 1000000000.0
#define EPSILON 0.2

#define DEF_SAMPLE_RATE  (8000)
#define DEF_SAMPLE_SIZE  (1)
#define DEF_FORMAT       (SUN_ULAW)
#define DEF_CYCLE_TIME   (0.5)
#define DEF_DELAY_TIME   (0.8)
#define DEF_GAIN         (0.5)


static int nextAudioInRsrcNum = 0;	/* Variable for next unique object
					 * name. */

/*
 * Prototypes for routines used internally
 */
int		AudioInRsrcModuleInit(Tcl_Interp *interp);
static AudioInRsrc *	CreateNewAudioInRsrc(Tcl_Interp *interp, char *name);

static int	AudioInRsrcCreateCmd(ClientData clientData,
		    Tcl_Interp *interp, int argc, char **argv);
static int	AudioInRsrcObjCmd(ClientData clientData,
		    Tcl_Interp *interp, int argc, char **argv);
static void	ChangeDeviceType(AudioInRsrc *objPtr, char *devtype);
static int	DestroyAudioInRsrc(AudioInRsrc *objPtr);
static int	ConfigureAudioInRsrc(AudioInRsrc *objPtr, int argc, char **argv);
static int	AudioInRsrcReady(AudioInRsrc *objPtr);
static int	AudioInRsrcUnready(AudioInRsrc *objPtr);
static void	AudioInRsrc_LTSChange(ClientData clientData);
static int	AudioInRsrcStats(AudioInRsrc *objPtr);
static int	AudioInChangeOutCmd(AudioInRsrc *objPtr, char *newOut);
static void	AudioInSendData(ClientData clientData);
static int	AudioInResourceDeviceList(AudioInRsrc *objPtr);
static void	AudioInScheduleFirstSend(AudioInRsrc *objPtr);

/*
 * -------------------------------------------------------
 *
 * AudioInRsrcModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 * -------------------------------------------------------
 */

int
AudioInRsrcModuleInit (Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "auRecord", AudioInRsrcCreateCmd,
		      NULL, (void (*)()) NULL);

    /* Tcl_Eval(interp, "cminfo add media {audio}"); */


    Tcl_VarEval(interp, "cminfo add objects audio {",
#ifdef SPARC_AUDIO
		"Sparc8 ",
#endif
		"} auRecord", NULL);

    Tcl_VarEval(interp, "cminfo add object_info auRecord {",
#ifdef SPARC_AUDIO
		"Sparc8 ",
#endif
		"}", NULL);
    
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * AudioInRsrcCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the "AudioInRsrc" Tcl command.
 *     A new object of type AudioInRsrc is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 * -------------------------------------------------------
 */

int
AudioInRsrcCreateCmd (ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv)
{
    char new_name[256];
    AudioInRsrc *new;
    
    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    
    nextAudioInRsrcNum++;
    sprintf(new_name, "%sauRecord%d", argv[1], nextAudioInRsrcNum);

    new = CreateNewAudioInRsrc(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    
    Tcl_CreateCommand(interp, new_name, AudioInRsrcObjCmd,
		      (ClientData) new, (void (*)()) NULL);
    
    if (argc > 2) {
	ConfigureAudioInRsrc(new, argc-2, argv+2);
    }
    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/*
 * -------------------------------------------------------
 *
 * AudioInRsrcObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     AudioInRsrc object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 * -------------------------------------------------------
 */

int
AudioInRsrcObjCmd (ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv)
{
    char c;
    int len;
    AudioInRsrc *objPtr = (AudioInRsrc *) clientData;
    
    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    
    c = argv[1][0];
    len = strlen(argv[1]);
    
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureAudioInRsrc(objPtr, argc-2, argv+2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyAudioInRsrc(objPtr);
    } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
	return AudioInRsrcStats(objPtr);
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return AudioInRsrcReady(objPtr);
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return AudioInRsrcUnready(objPtr);
    } else if ((c == 'd') && (strncmp("deviceList", argv[1], len) == 0)) {
	return AudioInResourceDeviceList(objPtr);
    }

    Tcl_AppendResult(interp, "Bad command option: ", argv[0], argv[1],
		     (char *) NULL);
 usage:
    Tcl_AppendResult(interp,
		     "\n<command> can be one of:\n",
		     "  configure ?-slot ?value??\n",
		     "  destroy\n",
		     "  ready\n",
		     "  unready\n",
		     "  stats\n",
		     "  deviceList\n",
		     (char *) NULL);

    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ChangeDeviceType --
 *
 *	Sets the device type for the AudioInRsrc.  Called whenever
 *	the "devtype" slot is changed, and on initialization of the
 *	resource.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The old device is shut down, if appropriate, and the new
 *	device is created.
 *
 * Parameters:
 *	AudioInRsrc *objPtr	Pointer to resource object.
 *	char *devtype		String giving name of new device
 *
 *--------------------------------------------------------------
 */
static void
ChangeDeviceType (AudioInRsrc *objPtr, char *devtype)
{
    int len;
    char c;

    len = strlen(devtype);
    c = devtype[0];

    /*
     * If there's no change, return
     */
    if ((objPtr->devtype != NULL) &&
	(strncmp (objPtr->devtype, devtype, len) == 0)) {
	return;
    }

    /*
     * Shut down old device, if necessary, ignoring any errors it
     * returns.
     */
    AudioInRsrcUnready(objPtr);
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
	objPtr->devtype = NULL;;
    }
    if (objPtr->devicePtr != NULL) {
        objPtr->devicePtr->shutdownProc(objPtr);
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
	objPtr->inited = 0;
    }

    /*
     * Create new device.
     */
    if ((c == 's') && (strncmp("sparc", devtype, len) == 0)) {
	objPtr->devicePtr = GetSparcAuInDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("sparc")+1);
	strcpy(objPtr->devtype, "sparc");
    } else {
	objPtr->devicePtr = GetDefaultAudioInDevice(objPtr);
	objPtr->devtype = ckalloc(strlen(DEFAULT_AU_RECORD_DEV_TYPE)+1);
	strcpy(objPtr->devtype, DEFAULT_AU_RECORD_DEV_TYPE);
    }
}

/*
 * -------------------------------------------------------
 *
 * CreateNewAudioInRsrc
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 * -------------------------------------------------------
 */

static AudioInRsrc *
CreateNewAudioInRsrc (Tcl_Interp *interp, char *name)
{
    AudioInRsrc *new;

    new = (AudioInRsrc *) ckalloc(sizeof(AudioInRsrc));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);

    new->inited = 0;
    new->ready = 0;

    new->sampleRate = DEF_SAMPLE_RATE;
    new->sampleSize = DEF_SAMPLE_SIZE;
    new->channels = 1;
    new->format = DEF_FORMAT;
    new->cycleTime = DEF_CYCLE_TIME;
    new->delayTime = DEF_DELAY_TIME;
    new->gain = DEF_GAIN;
    new->devtype = NULL;
    new->devicePtr = NULL;
    new->inputPort = MICROPHONE;
    new->lts = NULL;
    new->ltsTriggerNum = -1;
    new->ltsName = NULL;
    new->outCmd = NULL;
    new->atToken = -1;
    new->atWhen = 0.0;

    ChangeDeviceType (new, "default");

    return new;
}


/*
 * -------------------------------------------------------
 *
 * DestroyAudioInRsrc
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 * -------------------------------------------------------
 */

static int 
DestroyAudioInRsrc (AudioInRsrc *objPtr)
{
    int result = TCL_OK;
    int  devReturn;

    /*
     * If not being called by DecodeDone first delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Shut down the device.
     */
    if (objPtr->devicePtr != NULL) {
	devReturn = 0;
	if (objPtr->devicePtr->shutdownProc != NULL) {
	    devReturn = objPtr->devicePtr->shutdownProc(objPtr);
	}
	if (devReturn != 0) {
	    result = TCL_ERROR;
	    if (objPtr->devicePtr->errorProc != NULL) {
		Tcl_AppendResult(objPtr->interp,
				 objPtr->devicePtr->errorProc(objPtr, devReturn),
				 (char *) NULL);
	    } else {
		Tcl_AppendResult(objPtr->interp,
				 "Error sutting down audio device", NULL);
	    }
	}
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
    }

    /*
     * Clean up lts
     */
    if (objPtr->ltsName != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	}
	ckfree(objPtr->ltsName);
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs and such.
     */
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    if (objPtr->atToken) {
	if (Tcm_FindAtEvent(objPtr->atToken) != NULL) {
	    Tcm_DeleteAtHandler(objPtr->atToken);
	    objPtr->atToken = 0;
	}
    }

    ckfree((char *) objPtr);
    return result;
}


/*
 * -------------------------------------------------------
 *
 * ConfigureAudioInRsrc
 *
 * Parameters: 
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 * 
 * Return: 
 *     Standard TCL return value
 * 
 * Action: 
 *     Does the usual configuration thing
 *
 * -------------------------------------------------------
 */

static int
ConfigureAudioInRsrc (AudioInRsrc *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
    
    /*
     * Case 1: <obj> configure
     */
    if (argc == 0) {
	sprintf(str, "{%1.2lf} {%1.2lf}", DEF_GAIN, objPtr->gain);
	Tcl_AppendResult(interp, "{-gain ", str, "} ", NULL);

	sprintf(str, "{%d} {%d}", DEF_SAMPLE_RATE, objPtr->sampleRate);
	Tcl_AppendResult(interp, "{-sampleRate ", str, "} ", NULL);

	sprintf(str, "{%d}", objPtr->channels);
	Tcl_AppendResult(interp, "{-channels {1} ", str, "} ", (char *) NULL);

	if (objPtr->inputPort == CM_LINE_IN) {
	    Tcl_AppendResult(interp, "{-input {microphone} {line_in} } ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-input {microphone} {microphone} } ", NULL);
	}

	if (objPtr->format == SUN_ULAW) {
	    Tcl_AppendResult(interp, "{-format {mu-law} {mu-law} } ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-input {mu-law} {linear16} } ", (char *)NULL);
	}

	sprintf(str, "{%1.2lf} {%1.2lf}", DEF_CYCLE_TIME, objPtr->cycleTime);
	Tcl_AppendResult(interp, "{-cycleTime ", str, "} ", (char *) NULL);

	sprintf(str, "{%1.2lf} {%1.2lf}", DEF_DELAY_TIME, objPtr->delayTime);
	Tcl_AppendResult(interp, "{-delayTime ", str, "} ", (char *) NULL);

	Tcl_AppendResult(interp, "{", "-device ", "{", DEFAULT_AU_RECORD_DEV_TYPE, "} ",
			 objPtr->devtype, "} ", (char *)NULL);

	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {NULL} {NULL} } ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {NULL} {",
			     objPtr->ltsName, "} } ", (char *) NULL);
	}

	if (objPtr->outCmd != NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "} } ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {} } ", NULL);
	}

	return TCL_OK;
    }

    /*
     * Case 2: <obj> configure <-slot>
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 's') && (strncmp("-sampleRate", argv[0], len) == 0)) {
	    sprintf(str, "{%d} {%d}", DEF_SAMPLE_RATE, objPtr->sampleRate);
	    Tcl_AppendResult(interp, "-sampleRate ", str, NULL);

	    return TCL_OK;
	} else if ((c == 'g') && (strncmp("-gain", argv[0], len) == 0)) {
	    sprintf(str, "{%1.2lf} {%1.2lf}", DEF_GAIN, objPtr->gain);
	    Tcl_AppendResult(interp, "-gain ", str, NULL);
	    return TCL_OK;
	} else if ((c == 'c') && (strncmp("-channels", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->channels);
	    Tcl_AppendResult(interp, "-channels {1} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {

	    sprintf(str, "{%1.2lf} {%1.2lf}", DEF_CYCLE_TIME, objPtr->cycleTime);
	    Tcl_AppendResult(interp, "-cycleTime ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'i') && (strncmp("-input", argv[0], len) == 0)) {
	    if (objPtr->inputPort == MICROPHONE) {
		Tcl_AppendResult(interp, "-input {microphone} {microphone}",
				 NULL);
	    } else {
		Tcl_AppendResult(interp, "-input {microphone} {line_in}",
				 NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'f') && (strncmp("-format", argv[0], len) == 0)) {
	    if (objPtr->format == SUN_ULAW) {
		Tcl_AppendResult(interp, "-format {mu-law} {mu-law}",
				 NULL);
	    } else {
		Tcl_AppendResult(interp, "-format {mu-law} {linear16}",
				 NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-device", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-device ", "{", DEFAULT_AU_RECORD_DEV_TYPE, "} ",
			     objPtr->devtype, " ", (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-delayTime", argv[0], len) == 0)) {
	    sprintf(str, "{%1.2lf} {%1.2lf}", DEF_DELAY_TIME, objPtr->delayTime);
	    Tcl_AppendResult(interp, "-delayTime ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->ltsName == NULL) {
		Tcl_AppendResult(interp, "-lts {NULL} {NULL}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {NULL} {",
				 objPtr->ltsName, "}",  (char *) NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd != NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
				 "}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	    }
	    return TCL_OK;
	}

	Tcl_AppendResult(interp, "bad configuration option: ",
			 argv[0], (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: <obj> configure <-slot> <new value>
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	
	/* --------------------------- SAMPLERATE ------------------------- */
	if ((c == 's') && (strncmp("-sampleRate", argv[cntr], len) == 0)) {
	    int newSampleRate;

	    cntr++;
	    if (cntr == argc) {
	    noValueError:
		Tcl_AppendResult(interp,
				 "Must specify value when trying to set configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newSampleRate) != TCL_OK) {
	    badValueError:
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->sampleRate = newSampleRate;

	    /* --------------------------- GAIN ------------------------------- */
	} else if ((c == 'g') && (strncmp("-gain", argv[cntr], len) == 0)) {
	    double newGain;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (Tcl_GetDouble(interp, argv[cntr], &newGain) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->gain = newGain;

	    /*
	     * Call the device reinit procedure if we're ready
	     */
	    if (objPtr->ready) {
		if (objPtr->devicePtr->reinitProc != NULL) {
		    objPtr->devicePtr->reinitProc(objPtr);
		}
	    }
	    /* --------------------------- DELAYTIME --------------------------- */
	} else if ((c == 'd') && (strncmp("-delayTime", argv[cntr], len) == 0)) {
	    double newDelayTime;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (Tcl_GetDouble(interp, argv[cntr], &newDelayTime) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->delayTime = newDelayTime;
	    
	    /* --------------------------- CHANNELS --------------------------- */
	} else if ((c == 'c') && (strncmp("-channels", argv[cntr], len) == 0)) {
	    int newChannels;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newChannels) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->channels = newChannels;

	    /*
	     * Call the device reinit procedure if we're ready
	     */
	    if (objPtr->ready) {
		if (objPtr->devicePtr->reinitProc != NULL) {
		    objPtr->devicePtr->reinitProc(objPtr);
		}
	    }


	    /* --------------------------- CYCLETIME --------------------------- */
	} else if ((c == 'c') && (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	    double newCycleTime;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->cycleTime = newCycleTime;

	    /* --------------------------- DEVICE ----------------------------- */
	} else if ((c == 'd') && (strncmp("-device", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    ChangeDeviceType(objPtr, argv[cntr]);

	    /* --------------------------- INPUT ----------------------------- */
	} else if ((c == 'i') && (strncmp("-input", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    len = strlen(argv[cntr]);
	    if (strncmp("microphone", argv[cntr], len) == 0) {
		objPtr->inputPort = MICROPHONE;
	    } else if (strncmp("line_in", argv[cntr], len) == 0) {
		objPtr->inputPort = CM_LINE_IN;
	    } else {
		Tcl_AppendResult(interp,
				 "Illegal selection of AudioInput source to \"",
				 argv[cntr],
				 "\" must be one of \"microphone\" or \"line_in\"",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    /*
	     * Call the device reinit procedure if we're ready
	     */
	    if (objPtr->ready) {
		if (objPtr->devicePtr->reinitProc != NULL) {
		    objPtr->devicePtr->reinitProc(objPtr);
		}
	    }


	    /* --------------------------- FORMAT  ----------------------------- */
	} else if ((c == 'f') && (strncmp("-format", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    len = strlen(argv[cntr]);
	    if (strncmp("mu-law", argv[cntr], len) == 0) {
		objPtr->format = SUN_ULAW;
		objPtr->sampleSize = 1;
	    } else if (strncmp("linear16", argv[cntr], len) == 0) {
		objPtr->format = SUN_LIN_16;
		objPtr->sampleSize = 2;
	    } else {
		Tcl_AppendResult(interp,
				 "Illegal selection of audio format to \"",
				 argv[cntr],
				 "\" must be one of \"mu-law\" or \"linear16\"",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    /*
	     * Call the device reinit procedure if we're ready
	     */
	    if (objPtr->ready) {
		if (objPtr->devicePtr->reinitProc != NULL) {
		    objPtr->devicePtr->reinitProc(objPtr);
		}
	    }
	    /* --------------------------- LTS -------------------------------- */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    
	    if (strlen(argv[cntr]) == 0) {
		newLts = NULL;
	    } else {
		newLts = (ClientData) Tcm_FindAtom(interp, argv[cntr]);
		if (newLts == NULL) {
		    Tcl_AppendResult(interp, argv[cntr],
				     " is not a valid lts.", NULL);
		    return TCL_ERROR;
		}
	    }

	    /*
	     * Release triggers and name of old LTS.
	     */
	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }
	    
	    /*
	     * Release our lock on the old lts...
	     */
	    if (objPtr->lts != NULL) {
		if (objPtr->ltsTriggerNum != -1) {
		    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
		    objPtr->ltsTriggerNum = -1;
		}
		Tk_Release(objPtr->lts);
		objPtr->lts = NULL;
	    }

	    /*
	     * And set lock and arrange triggers for new one.
	     */
	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
		objPtr->ltsTriggerNum =
		    LTS_SetTrigger(objPtr->lts,
				   LTS_SPEED_FLAG|LTS_VALUE_FLAG|LTS_OFFSET_FLAG,
				   AudioInRsrc_LTSChange,
				   (ClientData) objPtr);
		objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr])+1);
		strcpy(objPtr->ltsName,argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }

	    /* ------------------------ OUT CMD ------------------------ */
	} else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (objPtr->outCmd != NULL) {
		ckfree(objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	    if (AudioInChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }

	    /* --------------------------- ERROR ------------------------------ */
	} else {
	    Tcl_AppendResult(interp, "Invalid configuration parameter ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}

	cntr++;
    }
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * AudioInRsrcReady
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *	Precomputes any values needed to get the object ready,
 *	and calls the device ready method, which does the same.
 *
 * -------------------------------------------------------
 */
static int
AudioInRsrcReady (AudioInRsrc *objPtr)
{
    int result;

    /*
     * Can only be made ready if
     *	1. We have a device.
     *	2. We have an lts
     */
    if (objPtr->devicePtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "Can't ready AudioIn resource \"",
			 objPtr->name, "\" ready: device type is not set.", (char *)NULL);
        return TCL_ERROR;
    }
    if (objPtr->lts == NULL) {
	Tcl_AppendResult(objPtr->interp, "Can't ready AudioIn resource \"",
			 objPtr->name, "\" ready: no lts has been attached.", (char *)NULL);
        return TCL_ERROR;
    }
    result = 0;

    /*
     * Call the device init/reinit procedure
     */
    if (!objPtr->inited) {
	if (objPtr->devicePtr->initProc != NULL) {
	    result = objPtr->devicePtr->initProc(objPtr);
	}
	if (result == 0)  {
	    objPtr->inited = 1;
	}
    } else {
	if (objPtr->devicePtr->reinitProc != NULL) {
	    result = objPtr->devicePtr->reinitProc(objPtr);
	}
    }	

    /*
     * If bad result, get error string, return error
     */
    if (result != 0) {
	char *errStr;
	
	if (objPtr->devicePtr->errorProc != NULL) {
	    errStr = objPtr->devicePtr->errorProc(objPtr, result);
	} else {
	    errStr = "Error with device initialization.";
	}
	
	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
    }
    
    objPtr->ready = 1;

    if (LTS_Speed(objPtr->lts) == 1.0) {
	AudioInScheduleFirstSend(objPtr);
    }

    return (result ? result : TCL_OK);
}


/*
 * -------------------------------------------------------
 *
 * AudioInRsrcUnready
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *	Unreadies the object,
 *	and calls the device unready method, which does the same.
 *
 * -------------------------------------------------------
 */

static int
AudioInRsrcUnready (AudioInRsrc *objPtr)
{
    int result = 0;
    
    if (objPtr->devicePtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "No device was found.", (char *)
			 NULL);
        return TCL_ERROR;
    }

    /*
     * Blow away "at" queue.
     */
    if (objPtr->atToken) {
	if (Tcm_FindAtEvent(objPtr->atToken) != NULL) {
	    Tcm_DeleteAtHandler(objPtr->atToken);
	    objPtr->atToken = 0;
	}
    }

    if (objPtr->devicePtr->unreadyProc != NULL) {
	result = objPtr->devicePtr->unreadyProc(objPtr);
    }
    
    if (result != 0) {
	char *errStr;
	if (objPtr->devicePtr->errorProc != NULL) {
	    errStr = objPtr->devicePtr->errorProc(objPtr, result);
	} else {
	    errStr = "Error in device unready routine.";
	}
	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
    }

    objPtr->ready = 0;
    
    return result;
}


/************************************************************
 *
 * AudioInChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *     newOut - new output command
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
AudioInChangeOutCmd(AudioInRsrc *objPtr, char *newOut)
{
    /*
     * Clear the previous entry. 
     */
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return. 
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set 
     * everything accordingly. 
     */
    if (newOut[0] == '@') {

	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
				       &(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			     "Could not find registered C procedure: ",
			     newOut, (char *) NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string. 
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut) + 1);
    strcpy(objPtr->outCmd, newOut);

    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * AudioInRsrc_LTSChange
 *
 * Parameters: 
 *	clientData - Really the object pointer
 * 
 * Return: 
 *	Standard Tcl return value
 * 
 * Action: 
 *	Called when the LTS changes value.  Dequeues any previously
 *	queued play requests, frees any memory we're holding, and
 *	returns.
 *
 * -------------------------------------------------------
 */
static void
AudioInRsrc_LTSChange(ClientData clientData)
{
    AudioInRsrc *objPtr = (AudioInRsrc *) clientData;
    
    /*
     * If not ready, do nothing.
     */
    if (objPtr->ready == 0) {
	return;
    }

    /*
     * Blow away "at" queue.
     */
    if (objPtr->atToken) {
	if (Tcm_FindAtEvent(objPtr->atToken) != NULL) {
	    Tcm_DeleteAtHandler(objPtr->atToken);
	    objPtr->atToken = 0;
	}
    }

    /*
     * Notify the device that the lts has changed.
     */
    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->ltsProc != NULL)) {
	objPtr->devicePtr->ltsProc(objPtr);
    }

    if (LTS_Speed(objPtr->lts) == 1.0) {
	AudioInScheduleFirstSend(objPtr);
    }

}

static void
AudioInScheduleFirstSend(AudioInRsrc *objPtr) 
{
    double logwhen, when;
    
    logwhen = LTS_Value(objPtr->lts) + objPtr->delayTime; 
    when = LTS_SystemTimeOf(objPtr->lts, logwhen);
#ifdef DEBUG_TIMING
    fprintf(stderr, "first trigger at log = %f  sys = %f\n", logwhen, when); 
#endif
    objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL, "audio",
					  AudioInSendData, objPtr, NULL,
					  NULL, NULL, NULL, NULL);
    objPtr->atWhen = logwhen; 
}

/* 
 * -------------------------------------------------------
 *
 * AudioInSendData
 *
 * Parameters:
 *     clientData - the audio input object pointer
 * Return: none
 * 
 * Action: 
 *    Reads all data that has accumulated on the audio device since the last
 * call to AudioInSendData, and then ships it out as one frame
 * and reschedules itself.
 *
 * Remarks:
 *    This is a very simplistic way of capturing audio input. Probably
 * will need to update this to buffer more data.
 * -------------------------------------------------------
 */
static void
AudioInSendData (ClientData clientData) 
{
    AudioInRsrc *objPtr = (AudioInRsrc *)clientData;
    int cnt, idx;
    int dataSegCode, dataLength;
    AuInBufInfo auInBuf;
    SparcAuBufferHeader bufHdr;   
    ScatterBuffer *sbPtr;
    SB_List *sbl;
    char *sbStr;
    double now, sTime, eTime, lsS, lsNs, leS, leNs, lognow;
    
    auInBuf.dlen = objPtr->cycleTime * objPtr->sampleRate * objPtr->sampleSize
	* objPtr->channels;
    
    /* add 8 (two ints) to length for CM_DATA_SEGMENT tag 
     *   and the length field.
     */
    auInBuf.len = sizeof(SparcAuBufferHeader) + 8 + auInBuf.dlen;
    auInBuf.objPtr = objPtr;
    auInBuf.offset = sizeof(SparcAuBufferHeader) + 8;
    auInBuf.bufPtr = MakeBuffer(objPtr->interp, 
				auInBuf.len,
				"audioInRsrc:AudioInSendData");
    
    if (objPtr->atToken != -1) {
	now = objPtr->atWhen;
    } else {
	now = LTS_Value(objPtr->lts);
    }
    objPtr->atToken = -1;
    
    
    /* When audio should be scheduled to play back */
    sTime = now + objPtr->delayTime;   
    eTime = sTime + objPtr->cycleTime;
#ifdef DEBUG_TIMING
    fprintf(stderr, "senddata: cur = %f sTime = %f  eTime = %f\n", 
	    LTS_Value(objPtr->lts), sTime, eTime); 
#endif
    /* Note: subtract 8 (two int's) from the size
     * of the buffer header for chunkFile support
     */
    
    bufHdr.hdrType = htonl(CM_SPARC_AU_BUF_HDR);
    bufHdr.device = htonl(SPARC_AU_DEV_CODE);    
    bufHdr.length = htonl(sizeof(SparcAuBufferHeader) - 8);
    bufHdr.dataFormat = htonl(objPtr->format);
    bufHdr.samplingRate = htonl(objPtr->sampleRate);
    bufHdr.channelCount = htonl(objPtr->channels);
    lsS = floor(sTime);
    lsNs = (sTime - floor(sTime)) * ONE_BILLION;
    leS = floor(eTime);
    leNs = (eTime - floor(eTime)) * ONE_BILLION;
    bufHdr.logStart_s = htonl((int) lsS);
    bufHdr.logStart_ns = htonl((int) lsNs);
    bufHdr.logEnd_s = htonl((int) leS);
    bufHdr.logEnd_ns = htonl((int) leNs);
    
    idx = 0;
    
    memcpy(auInBuf.bufPtr->data, &bufHdr, sizeof(SparcAuBufferHeader));
    idx += sizeof(SparcAuBufferHeader);
    
    dataSegCode = htonl(CM_DATA_SEGMENT);
    dataLength = htonl(auInBuf.dlen);
    
    memcpy(auInBuf.bufPtr->data + idx, &dataSegCode, 4);
    idx += 4;
    
    memcpy(auInBuf.bufPtr->data + idx, &dataLength, 4);
    idx += 4;
    
    /* invoke device dependent recording mechanism to actually grab 
     * some data
     */
    cnt = objPtr->devicePtr->recordProc(&auInBuf);
    
    idx += cnt;
    
    sbPtr = MakeScatterBuffer(auInBuf.bufPtr, idx, NULL);
    sbl = MakeSBList(sbPtr, NULL);
    
    
    /*
     * If out command is C, call it directly
     */
    if (objPtr->outCCmd != NULL) {

	/*
	 * Need to check if it still exists since object may have been
	 * destroyed.
	 */
	if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd + 1) != NULL) {
	    objPtr->outCCmd(objPtr->outCD, sbl);
	} else {
	    objPtr->outCCmd = NULL;
	    ckfree(objPtr->outCmd);
	    objPtr->outCmd = NULL;
	}

    } else {
	if (objPtr->outCmd != NULL) {
	    /*
	     * Out command is tcl, make sb list into a string and call using
	     * Tcl_Eval. 
	     */
	    sbStr = MakeSBListString(sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL);
	    ckfree(sbStr);
	    FreeSBList(sbl);
	}
    }
    
    /* we're done with the buffer. */
    DetachBuffer(objPtr->interp, auInBuf.bufPtr, "audioInRsrc:AudioInSendData");
    
    now = LTS_SystemTimeOf(objPtr->lts, objPtr->atWhen + objPtr->cycleTime);
    lognow = LTS_LogicalTimeOf(objPtr->lts, now);
#ifdef DEBUG_TIMING
    fprintf(stderr, "next trigger at log = %f  sys = %f  (cur = %f)\n", lognow, now, 
	    LTS_Value(objPtr->lts));
#endif
    objPtr->atToken = Tcm_CreateAtHandler(now, HUGE_VAL, "audio",
					  AudioInSendData, objPtr, NULL,
					  NULL, NULL, NULL, NULL);
    
    objPtr->atWhen = objPtr->atWhen + objPtr->cycleTime;
}


/*
 * -------------------------------------------------------
 *
 * AudioInRsrcStats
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *
 * -------------------------------------------------------
 */

static int
AudioInRsrcStats (AudioInRsrc *objPtr)
{
    Tcl_AppendResult(objPtr->interp, "This command is not yet implemented.",
		     NULL);
    return TCL_OK;
}

/******************************************************************
 *
 * AudioInResourceDeviceList
 *
 * Parameters:
 *     objPtr - pointer to audio input object.
 *
 * Results:
 *     Standard Tcl Result.
 *
 * Actions:
 *     Sets result to list of devices supported. Returns TCL_OK;
 *
 ******************************************************************/
static int
AudioInResourceDeviceList(AudioInRsrc *objPtr)
{
    Tcl_Interp *interp = objPtr->interp;
    Tcl_DString resStr;

    Tcl_DStringInit(&resStr);

#ifdef AUDIO_FILE
    Tcl_DStringAppendElement(&resStr, "af");
#endif
#ifdef SPARC_AU
    Tcl_DStringAppendElement(&resStr, "sparc");
#endif

    Tcl_SetResult(interp, Tcl_DStringValue(&resStr), TCL_VOLATILE);
    Tcl_DStringFree(&resStr);

    return TCL_OK;
}
