/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * This file contains the implementation of the cmBind command.
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <tcl.h>
#include <tk.h>
#include "dp.h"
#include "cmUtil.h"
#include "cmBind.h"

typedef struct {
    char *substSymbol;
    CM_BindProc *substProc;
    ClientData substClientdata;
} SubstRecord;

typedef struct {
    int isSubstFlag;
    CM_BindProc *substProc;
    ClientData substCD;
    char *componentBody;
    int componentLength;
} BindScriptComponent;

typedef struct BindScript {
    char *origScript;
    int numToSkip;
    double blankPeriod;
    int skipCounter;
    double lastTime;
    int numComponents;
    BindScriptComponent *components;
    struct BindScript *nextScript;
} BindScript;

typedef struct {
    BindScript *scripts;
} CMEventBinding;

typedef struct {
    int bindingId;
    char *objectName;
    Tcl_Interp *interp;
    Tcl_HashTable eventTable;
    Tcl_HashTable substTable;
} CMBinding;

static int bindingArraySize;
static CMBinding **bindingArray;
static Tcl_HashTable nameToBindingTable;
static Tcl_HashTable genericSubstTable;
static BindScript *currentEventScript;

static int CM_BindCmd(ClientData clientdata, Tcl_Interp *interp,
		      int argc, char **argv);
static int CM_UnbindCmd(ClientData clientdata, Tcl_Interp *interp,
			int arc, char **argv);

static void FreeEventScriptList(BindScript *scriptList);
static void ExecuteEventScript(Tcl_Interp *interp, BindScript *script);
static void InitGenericSubstTable(Tcl_Interp *interp, Tcl_HashTable *genSubstTable);
static char *GetNumSkipped(Tcl_Interp *interp, ClientData clientdata);
static char *GetCurrTime(Tcl_Interp *interp, ClientData clientdata);

int
InitCMBindModule(Tcl_Interp *interp) {
    
    bindingArray = NULL;
    bindingArraySize = 0;
    Tcl_InitHashTable(&nameToBindingTable, TCL_STRING_KEYS);
    Tcl_InitHashTable(&genericSubstTable, TCL_STRING_KEYS);

    InitGenericSubstTable(interp, &genericSubstTable);

    Tcl_CreateCommand(interp, "cmBind", CM_BindCmd, NULL, NULL);
    Tcl_CreateCommand(interp, "cmUnbind", CM_UnbindCmd, NULL, NULL);

    return TCL_OK;
}

int 
CreateCMBinding(Tcl_Interp *interp,
		char *objectName,
		int eventListLength,
		char **eventList,
		int substListLength,
		char **substSymbols,
		CM_BindProc **substProcs,
		ClientData *substClientdatas)
{
    int new_binding;
    Tcl_HashEntry *entryPtr;
    int binding_index;
    CMBinding *bindingPtr;
    int event_cntr, subst_cntr;

    entryPtr = Tcl_CreateHashEntry(&nameToBindingTable, objectName, &new_binding);

    if (!new_binding) {
	return TCL_ERROR;
    }

    for (binding_index = 0; binding_index < bindingArraySize; binding_index++) {
	if (bindingArray[binding_index] == NULL)
	  break;
    }

    if (binding_index == bindingArraySize) {
	if (bindingArraySize == 0) {
	    int i;

	    bindingArraySize = 50;
	    bindingArray = (CMBinding **) ckalloc(sizeof(CMBinding *)*bindingArraySize);

	    for (i=0; i<50; i++) {
		bindingArray[i] = NULL;
	    }
	    binding_index = 0;
	} else {
	    bindingArraySize++;
	    bindingArray = (CMBinding **) ckrealloc((char *) bindingArray,
						   sizeof(CMBinding *)*bindingArraySize);
	    binding_index = bindingArraySize-1;
	}
    }

    bindingArray[binding_index] = (CMBinding *) ckalloc(sizeof(CMBinding));
    bindingPtr = bindingArray[binding_index];

    bindingPtr->bindingId = binding_index;
    bindingPtr->objectName = (char *) ckalloc(strlen(objectName)+1);
    strcpy(bindingPtr->objectName, objectName);
    bindingPtr->interp = interp;
    Tcl_InitHashTable(&(bindingPtr->eventTable), TCL_STRING_KEYS);
    Tcl_InitHashTable(&(bindingPtr->substTable), TCL_STRING_KEYS);

    for (event_cntr = 0; event_cntr < eventListLength; event_cntr++) {
	Tcl_HashEntry *event_entry_ptr;
	int is_new;
	CMEventBinding *new_cm_event_binding;

	event_entry_ptr = Tcl_CreateHashEntry(&(bindingPtr->eventTable), 
				       eventList[event_cntr],
				       &is_new);

	if (is_new) {
	    new_cm_event_binding = (CMEventBinding *) ckalloc(sizeof(CMEventBinding));
	    new_cm_event_binding->scripts = NULL;

	    Tcl_SetHashValue(event_entry_ptr, (ClientData) new_cm_event_binding);
	}
    }

    for (subst_cntr = 0; subst_cntr < substListLength; subst_cntr++) {
	Tcl_HashEntry *subst_entry_ptr;
	int is_new;
	SubstRecord *new_subst_record;
	
	subst_entry_ptr = Tcl_CreateHashEntry(&(bindingPtr->substTable),
				       substSymbols[subst_cntr],
				       &is_new);

	if (is_new) {
	    new_subst_record = (SubstRecord *) ckalloc(sizeof(SubstRecord));

	    new_subst_record->substSymbol = (char *) 
	      ckalloc(strlen(substSymbols[subst_cntr] )+1);
	    strcpy(new_subst_record->substSymbol,
		   substSymbols[subst_cntr]);
	    new_subst_record->substProc = substProcs[subst_cntr];
	    new_subst_record->substClientdata = substClientdatas[subst_cntr];

	    Tcl_SetHashValue(subst_entry_ptr, (ClientData) new_subst_record);
	}
    }

    Tcl_SetHashValue(entryPtr, (ClientData) bindingPtr->bindingId);

    return bindingPtr->bindingId;
}
	
int
DestroyCMBinding(Tcl_Interp *interp, int bindingId)
{
    CMBinding *binding_ptr;
    Tcl_HashEntry *entry_ptr;
    Tcl_HashSearch hash_search;

    if (bindingId >= bindingArraySize) {
	return TCL_ERROR;
    }

    binding_ptr = bindingArray[bindingId];

    if (binding_ptr == NULL) {
	return TCL_OK;
    }

    entry_ptr = Tcl_FindHashEntry(&nameToBindingTable, binding_ptr->objectName);
    if (entry_ptr != NULL) {
	Tcl_DeleteHashEntry(entry_ptr);
    }

    entry_ptr = Tcl_FirstHashEntry(&(binding_ptr->substTable), &hash_search);

    while (entry_ptr != NULL) {
	SubstRecord *subst_rec;

	subst_rec = (SubstRecord *) Tcl_GetHashValue(entry_ptr);
	
	if (subst_rec->substSymbol != NULL) {
	    ckfree(subst_rec->substSymbol);
	}
	ckfree((char *) subst_rec);

	Tcl_DeleteHashEntry(entry_ptr);

	entry_ptr = Tcl_NextHashEntry(&hash_search);
    }

    entry_ptr = Tcl_FirstHashEntry(&(binding_ptr->eventTable), &hash_search);

    while (entry_ptr != NULL) {
	CMEventBinding *event_ptr;

	event_ptr = (CMEventBinding *) Tcl_GetHashValue(entry_ptr);

	FreeEventScriptList(event_ptr->scripts);

	ckfree((char *) event_ptr);

	Tcl_DeleteHashEntry(entry_ptr);

	entry_ptr = Tcl_NextHashEntry(&hash_search);
    }

    ckfree(binding_ptr->objectName);
    ckfree((char *) binding_ptr);

    bindingArray[bindingId] = NULL;

    return TCL_OK;
}

int
ExecuteCMBinding(Tcl_Interp *interp, int bindingId, char *eventSymbol)
{
    CMBinding *binding_ptr;
    Tcl_HashEntry *event_entry;
    CMEventBinding *event_binding_ptr;
    BindScript *event_script;

    if (bindingId >= bindingArraySize) {
	return TCL_ERROR;
    }

    binding_ptr = bindingArray[bindingId];

    if (binding_ptr == NULL) {
	return TCL_OK;
    }

    event_entry = Tcl_FindHashEntry(&(binding_ptr->eventTable), eventSymbol);

    if (event_entry == NULL) {
	return TCL_OK;
    }

    event_binding_ptr = (CMEventBinding *) Tcl_GetHashValue(event_entry);

    event_script = event_binding_ptr->scripts;

    while (event_script != NULL) {
	ExecuteEventScript(binding_ptr->interp, event_script);

	event_script = event_script->nextScript;
    }

    return TCL_OK;
}
    
static void
FreeEventScriptList(BindScript *scriptList) 
{
    BindScript *script_ptr;
    BindScript *next_script;

    script_ptr = scriptList;

    while(script_ptr != NULL) {
	ckfree(script_ptr->origScript);
	ckfree((char *) script_ptr->components);

	next_script = script_ptr->nextScript;

	ckfree((char *) script_ptr);

	script_ptr = next_script;
    }
}

static void
ExecuteEventScript(Tcl_Interp *interp, BindScript *script)
{
    int comp_cntr;
    Tcl_DString script_string;

    if (script->numComponents < 1) return;

    script->skipCounter++;
    if (script->skipCounter <= script->numToSkip) {
      return;
    }

    if (script->blankPeriod != 0.0) {
      double this_time;

      this_time = Tcm_ReadSysClock();
      if (this_time - script->lastTime < script->blankPeriod) {
	return;
      }

      script->lastTime = this_time;
    }

    currentEventScript = script;

    Tcl_DStringInit(&script_string);

    for (comp_cntr=0; comp_cntr<script->numComponents; comp_cntr++) {
	BindScriptComponent *comp_ptr;

	comp_ptr = &(script->components[comp_cntr]);

	if (comp_ptr->isSubstFlag) {
	    char *resStr;

	    resStr = comp_ptr->substProc(interp, comp_ptr->substCD);
	    
	    Tcl_DStringAppend(&script_string, resStr, -1);
    
	    ckfree(resStr);
	} else {
	    
	    Tcl_DStringAppend(&script_string, comp_ptr->componentBody, 
			      comp_ptr->componentLength);
	}
    }

    Tcl_GlobalEval(interp, Tcl_DStringValue(&script_string));

    Tcl_DStringFree(&script_string);

    script->skipCounter = 0;

    return;
}

static int CM_UnbindCmd(ClientData clientdata, 
			Tcl_Interp *interp,
			int argc, char **argv)
{
  char *object_name, *event_symbol, *script_body;
  Tcl_HashEntry *entryPtr, *event_entry;
  CMBinding *binding_ptr;
  CMEventBinding *event_binding;
  BindScript *script_ptr, *prev_script;
  int binding_id;

  if (argc < 4) {
    Tcl_AppendResult(interp, "Wrong # of arguments, should be: \"",
		     argv[0], " <objectName> <eventSymbol> <tclScript>\"",
		     (char *) NULL);
    return TCL_ERROR;
  }

  object_name = argv[1];
  event_symbol = argv[2];
  script_body = argv[3];

  entryPtr = Tcl_FindHashEntry(&nameToBindingTable, object_name);
  
  if (entryPtr == NULL) {
    return TCL_OK;
  }

  binding_id = (int) Tcl_GetHashValue(entryPtr);
  binding_ptr = bindingArray[binding_id];

  if (binding_ptr == NULL) 
    return TCL_OK;

  event_entry = Tcl_FindHashEntry(&(binding_ptr->eventTable), event_symbol);

  if (event_entry == NULL) {
    return TCL_OK;
  }

  event_binding = (CMEventBinding *) Tcl_GetHashValue(event_entry);

  if (event_binding == NULL) {
    return TCL_OK;
  }

  script_ptr = event_binding->scripts;
  prev_script = NULL;

  while (script_ptr != NULL) {
    if (strcmp(script_ptr->origScript, script_body) == 0) {
      ckfree(script_ptr->origScript);
      ckfree((char *) script_ptr->components);
      
      if (prev_script == NULL) {
	event_binding->scripts = script_ptr->nextScript;
      } else {
	prev_script->nextScript = script_ptr->nextScript;
      }
      ckfree((char *) script_ptr);
      return TCL_OK;
    }
    prev_script = script_ptr;
    script_ptr = script_ptr->nextScript;
  }

  return TCL_OK;
}

  

static int CM_BindCmd(ClientData clientdata, 
		      Tcl_Interp *interp,
		      int argc, char **argv)
{
    char *object_name, *event_symbol, *script_body;
    int append_flag, num_to_skip, counter;
    double blank_period;
    int len;
    Tcl_HashEntry *entryPtr, *event_entry, *subst_entry;
    CMBinding *binding_ptr;
    CMEventBinding *event_binding_ptr;
    BindScriptComponent *comp_array;
    int num_comp, sidx;
    SubstRecord *subst_rec;
    BindScript *new_bind_script;
    int binding_id;

    
    if (argc < 2) {
	Tcl_AppendResult(interp, "Wrong # of arguments, should be: \"",
			 argv[0], " <objectName> <eventSymbol> ?<tclScript> ",
			 "?-blankPeriod <numSec>? ?-skip <numToSkip>??\"",
			 (char *) NULL);
	return TCL_ERROR;
    }

    object_name = argv[1];

    entryPtr =  Tcl_FindHashEntry(&nameToBindingTable, object_name);

    if (entryPtr == NULL) {
	Tcl_AppendResult(interp, "Object \"", object_name, "\" has not registered any",
			 " CM bindings.", (char *) NULL);
	return TCL_ERROR;
    }
    
    binding_id = (int) Tcl_GetHashValue(entryPtr);
    binding_ptr = bindingArray[binding_id];
    
    if (binding_ptr == NULL) {
	return TCL_OK;
    }
    
    if (argc == 2) {
	Tcl_HashSearch hash_search;
	Tcl_DString dstr;

	Tcl_DStringInit(&dstr);

	event_entry = Tcl_FirstHashEntry(&(binding_ptr->eventTable), &hash_search);
	
	while (event_entry != NULL) {
	    Tcl_DStringAppendElement(&dstr, Tcl_GetHashKey(&(binding_ptr->eventTable),
							   event_entry));
	    event_entry = Tcl_NextHashEntry(&hash_search);
	}

	Tcl_DStringResult(interp, &dstr);
	Tcl_DStringFree(&dstr);

	return TCL_OK;
    }

    event_symbol = argv[2];
    event_entry = Tcl_FindHashEntry(&(binding_ptr->eventTable), event_symbol);
    
    if (event_entry == NULL) {
	Tcl_AppendResult(interp, "Event symbol \"", event_symbol, "\" is not valid.",
			 (char *) NULL);
	return TCL_ERROR;
    }
    event_binding_ptr = (CMEventBinding *) Tcl_GetHashValue(event_entry);


    if (argc == 3) {
	Tcl_DString dstr;
	BindScript *script_marker;

	Tcl_DStringInit(&dstr);

	script_marker = event_binding_ptr->scripts;

	while (script_marker != NULL) {
	    Tcl_DStringAppendElement(&dstr, script_marker->origScript);
	    script_marker = script_marker->nextScript;
	}

	Tcl_DStringResult(interp, &dstr);
	Tcl_DStringFree(&dstr);

	return TCL_OK;
    }

    script_body = (char *) ckalloc(strlen(argv[3])+1);
    strcpy(script_body, argv[3]);

    append_flag = 0;

    if (*script_body == '+') {
	script_body++;
	append_flag = 1;
    }

    blank_period = 0.0;
    num_to_skip = 0;
    if (argc > 4) {
	counter = 4;

	while (counter < argc) {
	    len = strlen(argv[counter]);

	    if (strncmp(argv[counter], "-blankPeriod", len) == 0) {
		counter++;
		
		if (counter == argc) {
		    Tcl_AppendResult(interp, "Need value after -blankPeriod option.",
				     (char *) NULL);
		    goto error;
		}
		
		if (Tcl_GetDouble(interp, argv[counter], &blank_period) != TCL_OK) {
		    Tcl_AppendResult(interp, "Bad value given after -blankPeriod option.",
				     (char *) NULL);
		    goto error;
		}
		
	    } else if (strncmp(argv[counter], "-skip", len) == 0) {
		counter++;
		
		if (counter == argc) {
		    Tcl_AppendResult(interp, "Need value after -skip option.",
				     (char *) NULL);
		    goto error;
		}
		
		if (Tcl_GetInt(interp, argv[counter], &num_to_skip) != TCL_OK) {
		    Tcl_AppendResult(interp, "Bad value given after -skip option.",
				     (char *) NULL);
		    goto error;
		}
	    }

	    counter++;
	}
    }

    /* 
     * Scan the script counting how many substitions are called for and
     * resolving them. 
     */

    if (strlen(script_body) > 0) {
	comp_array = (BindScriptComponent *) ckalloc(sizeof(BindScriptComponent));
	num_comp = 1;
	
	comp_array[0].isSubstFlag = 0;
	comp_array[0].substProc = NULL;
	comp_array[0].substCD = NULL;
	comp_array[0].componentBody = script_body;
	
	for (sidx = 0; sidx < strlen(script_body); sidx++) {
	    if (script_body[sidx] == '%') {
		
		/* End previous component. */
		
		comp_array[num_comp-1].componentLength = 
		  &(script_body[sidx]) - comp_array[num_comp-1].componentBody;
		
		sidx++;
		
		if (script_body[sidx] != '%') {
		    char subst_str[4];

		    subst_str[0] = script_body[sidx];
		    subst_str[1] = '\0';
		    
		    subst_entry = Tcl_FindHashEntry(&(binding_ptr->substTable), 
						    subst_str);
		    
		    if (subst_entry == NULL) {
		      subst_entry = Tcl_FindHashEntry(&genericSubstTable,
						      subst_str);

		      if (subst_entry == NULL) {
			ckfree((char *) comp_array);
			Tcl_AppendResult(interp, "Illegal substitution marker: %", 
					 subst_str, (char *) NULL);
			goto error;
		      }
		    }
		    
		    subst_rec = (SubstRecord *) Tcl_GetHashValue(subst_entry);
		    num_comp++;
		    comp_array = (BindScriptComponent *) 
		      ckrealloc((char *) comp_array, sizeof(BindScriptComponent)*num_comp);
		    
		    comp_array[num_comp-1].isSubstFlag = 1;
		    comp_array[num_comp-1].substProc = subst_rec->substProc;
		    comp_array[num_comp-1].substCD = subst_rec->substClientdata;
		    comp_array[num_comp-1].componentBody = NULL;
		    comp_array[num_comp-1].componentLength = 0;
		    
		    sidx++;
		}
		
		num_comp++;
		comp_array = (BindScriptComponent *) 
		  ckrealloc((char *) comp_array, sizeof(BindScriptComponent)*num_comp);
		
		comp_array[num_comp-1].isSubstFlag = 0;
		comp_array[num_comp-1].substProc = NULL;
		comp_array[num_comp-1].substCD = NULL;
		comp_array[num_comp-1].componentBody = &(script_body[sidx]);
	    }
	}

	comp_array[num_comp-1].componentLength = 
	  &(script_body[sidx]) - comp_array[num_comp-1].componentBody;

    }

    if (!append_flag) {
	
	if (event_binding_ptr->scripts != NULL) {
	    FreeEventScriptList(event_binding_ptr->scripts);
	    event_binding_ptr->scripts = NULL;
	}
    }

    if (strlen(script_body) == 0) {
	return TCL_OK;
    }

    new_bind_script = (BindScript *) ckalloc(sizeof(BindScript));

    new_bind_script->origScript = script_body;
    new_bind_script->numToSkip = num_to_skip;
    new_bind_script->blankPeriod = blank_period;
    new_bind_script->skipCounter = 0;
    new_bind_script->lastTime = Tcm_ReadSysClock();
    new_bind_script->numComponents = num_comp;
    new_bind_script->components = comp_array;
    new_bind_script->nextScript = event_binding_ptr->scripts;

    event_binding_ptr->scripts = new_bind_script;

    return TCL_OK;

  error:
    ckfree(script_body);
    return TCL_ERROR;
}

static void 
InitGenericSubstTable(Tcl_Interp *interp, Tcl_HashTable *genSubstTable)
{
  int dummy;
  Tcl_HashEntry *entry_ptr;
  SubstRecord *subst_rec;

  /* 
   * %N => num of events since last invocation of this script.
   * Depends on static variable currentEventScript being set before
   * substitutions are done.
   */

  entry_ptr = Tcl_CreateHashEntry(genSubstTable, "N", &dummy);
  
  subst_rec = (SubstRecord *) ckalloc(sizeof(SubstRecord));

  subst_rec->substSymbol = (char *) ckalloc(strlen("N")+1);
  strcpy(subst_rec->substSymbol, "N");
  
  subst_rec->substProc = GetNumSkipped;
  subst_rec->substClientdata = (ClientData) &currentEventScript;
  
  Tcl_SetHashValue(entry_ptr, (ClientData) subst_rec);

  /*
   * %D => current time.
   */

  entry_ptr = Tcl_CreateHashEntry(genSubstTable, "D", &dummy);
  
  subst_rec = (SubstRecord *) ckalloc(sizeof(SubstRecord));

  subst_rec->substSymbol = (char *) ckalloc(strlen("D")+1);
  strcpy(subst_rec->substSymbol, "D");
  
  subst_rec->substProc = GetCurrTime;
  subst_rec->substClientdata = NULL;
  
  Tcl_SetHashValue(entry_ptr, (ClientData) subst_rec);
}
  
static char *
GetNumSkipped(Tcl_Interp *interp, ClientData clientdata)
{
  char *tempStr;
  BindScript **curScript = (BindScript **) clientdata;

  tempStr = (char *) ckalloc(50);

  sprintf(tempStr, "%d", (*curScript)->skipCounter);

  return tempStr;
}

static char *
GetCurrTime(Tcl_Interp *interp, ClientData clientdata)
{
  char *tempStr;
  
  tempStr = (char *) ckalloc(50);

  sprintf(tempStr, "%f", Tcm_ReadSysClock());

  return tempStr;
}

