/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/* 
 *  This is the general bit parsing module.
 *  
 *  The module works like this:
 *      
 *      bp_Start(unsigned char *data, int offset) : Indicates that bit parsing is
 *                                         to be done from the byte pointed
 *                                         to by data and starting at bit 
 *                                         position indicated by offset (0-7)
 *                                         where 0 is the most sig. bit and
 *                                         7 is the least.
 *
 *      bp_Get(int num_bits, int *returnPtr) : retrieves  the next num_bits
 *                                             and puts them in *returnPtr.
 *                                             num_bits must be <= 32.
 *
 *      bp_Peek(int num_bits, int *returnPtr) : peeks at (i.e. doesn't parse off)
 *                                              nexts num_bits and puts them in
 *                                              *returnPtr. Again, num_bits must be
 *                                              <= 32.
 *
 *      bp_Flush(int num_bits) : Flushes (i.e., skips) the next num_bits. Num_bit
 *                               can be anything.
 *
 *      bp_End(unsigned char **data, int *offset) : Indicates bit parsing is finished and
 *                                         sets **data and *offset to the next
 *                                         byte to be parsed and bit position suitable
 *                                         for some future call to bp_Start.
 *
 *  NOTE: these macros do NOTHING to ensure the validity of the pointers. If you
 *        ask for more bits that the data pointer points to, results will be
 *        unpredictable.
 *
 *        Secondly, in order to deal with JPEG data correctly (i.e. FF followed by
 *        00 validates the FF as NOT a marker and the 00 is not actual data), you
 *        should define BP_JPEG_FLAG before including this file.
 *
 *        Thirdly, the bp macros work by caching data in static global variables
 *        which mean that you can't use the macros across .c files, but it does
 *        mean you don't have to call dp_End unless you want the state info (i.e.
 *        what byte and offset you're currently on). The static variables will be
 *        safe from procedures not in the same .c file.
 *     
 *
 */

#ifndef BP_JPEG_FLAG

#include <sys/types.h>
#include <netinet/in.h>

static unsigned int bp_curBits;
static unsigned int bp_bitOffset;
static unsigned int *bp_bitBuffer;

#define bp_Start(dataPtr, offset)            \
{                                            \
    int align_error;                                                        \
    int shift;                                                              \
    unsigned char *udataPtr;                                                \
                                                                            \
    udataPtr = dataPtr;                                                     \
					                                    \
    align_error = ((int) udataPtr)%4;                                       \
    bp_curBits = 0;                                                         \
    if (align_error) {                                                      \
       shift = 24;                                                          \
       for(; align_error%4 != 0; align_error++) {                           \
          bp_curBits |= (((unsigned int) *udataPtr) << shift);               \
          udataPtr++;                                                       \
          shift -= 8;                                                       \
       }                                                                    \
       bp_curBits <<= offset;                                   \
       bp_bitOffset = 8 + shift + offset;                                   \
    } else {                                                                \
       bp_curBits = ((((unsigned int) (udataPtr[0])) << 24) |               \
		     (((unsigned int) (udataPtr[1])) << 16) |               \
		     (((unsigned int) (udataPtr[2])) << 8)  |               \
		     (((unsigned int) (udataPtr[3]))));                     \
       udataPtr += 4;                                                       \
       bp_curBits <<= offset;                                   \
       bp_bitOffset = offset;                                               \
    }                                                                       \
                                                                            \
    bp_bitBuffer = (unsigned int *) udataPtr;                               \
}

#define bp_End(dataPtrPtr, offsetPtr)   \
{                                       \
    *(dataPtrPtr) = (char *) bp_bitBuffer;\
    if (bp_bitOffset < 8)               \
       *(dataPtrPtr) -= 4;               \
    else if (bp_bitOffset < 16) {       \
       *(dataPtrPtr) -= 3;               \
       bp_bitOffset -= 8;               \
    } else if (bp_bitOffset < 24) {     \
       *(dataPtrPtr) -= 2;               \
       bp_bitOffset -= 16;              \
    } else if (bp_bitOffset < 32) {     \
       *(dataPtrPtr) -= 1;               \
       bp_bitOffset -= 24;              \
    } else {                            \
       bp_bitOffset -= 32;              \
   }                                    \
    *(offsetPtr) = bp_bitOffset;        \
}

#define bp_GetBitsX(num, shift,  resultPtr)                              \
{                                                                         \
  bp_bitOffset += num;                                                       \
                                                                          \
  if (bp_bitOffset >= 32) {                                                 \
    bp_bitOffset -= 32;                                                      \
    if (bp_bitOffset) {                                                      \
      bp_curBits |= (((unsigned int) (htonl(*bp_bitBuffer))) >> (num - bp_bitOffset)); \
    }                                                                     \
    *(resultPtr) = (bp_curBits >> shift);                              \
    bp_curBits = (htonl(*bp_bitBuffer)) << bp_bitOffset;                 \
    bp_bitBuffer++;                                                       \
  } else {                                                                  \
   *(resultPtr) = (bp_curBits >> shift);                               \
   bp_curBits <<= num;                                                       \
  }                                                                       \
}

#define bp_Get(num_bits, returnPtr) bp_GetBitsX((num_bits), (32-(num_bits)), returnPtr)

#define bp_PeekBitsX(num, shift,  resultPtr)                           \
{                                                                       \
  int bO;                                                               \
  bO = bp_bitOffset + num;                                                 \
  if (bO > 32) {                                                        \
    bO -= 32;                                                           \
    *(resultPtr) = (bp_curBits >> shift) |                              \
                (((unsigned int) (htonl(*bp_bitBuffer))) >> (shift + (num - bO)));\
  }                                                                     \
  else {                                                                \
    *(resultPtr) = (bp_curBits >> shift);                               \
  }                                                                     \
}

#define bp_Peek(num_bits, returnPtr) bp_PeekBitsX((num_bits), (32-(num_bits)), returnPtr)

#define bp_Flush(num_bits)                                       \
{                                                                \
   bp_bitOffset += (num_bits);                                   \
   if (bp_bitOffset > 31) {                                      \
     while (bp_bitOffset > 31) {                                 \
       bp_bitOffset -= 32;                                       \
       bp_bitBuffer++;                                           \
     }                                                           \
     bp_curBits = (htonl(*(bp_bitBuffer-1))) << bp_bitOffset;    \
   } else {                                                      \
       bp_curBits <<= (num_bits);                                \
   }                                                             \
}

#else

/* 
 * The following routines are tailored for jpeg data. Here we must detect
 * the case of FF00 and ignore the 00 after the FF. This deals with false
 * markers (i.e. instances of 8 byte aligned 1's that look like a marker
 * but aren't).
 */

static unsigned int bp_curBits;
static unsigned int bp_nextBits;
static unsigned int bp_bitOffset;
static unsigned char *bp_bitBuffer;

#define bp_Start(dataPtr, offset)                          \
{                                                          \
                                                           \
    int shift;                                             \
                                                           \
    bp_bitBuffer = (unsigned char *) dataPtr;              \
    bp_bitOffset = offset;                                 \
                                                           \
    bp_curBits = 0;                                        \
    shift = 24;                                            \
                                                           \
    while (shift >= 0) {                                   \
       bp_curBits |= ((int) *bp_bitBuffer) << shift;       \
       shift -= 8;                                         \
       if (*bp_bitBuffer == 0xff) {                        \
	 if (*(bp_bitBuffer+1) == 0x00) {                  \
	   bp_bitBuffer += 2;                              \
	 } else {                                          \
	   bp_bitBuffer++;                                 \
	 }                                                 \
       } else {                                            \
	 bp_bitBuffer++;                                   \
       }                                                   \
     }                                                     \
                                                           \
     shift = 24;                                           \
     bp_nextBits = 0;                                      \
     while (shift >= 0) {                                  \
       bp_nextBits |= ((int) *bp_bitBuffer) << shift;      \
       shift -= 8;                                         \
       if (*bp_bitBuffer == 0xff) {                        \
	 if (*(bp_bitBuffer+1) == 0x00) {                  \
	   bp_bitBuffer += 2;                              \
	 } else {                                          \
	   bp_bitBuffer++;                                 \
	 }                                                 \
       } else {                                            \
	 bp_bitBuffer++;                                   \
       }                                                   \
     }                                                     \
}

#define bp_End(dataPtrPtr, offsetPtr)   \
{                                       \
                                        \
    int i;                              \
                                        \
    for (i=0; i<4; i++) {               \
      bp_bitBuffer--;                   \
      if (*bp_bitBuffer == 0x00) {      \
	if (*(bp_bitBuffer-1) == 0xff)  \
	  bp_bitBuffer--;               \
      }                                 \
    }                                   \
                                        \
    if (bp_bitOffset < 8)               \
       i = 4;                           \
    else if (bp_bitOffset < 16) {       \
       i = 3;                           \
       bp_bitOffset -= 8;               \
    } else if (bp_bitOffset < 24) {     \
       i = 2;                           \
       bp_bitOffset -= 16;              \
    } else if (bp_bitOffset < 32) {     \
       i = 1;                           \
       bp_bitOffset -= 24;              \
    } else {                            \
       i = 0;                           \
       bp_bitOffset -= 32;              \
   }                                    \
                                        \
   while (i > 0) {                      \
     bp_bitBuffer--;                    \
      if (*bp_bitBuffer == 0x00) {      \
	if (*(bp_bitBuffer-1) == 0xff)  \
	  bp_bitBuffer--;               \
      }                                 \
     i--;                               \
    }                                   \
                                        \
    *(dataPtrPtr) = bp_bitBuffer;       \
    *(offsetPtr) = bp_bitOffset;        \
}

#define bp_GetBitsX(num, off_shift,  resultPtr)                        \
{                                                                      \
  bp_bitOffset += num;                                                 \
                                                                       \
  if (bp_bitOffset >= 32) {                                            \
    int shift;                                                         \
    bp_bitOffset -= 32;                                                \
    if (bp_bitOffset) {                                                \
        bp_curBits |= (bp_nextBits >> (num - bp_bitOffset));           \
    }                                                                  \
    *(resultPtr) = (bp_curBits >> off_shift);                          \
    bp_curBits = bp_nextBits << bp_bitOffset;                          \
    shift = 24;                                                        \
    bp_nextBits = 0;                                                   \
    while (shift >= 0) {                                               \
	bp_nextBits |= ((int) *bp_bitBuffer) << shift;                 \
	shift -= 8;                                                    \
	if (*bp_bitBuffer == 0xff) {                                   \
	    if (*(bp_bitBuffer+1) == 0x00) {                           \
		bp_bitBuffer += 2;                                     \
	    } else {                                                   \
		bp_bitBuffer++;                                        \
	    }                                                          \
	 } else {                                                      \
	     bp_bitBuffer++;                                           \
	 }                                                             \
    }                                                                  \
  } else {                                                             \
     *(resultPtr) = (bp_curBits >> off_shift);                         \
     bp_curBits <<= num;                                               \
  }                                                                    \
}

#define bp_Get(num_bits, returnPtr) bp_GetBitsX((num_bits), (32-(num_bits)), returnPtr)

#define bp_PeekBitsX(num, shift,  resultPtr)                    \
{                                                               \
  int bO;                                                       \
  bO = bp_bitOffset + num;                                      \
  if (bO > 32) {                                                \
      bO -= 32;                                                 \
      *(resultPtr) = (bp_curBits >> shift) |                    \
                     (bp_nextBits >> (shift + (num - bO)));     \
  } else {                                                      \
      *(resultPtr) = (bp_curBits >> shift);                     \
  }                                                             \
}

#define bp_Peek(num_bits, returnPtr) bp_PeekBitsX((num_bits), (32-(num_bits)), returnPtr)

#define bp_Flush(num)                                                     \
{                                                                         \
  bp_bitOffset += num;                                                    \
                                                                          \
  if (bp_bitOffset > 31) {                                                \
    if (bp_bitOffset < 64) {                                              \
      bp_bitOffset -= 32;                                                 \
      bp_curBits = bp_nextBits << bp_bitOffset;                           \
      if ((bp_bitBuffer[0] != 0xff) && (bp_bitBuffer[1] != 0xff) &&       \
	  (bp_bitBuffer[2] != 0xff) && (bp_bitBuffer[3] != 0xff)) {       \
	bp_nextBits = (bp_bitBuffer[0]<<24) |                             \
		      (bp_bitBuffer[1]<<16) |                             \
		      (bp_bitBuffer[2]<< 8) |                             \
		       bp_bitBuffer[3];                                   \
	bp_bitBuffer+= 4;                                                 \
      } else {                                                            \
        int shift;                                                        \
	shift = 24;                                                       \
        bp_nextBits = 0;                                                  \
	while (shift >= 0) {                                              \
	  bp_nextBits |= ((int) *bp_bitBuffer) << shift;                  \
	  shift -= 8;                                                     \
	  if (*bp_bitBuffer == 0xff) {                                    \
	    if (*(bp_bitBuffer+1) == 0x00) {                              \
	      bp_bitBuffer += 2;                                          \
	    } else {                                                      \
	      bp_bitBuffer++;                                             \
	    }                                                             \
	  } else {                                                        \
	     bp_bitBuffer++;                                              \
	  }                                                               \
	}                                                                 \
      }                                                                   \
    } else {                                                              \
      unsigned char *tmp_bitBufferPtr;                                    \
      int shift;                                                          \
                                                                          \
      bp_bitOffset -= 32;                                                 \
                                                                          \
      while (bp_bitOffset > 31) {                                         \
	bp_bitOffset -=32;                                                \
	if ((*bp_bitBuffer == 0xff) && (*(bp_bitBuffer+1) == 0x00))       \
	  bp_bitBuffer += 2;                                              \
	else bp_bitBuffer += 1;                                           \
	if ((*bp_bitBuffer == 0xff) && (*(bp_bitBuffer+1) == 0x00))       \
	  bp_bitBuffer += 2;                                              \
	else bp_bitBuffer += 1;                                           \
	if ((*bp_bitBuffer == 0xff) && (*(bp_bitBuffer+1) == 0x00))       \
	  bp_bitBuffer += 2;                                              \
	else bp_bitBuffer += 1;                                           \
	if ((*bp_bitBuffer == 0xff) && (*(bp_bitBuffer+1) == 0x00))       \
	  bp_bitBuffer += 2;                                              \
	else bp_bitBuffer += 1;                                           \
      }                                                                   \
      tmp_bitBufferPtr = bp_bitBuffer;                                    \
      bp_curBits = 0;                                                     \
      tmp_bitBufferPtr--;                                                 \
      if ((*tmp_bitBufferPtr == 0x00) && (*(tmp_bitBufferPtr-1) == 0xff)) { \
        tmp_bitBufferPtr--;                                               \
      }                                                                   \
      bp_curBits |= ((int) *tmp_bitBufferPtr);                            \
      tmp_bitBufferPtr--;                                                 \
      if ((*tmp_bitBufferPtr == 0x00) && (*(tmp_bitBufferPtr-1) == 0xff)) { \
        tmp_bitBufferPtr--;                                               \
      }                                                                   \
      bp_curBits |= ((int) *tmp_bitBufferPtr) << 8;                       \
      tmp_bitBufferPtr--;                                                 \
      if ((*tmp_bitBufferPtr == 0x00) && (*(tmp_bitBufferPtr-1) == 0xff)) { \
        tmp_bitBufferPtr--;                                               \
      }                                                                   \
      bp_curBits |= ((int) *tmp_bitBufferPtr) << 16;                      \
      tmp_bitBufferPtr--;                                                 \
      if ((*tmp_bitBufferPtr == 0x00) && (*(tmp_bitBufferPtr-1) == 0xff)) { \
        tmp_bitBufferPtr--;                                               \
      }                                                                   \
      bp_curBits |= ((int) *tmp_bitBufferPtr) << 24;                      \
                                                                          \
      bp_curBits <<= bp_bitOffset;                                        \
      shift = 24;                                                         \
      bp_nextBits = 0;                                                    \
      while (shift >= 0) {                                                \
        bp_nextBits |= ((int) *bp_bitBuffer) << shift;                    \
        shift -= 8;                                                       \
        if (*bp_bitBuffer == 0xff) {                                      \
	  if (*(bp_bitBuffer+1) == 0x00) {                                \
	    bp_bitBuffer += 2;                                            \
          } else {                                                        \
            bp_bitBuffer++;                                               \
          }                                                               \
        } else {                                                          \
           bp_bitBuffer++;                                                \
        }                                                                 \
      }                                                                   \
    }                                                                     \
  } else {                                                                \
    bp_curBits <<= num;                                                   \
  }                                                                       \
}


#endif


