/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * cmBuffer.c --
 *	Manages the buffer abstraction used by the variou
 *	Dest modules.
 *
 *	Handles buffers in regular memory as well as shared
 *	memory, including creation, modification and searching.
 *	Used by the packetDest module to get buffers in to which
 *	incoming packets are stored and by the media Dest
 *	modules to manipulate those buffers.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif

#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif

#include <tcl.h>
/* #include <tkConfig.h> */
#include <cmConfig.h>
#include <limits.h>
#include <sys/types.h>
#ifdef HAVE_SHARED_MEMORY
#include <sys/ipc.h>
#include <sys/shm.h>
#endif
#ifndef NO_X11
#include <X11/Xlib.h>
#ifdef HAVE_SHARED_MEMORY
#include <X11/extensions/XShm.h>
#endif
#endif
#include <math.h>
#include "cmAtom.h"
#include "cmUtil.h"
#include "cmBuffer.h"

/*--------------------------------------------------------------------------*/

/*
 * The following hash tables are for buffers in different
 * states.  They are keyed by buffer ID
 */
static Tcl_HashTable *freeBuffs = NULL;	/* Buffers in state "free" */
static Tcl_HashTable *fillBuffs = NULL;	/* Buffers in state "filling" */
static Tcl_HashTable *otherBuffs = NULL;	/* all other buffers */

static int nextId = 0;			/* Next buffer ID */
static Buffer *errorBuf;

/*
 * Tcm_ConfigSpec for looking up info in config
 */

static Tcm_ConfigSpec bufferSpecs[] = {
    {TCM_CONFIG_BOOLEAN, "-shared",
	 "false", Tcm_Offset(Buffer, shared),
	 TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_BOOLEAN, "-xattach",
	 "false", Tcm_Offset(Buffer, xAttach),
	 TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_INT, "-size",
	 "0", Tcm_Offset(Buffer, size),
	 TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_STRING, "-state",
	 NULL, Tcm_Offset(Buffer, state), 0},
    {TCM_CONFIG_STRING, "-callback",
	 NULL, Tcm_Offset(Buffer, callback), 0},

    {TCM_CONFIG_END, (char *) NULL, (char *) NULL, 0, 0}
};

/*
 *  Tcm_ConfigSpec for parsing argv in create and config
 */

static Tcm_ConfigSpec configSpecs[] = {
    {TCM_CONFIG_BOOLEAN, "-shared",
	 "false", Tcm_Offset(ConfigureInfo, shared),
	 TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_BOOLEAN, "-xattach",
	 "false", Tcm_Offset(ConfigureInfo, xAttach),
	 TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_INT, "-size",
	 "0", Tcm_Offset(ConfigureInfo, size),
	 TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_STRING, "-state",
	 NULL, Tcm_Offset(ConfigureInfo, state), 0},
    {TCM_CONFIG_STRING, "-callback",
	 NULL, Tcm_Offset(ConfigureInfo, callback), 0},

    {TCM_CONFIG_END, (char *) NULL, (char *) NULL, 0, 0}
};

/*
 * Prototypes for routines declared later
 */
static void IdleDetachBuffer(ClientData clientdata);
int Tcm_BufferCmd (ClientData clientData,
		   Tcl_Interp *interp, int argc, char *argv[]);
static void FreeEverything(Tcl_Interp *interp);


/*
 *--------------------------------------------------------------
 *
 * Tcm_BufferInit
 *      Initializes the buffers module
 *
 * Results:
 *      A standard Tcl result
 *
 * Parameters:
 *      Tcl_Interp *interp              The current interpreter
 *
 * Side effects:
 *      Tcl commands are created
 *
 *--------------------------------------------------------------
 */
int
Tcm_BufferInit(Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "buffer", Tcm_BufferCmd,
                      (ClientData) NULL, (void (*) ()) NULL);

    /*
     * Moved to init command in cmInit.c
     *
     * Tcl_GlobalEval(interp, "dp_atexit append {buffer freeEverything}");
     */
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_BufferCmd --
 *
 *	This function implements the tcl "buffer" command.
 *	The possible options are:
 *
 *		buffer addtag bufferid tag
 *		buffer create ?options?
 *		buffer delete bufferid
 *		buffer dtag bufferid tag
 *		buffer config bufferid ?options?
 *		buffer find ?options?
 *		buffer findOrCreate ?options?
 *		buffer gettags bufferid
 *		buffer getString bufferid offset
 *		buffer setString bufferid offset string
 *		buffer attach bufferid
 *		buffer detach bufferid
 *
 * Results:
 *	A standard tcl result.
 *
 *
 * Parameters:
 *     ClientData clientData;          ignored
 *     Tcl_Interp *interp;             tcl interpreter
 *     int argc;                       Number of arguments
 *     char *argv[];                   Arg list
 * 
 * Side effects:
 *	
 *
 *--------------------------------------------------------------
 */
            /* ARGSUSED */
int
Tcm_BufferCmd (ClientData clientData,
	       Tcl_Interp *interp, int argc, char *argv[])
{
    static int resSize = 0;
    static char *result;

    char c;
    int len;
    Buffer *buffer;

    if (argc < 2) {
	Tcl_AppendResult(interp, "Wrong number of args\n", NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    
    /* ------------------------ ADDTAG --------------------------------- */
    if ((c == 'a') && (strncmp(argv[1], "addtag", len) == 0)) {
	if (argc < 4) {
	    Tcl_AppendResult(interp,
			     "Usage: buffer addtag bufferid tag",
			     NULL);
	    return TCL_ERROR;
	}
	
	/*
	 * Find the buffer to add a tag to
	 */
	if (GetBuffer(interp, argv[2], &buffer) != TCL_OK) {
	    return TCL_ERROR;
	}

	/*
	 * And add the tag
	 */
	return AddTag(interp, buffer, argv[3]);
	
    /* ------------------------ FREE EVERYTHING --------------------------------- */
    } else if ((c == 'f') && (strncmp(argv[1], "freeEverything", len) == 0)) {
      FreeEverything(interp);
      return TCL_OK;
      
    /*------------------------ ATTACH --------------------------------- */
    } else if ((c == 'a') && (strncmp(argv[1], "attach", len) == 0)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp,
			     "Usage: buffer attach <bufferid>",
			     NULL);
	    return TCL_ERROR;
	}
	
	/*
	 * Find the buffer to attach to
	 */
	if (GetBuffer(interp, argv[2], &buffer) != TCL_OK) {
	    return TCL_ERROR;
	}

	return AttachBuffer(interp, buffer, "tcmBuffer:Tcm_BufferCmd");
	
    /* ------------------------ CREATE --------------------------------- */
    } else if ((c == 'c') && (strncmp(argv[1], "create", len) == 0)) {
	char bufferName[16];
	int bufferNum;
	
	/*
	 * Create the buffer
	 */
	if (argc == 2) {
	    bufferNum = CreateBuffer(interp, NULL);
	} else {
	    ConfigureInfo info;
	
	    info.size = BUFFER_DONT_MODIFY;
	    info.shared = BUFFER_DONT_MODIFY;
	    info.xAttach = BUFFER_DONT_MODIFY;
	    info.state = NULL;
	    info.callback = NULL;
	    
	    if (Tcm_ConfigureObject(interp, configSpecs,
				    argc - 2, argv + 2, (char *) &info,
				    TCM_CONFIG_ARGV_ONLY) != TCL_OK) {
		return TCL_ERROR;
	    }

	    bufferNum = CreateBuffer(interp, &info);
	    
	    if (info.state != NULL) {
		ckfree(info.state);
	    }
	    if (info.callback != NULL) {
		ckfree(info.callback);
	    }
	}
	
	if (bufferNum == -1) {
	    Tcl_AppendResult(interp, "Can't create buffer", NULL);
	    return TCL_ERROR;
	}

	/*
	 * Form return value
	 */
	sprintf(bufferName, "buffer%d", bufferNum);
	Tcl_SetResult(interp, bufferName, TCL_VOLATILE);
	return TCL_OK;

    /* ------------------------ CONFIG --------------------------------- */
    } else if ((c == 'c') && (strncmp(argv[1], "config", len) == 0)) {
	int result;
	
	if (argc < 3) {
	    Tcl_AppendResult(interp,
			     "Usage: buffer config bufferid ?attributes?",
			     NULL);
	    return TCL_ERROR;
	}
	
	/*
	 * Find which buffer we're configuring
	 */
	if (GetBuffer(interp, argv[2], &buffer) != TCL_OK) {
	    return TCL_ERROR;
	}
	
	/*
	 * And configure it
	 */
	if (argc == 3) {
	    return (Tcm_ConfigureInfo(interp, bufferSpecs,
				      (char *) buffer, NULL, 0));
	}
	else if (argc == 4) {
	    return (Tcm_ConfigureInfo(interp, bufferSpecs,
				      (char *) buffer, argv[3], 0));
	}
	else {
	    ConfigureInfo info;
	
	    info.size = BUFFER_DONT_MODIFY;
	    info.shared = BUFFER_DONT_MODIFY;
	    info.xAttach = BUFFER_DONT_MODIFY;
	    info.state = NULL;
	    info.callback = NULL;
	    
	    if (Tcm_ConfigureObject(interp, configSpecs,
				    argc - 3, argv + 3, (char *) &info,
				    TCM_CONFIG_ARGV_ONLY) != TCL_OK) {
		return TCL_ERROR;
	    }
	    
	    result = ConfigureBuffer(interp, buffer, &info);

	    if (info.state != NULL) {
		ckfree(info.state);
	    }
	    if (info.callback != NULL) {
		ckfree(info.callback);
	    }
	    
	    return result;
	}

    /* ------------------------ DELETE --------------------------------- */
   } else if ((c == 'd') && (strncmp(argv[1], "delete", len) == 0)) {
       if (argc != 3) {
	   Tcl_AppendResult(interp,
			    "Usage: buffer delete bufferid",
			    NULL);
	   return TCL_ERROR;
       }

       /*
	* Find which buffer we're deleting
	*/
       if (GetBuffer(interp, argv[2], &buffer) != TCL_OK) {
	   return TCL_ERROR;
       }

       /*
	* And delete it
	*/
       DestroyBuffer(interp, buffer);
       return TCL_OK;
	
    /* ------------------------ DETACH --------------------------------- */
    } else if ((c == 'd') && (strncmp(argv[1], "detach", len) == 0)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp,
			     "Usage: buffer detach <bufferid>",
			     NULL);
	    return TCL_ERROR;
	}
	
	/*
	 * Find the buffer to detach
	 */
	if (GetBuffer(interp, argv[2], &buffer) != TCL_OK) {
	    return TCL_ERROR;
	}

	return DetachBuffer(interp, buffer, "tcmBuffer:Tcm_BufferCmd");
	
   /* ------------------------ DTAG --------------------------------- */
   } else if ((c == 'd') && (strncmp(argv[1], "dtag", len) == 0)) {
       if (argc != 4) {
	   Tcl_AppendResult(interp,
			    "Usage: buffer dtag bufferid tag",
			    NULL);
	   return TCL_ERROR;
       }
       
       /*
	* Find which buffer we're deleting tags from
	*/
       if (GetBuffer(interp, argv[2], &buffer) != TCL_OK) {
	   return TCL_ERROR;
       }

       DeleteTag(buffer, argv[3]);
       return TCL_OK;

    /* ------------------------ FIND --------------------------------- */
   } else if ((c == 'f') && (strncmp(argv[1], "find", len) == 0)) {
       FindQuery query;
       int *hits, count, i, resLen;
       char bufName[16];
       
       if (argc < 3) {
	   goto findUsage;
       }
       
       /*
	* Initialize the query
	*/
       query.minsize = BUFFER_DONT_CARE;
       query.maxsize = BUFFER_DONT_CARE;
       query.unique = 0;
       query.state = NULL;
       query.shared = BUFFER_DONT_CARE;
       query.xAttach = BUFFER_DONT_CARE;
       query.tag = NULL;

       /*
	* Parse the request and fill in the fields
	*/
       argc -= 2; argv += 2;
       while (argc > 0) {
	   if (argv[0][0] != '-') {
	       goto findUsage;
	   }
	   
	   switch (argv[0][1]) {
	     case 'a':
	       if (strncmp(argv[0], "-all", 4) == 0) {
		   query.unique = 0;
	       } else {
		   goto findUsage;
	       }
	       break;
	       
	     case 'm':
	       if (strncmp(argv[0], "-maxsize", 8) == 0) {
		   if (argc < 2) {
		       goto findUsage;
		   }
		   if (Tcl_GetInt(interp, argv[1], &query.maxsize) != TCL_OK) {
		       goto findUsage;
		   }
		   argc--; argv++;
	       } else if (strncmp(argv[0], "-minsize", 8) == 0) {
		   if (argc < 2) {
		       goto findUsage;
		   }
		   if (Tcl_GetInt(interp, argv[1], &query.minsize) != TCL_OK) {
		       goto findUsage;
		   }
		   argc--; argv++;
	       } else {
		   goto findUsage;
	       }
	       break;
	       
	     case 's':
	       if (strncmp(argv[0], "-shared", 7) == 0) {
		   if (argc < 2) {
		       goto findUsage;
		   }
		   if (Tcl_GetBoolean(interp, argv[1],
				      &query.shared) != TCL_OK) {
		       goto findUsage;
		   }
		   argc--; argv++;
	       } else if (strncmp(argv[0], "-state", 6) == 0) {
		   if (argc < 2) {
		       goto findUsage;
		   }
		   query.state = argv[1];
		   argc--; argv++;
	       } else {
		   goto findUsage;
	       }
	       break;
	       
	     case 't':
	       if (strncmp(argv[0], "-tag", 4) == 0) {
		   if (argc < 2) {
		       goto findUsage;
		   }
		   query.tag = argv[1];
		   argc--; argv++;
	       } else {
		   goto findUsage;
	       }
	       break;
	       
	     case 'u':
	       if (strncmp(argv[0], "-unique", 7) == 0) {
		   query.unique = 1;
	       } else {
		   goto findUsage;
	       }
	       break;
	       
	     case 'x':
	       if (strncmp(argv[0], "-xattach", 8) == 0) {
		   if (argc < 2) {
		       goto findUsage;
		   }
		   if (Tcl_GetBoolean(interp, argv[1],
				      &query.xAttach) != TCL_OK) {
		       goto findUsage;
		   }
		   argc--; argv++;
	       } else {
		   goto findUsage;
	       }
	       
	       break;
	       
	     default:
	       goto findUsage;
	   }
	   
	   argc--; argv++;
       }
       
       /*
	* Do the actual searching
	*/
       hits = FindBuffer(interp, &query);
       
       /*
	* Build the result string
	*/
#ifdef QUICKFIND
       Tcl_SetResult(interp, "done", TCL_STATIC);
#else
       if (resSize == 0) {
	   resSize = 1024;
	   if ((result = (char *) ckalloc(1024)) == NULL) {
	       Tcl_AppendResult(interp, "Out of memory", NULL);
	       return TCL_ERROR;
	   }
       }	

       resLen = 0;
       result[0] = '\0';

       count = hits[0];
       for (i=0; i < count; i++) {
	   sprintf(bufName, "buffer%d ", hits[i + 1]);
	   if ((resLen + strlen(bufName) + 1) > resSize) {
	       resSize *= 2;
	       if ((result = (char *) ckrealloc(result, resSize)) == NULL) {
		   Tcl_AppendResult(interp, "Out of memory", NULL);
		   return TCL_ERROR;
	       }
	   }
	   strcpy(result + resLen, bufName);
	   resLen += strlen(bufName);
       }
       
       Tcl_SetResult(interp, result, TCL_STATIC); 
#endif	
       if (query.state != NULL) {
	   ckfree(query.state);
       }
       if (query.tag != NULL) {
	   ckfree(query.tag);
       }
       
       return TCL_OK;
       
       
     findUsage:
       Tcl_AppendResult(interp, "Usage: \n",
			"buffer find -option ?value? ?more options?\n",
			"Options are:\n",
			"-all|-unique\n",
			"-minsize $size\n",
			"-maxsize $size\n",
			"-shared $value\n",
			"-xattach $value\n",
			"-state $state\n",
			"-tag $tag\n",
			NULL);
       return TCL_ERROR;
       
    /* -------------------- FIND_OR_CREATE ------------------------------ */
    } else if ((c == 'f') && (strncmp(argv[1], "findOrCreate", len) == 0)) {
       FindQuery query;
       int id;
       char bufferName[16];
       
       if (argc < 3) {
	   goto findOrCreateUsage;
       }
       
       /*
	* Initialize the query
	*/
       query.minsize = BUFFER_DONT_CARE;
       query.maxsize = BUFFER_DONT_CARE;
       query.unique = 0;
       query.state = NULL;
       query.shared = BUFFER_DONT_CARE;
       query.xAttach = BUFFER_DONT_CARE;
       query.tag = NULL;

       /*
	* Parse the request and fill in the fields
	*/
       argc -= 2; argv += 2;
       while (argc > 0) {
	   if (argv[0][0] != '-') {
	       goto findOrCreateUsage;
	   }
	   
	   switch (argv[0][1]) {
	     case 'm':
	       if (strncmp(argv[0], "-maxsize", 8) == 0) {
		   if (argc < 2) {
		       goto findOrCreateUsage;
		   }
		   if (Tcl_GetInt(interp, argv[1], &query.maxsize) != TCL_OK) {
		       goto findOrCreateUsage;
		   }
		   argc--; argv++;
	       } else if (strncmp(argv[0], "-minsize", 8) == 0) {
		   if (argc < 2) {
		       goto findOrCreateUsage;
		   }
		   if (Tcl_GetInt(interp, argv[1], &query.minsize) != TCL_OK) {
		       goto findOrCreateUsage;
		   }
		   argc--; argv++;
	       } else {
		   goto findOrCreateUsage;
	       }
	       break;
	       
	     case 's':
	       if (strncmp(argv[0], "-shared", 7) == 0) {
		   query.shared = 1;
	       } else {
		   goto findOrCreateUsage;
	       }
	       break;
	       
	     case 't':
	       if (strncmp(argv[0], "-tag", 4) == 0) {
		   if (argc < 2) {
		       goto findOrCreateUsage;
		   }
		   query.tag = argv[1];
		   argc--; argv++;
	       } else {
		   goto findOrCreateUsage;
	       }
	       break;
	       
	     case 'x':
	       if (strncmp(argv[0], "-xattach", 8) == 0) {
		   query.xAttach = 1;
	       } else {
		   goto findOrCreateUsage;
	       }
	       
	       break;
	       
	     default:
	       goto findOrCreateUsage;
	   }
	   
	   argc--; argv++;
       }
       
       /*
	* Find or create the buffer
	*/
       id = FindOrCreateBuffer(interp, &query);
       if (id == -1) {
	   Tcl_AppendResult(interp,
			    "Couldn't create buffer",
			    NULL);
	   return TCL_ERROR;
       }
	   
       /*
	* Form return value
	*/
       sprintf(bufferName, "buffer%d", id);
       Tcl_SetResult(interp, bufferName, TCL_VOLATILE);
       return TCL_OK;
       
     findOrCreateUsage:
       Tcl_AppendResult(interp, "Usage: \n",
			"buffer findOrCreate -option ?value? ?more options?\n",
			"Options are:\n",
			"-minsize $size\n",
			"-maxsize $size\n",
			"-shared $value\n",
			"-xattach $value\n",
			"-tag $tag\n",
			NULL);
       return TCL_ERROR;

    /* ----------------------- GETSTRING -------------------------------- */
    } else if ((c == 'g') && (strncmp(argv[1], "getString", len) == 0)) {
	Buffer *bufPtr;
	int offset;
	
	if (argc != 4) {
	    Tcl_AppendResult(interp,
			     "Usage: buffer getString bufferid offset",
			     NULL);
	    return TCL_ERROR;
	}

	if (GetBuffer(interp, argv[2], &bufPtr) != TCL_OK) {
	    return TCL_ERROR;
	}
	
	if (Tcl_GetInt(interp, argv[3], &offset) != TCL_OK) {
	    return TCL_ERROR;
	}
	
	if (offset > bufPtr->size) {
	    Tcl_AppendResult(interp,
			     "Offest out of bounds",
			     NULL);
	    return TCL_ERROR;
	}
	
	Tcl_SetResult(interp, (bufPtr->data + offset), TCL_STATIC);
	return TCL_OK;
	
    /* ------------------------ GETTAGS --------------------------------- */
    } else if ((c == 'g') && (strncmp(argv[1], "gettags", len) == 0)) {
	int i, resLen;
	char *tag;

	if (argc < 3) {
	    Tcl_AppendResult(interp,
			     "Usage: buffer gettags bufferid",
			     NULL);
	    return TCL_ERROR;
	}

	/*
	 * Find the buffer
	 */
	if (GetBuffer(interp, argv[2], &buffer) != TCL_OK) {
	    return TCL_ERROR;
	}

	/*
	 * Build the result
	 */
	if (resSize == 0) {
	    resSize = 1024;
	    if ((result = (char *) ckalloc(1024)) == NULL) {
		Tcl_AppendResult(interp, "Out of memory", NULL);
		return TCL_ERROR;
	    }
	}	
	
	resLen = 0;
	result[0] = '\0';
	
	for (i = 0; i < buffer->tagCount; i++) {
	    tag = buffer->tags[i];
	    if ((resLen + strlen(tag) + 1) > resSize) {
		resSize *= 2;
		if ((result = (char *) ckrealloc(result, resSize)) == NULL) {
		    Tcl_AppendResult(interp, "Out of memory", NULL);
		    return TCL_ERROR;
		}
	    }
	    strcpy(result + resLen, tag);
	    resLen += strlen(tag);
	    result[resLen++] = ' ';
	}
	
	result[resLen] = '\0';
	
	Tcl_SetResult(interp, result, TCL_STATIC); 
	return TCL_OK;
    /* ----------------------- SETSTRING -------------------------------- */
    } else if ((c == 's') && (strncmp(argv[1], "setString", len) == 0)) {
	Buffer *bufPtr;
	int offset;
	
	if (argc != 5) {
	    Tcl_AppendResult(interp,
			     "Usage: buffer setString bufferid offset string",
			     NULL);
	    return TCL_ERROR;
	}

	if (GetBuffer(interp, argv[2], &bufPtr) != TCL_OK) {
	    return TCL_ERROR;
	}
	
	if (Tcl_GetInt(interp, argv[3], &offset) != TCL_OK) {
	    return TCL_ERROR;
	}
	
	if ((offset + strlen(argv[4])) > bufPtr->size) {
	    Tcl_AppendResult(interp,
			     "Buffer ", argv[2],
			     " cannot hold the string \"",
			     argv[4], "\"",
			     NULL);
	    return TCL_ERROR;
	}
	
	strcpy(bufPtr->data + offset, argv[4]);
	return TCL_OK;
    }
	
      usage:
	Tcl_AppendResult(interp, " Possible usages:\n",
			 "\"buffer addtag bufferid tag\"\n",
			 "\"buffer create ?options?\"\n",
			 "\"buffer delete bufferid\"\n",
			 "\"buffer config bufferid ?options?\"\n",
			 "\"buffer find ?options?\"\n",
			 "\"buffer findOrCreate ?options?\"\n",
			 "\"buffer gettags bufferid\"\n",
			 "\"buffer getString bufferid offset\"\n",
			 "\"buffer setString bufferid offset string\"\n",
			 NULL);
	return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * CreateBuffer --
 *
 *	Create a new buffer and configure it properly
 *
 * Results:
 *	A new allocated and initialized buffer, -1 on error
 *
 * Parameters
 *	Tcl_Interp *interp		For error reporting
 *	ConfigureInfo *info		Set the initial values
 *
 * Side effects:
 *	Memory is allocated
 *
 *--------------------------------------------------------------
 */
int
CreateBuffer(Tcl_Interp *interp, ConfigureInfo *info)
{
#if 0
    int id, unused;
    Buffer *buffer;
    Tcl_HashEntry *hashEntry;
    
    /*
     * Allocate the various buffer hash tables if
     * they haven't already been allocated
     */
    if (otherBuffs == NULL) {
	freeBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	fillBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	otherBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(freeBuffs, TCL_ONE_WORD_KEYS);
	Tcl_InitHashTable(fillBuffs, TCL_ONE_WORD_KEYS);
	Tcl_InitHashTable(otherBuffs, TCL_ONE_WORD_KEYS);
    }
    
    /*
     * Allocate the space and initialize the values
     */
    buffer = (Buffer *) ckalloc(sizeof(Buffer));
    id = buffer->id = nextId++;
    buffer->tags = NULL;
    buffer->tagSize = 0;
    buffer->tagCount = 0;
    buffer->state = NULL;
    buffer->size = 0;
    buffer->shared = 0;
    buffer->key = -1;
    buffer->xAttach = 0;
    buffer->callback = NULL;
    buffer->data = NULL;
    buffer->refCount = 0;
    buffer->shmId = -1;
#ifndef NO_X11
#ifdef HAVE_SHARED_MEMORY
    memset(&buffer->shminfo, 0, sizeof(XShmSegmentInfo));
#endif
    buffer->dpy = Tcm_Display();
#endif



    if ((info != NULL) &&
	(ConfigureBuffer(interp, buffer, info) != TCL_OK)) {
	DestroyBuffer(interp, buffer);
	return -1;
    }
    
    /*
     * Add the buffer to the appropriate hash table
     */
    if ((buffer->state != NULL)
	&& (strncmp(buffer->state, "free", 4) == 0)) {
	hashEntry = Tcl_CreateHashEntry(freeBuffs, (char *) id, &unused);
    } else if ((buffer->state != NULL)
	&& (strncmp(buffer->state, "filling", 7) == 0)) {
	hashEntry = Tcl_CreateHashEntry(fillBuffs, (char *) id, &unused);
    } else {
	hashEntry = Tcl_CreateHashEntry(otherBuffs, (char *) id, &unused);
    }
    Tcl_SetHashValue(hashEntry, buffer);

    return id;
#else
    Buffer *b = MakeExistingBuffer(interp, info, NULL, 0);
    if (b == NULL) {
	return -1;
    } else {
	return b->id;
    }
#endif
}


/*
 * ------------------------------------------------------------
 *
 * MakeExistingBuffer
 *
 *	Create a CMT buffer from a previously allocated
 *	segment of memory
 *
 * Results:
 *	A pointer to a Buffer structure, NULL on error
 *
 * Parameters
 *	Tcl_Interp *interp		For error reporting
 *	ConfigureInfo *info		Set the initial values
 *	char *mem			The actual storage space
 *	int size			Size of the segment
 *
 * Side Effects:
 *	A new Buffer structure is allocated and the buffer
 *	manager state is updated to know about the new buffer
 *
 * ------------------------------------------------------------
 */
Buffer *
MakeExistingBuffer(Tcl_Interp *interp, ConfigureInfo *info,
		   char *mem, int size)
{
    int id, unused;
    Buffer *buffer;
    Tcl_HashEntry *hashEntry;
    
    /*
     * Allocate the various buffer hash tables if
     * they haven't already been allocated
     */
    if (otherBuffs == NULL) {
	freeBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	fillBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	otherBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(freeBuffs, TCL_ONE_WORD_KEYS);
	Tcl_InitHashTable(fillBuffs, TCL_ONE_WORD_KEYS);
	Tcl_InitHashTable(otherBuffs, TCL_ONE_WORD_KEYS);
    }
    
    /*
     * Allocate the space and initialize the values
     */
    buffer = (Buffer *) ckalloc(sizeof(Buffer));
    id = buffer->id = nextId++;
    buffer->tags = NULL;
    buffer->tagSize = 0;
    buffer->tagCount = 0;
    buffer->state = NULL;
    buffer->size = size;
    buffer->shared = 0;
    buffer->key = -1;
    buffer->xAttach = 0;
    buffer->callback = NULL;
    buffer->data = mem;
    buffer->refCount = 0;
    buffer->shmId = -1;
#ifndef NO_X11
#ifdef HAVE_SHARED_MEMORY
    memset(&buffer->shminfo, 0, sizeof(XShmSegmentInfo));
#endif
    buffer->dpy = Tcm_Display();
#endif

    if ((info != NULL) &&
	(ConfigureBuffer(interp, buffer, info) != TCL_OK)) {
	DestroyBuffer(interp, buffer);
	return NULL;
    }
    
    /*
     * Add the buffer to the appropriate hash table
     */
    if ((buffer->state != NULL)
	&& (strncmp(buffer->state, "free", 4) == 0)) {
	hashEntry = Tcl_CreateHashEntry(freeBuffs, (char *) id, &unused);
    } else if ((buffer->state != NULL)
	&& (strncmp(buffer->state, "filling", 7) == 0)) {
	hashEntry = Tcl_CreateHashEntry(fillBuffs, (char *) id, &unused);
    } else {
	hashEntry = Tcl_CreateHashEntry(otherBuffs, (char *) id, &unused);
    }
    Tcl_SetHashValue(hashEntry, buffer);

    return buffer;
}


/*
 * ------------------------------------------------------------
 *
 * MakeExistingShmBuffer
 *
 *	Create a CMT buffer from a previously allocated
 *	segment of shared memory
 *
 * Results:
 *	A pointer to a Buffer structure, NULL on error
 *
 * Parameters
 *	Tcl_Interp *interp		For error reporting
 *	ConfigureInfo *info		Set the initial values
 *	char *mem			The actual storage space
 *	int size			Size of the segment
 *
 * Side Effects:
 *	A new Buffer structure is allocated and the buffer
 *	manager state is updated to know about the new buffer
 *
 * ------------------------------------------------------------
 */
Buffer *
MakeExistingShmBuffer(Tcl_Interp *interp, ConfigureInfo *info,
		      char *addr, int shmid, int size)
{
#ifndef HAVE_SHARED_MEMORY
    return NULL;
#elif NO_X11
    return NULL;
#else

    int id, unused;
    Buffer *buffer;
    Tcl_HashEntry *hashEntry;
    struct shmid_ds shmStat;
    
    /*
     * Allocate the various buffer hash tables if
     * they haven't already been allocated
     */
    if (otherBuffs == NULL) {
	freeBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	fillBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	otherBuffs = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(freeBuffs, TCL_ONE_WORD_KEYS);
	Tcl_InitHashTable(fillBuffs, TCL_ONE_WORD_KEYS);
	Tcl_InitHashTable(otherBuffs, TCL_ONE_WORD_KEYS);
    }
    
    /*
     * Allocate the space and initialize the values
     */
    buffer = (Buffer *) ckalloc(sizeof(Buffer));
    id = buffer->id = nextId++;
    buffer->tags = NULL;
    buffer->tagSize = 0;
    buffer->tagCount = 0;
    buffer->state = NULL;
    buffer->size = size;
    buffer->shared = 1;
    buffer->key = -1;
    buffer->xAttach = 0;
    buffer->callback = NULL;
    buffer->data = addr;
    buffer->refCount = 0;
#ifndef NO_X11
    buffer->dpy = Tcm_Display();
#endif

    memset(&buffer->shminfo, 0, sizeof(XShmSegmentInfo));
    buffer->shmId = shmid;
    if (addr == NULL) {
	addr = shmat(shmid, NULL, 0);
    }

    buffer->shminfo.shmaddr = addr;
    buffer->shminfo.shmid = shmid;

    /* do we need this? */
    shmctl(shmid, IPC_STAT, &shmStat);
    if (shmStat.shm_nattch == 0) {
	shmctl(shmid, IPC_RMID, NULL);
	Tcl_AppendResult(interp,
			 "Failed to attach shared memory segment.", NULL);
	return NULL;
    }
    
    if ((info != NULL) &&
	(ConfigureBuffer(interp, buffer, info) != TCL_OK)) {
	DestroyBuffer(interp, buffer);
	return NULL;
    }
    
    /*
     * Add the buffer to the appropriate hash table
     */
    if ((buffer->state != NULL)
	&& (strncmp(buffer->state, "free", 4) == 0)) {
	hashEntry = Tcl_CreateHashEntry(freeBuffs, (char *) id, &unused);
    } else if ((buffer->state != NULL)
	&& (strncmp(buffer->state, "filling", 7) == 0)) {
	hashEntry = Tcl_CreateHashEntry(fillBuffs, (char *) id, &unused);
    } else {
	hashEntry = Tcl_CreateHashEntry(otherBuffs, (char *) id, &unused);
    }
    Tcl_SetHashValue(hashEntry, buffer);

    return GetBufferFromID(id);
#endif
}


/*
 *--------------------------------------------------------------
 *
 * FreeEverything --
 *
 *
 * Results:
 *	Frees all buffers in preparation for exit.
 *
 * Parameters:
 * 
 * Side effects:
 *	Everything assocated with all buffers is freed.
 *
 *--------------------------------------------------------------
 */
static void
FreeEverything(Tcl_Interp *interp)
{
  int counter;
  Buffer *bufPtr;

  for (counter=0; counter < nextId; counter++) {
    bufPtr = GetBufferFromID(counter);

    if (bufPtr != NULL) {
      if (bufPtr->shared) {
	DestroyBuffer(interp, bufPtr);
      }
    }
  }
}


/*
 *--------------------------------------------------------------
 *
 * DestroyBuffer --
 *
 *
 * Results:
 *	Gets rid of a buffer structure
 *
 * Parameters:
 *	Tcl_Interp *interp		For error reporting
 *	Buffer *buffer			The buffer to destroy
 * 
 * Side effects:
 *	Everything assocated with the buffer is freed.
 *
 *--------------------------------------------------------------
 */
void
DestroyBuffer(Tcl_Interp *interp, Buffer *buffer)
{
#ifdef HAVE_SHARED_MEMORY
    struct shmid_ds shmStat;
#endif

    Tcl_HashEntry *hashEntry;

    /*
     * Free the actual buffer
     */
    if ((buffer->shared) && (buffer->shmId != -1)) {
#ifdef HAVE_SHARED_MEMORY
	
	/*
	 * First stat the buffer to find out how many
	 * processes are attached
	 */
	shmctl(buffer->shmId, IPC_STAT, &shmStat);

#ifndef NO_X11
	/*
	 * Detach from X server if necessary.  Have to be careful, since
	 * we could've failed within a CreateBuffer() call, so the buffer
	 * may not be fully configured.
	 */
	if ((buffer->xAttach) && (buffer->dpy != NULL)) {
	    XShmDetach(buffer->dpy, &buffer->shminfo);
	}
#endif
	
	/*
	 * Detach the shared memory segment, deleting if
	 * we're the last process attached.
	 *
	 * The second clause prevents a race condition
	 * since the X server may not detach immediately
	 * when XShmDetach is called.
	 */
	shmdt(buffer->data);
	if ((shmStat.shm_nattch == 1)
	    || (buffer->xAttach && (shmStat.shm_nattch == 2))) {
	    shmctl(buffer->shmId, IPC_RMID, NULL);
	}
#endif
    } else {
	if (buffer->data != NULL) {
	    ckfree(buffer->data);
	    buffer->data = NULL;
	}
    }

    /*
     * Find the hashEntry for this buffer and remove it from
     * it's hash table
     */
    if ((buffer->state != NULL)
	&& (strncmp(buffer->state, "free", 4) == 0)) {
	hashEntry = Tcl_FindHashEntry(freeBuffs, (char *) buffer->id);
    } else if ((buffer->state != NULL)
	       && (strncmp(buffer->state, "filling", 7) == 0)) {
	hashEntry = Tcl_FindHashEntry(fillBuffs, (char *) buffer->id);
    } else {
	hashEntry = Tcl_FindHashEntry(otherBuffs, (char *) buffer->id);
    }	

    if (hashEntry != NULL) {
	Tcl_DeleteHashEntry(hashEntry);
    }
    
    /*
     * Free state and tags
     */
    if (buffer->state != NULL) {
	ckfree(buffer->state);
	buffer->state = NULL;
    }
    if (buffer->callback != NULL) {
	ckfree(buffer->callback);
	buffer->callback = NULL;
    }
    if (buffer->tags != NULL) {
	/*
	 * Don't need to free the individual tags since
	 * we are using Tk Uid's
	 */
	ckfree((void *) buffer->tags);
	buffer->tags = NULL;
    }

    ckfree((void *) buffer);
}

#ifndef NO_X11

static int AttachHandler (Display *display, XErrorEvent *err) 
{
    if (err->request_code == 134) {
	if (err->minor_code == 1) {
	    /* Found XAttachError. */
	    errorBuf->xAttach = 0;
	    return 0;
	}
    }
}

#endif


/*
 *----------------------------------------------------------------------
 *
 * ConfigureBuffer--
 *
 *      This procedure is called to process an argv/argc list,
 *      in order to configure (or reconfigure) a buffer
 *
 * Results:
 *      The return value is a standard Tcl result.  If TCL_ERROR is
 *      returned, then interp->result contains an error message.
 *
 * Parameters:
 *	Tcl_Interp *interp		Used for error reporting
 *	Buffer *buffer			The buffer to configure
 *	ConfigureInfo *info		How to configure it
 *
 * Side effects:
 *      Configuration information, such as the shared flag,
 *      xattach flag, size, tage, etc... get set.
 *
 *----------------------------------------------------------------------
 */
int
ConfigureBuffer(Tcl_Interp *interp, Buffer *buffer, ConfigureInfo *info)
{
    int oldSize, oldShared, oldAttach, unused;
    char oldState[32];
#ifdef HAVE_SHARED_MEMORY
    struct shmid_ds shmStat;
#endif
    Tcl_HashEntry *hashEntry;
    
    if (!buffer) {
        Tcl_AppendResult(interp, "bad buffer",
                         (char *) NULL);
        return TCL_ERROR;
    }

    /*
     * Save old values
     */
    oldShared = buffer->shared;
    oldAttach = buffer->xAttach;
    oldSize = buffer->size;
    
    if (buffer->state != NULL) {
	strcpy(oldState, buffer->state);
    } else {
	*oldState = '\0';
    }

    /*
     * Reconfigure
     */
    if (info->size != BUFFER_DONT_MODIFY) {
	/*
	 * In order to prevent a proliferation of buffers that are
	 * slightly too small, we want to heuristically pick the buffer
	 * size based on the requested size, guaranteeing that the buffer
	 * size selected is at least as big as the one we wanted.
	 *
	 * The heurstic we choose is to use a decibal scale of
	 * 10 log (x/1024) for sizes > 1024, 1024 for sizes < 1024;
	 */

	if (info->size < 1024) {
	    buffer->size = 1024;
	} else {
	    buffer->size = (int) (ceil(pow(10, ceil(10.0 * 
						    log10(((double) info->size)/1024.0))
					   /10.0)*1024.0));
	    /* NT Hack.  Sometimes, buffer->size is 0 after this line */
	    if (buffer->size < info->size) {
		buffer->size = info->size;
	    }
	}
    }
    
    if (info->shared != BUFFER_DONT_MODIFY) {
	buffer->shared = info->shared; 
    }
    
    if (info->xAttach != BUFFER_DONT_MODIFY) {
	buffer->xAttach = info->xAttach;
    }

    if (info->state != NULL) {
	if (buffer->state != NULL) {
	    ckfree(buffer->state);
	}
	buffer->state = (char *) ckalloc(strlen(info->state) + 1);
	strcpy(buffer->state, info->state);
    }

   
    if (info->callback != NULL) {
      if (buffer->callback != NULL) {
	ckfree(buffer->callback);
      }
      buffer->callback = (char *) ckalloc(strlen(info->callback)+1);
      strcpy(buffer->callback, info->callback);
    }

    /*
     * If the size used to be zero (i.e. there was no buffer)
     * and has increased, we allocate a new buffer
     */
    if ((oldSize == 0) && (buffer->size > 0)) {
	if (buffer->shared) {

#ifdef HAVE_SHARED_MEMORY
	    /*
	     * Get a shared memory segment
	     */
	    buffer->shmId = shmget(IPC_PRIVATE, buffer->size, IPC_CREAT|0777);
#ifndef NO_X11
	    buffer->shminfo.shmid = buffer->shmId;
#endif
	    
	    if (buffer->shmId < 0) {
		Tcl_AppendResult(interp, "Cannot create shared memory buffer.",
				 NULL);
		buffer->shared = 0;
		buffer->xAttach = 0;
		buffer->data = NULL;
		return TCL_ERROR;
	    }
	    
	    /*
	     * Attach it to this address space
	     */
	    buffer->data = (char *) shmat(buffer->shmId, NULL, 0);
#ifndef NO_X11
	    buffer->shminfo.shmaddr = buffer->data;
#endif
	    if ((buffer->data == NULL) || (((int) buffer->data) == -1)) {
		shmctl(buffer->shmId, IPC_STAT, &shmStat);
		if (shmStat.shm_nattch == 0) {
		    shmctl(buffer->shmId, IPC_RMID, NULL);
		}
		Tcl_AppendResult(interp,
				 "Failed to attach shared memory segment.",
				 NULL);
		buffer->shared = 0;
		buffer->xAttach = 0;
		buffer->data = NULL;
		return TCL_ERROR;
	    }
	    
	    /*
	     * Attach to X server if requested
	     */
	    if (buffer->xAttach) {
		int res;
#ifndef NO_X11
		buffer->dpy = Tcm_Display();
		if (cmXShmQueryExtension(buffer->dpy)) {
		    int (*old_handler) ();
		    
		    old_handler = XSetErrorHandler(AttachHandler);
		    buffer->shminfo.readOnly = 0;
		    errorBuf = buffer;
		    res = XShmAttach(buffer->dpy, &buffer->shminfo);
		    XSync(buffer->dpy, False);
		    XSetErrorHandler(old_handler);
		    if (buffer->xAttach == 0) {
			Tcl_AppendResult(interp,
					 "X Attach error", NULL);
			return TCL_ERROR;
		    }
		} else {
		    buffer->xAttach = 0;
		    Tcl_AppendResult(interp,
				     "Can't attach to X server\n", NULL);
		    return TCL_ERROR;
		}
#endif
	    }
	    
#else		/* HAVE_SHARED_MEMORY */
	    Tcl_AppendResult(interp,
			     "Shared memory is not available on this system\n",
			     NULL);
	    buffer->shared = 0;
	    buffer->xAttach = 0;
	    buffer->data = NULL;
	    return TCL_ERROR;
	    
#endif		/* HAVE_SHARED_MEMORY */
	    
	} else {
	    
	    /*
	     * Allocate non shared buffer
	     */
	    buffer->data = (char *) ckalloc(buffer->size);
	    if (buffer->data == NULL) {
		Tcl_AppendResult(interp, "Out of memory", NULL);
		return TCL_ERROR;
	    }
	}
    } else if (buffer->size == 0) {
	return TCL_OK;
    } else if (oldSize != buffer->size) {
	Tcl_AppendResult(interp, "Cannot dynamically change buffer size.",
			 NULL);
	return TCL_ERROR;
    } else if (oldShared != buffer->shared) {
	Tcl_AppendResult(interp,
			 "Cannot dynamically changed shared attribute",
			 NULL);
	return TCL_ERROR;
    } else if (oldAttach != buffer->xAttach) {
#ifndef NO_X11
#ifdef HAVE_SHARED_MEMORY
	if (oldAttach) {
	    XShmDetach(buffer->dpy, &buffer->shminfo);
	} else {
	    int res;
	    int (*old_handler)();

	    if (!buffer->shared) {
		Tcl_AppendResult(interp,
				 "Can't attach a non-shared buffer.",
				 NULL);
		return TCL_ERROR;
	    }
	    
	    buffer->dpy = Tcm_Display();

	    if (!cmXShmQueryExtension(buffer->dpy)) {
		Tcl_AppendResult(interp, "Can't attach to X server.",
				 NULL);
		return TCL_ERROR;
	    }

	    old_handler = XSetErrorHandler(AttachHandler);
	    buffer->shminfo.readOnly = 0;
	    errorBuf = buffer;
	    res = XShmAttach(buffer->dpy, &buffer->shminfo);
	    XSync(buffer->dpy, False);
	    XSetErrorHandler(old_handler);
	    if (buffer->xAttach == 0) {
		Tcl_AppendResult(interp, 
				 "X Attach error", NULL);
		return TCL_ERROR;
	    }
	}
#endif
#endif
    }
    
    /*
     * Move the buffer to a new hash table if necessary
     * because of the change in states
     */
    if (buffer->state != NULL) {
	if (strncmp(buffer->state, "free", 4) == 0) {
	    if (strncmp(oldState, "filling", 7) == 0) {
		/*
		 * Move from filling to free
		 */
		hashEntry = Tcl_FindHashEntry(fillBuffs, (char *) buffer->id);
		Tcl_DeleteHashEntry(hashEntry);
		hashEntry = Tcl_CreateHashEntry(freeBuffs,
						(char *) buffer->id, &unused);
		Tcl_SetHashValue(hashEntry, buffer);
	    } else if (strncmp(oldState, "free", 4) != 0) {
		/*
		 * Move from others to free
		 */
		hashEntry = Tcl_FindHashEntry(otherBuffs, (char *) buffer->id);
		Tcl_DeleteHashEntry(hashEntry);
		hashEntry = Tcl_CreateHashEntry(freeBuffs,
						(char *) buffer->id, &unused);
		Tcl_SetHashValue(hashEntry, buffer);
	    }
	} else if (strncmp(buffer->state, "filling", 7) == 0) {
	    if (strncmp(oldState, "free", 4) == 0) {
		/*
		 * Move from free to filling
		 */
		hashEntry = Tcl_FindHashEntry(freeBuffs, (char *) buffer->id);
		Tcl_DeleteHashEntry(hashEntry);
		hashEntry = Tcl_CreateHashEntry(fillBuffs,
						(char *) buffer->id, &unused);
		Tcl_SetHashValue(hashEntry, buffer);
	    } else if (strncmp(oldState, "filling", 7) != 0) {
		/*
		 * Move from others to free
		 */
		hashEntry = Tcl_FindHashEntry(otherBuffs, (char *) buffer->id);
		Tcl_DeleteHashEntry(hashEntry);
		hashEntry = Tcl_CreateHashEntry(fillBuffs,
						(char *) buffer->id, &unused);
		Tcl_SetHashValue(hashEntry, buffer);
	    }
	} else {
	    if (strncmp(oldState, "free", 4) == 0) {
		/*
		 * Move from free to others
		 */
		hashEntry = Tcl_FindHashEntry(freeBuffs, (char *) buffer->id);
		Tcl_DeleteHashEntry(hashEntry);
		hashEntry = Tcl_CreateHashEntry(otherBuffs,
						(char *) buffer->id, &unused);
		Tcl_SetHashValue(hashEntry, buffer);
	    } else if (strncmp(oldState, "filling", 7) == 0) {
		/*
		 * Move from filling to others
		 */
		hashEntry = Tcl_FindHashEntry(fillBuffs, (char *) buffer->id);
		Tcl_DeleteHashEntry(hashEntry);
		hashEntry = Tcl_CreateHashEntry(otherBuffs,
						(char *) buffer->id, &unused);
		Tcl_SetHashValue(hashEntry, buffer);
	    }
	}
    }

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * FindBuffer --
 *	This routine actually carries out the search for
 *	a buffer given a set of attributes to look for.
 *
 * Results:
 *	A vector of matching buffer IDs.  The first entry
 *    	contains the number of matches found.  The vector
 *	should be freed with ckfree after being used.
 *
 * Parameters:
 *	Tcl_Interp *interp		For error reporting
 *	FindQuery *query		The query being made
 *
 * Side effects:
 *	Nothing is modified.
 *
 *--------------------------------------------------------------
 */
int *
FindBuffer(Tcl_Interp *interp, FindQuery *query)
{
    int count, hit;
    Buffer *bufPtr;
    int tag, hitSize;
    char *tagAtom, *allAtom;
    Tcl_HashEntry *hashEntry;
    static int size = 0, *hits;
    Tcl_HashSearch search;

    /*
     * If this is the first time calling FindBuffer,
     * then set up the hits array
     */
    if (size == 0) {
	size = 16;
	hits = (int *) ckalloc(17 * sizeof(int));
    }
    
    /*
     * If no buffers have been created yet,
     * just quit right now
     */
    if (otherBuffs == NULL) {
	hits[0] = 0;
	return hits;
    }
    
    /*
     * Get the atom for the tag being searched for
     */
    if (query->tag == NULL) {
	tagAtom = NULL;
    } else {
	tagAtom = Tk_GetUid(query->tag);
    }
    
    allAtom = Tk_GetUid("all");
	
    count = 0;

    /*
     * Set up the search based on the state
     */
    if (query->state == NULL) {
	hashEntry = Tcl_FirstHashEntry(otherBuffs, &search);
    } else if (strncmp(query->state, "free", 4) == 0) {
	hashEntry = Tcl_FirstHashEntry(freeBuffs, &search);
    } else if (strncmp(query->state, "filling", 7) == 0) {
	hashEntry = Tcl_FirstHashEntry(fillBuffs, &search);
    } else {
	hashEntry = Tcl_FirstHashEntry(otherBuffs, &search);
    }
	
    while (hashEntry != NULL) {
	/*
	 * Get the buffer and see if it really exists
	 */
	bufPtr = (Buffer *) Tcl_GetHashValue(hashEntry);

	hit = 0;

	/*
	 * Check the size, state, shared, and X-attached
	 */
	if (tagAtom == allAtom) {
	    hit = 1;
	} else if (((query->minsize == BUFFER_DONT_CARE)
		    || (bufPtr->size >= query->minsize))
		   && ((query->maxsize == BUFFER_DONT_CARE)
		       || (bufPtr->size <= query->maxsize))
		   && ((query->shared == BUFFER_DONT_CARE)
		       || (bufPtr->shared == query->shared))
		   && ((query->xAttach == BUFFER_DONT_CARE)
		       || (bufPtr->xAttach == query->xAttach))
		   && ((query->state == NULL)
		       || ((bufPtr->state != NULL)
			   && (strcmp(bufPtr->state, query->state) == 0)))) {
	    /*
	     * Now, we just need to match a tag
	     */
	    
	    if (tagAtom == NULL) {
		hit = 1;
	    } else {
		for (tag = 0; tag < bufPtr->tagCount; tag++) {
		    if (tagAtom == bufPtr->tags[tag]) {
			hit = 1;
		    }
		}
	    }
	} 
#if 0
	else if  (((query->minsize != BUFFER_DONT_CARE) &&
		   (query->minsize > buffer->size)) &&
		  ((query->shared != BUFFER_DONT_CARE) &&
		   (bufPtr->shared == query->shared == 1)) &&
		  ((query->xAttach != BUFFER_DONT_CARE) &&
		   (bufPtr->xAttach == query->xAttach == 1)) &&
		  (strcmp(query->state, "free") == 0))
	  {
	    char logStr[100];

	    sprintf(logStr, "Couldn't use shared buf_id %d, size %d < %d", 
		    bufPtr->id, bufPtr->size, query->size);
	    Tcm_Log(logStr);
	}
#endif	
	
	hitSize = 0;
	if (hit) {
	    if ((query->unique) && (count > 0)) {
		/*
		 * It gets treated differently if the unique flag is on
		 */
		if (((query->minsize != BUFFER_DONT_CARE) &&
		     (bufPtr->size < hitSize)) ||
		    ((query->maxsize != BUFFER_DONT_CARE) &&
		     (bufPtr->size > hitSize))) {
		    hits[1] = bufPtr->id;
		    hitSize = bufPtr->size;
		}
	    } else {
		/*
		 * Otherwise simply add it to the list
		 */

		if (count == size) {
		    size += 16;
		    hits = (int *) ckrealloc((void *) hits,
					     (size + 1) * sizeof(int));
		}
	    
		count++;
		hits[count] = bufPtr->id;
		
		if (query->unique) {
		    hitSize = bufPtr->size;
		}
	    }
	}

	hashEntry = Tcl_NextHashEntry(&search);
    }
    hits[0] = count;
    
    return hits;
}

/*
 *--------------------------------------------------------------
 *
 * FindOrCreateBuffer --
 *
 *	Look for a buffer with the given configuration, and
 *	create one if one cannot be found
 *
 * Results:
 *	A buffer id, or -1 if a buffer is not found and one
 *	cannot be created.
 *
 * Side effects:
 *	A new buffer may be created
 *
 * Parameters:
 *	Tcl_Interp *interp	For error reporting
 *	FindQuery *query	The specs for the buffer
 *
 *--------------------------------------------------------------
 */
int
FindOrCreateBuffer(interp, query)
    Tcl_Interp *interp;
    FindQuery *query;
{
    int *hits;
    ConfigureInfo info;
    
    /*
     * First do the find
     */
    query->unique = 1;
    query->state = "free";
    
    hits = FindBuffer(interp, query);

    if (*hits > 0) {
	/*
	 * We found one
	 */
#if 0
      if (query->shared) {
	char logStr[50];
	sprintf(logStr, "Reusing shared buffer %d", hits[1]);
	Tcm_Log(logStr);
      }
#endif

#if 0
      {
	  char logStr[200];
	  Buffer *b;
	  b = GetBufferFromID(hits[1]);
	  sprintf(logStr, "FindOrCreateBuffer: Reusing buffer of size %d at %x\n", b->size, b->data);
	  Tcm_Log(logStr);
      }
#endif

	return hits[1];
    } else {
	int bid;

	/*
	 * We didn't find one; we need to create one
	 */
	info.size = query->minsize;
	info.shared = (query->shared == BUFFER_DONT_CARE) ? 0 : query->shared;
	info.xAttach = (query->xAttach == BUFFER_DONT_CARE) ? 0 : query->xAttach;
	info.state = NULL;
	info.callback = NULL;

	bid = CreateBuffer(interp, &info);
	
#if 0
      {
	  Buffer *b;
	  char logStr[200];
	  b = GetBufferFromID(bid);
	  sprintf(logStr, "FindOrCreateBuffer: Creating new buffer of size %d at %x\n", b->size, b->data);
	  Tcm_Log(logStr);
      }
#endif
	
	return bid;
    }
}

/*
 *--------------------------------------------------------------
 *
 * AddTag --
 *	This routine adds a tag to the list of tags
 *	associated with a buffer
 *
 * Results:
 *	A standard Tcl Result.
 *
 * Parameters:
 *	Tcl_Interp *interp		For error reporting
 *	Buffer *buffer			The buffer to add the tag to
 *	char *tag			The tag to add
 *
 * Side effects:
 *	The tag list for the buffer is modified
 *
 *--------------------------------------------------------------
 */
int
AddTag(Tcl_Interp *interp, Buffer *buffer, char *tag)
{
    /*
     * If we don't have any free space in the tags
     * array, allocate some more
     */
    if (buffer->tagSize == buffer->tagCount) {
	if (buffer->tagSize == 0) {
	    buffer->tags = (char **) ckalloc(16 * sizeof(char *));
	    if (buffer->tags == NULL) {
		Tcl_AppendResult(interp, "Out of memory", NULL);
		return TCL_ERROR;
	    }
	    buffer->tagSize = 16;
	} else {
	    buffer->tagSize += 16;
	    buffer->tags = (char **) ckrealloc((void *) buffer->tags,
					       buffer->tagSize * sizeof(char *));
	    if (buffer->tags == NULL) {
		Tcl_AppendResult(interp, "Out of memory", NULL);
		return TCL_ERROR;
	    }
	}
    }
    
    /*
     * Get a Uid for the tag
     */
    buffer->tags[buffer->tagCount] = Tk_GetUid(tag);
    buffer->tagCount++;
    
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * DeleteTag -- 
 *
 *	Deletes the given tag from the given buffer
 *
 * Results:
 *	none
 *
 * Side effects:
 *	The taglist for the given buffer is modified
 *
 * Parameters:
 *	Buffer *bufPtr		The buffer to remove a tag from
 *	char *tag		The tag to remove
 *
 *--------------------------------------------------------------
 */
void
DeleteTag(bufPtr, tag)
    Buffer *bufPtr;
    char *tag;
{
    int i;
    
    for (i = 0; i < bufPtr->tagCount; i++) {
	if (strcmp(bufPtr->tags[i], tag) == 0) {
	    /*
	     * Don't need to free the tag since we are using
	     * a Tk Uid
	     */
	    bufPtr->tagCount--;
	    bufPtr->tags[i] = bufPtr->tags[bufPtr->tagCount];
	    break;
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * GetBufferFromID
 *
 *	This procedure takes a buffer id and returns the full
 *	buffer structure
 *
 * Results:
 *	A pointer to the buffer structure associated with an ID.
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	int id			ID of the buffer to look up
 *
 *--------------------------------------------------------------
 */
Buffer *
GetBufferFromID(int id)
{
    Tcl_HashEntry *hashEntry;
    
    if (otherBuffs == NULL) {
	return NULL;
    }
    
    hashEntry = Tcl_FindHashEntry(freeBuffs, (char *) id);
    if (hashEntry != NULL) {
	return (Buffer *) Tcl_GetHashValue(hashEntry);
    }

    hashEntry = Tcl_FindHashEntry(fillBuffs, (char *) id);
    if (hashEntry != NULL) {
	return (Buffer *) Tcl_GetHashValue(hashEntry);
    }

    hashEntry = Tcl_FindHashEntry(otherBuffs, (char *) id);
    if (hashEntry != NULL) {
	return (Buffer *) Tcl_GetHashValue(hashEntry);
    }

    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * GetBuffer --
 *
 *	This procedure takes a bufferid string and converts
 *	it to a Buffer pointer usable from C.  The calling
 *	sequence looks just like Tcl_GetInt
 *
 * Results:
 *	A standard Tcl result
 *
 * Side effects:
 *	none
 *
 * Parameters:
 *	Tcl_Interp *interp    	For error reporting
 *	char *string		The string to convert
 *	Buffer **bufPtr		Buffer pointer to store the result in
 *
 *--------------------------------------------------------------
 */
int
GetBuffer(interp, string, bufPtr)
    Tcl_Interp *interp;
    char *string;
    Buffer **bufPtr;
{
    int id;
    
    if ((otherBuffs == NULL)
	|| (strncmp(string, "buffer", 6) != 0)
	||  (Tcl_GetInt(interp, string+6, &id) != TCL_OK)) {
	goto getError;
    }
    
    *bufPtr = GetBufferFromID(id);
    
    if (*bufPtr == NULL) {
	goto getError;
    }

    return TCL_OK;
    
  getError:
    Tcl_AppendResult(interp,
		     "\"", string, "\" is in inavlid buffer",
		     NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * AttachBuffer
 *
 *	Increases the reference count on the current buffer
 *
 * Results:
 *	Standard Tcl result
 *
 * Side effects:
 *	Reference count is increased
 *
 * Parameters:
 *	Tcl_Interp *interp	For error reporting
 *	Buffer *buffer		The buffer to increase the
 *				count on
 *	char *debugStr		String used for debugging (identifies
 *				caller)
 *
 *--------------------------------------------------------------
 */
int
AttachBuffer(Tcl_Interp *interp, Buffer *buffer, char *debugStr)
{
    char str[256];
    if (buffer == NULL) {
	Tcl_AppendResult(interp, "Invalid buffer", NULL);
	return TCL_ERROR;
    }
    
    buffer->refCount++;
    if (debugStr != NULL) {
      sprintf (str, "Attach %s %d %x", debugStr, buffer->refCount,
	       (unsigned int)buffer->data);
      Tcm_Log (str);
    }

    if (buffer->shared) {
      sprintf(str, "Shared buffer %d : %d", buffer->id, buffer->refCount);
      Tcm_Log(str);
    }

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * DetachBuffer
 *
 *      Schedules IdleDetachBuffer to be called when idle.
 *      Detaching from buffers is required to be on the idle
 *      queue in order to prevent buffer passing between objects
 *      from casuing the reference count to fall to 0 artificially.
 * 
 *
 * Results:
 *	Standard Tcl result
 *
 * Side effects:
 *      Call to IdleDetachBuffer put on idle queue.
 *
 * Parameters:
 *	Tcl_Interp *interp	For error reporting
 *	Buffer *buffer		The buffer to detach
 *	char *debugStr		String used for debugging (identifies
 *				caller)
 *
 *--------------------------------------------------------------
 */
int
DetachBuffer(Tcl_Interp *interp, Buffer *buffer, char *debugStr)
{
    char str[256];
    ClientData *params;

    if (buffer == NULL) {
	Tcl_AppendResult(interp, "Invalid buffer", NULL);
	return TCL_ERROR;
    }
    
    if (debugStr != NULL) {
      sprintf (str, "Detach %s %d %x", debugStr, buffer->refCount,
	       (unsigned int)buffer->data);
      Tcm_Log (str);
    }

    params = (ClientData *) ckalloc(sizeof(ClientData) * 2);

    params[0] = (ClientData) interp;
    params[1] = (ClientData) buffer;

    Tcm_DoWhenIdle("tk", IdleDetachBuffer, (ClientData) params);

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * IdleDetachBuffer
 *
 *	Decreases the reference count on the current buffer
 *	and calls the callback to free the buffer if the
 *     	count becomes zero.  If there is no callback, it
 *	is simply returned to the free list. 
 *
 * Results:
 *	Standard Tcl result
 *
 * Side effects:
 *	The reference count is decreased and the buffer may
 *	be freed or the callback may be called
 *
 * Parameters:
 *	Tcl_Interp *interp	For error reporting
 *	Buffer *buffer		The buffer to detach
 *	char *debugStr		String used for debugging (identifies
 *				caller)
 *
 *--------------------------------------------------------------
 */
static void
IdleDetachBuffer(ClientData clientdata)
{
    ClientData *params = (ClientData *) clientdata;
    Tcl_Interp *interp = (Tcl_Interp *) params[0];
    Buffer *buffer = (Buffer *) params[1];
    char str[256];

    ckfree((char *) params);

    if (buffer == NULL) {
	Tcl_AppendResult(interp, "Invalid buffer", NULL);
	return;
    }
    
    buffer->refCount--;

    if (buffer->shared) {
      sprintf(str, "Shared buffer %d : %d", buffer->id, buffer->refCount);
      Tcm_Log(str);
    }

    if (buffer->refCount <= 0) {
	/*
	 * Need to reclaim the buffer.  Three cases:
	 *
	 * o If there is no callback, simply mark the state as free
	 * o If the callback begins with the character '@' then look up
	 *   the corresponding atom to find the pointer to the callback
	 *   procedure.
	 * o Otherwise, the callback is a Tcl command
	 */
	if ((buffer->callback == NULL) || (strlen(buffer->callback) == 0)) {
	    ConfigureInfo configInfo;
	    
	    configInfo.size = BUFFER_DONT_MODIFY;
	    configInfo.shared = BUFFER_DONT_MODIFY;
	    configInfo.xAttach = BUFFER_DONT_MODIFY;
	    configInfo.state = "free";
	    configInfo.callback = NULL;
	    ConfigureBuffer(interp, buffer, &configInfo);
	} else if (*(buffer->callback) == '@') {
	    bufCallback callback;
	    ClientData dummy, callbackCD;

	    Tutil_GetRegisteredCommand(interp, buffer->callback, &callbackCD, &dummy);
	    callback = (bufCallback) callbackCD;
	    (callback)(buffer);
	    return;
	} else {
	    char bufferName[64];
		
	    sprintf(bufferName, "buffer%d", buffer->id);
	    Tcl_VarEval(interp, buffer->callback, " ", bufferName, NULL);
	    return;
	}
    }
}
