/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmBufferCapture.c,v 1.9 1996/02/06 22:40:25 kpatel Exp $
 *
 *  A CMT object that dumps buffers to disk (only audio currently).
 *  Configuration options determine the type of data snooped (and thus
 * the type of chunkfile created, as well as details like the filename to
 * capture to. Like the bufferPrinter object, can also pass the data stream
 * along to another object.
 *
 * ToDo:
 *  1.  install function pointers for the various write procs , rather than
 * lots of switch statements.
 *
 * HISTORY:
       5/29/95  Initial version (only write .au files, not clipfiles)
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <netinet/in.h>
#include "cmAtom.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmChunk.h"
#include "cmLTS.h"
#include "cmAuFile.h"
#include "cmUtil.h"

#include "libaudio.h"
#include "audio_device.h"

#define	ONE_BILLION	1000000000.0
#define CM_TEXT_SEGMENT 0xdabadaba

typedef enum  {
    NO_CF = -1, 
    CF_INITIAL = 1,     /* open but no data (or header) written */
    CF_OPEN,            /* open, header written, ready for data */
    CF_CLOSED          
} CF_Status;


/*
 * One of these structures is created for each bufferCapture object.
 */
typedef struct BufferCapture {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    char *outCmd;		/* Command to send buffers to. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;		/* ClientData for outCCmd. */
    char *acceptHandle;
    int verbose;		/* How verbose to be (0 = quiet) */
    ClientData lts;		/* LTS for object.  Used to print arrival
				 * time of message at accept interface. */
    char *ltsName;		/* Tcl name for lts */

    int fd;			/* file descriptor for file to dump data to */
    long fileLength;		/* length of file (currently unused) */
    char *fileName;
    enum {Overwrite, Append} fileMode;
    CF_Status curFileStatus;
    enum {automatic, au, auclipfile} fileType;
    tChunkFileState *cfState;
    MediaIndepHeader mediaHdr;
} BufferCapture;

static int nextBufferCaptureNum = 0;	/* Variable for next unique object
					 * name. */
/*
 * Procedures defined in this module
 */
int			BufferCaptureModuleInit(Tcl_Interp *interp);

static BufferCapture *	CreateNewBufferCapture(Tcl_Interp *interp, char *name);
static int		DestroyBufferCapture(BufferCapture *objPtr);
static int		ConfigureBufferCapture(BufferCapture *objPtr,
			    int argc, char **argv);
static int		BufferCaptureCreateCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		BufferCaptureObjCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		BufferCapture_Accept(BufferCapture *objPtr,
			    SB_List *sbl);
static int		BufferCapture_ChangeOutCmd(BufferCapture *objPtr,
			    char *newOut);

static int		BufferCapture_NewFileName(BufferCapture *objPtr,
			    char *filename);

static int		FinalizeClipFile(BufferCapture *bc);

static void		PrintMediaIndepHeader(char *ptr, int len);
static void		PrintMJPEGHeader(char *ptr);
static void		PrintSparcAuHdr(char *ptr);
static void		PrintSparcAuPktHdr(char *data);


/************************************************************
 *
 * CreateNewBufferCapture
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

BufferCapture *
CreateNewBufferCapture(Tcl_Interp *interp, char *name)
{

    BufferCapture *new;
    char *acceptName;

    new = (BufferCapture *) ckalloc(sizeof(BufferCapture));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy (new->name, name);
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->verbose = 2;
    new->lts = NULL;
    new->ltsName = NULL;

    new->fd = -1;
    new->fileName = NULL;
    new->fileMode = Overwrite;
    new->curFileStatus = NO_CF;
    new->fileType = au;

    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+9);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) BufferCapture_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    return new;
}

/************************************************************
 *
 * DestroyBufferCapture
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/
static int
DestroyBufferCapture(BufferCapture *objPtr)
{
    /*
     * Remove command name
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
        Tk_Release(objPtr->lts);
        objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
	
    if (objPtr->fd != -1) {
	FinalizeClipFile(objPtr);
    }
    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureBufferCapture
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

static int
ConfigureBufferCapture(BufferCapture * objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
  
    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}

	sprintf(str, "{%d}", objPtr->verbose);
	Tcl_AppendResult(interp, "{-verbose {2} ", str, "} ", NULL);

        if (objPtr->ltsName == NULL) {
            Tcl_AppendResult(interp, "{-lts {NULL} {NULL} } ", NULL);
        } else {
            Tcl_AppendResult(interp, "{-lts {NULL} {",
                             objPtr->ltsName, "} } ", (char *) NULL);
        }


	if (objPtr->fileName == NULL) {
	    Tcl_AppendResult(interp, "{-filename {NULL} {NULL} }", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-filename {NULL} {", objPtr->fileName,"} }",
			     NULL);
	}
	return TCL_OK;
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
				 "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'v') && (strncmp("-verbose", argv[0], len) == 0)) {
	    sprintf(str, "%d", objPtr->verbose);
	    Tcl_AppendResult(interp, "-verbose 2 ", str, NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->ltsName == NULL) {
		Tcl_AppendResult(interp, "-lts {NULL} {NULL}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {NULL} {",
				 objPtr->ltsName, "}",  (char *) NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'f') && (strncmp("-filename", argv[0], len) == 0)) {
	    if (objPtr->fileName == NULL) {
		Tcl_AppendResult(interp, "-filename {NULL} {NULL} ", NULL);
	    } else {
		Tcl_AppendResult(interp, "-filename {NULL} {", 
				 objPtr->fileName,"}", NULL);
	    }
	    return TCL_OK;
	}
	Tcl_AppendResult(interp, "bad configuration option: ",
			 argv[0], (char *)NULL);
	return TCL_ERROR;
	
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
noValueError:
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (BufferCapture_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }

	/* --------------------------- VERBOSE ---------------------------- */
	} else if ((c == 'v') && (strncmp("-verbose", argv[cntr], len) == 0)) {
	    int newVerbose;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newVerbose) != TCL_OK) {
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->verbose = newVerbose;

	/* --------------------------- FILENAME ---------------------------- */
	} else if ((c == 'f') && (strncmp("-filename", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (BufferCapture_NewFileName(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
        /* --------------------------- LTS -------------------------------- */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    
	    if (strlen(argv[cntr]) == 0) {
		newLts = NULL;
	    } else {
		newLts = Tcm_FindAtom(interp, argv[cntr]);
		if (newLts == NULL) {
		    Tcl_AppendResult(interp, argv[cntr],
				     " is not a valid lts.", NULL);
		    return TCL_ERROR;
		}
	    }

	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }

            /*
             * Release our lock on the old lts...
             */
            if (objPtr->lts != NULL) {
                Tk_Release(objPtr->lts);
                objPtr->lts = NULL;
            }
	    
            /*
             * And set lock and arrange triggers for new one.
             */
	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
		objPtr->ltsName = (char *) ckalloc(sizeof(argv[cntr])+1);
		strcpy(objPtr->ltsName,argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }
		
	/* --------------------------- ERROR ------------------------------ */
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			     argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}


/************************************************************
 *
 * BufferCapture_NewFileName
 *
 * Parameters:
 *    objPtr - buffer capture object
 *    filename - file to open
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *    Any currently open file is finalized and closed. New file
 * name is assigned.
 ************************************************************/
static int
BufferCapture_NewFileName(BufferCapture *objPtr, char *filename)
{
    if (objPtr->fd != -1) {
	FinalizeClipFile(objPtr);
    }

    objPtr->fileName = strdup(filename);
    objPtr->curFileStatus = CF_CLOSED;
 
    return TCL_OK;
}

/************************************************************
 *
 * BufferCapture_OpenFile
 *
 * Parameters:
 *    objPtr - buffer capture object
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     File is opened for output according to mode. 
 *  NOTE: Appending is currently incompletely supported!
 ************************************************************/
static int
BufferCapture_OpenFile(BufferCapture *objPtr)
{
    if (objPtr->fileName == NULL) {
	Tcl_AppendResult(objPtr->interp, objPtr->name, 
			 ": no file name specified to open\n", NULL);
	return TCL_ERROR;
    }

    if (objPtr->curFileStatus != CF_CLOSED) {
	Tcl_AppendResult(objPtr->interp, objPtr->name, 
			 ": file already open\n", NULL);
	return TCL_ERROR;
    }

    /* sparc au file */
    if (objPtr->fileType == au) {
	objPtr->fd = open(objPtr->fileName, 
			  (O_RDWR | O_CREAT | 
			   ((objPtr->fileMode == Append) ? 0 : O_TRUNC)), 0666);
    /* CMT AuClipFile */
    } else if (objPtr->fileType = auclipfile) {
	if ((objPtr->cfState = Chunk_OpenFile(objPtr->fileName, "cw")) == NULL) {
	    Tcl_AppendResult(objPtr->interp, "Couldn't open chunk file: ",
			     objPtr->fileName, NULL);
	    return TCL_ERROR;
	}
    }

    if (objPtr->fd < 0) {
	Tcl_AppendResult(objPtr->interp, objPtr->name, 
			 ": unable to open output file\n", NULL);
	return TCL_ERROR; 
    }

    if (objPtr->fileMode != Append) {
	objPtr->fileLength = 0;
    } else  {
	;  /* need to read file length and seek to eof --- */
    }

    if (objPtr->fileLength > 0) {
	objPtr->curFileStatus = CF_OPEN;     /* open, some data written */
    } else {
	objPtr->curFileStatus = CF_INITIAL;   /* no data written */
    }

    return TCL_OK;
}


/*
 * write media independent header 
 * initially sets length to zero. 
 * this must be updated when the file is finalized 
 */
static int
BufferCapture_WriteMediaIndHeader(BufferCapture *objPtr)
{
    objPtr->mediaHdr.media = htonl(AUDIO_MEDIA_CODE);
    objPtr->mediaHdr.format = htonl(SPARC_AU_FORMAT_CODE);
    objPtr->mediaHdr.device = htonl(SPARC_AU_DEV_CODE);
    objPtr->mediaHdr.version = htonl(3);
    objPtr->mediaHdr.num_frames = htonl(1);
    objPtr->mediaHdr.length_s = htonl(0);
    objPtr->mediaHdr.length_ns = htonl(0);
	
}

/*
 * write (or rewrite) audio clip file header 
 */
static int
BufferCapture_WriteAUClipFileHeader(BufferCapture *objPtr, char *data)
{

}


/*
 * write (or rewrite) au file header 
 */
static int
BufferCapture_WriteAUHeader(BufferCapture *objPtr, char *data)
{
    int format;
    unsigned int hdr_len = 24;
    RawSparcAuHeader au_hdr;         /* written to disk */
    SparcAuBufferHeader *bufHdr;      /* comes from scatter buffer */
	
    if (objPtr->verbose >= 2) printf("attempting to write au file header\n");

    au_hdr.magic = SUN_MAGIC;
    au_hdr.dataSize = SUN_UNSPEC;  /* Allows for audio file of unspecified
				      length */

    bufHdr = (SparcAuBufferHeader *)data;
	
    format = ntohl(bufHdr->dataFormat);
    if (format == SUN_ULAW) {
	au_hdr.dataFormat = AUDIO_ENCODING_ULAW;
    } else {
	au_hdr.dataFormat = AUDIO_ENCODING_LINEAR;
    }
    au_hdr.samplingRate = ntohl(bufHdr->samplingRate);
    au_hdr.channelCount = ntohl(bufHdr->channelCount);
    au_hdr.dataLocation = hdr_len;  /* skip 6 4-byte integer header */
	
    if (objPtr->curFileStatus != CF_OPEN) {  
	/* if the file is not in open state, then the file has no header */

	if ((write(objPtr->fd, &au_hdr, hdr_len)) != hdr_len) {
	    Tcl_AppendResult(objPtr->interp, objPtr->name, 
			     ": Error writing audio file header\n",
			     (char *)NULL);
			
	    return TCL_ERROR;
	}
	if (objPtr->verbose >= 2) printf("wrote au file header\n");

    } else { /* rewrite header */
	if (lseek(objPtr->fd, 0, SEEK_CUR) != -1) {
	    if ((write(objPtr->fd, &au_hdr, hdr_len)) != hdr_len) {
		Tcl_AppendResult(objPtr->interp, objPtr->name, 
				 ": Error rewriting audio file header\n",
				 (char *)NULL);
				
		return TCL_ERROR;
	    }
	    if (objPtr->verbose >= 2) printf("rewrote au file header\n");
	} else {
	    Tcl_AppendResult(objPtr->interp, objPtr->name, 
			     ": Error seeking to audio file header\n",
			     (char *)NULL);
	    return TCL_ERROR;
	}
    }

    objPtr->curFileStatus = CF_OPEN;
    return TCL_OK;
}

static int
BufferCapture_WriteHeader(BufferCapture *objPtr, char *data)
{
    if (objPtr->fd != -1) {
	switch(objPtr->fileType) {
	  case au:
	    return BufferCapture_WriteAUHeader(objPtr, data);
	    break; /* super defensive programming ;-) */
	  case auclipfile:
	    return BufferCapture_WriteAUClipFileHeader(objPtr, data);
	    break;
	  default:
	    return TCL_ERROR;
	    break;
	}
    } else {
	return TCL_ERROR;
    }
}


/************************************************************
 * 
 * BufferCapture_WriteBuffer
 *
 * Parameters:
 *   objPtr
 *   data
 *
 * Return: 
 *    Standard Tcl return value
 * 
 * Action:
 *    Data is written at the current location in the file. 
 *
 ************************************************************/
static int
BufferCapture_WriteBuffer(BufferCapture *objPtr, char *data, int len)
{
    int err;
	
    err = write(objPtr->fd, data, len);
    if (objPtr->verbose >= 2) {
	printf("wrote %i bytes\n", err);
    }

    return TCL_OK;
}

/************************************************************
 *
 * BufferCaptureCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     A new object of type BufferCapture is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

static int
BufferCaptureCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    BufferCapture *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextBufferCaptureNum++;
    sprintf(new_name, "%sBufferCapture%d", argv[1], nextBufferCaptureNum);

    new = CreateNewBufferCapture(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, BufferCaptureObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureBufferCapture(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * BufferCaptureObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     BufferCapture object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

static int
BufferCaptureObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    BufferCapture *objPtr = (BufferCapture *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureBufferCapture(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyBufferCapture(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl;
	int rv;
		
	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			     argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}

	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	    ckfree((char *) sbl);
	    Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
			     (char *) NULL);
	    return TCL_ERROR;
	}

	rv =  BufferCapture_Accept(objPtr, sbl);
	FreeSBList(sbl);
	return rv;
    } else if ((c == 'o') && (strncmp("open", argv[1], len) == 0)) {
	return BufferCapture_OpenFile(objPtr);
    } else if ((c == 'c') && (strncmp("close", argv[1], len) == 0)) {
	return FinalizeClipFile(objPtr);
    }
	
    Tcl_AppendResult(interp, "Bad command option: ", argv[0],
		     argv[1], (char *) NULL);
  usage:
    Tcl_AppendResult(interp,
		     "\n<command> can be one of:\n",
		     "  configure ?-slot ?value??\n",
		     "  destroy\n",
		     "  accept <ScatterBufferList>\n",
		     "  open\n",
		     "  close\n",
		     (char *) NULL);
    return TCL_ERROR;
}
	
/************************************************************
 *
 * BufferCaptureModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
BufferCaptureModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "bufferCapture", BufferCaptureCreateCmd,
			NULL, (void (*)()) NULL);
    return TCL_OK;
}

/************************************************************
 *
 * BufferCapture_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -	scatter buffer list containing frame data
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
BufferCapture_Accept(BufferCapture *objPtr, SB_List *sbl)
{
    int i, j, index;
    ScatterBuffer *sbPtr;
    Buffer *bufPtr;
    char *data;
    int offset, len;
    int cookie, cookieLen;
    double now;

    AttachToAllBuffers(objPtr->interp, sbl, "BufferCapture_accept");

    sbl = CopySBList(sbl);

    /*
     * Parse the buffers in the sbl.  Each buffer is formatted as
     *	cookie	<4>	Type of data, defined in cookies.h
     *	len	<4>	Length of segment.
     *	data	<len>	<len> bytes of data
     */
    if (objPtr->verbose >= 1) {
	if (objPtr->lts != NULL) {
	    now = LTS_Value(objPtr->lts);
	} else {
	    now = Tcm_ReadSysClock();
	}
	printf ("%s.accept (@%1.3lf)\n", objPtr->name, now);
    }

    for (i=0; i<sbl->count; i++) {
	sbPtr = sbl->sbs[i];
	for (j = 0; j<sbPtr->count; j++) {
	    bufPtr = sbPtr->buffers[j];
	    offset = sbPtr->offsets[j];
	    len = sbPtr->lens[j];
	    data = bufPtr->data+offset;
	    index = 0;
	    while (index < len) {
		memcpy(&cookie, data+index, 4);
		memcpy(&cookieLen, data+index+4, 4);
		cookie = ntohl(cookie);
		cookieLen = ntohl(cookieLen);
		switch(cookie) {
		  case CM_MEDIA_INDEP_CF:
		    if (objPtr->verbose >= 2) 
			printf ("Media Indepent Header: %d bytes\n",
				cookieLen);
		    if (objPtr->verbose >= 3) {
			PrintMediaIndepHeader (data+index+8, cookieLen);
		    }
		    break;
		  case CM_FOT:
		    if (objPtr->verbose >= 2) 
			printf ("CM_FOT: %d bytes\n", cookieLen);
		    break;
		  case CM_FTT:
		    if (objPtr->verbose >= 2) 
			printf ("CM_FTT: %d bytes\n", cookieLen);
		    break;
		  case CM_DATA_SEGMENT:
		    if (objPtr->verbose >= 2) 
			printf ("CM_DATA_SEGMENT: %d bytes\n", cookieLen);
		    if (objPtr->fd != -1) {
			char *bufptr; Buffer *dummy;
			bufptr = GetFromSB(sbPtr, CM_DATA_SEGMENT, &dummy, NULL);
			if (bufptr) {
			    BufferCapture_WriteBuffer(objPtr, bufptr, cookieLen);
			} else {
			    fprintf(stderr, "GetFromSB = NULL\n");
			}
		    }
		    break;
		  case CM_MJPEG_CF:
		    if (objPtr->verbose >= 2) 
			printf ("CM_MJPEG_CF: %d bytes\n", cookieLen);
		    break;
		  case CM_MJPEG_BUF_HDR:
		    if (objPtr->verbose >= 2) 
			printf ("Motion JPEG Header: %d bytes\n",
				cookieLen);
		    if (objPtr->verbose >= 3) {
			PrintMJPEGHeader (data);
		    }
		    break;
		  case CM_SPARC_AU_CF:
		    if (objPtr->verbose >= 2) 
			printf ("CM_SPARC_AU_CF: %d bytes\n", cookieLen);
		    break;
		  case CM_SPARC_AU_BUF_HDR:
		    if (objPtr->verbose >= 2) 
			printf ("CM_SPARC_AU_BUF_HDR: %d bytes\n",
				cookieLen);
		    if (objPtr->verbose >= 3) {
			PrintSparcAuHdr (data);
		    }
		    if (objPtr->curFileStatus == CF_INITIAL)
			BufferCapture_WriteHeader(objPtr, data);
		    break;
		  case CM_SPARC_AU_PKT_HDR:
		    if (objPtr->verbose >= 2)
			printf("CM_SPARC_AU_PKT_HDR: %d bytes\n",
			       cookieLen);
		    if (objPtr->verbose >= 3) {
			PrintSparcAuPktHdr (data);
		    }
		    break;
		  default:
		    if (objPtr->verbose >= 2) 
			printf ("Unknown cookie %x: %d bytes\n",
				cookie, cookieLen);
		    break;
		}
		index += cookieLen + 8;
	    }
	}
    }

    /*
     * Pass on buffers to outCmd
     */
    if (objPtr->outCmd != NULL) {
	if (objPtr->outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree(objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and 
	     * call using Tcl_Eval. 
	     */
	    char *sbStr = MakeSBListString(sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ", 
			sbStr, (char *) NULL); 
	    ckfree(sbStr);
	}
    }
    DetachToAllBuffers(objPtr->interp, sbl, "BufferCapture:acceptover");	
    FreeSBList(sbl);
    return TCL_OK;
}

/************************************************************
 *
 * BufferCapture_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
BufferCapture_ChangeOutCmd(BufferCapture *objPtr, char *newOut) 
{
    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

#ifdef SECURE_CMT
	if (newOut[0] != '@') {
	    Tcl_SetResult(objPtr->interp, "Cannot set Tcl outCmd because this program was compiled with -DSECURE_CMT.\nRecompile without security to use arbitrary Tcl outCmd's\n", TCL_STATIC);
	    return TCL_ERROR;
	}
#endif

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
	    &(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Couldn't find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    

/* 
 * FinalizeClipFile
 *
 * Parameters:
 *
 * Results:
 *
 */
static int
FinalizeClipFile(BufferCapture *bc)
{
    /* rewrite header(s) if necessary */

    /* close files */
    close(bc->fd);
    bc->fd = -1;

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * PrintMediaIndepHeader --
 *
 *	Pretty print the contents of a MediaIndepHeader structure.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *ptr	Pointer to buffer with MediaIndepHeader
 *	int len		Size of buffer (ignored)
 *
 *--------------------------------------------------------------
 */
static void
PrintMediaIndepHeader (char *ptr, int len)
{
    MediaIndepHeader hdr;
    char *media;
    char *format;
    char *device;

    memcpy (&hdr, ptr, sizeof(MediaIndepHeader));

    hdr.media = ntohl(hdr.media);
    hdr.format = ntohl(hdr.format);
    hdr.device = ntohl(hdr.device);

    if (hdr.media == VIDEO_MEDIA_CODE) {
	media = "Video";
    } else if (hdr.media == AUDIO_MEDIA_CODE) {
	media = "Audio";
    } else {
	media = "UNKNOWN";
    }
    if (hdr.format == MJPEG_FORMAT_CODE) {
	format = "Motion JPEG";
    } else if (hdr.format == MPEG_FORMAT_CODE) {
	format = "MPEG";
    } else if (hdr.format == SPARC_AU_FORMAT_CODE) {
	format = "Sparc Audio";
    } else {
	format = "UNKNOWN";
    }
    if (hdr.device == PARALLAX_DEV_CODE) {
	device = "Parallax";
    } else if (hdr.device == J300_DEV_CODE) {
	device = "J300";
    } else if (hdr.device == SPARC_AU_DEV_CODE) {
	device = "Sparc Audio";
    } else {
	device = "UNKNOWN";
    }
    printf ("  media = %s\n", media);
    printf ("  format = %s\n", format);
    printf ("  device = %s\n", device);
    printf ("  version = %d\n", ntohl(hdr.version));
    printf ("  num_frames = %d\n", ntohl(hdr.num_frames));
    printf ("  length_s = %d\n", ntohl(hdr.length_s));
    printf ("  length_ns = %d\n", ntohl(hdr.length_ns));
}

/*
 *--------------------------------------------------------------
 *
 * PrintMJPEGHeader --
 *
 *	Pretty print the contents of a MJPEGBufferHeader structure.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *ptr	Pointer to buffer with MJPEGBufferHeader
 *
 *--------------------------------------------------------------
 */
static void
PrintMJPEGHeader (char *ptr)
{
    MJPEGBufferHeader hdr;
    char *subformat;
    char *device;

    memcpy (&hdr, ptr, sizeof(MJPEGBufferHeader));

    hdr.width = ntohl(hdr.width);
    hdr.height = ntohl(hdr.height);
    hdr.device = ntohl(hdr.device);
    hdr.subformat = ntohl(hdr.subformat);
    hdr.qfactor = ntohl(hdr.qfactor);
    hdr.logStart_s = ntohl(hdr.logStart_s);
    hdr.logStart_ns = ntohl(hdr.logStart_ns);
    hdr.logEnd_s = ntohl(hdr.logEnd_s);
    hdr.logEnd_ns = ntohl(hdr.logEnd_ns);

    if (hdr.subformat == VID_FRAME_CODE) {
	subformat = "vid";
    } else if (hdr.subformat == JFIF_FRAME_CODE) {
	subformat = "JFIF";
    } else if (hdr.subformat == JFIF_NO_HUFFMAN_FRAME_CODE) {
	subformat = "JFIF without huffman";
    } else if (hdr.subformat == RAW_JPEG_NO_HEADERS_CODE) {
	subformat = "raw JPEG";
    } else {
	subformat = "UNKNOWN";
    }

    if (hdr.device == PARALLAX_DEV_CODE) {
	device = "Parallax";
    } else if (hdr.device == J300_DEV_CODE) {
	device = "J300";
    } else if (hdr.device == SPARC_AU_DEV_CODE) {
	device = "Sparc Audio";
    } else {
	device = "UNKNOWN";
    }
    printf ("  width = %d\n", hdr.width);
    printf ("  height = %d\n", hdr.height);
    printf ("  device = %s\n", device);
    printf ("  subformat = %s\n", subformat);
    printf ("  qfactor = %d\n", hdr.qfactor);
    printf ("  logStart = %1.3lf\n", 
		hdr.logStart_s+hdr.logStart_ns/ONE_BILLION);
    printf ("  logEnd = %1.3lf\n", 
		hdr.logEnd_s+hdr.logEnd_ns/ONE_BILLION);
}

/*
 *--------------------------------------------------------------
 *
 * PrintSparcAuHdr --
 *
 *	Pretty print the contents of a SparcAuBufferHeader structure.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *ptr	Pointer to buffer with SparcAuBufferHeader
 *	int len		Size of buffer (ignored)
 *
 *--------------------------------------------------------------
 */
static void
PrintSparcAuHdr (char *ptr)
{
    SparcAuBufferHeader hdr;
    char *dataFormat;
    char *device;

    memcpy (&hdr, ptr, sizeof(SparcAuBufferHeader));

    hdr.device = ntohl(hdr.device);
    hdr.dataFormat = ntohl(hdr.dataFormat);
    hdr.samplingRate = ntohl(hdr.samplingRate);
    hdr.channelCount = ntohl(hdr.channelCount);
    hdr.logStart_s = ntohl(hdr.logStart_s);
    hdr.logStart_ns = ntohl(hdr.logStart_ns);
    hdr.logEnd_s = ntohl(hdr.logEnd_s);
    hdr.logEnd_ns = ntohl(hdr.logEnd_ns);

    if (hdr.dataFormat == SUN_ULAW) {
	dataFormat = "Mu Law";
    } else if (hdr.dataFormat == SUN_LIN_8) {
	dataFormat = "Linear 8 bit";
    } else if (hdr.dataFormat == SUN_LIN_16) {
	dataFormat = "Linear 16 bit";
    } else {
	dataFormat = "UNKNOWN";
    }

    if (hdr.device == PARALLAX_DEV_CODE) {
	device = "Parallax";
    } else if (hdr.device == J300_DEV_CODE) {
	device = "J300";
    } else if (hdr.device == SPARC_AU_DEV_CODE) {
	device = "Sparc Audio";
    } else {
	device = "UNKNOWN";
    }

    printf ("Sparc Audio Buffer Header:\n");
    printf ("  device = %s\n", device);
    printf ("  dataFormat = %s\n", dataFormat);
    printf ("  samplingRate = %d\n", hdr.samplingRate);
    printf ("  channelCount = %d\n", hdr.channelCount);
    printf ("  logStart = %1.3lf\n", 
		hdr.logStart_s+hdr.logStart_ns/ONE_BILLION);
    printf ("  logEnd = %1.3lf\n", 
		hdr.logEnd_s+hdr.logEnd_ns/ONE_BILLION);
}


static void
PrintSparcAuPktHdr (char *data) {
    SparcAuPktHeader hdr;

    memcpy ((char *)&hdr, data, sizeof(SparcAuPktHeader));

    printf(" Sparc Au Segment Header: \n");
    PrintSparcAuHdr ((char *)&hdr);

    hdr.segID = ntohl(hdr.segID);
    hdr.pktFactor = ntohl(hdr.pktFactor);
    hdr.totDataLen = ntohl(hdr.totDataLen);
    hdr.pktID = ntohl(hdr.pktID);

    printf("   segID = %d\n", hdr.segID);
    printf("   pktFactor = %d\n", hdr.pktFactor);
    printf("   pktID = %d\n", hdr.pktID);
    printf("   totDataLen = %d\n", hdr.totDataLen);
}

