/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * tcmBufferPrint.c --
 *
 *  A simple CMT object (created using the bufferPrinter command)
 *  that prints the contents of buffers, which are then passed on
 *  to their outCmd.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <netinet/in.h>
#include "cmAtom.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmAuFile.h"
#include "cmLTS.h"
#include "cmUtil.h"
#define	ONE_BILLION	1000000000.0

/*
 * One of these structures is created for each bufferPrinter object.
 */
typedef struct BufferPrinter {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    char *outCmd;		/* Command to send buffers to. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;		/* ClientData for outCCmd. */
    char *acceptHandle;
    int verbose;		/* How verbose to be (0 = quiet) */
    ClientData lts;		/* LTS for object.  Used to print arrival
				 * time of message at accept interface. */
    char *ltsName;		/* Tcl name for lts */
} BufferPrinter;

static int nextBufferPrinterNum = 0;	/* Variable for next unique object
					 * name. */
/*
 * Procedures defined in this module
 */
int			BufferPrinterModuleInit(Tcl_Interp *interp);

static BufferPrinter *	CreateNewBufferPrinter(Tcl_Interp *interp, char *name);
static int		DestroyBufferPrinter(BufferPrinter *objPtr);
static int		ConfigureBufferPrinter(BufferPrinter *objPtr,
			    int argc, char **argv);
static int		BufferPrinterCreateCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		BufferPrinterObjCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		BufferPrinter_Accept(BufferPrinter *objPtr,
			    SB_List *sbl);
static int		BufferPrinter_ChangeOutCmd(BufferPrinter *objPtr,
			   char *newOut);
static void		PrintMediaIndepHeader(char *ptr, int len);
static void		PrintMJPEGHeader(char *ptr);
static void		PrintSparcAuHdr(char *ptr);

/************************************************************
 *
 * CreateNewBufferPrinter
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

BufferPrinter *
CreateNewBufferPrinter(Tcl_Interp *interp, char *name)
{

    BufferPrinter *new;
    char *acceptName;

    new = (BufferPrinter *) ckalloc(sizeof(BufferPrinter));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy (new->name, name);
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->verbose = 2;
    new->lts = NULL;
    new->ltsName = NULL;

    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+9);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) BufferPrinter_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    return new;
}

/************************************************************
 *
 * DestroyBufferPrinter
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

static int
DestroyBufferPrinter(BufferPrinter *objPtr)
{
    /*
     * Remove command name
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
        Tk_Release(objPtr->lts);
        objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureBufferPrinter
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

static int
ConfigureBufferPrinter(BufferPrinter * objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
  
    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}

	sprintf(str, "{%d}", objPtr->verbose);
	Tcl_AppendResult(interp, "{-verbose {2} ", str, "} ", NULL);

        if (objPtr->ltsName == NULL) {
            Tcl_AppendResult(interp, "{-lts {NULL} {NULL} } ", NULL);
        } else {
            Tcl_AppendResult(interp, "{-lts {NULL} {",
                             objPtr->ltsName, "} } ", (char *) NULL);
        }
	return TCL_OK;
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'v') && (strncmp("-verbose", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->verbose);
	    Tcl_AppendResult(interp, "-verbose {2} ", str, NULL);
	    return TCL_OK;
        } else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
            if (objPtr->ltsName == NULL) {
                Tcl_AppendResult(interp, "-lts {NULL} {NULL}", NULL);
            } else {
                Tcl_AppendResult(interp, "-lts {NULL} {",
                                 objPtr->ltsName, "}",  (char *) NULL);
            }
            return TCL_OK;
        }

	Tcl_AppendResult(interp, "bad configuration option: ",
			  argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
noValueError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (BufferPrinter_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }

	/* --------------------------- VERBOSE ---------------------------- */
	} else if ((c == 'v') && (strncmp("-verbose", argv[cntr], len) == 0)) {
	    int newVerbose;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newVerbose) != TCL_OK) {
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->verbose = newVerbose;

        /* --------------------------- LTS -------------------------------- */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    
	    if (strlen(argv[cntr]) == 0) {
		newLts = NULL;
	    } else {
		newLts = Tcm_FindAtom(interp, argv[cntr]);
		if (newLts == NULL) {
		    Tcl_AppendResult(interp, argv[cntr],
				     " is not a valid lts.", NULL);
		    return TCL_ERROR;
		}
	    }

	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }

            /*
             * Release our lock on the old lts...
             */
            if (objPtr->lts != NULL) {
                Tk_Release(objPtr->lts);
                objPtr->lts = NULL;
            }
	    
            /*
             * And set lock and arrange triggers for new one.
             */
	    objPtr->lts = newLts;
	    if (newLts != NULL) {
                Tk_Preserve (objPtr->lts);
		objPtr->ltsName = (char *) ckalloc(sizeof(argv[cntr])+1);
		strcpy(objPtr->ltsName,argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }

	/* --------------------------- ERROR ------------------------------ */
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			  argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}

/************************************************************
 *
 * BufferPrinterCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the mjpegResource command.
 *     A new object of type BufferPrinter is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

static int
BufferPrinterCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    BufferPrinter *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextBufferPrinterNum++;
    sprintf(new_name, "%sbufferPrinter%d", argv[1], nextBufferPrinterNum);

    new = CreateNewBufferPrinter(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, BufferPrinterObjCmd,
			(ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureBufferPrinter(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * BufferPrinterObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     BufferPrinter object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

static int
BufferPrinterObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    BufferPrinter *objPtr = (BufferPrinter *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureBufferPrinter(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyBufferPrinter(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}

	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	  ckfree((char *) sbl);
	  Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
		(char *) NULL);
	  return TCL_ERROR;
	}

	return BufferPrinter_Accept(objPtr, sbl);
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0],
		argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  accept <ScatterBufferList>\n",
                (char *) NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * BufferPrinterModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
BufferPrinterModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "bufferPrinter", BufferPrinterCreateCmd,
			NULL, (void (*)()) NULL);
    return TCL_OK;
}

/************************************************************
 *
 * BufferPrinter_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -	scatter buffer list containing frame data
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
BufferPrinter_Accept(BufferPrinter *objPtr, SB_List *sbl)
{
    int i, j, index;
    ScatterBuffer *sbPtr;
    Buffer *bufPtr;
    char *data;
    int offset, len;
    int cookie, cookieLen;
    double now;

    /*
     * Parse the buffers in the sbl.  Each buffer is formatted as
     *	cookie	<4>	Type of data, defined in cookies.h
     *	len	<4>	Length of segment.
     *	data	<len>	<len> bytes of data
     */
    if (objPtr->verbose >= 1) {
	if (objPtr->lts != NULL) {
	    now = LTS_Value(objPtr->lts);
	} else {
	    now = Tcm_ReadSysClock();
	}
	printf ("%s.accept (@%1.3lf)\n", objPtr->name, now);
    }
    if (objPtr->verbose >= 2) {
	for (i=0; i<sbl->count; i++) {
	    sbPtr = sbl->sbs[i];
	    for (j = 0; j<sbPtr->count; j++) {
		bufPtr = sbPtr->buffers[j];
		offset = sbPtr->offsets[j];
		len = sbPtr->lens[j];
		data = bufPtr->data+offset;
		index = 0;
		while (index < len) {
		    memcpy(&cookie, data+index, 4);
		    memcpy(&cookieLen, data+index+4, 4);
		    cookie = ntohl(cookie);
		    cookieLen = ntohl(cookieLen);
		    switch(cookie) {
			case CM_MEDIA_INDEP_CF:
			    printf ("Media Indepent Header: %d bytes\n",
				    cookieLen);
			    if (objPtr->verbose >= 3) {
				PrintMediaIndepHeader (data+index+8, cookieLen);
			    }
			    break;
			case CM_FOT:
			    printf ("CM_FOT: %d bytes\n", cookieLen);
			    break;
			case CM_FTT:
			    printf ("CM_FTT: %d bytes\n", cookieLen);
			    break;
			case CM_DATA_SEGMENT:
			    printf ("CM_DATA_SEGMENT: %d bytes\n", cookieLen);
			    break;
			case CM_MJPEG_CF:
			    printf ("CM_MJPEG_CF: %d bytes\n", cookieLen);
			    break;
			case CM_MJPEG_BUF_HDR:
			    printf ("Motion JPEG Header: %d bytes\n",
				    cookieLen);
			    if (objPtr->verbose >= 3) {
				PrintMJPEGHeader (data);
			    }
			    break;
			case CM_SPARC_AU_CF:
			    printf ("CM_SPARC_AU_CF: %d bytes\n", cookieLen);
			    break;
			case CM_SPARC_AU_BUF_HDR:
			    printf ("CM_SPARC_AU_BUF_HDR: %d bytes\n",
				    cookieLen);
			    if (objPtr->verbose >= 3) {
				PrintSparcAuHdr (data);
			    }
			    break;
			default:
			    printf ("Unknown cookie %x: %d bytes\n",
				     cookie, cookieLen);
			    break;
		    }
		    index += cookieLen + 8;
		}
	    }
	}
    }

    /*
     * Pass on buffers to outCmd
     */
    if (objPtr->outCmd == NULL) {
	return TCL_OK;
    }

    if (objPtr->outCCmd != NULL) {
	/*
	 * Need to check if it still exists since object may have
	 * been destroyed.
	 */
	if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
	    objPtr->outCCmd(objPtr->outCD, sbl);
	} else {
	    /*
	     * Out C command has unexpectedly disappeared.
	     */
	    objPtr->outCCmd = NULL;
	    ckfree(objPtr->outCmd);
	    objPtr->outCmd = NULL;
	}
    } else {
	/*
	 * Out command is tcl, make sb list into a string and 
	 * call using Tcl_Eval. 
	 */
	char *sbStr = MakeSBListString(sbl);
	Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ", sbStr, (char *) NULL); 
	ckfree(sbStr);
	FreeSBList(sbl);
    }
	
    return TCL_OK;
}

/************************************************************
 *
 * BufferPrinter_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
BufferPrinter_ChangeOutCmd(BufferPrinter *objPtr, char *newOut) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
	    &(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    


/*
 *--------------------------------------------------------------
 *
 * PrintMediaIndepHeader --
 *
 *	Pretty print the contents of a MediaIndepHeader structure.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *ptr	Pointer to buffer with MediaIndepHeader
 *	int len		Size of buffer (ignored)
 *
 *--------------------------------------------------------------
 */
static void
PrintMediaIndepHeader (char *ptr, int len)
{
    MediaIndepHeader hdr;
    char *media;
    char *format;
    char *device;

    memcpy (&hdr, ptr, sizeof(MediaIndepHeader));

    hdr.media = ntohl(hdr.media);
    hdr.format = ntohl(hdr.format);
    hdr.device = ntohl(hdr.device);

    if (hdr.media == VIDEO_MEDIA_CODE) {
	media = "Video";
    } else if (hdr.media == AUDIO_MEDIA_CODE) {
	media = "Audio";
    } else {
	media = "UNKNOWN";
    }
    if (hdr.format == MJPEG_FORMAT_CODE) {
	format = "Motion JPEG";
    } else if (hdr.format == MPEG_FORMAT_CODE) {
	format = "MPEG";
    } else if (hdr.format == SPARC_AU_FORMAT_CODE) {
	format = "Sparc Audio";
    } else {
	format = "UNKNOWN";
    }
    if (hdr.device == PARALLAX_DEV_CODE) {
	device = "Parallax";
    } else if (hdr.device == J300_DEV_CODE) {
	device = "J300";
    } else if (hdr.device == SPARC_AU_DEV_CODE) {
	device = "Sparc Audio";
    } else {
	device = "UNKNOWN";
    }
    printf ("  media = %s\n", media);
    printf ("  format = %s\n", format);
    printf ("  device = %s\n", device);
    printf ("  version = %d\n", ntohl(hdr.version));
    printf ("  num_frames = %d\n", ntohl(hdr.num_frames));
    printf ("  length_s = %d\n", ntohl(hdr.length_s));
    printf ("  length_ns = %d\n", ntohl(hdr.length_ns));
}

/*
 *--------------------------------------------------------------
 *
 * PrintMJPEGHeader --
 *
 *	Pretty print the contents of a MJPEGBufferHeader structure.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *ptr	Pointer to buffer with MJPEGBufferHeader
 *
 *--------------------------------------------------------------
 */
static void
PrintMJPEGHeader (char *ptr)
{
    MJPEGBufferHeader hdr;
    char *subformat;
    char *device;

    memcpy (&hdr, ptr, sizeof(MJPEGBufferHeader));

    hdr.width = ntohl(hdr.width);
    hdr.height = ntohl(hdr.height);
    hdr.device = ntohl(hdr.device);
    hdr.subformat = ntohl(hdr.subformat);
    hdr.qfactor = ntohl(hdr.qfactor);
    hdr.logStart_s = ntohl(hdr.logStart_s);
    hdr.logStart_ns = ntohl(hdr.logStart_ns);
    hdr.logEnd_s = ntohl(hdr.logEnd_s);
    hdr.logEnd_ns = ntohl(hdr.logEnd_ns);

    if (hdr.subformat == VID_FRAME_CODE) {
	subformat = "vid";
    } else if (hdr.subformat == JFIF_FRAME_CODE) {
	subformat = "JFIF";
    } else if (hdr.subformat == JFIF_NO_HUFFMAN_FRAME_CODE) {
	subformat = "JFIF without huffman";
    } else if (hdr.subformat == RAW_JPEG_NO_HEADERS_CODE) {
	subformat = "raw JPEG";
    } else {
	subformat = "UNKNOWN";
    }

    if (hdr.device == PARALLAX_DEV_CODE) {
	device = "Parallax";
    } else if (hdr.device == J300_DEV_CODE) {
	device = "J300";
    } else if (hdr.device == SPARC_AU_DEV_CODE) {
	device = "Sparc Audio";
    } else {
	device = "UNKNOWN";
    }
    printf ("  width = %d\n", hdr.width);
    printf ("  height = %d\n", hdr.height);
    printf ("  device = %s\n", device);
    printf ("  subformat = %s\n", subformat);
    printf ("  qfactor = %d\n", hdr.qfactor);
    printf ("  logStart = %1.3lf\n", 
		hdr.logStart_s+hdr.logStart_ns/ONE_BILLION);
    printf ("  logEnd = %1.3lf\n", 
		hdr.logEnd_s+hdr.logEnd_ns/ONE_BILLION);
}

/*
 *--------------------------------------------------------------
 *
 * PrintSparcAuHdr --
 *
 *	Pretty print the contents of a SparcAuBufferHeader structure.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *ptr	Pointer to buffer with SparcAuBufferHeader
 *	int len		Size of buffer (ignored)
 *
 *--------------------------------------------------------------
 */
static void
PrintSparcAuHdr (char *ptr)
{
    SparcAuBufferHeader hdr;
    char *dataFormat;
    char *device;

    memcpy (&hdr, ptr, sizeof(SparcAuBufferHeader));

    hdr.device = ntohl(hdr.device);
    hdr.dataFormat = ntohl(hdr.dataFormat);
    hdr.samplingRate = ntohl(hdr.samplingRate);
    hdr.channelCount = ntohl(hdr.channelCount);
    hdr.logStart_s = ntohl(hdr.logStart_s);
    hdr.logStart_ns = ntohl(hdr.logStart_ns);
    hdr.logEnd_s = ntohl(hdr.logEnd_s);
    hdr.logEnd_ns = ntohl(hdr.logEnd_ns);

    if (hdr.dataFormat == SUN_ULAW) {
	dataFormat = "Mu Law";
    } else if (hdr.dataFormat == SUN_LIN_8) {
	dataFormat = "Linear 8 bit";
    } else if (hdr.dataFormat == SUN_LIN_16) {
	dataFormat = "Linear 16 bit";
    } else {
	dataFormat = "UNKNOWN";
    }

    if (hdr.device == PARALLAX_DEV_CODE) {
	device = "Parallax";
    } else if (hdr.device == J300_DEV_CODE) {
	device = "J300";
    } else if (hdr.device == SPARC_AU_DEV_CODE) {
	device = "Sparc Audio";
    } else {
	device = "UNKNOWN";
    }

    printf ("Sparc Audio Buffer Header:\n");
    printf ("  device = %s\n", device);
    printf ("  dataFormat = %s\n", dataFormat);
    printf ("  samplingRate = %d\n", hdr.samplingRate);
    printf ("  channelCount = %d\n", hdr.channelCount);
    printf ("  logStart = %1.3lf\n", 
		hdr.logStart_s+hdr.logStart_ns/ONE_BILLION);
    printf ("  logEnd = %1.3lf\n", 
		hdr.logEnd_s+hdr.logEnd_ns/ONE_BILLION);
}
