/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#  define WIN32_LEAN_AND_MEAN
#  include <windows.h>
#  include <winsock.h>
#  undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>

#ifdef NO_STDLIB_H
#include "compat/stdlib.h"
#else
#include <stdlib.h>
#endif

#include <tcl.h>
/* #include <tkConfig.h> */
#include <cmConfig.h>
#include <limits.h>
#include <sys/types.h>
#ifdef HAVE_SHARED_MEMORY
#include <sys/ipc.h>
#include <sys/shm.h>
#endif
#ifndef NO_X11
#include <X11/Xlib.h>
#ifdef HAVE_SHARED_MEMORY
#include <X11/extensions/XShm.h>
#endif
#endif
#include <varargs.h>
#include "cmBufferlist.h"
#include <netinet/in.h>

void 
MergeSBLists(SB_List *mainSB, SB_List *subSB) 
{
  int newcount;
  int i;

  /* Quick sanity checks to avoid core dumping in badly called
     situations.
  */

  if (mainSB == NULL) {
      return;
  }
  if (subSB == NULL) {
      return;
  }

  newcount = mainSB->count + subSB->count;

  if (mainSB->sbs == NULL) {
      mainSB->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *)*newcount);
  } else {
      mainSB->sbs = (ScatterBuffer **) realloc((char *) mainSB->sbs, 
					       (sizeof(ScatterBuffer *)*newcount));
  }

  for (i=mainSB->count; i<newcount; i++) {
    mainSB->sbs[i] = subSB->sbs[i-mainSB->count];
    subSB->sbs[i-mainSB->count] = NULL;
  }
  mainSB->count = newcount;
  ckfree((char *) subSB->sbs);
  subSB->sbs = NULL;
  subSB->count = 0;

  return;
}

int
AttachScatterBuffer(Tcl_Interp *interp, ScatterBuffer *sb, char *debugStr)
{
    int j;

    if (sb == NULL) return TCL_OK;

    for (j=0; j<sb->count; j++) {
	if (sb->buffers[j] != NULL) {
	    AttachBuffer(interp, sb->buffers[j], debugStr);
	}
    }
    return TCL_OK;
}

int 
AttachToAllBuffers(Tcl_Interp *interp, SB_List *sb_list, char *debugStr) 
{
  int i,j;

  if (sb_list == NULL) return TCL_OK;

  for (i=0; i<sb_list->count; i++) {
      if (sb_list->sbs[i] != NULL) {
	  for (j=0; j<sb_list->sbs[i]->count; j++) {

	      if (sb_list->sbs[i]->buffers[j] != NULL) 
		AttachBuffer (interp, sb_list->sbs[i]->buffers[j], debugStr);
	  }
      }
  }

  return TCL_OK;
}

int
DetachScatterBuffer(Tcl_Interp *interp, ScatterBuffer *sb, char *debugStr)
{
    int j;

    if (sb == NULL) return TCL_OK;

    for (j=0; j<sb->count; j++) {
	if (sb->buffers[j] != NULL) {
	    DetachBuffer(interp, sb->buffers[j], debugStr);
	}
    }
    return TCL_OK;
}

int
DetachToAllBuffers(Tcl_Interp *interp, SB_List *sb_list, char *debugStr)
{
  int i,j;

  if (sb_list == NULL) return TCL_OK;

  for (i=0; i<sb_list->count; i++) {
      if (sb_list->sbs[i] != NULL) {
	  for (j=0; j<sb_list->sbs[i]->count; j++) {

	      if (sb_list->sbs[i]->buffers[j] != NULL) 
		DetachBuffer(interp, sb_list->sbs[i]->buffers[j], debugStr);
	  }
      }
  }
  return TCL_OK;
}

Buffer *
MakeBuffer(Tcl_Interp *interp, int size, char *debugStr)
{
    FindQuery query;
    ConfigureInfo info;
    int bufId;
    Buffer *bufPtr;

    /*
     * Create buffers with total_length number of bytes.
     */
    query.minsize = size;
    query.maxsize = BUFFER_DONT_CARE;
    query.unique = 1;
    query.state = NULL;
    query.shared = 0;
    query.xAttach = 0;
    query.tag = NULL;
    
    bufId = FindOrCreateBuffer(interp, &query);

    if (bufId == -1) return NULL;

    /*
     * Get the buffer, configure it and attach it.
     */
    bufPtr = GetBufferFromID(bufId);

    info.size = BUFFER_DONT_MODIFY;
    info.shared = BUFFER_DONT_MODIFY;
    info.xAttach = BUFFER_DONT_MODIFY;
    info.state = "";
    info.callback = NULL;

    ConfigureBuffer(interp, bufPtr, &info);
    AttachBuffer(interp, bufPtr, debugStr);

    return bufPtr;
}

Buffer *
MakeSharedBuffer(Tcl_Interp *interp, int size, char *debugStr)
{
    FindQuery query;
    ConfigureInfo info;
    int bufId;
    Buffer *bufPtr;
#ifdef HAVE_SHARED_MEMORY
    /*
     * Create buffers with total_length number of bytes.
     */
    query.minsize = size;
    query.maxsize = BUFFER_DONT_CARE;
    query.unique = 1;
    query.state = NULL;
    query.shared = 1;
    query.xAttach = 1;
    query.tag = NULL;
    
    bufId = FindOrCreateBuffer(interp, &query);

    if (bufId == -1) return NULL;

    /*
     * Get the buffer, fill it and attach it.
     */
    bufPtr = GetBufferFromID(bufId);

    info.size = BUFFER_DONT_MODIFY;
    info.shared = BUFFER_DONT_MODIFY;
    info.xAttach = BUFFER_DONT_MODIFY;
    info.state = "";
    info.callback = NULL;

    ConfigureBuffer(interp, bufPtr, &info);
    AttachBuffer(interp, bufPtr, debugStr);

    return bufPtr;
#else
    return NULL;
#endif
}


SB_List *MakeSBList(va_alist)
va_dcl
{
  va_list argList;
  ScatterBuffer *sbPtr;
  SB_List *newSBList;
  int idx, i;

  va_start(argList);

  idx = 0;
  while (1) {
    sbPtr = va_arg(argList, ScatterBuffer *);
    if (sbPtr == NULL) break;
    idx++;
  }
  
  newSBList = (SB_List *) ckalloc(sizeof(SB_List));

  newSBList->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *)*idx);
  
  va_start(argList);
  for (i=0; i<idx; i++) {
    newSBList->sbs[i] = va_arg(argList, ScatterBuffer *);
  }
  va_end(argList);

  newSBList->count = idx;

  return newSBList;
}

ScatterBuffer *MakeScatterBuffer(va_alist)
va_dcl
{
  va_list argList;
  Buffer *bufPtr;
  int dlength;
  ScatterBuffer *newSB;
  int idx, i;

  va_start(argList);
  
  idx = 0;
  while (1) {

    bufPtr = va_arg(argList, Buffer *);
    if (bufPtr == NULL) break;
    dlength = va_arg(argList, int);
    idx++;
  }

  va_end(argList);
  
  /* Make a scatterbuffer structure and return it. */

  newSB = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));

  newSB->count = idx;
  newSB->buffers = (Buffer **) ckalloc(sizeof(Buffer *)*idx);
  newSB->offsets = (int *) ckalloc(sizeof(int)*idx);
  newSB->lens = (int *) ckalloc(sizeof(int)*idx);

  va_start(argList);

  for(i=0; i<idx; i++) {
    newSB->buffers[i] = va_arg(argList, Buffer *);
    newSB->lens[i] = va_arg(argList, int);
    newSB->offsets[i] = 0;
  }

  va_end(argList);

  return newSB;
}


/*
 *--------------------------------------------------------------
 *
 * CopyScatterBuffer --
 *
 *	Makes a copy of the scatter buffer wrapper.  The buffers
 *	themselves are *not* copied.
 *
 * Results:
 *	Pointer to duplicate scatter buffer
 *
 * Side effects:
 *	Memory is allocated.
 *
 * Parameters:
 *	ScatterBuffer *sb	Scatter buffer to copy
 *
 *--------------------------------------------------------------
 */
ScatterBuffer *
CopyScatterBuffer(ScatterBuffer *sb) 
{
    ScatterBuffer *rv;

    if (sb == NULL) {
	return NULL;
    }
    rv = (ScatterBuffer *)ckalloc (sizeof(ScatterBuffer));
    rv->count = sb->count;
    if (sb->count == 0) {
	rv->buffers = NULL;
	rv->offsets = NULL;
	rv->lens = NULL;
	return rv;
    }
    rv->buffers = (Buffer **)ckalloc (sb->count * sizeof(Buffer *));
    rv->offsets = (int *)ckalloc (sb->count * sizeof(int));
    rv->lens = (int *)ckalloc (sb->count * sizeof(int));

    memcpy(rv->buffers, sb->buffers, sb->count * sizeof(Buffer *));
    memcpy(rv->offsets, sb->offsets, sb->count * sizeof(int));
    memcpy(rv->lens, sb->lens, sb->count * sizeof(int));
    return rv;
}


/*
 *--------------------------------------------------------------
 *
 * CopySBList --
 *
 *	Make a duplicate of the SB_List passed in.  Only the wrapper
 *	is duplicated -- the buffers themselves are *not* copied.
 *
 * Results:
 *	Pointer to new SB_List
 *
 * Side effects:
 *	Memory is allocated.
 *
 * Parameters:
 *	SB_List *sbList		SB_List to duplicate.
 *
 *--------------------------------------------------------------
 */
SB_List *
CopySBList(SB_List *sbList) 
{
    int i;
    SB_List *rv;

    if (sbList == NULL) {
	return NULL;
    }

    rv = (SB_List *)ckalloc (sizeof(SB_List));
    rv->count = sbList->count;
    if (sbList->count == 0) {
	rv->sbs = NULL;
	return rv;
    }
    rv->sbs = (ScatterBuffer **)ckalloc(sbList->count*sizeof(ScatterBuffer *));
    for (i=0; i<sbList->count; i++) {
	rv->sbs[i] = CopyScatterBuffer(sbList->sbs[i]);
    }
    return rv;
}

void
FreeScatterBuffer(ScatterBuffer *sb) 
{
    if (sb == NULL) return;

    if (sb->buffers != NULL) {
	ckfree((char *) sb->buffers);
	sb->buffers = NULL;
    }
    if(sb->offsets != NULL) {
	ckfree((char *) sb->offsets);
	sb->offsets = NULL;
    }
    if (sb->lens != NULL) {
	ckfree((char *) sb->lens);
	sb->lens = NULL;
    }
    ckfree((char *) sb);
    return;
}

void
FreeSBList(SB_List *sb_list) 
{
  int i;

  if (sb_list == NULL) return;

  if (sb_list->sbs != NULL) {
      for (i=0; i<sb_list->count; i++) {
	  if (sb_list->sbs[i] != NULL) {
	      FreeScatterBuffer(sb_list->sbs[i]);
	      sb_list->sbs[i] = NULL;
	  }
      }
      ckfree((char *) sb_list->sbs);
  }
  
  ckfree((char *) sb_list);
  return;
}

char *
MakeSBListString(SB_List *sb_list)
{
  int i, j;
  Tcl_DString retDStr;
  char *retStr;

  Tcl_DStringInit(&retDStr);

  Tcl_DStringStartSublist(&retDStr);

  for (i=0; i<sb_list->count; i++) {

      Tcl_DStringStartSublist(&retDStr);
      
      for (j=0; j<sb_list->sbs[i]->count; j++) {
	  char tmpStr[100];
	        
	  Tcl_DStringStartSublist(&retDStr);

	  sprintf(tmpStr, "buffer%d", sb_list->sbs[i]->buffers[j]->id);
	  Tcl_DStringAppendElement(&retDStr, tmpStr);

	  sprintf(tmpStr, "%d", sb_list->sbs[i]->offsets[j]);
	  Tcl_DStringAppendElement(&retDStr, tmpStr);

	  sprintf(tmpStr, "%d", sb_list->sbs[i]->lens[j]);
	  Tcl_DStringAppendElement(&retDStr, tmpStr);

	  Tcl_DStringEndSublist(&retDStr);
      }
      
      Tcl_DStringEndSublist(&retDStr);
  }

  Tcl_DStringEndSublist(&retDStr);

  retStr = (char *) ckalloc(Tcl_DStringLength(&retDStr)+1);
  strcpy(retStr, Tcl_DStringValue(&retDStr));
  Tcl_DStringFree(&retDStr);

  return retStr;
}

  
/*
 * GetFromSB
 *
 * Paramters:
 *    sb - ScatterBuffer pointer
 *    type - Type of data looking for.
 *    bufPtr - the bufPtr at this address is set to the buffer found
 *    sizePtr - pointer to an integer which is set to size of chunk found.
 * 
 * Action:
 *   Scans through scatter buffer looking for data type. If it
 *   finds it, returns char * to it. The pointer is set to the
 *   first byte of the chunk (i.e. past the chunk type and length).
 *   This allows the type and length to be in a different buffer than
 *   the actual chunk data which is important in many places.
 *
 *   NOTE: it does not do the ntohl for anything in the buffer
 *         so the data pointed to will still be big endian.
 *         It is important that the id type and length are not
 *         ntohl`d in this buffer space since future GetFromSB's
 *         on this scatter buffer list will fail since it expects
 *         the data type and length to be in big endian form.
 *
 *   Returns null on failure.
 */

char *
GetFromSB(ScatterBuffer *sb, int type, Buffer **bufPtr, int *sizePtr)
{
  int this_type;
  int this_length;
  int cur_buf;
  int idx = 0;

  if (sb->count <= 0) return NULL;

  cur_buf = 0;

  memcpy(&this_type, (sb->buffers[cur_buf]->data + 
		      sb->offsets[cur_buf]+idx), 4);

  this_type = ntohl(this_type);

  
  while (this_type != type) {
    idx += 4;

    memcpy(&this_length, (sb->buffers[cur_buf]->data+
			  sb->offsets[cur_buf]+idx), 4);
    idx += 4;

    this_length = ntohl(this_length); 
    idx += this_length;

    while (idx >= sb->lens[cur_buf]) {
      idx -= sb->lens[cur_buf];
      cur_buf++;
      if (cur_buf >= sb->count) return NULL;
    }

    memcpy(&this_type, (sb->buffers[cur_buf]->data + 
			sb->offsets[cur_buf]+idx), 4);

    this_type = ntohl(this_type);
  }

  idx += 4;
  memcpy(&this_length, (sb->buffers[cur_buf]->data+
			sb->offsets[cur_buf]+idx), 4);
  idx += 4;
  this_length = ntohl(this_length);

  while (idx >= sb->lens[cur_buf]) {
      idx -= sb->lens[cur_buf];
      cur_buf++;
      if (cur_buf >= sb->count) return NULL;
  }

  if (sizePtr) 
    *sizePtr = this_length;
  if (bufPtr) 
	  *bufPtr = sb->buffers[cur_buf];

  return  (sb->buffers[cur_buf]->data + sb->offsets[cur_buf]+idx);
}

