/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif

#include <tcl.h>
#include <sys/types.h>
#ifndef NO_X11
# include <X11/Xlib.h>
#endif
#include "cmBufferlist.h"


/*
 *--------------------------------------------------------------
 *
 * ParseScatterBuffer --
 *
 *	Given a tcl-formatted string specifying a scatter-buffer,
 *	parse it into an ScatterBuffer structure.
 *
 *	The resulting arrays belong to the caller.  When they're no
 *	longer needed, the caller should free them.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	Tcl_Interp *interp	Used for error reporting
 *	char *str;		Tcl-formatted string specifying ScatterBuffer
 *	ScatterBuffer *sb;	Scatter buffer to be filled in
 *
 *--------------------------------------------------------------
 */
static int
ParseScatterBuffer(Tcl_Interp *interp, char *str, ScatterBuffer *sb)
{
  char **spec, **seg;
  int count;
  Buffer *bufPtr;
  int t;
  
  seg = NULL;
  spec = NULL;
  
  /*
   * Parse into list of segments
   */
  if (Tcl_SplitList(interp, str, &(sb->count), &seg) != TCL_OK) {
    if (seg != NULL) {
      ckfree((char *) seg);
    }
    Tcl_AppendResult(interp, " Invalid scatter-buffer \"",
		     str, "\"", NULL);
    return TCL_ERROR;
  }
  
  sb->buffers = (Buffer **) ckalloc(sb->count * sizeof(Buffer *));
  sb->offsets = (int *) ckalloc(sb->count * sizeof(int));
  sb->lens = (int *) ckalloc(sb->count * sizeof(int));
  
  /*
   * For each segment, verify spec and add it to ScatterBuffer
   */
  for (t=0; t<sb->count; t++) {
    if (spec != NULL) {
      ckfree((char *) spec);
    }
    spec = NULL;
    if ((Tcl_SplitList(interp, seg[t], &count, &spec) == TCL_OK) &&
	(count == 3) &&
	(GetBuffer(interp, spec[0], &bufPtr) == TCL_OK) &&
	(Tcl_GetInt(interp, spec[1], &(sb->offsets[t])) == TCL_OK) &&
	(Tcl_GetInt(interp, spec[2], &(sb->lens[t])) == TCL_OK)) {
      
      sb->buffers[t] = bufPtr;
    } else {
      Tcl_AppendResult(interp, "Syntax error in scatter-buffer:\"",
		       seg[t], "\"", NULL);
      if (seg != NULL) {
	ckfree((char *) seg);
      }
      if (spec != NULL) {
	ckfree ((char *) spec);
      }
      ckfree((char *) sb->buffers);
      ckfree((char *) sb->offsets);
      ckfree((char *) sb->lens);
      
      sb->count = 0;
      sb->buffers = NULL;
      sb->offsets = NULL;
      sb->lens = NULL;
      
      return TCL_ERROR;
	}
  }
  if (spec != NULL) {
      ckfree((char *) spec);
  }
  if (seg != NULL) {
    ckfree((char *) seg);
  }
  return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * ParseSBList --
 *
 *	Given a tcl-formatted string specifying a scatter-buffer
 *	list, parse it into an SB_List structure.
 *
 *	The resulting arrays belong to the caller.  When they're no
 *	longer needed, the caller should free them.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	Tcl_Interp *interp	Used for error reporting
 *	char *str;		Tcl-formatted string specifying ScatterBuffer
 *	SB_List *sbl;		Scatter buffer list to be filled in
 *
 *--------------------------------------------------------------
 */
int
ParseSBList(Tcl_Interp *interp, char *str, SB_List *sbl)
{
    char **bufs;
    int t;

    /*
     * Parse into list of buffers
     */
    if (Tcl_SplitList(interp, str, &(sbl->count), &bufs) != TCL_OK) {
	Tcl_AppendResult(interp, " Invalid scatter-buffer list\"",
			str, "\"", NULL);
	return TCL_ERROR;
    }

    sbl->sbs =
	(ScatterBuffer **) ckalloc(sbl->count * sizeof(ScatterBuffer *));

    /*
     * For each segment, call ParseScatterBuffer
     */
    for (t=0; t<sbl->count; t++) {
	sbl->sbs[t] = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));
	if (ParseScatterBuffer(interp, bufs[t], sbl->sbs[t]) != TCL_OK) {
	    if (bufs != NULL) {
		ckfree((char *) bufs);
	    }

	    while (t >= 0) {
		ckfree((char*)(sbl->sbs[t--]));
	    }
	    ckfree((char *) sbl->sbs);
	    
	    sbl->count = 0;
	    sbl->sbs = NULL;

	    return TCL_ERROR;
	}
    } 
    if (bufs != NULL) {
      ckfree((char *) bufs);
    }
    return TCL_OK;
}

