/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*----------------------------------------------------------------
//	chunk.c
// author:  Eric14@cs.berkeley.edu
//
// purpose:  See "chunk.h"
//
----------------------------------------------------------------*/

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include "cmChunk.h"
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <stdio.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>

#ifndef O_BINARY
#    define O_BINARY 0
#endif

#define FILE_HDR_LEN		32
#define FILE_STATE_MAGIC_NUM	17

/*
#define ERR(s)  {fprintf(stderr,s); putc('\n',stderr);}
*/
#define ERR(s)  {}
#define C_	,

#ifndef TRUE
    #define TRUE 1
#endif

#ifndef FALSE
    #define FALSE 0
#endif

#define RND_QUAD(x)	((x) & ~3L)
#define RNDUP_QUAD(x)	(((x)+3) & ~3L)

/*--------------------------------------------------------------*/
/* Writes 0-3 bytes to the file to garenty that it ends on a quad
 * word boundry.  
 * Returns: EOF pos at start of operation if success else < 0.
 */

static int32
L_PadQuadWord (tChunkFileState* cf)
{
    int32	padBuf = 0;
    int		err;
    int32	pos, pad;

    pos = lseek(cf->fd,0,SEEK_END);
    if (pos < 0)
    {
	ERR("CHUNK:PadQuadWord - File end seek failed!");
	return kChunk_Error;
    }
    pad = pos % 4;
    if (pad)
    {
	pad = 4 - pad;
	if ((err = write(cf->fd,&padBuf,pad)) != pad)
	{
	    ERR("CHUNK:finish - Failed writing pad!");
	    return kChunk_Error;
	}
    }
    return pos;
 }

/*--------------------------------------------------------------*/
/* Finishes a started chunk write by calculating the data's size
 * and then modifying the chunk's header to contain the new size.
 * Finally the file is padded to a multipule of 4 and the file pos
 * moved to the EOF.
 * Effects -	writeFlag.
 */

static int 
L_FinishChunk (tChunkFileState* cf)
{
    int		err;
    int32	size, endPos, pos;

    if (!cf->writeFlag) 
    {
	ERR("CHUNK:finish - read only file");
	return kChunk_Error;
    }

    if (cf->writeHeaderStart == -1)
	return 0;

    if ((pos = L_PadQuadWord(cf)) < 0)
	return pos;

    size = pos - cf->writeHeaderStart;
    	
    	/* add size to header */
    pos= lseek(cf->fd,cf->writeHeaderStart+sizeof(int32),SEEK_SET);
    if (pos == -1)
    {
	ERR("CHUNK:finish - seeking size field");
	return kChunk_Error;
    }

    size = htonl(size);
    err = write(cf->fd,&size,sizeof(size));

    endPos = lseek(cf->fd,0,SEEK_END);
    if (err != 4)
    {
	ERR("CHUNK:finish - failed writing size");
	return kChunk_Error;
    }
    cf->writeHeaderStart = -1;
    return 0;
}

/*--------------------------------------------------------------*/
/* Advances the file read pos. to the next chunk's start pos.
 * effects -	curChunkNum (incremented)
 *		curHeaderStart
 */

static int
L_NextHeader (tChunkFileState* cf)
{
    int32	targ, pos;

    if (cf->readHeaderStart == -1 && cf->readIndex != -1)
    {
	ERR("CHUNK:next - past EOF");
	return kChunk_EndOfFile;
    }
    if (cf->readIndex == -1)
	targ = FILE_HDR_LEN;
    else
	targ = RNDUP_QUAD(cf->readHeaderStart + cf->chunkSize);

    pos = lseek(cf->fd,targ,SEEK_SET);
    if (pos != targ)
    {
	ERR("CHUNK: next - Failed: targ = %d, pos = %d"C_ targ C_ pos);
	return kChunk_Error;
    }
    cf->readHeaderStart = targ;
    return 0;
}

/*----{c2doc FUNCTION}------------------------------------------*/
/* Opens the given file and returns a ChunkFileState used by other
 * ChunkFile operations.  The options string may contain any of the
 * following characters in any order:
 *     r or R - file opened for reading (true by default)
 *     w or W - file opened for writing
 *     c or C - file created if not found.
 */

tChunkFileState*  
Chunk_OpenFile (char* name, char* options)
{
    char*		kFileHdrStr = "*cmtChunkFile v1.0";
    int			cnt;
    tChunkFileState	newState;
    tChunkFileState*	out;
    char		hdrStr[FILE_HDR_LEN];
    memset(&newState,0,sizeof(newState));

    for (; *options; options++)
	switch (*options)
	{
	  case 'r': case 'R':
	    break;
	  case 'w': case 'W':
	    newState.writeFlag = TRUE;
	    break;
	  case 'c': case 'C':
	    newState.createFlag = TRUE;
	    break;
	}

    if (newState.createFlag)
    {
	newState.fd = creat(name,0666);
    } else
	newState.fd = open(name,
			   (newState.writeFlag?O_RDWR:O_RDONLY)|O_BINARY,0);

    if (newState.fd == -1)
    {
	ERR("Chunk_OpenFile: Failed open: %s - '%s'"C_ name C_ options);
	return NULL;
    }

    if (newState.createFlag)
    {
	memset(hdrStr,0,FILE_HDR_LEN);
	strcpy(hdrStr,kFileHdrStr);
	cnt = write(newState.fd,hdrStr,FILE_HDR_LEN);
	if (cnt != FILE_HDR_LEN)
	{
	    ERR("Chunk_OpenFile: Failed header write: %s"C_ name);
	    close(newState.fd);
	    return NULL;
	}
    } else
    {
	cnt = read(newState.fd,hdrStr,FILE_HDR_LEN);
	if (cnt != FILE_HDR_LEN || strcmp(hdrStr,kFileHdrStr)!=0)
	{
	    ERR("Chunk_OpenFile: Failed header compare: %s"C_ name);
	    close(newState.fd);
	    return NULL;
	}
    }

    if ((out = (tChunkFileState *) malloc(sizeof(newState))) == NULL)
    {
	ERR("Chunk_OpenFile: State malloc faield!");
	close(newState.fd);
	return NULL;
    }	
    newState.writeHeaderStart = newState.readHeaderStart = -1;
    newState.readIndex = -1;
    *out = newState;
    return out;
}

/*----{c2doc FUNCTION}------------------------------------------*/
/* Closes the given chunk file.  This function MUST be called if
 * chunks have been added to the file.
 */

int
Chunk_CloseFile (tChunkFileState* cf)
{
    int			err;

    if (cf->writeFlag && (err = L_FinishChunk(cf)) != 0)
	return err;
    close(cf->fd);
    free((char *) cf);
    return 0;
}

/*----{c2doc FUNCTION}------------------------------------------*/
/* Appends a new header to the end of the chunk file.  The write
 * head is left positioned after the header, to allow chunk data
 * writes, using "write(cf->fd,buf,bufSz)".
 * Note: When a new header is writen to the chunk file or the file is
 *     closed, the size of the previous chunk can be determined
 *     automatically.  This data is computed and writen into the
 *     previous header.  This allows very free form chunk data
 *     definition, using any number of write calls.
 * Returns: 0 iff success */

int
Chunk_WriteHeader(tChunkFileState* cf, int32 type, char* tagStr)
{
    tChunkHeader	hdr;
    int			err;
    int32		siz;
    
    if ((err = L_FinishChunk(cf)) != 0)
	return err;
    hdr.type = type;
    hdr.chunkSize = 0xFBAD0BAD;
    hdr.stringSize = siz = 0;
    cf->writeHeaderStart = lseek(cf->fd,0,SEEK_END);
    if (cf->writeHeaderStart == -1)
    {
	ERR("Chunk_WriteHeader: seek file end failed!");
	return kChunk_Error;
    }
    if (cf->writeHeaderStart%4 != 0)
    {
	ERR("Chunk_WriteHeader: EOF is not 32bit aligned");
	return kChunk_Error;
    }
    if (tagStr != NULL)
	hdr.stringSize = siz = strlen(tagStr);

    hdr.chunkSize  = htonl(hdr.chunkSize);  /* host to net */
    hdr.stringSize = htonl(hdr.stringSize);
    hdr.type       = htonl(hdr.type);

    err = write(cf->fd,&hdr,sizeof(hdr));
    if (err != sizeof(hdr))
    {
	ERR("Chunk_WriteHeader: Failed on Header Write");
	return kChunk_Error;
    }
    if (siz != 0 && write(cf->fd,tagStr,siz) != siz)
    {
	ERR("Chunk_WriteHeader: String write");
	return kChunk_Error;
    }
    if ((err = L_PadQuadWord(cf)) < 0)
	return err;
    return 0;
}

/*--------------------------------------------------------------*/
/* Does an lseek to the right place in the file.
 * targ = 1  -->  string
 *	  0  -->  data
 */

static int
L_SeekForRead(tChunkFileState* cf, int targ)
{
    long	jump;

    if (cf->readHeaderStart == -1)
    {
	ERR("CHUNK: readSeek - Past EOF");
	return kChunk_EndOfFile;
    }

    jump = cf->readHeaderStart + sizeof(tChunkHeader);
    if (targ != 1)
	jump += RNDUP_QUAD(cf->stringSize);
    if (lseek(cf->fd,jump,SEEK_SET) != jump)
    {
	ERR("CHUNK:seekRead");
	return kChunk_Error;
    }
    return 0;
}

/*----{c2doc FUNCTION}------------------------------------------*/
/* Advances to the next chunk in the file.  Hdr is set to header 
 * info for the chunk.  This function leaves the read head at the
 * beginning of the chunk data.  To read this data call:
 *	read(cf->fd,&buffer,hdr->chunkSize)
 * Returns:	0 iff success
 *         	kChunk_EndOfFile if EOF reached.
 */

int
Chunk_ReadHeader (tChunkFileState* cf, tChunkHeader* hdr)
{
    int			err;

    if ((err = L_NextHeader(cf)) != 0)
	return err;

    if ((err = read(cf->fd,hdr,sizeof(*hdr))) != sizeof(*hdr))
    {
	if (err == 0 && 
	    lseek(cf->fd,0,SEEK_END) == cf->readHeaderStart)
	{
	    cf->readHeaderStart = -1;
	    return kChunk_EndOfFile;
	}
	ERR("Chunk_ReadHeader: header read failed");
	return kChunk_Error;
    }
    hdr->chunkSize  = ntohl(hdr->chunkSize);  /* net to host */
    hdr->stringSize = ntohl(hdr->stringSize);
    hdr->type       = ntohl(hdr->type);

    if (hdr->chunkSize < sizeof(*hdr)) {
	ERR("Chunk_ReadHeader: Bad Header size: %x"C_ hdr->chunkSize);
	return kChunk_Error;
    }

    cf->chunkSize = hdr->chunkSize;
    cf->stringSize = hdr->stringSize;
    if ((err= L_SeekForRead(cf,0)) != 0)
	return err;
    hdr->chunkSize -= sizeof(tChunkHeader) + RNDUP_QUAD(hdr->stringSize);
    ++cf->readIndex;
    return 0;
}

/*----{c2doc FUNCTION}------------------------------------------*/
/* Reads the current chunk's string.  Assumes that you have already
 * called a header reading function.  outBuf is filled with the
 * string.  It is assumed to be large enough!! It should be
 * (header->stringSize +1) bytes or larger!  If outBuf == NULL, then a
 * new buffer is malloc()ed for the string.
 * Returns:	NULL	if error
 * 			theString
 */

char*
Chunk_ReadString (tChunkFileState* cf, char* outBuf)
{
    int		err;
    int32	siz = cf->stringSize;

    if ((err = L_SeekForRead(cf,1)) != 0)
	return NULL;

    if (outBuf == NULL && (outBuf = malloc(siz+1))==NULL)
    {
	ERR("Chunk_ReadString: malloc failed!");
	return NULL;
    }
    if (read(cf->fd,outBuf,siz) != siz)
    {
	ERR("Chunk_ReadString: failed");
	return NULL;
    }
    outBuf[siz] = '\0';
    if ((err = L_SeekForRead(cf,0)) != 0)
	return NULL;
    return outBuf;
}

/*--------------------------------------------------------------*/
/* Does the obvious.  outStr must by at least 5 bytes.
 * Returns:	outStr
 */

void
Chunk_Type2Str (char* outStr, int32 type)
{
    outStr[0] = (type>>24) & 0xff;
    outStr[1] = (type>>16) & 0xff;
    outStr[2] = (type>>8) & 0xff;
    outStr[3] = type & 0xff;
    outStr[4] = '\0';
}

/*--------------------------------------------------------------*/
/* Does the obvious.
 */

int32
Chunk_Str2Type (char* str)
{
    char	s[4];
    int		i, fil = FALSE;
    
    memcpy(s,str,4);
    for (i= 0; i<4; i++)
	if (s[i]=='\0' || fil == TRUE)
	    s[i] = ' ', fil = TRUE;
    return C_TYPE(s[0],s[1],s[2],s[3]);
}

/*----{c2doc FUNCTION}------------------------------------------*/
/* Causes the next read header operation to return the first
 * header in the file.  Also effects Chunk_FindType().
 * Returns:	0	iff success
 */

int
Chunk_ReadRewind (tChunkFileState* cf)
{
    cf->readIndex = -1;
    cf->readHeaderStart = -1;
    return 0;
}

/*----{c2doc FUNCTION}------------------------------------------*/
/* Finds the idx's chunk in the file, numbered from 0.  The header
 * is stored in hdr.  Leaves read head at chunk data.  See
 * Chunk_ReadHeader().
 * Returns: 	0 iff success
 *          	kChunk_ErrIndexRange  	if idx is too small
 *		kChunk_EndOfFile	if idx too large
 */

int
Chunk_FindIndex (tChunkFileState* cf, int idx, tChunkHeader* hdr)
{
    int		err;

    if (idx < 0)
	return kChunk_ErrIndexRange;
    if (idx < cf->readIndex || cf->readIndex < 0)
	Chunk_ReadRewind(cf);

    while (cf->readIndex < idx)
    {
	if ((err= Chunk_ReadHeader(cf,hdr)) != 0)
	    return err;
    }
    return 0;
}

/*----{c2doc FUNCTION}------------------------------------------*/
/* Finds the 1st chunk of the given type in the file, starting
 * from the current start position.  Leaves read head at chunk 
 * data.  See Chunk_ReadHeader() and Chunk_FindIndex()
 * Returns: 	0 iff success
 *		kChunk_EndOfFile	if idx too large
 */

int
Chunk_FindType (tChunkFileState* cf, int32 type, tChunkHeader* hdr)
{
    int		err;

    while (1) {
	if ((err= Chunk_ReadHeader(cf,hdr)) != 0)
	    return err;
	if (hdr->type == type)
	    return 0;
    }
}

/*--------------------------------------------------------------*/
/* End Of File */
