/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * cmd.c --
 *
 *	This module implements routines to make Tcl callback
 *	programming easy.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#   include "compat/stdlib.h"
#else
#   include <stdlib.h>
#endif

#include "tcl.h"
#include "tk.h"

#include "cmUtil.h"

/*
 * For every command created, a structure of the following type
 * is created.
 */

typedef struct Cmd 
{
    Tcl_Interp *interp;
    int  free;		/* Automatically free after executing?	*/
    char *cmd;		/* Tcl callback. 			*/

} Cmd;

/*
 *------------------------------------------------------------------
 *
 * Tcm_MakeCmd --
 *
 *	Create a cmd struct and return a pointer to it.  
 *	The callback is copied into private memory.
 *
 * Results:
 *	Pointer to a cmd struct.
 *
 * Side effects:
 *	Memory is allocated.
 *
 *------------------------------------------------------------------
 */

ClientData
Tcm_MakeCmd(Tcl_Interp *interp, char *callback, int free)
{
    Cmd *rv;

    rv = (Cmd *) ckalloc(sizeof(Cmd));
    rv->interp = interp;
    rv->free   = free;
    rv->cmd    = (char *) ckalloc(strlen(callback) + 1);
    strcpy (rv->cmd, callback);

    /*  Preserve the interpreter, because we store a pointer to it.
     */
    Tk_Preserve ((ClientData) interp);

    return ((ClientData) rv);
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_FreeCmd --
 *
 *	Free up memory allocaed by Tcm_MakeCmd.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

void
Tcm_FreeCmd(ClientData clientData)
{
    Cmd *cmdPtr = (Cmd *) clientData;

    if (!cmdPtr)
      return;

    /*  Release the interpreter.
     */
    if (cmdPtr->interp)
      Tk_Release ((ClientData) cmdPtr->interp);

    if (cmdPtr->cmd)
      Tk_EventuallyFree ((ClientData) cmdPtr->cmd, (Tk_FreeProc *) free);

    cmdPtr->interp = (Tcl_Interp *) NULL;
    cmdPtr->free   = 0;
    cmdPtr->cmd    = NULL;

    Tk_EventuallyFree ((ClientData) cmdPtr, (Tk_FreeProc *) free);
}

/*
 *------------------------------------------------------------------
 *
 * Tcm_ExecCmd --
 *
 *    Evaluate the command passed in.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Calls Tcl code, so just about anything, including freeing
 *	the interpreter.
 *
 *------------------------------------------------------------------
 */

void
Tcm_ExecCmd(ClientData clientData)
{
    Cmd *cmdPtr = (Cmd *) clientData;

    if (cmdPtr)
      {
	Tcl_Interp *interp = cmdPtr->interp;

	Tk_Preserve ((ClientData) cmdPtr);

	if (interp && cmdPtr->cmd)
	  {
	    Tk_Preserve ((ClientData) interp);
	    Tcl_GlobalEval (interp, cmdPtr->cmd);
	    Tk_Release  ((ClientData) interp);
	  }

	if (cmdPtr->free) {
	  Tcm_FreeCmd (clientData);
	}

	Tk_Release ((ClientData) cmdPtr);
      }
}

/*
 *------------------------------------------------------------------
 *
 * Tcm_EvalCmd --
 *
 *    Evaluate the command passed in.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Calls Tcl code, so just about anything.
 *
 *------------------------------------------------------------------
 */

int
Tcm_EvalCmd(ClientData clientData, char *extra, char **interpResult)
{
  Cmd  *cmdPtr = (Cmd *) clientData;
  char *callback;
  int   result;

  result = TCL_OK;
  *interpResult = (char *) NULL;

  if (cmdPtr)
    {
      Tcl_Interp *interp = cmdPtr->interp;
      
      if (interp && cmdPtr->cmd)
	{
	  if (extra)
	    {
	      callback = (char *) ckalloc (strlen (cmdPtr->cmd) + strlen (extra) + 5);
	      sprintf (callback, "%s%s", cmdPtr->cmd, extra);
	    }
	  else
	    callback = cmdPtr->cmd;

	  Tk_Preserve ((ClientData) cmdPtr);
	  Tk_Preserve ((ClientData) interp);

	  result = Tcl_GlobalEval (interp, callback);

	  *interpResult = (char *) ckalloc (strlen (interp->result) + 1);
	  strcpy (*interpResult, interp->result);

	  Tk_Release ((ClientData) interp);
	  Tk_Release ((ClientData) cmdPtr);

	  if (extra)
	    ckfree (callback);
	}

      if (cmdPtr->free) {
	Tcm_FreeCmd (clientData);
	}
    }

  return result;
}
