/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * config.c --
 *
 *	This file contains the
 *      Tcm_ConfigureObject and Tcm_ConfigureInfo procedures.
 *      It was hacked from tkConfig.c to not depend on tkwin passing
 *      and thus will work for any records (structs/objects),
 *      not just widgets.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include "tk.h"

#include "cmConfig.h"

/*
 * Values for "flags" field of Tcm_ConfigSpec structures.  Be sure
 * to coordinate these values with those defined in tcmConfig.h.
 * There must not be overlap!
 *
 * INIT -		Non-zero means (char *) things have been
 *			converted to Tk_Uid's.
 */

#define INIT		0x20

/*
 * Forward declarations for procedures defined later in this file:
 */

static int DoConfig			_ANSI_ARGS_((Tcl_Interp *interp,
						     Tcm_ConfigSpec *specPtr,
						     Tk_Uid value,
						     int valueIsUid,
						     char *widgRec));
static Tcm_ConfigSpec *FindConfigSpec	_ANSI_ARGS_((Tcl_Interp *interp,
						     Tcm_ConfigSpec *specs,
						     char *argvName,
						     int needFlags,
						     int hateFlags));
static char *FormatConfigInfo		_ANSI_ARGS_((Tcm_ConfigSpec * specPtr,
						     char *widgRec));

/*
 *--------------------------------------------------------------
 *
 * Tcm_ConfigureObject --
 *
 *	Process command-line options (but not database options) to
 *	fill in fields of a record with resources and
 *	other parameters.
 *
 * Results:
 *	A standard Tcl return value.  In case of an error,
 *	interp->result will hold an error message.
 *
 * Side effects:
 *	The fields of objRec get filled in with information
 *	from argc/argv (but not the option database).
 *      Old information in objRec's fields gets recycled.
 *
 *--------------------------------------------------------------
 */

int
Tcm_ConfigureObject(Tcl_Interp *interp, Tcm_ConfigSpec *specs, int argc, char **argv, char *objRec, int flags)
                       		/* Interpreter for error reporting. */
                          	/* Describes legal options. */
             			/* Number of elements in argv. */
                		/* Command-line options. */
                 		/* Record whose fields are to be modified.
				 * Values must be properly initialized. */
              			/* Used to specify additional flags that must
				 * be present in config specs for them to be
				 * considered.  Also, may have
				 * TCM_CONFIG_ARGV_ONLY set. */
{
    register Tcm_ConfigSpec *specPtr;
    Tk_Uid value;		/* Value of option from database. */
    int needFlags;		/* Specs must contain this set of flags or
				 * else they are not considered. */
    int hateFlags;		/* If a spec contains any bits here, it's not
				 * considered. */

    needFlags = flags & ~(TCM_CONFIG_USER_BIT - 1);

    hateFlags = 0;

    /*
     * Pass one:  scan through all the option specs, replacing strings with
     * Tk_Uids (if this hasn't been done already) and clearing the
     * TCM_CONFIG_OPTION_SPECIFIED flags.
     */

    for (specPtr = specs; specPtr->type != TCM_CONFIG_END; specPtr++) {
	if (!(specPtr->specFlags & INIT) && (specPtr->argvName != NULL)) {
	    if (specPtr->defValue != NULL) {
		specPtr->defValue = Tk_GetUid(specPtr->defValue);
	    }
	}
	specPtr->specFlags = (specPtr->specFlags & ~TCM_CONFIG_OPTION_SPECIFIED)
	    | INIT;
    }

    /*
     * Pass two:  scan through all of the arguments, processing those that
     * match entries in the specs.
     */

    for (; argc > 0; argc -= 2, argv += 2) {
	specPtr = FindConfigSpec(interp, specs, *argv, needFlags, hateFlags);
	if (specPtr == NULL) {
	    return TCL_ERROR;
	}
	/*
	 * Process the entry.
	 */

	if (argc < 2) {
	    Tcl_AppendResult(interp, "value for \"", *argv,
			     "\" missing", (char *) NULL);
	    return TCL_ERROR;
	}
	if (DoConfig(interp, specPtr, argv[1], 0, objRec) != TCL_OK) {
	    char msg[100];

	    sprintf(msg, "\n    (processing \"%.40s\" option)",
		    specPtr->argvName);
	    Tcl_AddErrorInfo(interp, msg);
	    return TCL_ERROR;
	}
	specPtr->specFlags |= TCM_CONFIG_OPTION_SPECIFIED;
    }

    /*
     * Pass three:  scan through all of the specs again;  if no command-line
     * argument matched a spec, then check for info in the option database.
     * If there was nothing in the database, then use the default.
     */

    if (!(flags & TCM_CONFIG_ARGV_ONLY)) {
	for (specPtr = specs; specPtr->type != TCM_CONFIG_END; specPtr++) {
	    if ((specPtr->specFlags & TCM_CONFIG_OPTION_SPECIFIED)
		|| (specPtr->argvName == NULL)
		|| (specPtr->type == TCM_CONFIG_SYNONYM)) {
		continue;
	    }
	    if (((specPtr->specFlags & needFlags) != needFlags)
		|| (specPtr->specFlags & hateFlags)) {
		continue;
	    }
	    value = specPtr->defValue;
	    if ((value != NULL) && !(specPtr->specFlags
				     & TCM_CONFIG_DONT_SET_DEFAULT)) {
		if (DoConfig(interp, specPtr, value, 1, objRec) !=
		    TCL_OK) {
		    char msg[200];

		    sprintf(msg,
			    "\n    (default value in object)");

		    Tcl_AddErrorInfo(interp, msg);
		    return TCL_ERROR;
		}
	    }
	}
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * FindConfigSpec --
 *
 *	Search through a table of configuration specs, looking for
 *	one that matches a given argvName.
 *
 * Results:
 *	The return value is a pointer to the matching entry, or NULL
 *	if nothing matched.  In that case an error message is left
 *	in interp->result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static Tcm_ConfigSpec *
FindConfigSpec(Tcl_Interp *interp, Tcm_ConfigSpec *specs, char *argvName, int needFlags, int hateFlags)
                       		/* Used for reporting errors. */
                          	/* Pointer to table of configuration
				 * specifications for an object. */
                   		/* Name (suitable for use in a "config"
				 * command) identifying particular option. */
                  		/* Flags that must be present in matching
				 * entry. */
                  		/* Flags that must NOT be present in matching
				 * entry. */
{
    register Tcm_ConfigSpec *specPtr;
    register char c;		/* First character of current argument. */
    Tcm_ConfigSpec *matchPtr;	/* Matching spec, or NULL. */
    int length;

    c = argvName[1];
    length = strlen(argvName);
    matchPtr = NULL;
    for (specPtr = specs; specPtr->type != TCM_CONFIG_END; specPtr++) {
	if (specPtr->argvName == NULL) {
	    continue;
	}
	if ((specPtr->argvName[1] != c)
	    || (strncmp(specPtr->argvName, argvName, length) != 0)) {
	    continue;
	}
	if (((specPtr->specFlags & needFlags) != needFlags)
	    || (specPtr->specFlags & hateFlags)) {
	    continue;
	}
	if (specPtr->argvName[length] == 0) {
	    matchPtr = specPtr;
	    goto gotMatch;
	}
	if (matchPtr != NULL) {
	    Tcl_AppendResult(interp, "ambiguous option \"", argvName,
			     "\"", (char *) NULL);
	    return (Tcm_ConfigSpec *) NULL;
	}
	matchPtr = specPtr;
    }

    if (matchPtr == NULL) {
	Tcl_AppendResult(interp, "unknown option \"", argvName,
			 "\"", (char *) NULL);
	return (Tcm_ConfigSpec *) NULL;
    }
    /*
     * Found a matching entry.  If it's a synonym, then find the entry that
     * it's a synonym for.
     */

gotMatch:
    specPtr = matchPtr;
    if (specPtr->type == TCM_CONFIG_SYNONYM) {
	for (specPtr = specs;; specPtr++) {
	    if (specPtr->type == TCM_CONFIG_END) {
		Tcl_AppendResult(interp,
				 "couldn't find synonym for option \"",
				 argvName, "\"", (char *) NULL);
		return (Tcm_ConfigSpec *) NULL;
	    }
	    if ((specPtr->type != TCM_CONFIG_SYNONYM)
		&& ((specPtr->specFlags & needFlags) == needFlags)
		&& !(specPtr->specFlags & hateFlags)) {
		break;
	    }
	}
    }
    return specPtr;
}

/*
 *--------------------------------------------------------------
 *
 * DoConfig --
 *
 *	This procedure applies a single configuration option
 *	to a record.
 *
 * Results:
 *	A standard Tcl return value.
 *
 * Side effects:
 *	objRec is modified as indicated by specPtr and value.
 *	The old value is recycled, if that is appropriate for
 *	the value type.
 *
 *--------------------------------------------------------------
 */

 /* ARGSUSED */
static int
DoConfig(Tcl_Interp *interp, Tcm_ConfigSpec *specPtr, char *value, int valueIsUid, char *objRec)
                       		/* Interpreter for error reporting. */
                            	/* Specifier to apply. */
                		/* Value to use to fill in objRec. */
                   		/* Non-zero means value is a Tk_Uid; zero
				 * means it's an ordinary string. */
                 		/* Record whose fields are to be modified.
				 * Values must be properly initialized. */
{
    char *ptr;
    Tk_Uid uid;
    int nullValue;

    nullValue = 0;
    if ((*value == 0) && (specPtr->specFlags & TCM_CONFIG_NULL_OK)) {
	nullValue = 1;
    }
    do {
	ptr = objRec + specPtr->offset;
	switch (specPtr->type) {
	case TCM_CONFIG_BOOLEAN:
	    if (Tcl_GetBoolean(interp, value, (int *) ptr) != TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	case TCM_CONFIG_INT:
	    if (Tcl_GetInt(interp, value, (int *) ptr) != TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	case TCM_CONFIG_DOUBLE:
	    if (Tcl_GetDouble(interp, value, (double *) ptr) != TCL_OK) {
		return TCL_ERROR;
	    }
	    break;
	case TCM_CONFIG_STRING:{
		char *old, *new;

		if (nullValue) {
		    new = NULL;
		} else {
		    new = (char *) ckalloc((unsigned) (strlen(value) + 1));
		    strcpy(new, value);
		}
		old = *((char **) ptr);
		if (old != NULL) {
		    ckfree(old);
		}
		*((char **) ptr) = new;
		break;
	    }
	case TCM_CONFIG_UID:
	    if (nullValue) {
		*((Tk_Uid *) ptr) = NULL;
	    } else {
		uid = valueIsUid ? (Tk_Uid) value : Tk_GetUid(value);
		*((Tk_Uid *) ptr) = uid;
	    }
	    break;
	default:{
		sprintf(interp->result, "bad config table: unknown type %d",
			specPtr->type);
		return TCL_ERROR;
	    }
	}
	specPtr++;
    } while ((specPtr->argvName == NULL) && (specPtr->type != TCM_CONFIG_END));
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_ConfigureInfo --
 *
 *	Return information about the configuration options
 *	for an object, and their current values.
 *
 * Results:
 *	Always returns TCL_OK.  Interp->result will be modified
 *	hold a description of either a single configuration option
 *	available for "objRec" via "specs", or all the configuration
 *	options available.  In the "all" case, the result will
 *	available for "objRec" via "specs".  The result will
 *	be a list, each of whose entries describes one option.
 *	Each entry will itself be a list containing the option's
 *	name for use on command lines, database name, database
 *	class, default value, and current value (empty string
 *	if none).  For options that are synonyms, the list will
 *	contain only two values:  name and synonym name.  If the
 *	"name" argument is non-NULL, then the only information
 *	returned is that for the named argument (i.e. the corresponding
 *	entry in the overall list is returned).
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

int
Tcm_ConfigureInfo(Tcl_Interp *interp, Tcm_ConfigSpec *specs, char *objRec, char *argvName, int flags)
                       		/* Interpreter for error reporting. */
                          	/* Describes legal options. */
                 		/* Record whose fields contain current values
				 * for options. */
                   		/* If non-NULL, indicates a single option
				 * whose info is to be returned.  Otherwise
				 * info is returned for all options. */
              			/* Used to specify additional flags that must
				 * be present in config specs for them to be
				 * considered. */
{
    register Tcm_ConfigSpec *specPtr;
    int needFlags, hateFlags;
    char *list;
    char *leader = "{";

    needFlags = flags & ~(TCM_CONFIG_USER_BIT - 1);

    hateFlags = 0;

    /*
     * If information is only wanted for a single configuration spec, then
     * handle that one spec specially.
     */

    Tcl_SetResult(interp, (char *) NULL, TCL_STATIC);
    if (argvName != NULL) {
	specPtr = FindConfigSpec(interp, specs, argvName, needFlags,
				 hateFlags);
	if (specPtr == NULL) {
	    return TCL_ERROR;
	}
	interp->result = FormatConfigInfo(specPtr, objRec);
	interp->freeProc = (Tcl_FreeProc *) free;
	return TCL_OK;
    }
    /*
     * Loop through all the specs, creating a big list with all their
     * information.
     */

    for (specPtr = specs; specPtr->type != TCM_CONFIG_END; specPtr++) {
	if ((argvName != NULL) && (specPtr->argvName != argvName)) {
	    continue;
	}
	if (((specPtr->specFlags & needFlags) != needFlags)
	    || (specPtr->specFlags & hateFlags)) {
	    continue;
	}
	if (specPtr->argvName == NULL) {
	    continue;
	}
	list = FormatConfigInfo(specPtr, objRec);
	Tcl_AppendResult(interp, leader, list, "}", (char *) NULL);
	ckfree(list);
	leader = " {";
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * FormatConfigInfo --
 *
 *	Create a valid Tcl list holding the configuration information
 *	for a single configuration option.
 *
 * Results:
 *	A Tcl list, dynamically allocated.  The caller is expected to
 *	arrange for this list to be freed eventually.
 *
 * Side effects:
 *	Memory is allocated.
 *
 *--------------------------------------------------------------
 */

 /* ARGSUSED */
static char *
FormatConfigInfo(register Tcm_ConfigSpec *specPtr, char *objRec)
                                     	/* Pointer to information describing
					 * option. */
                 		/* Pointer to record holding current values
				 * of info for object. */
{
    char *argv[4], *ptr, *result;
    char buffer[200];
    Tcl_FreeProc *freeProc = (Tcl_FreeProc *) NULL;

    argv[0] = specPtr->argvName;
    argv[1] = specPtr->defValue;
    if (specPtr->type == TCM_CONFIG_SYNONYM) {
	return Tcl_Merge(2, argv);
    }
    ptr = objRec + specPtr->offset;
    argv[2] = "";
    switch (specPtr->type) {
    case TCM_CONFIG_BOOLEAN:
	if (*((int *) ptr) == 0) {
	    argv[2] = "false";
	} else {
	    argv[2] = "true";
	}
	break;
    case TCM_CONFIG_INT:
	sprintf(buffer, "%d", *((int *) ptr));
	argv[2] = buffer;
	break;
    case TCM_CONFIG_DOUBLE:
	sprintf(buffer, "%lf", *((double *) ptr));
	argv[2] = buffer;
	break;
    case TCM_CONFIG_STRING:
	argv[2] = (*(char **) ptr);
	break;
    case TCM_CONFIG_UID:{
	    Tk_Uid uid = *((Tk_Uid *) ptr);
	    if (uid != NULL) {
		argv[2] = uid;
	    }
	    break;
	}
    default:
	argv[2] = "?? unknown type ??";
    }
    if (argv[1] == NULL) {
	argv[1] = "";
    }
    if (argv[2] == NULL) {
	argv[2] = "";
    }
    result = Tcl_Merge(3, argv);
    if (freeProc != NULL) {
	if (freeProc == (Tcl_FreeProc *) free) {
	    ckfree(argv[2]);
	} else {
	    (*freeProc) (argv[2]);
	}
    }
    return result;
}
