/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/* This file contains the magic cookies (ID's) used
   by the clipfiles (i.e. chunk types) and the scatter
   buffer headers passed from object to object. 
   
   All the cookies are 4 bytes long and have the following
   form:
   
   0x434d????
   
   The 434d corresponds to "CM" in ASCII.
   
   This file also contains the various type codes and
   header structures used in all the various header types.
   */

#ifndef _cookies_h_
#define _cookies_h_

#define CMT_VERSION 300

/* 
 * Continuous Media Text Chunk Type 
 * 
 * This chunk type is used to include a human readable string
 * in a chunk file. The format of the string and/or information
 * in this segment is completely up to the utility/application/program
 * that produced it.
 */

#define CM_TEXT_SEGMENT 0xdabadaba

/* Clipfile media independent header. */

#define CM_MEDIA_INDEP_CF 0x434d0000

typedef struct {
    int media;
    int format;
    int device;
    int version;
    int num_frames;
    int length_s;
    int length_ns;
} MediaIndepHeader;

#define VIDEO_MEDIA_CODE 0x00000001
#define AUDIO_MEDIA_CODE 0x00000002

#define MJPEG_FORMAT_CODE 0x00000001
#define MPEG_FORMAT_CODE 0x00000002
#define SPARC_AU_FORMAT_CODE 0x00000004
#define NV_FORMAT_CODE 0x00000008
#define YUV_FORMAT_CODE 0x00000016

#define PARALLAX_DEV_CODE 0x00000001
#define J300_DEV_CODE 0x00000002
#define SPARC_AU_DEV_CODE 0x00000004
#define OTHER_DEV_CODE 0x00000000

/* Clipfile frame offset table. */

#define CM_FOT 0x434d0001

/* Clipfile frame time table. */

#define CM_FTT 0x434d0002

/* CM data segment. */

#define CM_DATA_SEGMENT 0x434d0003

/* Clipfile MPEG header. */
#define CM_MPEG_HDR_TABLE 0x434d0032
#define CM_MPEG_REF_TABLE 0x434d0033
#define CM_MPEG_CF 0x434d0030
#define CM_MPEG_BUF_HDR 0x434d0031
#define CM_MPEG_INTRA_MATRIX 0x434d0034
#define CM_MPEG_NON_INTRA_MATRIX 0x434d0035

typedef struct {
    int width;
    int height;
    int aspectRatioCode;
    int frameRateCode;
    int bitRate;
    int vbvBufferSize;
    int constrainedParamFlag;
} MPEGHeader;

typedef struct {
    int hdrType;
    int length;
    int width;
    int height;
    int type;
    int ref1;
    int ref2;
    int intra_matrix_flag;
    int non_intra_matrix_flag;
    int index;
    int logStart_s;
    int logStart_ns;
    int logEnd_s;
    int logEnd_ns;
    int displayFlag;
} MPEGBufferHeader;


/* Clipfile NV header. */

#define CM_NV_CF 0x434d0040
#define CM_NV_BUF_HDR 0x434d0041

typedef struct {
    int width;
    int height;
    int device;
    int subformat;
} NVHeader;

typedef struct {
    int hdrType;                /* Always equal to CM_NV_BUF_HDR */
    int length;                 /* Always equal to size of this structure. */
    int width;
    int height;
    int device;
    int subformat;
    int composit;		/* composit frame for random access */
    int logStart_s;
    int logStart_ns;
    int logEnd_s;
    int logEnd_ns;
} NVBufferHeader;

typedef struct {
    int iExist;
    int fdOffset;
    int fdLength;
    int fdPacketNum;
    int bdOffset;
    int bdLength;
    int bdPacketNum;
    int iOffset;
    int iLength;
    int iPacketNum;
    int direction;
    int iFrameNum;
} NVFrameOffset;

/* Clipfile MJPEG header. */

#define CM_MJPEG_CF 0x434d0010
#define CM_MJPEG_BUF_HDR 0x434d0011

typedef struct {
    int width;
    int height;
    int device;
    int subformat;
    int qfactor;
} MJPEGHeader;

typedef struct {
    int hdrType;                /* Always equal to CM_MJPEG_BUF_HDR */
    int length;                 /* Always equal to size of this structure - 8. */
    int width;
    int height;
    int device;
    int subformat;
    int qfactor;
    int logStart_s;
    int logStart_ns;
    int logEnd_s;
    int logEnd_ns;
} MJPEGBufferHeader;

#define VID_FRAME_CODE 0x00000001
#define JFIF_FRAME_CODE 0x00000002
#define JFIF_NO_HUFFMAN_FRAME_CODE 0x00000004
#define RAW_JPEG_NO_HEADERS_CODE 0x00000008

/* VQ Definitions (Vector Quant.  For info pad (allanl) */

#define CM_VQ_CF 0x434d0042
#define CM_VQ_BUF_HDR 0x434d0043
#define CM_VQ_CODEBOOKS 0x434d0044
#define CM_VQ_FDT		0x434d0045

#define VQ_FORMAT_CODE 0x00000008
#define INFOPAD_DEV_CODE 0x00000008

typedef struct {
    int width;			/* in pixels */
    int height;			/* in pixels */
    int numCodebooks;
    /* I don't know what use these will be for VQ, but for now I'll leave them in */
    int device;
    int subformat;
} VQHeader;

/* these are used in the subtype field in VQBufferHeader */

#define CM_VQ_FRAME_DATA	0
#define CM_VQ_CODEBOOK_DATA	1

typedef struct {
    int hdrType;		/* Always equal to CM_VQ_BUF_HDR in the file */
    int length;			/* excluding type & length */
    int width;
    int height;
    int codebookIndex;		/* which codebook this frame depends on
				 * (or which it is, if it is one) */
    int device;
    int subformat;		/* CM_VQ_FRAME_DATA or CM_VQ_CODEBOOK_DATA */
    int logStart_s;
    int logStart_ns;
    int logEnd_s;
    int logEnd_ns;
} VQBufferHeader;

/* Clipfile Sparc Audio header. */

#define CM_SPARC_AU_CF 0x434d0020
#define CM_SPARC_AU_BUF_HDR 0x434d0021

typedef struct {
    int dataFormat;		/* the data format code */
    int samplingRate;		/* the sampling rate */
    int channelCount;		/* the number of channels */
} SparcAuHeader;

typedef struct {
    int hdrType;		/* Always equal to CM_SPARC_AU_BUF_HDR */
    int length;                 
    int device;
    int dataFormat;        
    int samplingRate;      
    int channelCount;      
    int logStart_s;
    int logStart_ns;
    int logEnd_s;
    int logEnd_ns;
} SparcAuBufferHeader;


/* Segmented Sparc Audio headers */
#define CM_SPARC_AU_PKT_HDR        0x434d0030

/* note: initial fields in same order as SparcAuBufferHeader */
typedef struct {
    int hdrType;		/* Always equal to CM_SPARC_AU_PKT_HDR */
    int length;
    int device;
    int dataFormat;        
    int samplingRate;      
    int channelCount;      
    int logStart_s;
    int logStart_ns;
    int logEnd_s;
    int logEnd_ns;
  
    int segID;			/* segment identifier */
    int pktFactor;		/* packetization factor */
    int pktID;			/* which packet in segment this is */
    int totDataLen;		/* total amount of data in segment (bytes) */
} SparcAuPktHeader;

/* raw RTP packets (see cmRTP.h) */
#define CM_RTP_PACKET		0x434d0050

/* Clipfile YUV header. */

#define CM_YUV_CF 0x434d0060
#define CM_YUV_BUF_HDR 0x434d0061

#define FOUR_TWO_ZERO 0
#define FOUR_TWO_TWO 1

typedef struct {
  int width;
  int height;
  int format;
} YUVHeader;

typedef struct {
  int hdrType;                /* Always equal to CM_YUV_BUF_HDR */
  int length;                 /* Always equal to size of this structure - 8. */
  int device;
  int width;
  int height;
  int format;
  int logStart_s;
  int logStart_ns;
  int logEnd_s;
  int logEnd_ns;
} YUVBufferHeader;

/* MPEG System Stream Packs  */
#define CM_MPEG_SYS_PACK	0x434d0070

typedef struct {
  int hdrType;          	/* Always equal to CM_MPEG_SYS_PACK */
  int length;			/* header length */
  int packSeq;			/* sequential pack number */
  int clipStart_s;		/* current clip start */
  int clipStart_ns;
  int clipEnd_s;		/* current clip end */
  int clipEnd_ns;
  int logStart_s;		/* current logical start */
  int logStart_ns;
  int logEnd_s;		/* current logical end */
  int logEnd_ns;
} MPEGSysPackHeader;

/*
 * H.261 stuff
 */
#define CM_H261_BUF_HDR 0x434d0080

typedef struct {
/*    int type;                 /* Always CM_H261_BUF_HDR */
/*    int len;  */

    int logStart_s;
    int logStart_ns;
    int logEnd_s;
    int logEnd_ns;

    int options;                /* Options as in draft-ietf-avt-h261 */
} H261BufferHeader;

#endif
