/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmCosmoDevice.c,v 1.8 1996/02/06 22:40:43 kpatel Exp $
 *
 * Device to work with the CM MjpegPlay object to decode
 * a motion JPEG stream with the Cosmo card.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <tcl.h>
#include <tk.h>
#include <dp.h>
#include "cmMJPEGPlay.h"
#include "cmBuffer.h"

#include "cmUtil.h"
#include "cmBufferlist.h"
#include "cmLTS.h"

#ifdef COSMO

#include "cmCosmoDevice.h"

/*
 * Global definitions
 */
static _Codec codec;
static _Video video;
static _Image image;
static short decompressorOpen = 0;
static short videoPathExists = 0;

/*
 * Dimensions for video window
 */

static int winWidth;
static int winHeight;

/*
 * There is one of the following structs for every
 * unique device created
 */
typedef struct Cosmo_Device {
    int indecode;     	/* Is a decode currently in progress? */

    /*
     * Stuff needed for display
     */
    Window win;
    Display *dpy;
    GC gc;

    /*
     * Information about previous frame
     */
    Buffer *prevBuffer;
    int prevWidth;
    int prevHeight;

} Cosmo_Device;

/* Need to have a reference for a handler. Will not work if there are two
   Cosmo cards installed */

static Cosmo_Device *cosmoDeviceRef = NULL;

/*
 * Prototypes for internal routines
 */
static int needToScale (void);
static void clInit(int);
static void vlInit(void);
static void initglobals(void);
static void moveVideoWindow(ClientData, XEvent *);
static void destroyVideoPath(void);
   

/*
 *--------------------------------------------------------------
 *
 * vlInit --
 *
 *	Initialize the video path.
 *
 * Results:
 *	No return value
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static void
vlInit (void)
{
    VLDevice *devPtr;
    VLDevList devList;
    VLControlValue val;
    int n;

    if (image.display == GRAPHICS) 
	return;
    
    if (!(video.svr = vlOpenVideo("")) ) {
        vlPerror("Unable to open video -- ");
        exit(1);
    }
    
    vlGetDeviceList(video.svr, &devList);
    
    for (n = 0; n < devList.numDevices; n++) {
	devPtr = &(devList.devices[n]);
        if (strcmp (devPtr->name, "ev1") == 0)
	    break;
        else devPtr = NULL;
    }
    
    if (devPtr == NULL) {
	fprintf(stderr, "IndyVideo/Galileo not found\n");
	exit (1);
    }

    video.devNode = vlGetNode(video.svr, VL_DEVICE, 0, VL_ANY);

    video.path = vlCreatePath(video.svr, devPtr->dev, video.devNode,
			      video.devNode);
    if (video.path == -1) vlPerror("Bad path");

    vlSetupPaths(video.svr, (VLPathList)&video.path, 1, VL_SHARE,
		 VL_READ_ONLY);

    vlGetControl(video.svr, video.path, video.devNode, VL_SYNC, &val);
    if (val.intVal == VL_EV1_SYNC_SLAVE) {
	val.intVal = VL_SYNC_INTERNAL;
	if (vlSetControl(video.svr, video.path, video.devNode, VL_SYNC, &val))
	    vlPerror("Warning: unable to set video timing -- ");
    }

    vlDestroyPath(video.svr, video.path);
    
    /*
    ** setup the timing path
    ** (IndyVideo/Galileo -> Cosmo)
    */
    video.timingSrc = vlGetNode(video.svr, VL_SRC, VL_SCREEN, VL_ANY);
    video.timingDrn = vlGetNode(video.svr, VL_DRN, VL_VIDEO, 2);
    video.timingPath = 
	vlCreatePath(video.svr, devPtr->dev, video.timingSrc, video.timingDrn); 
    if (video.timingPath == -1) {
	vlPerror("vlCreatePath failed for timing path");
	exit(1);
    }

    if (vlSetupPaths(video.svr, (VLPathList)&video.timingPath, 1, 
						VL_SHARE, VL_SHARE) < 0) {
	vlPerror ("Unable to start video timing -- ");
	exit (1);
    }

    if (vlBeginTransfer(video.svr, video.timingPath, 0, NULL)) {
	vlPerror ("Unable to vlBeginTransfer video timing -- ");
	exit (1);
    }
    video.timingActive = 1;

    /*
    ** setup the data transfer path
    ** (Cosmo -> IndyVideo/Galileo --> screen, video)
    */
    video.src = vlGetNode(video.svr, VL_SRC, VL_VIDEO, 1);
    video.drn = vlGetNode(video.svr, VL_DRN, VL_SCREEN, VL_ANY);
    video.voutdrn = vlGetNode(video.svr, VL_DRN, VL_VIDEO, 0);
    video.dataTiming = vlGetNode(video.svr, VL_DRN, VL_VIDEO, 2);
    
    video.path = vlCreatePath(video.svr, devPtr->dev, video.src, video.drn); 
    if (video.path == -1) {
	vlPerror("vlCreatePath failed for data path");
	vlDestroyPath(video.svr, video.timingPath);
	exit(1);
    }

    if (vlAddNode(video.svr, video.path, video.voutdrn)) {
	vlPerror("voutdrn Node Error");
    }

    if (vlAddNode(video.svr, video.path, video.dataTiming)) {
	vlPerror("dataTiming Node Error");
    }

    if (vlSetupPaths(video.svr, (VLPathList)&video.path, 1,
						 VL_SHARE, VL_SHARE) < 0) {
	vlPerror ("Unable to setup data path");
	vlDestroyPath(video.svr, video.timingPath);
	exit(1);
    }
}


/*
 *--------------------------------------------------------------
 *
 * needToScale --
 *
 *	Determines if the image needs to be scaled to fit the
 *      video window.
 *
 * Results:
 *	Returns 1 if scaling is needed, otherwise 0.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static int
needToScale (void)
{
    int diff;
    int fudge = 5;
    
    if (image.interlacing != DM_IMAGE_NONINTERLACED) {
	if (winWidth == image.width || winWidth == image.width * 2) {
	    diff = abs (0.5*(winHeight+6) - image.height); /*6 trimmed lines*/
	    if (diff <= fudge) 
		return 1;
	}
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * clInit --
 *
 *	Set codec parameters.
 *
 * Results:
 *	No return value.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static void
clInit (int mode)
{
    int n, paramBuf [50];
    int status;
    int internalHeight;

    n = 0;
    if (mode == 0 || image.display == GRAPHICS) {
	paramBuf [n++] = CL_IMAGE_WIDTH;
	paramBuf [n++] = image.width;
	paramBuf [n++] = CL_IMAGE_HEIGHT;
	paramBuf [n++] = (image.interlacing != DM_IMAGE_NONINTERLACED)? 
	    image.height/2 : image.height;
    } else {
	paramBuf [n++] = CL_IMAGE_WIDTH;
	paramBuf [n++] = winWidth;
	internalHeight = needToScale () ? image.height*2:image.height;
	paramBuf [n++] = CL_IMAGE_HEIGHT;
	paramBuf [n++] = (image.interlacing != DM_IMAGE_NONINTERLACED)? 
	    internalHeight/2 : internalHeight;
	paramBuf [n++] = CL_INTERNAL_IMAGE_WIDTH;
	paramBuf [n++] = image.width;
	paramBuf [n++] = CL_INTERNAL_IMAGE_HEIGHT;
	paramBuf [n++] = (image.interlacing != DM_IMAGE_NONINTERLACED)? 
	    image.height/2 : image.height;
    }
    paramBuf [n++] = CL_STREAM_HEADERS;
    paramBuf [n++] = 1;
    if (image.display == GRAPHICS) {
	paramBuf [n++] = CL_ORIGINAL_FORMAT;
	paramBuf [n++] = codec.OriginalFormat;
    }
    if (codec.engine == CL_JPEG_COSMO) {
        if (image.orientation != DM_TOP_TO_BOTTOM) {
            fprintf(stderr, 
     "Cosmo JPEG can process only images with ``top to bottom'' orientation\n");
	    destroyVideoPath ();
        }
	paramBuf [n++] = CL_ENABLE_IMAGEINFO;
	paramBuf [n++] = 1;
        switch (image.display) {
            case GRAPHICS:
              break;
            case VIDEO_1_FIELD:
              paramBuf [n++] = CL_COSMO_VIDEO_TRANSFER_MODE;
              paramBuf [n++] = CL_COSMO_VIDEO_TRANSFER_AUTO_1_FIELD;
              break;
            case VIDEO_2_FIELD:
              paramBuf [n++] = CL_COSMO_VIDEO_TRANSFER_MODE;
              paramBuf [n++] = CL_COSMO_VIDEO_TRANSFER_AUTO_2_FIELD;
              break;
            default:
              fprintf (stderr, "Error in setting display parameters\n");
              break;
          }
    }
    else {
        paramBuf [n++] = CL_ORIENTATION;
        paramBuf [n++] = (image.orientation == DM_BOTTOM_TO_TOP) ? 
                                    CL_BOTTOM_UP : CL_TOP_DOWN;
    }

    /* Open  decompressor */
    if (mode) {
	if ((status=clOpenDecompressor(codec.engine, &codec.Hdl)) != SUCCESS) {
	    if (status == CL_SCHEME_NOT_AVAILABLE)
		fprintf (stderr, "Decompressor is not installed.\n");
	    else
	    if (status == CL_SCHEME_BUSY)
		fprintf (stderr, "Decompressor is in use.\n");
	    else
		fprintf (stderr, "Unable to open decompressor - %u\n",status);
	    destroyVideoPath ();
	}
    } else {
	if ((status=clOpenCompressor(codec.engine, &codec.Hdl)) != SUCCESS) {
	    if (status == CL_SCHEME_NOT_AVAILABLE)
		fprintf (stderr, "Compressor is not installed.\n");
	    else
	    if (status == CL_SCHEME_BUSY)
		fprintf (stderr, "Compressor is in use.\n");
	    else
		fprintf (stderr, "Unable to open compressor - %u\n",status);
	    destroyVideoPath ();
	}
    }

    if (clSetParams(codec.Hdl, paramBuf, n) != SUCCESS) {
	fprintf (stderr, "Error Setting Parameters.\tExiting\n");
	destroyVideoPath ();
    }
}


/*
 *--------------------------------------------------------------
 *
 * initglobals --
 *
 *	Initialize image, video, and codec structures.
 *
 * Results:
 *	No return value.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static void
initglobals( void )
{
    codec.singleFrame = 1;
    codec.OriginalFormat = CL_RGBX;
    codec.engine = CL_JPEG_COSMO;

    image.isTrack   = 0;
    image.interlacing = DM_IMAGE_NONINTERLACED;
    image.orientation = DM_TOP_TO_BOTTOM;
    image.width          = 320; /* will adjust it when the first frame comes */
    image.height         = 240; /* will adjust it when the first frame comes */
    image.display = VIDEO_2_FIELD;
    image.numberFrames = -1;

    video.dataActive = 0;
    video.dataFrozen = 0;
    video.timingActive = 0;
}


/*
 *--------------------------------------------------------------
 *
 * destroyVideoPath --
 *
 *	Attempts to gracefully exit all the threads.
 *
 * Results:
 *	Doesn't return.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

static void
destroyVideoPath (void)
{
    if (video.timingActive == 1) {
        video.timingActive = 0;
        vlDestroyPath(video.svr, video.timingPath);
    }
}

/*
 *--------------------------------------------------------------
 *
 * moveVideoWindow --
 *
 *	Adjust the position of the video window
 *      Not sure why this only works when the cursor is inside the window
 *
 * Results:
 *	No return value
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

static void
moveVideoWindow(ClientData clientData, XEvent *eventPtr)
{
    Window dummyWin;
    int x1, y1;
    VLControlValue val;

    Cosmo_Device *cosmoDevice = (Cosmo_Device*) clientData;

    XTranslateCoordinates(cosmoDevice->dpy, cosmoDevice->win,
			  RootWindow(cosmoDevice->dpy,
				     DefaultScreen(cosmoDevice->dpy)),
			  0, 0,&x1, &y1, &dummyWin);

    val.xyVal.x = x1+30;
    val.xyVal.y = y1+15;
    vlSetControl(video.svr, video.path, video.drn, VL_ORIGIN, &val);
}


/*
 *--------------------------------------------------------------
 *
 * Cosmo_Init --
 *
 *	Initialize the device.
 *
 * Results:
 *	0 for successful initilization, non-zero error code otherwise
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */
int
Cosmo_Init(MJPEGResource *rsrcPtr)
{
    Cosmo_Device *cosmoDevice;
    int result = 0;
    Window dummyWin;
    int x, y, x1, y1, w, h, bw, d;
    VLControlValue val;
    GC gc;                              /* X11 graphics context */
    XGCValues   gcvalue;                /* graphics context value */

    cosmoDevice = (Cosmo_Device *) rsrcPtr->devicePtr->deviceData;
    cosmoDevice->indecode = 0;
    cosmoDevice->win = rsrcPtr->xid;
    cosmoDevice->dpy = Tcm_Display();
    cosmoDevice->gc = XCreateGC(cosmoDevice->dpy, rsrcPtr->xid, 0, 0);
    XSetForeground(cosmoDevice->dpy, cosmoDevice->gc, 0);

    setreuid(geteuid(), getuid());
    setregid(getegid(), getgid());
    initglobals();
    vlInit();

    gcvalue.foreground = 0;         /* better be black ... */
    gcvalue.fill_style = FillSolid;
    gc=XCreateGC(cosmoDevice->dpy, cosmoDevice->win,
		 (GCForeground|GCFillStyle), &gcvalue);
    XFillRectangle(cosmoDevice->dpy, cosmoDevice->win, gc, 0, 0, val.xyVal.x,
		   val.xyVal.y);
    
    val.intVal = cosmoDevice->win;
    vlSetControl(video.svr, video.path, video.drn, VL_WINDOW, &val);

    XTranslateCoordinates(cosmoDevice->dpy, cosmoDevice->win,
			  RootWindow(cosmoDevice->dpy,
				     DefaultScreen(cosmoDevice->dpy)),
			  0, 0,&x1, &y1, &dummyWin);

    XGetGeometry(cosmoDevice->dpy, cosmoDevice->win, &dummyWin, &x, &y, &w,
		  &h, &bw, &d);

    val.xyVal.x = x1+30;
    val.xyVal.y = y1+15;
    vlSetControl(video.svr, video.path, video.drn, VL_ORIGIN, &val);

    /* Will set size when the first frame comes */

    if (vlBeginTransfer(video.svr, video.path,
			0, NULL)) {
       vlPerror("screenPath couldn't transfer");
       destroyVideoPath ();
    }

    Tk_CreateEventHandler(Tk_NameToWindow(rsrcPtr->interp, ".",
					  Tk_MainWindow(rsrcPtr->interp)),
			  StructureNotifyMask,
			  moveVideoWindow, (ClientData)cosmoDevice);

    cosmoDeviceRef = cosmoDevice;

    cosmoDevice->prevBuffer = NULL;
    return result;
}


/*
 *--------------------------------------------------------------
 *
 * Cosmo_Reinit --
 *
 *	Reinitialize the Cosmo device. Do just the display reinit.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
Cosmo_Reinit(MJPEGResource *rsrcPtr)
{
    Cosmo_Device *cosmoDevice;

    cosmoDevice = (Cosmo_Device *) rsrcPtr->devicePtr->deviceData;

    if (rsrcPtr->xid != cosmoDevice->win) {
	cosmoDevice->win = rsrcPtr->xid;

	if (cosmoDevice->gc)  {
	    XFreeGC(cosmoDevice->dpy, cosmoDevice->gc);
	}
	cosmoDevice->gc = XCreateGC(cosmoDevice->dpy, rsrcPtr->xid, 0, 0);
	XSetForeground(cosmoDevice->dpy, cosmoDevice->gc, 0);
	
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * Cosmo_Decode
 *
 *	Start the decoding process for a frame.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	This function will be periodcially called until frame
 *	decode is complete
 *
 *--------------------------------------------------------------
 */
int
Cosmo_Decode (MJPEGFrame *framePtr)
{
   return 0;
}

/*
 *--------------------------------------------------------------
 *
 * Cosmo_Display --
 *
 *	Toss a decoded frame up on the screen.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Could block if previous shared memory put image hasn't returned.
 *
 *--------------------------------------------------------------
 */
int
Cosmo_Display (MJPEGFrame *framePtr) 
{
    MJPEGResource *objPtr;
    MJPEGDevice *devicePtr;
    Cosmo_Device *cosmoDevice;
    int w, h;
    VLControlValue val;


    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    cosmoDevice = (Cosmo_Device *)devicePtr->deviceData;
    
    w = framePtr->frameHeader.width;
    h = framePtr->frameHeader.height;

    /* Init Cosmo here since now we know the image size */
    
    if (!decompressorOpen) {

       winWidth = w;
       winHeight = h;
       image.width = w;
       image.height = h;
       clInit(1);
       decompressorOpen = 1;
       videoPathExists = 1;

       /* Set the size of the video window */

       val.xyVal.x = w;
       val.xyVal.y = h;
       vlSetControl(video.svr, video.path, video.drn, VL_SIZE, &val);
    }

    /* Decompress the frame directly to the screen */

    if (clDecompress (codec.Hdl, 
		      ((image.interlacing != DM_IMAGE_NONINTERLACED)? 2:1), 
		      framePtr->dataLength-20, 
		      framePtr->data+20, 
		      CL_EXTERNAL_DEVICE) < SUCCESS) 
    {
       fprintf (stderr, 
		"Error Decompressing frame: \n\t Exiting");
       destroyVideoPath ();
    }
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * Cosmo_Refresh --
 *
 *	Called to refresh the display if part of it is exposed.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Display is repainted.  Could block if previous shared
 *	memory put image hasn't returned
 *
 *--------------------------------------------------------------
 */
int
Cosmo_Refresh(MJPEGResource *rsrcPtr)
{
    /* No need to refresh video window */

    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * clClose --
 *
 *	Called to destroy device.
 *
 * Results:
 *	No return value
 *
 * Side effects:
 *	The device must never be referenced again.
 *
 *--------------------------------------------------------------
 */
void
clClose(Tcl_Interp *interp)
{
   if (decompressorOpen) {
      clCloseDecompressor(codec.Hdl);
      decompressorOpen = 0;
   }
   if (videoPathExists) {
      Tk_DeleteEventHandler(Tk_NameToWindow(interp, ".",
					    Tk_MainWindow(interp)),
			    StructureNotifyMask,
			    moveVideoWindow, (ClientData)cosmoDeviceRef);
      cosmoDeviceRef = NULL;
      vlEndTransfer(video.svr, video.path);
      vlDestroyPath(video.svr, video.path);
      vlCloseVideo(video.svr);
      videoPathExists = 0;
   }

}

/*
 *--------------------------------------------------------------
 *
 * Cosmo_Shutdown --
 *
 *	Called to destroy device.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	The device must never be referenced again.
 *
 *--------------------------------------------------------------
 */
int
Cosmo_Shutdown(MJPEGResource *rsrcPtr) 
{
    Cosmo_Device *cosmoDevice = (Cosmo_Device *) rsrcPtr->devicePtr->deviceData;
    
    if (!rsrcPtr->inited) {
	ckfree((char *) cosmoDevice);
	return 0;
    }
    
    clClose(rsrcPtr->interp);
    if (cosmoDevice->gc) {
	XFreeGC(cosmoDevice->dpy, cosmoDevice->gc);
    }	

    ckfree((char *) cosmoDevice);
    rsrcPtr->devicePtr->deviceData = NULL;

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * Cosmo_FreeFrame --
 *
 *	The current frame is being taken away -- clean up any pointers
 *	we have to it.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
Cosmo_FreeFrame(MJPEGFrame *framePtr)
{
   /* No buffers are used */

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * Cosmo_NextFrame --
 *
 *	Chooses the next frame to decode
 *
 * Results:
 *	0 if successfull, non-zero error code otherwise
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	MJPEGResource *rsrcPtr	The device
 *	MJPEGFrame *frameList	List of frames to choose from
 *	MJPEGFrame **chosen	Set to point to the chosen frame
 *
 *--------------------------------------------------------------
 */
static int
Cosmo_NextFrame(MJPEGResource *rsrcPtr, MJPEGFrame *frameList,
		  MJPEGFrame **chosen)
{
    MJPEGFrame *currentFrame ;
    Cosmo_Device *cosmoDevice;
    double speed, now ;
    
    cosmoDevice = (Cosmo_Device *) rsrcPtr->devicePtr->deviceData;
    
    if (frameList == NULL) {
	*chosen = NULL;
	return 1;
    }
    
    speed = LTS_Speed((ClientData) rsrcPtr->lts);
    now = LTS_Value((ClientData) rsrcPtr->lts);
    *chosen = frameList;
    currentFrame = frameList;
    
    while (currentFrame != NULL) {
	if (speed < 0.0) {
	    if (currentFrame->logEnd < (*chosen)->logEnd) {
		*chosen = currentFrame;
	    }
	} else {
	    if (currentFrame->logEnd > (*chosen)->logEnd) {
		*chosen = currentFrame;
	    }
	}
	currentFrame = currentFrame->next;
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * Cosmo_Error --
 *
 *	Translates an error number (passed in) into a human readable
 *	string (returned).
 *
 * Results:
 *	Human readable string that indicates the error.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	MJPEGResource *rsrcPtr	The device
 *	int errNum		The error code
 *
 *--------------------------------------------------------------
 */
char *
Cosmo_Error(MJPEGResource *rsrcPtr, int errNum)
{
    switch (errNum) {
	
      default:
	return "Unknown Cosmo Device Error.\n";
    }
}


/*
 *--------------------------------------------------------------
 *
 * GetCosmoDevice --
 *
 *	Called by resource to determine default device.  We should
 *	only be called if there's no jpeg hardware available.
 *
 * Results:
 *	Pointer to new MJPEGDevice
 *
 * Side effects:
 *	A couple of structures are allocated.
 *
 *--------------------------------------------------------------
 */
MJPEGDevice *
GetCosmoDevice (MJPEGResource *objPtr)
{
    MJPEGDevice *new_dev;

    new_dev = (MJPEGDevice *) ckalloc(sizeof(MJPEGDevice));

    new_dev->asyncDecodeFlag = 0;
    new_dev->decodeProc = Cosmo_Decode;
    new_dev->refreshProc = Cosmo_Refresh;
    new_dev->displayProc = Cosmo_Display;
    new_dev->initProc = Cosmo_Init;
    new_dev->reinitProc = Cosmo_Reinit;
    new_dev->shutdownProc = Cosmo_Shutdown;
    new_dev->freeFrameProc = Cosmo_FreeFrame;
    new_dev->nextFrameProc = Cosmo_NextFrame;
    new_dev->configProc = NULL;		/* XXX Write this! */
    new_dev->errorProc = Cosmo_Error;
    new_dev->deviceData = (Cosmo_Device *) ckalloc(sizeof(Cosmo_Device));
    memset(new_dev->deviceData, 0, sizeof(Cosmo_Device));
    return new_dev;
}

#else /* COSMO */

/*
 * -------------------------------------------------------
 *
 * GetCosmoDevice
 *
 *	Stub routine for non Cosmo enabled machines
 *
 * -------------------------------------------------------
 */

MJPEGDevice *
GetCosmoDevice (MJPEGResource *objPtr)
{
    return NULL;
}

#endif	/* COSMO */
