/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * tcmDisplay.c --
 *
 *	This file provides the basic interface to create a conncetion
 *	to the display on the local machine, as needed by the cm server.
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <X11/Xlib.h>
#include "tcl.h"
#include "tkInt.h"

static Display *dpy = NULL;


/*
 *--------------------------------------------------------------
 *
 * Tcm_MakeDisplay --
 *
 *	Make a connection to X server unix:0.0 and squirrel it away.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	Open a connection to the server and creates a file handler on it.
 *
 *--------------------------------------------------------------
 */
int
Tcm_MakeDisplay(Tcl_Interp *interp)
{
    char *p, *screenName;
    int screenId;
    size_t length;

    /*
     * Separate the screen number from the rest of the display
     * name.  ScreenName is assumed to have the syntax
     * <display>.<screen> with the dot and the screen being
     * optional.
     */

    screenName = getenv("DISPLAY");
    if (screenName == NULL) {
#ifdef _WINDOWS
	screenName = "tkntDisplay:0.0";
#else	    	
	interp->result =
	    "no display name and no $DISPLAY environment variable";
	return TCL_ERROR;
#endif /* _WINDOWS */
    }
    length = strlen(screenName);
    screenId = 0;
    p = screenName+length-1;
    while (isdigit(UCHAR(*p)) && (p != screenName)) {
	p--;
    }
    if ((*p == '.') && (p[1] != '\0')) {
	length = p - screenName;
	screenId = strtoul(p+1, (char **) NULL, 10);
    }

    dpy = XOpenDisplay(screenName);
    if (dpy == NULL) {
	Tcl_AppendResult(interp, "couldn't connect to display \"",
			 screenName, "\"", (char *) NULL);
	return TCL_ERROR;
    }
#ifndef _WINDOWS
    Tcm_CreateFileHandler2("tk", ConnectionNumber(dpy), TkXFileProc,
			   TK_READABLE, (ClientData) dpy);
#endif
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_SetDisplay --
 *
 *      Like Tcm_MakeDisplay but uses a display pointer already created.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

int
Tcm_SetDisplay(Tcl_Interp *interp, Display *displayPtr)
{
    dpy = displayPtr;
    if (dpy == NULL) {
	Tcl_AppendResult(interp, "Failed to connect to X server",
			 (char *) NULL);
	return TCL_ERROR;
    }
#ifndef _WINDOWS
    Tcm_CreateFileHandler2("tk", ConnectionNumber(dpy), TkXFileProc,
			   TK_READABLE, (ClientData) dpy);
#endif
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_Display --
 *
 *	Return the current display.
 *
 * Results:
 *	A pointer to the display.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

Display *
Tcm_Display(void)
{
    return dpy;
}
