/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */



#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include "cmDither.h"
#include <stdio.h>
#include <math.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <assert.h>

/* #define DEBUG */

typedef unsigned char Boolean;

#ifndef FALSE
#    define FALSE 0
#endif

#ifdef _WINDOWS
HDC tcm_DitherHDC;
#endif

#define abs(x)	((x)<0?-(x):(x))

typedef unsigned char uchar;

/*
 * Prototypes for functions only used internally
 */
static int
InitGrayScale(int minColors, int maxColors, DitherRec *tempRef);

static int
InitOrderedColor(int minColors, int maxColors, DitherRec *tempRef);

static int
InitBrianDither (int minColors, int maxColors, DitherRec *tempRef);

static int
InitKetanDither (DitherRec *tempRef);

static int
InitTrueColor(int rgbOrder, DitherRec *tempRef);

static int
InitMono( DitherRec *tempRef);

static int
ConvertColor(unsigned char l, unsigned char cr, unsigned char cb,
	     unsigned char *r, unsigned char *g, unsigned char *b);

static void
make_square( double N, int divN[256], int modN[256], int magic[16][16] );    

static int
dithergb( int x, int y,
	     unsigned char r, unsigned char g, unsigned char b,
	     int levels, int divN[256], int modN[256], int magic[16][16] );

/* Given the plethora of color conversion equations we use the following 
   static variable to choose a set.
*/

#define JACK_GAMMA_CORRECTED 0
#define JACK_NOT_GAMMA_CORRECTED 1

static colorEquationSet = JACK_GAMMA_CORRECTED;


/*
 *--------------------------------------------------------------
 *
 * InitDither --
 *
 *	Initializes the dithering library by setting DitherProc
 *	to point to the appropriate function based on the display,
 *	window and requested dither.  Also, allocates colors in the
 *	global colormap to use in the dither.  Currently only
 *	dithers of YCrCb411 images are allowed.
 *
 * Results:
 *	returns a DitherRec which contains parameters
 *      that need to be passed to dither routines.
 *
 * Side effects:
 *	colorMap is the internal lookup table which translates
 *	values to pixels allocated in the global X colormap.
 *	myPixels is an array of the allocated pixels.  numPixels is
 *	the number of pixels in myPixels which are allocated.
 *	These are required to correctly de-allocate colors.
 *	display is set to point to dpy, and ditherWin to the
 *	passed in ditherWin.  All of these are in the
 *	returned DitherRec
 *
 * Parameters:
 *     DitherParam dithParams	structure containing the dither
 *                              parameters.
 *
 *         It contains the following:
 *
 *            DitherType dType	The type of dither to use.
 *				The valid types are listed in
 *				dither.h
 *	      int minColors	The minimum number of colors to
 *				use in the dither.  If fewer
 *				than this number are available,
 *				InitDither will fail.
 *	      PicFormat pFormat	The color format of the incoming
 *				picture.  Currently only
 *				YCRCB411 is supported.
 *	Display *dpy		A pointer to the display being
 *				dithered to.  Used to determine
 *				color depth and other display
 *				attributes.
 *      Window *ditherWin	The window which will be dithered
 *				into.
 *	DitherFunc *DitherProc  A pointer to the function pointer
 *				which will be used to call the
 *				appropriate dither routine
 *--------------------------------------------------------------
 */
#ifdef _WINDOWS
DitherRec 
*InitDither(DitherParam dithParams,Display *dpy, Window ditherWin,
	    DitherFunc *DitherProc) {
    
    DitherRec *tempRef;
    unsigned int depth;
    int rgbOrder = 1;

    tcm_DitherHDC = GetDC(ditherWin);
    depth = GetDeviceCaps(tcm_DitherHDC, BITSPIXEL);

    /*
     * Allocate the dither reference to be returned
     */
    tempRef=(DitherRec *)malloc(sizeof(DitherRec));

	tempRef->display = dpy;
    tempRef->ditherWin=ditherWin;
    tempRef->dType = dithParams.dType;
    tempRef->pFormat = dithParams.pFormat;
    if (depth == 8) {
	tempRef->depth = depth;
    } else {
	tempRef->depth = depth = 24;
    }

    if (tempRef->dType == CHOOSE_BEST) {
	if (depth == 8) {
/*	    tempRef->dType = BRIAN_DITHER; */
	    tempRef->dType = KETAN_DITHER;
	} else if (depth >= 24) {
	    tempRef->dType = TRUE_COLOR;
	} else if (depth == 1) {
	    tempRef->dType = MONO;
	}
    }

    /*
     * Figure out which picture format
     */ 
    if ( dithParams.pFormat == YCRCB411 ) {
	
	/*
	 * Determine what kind of dither is being requested
	 * and call the appropriate initialization.
	 */
	switch (tempRef->dType) {

	  case GRAY_SCALE : 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBGray8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitGrayScale(dithParams.minColors,
			  dithParams.maxColors,
			  tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case ORDERED_COLOR : 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBOrderedColor8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitOrderedColor(dithParams.minColors,
			     dithParams.maxColors,
			     tempRef)) {	      
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case BRIAN_DITHER: 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBBrian8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitBrianDither(dithParams.minColors, dithParams.maxColors,
				tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case KETAN_DITHER:
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBKetan8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitKetanDither(tempRef)) {
		return tempRef;
	    } else {
		free(tempRef);
		return NULL;
	    }
	    break;

	  case MONO : 
	    *DitherProc=DitherYCrCb411MSBMono;

	    if (InitMono(tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;
	    
	  case TRUE_COLOR :
	    *DitherProc=DitherYCrCb411TrueColor;

	    if (InitTrueColor(rgbOrder, tempRef)) {
		return tempRef;
	    } else {
		free(tempRef);
		return NULL;
	    }
	    break;

  	  default:free(tempRef);
	    return NULL;
	    break;
	    
	}
    } else if (dithParams.pFormat == YCRCB422) {
	
	/*
	 * Determine what kind of dither is being requested
	 * and call the appropriate initialization.
	 */
	switch (tempRef->dType) {

	  case GRAY_SCALE : 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBGray8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitGrayScale(dithParams.minColors,
			  dithParams.maxColors,
			  tempRef)) {	      
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case BRIAN_DITHER: 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb422MSBBrian8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitBrianDither(dithParams.minColors, dithParams.maxColors,
				tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case KETAN_DITHER: 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb422MSBKetan8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitKetanDither(tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case TRUE_COLOR :
	    *DitherProc=DitherYCrCb422TrueColor;

	    if (InitTrueColor(rgbOrder, tempRef)) {
		return tempRef;
	    } else {
		free(tempRef);
		return NULL;
	    }
	    break;
	    
  	  default:free(tempRef);
	    return NULL;
	    break;
	}	
    } else if ( dithParams.pFormat == FMT_RGB ) {

#ifdef DEBUG
	printf("Picture Format RGB\n");
#endif
	
	/*
	 * Determine what kind of dither is being requested
	 * and call the appropriate initialization.
	 */
	switch (tempRef->dType) {

	  case BRIAN_DITHER:
	  case KETAN_DITHER:
	  case ORDERED_COLOR : 
	    if (depth == 8) {
		*DitherProc=DitherRGBMSBOrderedColor8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitOrderedColor(dithParams.minColors,
			     dithParams.maxColors,
			     tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case GRAY_SCALE :
	  case MONO :
	  case TRUE_COLOR :
	  default:free(tempRef);
	    return NULL;
	    break;
	    
	}
	
    } else {

	free(tempRef);
	
	return NULL;
	
    }
} /* end of InitDither */
#else
DitherRec 
*InitDither(DitherParam dithParams,Display *dpy, Window ditherWin,
	    DitherFunc *DitherProc) {
    
    DitherRec *tempRef;
    unsigned int depth;
    unsigned char byteOrder;
    XWindowAttributes win_attrs;
    int visual_class;
    int rgbOrder;
    XVisualInfo *visual_list;

    /*
     * Allocate the dither reference to be returned
     */
    tempRef=(DitherRec *)malloc(sizeof(DitherRec));

    visual_list = NULL;
    visual_class = PseudoColor;

    XGetWindowAttributes(dpy, ditherWin, &win_attrs);

    depth = win_attrs.depth;
    byteOrder=ImageByteOrder(dpy);
    
    tempRef->display=dpy;
    tempRef->ditherWin=ditherWin;
    tempRef->dType = dithParams.dType;
    tempRef->pFormat = dithParams.pFormat;
    tempRef->depth = depth;
    tempRef->byteOrder = byteOrder;
    tempRef->visual = win_attrs.visual;

    if (tempRef->dType == CHOOSE_BEST) {
	XVisualInfo visual_template;
	int num_visuals, i;

	visual_template.depth = depth;

	visual_list = XGetVisualInfo(dpy, VisualDepthMask, &visual_template, &num_visuals);

	for (i=0; i<num_visuals; i++) {
	    if (visual_list[i].visual == win_attrs.visual) {
		visual_class = visual_list[i].class;
		break;
	    }
	}
	if (i==num_visuals) {
	    if (depth == 1) {
		visual_class = StaticGray;
	    } else if (depth < 24) {
		visual_class = PseudoColor;
	    } else if (depth >= 24) {
		visual_class = TrueColor;
	    }
	}
    
	switch (visual_class) {
	  case PseudoColor:
	  case StaticColor:
/*	    tempRef->dType = BRIAN_DITHER; */
	    tempRef->dType = KETAN_DITHER;
	    break;

	  case GrayScale:
	  case StaticGray:
	    if (depth > 1)
	      tempRef->dType = GRAY_SCALE;
	    else
	      tempRef->dType = MONO;
	    break;

	  case TrueColor:
	  case DirectColor:
	    tempRef->dType = TRUE_COLOR;
	    if (visual_list[i].red_mask & 0xff0000) {
		rgbOrder = 1;
	    } else {
		rgbOrder = 0;
	    }
	    break;
	}
    }
    if (visual_list != NULL) {
	XFree(visual_list);
    }

#ifdef DEBUG
    printf("Color Depth: %d ", depth);
    
    switch( byteOrder ) {
        case MSBFirst:
            printf("Byte Order: MSB\n");
            break;
        case LSBFirst:
            printf("Byte Order: LSB\n");
            break;
        default: ;
    }

#endif
    
    /*
     * Figure out which picture format
     */ 
    if ( dithParams.pFormat == YCRCB411 ) {
	
	/*
	 * Determine what kind of dither is being requested
	 * and call the appropriate initialization.
	 */
	switch (tempRef->dType) {

	  case GRAY_SCALE : 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBGray8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitGrayScale(dithParams.minColors,
			  dithParams.maxColors,
			  tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case ORDERED_COLOR : 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBOrderedColor8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitOrderedColor(dithParams.minColors,
			     dithParams.maxColors,
			     tempRef)) {	      
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case BRIAN_DITHER: 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBBrian8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitBrianDither(dithParams.minColors, dithParams.maxColors,
				tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case KETAN_DITHER: 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBKetan8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitKetanDither(tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case MONO : 
	    *DitherProc=DitherYCrCb411MSBMono;

	    if (InitMono(tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;
	    
	  case TRUE_COLOR :
	    *DitherProc=DitherYCrCb411TrueColor;

	    if (InitTrueColor(rgbOrder, tempRef)) {
		return tempRef;
	    } else {
		free(tempRef);
		return NULL;
	    }
	    break;

  	  default:free(tempRef);
	    return NULL;
	    break;
	    
	}
    } else if (dithParams.pFormat == YCRCB422) {
	
	/*
	 * Determine what kind of dither is being requested
	 * and call the appropriate initialization.
	 */
	switch (tempRef->dType) {

	  case GRAY_SCALE : 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb411MSBGray8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitGrayScale(dithParams.minColors,
			  dithParams.maxColors,
			  tempRef)) {	      
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;
	    
	  case BRIAN_DITHER: 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb422MSBBrian8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitBrianDither(dithParams.minColors, dithParams.maxColors,
				tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case KETAN_DITHER: 
	    if (depth == 8) {
		*DitherProc=DitherYCrCb422MSBKetan8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitKetanDither(tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case TRUE_COLOR :
	    *DitherProc=DitherYCrCb422TrueColor;

	    if (InitTrueColor(rgbOrder, tempRef)) {
		return tempRef;
	    } else {
		free(tempRef);
		return NULL;
	    }
	    break;
	    
  	  default:free(tempRef);
	    return NULL;
	    break;
	}	
    } else if ( dithParams.pFormat == FMT_RGB ) {

#ifdef DEBUG
	printf("Picture Format RGB\n");
#endif
	
	/*
	 * Determine what kind of dither is being requested
	 * and call the appropriate initialization.
	 */
	switch (tempRef->dType) {

	  case BRIAN_DITHER:
	  case KETAN_DITHER:
	  case ORDERED_COLOR : 
	    if (depth == 8) {
		*DitherProc=DitherRGBMSBOrderedColor8;
	    } else {
		free(tempRef);
		return NULL;
	    }

	    if (InitOrderedColor(dithParams.minColors,
			     dithParams.maxColors,
			     tempRef)) {
	      return tempRef;
	    } else {
	      free(tempRef);
	      return NULL;
	    }
	    break;

	  case GRAY_SCALE :
	  case MONO :
	  case TRUE_COLOR :
	  default:free(tempRef);
	    return NULL;
	    break;
	    
	}
	
    } else {

	free(tempRef);
	
	return NULL;
	
    }
} /* end of InitDither */
#endif /* ifdef _WINDOWS ..else .. */


/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb411MSBGray8 --
 *
 *	Grayscale dithers an incoming YCrCb411 formatted picture block 
 *	to an 8-bit MSB display.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane (only one used)
 *	unsigned char *cr      	the cr bit-plane (not used)
 *	unsigned char *cb   	the cb bit-plane (not used)
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane (not used)
 *	int cbStride		stride for cb-plane (not used)
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
DitherErr 
DitherYCrCb411MSBGray8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{ 
    int xPos, yPos;

    for(yPos=0;yPos<h;yPos++)
	    for(xPos=0;xPos<w;xPos++){
		
		out[xCoord+xPos+(yPos+yCoord)*outStride]=(unsigned char)
			ditherRef->colorMap[y[xPos+yStride*yPos]];

  }
  return 0;

} /* end of DitherYCrCb411MSBGray8 */


/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb411TrueColor --
 *
 *	Color dithers an incoming YCrCb411 formatted picture block 
 *      24 bit image.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane
 *	unsigned char *cr      	the cr bit-plane
 *	unsigned char *cb   	the cb bit-plane
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane
 *	int cbStride		stride for cb-plane
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
#ifndef _WINDOWS
DitherErr 
DitherYCrCb411TrueColor(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{
    unsigned int *row1, *row2;
    unsigned char *y1, *y2;
    unsigned char *cr_data, *cb_data;
    int L, CR, CB;
    TrueColorDitherRec *tcDithPtr;
    int cr_r, cr_g, cb_g, cb_b;
    int hpos, wpos;
    unsigned int v; /* used in DO_ASSIGN */

    tcDithPtr = (TrueColorDitherRec *) ditherRef->ditherData;
    if (tcDithPtr == NULL) {
	return BAD_DITHER_REC;
    }

    out = out + (outStride * yCoord) + (xCoord * sizeof(unsigned int));

    if (ditherRef->byteOrder) {
	if (tcDithPtr->rgbOrder) {

	    for (hpos=0; hpos<h; hpos+=2) {
		row1 = (unsigned int *) (out + (outStride * hpos));
		row2 = (unsigned int *) (out + (outStride * (hpos+1)));
		y1 = y + (yStride * hpos);
		y2 = y + (yStride * (hpos+1));
		cr_data = cr + (crStride * (hpos/2));
		cb_data = cb + (cbStride * (hpos/2));
	    
		for (wpos=0; wpos<w; wpos += 2) {
		    int R, G, B;
		
		    CR = *cr_data++;
		    CB = *cb_data++;
		    cr_r = tcDithPtr->cr_r_tab[CR];
		    cr_g = tcDithPtr->cr_g_tab[CR];
		    cb_g = tcDithPtr->cb_g_tab[CB];
		    cb_b = tcDithPtr->cb_b_tab[CB];
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row1,R,G,B);
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row1,R,G,B);
		
		    L = *y2++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		
		    DO_ASSIGN(row2,R,G,B);
		
		    L = *y2++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		
		    DO_ASSIGN(row2,R,G,B);
		}
	    }
	} else {

	    for (hpos=0; hpos<h; hpos+=2) {
		row1 = (unsigned int *) (out + (outStride * hpos));
		row2 = (unsigned int *) (out + (outStride * (hpos+1)));
		y1 = y + (yStride * hpos);
		y2 = y + (yStride * (hpos+1));
		cr_data = cr + (crStride * (hpos/2));
		cb_data = cb + (cbStride * (hpos/2));
	    
		for (wpos=0; wpos<w; wpos += 2) {
		    int R, G, B;
		
		    CR = *cr_data++;
		    CB = *cb_data++;
		    cr_r = tcDithPtr->cr_r_tab[CR];
		    cr_g = tcDithPtr->cr_g_tab[CR];
		    cb_g = tcDithPtr->cb_g_tab[CB];
		    cb_b = tcDithPtr->cb_b_tab[CB];
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;

		    DO_ASSIGN(row1,B,G,R);
		    
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row1,B,G,R);
		    
		    L = *y2++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row2,B,G,R);
		
		    L = *y2++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row2,B,G,R);
		}
	    }
	}
    } else {
	if (tcDithPtr->rgbOrder) {

	    for (hpos=0; hpos<h; hpos+=2) {
		row1 = (unsigned int *) (out + (outStride * hpos));
		row2 = (unsigned int *) (out + (outStride * (hpos+1)));
		y1 = y + (yStride * hpos);
		y2 = y + (yStride * (hpos+1));
		cr_data = cr + (crStride * (hpos/2));
		cb_data = cb + (cbStride * (hpos/2));
	    
		for (wpos=0; wpos<w; wpos += 2) {
		    int R, G, B;
		
		    CR = *cr_data++;
		    CB = *cb_data++;
		    cr_r = tcDithPtr->cr_r_tab[CR];
		    cr_g = tcDithPtr->cr_g_tab[CR];
		    cb_g = tcDithPtr->cb_g_tab[CB];
		    cb_b = tcDithPtr->cb_b_tab[CB];
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row1,B,G,R);
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row1,B,G,R);
		
		    L = *y2++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		
		    DO_TOP_ASSIGN(row2,B,G,R);
		
		    L = *y2++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		
		    DO_TOP_ASSIGN(row2,B,G,R);
		}
	    }
	} else {

	    for (hpos=0; hpos<h; hpos+=2) {
		row1 = (unsigned int *) (out + (outStride * hpos));
		row2 = (unsigned int *) (out + (outStride * (hpos+1)));
		y1 = y + (yStride * hpos);
		y2 = y + (yStride * (hpos+1));
		cr_data = cr + (crStride * (hpos/2));
		cb_data = cb + (cbStride * (hpos/2));
	    
		for (wpos=0; wpos<w; wpos += 2) {
		    int R, G, B;
		
		    CR = *cr_data++;
		    CB = *cb_data++;
		    cr_r = tcDithPtr->cr_r_tab[CR];
		    cr_g = tcDithPtr->cr_g_tab[CR];
		    cb_g = tcDithPtr->cb_g_tab[CB];
		    cb_b = tcDithPtr->cb_b_tab[CB];
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;

		    DO_TOP_ASSIGN(row1,R,G,B);
		    
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row1,R,G,B);
		    
		    L = *y2++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row2,R,G,B);
		
		    L = *y2++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row2,R,G,B);
		}
	    }
	}
    }

    return 0;
}
#else
DitherErr 
DitherYCrCb411TrueColor(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{
    unsigned char *row1, *row2;
    unsigned char *y1, *y2;
    unsigned char *cr_data, *cb_data;
    int L, CR, CB;
    TrueColorDitherRec *tcDithPtr;
    int cr_r, cr_g, cb_g, cb_b;
    int hpos, wpos;

    tcDithPtr = (TrueColorDitherRec *) ditherRef->ditherData;
    if (tcDithPtr == NULL) {
	return BAD_DITHER_REC;
    }

    out = out + (outStride * (imageHeight - yCoord - 1)) + (xCoord * 3);

    for (hpos=0; hpos<h; hpos+=2) {
        row1 = (out - (outStride * hpos));
	row2 = (out - (outStride * (hpos+1)));
	y1 = y + (yStride * hpos);
	y2 = y + (yStride * (hpos+1));
	cr_data = cr + (crStride * (hpos/2));
	cb_data = cb + (cbStride * (hpos/2));
	    
	for (wpos=0; wpos<w; wpos += 2) {
	    int R, G, B;
		
	    CR = *cr_data++;
	    CB = *cb_data++;
	    cr_r = tcDithPtr->cr_r_tab[CR];
	    cr_g = tcDithPtr->cr_g_tab[CR];
	    cb_g = tcDithPtr->cb_g_tab[CB];
	    cb_b = tcDithPtr->cb_b_tab[CB];
		
	    L = *y1++;
	    R = L + cr_r;
	    G = L + cr_g + cb_g;
	    B = L + cb_b;

	    DO_ASSIGN(row1,R,G,B);
		
	    L = *y1++;
	    R = L + cr_r;
	    G = L + cr_g + cb_g;
	    B = L + cb_b;
		
	    DO_ASSIGN(row1,R,G,B);
		
	    L = *y2++;
	    R = L + cr_r;
	    G = L + cr_g + cb_g;
	    B = L + cb_b;
		
	    DO_ASSIGN(row2,R,G,B);
		
	    L = *y2++;
	    R = L + cr_r;
	    G = L + cr_g + cb_g;
	    B = L + cb_b;
		
	    DO_ASSIGN(row2,R,G,B);
	}
    }

    return 0;
}
#endif



/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb422TrueColor --
 *
 *	Color dithers an incoming YCrCb411 formatted picture block 
 *      24 bit image.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane
 *	unsigned char *cr      	the cr bit-plane
 *	unsigned char *cb   	the cb bit-plane
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane
 *	int cbStride		stride for cb-plane
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
#ifndef _WINDOWS
DitherErr 
DitherYCrCb422TrueColor(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{
    unsigned int *row1;
    unsigned char *y1;
    unsigned char *cr_data, *cb_data;
    int L, CR, CB;
    TrueColorDitherRec *tcDithPtr;
    int cr_r, cr_g, cb_g, cb_b;
    int hpos, wpos;
    unsigned int v; /* used in DO_ASSIGN */

    tcDithPtr = (TrueColorDitherRec *) ditherRef->ditherData;
    if (tcDithPtr == NULL) {
	return BAD_DITHER_REC;
    }

    out = out + (outStride * yCoord) + (xCoord * sizeof(unsigned int));

    if (ditherRef->byteOrder) {
	if (tcDithPtr->rgbOrder) {

	    for (hpos=0; hpos<h; hpos++) {
		row1 = (unsigned int *) (out + (outStride * hpos));
		y1 = y + (yStride * hpos);
		cr_data = cr + (crStride * hpos);
		cb_data = cb + (cbStride * hpos);
	    
		for (wpos=0; wpos<w; wpos += 2) {
		    int R, G, B;
		
		    CR = *cr_data++;
		    CB = *cb_data++;
		    cr_r = tcDithPtr->cr_r_tab[CR];
		    cr_g = tcDithPtr->cr_g_tab[CR];
		    cb_g = tcDithPtr->cb_g_tab[CB];
		    cb_b = tcDithPtr->cb_b_tab[CB];
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row1,R,G,B);
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row1,R,G,B);
		}
	    }
	} else {

	    for (hpos=0; hpos<h; hpos++) {
		row1 = (unsigned int *) (out + (outStride * hpos));
		y1 = y + (yStride * hpos);
		cr_data = cr + (crStride * (hpos/2));
		cb_data = cb + (cbStride * (hpos/2));
		
		for (wpos=0; wpos<w; wpos += 2) {
		    int R, G, B;
		    
		    CR = *cr_data++;
		    CB = *cb_data++;
		    cr_r = tcDithPtr->cr_r_tab[CR];
		    cr_g = tcDithPtr->cr_g_tab[CR];
		    cb_g = tcDithPtr->cb_g_tab[CB];
		    cb_b = tcDithPtr->cb_b_tab[CB];
		    
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row1,B,G,R);
		    
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_ASSIGN(row1,B,G,R);
		}
	    }
	}
    } else {
	if (tcDithPtr->rgbOrder) {

	    for (hpos=0; hpos<h; hpos++) {
		row1 = (unsigned int *) (out + (outStride * hpos));
		y1 = y + (yStride * hpos);
		cr_data = cr + (crStride * hpos);
		cb_data = cb + (cbStride * hpos);
		
		for (wpos=0; wpos<w; wpos += 2) {
		    int R, G, B;
		    
		    CR = *cr_data++;
		    CB = *cb_data++;
		    cr_r = tcDithPtr->cr_r_tab[CR];
		    cr_g = tcDithPtr->cr_g_tab[CR];
		    cb_g = tcDithPtr->cb_g_tab[CB];
		    cb_b = tcDithPtr->cb_b_tab[CB];
		    
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row1,B,G,R);
		
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row1,B,G,R);
		}
	    }
	} else {

	    for (hpos=0; hpos<h; hpos++) {
		row1 = (unsigned int *) (out + (outStride * hpos));
		y1 = y + (yStride * hpos);
		cr_data = cr + (crStride * (hpos/2));
		cb_data = cb + (cbStride * (hpos/2));
		
		for (wpos=0; wpos<w; wpos += 2) {
		    int R, G, B;
		    
		    CR = *cr_data++;
		    CB = *cb_data++;
		    cr_r = tcDithPtr->cr_r_tab[CR];
		    cr_g = tcDithPtr->cr_g_tab[CR];
		    cb_g = tcDithPtr->cb_g_tab[CB];
		    cb_b = tcDithPtr->cb_b_tab[CB];
		    
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row1,R,G,B);
		    
		    L = *y1++;
		    L = UP(L);
		    R = L + cr_r;
		    G = L + cr_g + cb_g;
		    B = L + cb_b;
		    
		    DO_TOP_ASSIGN(row1,R,G,B);
		}
	    }
	}
    }

    return 0;
}

#else

DitherErr 
DitherYCrCb422TrueColor(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{
    unsigned char *row1;
    unsigned char *y1;
    unsigned char *cr_data, *cb_data;
    int L, CR, CB;
    TrueColorDitherRec *tcDithPtr;
    int cr_r, cr_g, cb_g, cb_b;
    int hpos, wpos;

    tcDithPtr = (TrueColorDitherRec *) ditherRef->ditherData;
    if (tcDithPtr == NULL) {
	return BAD_DITHER_REC;
    }

    out = out + (outStride * (imageHeight - yCoord - 1)) + (xCoord * 3);

    for (hpos=0; hpos<h; hpos++) {
	row1 = (out - (outStride * hpos));
	y1 = y + (yStride * hpos);
	cr_data = cr + (crStride * hpos);
	cb_data = cb + (cbStride * hpos);
	    
	for (wpos=0; wpos<w; wpos += 2) {
	    int R, G, B;
		
	    CR = *cr_data++;
	    CB = *cb_data++;
	    cr_r = tcDithPtr->cr_r_tab[CR];
	    cr_g = tcDithPtr->cr_g_tab[CR];
	    cb_g = tcDithPtr->cb_g_tab[CB];
	    cb_b = tcDithPtr->cb_b_tab[CB];
		
	    L = *y1++;
	    R = L + cr_r;
	    G = L + cr_g + cb_g;
	    B = L + cb_b;

	    DO_ASSIGN(row1,R,G,B);

	    L = *y1++;
	    R = L + cr_r;
	    G = L + cr_g + cb_g;
	    B = L + cb_b;

	    DO_ASSIGN(row1,R,G,B);
	}
    }

    return 0;
}

#endif


/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb411MSBBrian8 --
 *
 *	Color dithers an incoming YCrCb411 formatted picture block 
 *	to an 8-bit MSB display using Brian's ordered color method.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane
 *	unsigned char *cr      	the cr bit-plane
 *	unsigned char *cb   	the cb bit-plane
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane
 *	int cbStride		stride for cb-plane
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
static XColor currColors[256];
DitherErr 
DitherYCrCb411MSBBrian8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{
  unsigned char ***crCbArray;
  unsigned char **dithPosArray;
  unsigned char *crData, *cbData, *yDataFirst, *yDataSecond;
  unsigned char *outDataFirst, *outDataSecond;
  Boolean wDithQuad, hDithQuad;
  int hpos, wpos, crCbIndex;
  BrianDitherRec *ditherRec;
  
  ditherRec = (BrianDitherRec *) ditherRef->ditherData;
  crCbArray = (uchar ***) ditherRec->crCbArray2;

  hDithQuad = FALSE;
  wDithQuad = FALSE;

#ifdef _WINDOWS
    out = out + (outStride * (imageHeight - yCoord - 1)) + xCoord;
#else
    out = out + (outStride * yCoord) + xCoord;
#endif

  for (hpos=0; hpos<h; hpos+=2) {
    crData = cr + (crStride * (hpos/2));
    cbData = cb + (cbStride * (hpos/2));
#ifdef _WINDOWS
    outDataFirst = out - (outStride * hpos);
    outDataSecond = out - (outStride * (hpos+1));
#else
    outDataFirst = out + (outStride * hpos);
    outDataSecond = out + (outStride * (hpos+1));
#endif
    yDataFirst = y + (yStride * hpos);
    yDataSecond = y + (yStride * (hpos+1));
    
    for (wpos=0; wpos<w; wpos+=2) {
      crCbIndex = 256*(*crData) + (*cbData);
      dithPosArray = crCbArray[crCbIndex];
      crData++;
      cbData++;
      
      if (hDithQuad) {
	if (wDithQuad) {
	  outDataFirst[0] = (dithPosArray[1])[yDataFirst[0]];
	  outDataFirst[1] = (dithPosArray[9])[yDataFirst[1]];
	  outDataSecond[0] = (dithPosArray[13])[yDataSecond[0]];
	  outDataSecond[1] = (dithPosArray[5])[yDataSecond[1]];
	} else {
	  outDataFirst[0] = (dithPosArray[3])[yDataFirst[0]];
	  outDataFirst[1] = (dithPosArray[11])[yDataFirst[1]];
	  outDataSecond[0] = (dithPosArray[15])[yDataSecond[0]];
	  outDataSecond[1] = (dithPosArray[7])[yDataSecond[1]];
	}
      } else {
	if (wDithQuad) {
	  outDataFirst[0] = (dithPosArray[2])[yDataFirst[0]];
	  outDataFirst[1] = (dithPosArray[10])[yDataFirst[1]];
	  outDataSecond[0] = (dithPosArray[14])[yDataSecond[0]];
	  outDataSecond[1] = (dithPosArray[6])[yDataSecond[1]];
	} else {
	  outDataFirst[0] = (dithPosArray[0])[yDataFirst[0]];
	  outDataFirst[1] = (dithPosArray[8])[yDataFirst[1]];
	  outDataSecond[0] = (dithPosArray[12])[yDataSecond[0]];
	  outDataSecond[1] = (dithPosArray[4])[yDataSecond[1]];
	}
      }
      outDataFirst += 2;
      yDataFirst += 2;
      outDataSecond += 2;
      yDataSecond += 2;

      wDithQuad = !wDithQuad;
    }

    hDithQuad = !hDithQuad;
  }
  return 0;
}


/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb422MSBBrian8 --
 *
 *	Color dithers an incoming YCrCb422 formatted picture block 
 *	to an 8-bit MSB display using Brian's ordered color method.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane
 *	unsigned char *cr      	the cr bit-plane
 *	unsigned char *cb   	the cb bit-plane
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane
 *	int cbStride		stride for cb-plane
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
DitherErr 
DitherYCrCb422MSBBrian8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{
    BrianDitherRec *ditherRec;
    unsigned char ***crCbArray;
    unsigned char **dithPosArray;
    unsigned char *crData, *cbData, *yDataFirst;
    unsigned char *outDataFirst;
    Boolean wDithQuad, hDithQuad;
    int hpos, wpos, crCbIndex;

    ditherRec = (BrianDitherRec *) ditherRef->ditherData;
    crCbArray = (uchar ***) ditherRec->crCbArray2;

    hDithQuad = FALSE;
    wDithQuad = FALSE;

#ifdef _WINDOWS
    out = out + (outStride * (imageHeight - yCoord - 1)) + xCoord;
#else
    out = out + (outStride * yCoord) + xCoord;
#endif

    for(hpos=0; hpos<h; hpos++) {
	crData = cr + (crStride * hpos);
	cbData = cb + (cbStride * hpos);
#ifdef _WINDOWS
	outDataFirst = out - (outStride * hpos);
#else
	outDataFirst = out + (outStride * hpos);
#endif
	yDataFirst = y + (yStride * hpos);

	for (wpos=0; wpos<w; wpos+=2) {
	    crCbIndex = 256*(*crData) + (*cbData);
	    crData++;
	    cbData++;
	    dithPosArray = crCbArray[crCbIndex];

	    if (hDithQuad == 0) {
		if (wDithQuad) {
		    outDataFirst[0] = (dithPosArray[2])[yDataFirst[0]];
		    outDataFirst[1] = (dithPosArray[10])[yDataFirst[1]];
		} else {
		    outDataFirst[0] = (dithPosArray[0])[yDataFirst[0]];
		    outDataFirst[1] = (dithPosArray[8])[yDataFirst[1]];
		}
	    } else if (hDithQuad == 1) {
		if (wDithQuad) {
		    outDataFirst[0] = (dithPosArray[14])[yDataFirst[0]];
		    outDataFirst[1] = (dithPosArray[6])[yDataFirst[1]];
		} else {
		    outDataFirst[0] = (dithPosArray[12])[yDataFirst[0]];
		    outDataFirst[1] = (dithPosArray[4])[yDataFirst[1]];
		}
	    } else if (hDithQuad == 2) {
		if (wDithQuad) {
		    outDataFirst[0] = (dithPosArray[1])[yDataFirst[0]];
		    outDataFirst[1] = (dithPosArray[9])[yDataFirst[1]];
		} else {
		    outDataFirst[0] = (dithPosArray[3])[yDataFirst[0]];
		    outDataFirst[1] = (dithPosArray[11])[yDataFirst[1]];
		}
	    } else if (hDithQuad == 3) {
		if (wDithQuad) {
		    outDataFirst[0] = (dithPosArray[13])[yDataFirst[0]];
		    outDataFirst[1] = (dithPosArray[5])[yDataFirst[1]];
		} else {
		    outDataFirst[0] = (dithPosArray[15])[yDataFirst[0]];
		    outDataFirst[1] = (dithPosArray[7])[yDataFirst[1]];
		}	  
	    } 
	    outDataFirst += 2;
	    yDataFirst += 2;
	    wDithQuad = !wDithQuad;
	}
	hDithQuad = (hDithQuad+1)%4;
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb411MSBOrderedColor8 --
 *
 *	Color dithers an incoming YCrCb411 formatted picture block 
 *	to an 8-bit MSB display using the ordered color method.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane
 *	unsigned char *cr      	the cr bit-plane
 *	unsigned char *cb   	the cb bit-plane
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane
 *	int cbStride		stride for cb-plane
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
DitherErr 
DitherYCrCb411MSBOrderedColor8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{
    int xPos, yPos;
    int xPosCrCb, yPosCrCb;
    Boolean xCount=FALSE, yCount=FALSE;
    unsigned char r, g, b;

    yPosCrCb=0;

    for(yPos=0;yPos<h;yPos++) {

	/*
	 * only advance in Cr & Cb planes every other advance
	 * in Y plane
	 */
	if(yCount) yPosCrCb++;

	yCount= !yCount;
	
	/*
	 * initialize for loop through columns
	 */
	xPosCrCb=0;

	xCount=FALSE;
	
	for(xPos=0,xPosCrCb=0;xPos<w;xPos++){
	    
	    /*
	     * only advance in Cr & Cb planes every other advance
	     * in Y plane
	     */
	    if (xCount) xPosCrCb++;

	    xCount= !xCount;
	    
	    ConvertColor(y[xPos+yStride*yPos],
			 cr[xPosCrCb+crStride*yPosCrCb],
			 cb[xPosCrCb+cbStride*yPosCrCb],
			 &r, &g, &b);
	    
	    out[xPos+yPos*outStride]=(unsigned char)
		    ditherRef->pixels[dithergb( xCoord+xPos, yCoord+yPos,
					       r, g, b,
					       ditherRef->levels,
					       ditherRef->divN,
					       ditherRef->modN,
					       ditherRef->magic )];
/*
#ifdef DEBUG
	    printf("%d %d %d ",r,g,b);
#endif
  */     					       
	}
    }

    
    return 0;
} /* end of DitherYCrCb411MSBOrderedColor8 */ 


/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb411MSBMono --
 *
 *	Black and White dithers an incoming YCrCb411 formatted
 *	picture block.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane (only one used)
 *	unsigned char *cr      	the cr bit-plane (not used)
 *	unsigned char *cb   	the cb bit-plane (not used)
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane (not used)
 *	int cbStride		stride for cb-plane (not used)
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
DitherErr 
DitherYCrCb411MSBMono(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{ 
    int xPos, yPos;
    int row,col;

    char temp;

    unsigned char v;
    
    for(yPos=0;yPos<h;yPos++)
	    for(xPos=0;xPos<w;xPos++){

		row = (yCoord+yPos) % 16;
		col = (xCoord+xPos) % 16;

		v=y[xPos+yStride*yPos];
		
		temp=(ditherRef->modN[v] > ditherRef->magic[col][row] ?
		 ditherRef->divN[v] + 1 : ditherRef->divN[v]);

		/*if(y[xPos+yStride*yPos] > 128) temp=1;
		else temp=0;
		*/
		
		out[xPos+yPos*outStride]=(unsigned char)
			ditherRef->colorMap[temp];				       
	    }

    return 0;
} /* end of DitherYCrCb411MSBMono */


/*
 *--------------------------------------------------------------
 *
 * DitherRGBMSBOrderedColor8 --
 *
 *	Color dithers an incoming RGB formatted picture block 
 *	to an 8-bit MSB display using the ordered color method.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *r       	the r bit-plane
 *	unsigned char *g      	the g bit-plane
 *	unsigned char *b   	the b bit-plane
 *	unsigned char *out	the output bit-plane
 *	int rStride		distance to wrap in buffer to get to
 *				next row in block for r-plane
 *	int gStride		stride for g-plane
 *	int bStride		stride for b-plane
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
DitherErr 
DitherRGBMSBOrderedColor8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *r, unsigned char *g,
		       unsigned char *b, unsigned char *out, 
		       int rStride, int gStride, int bStride, int outStride,
		       int imageWidth, int imageHeight)
{
    int xPos, yPos;

    for(yPos=0;yPos<h;yPos++)
	    for(xPos=0;xPos<w;xPos++){
		
	  
		out[xPos+yPos*outStride]=(unsigned char)
		    ditherRef->pixels[dithergb( xCoord+xPos, yCoord+yPos,
					       r[xPos+yPos*rStride],
					       g[xPos+yPos*gStride],
					       b[xPos+yPos*bStride],
					       ditherRef->levels,
					       ditherRef->divN,
					       ditherRef->modN,
					       ditherRef->magic )];

	    }

    return 0;
} /* end of DitherRGBMSBOrderedColor8 */



/*
 *--------------------------------------------------------------
 *
 * FreeDither --
 *
 *	NULLifies the DitherProc pointer, and de-allocates the
 *	the colors that were allocated for the current dither
 *	type.  After this InitDither can be called again for a
 *	new dither type.
 *
 * Results:
 *	returns a DitherErr as specified in dither.h
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *
 *--------------------------------------------------------------
 */
DitherErr
FreeDither(DitherRec *ditherRef)
{
#ifndef _WINDOWS
    Colormap dcmap;
    
    /*
     * Free the colors that were allocated for dithering.
     */

    dcmap = XDefaultColormap(ditherRef->display,
			     DefaultScreen(ditherRef->display));

    if (ditherRef->numPixels > 0) {
      XFreeColors(ditherRef->display, dcmap, ditherRef->pixels,
		  ditherRef->numPixels, 0); 
    }
#endif

    /*
     * Dither specific freeing goes here.
     */

    if ((ditherRef->dType == TRUE_COLOR) &&
	(ditherRef->ditherData != NULL)) {
	TrueColorDitherRec *tcDithPtr;

	tcDithPtr = (TrueColorDitherRec *) ditherRef->ditherData;

	if (tcDithPtr->cb_b_tab != NULL) {
	    free((char *)tcDithPtr->cb_b_tab);
	    tcDithPtr->cb_b_tab = NULL;
	}

	if (tcDithPtr->cb_g_tab != NULL) {
	    free((char *)tcDithPtr->cb_g_tab);
	    tcDithPtr->cb_g_tab = NULL;
	}

	if (tcDithPtr->cr_r_tab != NULL) {
	    free((char *)tcDithPtr->cr_r_tab);
	    tcDithPtr->cr_r_tab = NULL;
	}

	if (tcDithPtr->cr_g_tab != NULL) {
	    free((char *)tcDithPtr->cr_g_tab);
	    tcDithPtr->cr_g_tab = NULL;
	}

	free(tcDithPtr);
	ditherRef->ditherData = NULL;
    }
    if ((ditherRef->dType == BRIAN_DITHER) &&
        (ditherRef->ditherData != NULL)) {
	BrianDitherRec *ditherRec;
	int i, j;
	int ncr, ncb;
	uchar ***crCbArray1;
	uchar ***crCbArray2;
	uchar **posArray1;
	uchar *yArray;

	ditherRec = (BrianDitherRec *) ditherRef->ditherData;
	ncr = ditherRec->ncr;
	ncb = ditherRec->ncb;
	crCbArray1 = ditherRec->crCbArray1;
	crCbArray2 = ditherRec->crCbArray2;
	for (i=0; i<ncr; i++) {
	    for (j=0; j<ncb; j++) {
		posArray1 = crCbArray1[ncb*i+j];
		free(posArray1);
	    }
	}
	for (i=0; i<256; i++) {
	    for (j=0; j<256; j++) {
		posArray1 = crCbArray2[(i<<8) + j];
		free(posArray1);
	    }
	}
	free(crCbArray1);
	free(crCbArray2);
	free(ditherRec);
	ditherRef->ditherData = NULL;
    }
    if ((ditherRef->dType == KETAN_DITHER) &&
	(ditherRef->ditherData != NULL)) {
	KetanDitherRec *ditherRec;
	int i;

	ditherRec = (KetanDitherRec *) ditherRef->ditherData;
	for (i=0; i<16; i++) {
	    free((char *)ditherRec->yIndexTables[i]);
	}
	free((char *)ditherRec->yIndexTables);

	for (i=0; i<4; i++) {
	  free((char *)ditherRec->crTables[i]);
	  free((char *)ditherRec->cbTables[i]);
	}
	free((char *)ditherRec->crTables);
	free((char *)ditherRec->cbTables);

	free((char *)ditherRec->yCrCbToPixel);
	free((char *)ditherRec);
	ditherRef->ditherData = NULL;
    }

    /*
     * Free the dither record
     */
    free(ditherRef);

    return FREE_OK;
    
} /* end of FreeDither */


/*
 *--------------------------------------------------------------
 *
 * InitGrayScale --
 *
 *	Initializes Gray Scale 8 bit dither by allocating an
 *	appropriate color map.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	colorMap is the internal lookup table which translates
 *	gray values to the allocated pixels.  myPixels is an array
 *	of the allocated pixels.  numPixels is the number of
 *	pixels in myPixels which are allocated.  All of these are
 *	set in the DitherRec passed in.
 *
 * Parameters:
 *
 *   	int minColors		The minimum number of gray levels to
 *				use in the dither.  If fewer
 *				than this number are available,
 *				InitGrayScale will fail.
 *	DitherRec *tempRef	A partially completed dither rec
 *				which will be completed by
 *				InitGrayScale
 *--------------------------------------------------------------
 */
static int
InitGrayScale(int minColors, int maxColors, DitherRec *tempRef) {
    
    XColor tempColor;
    int i,j;
    int colors=0;
    unsigned long plane_masks[256];
    int offset, step;
    Colormap dcmap;
    int startColors;
    
    /*
     * Allocate a block of colors from the global colormap
     */

    dcmap = XDefaultColormap(tempRef->display,
			     DefaultScreen(tempRef->display));

    for(startColors=2;startColors<=maxColors;startColors*=2);

    startColors/=2;
    
    for(colors=startColors;colors>=minColors;colors/=2)
	    if (XAllocColorCells(tempRef->display,dcmap,0,plane_masks,0,
				 tempRef->pixels,colors))
		    break;
#ifdef DEBUG
    printf("Got %d colors\n",colors);
#endif
    
    if(colors<minColors) {
	    return 0;
	}
    
    tempRef->numPixels=colors;
    
    /*
     * Now set the colors appropriately, and set the dither table up with
     * appropriate pixel values
     */
    tempColor.flags = DoRed | DoGreen | DoBlue;

    for(i=0;i<colors;i++) {

	step=65535 / (colors-1);

	offset=256 / colors;
	
	for(j=0;j< offset;j++)
		tempRef->colorMap[i*offset+j]=tempRef->pixels[i];

	tempColor.pixel=tempRef->pixels[i];

	tempColor.red=tempColor.green=tempColor.blue=i*step;

	XStoreColor(tempRef->display,dcmap,&tempColor);

    }

    return 1;
    
} /* end of InitGrayScale */

int ColorDistance(XColor *wanted, XColor *have)
{
  double y1, cr1, cb1;
  double y2, cr2, cb2;
  double r1, g1, b1;
  double r2, g2, b2;
  double dist;
  
  r1 = (double) wanted->red / 256.0;
  g1 = (double) wanted->green / 256.0;
  b1 = (double) wanted->blue / 256.0;

  y1 = 0.299 * r1 + 0.587 * g1 + 0.114 * b1;
  cr1 = (0.596 * r1 - 0.275 * g1 - 0.321 * b1) + 128.0;
  cb1 = (0.212 * r1 - 0.528 * g1 + 0.311 * b1) + 128.0;

  if (y1 > 256.0) y1 = 256.0;
  if (cr1 > 256.0) cr1 = 256.0;
  if (cb1 > 256.0) cb1 = 256.0;
  if (y1 < 0.0) y1 = 0.0;
  if (cr1 < 0.0) cr1 = 0.0;
  if (cb1 < 0.0) cb1 = 0.0;

  
  r2 = (double) have->red / 256.0;
  g2 = (double) have->green / 256.0;
  b2 = (double) have->blue / 256.0;

  y2 = 0.299 * r2 + 0.587 * g2 + 0.114 * b2;
  cr2 = (0.596 * r2 - 0.275 * g2 - 0.321 * b2) + 128.0;
  cb2 = (0.212 * r2 - 0.528 * g2 + 0.311 * b2) + 128.0;

  if (y2 > 256.0) y2 = 256.0;
  if (cr2 > 256.0) cr2 = 256.0;
  if (cb2 > 256.0) cb2 = 256.0;
  if (y2 < 0.0) y2 = 0.0;
  if (cr2 < 0.0) cr2 = 0.0;
  if (cb2 < 0.0) cb2 = 0.0;

  dist = ((y1 - y2) * (y1 - y2) +
	  (cr1 - cr2) * (cr1 - cr2) +
	  (cb1 - cb2) * (cb1 - cb2));

  return (int) dist;
}


/*
 *--------------------------------------------------------------
 *
 * InitTrueColor --
 *
 *	Initializes 24bit dither by creating appropriate y,cr,cb
 *      to r,g,b tables
 *
 * Results:
 *	None.
 *
 * Side effects:
 *      Puts tables into ditherData;
 *
 * Parameters:
 *
 *      int rgbOrder            1 = RGB, 0 = BGR;
 *	DitherRec *tempRef	A partially completed dither rec
 *				which will be completed by
 *				InitOrderedColor
 *--------------------------------------------------------------
 */
#ifndef _WINDOWS
static int
InitTrueColor(int rgbOrder, DitherRec *tempRef) 
{
    TrueColorDitherRec *tcDitherPtr;
    int CB, CR, i;

    tcDitherPtr = (TrueColorDitherRec *) malloc(sizeof(TrueColorDitherRec));
    
    tcDitherPtr->cb_b_tab = (int *)malloc(256*sizeof(int));
    tcDitherPtr->cr_g_tab = (int *)malloc(256*sizeof(int));
    tcDitherPtr->cb_g_tab = (int *)malloc(256*sizeof(int));
    tcDitherPtr->cr_r_tab = (int *)malloc(256*sizeof(int));
    
    for (i=0; i<256; i++) {
	CB = CR = i;
	
	CB -= 128; CR -= 128;
	
	tcDitherPtr->cr_r_tab[i] = FIX(1.596) * CR;
	tcDitherPtr->cr_g_tab[i] = -FIX(0.813) * CR;
	tcDitherPtr->cb_g_tab[i] = -FIX(0.391) * CB;   
	tcDitherPtr->cb_b_tab[i] = FIX(2.018) * CB;
    }
    tcDitherPtr->rgbOrder = rgbOrder;

    tempRef->ditherData = (void *) tcDitherPtr;
    tempRef->numPixels = 0;
    return 1;
}
#else
static int
InitTrueColor(int rgbOrder, DitherRec *tempRef) 
{
    TrueColorDitherRec *tcDitherPtr;
    int CB, CR, i;

    tcDitherPtr = (TrueColorDitherRec *) malloc(sizeof(TrueColorDitherRec));
    
    tcDitherPtr->cb_b_tab = (int *)malloc(256*sizeof(int));
    tcDitherPtr->cr_g_tab = (int *)malloc(256*sizeof(int));
    tcDitherPtr->cb_g_tab = (int *)malloc(256*sizeof(int));
    tcDitherPtr->cr_r_tab = (int *)malloc(256*sizeof(int));
    
    for (i=0; i<256; i++) {
	CB = CR = i;
	
	CB -= 128; CR -= 128;
	
	tcDitherPtr->cr_r_tab[i] = (int) ( 1.596 * CR + .5);
	tcDitherPtr->cr_g_tab[i] = (int) (-0.813 * CR + .5);
	tcDitherPtr->cb_g_tab[i] = (int) (-0.391 * CB + .5);   
	tcDitherPtr->cb_b_tab[i] = (int) ( 2.018 * CB + .5);
    }
    tcDitherPtr->rgbOrder = rgbOrder;

    tempRef->ditherData = (void *) tcDitherPtr;
    tempRef->numPixels = 0;
    return 1;
}
#endif


    
  

/*
 *--------------------------------------------------------------
 *
 * InitOrderedColor --
 *
 *	Initializes ordered color 8 bit dither by allocating an
 *	appropriate color map, and setting up the lookup tables.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	pixels is an array of the allocated pixels in the correct
 *	order to be picked by the dither routine. numPixels is
 *	the number of elements in pixels which are allocated.
 *	divN, modN and magic are also set up.  They are used
 *	to calculate the dithered value for a pixel.
 *	All of these are set in the DitherRec passed in.
 *
 * Parameters:
 *
 *   	int minColors		The minimum number of colors to
 *				use in the dither.  If fewer
 *				than this number are available,
 *				InitOrderedColor will fail.
 *	DitherRec *tempRef	A partially completed dither rec
 *				which will be completed by
 *				InitOrderedColor
 *--------------------------------------------------------------
 */
static int
InitOrderedColor(int minColors, int maxColors, DitherRec *tempRef) {
    
    XColor tempColor;
    int i;
    int colors;
    int levels;
    unsigned long plane_masks[256];
    Colormap dcmap;
    int step, levelsq, levelsc;
    double N;
    int maxLevels;
    
    /*
     * Allocate a block of colors from the global colormap
     */

    dcmap = XDefaultColormap(tempRef->display,
			     DefaultScreen(tempRef->display));

    for(maxLevels=2;
	maxLevels*maxLevels*maxLevels<=maxColors;
	maxLevels++);

    maxLevels--;
    
    for(levels=maxLevels;levels>1;levels--)
	    if (XAllocColorCells(tempRef->display,dcmap,0,plane_masks,0,
				 tempRef->pixels,levels*levels*levels))
		    break;
    
    tempRef->levels=levels;

    colors=levels*levels*levels;
    
#ifdef DEBUG
    printf("Got %d colors\n",colors);
#endif
    
    if(colors<minColors) {
	    return 0;
	}
    
    tempRef->numPixels=colors;
    
    /* 
     * Set up the color map entries.
     */

    levelsq = levels*levels;	/* squared */
    levelsc = levels*levelsq;	/* and cubed */

    N = 255.0 / (levels - 1);    /* Get size of each step */

    step = 65535 / (levels -1);
	    
    tempColor.flags = DoRed | DoGreen | DoBlue;
    
    for(i = 0; i < levelsc; i++) {

	tempColor.pixel=tempRef->pixels[i];

	tempColor.red = (i%levels) * step;
	tempColor.green = ((i/levels)%levels) * step;
	tempColor.blue = ((i/levelsq)%levels) * step;
	
	XStoreColor(tempRef->display,dcmap,&tempColor);
    }

    make_square( N, tempRef->divN, tempRef->modN, tempRef->magic );

    return 1;
    
} /* end of InitOrderedColor */


/*
 *--------------------------------------------------------------
 *
 * InitMono --
 *
 *	Initializes ordered color black and white dither by
 *	finding the X11 pixel values for black and white,
 *	and setting up the lookup tables.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	colorMap[0] is white and colorMap[1] is black.
 *	divN, modN and magic are also set up.  They are used
 *      All of these are set in the DitherRec passed in.
 *
 * Parameters:
 *
 *	DitherRec *tempRef	A partially completed dither rec
 *				which will be completed by
 *				InitOrderedColor
 *--------------------------------------------------------------
 */
static int
InitMono( DitherRec *tempRef) {

    tempRef->colorMap[1] =
	    WhitePixel (tempRef->display, DefaultScreen(tempRef->display));

    tempRef->colorMap[0] =
	    BlackPixel (tempRef->display, DefaultScreen(tempRef->display));
 
    make_square( 255.0, tempRef->divN, tempRef->modN, tempRef->magic );

    return 1;
    
} /* end of InitMono */


/*
 *--------------------------------------------------------------
 *
 * ConvertColor --
 *
 *      Given a l, cr, cb tuple, converts it to r,g,b.
 *
 * Results:
 *      r,g,b values returned in pointers passed as parameters.
 *	Valid if color is "close" to being a valid color.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */
static int
ConvertColor(unsigned char l, unsigned char cr, unsigned char cb,
	     unsigned char *r, unsigned char *g, unsigned char *b)
{
  double fl, fcr, fcb, fr, fg, fb;
  int valid;

  switch (colorEquationSet) {
    case JACK_GAMMA_CORRECTED:
      fl = 1.164*(((double) l)-16.0);
      fcr =  ((double) cr) - 128.0;
      fcb =  ((double) cb) - 128.0;

      fr = fl + (1.596 * fcr);
      fg = fl - (0.813 * fcr) - (0.391 * fcb);
      fb = fl + (2.018 * fcb);
      break;
      
    case JACK_NOT_GAMMA_CORRECTED:
      fl = (double) l;
      fcr =  ((double) cr) - 128.0;
      fcb =  ((double) cb) - 128.0;

      fr = fl + (1.366 * fcr) - (0.002 * fcb);
      fg = fl - (0.700 * fcr) - (0.334 * fcb);
      fb = fl - (0.006 * fcr) + (1.732 * fcb);
      break;
  }

#define THRESH 32.0
  valid = ((fr >= -THRESH) && (fr <= THRESH+255.0) &&
	   (fg >= -THRESH) && (fg <= THRESH+255.0) &&
	   (fb >= -THRESH) && (fb <= THRESH+255.0));

  if (fr < 0.0) {
      fr = 0.0;
  } else if (fr > 255.0) {
      fr = 255.0;
  }

  if (fg < 0.0) {
      fg = 0.0;
  } else if (fg > 255.0) {
      fg = 255.0;
  }

  if (fb < 0.0) {
      fb = 0.0;
  } else if (fb > 255.0) {
      fb = 255.0;
  }

  *r = (unsigned char) fr;
  *g = (unsigned char) fg;
  *b = (unsigned char) fb;
  return valid;

} /* end of ConvertColor */


/*
 *--------------------------------------------------------------
 *
 * ConvertColorBack --
 *
 *      Given a r, g, b tuple, converts it to l,cr,cb.
 *
 * Results:
 *      l,cr,cb values returned in pointers passed as parameters.
 *	Valid if color is "close" to being a valid color. Should
 *      always be valid.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */
static int
ConvertColorBack(unsigned char r, unsigned char g, unsigned char b,
	     unsigned char *l, unsigned char *cr, unsigned char *cb)
{
  double fl, fcr, fcb, fr, fg, fb;
  int valid;

  switch (colorEquationSet) {
    case JACK_GAMMA_CORRECTED:
      fr = (double) r;
      fg = (double) g;
      fb = (double) b;

      fl =   (0.257 * fr) + (0.504 * fg) + (0.098 * fb);
      fcr =  (0.439 * fr) - (0.368 * fg) - (0.071 * fb) + 128.0;
      fcb =  (-0.148 * fr) - (0.291 * fg) + (0.439 * fb) + 128.0;
      break;
      
    case JACK_NOT_GAMMA_CORRECTED:
      fr = (double) r;
      fg = (double) g;
      fb = (double) b;

      fl = ((77.0/256.0) * fr) + ((150.0/256.0) * fg) + ((29.0/256.0) * fb);
      fcr = ((131.0/256.0) * fr) - ((110.0/256.0) * fg) - ((21.0/256.0) * fb) + 128.0;
      fcb = ((-44.0/256.0) * fr) - ((87.0/256.0) * fg) + ((131.0/256.0) * fb) + 128.0;
      break;
  }

  if (fl > 255.0) {
      fl = 255.0;
  } else if (fl < 0.0) {
      fl = 0.0;
  }

  if (fcr > 255.0) {
      fcr = 255.0;
  } else if (fcr < 0.0) {
      fcr = 0.0;
  }

  if (fcb > 255.0) {
      fcb = 255.0;
  } else if (fcb < 0.0) {
      fcb = 0.0;
  }

  *l = (unsigned char) fl;
  *cr = (unsigned char) fcr;
  *cb = (unsigned char) fcb;

  return 1;
}

/*
 * All code following this point is copyrighted as noted below.
 */

/*
 * This software is copyrighted as noted below.  It may be freely copied,
 * modified, and redistributed, provided that the copyright notice is 
 * preserved on all copies.
 * 
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is".  Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a program or other software product
 * without supplying the source, or without informing the end-user that the 
 * source is available for no extra charge.
 *
 * If you modify this software, you should include a notice giving the
 * name of the person performing the modification, the date of modification,
 * and the reason for such modification.
 *
 *  Modified at BRL 16-May-88 by Mike Muuss to avoid Alliant STDC desire
 *  to have all "void" functions so declared.
 */

/* 
 * dither.c - Functions for RGB color dithering.
 * 
 * Author:	Spencer W. Thomas
 * 		Computer Science Dept.
 * 		University of Utah
 * Date:	Mon Feb  2 1987
 * Copyright (c) 1987, University of Utah
 */

static int magic4x4[4][4] =  {
 	 0, 14,  3, 13,
	11,  5,  8,  6,
	12,  2, 15,  1,
	 7,  9,  4, 10
};

/*****************************************************************
 * TAG( make_square )
 * 
 * Build the magic square for a given number of levels.
 * Inputs:
 * 	N:		Pixel values per level (255.0 / levels).
 * Outputs:
 * 	divN:		Integer value of pixval / N
 *	modN:		Integer remainder between pixval and divN[pixval]*N
 *	magic:		Magic square for dithering to N sublevels.
 * Assumptions:
 * 	
 * Algorithm:
 *	divN[pixval] = (int)(pixval / N) maps pixval to its appropriate level.
 *	modN[pixval] = pixval - (int)(N * divN[pixval]) maps pixval to
 *	its sublevel, and is used in the dithering computation.
 *	The magic square is computed as the (modified) outer product of
 *	a 4x4 magic square with itself.
 *	magic[4*k + i][4*l + j] = (magic4x4[i][j] + magic4x4[k][l]/16.0)
 *	multiplied by an appropriate factor to get the correct dithering
 *	range.
 */
static void
make_square( double N, int divN[256], int modN[256], int magic[16][16] )
{
    register int i, j, k, l;
    double magicfact;

    for ( i = 0; i < 256; i++ )
    {
	divN[i] = (int)(i / N);
	modN[i] = i - (int)(N * divN[i]);
    }
    modN[255] = 0;		/* always */

    /*
     * Expand 4x4 dither pattern to 16x16.  4x4 leaves obvious patterning,
     * and doesn't give us full intensity range (only 17 sublevels).
     * 
     * magicfact is (N - 1)/16 so that we get numbers in the matrix from 0 to
     * N - 1: mod N gives numbers in 0 to N - 1, don't ever want all
     * pixels incremented to the next level (this is reserved for the
     * pixel value with mod N == 0 at the next level).
     */
    magicfact = (N - 1) / 16.;
    for ( i = 0; i < 4; i++ )
	for ( j = 0; j < 4; j++ )
	    for ( k = 0; k < 4; k++ )
		for ( l = 0; l < 4; l++ )
		    magic[4*k+i][4*l+j] =
			(int)(0.5 + magic4x4[i][j] * magicfact +
			      (magic4x4[k][l] / 16.) * magicfact);
}

/* basic dithering macro */
#define DMAP(v,x,y)	(modN[v]>magic[x][y] ? divN[v] + 1 : divN[v])

/*****************************************************************
 * TAG( dithergb )
 * 
 * Return dithered RGB value.
 * Inputs:
 * 	x:		X location on screen of this pixel.
 *	y:		Y location on screen of this pixel.
 *	r, g, b:	Color at this pixel (0 - 255 range).
 *	levels:		Number of levels in this map.
 *	divN, modN:	From dithermap.
 *	magic:		Magic square from dithermap.
 * Outputs:
 * 	Returns color map index for dithered pixelv value.
 * Assumptions:
 * 	divN, modN, magic were set up properly.
 * Algorithm:
 * 	see "Note:" in dithermap comment.
 */
static int
dithergb( int x, int y,
	     unsigned char r, unsigned char g, unsigned char b,
	     int levels, int divN[256], int modN[256], int magic[16][16] )
{
    int col = x % 16, row = y % 16;

    return DMAP(r, col, row) +
	DMAP(g, col, row) * levels +
	    DMAP(b, col, row) * levels*levels;
}


/*
 *--------------------------------------------------------------
 *
 * InitBrianDither --
 *
 *	This function initializes Brian's dithering method.  The
 *	basic idea is the same as Ketan's original FastOrderedColor,
 *	but the initialization is different.  The initialization
 *	works as follows:
 *
 *	Step one is to precompute an array (the crCbArray) indexed
 *	for a subset of the possible Cr/Cb values.  We chose 4 bits
 *	each for Cr/Cb (divide by 16), but this is settable.  Each
 *	entry in the crCbArray contains a pointer to what is
 *	conceptually a 256 x 16 table of pixel values (the table is
 *	called the "yArray").  The larger entry (256) is the Y
 *	value of the pixel, the smaller entry (16) is the position
 *	in the pixel block.  The actual storage structure is an
 *	array of pointers (indexed on pos and called posArray) to
 *	arrays of pixels (indexed on Y).
 *
 *	The basic idea is that the 16 pixels corresponding to
 *	constant Y in the yArray should make a reasonable
 *	approximation to the color.  Thus, this row is filled in
 *	using a standard dithering algorithm, which allows us to
 *	initially allocate the color map any way we want.  I chose
 *	to do a standard 6x6x6 color map initially.
 *
 *	Once the yArrays are computed, a larger table (indexed on 8
 *	bits of Cr/Cb) is built.  For each entry, we calculate the
 *	two nearest Cr (cr1 and cr2) and two nearest Cb (cb1 and
 *	cb2) values in the crCbArray and their corresponding
 *	"weight" (wcr and wcb).  Wcr is 4 if Cr = cr1, 2 if Cr is
 *	halfway between cr1 and cr2, and zero if it's nearest cr2.
 *	The method of choosing wcb is similar.  Each entry in the
 *	table is set to a pointer to an array of 16 values indexed
 *	on pos.  The first wcr*wcb entries are copied from the
 *	posArray entries for Cr1,Cb1, the next wcr*(4-wcb) entries
 *	entries are copied from the posArray entries for Cr1,Cb2,
 *	the next (4-wcr)*wcb entries entries are copied from the
 *	posArray entries for Cr2,Cb1, and the last (4-wcr)*(4-wcb)
 *	entries entries are copied from the posArray entries for
 *	Cr2,Cb2.
 *
 *	The net result is a structure that is a 256/256 array
 *	indexed by Cr/Cb.  Each entry of this array points to an
 *	array of 16 values indexed by pos, and each of these values
 *	points to 256 element vector of indexed by Y.
 *
 * Results:
 *	1 for success, 0 for failure
 *
 * Side effects:
 *	From now on, this dither will be used.
 *
 *--------------------------------------------------------------
 */

int
InitBrianDither (int minColors, int maxColors, DitherRec *tempRef)
{
    /*
     * This table gives the number of y/cr/cb levels for the
     * available number of colors.  For example, if 8 colors
     * are available, {ny,ncr,ncb} = {7,3,2}.
     */
    static numYCrCb [256][3] = {
     /*  ny   ncr  ncb */
	{ 4,  3,  2}, { 4,  3,  2}, { 4,  2,  3}, { 4,  2,  3}, 
	{ 4,  2,  3}, { 4,  2,  3}, { 4,  3,  3}, { 5,  3,  3}, 
	{ 5,  3,  3}, { 5,  3,  3}, { 5,  4,  3}, { 5,  3,  4}, 
	{ 5,  3,  4}, { 5,  3,  4}, { 5,  3,  4}, { 5,  3,  4}, 
	{ 5,  3,  4}, { 5,  3,  4}, { 5,  4,  4}, { 6,  4,  4}, 
	{ 6,  4,  4}, { 6,  4,  4}, { 6,  4,  4}, { 6,  5,  4}, 
	{ 6,  5,  4}, { 6,  5,  4}, { 6,  5,  4}, { 6,  5,  4}, 
	{ 6,  5,  4}, { 6,  4,  5}, { 6,  4,  5}, { 6,  4,  5}, 
	{ 6,  4,  5}, { 6,  4,  5}, { 7,  5,  5}, { 7,  5,  5}, 
	{ 6,  5,  5}, { 6,  5,  5}, { 6,  5,  5}, { 6,  5,  5}, 
	{ 6,  5,  5}, { 6,  5,  5}, { 6,  5,  5}, { 6,  5,  5}, 
	{ 7,  6,  5}, { 7,  6,  5}, { 7,  6,  5}, { 7,  5,  6}, 
	{ 7,  5,  6}, { 7,  5,  6}, { 7,  5,  6}, { 7,  5,  6}, 
	{ 7,  5,  6}, { 7,  5,  6}, { 7,  5,  6}, { 7,  5,  6}, 
	{ 7,  5,  6}, { 7,  5,  6}, { 7,  5,  6}, { 7,  5,  6}, 
	{ 7,  5,  6}, { 7,  5,  6}, { 7,  6,  6}, { 7,  6,  6}, 
	{ 7,  6,  6}, { 7,  6,  6}, { 7,  6,  6}, { 7,  6,  6}, 
	{ 7,  6,  6}, { 7,  6,  6}, { 7,  6,  6}, { 7,  6,  6}, 
	{ 7,  6,  6}, { 7,  6,  6}, { 7,  6,  6}, { 7,  6,  6}, 
	{ 7,  6,  6}, { 7,  6,  6}, { 7,  6,  6}, { 7,  6,  6}, 
	{ 7,  6,  6}, { 8,  6,  6}, { 8,  6,  6}, { 8,  6,  6}, 
	{ 8,  6,  6}, { 8,  6,  6}, { 8,  6,  6}, { 8,  6,  6}, 
	{ 8,  6,  7}, { 8,  6,  7}, { 8,  6,  7}, { 8,  6,  7}, 
	{ 8,  6,  7}, { 8,  6,  7}, { 8,  6,  7}, { 8,  6,  7}, 
	{ 8,  7,  6}, { 8,  7,  6}, { 8,  7,  6}, { 8,  7,  6}, 
	{ 8,  7,  6}, { 8,  7,  6}, { 8,  7,  6}, { 8,  7,  6}, 
	{ 8,  7,  6}, { 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, 
	{ 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, 
	{ 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, 
	{ 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, 
	{ 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, { 8,  7,  7}, 
	{ 8,  7,  7}, { 8,  7,  7}, { 9,  7,  7}, { 9,  7,  7}, 
	{ 9,  7,  7}, { 9,  7,  7}, { 9,  7,  7}, { 9,  7,  7}, 
	{ 9,  7,  7}, { 9,  7,  7}, { 9,  7,  7}, { 9,  7,  7}, 
	{ 9,  7,  7}, { 9,  7,  7}, { 9,  7,  7}, { 9,  7,  7}, 
	{ 9,  7,  7}, { 9,  7,  7}, { 9,  7,  7}, { 9,  8,  7}, 
	{ 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, 
	{ 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, 
	{ 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, 
	{ 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, 
	{ 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, { 9,  7,  8}, 
	{ 9,  7,  8}, { 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, 
	{ 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, 
	{ 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, 
	{ 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, 
	{ 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, 
	{ 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, { 9,  8,  8}, 
	{ 9,  8,  8}, {10,  8,  8}, {10,  8,  8}, {10,  8,  8}, 
	{10,  8,  8}, {10,  8,  8}, {10,  8,  8}, {10,  8,  8}, 
	{10,  8,  8}, {10,  8,  8}, {10,  8,  8}, {10,  8,  8}, 
	{10,  8,  8}, {10,  8,  8}, {10,  8,  8}, {10,  8,  8}, 
	{10,  8,  8}, {10,  8,  8}, {10,  8,  8}, {10,  8,  8}, 
	{10,  8,  8}, {10,  8,  8}, {10,  8,  8}, {10,  8,  8}, 
	{10,  8,  8}, {10,  8,  9}, {10,  8,  9}, {10,  8,  9}, 
	{10,  8,  9}, {10,  9,  8}, {10,  9,  8}, {10,  9,  8}, 
	{10,  9,  8}, {10,  9,  8}, {10,  9,  8}, {10,  9,  8}, 
	{10,  9,  8}, {10,  9,  8}, {10,  9,  8}, {10,  9,  8}, 
	{10,  9,  8}, {10,  9,  8}, {10,  9,  8}, {10,  9,  8}, 
	{10,  9,  8}, {10,  9,  8}, {10,  9,  8}, {10,  9,  8}, 
	{10,  9,  8}, {10,  9,  8}, {10,  9,  8}, {10,  9,  8}, 
	{10,  9,  8}, {10,  9,  8}, {10,  9,  9}, {10,  9,  9}, 
	{10,  9,  9}, {10,  9,  9}, {11,  9,  9}, {11,  9,  9}, 
	{11,  9,  9}, {11,  9,  9}, {11,  9,  9}, {11,  9,  9}, 
	{11,  9,  9}, {11,  9,  9}, {11,  9,  9}, {11,  9,  9}, 
	
	};
	

    uchar ***crCbArray1;
    uchar ***crCbArray2;
    uchar **posArray1;
    uchar **posArray2;
    uchar *yArray;
    short *nearestColor;
    int y, cr, cb;
    int y1, y2, cr1, cr2, cb1, cb2;
    int wcr, wcb, v1, v2;
    int count, i, j, k, x;

    /* These variables are used for color allocation... */
    uchar r, g, b;
    unsigned long planeMasks[256];
    Colormap dcmap;
    XColor colorsWanted[256];
    int ny, ncr, ncb;
    int numReq, numFound;
    double dblNumCbLevels;
    int colorMap[256];
    int invMap[256];
    BrianDitherRec *ditherRec;

    /*
     * Determine how many colors are available
     */
    dcmap = XDefaultColormap(tempRef->display, DefaultScreen(tempRef->display));
    for (numReq = maxColors; numReq >= minColors; numReq--) {
	if (XAllocColorCells(tempRef->display, dcmap, 0, planeMasks, 0,
			     tempRef->pixels, numReq)) {
	    break;
	}
    }
    if (numReq < minColors) {
        return 0;
    }

    /*
     * Step 1: Split them into a reasonable color cube.  This is a cube
     * of size ny/ncr/ncb, but only the valid colors are set.  The
     * rest of the colors are set to the nearest available color that's
     * valid.
     */
    ny = numYCrCb[numReq][0];
    ncr = numYCrCb[numReq][1];
    ncb = numYCrCb[numReq][2];
    numFound = 0;
    nearestColor = (short *)malloc(ny*ncr*ncb*sizeof(short));
    for (i=0; i<ncr; i++) {
	cr = 255*i/(ncr-1);
	for (j=0; j<ncb; j++) {
	    cb = 255*j/(ncb-1);
	    for (k = 0; k < ny; k++) {
		y = 255*k/(ny-1);
		if (ConvertColor((uchar)y, (uchar)cr, (uchar)cb, &r, &g, &b)) {
		    colorsWanted[numFound].red = r << 8;
		    colorsWanted[numFound].green = g << 8;
		    colorsWanted[numFound].blue = b << 8;
		    colorsWanted[numFound].flags = DoRed | DoGreen | DoBlue;
		    nearestColor[k + j*ny + i*ny*ncb] = numFound;
		    numFound++;
		} else {
		    nearestColor[k + j*ny + i*ny*ncb] = -1;
		}
	    }
	}
    }

    /*
     * Make a second pass through the map, mapping the invalid colors
     * to the closest valid color.  This could be faster.
     */
    for (i=0; i<ncr; i++) {
	cr = 255*i/(ncr-1);
	for (j=0; j<ncb; j++) {
	    cb = 255*j/(ncb-1);
	    for (k = 0; k < ny; k++) {
		int d, dr, dg, db;
		int best, bestD;
		y = 255*k/(ny-1);
		if (nearestColor[k + j*ny + i*ny*ncb] == -1) {
		    ConvertColor((uchar)y, (uchar)cr, (uchar)cb, &r, &g, &b);
		    bestD = (1<<30);
		    for (x=0; x<numFound; x++) {
			dr = (colorsWanted[x].red >> 8) - r;
			dg = (colorsWanted[x].green >> 8) - g;
			db = (colorsWanted[x].blue >> 8) - b;
			dr = abs(dr);
			dg = abs(dg);
			db = abs(db);
			d = dr + dg + db;
			if (d < bestD) {
			    bestD = d;
			    best = x;
			}
		    }
		    nearestColor[k + j*ny + i*ny*ncb] = best;
		}
	    }
	}
    }

    /*
     * Build a map with cr/cb axes.  The table size is
     * ncr x ncb, each element of which points to a 16x256 array indexed
     * on (pos,y), where pos = [0..15] and y = [0..255]
     */
    crCbArray1 = (uchar ***)malloc(ncr*ncb*sizeof(uchar **));
    for (i=0; i<ncr; i++) {
	cr = 255*i/(ncr-1);
	for (j=0; j<ncb; j++) {
	    uchar foundMap[256];
	    int valid, dr, dg, db;
	    cb = 255*j/(ncb-1);
	    for (k = 0; k < ny; k++) {
		foundMap[k] = nearestColor[k + j*ny + i*ny*ncb];
	    }

	    /*
	     * Fill in the y/cr/cb base array for this cr/cb combo
	     */
	    posArray1 = (uchar **)malloc(16*sizeof(uchar *));
	    crCbArray1[ncb*i+j] = posArray1;
	    for (k=0; k<16; k++) {
		posArray1[k] = (uchar *)malloc(256);
	    }
	    for (k = 0; k < (ny-1); k++) {
		v1 = foundMap[k];
		v2 = foundMap[k+1];
		y1 = 255*k/(ny-1);
		y2 = 255*(k+1)/(ny-1);
		for (y=y1; y<=y2; y++) {
		    int n1 = 16 - 16*(y-y1)/(y2-y1);
		    for (x=0; x<n1; x++) {
			posArray1[x][y] = v1;
		    }
		    for (x=n1; x<16; x++) {
			posArray1[x][y] = v2;
		    }
		}
	    }
	}
    }

    /*
     * Store the colors we got, free the ones we're not using,
     * and fix map color indices above to pixel values.
     */
    if (numFound < numReq) {
	XFreeColors(tempRef->display, dcmap, tempRef->pixels+numFound,
		    numReq-numFound, 0);
    }
    for (i = 0; i < numFound; i++) {
	int pix;
	pix = colorsWanted[i].pixel = tempRef->pixels[i];
	colorMap[pix] = i;
	invMap[i] = pix;
	currColors[pix].pixel = pix;
	currColors[pix].red = colorsWanted[i].red;
	currColors[pix].green = colorsWanted[i].green;
	currColors[pix].blue = colorsWanted[i].blue;
	XStoreColor(tempRef->display, dcmap, colorsWanted+i);

    }
    for (i=0; i<ncr; i++) {
	for (j=0; j<ncb; j++) {
	    posArray1 = crCbArray1[ncb*i+j];
	    for (x=0; x<16; x++) {
		for (y=0; y<256; y++) {
		    posArray1[x][y] = invMap[posArray1[x][y]];
		}
	    }
	}
    }

    /*
     * Step 2: Build a table dithered on Cr/Cb.  This is built out
     * of pointers to entries in posArray1 allocated above.
     * In general, there will be 4 relevant crCbArray1 entries
     * "surrounding" the cr/cb value, since the cr/cb value lies
     * off the grid.  We compute wcr/wcb as the "weight" of the
     * one above and to the left of the entry we're trying to
     * dither.
     */
    crCbArray2 = (uchar ***)malloc(256*256*sizeof(uchar **));
    for (i=0; i<(ncr-1); i++) {
	cr1 = 255*i/(ncr-1);
	cr2 = 255*(i+1)/(ncr-1);
	for (cr=cr1; cr<=cr2; cr++) {
	    wcr = 5 - 5*(cr-cr1)/(cr2-cr1);
	    if (wcr > 4) {
		wcr = 4;
	    }
	    for (j=0; j<(ncb-1); j++) {
		cb1 = 255*j/(ncb-1);
		cb2 = 255*(j+1)/(ncb-1);
		for (cb=cb1; cb<=cb2; cb++) {
		    wcb = 5 - 5*(cb-cb1)/(cb2-cb1);
		    if (wcb > 4) {
			wcb = 4;
		    }
		    posArray2 = (uchar **)malloc(16*sizeof(uchar *));
		    crCbArray2[(cr<<8) + cb] = posArray2;
		    count = 0;
		    posArray1 = crCbArray1[ncb*i+j];
		    for (k=0; k<wcr*wcb; k++) {
			posArray2[count] = posArray1[count];
			count++;
		    }
		    posArray1 = crCbArray1[ncb*i+j+1];
		    for (k=0; k<wcr*(4-wcb); k++) {
			posArray2[count] = posArray1[count];
			count++;
		    }
		    posArray1 = crCbArray1[ncb*(i+1)+j];
		    for (k=0; k<(4-wcr)*wcb; k++) {
			posArray2[count] = posArray1[count];
			count++;
		    }
		    posArray1 = crCbArray1[ncb*(i+1)+(j+1)];
		    for (k=0; k<(4-wcr)*(4-wcb); k++) {
			posArray2[count] = posArray1[count];
			count++;
		    }
		}
	    }
	}
    }


    /*
     * Free up the temporary variables...
     */
    free(nearestColor);

    /*
     * Construct the ordered dither record and return
     */
    ditherRec = (BrianDitherRec *)malloc(sizeof(BrianDitherRec));
    ditherRec->crCbArray1 = crCbArray1;
    ditherRec->crCbArray2 = crCbArray2;
    ditherRec->ncr = ncr;
    ditherRec->ncb = ncb;
    ditherRec->ny = ny;
    tempRef->numPixels = numFound;
    tempRef->ditherData = (void *) ditherRec;

    return 1;
}


/*
 *--------------------------------------------------------------
 *
 * InitKetanDither --
 *
 *    This function initializes Ketan's dithering method. 
 *    This dither is essentially a modification of Brian's dither
 *    but it uses much less space (about 25K) and can reuse colors
 *    allocated by other applications. In addition, the Y, Cr, Cb
 *    space is always cut into 16 y levels,  and 8 cr and cb levels
 *    regardless of the number of colors available.
 *    One drawback is that it will probably take more time to initialize.
 *
 *    The basic idea is that 8 bits of Cr and Cb  data is used to find
 *    a 4x4 dither pattern of the closest Cr/Cb combinations available where
 *    there are 8 distinct possibilities for Cr and Cb. This is done by using
 *    a separate lookup table for cr and cb (i.e., the crIndex and the cbIndex)
 *    to map between cr and cb data values to 4 cr and cb index values. Since
 *    each index value is between 0 and 7 and only requires three bits, the
 *    cr indexes are packed into bits 0-2 in 4 successive bytes that are treated
 *    as an integer. The cb values are similar but are packed into bits 3-5.
 *    The two integers are added resulting in an integer whose bytes consist of
 *    the correctly dithered cr and cb index values for a quadrant of the 4x4
 *    dither pattern. There are 4 sets of these tables corresponding to each
 *    quadrant.
 * 
 *    All 8 bits of the Y data is used to map to the y index preshifted by 6 bits
 *    to the left. There are 16 tables used to do this lookup depending on the
 *    position of the pixel in the 4x4 dither matrix.
 *
 *    The proper 6 bit cr/cb index from the 4 quadrant values already calculated
 *    is shifted into the correct bits (i.e., 0-7) and then combined with the 
 *    preshifted y index to create a 10 bit cr/cb/y index which is then used to 
 *    find that actual pixel value that maps to this y/cr/cb combination.
 *
 *    This initialization routine does the following:
 *
 *    -- Allocates colors depending on what colors are alread allocated by other
 *       applications and which colors are needed.
 *    -- Builds the 4 cr and 4 cb index tables.
 *    -- Builds the y index table.
 *    -- Builds the cr/cb/y to pixel value table.
 *
 * Results:
 *    1 on success, 0 on failure.
 *
 *------------------------------------------------------------------------------*/

typedef struct {
  int is_valid;
  int pixel_val;
  long int distance;
  unsigned char y;
  unsigned char cr;
  unsigned char cb;
  unsigned char r;
  unsigned char g;
  unsigned char b;
} KetanColor;

#define NUM_Y_VALS 15
#define NUM_CR_VALS 7
#define NUM_CB_VALS 7
#ifndef _WINDOWS
#  define MAX_COLORS_USED 256
#else
#  define MAX_COLORS_USED 200
#endif


static int
InitKetanDither (DitherRec *tempRef)
{
  unsigned char yvals[NUM_Y_VALS];
  unsigned char crvals[NUM_CR_VALS];
  unsigned char cbvals[NUM_CB_VALS];

  unsigned char *yCrCbToPixel;
  unsigned int **crTables;
  unsigned int **cbTables;
  unsigned int cr_pos_val[16];
  unsigned int cb_pos_val[16];

  unsigned int **yIndexTables;
  KetanDitherRec *ditherRec;

  Colormap dcmap;
  XColor *xcolor_array, txcolor;
  int num_in_palette;

  unsigned long planeMasks[256];

  int max_colors;
  int i, yidx, cridx, cbidx, idx;
  int vyidx, vcridx, vcbidx, vidx;
  int cr_val, cb_val, cr_val_idx, cb_val_idx;
  int cr1, cr2, cb1, cb2, wcr, wcb;
  int y_val, y_val_idx, y1, y2, wy;
  int pos_idx, dpos, pos;
  int firstTime;
  int xc_red, xc_green, xc_blue;
  long int distance, max_dist;
  int max_dist_idx, pix_num;
  int min_dist, min_dist_idx;
  int num_combos, num_valid, numAvail;
  KetanColor *color_info_array, *cptr, **valid_colors;
  int num_allocated, num_reused;
  unsigned char *crCbPosPtr;
  int closest_valid_color;
  int cidx;

  /* Step One: Allocate colors. 
   *
   *    We start by cutting y value range into 16 possibilities and the cr and cb
   *    ranges into 8 possibilities.
   *
   *    Of these 1K possible combinations, we create two arrays: 
   *       An array of all color combinations. 
   *       An array of pointers to all valid color combinations.
   *
   *    We map the invalid color combinations to valid color combinations by converting
   *    from y/cr/cb to r/g/b, capping the resulting components to the range 0-255,
   *    converting back to y/cr/cb and then calculating the y/cr/cb indexes of the
   *    closest valid color. 
   *
   *    We then construct a list of colors that have been allocated by other applications
   *    as well as finding out which pixel values we are able to allocate ourselves.
   *
   *    For each pre-allocated color, we calculate the distance to each valid 
   *    y/cr/cb combo and for each of these combonations store which color is closest
   *    and its associated distance.
   *
   *    For each color we have free to allocate, we choose the y/cr/cb combo that is
   *    furthest from any pre-allocated color and allocate a color to match the combo.
   *    We then recalculate the distance between all other valid combos and the newly
   *    allocated color possibly associating other combos to this new color if its
   *    distance is smaller than the previous closest color.
   *    This continues until all colors are allocated.
   *
   *    Now we construct a pixel value list of all the colors used for the corresponding
   *    field in the tempRef structure passed in. This is required to free these colors
   *    later on. 
   * 
   *    Finally, we build the y/cr/cb to pixel value table and free any temporary
   *    memory.
   *
   *    The procedure for allocating colors for Windows is sufficiently different
   *    that I've blocked out the code based on the macro _WINDOWS even though
   *    much of what it does is effectively the same.
   *
   */

  for (i=0; i<NUM_Y_VALS; i++) {
    yvals[i] = (unsigned char) (((((double) i) * 256.0) / ((double) NUM_Y_VALS)) + 
		      (128.0 / ((double) NUM_Y_VALS)));
  }

  for (i=0; i<NUM_CR_VALS; i++) {
    crvals[i] = (unsigned char) (((((double) i) * 256.0) / ((double) NUM_CR_VALS)) + 
		      (128.0 / ((double) NUM_CR_VALS)));
  }

  for (i=0; i<NUM_CB_VALS; i++) {
    cbvals[i] = (unsigned char) (((((double) i) * 256.0) / ((double) NUM_CB_VALS)) + 
		      (128.0 / ((double) NUM_CB_VALS)));
  }

  num_combos = NUM_Y_VALS * NUM_CR_VALS * NUM_CB_VALS;

  color_info_array = (KetanColor *) malloc(NUM_Y_VALS*NUM_CR_VALS*NUM_CB_VALS*
					   sizeof(KetanColor));

  num_valid = 0;

  for (yidx=0; yidx<NUM_Y_VALS; yidx++) {
    for (cridx=0; cridx<NUM_CR_VALS; cridx++) {
      for (cbidx=0; cbidx<NUM_CB_VALS; cbidx++) {
	idx = (yidx*NUM_CR_VALS*NUM_CB_VALS)+(cridx*NUM_CB_VALS)+cbidx;
	cptr = &(color_info_array[idx]);
	cptr->y = yvals[yidx];
	cptr->cr = crvals[cridx];
	cptr->cb = cbvals[cbidx];
	
	cptr->is_valid = ConvertColor(cptr->y, cptr->cr, cptr->cb, 
				      &(cptr->r), &(cptr->g), &(cptr->b));
	if (cptr->is_valid) {
	  num_valid++;
	} 
      }
    }
  }

  if (num_valid < MAX_COLORS_USED) {
      max_colors = num_valid;
  } else {
      max_colors = MAX_COLORS_USED;
  }


  valid_colors = (KetanColor **) malloc(sizeof(KetanColor)*num_valid);
  vidx = 0;
  for (idx=0; idx<num_combos; idx++) {
    if (color_info_array[idx].is_valid) {
      valid_colors[vidx++] = &(color_info_array[idx]);
    }
  }

  assert(vidx == num_valid);

  dcmap = XDefaultColormap(tempRef->display, DefaultScreen(tempRef->display));

#ifndef _WINDOWS

  for (numAvail=256; numAvail>0; numAvail--) {
    if (XAllocColorCells(tempRef->display, dcmap, 0, planeMasks, 0,
			 tempRef->pixels, numAvail)) {
      break;
    }
  }
  xcolor_array = (XColor *) malloc(sizeof(XColor)*256);
  
  for (i=0; i<256; i++) {
    xcolor_array[i].pixel = -1;
  }

  for (i=0; i<numAvail; i++) {
    if ((tempRef->pixels[i] < 256) && (tempRef->pixels[i] >= 0)) {
      xcolor_array[tempRef->pixels[i]].pixel = tempRef->pixels[i];
    }
  }

  firstTime = 1;

  num_reused = 0;
  for (i=0; i<256; i++) {
    if (xcolor_array[i].pixel == -1) {
      xcolor_array[i].pixel = i;
      xcolor_array[i].flags = DoRed | DoGreen | DoBlue;
      
      XQueryColor(tempRef->display, dcmap, &(xcolor_array[i]));
      XAllocColor(tempRef->display, dcmap, &(xcolor_array[i]));

      if (xcolor_array[i].pixel != i) 
	continue;

      num_reused++;

      xc_red = xcolor_array[i].red >> 8;
      xc_green = xcolor_array[i].green >> 8;
      xc_blue = xcolor_array[i].blue >> 8;

      for (vidx=0; vidx<num_valid; vidx++) {
	  distance =  (((xc_red - valid_colors[vidx]->r) *
			(xc_red - valid_colors[vidx]->r)) +
		       ((xc_green - valid_colors[vidx]->g) *
			(xc_green - valid_colors[vidx]->g)) +
		       ((xc_blue - valid_colors[vidx]->b) *
			(xc_blue - valid_colors[vidx]->b)));

	if ((firstTime) || (distance < valid_colors[vidx]->distance)) {
	  valid_colors[vidx]->distance = distance;
	  valid_colors[vidx]->pixel_val = xcolor_array[i].pixel;
	}
      }
      firstTime = 0;
    }
  }

  for (i=0; i<numAvail; i++) {
      xcolor_array[tempRef->pixels[i]].pixel = -1;
  }
  XFreeColors(tempRef->display, dcmap, tempRef->pixels, numAvail, 0);

  max_dist = -1;
  max_dist_idx = 0;

  for (vidx=0; vidx<num_valid; vidx++) {
    if (valid_colors[vidx]->distance > max_dist) {
      max_dist = valid_colors[vidx]->distance;
      max_dist_idx = vidx;
    }
  }

  num_allocated = 0;
  while (num_allocated+num_reused < max_colors) {

    txcolor.red = ((unsigned short) valid_colors[max_dist_idx]->r) << 8;
    txcolor.green = ((unsigned short) valid_colors[max_dist_idx]->g) << 8;
    txcolor.blue = ((unsigned short) valid_colors[max_dist_idx]->b) << 8;
    txcolor.flags = DoRed | DoGreen | DoBlue;

    if (!XAllocColor(tempRef->display, dcmap, &txcolor)) {
	break;
    }

    pix_num = txcolor.pixel;
    num_allocated++;
    if (xcolor_array[pix_num].pixel == -1) {
	xcolor_array[pix_num].pixel = pix_num;
	xcolor_array[pix_num].red = txcolor.red;
	xcolor_array[pix_num].green = txcolor.green;
	xcolor_array[pix_num].blue = txcolor.blue;
    }

    max_dist = -1;
    max_dist_idx = -1;

    xc_red = txcolor.red >> 8;
    xc_green = txcolor.green >> 8;
    xc_blue = txcolor.blue >> 8;

    for (vidx=0; vidx<num_valid; vidx++) {
	distance = (((xc_red - valid_colors[vidx]->r) *
		     (xc_red - valid_colors[vidx]->r)) +
		    ((xc_green - valid_colors[vidx]->g) *
		     (xc_green - valid_colors[vidx]->g)) +
		    ((xc_blue - valid_colors[vidx]->b) *
		     (xc_blue - valid_colors[vidx]->b)));

      if ((firstTime) || (distance < valid_colors[vidx]->distance)) {
	valid_colors[vidx]->distance = distance;
	valid_colors[vidx]->pixel_val = txcolor.pixel;
      }
      if (valid_colors[vidx]->distance > max_dist) {
	max_dist = valid_colors[vidx]->distance;
	max_dist_idx = vidx;
      }
    }
    firstTime = 0;
  }

  cidx = 0;
  for (i=0; i<256; i++) {
      if (xcolor_array[i].pixel != -1) {
	  tempRef->pixels[cidx++] = xcolor_array[i].pixel;
      }
  }
  free((char *) xcolor_array);

  tempRef->numPixels = cidx;

#else /* ifndef _WINDOWS */

  if (max_colors > 256) {
      return 0;
  }

  /* Deal with colors already in the palette */

  xcolor_array = (XColor *) malloc(sizeof(XColor)*256);

  num_in_palette = XQueryColorPalette(tempRef->display, dcmap, xcolor_array, 256);

  for (vidx=0; vidx<num_valid; vidx++) {
    valid_colors[vidx]->distance = 300 * 300 * 3;
  }

  num_reused = 0;
  for (i=0; i<num_in_palette; i++) {
      if ((xcolor_array[i].flags == 0x01) || (xcolor_array[i].flags == 0x04)) {

	  XAllocColor(tempRef->display, dcmap, &(xcolor_array[i]));
	  num_reused++;

	  xc_red = xcolor_array[i].red >> 8;
	  xc_green = xcolor_array[i].green >> 8;
	  xc_blue = xcolor_array[i].blue >> 8;
	  
	  for (vidx=0; vidx<num_valid; vidx++) {
	      distance =  (((xc_red - valid_colors[vidx]->r) *
			    (xc_red - valid_colors[vidx]->r)) +
			   ((xc_green - valid_colors[vidx]->g) *
			  (xc_green - valid_colors[vidx]->g)) +
			   ((xc_blue - valid_colors[vidx]->b) *
			    (xc_blue - valid_colors[vidx]->b)));
	      if (distance < valid_colors[vidx]->distance) {
		  valid_colors[vidx]->distance = distance;
		  valid_colors[vidx]->pixel_val = xcolor_array[i].pixel;
	      }
	  }
      } else {
	  xcolor_array[i].pixel = -1;
      }
  }

  max_dist = -1;
  max_dist_idx = 0;

  for (vidx=0; vidx<num_valid; vidx++) {
      if (valid_colors[vidx]->distance > max_dist) {
	  max_dist = valid_colors[vidx]->distance;
	  max_dist_idx = vidx;
      }
  }

  num_allocated = 0;
  while (num_allocated +  num_reused < max_colors) {
      txcolor.red = valid_colors[max_dist_idx]->r << 8; 
      txcolor.green = valid_colors[max_dist_idx]->g << 8;
      txcolor.blue = valid_colors[max_dist_idx]->b << 8;
      txcolor.flags = DoRed | DoGreen | DoBlue;

      if (!XAllocColor(tempRef->display, dcmap, &(txcolor))) {
	  break;
      }

      num_allocated++;
      if (xcolor_array[txcolor.pixel].pixel == -1) {
	  xcolor_array[txcolor.pixel].pixel = txcolor.pixel;
      }

      max_dist = -1;
      max_dist_idx = -1;

      xc_red = txcolor.red >> 8;
      xc_green = txcolor.green >> 8;
      xc_blue = txcolor.blue >> 8;

      for (vidx=0; vidx<num_valid; vidx++) {
	  distance = (((xc_red - valid_colors[vidx]->r) *
		       (xc_red - valid_colors[vidx]->r)) +
		      ((xc_green - valid_colors[vidx]->g) *
		       (xc_green - valid_colors[vidx]->g)) +
		      ((xc_blue - valid_colors[vidx]->b) *
		       (xc_blue - valid_colors[vidx]->b)));
	  
	  if (distance < valid_colors[vidx]->distance) {
	      valid_colors[vidx]->distance = distance;
	      valid_colors[vidx]->pixel_val = txcolor.pixel;
	  }
	  if (valid_colors[vidx]->distance > max_dist) {
	      max_dist = valid_colors[vidx]->distance;
	      max_dist_idx = vidx;
	  }
      }
  }

  cidx = 0;
  for (i=0; i<256; i++) {
      if (xcolor_array[i].pixel != -1) {
	  tempRef->pixels[cidx++] = xcolor_array[i].pixel;
      }
  }
  tempRef->numPixels = cidx;

  free((char *) xcolor_array);

#endif /* else of ifndef _WINDOWS */

  /* Deal with invalid combinations. */

  for (i=0; i<num_combos; i++) {
    if (!color_info_array[i].is_valid) {
      min_dist = 256*256*4;
      min_dist_idx = 0;

      for (vidx=0; vidx<num_valid; vidx++) {
	  distance = (((valid_colors[vidx]->r - color_info_array[i].r) *
		       (valid_colors[vidx]->r - color_info_array[i].r)) +
		      ((valid_colors[vidx]->g - color_info_array[i].g) *
		       (valid_colors[vidx]->g - color_info_array[i].g)) +
		      ((valid_colors[vidx]->b - color_info_array[i].b) *
		       (valid_colors[vidx]->b - color_info_array[i].b)));
	
	if (distance < min_dist) {
	  min_dist = distance;
	  min_dist_idx = vidx;
	}
      }

      color_info_array[i].pixel_val = valid_colors[min_dist_idx]->pixel_val;
    }
  }

  yCrCbToPixel = (unsigned char *) malloc(num_combos);

  for (yidx=0; yidx<NUM_Y_VALS; yidx++) {
    for (cridx=0; cridx<NUM_CR_VALS; cridx++) {
      for (cbidx=0; cbidx<NUM_CB_VALS; cbidx++) {

	idx = ((yidx*NUM_CR_VALS*NUM_CB_VALS)+
	       (cridx*NUM_CB_VALS)+
	       cbidx);
	
	yCrCbToPixel[idx] = color_info_array[idx].pixel_val;
      }
    }
  }  

  free((char*) color_info_array);
  free((char *) valid_colors);

  /* Step Two: Build the cr and cb index tables.
   *
   *   These tables are indexed by 8 bits of either cr or cb data. For each
   *   entry, we calculate the two nearest Cr (cr1 and cr2) and two nearest Cb
   *   (cb1 and cb2) indexes and their corresponding "weight" (wcr and wcb).
   *   Wcr is 16 if Cr = cr1, 8 if Cr is halfway between cr1 and cr2, and 0 if it's
   *   nearest cr2. The method of choosing wcb is similar. 
   *   Each table represents a quadrant of the dither (i.e., 4 particular positions).
   *   The first wcr position entries are set to the indexes for cr1. 
   *   The next (16-wcr) entries are set to cr2. Similar methods are used for cb.
   *
   */

  crTables = (unsigned int **) malloc(sizeof(unsigned int *) * 4);
  cbTables = (unsigned int **) malloc(sizeof(unsigned int *) * 4);

  for (i=0; i<4; i++) {
    crTables[i] = (unsigned int *) malloc(sizeof(unsigned int) * 256);
    cbTables[i] = (unsigned int *) malloc(sizeof(unsigned int) * 256);
  }
  
  for (cr_val=0; cr_val<256; cr_val++) {

    cr1 = 0;
    while (cr_val > crvals[cr1]) {
      cr1++;
      if (cr1 >= NUM_CR_VALS) 
	break;
    }
    cr1--;
    cr2 = cr1+1;
    if (cr1 < 0) {
      cr1 = 0;
    }
    if (cr2 >= NUM_CR_VALS) {
      cr2 = NUM_CR_VALS-1;
    }

    if (cr1 == cr2) {
      wcr = 16;
    } else {
      wcr = 16 - (16*(cr_val - crvals[cr1]))/(crvals[cr2]-crvals[cr1]);
    }

    for (i=0; i<wcr; i++) {
      cr_pos_val[i] = cr1 * NUM_CB_VALS;
    }
    for (i=wcr; i<16; i++) {
      cr_pos_val[i] = cr2 * NUM_CB_VALS;
    }

    (crTables[0])[cr_val] = ((cr_pos_val[0] << 24) | 
			     (cr_pos_val[12] << 16) | 
			     (cr_pos_val[8] << 8) | 
			     cr_pos_val[4]);

    (crTables[1])[cr_val] = ((cr_pos_val[3] << 24) | 
			     (cr_pos_val[15] << 16) | 
			     (cr_pos_val[11] << 8) | 
			     cr_pos_val[7]);

    (crTables[2])[cr_val] = ((cr_pos_val[2] << 24) | 
			     (cr_pos_val[14] << 16) | 
			     (cr_pos_val[10] << 8) | 
			     cr_pos_val[6]);

    (crTables[3])[cr_val] = ((cr_pos_val[1] << 24) | 
			     (cr_pos_val[13] << 16) | 
			     (cr_pos_val[9] << 8) | 
			     cr_pos_val[5]);
  }

  for (cb_val=0; cb_val<256; cb_val++) {

    cb1 = 0;
    while (cb_val > cbvals[cb1]) {
      cb1++;
      if (cb1 >= NUM_CB_VALS) 
	break;
    }
    cb1--;
    cb2 = cb1+1;
    if (cb1 < 0) {
      cb1 = 0;
    }
    if (cb2 >= NUM_CB_VALS) {
      cb2 = NUM_CB_VALS-1;
    }

    if (cb1 == cb2) {
      wcb = 16;
    } else {
      wcb = 16 - (16*(cb_val - cbvals[cb1]))/(cbvals[cb2]-cbvals[cb1]);
    }

    for (i=0; i<wcb; i++) {
      cb_pos_val[i] = cb1;
    }
    for (i=wcb; i<16; i++) {
      cb_pos_val[i] = cb2;
    }

    (cbTables[0])[cb_val] = ((cb_pos_val[15] << 24) | 
			     (cb_pos_val[7] << 16) | 
			     (cb_pos_val[3] << 8) | 
			     cb_pos_val[11]);

    (cbTables[1])[cb_val] = ((cb_pos_val[13] << 24) | 
			     (cb_pos_val[5] << 16) | 
			     (cb_pos_val[1] << 8) | 
			     cb_pos_val[9]);

    (cbTables[2])[cb_val] = ((cb_pos_val[12] << 24) | 
			     (cb_pos_val[4] << 16) | 
			     (cb_pos_val[0] << 8) | 
			     cb_pos_val[8]);

    (cbTables[3])[cb_val] = ((cb_pos_val[14] << 24) | 
			     (cb_pos_val[6] << 16) | 
			     (cb_pos_val[2] << 8) | 
			     cb_pos_val[10]);
  }


  /* Step Three: Build the y index tables.
   *
   *   These tables map from y value to one of the 16 y indexes. Each table is
   *   calculated for a different dither position. The tables are indexed on 
   *   8 bits of y data and produce a 16 bit value with bits 6-9 set to the 
   *   corresponding y index. 
   */

  yIndexTables = (unsigned int **) malloc(sizeof(int *) * 16);
  for (dpos=0; dpos<16; dpos++) {
    yIndexTables[dpos] = (int *) malloc(sizeof(int)*256);
  }

  for (y_val=0; y_val<256; y_val++) {
    
    y1 = 0;
    while (y_val > yvals[y1]) {
      y1++;
      if (y1 >= NUM_Y_VALS)
	break;
    }
    y1--;
    y2 = y1+1;
    if (y1 < 0) {
      y1 = 0;
    } 
    if (y2 >= NUM_Y_VALS) {
      y2 = NUM_Y_VALS-1;
    }

    y_val_idx = (y_val * NUM_Y_VALS) / 256;
    
    if (y1 == y2) {
      wy = 16;
    }
    else {
      wy = 16 - (16*(y_val - yvals[y1]))/(yvals[y2]-yvals[y1]);
    }

    for (dpos=0; dpos<wy; dpos++) {
      (yIndexTables[dpos])[y_val] = (y1 * NUM_CR_VALS * NUM_CB_VALS);
    }
    for (dpos=wy; dpos<16; dpos++) {
      (yIndexTables[dpos])[y_val] = (y2 * NUM_CR_VALS * NUM_CB_VALS);
    }
  }
    
  /* Step Four: Cleanup
   *
   *   A dither specific data structure is constructed to point to the newly
   *   built tables. The various fields in the the tempRef dither structure
   *   are properly filled and the value 1 is returned for success.
   */

  ditherRec = (KetanDitherRec *) malloc(sizeof(KetanDitherRec));
  ditherRec->yCrCbToPixel = yCrCbToPixel;
  ditherRec->crTables = crTables;
  ditherRec->cbTables = cbTables;
  ditherRec->yIndexTables = yIndexTables;
  tempRef->ditherData = (void *) ditherRec;

  return 1;
}


/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb411MSBKetan8 --
 *
 *	Color dithers an incoming YCrCb411 formatted picture block 
 *	to an 8-bit MSB display using Ketan's ordered color method.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane
 *	unsigned char *cr      	the cr bit-plane
 *	unsigned char *cb   	the cb bit-plane
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane
 *	int cbStride		stride for cb-plane
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
DitherErr 
DitherYCrCb411MSBKetan8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{
    KetanDitherRec *ditherRec;
    unsigned char *yCrCbToPixel;
    unsigned int *yIndex0, *yIndex1, *yIndex2, *yIndex3;
    unsigned int *yIndex4, *yIndex5, *yIndex6, *yIndex7;
    unsigned int *yIndex8, *yIndex9, *yIndex10, *yIndex11;
    unsigned int *yIndex12, *yIndex13, *yIndex14, *yIndex15;
    unsigned int *crIndexQ1, *crIndexQ2, *crIndexQ3, *crIndexQ4;
    unsigned int *cbIndexQ1, *cbIndexQ2, *cbIndexQ3, *cbIndexQ4;
    Boolean wDithQuad, hDithQuad;
    int hpos, wpos;
    unsigned char *crData, *cbData, *yDataFirst, *yDataSecond;
    unsigned char *outDataFirst, *outDataSecond;
    register unsigned int cr_cb_index;
    
    ditherRec = (KetanDitherRec *) ditherRef->ditherData;
    
    yCrCbToPixel = ditherRec->yCrCbToPixel;
    yIndex0 = ditherRec->yIndexTables[0];
    yIndex1 = ditherRec->yIndexTables[1];
    yIndex2 = ditherRec->yIndexTables[2];
    yIndex3 = ditherRec->yIndexTables[3];
    yIndex4 = ditherRec->yIndexTables[4];
    yIndex5 = ditherRec->yIndexTables[5];
    yIndex6 = ditherRec->yIndexTables[6];
    yIndex7 = ditherRec->yIndexTables[7];
    yIndex8 = ditherRec->yIndexTables[8];
    yIndex9 = ditherRec->yIndexTables[9];
    yIndex10 = ditherRec->yIndexTables[10];
    yIndex11 = ditherRec->yIndexTables[11];
    yIndex12 = ditherRec->yIndexTables[12];
    yIndex13 = ditherRec->yIndexTables[13];
    yIndex14 = ditherRec->yIndexTables[14];
    yIndex15 = ditherRec->yIndexTables[15];
    crIndexQ1 = ditherRec->crTables[0];
    crIndexQ2 = ditherRec->crTables[1];
    crIndexQ3 = ditherRec->crTables[2];
    crIndexQ4 = ditherRec->crTables[3];
    cbIndexQ1 = ditherRec->cbTables[0];
    cbIndexQ2 = ditherRec->cbTables[1];
    cbIndexQ3 = ditherRec->cbTables[2];
    cbIndexQ4 = ditherRec->cbTables[3];

    hDithQuad = FALSE;
    wDithQuad = FALSE;
    
#ifdef _WINDOWS
    out = out + (outStride * (imageHeight - yCoord - 1)) + xCoord;
#else
    out = out + (outStride * yCoord) + xCoord;
#endif

    for (hpos=0; hpos<h; hpos+=2) {
	crData = cr + (crStride * (hpos/2));
	cbData = cb + (cbStride * (hpos/2));
#ifdef _WINDOWS
	outDataFirst = out - (outStride * hpos);
	outDataSecond = out - (outStride * (hpos+1));
#else
	outDataFirst = out + (outStride * hpos);
	outDataSecond = out + (outStride * (hpos+1));
#endif
	yDataFirst = y + (yStride * hpos);
	yDataSecond = y + (yStride * (hpos+1));
	
	for (wpos=0; wpos<w; wpos+=2) {
	    
	    if (hDithQuad) {
		if (wDithQuad) {
		    cr_cb_index = crIndexQ4[*crData] + cbIndexQ3[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex1[yDataFirst[0]]+
						   (cr_cb_index >> 24)];
		    outDataFirst[1] = yCrCbToPixel[yIndex13[yDataFirst[1]]+
						   ((cr_cb_index & 0xff0000)>>16)];
		    outDataSecond[0] = yCrCbToPixel[yIndex9[yDataSecond[0]]+
						    ((cr_cb_index & 0xff00)>>8)];
		    outDataSecond[1] = yCrCbToPixel[yIndex5[yDataSecond[1]]+
						    (cr_cb_index & 0xff)];
		} else {
		    cr_cb_index = crIndexQ3[*crData] + cbIndexQ1[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex2[yDataFirst[0]]+
						   (cr_cb_index >> 24)];
		    outDataFirst[1] = yCrCbToPixel[yIndex14[yDataFirst[1]]+
						   ((cr_cb_index & 0xff0000)>>16)];
		    outDataSecond[0] = yCrCbToPixel[yIndex10[yDataSecond[0]]+
						    ((cr_cb_index & 0xff00)>>8)];
		    outDataSecond[1] = yCrCbToPixel[yIndex6[yDataSecond[1]]+
						    (cr_cb_index & 0xff)];
		}
	    } else {
		if (wDithQuad) {
		    cr_cb_index = crIndexQ2[*crData] + cbIndexQ2[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex3[yDataFirst[0]]+
						   (cr_cb_index >> 24)];
		    outDataFirst[1] = yCrCbToPixel[yIndex15[yDataFirst[1]]+
					       ((cr_cb_index & 0xff0000) >> 16)];
		    outDataSecond[0] = yCrCbToPixel[yIndex11[yDataSecond[0]]+
						    ((cr_cb_index & 0xff00) >> 8)];
		    outDataSecond[1] = yCrCbToPixel[yIndex7[yDataSecond[1]]+
						    (cr_cb_index & 0xff)];
		} else {
		    cr_cb_index = crIndexQ1[*crData] + cbIndexQ1[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex0[yDataFirst[0]]+
						   (cr_cb_index >> 24)];
		    outDataFirst[1] = yCrCbToPixel[yIndex12[yDataFirst[1]]+
						   ((cr_cb_index & 0xff0000) >> 16)];
		    outDataSecond[0] = yCrCbToPixel[yIndex8[yDataSecond[0]]+
						    ((cr_cb_index & 0xff00) >> 8)];
		    outDataSecond[1] = yCrCbToPixel[yIndex4[yDataSecond[1]]+
						    (cr_cb_index & 0xff)];
		}
	    }
	    crData++;
	    cbData++;
	    outDataFirst += 2;
	    yDataFirst += 2;
	    outDataSecond += 2;
	    yDataSecond += 2;
	    
	    wDithQuad = !wDithQuad;
	}
	
	hDithQuad = !hDithQuad;
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * DitherYCrCb422MSBKetan8 --
 *
 *	Color dithers an incoming YCrCb422 formatted picture block 
*	to an 8-bit MSB display using Ketan's ordered color method.
 *
 * Results:
 *	If there is a problem, it returns one of the DitherErrs as
 *	defined in dither.h
 *
 * Side effects:
 *	the appropriate block in the output bit plane contains the
 *	appropriate X11 pixel values for display.
 *
 * Parameters:
 *	DitherRec *ditherRef	The dither reference returned
 *				initially by InitDither
 *	int xCoord		x-coord of top left corner
 *	int yCoord		y-coord of top left corner
 *	int w			the width of the block
 *	int h			the height of the block
 *	unsigned char *y       	the y bit-plane
 *	unsigned char *cr      	the cr bit-plane
 *	unsigned char *cb   	the cb bit-plane
 *	unsigned char *out	the output bit-plane
 *	int yStride		distance to wrap in buffer to get to
 *				next row in block for y-plane
 *	int crStride		stride for cr-plane
 *	int cbStride		stride for cb-plane
 *	int outStride		stride for output plane
 *
 *--------------------------------------------------------------
 */
DitherErr 
DitherYCrCb422MSBKetan8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight)
{

    KetanDitherRec *ditherRec;
    unsigned char *yCrCbToPixel;
    unsigned int *yIndex0, *yIndex1, *yIndex2, *yIndex3;
    unsigned int *yIndex4, *yIndex5, *yIndex6, *yIndex7;
    unsigned int *yIndex8, *yIndex9, *yIndex10, *yIndex11;
    unsigned int *yIndex12, *yIndex13, *yIndex14, *yIndex15;
    unsigned int *crIndexQ1, *crIndexQ2, *crIndexQ3, *crIndexQ4;
    unsigned int *cbIndexQ1, *cbIndexQ2, *cbIndexQ3, *cbIndexQ4;
    register unsigned int cr_cb_index;

    Boolean wDithQuad, hDithQuad;
    int hpos, wpos;
    unsigned char *crData, *cbData, *yDataFirst;
    unsigned char *outDataFirst;

    ditherRec = (KetanDitherRec *) ditherRef->ditherData;

    yCrCbToPixel = ditherRec->yCrCbToPixel;
    yIndex0 = ditherRec->yIndexTables[0];
    yIndex1 = ditherRec->yIndexTables[1];
    yIndex2 = ditherRec->yIndexTables[2];
    yIndex3 = ditherRec->yIndexTables[3];
    yIndex4 = ditherRec->yIndexTables[4];
    yIndex5 = ditherRec->yIndexTables[5];
    yIndex6 = ditherRec->yIndexTables[6];
    yIndex7 = ditherRec->yIndexTables[7];
    yIndex8 = ditherRec->yIndexTables[8];
    yIndex9 = ditherRec->yIndexTables[9];
    yIndex10 = ditherRec->yIndexTables[10];
    yIndex11 = ditherRec->yIndexTables[11];
    yIndex12 = ditherRec->yIndexTables[12];
    yIndex13 = ditherRec->yIndexTables[13];
    yIndex14 = ditherRec->yIndexTables[14];
    yIndex15 = ditherRec->yIndexTables[15];
    crIndexQ1 = ditherRec->crTables[0];
    crIndexQ2 = ditherRec->crTables[1];
    crIndexQ3 = ditherRec->crTables[2];
    crIndexQ4 = ditherRec->crTables[3];
    cbIndexQ1 = ditherRec->cbTables[0];
    cbIndexQ2 = ditherRec->cbTables[1];
    cbIndexQ3 = ditherRec->cbTables[2];
    cbIndexQ4 = ditherRec->cbTables[3];

    hDithQuad = FALSE;
    wDithQuad = FALSE;

#ifdef _WINDOWS
    out = out + (outStride * (imageHeight - yCoord - 1)) + xCoord;
#else
    out = out + (outStride * yCoord) + xCoord;
#endif

    for(hpos=0; hpos<h; hpos++) {
	crData = cr + (crStride * hpos);
	cbData = cb + (cbStride * hpos);
#ifdef _WINDOWS
	outDataFirst = out - (outStride * hpos);
#else
	outDataFirst = out + (outStride * hpos);
#endif
	yDataFirst = y + (yStride * hpos);

	for (wpos=0; wpos<w; wpos+=2) {
	    if (hDithQuad == 0) {
		if (wDithQuad) {
		    cr_cb_index = crIndexQ2[*crData] + cbIndexQ2[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex3[yDataFirst[0]]+
						   (cr_cb_index >> 24)];
		    outDataFirst[1] = yCrCbToPixel[yIndex15[yDataFirst[1]]+
						   ((cr_cb_index & 0xff0000) >> 16)];
		} else {
		    cr_cb_index = crIndexQ1[*crData] + cbIndexQ1[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex0[yDataFirst[0]]+
						   (cr_cb_index >> 24)];
		    outDataFirst[1] = yCrCbToPixel[yIndex12[yDataFirst[1]]+
						   ((cr_cb_index & 0xff0000) >> 16)];
		}
	    } else if (hDithQuad == 1) {
		if (wDithQuad) {
		    cr_cb_index = crIndexQ2[*crData] + cbIndexQ2[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex11[yDataFirst[0]]+
						    ((cr_cb_index & 0xff00) >> 8)];
		    outDataFirst[1] = yCrCbToPixel[yIndex7[yDataFirst[1]]+
						    (cr_cb_index & 0xff)];
		} else {
		    cr_cb_index = crIndexQ1[*crData] + cbIndexQ1[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex8[yDataFirst[0]]+
						    ((cr_cb_index & 0xff00) >> 8)];
		    outDataFirst[1] = yCrCbToPixel[yIndex4[yDataFirst[1]]+
						    (cr_cb_index & 0xff)];
		}
	    } else if (hDithQuad == 2) {
		if (wDithQuad) {
		    cr_cb_index = crIndexQ4[*crData] + cbIndexQ4[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex1[yDataFirst[0]]+
						   (cr_cb_index >> 24)];
		    outDataFirst[1] = yCrCbToPixel[yIndex13[yDataFirst[1]]+
						   ((cr_cb_index & 0xff0000) >> 16)];
		} else {
		    cr_cb_index = crIndexQ3[*crData] + cbIndexQ3[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex2[yDataFirst[0]]+
						   (cr_cb_index >> 24)];
		    outDataFirst[1] = yCrCbToPixel[yIndex14[yDataFirst[1]]+
						   ((cr_cb_index & 0xff0000) >> 16)];
		}
	    } else if (hDithQuad == 3) {
		if (wDithQuad) {
		    cr_cb_index = crIndexQ4[*crData] + cbIndexQ4[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex9[yDataFirst[0]]+
						    ((cr_cb_index & 0xff00) >> 8)];
		    outDataFirst[1] = yCrCbToPixel[yIndex5[yDataFirst[1]]+
						    (cr_cb_index & 0xff)];
		} else {
		    cr_cb_index = crIndexQ3[*crData] + cbIndexQ3[*cbData];
		    outDataFirst[0] = yCrCbToPixel[yIndex10[yDataFirst[0]]+
						    ((cr_cb_index & 0xff00) >> 8)];
		    outDataFirst[1] = yCrCbToPixel[yIndex6[yDataFirst[1]]+
						    (cr_cb_index & 0xff)];
		}	  
	    } 
	    crData++;
	    cbData++;
	    outDataFirst += 2;
	    yDataFirst += 2;
	    wDithQuad = !wDithQuad;
	}
	hDithQuad = (hDithQuad+1)%4;
    }
    return 0;
}
