/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */




/*
 *--------------------------------------------------------------
 *
 * dither.h dither.c --
 *
 *	This library contains all of the routines related to dithering for 
 *      the CMT v3 system.
 *
 *--------------------------------------------------------------
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

typedef int DitherErr;
typedef int DitherType;
typedef int PicFormat;

/*
 * Parameters structure used to pass options to InitDither
 */
typedef struct DitherParam {
    DitherType dType;	/* the dither type (see below) */
    int minColors;	/* minimum colors to use */
    int maxColors;	/* maximum colors to use */
    PicFormat pFormat;	/* format of the bit planes */
} DitherParam;

/*
 *  Dither type definitions (for DitherType)
 */
#define CHOOSE_BEST 0
#define GRAY_SCALE 1
#define ORDERED_COLOR 2
#define MONO 3
#define TRUE_COLOR 4
#define BRIAN_DITHER 5
#define KETAN_DITHER 6

/*
 *  Picture format definitions (for PicFormat)
 */ 
#define YCRCB411 1
#define FMT_RGB 2
#define YCRCB422 3

/*
 * Record structure returned by InitDither, and a parameter for other
 * dither functions
 */
typedef struct DitherRec {
    unsigned long colorMap[256];
    unsigned long pixels[256];
    int numPixels;
    Display *display;
    Window ditherWin;
    int levels;
    int divN[256];
    int modN[256];
    int magic[16][16];
    void *ditherData;
    DitherType dType;
    PicFormat pFormat;
    int depth;
    int byteOrder;
    Visual *visual;
} DitherRec;

typedef struct {
  unsigned char ***crCbArray1;
  unsigned char ***crCbArray2;
  int ncr, ncb, ny;
} BrianDitherRec;

typedef struct {
  unsigned char *yCrCbToPixel;
  unsigned int **crTables;
  unsigned int **cbTables;
  unsigned int **yIndexTables;
} KetanDitherRec;

typedef struct {
  int *cr_r_tab;
  int *cr_g_tab;
  int *cb_b_tab;
  int *cb_g_tab;
  int rgbOrder;
} TrueColorDitherRec;

/*
 * Prototype for Dither Procedure
 */
typedef int (*DitherFunc)(DitherRec *ditherRef,
			  int xCoord, int yCoord,
			  int w, int h, 
			  unsigned char *bp1, unsigned char *bp2,
			  unsigned char *bp3, unsigned char *out,
			  int bp1Stride, int bp2Stride, int bp3Stride, 
			  int outStride, int, int);


/*
 *  Possible Dither Errors
 */
#define INVALID_DITHER 1
#define NOT_ENOUGH_COLORS 2
#define INIT_OK 0
#define FREE_OK 0    
#define BAD_DITHER_REC 3
    
/*
 * Dither Routine Prototypes
 */

DitherRec 
*InitDither(DitherParam dithParams, Display *dpy, Window ditherWin,
	        DitherFunc *DitherProc); 
	
DitherErr 
DitherYCrCb411MSBGray8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);

DitherErr 
DitherYCrCb411MSBOrderedColor8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);

DitherErr 
DitherYCrCb411MSBBrian8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);
 
DitherErr 
DitherYCrCb422MSBBrian8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);

DitherErr 
DitherYCrCb411MSBKetan8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);
 
DitherErr 
DitherYCrCb422MSBKetan8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);

DitherErr 
DitherYCrCb411TrueColor(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);

DitherErr 
DitherYCrCb422TrueColor(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);


DitherErr 
DitherYCrCb411MSBMono(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *y, unsigned char *cr,
		       unsigned char *cb, unsigned char *out, 
		       int yStride, int crStride, int cbStride, int outStride,
		       int imageWidth, int imageHeight);

DitherErr 
DitherRGBMSBOrderedColor8(DitherRec *ditherRef,
		       int xCoord, int yCoord,
		       int w, int h, 
		       unsigned char *r, unsigned char *g,
		       unsigned char *b, unsigned char *out, 
		       int rStride, int gStride, int bStride, int outStride,
		       int imageWidth, int imageHeight);

DitherErr
FreeDither(DitherRec *ditherRef);

/* Defines for 24 bit true color dither. */
#define NBITS	8
#define ONE     ((int) 1)
#define CONST_SCALE	(ONE << NBITS)

#define UP(x)	(((int)(x)) << NBITS)
 
#define FIX(x)  ((int) ((x)*CONST_SCALE + 0.5))

#ifndef _WINDOWS
#define DO_ASSIGN(row,C1,C2,C3)                                          \
 	    if ((unsigned) C1 < (unsigned) UP(255) &&                    \
 	        (unsigned) C2 < (unsigned) UP(255) &&                    \
 	        (unsigned) C3 < (unsigned) UP(255))                      \
 	      /* Fast common case (no overflow). */                      \
 	      *row++ = (C3 >> NBITS) + (C2 & 0xff00) + ((C1 & 0xff00) <<NBITS);\
 	    else {                                                       \
 	      /* Overflow in C3, C2 or C1. */                            \
	      v = 0;                                                     \
  	      if (C3 >= 0)                                               \
 	        if (C3 > UP(255))                                        \
 	    	  v = 255;                                               \
 	        else                                                     \
 	      	  v = C3 >> NBITS;                                       \
 	      if (C2 >= 0)                                               \
 	        if (C2 > UP(255))                                        \
 	    	  v += 255 << NBITS;                                     \
 	        else                                                     \
 	      	  v += C2 & 0xff00;                                      \
 	      if (C1 >= 0)                                               \
 	        if (C1 > UP(255))                                        \
 	    	  v += 255 << (NBITS + NBITS);                           \
 	        else                                                     \
 	      	  v += (C1 & 0xff00) << NBITS;                           \
 	      *row++ = v;}        

#define DO_TOP_ASSIGN(row,C1,C2,C3)                                          \
 	    if ((unsigned) C1 < (unsigned) UP(255) &&                    \
 	        (unsigned) C2 < (unsigned) UP(255) &&                    \
 	        (unsigned) C3 < (unsigned) UP(255))                      \
 	      /* Fast common case (no overflow). */                      \
 	      *row++ = (C3 & 0xff00) + ((C2 & 0xff00) << NBITS) + ((C1 & 0xff00) <<(NBITS+NBITS));\
 	    else {                                                       \
 	      /* Overflow in C3, C2 or C1. */                            \
	      v = 0;                                                     \
  	      if (C3 >= 0)                                               \
 	        if (C3 > UP(255))                                        \
 	    	  v = (255 << NBITS);                                               \
 	        else                                                     \
 	      	  v = C3 & 0xff00;                                       \
 	      if (C2 >= 0)                                               \
 	        if (C2 > UP(255))                                        \
 	    	  v += 255 << (NBITS+NBITS);                                     \
 	        else                                                     \
 	      	  v += (C2 & 0xff00) << NBITS;                                      \
 	      if (C1 >= 0)                                               \
 	        if (C1 > UP(255))                                        \
 	    	  v += 255 << (NBITS + NBITS + NBITS);                           \
 	        else                                                     \
 	      	  v += (C1 & 0xff00) << (NBITS+NBITS);                           \
 	      *row++ = v;}        


#else
#define DO_ASSIGN(row,R,G,B)						 \
 	    if (!((unsigned) R <= (unsigned) 255 &&			 \
 	          (unsigned) G <= (unsigned) 255 &&			 \
 	          (unsigned) B <= (unsigned) 255)) {			 \
 	      /* Overflow in R, G or B. */				 \
  	      if (R > 255)                                               \
 	    	  R = 255;                                               \
	      else if (R < 0)                                            \
 	      	  R = 0; 						 \
 	      if (G > 255)						 \
 	    	  G = 255;                                               \
 	      else if (G < 0)                                            \
 	      	  G = 0; 						 \
 	      if (B > 255)						 \
 	    	  B = 255;                                               \
 	      else if (B < 0)                                            \
 	      	  B = 0; 						 \
	    }								 \
            *row++ = (unsigned char) B;				 	 \
            *row++ = (unsigned char) G;				 	 \
            *row++ = (unsigned char) R;
#endif
