/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * every.c
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <math.h>
#include "tcl.h"
#include "tk.h"
#include "cmUtil.h"
#include "cmCmd.h"
#include "cmEvery.h"
#include "tcmEvent.h"

/*
 * One record of the following type is kept for each every record:
 */

typedef struct EveryReq {
    char name[10];
    int interval;				/* in milleseconds */
    char *priorityGroup;			/* Priority group */ 
    int token;					/* Of timer handler */
    void (*proc)(ClientData clientData);	/* call when timer fires */
    ClientData clientData;			/* Passed to proc */
    struct EveryReq *next;
} EveryReq;

static EveryReq *reqList;

/*
 *--------------------------------------------------------------
 *
 * Tcm_EveryExec --
 *
 *	This procedure is called by the timer handler routine as
 *	a callback.  It just called the request callback and then creates a
 *	new timer handler.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	A tcl command is executed, and another entry is queued up.
 *
 *--------------------------------------------------------------
 */
static void
Tcm_EveryExec(ClientData clientData)
{
    EveryReq *reqPtr = (EveryReq *) clientData;

    Tk_Preserve ((ClientData) reqPtr);

    (*reqPtr->proc) (reqPtr->clientData);

    /*  If the callback deleted the EveryReq, then don't 
     *  requeue ourselves on timer queue.
     */

    if (reqPtr->proc) {
	struct timeval tv;
	double time;

	/*
	 * Compute when the event should fire.
	 */

	(void) gettimeofday(&tv, (struct timezone *) NULL);
	TimevalToTime (tv, time);
	time += reqPtr->interval / 1000.0;

	reqPtr->token = Tcm_CreateAtHandler(time, HUGE_VAL,
					    reqPtr->priorityGroup,
					    Tcm_EveryExec,
					    (ClientData) reqPtr,
					    NULL,
					    (char *)NULL,
					    (Tk_TimerProc *)NULL,
					    (ClientData)NULL,
					    NULL);
    }
					  
    Tk_Release ((ClientData) reqPtr);
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_CreateEvery --
 *
 *	Create and initialize a request for an "every" structure.
 *
 * Results:
 *	String giving handle for newly created struct.
 *
 * Side effects:
 *	A timer handler is created and memory is allocated.
 *
 *--------------------------------------------------------------
 */
char *
Tcm_CreateEvery(int interval, char *priorityGroup,
		void (*proc)(ClientData clientData), ClientData clientData)
{
    static int everyNum;
    EveryReq *newReq;
    int len;
    struct timeval tv;
    double time;


    /*
     * Alloc it, give it a name, and link it into the list.
     */
    newReq = (EveryReq *) ckalloc(sizeof(EveryReq));
    sprintf(newReq->name, "every%d", everyNum++);
    newReq->proc = proc;
    newReq->clientData = clientData;

    newReq->interval = interval;

    len = strlen(priorityGroup) + 1;
    newReq->priorityGroup = ckalloc(len);
    memcpy(newReq->priorityGroup, priorityGroup, len);
    newReq->next = reqList;
    reqList = newReq;

    /*
     * Compute when the event should fire.
     */

    (void) gettimeofday(&tv, (struct timezone *) NULL);
    TimevalToTime (tv, time);
    time += interval / 1000.0;

    newReq->token = Tcm_CreateAtHandler(time, HUGE_VAL,
					priorityGroup,
					Tcm_EveryExec,
					(ClientData) newReq,
					NULL, (char *)NULL,
					(Tk_TimerProc *)NULL,
					(ClientData)NULL,
					NULL);
    return (newReq->name);
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_DeleteEvery --
 *
 *	Delete the command created by the procedure above, and free
 *	any associated memory.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Memory is freed.
 *
 *--------------------------------------------------------------
 */
int
Tcm_DeleteEvery(Tcl_Interp *interp, char *name, void (*freeProc)())
{
    EveryReq *reqPtr;
    EveryReq *prevPtr;

    if (strcmp(name, "all") == 0) {
	reqPtr = reqList;
	while (reqPtr != NULL) {
	    if (freeProc != NULL) {
	      (*freeProc) (reqPtr->clientData);
	    }

	    ckfree((char *) reqPtr->priorityGroup);
	    reqPtr->proc = NULL;
	    reqPtr->clientData = (ClientData) NULL;

	    Tcm_DeleteAtHandler(reqPtr->token);
	    prevPtr = reqPtr;
	    reqPtr = reqPtr->next;
	    Tk_EventuallyFree ((ClientData) prevPtr, (Tk_FreeProc *) free);
	}
	reqList = NULL;
    } else {
	/*
	 * Find the entry (error if there isn't one).
	 */
	for (prevPtr = NULL, reqPtr = reqList;;
	     prevPtr = reqPtr, reqPtr = reqPtr->next) {
	    if (reqPtr == NULL) {
		Tcl_AppendResult(interp, " DeleteEvery: invalid handle \"",
				 name, "\"", (char *) NULL);
		return TCL_ERROR;
	    }
	    if (strcmp(reqPtr->name, name) == 0) {
		break;
	    }
	}

	/*
	 * Delete this one from the list.
	 */
	if (prevPtr == NULL) {
	    reqList = reqPtr->next;
	} else {
	    prevPtr->next = reqPtr->next;
	}

	/*
	 * Free up the memory
	 */
	if (freeProc != NULL) {
	  (*freeProc) (reqPtr->clientData);
	}
	ckfree((char *) reqPtr->priorityGroup);
	reqPtr->proc = NULL;
	reqPtr->clientData = (ClientData) NULL;

	Tcm_DeleteAtHandler(reqPtr->token);
	Tk_EventuallyFree ((ClientData) reqPtr,  (Tk_FreeProc *) free);
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_EveryCmd --
 *
 *	This procedure is invoked to process the "every" Tcl/Tk
 *	command.  See the user documentation for details on what
 *	it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
int
Tcm_EveryCmd(ClientData notUsed, Tcl_Interp *interp, int argc, char **argv)
{
    char c;
    int len;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    len = strlen(argv[1]);
    if ((c == 'd') && (strncmp(argv[1], "delete", len) == 0)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " delete handle\"", (char *) NULL);
	    return TCL_ERROR;
	}
	return Tcm_DeleteEvery(interp, argv[2], Tcm_FreeCmd);
    } else {
	int interval;
	char *name;
	char *groupName;

	/*
	 * Check args
	 */
	if (argc != 3 && argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " interval cmd ?priorityGroup?\"",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[1], &interval) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (argc == 3) {
	    groupName = "tk";
	} else {
	    groupName = argv[3];
	}
	/*
	 * Looks good -- alloc a new struct and initialize it.
	 */
	name = Tcm_CreateEvery(interval, groupName, Tcm_ExecCmd,
			       Tcm_MakeCmd(interp, argv[2], 0));
	strcpy(interp->result, name);
    }
    return TCL_OK;
}
