/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/* cmFilterUtils.c
   
   perform filtering operations on arrays representing frames of video.
   currently supported filters:

   TextOverlay
   Fade
   PictInPict
   */

/* Revision History:
**
** rev	who	when		what
** ---	---	-----------	---------------------------------------------------------------------------------------------
** 3    rmf     28-nov-1995     changed to support real storage of 4:2:2 YUV arrays, INTERLEAVED and NON-INTERLEAVED
**                              more gracefully handle conditions where a text overlay or p-in-p go beyond the borders
** 2    drs     27-nov-1995     Just some syntax changes so that it would mesh with CMT better.
** 1    rmf     20-nov-1995     created
*/

#include <stdio.h>
#include "cmFilterUtils.h"

/* x distances = horizontal
   y distances = vertical
   assume (0,0) is the upper left corner
   an image with (m-1, n-1) as its lower right corner therefore has Xdist = m and Ydist = n */

/* in general, assume there are up to two input video sources which must be of the same dimensions, Xdist, Ydist
   pixels stored as an array of char, left to right, then top to bottom.
   Ydist rows are stored in order.  each row has Xdist pixels.  each pixel has 3 bytes, for Y, U, and V */

/* ********************************************************************************************************************** */

/* TextOverlay
   
   given an image, and a run length encoded bitmap for a text string, overlay the text on top of the image
   
   TextRlc is a run length encoding of bitmap for a text string, created by bdftoxbm.
   it contains the following information:
   TextRlc[0] = RlcIndex, the total number of elements in the array, starting count at 1
   (so the final -1 indicating the end of the array is at TextRlc[RlcIndex-1])
   TextRlc[1] = Row, total number of rows, starting count from 1
   TextRlc[2] = Col, total number of bits in each row, starting count from 1
   following that is the bitmap information, encoded as such:
   each row starts with a 0 or a 1, indicating the status of the first pixel
   each successive number represents a run length sequence of 0s and 1s, alternating.
   -1 indicates the end of a row.
   -1 at the start of a row indicates the end of the bitmap
   
   Source1 is a pointer to an array that represents an image
   Xdist and Ydist are the pixel dimensions of the image
   Xstart and Ystart are the coordinates to place the upper left corner of the text bounding box
   TextY, TextU, and TextV represent the 24-bit color for the text to be overlaid
   
   the source array is modified in place
   return address on success
   return NULL on failure */

/* Mode = NON_INTERLEAVED means separate Y, U, and V arrays
   
   Mode = INTERLEAVED means one continuous array with of the form:
   Y1 U12 Y2 V12 Y3 U34 Y4 V34 ...

   if Mode = INTERLEAVED, the U and V array pointers passed to the function are NULL.
   to address an INTERLEAVED array, use this handy trick:
   if SourceU is set to SourceY+1 and SourceV is set to SourceY+3,
   can address Source arrays in the same way as the NON_INTERLEAVED case,
   except multiply Y array indices by 2 and U and V array indices by 4. */

unsigned char *
TextOverlay (short *TextRlc, unsigned char *Source1Y, unsigned char *Source1U, unsigned char *Source1V,
             int Xdist, int Ydist, int Xstart, int Ystart,
             unsigned char TextY, unsigned char TextU, unsigned char TextV, int Mode)
{
    int TextSize = TextRlc[0];  /* total number of elements (starting counting from 1) of TextRlc array */
    int TextYdist = TextRlc[1]; /* vertical size of text bitmap */
    int TextXdist = TextRlc[2]; /* horizontal size of text bitmap */
    int RlcIndex = 3;           /* first array element in TextRlc with bitmap information */
    int Val;                    /* other values in the TextRlc array */
    int Status;                 /* used to check for end of row and end of array */
    int i;                      /* miscellaneous loop counting variable */
    
    int RowLength = Xdist;      /* length in bytes of a row in the source array */ 
    int Src1Index;              /* location in source array to start text overlay */ 
    int RowStart;               /* used to mark the starting position of each row of text */
    
    int PixelsToSpare;
    int LoopDone = FALSE;

    /* because U and V are downsampled by 2 in a 4:2:2 encoding, it's easier if we start on an even pixel boundary */
    /* i'm not positive if this is definitely needed, but it might help, and it's unlikely to be noticeable to the user */
    if ((Xstart % 2) != 0)
        Xstart--;

    Src1Index = Ystart*RowLength + Xstart;

    /* check to see if we're off the top or left borders - sorry, we don't truncate here */
    if (Xstart<0 || Ystart<0) {
        /* only return NULL on absolute failure */
        /* return(NULL); */
        return (Source1Y);
    }
        
    if (Mode == NON_INTERLEAVED) {
        /* check to see if we've gone beyond the bottom border */
        while (!LoopDone && (Src1Index < Xdist*Ydist)) {
            /* mark start of row in source array */
            RowStart = Src1Index;
            /* get first value in a row */
            Val = TextRlc[RlcIndex++];
            /* does the row start with a 1 (overlay), 0 (skip), or -1 (done) */
            if (Val != -1) {
                Status = Val;
                while (Val != -1) {
                    if (!Status) {
                        /* check to see if we've gone beyond the right border */
                        PixelsToSpare = (Xdist - Xstart) - (Src1Index - RowStart);
                        if (PixelsToSpare > 0) {
                            if (Val > PixelsToSpare)
                                Val = PixelsToSpare;
                            /* overlay text */
                            for (i=0; i<Val; i++) {
                                if ((Src1Index % 2) == 0) {
                                    Source1U[Src1Index/2] = TextU;
                                    Source1V[Src1Index/2] = TextV;
                                }
                                Source1Y[Src1Index++] = TextY;
                            }
                        }
                    }
                    else
                        /* skip */
                        Src1Index = Src1Index + Val;
                    /* alternate text overlays with skips, until row is done */
                    Status = !Status;
                    Val = TextRlc[RlcIndex++];
                }
                /* advance to next row in source array */
                Src1Index = RowStart + RowLength;
            }
            else
                LoopDone = TRUE;
        } /* while !LoopDone */
    }
    else /* (MODE == INTERLEAVED) */ {
        Source1U = Source1Y + 1;
        Source1V = Source1Y + 3;
        /* check to see if we've gone beyond the bottom border */
        while (!LoopDone && (Src1Index < Xdist*Ydist)) {
            /* mark start of row in source array */
            RowStart = Src1Index;
            /* get first value in a row */
            Val = TextRlc[RlcIndex++];
            /* does the row start with a 1 (overlay), 0 (skip), or -1 (done) */
            if (Val != -1) {
                Status = Val;
                while (Val != -1) {
                    if (!Status) {
                        /* check to see if we've gone beyond the right border */
                        PixelsToSpare = (Xdist - Xstart) - (Src1Index - RowStart);
                        if (PixelsToSpare > 0) {
                            if (Val > PixelsToSpare)
                                Val = PixelsToSpare;
                            /* overlay text */
                            for (i=0; i<Val; i++) {
                                if ((Src1Index % 2) == 0) {
                                    Source1U[Src1Index*2] = TextU;
                                    Source1V[Src1Index*2] = TextV;
                                }
                                Source1Y[Src1Index++*2] = TextY;
                            }
                        }
                    }
                    else
                        /* skip */
                        Src1Index = Src1Index + Val;
                    /* alternate text overlays with skips, until row is done */
                    Status = !Status;
                    Val = TextRlc[RlcIndex++];
                }
                /* advance to next row in source array */
                Src1Index = RowStart + RowLength;
            }
            else
                LoopDone = TRUE;
        } /* while !LoopDone */
    }
    /* check to make sure that we've properly read the whole array */
    if (RlcIndex != TextSize) {
        /* fprintf(stderr, "%%FILTER-E-RLCLEN, Actual length %d of text bitmap array doesn't match expected value %d\n",
                RlcIndex, TextSize); */
        /* only return NULL on absolute failure */
        /* return (NULL); */
        return (Source1Y);
    }
    else
        return (Source1Y);
} /* TextOverlay */

/* ********************************************************************************************************************** */

/* given two sources, of dimensions Xdist by Ydist, produce a composite image that fades between Source1 and Source2
   Percent = 0 means just output Source1; Percent = 1 means just output Source2
   the operation is destructive to the Source2 array(s). */

/* Mode = NON_INTERLEAVED means separate Y, U, and V arrays
   
   Mode = INTERLEAVED means one continuous array with of the form:
   Y1 U12 Y2 V12 Y3 U34 Y4 V34 ...
   
   if Mode = INTERLEAVED, the U and V array pointers passed to the function are NULL. */

unsigned char *
Fade (unsigned char *Source1Y, unsigned char *Source1U, unsigned char *Source1V,
      unsigned char *Source2Y, unsigned char *Source2U, unsigned char *Source2V,
      int Xdist, int Ydist, float Percent, int Mode)
{
    float NotPercent = 1.0 - Percent;
    unsigned char P[256];
    unsigned char NotP[256];
    int i;

    /* complete fade, don't have to do anything */
    if (Percent == 1.0)
        return(Source2Y);

    /* no fade at all, just copy Source1 array(s) to Source2 array(s) */
    if (Percent == 0.0) {
        if (Mode == NON_INTERLEAVED) {
            for (i=0; i < Xdist*Ydist/2; i++) {
                Source2Y[i] = Source1Y[i];
                Source2U[i] = Source1U[i];
                Source2V[i] = Source1V[i];
            }
            for (i=Xdist*Ydist/2; i < Xdist*Ydist; i++)
                Source2Y[i] = Source1Y[i];
        }
        else /* (MODE == INTERLEAVED) */ {
            for (i=0; i < Xdist*Ydist*2; i++)
                Source2Y[i] = Source1Y[i];
        }
        return(Source2Y);
    }
    
/* precompute Percent and (1-Percent) times all legal image values (0-255) to save time
   put these in arrays P and NotP
   for the cost of 512 multiplications up front, all of the parsing thru the arrays involves just integer arithmetic */
    for (i=0; i<256; i++) {
        P[i] = (unsigned char) (Percent * (float) i);
        NotP[i] = (unsigned char) (NotPercent * (float) i);
    }
    
/* do the actual fade operation, destructively in the Source2 array(s) */
    if (Mode == NON_INTERLEAVED) {
        for (i=0; i < Xdist*Ydist/2; i++) {
            Source2Y[i] = P[(int) Source2Y[i]] + NotP[(int) Source1Y[i]];
            Source2U[i] = P[(int) Source2U[i]] + NotP[(int) Source1U[i]];
            Source2V[i] = P[(int) Source2V[i]] + NotP[(int) Source1V[i]];
        }
        for (i=Xdist*Ydist/2; i < Xdist*Ydist; i++)
            Source2Y[i] = P[(int) Source2Y[i]] + NotP[(int) Source1Y[i]];
    }
    else /* (MODE == INTERLEAVED) */ {
        for (i=0; i < Xdist*Ydist*2; i++)
            Source2Y[i] = P[(int) Source2Y[i]] + NotP[(int) Source1Y[i]];
    }

    return(Source2Y);
} /* Fade */

/* ********************************************************************************************************************** */

/* shrink Source1 image by an integer factor and place within the Source2 image.
   shrinking is done simply by downsampling every Shrink'th (is that a word?) pixel.
   both images must be of size Xdist by Ydist.
   the upper left corner of the Source1 image goes at point Xstart,Ystart in the Source2 image.
   the operation is destructive to the Source2 array. */

/* Mode = NON_INTERLEAVED means separate Y, U, and V arrays
   
   Mode = INTERLEAVED means one continuous array with of the form:
   Y1 U12 Y2 V12 Y3 U34 Y4 V34 ...
   
   if Mode = INTERLEAVED, the U and V array pointers passed to the function are NULL.
   to address an INTERLEAVED array, use this handy trick:
   if SourceU is set to SourceY+1 and SourceV is set to SourceY+3,
   can address Source arrays in the same way as the NON_INTERLEAVED case,
   except multiply Y array indices by 2 and U and V array indices by 4. */

unsigned char *
PictInPict (unsigned char *Source1Y, unsigned char *Source1U, unsigned char *Source1V,
            unsigned char *Source2Y, unsigned char *Source2U, unsigned char *Source2V,
            int Xdist, int Ydist, int Xstart, int Ystart, int Shrink, int Mode)
{
    int XShrink = Xdist/Shrink;
    int YShrink = Ydist/Shrink;
    int i, j;
    
    int RowLength = Xdist;
    int XSkip = Shrink;
    int YSkip = Shrink * RowLength;
    int Src1Index = 0;
    int Src2Index;
    int Row1Start, Row2Start;
    int MaxXPixels, MaxYPixels;
    
    /* because U and V are downsampled by 2 in a 4:2:2 encoding, it's easier if we start on an even pixel boundary */
    /* i'm not positive if this is definitely needed, but it might help, and it's unlikely to be noticeable to the user */
    if ((Xstart % 2) != 0)
        Xstart--;

    Src2Index = Ystart*RowLength + Xstart;
    
    /* check to see if we're off the top or left borders - sorry, we don't truncate here */
    if (Xstart<0 || Ystart<0) {
        /* only return NULL on absolute failure */
        /* return(NULL); */
        return (Source2Y);
    }

    /* make sure that the shrunken Source1 fits within the boundaries of Source2; if not, compensate */
    MaxXPixels = Xdist - Xstart;
    if (XShrink > MaxXPixels)
        XShrink = MaxXPixels;
    
    MaxYPixels = Ydist - Ystart;
    if (YShrink > MaxYPixels)
        YShrink = MaxYPixels;
    
    if (Mode == NON_INTERLEAVED) {
        for (i=0; i<YShrink; i++) {            
            Row1Start = Src1Index;
            Row2Start = Src2Index;
            for (j=0; j<XShrink; j++) {
                if ((Src2Index % 2) == 0) {
                    Source2U[Src2Index/2] = Source1U[Src1Index/2];
                    Source2V[Src2Index/2] = Source1V[Src1Index/2];
                }
                Source2Y[Src2Index++] = Source1Y[Src1Index];
                Src1Index = Src1Index + XSkip;
            }
            Src1Index = Row1Start + YSkip;
            Src2Index = Row2Start + RowLength;
        }
    }
    else /* (Mode == INTERLEAVED) */ {
        Source1U = Source1Y + 1;
        Source1V = Source1Y + 3;
        for (i=0; i<YShrink; i++) {            
            Row1Start = Src1Index;
            Row2Start = Src2Index;
            for (j=0; j<XShrink; j++) {
                if ((Src2Index % 2) == 0) {
                    Source2U[Src2Index*2] = Source1U[Src1Index*2];
                    Source2V[Src2Index*2] = Source1V[Src1Index*2];
                }
                Source2Y[Src2Index++*2] = Source1Y[Src1Index*2];
                Src1Index = Src1Index + XSkip;
            }
            Src1Index = Row1Start + YSkip;
            Src2Index = Row2Start + RowLength;
        }
    }
    return(Source2Y);
} /* PictInPict */
