/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/* cmFontUtils.c

   convert a string, given a font in a bdf format, into a bitmap
   the final internal format is a run length coded form of the bitmap
   also optionally output an .xbm file
   */

/* Revision History:
**
** rev	who	when		what
** ---	---	-----------	---------------------------------------------------------------------------------------------
** 2    rmf     30-nov-1995     enable handling of any width text string - no preset maximum value
** 1    rmf     20-nov-1995     created
*/

/*
  FONTBOUNDINGBOX FBBx FBBy Xoff Yoff

  reference points by (row,col), i.e. (y,x)
  our array will have (0,0) at the lower left hand corner.
  row can range from 0 to FBBy-Yoff-1
  start col at 0.  it will grow as characters are added.
  start origin at (-Yoff,-Xoff)
  
  DWIDTH dwx dwy
  BBX BBw BBh BBxoff BByoff

  pixels get read in from top to bottom (decreasing row).
  within a row, pixels get read in from left to right (increasing col).

  BBh is a check on how many rows to read between BITMAP and ENDCHAR.
  BBw is how many pixels to read in on each row.
  these are represented by (BBw div 8) hex characters, padded on the right
  with zeros to the nearest byte.
  the first bit for a character is at an offset of (BByoff+BBh, BBxoff)
  from the origin.
  after we're done with a character, move origin by an offset of (+dwy, +dwx).
  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cmFontUtils.h"

/* internal, intermediate, inefficient, representation of bitmap */
/* char TitleBits[FBBy-Yoff][MaxWidth]; */
static char **TitleBits;

/* final, efficient, globally defined, run length encoding of bitmap */
short *RlcBits;

/* get global information about an entire font */
static int
GetFontInfo(FILE *fileptr, FontParameters *FontInfo)
{
    int FileDone;
    char StrData[MAXSTRING];

    do
        FileDone = (fscanf(fileptr, "%s", StrData) == EOF);
    /* loop until end of file, or we've found the font bounding box */    
    while (!FileDone && (strcmp(StrData, "FONTBOUNDINGBOX") != 0));
    
    if (!FileDone)
        /* get the global parameters */
        if (fscanf(fileptr, "%d %d %d %d", &FontInfo->FBBx, &FontInfo->FBBy, &FontInfo->Xoff, &FontInfo->Yoff) == EOF)
            return(NULL);
        else
            return(1);
    else
        return(NULL);
} /* GetFontInfo */

/* allocate space for and initialize internal bitmap array */
static int
InitBitmap(FontParameters *FontInfo, unsigned int MaxWidth)
{
    /* extern char **TitleBits; */
    unsigned int i;
    
    TitleBits = (char **) calloc((FontInfo->FBBy - FontInfo->Yoff), sizeof(char *));
    if (TitleBits == NULL) {
        fprintf(stdout, "%%BDFTOXBM-E-CALLOC, Error dynamically allocating TitleBits array with calloc\n");
        return (NULL);
    }

    for (i=0; i < (FontInfo->FBBy - FontInfo->Yoff); i++) {
        TitleBits[i] = (char *) calloc(MaxWidth, sizeof(char));
        if (TitleBits[i] == NULL) {
            fprintf(stdout, "%%BDFTOXBM-E-CALLOC, Error dynamically allocating TitleBits array with calloc\n");
            return (NULL);
        }
    }
    return(1);
} /* InitBitmap */

/* scan thru the bitmap file and find the positions in the file of all of the character encodings */
static int
GetFilePos(FILE *fileptr, long *filepos)
{
    int Encoding;
    int FileDone;
    char StrData[MAXSTRING];
    long CurPos;
    
    /* any character not specified in file will be assigned a position of -1 */
    for (Encoding=0; Encoding < MAXCHAR; Encoding++)
        filepos[Encoding] = -1;

    do {
        do
            FileDone = (fscanf(fileptr, "%s", StrData) == EOF);
        /* loop until end of file, or start of a character */
        while (!FileDone && (strcmp(StrData, "STARTCHAR") != 0));
        
        /* we want to save this file position */
        CurPos = ftell(fileptr);
        if (CurPos == EOF)
            FileDone = TRUE;
        else {
            do
                FileDone = (fscanf(fileptr, "%s", StrData) == EOF);
            /* loop until end of file, or we've found the encoding */    
            while (!FileDone && (strcmp(StrData, "ENCODING") != 0));
            
            /* get the encoding */
            FileDone = (fscanf(fileptr, "%d", &Encoding) == EOF);
            if (!FileDone && (Encoding>=0) && (Encoding<=MAXCHAR))
                filepos[Encoding] = CurPos;
        } /* else */
    }
    while (!FileDone);
    return(1);
} /* GetFilePos */

/* get information about a particular glyph (i.e. character) within a font */
static int
GetGlyphInfo(FILE *fileptr, long filepos, GlyphParameters *GlyphInfo)
{
    int FileDone;
    char StrData[MAXSTRING];

    /* position pointer in file to start of appropriate character */
    if (fseek(fileptr, filepos, 0) == EOF)
        return(NULL);
    else {
        /* get the glyph information, up to, but not including, the bitmap itself, which is from BITMAP to ENDCHAR */
        do
            FileDone = (fscanf(fileptr, "%s", StrData) == EOF);
        /* loop until end of file, or we've found the DWIDTH */    
        while (!FileDone && (strcmp(StrData, "DWIDTH") != 0));
        if (!FileDone)
            /* get the DWIDTH info */
            fscanf(fileptr, "%d %d", &GlyphInfo->dwx, &GlyphInfo->dwy);

        do
            FileDone = (fscanf(fileptr, "%s", StrData) == EOF);
        /* loop until end of file, or we've found the BBX */    
        while (!FileDone && (strcmp(StrData, "BBX") != 0));
        if (!FileDone)
            /* get the BBX info */
            fscanf(fileptr, "%d %d %d %d", &GlyphInfo->BBw, &GlyphInfo->BBh, &GlyphInfo->BBxoff, &GlyphInfo->BByoff);
        return(1);
    } /* else */
} /* GetGlyphInfo */

/* add the next character to the bitmap contained in the global variable TitleBits */
static int
UpdateBitmap(FILE *fileptr, FontParameters *FontInfo, GlyphParameters *GlyphInfo,
             Location *Origin, unsigned int *MaxCol, unsigned int *MaxWidth, unsigned int debug)
{
    /* extern char **TitleBits; */

    int FileDone;
    char StrData[MAXSTRING];
    int CharVal;
    int Row, Col;
    char SingleCharStr[1];
    int Mask;
    unsigned int i, j, k;
    
    /* go to the beginning of the bitmap */
    do
        FileDone = (fscanf(fileptr, "%s", StrData) == EOF);
    /* loop until end of file, or we've found the BITMAP */    
    while (!FileDone && (strcmp(StrData, "BITMAP") != 0));

    /* read in the bitmap for the character and process, one line at a time */
    if (!FileDone) {
        Row = Origin->Row + GlyphInfo->BByoff + GlyphInfo->BBh;
        for (i=0; i < GlyphInfo->BBh; i++) {            
            FileDone = (fscanf(fileptr, "%s", StrData) == EOF);
            Col = Origin->Col + GlyphInfo->BBxoff;
            if (!FileDone && (strcmp(StrData, "ENDCHAR") !=0)) {
                /* process the line, one bit at a time */
                for (j=0; j < GlyphInfo->BBw; j++) {
                    /* search for a 1 at bit position (j MOD 4) in character position (j DIV 4) */
                    /* use bitwise-AND to mask just that bit */
                    *SingleCharStr = StrData[j/4];
                    sscanf(SingleCharStr, "%x", &CharVal);
                    Mask = (1 << (3 - (j % 4)));
                    if (CharVal & Mask) {
                        /* make array bigger as needed */
                        if (Col > *MaxWidth) {
                            *MaxWidth = *MaxWidth + 100;
                            for (k=0; k < (FontInfo->FBBy - FontInfo->Yoff); k++) {
                                /* need to make every row bigger */
                                TitleBits[k] = (char *) realloc(TitleBits[k], *MaxWidth * sizeof(char));
                                if (TitleBits[k] == NULL) {
                                    fprintf(stdout, "%%BDFTOXBM-E-REALLOC, Error dynamically reallocating TitleBits array with calloc\n");
                                    return (NULL);
                                }
                                /* make sure newly allocated bits are cleared */
                                memset((void *) (TitleBits[k]+(*MaxWidth-100)), NULL, 100); 
                            }
                        }
                        TitleBits[Row][Col] = 1;
                        if (debug) printf("*");
                        /* keep track of the farthest right column */
                        if (Col > *MaxCol)
                            *MaxCol = Col;
                    }
                    else
                        if (debug) printf(" ");
                    Col++;
                }
                Row--;
                if (debug) printf("\n");
            }
        }
        if (debug) printf("\n");
    }

    /* update the origin to be ready for the next character */
    Origin->Row = Origin->Row + GlyphInfo->dwy;
    Origin->Col = Origin->Col + GlyphInfo->dwx;
        
    /* check that we're really at the end of a character */
    if (!FileDone)
        FileDone = (fscanf(fileptr, "%s", StrData) == EOF);
    if (!FileDone && (strcmp(StrData, "ENDCHAR") ==0))
        return(1);
    else
        return(NULL);
} /* UpdateBitmap */

/* output bitmap contained in TitleBits array */
static int
GenXbmFile(FILE *fileptr, char *title, FontParameters *FontInfo, unsigned int *MaxCol)
{
    /* extern char **TitleBits; */

    int Row, Col;
    unsigned int Val = 0;
    int Done;
    unsigned int NewBit, RotOldBits;

    /* pad blank bits on right hand side so that the bitmap ends on an 8-pixel block boundary */
    if (((*MaxCol+1) % 8) != 0)
        *MaxCol = *MaxCol + (8 - ((*MaxCol+1) % 8));
    
    fprintf(fileptr, "#define %s_width %d\n", title, *MaxCol+1);
    fprintf(fileptr, "#define %s_height %d\n", title, (FontInfo->FBBy - FontInfo->Yoff));
    fprintf(fileptr, "static char %s_bits[] = {\n", title);
    for (Row = (FontInfo->FBBy - FontInfo->Yoff - 1); Row >= 0; Row--) {
        fprintf(fileptr, "   ");
        for (Col = 0; Col <= *MaxCol; Col++) {
            NewBit = ((unsigned int) TitleBits[Row][Col]) << 7;
            RotOldBits = ((unsigned int) Val) >> 1;
            Val = NewBit | RotOldBits;
            if (((Col+1) % 8) == 0) {
                fprintf(fileptr, "0x%02x", (int) Val);
                Done = ((Row == 0) && (Col == *MaxCol));
                if (!Done) {
                    fprintf(fileptr, ", ");
                }
            }
        }
        /* print final right hand bits if the bitmap doesn't end on a divisible by 8 boundary */
        /*
        if ((Col-- % 8) != 0) {
            fprintf(fileptr, "0x%02x", (int) Val);
            Done = ((Row == 0) && (Col == *MaxCol));
            if (!Done) {
                fprintf(fileptr, ", ");
            }
        }
        */
        if (!Done) {
            fprintf(fileptr, "\n"); 
        }
    }
    fprintf(fileptr, "};\n");
    return(1);
} /* GenXbmFile */

/* check size of global array RlcBits - reallocate memory 1000 elements at a time as needed */
static int
RlcAryChk(RlcIndex)
{
    /* extern short *RlcBits; */
    
    /* if # of elements is currently divisible by 1000 ... */
    if ((RlcIndex % 1000) == 0) {
        /* ... increase size of array by 1000 more elements */
        RlcBits = realloc (RlcBits, (RlcIndex + 1000) * sizeof(short));
        if (RlcBits == NULL) {
            printf("%%BDFTOXBM-E-REALLOC, Error dynamically allocating RlcBits array with realloc\n");
            return (NULL);
        }
    }
    return(1);
} /* RlcAryChk */

/* generate a run length encoding of the bitmap */
static int
GenRlc(FontParameters *FontInfo, unsigned int *MaxCol)
{
    /* extern char **TitleBits; */
    /* extern short *RlcBits; */
    
    unsigned int RlcIndex = 1;
    int Row, Col;
    char RunBit, CurrentBit;
    short RunCount;
    
/* store the following items in order:
   
   RlcBits[0] = RlcIndex, the total number of elements in the array, starting count at 1
                          (so the final -1 indicating the end of the array is at RlcBits[RlcIndex-1])
   RlcBits[1] = Row, total number of rows, starting count from 1
   RlcBits[2] = Col, total number of bits in each row, starting count from 1
   each row starts with a 0 or a 1, indicating the status of the first pixel
   each successive number represents a run length sequence of 0s and 1s, alternating.
   -1 indicates the end of a row.
   -1 at the start of a row indicates the end of the bitmap
   */
    
    /* start out with RlcBits having 1000 elements.  increase size as needed. */
    RlcBits = (short *) calloc(1000, sizeof(short));
    if (RlcBits == NULL) {
        printf("%%BDFTOXBM-E-CALLOC, Error dynamically allocating RlcBits array with calloc\n");
        return (NULL);
    }

    RlcBits[RlcIndex++] = (short) (FontInfo->FBBy - FontInfo->Yoff); /* number of rows, starting count from 1 */
    RlcBits[RlcIndex++] = (short) (*MaxCol + 1); /* number of columns, starting count from 1 */
    for (Row = (FontInfo->FBBy - FontInfo->Yoff - 1); Row >= 0; Row--) {
        /* beginning of new row */
        RunBit = TitleBits[Row][0];
        RlcBits[RlcIndex++] = (short) RunBit;
        RlcAryChk(RlcIndex);
        RunCount = 1;
        for (Col = 1; Col <= *MaxCol; Col++) {
            CurrentBit = TitleBits[Row][Col];
            /* run of same bit still going */
            if (CurrentBit == RunBit)
                RunCount++;
            /* switch from 0 to 1 or vice versa */
            else {
                RlcBits[RlcIndex++] = RunCount;
                RlcAryChk(RlcIndex);
                RunBit = CurrentBit;
                RunCount = 1;
            }
        }
        /* end of row */
        RlcBits[RlcIndex++] = RunCount;
        RlcAryChk(RlcIndex);
        RlcBits[RlcIndex++] = -1;
        RlcAryChk(RlcIndex);
    }
    /* end of array */
    RlcBits[RlcIndex++] = -1;
    RlcAryChk(RlcIndex);

    /* RlcIndex now contains the total number of elements in the RlcBits array, starting count at 1
       store this as the first element in the array */
    RlcBits[0] = (short) RlcIndex;
    /* make RlcBits only as big as it needs to be */
    RlcBits = realloc (RlcBits, RlcIndex * sizeof(short));
    if (RlcBits == NULL) {
        printf("%%BDFTOXBM-E-REALLOC, Error dynamically allocating RlcBits array with realloc\n");
        return (NULL);
    }

    return(RlcIndex);
} /* GenRlc */

/* print out final run-length-coded representation of bitmap */
static int
PrintRlc()
{
    /* extern short *RlcBits; */
    
    int i;
    int max = RlcBits[0];
    
    for (i=0; i<max; i++)
        fprintf(stdout, "RlcBits[%d] = %d\n", i, RlcBits[i]);
                
    return(1);
} /* PrintRlc */

short *
bdftorlc(char *BDFFileStr, char *XBMFileStr, char *XBMStr, char *TitleStr, unsigned int debug)
{
/* keep track of the width of the bitmap (i.e. the maximum column reached) */
    unsigned int MaxCol = 0;

/* start with the bitmap array being 100 bits wide, increase as needed with realloc */
    unsigned int MaxWidth = 100;
    
/* input BDF font file */
    FILE *BDFFile;

/* output XBM X11 bitmap file */
    FILE *XBMFile;

    FontParameters FontInfo;
    GlyphParameters GlyphInfo;    
    Location Origin;
    long BDFFilePos[MAXCHAR+1]; /* file position for start of every character */

    unsigned int i;
    
    if ((BDFFile = fopen(BDFFileStr, "r")) == NULL) {
        fprintf(stdout, "%%BDFTOXBM-E-FILE, Error opening input font file %s with fopen\n", BDFFileStr);
        return NULL;
    }

    GetFontInfo(BDFFile, &FontInfo);
    InitBitmap(&FontInfo, MaxWidth);
    GetFilePos(BDFFile, BDFFilePos);
    
    Origin.Row = -(FontInfo.Yoff);
    Origin.Col = -(FontInfo.Xoff);
    
    for (i=0; i < strlen(TitleStr); i++) {
        GetGlyphInfo(BDFFile, BDFFilePos[(int) TitleStr[i]], &GlyphInfo);
        UpdateBitmap(BDFFile, &FontInfo, &GlyphInfo, &Origin, &MaxCol, &MaxWidth, debug);
    }
    
    if (XBMFileStr != NULL) {
        /* output results to an XBM bitmap file */
        if ((XBMFile = fopen(XBMFileStr, "w")) == NULL) {
            fprintf(stdout, "%%BDFTOXBM-E-FILE, Error opening output bitmap file %s with fopen\n", XBMFileStr);
            return NULL;
        }
        GenXbmFile(XBMFile, XBMStr, &FontInfo, &MaxCol);
    }
    
    /* generate and optionally print out final run-length-coded representation of bitmap */
    GenRlc(&FontInfo, &MaxCol);
    if (debug)
        PrintRlc();

    return(RlcBits);
}
