/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */



/*
 * --------------------------------------------------
 *
 * cmH261Play.c
 *
 * Implements the "h261Play" object
 *
 * --------------------------------------------------
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <math.h>
#include <tcl.h>
#include <tk.h>

#include "cmBuffer.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmLTS.h"

#include "cmH261Play.h"

#ifndef DEFAULT_H261_PLAY_DEV_TYPE
#  define DEFAULT_H261_PLAY_DEV_TYPE "sw"
#endif

#define ONE_BILLION (1000000000.)

static int nextH261PlayNum = 0;	/* Used for next unique object name */

/*
 * Prototypes for routines used internally
 */
int H261PlayCreateCmd (ClientData clientData, Tcl_Interp *interp,
		       int argc, char **argv);
static H261Play *CreateNewH261Play (Tcl_Interp *interp, char *name);
int H261PlayObjCmd (ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv);
static void ChangeDeviceType (H261Play *objPtr, char *devtype);
static int DestroyH261Play (H261Play *objPtr);
static int ConfigureH261Play (H261Play *objPtr, int argc, char **argv);
static int H261PlayReady (H261Play *objPtr);
static int H261PlayUnready (H261Play *objPtr);
static void H261Play_LTSChange(ClientData clientData);
static int H261PlayAccept (H261Play *objPtr, SB_List *sbl);
static void H261PlayFreePacket(H261Play *objPtr, H261Packet *pktPtr);
static void H261PlayStartDecode(H261Play *objPtr);
static int H261PlayContinueDecode(H261Play *objPtr);
static int H261PlayDisplay(H261Play *objPtr);
static int H261PlayStats (H261Play *objPtr);
static int H261PlayDeviceList(H261Play *objPtr);


/*
 * -------------------------------------------------------
 *
 * H261PlayModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     creates Tcl command "h261Play" for object creation
 *
 * -------------------------------------------------------
 */

int
H261PlayModuleInit (Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "h261Play", H261PlayCreateCmd,
		      NULL, (void (*)()) NULL);
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * H261PlayCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the "h261Play" Tcl command.
 *     A new object of type H261Play is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 * -------------------------------------------------------
 */

int
H261PlayCreateCmd (ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv)
{
    char new_name[256];
    H261Play *new;
    
    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    
    nextH261PlayNum++;
    sprintf(new_name, "%sh261Play%d", argv[1], nextH261PlayNum);

    new = CreateNewH261Play(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    
    Tcl_CreateCommand(interp, new_name, H261PlayObjCmd,
		      (ClientData) new, (void (*)()) NULL);
    
    if (argc > 2) {
	ConfigureH261Play(new, argc-2, argv+2);
    }
    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * CreateNewH261Play
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 * -------------------------------------------------------
 */

static H261Play *
CreateNewH261Play (Tcl_Interp *interp, char *name)
{
    H261Play *new;
    char *atomName;

    new = (H261Play *) ckalloc(sizeof(H261Play));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    new->inited = 0;
    new->ready = 0;
    new->indecode = 0;
    new->decodePkt = NULL;
    new->displayPkt = NULL;
    new->ltsName = NULL;
    new->lts = 0;
    new->ltsTriggerNum = 0;
    new->qFirst = NULL;
    new->qLast = NULL;
    new->queueLen = 5;
    new->enqueued = 0;
    new->atTokens = NULL;
    new->numAtTokens = 0;
    new->xid = 0;

    new->devicePtr = NULL;
    new->devtype = NULL;
    ChangeDeviceType(new, "default");

    atomName = ckalloc(strlen(name) + 10);
    sprintf(atomName, "%s.accept", name);
    Tutil_RegisterCommand(interp, atomName, (ClientData) H261PlayAccept,
			  (ClientData) new);
    new->acceptHandle = atomName;

    return new;
}


/*
 * -------------------------------------------------------
 *
 * H261PlayObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     H261Play object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 * -------------------------------------------------------
 */

int
H261PlayObjCmd (ClientData clientData, Tcl_Interp *interp,
		int argc, char **argv)
{
    char c;
    int len;
    H261Play *objPtr = (H261Play *) clientData;
    
    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    
    c = argv[1][0];
    len = strlen(argv[1]);
    
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureH261Play(objPtr, argc-2, argv+2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyH261Play(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl;

	if (argc < 3) {
	    Tcl_AppendResult (interp, "wrong # of args, should be: ", argv[0],
			      " accept <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}
	
	/*
	 * Parse the ScatterBufferList
	 */
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	    ckfree( (char *)sbl);
	    Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list ",
			     argv[2], NULL);
	    return TCL_ERROR;
	}

	return H261PlayAccept(objPtr, sbl);

    } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
	return H261PlayStats(objPtr);
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return H261PlayReady(objPtr);
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return H261PlayUnready(objPtr);
    } else if ((c == 'd') && (strncmp("deviceList", argv[1], len) == 0)) {
	return H261PlayDeviceList(objPtr);
    }

    Tcl_AppendResult(interp, "Bad command option: ", argv[0], argv[1],
		     (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  accept <ScatterBufferList>\n",
                "  ready\n",
                "  unready\n",
                "  stats\n",
		"  deviceList\n",
                (char *) NULL);

    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * ChangeDeviceType --
 *
 *	Sets the device type for the H261Play.  Called whenever
 *	the "devtype" slot is changed, and on initialization of the
 *	resource.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The old device is shut down, if appropriate, and the new
 *	device is created.
 *
 * Parameters:
 *	H261Play *objPtr	Pointer to resource object.
 *	char *devtype		String giving name of new device
 *
 *--------------------------------------------------------------
 */
static void
ChangeDeviceType(H261Play *objPtr, char *devtype)
{
    int len;
    char c;

    len = strlen(devtype);
    c = devtype[0];

    /*
     * If there's no change, return
     */
    if ((objPtr->devtype != NULL) &&
	(strncmp(objPtr->devtype, devtype, len) == 0)) {
	return;
    }

    /*
     * Shut down old device, if necessary, ignoring any errors it
     * returns.
     */
    H261PlayUnready(objPtr);
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
	objPtr->devtype = NULL;
    }
    if (objPtr->devicePtr != NULL) {
        objPtr->devicePtr->shutdownProc(objPtr);
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
	objPtr->inited = 0;
    }

    /*
     * Create new device.
     */
    if ((c == 's') && (strncmp("sw", devtype, len) == 0)) {
	objPtr->devicePtr = GetSwH261Device(objPtr);
	objPtr->devtype = ckalloc(strlen("sw")+1);
	strcpy(objPtr->devtype, "sw");
    } else {
	objPtr->devicePtr = GetDefaultH261Device(objPtr);
	objPtr->devtype = ckalloc(strlen(DEFAULT_H261_PLAY_DEV_TYPE)+1);
	strcpy(objPtr->devtype, DEFAULT_H261_PLAY_DEV_TYPE);
    }
}


/*
 * -------------------------------------------------------
 *
 * DestroyH261Play
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 * -------------------------------------------------------
 */

static int 
DestroyH261Play(H261Play *objPtr)
{
    int result = TCL_OK;
    int i, devReturn;

    /*
     * If not being called by DecodeDone first delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
    }

    /*
     * Shut down the device.
     */
    if (objPtr->devicePtr != NULL) {
	devReturn = 0;
	if (objPtr->devicePtr->shutdownProc != NULL) {
	    devReturn = objPtr->devicePtr->shutdownProc(objPtr);
	}
	if (devReturn != 0) {
	    result = TCL_ERROR;
	    if (objPtr->devicePtr->errorProc != NULL) {
		Tcl_AppendResult(objPtr->interp,
			 objPtr->devicePtr->errorProc(objPtr, devReturn),
				 (char *) NULL);
	    } else {
		Tcl_AppendResult(objPtr->interp,
				 "Error sutting down h.261 device", NULL);
	    }
	}
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
    }

    /*
     * Clean up lts
     */
    if (objPtr->ltsName != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	}
	ckfree(objPtr->ltsName);
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs and such.
     */
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    /*
     * Clean up previously scheduled display requests
     */
    if (objPtr->atTokens != NULL) {
	for (i = 0; i < objPtr->numAtTokens; i++) {
	    if (objPtr->atTokens[i] > 0) {
		AtEvent *event;
		int token;

		token = objPtr->atTokens[i];
		event = Tcm_FindAtEvent(token);
		if (token && (event != NULL)) {
		    H261PlayFreePacket(objPtr,
				       (H261Packet *) event->fireClientData);
		    Tcm_DeleteAtHandler(token);
		    objPtr->atTokens[i] = 0;
		}
	    }
	}
	ckfree((char *) objPtr->atTokens);
    }

    ckfree((char *) objPtr);
    return result;
}


/*
 * -------------------------------------------------------
 *
 * ConfigureH261Play
 *
 * Parameters: 
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 * 
 * Return: 
 *     Standard TCL return value
 * 
 * Action: 
 *     Does the usual configuration thing
 *
 * -------------------------------------------------------
 */

static int
ConfigureH261Play(H261Play *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
    
    /*
     * Case 1: <obj> configure
     */
    if (argc == 0) {
	if (objPtr->devtype == NULL) {
	    Tcl_AppendResult(interp, "{-device {} {}} ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-device {} {",
			     objPtr->devtype, "}} ", NULL);
	}
	
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {} {}} ", NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName,
			     "}} ", NULL);
	}
	
	sprintf(str, "0x%x", objPtr->xid);
	Tcl_AppendResult(interp, "{-xid {0} {", str, "}} ", NULL);
    }

    /*
     * Case 2: <obj> configure <-slot>
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'd') && (strncmp("-device", argv[0], len) == 0)) {
	    if (objPtr->devtype == NULL) {
		Tcl_AppendResult(interp, "-device {} {}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-device {} {",
				 objPtr->devtype, "} ", NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->ltsName == NULL) {
		Tcl_AppendResult(interp, "-lts {} {}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName,
				 "}", NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'x') && (strncmp("-xid", argv[0], len) == 0)) {
	    sprintf(str, "0x%x", objPtr->xid);
	    Tcl_AppendResult(interp, "-xid {0} {", str, "}", NULL);
	    return TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "bad configuration option: ",
			     argv[0], NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Case 3: <obj> configure <-slot> <new value>
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	
	/* --------------------------- DEVICE ----------------------------- */
	if ((c == 'd') && (strncmp("-device", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "No argument give for -device", NULL);
		return TCL_ERROR;
	    }
	    ChangeDeviceType(objPtr, argv[cntr]);
	/* --------------------------- LTS -------------------------------- */
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "No value given for slot -lts",
				 NULL);
		return TCL_ERROR;
	    }
	    
	    if (strlen(argv[cntr]) == 0) {
		newLts = NULL;
	    } else {
		newLts = (ClientData) Tcm_FindAtom(interp, argv[cntr]);
		if (newLts == NULL) {
		    Tcl_AppendResult(interp, argv[cntr],
				     " is not a valid lts.", NULL);
		    return TCL_ERROR;
		}
	    }

	    /*
	     * Release triggers and name of old LTS.
	     */
	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }
	    
	    /*
	     * Release our lock on the old lts...
	     */
	    if (objPtr->lts != NULL) {
		if (objPtr->ltsTriggerNum != -1) {
		    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
		    objPtr->ltsTriggerNum = -1;
		}

		Tk_Release(objPtr->lts);
		objPtr->lts = NULL;
	    }
		
	    /*
	     * And set lock and arrange triggers for new one.
	     */
	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
		objPtr->ltsTriggerNum =
			LTS_SetTrigger(objPtr->lts,
					LTS_SPEED_FLAG|LTS_VALUE_FLAG|LTS_OFFSET_FLAG,
				        H261Play_LTSChange,
				       (ClientData) objPtr);
		objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr])+1);
		strcpy(objPtr->ltsName,argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }

	/* ---------------------------- XID ------------------------------- */
	} else if ((c == 'x') && (strncmp("-xid", argv[cntr], len) == 0)) {
	    int newxid;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "No value given for slot -xid",
				 NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newxid) == TCL_ERROR) {
		Tcl_AppendResult(interp, argv[cntr],
				 " is not a valid xid.", NULL);
		return TCL_ERROR;
	    }
	    objPtr->xid = newxid;
	/* --------------------------- ERROR ------------------------------ */
	} else {
	    Tcl_AppendResult(interp, "Invalid configuration parameter ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}

	cntr++;
    }
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * H261PlayReady
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *	Precomputes any values needed to get the object ready,
 *	and calls the device resady method, which does the same.
 *
 * -------------------------------------------------------
 */
static int
H261PlayReady (H261Play *objPtr)
{
    int result;
    
    /*
     * Can only be made ready if both device and lts are set
     */
    if (objPtr->devicePtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "Can't make h261 player \"",
			 objPtr->name, "\" ready: device type is not set.",
			 (char *)NULL);
        return TCL_ERROR;
    }
    if (objPtr->lts == NULL) {
	Tcl_AppendResult(objPtr->interp, "Can't make h261 player \"",
			 objPtr->name, "\" ready: no lts has been attached.",
			 (char *)NULL);
        return TCL_ERROR;
    }
    result = 0;

    /*
     * Call the device init/reinit procedure
     */
    if (!objPtr->inited) {
	if (objPtr->devicePtr->initProc != NULL) {
	    result = objPtr->devicePtr->initProc(objPtr);
	}
	if (result == 0)  {
	    objPtr->inited = 1;
	}
    } else {
	if (objPtr->devicePtr->reinitProc != NULL) {
	    result = objPtr->devicePtr->reinitProc(objPtr);
	}
    }

    /*
     * If bad result, get error string, return error
     */
    if (result != 0) {
	char *errStr;

	if (objPtr->devicePtr->errorProc != NULL) {
	    errStr = objPtr->devicePtr->errorProc(objPtr, result);
	} else {
	    errStr = "Error with device initialization.";
	}

	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
    }

    objPtr->ready = 1;
    
    return result;
}


/*
 * -------------------------------------------------------
 *
 * H261PlayUnready
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *	Performs any operations needed to make the object unready,
 *	and calls the device unready method, which does the same.
 *
 * -------------------------------------------------------
 */

static int
H261PlayUnready (H261Play *objPtr)
{
    int i, result = 0;
    
    if (objPtr->devicePtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "No device was found.", (char *)
			 NULL);
        return TCL_ERROR;
    }
    
    /*
     * Blow away "at" queue.
     */
    for (i=0; i<objPtr->numAtTokens; i++) {
	if (objPtr->atTokens[i] > 0) {
	    AtEvent *at_event;
	    int token;

	    token = objPtr->atTokens[i];
	    at_event = Tcm_FindAtEvent(token);
	    if (token && (at_event != NULL)) {
		H261PlayFreePacket(objPtr,
				   (H261Packet *) at_event->fireClientData);
		Tcm_DeleteAtHandler(token);
		objPtr->atTokens[i] = 0;
	    }
	}
    }
    
    /* XXX empty packet queue? */

    objPtr->ready = 0;
    
    return result;
}


/*
 * -------------------------------------------------------
 *
 * H261Play_LTSChange
 *
 * Parameters: 
 *	clientData - Really the object pointer
 * 
 * Return: 
 *	Standard Tcl return value
 * 
 * Action: 
 *	Called when the LTS changes value.  Dequeues any previously
 *	queued play requests, frees any memory we're holding, and
 *	returns.
 *
 * -------------------------------------------------------
 */

static void
H261Play_LTSChange(ClientData clientData)
{
    H261Play *objPtr = (H261Play *) clientData;
    int i;
    
    /*
     * If not ready, do nothing.
     */
    if (objPtr->ready == 0) {
	return;
    }

    /*
     * Blow away "at" queue.
     */
    for (i=0; i<objPtr->numAtTokens; i++) {
	if (objPtr->atTokens[i] > 0) {
	    AtEvent *at_event;
	    int token;

	    token = objPtr->atTokens[i];
	    at_event = Tcm_FindAtEvent(token);
	    if (token && (at_event != NULL)) {
		H261PlayFreePacket(objPtr,
				   (H261Packet *) at_event->fireClientData);
		Tcm_DeleteAtHandler(token);
		objPtr->atTokens[i] = 0;
	    }
	}
    }
}


/*
 * -------------------------------------------------------
 *
 * H261PlayAccept
 *
 * Parameters: 
 *     objPtr - pointer to object
 *     sbl -    ScatterBuffer list
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *
 * -------------------------------------------------------
 */

static int
H261PlayAccept (H261Play *objPtr, SB_List *sbl)
{
    int i, len;
    H261Packet *newPkt, *oldPkt;
    H261BufferHeader *hdr;
    Buffer *buf;
    unsigned char *data;

    if (!objPtr->ready) {
	Tcl_AppendResult(objPtr->interp, "Cannot send data to ", objPtr->name,
			 " since it has not been readied", NULL);
	return TCL_ERROR;
    }
    
    for (i=0; i<sbl->count; i++) {
	newPkt = (H261Packet *) ckalloc(sizeof(H261Packet));
	
	hdr = (H261BufferHeader *) GetFromSB(sbl->sbs[i],
					     CM_H261_BUF_HDR, &buf, &len);
	if (hdr == NULL) {
	    /* XXX cmbinding for bad buffer */
	    ckfree((char *) newPkt);
	    continue;
	}

	newPkt->start = ((double) hdr->logStart_s) +
	    (((double) hdr->logStart_ns) / ONE_BILLION);
	newPkt->end = ((double) hdr->logEnd_s) +
	    (((double) hdr->logEnd_ns) / ONE_BILLION);
	newPkt->deviceData = NULL;
	newPkt->options = hdr->options;
	
	data = GetFromSB(sbl->sbs[i], CM_DATA_SEGMENT, &buf, &len);
	if (data == NULL) {
	    /* XXX cmbinding for bad buffer */
	    ckfree((char *) newPkt);
	    continue;
	}

	newPkt->buf = buf;
	AttachBuffer(objPtr->interp, buf, "H261Play:H261PlayAccept");

	newPkt->data = data;
	newPkt->len = len;
	newPkt->atToken = -1;

	if (objPtr->enqueued == 0) {
	    objPtr->qFirst = objPtr->qLast = newPkt;
	    newPkt->next = newPkt->prev = NULL;
	    objPtr->queueLen = 1;
	} else {
	    if (objPtr->enqueued == objPtr->queueLen) {
		/*
		 * Remove the oldest packet from the queue
		 */
		oldPkt = objPtr->qFirst;
		objPtr->qFirst = objPtr->qFirst->next;
		objPtr->qFirst->prev = NULL;
		H261PlayFreePacket(objPtr, oldPkt);
	    }

	    /*
	     * Add the new packet
	     */
	    newPkt->prev = objPtr->qLast;
	    newPkt->next = NULL;
	    objPtr->qLast->next = newPkt;
	    objPtr->qLast = newPkt;
	}
    }

    /*
     * If no decode in process, start one
     */
    if (!objPtr->indecode) {
	H261PlayStartDecode(objPtr);
    }
    
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * H261PlayFreePacket
 *
 * Parameters:
 *	objPtr - pointer to H261Play object
 *	pktPtr - pointer to the packet to free
 *
 * Action:
 *	Releases memory held for the packet
 *
 *
 * --------------------------------------------------
 */

static void
H261PlayFreePacket(H261Play *objPtr, H261Packet *pktPtr)
{
    /*
     * Call device freeProc
     */
    if ((objPtr->devicePtr != NULL)
	&& (objPtr->devicePtr->freePktProc != NULL)) {
	objPtr->devicePtr->freePktProc(objPtr, pktPtr);
    }
    
    DetachBuffer(objPtr->interp, pktPtr->buf, "H261Play:H261PlayFreePacket");
    
    if (pktPtr->atToken >= 0) {
	Tcm_DeleteAtHandler(pktPtr->atToken);
	objPtr->atTokens[pktPtr->atToken] = 0;
    }
    
    ckfree((char *) pktPtr);
}


/*
 * -------------------------------------------------------
 *
 * H261PlayStartDecode
 *
 * Parameters:
 *	objPtr - pointer to H261Play object
 *
 * Action:
 *	Begins decoding a packet, displays when finished
 *
 * -------------------------------------------------------
 */

static void
H261PlayStartDecode(H261Play *objPtr)
{
    H261Packet *pktPtr;
    
    if ((objPtr->ready == 0) || (objPtr->indecode != 0)) {
	return;
    }

    if (objPtr->qFirst == NULL) {
	return;
    }

    pktPtr = objPtr->qFirst;
    objPtr->qFirst = objPtr->qFirst->next;
    if (objPtr->qFirst != NULL) {
	objPtr->qFirst->prev = NULL;
    }

    objPtr->indecode = 1;
    objPtr->decodePkt = pktPtr;
    H261PlayContinueDecode(objPtr);
}


/*
 * -------------------------------------------------------
 *
 * H261PlayContinueDecode
 *
 * Parameters:
 *	objPtr - pointer to H261Play object
 *
 * Action:
 *	Calls the device decode routine to continue the
 *	decoding process
 *
 * -------------------------------------------------------
 */

static int
H261PlayContinueDecode(H261Play *objPtr)
{
    H261Packet *pktPtr;
    int i, newNum;
    double start, end, t;
    
    if ((objPtr->ready == 0) || (objPtr->indecode == 0)) {
	return;
    }
    
    pktPtr = objPtr->decodePkt;
    
    if ((objPtr->devicePtr != NULL)
	&& (objPtr->devicePtr->decodeProc != NULL)) {

	if (objPtr->devicePtr->decodeProc(objPtr, pktPtr) != 0) {
	    H261PlayFreePacket(objPtr, pktPtr);
	    objPtr->decodePkt = NULL;

	    objPtr->indecode = 0;
	    if (objPtr->enqueued > 0) {
		Tcm_DoWhenIdle("video", (Tk_IdleProc *) H261PlayStartDecode,
			       objPtr);
	    }
	    return TCL_OK;
	}

	/*
	 * Schedule further decoding if not done
	 */
	if (objPtr->indecode != 0) {
	    Tcm_DoWhenIdle("video", (Tk_IdleProc *) H261PlayContinueDecode,
			   objPtr);
	    return;
	}
    }
    
    if (objPtr->ready == 0) {
	H261PlayFreePacket(objPtr, pktPtr);
	return TCL_OK;
    }

    objPtr->decodePkt = NULL;

    /* XXX gather stats */
    
    if (LTS_Speed(objPtr->lts) == 0.0) {
	pktPtr->atToken = -1;
	objPtr->displayPkt = pktPtr;
	H261PlayDisplay(objPtr);
    } else {
	/*
	 * Find an at token
	 */
	for (i=0; i<objPtr->numAtTokens; i++) {
	    if (objPtr->atTokens[i] == 0) {
		break;
	    }
	}
	
	if (i == objPtr->numAtTokens) {
	    if (objPtr->numAtTokens == 0) {
		newNum = 32;
		objPtr->atTokens = (int *) ckalloc(30 * sizeof(int));
	    } else {
		newNum = objPtr->numAtTokens * 2;
		objPtr->atTokens = (int *) realloc(objPtr->atTokens,
						   newNum * sizeof(int));
	    }

	    for (i=objPtr->numAtTokens; i<newNum; i++) {
		objPtr->atTokens[i] = 0;
	    }
	    
	    i = objPtr->numAtTokens;
	    objPtr->numAtTokens = newNum;
	}
	
	start = LTS_SystemTimeOf(objPtr->lts, pktPtr->start);
	end = LTS_SystemTimeOf(objPtr->lts, pktPtr->end);
	t = (start < end) ? start : end;

	objPtr->displayPkt = pktPtr;
	objPtr->atTokens[i] =
	    Tcm_CreateAtHandler(t, HUGE_VAL, "video",
				(Tk_TimerProc *) H261PlayDisplay,
				objPtr, NULL, NULL, NULL, NULL, NULL);
	pktPtr->atToken = i;
    }

    /*
     * Schedule further decoding
     */
    if (objPtr->enqueued > 0) {
	Tcm_DoWhenIdle("video", (Tk_IdleProc *) H261PlayStartDecode,
		       objPtr);
    }
    
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * H261PlayDisplay
 *
 * Parameters:
 *	objPtr - pointer to H261Play object
 *
 * Action:
 *	Calls the device display routine to display
 *	the packet
 *
 * -------------------------------------------------------
 */

static int
H261PlayDisplay(H261Play *objPtr)
{
    H261Packet *pktPtr;
    
    pktPtr = objPtr->displayPkt;
    
    if ((objPtr->devicePtr != NULL) 
	&& (objPtr->devicePtr->displayProc != NULL)
	&& (pktPtr != NULL)) {
	
	objPtr->devicePtr->displayProc(objPtr, pktPtr);
    }
    
    H261PlayFreePacket(objPtr, pktPtr);
    
    return TCL_OK;
}


/*
 * -------------------------------------------------------
 *
 * H261PlayStats
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *
 * -------------------------------------------------------
 */

static int
H261PlayStats(H261Play *objPtr)
{
    Tcl_AppendResult(objPtr->interp, "This command is not yet implemented.",
		     NULL);
    return TCL_OK;
}

/*
 * --------------------------------------------------
 *
 * H261PlayDeviceList
 *
 * Parameters:
 *     objPtr - pointer to h261play object.
 *
 * Results:
 *     Standard Tcl Result.
 *
 * Actions:
 *     Sets result to list of devices supported. Returns TCL_OK;
 *
 * --------------------------------------------------
 */
static int
H261PlayDeviceList(H261Play *objPtr)
{
    Tcl_SetResult(objPtr->interp, "sw", TCL_STATIC);
    return TCL_OK;
}
