/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


 /*
 * --------------------------------------------------
 *
 * cmH261SwDevice.c
 *
 * Implements the software h.261 decoder device
 *
 * --------------------------------------------------
 */
#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <string.h>

#include <tcl.h>
#include <tk.h>

#ifndef _WINDOWS
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#else
extern HDC tcm_DitherHDC;
#endif

#include "cmDither.h"
#include "cmH261Play.h"
#include "cmH261Tables.h"
#include "cmUtil.h"

static unsigned int bp_curBits;
static int bp_bitCount;
static unsigned short *bp_bitBuffer;

#ifdef H261_SW

/*
 * There is one of the following structs for every
 * unique software device created
 */
typedef struct H261SwDevice {
    int indecode;     	/* Is a decode currently in progress? */

    /*
     * Stuff needed for display
     */
    Window win;
    Display *dpy;
    GC gc;
    int hasShm;
#ifdef HAVE_SHARED_MEMORY
    XShmSegmentInfo shminfo;
#endif

#ifdef _WINDOWS
    BITMAPINFO *image;
    unsigned char *image_data;
#else
    XImage *image;	/* The current image */
#endif

    int destroyImage;	/* Should the current XImage be destroyed */
    int waiting;	/* Waiting for shared memory completion event */
    int ShmEvent;	/* X Event number for shared memory events */

    /*
     * Dithering stuff
     */
    DitherFunc ditherFunc;	/* Routine to do dithering */
    DitherRec *ditherRec;	/* Information for ditherFunc */
    
    /*
     * Frame currently being displayed
     */
    int w;
    int h;
    Buffer *buf;		/* Buffer for decoded image */
    unsigned char *dithered;
    int stride;

} H261SwDevice;

/*
 * Error values
 */
static const int H261_INIT_DITHER_FAILED = 1;
static const int H261_BAD_FORMAT = 2;
static const int H261_NOT_IMPLEMENTED = 3;
static const int H261_OTHER_ERR = 4;

/*
 * Prototypes for internal routines
 */
int H261SwDevice_Init(H261Play *objPtr);
int H261SwDevice_Reinit(H261Play *objPtr);
int H261SwDevice_Decode(H261Play *objPtr, H261Packet *pkt);
int H261SwDevice_Display(H261Play *objPtr, H261Packet *pkt);
int H261SwDevice_Refresh(H261Play *objPtr, H261Packet *pkt);
int H261SwDevice_Shutdown(H261Play *objPtr);
int H261SwDevice_FreePacket(H261Play *objPtr, H261Packet *pkt);
char *H261SwDevice_Error(H261Play *objPtr, int errNum);


/*
 * Bounds check array used after IDCT, biased by 256
 */
static unsigned char H261_BoundsFix[] = {
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
    16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
    32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
    48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
    64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
    80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,
    96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108,
    109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120,
    121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132,
    133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144,
    145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156,
    157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168,
    169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180,
    181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192,
    193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204,
    205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216,
    217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228,
    229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240,
    240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240, 240 };

static char zigzag[] = {
    0, 1, 8, 16, 9, 2, 3, 10, 17, 24, 32, 25, 18, 11, 4, 5,
    12, 19, 26, 33, 40, 48, 41, 34, 27, 20, 13, 6, 7, 14, 21, 28,
    35, 42, 49, 56, 57, 50, 43, 36, 29, 22, 15, 23, 30, 37, 44, 51,
    58, 59, 52, 45, 38, 31, 39, 46, 53, 60, 61, 54, 47, 55, 62, 63 };

#define bp_Start(dataPtr, offset)		\
{						\
    bp_bitBuffer = (unsigned short *) dataPtr;	\
    bp_curBits = *bp_bitBuffer++;		\
						\
    bp_bitCount = 16 - offset;			\
}

#define bp_Get(num, retval)			\
{						\
    bp_bitCount -= (num);			\
    if (bp_bitCount < 0) {			\
	bp_curBits <<= 16;			\
	bp_curBits |= *bp_bitBuffer++;		\
	bp_bitCount += 16;			\
    }						\
    *(retval) = ((bp_curBits >> bp_bitCount) & ((1<<(num)) - 1));	\
}

#define bp_Peek(num, retval)			\
{						\
    if (bp_bitCount < (num)) {			\
	bp_curBits <<= 16;			\
	bp_curBits |= *bp_bitBuffer++;		\
	bp_bitCount += 16;			\
    }						\
    *(retval) = ((bp_curBits >> (bp_bitCount-(num))) & ((1<<(num)) - 1)); \
}

#define bp_Flush(num)				\
{						\
    bp_bitCount -= (num);			\
    while (bp_bitCount < 0) {			\
	bp_curBits <<= 16;			\
	bp_curBits |= *bp_bitBuffer++;		\
	bp_bitCount += 16;			\
    }						\
}


/*
 *--------------------------------------------------------------
 *
 * ProcessShmEvent --
 *
 *	Registered as a Tk generic handler to unset the
 *	waiting flag on a shared memory event so that we
 *	don't miss shared memory events
 *
 * Results:
 *	1 if we handle the event, 0 otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static int
ProcessShmEvent (ClientData clientData, XEvent *e)
{
    H261Play *objPtr;
    H261SwDevice *swDevice;
    XAnyEvent *shmEventPtr = (XAnyEvent *) e;

    objPtr = (H261Play *) clientData;
    swDevice = (H261SwDevice *) objPtr->devicePtr->deviceData;
    
    if (swDevice->waiting == 0) {
	return 0;
    } else if (objPtr->xid == 0) {
	swDevice->waiting = 0;
	return 0;
    } else if ((swDevice->waiting)
	       && (shmEventPtr->type == swDevice->ShmEvent)
	       && (shmEventPtr->window = objPtr->xid)) {
	swDevice->waiting = 0;
	return 1;
    }
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * LookForShmEvent --
 *
 *	Helper function for XIfEvent call  that waits for the
 *	shared memory operation to complete.
 *
 * Results:
 *	True (1) if event is a ShmEvent for the right window, 
 *	False (0) otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static int
LookForShmEvent(Display *display, XAnyEvent *eventPtr, char *arg)
{
    H261SwDevice *swDevice;

    swDevice = (H261SwDevice *) arg;

    return ((eventPtr->type == swDevice->ShmEvent) && 
	    (eventPtr->window == swDevice->win));
}


/*
 *--------------------------------------------------------------
 *
 * WaitForCompletion --
 *
 *	If we've issued a shared memory put image call, block
 *	until X server sends us a message that it's complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Will block until appropriate shared memory event is
 *	returned from X server.
 *
 *--------------------------------------------------------------
 */
static void
WaitForCompletion (H261SwDevice *swDevice)
{
#ifdef HAVE_SHARED_MEMORY
    XEvent shmEvent;

    if (!swDevice->win) {
	swDevice->waiting = 0;
	return;
    }
    if (swDevice->waiting) {
	XIfEvent(swDevice->dpy, &shmEvent, LookForShmEvent, (char *) swDevice);
	swDevice->waiting = 0;
    }
#else
    swDevice->waiting = 0;
#endif
}


/*
 *--------------------------------------------------------------
 *
 * H261SwDevice_Init --
 *
 *	Initialize the device.
 *
 * Results:
 *	0 for successful initilization, non-zero error code otherwise
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */
int
H261SwDevice_Init(H261Play *objPtr)
{
    H261SwDevice *swDevice;
    Tk_ErrorHandler handler;
    int result = 0;
    DitherParam ditherInfo;

    swDevice = (H261SwDevice *) objPtr->devicePtr->deviceData;

    swDevice->indecode = 0;

    swDevice->win = objPtr->xid;
    swDevice->dpy = Tcm_Display();
#ifndef _WINDOWS
    swDevice->gc = XCreateGC(swDevice->dpy, objPtr->xid, 0, 0);
    XSetForeground(swDevice->dpy, swDevice->gc, 0);
#endif
    swDevice->image = NULL;
#ifdef _WINDOWS
    swDevice->image_data = NULL;
#endif
    swDevice->waiting = 0;

#ifdef HAVE_SHARED_MEMORY
    if (cmXShmQueryExtension(swDevice->dpy)) {
	swDevice->hasShm = 1;
	swDevice->ShmEvent = XShmGetEventBase(swDevice->dpy)+ShmCompletion;
	Tk_CreateGenericHandler(ProcessShmEvent, (ClientData) objPtr);
    } else {
	swDevice->hasShm = 0;
    }
#else
    swDevice->hasShm = 0;
#endif
    
    /*
     * Initialize the dithering module
     */
    ditherInfo.dType = CHOOSE_BEST;
    ditherInfo.minColors = 2;
    ditherInfo.maxColors = 256;
    ditherInfo.pFormat = YCRCB411;

    swDevice->ditherRec = InitDither(ditherInfo, swDevice->dpy,
				     swDevice->win,
				     &(swDevice->ditherFunc));
    if (swDevice->ditherRec == NULL) {
	result = H261_INIT_DITHER_FAILED;
    }

    return result;
}


/*
 *--------------------------------------------------------------
 *
 * H261SwDevice_Reinit --
 *
 *	Reinitialize the software device.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
H261SwDevice_Reinit(H261Play *objPtr)
{
    H261SwDevice *swDevice;

    swDevice = (H261SwDevice *) objPtr->devicePtr->deviceData;

    if (objPtr->xid != swDevice->win) {
	swDevice->win = objPtr->xid;

	WaitForCompletion(swDevice);
	
#ifndef _WINDOWS
	if (swDevice->gc)  {
	    XFreeGC(swDevice->dpy, swDevice->gc);
	}
	swDevice->gc = XCreateGC(swDevice->dpy, objPtr->xid, 0, 0);
	XSetForeground(swDevice->dpy, swDevice->gc, 0);
#endif
	
	swDevice->waiting = 0;    
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * H261SwDevice_Decode
 *
 *	Start the decoding process for a packet.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	This function will be periodcially called until packet
 *	decode is complete
 *
 *--------------------------------------------------------------
 */
int
H261SwDevice_Decode(H261Play *objPtr, H261Packet *pkt) 
{
    H261Device *dev;
    H261SwDevice *swDevice;
    unsigned int *end;
    int gob, mb, quant, sbit;
    short *qt;

    dev = objPtr->devicePtr;
    swDevice = (H261SwDevice *) dev->deviceData;

    if (H261_INTRA(pkt->options) == 0) {
	return H261_NOT_IMPLEMENTED;
	fprintf(stderr, "Intra set in packet header\n");
	/* XXX should be cmbind for error */
    }

    sbit = H261_SBIT(pkt->options);
    bp_Start(pkt->data, sbit);

    mb = H261_MBAP(pkt->options) + 1;
    gob = H261_GOBN(pkt->options);
    quant = H261_QUANT(pkt->options);
    
    end = (int *) (pkt->data + pkt->len);
    while (bp_bitBuffer < end) {
	int code, gobn, mba;
	int w, h;

	bp_Peek(11, &code);
	mba = H261_mbaTable[code];

	if (mba == MBA_STUFF_CODE) {
	    bp_Flush(11);
	} else if (mba == MBA_START_CODE) {
	    bp_Get(16, &code);
	    if (code != 0x0001) {

		if (code == 0) {
		    /* XXX */
		    goto done;
		}

		fprintf(stderr, "invalid start code\n");
		return H261_BAD_FORMAT;
	    }
	    
	    bp_Get(4, &gobn);
	    if (gobn > 12) {
		fprintf(stderr, "gob > 12 (%d)\n", gobn);
		return H261_BAD_FORMAT;
	    } else if (gobn == 0) {
		/*
		 * Got Picture Start Code (PCS)
		 *
		 * First reset decoder state, then parse
		 * picture header
		 */
		mb = 0;
		gob = 1;

		/* Temporal Reference (TR) */
		bp_Flush(5);
		
		/* Type Information (PTYPE) */
		bp_Get(6, &code);
		if (code == 0x00) {
		    w = 176;
		    h = 144;
		} else if (code = 0x04) {
		    w = 352;
		    h = 288;
		} else {
		    fprintf(stderr, "invalid ptype (0x%x)\n", code);
		    return H261_BAD_FORMAT;
		}
		
		if ((w != swDevice->w) || (h != swDevice->h)) {
		    /*
		     * Reallocate buffer for decoded image
		     */
		    if (swDevice->dithered != NULL) {
			DetachBuffer(objPtr->interp, swDevice->buf,
				     "H261SwDevice:Decode");
		    }
		    swDevice->buf = MakeSharedBuffer(objPtr->interp,
						     w * h,
						     "H261SwDevice:Decode");
		    if (swDevice->buf == NULL) {
			fprintf(stderr, "couldn't allocate buffer\n");
			return H261_OTHER_ERR;
		    }
		    swDevice->w = w;
		    swDevice->h = h;
		    swDevice->dithered = swDevice->buf->data;
		    swDevice->stride = w;
		    swDevice->destroyImage = 1;
		}
		
		/* Extra Insertion Information (PEI) */
		bp_Get(1, &code);
		while (code) {
		    bp_Flush(8);
		    bp_Get(1, &code);
		}
	    } else {
		/* Quantizer Information (GQUANT) */
		bp_Get(5, &quant);
		
		/* Extra insertion Information (GEI) */
		bp_Get(1, &code);
		while (code) {
		    bp_Flush(8);
		    bp_Get(1, &code);
		}

		gob = gobn;
		mb = 0;
	    }
	} else {
	    int type, block, mbx, mby, flushbits;
	    unsigned char y[16*16], cr[8*8], cb[8*8];
	    
	    if ((gob == 0) || (gob > 12)) {
		fprintf(stderr, "Bad gob (%d)\n", gob);
		return H261_OTHER_ERR;
	    }

	    flushbits = mba >> 8;
	    bp_Flush(flushbits);
	    mba &= 0xff;
	    mb += mba;
	    
	    if (mb > 33) {
		fprintf(stderr, "mb > 33 (%d)\n", mb);
		return H261_BAD_FORMAT;
	    }
	    
	    /* Macroblock type information (MTYPE) */
	    bp_Peek(10, &code);
	    type = H261_typeTable[code];
	    flushbits = type >> 8;
	    bp_Flush(flushbits);
	    type &= 0xff;
	    
	    /* Macroblock quantizer (MQUANT) */
	    if (type & H261_MQUANT_MASK) {
		bp_Get(5, &quant);
	    }

	    qt = H261_Quant + (quant << 8) + 128;
	    
	    if (type & H261_MVD_MASK) {
		fprintf(stderr, "block has motion vectors (type 0x%x)\n", type);
		return H261_NOT_IMPLEMENTED;
	    }

	    /*
	     * Since motion vectors are not implemented, the
	     * CBP bit is not set and the TCOEFF bit is set so
	     * all of the actual blocks must be present.
	     */
	    for (block = 0; block < 6; block++) {
		short bits[64];
		int loc, run, level, sign;
		
		memset(bits, 0, sizeof(bits));

		/*
		 * Intra coded blocks begin with 8 bits for the
		 * DC coefficient
		 */
		bp_Get(8, &code);
		if (code == 255) {
		    code = 128;
		}
		bits[0] = code << 3;
		loc = 1;
		
		while (loc < 64) {
		    bp_Peek(14, &code);
		    level = H261_coeffTable[code];
		    flushbits = level >> 12;
		    bp_Flush(flushbits);
		    
		    level &= 0xfff;
		    if (level == H261_EOB) {
			loc = 64;
		    } else {
			if (level == H261_ESCAPE) {
			    bp_Get(6, &run);
			    bp_Get(8, &level);

			    /*XXX*/
			    if (level > 127) {
				level -= 256;
			    }
			} else {
			    run = (level >> 5) & 0xff;

			    /* sign extend 5 bit level */
			    level &= 0x1f;
			    if (level & 0x10) {
				/* level |= ~(0x1f); */
				level = ~(level & ~0x10) + 1;
			    }
			}

			loc += run;
			bits[zigzag[loc++]] = qt[level];
		    }
		}
		
		j_rev_dct(bits);

		{
		    int row, stride;
		    short *inPtr;
		    unsigned char *outPtr, *fix = H261_BoundsFix + 512;

		    if (block < 2) {
			/* Top Row Luminance Block */
			outPtr = y + 8*block;
			stride = 16;
		    } else if (block < 4) {
			/* Bottom Row Luminance Block */
			outPtr = y + 16*8 + 8*(block - 2);
			stride = 16;
		    } else if (block == 4) {
			/* First Chrominance Block */
			outPtr = cr;
			stride = 8;
		    } else {
			/* Second Chrominance Block */
			outPtr = cb;
			stride = 8;
		    }
		    
		    inPtr = bits;
		    for (row=0; row<8; row++) {
			outPtr[0] = fix[inPtr[0]];
			outPtr[1] = fix[inPtr[1]];
			outPtr[2] = fix[inPtr[2]];
			outPtr[3] = fix[inPtr[3]];
			outPtr[4] = fix[inPtr[4]];
			outPtr[5] = fix[inPtr[5]];
			outPtr[6] = fix[inPtr[6]];
			outPtr[7] = fix[inPtr[7]];
			
			outPtr += stride;
			inPtr += 8;
		    }
		}
	    }

	    mbx = 11*16*((gob-1) % 2);
	    mbx += 16 * ((mb-1) % 11);
	    mby = 3*16*((gob-1) / 2);
	    mby += 16 * ((mb-1) / 11);
	    swDevice->ditherFunc(swDevice->ditherRec, mbx, mby, 16, 16,
				 y, cb, cr, swDevice->dithered,
				 16, 8, 8, swDevice->stride,
				 swDevice->w, swDevice->h);

	}
    }

  done:
    objPtr->indecode = 0;
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * H261SwDevice_Display --
 *
 *	Toss the current frame up on the screen.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Could block if previous shared memory put image hasn't returned.
 *
 *--------------------------------------------------------------
 */
int
H261SwDevice_Display(H261Play *objPtr, H261Packet *pkt)
{
    H261Device *dev;
    H261SwDevice *swDevice;
    XImage *image;
    int x, y;
    
    dev = objPtr->devicePtr;
    swDevice = (H261SwDevice *) dev->deviceData;

    /* TODO calculate position */
    x = y = 0;
    
    WaitForCompletion(swDevice);
    
    /*
     * Create XImage structure if necessary
     */
    if ((swDevice->destroyImage) && (swDevice->image != NULL)) {

#ifndef _WINDOWS
	XDestroyImage(swDevice->image);
	swDevice->image = NULL;
#else
	swDevice->image_data = NULL;
	free((char *) swDevice->image);
#endif
    }
	
    if (swDevice->image == NULL) {
#ifndef _WINDOWS
	image = XCreateImage(swDevice->dpy, swDevice->ditherRec->visual,
			     swDevice->ditherRec->depth, ZPixmap, 0,
			     swDevice->dithered, swDevice->w, swDevice->h,
			     32, 0);
	swDevice->image = image;

#else
	int i;
	short *pPalIndex;
	int allocSize;

	allocSize = sizeof(BITMAPINFO);
	if (swDevice->ditherRec->depth == 8) {
	    allocSize += 256 * sizeof(short);
	}

	swDevice->image = (BITMAPINFO *) malloc(allocSize);

	swDevice->image->bmiHeader.biSize = (LONG)sizeof(BITMAPINFOHEADER);
	swDevice->image->bmiHeader.biWidth = imWidth;
	swDevice->image->bmiHeader.biHeight = imHeight;
	swDevice->image->bmiHeader.biPlanes = 1;
	swDevice->image->bmiHeader.biBitCount = swDevice->ditherRec->depth;
	swDevice->image->bmiHeader.biCompression = 0L;
	swDevice->image->bmiHeader.biSizeImage = 0L;
	swDevice->image->bmiHeader.biXPelsPerMeter = 0L;
	swDevice->image->bmiHeader.biYPelsPerMeter = 0L;
	swDevice->image->bmiHeader.biClrUsed = 0L;
	swDevice->image->bmiHeader.biClrImportant = 0L;
	pPalIndex = (short *) swDevice->image->bmiColors;
	if (swDevice->ditherRec->depth == 8) {
	    for (i = 0; i < 256; i ++) {
		pPalIndex[i] = i;
	    }
	}
	swDevice->image_data = swDevice->dithered;
#endif
    }
    
    /*
     * Display the image
     */
#ifndef _WINDOWS
    XPutImage(swDevice->dpy, swDevice->win, swDevice->gc, image,
	      0, 0, x, y, swDevice->w, swDevice->h);
#else
    {
	HPALETTE pal, palPrev;

	pal = XGetWindowsPalette(swDevice->dpy, 
				 XDefaultColormap(swDevice->dpy, 
						  DefaultScreen(swDevice->dpy)));
	palPrev = SelectPalette(tcm_DitherHDC, pal, FALSE);
	
	SetDIBitsToDevice(tcm_DitherHDC, x, y, swDevice->w, swDevice->h,
			  0, 0, 0, swDevice->h,
			  swDevice->image_data,
			  swDevice->image,
			  ((swDevice->ditherRef->depth == 8) ?
			   DIB_PAL_COLORS : DIB_RGB_COLORS));
	SelectPalette(tcm_DitherHDC, palPrev, FALSE);
    }
#endif

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * H261SwDevice_Refresh --
 *
 *	Called to refresh the display if part of it is exposed.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Display is repainted.  Could block if previous shared
 *	memory put image hasn't returned
 *
 *--------------------------------------------------------------
 */
int
H261SwDevice_Refresh(H261Play *objPtr, H261Packet *pkt)
{
    /* refresh */
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * H261SwDevice_Shutdown --
 *
 *	Called to destroy device.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	The device must never be referenced again.
 *
 *--------------------------------------------------------------
 */
int
H261SwDevice_Shutdown(H261Play *objPtr) 
{
    /* shutdown */

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * H261SwDevice_FreePacket --
 *
 *	The current packet is being taken away -- clean up any
 *	pointers we have to it.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
H261SwDevice_FreePacket(H261Play *objPtr, H261Packet *pkt)
{
    /* free */
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * H261SwDevice_Error --
 *
 *	Translates an error number (passed in) into a human
 *	readable string (returned).
 *
 * Results:
 *	Human readable string that indicates the error.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	H261Play *objPtr	The device
 *	int errNum		The error code
 *
 *--------------------------------------------------------------
 */
char *
H261SwDevice_Error(H261Play *objPtr, int errNum)
{
    switch (errNum) {
	/* errors */
	
      default:
	return "Unknown Software h.261 Device Error.\n";
    }
}


/*
 *--------------------------------------------------------------
 *
 * GetSoftwareH261Device --
 *
 *	Called by resource to determine default device.  We should
 *	only be called if there's no jpeg hardware available.
 *
 * Results:
 *	Pointer to new H261Device
 *
 * Side effects:
 *	A couple of structures are allocated.
 *
 *--------------------------------------------------------------
 */
H261Device *
GetSwH261Device(H261Play *objPtr)
{
    H261Device *new;
    H261SwDevice *newsw;
    
    newsw = (H261SwDevice *) ckalloc(sizeof(H261SwDevice));
    newsw->indecode = 0;
    newsw->win = 0;
    newsw->dpy = NULL;
    newsw->gc = NULL;
    newsw->hasShm = 0;
    newsw->image = NULL;
    newsw->destroyImage = 0;
    newsw->waiting = 0;
    newsw->ShmEvent = 0;
    newsw->ditherFunc = NULL;
    newsw->ditherRec = NULL;
    newsw->w = 0;
    newsw->h = 0;
    newsw->buf = NULL;
    newsw->dithered = NULL;
    newsw->stride = 0;
    
    new = (H261Device *) ckalloc(sizeof(H261Device));
    new->deviceData = (ClientData) newsw;
    new->initProc = H261SwDevice_Init;
    new->reinitProc = H261SwDevice_Reinit;
    new->shutdownProc = H261SwDevice_Shutdown;
    new->decodeProc = H261SwDevice_Decode;
    new->displayProc = H261SwDevice_Display;
    new->refreshProc = H261SwDevice_Refresh;
    new->freePktProc = H261SwDevice_FreePacket;
    new->configProc = NULL;
    new->errorProc = H261SwDevice_Error;

    return new;
}


#else

/*
 *--------------------------------------------------------------
 *
 * GetSoftwareH261Device --
 *
 *	Stub routine for machines without compiled in software
 *	h.261 decoder.
 *
 * Results:
 *	NULL
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
H261Device *
GetSwH261Device(H261Play *objPtr)
{
    return NULL;
}

#endif
