/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * cmIrixAuDevice.c
 *
 *       This file provides an auPlay device for SGI machines running Irix.
 *       This device should be able to play 8-bit ulaw, 16 linear PCM, 
 *       24-bit linear PCM, and music with 1, 2, or 4 channels on an Indy.
 *       It provides a much smoother playback than using audiofile.
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <math.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>
#include "cmAuPlay.h"

#ifdef IRIX_AU

/*#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/filio.h>
#include <stropts.h>*/
#include "cmAuFile.h"
#include "cmLTS.h"
#include "cmAuUtil.h"

/* Irix specific include */
#include <dmedia/audio.h>

static ALport audioPort ;       /* Irix audio port structure */
static ALconfig audioPortConfig ; /* structure used to configure port */
static int inUse = 0;           /* 1 if the audio device is being used*/
static int atToken = 0;		/* Token for atEvent for playback when
				 * audio device queue is empty */
static double lastStartTime=0;	/* Last time the IrixAuStartPlay()
				 * function was called. */
static int errorCode;		/* Current error code */
static long origSampleSize ;    /* number of bytes in arriving sample */
static long playSampleSize ;    /* number of bytes that get played */
static int samplingRate ;       /* Samples per second */
static long channels     ;      /* 1 - mono, 2 - stereo, 4 - 4 channel */
static int encoding     ;       /* u-law, 16bit lin */

/*
 * Interface routines
 */
static int IrixAuDeviceInit		(SpeakerRsrc *objPtr);
static int IrixAuDeviceReinit		(SpeakerRsrc *objPtr);
static int IrixAuDeviceShutdown	(SpeakerRsrc *objPtr);
static int IrixAuDeviceUnready		(SpeakerRsrc *objPtr);
static int IrixAuDeviceLTSChanged	(SpeakerRsrc *objPtr);
static int IrixAuDevicePlay		(CBInfo *cbPtr);
static char *IrixAuDeviceError		(SpeakerRsrc *objPtr, int errNum);
SpeakerDevice *GetIrixAuDevice		(SpeakerRsrc *objPtr);

/*
 * Internal Routines
 */
static void IrixAuSilence		(void);
static void IrixAuStartPlay 		(CBInfo *cbPtr);
static void IrixAuDeviceSetGain	(SpeakerRsrc *objPtr);
static int IrixAuDeviceSetPlayParams	(int sample_rate, int channels,
					 int encoding);

/* a safety factor when forcing restarts and syncing the clock */
#define SKEW		0.005

#define SYNC_TIME	10.0

#define IN_USE_ERROR	1
#define OPEN_ERROR	2
#define FORMAT_ERROR	3
#define TOO_EARLY_ERROR	4
#define NOT_OPEN_ERROR	5
#define BAD_CHANNELS    6
#define BAD_CONFIG      7
#define BAD_PARAMS      8
#define WRITE           9

#ifndef min
#  define min(a,b)	((a)<(b)?(a):(b))
#endif


/*
 * -------------------------------------------------------
 *
 * GetIrixAuDevice
 *
 *	Allocates a new SpeakerDevice structure, fills
 *	in to handle Irix Au and return it.
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetIrixAuDevice(SpeakerRsrc *objPtr)
{
    SpeakerDevice *sdev;
    
    sdev = (SpeakerDevice *) ckalloc(sizeof(SpeakerDevice));
    sdev->deviceData = NULL;
    sdev->initProc = IrixAuDeviceInit;
    sdev->reinitProc = IrixAuDeviceReinit;
    sdev->shutdownProc = IrixAuDeviceShutdown;
    sdev->unreadyProc = IrixAuDeviceUnready;
    sdev->ltsProc = IrixAuDeviceLTSChanged;
    sdev->playProc = IrixAuDevicePlay;
    sdev->errorProc = IrixAuDeviceError;
    errorCode = 0;

    return sdev;
}

/*
 * -------------------------------------------------------
 *
 * IrixAuSilence
 *
 * Flush any pending output (i.e., silence the speaker)
 *
 * -------------------------------------------------------
 */

static void
IrixAuSilence()
{
    if (inUse) {
	/*
	 * As far as I can tell, there is no way to 
	 * flush the audio buffer once that it has 
	 * been written
	 */
    }
}


/*
 * -------------------------------------------------------
 *
 * IrixAuDeviceInit
 *
 * Initialize this device. Open an Irix audio port, configured with
 * a default number of channels.

 *
 * -------------------------------------------------------
 */

static int
IrixAuDeviceInit(SpeakerRsrc *objPtr)
{
    if (inUse) {
	return IN_USE_ERROR;
    }     

    audioPortConfig = ALnewconfig() ;

    /* 
     * Determine the current settings of the audio config
     */
    
    /* 
     *  We need the save the channels, because we cannot change them
     *  without reopening the port
     */  
    channels = ALgetchannels(audioPortConfig) ;
    
    /* Can't set the audio parameters until we have an audio stream */
       
    audioPort = ALopenport("CMT", "w", audioPortConfig) ;
    
    if (audioPort == NULL) {
	return OPEN_ERROR ;
    }
    
    inUse = 1 ;
    IrixAuDeviceSetGain(objPtr);
    errorCode = 0;
    return 0;
}

/*
 ****************************************************************
 *
 * IrixAuDeviceChangeChannels
 *
 *     Once you open an Irix audio port, you cannot change the number
 *     of channels. So if the sound changes from streo to mono in 
 *     midstream, we are forced to close and reopen the ALport.
 *
 * Parameters: number of channels 
 *
 * Returns: 0 if success, >0 if failure.
 *
 ****************************************************************
 */

static int
IrixAuDeviceChangeChannels(long ch) 
{
  if (inUse) {
      ALcloseport(audioPort) ;
  }
  
  if (ALsetchannels(audioPortConfig, ch) < 0) {
    return BAD_CHANNELS ;
  }
  
  audioPort = ALopenport("CMT", "w", audioPortConfig) ;
  
  if (audioPort == NULL) {
    return OPEN_ERROR ;
  }
  channels = ch ;
  return 0 ; 
}

/*
 * -------------------------------------------------------
 *
 * IrixAuDeviceShutdown
 *
 * Destroys the memory that has been allocated for a
 * given IrixAuDevice
 *
 * -------------------------------------------------------
 */

static int
IrixAuDeviceShutdown(SpeakerRsrc *objPtr)
{
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	atToken = 0;
    }
    IrixAuSilence();

    if (inUse) {
	ALcloseport(audioPort) ;
	ALfreeconfig(audioPortConfig) ;

	inUse = 0 ;
    }
    
        
    errorCode = 0;
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * IrixAuDeviceUnready
 *
 *	Stub routine
 *
 * -------------------------------------------------------
 */

static int
IrixAuDeviceUnready(SpeakerRsrc *objPtr)
{ 
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	atToken = 0;
    }
    IrixAuSilence();
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * IrixAuDeviceReinit
 *
 *	Stub routine
 *
 * -------------------------------------------------------
 */
static int
IrixAuDeviceReinit(SpeakerRsrc *objPtr)
{ 

    IrixAuDeviceSetGain(objPtr);
    errorCode = 0;
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * IrixAuDeviceSetPlayParams
 *
 *	Set the sample rate for audio output.  Returns 1
 *	if successfully set, 0 otherwise
 *
 * Parameters:
 *	int sample_rate;         samples per second
 *	int channels;            # of interleaved channels
 *	int encoding;            data encoding format
 *
 *  NOTE: if any parameter is -1, the value is left unchanged.
 *
 * -------------------------------------------------------
 */
static int
IrixAuDeviceSetPlayParams (int sampleRate, int numChannels, int coding)
{
    long pvBuffer[10] ;
    
    /*
     * If not initiatilzed, return
     */
    if (!inUse) {
	return 1;
      }
    

    /* 
     *  Although sun mu-law is only 1 byte, Irix au does not 
     *  directly play m-law samples. So the 8-bit mu-law gets
     *  converted into 16-bit linear before playing
     */
    
    if (coding == SUN_ULAW) {
	playSampleSize = 2 ;
	origSampleSize = 1 ;
    } else {
	/*
	 * Currently CMT only support ulaw and 16-bit linear
	 */
	playSampleSize = 2; 
	origSampleSize = 2 ;
    }
    
    /* 
     *  Configure the size of the device queue. Right now we
     *  will configure it for 2.0 seconds.
     */
    ALsetqueuesize(audioPortConfig, sampleRate*numChannels*2) ;
    
    if (numChannels != -1) {
        if (IrixAuDeviceChangeChannels(numChannels) != 0) {
	  return BAD_CONFIG ;
	}
	if (ALsetwidth(audioPortConfig, playSampleSize) < 0) {
	  return BAD_CONFIG ;
	}
	if (ALsetconfig(audioPort, audioPortConfig) < 0) {
	  return BAD_CONFIG ;
	}
    }
    
    pvBuffer[0] = AL_OUTPUT_RATE ;
    pvBuffer[1] = sampleRate ;
    
    if (ALsetparams(AL_DEFAULT_DEVICE, pvBuffer, 2) < 0) {
      return BAD_PARAMS ;
    }

    /*
     * Save the parameters, so that we can detect a change in 
     * format
     */
    channels = numChannels ;

    encoding = coding ;
    
    samplingRate = sampleRate ;
    
    fprintf(stderr, "Rate: %d\nWidth:%d\nchannels:%d\n", samplingRate, origSampleSize,
	    channels) ;
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * IrixAuDeviceSetGain
 *
 *	Set the gain of the audio device
 *
 * -------------------------------------------------------
 */

static void
IrixAuDeviceSetGain (SpeakerRsrc *objPtr)
{
    long pvBuffer[4] ;
    
    /* 
     *  Gain on an Irix machine is in the range 0-255 
     */
    if ((objPtr != NULL) && (inUse)) {
	pvBuffer[0] = AL_LEFT_SPEAKER_GAIN ;
	pvBuffer[1] = objPtr->gain * 255;
	pvBuffer[2] = AL_RIGHT_SPEAKER_GAIN ;
	pvBuffer[3] = objPtr->gain * 255;
	/* 
	 * Currently Irix only supports one device, the default 
	 */
	if (ALsetparams(AL_DEFAULT_DEVICE, pvBuffer, 4) < 0) {
	  return ;
	}
    }
    fprintf(stderr, "gain: %lf\n", objPtr->gain * 255) ;
}


/*
 * -------------------------------------------------------
 *
 * IrixAuDeviceLTSChanged
 *
 *	Notify the IrixAuDevice that the LTS has changed
 *
 * -------------------------------------------------------
 */

static int
IrixAuDeviceLTSChanged (SpeakerRsrc *objPtr)
{
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	atToken = 0;
    }
    IrixAuSilence();
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * IrixAuPlaySamples
 *
 * Parameters: 
 *     	objPtr - pointer to spearker resource
 *      cbPtr  - pointer to audio info structure
 *      n      - number of samples.  
 * 
 * Return: 
 *	None
 * 
 * Action: 
 *	Converts 8-bit mulaw samples to 16-bit linear if necessary.
 *	Then writes the samples to the device.
 *
 * -------------------------------------------------------
 */
static int 
IrixAuPlaySamples(SpeakerRsrc *objPtr, CBInfo *cbPtr, int n)
{
    Buffer *linBuf ;   /* Holds the uncompressed ulaw data */
    short *linArray ;  /* Used to access the linBuf->data like an array */
    char *ulawArray ;  /* Used to access the cbPtr->data like an array */
    int i ;
    int numSlots ;     /* Number of free slots in device queue for samples */

    /* SGI audio device does not automatically decompress U-LAW like sun */
    if (encoding == SUN_ULAW) {
	/* 
	 *  Use cm's buffering utilities 
	 *  ulaw compresses 2 to 1.
	 *  width of sample had better be 1 byte.
	 */
	linBuf = MakeBuffer(objPtr->interp, n * 2, 
			    "IrixAuDevice:IrixAuDevicePlay") ;
	
	linArray = (short *) linBuf->data ;
	ulawArray = (char *) cbPtr->data ;
	
	for (i=0; i < n; i++) {
	    linArray[i] = tcm_UlawToLinearTable[ulawArray[i]] ;
	}

	/*
	 * Only write as many samples as there are slots for the 
	 * samples.
	 */
	
	/*
	 * Determine the number of samples we can write
	 */
	numSlots = ALgetfillable(audioPort) ;
	
	if (ALwritesamps(audioPort, linBuf->data, 
			 (n > numSlots) ? numSlots : n) < 0) {
	    return WRITE ;
	}
	
	DetachBuffer(objPtr->interp, linBuf, 
		     "IrixAuDevice:IrixAuPlaySamples") ;
    } else {
	/*
	 * Only write as many samples as there are slots for the 
	 * samples.
	 */
	numSlots = ALgetfillable(audioPort) ;
	
	if (ALwritesamps(audioPort, cbPtr->data, 
			 (n > numSlots) ? numSlots : n) < 0) {
	    return WRITE ;
	}
	
    }
    fprintf(stderr, "number of Samples Written: %d\n", n) ;
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * IrixAuDevicePlay
 *
 * Parameters: 
 *	cbPtr - Frame to be played
 * 
 * Return: 
 *	None
 * 
 * Action: 
 *	Check to make sure the AC is set up properly,
 *	then gives the data to the Irix AudioLibrary.
 *
 * -------------------------------------------------------
 */
static int
IrixAuDevicePlay(CBInfo *cbPtr)
{
    double now;
    int forceRestart;
    SpeakerRsrc *objPtr;


    if (!inUse) {
	return NOT_OPEN_ERROR;
    }

    /*
     * Check for earlier failures that need special recovery (e.g.
     * unplayable format)
     */
    if (errorCode != 0) {
	return errorCode;
    }

    if (atToken != 0) {
	return TOO_EARLY_ERROR;
    }

    /*
     *  1. if the device parameters have changed (e.g., a different
     *	   data format), wait until the real start time to reset the
     *	   parameters and then write the data.
     *  2. If we've been writing data continuously in our attempts
     *	   to keep the queue full for a long time (SYNC_TIME), the
     *	   synchronization might have drifted.  So, wait until the
     *	   real start time, flush any remaining data, and then write
     *	   the data.
     *
     * Actions 1,2 are performed by the IrixAuStartPlay() funciton
     * below, since they are fired by an at event.  They care signaled
     * by setting the forceRestart variable to 1.
     *
     */
    forceRestart = 0;

    /*
     * Check to see if the format has changed.  If so, wait for
     * device to drain and force restart.
     */
    if ((cbPtr->hdr.samplingRate != samplingRate) ||
	(cbPtr->hdr.channelCount != channels) ||
	(cbPtr->hdr.dataFormat != encoding)) {
	forceRestart = 1;
    }

    /*
     * Check to see if it's been awhile (SYNC_TIME) since we did
     * the last restart.  If so, force a restart to keep synchronization
     * from drifting
     */
    objPtr = cbPtr->objPtr;
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));
    if (now > lastStartTime + SYNC_TIME) {
	forceRestart = 1;
    }

    /*
     * Play it directly or wait.
     */
    if (forceRestart) {
	Tk_Preserve(cbPtr);
	AttachBuffer (objPtr->interp, cbPtr->buffer,
			"IrixAuDevice:IrixAuDevicePlay");
	
	atToken = Tcm_CreateAtHandler(cbPtr->start-SKEW, HUGE_VAL,
			    "audio", (Tk_TimerProc *)IrixAuStartPlay,
			    (ClientData) cbPtr, NULL,
			    NULL, (Tk_IdleProc *)NULL, (ClientData)NULL, NULL);

    } else {
	int numSamps ;     /* Number of samples arriving */
	
	/* We only write as much data as we have buffers for */
	
	
	/*
	 * Now determine how many samples we have to write
	 * If there are multiple channles, there is a sample
	 * for each channel
	 */
	numSamps = cbPtr->len / origSampleSize ;	
	fprintf(stderr, "number of Samples        : %d\n", numSamps) ;

	IrixAuPlaySamples(objPtr, cbPtr, numSamps) ;

    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * IrixAuStartPlay --
 *
 *	Starts off the initial playback to the device.  Only
 *	called when teh device queu is thought to be empty.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Writes data to the device.
 *
 *--------------------------------------------------------------
 */
static void
IrixAuStartPlay (CBInfo *cbPtr)
{
    double now, fraction;
    int offset;
    SpeakerRsrc *objPtr;
    int numSamps ;        /* Number of arriving audio samples */
    atToken = 0;
    objPtr = cbPtr->objPtr;
    DetachBuffer (objPtr->interp, cbPtr->buffer,
		  "IrixAuDevice:IrixAuStartPlay");
    if (!inUse) {
	Tk_Release(cbPtr);
	return;
    }

    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));

    /*
     * Flush any data in the output queue
     */
    IrixAuSilence();

    /*
     * Set up the device parameters
     */
        
    if ((cbPtr->hdr.samplingRate != samplingRate) ||
	(cbPtr->hdr.channelCount != channels) ||
	(cbPtr->hdr.dataFormat != encoding)) {
	if (IrixAuDeviceSetPlayParams (cbPtr->hdr.samplingRate,
				       cbPtr->hdr.channelCount, 
				       cbPtr->hdr.dataFormat)) {
	    errorCode = FORMAT_ERROR;
	    Tk_Release(cbPtr);
	    return;
	}
    }
    

    /*
     * If callback is late, just play the appropriate fraction of the data
     */
    if (now > cbPtr->start) {
	fraction = (now - cbPtr->start) / (cbPtr->end - cbPtr->start);
	if (fraction > 1.0) {
	    Tk_Release(cbPtr);
	    return;
	}
	if (fraction < SKEW) {
	    fraction = 0.0;
	}
    } else {
	fraction = 0.0;
    }
    offset = fraction * cbPtr->len + 0.499;

    /*
     * Byte align the offset correctly
     */
    offset = origSampleSize*(offset/origSampleSize);

    /* We only write as much data as we have buffers for */
	
    /*
     * Now determine how many samples we have to write
     */
    numSamps = (cbPtr->len - offset) / origSampleSize ;
	
    IrixAuPlaySamples(objPtr, cbPtr, numSamps) ;

    fprintf(stderr, "Restarted\n") ;
    fprintf(stderr, "number of Samples        : %d\n", numSamps) ;
    
    
    Tk_Release(cbPtr);
    lastStartTime = now;
    return;
}

/*
 *--------------------------------------------------------------
 *
 * IrixAuDeviceError --
 *
 *      Map an error code (an integer returned by one of the functions
 *      above) into a human readable string.
 *
 * Results:
 *      Constant string.
 *
 * Side effects:
 *      None
 *
 *--------------------------------------------------------------
 */
static char *
IrixAuDeviceError(SpeakerRsrc *rsrcPtr, int errNum)
{
    switch (errNum) {
      case IN_USE_ERROR:
	  return "Audio Device is already in use by another object";

      case NOT_OPEN_ERROR:
	  return "Attempt to play to unopen audio device";

      case OPEN_ERROR:
	  return "Error opening /dev/audio";

      case TOO_EARLY_ERROR:
	  return "Send-ahead time must be less than data length";

      case FORMAT_ERROR:
	  return "Audio data is of wrong format";
      
      case BAD_CHANNELS:
	  return "Error setting channels" ;
	  
      case BAD_CONFIG:
	  return "Error configuring port" ;

      case BAD_PARAMS:
	  return "Error configureing device" ;
      case WRITE:
	  return "Error writing samples" ;

      default:
	  return "Unknown Irix Audio Error";
    }
}


#else /* IRIX_AU */

/*
 * -------------------------------------------------------
 *
 * GetIrixDevice
 *
 *	Stub routine for non sparc machines
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetIrixAuDevice(SpeakerRsrc *objPtr)
{
    return (SpeakerDevice *)NULL;
}

#endif /* IRIX_AU */
