/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include "cmCookies.h"

/* 
 * Definition of XPlxCImage copied from appropriate
 * parallax header. Avoids having to worry about distributing
 * the parallax header.
 */

typedef struct _XPlxCImage {
    int width, height;		/* size of image */
    int size;			/* size in bytes of image */
    char *data;			/* pointer to image data */
    char *info;			/* shared memory segment info */
} XPlxCImage;

/*
 * The next array represents the default JFIF header for
 * quality = 100 and size = 640x480. The values are
 * adjusted as the J_Movie header is read.  
 *
 * The buffer contains the jfif header with the huffman table
 */
static unsigned char jfif100Header[] = {
    /* SOI */
    0xFF, 0xD8,

    /* JFIF header */
    0xFF, 0xE0,			/* Marker */
    0x00, 0x10,			/* Length = 16 bytes */
    0x4A, 0x46, 0x49, 0x46,	/* "JFIF" */
    0x00, 0x01, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00,

    /* Start of frame (section B.2.2) */
    0xFF, 0xC0,		/* Baseline DCT */
    0x00, 0x11,		/* Length = 17 bytes */
    0x08,		/* Sample precision */
    0x01, 0xE0,		/* Height */
    0x02, 0x80,		/* Width */
    0x03,		/* Number of components = 3 */
    0x01, 0x21, 0x00,	/* Scan 1: 2:1 horiz, 1:1 vertical, use QT 0 */
    0x02, 0x11, 0x01,	/* Scan 2: 1:1 horiz, 1:1 vertical, use QT 1 */
    0x03, 0x11, 0x01,	/* Scan 3: 1:1 horiz, 1:1 vertical, use QT 1 */

    /* Define Quant table (section B.2.4.1) */
    0xFF, 0xDB,		/* Marker */
    0x00, 0x84,		/* Length (both tables) */
    0x00,		/* 8 bit values, table 0 */
    0x10, 0x0B, 0x0C, 0x0E, 0x0C, 0x0A, 0x10, 0x0E,
    0x0D, 0x0E, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
    0x1A, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
    0x1D, 0x28, 0x3A, 0x33, 0x3D, 0x3C, 0x39, 0x33,
    0x38, 0x37, 0x40, 0x48, 0x5C, 0x4E, 0x40, 0x44,
    0x57, 0x45, 0x37, 0x38, 0x50, 0x6D, 0x51, 0x57,
    0x5F, 0x62, 0x67, 0x68, 0x67, 0x3E, 0x4D, 0x71,
    0x79, 0x70, 0x64, 0x78, 0x5C, 0x65, 0x67, 0x63,

    0x01,		/* 8 bit values, table 1 */
    0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2F, 0x1A,
    0x1A, 0x2F, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,

    /* Define huffman table (section B.2.4.1) */
    0xFF, 0xC4,		/* Marker */
    0x00, 0x1F,		/* Length (31 bytes) */
    0x00,		/* DC, table 0 */
    0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0A, 0x0B,

    /* Define huffman table (section B.2.4.1) */
    0xFF, 0xC4,		/* Marker */
    0x00, 0xB5,		/* Length (181 bytes) */
    0x10,		/* AC, table 0 */
    0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05, 0x05, 0x04, 0x04,
    0x00, 0x00, 0x01, 0x7D, 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
    0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32,
    0x81, 0x91, 0xA1, 0x08, 0x23, 0x42, 0xB1, 0xC1, 0x15, 0x52, 0xD1, 0xF0,
    0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0A, 0x16, 0x17, 0x18, 0x19, 0x1A,
    0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
    0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53, 0x54, 0x55,
    0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
    0x6A, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x83, 0x84, 0x85,
    0x86, 0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
    0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2,
    0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5,
    0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8,
    0xD9, 0xDA, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA,
    0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA,

    /* Define huffman table (section B.2.4.1) */
    0xFF, 0xC4,		/* Marker */
    0x00, 0x1F,		/* Length (31 bytes) */
    0x01,		/* DC, table 1 */
    0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0A, 0x0B,

    /* Define huffman table (section B.2.4.1) */
    0xFF, 0xC4,		/* Marker */
    0x00, 0xB5,		/* Length (181 bytes) */
    0x11,		/* AC, table 1 */
    0x00, 0x02, 0x01, 0x02, 0x04, 0x04, 0x03, 0x04, 0x07, 0x05, 0x04, 0x04,
    0x00, 0x01, 0x02, 0x77, 0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
    0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81,
    0x08, 0x14, 0x42, 0x91, 0xA1, 0xB1, 0xC1, 0x09, 0x23, 0x33, 0x52, 0xF0,
    0x15, 0x62, 0x72, 0xD1, 0x0A, 0x16, 0x24, 0x34, 0xE1, 0x25, 0xF1, 0x17,
    0x18, 0x19, 0x1A, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x35, 0x36, 0x37, 0x38,
    0x39, 0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53, 0x54,
    0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
    0x69, 0x6A, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x82, 0x83,
    0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96,
    0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9,
    0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3,
    0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6,
    0xD7, 0xD8, 0xD9, 0xDA, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9,
    0xEA, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA,

    /* Start of Scan (section B.2.3) */
    0xFF, 0xDA,		/* Marker */
    0x00, 0x0C,		/* Length of header */
    0x03,		/* Number of image components in scan */
    0x01, 0x00,		/* Scan 1: use DC/AC huff tables 0/0 */
    0x02, 0x11,		/* Scan 2: use DC/AC huff tables 1/1 */
    0x03, 0x11,		/* Scan 3: use DC/AC huff tables 1/1 */
    0x00, 0x3F, 0x00	/* Not used */
};

/*
 * The next array represents the default JFIF header for
 * quality = 100 and size = 640x480. The values are
 * adjusted as the J_Movie header is read.  
 *
 * The buffer contains the jfif header without the huffman table
 */
static unsigned char jfifNoHuff100Header[] = {
    /* SOI */
    0xFF, 0xD8,

    /* JFIF header */
    0xFF, 0xE0,			/* Marker */
    0x00, 0x10,			/* Length = 16 bytes */
    0x4A, 0x46, 0x49, 0x46,	/* "JFIF" */
    0x00, 0x01, 0x01, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00,

    /* Start of frame (section B.2.2) */
    0xFF, 0xC0,		/* Baseline DCT */
    0x00, 0x11,		/* Length = 17 bytes */
    0x08,		/* Sample precision */
    0x01, 0xE0,		/* Height */
    0x02, 0x80,		/* Width */
    0x03,		/* Number of components = 3 */
    0x01, 0x21, 0x00,	/* Scan 1: 2:1 horiz, 1:1 vertical, use QT 0 */
    0x02, 0x11, 0x01,	/* Scan 2: 1:1 horiz, 1:1 vertical, use QT 1 */
    0x03, 0x11, 0x01,	/* Scan 3: 1:1 horiz, 1:1 vertical, use QT 1 */

    /* Define Quant table (section B.2.4.1) */
    0xFF, 0xDB,		/* Marker */
    0x00, 0x84,		/* Length (both tables) */
    0x00,		/* 8 bit values, table 0 */
    0x10, 0x0B, 0x0C, 0x0E, 0x0C, 0x0A, 0x10, 0x0E,
    0x0D, 0x0E, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
    0x1A, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
    0x1D, 0x28, 0x3A, 0x33, 0x3D, 0x3C, 0x39, 0x33,
    0x38, 0x37, 0x40, 0x48, 0x5C, 0x4E, 0x40, 0x44,
    0x57, 0x45, 0x37, 0x38, 0x50, 0x6D, 0x51, 0x57,
    0x5F, 0x62, 0x67, 0x68, 0x67, 0x3E, 0x4D, 0x71,
    0x79, 0x70, 0x64, 0x78, 0x5C, 0x65, 0x67, 0x63,

    0x01,		/* 8 bit values, table 1 */
    0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2F, 0x1A,
    0x1A, 0x2F, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
    0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,

    /* Start of Scan (section B.2.3) */
    0xFF, 0xDA,		/* Marker */
    0x00, 0x0C,		/* Length of header */
    0x03,		/* Number of image components in scan */
    0x01, 0x00,		/* Scan 1: use DC/AC huff tables 0/0 */
    0x02, 0x11,		/* Scan 2: use DC/AC huff tables 1/1 */
    0x03, 0x11,		/* Scan 3: use DC/AC huff tables 1/1 */
    0x00, 0x3F, 0x00	/* Not used */
};


/*
 *--------------------------------------------------------------
 *
 * CreateVidHeader --
 *
 *	Create a vid frame header, which is required for parallax.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *buf	(out)	Buffer with resulting header
 *	int width	(in)	Width of image
 *	int height	(in)	Height of image
 *      char *data      (in)    Pointer to data after any headers.
 *      int dataSize    (in)    Size of data without any headers
 *
 *--------------------------------------------------------------
 */
static int
CreateVidHeader (char **buf, int width, int height, char *data, int dataSize)
{
    static XPlxCImage vid_header;

    vid_header.width = width;
    vid_header.height = height;
    vid_header.size = dataSize;
    vid_header.data = data;
    vid_header.info = NULL;

    *buf = &vid_header;

    return sizeof(XPlxCImage);
}


/*
 *--------------------------------------------------------------
 *
 * CreateJfif --
 *
 *	Create a JFIF header, which is missing in some video
 *	formats (esp, the parallax format)
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *buf	(out)	Buffer with resulting header
 *	int width	(in)	Width of image
 *	int height	(in)	Height of image
 *	int qfactor	(in)	Quality factor for JPEG compression
 *
 *--------------------------------------------------------------
 */
static int
CreateJfif (char **buf, int width, int height, int qfactor)
{
	
    int i;
    int temp = 0;
    static char newHeader[sizeof(jfif100Header)];
    static int newHeaderQfactor = 100;
    static int init = 0;

    if (!init) {
	memcpy (newHeader, jfif100Header, sizeof(jfif100Header));
	init++;
    }

    /*
     * set image width in JFIF header
     */
    newHeader[27] = (char)(0xFF & (width >> 8));
    newHeader[28] = (char)(0xFF & width);

    /*
     * set image height in JFIF header
     */
    newHeader[25] = (char)(0xFF & (height >> 8));
    newHeader[26] = (char)(0xFF & height);

    /*
     * Adjust the quality factor.
     *
     * The default quality factor = 100, therefore, if
     * our quality factor does not equal 100 we must
     * scale the quantization matrices in the JFIF header
     * Note values are clipped to a max of 255
     */
    if (qfactor != newHeaderQfactor) {
        for (i=44; i<108; i++) {
            temp = (int)(jfif100Header[i]*qfactor)/100;
	    jfif100Header[i] = (char)((temp<255) ? temp : 255);
	}
	for (i=109; i<173; i++){
	    temp = (int)(jfif100Header[i]*qfactor)/100;
	    jfif100Header[i] = (char)((temp<255) ? temp : 255);
	}    
    }

    /*
     * Set result and return
     */
    *buf = newHeader;
    return (sizeof(jfif100Header));
}


/*
 *--------------------------------------------------------------
 *
 * CreateJfifNoHuff --
 *
 *	Create a JFIF header without the huffman table info (implies
 *	default huffman table).
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *buf	(out)	Buffer with resulting header
 *	int width	(in)	Width of image
 *	int height	(in)	Height of image
 *	int qfactor	(in)	Quality factor for JPEG compression
 *
 *--------------------------------------------------------------
 */
static int
CreateJfifNoHuff (char **buf, int width, int height, int qfactor)
{
	
    int i;
    int temp = 0;
    static char newHeader[sizeof(jfifNoHuff100Header)];
    static int newHeaderQfactor = 100;
    static int init = 0;

    if (!init) {
	memcpy (newHeader, jfifNoHuff100Header, sizeof(jfifNoHuff100Header));
	init++;
    }

    /*
     * set image width in JFIF header
     */
    newHeader[27] = (char)(0xFF & (width >> 8));
    newHeader[28] = (char)(0xFF & width);

    /*
     * set image height in JFIF header
     */
    newHeader[25] = (char)(0xFF & (height >> 8));
    newHeader[26] = (char)(0xFF & height);

    /*
     * Adjust the quality factor.
     *
     * The default quality factor = 100, therefore, if
     * our quality factor does not equal 100 we must
     * scale the quantization matrices in the JFIF header
     * Note values are clipped to a max of 255
     */
    if (qfactor != newHeaderQfactor) {
        for (i=44; i<108; i++) {
            temp = (int)(jfifNoHuff100Header[i]*qfactor)/100;
	    jfifNoHuff100Header[i] = (char)((temp<255) ? temp : 255);
	}
	for (i=109; i<173; i++){
	    temp = (int)(jfifNoHuff100Header[i]*qfactor)/100;
	    jfifNoHuff100Header[i] = (char)((temp<255) ? temp : 255);
	}    
    }

    /*
     * Set result and return
     */
    *buf = newHeader;
    return (sizeof(jfifNoHuff100Header));
}


/*
 *--------------------------------------------------------------
 *
 * JfifToNoHuff --
 *
 *	Helper function for JpegConvert: converts from
 *	JFIF_FRAME_CODE to JFIF_NO_HUFFMAN_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
JfifToNoHuff (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
}


/*
 *--------------------------------------------------------------
 *
 * JfifToVid --
 *
 *	Helper function for JpegConvert: converts from
 *	JFIF_FRAME_CODE to VID_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
JfifToVid (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
}


/*
 *--------------------------------------------------------------
 *
 * JfifToRaw --
 *
 *	Helper function for JpegConvert: converts from
 *	JFIF_FRAME_CODE to RAW_JPEG_NO_HEADERS_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
JfifToRaw (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
}


/*
 *--------------------------------------------------------------
 *
 * NoHuffToJfif --
 *
 *	Helper function for JpegConvert: converts from
 *	JFIF_NO_HUFFMAN_FRAME_CODE to JFIF_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
NoHuffToJfif (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
}


/*
 *--------------------------------------------------------------
 *
 * NoHuffToVid --
 *
 *	Helper function for JpegConvert: converts from
 *	JFIF_NO_HUFFMAN_FRAME_CODE to VID_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
NoHuffToVid (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
}


/*
 *--------------------------------------------------------------
 *
 * NoHuffToRaw --
 *
 *	Helper function for JpegConvert: converts from
 *	JFIF_NO_HUFFMAN_FRAME_CODE to
 *      RAW_JPEG_NO_HEADERS_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
NoHuffToRaw (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
}


/*
 *--------------------------------------------------------------
 *
 * VidToNoHuff --
 *
 *	Helper function for JpegConvert: converts from
 *	VID_FRAME_CODE to JFIF_NO_HUFFMAN_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
VidToNoHuff (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
    *headerSize = CreateJfifNoHuff (header, width, height, qfactor);
    *data = inData+20;
    *dataSize = inSize-20;
}


/*
 *--------------------------------------------------------------
 *
 * VidToJfif --
 *
 *	Helper function for JpegConvert: converts from
 *	VID_FRAME_CODE to JFIF_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
VidToJfif (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
    *headerSize = CreateJfif (header, width, height, qfactor);
    *data = inData+20;
    *dataSize = inSize-20;
}


/*
 *--------------------------------------------------------------
 *
 * VidToRaw --
 *
 *	Helper function for JpegConvert: converts from
 *	VID_FRAME_CODE to RAW_JPEG_NO_HEADERS_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
VidToRaw (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
    *headerSize = 0;
    *header = NULL;
    *data = inData+20;
    *dataSize = inSize-20;
}


/*
 *--------------------------------------------------------------
 *
 * RawToNoHuff --
 *
 *	Helper function for JpegConvert: converts from
 *	RAW_JPEG_NO_HEADERS_CODE to JFIF_NO_HUFFMAN_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
RawToNoHuff (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
    *headerSize = CreateJfifNoHuff (header, width, height, qfactor);
    *data = inData;
    *dataSize = inSize;
}


/*
 *--------------------------------------------------------------
 *
 * RawToJfif --
 *
 *	Helper function for JpegConvert: converts from
 *	RAW_JPEG_NO_HEADERS_CODE to JFIF_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
RawToJfif (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
    *headerSize = CreateJfif (header, width, height, qfactor);
    *data = inData;
    *dataSize = inSize;
}


/*
 *--------------------------------------------------------------
 *
 * RawToVid --
 *
 *	Helper function for JpegConvert: converts from
 *	RAW_JPEG_NO_HEADERS_CODE to VID_FRAME_CODE
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See JpegConvert
 *
 *--------------------------------------------------------------
 */
static void
RawToVid (char *inData, int inSize, int width, int height, int qfactor,
	      char **header, int *headerSize, char **data, int *dataSize)
{
    *headerSize = CreateVidHeader (header, width, height, inData, inSize);
    *data = inData;
    *dataSize = inSize;
}


/*
 *--------------------------------------------------------------
 *
 * JpegConvert --
 *
 *	Find the header necessary to convert between video subformats.
 *	Two pointers are returned, one to the JPEG header (which might be
 *	NULL, depending on the format) and one to the JPEG data.  The
 *	concatenation of the header and the data results in a JPEG structure
 *	in the subformat requested.
 *
 * Results:
 *	1 for success, 0 for failure (unknown format).
 *
 * Side effects:
 *	The pointer returned is a pointer to a read-only buffer.
 *	It should not be modified by the caller, but is guarenteed
 *	not to be modified until the next call to JpegConvert()
 *
 * Parameters:
 *
 *   char *inData	(in)	Data to be "converted"
 *   int inSize		(in)	Number of valid data bytes
 *   int width		(in)	Width of image, in pixels
 *   int height		(in)	Height of image, in pixels
 *   int qfactor	(in)	Qfactor (depends on sub-format)
 *   int inFormat	(in)	Current format of JPEG data
 *   int outFormat	(in)	Desired format of JPEG data
 *   char **header	(out)	Pointer to new header
 *   int *headerSize	(out)	Number of valid bytes in header
 *   char **data	(out)	Pointer to new data
 *   int *dataSize	(out)	Number of valid bytes in data
 *
 *--------------------------------------------------------------
 */
int
JpegConvert (char *inData, int inSize, int width, int height, int qfactor,
	     int inFormat, int outFormat, char **header, int *headerSize,
	     char **data, int *dataSize)
{
    /*
     * Special case: no change
     */
    if (inFormat == outFormat) {
	*header = NULL;
	*headerSize = 0;
	*data = inData;
	*dataSize = inSize;
	return 0;
    }

    /*
     * Need to convert:  Figure out what function to call, and do it!
     */
    switch (inFormat) {
    case JFIF_FRAME_CODE:
	if (outFormat == JFIF_NO_HUFFMAN_FRAME_CODE) {
	    JfifToNoHuff (inData, inSize, width, height, qfactor,
			  header, headerSize, data, dataSize);
	} else if (outFormat == VID_FRAME_CODE) {
	    JfifToVid (inData, inSize, width, height, qfactor,
		       header, headerSize, data, dataSize);
	} else if (outFormat == RAW_JPEG_NO_HEADERS_CODE) {
	    JfifToRaw (inData, inSize, width, height, qfactor,
		       header, headerSize, data, dataSize);
	} else {
	    return 0;
	}
	break;

    case JFIF_NO_HUFFMAN_FRAME_CODE:
	if (outFormat == JFIF_FRAME_CODE) {
	    NoHuffToJfif (inData, inSize, width, height, qfactor,
			  header, headerSize, data, dataSize);
	} else if (outFormat == VID_FRAME_CODE) {
	    NoHuffToVid (inData, inSize, width, height, qfactor,
		       header, headerSize, data, dataSize);
	} else if (outFormat == RAW_JPEG_NO_HEADERS_CODE) {
	    NoHuffToRaw (inData, inSize, width, height, qfactor,
		       header, headerSize, data, dataSize);
	} else {
	    return 0;
	}
	break;

    case VID_FRAME_CODE:
	if (outFormat == JFIF_NO_HUFFMAN_FRAME_CODE) {
	    VidToNoHuff (inData, inSize, width, height, qfactor,
			  header, headerSize, data, dataSize);
	} else if (outFormat == JFIF_FRAME_CODE) {
	    VidToJfif (inData, inSize, width, height, qfactor,
		       header, headerSize, data, dataSize);
	} else if (outFormat == RAW_JPEG_NO_HEADERS_CODE) {
	    VidToRaw (inData, inSize, width, height, qfactor,
		       header, headerSize, data, dataSize);
	} else {
	    return 0;
	}
	break;

    case RAW_JPEG_NO_HEADERS_CODE:
	if (outFormat == JFIF_NO_HUFFMAN_FRAME_CODE) {
	    RawToNoHuff (inData, inSize, width, height, qfactor,
			  header, headerSize, data, dataSize);
	} else if (outFormat == JFIF_FRAME_CODE) {
	    RawToJfif (inData, inSize, width, height, qfactor,
		       header, headerSize, data, dataSize);
	} else if (outFormat == VID_FRAME_CODE) {
	    RawToVid (inData, inSize, width, height, qfactor,
		       header, headerSize, data, dataSize);
	} else {
	    return 0;
	}
	break;

    }
    return 1;
}
