/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmJPEGSoftwareDevice.c,v 1.37 1996/02/07 22:00:10 kpatel Exp $
 *
 * Device to work with the CM MjpegPlay object to decode
 * a motion JPEG stream in software.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <tcl.h>
#include <tk.h>
#include <dp.h>
#include "cmMJPEGPlay.h"
#include "cmBuffer.h"

#ifdef JPEG_SW

#include "cmUtil.h"
#include "cmDither.h"
#include "cmBufferlist.h"
#include "cmLTS.h"

#ifdef _WINDOWS
extern HDC tcm_DitherHDC;
#endif

extern void init_pre_idct();
extern void j_rev_dct();

#define BP_JPEG_FLAG
#include "cmBitParse.h"
#undef BP_JPEG_FLAG

#include <X11/Xlib.h>

#ifdef HAVE_SHARED_MEMORY
#  include <sys/ipc.h>
#  include <sys/shm.h>
#  include <X11/extensions/XShm.h>
#endif

/*
 * There is one of the following structs for every
 * unique software device created
 */
typedef struct JPEG_SW_Device {
    int indecode;     	/* Is a decode currently in progress? */

    /*
     * Stuff needed for display
     */
    Window win;
    Display *dpy;
    GC gc;
    int hasShm;
#ifdef HAVE_SHARED_MEMORY
    XShmSegmentInfo shminfo;
#endif
#ifdef _WINDOWS
    BITMAPINFO *image;
    unsigned char *image_data;
#else
    XImage *image;	/* The current frame */
#endif
    int waiting;	/* Waiting for shared memory completion event */
    int ShmEvent;	/* X Event number for shared memory events */

    /*
     * Stuff needed for decoding
     */
    unsigned short *ldht;	/* Luminance DC Huffman Table */
    unsigned short *laht;	/* Luminance AC Huffman Table */
    unsigned short *cdht;	/* Chrominance DC Huffman Table */
    unsigned short *caht;	/* Chrominance AC Huffman Table */

    int qfactor;      		/* Q-Factor for current Quantization Table */
    unsigned char lumqt[64];	/* Luminance Quantization Table */
    unsigned char chqt[64];	/* Chrominance Quantization Table */

    DitherFunc ditherFunc;	/* Routine to do dithering */
    DitherRec *ditherRec;	/* Information for ditherFunc */

    /*
     * Stuff for estimating how long it takes to decode
     * frames which is then used for scheduling.  The model is
     * time = m * (bytes) + b.  We calculate m and b using a
     * weighted linear regression scheme.
     */
    double n, sx, sy, sxy, sxx;	/* Stats on measured variables in the
				 * regression analysis */
    double m, b;		/* Regression coefficients */
    double secMcu;   		/* How long does it take to decode a block */
    double secFrame;		/* Time it took to decode/dither this frame */
    double secIdle;		/* How long does it take to get called again
				   after giving up the cpu during a decode */
    double lastTime;		/* Temporary time variable */

    /*
     * Information about previous frame
     */
    Buffer *prevBuffer;
    int prevWidth;
    int prevHeight;

} JPEG_SW_Device;

/*
 * Instances of the following struct are created
 * for every frame that is decoded and put in to
 * the deviceData slot in the MJPEGFrame structure
 */
typedef struct {
    short block[64];		/* Block currently being decoded */
    short yPredictor;		/* Predictor for DC component of Y plane */
    short crPredictor;		/* Predictor for DC component of Cr plane */
    short cbPredictor;		/* Predictor for DC component of Cb plane*/
    int mcuNum;			/* Which MCU we are currently decoding */

    Buffer *decodeBuffer;	/* Buffer used temporarily to hold
				 * Y-Cr-Cb planes (may be NULL if the
				 * same buffer as being used as the
				 * viewBuffer */
    unsigned char *y;		/* Y plane */
    unsigned char *cr;		/* Cr plane */
    unsigned char *cb;		/* Cb plane */
    
    Buffer *viewBuffer;		/* Buffer holding the dithered image */
    unsigned char *dithered;	/* The final dithered image */
} DecodeInfo;

/*
 * Prototypes for internal routines
 */
static int JPEG_SW_NewFrame(MJPEGFrame *framePtr);
static unsigned short *JPEG_SW_CreateHufftab(unsigned char *bits,
					     unsigned char *vals);
static void JPEG_SW_FreeHufftab(unsigned short *table);
static int ProcessShmEvent (ClientData clientData, XEvent *e);
static int LookForShmEvent (Display *display, XAnyEvent *eventPtr, char *arg);
static void WaitForCompletion (JPEG_SW_Device *swDevice);

typedef enum { INIT_DITHER_FAILED = 1, NEW_FRAME_FAILED, BAD_PRECISION,
		   BAD_GEOMETRY, BAD_DECIMATION, SHM_ERROR } JPEG_SW_ErrorValue;

/*
 * Standard Quantization Tables
 */
static unsigned char JPEG_SW_StandardLumQT[] = { 
    16, 11, 10, 16, 24, 40, 51, 61,
    12, 12, 14, 19, 26, 58, 60, 55,
    14, 13, 16, 24, 40, 57, 69, 56,
    14, 17, 22, 29, 51, 87, 80, 62,
    18, 22, 37, 56, 68, 109, 103, 77,
    24, 35, 55, 64, 81, 104, 113, 92,
    49, 64, 78, 87, 103, 121, 120, 101,
    72, 92, 95, 98, 112, 100, 103, 99 };

static unsigned char JPEG_SW_StandardChQT[] = {
    17, 18, 24, 47, 99, 99, 99, 99,
    18, 21, 26, 66, 99, 99, 99, 99,
    24, 26, 56, 99, 99, 99, 99, 99,
    47, 66, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99 };

/*
 * Standard Huffman Tables
 */
static unsigned char JPEG_LDC_HuffBits[]
    = { 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 };
static unsigned char JPEG_LDC_HuffVals[]
    = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };

static unsigned char JPEG_CDC_HuffBits[]
    = { 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 };
static unsigned char JPEG_CDC_HuffVals[]
    = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };

static unsigned char JPEG_LAC_HuffBits[]
    = { 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d };
static unsigned char JPEG_LAC_HuffVals[] = {
    0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
    0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
    0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
    0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
    0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
    0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
    0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
    0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
    0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
    0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
    0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
    0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
    0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
    0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
    0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
    0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
    0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
    0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
    0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
    0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
    0xf9, 0xfa };

static unsigned char JPEG_CAC_HuffBits[]
    = { 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77 };
static unsigned char JPEG_CAC_HuffVals[] = {
    0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
    0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
    0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
    0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
    0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
    0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
    0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
    0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
    0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
    0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
    0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
    0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
    0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
    0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
    0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
    0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
    0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
    0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
    0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
    0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
    0xf9, 0xfa };

/*
 * Bounds checking array for after DCT
 * (biased by 16)
 */
static unsigned char JPEG_BoundsFix[] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
    17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32,
    33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
    49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64,
    65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80,
    81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96,
    97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108,
    109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120,
    121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132,
    133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144,
    145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156,
    157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168,
    169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180,
    181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192,
    193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204,
    205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216,
    217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228,
    229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240,
    241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252,
    253, 254, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 };


/*
 *--------------------------------------------------------------
 *
 * ProcessShmEvent --
 *
 *	Registered as a Tk generic handler to unset the
 *	waiting flag on a shared memory event so that we
 *	don't miss shared memory events
 *
 * Results:
 *	1 if we handle the event, 0 otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static int
ProcessShmEvent (ClientData clientData, XEvent *e)
{
    MJPEGResource *objPtr;
    JPEG_SW_Device *swDevice;
    XAnyEvent *shmEventPtr = (XAnyEvent *) e;

    objPtr = (MJPEGResource *) clientData;
    swDevice = (JPEG_SW_Device *) objPtr->devicePtr->deviceData;
    
    if (swDevice->waiting == 0) {
	return 0;
    } else if (objPtr->xid == 0) {
	swDevice->waiting = 0;
	return 0;
    } else if ((swDevice->waiting)
	       && (shmEventPtr->type == swDevice->ShmEvent)
	       && (shmEventPtr->window = objPtr->xid)) {
	swDevice->waiting = 0;
	return 1;
    }
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * LookForShmEvent --
 *
 *	Helper function for XIfEvent call  that waits for the
 *	shared memory operation to complete.
 *
 * Results:
 *	True (1) if event is a ShmEvent for the right window, 
 *	False (0) otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static int
LookForShmEvent (Display *display, XAnyEvent *eventPtr, char *arg)
{
    JPEG_SW_Device *swDevice;

    swDevice = (JPEG_SW_Device *) arg;

    return ((eventPtr->type == swDevice->ShmEvent) && 
	    (eventPtr->window == swDevice->win));
}


/*
 *--------------------------------------------------------------
 *
 * WaitForCompletion --
 *
 *	If we've issued a shared memory put image call, block
 *	until X server sends us a message that it's complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Will block until appropriate shared memory event is
 *	returned from X server.
 *
 *--------------------------------------------------------------
 */
static void
WaitForCompletion (JPEG_SW_Device *swDevice)
{
#ifdef HAVE_SHARED_MEMORY
    XEvent shmEvent;

    if (!swDevice->win) {
	swDevice->waiting = 0;
	return;
    }
    if (swDevice->waiting) {
	XIfEvent(swDevice->dpy, &shmEvent, LookForShmEvent, (char *) swDevice);
	swDevice->waiting = 0;
    }
#else
    swDevice->waiting = 0;
#endif
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_Init --
 *
 *	Initialize the device.
 *
 * Results:
 *	0 for successful initilization, non-zero error code otherwise
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */
int
JPEG_SW_Init(MJPEGResource *rsrcPtr)
{
    JPEG_SW_Device *swDevice;
#ifdef HAVE_SHARED_MEMORY
    XShmSegmentInfo si;
#endif
    Tk_ErrorHandler handler;
    int result = 0;
    DitherParam ditherInfo;

    swDevice = (JPEG_SW_Device *) rsrcPtr->devicePtr->deviceData;

    swDevice->indecode = 0;

    swDevice->win = rsrcPtr->xid;
    swDevice->dpy = Tcm_Display();
    swDevice->gc = XCreateGC(swDevice->dpy, rsrcPtr->xid, 0, 0);
    XSetForeground(swDevice->dpy, swDevice->gc, 0);
    swDevice->image = NULL;
    swDevice->waiting = 0;

#ifdef HAVE_SHARED_MEMORY
    if (cmXShmQueryExtension(Tcm_Display())) {
	swDevice->hasShm = 1;
	swDevice->ShmEvent = XShmGetEventBase(swDevice->dpy)+ShmCompletion;
	Tk_CreateGenericHandler(ProcessShmEvent, (ClientData) rsrcPtr);
    } else {
	swDevice->hasShm = 0;
    }
#else
    swDevice->hasShm = 0;
#endif

    swDevice->ldht = JPEG_SW_CreateHufftab(JPEG_LDC_HuffBits,
					   JPEG_LDC_HuffVals);
    swDevice->laht = JPEG_SW_CreateHufftab(JPEG_LAC_HuffBits,
					   JPEG_LAC_HuffVals);
    swDevice->cdht = JPEG_SW_CreateHufftab(JPEG_CDC_HuffBits,
					   JPEG_CDC_HuffVals);
    swDevice->caht = JPEG_SW_CreateHufftab(JPEG_CAC_HuffBits,
					   JPEG_CAC_HuffVals);

    swDevice->qfactor = 100;
    memcpy(swDevice->lumqt, JPEG_SW_StandardLumQT,
	   sizeof(JPEG_SW_StandardLumQT));
    memcpy(swDevice->chqt, JPEG_SW_StandardChQT,
	   sizeof(JPEG_SW_StandardChQT));

    swDevice->n = 0.0;
    swDevice->sx = 0.0;
    swDevice->sy = 0.0;
    swDevice->sxy = 0.0;
    swDevice->sxx = 0.0;
    swDevice->m = 0.0;
    swDevice->b = 0.3;	/* Conservative guess -- 300 ms to decode a picture */

    swDevice->secMcu = 0.0;
    swDevice->secFrame = 0.0;
    swDevice->secIdle = 0.0;
    swDevice->lastTime = 0.0;

    swDevice->prevBuffer = NULL;

    /*
     * Initialize the dithering module
     */
    ditherInfo.dType = CHOOSE_BEST;
    ditherInfo.minColors = 2;
    ditherInfo.maxColors = 256;
    ditherInfo.pFormat = YCRCB422;

    swDevice->ditherRec = InitDither(ditherInfo, swDevice->dpy,
				     swDevice->win,
				     &(swDevice->ditherFunc));
    if (swDevice->ditherRec == NULL) {
	result = INIT_DITHER_FAILED;
    }

    /*
     * Init the reverse DCT module
     */
    init_pre_idct();

    return result;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_Reinit --
 *
 *	Reinitialize the software device.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
JPEG_SW_Reinit(MJPEGResource *rsrcPtr)
{
    JPEG_SW_Device *swDevice;

    swDevice = (JPEG_SW_Device *) rsrcPtr->devicePtr->deviceData;

    if (rsrcPtr->xid != swDevice->win) {
	swDevice->win = rsrcPtr->xid;

	WaitForCompletion(swDevice);
	
	if (swDevice->gc)  {
	    XFreeGC(swDevice->dpy, swDevice->gc);
	}
	swDevice->gc = XCreateGC(swDevice->dpy, rsrcPtr->xid, 0, 0);
	XSetForeground(swDevice->dpy, swDevice->gc, 0);
	
	swDevice->waiting = 0;    
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_CreateHufftab
 *
 *	Convert a list of bits and values for a Huffman
 *	table in to a table that can be used in decoding.
 *
 * Results:
 *	The table itself is returned.  It should be freed
 *	with JPEG_SW_FreeHufftab
 *
 * Side effects:
 *	Memory is allocated for the table
 *
 *--------------------------------------------------------------
 */
static unsigned short *
JPEG_SW_CreateHufftab(unsigned char *bits, unsigned char *vals)
{
    int symbols, sizes[256], codes[256], code, i, j;
    unsigned short *table;

    /*
     * Allocate some memory
     */
    table = (unsigned short *) ckalloc(65536 * sizeof(short));
    if (table == NULL) {
	return NULL;
    }
    
    memset(table, 0, 65536 * sizeof(short) );

    /* 
     * Count the total number of symbols
     * and generate the codes
     */
    symbols = 0;
    code = 0;
    for (i=0; i<16; i++) {
	for (j=0; j<(int)(bits[i]); j++) {
	    codes[symbols] = code++;
	    sizes[symbols] = i + 1;
	    symbols++;
	}
	code <<= 1;
    }
    
    /*
     * Generate the values in the final table
     */
    for (i=0; i<symbols; i++) {
	int val, index, bits;
	
	val = (sizes[i] << 8) | vals[i];
	bits = 16 - sizes[i];
	index = codes[i] << bits;
	
	for (j=0; j< 1<<bits; j++) {
	    table[index | j] = val ;
	}
    }

    return table;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_FreeHufftab
 *
 *	Destory a Huffman table created with
 *	JPEG_SW_CreateHufftab
 *
 * Results:
 *	No return value
 *
 * Side effects:
 *	Memory is freed
 *
 *--------------------------------------------------------------
 */
static void
JPEG_SW_FreeHufftab(unsigned short *table)
{
    if (table != NULL) {
	ckfree((char *) table);
    }
}


/*
 * The following macro decodes a number is the format
 * that jpeg stores it.  Basically, negative numbers
 * are stored as a difference from minint
 */
#define JPEG_EXTEND(v, t) \
  ((v) < (1 << (t-1) ) ? (v) + (((-1) << t) + 1) : (v))

static const char zz[] = { 0, 1, 8, 16, 9, 2, 3, 10, 17, 24, 32, 25, 18,
			       11, 4, 5, 12, 19, 26, 33, 40, 48, 41, 34,
			       27, 20, 13, 6, 7, 14, 21, 28, 35, 42, 49,
			       56, 57, 50, 43, 36, 29, 22, 15, 23, 30,
			       37, 44, 51, 58, 59, 52, 45, 38, 31, 39,
			       46, 53, 60, 61, 54, 47, 55, 62, 63 };

/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_Decode
 *
 *	Start the decoding process for a frame.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	This function will be periodcially called until frame
 *	decode is complete
 *
 *--------------------------------------------------------------
 */
int
JPEG_SW_Decode (MJPEGFrame *framePtr) 
{
    MJPEGResource *objPtr;
    MJPEGDevice *devicePtr;
    JPEG_SW_Device *swDevice;
    DecodeInfo *dInfo;
    double start, end;
    int mcu, firstmcu, lastmcu, maxmcu;
    int w, h;
    int ditherStride;

    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    swDevice = (JPEG_SW_Device *)devicePtr->deviceData;
    dInfo = (DecodeInfo *) framePtr->deviceData;
    
    w = framePtr->frameHeader.width;
    h = framePtr->frameHeader.height;
    
    devicePtr->asyncDecodeFlag = 1;

    /*
     * Update timing estimates if we can
     */
    end = Tcm_ReadSysClock();
    if ((swDevice->indecode) && (swDevice->lastTime != 0.0)) {
	if (swDevice->secIdle == 0.0) {
	    swDevice->secIdle = end - swDevice->lastTime;
	} else {
	    swDevice->secIdle = (0.95*swDevice->secIdle)
		+ 0.05*(end - swDevice->lastTime);
	}
    }
    
    /*
     * Initialize data structures if needed
     */
    if (swDevice->indecode == 0) {
	int marker;
	
	/* 
	 * Keep track of how much time it takes to do this frame.
	 */
	swDevice->secFrame = 0.0;

	if (JPEG_SW_NewFrame(framePtr) != 0) {
	    return NEW_FRAME_FAILED;
	}
	swDevice->indecode = 1;
	dInfo = (DecodeInfo *) framePtr->deviceData;

	/*
	 * Handle any jpeg markers in the stream
	 */
	bp_Peek(8, &marker);
	while (marker == 0xff) {
	    bp_Flush(8);
	    bp_Get(8, &marker);

	    /*
	     * Right now, don't handle any markers
	     * so just drop through to ignore the marker
	     */
	    switch (marker) {
	      case 0xc0:
		{
		    /* Start of Frame tag */
		    int value, i;
		    
		    /* Flush the length header */
		    bp_Flush(16);
		    
		    /*
		     * Check sample precision
		     */
		    bp_Get(8, &value);
		    if (value != 8) {
			return BAD_PRECISION;
		    }
		    
		    /*
		     * Check w and h
		     */
		    bp_Get(16, &value);
		    if (value != h) {
			return BAD_GEOMETRY;
		    }
		    
		    bp_Get(16, &value);
		    if (value != w) {
			return BAD_GEOMETRY;
		    }
		    
		    /*
		     * Check number of components
		     */
		    bp_Get(8, &value);
		    if (value != 3) {
			return BAD_DECIMATION;
		    }
		    for (i=0; i<3; i++) {
			/* Ignore ID number */
			bp_Flush(8);
		      
			/* Check decimation */
			bp_Get(8, &value);
			if (((i == 0) && (value != 0x21))
			    || ((i > 0) && (value != 0x11))) {
			    return BAD_DECIMATION;
			}
		      
			/* Get Q table id and ignore it*/
			bp_Flush(8);
		    }
		}
		break;
		
	      case 0xc4:	/* Define Huffman Table */
	      case 0xda:	/* Start of Scan */
	      case 0xdb:	/* Define Quant Table */
	      case 0xdd:	/* Define Restart Interval */
		/*
		 * Ignore all of those codes, skipping the contents
		 */
		bp_Peek(16, &marker);
		bp_Flush(8 * marker);
		break;

	      case 0xd8:	/* Start of Image */
	      case 0xd9:	/* End of Image */
		/*
		 * Ignore these codes, and don't even
		 * look for length
		 */
		break;
		
	      default:
		/*
		 * Unrecognized code
		 * Skip the contents and ignore it
		 */
		bp_Peek(16, &marker);
		bp_Flush(8 * marker);
		break;
	    }
	    
	    bp_Peek(8, &marker);
	}
    }
    
    /*
     * Check whether quantization tables have changed
     */
    if (framePtr->frameHeader.qfactor != swDevice->qfactor) {
	int i, qfactor;
	
	qfactor = framePtr->frameHeader.qfactor;
	
	for (i=0; i<64; i++) {
	    swDevice->lumqt[i] = (int)(JPEG_SW_StandardLumQT[i])*qfactor / 100;
	    swDevice->chqt[i] = (int)(JPEG_SW_StandardChQT[i]) * qfactor / 100;
	}
	
	swDevice->qfactor = framePtr->frameHeader.qfactor;
    }
    
    /*
     * Calculate how many mcus to decode before returning
     */
    firstmcu = mcu = dInfo->mcuNum;
    if (swDevice->secMcu != 0.0) {
	lastmcu = mcu + ((int) (0.010 / swDevice->secMcu));
    } else {
	lastmcu = mcu + 100;
    }

    /* Always decode at least 1 block */
    if (lastmcu == mcu) {
	lastmcu++;
    }

    maxmcu = (w/16) * (h/8);
    if (lastmcu > maxmcu) {
	lastmcu = maxmcu;
    }
    
    start = Tcm_ReadSysClock();
    
    if (swDevice->ditherRec->depth == 8) {
	ditherStride = w;
#ifdef _WINDOWS
    } else if (swDevice->ditherRec->depth == 24) {
        ditherStride = (((w*3) + 3) / 4) * 4;
#endif
    } else {
	ditherStride = w << 2;
    }
    while (mcu < lastmcu) {
	short block[64];
	unsigned short *dcht;
	unsigned short *acht;
	unsigned char *qt;
	short *predictor;
	int stride;

	int piece;
	int code, value;
	unsigned char difflen;
	signed char diff;
	int k;
	unsigned char mixed, run, coefflen, coeff;
	unsigned char *dest;
	int mcux, mcuy;
	unsigned char y[8*16], cr[8*8], cb[8*8];

	/*
	 * Each MCU consists of 2 pieces -- 2 Y, 1 Cr, and 1 Cb
	 * If we want to change this to handle different
	 * decimation, this would be one of the first things
	 * to modify
	 */
	mcux = mcu % (w / 16);
	mcuy = mcu / (w / 16);
	for (piece = 0; piece < 4; piece++) {
	    /*
	     * Zero the piece we are about to decode
	     */
	    memset(block, 0, sizeof(block));

	    /*
	     * Calculate everything specific to which plane we
	     * are decoding to avoid future conditionals
	     */

	    if (piece < 2) {
		dcht = swDevice->ldht;
		acht = swDevice->laht;
		qt = swDevice->lumqt;
		predictor = &dInfo->yPredictor;

		dest = y + (8 * piece);
		stride = 16;
	    } else if (piece == 2) {
		dcht = swDevice->cdht;
		acht = swDevice->caht;
		qt = swDevice->chqt;
		predictor = &dInfo->cbPredictor;

		dest = cb;
		stride = 8;
	    } else {
		dcht = swDevice->cdht;
		acht = swDevice->caht;
		qt = swDevice->chqt;
		predictor = &dInfo->crPredictor;
		
		dest = cr;
		stride = 8;
	    }
	    
	    /*
	     * Get the DC value
	     */
	    bp_Peek(16, &code);
	    value = dcht[code];
	    bp_Flush(value >> 8);
	    difflen = value & 0xff;
	    if (difflen) {
		bp_Get(difflen, &diff);
		diff = JPEG_EXTEND(diff, difflen);
		*predictor += diff;
	    }
	    block[0] = (*predictor) * qt[0];

	    /*
	     * Get the AC values
	     */
	    for (k=1; k<64; ) {
		bp_Peek(16, &code);
		value = acht[code];
		bp_Flush(value >> 8);
		mixed = value & 0xff;
		run = mixed >> 4;
		coefflen = mixed & 0xf;
		if (coefflen == 0) {
		    if (run == 15) {
			k += 16;
		    } else {
			k = 64;
		    }
		} else {
		    k += run;
		    bp_Get(coefflen, &coeff);
		    block[zz[k]] = (JPEG_EXTEND(coeff, coefflen)) * qt[zz[k]];
		    k++;
		}
	    }

	    /*
	     * Do Reverse DCT and put the data where it belongs
	     */
	    j_rev_dct(block);
	    {
		int row;
		unsigned char *outPtr;
		short *inPtr;
		unsigned char *BoundsFix = JPEG_BoundsFix + 128 + 16;
		
		inPtr = block;
		outPtr = dest;
		for (row=0; row<8; row++) {
		    outPtr[0] = BoundsFix[inPtr[0]];
		    outPtr[1] = BoundsFix[inPtr[1]];
		    outPtr[2] = BoundsFix[inPtr[2]];
		    outPtr[3] = BoundsFix[inPtr[3]];
		    outPtr[4] = BoundsFix[inPtr[4]];
		    outPtr[5] = BoundsFix[inPtr[5]];
		    outPtr[6] = BoundsFix[inPtr[6]];
		    outPtr[7] = BoundsFix[inPtr[7]];
		    
		    outPtr += stride;
		    inPtr += 8;
		}
	    }
	}
        swDevice->ditherFunc(swDevice->ditherRec, 2*8*mcux, 8*mcuy,
                         16, 8, y, cr, cb,
                         dInfo->dithered,
                         16, 8, 8, ditherStride,
                         w, h);
	mcu++;
    }
    end = Tcm_ReadSysClock();
    dInfo->mcuNum = mcu;
    swDevice->secFrame += end-start;
    
    /*
     * Update time estimates per mcu if possible
     */
    if (mcu != firstmcu) {
	if (swDevice->secMcu == 0.0) {
	    swDevice->secMcu = (end - start)/((double) (mcu - firstmcu));
	} else {
	    swDevice->secMcu = (0.8 * swDevice->secMcu)
		+ (0.2 * ((end - start) / ((double) (mcu - firstmcu))));
	}
    }
    
    /*
     * Signal decode is done if it is...
     */
    if (mcu == maxmcu) {
	float x, y;

	/*
	 * We've finished decoding so unset the
	 * async decode flag
	 */
	devicePtr->asyncDecodeFlag = 0;
	swDevice->indecode = 0;

	/*
	 * Update the statistics of how long it took to decode
	 */
	swDevice->n = (1+swDevice->n)*0.9;
	x = framePtr->dataLength;
	y = swDevice->secFrame;
	swDevice->sx = swDevice->sx*0.9 + 0.1*x;
	swDevice->sxx = swDevice->sxx*0.9 + 0.1*x*x;
	swDevice->sy = swDevice->sy*0.9 + 0.1*y;
	swDevice->sxy = swDevice->sxy*0.9 + 0.1*x*y;
	swDevice->m = (swDevice->n*swDevice->sxy - swDevice->sx*swDevice->sy) /
		      (swDevice->n*swDevice->sxx - swDevice->sx*swDevice->sx);
	swDevice->b = (swDevice->sy - swDevice->m*swDevice->sy) / swDevice->n;
    }
    
    swDevice->lastTime = Tcm_ReadSysClock();
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_NewFrame
 *
 * Create a new device structure to encapsulate the
 * state of a decode currently in progres
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Allocates memory 
 *
 *--------------------------------------------------------------
 */
static int
JPEG_SW_NewFrame(MJPEGFrame *framePtr)
{
    MJPEGResource *objPtr;
    MJPEGDevice *devicePtr;
    JPEG_SW_Device *swDevice;
    DecodeInfo *dInfo;
    int size, finalsize;
    Buffer *tempBuffer;
    signed char *data;

    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    swDevice = (JPEG_SW_Device *)devicePtr->deviceData;

    /*
     * Find the actual beginning of the bitstream in
     * the data
     */
    data = framePtr->data;

    if (framePtr->frameHeader.subformat == VID_FRAME_CODE) {
	data += 20;
    } else if (framePtr->frameHeader.subformat == RAW_JPEG_NO_HEADERS_CODE) {
	/* Don't need to do anything. */
    } else if (framePtr->frameHeader.subformat == JFIF_NO_HUFFMAN_FRAME_CODE) {
	/* Don't need to do anything... */
    } else if (framePtr->frameHeader.subformat == JFIF_FRAME_CODE) {
	/* Handle custom Huffman tables? */
    }
    
    /*
     * Allocate and initialize a DecodeInfo structure
     */
    dInfo = (DecodeInfo *) ckalloc(sizeof(DecodeInfo));
    framePtr->deviceData = (ClientData) dInfo;

    /*
     * Initialize the bit parser, the current mcu
     * counter, and the dc predictor
     */
    bp_Start(data, 0);
    dInfo->yPredictor = 0;
    dInfo->crPredictor = 0;
    dInfo->cbPredictor = 0;
    dInfo->mcuNum = 0;

    /*
     * Allocate temporary space for the Y, Cr, and Cb
     * planes and the final dithered image
     */
    size = framePtr->frameHeader.width * framePtr->frameHeader.height;
    if (swDevice->ditherRec->depth == 8) {
	finalsize = size;
    } else {
	finalsize = 4 * size;
    }
    
    if (swDevice->hasShm) {
	tempBuffer = MakeBuffer(objPtr->interp, size*2, "jpegSw:NewFrame");
	dInfo->decodeBuffer = tempBuffer;
	dInfo->y = tempBuffer->data;
	dInfo->cr = tempBuffer->data + size;
	dInfo->cb = tempBuffer->data + (size * 3/2);
	
	tempBuffer = MakeSharedBuffer(objPtr->interp, finalsize,
				      "jpegSw:NewFrame");
	if (tempBuffer == NULL) {
	    tempBuffer = MakeBuffer(objPtr->interp, finalsize,
				    "jpegSw:NewFrame");
	}

	dInfo->viewBuffer = tempBuffer;
	dInfo->dithered = tempBuffer->data;
    } else {
	tempBuffer = MakeBuffer(objPtr->interp, (size * 2) + finalsize,
				"jpegSw:NewFrame");
	dInfo->decodeBuffer = NULL;
	dInfo->viewBuffer = tempBuffer;

	dInfo->y = (unsigned char *)tempBuffer->data;
	dInfo->cr = (unsigned char *)tempBuffer->data + size;
	dInfo->cb = (unsigned char *)tempBuffer->data + (size * 3/2);
	dInfo->dithered = (unsigned char *)tempBuffer->data + (size * 2);
    }

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_Display --
 *
 *	Toss a decoded frame up on the screen.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Could block if previous shared memory put image hasn't returned.
 *
 *--------------------------------------------------------------
 */
int
JPEG_SW_Display (MJPEGFrame *framePtr)
{
    MJPEGResource *objPtr;
    JPEG_SW_Device *swDevice;
    DecodeInfo *dInfo;
    int imWidth, imHeight;
    int winWidth, winHeight;
    int xpos, ypos;
    int depth;
    Display *dpy;
#ifdef _WINDOWS
    BITMAPINFO *image;
#else
    XImage *image = NULL;	/* Quiet gcc */
#endif
    
    objPtr = framePtr->objPtr;
    swDevice = (JPEG_SW_Device *) framePtr->objPtr->devicePtr->deviceData;
    dInfo = (DecodeInfo *) framePtr->deviceData;
    dpy = swDevice->dpy;
    depth = swDevice->ditherRec->depth;
    imWidth = framePtr->frameHeader.width;
    imHeight = framePtr->frameHeader.height;
    winWidth = objPtr->width;
    winHeight = objPtr->height;

    xpos = objPtr->xpos;
    ypos = objPtr->ypos;

    if (winWidth == 0) {
      winWidth = imWidth;
      xpos = 0;
    }
    if (winHeight == 0) {
      winHeight = imHeight;
      ypos = 0;
    }

    if (xpos == -1) {
	xpos = (winWidth-imWidth)/2;
    }
    if (ypos == -1) {
	ypos = (winHeight-imHeight)/2;
    }

    WaitForCompletion(swDevice);

    /*
     * Create (or recreate) the XImage structure if necessary
     */
    if ((swDevice->prevWidth != imWidth) || (swDevice->prevHeight != imHeight)
	|| (swDevice->image == NULL)) {
	
	if (swDevice->image != NULL) {
#ifndef _WINDOWS
	    swDevice->image->data = NULL;
	    XDestroyImage(swDevice->image);
#else
	    free((char *) swDevice->image);
	    swDevice->image_data = NULL;
#endif
	}
	
	if (swDevice->hasShm) {
#ifdef HAVE_SHARED_MEMORY
	    image = XShmCreateImage(dpy, swDevice->ditherRec->visual,
				    depth,
				    ZPixmap, NULL, &(swDevice->shminfo),
				    imWidth, imHeight);
#endif
	} else {
#ifndef _WINDOWS
	    image = XCreateImage(dpy, swDevice->ditherRec->visual,
				 depth, ZPixmap, 0,
				 dInfo->dithered, imWidth, imHeight, 32, 0);
#else
	    int i;
	    short *pPalIndex;
	    int allocSize;

	    if (dInfo->dithered == NULL) {
		dInfo->dithered = malloc(imWidth * imHeight * depth);
	    }
	    allocSize = sizeof(BITMAPINFO);
	    if (depth == 8) {
	        allocSize += 256 * sizeof(short);
	    }
	    image = (BITMAPINFO *) malloc(allocSize);

	    image->bmiHeader.biSize = (LONG)sizeof(BITMAPINFOHEADER);
	    image->bmiHeader.biWidth = imWidth;
	    image->bmiHeader.biHeight = imHeight;
	    image->bmiHeader.biPlanes = 1;
	    image->bmiHeader.biBitCount = depth;
	    image->bmiHeader.biCompression = 0L;
	    image->bmiHeader.biSizeImage = 0L;
	    image->bmiHeader.biXPelsPerMeter = 0L;
	    image->bmiHeader.biYPelsPerMeter = 0L;
	    image->bmiHeader.biClrUsed = 0L;
	    image->bmiHeader.biClrImportant = 0L;
	    if (depth == 8) {
		pPalIndex=(short *) image->bmiColors;
		for (i = 0; i < 256; i++) {
		    pPalIndex[i] = i;
		}
	    }
#endif
	}
	
	swDevice->prevWidth = imWidth;
	swDevice->prevHeight = imHeight;
    } else {
	image = swDevice->image;
    }

    /*
     * Get rid of the old frame if there is one
     * and set up the new frame as the current one
     */
    if (swDevice->prevBuffer != NULL) {
	DetachBuffer(framePtr->objPtr->interp, swDevice->prevBuffer,
		     "jpegSW:JPEG_SW_Display");
    }

    swDevice->prevBuffer = dInfo->viewBuffer;
    AttachBuffer(objPtr->interp, swDevice->prevBuffer,
		 "jpegSW:JPEG_SW_Display");
#ifndef _WINDOWS
    image->data = (signed char *)dInfo->dithered;
#else
    swDevice->image_data = dInfo->dithered;
#endif
    
    /*
     * Actually display the image
     */
    if ((swDevice->hasShm) && (swDevice->prevBuffer->shared)
	&& (swDevice->prevBuffer->xAttach)) {
#ifdef HAVE_SHARED_MEMORY
	memcpy( &(swDevice->shminfo), &(swDevice->prevBuffer->shminfo),
	       sizeof(XShmSegmentInfo) );
	XShmPutImage(dpy, swDevice->win, swDevice->gc,
		     image, 0, 0, xpos, ypos, imWidth, imHeight, 1);
	swDevice->waiting = 1;
#endif
    } else {
#ifndef _WINDOWS
	XPutImage(dpy, swDevice->win, swDevice->gc, image, 0, 0, xpos, ypos,
		  imWidth, imHeight);
#else
	{
	    HPALETTE pal, palPrev;
	    
	    pal = XGetWindowsPalette(swDevice->dpy, 
				     XDefaultColormap(swDevice->dpy, 
						      DefaultScreen(swDevice->dpy)));
	    palPrev = SelectPalette(tcm_DitherHDC, pal, FALSE);
	    
	    SetDIBitsToDevice(tcm_DitherHDC, xpos, ypos,
			      imWidth, imHeight, 
			      0, 0, 0, imHeight,
			      dInfo->dithered,
			      image,
			      ((depth == 8) ?  DIB_PAL_COLORS : DIB_RGB_COLORS));

	    SelectPalette(tcm_DitherHDC, palPrev, FALSE);
	}

#endif
	swDevice->waiting = 0;
    }
    
    swDevice->image = image;
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_Refresh --
 *
 *	Called to refresh the display if part of it is exposed.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Display is repainted.  Could block if previous shared
 *	memory put image hasn't returned
 *
 *--------------------------------------------------------------
 */
int
JPEG_SW_Refresh(MJPEGResource *rsrcPtr)
{
    JPEG_SW_Device *swDevice;
    double speed;
    int imWidth, imHeight;
    int winWidth, winHeight;
    int xpos, ypos;

    swDevice = (JPEG_SW_Device *) rsrcPtr->devicePtr->deviceData;

    if (swDevice->prevBuffer == NULL) {
	return 0;
    }

    speed = LTS_Speed(rsrcPtr->lts);
    if (speed != 0.0) {
	return 0;
    }

    imWidth = swDevice->prevWidth;
    imHeight = swDevice->prevHeight;
    winWidth = rsrcPtr->width;
    winHeight = rsrcPtr->height;

    xpos = rsrcPtr->xpos;
    ypos = rsrcPtr->ypos;

    if (winWidth == 0) {
      winWidth = imWidth;
      xpos = 0;
    }
    if (winHeight == 0) {
      winHeight = imHeight;
      ypos = 0;
    }

    if (xpos == -1) {
	xpos = (winWidth-imWidth)/2;
    }
    if (ypos == -1) {
	ypos = (winHeight-imHeight)/2;
    }

    if ((swDevice->hasShm) && (swDevice->prevBuffer->shared)
	&& (swDevice->prevBuffer->xAttach)) {
#ifdef HAVE_SHARED_MEMORY
	XShmPutImage(swDevice->dpy, swDevice->win, swDevice->gc,
		     swDevice->image, 0, 0, xpos, ypos,
		     swDevice->prevWidth, swDevice->prevHeight, 1);
	swDevice->waiting = 1;
#endif
    } else {
#ifndef _WINDOWS
	XPutImage(swDevice->dpy, swDevice->win, swDevice->gc,
		  swDevice->image, 0, 0, xpos, ypos,
		  swDevice->prevWidth, swDevice->prevHeight);
#else
	{
	    HPALETTE pal, palPrev;
	    
	    pal = XGetWindowsPalette(swDevice->dpy, 
				     XDefaultColormap(swDevice->dpy, 
						      DefaultScreen(swDevice->dpy)));
	    palPrev = SelectPalette(tcm_DitherHDC, pal, FALSE);
	    
	    SetDIBitsToDevice(tcm_DitherHDC, xpos, ypos,
			      swDevice->prevWidth, swDevice->prevHeight,
			      0, 0, 0, swDevice->prevHeight,
			      swDevice->image_data,
			      swDevice->image,
			      ((swDevice->ditherRec->depth == 8) ?
			       DIB_PAL_COLORS : DIB_RGB_COLORS));

	    SelectPalette(tcm_DitherHDC, palPrev, FALSE);
	}

#endif
	swDevice->waiting = 0;
    }
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_Shutdown --
 *
 *	Called to destroy device.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	The device must never be referenced again.
 *
 *--------------------------------------------------------------
 */
int
JPEG_SW_Shutdown(MJPEGResource *rsrcPtr) 
{
    JPEG_SW_Device *swDevice = (JPEG_SW_Device *) rsrcPtr->devicePtr->deviceData;
    
    if (!rsrcPtr->inited) {
	ckfree((char *) swDevice);
	return 0;
    }

    WaitForCompletion(swDevice);

    if (swDevice->prevBuffer) {
	DetachBuffer(rsrcPtr->interp, swDevice->prevBuffer,
		     "jpegSW:JPEG_SW_Shutdown");
#ifndef _WINDOWS
	swDevice->image->data = NULL;	/* Prevent XDestroyImage
						 * from trying to free it */
	XDestroyImage(swDevice->image);
#else
	free((char *) swDevice->image);
	swDevice->image_data = NULL;
#endif
    }
    
#ifdef HAVE_SHARED_MEMORY
    Tk_DeleteGenericHandler(ProcessShmEvent, (ClientData) rsrcPtr);
#endif
    if (swDevice->ditherRec != NULL ) {
	FreeDither(swDevice->ditherRec);
	swDevice->ditherRec = NULL;
    }
    if (swDevice->gc) {
	XFreeGC(swDevice->dpy, swDevice->gc);
    }	

    if (swDevice->ldht) {
	JPEG_SW_FreeHufftab(swDevice->ldht);
    }
    if (swDevice->laht) {
	JPEG_SW_FreeHufftab(swDevice->laht);
    }
    if (swDevice->cdht) {
	JPEG_SW_FreeHufftab(swDevice->cdht);
    }
    if (swDevice->caht) {
	JPEG_SW_FreeHufftab(swDevice->caht);
    }

    ckfree((char *) swDevice);
    rsrcPtr->devicePtr->deviceData = NULL;

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_FreeFrame --
 *
 *	The current frame is being taken away -- clean up any pointers
 *	we have to it.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
JPEG_SW_FreeFrame(MJPEGFrame *framePtr)
{
    DecodeInfo *dInfo = (DecodeInfo *) framePtr->deviceData;
    
    if (dInfo == NULL) {
	return 0;
    }
    
    if (dInfo->decodeBuffer != NULL) {
	DetachBuffer(framePtr->objPtr->interp, dInfo->decodeBuffer,
		     "jpegSw:FreeFrame");
    }
    
    if (dInfo->viewBuffer != NULL) {
	DetachBuffer(framePtr->objPtr->interp, dInfo->viewBuffer,
		     "jpegSw:FreeFrame");
    }
    
    ckfree((void *) dInfo);
    framePtr->deviceData = NULL;
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_NextFrame --
 *
 *	Chooses the next frame to decode
 *
 * Results:
 *	0 if successfull, non-zero error code otherwise
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	MJPEGResource *rsrcPtr	The device
 *	MJPEGFrame *frameList	List of frames to choose from
 *	MJPEGFrame **chosen	Set to point to the chosen frame
 *
 *--------------------------------------------------------------
 */
static int
JPEG_SW_NextFrame(MJPEGResource *rsrcPtr, MJPEGFrame *frameList,
		  MJPEGFrame **chosen)
{
    MJPEGFrame *currentFrame, *longestFrame;
    JPEG_SW_Device *swDevice;
    double speed, now, end, decode;
    
    swDevice = (JPEG_SW_Device *) rsrcPtr->devicePtr->deviceData;
    
    if (frameList == NULL) {
	*chosen = NULL;
	return 1;
    }
    
    speed = LTS_Speed((ClientData) rsrcPtr->lts);
    now = LTS_Value((ClientData) rsrcPtr->lts);

#if 1
    /*
     * Go through the available frames, looking for one
     * we have time to decode
     */
    longestFrame = frameList;
    for (currentFrame = frameList; currentFrame != NULL;
	 currentFrame = currentFrame->next) {

	end = (speed < 0.0) ? currentFrame->logStart : currentFrame->logEnd;
	decode = swDevice->m*currentFrame->dataLength + swDevice->b;
	decode += (decode / 0.010) * swDevice->secIdle;
	
	if ((now + decode*speed) <= end) {
	    *chosen = currentFrame;
	    return 0;
	}
	
	if (speed < 0.0) {
	    if (currentFrame->logEnd < longestFrame->logEnd) {
		longestFrame = currentFrame;
	    }
	} else {
	    if (currentFrame->logEnd > longestFrame->logEnd) {
		longestFrame = currentFrame;
	    }
	}
    }
    
    *chosen = longestFrame;
    return 0;
#else
    *chosen = frameList;
    currentFrame = frameList;
    
    while (currentFrame != NULL) {
	if (speed < 0.0) {
	    if (currentFrame->logEnd < (*chosen)->logEnd) {
		*chosen = currentFrame;
	    }
	} else {
	    if (currentFrame->logEnd > (*chosen)->logEnd) {
		*chosen = currentFrame;
	    }
	}
	currentFrame = currentFrame->next;
    }
    return 0;
#endif
}


/*
 *--------------------------------------------------------------
 *
 * JPEG_SW_Error --
 *
 *	Translates an error number (passed in) into a human readable
 *	string (returned).
 *
 * Results:
 *	Human readable string that indicates the error.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	MJPEGResource *rsrcPtr	The device
 *	int errNum		The error code
 *
 *--------------------------------------------------------------
 */
char *
JPEG_SW_Error(MJPEGResource *rsrcPtr, int errNum)
{
    switch (errNum) {
      case INIT_DITHER_FAILED:
	return "Dither initialization failed.\n";
	
      case NEW_FRAME_FAILED:
	return "Initialization for new frame failed.\n";
	
      case BAD_PRECISION:
	return "JPEG Stream is not at 8 bit precision.\n";
	
      case BAD_GEOMETRY:
	return "Image size in JPEG header does not match CMT header.\n";
	
      case BAD_DECIMATION:
	return "Image is not 422.\n";
	
      case SHM_ERROR:
	return "JPEG Shared Memory Error.\n";
	
      default:
	return "Unknown Software JPEG Device Error.\n";
    }
}


/*
 *--------------------------------------------------------------
 *
 * GetSoftwareJpegDevice --
 *
 *	Called by resource to determine default device.  We should
 *	only be called if there's no jpeg hardware available.
 *
 * Results:
 *	Pointer to new MJPEGDevice
 *
 * Side effects:
 *	A couple of structures are allocated.
 *
 *--------------------------------------------------------------
 */
MJPEGDevice *
GetSoftwareJpegDevice (MJPEGResource *objPtr)
{
    MJPEGDevice *new_dev;

    new_dev = (MJPEGDevice *) ckalloc(sizeof(MJPEGDevice));

    new_dev->asyncDecodeFlag = 1;
    new_dev->decodeProc = JPEG_SW_Decode;
    new_dev->refreshProc = JPEG_SW_Refresh;
    new_dev->displayProc = JPEG_SW_Display;
    new_dev->initProc = JPEG_SW_Init;
    new_dev->reinitProc = JPEG_SW_Reinit;
    new_dev->shutdownProc = JPEG_SW_Shutdown;
    new_dev->freeFrameProc = JPEG_SW_FreeFrame;
    new_dev->nextFrameProc = JPEG_SW_NextFrame;
    new_dev->configProc = NULL;		/* XXX Write this! */
    new_dev->errorProc = JPEG_SW_Error;
    new_dev->deviceData = (ClientData) ckalloc(sizeof(JPEG_SW_Device));
    memset(new_dev->deviceData, 0, sizeof(JPEG_SW_Device));
    return new_dev;
}


#else /* JPEG_SW */

/*
 *--------------------------------------------------------------
 *
 * GetSoftwareJpegDevice --
 *
 *	Stub routine for machines without compiled in software
 *	JPEG decoder.
 *
 * Results:
 *	NULL
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
MJPEGDevice *
GetSoftwareJpegDevice (MJPEGResource *objPtr)
{
    return NULL;
}

#endif	/* JPEG_SW */
