/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * tcmLTS.c
 *
 *	This file contains code to implement the lts tcl command,
 *	as described in the cm user documentation.  
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <math.h>
#include <sys/time.h>
#include "default.h"
#include "tcl.h"
#include "tk.h"
#include "cmConfig.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmUtil.h"

/* 
 * A few convieninet constants are defined here. 
 */

#define DEF_LTS_ZERO	"0.0"
#define DEF_LTS_COMMAND	""

#define LTS_IN_SYNCH_WINDOW 0.005
#define LTS_MIN_RTRIP_AVG_NUM 10
#define LTS_MIN_RESYNCH_TIME 1.0
#define LTS_RESYNCH_INCREMENT 1.0

/*
 * Trigger callback structure.
 */
static int nextTrigId = 0;          /* Counter for generating trigger id's. */

struct Trigger {     
  void (*callback)();               /* Pointer to trigger callback. */
  ClientData clientData;            /* Pointer to trigger callback parameter. */
  int flags;                        /* Indicates which field changes cause callback. */
  int trigId;                       /* Id of this trigger. */
  struct Trigger *next;             /* Pointer to next trigger in linked list. */
};
typedef struct Trigger Trigger;

/*
 * LTS Slave structure.
 */
  
typedef struct LtsSlave {
  char *slaveName;                   /* Name of slave object. */
  struct LtsSlave *next;             /* Pointer to next slave in list. */
} LtsSlave;


/*
 * One record of the following type is kept for each lts:
 */

typedef struct {
  Tcl_Interp *interp;    /* Pointer to Tcl interpreter. */
  char *name;            /* Name of lts object. */
  double speed;		 /* Current speed of the lts */
  double offset;	 /* Current offset of the lts */
  double value;		 /* Current value of the lts */
  Trigger *triggers;     /* C procedure triggers registered. */
  double resynchTime;    /* Time between resynching if necessary. 0 = never. */
  double synchStart;     /* Used to store value of LTS when doing a remote synch. */
  double rtrip_avg;
  int rtrip_avg_num;
  LtsSlave *slaves;      /* Pointer to lts slaves if any. */
  char *masterName;      /* Master name if this lts is a slave to someone. */
  char *masterRPCFile;   /* Rpc file descriptor of master if not local. */
  int resynchAtToken;    /* At token for pending resynch command. */
  char source[32];       /* Indicates source of timing info (local, stream, ntp). */
} Lts;

/* Counter for creating uid's for new lts's. */

static int nextLtsNum = 0;

/*
 * Information used for configuration option parsing:
 */

static Tcm_ConfigSpec configSpecs[] = {

    {TCM_CONFIG_DOUBLE, "-speed",
	DEF_LTS_ZERO, Tcm_Offset(Lts, speed),
    TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_DOUBLE, "-offset",
	DEF_LTS_ZERO, Tcm_Offset(Lts, offset),
    TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_DOUBLE, "-value",
	DEF_LTS_ZERO, Tcm_Offset(Lts, value),
    TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_DOUBLE, "-resynchTime",
	DEF_LTS_ZERO, Tcm_Offset(Lts, resynchTime),
    TCM_CONFIG_NULL_OK},
    {TCM_CONFIG_END, (char *) NULL, (char *) NULL, 0, 0}
};


/*
 * Forward declarations for procedures defined later in this file:
 */

static int ConfigureLts		_ANSI_ARGS_((Tcl_Interp * interp,
					     Lts * ltsPtr, int argc,
					     char **argv, int flags));

static int LtsCmd		_ANSI_ARGS_((ClientData clientData,
					     Tcl_Interp * interp,
					     int argc, char **argv));

static void DestroyLts		_ANSI_ARGS_((ClientData clientData));

static void DoSynch _ANSI_ARGS_((ClientData clientData));

static void DoSynchReturn _ANSI_ARGS_((Lts *ltsPtr, double remTime));


/*
 *----------------------------------------------------------------------
 *
 * UpdateLTSValue --
 *
 *	This procedure recalculates the lts "value" slot based on the
 *	current system clock.
 *
 * Parameters:
 *      ltsPtr (in) - pointer to lts structure.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *	"value" field of lts structure is changed.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateLTSValue(Lts *ltsPtr)
{
    ltsPtr->value = ltsPtr->speed * Tcm_ReadSysClock() + ltsPtr->offset;
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_LtsCmd --
 *
 *	This procedure is invoked to process the "lts" Tcl/Tk
 *	command.  See the user documentation for details on what
 *	it does.
 *
 * Parameters:
 *      Standard Tcl C command interface.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
Tcm_LtsCmd(ClientData clientData,        /* Unused. */
	   Tcl_Interp *interp,           /* Current interpreter. */
	   int argc,                     /* Number of arguements. */
	   char **argv)                  /* Argument strings. */
{
    Lts *ltsPtr;		/* New lts. */

    /* 
     * Check for minimum number of arguments. 
     */

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " namePrefix ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Initialize the fields of the structure that won't be initialized by
     * ConfigureLts or that ConfigureLts requires to be initialized already.
     */

    ltsPtr = (Lts *) ckalloc(sizeof(Lts));
    ltsPtr->interp = interp;
    ltsPtr->speed = 0.0;
    ltsPtr->offset = 0.0;
    ltsPtr->triggers = NULL;
    ltsPtr->value = 0.0;
    ltsPtr->resynchTime = 0.0;
    ltsPtr->rtrip_avg = 0.0;
    ltsPtr->rtrip_avg_num = 0;
    ltsPtr->slaves = NULL;
    ltsPtr->masterName = NULL;
    ltsPtr->masterRPCFile = NULL;
    ltsPtr->resynchAtToken = 0;
    ltsPtr->name = (char *) ckalloc(strlen(argv[1])+10);
    sprintf(ltsPtr->name, "%slts%d", argv[1], nextLtsNum);
    nextLtsNum++;

    /* 
     * Current source of lts is always local. This may change in the
     * future.
     */

    sprintf(ltsPtr->source, "local");

    Tcl_CreateCommand(interp, ltsPtr->name, LtsCmd, (ClientData) ltsPtr,
		      (void (*) ()) NULL);

    if (ConfigureLts(interp, ltsPtr, argc - 2, argv + 2, 0) != TCL_OK)
	return TCL_ERROR;

    strcpy(interp->result, ltsPtr->name);

    /* 
     * Register the new lts name as an atom. This allows other objects
     * to get a C pointer to the lts structure and use the C interface
     * instead of the slower Tcl interface. 
     */

    Tcm_RegisterAtom(ltsPtr->name, (ClientData)ltsPtr);

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * LtsCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a object managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Parameters:
 *      Standard Tcl C command interface.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
LtsCmd(ClientData clientData,   /* Pointer to LTS object. */
       Tcl_Interp *interp,      /* Current interpreter. */
       int argc,                /* Number of arguments. */
       char **argv)             /* Argument strings. */
{
    Lts *ltsPtr = (Lts *) clientData;     /* Pointer to lts structure. */
    double time;                          /* Used to hold time values passed
					     in as parameters to lts commands. */
    int result = TCL_OK;                  /* Return value of command. */
    int length;                           /* Lenght of command. */
    char c;                               /* First character of command. */
    char str[256];                        /* Temp string storage. */
    int status;                           /* Used to hold return value of other
					     Tcl subroutines called. */

    /*
     * Check for minimum number of arguments.
     */

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg arg ...?\"",
			 (char *) NULL);
	goto usage;
    }

    /*
     * Parse command name and switch to appropriate action.
     */

    c = argv[1][0];
    length = strlen(argv[1]);

    switch (c) {
    case 'c':
	/*------------------------ CONFIGURE ---------------------------*/

	UpdateLTSValue(ltsPtr);
	if (strncmp(argv[1], "configure", length) == 0) {
	    if (argc == 2) {
		result = Tcm_ConfigureInfo(interp, configSpecs,
					 (char *) ltsPtr, (char *) NULL, 0);
	    } else if (argc == 3) {
		result = Tcm_ConfigureInfo(interp, configSpecs,
					   (char *) ltsPtr, argv[2], 0);
	    } else {
		result = ConfigureLts(interp, ltsPtr, argc - 2,
				      argv + 2, TCM_CONFIG_ARGV_ONLY);
	    }

	/*---------------- CLEAR SLAVES ---------------------------------*/
	} else if (strncmp(argv[1], "clearSlaves", length) == 0) {

	  LtsSlave *slavePtr, *tmpSlavePtr;

	  slavePtr = ltsPtr->slaves;

	  while (slavePtr != NULL) {
	    if (slavePtr->slaveName != NULL) ckfree(slavePtr->slaveName);
	    tmpSlavePtr = slavePtr;
	    slavePtr = slavePtr->next;
	    ckfree((char *) tmpSlavePtr);
	  }
	  result = TCL_OK;
	} else {
	    goto unknownCommand;
	}
	break;

    case 'd':
	/*------------------------ DESTROY -----------------------------*/

	if (strncmp(argv[1], "destroy", length) == 0) {
	    Tcl_DeleteCommand(interp, argv[0]);
	    Tcm_UnregisterAtom(argv[0]);
	    if (ltsPtr->resynchAtToken != 0) {
		Tcm_DeleteAtHandler(ltsPtr->resynchAtToken);
		ltsPtr->resynchAtToken = 0;
	    }
	    Tk_EventuallyFree((ClientData)ltsPtr, DestroyLts);
	    strcpy(interp->result, "");
	    result = TCL_OK;
	}
	/*----------------------DEL SLAVE ----------------------------*/
	else if (strncmp(argv[1], "delSlave", length) == 0) {
	  if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: \"", argv[0],
			     "delSlave <slave_name>\"", (char *) NULL);
	    result = TCL_ERROR;
	  }
	  else {
	    LtsSlave *slavePtr, *prevSlavePtr;

	    prevSlavePtr = NULL;
	    slavePtr = ltsPtr->slaves;
	    while(slavePtr != NULL) {
	      if (strcmp(slavePtr->slaveName, argv[2]) == 0) 
		break;
	      prevSlavePtr = slavePtr;
	      slavePtr = slavePtr->next;
	    }

	    if (slavePtr != NULL) {
	      if (prevSlavePtr == NULL) {
		ltsPtr->slaves = slavePtr->next;
	      } else {
		prevSlavePtr->next = slavePtr->next;
	      }
	      if (slavePtr->slaveName != NULL) ckfree(slavePtr->slaveName);
	      ckfree((char *) slavePtr);
	    }

	    result = TCL_OK;
	  }
	}
	    
	break;

    case 's':
	/*-----------------------SYNCHRETURN-------------------------*/
	/* 
	 * The synchReturn command is not intended to be called by
	 * the user or creator of the lts object. This command is used
	 * by the synch command when synchronizing with remote lts's in
	 * order to avoid blocking.
	 */

	if ((argv[1][1] = 'y') && (strcmp(argv[1], "synchReturn") == 0)) {
	  double remTime;

	  if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: \"",
			     argv[0], "synchReturn <remote_time>\"",
			     (char *) NULL);
	    return TCL_ERROR;
	  }

	  if (Tcl_GetDouble(interp, argv[2], &remTime) != TCL_OK) {
	    Tcl_AppendResult(interp, "bad remote time value: \"",
			     argv[2], "\"", (char *) NULL);
	    return TCL_ERROR;
	  }

	  DoSynchReturn(ltsPtr, remTime);
	  result = TCL_OK;
	}
	/*----------------------- SYNCH -----------------------------*/
	else if ((argv[1][1] = 'y') && (strcmp(argv[1], "synch") == 0)) {
	    double newSpeed, newValue;
	    char *tmp;
	    Trigger *trigPtr, *nextTrig;

	    if (argc != 7) {
	        Tcl_AppendResult(interp, "wrong # of args, should be: \"",
			 argv[0],
			 " synch <source> <speed> <value> <offset> <master>\"",
		         (char *) NULL);
	        return TCL_ERROR;
	    }

	    if (Tcl_GetDouble(interp, argv[3], &newSpeed) != TCL_OK) {
	        Tcl_AppendResult(interp, "bad speed value in synch command",
				 (char *) NULL);
	    }

	    if (Tcl_GetDouble(interp, argv[4], &newValue) != TCL_OK) {
	        Tcl_AppendResult(interp, "bad value value in synch command",
				 (char *) NULL);
	    }

	    /*
	     * Avoid any pending resynching command on at queue.
	     */

	    if (ltsPtr->resynchAtToken != 0) {
	        Tcm_DeleteAtHandler(ltsPtr->resynchAtToken);
	        ltsPtr->resynchAtToken = 0;
	    }

	    /*
	     * Set speed and value, recalculate offset based on given values.
	     */

	    ltsPtr->speed = newSpeed;
	    ltsPtr->value = newValue;

	    /*
	     * Reset synching information such as master object name, rpc_file
	     * if any, resynching time, etc.
	     */

	    if (ltsPtr->masterName != NULL) {
	        ckfree(ltsPtr->masterName);
	        ltsPtr->masterName = NULL;
	    }
	    ltsPtr->masterName = (char *) ckalloc(strlen(argv[6])+1);
	    strcpy(ltsPtr->masterName, argv[6]);

	    tmp = Tcl_GetVar(interp, "dp_rpcFile", TCL_GLOBAL_ONLY);
	    if (ltsPtr->masterRPCFile != NULL) {
	        ckfree(ltsPtr->masterRPCFile);
	        ltsPtr->masterRPCFile = NULL;
	    }
	    if (tmp != NULL) {
	        ltsPtr->masterRPCFile = (char *) ckalloc(strlen(tmp)+1);
	        strcpy(ltsPtr->masterRPCFile, tmp);
	    }

	    if (ltsPtr->masterRPCFile == NULL) {
		double newOffset;

		if (Tcl_GetDouble(interp, argv[5], &newOffset) != TCL_OK) {
		    Tcl_AppendResult(interp, "bad offset value in synch command",
				     (char *) NULL);
		}
		ltsPtr->offset = newOffset;
	    } else {
		ltsPtr->offset = ltsPtr->value - ltsPtr->speed * Tcm_ReadSysClock();
		ltsPtr->resynchTime = 1.0;
		ltsPtr->rtrip_avg = 0.0;
		ltsPtr->rtrip_avg_num = 0;

	    /*
	     * If speed is not 0.0, schedule a resynch event with at queue.
             */

		if (ltsPtr->speed != 0.0) {
		    ltsPtr->resynchAtToken = Tcm_CreateAtHandler(Tcm_ReadSysClock()+
								 ltsPtr->resynchTime, HUGE_VAL,
								 "tk", DoSynch, ltsPtr, NULL,
								 NULL, NULL, NULL, NULL);
		}
	    }

	    /*
	     * Send changes to any slaves.
	     */

	    if (ltsPtr->slaves != NULL) {
	        LtsSlave *slavePtr;
	        char synchCmd[256];
	      
	        slavePtr = ltsPtr->slaves;
	      
	        sprintf(synchCmd, " synch %s %f %f %f %s", ltsPtr->source,
			ltsPtr->speed, ltsPtr->value, ltsPtr->offset,
			ltsPtr->name);
	        while (slavePtr != NULL) {
		    Tcl_VarEval(interp, slavePtr->slaveName, synchCmd,
				(char *) NULL);
		    slavePtr = slavePtr->next;
	        }
	    }

	    /*
	     * Call callbacks if trigger on specific change is true.
	     */

	    trigPtr = ltsPtr->triggers;
	    while (trigPtr != NULL) {
	        nextTrig = trigPtr->next;
	        trigPtr->callback(trigPtr->clientData);
	        trigPtr = nextTrig;
	    }
	    strcpy(interp->result, "");
	    result = TCL_OK;

      	/*------------------------ SYSTEM TIME OF ---------------------------*/
	} else if ((argv[1][1] == 'y') &&
		   (strcmp(argv[1], "systemTimeOf") == 0)) {
	    if (argc == 3) {
		status = Tcl_GetDouble(interp, argv[2], &time);
		if (status != TCL_OK)
		    return status;
		if (ltsPtr->speed == 0) {
		    Tcl_AppendResult(interp, "undefined", (char *) NULL);
		    result = TCL_OK;
		} else {
		    sprintf(str, "%lf", (time - ltsPtr->offset) / ltsPtr->speed);
		    Tcl_AppendResult(interp, str, (char *) NULL);
		    result = TCL_OK;
		}
		result = TCL_OK;
	    } else {
		Tcl_AppendResult(interp, "option \"", argv[1],
				 "\" requires an argument", (char *) NULL);
		result = TCL_ERROR;
	    }
	} else {
	    goto unknownCommand;
	}
	break;

    case 'l':
	/*------------------------ LOGICAL TIME OF --------------------------*/
	if (strncmp(argv[1], "logicalTimeOf", length) == 0) {
	    if (argc == 3) {
		if (strcmp(argv[2], "now") == 0) {
		    UpdateLTSValue(ltsPtr);
		    sprintf(str, "%lf", ltsPtr->value);
		    Tcl_AppendResult(interp, str, (char *) NULL);
		    result = TCL_OK;
		} else {
		    status = Tcl_GetDouble(interp, argv[2], &time);
		    if (status != TCL_OK)
			return status;
		    sprintf(str, "%lf", time * ltsPtr->speed + ltsPtr->offset);
		    Tcl_AppendResult(interp, str, (char *) NULL);
		    result = TCL_OK;
		}
	    } else {
		Tcl_AppendResult(interp, "option \"", argv[1],
				 "\" requires an argument", (char *) NULL);
		result = TCL_ERROR;
	    }
	} else {
	    goto unknownCommand;
	}
	break;
	
      case 'a':
	/*----------------------ADD SLAVE--------------------------------*/
	if (strncmp(argv[1], "addSlave", length) == 0) {
	  if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: \"", argv[0],
			     " addSlave <slave_name>\"", (char *) NULL);
	    result = TCL_ERROR;
	  }
	  else {
	    LtsSlave *newSlave;

	    newSlave = (LtsSlave *) ckalloc(sizeof(LtsSlave));
	    newSlave->slaveName = (char *) ckalloc(strlen(argv[2])+1);
	    strcpy(newSlave->slaveName, argv[2]);
	    newSlave->next = ltsPtr->slaves;
	    ltsPtr->slaves = newSlave;
	    strcpy(interp->result, "");
	    result = TCL_OK;
	  }
	}

	break;

    default:
	goto unknownCommand;
    }

    return result;

unknownCommand:
    Tcl_AppendResult(interp, "bad <command> \"", argv[1], "\"",
		     (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  synch <source> <speed> <value> <offset> <master>",
                "  systemTimeOf <time>\n",
                "  logicalTimeOf <time>\n",
                "  addSlave <slave_name>\n",
                (char *) NULL);
    return TCL_ERROR;
}


/*
 *----------------------------------------------------------------------
 *
 * DestroyLts --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a lts at a safe time
 *	(when no-one is using it anymore).
 *
 * Parameters:
 *      clientData (in) - pointer to lts structure being destroyed.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the lts is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyLts(ClientData clientData       /* Pointer to LTS structure. */)
{
    Lts *ltsPtr = (Lts *) clientData;
    Trigger *trigPtr, *tmpPtr;

    trigPtr = ltsPtr->triggers;
    while (trigPtr != NULL) {
      tmpPtr = trigPtr;
      trigPtr = trigPtr->next;
      ckfree((char *) tmpPtr);
    }

    if (ltsPtr->name != NULL) {
      ckfree(ltsPtr->name);
      ltsPtr->name = NULL;
    }

    while (ltsPtr->slaves != NULL) {
      LtsSlave *tmp;
      
      tmp = ltsPtr->slaves;
      ltsPtr->slaves = ltsPtr->slaves->next;

      if (tmp->slaveName != NULL) {
	ckfree(tmp->slaveName);
      }

      ckfree((char *) tmp);
    }

    if (ltsPtr->masterName != NULL) {
      ckfree(ltsPtr->masterName);
    }

    if (ltsPtr->masterRPCFile != NULL) {
      ckfree(ltsPtr->masterRPCFile);
    }

    if (ltsPtr->resynchAtToken != 0) {
      Tcm_DeleteAtHandler(ltsPtr->resynchAtToken);
    }

    ckfree((char *) ltsPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * ConfigureLts --
 *
 *	This procedure is called to process an argv/argc list,
 *	in order to configure (or reconfigure) a lts.
 *
 * Parameters:
 *      interp (in) - pointer to interpreter.
 *      ltsPtr (in) - pointer to lts structure.
 *      argc (in) - number of arguments
 *      argv (in) - pointer to argument strings.
 *      flags (in) - configuration flags passed on to Tcm_ConfigureObject
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for ltsPtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */
static int
ConfigureLts(Tcl_Interp *interp,    /* Used for error reporting. */
	     register Lts *ltsPtr,  /* Pointer to lts structure. */
	     int argc,              /* Number of arguements.     */
	     char **argv,           /* Arguments.                */
	     int flags              /* Flags to pass to Tcm_ConfigureObject */)
{
    double currValue;
    double currSpeed;
    double currOffset;
    Trigger *trigPtr, *nextTrig;
    int changeFlag;

    /*
     * Save current state of LTS.
     */

    UpdateLTSValue(ltsPtr);
    currValue = ltsPtr->value;
    currSpeed = ltsPtr->speed;
    currOffset = ltsPtr->offset;

    if (Tcm_ConfigureObject(interp, configSpecs,
			    argc, argv, (char *) ltsPtr, flags) != TCL_OK)
	return TCL_ERROR;

    /* 
     * Construct flag for what changed. 
     */

    changeFlag = 0;
    if (currOffset != ltsPtr->offset) changeFlag = changeFlag | LTS_OFFSET_FLAG;
    if (currSpeed != ltsPtr->speed) changeFlag = changeFlag | LTS_SPEED_FLAG;
    if (currValue != ltsPtr->value) changeFlag = changeFlag | LTS_VALUE_FLAG;
    
    /*
     * If the user sets the speed or value, this requires special processing
     * to set the offset.
     */

    if ((currSpeed != ltsPtr->speed) || (ltsPtr->value != currValue)) {
	ltsPtr->offset = ltsPtr->value - ltsPtr->speed * Tcm_ReadSysClock();
    }

    /* 
     * Call callbacks if trigger on specific change is true. 
     * Use a two phase trigger.
     */

    if (changeFlag) {
      /* Phase 1 */
      trigPtr = ltsPtr->triggers;
      while (trigPtr != NULL) {
	nextTrig = trigPtr->next;
	if ((trigPtr->flags & LTS_EARLY_PHASE) &&
	    (trigPtr->flags & changeFlag)) {
	  trigPtr->callback(trigPtr->clientData);
	}
	trigPtr = nextTrig;
      }
      /* Phase 2 */
      trigPtr = ltsPtr->triggers;
      while (trigPtr != NULL) {
	nextTrig = trigPtr->next;
	if ((!(trigPtr->flags & LTS_EARLY_PHASE)) &&
	    (trigPtr->flags & changeFlag)) {
	  trigPtr->callback(trigPtr->clientData);
	}
	trigPtr = nextTrig;
      }
    }

    /* 
     * Tell all slaves about changes. 
     */

    if (changeFlag) {
      LtsSlave *slavePtr;
      char synchCmd[256];

      slavePtr = ltsPtr->slaves;

      sprintf(synchCmd, " synch %s %f %f %f %s", ltsPtr->source, ltsPtr->speed,
	      ltsPtr->value, ltsPtr->offset, ltsPtr->name);
      while (slavePtr != NULL) {
	Tcl_VarEval(interp, slavePtr->slaveName, synchCmd, (char *) NULL);
	slavePtr = slavePtr->next;
      }
    }

    interp->result[0] = '\0';
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * LTS_Speed --
 *
 *	C level access function for speed of lts.
 *
 * Parameters:
 *      clientData (in) - pointer to lts structure.
 *
 * Results:
 *	Speed of the lts.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

double
LTS_Speed(ClientData clientData   /* Pointer to lts structure.*/)
{
    Lts *ltsPtr = (Lts *) clientData;
    return ltsPtr->speed;
}


/*
 *--------------------------------------------------------------
 *
 * LTS_Offset --
 *
 *	C level access function for offset of lts.
 *
 * Parameters:
 *      clientData (in) - pointer to lts structure.
 *
 * Results:
 *	Offset of the lts.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

double
LTS_Offset(ClientData clientData       /* Pointer to lts structure.*/)
{
    Lts *ltsPtr = (Lts *) clientData;
    return ltsPtr->offset;
}


/*
 *--------------------------------------------------------------
 *
 * LTS_Value --
 *
 *	C level access function for value of lts.
 *
 * Parameters:
 *      clientData (in) - pointer to lts structure.
 * Results:
 *	Value of the lts.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
double
LTS_Value(ClientData clientData        /* Pointer to lts strucutre.*/)
{
    Lts *ltsPtr = (Lts *) clientData;
    UpdateLTSValue(ltsPtr);
    return ltsPtr->value;
}


/*
 *--------------------------------------------------------------
 *
 * LTS_SystemTimeOf --
 *
 *	C level access function for converting to system time from
 *	logical time.
 *
 * Parameters:
 *      clientData (in) - pointer to lts structure.
 *      lTime (in) - logical time value.
 *
 * Results:
 *	System time corresponding to logical time specified.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

double
LTS_SystemTimeOf(ClientData clientData,       /* Pointer to lts structure. */
		 double lTime                 /* Logical time value.*/ )
{
    Lts *ltsPtr = (Lts *) clientData;
    if (ltsPtr->speed != 0.0) {
	return (lTime - ltsPtr->offset) / ltsPtr->speed;
    } else {
	return HUGE_VAL;
    }
}


/*
 *--------------------------------------------------------------
 *
 * LTS_LogicalTimeOf --
 *
 *	C level access function for converting to logical time
 *	from system time.
 *
 * Parameters:
 *      clientData (in) - pointer to lts structure.
 *      sTime (in) - system time value.
 * 
 * Results:
 *	Logical time corresponding to system time specified.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

double
LTS_LogicalTimeOf(ClientData clientData, /* Pointer to lts structure. */
		  double sTime           /* System time value. */)
{
    Lts *ltsPtr = (Lts *) clientData;
    return sTime * ltsPtr->speed + ltsPtr->offset;
}

/*
 *---------------------------------------------------------------
 *
 * LTS_SetTrigger
 *
 *     C level function to associate a callback with the changing
 *     of one or more of the lts parameters.
 *
 * Parameters:
 *     clientData (in) - pointer to lts structure.
 *     eventFlags (in) - specifies which slot values trigger is associated with.
 *     callback (in) - pointer to callback procedure.
 *     triggerData (in) - pointer to data to be given as parameter to callback.
 *
 * Results:
 *     Trigger id number.
 * 
 * SideEffects:
 *
 *     Trigger structure allocated and filled. 
 *     Pointer in lts's trigger pointer array set. 
 *
 *----------------------------------------------------------------
 */

int 
LTS_SetTrigger(ClientData clientData,     /* Pointer to lts structure. */
	       int eventFlags,            /* Which slots to associate trigger with. */
	       void (*callback)(),        /* Pointer to callback proc. */
	       ClientData triggerData     /* Pointer to data for callback proc. */)
{
  Trigger *newTrigger;
  Lts *lts = (Lts *) clientData;

  newTrigger = (Trigger *) ckalloc(sizeof(Trigger));
  newTrigger->flags = eventFlags;
  newTrigger->callback = callback;
  newTrigger->clientData = triggerData;
  newTrigger->trigId = nextTrigId++;
  newTrigger->next = lts->triggers;
  lts->triggers = newTrigger;
  return newTrigger->trigId;
}

/*
 *---------------------------------------------------------------
 *
 * LTS_UnsetTrigger
 *
 *     C level function to disassociate a callback with the changing
 *     of one or more of the lts slot values.
 *
 * Parameters:
 *     clientData (in) - pointer to lts structure.
 *     triggerNum (in) - trigger id.
 *
 * Results:
 *     None.
 *
 * Side Effects:
 *     Trigger structure freed. 
 *     Pointer in lts's trigger pointer array unset. 
 *
 *----------------------------------------------------------------
 */

void
LTS_UnsetTrigger(ClientData clientData,  /* Pointer to lts structure. */
		 int triggerNum          /* Trigger id. */)
{
  Lts *lts = (Lts *) clientData;
  Trigger *trigPtr, *prevTrig;

  prevTrig = NULL;
  trigPtr = lts->triggers;
  while (trigPtr != NULL) {
    if (trigPtr->trigId == triggerNum) break;
    prevTrig = trigPtr;
    trigPtr = trigPtr->next;
  }

  if (trigPtr == NULL) return;
  if (prevTrig == NULL) {
    lts->triggers = trigPtr->next;
  }
  else {
    prevTrig->next = trigPtr->next;
  }
  
  ckfree((char *) trigPtr);
  return;
}


/*
 *---------------------------------------------------------------
 *
 * DoSynchReturn
 *     
 *     Calculates offset for synchronization with another LTS.
 *     This function finishes the synchronization process started
 *     by DoSynch. It can be called through the lts command synchReturn
 *     but is not expected to be used by users. Instead, the tcl 
 *     interface exists solely to implement a non blocking synch mechanism
 *     that utilizes dp_RDO instead of dp_RPC.
 * 
 * Parameters:
 *
 *     ltsPtr (in) - pointer to lts structure.
 *     remTime (in) - value of master lts.
 *
 * Results:
 *
 *     Standade Tcl result value.
 *
 * Side Effects:
 *
 *     May adjust the lts offset slot based on synch. calculation.
 *     Reschedules next synching time.
 *    
 *---------------------------------------------------------------
 */

static void
DoSynchReturn (Lts *ltsPtr,           /* Pointer to lts structure. */
	       double remTime         /* Time value returned from master. */)
{ 
    double nowTime, retTime, alpha;
    double rtripTime;
    
    nowTime = ltsPtr->synchStart;
    
    retTime = LTS_Value((ClientData) ltsPtr);
  
    rtripTime = retTime - nowTime;
    
    if ((rtripTime < 1.5*ltsPtr->rtrip_avg) || 
	(ltsPtr->rtrip_avg_num < LTS_MIN_RTRIP_AVG_NUM)) {

	ltsPtr->rtrip_avg = (ltsPtr->rtrip_avg * ltsPtr->rtrip_avg_num + rtripTime) /
	  (ltsPtr->rtrip_avg_num+1);
	ltsPtr->rtrip_avg_num++;
	
	alpha = remTime - ((retTime - nowTime)/2.0) - nowTime;
	
	if ((alpha > LTS_IN_SYNCH_WINDOW) || (alpha < -1.0*LTS_IN_SYNCH_WINDOW)) {
	    if (ltsPtr->resynchTime > LTS_MIN_RESYNCH_TIME) {
		ltsPtr->resynchTime = LTS_MIN_RESYNCH_TIME;
	    }
	    else {
		ltsPtr->offset += alpha;
	    }
	} else {
	    ltsPtr->resynchTime += LTS_RESYNCH_INCREMENT;
	}
    }
  
    if (ltsPtr->speed != 0.0) {
	ltsPtr->resynchAtToken =
	    Tcm_CreateAtHandler(Tcm_ReadSysClock()+ltsPtr->resynchTime,
				HUGE_VAL, "tk", DoSynch, ltsPtr, NULL,
				NULL, NULL, NULL, NULL);
    }
}


/*
 *---------------------------------------------------------------
 *
 * DoSynch
 *     
 *     Starts the synchronization mechanism. Saves current time
 *     in lts structure and then sets up call to master object 
 *     to query its time. If master object is remote, uses dp_RDO,
 *     with DoSynchReturn as the callback, otherwise, does query
 *     to local master object and calls DoSynchReturn directly.
 * 
 * Parameters:
 *
 *     clientData (in) - pointer to lts structure.
 *
 * Results:
 *
 *     None
 *
 * Side Effects:
 *
 *     May initiate a dp_RDO command if master object is remote.
 *    
 *---------------------------------------------------------------
 */

static void
DoSynch(ClientData clientData        /* Pointer to lts structure. */)
{
  Lts *ltsPtr;
  double remTime; 
  Tcl_Interp *interp;

  ltsPtr = (Lts *) clientData;
  interp = ltsPtr->interp;

  ltsPtr->resynchAtToken = 0;

  if (ltsPtr->speed == 0.0) return;

  ltsPtr->synchStart = LTS_Value((ClientData) ltsPtr);
    

  if (ltsPtr->masterRPCFile != NULL) {
    if (Tcl_VarEval(interp, "catch { dp_RDO ", ltsPtr->masterRPCFile, " -callback \"",
		    ltsPtr->name, " synchReturn ", "\" ", 
		    ltsPtr->masterName, " logicalTimeOf now }",(char *) NULL)!= TCL_OK) {
      Tcl_AppendResult(interp, "Bad lts master: ", ltsPtr->masterName, (char *) NULL);
      Tk_BackgroundError(interp);
      return;
    }
    return;
  } else {
    if (Tcl_VarEval(interp, "catch { ", ltsPtr->masterName, 
		    " logicalTimeOf now }", (char *) NULL) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad lts master: ", ltsPtr->masterName, (char *) NULL);
      Tk_BackgroundError(interp);
      return;
    }
    Tcl_GetDouble(interp, interp->result, &remTime);
  }

  DoSynchReturn(ltsPtr, remTime);
}
