/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <math.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>
#include "cmAuPlay.h"

#ifdef LINUX_AU

#include <linux/soundcard.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "cmAuFile.h"
#include "cmLTS.h"

typedef struct WinAudioDevice {
    int fd;			/* File descriptor for /dev/audio */
    int mixerFd;		/* File descriptor for /dev/mixer */
    SparcAuBufferHeader currFmt;/* The last format we changed to */
    int resampling;		/* Are we resampling to new rate? */
    int errorCode;		/* Current error code */
    int atToken;		/* Token for atEvent for playback when
				 * audio device queue is empty */
    int stopToken;		/* Token for atEvent to silence the
				 * device when playback queue empties */
    double lastStartTime;	/* Last time the LinuxAuStartPlay()
				 * function was called. */
    double emptyTime;		/* System time when the data currently
				 * in the stream queue should flush */
} LinuxAudioDevice;

/*
 * Interface routines
 */
static int LinuxAuDeviceInit		(SpeakerRsrc *objPtr);
static int LinuxAuDeviceReinit		(SpeakerRsrc *objPtr);
static int LinuxAuDeviceShutdown	(SpeakerRsrc *objPtr);
static int LinuxAuDeviceUnready		(SpeakerRsrc *objPtr);
static int LinuxAuDeviceLTSChanged	(SpeakerRsrc *objPtr);
static int LinuxAuDevicePlay		(CBInfo *cbPtr);
static char *LinuxAuDeviceError		(SpeakerRsrc *objPtr, int errNum);
SpeakerDevice *GetLinuxAuDevice		(SpeakerRsrc *objPtr);

/*
 * Internal Routines
 */
static void LinuxAuSilence		(LinuxAudioDevice *dev);
static void LinuxAuStartPlay 		(CBInfo *cbPtr);
static void LinuxAuDeviceSetGain	(SpeakerRsrc *objPtr);
static void LinuxAuDeviceSetOutput	(SpeakerRsrc *objPtr);
static int LinuxAuDeviceSetPlayParams	(LinuxAudioDevice *dev,
					 int sample_rate, int channels,
					 int encoding);

#define SKEW		0.05
#define SYNC_TIME	10.0

/*
 * Error codes: start at 100000 so we don't conflict with Windows errors
 */
#define CANT_OPEN_HANDLE_BUSY_ERROR	100000
#define CANT_OPEN_ERROR			100001
#define BAD_DEVICE_ERROR		100002
#define	BAD_SAMPLE_RATE_ERROR		100003
#define NO_AUDIO_DEVICES_ERROR		100004
#define NO_MATCHING_DEVICE_ERROR	100005
#define NOT_OPEN_ERROR			100006
#define TOO_EARLY_ERROR			100007
#define BAD_FORMAT_ERROR		100008
#define TOO_LAZY_ERROR			100009
#define NO_MEMORY_ERROR			100010
#define MEMORY_LOCK_ERROR		100011
#define UNSUPPORTED_FORMAT_ERROR	100012
#define FORMAT_ERROR			100013
#define IN_USE_ERROR			100014
#define OPEN_ERROR			100015
#define OPEN_MIX_ERROR			100016

#ifndef min
#  define min(a,b)	((a)<(b)?(a):(b))
#endif


/*
 * -------------------------------------------------------
 *
 * GetLinuxAuDevice
 *
 *	Allocates a new SpeakerDevice structure, fills
 *	in to handle Sparc Au and return it.
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetLinuxAuDevice(SpeakerRsrc *objPtr)
{
    SpeakerDevice *sdev;
    LinuxAudioDevice *dev;
    
    dev = (LinuxAudioDevice *) ckalloc(sizeof(LinuxAudioDevice));
    if (dev == NULL) {
	return NULL;
    }
    dev->fd = -1;

    sdev = (SpeakerDevice *) ckalloc(sizeof(SpeakerDevice));
    sdev->deviceData = (ClientData) dev;
    sdev->initProc = LinuxAuDeviceInit;
    sdev->reinitProc = LinuxAuDeviceReinit;
    sdev->shutdownProc = LinuxAuDeviceShutdown;
    sdev->unreadyProc = LinuxAuDeviceUnready;
    sdev->ltsProc = LinuxAuDeviceLTSChanged;
    sdev->playProc = LinuxAuDevicePlay;
    sdev->errorProc = LinuxAuDeviceError;
    dev->errorCode = 0;

    return sdev;
}

/*
 * -------------------------------------------------------
 *
 * LinuxAuSilence
 *
 * Flush any pending output (i.e., silence the speaker)
 *
 * -------------------------------------------------------
 */

static void
LinuxAuSilence(LinuxAudioDevice *dev)
{
    if (dev->fd > 0) {
	ioctl(dev->fd, AFMT_MU_LAW, NULL);
    }
}


/*
 * -------------------------------------------------------
 *
 * LinuxAuDeviceInit
 *
 * Initialize this device.
 *
 * -------------------------------------------------------
 */

static int
LinuxAuDeviceInit(SpeakerRsrc *objPtr)
{
    LinuxAudioDevice *dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;
    int nonblock = 1;
    int frag = 0x00100008;	/* 16 fragments of 2^8=256 bytes */
    int rate = 8000;
    int channels = 1;
    int bits = 8;
    int format = AFMT_MU_LAW;
    
    if (dev->fd > 0) {
	return IN_USE_ERROR;
    }

    /* The normal Linux sound device ignores O_NDELAY.  This
     * will eventually change.
     */
    dev->fd = open("/dev/audio", (O_WRONLY | O_NDELAY), 0);
    if (dev->fd < 0) {
	dev->fd = -1;
	return OPEN_ERROR;
    }
    dev->mixerFd = open("/dev/mixer", O_RDWR, 0);
    if (dev->mixerFd < 0) {
	dev->mixerFd = -1;
	return OPEN_MIX_ERROR;
    }

    ioctl(dev->fd, SNDCTL_DSP_SETFRAGMENT, &frag);
    /*
     * allow for audio devices that don't do exactly 8000 Hz
     */
    if ((ioctl(dev->fd, SOUND_PCM_WRITE_RATE, &rate) < 0)
        || (rate < 7920) || (rate > 8080)) {
	close(dev->fd);
	dev->fd = -1;
	return BAD_SAMPLE_RATE_ERROR;
    }

    if ((ioctl(dev->fd, SOUND_PCM_WRITE_CHANNELS, &channels) < 0) ||
	(channels != 1)) {
	close(dev->fd);
	dev->fd = -1;
	return BAD_FORMAT_ERROR;
    }

    if ((ioctl(dev->fd, SOUND_PCM_WRITE_BITS, &bits) < 0) || (bits != 8)) {
	close(dev->fd);
	dev->fd = -1;
	return BAD_FORMAT_ERROR;
    }

    ioctl(dev->fd, SNDCTL_DSP_SETFMT, &format);

    dev->currFmt.samplingRate = rate;
    dev->currFmt.channelCount = channels;
    dev->currFmt.dataFormat = SUN_ULAW;

    LinuxAuDeviceSetGain(objPtr);
    LinuxAuDeviceSetOutput(objPtr);
    dev->errorCode = 0;
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * StopTokenCallback --
 *
 *	Helper function for ResetStopToken
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
void
StopTokenCallback (ClientData clientData)
{
    SpeakerRsrc *objPtr = (SpeakerRsrc *) clientData;
    LinuxAudioDevice *dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;

    LinuxAuSilence(dev);
    dev->stopToken = 0;
}

/*
 *--------------------------------------------------------------
 *
 * ResetStopToken --
 *
 *	Called whenever the emptyTime variable changes.  Resets
 *	the at event associated with the queue becoming empty.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	When the system clock reaches the value specified by
 *	the global variable emptyTime, the speaker will be
 *      silenced.  This is necessary because the speaker otherwise
 *	may get left in a state where it's producing a high
 *	pitched whine.
 *
 *--------------------------------------------------------------
 */
static void
ResetStopToken(SpeakerRsrc *objPtr)
{
    LinuxAudioDevice *dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;

    if (dev->stopToken != 0) {
	Tcm_DeleteAtHandler(dev->stopToken);
	dev->stopToken = 0;
    }
    if (dev->emptyTime != 0.0) {
	dev->stopToken = Tcm_CreateAtHandler(dev->emptyTime, HUGE_VAL,
			"audio", (Tk_TimerProc *)StopTokenCallback,
			(ClientData) objPtr, NULL,
			NULL, (Tk_IdleProc *)NULL, (ClientData)NULL, NULL);
    }
}

/*
 * -------------------------------------------------------
 *
 * LinuxAuDeviceShutdown
 *
 * Destroys the memory that has been allocated for a
 * given LinuxAuDevice
 *
 * -------------------------------------------------------
 */

static int
LinuxAuDeviceShutdown(SpeakerRsrc *objPtr)
{
    LinuxAudioDevice *dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;

    if (dev == NULL) {
	return 0;
    }

    if (dev->atToken != 0) {
	Tcm_DeleteAtHandler(dev->atToken);
	dev->emptyTime = 0;
	dev->atToken = 0;
    }
    if (dev->stopToken != 0) {
	Tcm_DeleteAtHandler(dev->stopToken);
	dev->stopToken = 0;
    }
    LinuxAuSilence(dev);
    if (dev->fd > 0) {
	close(dev->fd);
	dev->fd = -1;
    }
    dev->errorCode = 0;
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * LinuxAuDeviceUnready
 *
 *	Stub routine
 *
 * -------------------------------------------------------
 */

static int
LinuxAuDeviceUnready(SpeakerRsrc *objPtr)
{ 
    LinuxAudioDevice *dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;

    if (dev->atToken != 0) {
	Tcm_DeleteAtHandler(dev->atToken);
	dev->emptyTime = 0;
	dev->atToken = 0;
    }
    if (dev->stopToken != 0) {
	Tcm_DeleteAtHandler(dev->stopToken);
	dev->stopToken = 0;
    }
    LinuxAuSilence(dev);
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * LinuxAuDeviceReinit
 *
 *	Stub routine
 *
 * -------------------------------------------------------
 */
static int
LinuxAuDeviceReinit(SpeakerRsrc *objPtr)
{ 
    LinuxAudioDevice *dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;

    LinuxAuDeviceSetGain(objPtr);
    LinuxAuDeviceSetOutput(objPtr);
    dev->errorCode = 0;
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * LinuxAuDeviceSetPlayParams
 *
 *	Set the sample rate for audio output.  Returns 1
 *	if successfully set, 0 otherwise
 *
 * Parameters:
 *	int sample_rate;         samples per second
 *	int channels;            # of interleaved channels
 *	int encoding;            data encoding format
 *
 *  NOTE: if any parameter is -1, the value is left unchanged.
 *
 * -------------------------------------------------------
 */
static int
LinuxAuDeviceSetPlayParams (LinuxAudioDevice *dev,
			    int sample_rate, int channels, int encoding)
{
    int err;

    /*
     * If not initiatilzed, return
     */
    if (dev->fd == -1) {
	return 0;
    }

    /* XXX: Need to handle additional formats */
    if (encoding != SUN_ULAW) {
	return BAD_FORMAT_ERROR;
    }
    dev->currFmt.dataFormat = encoding;

    if (channels != -1) {
	if (channels == 1) {
	    dev->currFmt.channelCount = channels;
	} else {
	    return BAD_FORMAT_ERROR;
	}
    }

    if (sample_rate == -1) {
	return 0;
    }
    if (sample_rate == dev->currFmt.samplingRate) {
	return 0;
    }

    if (sample_rate != 8000) {
	return BAD_FORMAT_ERROR;
    }
    dev->currFmt.samplingRate = sample_rate;
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * LinuxAuDeviceSetGain
 *
 *	Set the gain of the audio device
 *
 * -------------------------------------------------------
 */

static void
LinuxAuDeviceSetGain (SpeakerRsrc *objPtr)
{
    LinuxAudioDevice *dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;
    int vol;

    if ((objPtr != NULL) && (dev->mixerFd != -1)) {
	vol = (128 * objPtr->gain);
	if (vol > 128) { vol = 128; }
	if (vol < 0) { vol = 0; }
	vol = vol | (vol << 8);
	ioctl(dev->mixerFd, SOUND_MIXER_WRITE_VOLUME, &vol);
    }
}


/*
 * -------------------------------------------------------
 *
 * LinuxAuDeviceSetOutput
 *
 *	Direct the output of the audio device.  On Linux,
 *	this does nothing, at least for now.
 *
 * -------------------------------------------------------
 */

static void
LinuxAuDeviceSetOutput (SpeakerRsrc *objPtr)
{
/* XXX: Need to implement */
#if 0
    int port;

    if ((objPtr != NULL) && (dev->fd != -1)) {
	if (objPtr->headphone) {
	    port = AUDIO_HEADPHONE;
	} else {
	    port = AUDIO_SPEAKER;
	}
	audio_set_play_port (dev->fd, &port);
    }
#endif
}


/*
 * -------------------------------------------------------
 *
 * LinuxAuDeviceLTSChanged
 *
 *	Notify the LinuxAuDevice that the LTS has changed
 *
 * -------------------------------------------------------
 */

static int
LinuxAuDeviceLTSChanged (SpeakerRsrc *objPtr)
{
    LinuxAudioDevice *dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;

    if (dev->atToken != 0) {
	Tcm_DeleteAtHandler(dev->atToken);
	dev->emptyTime = 0;
	dev->atToken = 0;
    }
    if (dev->stopToken != 0) {
	Tcm_DeleteAtHandler(dev->stopToken);
	dev->stopToken = 0;
    }
    LinuxAuSilence(dev);
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * LinuxAuDevicePlay
 *
 * Parameters: 
 *	cbPtr - Frame to be played
 * 
 * Return: 
 *	None
 * 
 * Action: 
 *	Check to make sure the AC is set up properly,
 *	then gives the data to AF to be played.
 *
 * -------------------------------------------------------
 */
static int
LinuxAuDevicePlay(CBInfo *cbPtr)
{
    double now;
    int forceRestart;
    SpeakerRsrc *objPtr;
    LinuxAudioDevice *dev;
    
    objPtr = cbPtr->objPtr;
    dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;

    if (dev->fd < 0) {
	return NOT_OPEN_ERROR;
    }

    /*
     * Check for earlier failures that need special recovery (e.g.
     * unplayable format)
     */
    if (dev->errorCode != 0) {
	return dev->errorCode;
    }

    if (dev->atToken != 0) {
	return TOO_EARLY_ERROR;
    }

    /*
     * The following code s tricky because it attempts to keep the
     * device queue full.  Since there's no way to directly query
     * the state of the device queue, we use the following strategy.
     * The Rsrc will normally call us a little while (SKEW_TIME, about
     * 200 milliseconds) before we're ready to play.  There are 2 cases:
     *  1. if the device queue has data in it, just write the data
     *	   to the device.
     *	2. if the device queue is empty, wait until the real start
     *	   time to write the data.
     *  3. if the device parameters have changed (e.g., a different
     *	   data format), wait until the real start time to reset the
     *	   parameters and then write the data.
     *  4. If we've been writing data continuously in our attempts
     *	   to keep the queue full for a long time (SYNC_TIME), the
     *	   synchronization might have drifted.  So, wait until the
     *	   real start time, flush any remaining data, and then write
     *	   the data.
     *
     * Actions 2-4 are performed by the LinuxAuStartPlay() funciton
     * below, since they are fired by an at event.  They care signaled
     * by setting the forceRestart variable to 1.
     *
     * In any case, the emptyTime variable is set to when the
     * queue will become empty again.
     */
    forceRestart = 0;

    /*
     * Check to see if the format has changed.  If so, wait for
     * device to drain and force restart.
     */
    if ((cbPtr->hdr.samplingRate != dev->currFmt.samplingRate) ||
	(cbPtr->hdr.channelCount != dev->currFmt.channelCount) ||
	(cbPtr->hdr.dataFormat != dev->currFmt.dataFormat)) {
	forceRestart = 1;
    }

    /*
     * Check to see if it's been awhile (SYNC_TIME) since we did
     * the last restart.  If so, force a restart to keep synchronization
     * from drifting
     */
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));
    if (now > dev->lastStartTime + SYNC_TIME) {
	forceRestart = 1;
    }

    /*
     * If we think the device queue is empty (now > emptyTime), force
     * a restart so we start as close to play time as possible.
     */
    if (now > dev->emptyTime) {
	forceRestart = 1;
    }

    /*
     * Play it directly or wait.
     */
    if (forceRestart) {
	Tk_Preserve(cbPtr);
	AttachBuffer (objPtr->interp, cbPtr->buffer,
			"LinuxAuDevice:LinuxAuDevicePlay");
	
	dev->atToken = Tcm_CreateAtHandler(cbPtr->start-SKEW, HUGE_VAL,
			    "audio", (Tk_TimerProc *)LinuxAuStartPlay,
			    (ClientData) cbPtr, NULL,
			    NULL, (Tk_IdleProc *)NULL, (ClientData)NULL, NULL);

    } else {
	write(dev->fd, cbPtr->data, cbPtr->len);
	dev->emptyTime += (cbPtr->end - cbPtr->start);
	ResetStopToken(objPtr);
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * LinuxAuStartPlay --
 *
 *	Starts off the initial playback to the device.  Only
 *	called when teh device queu is thought to be empty.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Writes data to the device and updates emptyTime
 *
 *--------------------------------------------------------------
 */
static void
LinuxAuStartPlay (CBInfo *cbPtr)
{
    double now, fraction;
    int offset, n, width;
    SpeakerRsrc *objPtr;
    LinuxAudioDevice *dev;

    objPtr = cbPtr->objPtr;
    DetachBuffer (objPtr->interp, cbPtr->buffer,
		  "LinuxAuDevice:LinuxAuStartPlay");
    dev = (LinuxAudioDevice *) objPtr->devicePtr->deviceData;
    dev->atToken = 0;
    if (dev->fd < 0) {
	Tk_Release(cbPtr);
	return;
    }

    /*
     * After either action, the emptyTime variable is set to when the
     * queue will become empty again.
     */
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));

    /*
     * Flush any data in the output queue
     */
    LinuxAuSilence(dev);

    /*
     * Set up the device parameters
     */
    if (LinuxAuDeviceSetPlayParams (dev, cbPtr->hdr.samplingRate,
			cbPtr->hdr.channelCount, cbPtr->hdr.dataFormat)) {
	dev->errorCode = FORMAT_ERROR;
	Tk_Release(cbPtr);
	return;
    }

    if (cbPtr->hdr.dataFormat == SUN_LIN_16) {
        width = 2;
    } else {
        width = 1;
    }

    /*
     * If callback is late, just play the appropriate fraction of the data
     */
    if (now > cbPtr->start) {
	fraction = (now - cbPtr->start) / (cbPtr->end - cbPtr->start);
	if (fraction > 1.0) {
	    Tk_Release(cbPtr);
	    return;
	}
	if (fraction < SKEW) {
	    fraction = 0.0;
	}
    } else {
	fraction = 0.0;
    }
    offset = fraction * cbPtr->len + 0.499;

    /*
     * Byte align the offset correctly
     */
    offset = width*(offset/width);
    n = write(dev->fd, cbPtr->data + offset, cbPtr->len - offset);
    dev->emptyTime = now + (cbPtr->end - cbPtr->start)*(1.0 - fraction);
    ResetStopToken(objPtr);
    Tk_Release(cbPtr);
    dev->lastStartTime = now;
    return;
}
/*
 *--------------------------------------------------------------
 *
 * LinuxAuDeviceError --
 *
 *      Map an error code (an integer returned by one of the functions
 *      above) into a human readable string.
 *
 * Results:
 *      Constant string.
 *
 * Side effects:
 *      None
 *
 *--------------------------------------------------------------
 */
static char *
LinuxAuDeviceError(SpeakerRsrc *rsrcPtr, int errNum)
{
    switch (errNum) {
      case CANT_OPEN_HANDLE_BUSY_ERROR:
	return
	    "Unable to properly close currently open device: handle is busy";

      case CANT_OPEN_ERROR:
	return "Unable to open device";

      case NO_MATCHING_DEVICE_ERROR:
	return "No useful PCM audio device was found";

      case BAD_DEVICE_ERROR:
	return "Invalid device";
	
      case BAD_SAMPLE_RATE_ERROR:
	return "Sampling rate incompatible with audio device";

      case BAD_FORMAT_ERROR:
	return "Audio format is incompatible with audio device";

      case UNSUPPORTED_FORMAT_ERROR:
	return "Audio format is currently unsupported";

      case NO_AUDIO_DEVICES_ERROR:
	return "No audio devices are available on this machine";
	
      case NOT_OPEN_ERROR:
	return "Audio device is not open";

      case TOO_EARLY_ERROR:
	return "Send-ahead time must be less than data length";

      case FORMAT_ERROR:
	  return "Audio data is of an unhandled format";

      case NO_MEMORY_ERROR:
	return "Out of memory";

      case MEMORY_LOCK_ERROR:
	return "Failed to lock memory for data chunk";

      case IN_USE_ERROR:
	  return "Audio Device is already in use by another object";

      case OPEN_ERROR:
	  return "Error opening /dev/audio";

      case OPEN_MIX_ERROR:
	  return "Error opening /dev/mixer";

      default:
	return "Unknown LinuxAudio error";
    }
}

#else /* LINUX_AU */

/*
 * -------------------------------------------------------
 *
 * GetLinuxAuDevice
 *
 *	Stub routine for non sparc machines
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetLinuxAuDevice(SpeakerRsrc *objPtr)
{
    return (SpeakerDevice *)NULL;
}

#endif /* SPARC_AU */
