/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * log.c
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "tcl.h"
#include "dp.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmUtil.h"

static FILE *logFile;
static char tag[256];
static int tagLen;
static ClientData ltsPtr;


/*
 *--------------------------------------------------------------
 *
 * Tcm_LogCmd --
 *
 *	This procedure is invoked to process the "log" Tcl/Tk
 *	command.  See the user documentation for details on what
 *	it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
int
Tcm_LogCmd(ClientData notUsed, Tcl_Interp *interp, int argc, char **argv)
                       
                       		/* Current interpreter. */
             			/* Number of arguments. */
                		/* Argument strings. */
{
    char c;
    int len;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    /*---------------------------- OPEN -----------------------------*/
    if ((c == 'o') && (strncmp(argv[1], "open", len) == 0)) {
	FILE *filePtr;

	/*
	 * Check syntax and validity of args.
	 */
	if ((argc < 4) || (argc > 5)) {
	    sprintf(interp->result,
		    "wrong # args: should be \"%s open fileId tag ?lts?\"",
		    argv[0]);
	    return TCL_ERROR;
	}
	
	if (argc == 5) {
	  ltsPtr = (ClientData) Tcm_FindAtom(interp, argv[4]);
	  if (ltsPtr == NULL) {
	    sprintf(interp->result, "%s open: can't find lts \"%s\"",
		    argv[0], argv[4]);
	    return TCL_ERROR;
	  }
	}
	else ltsPtr = NULL;

	if (Tcl_GetOpenFile(interp, argv[2], 1, 1, &filePtr) != TCL_OK) {
	    return TCL_ERROR;
	}
	if ((logFile != NULL) && (fileno(logFile) > 2)) {
	    fclose(logFile);
	}

	logFile = filePtr;

	sprintf(tag, "%s\t", argv[3]);
	tagLen = strlen(tag);

    /*---------------------------- WRITE ----------------------------*/
    } else if ((c == 'w') && (strncmp(argv[1], "write", len) == 0)) {
	char buffer[4096];
	char *dst;
	int i, numWritten, toWrite;
	/*
	 * Format up the output and send it off
	 */
	if (logFile != NULL) {
	    dst = buffer;

	    if (ltsPtr == NULL)
	      sprintf(dst, "%lf\t", Tcm_ReadSysClock());
	    else
	      sprintf(dst, "%lf\t", LTS_Value(ltsPtr));

	    dst += strlen(dst);
	    strcpy(dst, tag);
	    dst += tagLen;
	    for (i = 2; i < argc; i++) {
		strcpy(dst, argv[i]);
		dst += strlen(argv[i]);
		*dst++ = '\t';
	    }
	    dst--;
	    *dst++ = '\n';
	    toWrite = dst - buffer;
	    numWritten = fwrite(buffer, 1, toWrite, logFile);
	    if (numWritten != toWrite) {
		sprintf(interp->result, "Error writing log: %s\n",
			Tcl_PosixError(interp));
		return TCL_ERROR;
	    }
	}
    /*---------------------------- CLOSE ----------------------------*/
    } else if ((c == 'c') && (strncmp(argv[1], "close", len) == 0)) {
	/*
	 * Reset all the variables.
	 */
	if (logFile != NULL) {
	    tag[0] = 0;
	    tagLen = 0;
	    logFile = NULL;
	    ltsPtr = NULL;
	}
    } else {
	sprintf(interp->result, "%s: must be \"open\" \"close\" or \"write\"",
		argv[0]);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_Log --
 *
 *	Write a string out to the log file.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Log file is written.
 *
 *--------------------------------------------------------------
 */
void
Tcm_Log(char *str)
{
    char buffer[4096];
    char *dst;
    static int first = 1;

    /*
     * Format up the output and send it off
     */
    if (logFile != NULL) {
	dst = buffer;

	if (ltsPtr == NULL)  {
	    sprintf(dst, "%lf\t", Tcm_ReadSysClock());
	} else {
	    sprintf(dst, "%lf\t", LTS_Value(ltsPtr));
	}

	if (first) {
	    write (2, "Logging using gettimeofday\n", 28);
	    first--;
	}
	dst += strlen(dst);
	strcpy(dst, tag);
	dst += tagLen;
	strcpy(dst, str);
	dst += strlen(str);
	*dst++ = '\n';
	fwrite(buffer, 1, dst - buffer, logFile);
    }
}
