/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <math.h>

#ifdef NO_STRING_H
#   include "compat/string.h"
#else
#   include <string.h>
#endif

#include <sys/types.h>
#include <netinet/in.h>
#include <dp.h>
#include "cmMJPEGCapture.h"
#include "tcmEvent.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmBind.h"
#include "cmBufferlist.h"
#include "cmUtil.h"

#include <sys/types.h>
#include <netinet/in.h>

#define ONE_BILLION 1000000000.0
#define DEFAULT_QFACTOR 150
#define DEFAULT_DELAY 0.100

static int nextMJPEGCaptureNum = 0;	/* Variable for next unique object
					 * name. */

/*
 * Prototypes for routines used internally
 */
static MJPEGCapture *CreateNewMJPEGCapture(Tcl_Interp *interp, char *name);
static int	MJPEGCaptureCreateCmd(ClientData clientData,
		    Tcl_Interp *interp, int argc, char **argv);
static int	MJPEGCaptureObjCmd(ClientData clientData,
		    Tcl_Interp *interp, int argc, char **argv);

static void 	ChangeDeviceType(MJPEGCapture *objPtr, char *devtype);
static int	ConfigureObject(MJPEGCapture *objPtr,
		    int argc, char **argv);
static int	DestroyObject(MJPEGCapture *objPtr);
static char *	GetCumulLenEncoded(Tcl_Interp *interp, ClientData clientdata);
static char *	GetDataLen(Tcl_Interp *interp, ClientData clientdata);
static char *	GetHeight(Tcl_Interp *interp, ClientData clientdata);
static char *	GetLogEnd(Tcl_Interp *interp, ClientData clientdata);
static char *	GetLogStart(Tcl_Interp *interp, ClientData clientdata);
static char *	GetQFactor(Tcl_Interp *interp, ClientData clientdata);
static char *	GetWidth(Tcl_Interp *interp, ClientData clientdata);
static int	DeviceListFunc(MJPEGCapture *objPtr);
static void	FreeFrame(MJPEGFrame *framePtr);
static void	LTSChanged(ClientData clientData);
static void     DoCapture(MJPEGCapture *objPtr, int asynchFlag);
static void     AtQueueCapture(MJPEGCapture *objPtr);
static int      DoEncode(MJPEGFrame *framePtr, int asynchFlag);
static int      DoneEncode(MJPEGFrame *framePtr);
static void     ContinueEncode (MJPEGFrame *framePtr);
static int	Ready(MJPEGCapture *objPtr);
static int	Stats(MJPEGCapture *objPtr);
static int	Unready(MJPEGCapture *objPtr);
static int	Trigger(MJPEGCapture *objPtr);
static int      InternalTrigger(MJPEGCapture *objPtr, SB_List **sblPtr);
static int      MJpegCaptureRefresh(MJPEGCapture *objPtr);



/*
 *----------------------------------------------------------------------
 *
 * MJPEGCaptureModuleInit
 *
 * Parameters:
 *	interp - Tcl interpreter
 *
 * Return:
 *	Standard Tcl return value.
 *
 * Action:
 *	Initializes the object module, registering
 *	media, format, and object keywords as well
 *	as creating Tcl command for object type creation
 *
 *----------------------------------------------------------------------
 */

int
MJPEGCaptureModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mjpegCapture", MJPEGCaptureCreateCmd,
		      NULL, (void (*)()) NULL);
    Tcl_GlobalEval(interp, "cminfo add media {video}");
    Tcl_GlobalEval(interp, "cminfo add formats video {jpeg}");
    Tcl_GlobalEval(interp, "cminfo add objects video jpeg mjpegCapture");
    Tcl_VarEval(interp, "cminfo add object_info mjpegCapture {capture",
#ifdef PARALLAX
		" parallax",
#endif
		"}", NULL);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DeviceListFunc
 *
 * Parameters:
 *	objPtr - pointer to mpeg capture object.
 *
 * Results:
 *	Standard Tcl Result.
 *
 * Actions:
 *	Sets result to list of devices supported. Returns TCL_OK;
 *
 *----------------------------------------------------------------------
 */
static int
DeviceListFunc(MJPEGCapture *objPtr)
{
    Tcl_Interp *interp = objPtr->interp;
    Tcl_DString resStr;

    Tcl_DStringInit(&resStr);

#ifdef PARALLAX
    Tcl_DStringAppendElement(&resStr, "parallax");
#endif

    Tcl_SetResult(interp, Tcl_DStringValue(&resStr), TCL_VOLATILE);
    Tcl_DStringFree(&resStr);

    return TCL_OK;
}


/************************************************************
 *
 * ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *     newOut - new out command
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
ChangeOutCmd(MJPEGCapture *objPtr, char *newOut) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
		&(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    


/*
 *--------------------------------------------------------------
 *
 * ChangeDeviceType --
 *
 *	Sets the device type for the MJPEGReseource.  Called whenever
 *	the "devtype" slot is changed, and on initialization of the
 *	capture.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The old device is shut down, if appropriate, and the new
 *	device is created.
 *
 * Parameters:
 *	MJPEGCapture *objPtr	Pointer to capture object.
 *	char *devtype		String giving name of new device
 *
 *--------------------------------------------------------------
 */
static void
ChangeDeviceType(MJPEGCapture *objPtr, char *devtype)
{
    int len;
    char c;

    len = strlen(devtype);
    c = devtype[0];

    /*
     * If there's no change, return
     */
    if ((objPtr->devtype != NULL) &&
	(strncmp(objPtr->devtype, devtype, len) == 0)) {
	return;
    }

    /*
     * Shut down old device, if necessary, ignoring any errors it
     * returns.
     */
    Unready(objPtr);

    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
	objPtr->devtype = NULL;;
    }
    if (objPtr->devicePtr != NULL) {
	if (objPtr->devicePtr->shutdownProc != NULL) {
	    objPtr->devicePtr->shutdownProc(objPtr);
	}
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
	objPtr->inited = 0;
    }

    /*
     * Create new device.
     */
#ifdef PARALLAX
    if ((c == 'p') && (strncmp("parallax", devtype, len) == 0)) {
	objPtr->devicePtr = GetParallaxCaptureDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("parallax")+1);
	strcpy(objPtr->devtype, "parallax");

    } else 
#endif
      if (strlen(DEFAULT_JPEG_CAPTURE_DEV_TYPE) != 0) {
	  objPtr->devicePtr = GetDefaultJpegCaptureDevice(objPtr);
	  objPtr->devtype = ckalloc(strlen(DEFAULT_JPEG_CAPTURE_DEV_TYPE)+1);
	  strcpy(objPtr->devtype, DEFAULT_JPEG_CAPTURE_DEV_TYPE);
      } else {
	  objPtr->devicePtr = NULL;
	  objPtr->devtype = ckalloc(1);
	  objPtr->devtype[0] = '\0';
      }
}

/*
 *----------------------------------------------------------------------
 *
 * MJPEGCaptureCreateCmd --
 *
 * Parameters:
 *	clientData - pointer to data when command was created.
 *		     Not used here.
 *	interp - pointer to Tcl interpreter
 *	argc - argument count
 *	argv - arguments
 *
 * Return:
 *	Standard Tcl return value.
 *
 * Action:
 *	This command implements the "mjpegCapture" Tcl command.
 *	A new object of type MJPEGCapture is created and initialized.
 *	Naming of the new object follows the proper object
 *	naming scheme for all cm objects. A Tcl command is bound
 *	to the object name to implement the objects methods.
 *
 *----------------------------------------------------------------------
 */
static int
MJPEGCaptureCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{
    char new_name[256];
    MJPEGCapture *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: \"", argv[0],
			 " parent ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }
    nextMJPEGCaptureNum++;
    sprintf(new_name, "%smjpegCapture%d", argv[1], nextMJPEGCaptureNum);

    new = CreateNewMJPEGCapture(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MJPEGCaptureObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureObject(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * MJPEGCaptureObjCmd --
 *	This command implements the actual methods of MJPEGCapture object.
 *	It is passed a pointer to the actual data structure of the particular
 *	object instance.
 *
 * Parameters:
 *	clientData - pointer to object.
 *	interp - pointer to Tcl interpreter
 *	argc - argument count
 *	argv - arguments
 *
 * Results:
 *	Standard Tcl return value.
 *
 * Side Effects:
 *
 *----------------------------------------------------------------------
 */
static int
MJPEGCaptureObjCmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char **argv)
{
    char c;
    int len;
    MJPEGCapture *objPtr = (MJPEGCapture *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: \"",
		argv[0], " <command> ?arguments?\"", (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureObject(clientData, argc - 2, argv + 2);

    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0) && (len > 2)) {
	return DestroyObject(clientData);

    } else if ((c == 'd') && (strncmp("deviceList", argv[1], len) == 0) &&
	       (len > 2)) {
	return DeviceListFunc(objPtr);

    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return Ready(objPtr);

    } else if ((c == 'r') && (strncmp("refresh", argv[1], len) == 0)) {
	return MJpegCaptureRefresh(objPtr);

    } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
	return Stats(objPtr);

    } else if ((c == 't') && (strncmp("trigger", argv[1], len) == 0)) {
	/* Should a couple of time arguments, but not for now */
	return Trigger(objPtr);

    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return Unready(objPtr);

    } else {
	Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ", argv[1], 
			 "\n <command> can be one of:\n",
			 "configure ?-slot ?value??\n",
			 "destroy\n",
			 "deviceList\n",
			 "ready\n",
			 "refresh\n",
			 "stats\n",
			 "trigger\n",
			 "unready\n", 
			 (char *) NULL);
	return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CreateNewMJPEGCapture
 *
 * Parameters:
 *	interp - pointer to Tcl interpreter
 *	name - Tcl name for new object
 *
 * Results:
 *	Pointer to new object with fields initialized to default values.
 *
 * Action:
 *	Allocates memory for new object and initializes it.
 *
 *----------------------------------------------------------------------
 */
static MJPEGCapture *
CreateNewMJPEGCapture(Tcl_Interp *interp, char *name)
{
    MJPEGCapture *new;
    char **event_list, **subst_list;
    CM_BindProc **subst_procs;
    ClientData *subst_cds;
    char *suffix;
    int i;

    new = (MJPEGCapture *) ckalloc(sizeof(MJPEGCapture));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);

    new->width = 0;
    new->height = 0;
    new->xpos = -1;
    new->ypos = -1;
    new->squeeze = 0;
    new->devtype = NULL;
    new->devicePtr = NULL;
    new->qfactor = DEFAULT_QFACTOR;
    new->lts = NULL;
    new->ltsTriggerNum = -1;
    new->ltsName = NULL;
    new->maxFrameRate = 0.0;
    new->statStartTime = 0.0;
    new->fps = 0.0;
    new->captured = 0;
    new->bytes = 0;
    new->ready = 0;
    new->inited = 0;
    new->inAsynchEncode = 0;
    new->atToken = 0;
    new->currFrame = NULL;
    new->encodeFrame = NULL;

    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;

    new->delay = DEFAULT_DELAY;

    suffix = "trigger";
    new->triggerHandle = ckalloc(strlen(new->name) + strlen(suffix) + 3);
    sprintf(new->triggerHandle, "%s.%s", new->name, suffix);
    Tutil_RegisterCommand(interp, new->triggerHandle, (ClientData) InternalTrigger,
			  (ClientData) new);

    /*
     * Set up bindings with binding mechanism.
     */

    event_list = (char **) ckalloc(3*sizeof(char *));
    event_list[0] = (char *) ckalloc(strlen("<FRAME_CAPTURED>")+1);
    strcpy(event_list[0], "<FRAME_CAPTURED>");
    event_list[1] = (char *) ckalloc(strlen("<FRAME_ENCODE_BEGIN>")+1);
    strcpy(event_list[1], "<FRAME_ENCODE_BEGIN>");
    event_list[2] = (char *) ckalloc(strlen("<FRAME_ENCODE_DONE>")+1);
    strcpy(event_list[2], "<FRAME_ENCODE_DONE>");

    subst_list = (char **) ckalloc(7*sizeof(char *));
    subst_procs = (CM_BindProc **) ckalloc(7*sizeof(CM_BindProc *));
    subst_cds = (ClientData *) ckalloc(7*sizeof(ClientData));

    subst_list[0] = (char *) ckalloc(2);
    (subst_list[0])[0] = 's';
    (subst_list[0])[1] = '\0';
    subst_procs[0] = GetLogStart;
    subst_cds[0] = (ClientData) new;

    subst_list[1] = (char *) ckalloc(2);
    (subst_list[1])[0] = 'e';
    (subst_list[1])[1] = '\0';
    subst_procs[1] = GetLogEnd;
    subst_cds[1] = (ClientData) new;

    subst_list[2] = (char *) ckalloc(2);
    (subst_list[2])[0] = 'l';
    (subst_list[2])[1] = '\0';
    subst_procs[2] = GetDataLen;
    subst_cds[2] = (ClientData) new;

    subst_list[3] = (char *) ckalloc(2);
    (subst_list[3])[0] = 'w';
    (subst_list[3])[1] = '\0';
    subst_procs[3] = GetWidth;
    subst_cds[3] = (ClientData) new;

    subst_list[4] = (char *) ckalloc(2);
    (subst_list[4])[0] = 'h';
    (subst_list[4])[1] = '\0';
    subst_procs[4] = GetHeight;
    subst_cds[4] = (ClientData) new;

    subst_list[5] = (char *) ckalloc(2);
    (subst_list[5])[0] = 'b';
    (subst_list[5])[1] = '\0';
    subst_procs[5] = GetCumulLenEncoded;
    subst_cds[5] = (ClientData) new;

    subst_list[6] = (char *) ckalloc(2);
    (subst_list[6])[0] = 'q';
    (subst_list[6])[1] = '\0';
    subst_procs[6] = GetQFactor;
    subst_cds[6] = (ClientData) new;

    new->bindToken =
	CreateCMBinding(interp, new->name, 3, event_list,
			7, subst_list, subst_procs, subst_cds);
    new->bindFramePtr = NULL;

    for (i=0; i<3; i++) {
	ckfree(event_list[i]);
    }
    ckfree((char *) event_list);

    for (i=0; i<7; i++) {
	ckfree(subst_list[i]);
    }
    ckfree((char *) subst_list);
    ckfree((char *) subst_procs);
    ckfree((char *) subst_cds);

    ChangeDeviceType(new, DEFAULT_JPEG_CAPTURE_DEV_TYPE);

    return new;
}


/*
 *----------------------------------------------------------------------
 *
 * DestroyObject
 *	Destroys previously created object
 *
 * Parameters:
 *	objPtr - pointer to object.
 *
 * Results:
 *	Standard TCL Result
 *
 *----------------------------------------------------------------------
 */
static int
DestroyObject(MJPEGCapture *objPtr)
{
    int result = TCL_OK;
    int i, devReturn;

    Unready(objPtr);

    Tcl_DeleteCommand(objPtr->interp, objPtr->name);


    /*
     * Shut down the device
     */
    if (objPtr->devicePtr != NULL) {
	if (objPtr->devicePtr->shutdownProc != NULL) {
	    devReturn = objPtr->devicePtr->shutdownProc(objPtr);
	    if (devReturn != 0) {
		result = TCL_ERROR;
		if (objPtr->devicePtr->errorProc != NULL) {
		    Tcl_AppendResult(objPtr->interp,
				     objPtr->devicePtr->errorProc(objPtr, devReturn),
				     (char *) NULL);
		}
	    }
	}
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
    }
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
    }

    /*
     * Clean up lts
     */
    if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
        Tk_Release(objPtr->lts);
        objPtr->lts = NULL;
    }

    /*
     * Unregister the trigger atom
     */
    if (objPtr->triggerHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->triggerHandle);
	ckfree(objPtr->triggerHandle);
	objPtr->triggerHandle = NULL;
    }

    /*
     * Free up space associated with out command.
     */

    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
	objPtr->outCmd = NULL;
    }

    /*
     * Free up previous allocs and such.
     */
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    /*
     * Unregister with bind mechanism.
     */

    DestroyCMBinding(objPtr->interp, objPtr->bindToken);

    ckfree((char *) objPtr);
    return result;
}


/*
 *----------------------------------------------------------------------
 *
 * ConfigureObject
 *
 * Parameters:
 *	objPtr - pointer to the object
 *	argc - num of config arguments
 *	argv - config arguments
 *
 * Return:
 *	Standard TCL return value
 *
 * Action:
 *	Does the usual configuration thing
 *
 *----------------------------------------------------------------------
 */
static int
ConfigureObject(MJPEGCapture *objPtr, int argc, char **argv)
{

    Tcl_Interp *interp;
    int cntr;
    char str[100], str2[100];

    interp = objPtr->interp;

    /*
     * Case 1: <obj> configure
     */
    if (argc == 0) {
	Tcl_PrintDouble(interp, objPtr->fps, str);
	Tcl_AppendResult(interp, "{", "-fps ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "%d", objPtr->width);
	Tcl_AppendResult(interp, "{", "-width ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "%d", objPtr->height);
	Tcl_AppendResult(interp, "{", "-height ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "%d", objPtr->xpos);
	Tcl_AppendResult(interp, "{", "-xpos ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "%d", objPtr->ypos);
	Tcl_AppendResult(interp, "{", "-ypos ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "%d", objPtr->squeeze);
	Tcl_AppendResult(interp, "{", "-squeeze ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "%d %d", DEFAULT_QFACTOR, objPtr->qfactor);
	Tcl_AppendResult(interp, "{", "-qfactor ", str, "} ",
			 (char *)NULL);
	Tcl_AppendResult(interp, "{", "-device ", "{",
			 DEFAULT_JPEG_CAPTURE_DEV_TYPE, "} ",
			 objPtr->devtype, "} ", (char *)NULL);
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *) NULL);
	}
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{}", "} ",
			 (char *)NULL);
	} else {
	  Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{",
			   objPtr->ltsName, "}", "} ", (char *) NULL);
	}
	Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	Tcl_AppendResult(interp, "{", "-maxFrameRate ", "0 ", str, "} ",
			 (char *) NULL);

	Tcl_PrintDouble(interp, objPtr->delay, str);
	Tcl_PrintDouble(interp, DEFAULT_DELAY, str2);
	Tcl_AppendResult(interp, "{", "-delay ", str2, " ", str, "} ",
			 (char *) NULL);
	/*
	 * Give the device a chance to add any options of its own.
	 */

	if (objPtr->devicePtr != NULL) {
	    if (objPtr->devicePtr->configProc != NULL) {
		objPtr->devicePtr->configProc(objPtr, NULL, NULL);
	    }
	}
	return TCL_OK;
    }

    /*
     * Case 2: <obj> configure <-slot>
     */
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 'f') && (strncmp("-fps", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->fps, str);
	    Tcl_AppendResult(interp, "-fps ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'w') && (strncmp("-width", argv[0], len) == 0)) {
	    sprintf(str, "%d", objPtr->width);
	    Tcl_AppendResult(interp, "-width ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'h') && (strncmp("-height", argv[0], len) == 0)) {
	    sprintf(str, "%d", objPtr->height);
	    Tcl_AppendResult(interp, "-height ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'x') && (strncmp("-xpos", argv[0], len) == 0)) {
	    sprintf(str, "%d", objPtr->xpos);
	    Tcl_AppendResult(interp, "-xpos ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'y') && (strncmp("-ypos", argv[0], len) == 0)) {
	    sprintf(str, "%d", objPtr->ypos);
	    Tcl_AppendResult(interp, "-ypos ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 's') && (strncmp("-squeeze", argv[0], len) == 0)) {
	    sprintf(str, "%d", objPtr->squeeze);
	    Tcl_AppendResult(interp, "-squeeze ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'q') && (strncmp("-qfactor", argv[0], len) == 0)) {
	    sprintf(str, "%d %d", DEFAULT_QFACTOR, objPtr->qfactor);
	    Tcl_AppendResult(interp, "-qfactor ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-device", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-device ", "{", DEFAULT_JPEG_CAPTURE_DEV_TYPE, "} ", 
			     objPtr->devtype,
			     (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	  if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "-lts ", "{} ", "{}", (char *) NULL);
	  } else {
	    Tcl_AppendResult(interp, "-lts ", "{} ", "{", objPtr->ltsName, "}", (char *) NULL);
	  }
	    return TCL_OK;
	} else if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	    Tcl_AppendResult(interp, "-maxFrameRate ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-delay", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->delay, str);
	    Tcl_PrintDouble(interp, objPtr->delay, str2);
	    Tcl_AppendResult(interp, "-delay ", str2, " ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *) NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd, "} ",
				 (char *) NULL);
	    }
	    return TCL_OK;
	} else if ((objPtr->devicePtr != NULL) && 
		   (objPtr->devicePtr->configProc != NULL)) {
	    if (objPtr->devicePtr->configProc(objPtr, argv[0], NULL) != 0) 
	      goto bad_option;
	    return TCL_OK;
	}
      bad_option:
	Tcl_AppendResult(interp, "unknown configuration option \"", argv[0],
			 "\": must be one of -width, -height, -xpos, -ypos,"
			 " -squeeze, -xid, -squeeze, -device, -lts, "
			 " -maxFrameRate, -qfactor,"
			 " or a device configuration option",
			 (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: <obj> configure <-slot> <newValue>
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;

	/*
	 * Make capture unready since something is changing.
	 * Should change this so that only options that need unready
	 * actually get unready.
	 */

	Unready(objPtr);

	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	if ((c == 'f') && (strncmp("-fps", argv[cntr], len) == 0)) {
	    double newFps;

	    cntr++;
	    if (cntr == argc) {
noValueError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newFps) != TCL_OK) {
badValueError:
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->fps = newFps;
	} else if ((c == 'w') && (strncmp("-width", argv[cntr], len) == 0)) {
	    int newWidth;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newWidth) != TCL_OK) {
	        goto badValueError;
	    }
	    objPtr->width = newWidth;
	} else if ((c == 'h') && (strncmp("-height", argv[cntr], len) == 0)) {
	    int newHeight;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newHeight) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->height = newHeight;
	} else if ((c == 'x') && (strncmp("-xpos", argv[cntr], len) == 0)) {
	    int newXpos;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newXpos) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->xpos = newXpos;
	} else if ((c == 'y') && (strncmp("-ypos", argv[cntr], len) == 0)) {
	    int newYpos;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newYpos) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->ypos = newYpos;
	} else if ((c == 's') && (strncmp("-squeeze", argv[cntr], len) == 0)) {
	    int newSqueeze;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newSqueeze) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->squeeze = newSqueeze;
	} else if ((c == 'q') && (strncmp("-qfactor", argv[cntr], len) == 0)) {
	    int newQfactor;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newQfactor) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->qfactor = newQfactor;
	} else if ((c == 'd') && (strncmp("-device", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    ChangeDeviceType(objPtr, argv[cntr]);
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (strlen(argv[cntr]) == 0) {
	        newLts = NULL;
	    }
	    else if ((newLts = Tcm_FindAtom(interp, argv[cntr])) == NULL) {
		goto badValueError;
	    }
	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }

            /*
             * Release our lock on the old lts...
             */
            if (objPtr->lts != NULL) {
                Tk_Release(objPtr->lts);
            }

	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
	        objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
	        strcpy(objPtr->ltsName, argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }
	} else if ((c == 'm') &&
		   (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
	    double newMaxFrameRate;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->maxFrameRate = newMaxFrameRate;
	} else if ((c == 'd') &&
		   (strncmp("-delay", argv[cntr], len) == 0)) {
	    double newDelay;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newDelay) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->delay = newDelay;
	} else if ((c == 'o') &&
		   (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	} else if ((objPtr->devicePtr != NULL) && 
		   (objPtr->devicePtr->configProc != NULL)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]) != 0) {
		return TCL_ERROR;
	    }
	}
	cntr++;
    }
    return TCL_OK;
}



/*
 *----------------------------------------------------------------------
 *
 * Ready
 *
 * Parameters:
 *	objPtr - pointer to object
 *
 * Return:
 *	Standard Tcl return value
 *
 * Action:
 *
 *----------------------------------------------------------------------
 */
static int
Ready(MJPEGCapture *objPtr)
{
    int init_res;

    /*
     * Can only be made ready if
     *  1. We have a device.
     *  2. We have an lts
     */

    /* If already ready, just return. */

    if (objPtr->ready) {
	return TCL_OK;
    }

    if (objPtr->devicePtr == NULL) {
        Tcl_AppendResult(objPtr->interp, "Can't make MJPEG capture \"",
			 objPtr->name, "\" ready: device type is not set.",
			 (char *)NULL);
        return TCL_ERROR;
    }

    /*
     * If not inited, do initProc. If inited, do reinitProc. Catch result.  
     */
    init_res = 0;
    if (!objPtr->inited) {
	if (objPtr->devicePtr->initProc != NULL) {
	    init_res = objPtr->devicePtr->initProc(objPtr);
	}
	if (init_res == 0) {
	    objPtr->inited = 1;
	}
    } else {
	if (objPtr->devicePtr->reinitProc != NULL) {
	    init_res = objPtr->devicePtr->reinitProc(objPtr);
	}
    }
    
    /*
     * If bad result, get error string, return error.
     */
    if (init_res != 0) {
	char *errStr;

	if (objPtr->devicePtr->errorProc != NULL) {
	    errStr = objPtr->devicePtr->errorProc(objPtr, init_res);
	} else {
	    errStr = "Error with device initialization.";
	}

	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
    }

    if (objPtr->lts) {
	objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts, LTS_SPEED_FLAG,
					       LTSChanged, objPtr);
    } else {
	Tcl_AppendResult(objPtr->interp, "Can't make MJPEG catpure \"",
			 objPtr->name, "\" ready: lts not set.",
			 (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Reset stat stuff.
     */

    objPtr->statStartTime = Tcm_ReadSysClock();
    objPtr->bytes = 0;
    objPtr->captured = 0;

    objPtr->ready = 1;

    if (LTS_Speed(objPtr->lts) != 0.0) {
	DoCapture(objPtr, 1);
    }

    
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Unready
 *
 * Parameters:
 *	objPtr - pointer to object
 *
 * Return:
 *	Standard Tcl return value
 *
 * Action:
 *
 *----------------------------------------------------------------------
 */

static int
Unready(MJPEGCapture *objPtr)
{
    int i;

    if (!objPtr->ready) return TCL_OK;

    objPtr->ready = 0;

    if (objPtr->atToken) {
	if (Tcm_FindAtEvent(objPtr->atToken) != NULL) {
	    Tcm_DeleteAtHandler(objPtr->atToken);
	}
	objPtr->atToken = 0;
    }

    if (objPtr->ltsTriggerNum != -1) {
	LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	objPtr->ltsTriggerNum = -1;
    }

    if (objPtr->currFrame) {
	FreeFrame(objPtr->currFrame);
	objPtr->currFrame = NULL;
    }

    if (objPtr->inAsynchEncode) {
	Tcm_CancelIdleCall("video", (Tk_IdleProc *) ContinueEncode,
			   objPtr->encodeFrame);
	FreeFrame(objPtr->encodeFrame);
	objPtr->encodeFrame = NULL;
	objPtr->inAsynchEncode = 0;
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * LTSChanged
 *
 * Parameters:
 *	clientdata - pointer to object.
 *
 * Return:
 *	Standard Tcl return value
 *
 * Action:
 *	Called when the LTS changes value.  Frees any previously
 *	captured frames, frees any memory we're holding, and
 *	returns.
 *
 *----------------------------------------------------------------------
 */
static void
LTSChanged(ClientData clientData)
{
    MJPEGCapture *objPtr = (MJPEGCapture *) clientData;
    int i;

    /* The following sequence gets rid of idle calls to encode and
     * at queued calls to DoCapture, but unfortunately also blows away
     * currFrame which we may need for refresh, so we should find a way
     * around this.
     */

    Unready(objPtr);
    Ready(objPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * CreateSBListFromCurr
 *      This routine creates a SB_List structure out of the scatter buffer
 *      pointed in the objPtr->currFrame data structure. Since the scatter
 *      buffer should already be well formated and ready to go, all I need
 *      to do is allocate a scatter buffer list structure and fill it. If
 *      there is anything wrong with what is in currFrame, return an empty
 *      SB_List.
 *
 * 
 * Paramters:
 *      objPtr - pointer to object.
 * 
 * Return:
 *      Pointer to SB_List structure.
 *
 * Action:
 *      Described in description.
 *
 *----------------------------------------------------------------------
 */

static SB_List *
CreateSBListFromCurr(MJPEGCapture *objPtr)
{
    SB_List *sb_list;
    
    sb_list = (SB_List *) ckalloc(sizeof(SB_List));

    /* 
     * Check for valid currFrame. If not valid return empty sb list. 
     */

    if ((objPtr->currFrame == NULL) || (objPtr->currFrame->sb == NULL)) {
	sb_list->count = 0;
	sb_list->sbs = NULL;
    } else {
	sb_list->count = 1;
	sb_list->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *)*1);
	sb_list->sbs[0] = objPtr->currFrame->sb;
    }

    return sb_list;
}
	

/*
 *----------------------------------------------------------------------
 *
 * InternalTrigger
 *	This routine is the registered C equivalent of the trigger command.
 *
 *	Essentially, it causes the device to capture a frame and construct
 *      a C scatter buffer list representation which is passed back in the
 *      pointer pased in.
 *	
 *
 * Parameters: 
 *     objPtr - pointer to object
 *     sblPtr - pointer to scatter buffer list to produce.
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *
 *----------------------------------------------------------------------
 */

static int
InternalTrigger(MJPEGCapture *objPtr, SB_List **sblPtr)
{
    /*
     * If object is not ready, return empty SB List.
     */

    if (!objPtr->ready) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;
	
	return TCL_OK;
    }

    DoCapture(objPtr, 0);

    *sblPtr = CreateSBListFromCurr(objPtr);

    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Trigger
 *	This routine gets call when a trigger gets fired.
 *	This probably should take a parameter that has some information
 *	about the time of the trigger, the duration of the trigger, 
 *	etc.
 *
 *	Essentially, it causes the device to capture a frame and construct
 *      a tcl scatter buffer representation which is produces as the result.
 *	
 *
 * Parameters: 
 *     objPtr - pointer to object
 * 
 * Return: 
 *     Standard Tcl return value
 * 
 * Action: 
 *
 *----------------------------------------------------------------------
 */

static int
Trigger(MJPEGCapture *objPtr)
{
    SB_List *sb_list;
    char *sbStr;

    /*
     * Check for object readiness. If not 
     * ready, return empty list.
     */

    if (!objPtr->ready) {
	Tcl_SetResult(objPtr->interp, "{}", TCL_VOLATILE);
	return TCL_OK;
    }

    /* 
     * Update what is in currFrame. 
     */

    DoCapture(objPtr, 0);

    /*
     * Package in SB List.
     */

    sb_list = CreateSBListFromCurr(objPtr);

    sbStr = MakeSBListString(sb_list);

    Tcl_SetResult(objPtr->interp, sbStr, TCL_VOLATILE);

    ckfree(sbStr);
    ckfree((char *) sb_list);

    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * MJpegCaptureRefresh
 *
 * Parameters:
 *	objPtr - pointer to object
 *
 * Return:
 *	Standard Tcl return value
 *
 * Action:
 *
 *----------------------------------------------------------------------
 */
static int
MJpegCaptureRefresh(MJPEGCapture *objPtr)
{
    if (objPtr->devicePtr != NULL) {
	if (objPtr->devicePtr->refreshProc != NULL) {
	    int res_code;
	    
	    res_code = objPtr->devicePtr->refreshProc(objPtr);
	    
	    if (res_code != 0) {
		char *errStr;
		
		if (objPtr->devicePtr->errorProc != NULL) {
		    errStr = objPtr->devicePtr->errorProc(objPtr, res_code);
		} else {
		    errStr = "Error during capture device refresh";
		}
		
		Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
		return TCL_ERROR;
	    }
	}
    }
    return TCL_OK;
}



/*
 *----------------------------------------------------------------------
 *
 * AtQueueCapture
 *
 * Parameters:
 *	objPtr - pointer to object.
 *
 * Return:
 *      Nothing.
 *
 * Action:
 *      Procedure called by at queue for sending frames.
 *
 *----------------------------------------------------------------------
 */
static void
AtQueueCapture(MJPEGCapture *objPtr)
{
    /* 
     * Be sure to clear at token field so that next capture time
     * will get queued (i.e., won't think we're already queued).
     */

    objPtr->atToken = 0;
    DoCapture(objPtr, 1);
}


/*
 *----------------------------------------------------------------------
 *
 * DoCapture
 *
 * Parameters:
 *	objPtr - pointer to object.
 *      asynchFlag - flag indicating whether capture should be done
 *                   synchronously (i.e., for a trigger command) or
 *                   asynchronously (i.e., for a out command).
 *
 * Return:
 *      Nothing.
 *
 * Action:
 *      Determines if a new frame should be captured or not depending
 *      on the values of the max frame rate attribute, etc. If so, then
 *      starts the process of capturing and encoding the frame.
 *
 *----------------------------------------------------------------------
 */
static void
DoCapture(MJPEGCapture *objPtr, int asynchFlag)
{
    double now, sys_now, speed;
    MJPEGFrame *framePtr;
    
    /* Deal with unreadiness. */

    if ((!objPtr->ready) || (objPtr->inAsynchEncode)) return;

    /* Deal with valid out command. */
    
    if (asynchFlag != 0) {
	if (objPtr->outCmd == NULL) {
	    return;
	}
    }

    now = LTS_Value(objPtr->lts);
    speed = LTS_Speed(objPtr->lts);

    /* Deal with max frame rate issues. */
    
    if (objPtr->maxFrameRate != 0.0) {
	if (objPtr->currFrame != NULL) {
	    if (((now + (objPtr->delay*speed)) - objPtr->currFrame->logStart) < 
		(1.0/objPtr->maxFrameRate)) {
		return;
	    }
	}
    }

    framePtr = (MJPEGFrame *) ckalloc(sizeof(MJPEGFrame));

    framePtr->objPtr = objPtr;
    framePtr->next = framePtr->prev = NULL;
    framePtr->deviceData = NULL;
    framePtr->data = NULL;
    framePtr->dataBuffer = NULL;
    framePtr->sb = NULL;
    framePtr->dataLength = 0;
    framePtr->logStart = now + objPtr->delay*speed;

    /* 
     * Set the start and end times. Here things are a little tricky.
     * If fps is set, then use that to tell you how far in the future
     * end time should be. If fps is not set, however, use the statistics
     * in order to judge how often frames are being asked for.
     */

    if (objPtr->fps > 0.0) {
	framePtr->logEnd = framePtr->logStart + (1.0/objPtr->fps);
    } else {
	if (objPtr->captured > 0) {
	    sys_now = LTS_SystemTimeOf(objPtr->lts, now);
	    framePtr->logEnd = framePtr->logStart + ((sys_now - objPtr->statStartTime)/
						     objPtr->captured);
	} else {
	    framePtr->logEnd = HUGE_VAL;
	}
    }

    framePtr->frameHeader.hdrType = CM_MJPEG_BUF_HDR;
    framePtr->frameHeader.length = sizeof(MJPEGBufferHeader)-8;
    framePtr->frameHeader.width = 0;
    framePtr->frameHeader.height = 0;
    framePtr->frameHeader.device = 0;
    framePtr->frameHeader.subformat = 0;
    framePtr->frameHeader.qfactor = objPtr->qfactor;
    framePtr->frameHeader.logStart_s = floor(framePtr->logStart);
    framePtr->frameHeader.logStart_ns = floor((framePtr->logStart 
					      - floor(framePtr->logStart))
					      *ONE_BILLION);
    framePtr->frameHeader.logEnd_s = floor(framePtr->logEnd);
    framePtr->frameHeader.logEnd_ns = floor((framePtr->logEnd 
					      - floor(framePtr->logEnd))
					      *ONE_BILLION);

    /* 
     * We expect the following from the capture/encode process of the device:
     *
     * To be accurately filled:
     *      framePtr->frameHeader.width;
     *                           .height;
     *                           .device;
     *                           .subformat;
     *                           .qfactor;
     *      framePtr->deviceData (not used by us)
     *      framePtr->data 
     *      framePtr->dataBuffer
     *      framePtr->sb (should be all ready to go to next object)
     *      framePtr->dataLength
     */

    if (objPtr->devicePtr->captureProc != NULL) {
	if (objPtr->devicePtr->captureProc(framePtr) != 0) {
	    ckfree((char *) framePtr);
	    return;
	}
    }

    objPtr->encodeFrame = framePtr;

    DoEncode(framePtr, asynchFlag);


    /*
     * Generate FRAME_CAPTURED event and do stats.
     */
	
    objPtr->captured++;
    objPtr->bindFramePtr = framePtr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken, "<FRAME_CAPTURED>");
    
    return;
}
    

/*
 *----------------------------------------------------------------------
 *
 * DoEncode
 *
 * Parameters:
 *	framePtr - frame to encode
 *      asynchFlag - indicates whether should be done asynchronously.
 *
 * Return:
 *	Standard Tcl return value
 *
 * Action:
 *
 *----------------------------------------------------------------------
 */
static int
DoEncode(MJPEGFrame *framePtr, int asynchFlag)
{
    MJPEGCapture *objPtr = framePtr->objPtr;

    /* 
     * If not ready or alraady in encode, do nothing.
     */
    if ((!objPtr->ready) || (objPtr->inAsynchEncode)) {
	return TCL_OK;
    }

    /*
     * Record the fact that we're decoding a frame asynchronously
     * if that's what we're doing.
     */

    if (asynchFlag) {
	objPtr->inAsynchEncode = 1;
    }

    /*
     * Generate FRAME_ENCODE_BEGIN event.
     */

    objPtr->bindFramePtr = framePtr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		     "<FRAME_ENCODE_BEGIN>");

    /*
     * Start the decoding loop
     */
    ContinueEncode(framePtr);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * ContinueEncode --
 *
 *	Continues an encode until it's finished.
 *
 * Parameters: 
 *      framePtr - pointer to frame being done.
 *      
 * Results:
 *	None
 *
 * Side effects:
 *	May schedule a frame for capture if the decoding is finished,
 *	otherwise, schedules itself to be called again when processor
 *	is idle.
 *
 *--------------------------------------------------------------
 */
static void
ContinueEncode (MJPEGFrame *framePtr)
{
    MJPEGCapture *objPtr;

    objPtr = framePtr->objPtr;

    if (!objPtr->ready) {
	return;
    }

  encode_loop:

    /*
     * Call device encode procedure.
     */
    if (objPtr->devicePtr->encodeProc != NULL) {
	objPtr->devicePtr->encodeProc(framePtr);
    }

    /*
     * If decoding is not finished, schedule another call.  Otherwise
     * we're done.
     */
    if (objPtr->devicePtr->asyncEncodeFlag) {
	if (objPtr->inAsynchEncode) {
	    Tcm_DoWhenIdle("video", (Tk_IdleProc *) ContinueEncode, framePtr);
	    return;
	} else {
	    goto encode_loop;
	}
    }

    DoneEncode(framePtr);
    return;
}


/*
 *----------------------------------------------------------------------
 *
 * DoneEncode
 *
 * Parameters:
 *	framePtr - pointer to frame just finished with decoding.
 *
 * Return:
 *	Standard Tcl return value
 *
 * Action:
 *
 *----------------------------------------------------------------------
 */
static int
DoneEncode(MJPEGFrame *framePtr)
{
    int i, new_num;
    MJPEGCapture *objPtr;
    double nextTime, currTime, speed;

    objPtr = framePtr->objPtr;

    /*
     * Gather stats.
     */
    objPtr->bytes += framePtr->dataLength;

    /*
     * Generate FRAME_ENCODE_DONE event.
     */

    objPtr->bindFramePtr = framePtr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		     "<FRAME_ENCODE_DONE>");

    if (objPtr->currFrame) {
	FreeFrame(objPtr->currFrame);
    }
    objPtr->currFrame = framePtr;
    objPtr->encodeFrame = NULL;


    /*
     * Create a scatter buffer list and pass the frame along to
     * the out command if call was done asynchronously.
     */

    if (objPtr->inAsynchEncode) {
	if (objPtr->outCmd != NULL) {
	    SB_List *sb_list;
	    char *sbStr;

	    sb_list = CreateSBListFromCurr(objPtr);

	    if (objPtr->outCCmd != NULL) {
		if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		    objPtr->outCCmd(objPtr->outCD, sb_list);
		} else {
		    objPtr->outCCmd = NULL;
		    ckfree((char *) objPtr->outCmd);
		    objPtr->outCmd = NULL;
		    objPtr->outCD = NULL;
		}
	    } else {
		sbStr = MakeSBListString(sb_list);

		Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			    sbStr, (char *) NULL);
		ckfree((char *) sbStr);
	    }
	    /* 
	     * Don't want to use regular FreeSBList since that will free 
	     * recursively and we don't want to do that.
	     */
	    ckfree((char *) sb_list->sbs);
	    ckfree((char *) sb_list);
	}
	objPtr->inAsynchEncode = 0;
    }

    /*
     * Schedule the next capture time if fps and outCmd are
     * both set.
     */ 

    speed = LTS_Speed(objPtr->lts);

    if ((objPtr->fps != 0.0) && (objPtr->outCmd != NULL) && 
	(objPtr->atToken == 0) && (speed != 0.0)) {
	double when;

	when = LTS_SystemTimeOf(objPtr->lts,
				objPtr->currFrame->logStart - objPtr->delay*speed + 
				(1.0 / objPtr->fps));
	objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					      "video", (Tk_TimerProc *) AtQueueCapture,
					      (ClientData) objPtr,
					      NULL, NULL, NULL, NULL, NULL);
    }
     
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Stats
 *
 * Parameters:
 *	objPtr - pointer to object
 *
 * Return:
 *	Standard Tcl return value
 *
 * Action:
 *
 *----------------------------------------------------------------------
 */

static int
Stats(MJPEGCapture *objPtr)
{

    /*
     * This is probably a bad way of doing this, but seems to work.
     */

    Tcl_ResetResult(objPtr->interp);
    sprintf (objPtr->interp->result, "%d %d",
 	     objPtr->captured, objPtr->bytes);
    
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FreeFrame
 *
 * Parameters:
 *	framePtr - frame pointer.
 *
 * Return:
 *	Standard Tcl return value
 *
 * Action:
 *
 *----------------------------------------------------------------------
 */

static void
FreeFrame(MJPEGFrame *framePtr)
{
    int i;

    /*
     * Call device frame freeer.
     */
    if (framePtr->objPtr->devicePtr->freeFrameProc != NULL) {
	framePtr->objPtr->devicePtr->freeFrameProc(framePtr);
    }

#if 0
    {
	char logStr[100];
	sprintf(logStr, "MJPEGFrame FreeingFrame %x %x", framePtr, framePtr->dataBuffer->data);
	Tcm_Log(logStr);
    }
#endif

    /*
     * Detach from buffer.
     */

    for (i=0; i<framePtr->sb->count; i++) {
	DetachBuffer(framePtr->objPtr->interp, framePtr->sb->buffers[i],
		     "tcmMJPEGCapture:FreeFrame");
    }

    FreeScatterBuffer(framePtr->sb);
    framePtr->sb = NULL;

    ckfree((char *) framePtr);

    return;
}

/***********************************************************************
 *
 * GetLogStart
 * GettLogEnd
 * GetCumulLenEncoded
 * GetDataLen
 * GetWidth
 * GetHeight
 * GetQFactor
 *
 * These functions are registered with the CM Bind mechanism to
 * access different values of the current frame's data structure.
 * for bind script % substitution.
 * 
 *********************************************************************/


static char *
GetLogStart(Tcl_Interp *interp, ClientData clientdata)
{
    MJPEGCapture *objPtr;
    char *resStr;

    objPtr = (MJPEGCapture *) clientdata;

    resStr = (char *) ckalloc(50);

    sprintf(resStr, "%f", objPtr->bindFramePtr->logStart);

    return resStr;
}

static char *
GetLogEnd(Tcl_Interp *interp, ClientData clientdata)
{
    MJPEGCapture *objPtr;
    char *resStr;

    objPtr = (MJPEGCapture *) clientdata;

    resStr = (char *) ckalloc(50);

    sprintf(resStr, "%f", objPtr->bindFramePtr->logEnd);

    return resStr;
}

static char *
GetCumulLenEncoded(Tcl_Interp *interp, ClientData clientdata)
{
    MJPEGCapture *objPtr;
    char *resStr;

    objPtr = (MJPEGCapture *) clientdata;

    resStr = (char *) ckalloc(50);
    sprintf(resStr, "%d", objPtr->bytes);
    objPtr->bytes = 0;
    return resStr;
}

static char *
GetDataLen(Tcl_Interp *interp, ClientData clientdata)
{
    MJPEGCapture *objPtr;
    char *resStr;

    objPtr = (MJPEGCapture *) clientdata;

    resStr = (char *) ckalloc(50);

    sprintf(resStr, "%d", objPtr->bindFramePtr->dataLength);

    return resStr;
}

static char *
GetWidth(Tcl_Interp *interp, ClientData clientdata)
{
    MJPEGCapture *objPtr;
    char *resStr;

    objPtr = (MJPEGCapture *) clientdata;

    resStr = (char *) ckalloc(50);

    sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.width);

    return resStr;
}

static char *
GetHeight(Tcl_Interp *interp, ClientData clientdata)
{
    MJPEGCapture *objPtr;
    char *resStr;

    objPtr = (MJPEGCapture *) clientdata;

    resStr = (char *) ckalloc(50);

    sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.height);

    return resStr;
}

static char *
GetQFactor(Tcl_Interp *interp, ClientData clientdata)
{
    MJPEGCapture *objPtr;
    char *resStr;

    objPtr = (MJPEGCapture *) clientdata;

    resStr = (char *) ckalloc(50);

    sprintf(resStr, "%d", objPtr->qfactor);

    return resStr;
}

