/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include "cmCookies.h"
#include "cmBufferlist.h"
#include "cmChunk.h"
#include "cmLTS.h"
#include "cmAtom.h"

#ifndef O_BINARY
#    define O_BINARY 0
#endif

#define ONE_BILLION 1000000000.0
#define CF_READ_BUFFER_SIZE 150000


/* 
 * Structure kept for each segment in segment list.
 */

typedef struct Segment {
    double clipStart;
    double clipEnd;
    double logicalStart;
    double logicalEnd;
    struct Segment *next;
    struct Segment *prev;
    int *FOT;
    double *FTT;
    int numFrames;
    int maxSize;
    int segId;
} Segment;

/*
 * Structure used when picking next set of frames to be sent.
 */

typedef struct {
    Buffer *bufPtr;
    Segment *segPtr;
    int index;
    int size;
    double sTime;
    double eTime;
    int sendFlag;
    int offset;
} FramePick;

/*
 * Structure kept for each clipfile object. 
 */

typedef struct MJPEG_CF {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    ClientData lts;		/* LTS for object to use. */
    char *ltsName;
    int ltsTriggerNum;             /* LTS trigger for changed lts. */
    double sendAhead;		/* Send ahead time. */
    int bandwidth;		/* Bandwidth available */
    double cycleTime;		/* Cycle period. */
    char *filename;		/* Filename of clipfile. */
    tChunkFileState *cfState;   /* Pointer to chunkfile structure. */
    int *rawFOT;                /* Pointer to FOT if file is raw data. */
    double *rawFTT;             /* Pointer to FTT if file is raw data. */
    int rawfd;                  /* File descriptor for raw file. -1 if not open. */
    char *outCmd;		/* Command to send buffers to. */
    double maxFrameRate;	/* Maximum frame rate. */
    double extraFrameCounter;   /* Used to adjust number of frames sent per cycleTime
				   to match with maxFrameRate */
    int extraBWCounter;         /* Used to adjust number of bytes send per cycleTime
				   to mactch with bandwidth restrictions. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;           /* ClientData for outCCmd. */
    int nextSegId;		/* Next segment id. */
    int segmentListChangeFlag;  /* Indicates that segment list has changed. */
    Segment *segments;		/* List of segment structures. */
    Segment *lastSeg;           /* Segment from which data was sent last. */
    int lastFrameNum;           /* Frame number of last frame sent. */
    MJPEGHeader *cfInfo;	/* MJPEG Header from clipfile. */
    int numFrames;		/* Number of frames in clipfile. */
    double clipLength;		/* Length of clipfile. */
    int dataStart;              /* Offset to start of data. */
    int ready;                  /* Ready flag. */
    int at_token;               /* Token for at event. */
} MJPEG_CF;

/* Globals used by this object module. */

static int nextMJPEG_CFNum = 0;	/* Variable for next unique object name. */

/* Forward declarations. */

int MJPEG_CFObjCmd(ClientData clientData, Tcl_Interp * interp, int argc, char **argv);
static int MJPEG_CF_AddSegment(MJPEG_CF * objPtr, double clipStart, double clipEnd, 
				  double logStart, double logEnd, 
				  int logStartFlag, int logEndFlag);
static int MJPEG_CF_RemoveSegment(MJPEG_CF *objPtr, int segId);
static int MJPEG_CF_ClearSegments(MJPEG_CF *objPtr);
static int MJPEG_CF_SegmentInfo(MJPEG_CF *objPtr, int segId);
static int MJPEG_CF_Info(MJPEG_CF *objPtr);
static int MJPEG_CF_Ready(MJPEG_CF *objPtr);
static int MJPEG_CF_Unready(MJPEG_CF *objPtr);
static int MJPEG_CF_ChangeOutCmd(MJPEG_CF *objPtr, char *newOut);
static int MJPEG_CF_OpenClipfile(MJPEG_CF *objPtr, char *filename);
static int MJPEG_CF_SegmentList(MJPEG_CF *objPtr); 
static int MJPEG_CF_ConfigureSegment(MJPEG_CF *objPtr, int segmentId,
				     double clipStart, double clipEnd, 
				     double logStart, double logEnd, 
				     int logStartFlag, int logEndFlag);

void MJPEG_CF_LTSChanged(ClientData clientdata);
void MJPEG_CF_SendData(ClientData clientdata);

/************************************************************
 *
 * CreateNewMJPEG_CF
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

MJPEG_CF *
CreateNewMJPEG_CF(Tcl_Interp *interp,  /* Pointer to interpreter. */
		  char *name           /* Name of new object. */)
{

    MJPEG_CF *new;

    new = (MJPEG_CF *) ckalloc(sizeof(MJPEG_CF));

    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);
    new->lts = NULL;
    new->ltsName = NULL;
    new->ltsTriggerNum = -1;
    new->sendAhead = 0.2;
    new->bandwidth = 0;
    new->cycleTime = 0.0;
    new->filename = NULL;
    new->cfState = NULL;
    new->rawFOT = NULL;
    new->rawFTT = NULL;
    new->rawfd = -1;
    new->outCmd = NULL;
    new->maxFrameRate = 0.0;
    new->extraFrameCounter = 0.0;
    new->extraBWCounter = 0;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->nextSegId = 0;
    new->segmentListChangeFlag = 0;
    new->segments = NULL;
    new->lastSeg = NULL;
    new->lastFrameNum = -1;
    new->cfInfo = NULL;
    new->numFrames = 0;
    new->clipLength = 0.0;
    new->dataStart = 0;
    new->ready = 0;
    new->at_token = 0;
    return new;
}

/************************************************************
 *
 * DestroyMJPEG_CF
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

int
DestroyMJPEG_CF(MJPEG_CF *objPtr       /* Pointer to clipfile object. */)
{

    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
	objPtr->at_token = 0;
    }

    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    if (objPtr->filename != NULL) {
	ckfree(objPtr->filename);
	objPtr->filename = NULL;
    }

    if (objPtr->cfState != NULL) {
	Chunk_CloseFile(objPtr->cfState);
	objPtr->cfState = NULL;
    }

    if (objPtr->rawfd > -1) {
	close(objPtr->rawfd);
	objPtr->rawfd = -1;
    }

    if (objPtr->rawFTT != NULL) {
	ckfree((char *) objPtr->rawFTT);
	objPtr->rawFTT = NULL;
    }

    if (objPtr->rawFOT != NULL) {
	ckfree((char *) objPtr->rawFOT);
	objPtr->rawFOT = NULL;
    }

    if (objPtr->outCmd != NULL)
	ckfree((char *) objPtr->outCmd);

    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);
    }

    if (objPtr->ltsName != NULL) {
	ckfree((char *) objPtr->ltsName);
    }

    /*
     * Free all segments.
     */
    MJPEG_CF_ClearSegments(objPtr);

    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureMJPEG_CF
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

int
ConfigureMJPEG_CF(MJPEG_CF * objPtr, int argc, char **argv)
{
  
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
  
    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {} {}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName, "}} ",
		    (char *) NULL);
	}
	Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	Tcl_AppendResult(interp, "{-sendAhead {0.2} ", str, "} ", (char *)NULL);

	sprintf(str, "{%d}", objPtr->bandwidth);
	Tcl_AppendResult(interp, "{-bandwidth {0} ", str, "} ", (char *)NULL);

	Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	Tcl_AppendResult(interp, "{-cycleTime {0.0} ", str, "} ", (char *)NULL);

	if (objPtr->filename == NULL) {
	    Tcl_AppendResult(interp, "{-filename {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-filename {} {", objPtr->filename,
			    "}} ",(char *)NULL);
	}
	
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}

	Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	Tcl_AppendResult(interp, "{-maxFrameRate {0.0} ", str,
			  "} ", (char *)NULL);
	return TCL_OK;
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->lts == NULL) {
		Tcl_AppendResult(interp, "-lts {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName, "} ",
				  (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	    Tcl_AppendResult(interp, "-sendAhead {0.2} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'b') && (strncmp("-bandwidth", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->bandwidth);
	    Tcl_AppendResult(interp, "-bandwidth {0} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	    Tcl_AppendResult(interp, "-cycleTime {0.0} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'f') && (strncmp("-filename", argv[0], len) == 0)) {
	    if (objPtr->filename == NULL) {
		Tcl_AppendResult(interp, "-filename {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-filename {} {",
			  objPtr->filename, "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	    Tcl_AppendResult(interp, "-maxFrameRate {0.0} ", str, (char *)NULL);
	    return TCL_OK;
	}
	Tcl_AppendResult(interp, "bad configuration option: ",
			  argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	/* ------------------------- LTS ------------------------- */
	if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newlts;
	    
	    /*
	     * Check args
	     */
	    cntr++;
	    if (cntr == argc) {
argcError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }

	    /*
	     * If resetting to null value, just free up old lts.
	     */
	    if (strlen(argv[cntr]) == 0) {
		if (objPtr->lts != NULL) {
		    if (objPtr->ltsTriggerNum != -1) {
			LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
			objPtr->ltsTriggerNum = -1;
		    }
		    Tk_Release(objPtr->lts);
		    objPtr->lts = NULL;
		}
		if (objPtr->ltsName != NULL) {
		    ckfree((char *) objPtr->ltsName);
		    objPtr->ltsName = NULL;
		}
	    } else {
		newlts = (ClientData)Tcm_FindAtom(interp, argv[cntr]);
		if (newlts == NULL) {
valueError:	    
		    Tcl_AppendResult(interp, "bad value \"", argv[cntr],
				  "\" given for configuration option ",
				  argv[cntr-1], (char *)NULL);
		    return TCL_ERROR;
		}
		
		/*
		 * Blow away old callback.
		 */
		if (objPtr->lts != NULL) {
		    if (objPtr->ltsTriggerNum != -1) {
			LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
			objPtr->ltsTriggerNum = -1;
		    }
		    Tk_Release(objPtr->lts);
		    objPtr->lts = NULL;
		}
		objPtr->lts = newlts;
		Tk_Preserve (objPtr->lts);

		/*
		 * Copy in new lts name
		 */
		if (objPtr->ltsName != NULL) {
		    ckfree((char *) objPtr->ltsName);
		}
		objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
		strcpy(objPtr->ltsName, argv[cntr]);
		
		/*
		 * Set up new lts trigger
		 */
		objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts,
				LTS_SPEED_FLAG | LTS_OFFSET_FLAG | LTS_VALUE_FLAG,
			        MJPEG_CF_LTSChanged, (ClientData) objPtr);
	    }
	    MJPEG_CF_Unready(objPtr);

	/* ------------------------- SEND_AHEAD ------------------------- */
	} else if ((c=='s') && (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	    double newSendAhead;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->sendAhead = newSendAhead;

	/* ------------------------- BANDWIDTH -------------------------- */
	} else if ((c=='b') && (strncmp("-bandwidth", argv[cntr], len) == 0)) {
	    int newBandwidth;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newBandwidth) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->bandwidth = newBandwidth;
	    objPtr->extraBWCounter = 0;

	/* ------------------------- CYCLE_TIME ------------------------- */
	} else if ((c=='c') && (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	    double newCycleTime;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->cycleTime = newCycleTime;

	/* ------------------------- FILE_NAME -------------------------- */
	} else if ((c=='f') && (strncmp("-filename", argv[cntr], len) == 0)) {
	    Tcl_DString fname_dstring;
	    char *fname;

	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (objPtr->filename != NULL) {
		Tcl_AppendResult(interp,
				"Can not set clipfile name more than once.",
				(char *)NULL);
		return TCL_ERROR;
	    }

	    fname = Tcl_TildeSubst(objPtr->interp, argv[cntr], &fname_dstring);

	    if (MJPEG_CF_OpenClipfile(objPtr, fname) != TCL_OK) {
		Tcl_DStringFree(&fname_dstring);
		return TCL_ERROR;
	    }
	    
	    objPtr->filename = (char *) ckalloc(strlen(fname) + 1);
	    strcpy(objPtr->filename, fname);
	    Tcl_DStringFree(&fname_dstring);

	/* ------------------------- OUT_CMD ---------------------------- */
	} else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    
	    if (MJPEG_CF_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	    MJPEG_CF_Unready(objPtr);
	    
	/* ------------------------- MAX_FRAME_RATE --------------------- */
	} else if ((c=='m') &&
		   (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
	    double newMaxFrameRate;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->maxFrameRate = newMaxFrameRate;
	    objPtr->extraFrameCounter = 0.0;

	/* ------------------------- UNKNOWN OPTION ----------------------*/
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			  argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CFCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the mjpegCF command.
 *     A new object of type MJPEG_CF is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
MJPEG_CFCreateCmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char **argv)
{
    char new_name[256];
    MJPEG_CF *new;
    
    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextMJPEG_CFNum++;
    sprintf(new_name, "%smjpegFile%d", argv[1], nextMJPEG_CFNum);
    
    new = CreateNewMJPEG_CF(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MJPEG_CFObjCmd,
		      (ClientData) new, (void (*)()) NULL);
    
    if (argc > 2) {
	ConfigureMJPEG_CF(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CFObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     MJPEG_CF object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
MJPEG_CFObjCmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char **argv)
{
    char c;
    int len, i;
    MJPEG_CF *objPtr = (MJPEG_CF *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    /*------------------------- CONFIGURE ------------------------------*/
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureMJPEG_CF(objPtr, argc - 2, argv + 2);

    /*------------------------- DESTROY --------------------------------*/
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyMJPEG_CF(objPtr);

    /*------------------------- ADD_SEGMENT -------------------------*/
    } else if ((c == 'a') && (strncmp("addSegment", argv[1], len) == 0)) {
	double clipStart;
	double clipEnd;
	double logStart = 0.0;
	double logEnd = 0.0;
	int logStartFlag = 0;
	int logEndFlag = 0;

	if (argc < 4) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			" addSegment <clipStart> <clipEnd> ",
			"?-start <logicalStart>? ?-end <logicalEnd>? ",
			(char *) NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetDouble(interp, argv[2], &clipStart) != TCL_OK) {
	    Tcl_AppendResult(interp,
		"error found while parsing clip start time ", (char *)NULL);
	    return TCL_ERROR;
	}

	if (strcmp(argv[3], "end") == 0) {
	    clipEnd = HUGE_VAL;
	} else if (Tcl_GetDouble(interp, argv[3], &clipEnd) != TCL_OK) {
	    Tcl_AppendResult(interp,
		  "error found while parsing clip end time ", (char *)NULL);
	    return TCL_ERROR;
	}

	for (i = 4; i < argc; i++) {
	    char ac;
	    int alen;

	    ac = argv[i][1];
	    alen = strlen(argv[i]);

	    if ((ac == 's') && (strncmp("-start", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Must specify parameter value ",
			      "after optional argument flag ", argv[i],
			      " passed to ", argv[0], " ", argv[1],
			      (char *) NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logStart) != TCL_OK) {
		    Tcl_AppendResult(interp,
				"error found while logical start end time ",
				(char *)NULL);
		    return TCL_ERROR;
		}
		logStartFlag = 1;
	    } else if ((ac == 'e') && (strncmp("-end", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Must specify parameter value ",
			      "after optional argument flag ", argv[i],
			      " passed to ", argv[0], " ", argv[1],
			      (char *) NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logEnd) != TCL_OK) {
		    Tcl_AppendResult(interp,
				"error found while logical end end time ",
				(char *)NULL);
		    return TCL_ERROR;
		}
		logEndFlag = 1;
	    }
	  }
	  return MJPEG_CF_AddSegment(objPtr, clipStart, clipEnd,
			logStart, logEnd, logStartFlag, logEndFlag);

    /*------------------------- CONFIGURE_SEGMENT -------------------------*/
    } else if ((c == 'c') && (strncmp("configureSegment", argv[1], len) == 0)) {
	double clipStart;
	double clipEnd;
	double logStart = 0.0;
	double logEnd = 0.0;
	int logStartFlag = 0;
	int logEndFlag = 0;
	int segmentId;

	if (argc < 5) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			" configureSegment <segmentId> <clipStart> <clipEnd> ",
			"?-start <logicalStart>? ?-end <logicalEnd>? ",
			(char *) NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[2], &segmentId) != TCL_OK) {
	  Tcl_AppendResult(interp, "error found while parsing segment id ",
			   (char *) NULL);
	  return TCL_ERROR;
	}

	if (Tcl_GetDouble(interp, argv[3], &clipStart) != TCL_OK) {
	    Tcl_AppendResult(interp,
		"error found while parsing clip start time ", (char *)NULL);
	    return TCL_ERROR;
	}

	if (strcmp(argv[4], "end") == 0) {
	    clipEnd = HUGE_VAL;
	} else if (Tcl_GetDouble(interp, argv[4], &clipEnd) != TCL_OK) {
	    Tcl_AppendResult(interp,
		  "error found while parsing clip end time ", (char *)NULL);
	    return TCL_ERROR;
	}

	for (i = 5; i < argc; i++) {
	    char ac;
	    int alen;

	    ac = argv[i][1];
	    alen = strlen(argv[i]);

	    if ((ac == 's') && (strncmp("-start", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Must specify parameter value ",
			      "after optional argument flag ", argv[i],
			      " passed to ", argv[0], " ", argv[1],
			      (char *) NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logStart) != TCL_OK) {
		    Tcl_AppendResult(interp,
				"error found while logical start end time ",
				(char *)NULL);
		    return TCL_ERROR;
		}
		logStartFlag = 1;
	    } else if ((ac == 'e') && (strncmp("-end", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Must specify parameter value ",
			      "after optional argument flag ", argv[i],
			      " passed to ", argv[0], " ", argv[1],
			      (char *) NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logEnd) != TCL_OK) {
		    Tcl_AppendResult(interp,
				"error found while logical end end time ",
				(char *)NULL);
		    return TCL_ERROR;
		}
		logEndFlag = 1;
	    }
	  }

	return MJPEG_CF_ConfigureSegment(objPtr, segmentId, clipStart, clipEnd,
				  logStart, logEnd, logStartFlag, logEndFlag);

    /*------------------------- REMOVE_SEGMENT -------------------------*/
    } else if ((c == 'r') && (strncmp("removeSegment", argv[1], len) == 0)) {
	int segId;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			argv[0], " removeSegment <segmentId> ", (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &segId) != TCL_OK) {
	    Tcl_AppendResult(interp, "Bad integer parameter value ",
			argv[2], " given to: ", argv[0], " ", argv[1],
			(char *) NULL);
	    return TCL_ERROR;
	}
	return MJPEG_CF_RemoveSegment(objPtr, segId);

    /*------------------------ SEGMENT LIST ----------------------------*/
    } else if ((c == 's') && (strncmp("segmentList", argv[1], len) == 0)) {
	return MJPEG_CF_SegmentList(objPtr);

    /*------------------------- CLEAR_SEGMENTS -------------------------*/
    } else if ((c == 'c') && (strncmp("clearSegments", argv[1], len) == 0)) {
	return MJPEG_CF_ClearSegments(objPtr);

    /*------------------------- SEGMENT_INFO ---------------------------*/
    } else if ((c == 's') && (strncmp("segmentInfo", argv[1], len) == 0)) {
	int segId;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			argv[0], " segmentInfo <segmentId> ", (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &segId) != TCL_OK) {
	    Tcl_AppendResult(interp, "Bad integer parameter value ", argv[2],
			" given to: ", argv[0], " ", argv[1], (char *) NULL);
	    return TCL_ERROR;
	}
	return MJPEG_CF_SegmentInfo(objPtr, segId);


    /*------------------------- INFO -----------------------------------*/
    } else if ((c == 'i') && (strncmp("info", argv[1], len) == 0)) {
	return MJPEG_CF_Info(objPtr);

    /*------------------------- READY ----------------------------------*/
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return MJPEG_CF_Ready(objPtr);

    /*------------------------- UNREADY --------------------------------*/
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return MJPEG_CF_Unready(objPtr);
    }

    Tcl_AppendResult(interp, "Bad command option: ",
		argv[0], argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  addSegment <clipStart> <clipEnd>",
			" ?-start <logicalStart>? ?-end <logicalEnd>?\n",
                "  removeSegment <segmentId>\n",
                "  clearSegments\n",
		"  configureSegment <segmentId> <clipStart> <clipEnd>",
			" ?-start <logicalStart>? ?-end <logicalEnd>?\n",
		"  segmentList\n",
                "  segmentInfo <segmentId>\n",
                "  ready\n",
                "  unready\n",
                "  info\n",
                (char *) NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * MJPEG_CFModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
MJPEG_CFModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mjpegFile", MJPEG_CFCreateCmd, NULL,
		(void (*)()) NULL);
    Tcl_GlobalEval(interp, "cminfo add object_info mjpegFile {motion jpeg file}");
    return TCL_OK;
}

static double
GetTimeFromFile(int fd)
{
    int sec, nsec;
    double res;

    read(fd, &sec, 4);
    read(fd, &nsec, 4);
    
    sec = ntohl(sec);
    nsec = ntohl(nsec);

    res = ((double) sec) + (((double) nsec) / ONE_BILLION);
    
    return res;
}

static void
GetOffsetFromFile(int fd, int *start, int *length)
{
    int tstart, tlength;

    read(fd, &tstart, 4);
    read(fd, &tlength, 4);
    
    (*start) = ntohl(tstart);
    (*length) = ntohl(tlength);

    return;
}

/************************************************************
 *
 * MJPEG_CF_AddSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     clipStart -
 *     clipEnd -
 *     logStart -
 *     logEnd -
 *     logStartFlag - indicates whetehr logStart value set or not.
 *     logEndFlag - indicates whether logEnd value set or not.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MJPEG_CF_AddSegment(MJPEG_CF *objPtr, double clipStart, double clipEnd, 
		       double logStart, double logEnd, 
		       int logStartFlag, int logEndFlag)
{
  Segment *newSeg;
  char retStr[32];
  double clip_dur;
  Segment *nextSeg, *prevSeg;
  tChunkHeader chunk_hdr;
  int beg_frame, end_frame;
  double this_time, next_time;
  int count, i, maxSize, j;
  double scale;

  /*
   * Unready object saving ready state so that it can be
   * restored later. 
   */

  MJPEG_CF_Unready(objPtr);

  if (clipStart > clipEnd) {
    Tcl_AppendResult(objPtr->interp, "Clip start can not be after clip end.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  if (objPtr->filename == NULL) {
    Tcl_AppendResult(objPtr->interp, "Must specify clipfile first.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  if (clipStart < 0.0) clipStart = 0.0;
  if (clipStart > objPtr->clipLength) {
    Tcl_AppendResult(objPtr->interp, "Clip start too large.", (char *) NULL);
    return TCL_ERROR;
  }
  if (clipEnd > objPtr->clipLength) clipEnd = objPtr->clipLength;
  if (clipEnd < 0.0) {
    Tcl_AppendResult(objPtr->interp, "Clip end too small.", (char *) NULL);
    return TCL_ERROR;
  }
  
  newSeg = (Segment *) ckalloc(sizeof(Segment));
  
  newSeg->clipStart = clipStart;
  newSeg->clipEnd = clipEnd;
  
  clip_dur = clipEnd - clipStart;
  
  if (clip_dur <= 0.0) {
    ckfree((char *) newSeg);
    Tcl_AppendResult(objPtr->interp, "Clip segment duration too small.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  
  /*
   * Do both start and end specified.
   */
  if (logStartFlag && logEndFlag) {
    /*
     * Set logStart and logEnd.
     */
    if (logStart > logEnd) {
      newSeg->logicalStart = logEnd;
      newSeg->logicalEnd = logStart;
    } else {
      newSeg->logicalStart = logStart;
      newSeg->logicalEnd = logEnd;
    }
  } else if (logStartFlag) {
    newSeg->logicalStart = logStart;
    logEnd = newSeg->logicalEnd = logStart+clip_dur;
  } else if (logEndFlag) {
    newSeg->logicalEnd = logEnd;
    logStart = newSeg->logicalStart = logEnd - clip_dur;
  } else {
    logStart = newSeg->logicalStart = 0.0;
    logEnd = newSeg->logicalEnd = clip_dur;
  }
  
  
  /*
   * Check for zero durations.
   */
  if (newSeg->logicalStart - newSeg->logicalEnd == 0.0) {
    Tcl_AppendResult(objPtr->interp, "Zero length logical time.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  if (newSeg->clipStart - newSeg->clipEnd == 0.0) {
    Tcl_AppendResult(objPtr->interp, "Zero length clip time.",
		     (char *) NULL);
    return TCL_ERROR;
  }

  
  /*
   * Find place in list. Sort by start times.
   */
  nextSeg = objPtr->segments;
  prevSeg = NULL;
  
  while (nextSeg != NULL) {
    if (nextSeg->logicalStart < newSeg->logicalStart) {
      prevSeg = nextSeg;
      nextSeg = nextSeg->next;
    } else {
      break;
    }
  }
  
  /* 
   * Create frame offset and time tables for this segment.
   * This means skipping frames until the correct beginning and
   * then finding correct end. Then generating segment FOT and FTT
   * for these frames with time scaled properly. 
   * 
   * If rawFOT is not NULL, use rawFOT and rawFTT instead of
   * opening chunk files (i.e. data is in raw format and not clipfile format).
   */
  
  if (objPtr->rawFOT == NULL) {
    if (objPtr->cfState != NULL) {
      Chunk_ReadRewind(objPtr->cfState);
    } else {
      objPtr->cfState = Chunk_OpenFile(objPtr->filename, "r");
    }
    
    if (Chunk_FindType(objPtr->cfState, CM_FTT, &chunk_hdr)) {
      ckfree((char *) newSeg);
      Tcl_AppendResult(objPtr->interp,
		       "Problem reading clipfile frame time table.", (char *) NULL);
      return TCL_ERROR;
    }
    
    beg_frame = 0;
    this_time = GetTimeFromFile(objPtr->cfState->fd);
    next_time = GetTimeFromFile(objPtr->cfState->fd);
    count = 0;
    while ((this_time < newSeg->clipStart) &&
	   (next_time <= newSeg->clipStart)) {
      this_time = next_time;
      next_time = GetTimeFromFile(objPtr->cfState->fd);
      count++;
      if (count == objPtr->numFrames) {
	Tcl_AppendResult(objPtr->interp,
			 "Error constructing segment frame time table.", (char *) NULL);
	return TCL_ERROR;
      }
    }
    
    beg_frame = count;
    
    while ((this_time < newSeg->clipEnd) && (next_time < newSeg->clipEnd)) {
      this_time = next_time;
      next_time = GetTimeFromFile(objPtr->cfState->fd);
      count++;
      if (count == objPtr->numFrames) {
	Tcl_AppendResult(objPtr->interp,
			 "Error constructing segment frame time table.", (char *) NULL);
	return TCL_ERROR;
      }
    }
    
    end_frame = count;
    newSeg->numFrames = end_frame - beg_frame + 1;
    
    /*
     * Read current FTT values.
     */
    newSeg->FTT = (double *) ckalloc((newSeg->numFrames+1)*sizeof(double));
    
    Chunk_ReadRewind(objPtr->cfState);
    if (Chunk_FindType(objPtr->cfState, CM_FTT, &chunk_hdr)) {
      if (newSeg->FTT != NULL) {
	ckfree((char *) newSeg->FTT);
      }
      ckfree((char *) newSeg);
      Tcl_AppendResult(objPtr->interp,
		       "Error constructing segment frame time table.", (char *) NULL);
      return TCL_ERROR;
    }
    
    if (beg_frame != 0) {
      lseek(objPtr->cfState->fd, beg_frame * 8, SEEK_CUR);
    }
    
    for (i=0; i<newSeg->numFrames; i++) {
      newSeg->FTT[i] = GetTimeFromFile(objPtr->cfState->fd);
    }
    
    scale = (newSeg->logicalEnd - newSeg->logicalStart) / 
      (newSeg->clipEnd - newSeg->clipStart);
    
    newSeg->FTT[0] = newSeg->logicalStart;
    
    for (i=1; i<newSeg->numFrames; i++) {
      newSeg->FTT[i] = scale * (newSeg->FTT[i]-newSeg->clipStart) +
	newSeg->logicalStart;
    }
    newSeg->FTT[i] = newSeg->logicalEnd;
    
    /*
     * Construct segment frame offset table.
     */
    Chunk_ReadRewind(objPtr->cfState);
    if (Chunk_FindType(objPtr->cfState, CM_FOT, &chunk_hdr)) {
      if (newSeg->FTT != NULL) {
	ckfree((char *) newSeg->FTT);
      }
      ckfree((char *) newSeg);
      Tcl_AppendResult(objPtr->interp,
		       "Error constructing segment frame offset table.", (char *)NULL);
      return TCL_ERROR;
    }
    
    newSeg->FOT = (int *) ckalloc(sizeof(int)*2*newSeg->numFrames);
    
    if (beg_frame != 0) {
      lseek(objPtr->cfState->fd, beg_frame * 8, SEEK_CUR);
    }
    
    if (logStart < logEnd) {
      for (i=0; i<newSeg->numFrames; i++) {
	GetOffsetFromFile (objPtr->cfState->fd, &(newSeg->FOT[i*2]),
			   &(newSeg->FOT[(i*2)+1]));
      }
    } else {
      for (i=newSeg->numFrames-1; i>=0; i--) {
	GetOffsetFromFile( objPtr->cfState->fd, &(newSeg->FOT[i*2]),
			  &(newSeg->FOT[(i*2)+1]));
      }
    }
  } else {
    int ftt_index;
    
    /* 
     * Construct segment FOT and FTT from rawFOT and rawFTT. 
     * First find out which frames are needed.
     */
    
    beg_frame = 0;
    ftt_index = 0;
    
    this_time = objPtr->rawFTT[ftt_index++];
    next_time = objPtr->rawFTT[ftt_index++];
    
    count = 0;
    
    while ((this_time < newSeg->clipStart) &&
	   (next_time <= newSeg->clipStart)) {
      this_time = next_time;
      next_time = objPtr->rawFTT[ftt_index++];
      count++;
      if (count == objPtr->numFrames) {
	Tcl_AppendResult(objPtr->interp,
			 "Error constructing segment frame time table.", 
			 (char *) NULL);
	return TCL_ERROR;
      }
    }
    
    beg_frame = count;
    
    while ((this_time < newSeg->clipEnd) && (next_time < newSeg->clipEnd)) {
      this_time = next_time;
      next_time = objPtr->rawFTT[ftt_index++];
      count++;
      if (count == objPtr->numFrames) {
	Tcl_AppendResult(objPtr->interp,
			 "Error constructing segment frame time table.", 
			 (char *) NULL);
	return TCL_ERROR;
      }
    }
    
    end_frame = count;
    newSeg->numFrames = end_frame - beg_frame + 1;
    
    /* 
     * Construct FTT for new segment from rawFTT. 
     */
    
    newSeg->FTT = (double *) ckalloc((newSeg->numFrames+1)*sizeof(double));
    
    for (i=0; i<newSeg->numFrames; i++) {
      newSeg->FTT[i] = objPtr->rawFTT[beg_frame+i];
    }
    
    scale = (newSeg->logicalEnd - newSeg->logicalStart) /
      (newSeg->clipEnd - newSeg->clipStart);
    
    newSeg->FTT[0] = newSeg->logicalStart;
    
    for (i=1; i<newSeg->numFrames; i++) {
      newSeg->FTT[i] = scale * (newSeg->FTT[i]-newSeg->clipStart) +
	newSeg->logicalStart;
    }
    newSeg->FTT[i] = newSeg->logicalEnd;
    
    /* 
     * Construct seg. FOT.
     */
    
    newSeg->FOT = (int *) ckalloc(sizeof(int)*2*newSeg->numFrames);
    
    if (logStart < logEnd) {
      for (i=0; i<newSeg->numFrames; i++) {
	newSeg->FOT[i*2] = objPtr->rawFOT[(beg_frame+i)*2];
	newSeg->FOT[(i*2)+1] = objPtr->rawFOT[((beg_frame+i)*2)+1];
      }
    } else {
      for (i=newSeg->numFrames-1,j=0; i>=0; i--,j++) {
	newSeg->FOT[i*2] = objPtr->rawFOT[(beg_frame+j)*2];
	newSeg->FOT[(i*2)+1] = objPtr->rawFOT[((beg_frame+j)*2)+1];
      }
    }
  }
  
  /*
   * Set segment id.
   */
  newSeg->segId = objPtr->nextSegId++;
  
  /*
   * Find max frame size.
   */
  maxSize = -1;
  for (i=0; i<newSeg->numFrames; i++) {
    if (newSeg->FOT[(i*2)+1] > maxSize) {
      maxSize = newSeg->FOT[(i*2)+1];
    }
  }
  newSeg->maxSize = maxSize;
  
  /*
   * Set up next and prev pointers and return.
   */
  newSeg->next = nextSeg;
  newSeg->prev = prevSeg;
  if (prevSeg == NULL) {
    objPtr->segments = newSeg;
  } else {
    prevSeg->next = newSeg;
    }
  if (nextSeg != NULL) {
    nextSeg->prev = newSeg;
  }
  
  sprintf(retStr, "%d", newSeg->segId);
  Tcl_SetResult(objPtr->interp, retStr, TCL_VOLATILE);

  objPtr->segmentListChangeFlag = 1;

  return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_RemoveSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segId -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MJPEG_CF_RemoveSegment(MJPEG_CF * objPtr, int segId)
{
  Segment *segPtr;
  
  objPtr->lastSeg = NULL;
  objPtr->lastFrameNum = -1;

  segPtr = objPtr->segments;
  
  while (segPtr != NULL) {
    if (segPtr->segId == segId) break;
    segPtr = segPtr->next;
  }
  
  if (segPtr == NULL) return TCL_OK;
  
  if (segPtr->prev == NULL) {
    objPtr->segments = segPtr->next;
  } else {
    segPtr->prev->next = segPtr->next;
  }

  if (segPtr->next != NULL) {
    segPtr->next->prev = segPtr->prev;
  }
  
  if (segPtr->FOT != NULL)  {
    ckfree((char *) segPtr->FOT);
  }
  if (segPtr->FTT != NULL) {
    ckfree((char *) segPtr->FTT);
  }
  ckfree((char *) segPtr);
  
  return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_ClearSegments
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MJPEG_CF_ClearSegments(MJPEG_CF * objPtr)
{
    while (objPtr->segments != NULL) {
	MJPEG_CF_RemoveSegment(objPtr, objPtr->segments->segId);
    }
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_SegmentInfo
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segId -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sets the Tcl result string to {clipStart clipEnd logStart logEnd 
 *                                    numFrames maxSize} 
 *
 ************************************************************/

static int
MJPEG_CF_SegmentInfo(MJPEG_CF * objPtr, int segId)
{
    Segment *segPtr;
    char retStr[256];

    segPtr = objPtr->segments;

    while (segPtr != NULL) {
	if (segPtr->segId == segId) {
	    break;
	}
	segPtr = segPtr->next;
    }

    if (segPtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "Segment doesn't exist.",
		(char *) NULL);
	return TCL_ERROR;
    }

    sprintf(retStr, "%g %g %g %g %d %d", segPtr->clipStart, segPtr->clipEnd,
	    segPtr->logicalStart, segPtr->logicalEnd,
	    segPtr->numFrames, segPtr->maxSize);

    Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_Info
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sets return string to {numFrames clipLength maxSize width height}
 *     where maxSize is maximum size of segments so far.
 *
 ************************************************************/
static int
MJPEG_CF_Info(MJPEG_CF * objPtr)
{
    Segment *segPtr;
    int maxSize;
    char retStr[256];

    segPtr = objPtr->segments;
    maxSize = 0;

    while(segPtr != NULL) {
	if (segPtr->maxSize > maxSize)
	    maxSize = segPtr->maxSize;
	segPtr = segPtr->next;
    }

    if (objPtr->cfInfo != NULL) {
	sprintf(retStr, "%d %g %d %d %d", objPtr->numFrames,
		objPtr->clipLength, maxSize,
		objPtr->cfInfo->width, objPtr->cfInfo->height);
    } else {
	sprintf(retStr, "%d %g %d 0 0", objPtr->numFrames,
		  objPtr->clipLength, maxSize);
    }

    Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_Ready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Gets the object ready and starts sending data.
 ************************************************************/

static int
MJPEG_CF_Ready(MJPEG_CF * objPtr)
{

    /*
     * Blow away the old stuff.
     */

    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
    }
    objPtr->at_token = 0;

    /*
     * Check for overlaps if necessary.
     */

    if (objPtr->segmentListChangeFlag) {
      Segment *seg_ptr;

      seg_ptr = objPtr->segments;

      while (seg_ptr != NULL) {
	if (seg_ptr->next != NULL) {
	  if (seg_ptr->next->logicalStart < seg_ptr->logicalEnd) {
	    
	    /* Overlap detected. Definitely not ready. Return error. */

	    objPtr->ready = 0;
	    Tcl_AppendResult(objPtr->interp, "Segment overlap detected.", (char *) NULL);
	    return TCL_ERROR;
	  }
	}
	seg_ptr = seg_ptr->next;
      }

      objPtr->segmentListChangeFlag = 0;
    }

    /*
     * Set the ready flag.
     */
    objPtr->ready = 1;

    if (objPtr->lts != NULL) {
	MJPEG_CF_SendData((ClientData) objPtr);
    }
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_Unready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Stops sending data and puts the object in the unready state.
 ************************************************************/

static int
MJPEG_CF_Unready(MJPEG_CF * objPtr)
{
    objPtr->lastSeg = NULL;
    objPtr->lastFrameNum = -1;
    objPtr->ready = 0;
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_LTSChanged
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Procedure called whenever the LTS changes.
 ************************************************************/

void
MJPEG_CF_LTSChanged(ClientData clientdata)
{
    MJPEG_CF *objPtr = (MJPEG_CF *) clientdata;

    objPtr->extraFrameCounter = 0.0;
    objPtr->extraBWCounter = 0;

    if (objPtr->ready == 1) {
	MJPEG_CF_Ready(objPtr);
    }
    return;
}

/************************************************************
 *
 * GetInt32FromAVI
 *
 * Parameters:
 *     buffer - pointer to data read from AVI file.
 *     buf_index - index where byte inverted AVI 32 bit integer
 *                 starts in buffer.
 *
 * Return:
 *     32-bit int value calculated from buf_index through buf_index+3
 *     bytes of buffer.
 *
 * Action:
 *     Since integers in avi files are LSB order, this function calculates
 *     the integer value constructed from the bytes buf_index to buf_index+3
 *     of the buffer passed in. This function is used as a helper function
 *     of MJPEG_CF_ReadAsAVI.
 *
 ************************************************************/

static unsigned int
GetInt32FromAVI(unsigned char *buffer,     /* Pointer to actual bytes. */
		int buf_index     /* Index where 32-bit value starts. */)
{
  unsigned int result = 0;

  result = (((unsigned int) buffer[buf_index]) | 
	    (((unsigned int) (buffer[buf_index+1])) << 8) |
	    (((unsigned int) (buffer[buf_index+2])) << 16) | 
	    (((unsigned int) (buffer[buf_index+3])) << 24));

  return result;
}

#define FIND_AVI_LIST(tchar1, tchar2, tchar3, tchar4)                         \
  while ((!((buffer[buf_index] == 'L') && (buffer[buf_index+1] == 'I') &&     \
	  (buffer[buf_index+2] == 'S') && (buffer[buf_index+3] == 'T'))) ||   \
	 (!((buffer[buf_index+8] == tchar1) && (buffer[buf_index+9] == tchar2) && \
	    (buffer[buf_index+10] == tchar3) &&                               \
	    (buffer[buf_index+11] == tchar4)))) {                             \
    buf_index += 4;                                                           \
    junk_size = GetInt32FromAVI(buffer, buf_index);                           \
    buf_index += 4 + junk_size;                                               \
                                                                              \
    if (buf_index > num_read - 12) {                                          \
      lseek(objPtr->rawfd, buf_index+file_index, SEEK_SET);                   \
      file_index += buf_index;                                                \
      buf_index = 0;                                                          \
      num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);            \
      if (num_read < 12) goto error;                                          \
    }                                                                         \
  }

#define FIND_AVI_SUBCHUNK(tchar1, tchar2, tchar3, tchar4)                     \
  while (!((buffer[buf_index] == tchar1) && (buffer[buf_index+1] == tchar2) &&\
	    (buffer[buf_index+2] == tchar3) &&                                \
	    (buffer[buf_index+3] == tchar4))) {                               \
    buf_index += 4;                                                           \
    junk_size = GetInt32FromAVI(buffer, buf_index);                           \
    buf_index += 4 + junk_size;                                               \
                                                                              \
    if (buf_index > num_read - 8) {                                           \
      lseek(objPtr->rawfd, buf_index+file_index, SEEK_SET);                   \
      file_index += buf_index;                                                \
      buf_index = 0;                                                          \
      num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);            \
      if (num_read < 8) goto error;                                           \
    }                                                                         \
  }

/************************************************************
 *
 * MJPEG_CF_ReadAsJMovie
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Attempts to read file descriptor objPtr->rawfd as a Parallax JMovie file,
 *     constructing the necessary data structures and filling in the
 *     required info (i.e. whatever OpenClipfile would have done.)
 *     Returns TCL_OK on success, TCL_ERROR if not an JMovie file or if
 *     the file was not correct.
 *
 ************************************************************/

static int
MJPEG_CF_ReadAsJMovie(MJPEG_CF *objPtr       /* Pointer to mjpeg clipfile object. */)
{
    char magic_str[8];
    int fps, num_frames, width, height, qfactor;
    int indexbuf, au_slice;
    unsigned char j_movie_cmd;
    int cntr, raw_fot_frame_count;
    int skipped, frame_size;
    int *j_movie_offset_table;

    /*
     * The following are definitions matching the Parallax JMovie macros.
     * They are included here to avoid having to distribute the Parallax
     * include files.
     */

    int LOAD_AUDIO_0 = 229;
    int LOAD_AUDIO_1 = 228;
    int LOAD_AUDIO_2 = 227;
    int LOAD_AUDIO_3 = 226;
    int LOAD_JPEG = 236;
    int END_FRAME = 237;


    if (objPtr->rawfd < 0) return TCL_ERROR;

    lseek(objPtr->rawfd, 0, SEEK_SET);

    /* 
     * Confirm it is a j_movie file by looking for "j_movie" as first
     * 8 bytes.
     */

    read(objPtr->rawfd, magic_str, 8);
    
    if (strcmp(magic_str,"j_movie") != 0) {
	return TCL_ERROR;
    }

    /*
     * Get info out of jpheader. We do this without actually creating
     * a jpheader in order to avoid having to compile with the XVideo 
     * include files.
     */

    lseek(objPtr->rawfd, 4, SEEK_CUR);       /* Skip version number. */
    read(objPtr->rawfd, &fps, 4);
    fps = ntohl(fps);
    read(objPtr->rawfd, &num_frames, 4);
    num_frames = ntohl(num_frames);
    read(objPtr->rawfd, &width, 4);
    width = ntohl(width);
    read(objPtr->rawfd, &height, 4);
    height = ntohl(height);
    lseek(objPtr->rawfd, 4, SEEK_CUR);       /* Skip bandwidth. */
    read(objPtr->rawfd, &qfactor, 4);
    qfactor = ntohl(qfactor);
    lseek(objPtr->rawfd, 4, SEEK_CUR);       /* Skip mapsize. */
    read(objPtr->rawfd, &indexbuf, 4);
    indexbuf = ntohl(indexbuf); 
    lseek(objPtr->rawfd, 8, SEEK_CUR);       /* Skip tracks, volbase. */
    read(objPtr->rawfd, &au_slice, 4);

    /* 
     * Create raw frame offset table. 
     * Look in each j_movie frame for actual jpeg data.
     */

    j_movie_offset_table = (int *) ckalloc(num_frames*sizeof(int));
    lseek(objPtr->rawfd, indexbuf, SEEK_SET);
    for (cntr=0; cntr<num_frames; cntr++) {
	read(objPtr->rawfd, &(j_movie_offset_table[cntr]), 4);
	j_movie_offset_table[cntr] = ntohl(j_movie_offset_table[cntr]);
    }

    objPtr->rawFOT = (int *) ckalloc(num_frames*sizeof(int)*2);
    raw_fot_frame_count = 0;

    for (cntr=0; cntr<num_frames; cntr++) {
	lseek(objPtr->rawfd, j_movie_offset_table[cntr], SEEK_SET);
	
	read(objPtr->rawfd, &j_movie_cmd, 1);
	skipped = 0;

	while(j_movie_cmd != LOAD_JPEG) {
	    if ((j_movie_cmd == LOAD_AUDIO_0) ||
		(j_movie_cmd == LOAD_AUDIO_1) ||
		(j_movie_cmd == LOAD_AUDIO_2) ||
		(j_movie_cmd == LOAD_AUDIO_3)) {
		
		lseek(objPtr->rawfd, au_slice, SEEK_CUR);
	    } else if (j_movie_cmd == END_FRAME) {
		skipped = 1;
		break;
	    }

	    read(objPtr->rawfd, &j_movie_cmd, 1);
	}

	if (!skipped) {
	    read(objPtr->rawfd, &frame_size, 4);
	    frame_size = ntohl(frame_size);

	    objPtr->rawFOT[raw_fot_frame_count*2] = lseek(objPtr->rawfd, 0, SEEK_CUR);
	    objPtr->rawFOT[raw_fot_frame_count*2+1] = frame_size;
	    raw_fot_frame_count++;
	}
    }
	
    ckfree((char *) j_movie_offset_table);

    num_frames = raw_fot_frame_count;

    objPtr->rawFOT = (int *) ckrealloc((char *) objPtr->rawFOT, num_frames * 2 * sizeof(int));

    /*
     * Create frame time table.
     */

    objPtr->rawFTT = (double *) ckalloc((num_frames+1)*sizeof(double));

    for (cntr=0; cntr<=num_frames; cntr++) {
	objPtr->rawFTT[cntr] = (((double) cntr) * (1.0 / ((double) fps)));
    }

    objPtr->numFrames = num_frames;
    objPtr->clipLength = objPtr->rawFTT[num_frames];
    
    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);
    }
    objPtr->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));

    objPtr->cfInfo->width = width;
    objPtr->cfInfo->height = height;
    objPtr->cfInfo->device = PARALLAX_DEV_CODE;
    objPtr->cfInfo->subformat = RAW_JPEG_NO_HEADERS_CODE;
    objPtr->cfInfo->qfactor = qfactor;
    
    objPtr->dataStart = 0;
    
    return TCL_OK;
}
	
	 
    
   
    
/************************************************************
 *
 * MJPEG_CF_ReadAsVid
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Attempts to read file descriptor objPtr->rawfd as a .vid file,
 *     constructing the necessary data structures and filling in the
 *     required info (i.e. whatever OpenClipfile would have done.)
 *     Returns TCL_OK on success, TCL_ERROR if not an .vid file or if
 *     the file was not correct.
 *
 ************************************************************/

static int
MJPEG_CF_ReadAsVid(MJPEG_CF *objPtr       /* Pointer to mjpeg clipfile object. */)
{
    int width, height, size;
    int width2, height2;
    int frame_counter, num_read;
    int cur_pos, new_pos, i;
    int fot_allocated;
    
  if (objPtr->rawfd < 0) return TCL_ERROR;

  lseek(objPtr->rawfd, 0, SEEK_SET);

  read(objPtr->rawfd, &width, 4);
  width = ntohl(width);

  read(objPtr->rawfd, &height, 4);
  height = ntohl(height);

  read(objPtr->rawfd, &size, 4);
  size = ntohl(size);

  if ((width <= 0) || (height <= 0) || (size <= 0)) {
      return TCL_ERROR;
  }

  lseek(objPtr->rawfd, 20+size, SEEK_SET);

  read(objPtr->rawfd, &width2, 4);
  width2 = ntohl(width2);

  read(objPtr->rawfd, &height2, 4);
  height2 = ntohl(height2);

  if (width != width2) return TCL_ERROR;
  if (height != height2) return TCL_ERROR;

  /*
   * Seems to be a .vid file. No real guarantee, but if width and height
   * ever do not agree, abort.
   */

  lseek(objPtr->rawfd, 0, SEEK_SET);

  /* 
   * Create frame offset table. 
   */
  
  fot_allocated = 128;
  objPtr->rawFOT = (int *) ckalloc(fot_allocated*sizeof(int)*2);

  frame_counter = 0;
  num_read = 0;

  num_read += read(objPtr->rawfd, &width2, 4);
  num_read += read(objPtr->rawfd, &height2, 4);
  num_read += read(objPtr->rawfd, &size, 4);

  width2 = ntohl(width2);
  height2 = ntohl(height2);
  size = ntohl(size);
  
  while (num_read == 12) {

      if ((width2 != width) || (height2 != height))
	goto error;

      if (size < 0)
	goto error;

      cur_pos = lseek(objPtr->rawfd, 0, SEEK_CUR);
      lseek(objPtr->rawfd, 8+size, SEEK_CUR);
      new_pos = lseek(objPtr->rawfd, 0, SEEK_CUR);

      if (new_pos - cur_pos != 8+size)
	goto error;
      
      if (frame_counter >= fot_allocated) {
	  fot_allocated += 128;
	  objPtr->rawFOT = (int *) ckrealloc((char *) objPtr->rawFOT, 
					   fot_allocated*2*sizeof(int));
      }

      objPtr->rawFOT[frame_counter*2] = cur_pos-12;
      objPtr->rawFOT[(frame_counter*2)+1] = size+20;
      frame_counter++;

      num_read = 0;
      num_read += read(objPtr->rawfd, &width2, 4);
      num_read += read(objPtr->rawfd, &height2, 4);
      num_read += read(objPtr->rawfd, &size, 4);

      width2 = ntohl(width2);
      height2 = ntohl(height2);
      size = ntohl(size);
  }

  if (fot_allocated > frame_counter) {
      objPtr->rawFOT = (int *) ckrealloc((char *) objPtr->rawFOT,
					 frame_counter*2*sizeof(int));
  }

  /* Now create FTT. */
  
  objPtr->rawFTT = (double *) ckalloc((frame_counter+1)*sizeof(double));

  for (i=0; i<=frame_counter; i++) {
      objPtr->rawFTT[i] = (((double) i) * (1.0 / 30.0));
  }

  /* Now fill in cm info, etc. */

  objPtr->numFrames = frame_counter + 1;
  objPtr->clipLength = objPtr->rawFTT[frame_counter];

  if (objPtr->cfInfo != NULL) {
      ckfree((char *) objPtr->cfInfo);
  }
  objPtr->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));

  objPtr->cfInfo->width = width;
  objPtr->cfInfo->height = height;
  objPtr->cfInfo->device = PARALLAX_DEV_CODE;
  objPtr->cfInfo->subformat = VID_FRAME_CODE;
  objPtr->cfInfo->qfactor = 150;

  objPtr->dataStart = 0;

  return TCL_OK;

 error:
  if (objPtr->rawFOT != NULL) {
      ckfree((char *) objPtr->rawFOT);
      objPtr->rawFOT = NULL;
  }
  if (objPtr->rawFTT != NULL) {
      ckfree((char *) objPtr->rawFTT);
      objPtr->rawFTT = NULL;
  }
  return TCL_ERROR;
}

  



/************************************************************
 *
 * MJPEG_CF_ReadAsAVI
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Attempts to read file descriptor objPtr->rawfd as an AVI file,
 *     constructing the necessary data structures and filling in the
 *     required info (i.e. whatever OpenClipfile would have done.)
 *     Returns TCL_OK on success, TCL_ERROR if not an AVI file or if
 *     the file was not correct.
 *
 ************************************************************/

static int
MJPEG_CF_ReadAsAVI(MJPEG_CF *objPtr       /* Pointer to mjpeg clipfile object. */)
{
  char buffer[CF_READ_BUFFER_SIZE];
  int num_read;
  int i;
  int buf_index, file_index, strl_start, strl_size;
  int stream_length, stream_rate, stream_scale;
  int avi_success, junk_size, main_avi_hdr_size;
  int stream_num, found_stream, stream_header_size;
  int strf_header_size, stream_width, stream_height;
  char stream_id[10];
  int frame_counter, movi_length, movi_end, list_size;
  int frame_length;

  /* 
   * Assume the objPtr->rawfd is set at beginning of file. Resolve
   * whether file is in fact an avi file.
   */

  if (objPtr->rawfd < 0) return TCL_ERROR;

  lseek(objPtr->rawfd, 0, SEEK_SET);
  num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
  file_index = 0;
  buf_index = 0;

  if (num_read < 24) goto error;
  avi_success = 0;
  
  while (buf_index < num_read-24) {
    if ((buffer[buf_index] == 'R') && (buffer[buf_index+1] == 'I') &&
	(buffer[buf_index+2] == 'F') && (buffer[buf_index+3] == 'F') &&
	(buffer[buf_index+8] == 'A') && (buffer[buf_index+9] == 'V') &&
	(buffer[buf_index+10] == 'I') && (buffer[buf_index+11] == ' ')) {
      avi_success = 1;
      break;
    }
    buf_index++;
  }

  if (!avi_success) return TCL_ERROR;

  /*
   * Now we know its an avi, lets make sure that there is a mjpeg stream 
   * in the file and find the stream number.
   */

  buf_index += 12;

  FIND_AVI_LIST('h', 'd', 'r', 'l');
  
  buf_index += 12;

  if (buf_index > num_read-8) {
    lseek(objPtr->rawfd, buf_index+file_index, SEEK_SET);
    file_index += buf_index;
    buf_index = 0;
    num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
    if (num_read < 8) goto error;
  }

  FIND_AVI_SUBCHUNK('a', 'v', 'i', 'h');
  buf_index += 4;

  main_avi_hdr_size = GetInt32FromAVI(buffer, buf_index);
  buf_index += 4;

  buf_index += main_avi_hdr_size;

  if (buf_index > num_read-8) {
    lseek(objPtr->rawfd, file_index+buf_index, SEEK_SET);
    file_index += buf_index;
    buf_index = 0;
    num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
    if (num_read < 8) goto error;
  }

  FIND_AVI_LIST('s', 't', 'r', 'l');

  buf_index += 4;

  strl_size = GetInt32FromAVI(buffer, buf_index);
  buf_index += 4;
  
  strl_start = file_index + buf_index;
  buf_index += 4;


  if (buf_index > num_read-8) {
    lseek(objPtr->rawfd, file_index+buf_index, SEEK_SET);
    file_index += buf_index;
    buf_index = 0;
    num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
    if (num_read < 8) goto error;
  }

  stream_num = -1;
  found_stream = 0;

  stream_length = 0;
  stream_rate = stream_scale = 0;
  stream_width = stream_height = 0;
  while (1) {
    FIND_AVI_SUBCHUNK('s', 't', 'r', 'h');
    
    stream_num++;
    buf_index += 4;
    
    stream_header_size = GetInt32FromAVI(buffer, buf_index);

    buf_index += 4;
    
    if ((buffer[buf_index] == 'v') && (buffer[buf_index+1] == 'i') &&
	(buffer[buf_index+2] == 'd') && (buffer[buf_index+3] == 's')) {
      
      stream_scale = GetInt32FromAVI(buffer, buf_index+20);
      stream_rate = GetInt32FromAVI(buffer, buf_index+24);
      stream_length = GetInt32FromAVI(buffer, buf_index+32);
      
      buf_index += stream_header_size;
      
      if (buf_index > num_read - 8) {
	lseek(objPtr->rawfd, file_index+buf_index, SEEK_SET);
	file_index += buf_index;
	buf_index = 0;
	num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
	if (num_read < 8) goto error;
      }
      
      FIND_AVI_SUBCHUNK('s', 't', 'r', 'f');

      buf_index += 4;
      
      strf_header_size = GetInt32FromAVI(buffer, buf_index);
      buf_index += 4;
      
      if ((buffer[buf_index+16] == 'M') && (buffer[buf_index+17] == 'J') &&
	  (buffer[buf_index+18] == 'P') && (buffer[buf_index+19] == 'G')) {
	
	stream_width = GetInt32FromAVI(buffer, buf_index+4);
	stream_height = GetInt32FromAVI(buffer, buf_index+8);

	found_stream = 1;
      }

      buf_index += strf_header_size;
    
    } else {
      buf_index += stream_header_size;
    }

    if (found_stream) break;
    
    if (buf_index > num_read - 8) {
      lseek(objPtr->rawfd, file_index+buf_index, SEEK_SET);
      file_index += buf_index;
      buf_index = 0;
      num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
      if (num_read < 8) goto error;
    }
  }

  if (!found_stream) {
    Tcl_AppendResult(objPtr->interp, "AVI file did not contain MJPG data stream.",
		     (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Found a motion jpeg stream at this point as well as general things like:
   * width, height, how many frames, frame rate.
   * Now construct the data structures by doing:
   * 1) Find all the frames in the avi file, creating an offset table.
   * 2) Adjust number of frames to match number of frames found.
   * 3) Create a frame time table.
   * 4) Fill in cf info structure and general info.
   * 5) Return.
   */

  /* 
   * Find all frames, create FOT.
   */

  sprintf(stream_id, "%.2d", stream_num);

  if (stream_length < 1) {
    /* Stream length does not contain valid info. Heuristically use 128 to start. */

    stream_length = 128;
  }

  objPtr->rawFOT = (int *) ckalloc(stream_length*sizeof(int)*2);

  frame_counter = 0;

  lseek(objPtr->rawfd, strl_start+strl_size, SEEK_SET);
  file_index = strl_start+strl_size;
  buf_index = 0;
  num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
  if (num_read < 12) goto error;

  FIND_AVI_LIST('m', 'o', 'v', 'i');

  buf_index += 4;

  movi_length = GetInt32FromAVI(buffer, buf_index);
  buf_index += 4;
  movi_end = file_index + buf_index + movi_length;
  buf_index += 4;

  if (buf_index > num_read - 12) {
    lseek(objPtr->rawfd, buf_index+file_index, SEEK_SET);
    file_index += buf_index;
    buf_index = 0;
    num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
    if (num_read < 12) goto error;
  }

  while (buf_index + file_index < movi_end) {
    
    if ((buffer[buf_index] == 'L') && (buffer[buf_index+1] == 'I') &&
	(buffer[buf_index+2] == 'S') && (buffer[buf_index+3] == 'T')) {
      buf_index += 4;

      if ((buffer[buf_index+4] == 'r') && (buffer[buf_index+5] == 'e') &&
	  (buffer[buf_index+6] == 'c') && (buffer[buf_index+7] == ' ')) {
	buf_index += 8;
      } else {
	list_size = GetInt32FromAVI(buffer, buf_index);
	buf_index += 4 + list_size;
      }
    } else {
      
      frame_length = GetInt32FromAVI(buffer, buf_index+4);

      if ((buffer[buf_index] == stream_id[0]) && 
	  (buffer[buf_index+1] == stream_id[1]) &&
	  (buffer[buf_index+2] == 'd') &&
	  (buffer[buf_index+3] == 'c')) {

	if (frame_counter >= stream_length) {
	  stream_length += 128;
	  objPtr->rawFOT = (int *) ckrealloc((char *) objPtr->rawFOT, sizeof(int)*2*stream_length);
	}

	objPtr->rawFOT[frame_counter*2] = file_index + buf_index + 8;
	objPtr->rawFOT[(frame_counter*2)+1] = frame_length;
	frame_counter++;
      }

      buf_index += 8;
      if (frame_length%2) {
	buf_index += frame_length + 1;
      } else {
	buf_index += frame_length;
      }
    }
  
    if (buf_index > num_read - 12) {
/*      lseek(objPtr->rawfd, buf_index+file_index, SEEK_SET); */
	lseek(objPtr->rawfd, buf_index-num_read, SEEK_CUR);
      file_index += buf_index;
      buf_index = 0;
      num_read = read(objPtr->rawfd, buffer, CF_READ_BUFFER_SIZE);
      if (num_read < 12) break;
    }
  }

  /* 
   * Done computing FOT, now fill in general info.
   * NOTE: stream_length has number of frames found.
   */

  objPtr->numFrames = stream_length;
  objPtr->clipLength = (((double) stream_length) /
			(((double) stream_rate) / ((double) stream_scale)));
  
  if (objPtr->cfInfo != NULL) {
    ckfree((char *) objPtr->cfInfo);
  }
  objPtr->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));

  objPtr->cfInfo->width = stream_width;
  objPtr->cfInfo->height = stream_height;
  objPtr->cfInfo->device = J300_DEV_CODE;
  objPtr->cfInfo->subformat = JFIF_NO_HUFFMAN_FRAME_CODE;
  objPtr->cfInfo->qfactor = 100;
  objPtr->dataStart = 0;

  
  /* 
   * Now create FTT.
   */

  objPtr->rawFTT = (double *) ckalloc(sizeof(double)*(stream_length+1));

  for (i=0; i<=stream_length; i++) {
    objPtr->rawFTT[i] = (((double) i) / ((double) stream_length)) * objPtr->clipLength;
  }

  /* 
   * Ok, we're done.
   */

  return TCL_OK;

  /*
   * Error handling. We jump here if reading data went awry. 
   */

 error:

  Tcl_AppendResult(objPtr->interp, "Error reading data from avi file.", (char *) NULL);
  if (objPtr->rawFTT != NULL) {
    ckfree((char *) objPtr->rawFTT);
    objPtr->rawFTT = NULL;
  }
  if (objPtr->rawFOT != NULL) {
    ckfree((char *) objPtr->rawFOT);
    objPtr->rawFOT = NULL;
  }
  if (objPtr->cfInfo != NULL) {
    ckfree((char *) objPtr->cfInfo);
    objPtr->cfInfo = NULL;
  }
  return TCL_ERROR;
}


  
  


/************************************************************
 *
 * MJPEG_CF_OpenRawfile
 *
 * Parameters:
 *     objPtr - pointer to object
 *     filename - filename
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Opens the file as a raw data file, constructs  basic info data.
 *     Closes file when done.
 ************************************************************/

static int
MJPEG_CF_OpenRawfile(MJPEG_CF *objPtr,         /* Pointer to mjpeg clipfile object. */
		     char *filename            /* File to be opened as raw data. */)
{

  objPtr->rawfd = open(filename, O_RDONLY|O_BINARY);

  if (objPtr->rawfd < 0) {
    objPtr->rawfd = -1;
    return TCL_ERROR;
  }

  /* 
   * Send object pointer to routines that attempt
   * to read the file as a particular format. Each routine is responsible
   * for constructing the rawFTT and rawFOT as well as filling in any
   * info in the object structure that OpenClipfile would have done.
   * If the routine succeeds, close the fd and return TCL OK. If not,
   * reset the fd with an lseek and try the next format routine.
   */

  if (MJPEG_CF_ReadAsAVI(objPtr) == TCL_OK) {
    if (objPtr->rawfd > -1) {
      close(objPtr->rawfd);
      objPtr->rawfd = -1;
    }
    return TCL_OK;
  }

  lseek(objPtr->rawfd, 0, SEEK_SET);

  if (MJPEG_CF_ReadAsVid(objPtr) == TCL_OK) {
      if (objPtr->rawfd > -1) {
	  close(objPtr->rawfd);
	  objPtr->rawfd = -1;
      }
      return TCL_OK;
  }

  lseek(objPtr->rawfd, 0, SEEK_SET);

  if (MJPEG_CF_ReadAsJMovie(objPtr) == TCL_OK) {
      if (objPtr->rawfd > -1) {
	  close(objPtr->rawfd);
	  objPtr->rawfd = -1;
      }
      return TCL_OK;
  }

  /* 
   * No more formats. Close rawfd and return TCL_ERROR.
   */

  if (objPtr->rawfd > -1) {
    close(objPtr->rawfd);
    objPtr->rawfd = -1;
  }

  return TCL_ERROR;
}
  

/************************************************************
 *
 * MJPEG_CF_OpenClipfile
 *
 * Parameters:
 *     objPtr - pointer to object
 *     filename - file name
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Opens the clipfile, reads basic info data.
 *     Closes file when done.
 ************************************************************/

static int
MJPEG_CF_OpenClipfile(MJPEG_CF *objPtr, char *filename)
{
    tChunkHeader chunk_hdr;
    MediaIndepHeader mi_hdr;

    if (objPtr->cfState != NULL) {
	Tcl_AppendResult(objPtr->interp, "Problem: chunk file already open", 
			 (char *) NULL);
	return TCL_ERROR;
    }

    /* 
     * Attempt to open the file specified as a chunk file (i.e. pre formatted
     * clipfile). If this fails, attempt to open the file as a raw data file.
     */

    objPtr->cfState = Chunk_OpenFile(filename, "r");
    if (objPtr->cfState == NULL) {

      if (MJPEG_CF_OpenRawfile(objPtr, filename) != TCL_OK) {
	Tcl_AppendResult(objPtr->interp, "Couldn't open file: ", filename,
			 " as either mjpeg clipfile or raw mjpeg data file.", 
			 (char *) NULL);
	return TCL_ERROR;
      }
      return TCL_OK;
    }

    /*
     * Get information from media independent header. 
     */

    if (Chunk_FindType(objPtr->cfState, CM_MEDIA_INDEP_CF, &chunk_hdr) != 0) {
	Tcl_AppendResult(objPtr->interp,
		      "Couldn't find media independent header.", (char *) NULL);
	return TCL_ERROR;
    }

    read(objPtr->cfState->fd, &mi_hdr, sizeof(mi_hdr));

    mi_hdr.media = ntohl(mi_hdr.media);
    mi_hdr.format = ntohl(mi_hdr.format);
    mi_hdr.device = ntohl(mi_hdr.device);
    mi_hdr.version = ntohl(mi_hdr.version);
    mi_hdr.num_frames = ntohl(mi_hdr.num_frames);
    mi_hdr.length_s = ntohl(mi_hdr.length_s);
    mi_hdr.length_ns = ntohl(mi_hdr.length_ns);

    /*
     * Sanity check on media and format type.
     */

    if ((mi_hdr.media != VIDEO_MEDIA_CODE) ||
	(mi_hdr.format != MJPEG_FORMAT_CODE)) {
      Tcl_AppendResult(objPtr->interp,
		       "File is not mjpeg chunkfile.", (char *) NULL);
      return TCL_ERROR;
    }

    objPtr->numFrames = mi_hdr.num_frames;
    objPtr->clipLength = mi_hdr.length_s + mi_hdr.length_ns/1000000000.0;

    /*
     * Get information from mjpeg clipfile header.
     */

    if (Chunk_FindType(objPtr->cfState, CM_MJPEG_CF, &chunk_hdr) != 0) {
	Chunk_ReadRewind(objPtr->cfState);
	if (Chunk_FindType(objPtr->cfState, CM_MJPEG_CF, &chunk_hdr) != 0) {
	    Tcl_AppendResult(objPtr->interp,
			"Couldn't find motion jpeg header.", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);;
    }
    objPtr->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));

    read(objPtr->cfState->fd, objPtr->cfInfo, sizeof(MJPEGHeader));

    objPtr->cfInfo->width = ntohl(objPtr->cfInfo->width);
    objPtr->cfInfo->height = ntohl(objPtr->cfInfo->height);
    objPtr->cfInfo->device = ntohl(objPtr->cfInfo->device);
    objPtr->cfInfo->subformat = ntohl(objPtr->cfInfo->subformat);
    objPtr->cfInfo->qfactor = ntohl(objPtr->cfInfo->qfactor);

    /*
     * Find data segment and store location in file.
     */

    if (Chunk_FindType(objPtr->cfState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
	Chunk_ReadRewind(objPtr->cfState);
	if (Chunk_FindType(objPtr->cfState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
	    Tcl_AppendResult(objPtr->interp, "Couldn't find data segment.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
    }

    objPtr->dataStart = lseek(objPtr->cfState->fd, 0, SEEK_CUR);

    Chunk_CloseFile(objPtr->cfState);
    objPtr->cfState = NULL;
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
MJPEG_CF_ChangeOutCmd(MJPEG_CF *objPtr, char *newOut) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
		&(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    

/************************************************************
 *
 * MJPEG_CF_SendData
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sends data according to current value of LTS, queues
 *     itself on the at queue for next send time.
 ************************************************************/

void
MJPEG_CF_SendData(ClientData clientData) 
{
    MJPEG_CF *objPtr = (MJPEG_CF *) clientData;
    double now, speed, when;
    double eTime, sTime, lastETime;
    Segment *segPtr;
    int i, curBw;
    MJPEGBufferHeader hdrPtr;
    double frac_want_to_send;
    int want_to_send;
    int bw_want_to_send;
    Buffer *bufPtr;
    int idx;
    int dataSegCode, dataLength;
    ScatterBuffer *sbPtr;
    SB_List *sbl;
    char *sbStr;
    int numToSend, bwToSend, origNumToSend;
    int frameNum, curIndex;
    int sbIdx;
    static FramePick *fPickList;
    static int fPickNum = 0;
    extern void MJPEG_Prioritize (int n, int m, int k, int *pl);
    int data_fd;

    /*
     * Return if not ready, no LTS, or no out command.
     */
    if ((!objPtr->ready) ||
	(objPtr->lts == NULL) ||
	(objPtr->outCmd == NULL)) {
	goto done;
    }

    /*
     * Find segment we are in.  Two cases: forward and reverse play.
     */
    now = LTS_Value(objPtr->lts);
    speed = LTS_Speed(objPtr->lts);
    now += objPtr->sendAhead * speed;
    if (objPtr->lastSeg == NULL) 
      segPtr = objPtr->segments;
    else segPtr = objPtr->lastSeg;

    if (speed >= 0.0) {
	while (segPtr != NULL) {
	    if (segPtr->logicalEnd > now) {
		if (segPtr->logicalStart <= now)
		  break;
		if (segPtr->prev == NULL)
		  break;
		segPtr = segPtr->prev;
	    } else {
	      if (segPtr->next != NULL) {
		if (segPtr->next->logicalStart > now) {
		  segPtr = segPtr->next;
		  break;
		}
	      }
	      segPtr = segPtr->next;
	    }
	}
    } else {
	/*
	 * Speed is backwards so now should be less or equal to segment end.
	 * and strictly greater than beginning. Almost identical to above
	 * code but with different boundary conditions.
	 */

	while (segPtr != NULL) {
	    if (segPtr->logicalEnd >= now) {
		if (segPtr->logicalStart < now)
		  break;
		segPtr = segPtr->prev;
	    } else {
	      if (segPtr->next != NULL) {
		if (segPtr->next->logicalStart >= now) {
		    break;
		}
	      } else break;
		segPtr = segPtr->next;
	    }
	}
    }

    /*
     * If segment pointer is NULL, means we will never be
     * needed (i.e. time past all segments or going backwards
     * before all segments).
     */

    if (segPtr == NULL) {
	objPtr->lastSeg = NULL;
	objPtr->lastFrameNum = -1;
	goto done;
    }

    /*
     * If we are in the same segment we were in previously, lastFrameNum is
     * probably still good. If not, blow it away.
     */

    if (segPtr != objPtr->lastSeg) {
	objPtr->lastFrameNum = -1;
    }
    objPtr->lastSeg = segPtr;

    /*
     * If not in a segment, calculate when we will be and set up at queue.
     */
    if (speed > 0.0) {
	if (now < segPtr->logicalStart) {
	    when = LTS_SystemTimeOf(objPtr->lts,
			(segPtr->logicalStart-(objPtr->sendAhead*speed)));
	    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				       "video", MJPEG_CF_SendData, objPtr, NULL,
				       NULL, NULL, NULL, NULL);
	    return;
	}
    } else if (speed < 0.0) {
	if (now > segPtr->logicalEnd) {
	    when = LTS_SystemTimeOf(objPtr->lts,
			segPtr->logicalEnd+(objPtr->sendAhead*speed));
	    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				     "video", MJPEG_CF_SendData, objPtr, NULL,
				     NULL, NULL, NULL, NULL);
	  return;
	}
    } else {
	if ((now < segPtr->logicalStart) || (now >= segPtr->logicalEnd)) {
	    return;
	}
    }

    /*
     * If chunk file not open, open it now. Set data_fd to chunk file descriptor.
     * If raw file, open raw file and set data_fd.
     */
    
    if (objPtr->rawFOT != NULL) {
      if (objPtr->rawfd < 0) {
	objPtr->rawfd = open(objPtr->filename, O_RDONLY|O_BINARY);
      }
      data_fd = objPtr->rawfd;
    } else {      
      if (objPtr->cfState == NULL) {
	objPtr->cfState = Chunk_OpenFile(objPtr->filename, "r");
      }
      data_fd = objPtr->cfState->fd;
    }

    /*
     * If only sending one frame because speed = 0.0 or
     * cycle time = 0.0, do this now. 
     */
    if ((objPtr->cycleTime == 0.0) || (speed == 0.0)) {
	
	/*
	 * Find frame to send.
	 */
	if (speed > 0.0) {
	    for (i=objPtr->lastFrameNum+1; i<segPtr->numFrames; i++)  {
		if (segPtr->FTT[i+1] > now) 
		  break;
	    }
	    for (; i>=0; i--) {
		if (segPtr->FTT[i] <= now) 
		  break;
	    }
	} else if (speed < 0.0) {
	    if (objPtr->lastFrameNum == -1) 
	      i = segPtr->numFrames;
	    else i = objPtr->lastFrameNum;

	    for (; i>0; i--) {
		if (segPtr->FTT[i-1] < now) 
		    break;
	    }
	    for (; i<=objPtr->lastFrameNum; i++) {
		if (segPtr->FTT[i] >= now)
		  break;
	    }
	} else {
	    for (i=0; i<segPtr->numFrames; i++) {
		if ((segPtr->FTT[i] <= now) && (segPtr->FTT[i+1] > now)) {
		    break;
		}
	    }
	}
	
	if (speed > 0.0) {
	    if (i < 0) i = 0;
	    if (i >= segPtr->numFrames) {
		i = segPtr->numFrames-1;
	    }

	    sTime = segPtr->FTT[i];
	    eTime = segPtr->FTT[i+1];
	    frameNum = i;
	} else if (speed < 0.0) {
	    if (i <= 0) i = 1;
	    if (i > segPtr->numFrames) {
		i = segPtr->numFrames;
	    }

	    sTime = segPtr->FTT[i];
	    eTime = segPtr->FTT[i-1];
	    frameNum = i-1;
	} else {
	  if (i < 0) i = 0;
	  if (i >= segPtr->numFrames) {
	      i = segPtr->numFrames-1;
	  }

	  sTime = eTime = now;
	  frameNum = i;
	}

	objPtr->lastFrameNum = frameNum;

	hdrPtr.hdrType = htonl(CM_MJPEG_BUF_HDR);
	hdrPtr.length = htonl(36);
	hdrPtr.width = htonl(objPtr->cfInfo->width);
	hdrPtr.height = htonl(objPtr->cfInfo->height);
	hdrPtr.device = htonl(objPtr->cfInfo->device);
	hdrPtr.subformat = htonl(objPtr->cfInfo->subformat);
	hdrPtr.qfactor = htonl(objPtr->cfInfo->qfactor);
	hdrPtr.logStart_s = htonl(floor(sTime));
	hdrPtr.logStart_ns = htonl(floor((sTime - floor(sTime))* ONE_BILLION));
	hdrPtr.logEnd_s = htonl(floor(eTime));
	hdrPtr.logEnd_ns = htonl(floor((eTime - floor(eTime))* ONE_BILLION));

	lseek(data_fd,
		objPtr->dataStart+segPtr->FOT[frameNum*2], SEEK_SET);

	bufPtr = MakeBuffer(objPtr->interp,
		    sizeof(MJPEGBufferHeader)+8+segPtr->FOT[(frameNum*2)+1],
		    "tcmMJPEGClipfile:MJPEG_CF_SendData.1");
	idx = 0;

	memcpy(bufPtr->data+idx, &hdrPtr, sizeof(MJPEGBufferHeader));
	idx += sizeof(MJPEGBufferHeader);

	dataSegCode = htonl(CM_DATA_SEGMENT);
	dataLength = htonl(segPtr->FOT[(frameNum*2)+1]);

	memcpy(bufPtr->data+idx, &dataSegCode, 4);
	idx += 4;

	memcpy(bufPtr->data+idx, &dataLength, 4);
	idx += 4;

	read(data_fd, bufPtr->data+idx,
		segPtr->FOT[(frameNum*2)+1]); 
	idx += segPtr->FOT[(frameNum*2)+1];

	/*
	 * Create sbl.
	 */
	sbPtr = MakeScatterBuffer(bufPtr, idx, NULL);
	sbl = MakeSBList(sbPtr, NULL);

	/*
	 * If out command is C, create scatter buffer list and call.
	 */
	if (objPtr->outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, sbl);	       
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree((char *) objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and 
	     * call using Tcl_Eval. 
	     */
	    sbStr = MakeSBListString(sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree((char *) sbStr);
	}
	FreeSBList(sbl);

	DetachBuffer(objPtr->interp, bufPtr, 
		     "tcmMJPEGClipfile:MJPEG_CF_SendData");

	/*
	 * Figure out when we need to send again, queue it up and return.
	 */
	if (speed == 0.0) {
	    return;
	}

	when = LTS_SystemTimeOf(objPtr->lts,
		(eTime-(objPtr->sendAhead*speed)));
	objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				     "video", MJPEG_CF_SendData, objPtr, NULL,
				     NULL, NULL, NULL, NULL);
	return;
    } 
    
    /*
     * Cycle time is not 0.0, so send multiple frames.
     * Find frames to be sent.
     */
    numToSend = 0;
    bwToSend = 0;
    if (fPickNum == 0) {
	fPickList = (FramePick *) ckalloc(sizeof(FramePick)*100);
	fPickNum = 100;
    }

    /*
     * Find frame to send.
     */
    if (speed >= 0.0) {
	for (i=objPtr->lastFrameNum+1; i<segPtr->numFrames; i++)  {
	    if (segPtr->FTT[i+1] > now) 
	      break;
	}
	for (; i>=0; i--) {
	    if (segPtr->FTT[i] <= now) 
	      break;
	}
    } else if (speed < 0.0) {
	if (objPtr->lastFrameNum == -1) 
	  i = segPtr->numFrames;
	else i = objPtr->lastFrameNum;
	
	for (; i>0; i--) {
	    if (segPtr->FTT[i-1] < now) 
	      break;
	}
	for (; i<=objPtr->lastFrameNum; i++) {
	    if (segPtr->FTT[i] >= now)
	      break;
	}
    }

    if (speed >= 0.0) {
	if (i < 0) {
	   i = 0;
	} else if (i >= segPtr->numFrames) {
	    i = segPtr->numFrames-1;
	}

	fPickList[0].sTime = segPtr->FTT[i];
	fPickList[0].eTime = segPtr->FTT[i+1];
	fPickList[0].index = i;
	fPickList[0].segPtr = segPtr;
	fPickList[0].offset = segPtr->FOT[i*2];
	fPickList[0].size = segPtr->FOT[(i*2)+1];
	fPickList[0].sendFlag = 1;
    } else {
	if (i <= 0) {
	   i = 1;
	} else if (i > segPtr->numFrames) {
	    i = segPtr->numFrames;
	}

	fPickList[0].sTime = segPtr->FTT[i];
	fPickList[0].eTime = segPtr->FTT[i-1];
	fPickList[0].index = i-1;
	fPickList[0].segPtr = segPtr;
	fPickList[0].offset = segPtr->FOT[(i-1)*2];
	fPickList[0].size = segPtr->FOT[((i-1)*2)+1];
	fPickList[0].sendFlag = 1;
    }
    numToSend = 1;
    bwToSend = fPickList[0].size;
    curIndex = fPickList[0].index;

    objPtr->lastSeg = segPtr;
    objPtr->lastFrameNum = curIndex;

    if (speed > 0.0) {
	while (1) {
	    curIndex++;
	    if (curIndex >= segPtr->numFrames) {
		segPtr = segPtr->next;
		if (segPtr == NULL) {
		    break;
		}
		curIndex = 0;
	    }
	    if (segPtr->FTT[curIndex] < now+objPtr->cycleTime*speed) {
		objPtr->lastSeg = segPtr;
		objPtr->lastFrameNum = curIndex;
		if (numToSend >= fPickNum) {
		    fPickNum += 100;
		    fPickList = (FramePick *) ckrealloc((char *) fPickList,
				    sizeof(FramePick)*fPickNum);
		}

		fPickList[numToSend].sTime = segPtr->FTT[curIndex];
		fPickList[numToSend].eTime = segPtr->FTT[curIndex+1];
		fPickList[numToSend].index = curIndex;
		fPickList[numToSend].segPtr = segPtr;
		fPickList[numToSend].offset = segPtr->FOT[curIndex*2];
		fPickList[numToSend].size = segPtr->FOT[(curIndex*2)+1];
		fPickList[numToSend].sendFlag = 1;
	
		bwToSend += fPickList[numToSend].size;
		numToSend++;
	    } else {
		break;
	    }
	}
    } else if (speed < 0.0) {
	while (1) {
	    curIndex--;
	    if (curIndex < 0) {
		segPtr = segPtr->prev;
		if (segPtr == NULL) {
		    break;
		}
		curIndex = segPtr->numFrames-1;
	    }
	    if (segPtr->FTT[curIndex+1] > now+objPtr->cycleTime*speed) {
		objPtr->lastSeg = segPtr;
		objPtr->lastFrameNum = curIndex;

		if (numToSend >= fPickNum) {
		    fPickNum += 100;
		    fPickList = (FramePick *) ckrealloc((char *) fPickList,
				    sizeof(FramePick)*fPickNum);
		}
		
		fPickList[numToSend].sTime = segPtr->FTT[curIndex+1];
		fPickList[numToSend].eTime = segPtr->FTT[curIndex];
		fPickList[numToSend].index = curIndex;
		fPickList[numToSend].segPtr = segPtr;
		fPickList[numToSend].offset = segPtr->FOT[curIndex*2];
		fPickList[numToSend].size = segPtr->FOT[(curIndex*2)+1];
		fPickList[numToSend].sendFlag = 1;
	
		bwToSend += fPickList[numToSend].size;
		numToSend++;
	    } else {
		break;
	    }
	}
    }

   /*
    * If number of frames to be sent greater than the number of
    * frames we can send (due to bandwidth or max frame rate),
    * mark low priority frames so we don't bother reading them.
    */
    frac_want_to_send = (objPtr->maxFrameRate * objPtr->cycleTime) -
      objPtr->extraFrameCounter;
    want_to_send = (int) (ceil(frac_want_to_send));

    if (objPtr->maxFrameRate != 0.0) {
	objPtr->extraFrameCounter = ((double) want_to_send) - frac_want_to_send;
    }

    bw_want_to_send = objPtr->bandwidth - ((int) ((((double) objPtr->extraBWCounter) /
						   objPtr->cycleTime)));
    curBw = (int) (((double) bwToSend) / objPtr->cycleTime);
    origNumToSend = numToSend;
    if ((numToSend > 0) &&
	(((numToSend > want_to_send) && objPtr->maxFrameRate) ||
	 ((curBw  > bw_want_to_send)    && objPtr->bandwidth &&
	  ((numToSend > 1) || (objPtr->extraBWCounter > 0))))) {
	
	int *orderVec;

	orderVec = (int *)ckalloc(numToSend*sizeof(int));
	MJPEG_Prioritize (numToSend, 0, 1, orderVec);
	while ((numToSend > 0) &&
	       (((numToSend > want_to_send) && objPtr->maxFrameRate) ||
		((curBw  > bw_want_to_send)    && objPtr->bandwidth &&
		 ((numToSend > 1) || (objPtr->extraBWCounter > 0))))) {
	    numToSend--;
	    fPickList[orderVec[numToSend]].sendFlag = 0;
	    bwToSend -= fPickList[orderVec[numToSend]].size;
	    curBw  = (int) (((double) bwToSend) / objPtr->cycleTime);
	}
	ckfree ((char *)orderVec);
    }

    /*
     * Adjust bandwidth counter.
     */
    
    if (objPtr->bandwidth > 0) {
      objPtr->extraBWCounter += bwToSend - ((int) (objPtr->bandwidth * 
						   objPtr->cycleTime));
    }

    if (numToSend > 0) {

	/*
	 * Adjust start and end times to take up space for frames
	 * that will not be sent. Do this only if frames were
	 * actually dropped.
	 */

	if (numToSend != origNumToSend) {
	    double midpoint;
	    int j;

	    
	    for (i=0; i<origNumToSend; i++) {
		if (fPickList[i].sendFlag == 1)
		  break;
	    }
	    if (i != 0) {
		fPickList[i].sTime = fPickList[0].sTime;
	    }
	    while (i<origNumToSend-1) {
		if (fPickList[i+1].sendFlag == 1) {
		    i++;
		} else {
		    j = i+1;
		    while(j<origNumToSend) {
			if (fPickList[j].sendFlag == 1) {
			    break;
			}
			j++;
		    }
		    if (j==origNumToSend) {
			fPickList[i].eTime = fPickList[j-1].eTime;
			break;
		    }

		    midpoint = (fPickList[i].eTime + fPickList[j].sTime)/2;
		    fPickList[i].eTime = midpoint;
		    fPickList[j].sTime = midpoint;
		    i = j;
		}
	    }
	}

	/*
	 * For each frame chosen, create buffers with headers and data.
	 */
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	sbl->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *) *
					      numToSend);
	sbl->count = numToSend;
	sbIdx = 0;
	
	for (i=0; i<origNumToSend; i++) {
	    if (fPickList[i].sendFlag == 1) {
		segPtr = fPickList[i].segPtr;
		
		sTime = fPickList[i].sTime;
		eTime = fPickList[i].eTime;
		
		hdrPtr.hdrType = htonl(CM_MJPEG_BUF_HDR);
		hdrPtr.length = htonl(36);
		hdrPtr.width = htonl(objPtr->cfInfo->width);
		hdrPtr.height = htonl(objPtr->cfInfo->height);
		hdrPtr.device = htonl(objPtr->cfInfo->device);
		hdrPtr.subformat = htonl(objPtr->cfInfo->subformat);
		hdrPtr.qfactor = htonl(objPtr->cfInfo->qfactor);
		hdrPtr.logStart_s = htonl(floor(sTime));
		hdrPtr.logStart_ns = htonl(floor((sTime - floor(sTime))*ONE_BILLION));
		hdrPtr.logEnd_s = htonl(floor(eTime));
		hdrPtr.logEnd_ns = htonl(floor((eTime - floor(eTime))*ONE_BILLION));
		
		lseek(data_fd,
		      objPtr->dataStart+fPickList[i].offset, SEEK_SET);
		
		bufPtr = MakeBuffer(objPtr->interp,
				    sizeof(MJPEGBufferHeader)+8+fPickList[i].size,
				    "tcmMJPEGClipfile:MJPEG_CF_SendData.2");
		fPickList[i].bufPtr = bufPtr;
		idx = 0;
		

		memcpy(bufPtr->data+idx, &hdrPtr, sizeof(MJPEGBufferHeader));
		idx += sizeof(MJPEGBufferHeader);
		
		dataSegCode = htonl(CM_DATA_SEGMENT);
		dataLength = htonl(fPickList[i].size);
		
		memcpy(bufPtr->data+idx, &dataSegCode, 4);
		idx += 4;
		
		memcpy(bufPtr->data+idx, &dataLength, 4);
		idx += 4;
		
		read(data_fd, bufPtr->data+idx, fPickList[i].size); 
		idx += fPickList[i].size;
		
		sbl->sbs[sbIdx] = MakeScatterBuffer(bufPtr, idx, NULL);
		sbIdx++;
	    }
	}
	
	/*
	 * If out command is C, create scatter buffer list and call.
	 */
	if (objPtr->outCCmd != NULL) {
	    
	    
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree((char *) objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and
	     * call using Tcl_Eval. 
	     */
	    sbStr = MakeSBListString(sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree((char *) sbStr);
	}
	FreeSBList (sbl);
    }

    lastETime = 0;
    for (i=0; i<origNumToSend; i++) {

	if (fPickList[i].sendFlag == 1) {
/*	    lastETime = fPickList[i].eTime; */
	    DetachBuffer(objPtr->interp, fPickList[i].bufPtr,
			 "tcmMJPEGClipfile:MJPEG_CF_SendData");
	}
	lastETime = fPickList[i].eTime;
    }
  
    /*
     * Find next wake up time and set up atqueue.
     */
    when = LTS_SystemTimeOf(objPtr->lts,
		    (lastETime-(objPtr->sendAhead*speed)));
    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
			       "video", MJPEG_CF_SendData, objPtr, NULL,
			       NULL, NULL, NULL, NULL);
    return;

done:
    if (objPtr->cfState != NULL) {
	Chunk_CloseFile(objPtr->cfState);
	objPtr->cfState = NULL;
    } 
    if (objPtr->rawfd > -1) {
      close(objPtr->rawfd);
      objPtr->rawfd = -1;
    }
    
    return;
}

/************************************************************
 *
 * MJPEG_CF_SegmentList
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Returns list of current segment id's. If no segments,
 *     list is empty.
 ************************************************************/

static int
MJPEG_CF_SegmentList(MJPEG_CF *objPtr)
{
    Tcl_DString res_list;
    Segment *seg_ptr;
    char temp_str[100];

    seg_ptr = objPtr->segments;

    Tcl_DStringInit(&res_list);

    while (seg_ptr != NULL) {
	sprintf(temp_str, "%d", seg_ptr->segId);
    
	Tcl_DStringAppendElement(&res_list, temp_str);
	
	seg_ptr = seg_ptr->next;
    }

    Tcl_SetResult(objPtr->interp, Tcl_DStringValue(&res_list), TCL_VOLATILE);
    
    Tcl_DStringFree(&res_list);

    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_CF_ConfigureSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segmentId - id of segment to be reconfigured
 *     clipStart - new clip start.
 *     clipEnd - new clip end.
 *     logStart - new logical start.
 *     logEnd - new logical end.
 *     logStartFlag - indicates whetehr logStart value set or not.
 *     logEndFlag - indicates whether logEnd value set or not.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Reconfigures a previously created segment. Allows all
 *     aspects of the segment to be changed (i.e. clip start/end
 *     and logical start/end). Comparisons made with current
 *     segment info to decide whether change requires rebuilding
 *     tables or if current tables can be slightly altered.
 *     Segment removed from list and re-inserted to ensure proper
 *     ordering.
 *
 ************************************************************/

static int
MJPEG_CF_ConfigureSegment(MJPEG_CF *objPtr, int segmentId,
			 double clipStart, double clipEnd, 
			 double logStart, double logEnd, 
			 int logStartFlag, int logEndFlag)
{
  int new_seg_id;
  int result;
  Segment *segment_ptr;

  /*
   * This function could be optimized to recognize cases where
   * the clip only moves in logical time or the new clip is a 
   * proper subset of the original. In interest of getting this
   * function working as soon as possible, however, this code
   * currently just calls MJPEG_CF_AddSegment to create a new
   * segment with the new clip parameters, deletes the old
   * segment and changes the segmentId of the new segment to be
   * the same as the old segment.
   */

  /* Unready the object. */

  MJPEG_CF_Unready(objPtr);

  /* 
   * Find the original segment to be sure that it
   * existed.
   */

  segment_ptr = objPtr->segments;

  while (segment_ptr != NULL) {
    if (segment_ptr->segId == segmentId) 
      break;
    segment_ptr = segment_ptr->next;
  }

  if (segment_ptr == NULL) {
    char tempStr[40];

    sprintf(tempStr, "%d", segmentId);
    Tcl_AppendResult(objPtr->interp, "Segment ", tempStr, " does not exist.",
		     (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Remove the original segment.
   */

  MJPEG_CF_RemoveSegment(objPtr, segmentId);

  /* 
   * Create a segment with new parameters.
   */

  result = MJPEG_CF_AddSegment(objPtr, clipStart, clipEnd, logStart,
			       logEnd, logStartFlag, logEndFlag);

  /*
   * Change segment id of new segment to be segment id of original segment.
   */

  if (result == TCL_OK) {
    char resStr[40];

    if (Tcl_GetInt(objPtr->interp, objPtr->interp->result, &new_seg_id) != TCL_OK) {
      Tcl_AppendResult(objPtr->interp, "Couldn't reconfigure segment.", (char *) NULL);
      return TCL_ERROR;
    }

    segment_ptr = objPtr->segments;
    while (segment_ptr != NULL) {
      if (segment_ptr->segId == new_seg_id) {
	break;
      }
      segment_ptr = segment_ptr->next;
    }

    if (segment_ptr == NULL) {
      Tcl_AppendResult(objPtr->interp, "Couldn't reconfigure segment.", (char *) NULL);
      return TCL_ERROR;
    }

    segment_ptr->segId = segmentId;

    sprintf(resStr, "%d", segmentId);
    Tcl_SetResult(objPtr->interp, resStr, TCL_VOLATILE);

    /* 
     * Set segment list change flag so that next ready will check for overlaps.
     */

    objPtr->segmentListChangeFlag = 1;

    return TCL_OK;
  }

  return TCL_ERROR;
}
