/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMJPEGFileInterface.c,v 1.8 1996/02/20 23:00:38 aswan Exp $
 *
 * Contains the code for the MJPEG File Interface module
 * used by the MJPEGSegment object.
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <tcl.h>
#include <math.h>

#include "cmMJPEGFileInterface.h"

#ifndef O_BINARY
#    define O_BINARY 0
#endif

/*
 * Valid values for the file type. 
 */

#define UNKNOWN_MJPEG_FILE_TYPE 0
#define MJPEG_CHUNK_FILE_TYPE 1
#define MJPEG_VID_FILE_TYPE 2
#define MJPEG_AVI_FILE_TYPE 4
#define MJPEG_JMOVIE_FILE_TYPE 8

/*
 * Error codes.
 */

#define NO_FOT_ERROR 1
#define NO_FTT_ERROR 2
#define AVI_MISREAD_ERROR 4

/*
 * Misc. macros. 
 */

#define ONE_BILLION 1000000000.0
#define CF_READ_BUFFER_SIZE 2000
#define NUM_FRAME_TO_PARSE_PER_IDLE 1000

/* 
 * Initialization flag indicates whether hash table has been set up.
 */

static moduleInitFlag = 0;
static Tcl_HashTable fileHashTable;

/* 
 * Forward declarations. 
 */

static int MJPEG_FIP_FillBasicInfo(MJPEGDataFile *dataFile);
static int MJPEG_FIP_VidBasicInfo(MJPEGDataFile *dataFile);
static int MJPEG_FIP_AVIBasicInfo(MJPEGDataFile *dataFile);
static int MJPEG_FIP_JMovieBasicInfo(MJPEGDataFile *dataFile);
static void MJPEG_FIP_StartParse(MJPEGDataFile *dataFile);
static void MJPEG_FIP_ChunkFileStartParse(MJPEGDataFile *dataFile);
static void MJPEG_FIP_ChunkFileParse(MJPEGDataFile *dataFile);
static void MJPEG_FIP_VidFileStartParse(MJPEGDataFile *dataFile);
static void MJPEG_FIP_VidFileParse(MJPEGDataFile *dataFile);
static int MJPEG_FIP_AVIFileParse(MJPEGDataFile *dataFile);
static int MJPEG_FIP_AVIFileStartParse(MJPEGDataFile *dataFile);
static void MJPEG_FIP_JMovieFileStartParse(MJPEGDataFile *dataFile);
static void MJPEG_FIP_JMovieFileParse(MJPEGDataFile *dataFile);
static void MJPEG_FIP_ContinueParse(MJPEGDataFile *dataFile);


/*********************************************************************
 * MJPEG_FIP_ModuleInit
 *
 * Parameters:
 *    None
 *
 * Results:
 *    None
 * 
 * Action:
 *    Initializes MJPEG File Interface module by setting up hash
 *    table and setting init flag.
 *********************************************************************/
static void
MJPEG_FIP_ModuleInit()
{
    moduleInitFlag = 1;

    Tcl_InitHashTable(&fileHashTable, TCL_STRING_KEYS);
}


/********************************************************************
 * MJPEG_FIP_Open
 *
 * Parameters:
 *    fname - filename for desired file.
 *
 * Results:
 *    Pointer to MJPEGFileInterface structure.
 * 
 * Action:
 *    Finds underlying file data structure in hash table, creating
 *    it if not found. If the file data structure had to be created
 *    table parsing is started. Creates a file interface structure 
 *    for the segment object and returns it.
 *********************************************************************/

MJPEGFileInterface *
MJPEG_FIP_Open(char *fname)
{
    MJPEGFileInterface *new_fip;
    MJPEGDataFile *dataFile;
    Tcl_HashEntry *entry_ptr;
    int newFile;

    /* 
     *  Find data file structure if it already exists. 
     */

    if (!moduleInitFlag) {
	MJPEG_FIP_ModuleInit();
    }

    entry_ptr = Tcl_CreateHashEntry(&fileHashTable, fname, &newFile);

    if (newFile) {

	/* 
         * Didn't find one, so create one. 
	 */

	dataFile = (MJPEGDataFile *) ckalloc(sizeof(MJPEGDataFile));
	dataFile->filename = (char *) ckalloc(strlen(fname)+1);
	strcpy(dataFile->filename, fname);
	dataFile->cfState = NULL;
	dataFile->rawfd = -1;
	dataFile->file_refCount = 1;
	dataFile->open_refCount = 0;
	dataFile->FOT = NULL;
	dataFile->FTT = NULL;
	dataFile->inParse = 0;
	dataFile->numFrames = 0;
	dataFile->clipLength = 0.0;
	dataFile->dataStart = 0;
	dataFile->cfInfo = NULL;
	dataFile->fileType = UNKNOWN_MJPEG_FILE_TYPE;
	dataFile->fatalError = 0;
	dataFile->pinfo = (ParseInfo *) ckalloc(sizeof(ParseInfo));

	/* 
	 * Get basic info (i.e., width, height, etc.)
	 */

	if (MJPEG_FIP_FillBasicInfo(dataFile) != TCL_OK) {
	    Tcl_DeleteHashEntry(entry_ptr);
	    ckfree(dataFile->filename);
	    ckfree((char*) dataFile->pinfo);
	    ckfree((char *) dataFile);
	    return NULL;
	}
	
	/* 
	 * Begin parsing for offset and time tables.
	 */

	MJPEG_FIP_StartParse(dataFile);

	Tcl_SetHashValue(entry_ptr, (ClientData) dataFile);
    } else {
	/* 
	 * Found one that already existed.
	 */
	dataFile = (MJPEGDataFile *) Tcl_GetHashValue(entry_ptr);
	dataFile->file_refCount++;
    }

    /* 
     * Create file interface structure for segment, fill and return. 
     */

    new_fip = (MJPEGFileInterface *) ckalloc(sizeof(MJPEGFileInterface));
    
    new_fip->data_file = dataFile;
    new_fip->is_open = 0;

    return new_fip;
}


/********************************************************************
 * MJPEG_FIP_FDOpen
 *
 * Parameters:
 *    fip - Pointer to file interface structure.
 *
 * Results:
 *    None
 * 
 * Action:
 *    Used by segment object to indicate that it would like the
 *    file to have been "opened". File may or may not already be
 *    open. Reference count needs to be maintained.
 *********************************************************************/

void
MJPEG_FIP_FDOpen(MJPEGFileInterface *fip)
{
    MJPEGDataFile *dataFile;

    /* 
     * If file is already open in context to this segment object,
     * do nothing. This prevents the reference count for increasing
     * if a segment object does more than one open call.
     */

    if (fip->is_open == 1) {
	return;
    }

    /*
     * Set open flag in file interface structure and advance
     * reference count.
     */

    fip->is_open = 1;

    dataFile = fip->data_file;
    dataFile->open_refCount++;

    /*
     * Make sure file is actually open.
     */

    if (dataFile->fileType == MJPEG_CHUNK_FILE_TYPE) {
	if (dataFile->cfState == NULL) {
	    dataFile->cfState = Chunk_OpenFile(dataFile->filename, "r");
	}
    } else {
	if (dataFile->rawfd < 0) {
	    dataFile->rawfd = open(dataFile->filename,O_RDONLY|O_BINARY);
	}
    }
}


/********************************************************************
 * MJPEG_FIP_FDClose
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    None
 * 
 * Action:
 *    Used by segment object to close file. Doesn't actually happen
 *    unless all segment objects that share this file close.
 *********************************************************************/
void
MJPEG_FIP_FDClose(MJPEGFileInterface *fip) 
{
    MJPEGDataFile *dataFile;

    /*
     * Check for open flag. Maintain reference count.
     */

    if (fip->is_open == 0) {
	return;
    } 

    fip->is_open = 0;
    dataFile = fip->data_file;
    dataFile->open_refCount--;

    /*
     * If reference count falls to zero, actually do the close.
     */

    if (dataFile->open_refCount <= 0) {
	if (dataFile->cfState != NULL) {
	    Chunk_CloseFile(dataFile->cfState);
	    dataFile->cfState = NULL;
	} 
	if (dataFile->rawfd != -1) {
	    close(dataFile->rawfd);
	    dataFile->rawfd = -1;
	}
	dataFile->open_refCount = 0;
    }
}


/********************************************************************
 * MJPEG_FIP_Close
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    None
 * 
 * Action:
 *    Called by segment object to free file interface structure.
 *    Indicates that segment object is now totally divorced from
 *    this file.
 *********************************************************************/

void
MJPEG_FIP_Close(MJPEGFileInterface *fip) 
{
    Tcl_HashEntry *entry_ptr;
    MJPEGDataFile *dataFile;

    /* 
     * Close the file.
     */

    MJPEG_FIP_FDClose(fip);

    dataFile = fip->data_file;

    /*
     * Decrement file structure reference count, free if falls to 0.
     */

    dataFile->file_refCount--;
    ckfree((char *) fip);

    if (dataFile->file_refCount <= 0) {
	entry_ptr = Tcl_FindHashEntry(&fileHashTable, dataFile->filename);

	if (entry_ptr != NULL) {
	    Tcl_DeleteHashEntry(entry_ptr);
	}

	/*
	 * Cancel any scheduled parsing action.
	 */

	if (dataFile->inParse) {
	    Tcm_CancelIdleCall("tk", MJPEG_FIP_ContinueParse, dataFile);
	}

	if (dataFile->filename != NULL) {
	    ckfree(dataFile->filename);
	}
	if (dataFile->FOT != NULL) {
	    ckfree((char *) dataFile->FOT);
	}
	if (dataFile->FTT != NULL) {
	    ckfree((char *) dataFile->FTT);
	}
	if (dataFile->cfInfo != NULL) {
	    ckfree((char *) dataFile->cfInfo);
	}
	if (dataFile->pinfo != NULL) {
	    ckfree((char *) dataFile->pinfo);
	}
	ckfree((char *) dataFile);
    }
}


/********************************************************************
 * MJPEG_FIP_Subformat
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer subformat code.
 * 
 * Action:
 *    Used by segment object to access subformat information from file.
 *********************************************************************/	

int
MJPEG_FIP_Subformat(MJPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->subformat;
}


/********************************************************************
 * MJPEG_FIP_Device
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer device code.
 * 
 * Action:
 *    Used by segment object to access device information from file.
 *********************************************************************/	
int 
MJPEG_FIP_Device(MJPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->device;
}


/********************************************************************
 * MJPEG_FIP_Width
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer width value.
 * 
 * Action:
 *    Used by segment object to access width information from file.
 *********************************************************************/	

int
MJPEG_FIP_Width(MJPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->width;
}


/********************************************************************
 * MJPEG_FIP_Height
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer height value.
 * 
 * Action:
 *    Used by segment object to access height information from file.
 *********************************************************************/	

int
MJPEG_FIP_Height(MJPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->height;
}


/********************************************************************
 * MJPEG_FIP_QFactor
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer qfactor value.
 * 
 * Action:
 *    Used by segment object to access qfactor information from file.
 *********************************************************************/	

int
MJPEG_FIP_QFactor(MJPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->qfactor;
}


/********************************************************************
 * MJPEG_FIP_ClipLength
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Double clip length value.
 * 
 * Action:
 *    Used by segment object to access cliplength information from file.
 *********************************************************************/	

double
MJPEG_FIP_ClipLength(MJPEGFileInterface *fip) {
    MJPEGDataFile *dataFile;

    dataFile = fip->data_file;

    if (dataFile->clipLength == 0.0) {
	if (dataFile->inParse) {
	    Tcm_CancelIdleCall("tk", MJPEG_FIP_ContinueParse, dataFile);
	}

	while(dataFile->inParse) {
	    MJPEG_FIP_ContinueParse(dataFile);
	}
    }

    return dataFile->clipLength;
}


/********************************************************************
 * MJPEG_FIP_FrameFromTime
 *
 * Parameters:
 *    fip - file interface pointer
 *    time - clip time value
 *    speed - lts speed value (affects boundary conditions)
 *
 * Results:
 *    Integer frame number.
 * 
 * Action:
 *    Used by segment object to map frame number to clip time.
 *********************************************************************/	

int
MJPEG_FIP_FrameFromTime(MJPEGFileInterface *fip, double time, double speed) {
    MJPEGDataFile *dataFile;
    int low_frame, high_frame, mid_frame;

    dataFile = fip->data_file;
    
    if (time < 0.0) {
	return 0;
    }

    if ((dataFile->clipLength != 0.0) && (time > dataFile->clipLength)) {
	return dataFile->numFrames-1;
    }

    if (dataFile->inParse) {
	if (dataFile->FTT[dataFile->pinfo->highFrameNum] < time) {
	    Tcm_CancelIdleCall("tk", MJPEG_FIP_ContinueParse, dataFile);

	    while (dataFile->inParse) {
		MJPEG_FIP_ContinueParse(dataFile);
	    }

	    if (time > dataFile->clipLength) {
		return dataFile->numFrames-1;
	    }
	    high_frame = dataFile->numFrames;
	} else {
	    high_frame = dataFile->pinfo->highFrameNum;
	}
    } else {
	high_frame = dataFile->numFrames;
    }

    low_frame = 0;

    if (time <= dataFile->FTT[low_frame]) {
	return low_frame;
    }

    if (time >= dataFile->FTT[high_frame]) {
	return high_frame-1;
    }

    while (high_frame-low_frame > 1) {
	mid_frame = (high_frame+low_frame) / 2;
	
	if (time < dataFile->FTT[mid_frame]) {
	    high_frame = mid_frame;
	} else if (time > dataFile->FTT[mid_frame]) {
	    low_frame = mid_frame;
	} else {
	    if (speed >= 0.0) {
		return mid_frame;
	    } else {
		return mid_frame-1;
	    }
	}
    }

    return low_frame;
}


/********************************************************************
 * MJPEG_FIP_TimeFromFrame
 *
 * Parameters:
 *    fip - file interface pointer
 *    frameNum - frame number
 *    stime - pointer to double. Set to start time of frame.
 *    etime - pointer to double. Set to end time of frame.
 *
 * Results:
 *    None
 * 
 * Action:
 *    Used by segment object to map clip time to frame number.
 *********************************************************************/

void
MJPEG_FIP_TimeFromFrame(MJPEGFileInterface *fip, int frameNum, double *stime,
			double *etime)
 {
    MJPEGDataFile *dataFile;

    dataFile = fip->data_file;

    if (frameNum < 0) frameNum = 0;

    if (dataFile->inParse) {
	if (frameNum > dataFile->pinfo->highFrameNum) {
	    Tcm_CancelIdleCall("tk", MJPEG_FIP_ContinueParse, dataFile);
	    
	    while(dataFile->inParse) {
		MJPEG_FIP_ContinueParse(dataFile);
	    }
	    if (frameNum >= dataFile->numFrames) {
		frameNum = dataFile->numFrames-1;
	    }
	}
    } else if (frameNum >= dataFile->numFrames) {
	frameNum = dataFile->numFrames-1;
    }
	

    *stime = dataFile->FTT[frameNum];
    *etime = dataFile->FTT[frameNum+1];

    return;
}


/********************************************************************
 * MJPEG_FIP_LengthFromFrame
 *
 * Parameters:
 *    fip - file interface pointer
 *    frameNum - integer frame number.
 *
 * Results:
 *    Integer length of frame in bytes.
 *
 * Action:
 *    Used by segment object to access the size of a particular frame.
 *********************************************************************/	

int
MJPEG_FIP_LengthFromFrame(MJPEGFileInterface *fip, int frameNum)
{
    MJPEGDataFile *dataFile;

    dataFile = fip->data_file;

    if (frameNum < 0) frameNum = 0;

    if (dataFile->inParse) {
	if (frameNum > dataFile->pinfo->highFrameNum) {
	    Tcm_CancelIdleCall("tk", MJPEG_FIP_ContinueParse, dataFile);
	    
	    while(dataFile->inParse) {
		MJPEG_FIP_ContinueParse(dataFile);
	    }
	    if (frameNum >= dataFile->numFrames) {
		frameNum = dataFile->numFrames-1;
	    }
	}
    } else if (frameNum >= dataFile->numFrames) {
	frameNum = dataFile->numFrames-1;
    }
	
    return dataFile->FOT[frameNum*2+1];
}


/********************************************************************
 * MJPEG_FIP_ReadFrame
 *
 * Parameters:
 *    fip - file interface pointer
 *    buffer - pointer to memory to read frame into.
 *    frameNum - integer frame number of frame to be read.
 *
 * Results:
 *    Integer number of bytes read.
 * 
 * Action:
 *    Used by segment object to read data for a particular frame out
 *    of file into a buffer.
 *********************************************************************/	

int
MJPEG_FIP_ReadFrame(MJPEGFileInterface *fip, char *buffer, int frameNum)
{
    MJPEGDataFile *dataFile;
    int close_file_flag;
    int fd;
    int num_read;

    dataFile = fip->data_file;
    
    if (frameNum < 0) frameNum = 0;

    if (dataFile->inParse) {
	if (frameNum > dataFile->pinfo->highFrameNum) {
	    Tcm_CancelIdleCall("tk", MJPEG_FIP_ContinueParse, dataFile);
	    
	    while(dataFile->inParse) {
		MJPEG_FIP_ContinueParse(dataFile);
	    }
	    if (frameNum >= dataFile->numFrames) {
		frameNum = dataFile->numFrames-1;
	    }
	}
    } else if (frameNum >= dataFile->numFrames) {
	frameNum = dataFile->numFrames-1;
    }

    close_file_flag = 0;
    if (dataFile->fileType == MJPEG_CHUNK_FILE_TYPE) {
	if (dataFile->cfState == NULL) {
	    close_file_flag = 1;
	    dataFile->cfState = Chunk_OpenFile(dataFile->filename, "r");
	}
	fd = dataFile->cfState->fd;
    } else {
	if (dataFile->rawfd < 0) {
	    close_file_flag = 1;
	    dataFile->rawfd = open(dataFile->filename, O_RDONLY|O_BINARY);
	}
	fd = dataFile->rawfd;
    }

    lseek(fd, dataFile->dataStart + dataFile->FOT[frameNum*2], SEEK_SET);
    num_read = read(fd, buffer, dataFile->FOT[frameNum*2+1]);

    if (close_file_flag) {
	if (dataFile->cfState != NULL) {
	    Chunk_CloseFile(dataFile->cfState);
	    dataFile->cfState = NULL;
	} 
	if (dataFile->rawfd >= 0) {
	    close(dataFile->rawfd);
	    dataFile->rawfd = -1;
	}
    }

    return num_read;
}


/********************************************************************
 * MJPEG_FIP_FrameListFromTime
 *
 * Parameters:
 *    fip - file interface pointer
 *    start - double clip time value for start of interval.
 *    end - double clip time value for end of interval.
 *    frameList - pointer to integer pointer. Set to beginning of
 *                array of integer frame numbers corresponding to
 *                frame within specified interval.
 *
 * Results:
 *    Integer number of frame numbers in constructed frame list.
 * 
 * Action:
 *    Used by segment object to retrieve a list of frame numbers that
 *    correspond to the requested interval. The list of frame numbers
 *    should be free'd by the segment object.
 *********************************************************************/	

int
MJPEG_FIP_FrameListFromTime(MJPEGFileInterface *fip, double start, double end,
			    int **frameList)
{
    int start_frame, end_frame;
    int index, cntr;

    start_frame = MJPEG_FIP_FrameFromTime(fip, start, 0.0);
    end_frame = MJPEG_FIP_FrameFromTime(fip, end, 0.0);

    if (start_frame == end_frame) {
	*frameList = (int *) ckalloc(sizeof(int));
	(*frameList)[0] = start_frame;
	return 1;
    }

    if (end_frame < start_frame) {
	/* Speed is less than 0. */
	*frameList = (int *) ckalloc(sizeof(int)*(start_frame-end_frame));
	for (index=0,cntr = start_frame-1; cntr >= end_frame; cntr--,index++) {
	    (*frameList)[index] = cntr;
	}
	return index;
    } else {
	*frameList = (int *) ckalloc(sizeof(int*)*(end_frame-start_frame));
	for (index=0,cntr=start_frame; cntr < end_frame; cntr++,index++) {
	    (*frameList)[index] = cntr;
	}
	return index;
    }
}


/********************************************************************
 * MJPEG_FIP_Info
 *
 * Parameters:
 *    fip - file interface pointer
 *    start - double clip time value for start of interval.
 *    end - double clip time value for end of interval.
 *    numFrames - pointer to integer. Value of integer will be set to
 *                number of frames in the requested interval.
 *    maxSize - pointer to integer. Value of integer will be set to
 *              maximum frame size of frames in the requested interval.
 *
 * Results:
 *    None
 * 
 * Action:
 *    Used by segment object to access information about frames in
 *    particular interval of the clip.
 *********************************************************************/	

void
MJPEG_FIP_Info(MJPEGFileInterface *fip, double start, double end,
	       int *numFrames, int *maxSize)
{
    MJPEGDataFile *dataFile;
    double tmp;
    int start_frame, end_frame;
    int max_size, index;

    dataFile = fip->data_file;

    if (start > end) {
	tmp = start;
	start = end;
	end = tmp;
    }

    start_frame = MJPEG_FIP_FrameFromTime(fip, start, 0.0);

    if (start == end) {
	*numFrames = 1;
	*maxSize = dataFile->FOT[start_frame*2+1];
	return;
    }

    end_frame = MJPEG_FIP_FrameFromTime(fip, end, 0.0);

    *numFrames = (end_frame-start_frame+1);
    
    max_size = dataFile->FOT[start_frame*2+1];

    for (index = start_frame+1; index <= end_frame; index++) {
	if (dataFile->FOT[index*2+1] > max_size) 
	  max_size = dataFile->FOT[index*2+1];
    }

    *maxSize = max_size;
    return;
}


/********************************************************************
 * MJPEG_FIP_FillBasicInfo
 *
 * Parameters:
 *    dataFile - file data structure
 *
 * Results:
 *    Integer Tcl return code.
 * 
 * Action:
 *    Called when file is first opened to fill in basic information
 *    fields in the file data structure. 
 *********************************************************************/	

static int
MJPEG_FIP_FillBasicInfo(MJPEGDataFile *dataFile)
{
    tChunkHeader chunk_hdr;
    MediaIndepHeader mi_hdr;
    
    dataFile->cfState = Chunk_OpenFile(dataFile->filename, "r");

    if (dataFile->cfState == NULL) {

	/* 
	 * File doesn't seem to be chunk file, try raw file types.
	 */

	int result;

	dataFile->rawfd = open(dataFile->filename,O_RDONLY|O_BINARY);

	if (dataFile->rawfd < 0) {
	    dataFile->rawfd = -1;
	    return TCL_ERROR;
	}

	if (MJPEG_FIP_AVIBasicInfo(dataFile) == TCL_OK) {
	    close(dataFile->rawfd);
	    result = TCL_OK;
	    goto raw_done;
	} 
	lseek(dataFile->rawfd, 0, SEEK_SET);
	if (MJPEG_FIP_VidBasicInfo(dataFile) == TCL_OK) {
	    result = TCL_OK;
	    goto raw_done;
	} 
	lseek(dataFile->rawfd, 0, SEEK_SET);
	if (MJPEG_FIP_JMovieBasicInfo(dataFile) == TCL_OK) {
	    result = TCL_OK;
	    goto raw_done;
	} 

	/* 
	 * File doesn't seem to be any type that we know about,
	 * return error.
	 */

	result = TCL_ERROR;

      raw_done:
	close(dataFile->rawfd);
	dataFile->rawfd = -1;
	return result;
    }
    
    /*
     * Get information from media independent header. 
     */

    if (Chunk_FindType(dataFile->cfState, CM_MEDIA_INDEP_CF, &chunk_hdr) != 0) {
	Chunk_CloseFile(dataFile->cfState);
	dataFile->cfState = NULL;
	return TCL_ERROR;
    }

    read(dataFile->cfState->fd, &mi_hdr, sizeof(mi_hdr));

    mi_hdr.media = ntohl(mi_hdr.media);
    mi_hdr.format = ntohl(mi_hdr.format);
    mi_hdr.device = ntohl(mi_hdr.device);
    mi_hdr.version = ntohl(mi_hdr.version);
    mi_hdr.num_frames = ntohl(mi_hdr.num_frames);
    mi_hdr.length_s = ntohl(mi_hdr.length_s);
    mi_hdr.length_ns = ntohl(mi_hdr.length_ns);

    /*
     * Sanity check on media and format type.
     */

    if ((mi_hdr.media != VIDEO_MEDIA_CODE) ||
	(mi_hdr.format != MJPEG_FORMAT_CODE)) {
	Chunk_CloseFile(dataFile->cfState);
	dataFile->cfState == NULL;
	return TCL_ERROR;
    }

    dataFile->numFrames = mi_hdr.num_frames;
    dataFile->clipLength = mi_hdr.length_s + mi_hdr.length_ns/1000000000.0;

    /*
     * Get information from mjpeg clipfile header.
     */

    if (Chunk_FindType(dataFile->cfState, CM_MJPEG_CF, &chunk_hdr) != 0) {
	Chunk_ReadRewind(dataFile->cfState);
	if (Chunk_FindType(dataFile->cfState, CM_MJPEG_CF, &chunk_hdr) != 0) {
	    Chunk_CloseFile(dataFile->cfState);
	    dataFile->cfState == NULL;
	    return TCL_ERROR;
	}
    }

    if (dataFile->cfInfo != NULL) {
	ckfree((char *) dataFile->cfInfo);;
    }
    dataFile->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));

    read(dataFile->cfState->fd, dataFile->cfInfo, sizeof(MJPEGHeader));

    dataFile->cfInfo->width = ntohl(dataFile->cfInfo->width);
    dataFile->cfInfo->height = ntohl(dataFile->cfInfo->height);
    dataFile->cfInfo->device = ntohl(dataFile->cfInfo->device);
    dataFile->cfInfo->subformat = ntohl(dataFile->cfInfo->subformat);
    dataFile->cfInfo->qfactor = ntohl(dataFile->cfInfo->qfactor);
    
    /*
     * Find data segment and store location in file.
     */
    
    if (Chunk_FindType(dataFile->cfState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
	Chunk_ReadRewind(dataFile->cfState);
	if (Chunk_FindType(dataFile->cfState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
	    Chunk_CloseFile(dataFile->cfState);
	    dataFile->cfState == NULL;
	    ckfree((char *) dataFile->cfInfo);
	    dataFile->cfInfo = NULL;
	    return TCL_ERROR;
	}
    }
    
    dataFile->dataStart = lseek(dataFile->cfState->fd, 0, SEEK_CUR);
    dataFile->fileType = MJPEG_CHUNK_FILE_TYPE;

    Chunk_CloseFile(dataFile->cfState);
    dataFile->cfState = NULL;
    return TCL_OK;
}


/********************************************************************
 * MJPEG_FIP_VidBasicInfo
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    Integer Tcl return code.
 * 
 * Action:
 *    Used by FillBasicInfo in order to open the file as a .vid file.
 *    If successful, basic information parsed and fields in data structure
 *    set as appropriate. Parsing information squirreled away in 
 *    parsing structure.
 *********************************************************************/	
    
static int
MJPEG_FIP_VidBasicInfo(MJPEGDataFile *dataFile)
{
 int width, height, size;
    int width2, height2;
    int num_tests;

    /* Assume dataFile->rawfd is open and rewound. */

    read(dataFile->rawfd, &width, 4);
    width = ntohl(width);

    read(dataFile->rawfd, &height, 4);
    height = ntohl(height);

    read(dataFile->rawfd, &size, 4);
    size = ntohl(size);

    if ((width <= 0) || (height <= 0) || (size <= 0)) {
	return TCL_ERROR;
    }
    
    for (num_tests=0; num_tests<4; num_tests++) {
	lseek(dataFile->rawfd, 8+size, SEEK_CUR);

	read(dataFile->rawfd, &width2, 4);
	width2 = ntohl(width2);
	
	read(dataFile->rawfd, &height2, 4);
	height2 = ntohl(height2);
	
	read(dataFile->rawfd, &size, 4);
	size = ntohl(size);
	
	if (width != width2) return TCL_ERROR;
	if (height != height2) return TCL_ERROR;
    }
    
    /* Seems OK. */
    
    if (dataFile->cfInfo != NULL) {
	ckfree((char *) dataFile->cfInfo);
    }
    dataFile->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));
    
    dataFile->cfInfo->width = width;
    dataFile->cfInfo->height = height;
    dataFile->cfInfo->device = PARALLAX_DEV_CODE;
    dataFile->cfInfo->subformat = VID_FRAME_CODE;
    dataFile->cfInfo->qfactor = 150;
    
    dataFile->dataStart = 0;
    dataFile->fileType = MJPEG_VID_FILE_TYPE;
    
    return TCL_OK;
}


/************************************************************
 *
 * GetInt32FromAVI
 *
 * Parameters:
 *     buffer - pointer to data read from AVI file.
 *     buf_index - index where byte inverted AVI 32 bit integer
 *                 starts in buffer.
 *
 * Return:
 *     32-bit int value calculated from buf_index through buf_index+3
 *     bytes of buffer.
 *
 * Action:
 *     Since integers in avi files are LSB order, this function calculates
 *     the integer value constructed from the bytes buf_index to buf_index+3
 *     of the buffer passed in. This function is used as a helper function
 *     of MJPEG_CF_ReadAsAVI.
 *
 ************************************************************/

static unsigned int
GetInt32FromAVI(unsigned char *buffer,     /* Pointer to actual bytes. */
		int buf_index     /* Index where 32-bit value starts. */)
{
  unsigned int result = 0;

  result = (((unsigned int) buffer[buf_index]) | 
	    (((unsigned int) (buffer[buf_index+1])) << 8) |
	    (((unsigned int) (buffer[buf_index+2])) << 16) | 
	    (((unsigned int) (buffer[buf_index+3])) << 24));

  return result;
}

#define FIND_AVI_LIST(tchar1, tchar2, tchar3, tchar4)                         \
  while ((!((buffer[buf_index] == 'L') && (buffer[buf_index+1] == 'I') &&     \
	  (buffer[buf_index+2] == 'S') && (buffer[buf_index+3] == 'T'))) ||   \
	 (!((buffer[buf_index+8] == tchar1) && (buffer[buf_index+9] == tchar2) && \
	    (buffer[buf_index+10] == tchar3) &&                               \
	    (buffer[buf_index+11] == tchar4)))) {                             \
    buf_index += 4;                                                           \
    junk_size = GetInt32FromAVI(buffer, buf_index);                           \
    buf_index += 4 + junk_size;                                               \
                                                                              \
    if (buf_index > num_read - 12) {                                          \
      lseek(dataFile->rawfd, buf_index+file_index, SEEK_SET);                   \
      file_index += buf_index;                                                \
      buf_index = 0;                                                          \
      num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);            \
      if (num_read < 12) return TCL_ERROR;                                          \
    }                                                                         \
  }

#define FIND_AVI_SUBCHUNK(tchar1, tchar2, tchar3, tchar4)                     \
  while (!((buffer[buf_index] == tchar1) && (buffer[buf_index+1] == tchar2) &&\
	    (buffer[buf_index+2] == tchar3) &&                                \
	    (buffer[buf_index+3] == tchar4))) {                               \
    buf_index += 4;                                                           \
    junk_size = GetInt32FromAVI(buffer, buf_index);                           \
    buf_index += 4 + junk_size;                                               \
                                                                              \
    if (buf_index > num_read - 8) {                                           \
      lseek(dataFile->rawfd, buf_index+file_index, SEEK_SET);                   \
      file_index += buf_index;                                                \
      buf_index = 0;                                                          \
      num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);            \
      if (num_read < 8) return TCL_ERROR;                                           \
    }                                                                         \
  }


/********************************************************************
 * MJPEG_FIP_AVIBasicInfo
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    Integer Tcl return code.
 * 
 * Action:
 *    Used by FillBasicInfo in order to open the file as a .avi file.
 *    If successful, basic information parsed and fields in data structure
 *    set as appropriate. Parsing information squirreled away in parsing
 *    structre.
 *********************************************************************/	

static int
MJPEG_FIP_AVIBasicInfo(MJPEGDataFile *dataFile)
 {
    unsigned char buffer[CF_READ_BUFFER_SIZE];
    int num_read;
    int file_index, buf_index, avi_success, junk_size;
    int main_avi_hdr_size, strl_size, strl_start, stream_num;
    int found_stream, stream_length, stream_rate, stream_scale;
    int stream_width, stream_height, stream_header_size;
    int strf_header_size;

    num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
    file_index = 0;
    buf_index = 0;

    if (num_read < 24) {
	return TCL_ERROR;
    }
	 
    avi_success = 0;
  
    while (buf_index < num_read-24) {
	if ((buffer[buf_index] == 'R') && (buffer[buf_index+1] == 'I') &&
	    (buffer[buf_index+2] == 'F') && (buffer[buf_index+3] == 'F') &&
	    (buffer[buf_index+8] == 'A') && (buffer[buf_index+9] == 'V') &&
	    (buffer[buf_index+10] == 'I') && (buffer[buf_index+11] == ' ')) {
	    avi_success = 1;
	    break;
	}
	buf_index++;
    }

    if (!avi_success) return TCL_ERROR;

    /*
     * Now we know its an avi, lets make sure that there is a mjpeg stream 
     * in the file and find the stream number.
     */

    buf_index += 12;

    FIND_AVI_LIST('h', 'd', 'r', 'l');
  
    buf_index += 12;

    if (buf_index > num_read-8) {
	lseek(dataFile->rawfd, buf_index+file_index, SEEK_SET);
	file_index += buf_index;
	buf_index = 0;
	num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
	if (num_read < 8) {
	    return TCL_ERROR;
	}
    }

    FIND_AVI_SUBCHUNK('a', 'v', 'i', 'h');
    buf_index += 4;

    main_avi_hdr_size = GetInt32FromAVI(buffer, buf_index);
    buf_index += 4;

    buf_index += main_avi_hdr_size;
    
    if (buf_index > num_read-8) {
	lseek(dataFile->rawfd, file_index+buf_index, SEEK_SET);
	file_index += buf_index;
	buf_index = 0;
	num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
	if (num_read < 8) {
	    return TCL_ERROR;
	}
    }
    
    FIND_AVI_LIST('s', 't', 'r', 'l');
    
    buf_index += 4;
    
    strl_size = GetInt32FromAVI(buffer, buf_index);
    buf_index += 4;
    
    strl_start = file_index + buf_index;
    buf_index += 4;
    
    
    if (buf_index > num_read-8) {
	lseek(dataFile->rawfd, file_index+buf_index, SEEK_SET);
	file_index += buf_index;
	buf_index = 0;
	num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
	if (num_read < 8) {
	    return TCL_ERROR;
	}
    }
    
    stream_num = -1;
    found_stream = 0;
    
    stream_length = 0;
    stream_rate = stream_scale = 0;
    stream_width = stream_height = 0;
    while (1) {
	FIND_AVI_SUBCHUNK('s', 't', 'r', 'h');
	
	stream_num++;
	buf_index += 4;
	
	stream_header_size = GetInt32FromAVI(buffer, buf_index);
	
	buf_index += 4;
	
	if ((buffer[buf_index] == 'v') && (buffer[buf_index+1] == 'i') &&
	    (buffer[buf_index+2] == 'd') && (buffer[buf_index+3] == 's')) {
      
	    stream_scale = GetInt32FromAVI(buffer, buf_index+20);
	    stream_rate = GetInt32FromAVI(buffer, buf_index+24);
	    stream_length = GetInt32FromAVI(buffer, buf_index+32);
	    
	    buf_index += stream_header_size;
	    
	    if (buf_index > num_read - 8) {
		lseek(dataFile->rawfd, file_index+buf_index, SEEK_SET);
		file_index += buf_index;
		buf_index = 0;
		num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
		if (num_read < 8) {
		    return TCL_ERROR;
		}
	    }
	    
	    FIND_AVI_SUBCHUNK('s', 't', 'r', 'f');
	    
	    buf_index += 4;
	    
	    strf_header_size = GetInt32FromAVI(buffer, buf_index);
	    buf_index += 4;
	    
	    if ((buffer[buf_index+16] == 'M') && (buffer[buf_index+17] == 'J') &&
		(buffer[buf_index+18] == 'P') && (buffer[buf_index+19] == 'G')) {
		
		stream_width = GetInt32FromAVI(buffer, buf_index+4);
		stream_height = GetInt32FromAVI(buffer, buf_index+8);
		
		found_stream = 1;
	    }
	    
	    buf_index += strf_header_size;
    
	} else {
	    buf_index += stream_header_size;
	}

	if (found_stream) break;
    
	if (buf_index > num_read - 8) {
	    lseek(dataFile->rawfd, file_index+buf_index, SEEK_SET);
	    file_index += buf_index;
	    buf_index = 0;
	    num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
	    if (num_read < 8) {
		return TCL_ERROR;
	    }
	}
    }
    
    if (!found_stream) {
	return TCL_ERROR;
    }
    
    /*
     * Found a motion jpeg stream at this point as well as general things like:
     * width, height, how many frames, frame rate. Fill in cfInfo and return.
     */

    if (dataFile->cfInfo != NULL) {
	ckfree((char *) dataFile->cfInfo);
    }

    dataFile->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));

    dataFile->cfInfo->width = stream_width;
    dataFile->cfInfo->height = stream_height;
    dataFile->cfInfo->device = J300_DEV_CODE;
    dataFile->cfInfo->subformat = JFIF_NO_HUFFMAN_FRAME_CODE;
    dataFile->cfInfo->qfactor = 100;
    dataFile->dataStart = 0;

    dataFile->fileType = MJPEG_AVI_FILE_TYPE;

    /* Store away parsing information. */

    dataFile->pinfo->strl_start = strl_start;
    dataFile->pinfo->strl_size = strl_size;
    sprintf(dataFile->pinfo->stream_id, "%.2d", stream_num);
    dataFile->pinfo->tableSize = stream_length;
    dataFile->pinfo->stream_rate = stream_rate;
    dataFile->pinfo->stream_scale = stream_scale;

    return TCL_OK;
}


/********************************************************************
 * MJPEG_FIP_JMovieBasicInfo
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    Integer Tcl return code.
 * 
 * Action:
 *    Used by FillBasicInfo in order to open the file as a JMovie file.
 *    If successful, basic information parsed and fields in data structure
 *    set as appropriate. Parsing information squirreled away in parsing
 *    structure.
 *********************************************************************/	

static int
MJPEG_FIP_JMovieBasicInfo(MJPEGDataFile *dataFile)
{
    char magic_str[8];
    int fps, num_frames, width, height, qfactor;
    int indexbuf, au_slice;
    unsigned char j_movie_cmd;
    int cntr, raw_fot_frame_count;
    int skipped, frame_size;


    /* 
     * Confirm it is a j_movie file by looking for "j_movie" as first
     * 8 bytes.
     */

    read(dataFile->rawfd, magic_str, 8);
    
    if (strcmp(magic_str,"j_movie") != 0) {
	return TCL_ERROR;
    }

    /*
     * Get info out of jpheader. We do this without actually creating
     * a jpheader in order to avoid having to compile with the XVideo 
     * include files.
     */

    lseek(dataFile->rawfd, 4, SEEK_CUR);       /* Skip version number. */
    read(dataFile->rawfd, &fps, 4);
    fps = ntohl(fps);
    read(dataFile->rawfd, &num_frames, 4);
    num_frames = ntohl(num_frames);
    read(dataFile->rawfd, &width, 4);
    width = ntohl(width);
    read(dataFile->rawfd, &height, 4);
    height = ntohl(height);
    lseek(dataFile->rawfd, 4, SEEK_CUR);       /* Skip bandwidth. */
    read(dataFile->rawfd, &qfactor, 4);
    qfactor = ntohl(qfactor);
    lseek(dataFile->rawfd, 4, SEEK_CUR);       /* Skip mapsize. */
    read(dataFile->rawfd, &indexbuf, 4);
    indexbuf = ntohl(indexbuf); 
    lseek(dataFile->rawfd, 8, SEEK_CUR);       /* Skip tracks, volbase. */
    read(dataFile->rawfd, &au_slice, 4);

    if (dataFile->cfInfo != NULL) {
	ckfree((char *) dataFile->cfInfo);
    }
    dataFile->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));
    
    dataFile->cfInfo->width = width;
    dataFile->cfInfo->height = height;
    dataFile->cfInfo->device = PARALLAX_DEV_CODE;
    dataFile->cfInfo->subformat = RAW_JPEG_NO_HEADERS_CODE;
    dataFile->cfInfo->qfactor = qfactor;
    
    dataFile->dataStart = 0;

    dataFile->fileType = MJPEG_JMOVIE_FILE_TYPE;

    dataFile->numFrames = num_frames;
    dataFile->clipLength = ((double) num_frames) / ((double) fps);
    dataFile->pinfo->auSlice = au_slice;

    dataFile->pinfo->fps = fps;
    dataFile->pinfo->indexbuf = indexbuf;

    return TCL_OK;
}


/********************************************************************
 * MJPEG_FIP_StartParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Used by MJPEG_FIP_Open to start the background process of parsing
 *    a file for FOT and FTT tables. Essentially just switches to
 *    file type specific parse procedure.
 *********************************************************************/	

static void
MJPEG_FIP_StartParse(MJPEGDataFile *dataFile)
{
    dataFile->pinfo->highFrameNum = 0;
    dataFile->inParse = 1;

    switch(dataFile->fileType) {
      case MJPEG_CHUNK_FILE_TYPE:
	MJPEG_FIP_ChunkFileStartParse(dataFile);
	break;

      case MJPEG_VID_FILE_TYPE:
	MJPEG_FIP_VidFileStartParse(dataFile);
	break;

      case MJPEG_AVI_FILE_TYPE:
	MJPEG_FIP_AVIFileStartParse(dataFile);
	break;

      case MJPEG_JMOVIE_FILE_TYPE:
	MJPEG_FIP_JMovieFileStartParse(dataFile);
	break;
    }
}


/********************************************************************
 * MJPEG_FIP_ContinueParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Incremental parsing procedure that gets queued to parse file.
 *    Essentially just switches to file type specific parsing procedure.
 *********************************************************************/	

static void
MJPEG_FIP_ContinueParse(MJPEGDataFile *dataFile)
{
    /* 
     * Protects us from possibility that some object command
     * forced parsing to be finished.
     */

    if (!dataFile->inParse) return;

    switch(dataFile->fileType) {
      case MJPEG_CHUNK_FILE_TYPE:
	MJPEG_FIP_ChunkFileParse(dataFile);
	break;

      case MJPEG_VID_FILE_TYPE:
	MJPEG_FIP_VidFileParse(dataFile);
	break;

      case MJPEG_AVI_FILE_TYPE:
	MJPEG_FIP_AVIFileParse(dataFile);
	break;

      case MJPEG_JMOVIE_FILE_TYPE:
	MJPEG_FIP_JMovieFileParse(dataFile);
	break;
    }
}


/********************************************************************
 * MJPEG_FIP_JMovieFileStartParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    JMovie specific procedure to start parsing process for JMovie
 *    file. Allocates tables and then calls file type specific parse
 *    procedure.
 *********************************************************************/	

static void
MJPEG_FIP_JMovieFileStartParse(MJPEGDataFile *dataFile)
{
    dataFile->FOT = (int *) ckalloc(sizeof(int)*2*dataFile->numFrames);
    dataFile->FTT = (double *) ckalloc(sizeof(double)*(dataFile->numFrames+1));

    if (dataFile->rawfd < 0) {
	dataFile->rawfd = open(dataFile->filename, O_RDONLY | O_BINARY);
    }
    dataFile->open_refCount++;

    MJPEG_FIP_JMovieFileParse(dataFile);
}


/********************************************************************
 * MJPEG_FIP_JMovieFileParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Procedure used to incrementally parse JMovie file. Picks up where
 *    it last left off. Parses information for some small number of frames,
 *    then reschedules itself on idle queue.
 *********************************************************************/	

static void
MJPEG_FIP_JMovieFileParse(MJPEGDataFile *dataFile)
{
    int num_to_read = NUM_FRAME_TO_PARSE_PER_IDLE;
    int last_time_flag = 0;
    int *j_movie_offset_table;
    int frame_counter;
    int cntr, skipped;
    unsigned char j_movie_cmd;
    int frame_size;
    ParseInfo *pinfo;

    /*
     * The following are definitions matching the Parallax JMovie macros.
     * They are included here to avoid having to distribute the Parallax
     * include files.
     */

    int LOAD_AUDIO_0 = 229;
    int LOAD_AUDIO_1 = 228;
    int LOAD_AUDIO_2 = 227;
    int LOAD_AUDIO_3 = 226;
    int LOAD_JPEG = 236;
    int END_FRAME = 237;

    pinfo = dataFile->pinfo;
    
    if (pinfo->highFrameNum+num_to_read >= dataFile->numFrames) {
	num_to_read = dataFile->numFrames-pinfo->highFrameNum;
	last_time_flag = 1;
    }


    lseek(dataFile->rawfd, pinfo->indexbuf+(4*pinfo->highFrameNum), SEEK_SET);
    
    j_movie_offset_table = (int *) ckalloc(num_to_read*sizeof(int));

    read(dataFile->rawfd, j_movie_offset_table, 4*num_to_read);
    
    frame_counter = 0;

    for (cntr=0; cntr < num_to_read; cntr++) {
	j_movie_offset_table[cntr] = ntohl(j_movie_offset_table[cntr]);

	lseek(dataFile->rawfd, j_movie_offset_table[cntr], SEEK_SET);
	
	read(dataFile->rawfd, &j_movie_cmd, 1);
	skipped = 0;
	
	while (j_movie_cmd != LOAD_JPEG) {
	    if ((j_movie_cmd == LOAD_AUDIO_0) ||
		(j_movie_cmd == LOAD_AUDIO_1) ||
		(j_movie_cmd == LOAD_AUDIO_2) ||
		(j_movie_cmd == LOAD_AUDIO_3)) {
		
		lseek(dataFile->rawfd, pinfo->auSlice, SEEK_CUR);
	    } else if (j_movie_cmd == END_FRAME) {
		skipped = 1;
		break;
	    }
	    
	    read(dataFile->rawfd, &j_movie_cmd, 1);
	}
	
	if (!skipped) {
	    read(dataFile->rawfd, &frame_size, 4);
	    frame_size = ntohl(frame_size);
	    
	    dataFile->FOT[(pinfo->highFrameNum+frame_counter)*2] =
	      lseek(dataFile->rawfd, 0, SEEK_CUR);
	    dataFile->FOT[(pinfo->highFrameNum+frame_counter)*2+1] =
	      frame_size;
	    dataFile->FTT[pinfo->highFrameNum+frame_counter] =
	      ((double) pinfo->highFrameNum+frame_counter) / 
		((double) pinfo->fps);
	    frame_counter++;
	}
    }

    pinfo->highFrameNum += frame_counter;

    if (!last_time_flag) {
	Tcm_DoWhenIdle("tk", MJPEG_FIP_ContinueParse, dataFile);
    } else {
	dataFile->inParse = 0;

	dataFile->open_refCount--;
	if (dataFile->open_refCount <= 0) {
	    close(dataFile->rawfd);
	    dataFile->rawfd = -1;
	}

	if (dataFile->pinfo != NULL) {
	    ckfree((char *) dataFile->pinfo);
	}
    }
    ckfree((char *) j_movie_offset_table);
}


/********************************************************************
 * MJPEG_FIP_AVIFileStartParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Used to initiate file parsing for .avi files.
 *********************************************************************/	
    
static int
MJPEG_FIP_AVIFileStartParse(MJPEGDataFile *dataFile)
{
    int file_index, buf_index;
    unsigned char buffer[CF_READ_BUFFER_SIZE];
    int num_read, junk_size;
    int movi_length, movi_end;
    ParseInfo *pinfo;

    pinfo = dataFile->pinfo;

    if (dataFile->rawfd < 0) {
	dataFile->rawfd = open(dataFile->filename, O_RDONLY | O_BINARY);
    }
    dataFile->open_refCount++;

    if (pinfo->tableSize < 1) {
	pinfo->tableSize = 128;
    }

    dataFile->FOT = (int *) ckalloc(pinfo->tableSize*sizeof(int)*2);
    dataFile->FTT = (double *) ckalloc((pinfo->tableSize+1)*sizeof(double));

    lseek(dataFile->rawfd, pinfo->strl_start+pinfo->strl_size, SEEK_SET);

    file_index = pinfo->strl_start+pinfo->strl_size;
    buf_index = 0;

    num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
    if (num_read < 12) {
	dataFile->fatalError = AVI_MISREAD_ERROR;
	dataFile->inParse = 0;
	if (dataFile->pinfo != NULL) {
	    ckfree((char *) dataFile->pinfo);
	    dataFile->pinfo = NULL;
	}
	return;
    }

    FIND_AVI_LIST('m', 'o', 'v', 'i');

    buf_index += 4;

    movi_length = GetInt32FromAVI(buffer, buf_index);
    buf_index += 4;
    movi_end = file_index + buf_index + movi_length;
    buf_index += 4;
    lseek(dataFile->rawfd, buf_index+file_index, SEEK_SET);
    pinfo->lastParsePos = lseek(dataFile->rawfd, 0, SEEK_CUR);
    pinfo->movi_length = movi_length;
    pinfo->movi_end = movi_end;

    MJPEG_FIP_AVIFileParse(dataFile);
    return TCL_OK;
}


/********************************************************************
 * MJPEG_FIP_AVIFileParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Procedure that actually does incremental .avi file parsing.
 *    Parses information about some small number of frames and then
 *    schedules furthur parsing on idle queue until done.
 *********************************************************************/	

static int
MJPEG_FIP_AVIFileParse(MJPEGDataFile *dataFile)
{
    int num_to_read = NUM_FRAME_TO_PARSE_PER_IDLE;
    int file_index;
    int buf_index;
    int num_read;
    unsigned char buffer[CF_READ_BUFFER_SIZE];
    int frame_counter;
    int list_size;
    int frame_length;
    ParseInfo *pinfo;

    pinfo = dataFile->pinfo;
    
    lseek(dataFile->rawfd, pinfo->lastParsePos, SEEK_SET);
    file_index = pinfo->lastParsePos;
    buf_index = 0;
    num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
    if (num_read < 12) {
	dataFile->fatalError = AVI_MISREAD_ERROR;
	dataFile->inParse = 0;
	if (dataFile->pinfo != NULL) {
	    ckfree((char *) dataFile->pinfo);
	    dataFile->pinfo = NULL;
	}
	dataFile->open_refCount--;
	if (dataFile->open_refCount <= 0) {
	    close(dataFile->rawfd);
	    dataFile->rawfd = -1;
	}
	return TCL_ERROR;
    }


    frame_counter = 0;

    while (buf_index + file_index < pinfo->movi_end) {
	if ((buffer[buf_index] == 'L') && (buffer[buf_index+1] == 'I') &&
	    (buffer[buf_index+2] == 'S') && (buffer[buf_index+3] == 'T')) {
	    buf_index += 4;
	    
	    if ((buffer[buf_index+4] == 'r') && (buffer[buf_index+5] == 'e') &&
		(buffer[buf_index+6] == 'c') && (buffer[buf_index+7] == ' ')) {
		buf_index += 8;
	    } else {
		list_size = GetInt32FromAVI(buffer, buf_index);
		buf_index += 4 + list_size;
	    }
	} else {
	    
	    frame_length = GetInt32FromAVI(buffer, buf_index+4);
	    
	    if ((buffer[buf_index] == pinfo->stream_id[0]) && 
		(buffer[buf_index+1] == pinfo->stream_id[1]) &&
		(buffer[buf_index+2] == 'd') &&
		(buffer[buf_index+3] == 'c')) {
		
		if (pinfo->highFrameNum+frame_counter >= pinfo->tableSize) {
		    pinfo->tableSize += 128;
		    dataFile->FOT = (int *) ckrealloc((char *) dataFile->FOT,
						      sizeof(int)*2*
						      pinfo->tableSize);
		    dataFile->FTT = (double *) ckrealloc((char *) dataFile->FTT,
							 sizeof(double)*
							 (pinfo->tableSize+1));
		}

		dataFile->FOT[(pinfo->highFrameNum+frame_counter)*2] = 
		  file_index + buf_index + 8;
		dataFile->FOT[(pinfo->highFrameNum+frame_counter)*2+1] =
		  frame_length;
		
		dataFile->FTT[pinfo->highFrameNum+frame_counter] =
		  ((double) (pinfo->highFrameNum+frame_counter)) / 
		    (((double) pinfo->stream_rate) /
		     ((double) pinfo->stream_scale));
		
		frame_counter++;
	    }

	    buf_index += 8;
	    if (frame_length%2) {
		buf_index += frame_length + 1;
	    } else {
		buf_index += frame_length;
	    }
		
	    if (frame_counter >= num_to_read) {
		dataFile->FTT[pinfo->highFrameNum+frame_counter] =
		  ((double) (pinfo->highFrameNum+frame_counter)) / 
		    ((double) pinfo->stream_rate) /
		      ((double) pinfo->stream_scale);
		pinfo->lastParsePos = file_index + buf_index;
		pinfo->highFrameNum += num_to_read;
		Tcm_DoWhenIdle("tk", MJPEG_FIP_ContinueParse, dataFile);
		return TCL_OK;
	    }
	}
	if (buf_index > num_read - 12) {
	    lseek(dataFile->rawfd, buf_index+file_index, SEEK_SET);
	    file_index += buf_index;
	    buf_index = 0;
	    num_read = read(dataFile->rawfd, buffer, CF_READ_BUFFER_SIZE);
	    if (num_read < 12) break;
	}
    }

    dataFile->numFrames = pinfo->highFrameNum + frame_counter;
    dataFile->clipLength = dataFile->FTT[dataFile->numFrames] =
      (((double) dataFile->numFrames) /
       (((double) pinfo->stream_rate) /
	((double) pinfo->stream_scale)));

    dataFile->inParse = 0;
    if (dataFile->pinfo != NULL) {
	ckfree((char *) dataFile->pinfo);
	dataFile->pinfo = NULL;
    }

    dataFile->FOT = (int *) ckrealloc((char *) dataFile->FOT, 
				      sizeof(int)*2*dataFile->numFrames);
    dataFile->FTT = (double *) ckrealloc((char *) dataFile->FTT,
					 sizeof(double)*dataFile->numFrames+1);
    
    dataFile->open_refCount--;
    if (dataFile->open_refCount<=0) {
	close(dataFile->rawfd);
	dataFile->rawfd = -1;
    }

    return TCL_OK;
}


/********************************************************************
 * MJPEG_FIP_VidFileStartParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Procedure used to initiate file parsing for .vid files.
 *    Allocates tables, sets up parsing information, then calls 
 *    .vid file specific file parsing procedure.
 *********************************************************************/	
	
static void
MJPEG_FIP_VidFileStartParse(MJPEGDataFile *dataFile)
{
    ParseInfo *pinfo;

    pinfo = dataFile->pinfo;

    pinfo->tableSize = 1000;

    dataFile->FOT = (int *) ckalloc(2*sizeof(int)*pinfo->tableSize);
    dataFile->FTT = (double *) ckalloc(sizeof(double)*(pinfo->tableSize+1));

    pinfo->lastParsePos = 0;

    if (dataFile->rawfd < 0) {
	dataFile->rawfd = open(dataFile->filename, O_RDONLY|O_BINARY);
    }
    dataFile->open_refCount++;

    MJPEG_FIP_VidFileParse(dataFile);
}


/********************************************************************
 * MJPEG_FIP_VidFileParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Procedure that does incremental file parsing of .vid files.
 *    Parses information about some small number of frames and then
 *    reschedules parsing on idle queue, until done.
 *********************************************************************/	

static void
MJPEG_FIP_VidFileParse(MJPEGDataFile *dataFile)
{
    int num_to_read = NUM_FRAME_TO_PARSE_PER_IDLE;
    int i, cur_pos, new_pos, num_read;
    int width, height, size;
    ParseInfo *pinfo;

    pinfo = dataFile->pinfo;

    lseek(dataFile->rawfd, pinfo->lastParsePos, SEEK_SET);

    for (i=0; i<num_to_read; i++) {

	cur_pos = lseek(dataFile->rawfd, 0, SEEK_CUR);

	num_read = 0;
	num_read += read(dataFile->rawfd, &width, 4);
	num_read += read(dataFile->rawfd, &height, 4);
	num_read += read(dataFile->rawfd, &size, 4);
	
	width = ntohl(width);
	height = ntohl(height);
	size = ntohl(size);
	
	lseek(dataFile->rawfd, 8+size, SEEK_CUR);
	new_pos = lseek(dataFile->rawfd, 0, SEEK_CUR);

	if ((num_read != 12) || (new_pos - cur_pos != 20+size) ||
	    (width != dataFile->cfInfo->width) ||
	    (height != dataFile->cfInfo->height) ||
	    (size < 0)) {

	    /* Must be done. */
	    
	    dataFile->numFrames = pinfo->highFrameNum+i;
	    dataFile->clipLength = ((double) dataFile->numFrames) / 30.0;
	    
	    dataFile->FOT = (int *) ckrealloc((char *) dataFile->FOT,
					      sizeof(int)*2*dataFile->numFrames);
	    dataFile->FTT = (double *) ckrealloc((char *) dataFile->FTT,
					      sizeof(double)*(dataFile->numFrames+1));
	    dataFile->FTT[dataFile->numFrames] = dataFile->clipLength;

	    dataFile->inParse = 0;
	    
	    if (dataFile->pinfo != NULL) {
		ckfree((char *) dataFile->pinfo);
		dataFile->pinfo = NULL;
	    }

	    dataFile->open_refCount--;
	    if (dataFile->open_refCount <= 0) {
		close(dataFile->rawfd);
		dataFile->rawfd = -1;
	    }

	    return;
	}

	if (pinfo->highFrameNum+i >= pinfo->tableSize) {
	    pinfo->tableSize += 1000;
	    dataFile->FOT = (int *) ckrealloc((char *) dataFile->FOT,
					      2*sizeof(int)*pinfo->tableSize);
	    dataFile->FTT = (double *) ckrealloc((char *) dataFile->FTT,
						 sizeof(double)*(pinfo->tableSize+1));
	}

	dataFile->FOT[(pinfo->highFrameNum+i)*2] = cur_pos;
	dataFile->FOT[(pinfo->highFrameNum+i)*2+1] = size;
	dataFile->FTT[pinfo->highFrameNum+i] = ((1.0/30.0)*
						   ((double) pinfo->highFrameNum+i));
    }

    dataFile->FTT[pinfo->highFrameNum+i] = ((1.0/30.0)*
					       ((double) pinfo->highFrameNum+i));

    pinfo->lastParsePos = new_pos;
    pinfo->highFrameNum += num_to_read;

    Tcm_DoWhenIdle("tk", MJPEG_FIP_ContinueParse, dataFile);
}


/********************************************************************
 * MJPEG_FIP_ChunkFileStartParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Initiates parsing on chunk file types. Allocates tables and then
 *    calls procedure to actually parse frames.
 *********************************************************************/	

static void
MJPEG_FIP_ChunkFileStartParse(MJPEGDataFile *dataFile)
{
    dataFile->FOT = (int *) ckalloc(2*sizeof(int)*dataFile->numFrames);
    dataFile->FTT = (double *) ckalloc(sizeof(double)*(dataFile->numFrames+1));

    if (dataFile->cfState == NULL) {
	dataFile->cfState = Chunk_OpenFile(dataFile->filename, "r");
    }
    dataFile->open_refCount++;

    MJPEG_FIP_ChunkFileParse(dataFile);
}


/********************************************************************
 * MJPEG_FIP_ChunkFileParse
 *
 * Parameters:
 *    dataFile - file data structure.
 *
 * Results:
 *    None;
 * 
 * Action:
 *    Procedure that incrementally parses information about frames from
 *    chunk file. Queues furthur parsing on idle queue until done.
 *********************************************************************/	

static void
MJPEG_FIP_ChunkFileParse(MJPEGDataFile *dataFile) 
{
    int num_to_read = NUM_FRAME_TO_PARSE_PER_IDLE;
    tChunkHeader chunk_hdr;
    int i, offset, length, sec, nsec;
    int qagain_flag;
    ParseInfo *pinfo;

    pinfo = dataFile->pinfo;

    Chunk_ReadRewind(dataFile->cfState);
    if (Chunk_FindType(dataFile->cfState, CM_FOT, &chunk_hdr)) {
	dataFile->fatalError = NO_FOT_ERROR;
	return;
    }

    /* Skip entries we've already read. */

    lseek(dataFile->cfState->fd, pinfo->highFrameNum * 8, SEEK_CUR);

    /* Read some entries. */

    for (i=0; i<num_to_read; i++) {
	if (pinfo->highFrameNum+i >= dataFile->numFrames) break;
	
	read(dataFile->cfState->fd, &offset, 4);
	offset = ntohl(offset);

	read(dataFile->cfState->fd, &length, 4);
	length = ntohl(length);


	dataFile->FOT[(pinfo->highFrameNum+i)*2] = offset;
	dataFile->FOT[(pinfo->highFrameNum+i)*2+1] = length;
    }

    Chunk_ReadRewind(dataFile->cfState);
    if (Chunk_FindType(dataFile->cfState, CM_FTT, &chunk_hdr)) {
	dataFile->fatalError = NO_FTT_ERROR;
	return;
    }

    lseek(dataFile->cfState->fd, pinfo->highFrameNum * 8, SEEK_CUR);
    
    for (i=0; i<num_to_read+1; i++) {
	if (pinfo->highFrameNum+i >= dataFile->numFrames+1) break;

	read(dataFile->cfState->fd, &sec, 4);
	read(dataFile->cfState->fd, &nsec, 4);

	sec = ntohl(sec);
	nsec = ntohl(nsec);

	dataFile->FTT[(pinfo->highFrameNum+i)] = (((double) sec) + 
						     (((double) nsec) / ONE_BILLION));
    }

    if (i < num_to_read) {
	/* Read all frames. Do not queue again. */
	qagain_flag = 0;
    } else {
	qagain_flag = 1;
	pinfo->highFrameNum += num_to_read;
    }

    if (qagain_flag) {
	Tcm_DoWhenIdle("tk", MJPEG_FIP_ContinueParse, dataFile);
	return;
    } else {

	dataFile->open_refCount--;
	if (dataFile->open_refCount <= 0) {
	    Chunk_CloseFile(dataFile->cfState);
	    dataFile->cfState = NULL;
	}

	dataFile->inParse = 0;
	
	if (dataFile->pinfo != NULL) {
	    ckfree((char *) dataFile->pinfo);
	    dataFile->pinfo = NULL;
	}
    }
}
