/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMJPEGFileInterface.h,v 1.3 1996/02/06 22:41:39 kpatel Exp $
 *
 * Contains data structures used by the MJPEG File Segment object
 * and the MJPEG File Interface.
 */


#include "cmCookies.h"
#include "cmChunk.h"

/*
 * The following record is used as the file is parsed. Once all tables
 * required are completely built, this data structure is no longer kept.
 * Since many different types of files need to be dealt with, this data
 * structure has many fields, but only a small subset of them will be used
 * to parse any given file type (i.e., .avi, .vid, JMovie, .mjpgc, etc).
 */

typedef struct ParseInfo {
    int highFrameNum;         /* Next frame number for which information
			       * needs to be constructed. */
    int tableSize;            /* Current size of allocated tables */
    int lastParsePos;         /* Position in file where parsing last ended. */
    int stream_length;        /* Used when reading .avi files. */
    int strl_start;           /* Used when reading .avi files. */
    int strl_size;            /* Used when reading .avi files. */
    int movi_length;          /* Used when reading .avi files. */
    int movi_end;             /* Used when reading .avi files. */
    char stream_id[60];       /* Used when reading .avi files. */
    int stream_rate;          /* Used when reading .avi files. */
    int stream_scale;         /* Used when reading .avi files. */
    int auSlice;              /* Used when reading JMovie files. */
    int fps;                  /* Frame rate as indicated within file. */
    int indexbuf;             /* Indicates offset to JMovie index from start of file. */
} ParseInfo;

/* 
 * The following data structure is kept per file that has been opened. This is
 * the underlying data structure that is essentially shared by all mjpegSegment
 * objects that are trying to access the same file. These objects, however,
 * should not directly change or access this data structure, but should
 * instead, access the data structure through the file interface routines
 * provided.
 */

typedef struct MJPEGDataFile {
    char *filename;                /* Name of file. Must be full path name. */
    tChunkFileState *cfState;      /* Pointer to chunk file interface structure.
				      Used only if the file is .mjpgc type */
    int rawfd;                     /* File descriptor for file, used only if not a 
				      chunk file. */
    int file_refCount;             /* Number of segment objects using this structure. */
    int open_refCount;             /* Number of segment objects that want
				      the file to be "open" */
    int *FOT;                      /* Frame offset table. */
    double *FTT;                   /* Frame time table. */
    int fileType;                  /* File type (i.e., .avi, .mjpgc, etc.) */
    int inParse;                   /* Indicates that parsing is still in progress. */
    ParseInfo *pinfo;              /* Pointer to parse info structure if parsing. */
    int numFrames;                 /* Total number of frames in file. */
    double clipLength;             /* Length in seconds of clip. */
    int dataStart;                 /* Offset from beginning of file to data. */
    MJPEGHeader *cfInfo;           /* Header structure with standard info like
				      width, height, qfactor, etc. */
    int fatalError;                /* Indicates something very bad occured
				      during parse or reading of data. */
} MJPEGDataFile;


/*
 * This structure is kept for each MJPEG Segment object that is accessing any 
 * given file. It contains a pointer to the real file data structure defined
 * above that is actually being shared by possibly many objects. This is the
 * structure that is passed to the routines that provide the Segment objects
 * access to the file.
 */

typedef struct MJPEGFileInterface {
    MJPEGDataFile *data_file;  /* Pointer to shared file structure. */
    int is_open;               /* Indication as to whether this Segment object
				* thinks that the file is open or not. */
} MJPEGFileInterface;

/* 
 * Declarations of routines provided to the Segment objects in order to open
 * and access files.
 */

MJPEGFileInterface *MJPEG_FIP_Open(char *fname);
void MJPEG_FIP_Close(MJPEGFileInterface *fip);
void MJPEG_FIP_FDClose(MJPEGFileInterface *fip);
void MJPEG_FIP_FDOpen(MJPEGFileInterface *fip);
int MJPEG_FIP_Subformat(MJPEGFileInterface *fip);
int MJPEG_FIP_Device(MJPEGFileInterface *fip);
int MJPEG_FIP_Width(MJPEGFileInterface *fip);
int MJPEG_FIP_Height(MJPEGFileInterface *fip);
int MJPEG_FIP_QFactor(MJPEGFileInterface *fip);
double MJPEG_FIP_ClipLength(MJPEGFileInterface *fip);
int MJPEG_FIP_FrameFromTime(MJPEGFileInterface *fip, double time, double speed);
void MJPEG_FIP_TimeFromFrame(MJPEGFileInterface *fip, int frameNum, double *stime, double *etime);
int MJPEG_FIP_LengthFromFrame(MJPEGFileInterface *fip, int frameNum);
int MJPEG_FIP_ReadFrame(MJPEGFileInterface *fip, char *buffer, int frameNum);
int MJPEG_FIP_FrameListFromTime(MJPEGFileInterface *fip, double start, double end, int **frameList);
void MJPEG_FIP_Info(MJPEGFileInterface *fip, double start, double end, int *numFrames, int *maxSize);

