/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMJPEGFileSegment.c,v 1.7 1996/02/13 00:49:04 kpatel Exp $
 *
 * Implementation of the "mjpegSegment" command, which reads
 * files containg MJPEG compressed video
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include "cmCookies.h"
#include "cmBufferlist.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmMJPEGFileInterface.h"

#ifndef O_BINARY
#    define O_BINARY 0
#endif

#define ONE_BILLION 1000000000.0
#define CF_READ_BUFFER_SIZE 512



/*
 * Structure used when picking next set of frames to be sent.
 */

typedef struct {
    Buffer *bufPtr;
    int index;
    int size;
    double sTime;
    double eTime;
    int sendFlag;
    int offset;
} FramePick;

/*
 * Structure kept for each clipfile object. 
 */

typedef struct MJPEGSegment {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */

    ClientData lts;		/* LTS for object to use. */
    char *ltsName;
    int ltsTriggerNum;          /* LTS trigger for changed lts. */

    double sendAhead;		/* Send ahead time. */
    double cycleTime;		/* Cycle period. */

    int bandwidth;		/* Bandwidth available */
    int extraBWCounter;         /* Used to adjust number of bytes send per cycleTime
				   to mactch with bandwidth restrictions. */
    double maxFrameRate;	/* Maximum frame rate. */
    double extraFrameCounter;   /* Used to adjust number of frames sent per cycleTime
				   to match with maxFrameRate */

    char *filename;		/* Filename of clipfile. */
    MJPEGFileInterface *fip;    /* File interface pointer. */
    MJPEGHeader *cfInfo;	/* MJPEG Header from clipfile. */

    char *outCmd;		/* Command to send buffers to. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;           /* ClientData for outCCmd. */

    int lastFrameNum;           /* Frame number of last frame sent. */

    int ready;                  /* Ready flag. */

    int at_token;               /* Token for at event. */

    double clipStart;           /* Start time relative to clip beginning = 0.*/
    double clipEnd;             /* End time relative to clip beginning = 0. */
    double logicalStart;        /* Start time relative to lts. */
    double logicalEnd;          /* End time relative to lts. */
} MJPEGSegment;

/* Globals used by this object module. */

static int nextMJPEGSegmentNum = 0;	/* 
					 *  Variable for next unique 
					 *  object name. 
					 */

/* Macros for converting between object clip time and object logical time. */

#define LOG_TIME_FROM_CLIP_TIME(objPtr,time) ((((time - objPtr->clipStart) *          \
						(objPtr->logicalEnd -                 \
						 objPtr->logicalStart)) /             \
					       (objPtr->clipEnd - objPtr->clipStart)) \
					      + objPtr->logicalStart)

#define CLIP_TIME_FROM_LOG_TIME(objPtr,time) ((((time - objPtr->logicalStart) *    \
						(objPtr->clipEnd -                 \
						 objPtr->clipStart)) /             \
					       (objPtr->logicalEnd -               \
						objPtr->logicalStart))             \
					      + objPtr->clipStart)

/* Forward declarations. */

int MJPEGSegmentObjCmd(ClientData clientData, Tcl_Interp * interp, int argc, char **argv);
static int MJPEGSegmentInfo(MJPEGSegment *objPtr, double start, double end);
static int MJPEGSegmentReady(MJPEGSegment *objPtr);
static int MJPEGSegmentUnready(MJPEGSegment *objPtr);
static int MJPEGSegmentChangeOutCmd(MJPEGSegment *objPtr, char *newOut);
static MJPEGFileInterface *MJPEGSegmentOpenClipfile(MJPEGSegment *objPtr, char *filename);
void MJPEGSegmentLTSChanged(ClientData clientdata);
void MJPEGSegmentSendData(ClientData clientdata);

/************************************************************
 *
 * CreateNewMJPEGSegment
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

MJPEGSegment *
CreateNewMJPEGSegment(Tcl_Interp *interp,  /* Pointer to interpreter. */
		      char *name           /* Name of new object. */)
{

    MJPEGSegment *new;

    new = (MJPEGSegment *) ckalloc(sizeof(MJPEGSegment));

    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);

    new->lts = NULL;
    new->ltsName = NULL;
    new->ltsTriggerNum = -1;

    new->sendAhead = 0.2;
    new->cycleTime = 0.0;

    new->bandwidth = 0;
    new->extraBWCounter = 0;
    new->maxFrameRate = 0.0;
    new->extraFrameCounter = 0.0;

    new->filename = NULL;
    new->fip = NULL;

    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;

    new->lastFrameNum = -1;

    new->cfInfo = NULL;

    new->ready = 0;

    new->at_token = 0;

    new->clipStart = 0.0;
    new->clipEnd = 0.0;
    new->logicalStart = 0.0;
    new->logicalEnd = 0.0;

    return new;
}

/************************************************************
 *
 * DestroyMJPEGSegment
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

int
DestroyMJPEGSegment(MJPEGSegment *objPtr       /* Pointer to seg. object. */)
{

    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
	objPtr->at_token = 0;
    }

    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    if (objPtr->filename != NULL) {
	ckfree(objPtr->filename);
	objPtr->filename = NULL;
    }

    if (objPtr->fip != NULL) {
	MJPEG_FIP_Close(objPtr->fip);
	objPtr->fip = NULL;
    }

    if (objPtr->outCmd != NULL)
	ckfree((char *) objPtr->outCmd);

    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);
    }

    if (objPtr->ltsName != NULL) {
	ckfree((char *) objPtr->ltsName);
    }

    ckfree((char *) objPtr);

    return TCL_OK;
}

/************************************************************
 *
 * ConfigureMJPEGSegment
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

int
ConfigureMJPEGSegment(MJPEGSegment *objPtr, int argc, char **argv)
{
  
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
  
    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {} {}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName, "}} ",
		    (char *) NULL);
	}
	Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	Tcl_AppendResult(interp, "{-sendAhead {0.2} ", str, "} ", (char *)NULL);

	sprintf(str, "{%d}", objPtr->bandwidth);
	Tcl_AppendResult(interp, "{-bandwidth {0} ", str, "} ", (char *)NULL);

	Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	Tcl_AppendResult(interp, "{-cycleTime {0.0} ", str, "} ", (char *)NULL);

	if (objPtr->filename == NULL) {
	    Tcl_AppendResult(interp, "{-filename {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-filename {} {", objPtr->filename,
			    "}} ",(char *)NULL);
	}
	
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}

	Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	Tcl_AppendResult(interp, "{-maxFrameRate {0.0} ", str,
			  "} ", (char *)NULL);

	Tcl_PrintDouble(interp, objPtr->clipStart, str);
	Tcl_AppendResult(interp, "{-clipStart 0.0 ", str, "} ", (char *) NULL);

	Tcl_PrintDouble(interp, objPtr->clipEnd, str);
	Tcl_AppendResult(interp, "{-clipEnd 0.0 ", str, "} ", (char *) NULL);

	Tcl_PrintDouble(interp, objPtr->logicalStart, str);
	Tcl_AppendResult(interp, "{-logicalStart 0.0 ", str, "} ", (char *) NULL);

	Tcl_PrintDouble(interp, objPtr->logicalEnd, str);
	Tcl_AppendResult(interp, "{-logicalEnd 0.0 ", str, "} ", (char *) NULL);
	
	return TCL_OK;
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->lts == NULL) {
		Tcl_AppendResult(interp, "-lts {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName, "} ",
				  (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	    Tcl_AppendResult(interp, "-sendAhead {0.2} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'b') && (strncmp("-bandwidth", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->bandwidth);
	    Tcl_AppendResult(interp, "-bandwidth {0} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	    Tcl_AppendResult(interp, "-cycleTime {0.0} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'f') && (strncmp("-filename", argv[0], len) == 0)) {
	    if (objPtr->filename == NULL) {
		Tcl_AppendResult(interp, "-filename {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-filename {} {",
			  objPtr->filename, "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	    Tcl_AppendResult(interp, "-maxFrameRate {0.0} ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'c') && (strncmp("-clipStart", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->clipStart, str);
	    Tcl_AppendResult(interp, "-clipStart 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'c') && (strncmp("-clipEnd", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->clipEnd, str);
	    Tcl_AppendResult(interp, "-clipEnd 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'l') && (strncmp("-logicalStart", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->logicalStart, str);
	    Tcl_AppendResult(interp, "-logicalStart 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'l') && (strncmp("-logicalEnd", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->logicalEnd, str);
	    Tcl_AppendResult(interp, "-logicalEnd 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}


	Tcl_AppendResult(interp, "bad configuration option: ",
			  argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	/* ------------------------- LTS ------------------------- */
	if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newlts;
	    
	    /*
	     * Check args
	     */
	    cntr++;
	    if (cntr == argc) {
argcError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }

	    /*
	     * If resetting to null value, just free up old lts.
	     */
	    if (strlen(argv[cntr]) == 0) {
		if (objPtr->lts != NULL) {
		    if (objPtr->ltsTriggerNum != -1) {
			LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
			objPtr->ltsTriggerNum = -1;
		    }
		    Tk_Release(objPtr->lts);
		    objPtr->lts = NULL;
		}
		if (objPtr->ltsName != NULL) {
		    ckfree((char *) objPtr->ltsName);
		    objPtr->ltsName = NULL;
		}
	    } else {
		newlts = (ClientData)Tcm_FindAtom(interp, argv[cntr]);
		if (newlts == NULL) {
valueError:	    
		    Tcl_AppendResult(interp, "bad value \"", argv[cntr],
				  "\" given for configuration option ",
				  argv[cntr-1], (char *)NULL);
		    return TCL_ERROR;
		}
		
		/*
		 * Blow away old callback.
		 */
		if (objPtr->lts != NULL) {
		    if (objPtr->ltsTriggerNum != -1) {
			LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
			objPtr->ltsTriggerNum = -1;
		    }
		    Tk_Release(objPtr->lts);
		    objPtr->lts = NULL;
		}
		objPtr->lts = newlts;
		Tk_Preserve (objPtr->lts);

		/*
		 * Copy in new lts name
		 */
		if (objPtr->ltsName != NULL) {
		    ckfree((char *) objPtr->ltsName);
		}
		objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
		strcpy(objPtr->ltsName, argv[cntr]);
		
		/*
		 * Set up new lts trigger
		 */
		objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts,
				LTS_SPEED_FLAG | LTS_OFFSET_FLAG | LTS_VALUE_FLAG,
			        MJPEGSegmentLTSChanged, (ClientData) objPtr);
	    }
	    MJPEGSegmentUnready(objPtr);

	/* ------------------------- SEND_AHEAD ------------------------- */
	} else if ((c=='s') && (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	    double newSendAhead;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->sendAhead = newSendAhead;

	/* ------------------------- BANDWIDTH -------------------------- */
	} else if ((c=='b') && (strncmp("-bandwidth", argv[cntr], len) == 0)) {
	    int newBandwidth;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newBandwidth) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->bandwidth = newBandwidth;
	    objPtr->extraBWCounter = 0;

	/* ------------------------- CYCLE_TIME ------------------------- */
	} else if ((c=='c') && (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	    double newCycleTime;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->cycleTime = newCycleTime;

	/* ------------------------- FILE_NAME -------------------------- */
	} else if ((c=='f') && (strncmp("-filename", argv[cntr], len) == 0)) {
	    Tcl_DString fname_dstring;
	    char *fname;

	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (objPtr->filename != NULL) {
		Tcl_AppendResult(interp,
				"Can not set clipfile name more than once.",
				(char *)NULL);
		return TCL_ERROR;
	    }

	    fname = Tcl_TildeSubst(objPtr->interp, argv[cntr], &fname_dstring);

	    objPtr->fip = MJPEGSegmentOpenClipfile(objPtr, fname);

	    if (objPtr->fip == NULL) {
		Tcl_DStringFree(&fname_dstring);
		Tcl_AppendResult(interp, "Couldn't open file \"",
				 fname, "\" as mjpeg file",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    objPtr->filename = (char *) ckalloc(strlen(fname) + 1);
	    strcpy(objPtr->filename, fname);
	    Tcl_DStringFree(&fname_dstring);

	/* ------------------------- OUT_CMD ---------------------------- */
	} else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    
	    if (MJPEGSegmentChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	    MJPEGSegmentUnready(objPtr);
	    
	/* ------------------------- MAX_FRAME_RATE --------------------- */
	} else if ((c=='m') &&
		   (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
	    double newMaxFrameRate;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->maxFrameRate = newMaxFrameRate;
	    objPtr->extraFrameCounter = 0.0;

	/* ------------------------- CLIP_START --------------------- */
	} else if ((c=='c') &&
		   (strncmp("-clipStart", argv[cntr], len) == 0)) {
	    double newClipStart;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (argv[cntr][0] == '?') {
		objPtr->clipStart = objPtr->clipEnd - 
		  (objPtr->logicalEnd - objPtr->logicalStart);
	    } else {

		if (Tcl_GetDouble(interp, argv[cntr], &newClipStart) != TCL_OK) {
		    goto valueError;
		}
		objPtr->clipStart = newClipStart;
	    }

	    MJPEGSegmentUnready(objPtr);

	/* ------------------------- CLIP_END --------------------- */
	} else if ((c=='c') &&
		   (strncmp("-clipEnd", argv[cntr], len) == 0)) {
	    double newClipEnd;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (argv[cntr][0] == 'e') {
		if (objPtr->fip != NULL) {
		    objPtr->clipEnd = MJPEG_FIP_ClipLength(objPtr->fip);
		}
	    } else if (argv[cntr][0] == '?') {
		objPtr->clipEnd = objPtr->clipStart + 
		  (objPtr->logicalEnd - objPtr->logicalStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newClipEnd) != TCL_OK) {
		    goto valueError;
		}
		objPtr->clipEnd = newClipEnd;
	    }

	    MJPEGSegmentUnready(objPtr);

	/* ------------------------- LOGICAL_START --------------------- */
	} else if ((c=='l') &&
		   (strncmp("-logicalStart", argv[cntr], len) == 0)) {
	    double newLogicalStart;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (argv[cntr][0] == '?') {
		objPtr->logicalStart = objPtr->logicalEnd - 
		  fabs(objPtr->clipEnd - objPtr->clipStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newLogicalStart) != TCL_OK) {
		    goto valueError;
		}
		objPtr->logicalStart = newLogicalStart;
	    }

	    MJPEGSegmentUnready(objPtr);

	/* ------------------------- LOGICAL_END --------------------- */
	} else if ((c=='l') &&
		   (strncmp("-logicalEnd", argv[cntr], len) == 0)) {
	    double newLogicalEnd;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    
	    if (argv[cntr][0] == '?') {
		objPtr->logicalEnd = objPtr->logicalStart +
		  fabs(objPtr->clipEnd - objPtr->clipStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newLogicalEnd) != TCL_OK) {
		    goto valueError;
		}
		objPtr->logicalEnd = newLogicalEnd;
	    }

	    MJPEGSegmentUnready(objPtr);

	/* ------------------------- UNKNOWN OPTION ----------------------*/
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			  argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}

/************************************************************
 *
 * MJPEGSegmentCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the mjpegSegment command.
 *     A new object of type MJPEGSegment is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
MJPEGSegmentCreateCmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char **argv)
{
    char new_name[256];
    MJPEGSegment *new;
    
    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextMJPEGSegmentNum++;
    sprintf(new_name, "%smjpegSeg%d", argv[1], nextMJPEGSegmentNum);
    
    new = CreateNewMJPEGSegment(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MJPEGSegmentObjCmd,
		      (ClientData) new, (void (*)()) NULL);
    
    if (argc > 2) {
	ConfigureMJPEGSegment(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MJPEGSegmentObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     MJPEGSegment object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
MJPEGSegmentObjCmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char **argv)
{
    char c;
    int len, i;
    MJPEGSegment *objPtr = (MJPEGSegment *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    /*------------------------- CONFIGURE ------------------------------*/
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureMJPEGSegment(objPtr, argc - 2, argv + 2);

    /*------------------------- DESTROY --------------------------------*/
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyMJPEGSegment(objPtr);

    /*------------------------- INFO -----------------------------------*/
    } else if ((c == 'i') && (strncmp("info", argv[1], len) == 0)) {
	double start_param, end_param;
	
	start_param = 0.0;
	end_param = 0.0;
	
	if (argc > 2) {
	    if (Tcl_GetDouble(interp, argv[2], &start_param) != TCL_OK) {
		Tcl_AppendResult(interp, "Bad value: \"", argv[2], "\" given",
				 " as argument to: ", argv[0], " ", argv[1],
				 (char *) NULL);
	    }
	    if (argc > 3) {
		if (Tcl_GetDouble(interp, argv[3], &end_param) != TCL_OK) {
		    Tcl_AppendResult(interp, "Bad value: \"", argv[3], 
				     "\" given as argument to: ", 
				     argv[0], " ", argv[1],
				     (char *) NULL);
		}
	    } else {
		end_param = start_param;
	    }
	} else {
	    start_param = objPtr->clipStart;
	    end_param = objPtr->clipEnd;
	}
	return MJPEGSegmentInfo(objPtr, start_param, end_param);
	
    /*------------------------- READY ----------------------------------*/
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return MJPEGSegmentReady(objPtr);

    /*------------------------- UNREADY --------------------------------*/
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return MJPEGSegmentUnready(objPtr);

    /*------------------------- WIDTH ----------------------------------*/
    } else if ((c == 'w') && (strncmp("width", argv[1], len) == 0)) {
	if (objPtr->fip == NULL) {
	    Tcl_SetResult(interp, "0", TCL_VOLATILE);
	    return TCL_OK;
	} else {
	    char retStr[100];
	    sprintf(retStr, "%d", MJPEG_FIP_Width(objPtr->fip));
	    Tcl_SetResult(interp, retStr, TCL_VOLATILE);
	    return TCL_OK;
	}
    /*------------------------- HEIGHT ----------------------------------*/
    } else if ((c == 'h') && (strncmp("height", argv[1], len) == 0)) {
	if (objPtr->fip == NULL) {
	    Tcl_SetResult(interp, "0", TCL_VOLATILE);
	    return TCL_OK;
	} else {
	    char retStr[100];
	    sprintf(retStr, "%d", MJPEG_FIP_Height(objPtr->fip));
	    Tcl_SetResult(interp, retStr, TCL_VOLATILE);
	    return TCL_OK;
	}
    }

    Tcl_AppendResult(interp, "Bad command option: ",
		argv[0], " ", argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  ready\n",
                "  unready\n",
                "  info ?clip_time ?clip_time??\n",
		"  width\n",
		"  height\n",
                (char *) NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * MJPEGSegmentModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
MJPEGSegmentModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mjpegSegment", MJPEGSegmentCreateCmd, NULL,
		(void (*)()) NULL);
    Tcl_GlobalEval(interp, "cminfo add object_info mjpegSegment {motion jpeg file}");
    return TCL_OK;
}


/************************************************************
 *
 * MJPEGSegmentReady
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Gets the object ready and starts sending data.
 ************************************************************/

static int
MJPEGSegmentReady(MJPEGSegment * objPtr)
{
    double clip_length;

    if (objPtr->ready) {
	return TCL_OK;
    }

    if (objPtr->fip == NULL) {
	Tcl_AppendResult(objPtr->interp, "Must specify a filename before readying this",
			 " object", (char *) NULL);
	return TCL_ERROR;
    }
    
    if (objPtr->lts == NULL) {
	Tcl_AppendResult(objPtr->interp, "Must specify an LTS before readying",
			 NULL);
	return TCL_ERROR;
    }

    /*
     * Blow away the old stuff.
     */

    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
    }
    objPtr->at_token = 0;

    /*
     * Check for clip start, end and logical start, end 
     * consistency, etc.
     */

    if (objPtr->logicalEnd - objPtr->logicalStart <= 0.0) {
	Tcl_AppendResult(objPtr->interp, "Logical end time must be greater than ",
			 "logical start time.", (char *) NULL);
	return TCL_ERROR;
    }

    if (objPtr->clipStart < 0.0) {
	objPtr->clipStart = 0.0;
    }
    if (objPtr->clipEnd < 0.0) {
	objPtr->clipEnd = 0.0;
    }
    
    clip_length = MJPEG_FIP_ClipLength(objPtr->fip);

    if (objPtr->clipStart > clip_length) {
	objPtr->clipStart = clip_length;
    }
    if (objPtr->clipEnd > clip_length) {
	objPtr->clipEnd = clip_length;
    }

    if (objPtr->clipEnd - objPtr->clipStart == 0.0) {
	Tcl_AppendResult(objPtr->interp, "Clip start and end times specified create ",
			 "empty segment.", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Set the ready flag.
     */
    objPtr->ready = 1;

    if (objPtr->lts != NULL) {
	MJPEGSegmentSendData((ClientData) objPtr);
    }
    return TCL_OK;
}

/************************************************************
 *
 * MJPEGSegmentUnready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Stops sending data and puts the object in the unready state.
 ************************************************************/

static int
MJPEGSegmentUnready(MJPEGSegment * objPtr)
{
    objPtr->lastFrameNum = -1;
    objPtr->ready = 0;
    return TCL_OK;
}

/************************************************************
 *
 * MJPEGSegmentLTSChanged
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Procedure called whenever the LTS changes.
 ************************************************************/

void
MJPEGSegmentLTSChanged(ClientData clientdata)
{
    MJPEGSegment *objPtr = (MJPEGSegment *) clientdata;

    objPtr->extraFrameCounter = 0.0;
    objPtr->extraBWCounter = 0;

    if (objPtr->ready == 1) {
	MJPEGSegmentUnready(objPtr);
	MJPEGSegmentReady(objPtr);
    }
    return;
}


/************************************************************
 *
 * MJPEGSegmentChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
MJPEGSegmentChangeOutCmd(MJPEGSegment *objPtr, char *newOut) 
{
#ifdef SECURE_CMT
    if ((newOut != NULL) && (newOut[0] != '@')) {
	Tcl_SetResult(objPtr->interp, "Cannot set Tcl outCmd because this program was compiled with -DSECURE_CMT.\nRecompile without security to use arbitrary Tcl outCmd's\n", TCL_STATIC);
	return TCL_ERROR;
    }
#endif

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
		&(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    

/************************************************************
 *
 * MJPEGSegmentSendData
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sends data according to current value of LTS, queues
 *     itself on the at queue for next send time.
 ************************************************************/

void
MJPEGSegmentSendData(ClientData clientData) 
{
    MJPEGSegment *objPtr = (MJPEGSegment *) clientData;
    double now, speed, when, midpoint;
    double eTime, sTime, lastETime, tmpTime;
    double clip_now, clip_cycle_end;
    int i, curBw, j;
    int length;
    MJPEGBufferHeader hdrPtr;
    double frac_want_to_send;
    int want_to_send;
    int bw_want_to_send;
    Buffer *bufPtr;
    int idx;
    int dataSegCode, dataLength;
    ScatterBuffer *sbPtr;
    SB_List *sbl;
    char *sbStr;
    int numToSend, bwToSend, origNumToSend;
    int frameNum;
    int *framelist;
    int sbIdx;
    static FramePick *fPickList;
    static int fPickNum = 0;
    extern void MJPEG_Prioritize (int n, int m, int k, int *pl);

    /*
     * Return if not ready, no LTS, or no out command.
     */
    if ((!objPtr->ready) ||
	(objPtr->lts == NULL) ||
	(objPtr->outCmd == NULL)) {
	goto done;
    }

    /*
     * Find out if we should be sending data or going to sleep.
     * If going to sleep, for how long.
     */

    now = LTS_Value(objPtr->lts);
    speed = LTS_Speed(objPtr->lts);
    now += objPtr->sendAhead * speed;

    /* Slightly different for forwards and backwards. */

    if (speed > 0.0) {
	if (objPtr->logicalEnd <= now) {
	    /* Sleep forever case. */
	    return;
	} else if (objPtr->logicalStart > now) {
	    /* Sleep until expected to be here. */

	    when = LTS_SystemTimeOf(objPtr->lts,
			(objPtr->logicalStart-(objPtr->sendAhead*speed)));
	    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				       "video", MJPEGSegmentSendData, objPtr, NULL,
				       NULL, NULL, NULL, NULL);
	    return;
	}
    } else if (speed < 0.0) {
	if (objPtr->logicalStart >= now) {
	    /* Sleep forever case. */
	    return;
	} else if (objPtr->logicalEnd < now) {
	    /* Sleep until expected to be here. */
	    
	    when = LTS_SystemTimeOf(objPtr->lts,
				    objPtr->logicalEnd+(objPtr->sendAhead*speed));
	    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				     "video", MJPEGSegmentSendData, objPtr, NULL,
				     NULL, NULL, NULL, NULL);
	  return;

	}
    } else {
	if (!((objPtr->logicalStart <= now) && 
	      (objPtr->logicalEnd > now)))  {
	    /* Sleep forever. */
	    MJPEG_FIP_FDClose(objPtr->fip);
	    return;
	}
    }

    /* 
     * Currently active, need to send data. 
     * Tell file interface to be open (i.e., 
     */

    MJPEG_FIP_FDOpen(objPtr->fip);

    /*
     * Calculate time in terms of clip time.
     */

    clip_now = CLIP_TIME_FROM_LOG_TIME(objPtr, now);

    /*
     * If only sending one frame because speed = 0.0 or
     * cycle time = 0.0, do this now. 
     */
    if ((objPtr->cycleTime == 0.0) || (speed == 0.0)) {
	
	/*
	 * Find frame to send.
	 */

	frameNum = MJPEG_FIP_FrameFromTime(objPtr->fip, clip_now, speed);
	if (speed != 0.0) {
	    MJPEG_FIP_TimeFromFrame(objPtr->fip, frameNum, &sTime, &eTime);
	    if (speed < 0.0) {
		tmpTime = eTime;
		eTime = sTime;
		sTime = tmpTime;
	    }
	    eTime = LOG_TIME_FROM_CLIP_TIME(objPtr, eTime);
	    sTime = LOG_TIME_FROM_CLIP_TIME(objPtr, sTime);
	} else {
	    sTime = eTime = now;
	}

	length = MJPEG_FIP_LengthFromFrame(objPtr->fip, frameNum);

	objPtr->lastFrameNum = frameNum;

	hdrPtr.hdrType = htonl(CM_MJPEG_BUF_HDR);
	hdrPtr.length = htonl(36);
	hdrPtr.width = htonl(objPtr->cfInfo->width);
	hdrPtr.height = htonl(objPtr->cfInfo->height);
	hdrPtr.device = htonl(objPtr->cfInfo->device);
	hdrPtr.subformat = htonl(objPtr->cfInfo->subformat);
	hdrPtr.qfactor = htonl(objPtr->cfInfo->qfactor);
	hdrPtr.logStart_s = htonl(floor(sTime));
	hdrPtr.logStart_ns = htonl(floor((sTime - floor(sTime))* ONE_BILLION));
	hdrPtr.logEnd_s = htonl(floor(eTime));
	hdrPtr.logEnd_ns = htonl(floor((eTime - floor(eTime))* ONE_BILLION));

	bufPtr = MakeBuffer(objPtr->interp,
			    sizeof(MJPEGBufferHeader)+8+length,
 			    "tcmMJPEGClipfile:MJPEG_CF_SendData.1");
	idx = 0;

	memcpy(bufPtr->data+idx, &hdrPtr, sizeof(MJPEGBufferHeader));
	idx += sizeof(MJPEGBufferHeader);

	dataSegCode = htonl(CM_DATA_SEGMENT);
	dataLength = htonl(length);
	memcpy(bufPtr->data+idx, &dataSegCode, 4);
	idx += 4;
	memcpy(bufPtr->data+idx, &dataLength, 4);
	idx += 4;

	MJPEG_FIP_ReadFrame(objPtr->fip, bufPtr->data+idx, frameNum);
	idx += length;

	/*
	 * Create sbl.
	 */
	sbPtr = MakeScatterBuffer(bufPtr, idx, NULL);
	sbl = MakeSBList(sbPtr, NULL);

	/*
	 * If out command is C, create scatter buffer list and call.
	 */
	if (objPtr->outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, sbl);	       
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree((char *) objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and 
	     * call using Tcl_Eval. 
	     */
	    sbStr = MakeSBListString(sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree((char *) sbStr);
	}
	FreeSBList(sbl);

	DetachBuffer(objPtr->interp, bufPtr, 
		     "tcmMJPEGClipfile:MJPEG_CF_SendData");

	/*
	 * Figure out when we need to send again, queue it up and return.
	 */
	if (speed == 0.0) {
	    return;
	}

	when = LTS_SystemTimeOf(objPtr->lts,
		(eTime-(objPtr->sendAhead*speed)));
	objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
					       "video", 
					       MJPEGSegmentSendData, objPtr, 
					       NULL, NULL, NULL, NULL, NULL);
	return;
    } 
    
    /*
     * Cycle time is not 0.0, so send multiple frames.
     * Find frames to be sent.
     */
    numToSend = 0;
    bwToSend = 0;
    if (fPickNum == 0) {
	fPickList = (FramePick *) ckalloc(sizeof(FramePick)*100);
	fPickNum = 100;
    }

    if (speed > 0.0) {
      if (now+objPtr->cycleTime*speed > objPtr->logicalEnd) {
	clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, objPtr->logicalEnd);
      } else
	clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, now+objPtr->cycleTime*speed);
    } else {
      if (now+objPtr->cycleTime*speed < objPtr->logicalStart) {
	clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, objPtr->logicalStart);
      } else
	clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, now+objPtr->cycleTime*speed);
    }

    numToSend = MJPEG_FIP_FrameListFromTime(objPtr->fip, clip_now, 
					    clip_cycle_end, &framelist);

    if (numToSend > fPickNum) {
	fPickNum = numToSend+1;
	fPickList = (FramePick *) ckrealloc((char *) fPickList,
					    sizeof(FramePick)*fPickNum);
    }

    for (i = 0; i<numToSend; i++) {
	MJPEG_FIP_TimeFromFrame(objPtr->fip, framelist[i], &sTime, &eTime);
	if (speed < 0) {
	    tmpTime = eTime;
	    eTime = sTime;
	    sTime = tmpTime;
	}
	
	fPickList[i].sTime = LOG_TIME_FROM_CLIP_TIME(objPtr, sTime);
	fPickList[i].eTime = LOG_TIME_FROM_CLIP_TIME(objPtr, eTime);
	fPickList[i].index = framelist[i];
	fPickList[i].size = MJPEG_FIP_LengthFromFrame(objPtr->fip, 
						      framelist[i]);
	fPickList[i].sendFlag = 1;
	bwToSend += fPickList[i].size;
    }
    ckfree((char *) framelist);

   /*
    * If number of frames to be sent greater than the number of
    * frames we can send (due to bandwidth or max frame rate),
    * mark low priority frames so we don't bother reading them.
    */
    frac_want_to_send = (objPtr->maxFrameRate * objPtr->cycleTime) -
      objPtr->extraFrameCounter;
    want_to_send = (int) (ceil(frac_want_to_send));

    if (objPtr->maxFrameRate != 0.0) {
	objPtr->extraFrameCounter = ((double) want_to_send) - frac_want_to_send;
    }

    bw_want_to_send = objPtr->bandwidth - ((int) ((((double) objPtr->extraBWCounter) /
						   objPtr->cycleTime)));
    curBw = (int) (((double) bwToSend) / objPtr->cycleTime);
    origNumToSend = numToSend;
    if ((numToSend > 0) &&
	(((numToSend > want_to_send) && objPtr->maxFrameRate) ||
	 ((curBw  > bw_want_to_send)    && objPtr->bandwidth &&
	  ((numToSend > 1) || (objPtr->extraBWCounter > 0))))) {
	
	int *orderVec;

	orderVec = (int *)ckalloc(numToSend*sizeof(int));
	MJPEG_Prioritize (numToSend, 0, 1, orderVec);
	while ((numToSend > 0) &&
	       (((numToSend > want_to_send) && objPtr->maxFrameRate) ||
		((curBw  > bw_want_to_send)    && objPtr->bandwidth &&
		 ((numToSend > 1) || (objPtr->extraBWCounter > 0))))) {
	    numToSend--;
	    fPickList[orderVec[numToSend]].sendFlag = 0;
	    bwToSend -= fPickList[orderVec[numToSend]].size;
	    curBw  = (int) (((double) bwToSend) / objPtr->cycleTime);
	}
	ckfree ((char *)orderVec);
    }

    /*
     * Adjust bandwidth counter.
     */
    
    if (objPtr->bandwidth > 0) {
      objPtr->extraBWCounter += bwToSend - ((int) (objPtr->bandwidth * 
						   objPtr->cycleTime));
    }

    if (numToSend > 0) {

	/*
	 * Adjust start and end times to take up space for frames
	 * that will not be sent. Do this only if frames were
	 * actually dropped.
	 */

	if (numToSend != origNumToSend) {
	    for (i=0; i<origNumToSend; i++) {
		if (fPickList[i].sendFlag == 1)
		  break;
	    }
	    if (i != 0) {
		fPickList[i].sTime = fPickList[0].sTime;
	    }
	    while (i<origNumToSend-1) {
		if (fPickList[i+1].sendFlag == 1) {
		    i++;
		} else {
		    j = i+1;
		    while(j<origNumToSend) {
			if (fPickList[j].sendFlag == 1) {
			    break;
			}
			j++;
		    }
		    if (j==origNumToSend) {
			fPickList[i].eTime = fPickList[j-1].eTime;
			break;
		    }

		    midpoint = (fPickList[i].eTime + fPickList[j].sTime)/2;
		    fPickList[i].eTime = midpoint;
		    fPickList[j].sTime = midpoint;
		    i = j;
		}
	    }
	}

	/*
	 * For each frame chosen, create buffers with headers and data.
	 */
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	sbl->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *) *
					      numToSend);
	sbl->count = numToSend;
	sbIdx = 0;
	
	for (i=0; i<origNumToSend; i++) {
	    if (fPickList[i].sendFlag == 1) {
		sTime = fPickList[i].sTime;
		eTime = fPickList[i].eTime;
		
		hdrPtr.hdrType = htonl(CM_MJPEG_BUF_HDR);
		hdrPtr.length = htonl(36);
		hdrPtr.width = htonl(objPtr->cfInfo->width);
		hdrPtr.height = htonl(objPtr->cfInfo->height);
		hdrPtr.device = htonl(objPtr->cfInfo->device);
		hdrPtr.subformat = htonl(objPtr->cfInfo->subformat);
		hdrPtr.qfactor = htonl(objPtr->cfInfo->qfactor);
		hdrPtr.logStart_s = htonl(floor(sTime));
		hdrPtr.logStart_ns = htonl(floor((sTime - floor(sTime))*ONE_BILLION));
		hdrPtr.logEnd_s = htonl(floor(eTime));
		hdrPtr.logEnd_ns = htonl(floor((eTime - floor(eTime))*ONE_BILLION));
		
		bufPtr = MakeBuffer(objPtr->interp,
				    sizeof(MJPEGBufferHeader)+8+fPickList[i].size,
				    "tcmMJPEGClipfile:MJPEG_CF_SendData.2");
		fPickList[i].bufPtr = bufPtr;
		idx = 0;
		

		memcpy(bufPtr->data+idx, &hdrPtr, sizeof(MJPEGBufferHeader));
		idx += sizeof(MJPEGBufferHeader);
		
		dataSegCode = htonl(CM_DATA_SEGMENT);
		dataLength = htonl(fPickList[i].size);
		memcpy(bufPtr->data+idx, &dataSegCode, 4);
		idx += 4;
		memcpy(bufPtr->data+idx, &dataLength, 4);
		idx += 4;

		MJPEG_FIP_ReadFrame(objPtr->fip, bufPtr->data+idx,
				    fPickList[i].index);
		idx += fPickList[i].size;
		
		sbl->sbs[sbIdx] = MakeScatterBuffer(bufPtr, idx, NULL);
		sbIdx++;
	    }
	}
	
	/*
	 * If out command is C, create scatter buffer list and call.
	 */
	if (objPtr->outCCmd != NULL) {
	    
	    
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree((char *) objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and
	     * call using Tcl_Eval. 
	     */
	    sbStr = MakeSBListString(sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree((char *) sbStr);
	}
	FreeSBList (sbl);
    }

    for (i=0; i<origNumToSend; i++) {
	if (fPickList[i].sendFlag == 1) {
	    DetachBuffer(objPtr->interp, fPickList[i].bufPtr,
			 "tcmMJPEGClipfile:MJPEG_CF_SendData");
	}
    }
  
    /*
     * Find next wake up time and set up atqueue.
     */
    lastETime = fPickList[origNumToSend-1].eTime;
    when = LTS_SystemTimeOf(objPtr->lts,
		    (lastETime-(objPtr->sendAhead*speed)));
    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
			       "video", MJPEGSegmentSendData, objPtr, NULL,
			       NULL, NULL, NULL, NULL);
    return;

done:
    MJPEG_FIP_FDClose(objPtr->fip);
    return;
}


/**************************************************************
 * MJPEGSegmentOpenClipfile
 *
 * Parameters:
 *    objPtr - pointer to Segment object.
 *    filename - character string file name.
 * 
 * Results:
 *    Returns a MJPEGFileInterface pointer valid for filename on success,
 *    NULL on failure.
 * 
 * Actions:
 *    Interfaces with MJPEG File Interface module to get access to file.
 *    Queries the module in order to fill the cfInfo structure in the object.
 *    
 ******************************************************************/

static MJPEGFileInterface *
MJPEGSegmentOpenClipfile(MJPEGSegment *objPtr, char *filename)
{
    MJPEGFileInterface *fip;

    fip = MJPEG_FIP_Open(filename);

    if (fip == NULL) 
      return NULL;

    objPtr->cfInfo = (MJPEGHeader *) ckalloc(sizeof(MJPEGHeader));

    objPtr->cfInfo->width = MJPEG_FIP_Width(fip);
    objPtr->cfInfo->height = MJPEG_FIP_Height(fip);
    objPtr->cfInfo->device = MJPEG_FIP_Device(fip);
    objPtr->cfInfo->subformat = MJPEG_FIP_Subformat(fip);
    objPtr->cfInfo->qfactor = MJPEG_FIP_QFactor(fip);

    return fip;
}  

/*************************************************************
 * MJPEGSegmentInfo
 * 
 * Parameters:
 *     objPtr - pointer to MJPEG Segment object.
 *     start - start time of segment that info is needed.
 *     end - end time of segment that info is needed.
 * 
 * Result:
 *     Standard tcl result.
 *
 * Action:
 *     Returns in result a list of two numbers: numFrames in segment, 
 *     and maxSize of those frames.
 *
 *************************************************************/

static int 
MJPEGSegmentInfo(MJPEGSegment *objPtr, double start, double end)
{
    int numFrames;
    int maxSize;
    char retStr[500];

    if (objPtr->fip == NULL) {
	sprintf(retStr, "%d %d", 0, 0);
	Tcl_SetResult(objPtr->interp, retStr, TCL_VOLATILE);
	return TCL_OK;
    }

    MJPEG_FIP_Info(objPtr->fip, start, end, &numFrames, &maxSize);

    sprintf(retStr, "%d %d", numFrames, maxSize);
    Tcl_SetResult(objPtr->interp, retStr, TCL_VOLATILE);
    return TCL_OK;
}


   
