/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMJPEGPlay.c,v 1.48 1996/02/06 22:41:43 kpatel Exp $
 *
 * Implementation of the "mjpegPlay" object.  Takes incoming
 * frames, manages a queue of playable frames, then uses a
 * jpeg decoding device to decode and display frames
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <math.h>

#ifdef NO_STRING_H
#   include "compat/string.h"
#else
#   include <string.h>
#endif

#include <sys/types.h>
#include <netinet/in.h>
#include <dp.h>
#include "cmEvery.h"
#include "cmMJPEGPlay.h"
#include "tcmEvent.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmBind.h"
#include "cmUtil.h"

#ifndef DEFAULT_JPEG_PLAY_DEV_TYPE
#    define DEFAULT_JPEG_PLAY_DEV_TYPE "software"
#endif

#define ONE_BILLION 1000000000.0

static int nextMJPEGResourceNum = 0;	/* Variable for next unique object
					 * name. */

/*
 * Prototypes for routines used internally
 */
static MJPEGResource *CreateNewMJPEGResource	(Tcl_Interp *interp,
						 char *name);
static void ChangeDeviceType			(MJPEGResource *objPtr,
						 char *devtype);
static int DestroyMJPEGResource			(MJPEGResource *objPtr);
static int MJPEGResourceObjCmd			(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc, char **argv);
static int MJPEG_RSRC_Accept			(MJPEGResource *objPtr,
						 SB_List * sbl);
static int MJPEG_RSRC_Ready			(MJPEGResource *objPtr);
static int MJPEG_RSRC_Unready			(MJPEGResource *objPtr);
static int MJPEG_RSRC_Stats			(MJPEGResource *objPtr);
static int MJPEG_RSRC_Refresh			(MJPEGResource *objPtr);
static void MJPEG_RSRC_LTSChanged		(ClientData clientData);
static int MJPEG_RSRC_DoDecode			(MJPEGResource *objPtr);
static void MJPEG_RSRC_FreeFrame		(MJPEGFrame *framePtr);
static int MJPEGResourceCreateCmd		(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc, char **argv);
static void MJPEG_RSRC_ContinueDecode		(MJPEGFrame *framePtr);
static int ConfigureMJPEGResource		(MJPEGResource *objPtr,
						 int argc, char **argv);
static int MJPEG_RSRC_DeviceList                (MJPEGResource *objPtr);
static char * GetHeight(Tcl_Interp *interp, ClientData clientdata);
static char * GetWidth(Tcl_Interp *interp, ClientData clientdata);
static char * GetDataLen(Tcl_Interp *interp, ClientData clientdata);
static char * GetLogEnd(Tcl_Interp *interp, ClientData clientdata);
static char * GetLogStart(Tcl_Interp *interp, ClientData clientdata);
static char * GetCumulLenDecoded(Tcl_Interp *interp, ClientData clientdata);


/************************************************************
 *
 * MJPEGResourceCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the "mjpegResource" Tcl command.
 *     A new object of type MJPEGResource is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
MJPEGResourceCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    MJPEGResource *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextMJPEGResourceNum++;
    sprintf(new_name, "%smjpegPlay%d", argv[1], nextMJPEGResourceNum);

    new = CreateNewMJPEGResource(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MJPEGResourceObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureMJPEGResource(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MJPEGResourceObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     MJPEGResource object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
MJPEGResourceObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    MJPEGResource *objPtr = (MJPEGResource *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureMJPEGResource(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyMJPEGResource(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl = NULL;
	int result;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}

	/*
	 * Parse the ScatterBufferList
	 */
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	  ckfree((char *) sbl);
	  Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
			   argv[2],  (char *)NULL);
	  return TCL_ERROR;
	}

	result = MJPEG_RSRC_Accept(objPtr, sbl);
	if (sbl != NULL) {
	    FreeSBList(sbl);
	}
	return result;

    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return MJPEG_RSRC_Ready(objPtr);

    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return MJPEG_RSRC_Unready(objPtr);

    } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
	return MJPEG_RSRC_Stats(objPtr);

    } else if ((c == 'r') && (strncmp("refresh", argv[1], len) == 0)) {
	return MJPEG_RSRC_Refresh(objPtr);

    } else if ((c == 'd') && (strncmp("deviceList", argv[1], len) == 0)) {
	return MJPEG_RSRC_DeviceList(objPtr);
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ",
		argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
		"\n<command> can be one of:\n",
		"  configure ?-slot ?value??\n",
		"  destroy\n",
		"  accept <ScatterBufferList>\n",
		"  ready\n",
		"  unready\n",
		"  stats\n",
		"  refresh\n",
		"  deviceList\n",
		(char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ChangeDeviceType --
 *
 *	Sets the device type for the MJPEGReseource.  Called whenever
 *	the "devtype" slot is changed, and on initialization of the
 *	resource.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The old device is shut down, if appropriate, and the new
 *	device is created.
 *
 * Parameters:
 *	MJPEGResource *objPtr	Pointer to resource object.
 *	char *devtype		String giving name of new device
 *
 *--------------------------------------------------------------
 */
static void
ChangeDeviceType (MJPEGResource *objPtr, char *devtype)
{
    int len;
    char c;

    len = strlen(devtype);
    c = devtype[0];

    /*
     * If there's no change, return
     */
    if ((objPtr->devtype != NULL) &&
	(strncmp (objPtr->devtype, devtype, len) == 0)) {
	return;
    }

    /*
     * Shut down old device, if necessary, ignoring any errors it
     * returns.
     */
    MJPEG_RSRC_Unready(objPtr);
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
	objPtr->devtype = NULL;;
    }
    if (objPtr->devicePtr != NULL) {
	objPtr->devicePtr->shutdownProc(objPtr);
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
	objPtr->inited = 0;
    }

    /*
     * Create new device.
     */
    if ((c == 'p') && (strncmp("parallax", devtype, len) == 0)) {
	objPtr->devicePtr = GetParallaxDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("parallax")+1);
	strcpy(objPtr->devtype, "parallax");
    } else if ((c == 'j') && (strncmp("j300", devtype, len) == 0)) {
	objPtr->devicePtr = GetJ300Device(objPtr);
	objPtr->devtype = ckalloc(strlen("j300")+1);
	strcpy(objPtr->devtype, "j300");
    } else if ((c == 's') && (strncmp("software", devtype, len) == 0)) {
	objPtr->devicePtr = GetSoftwareJpegDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("software")+1);
	strcpy(objPtr->devtype, "software");
    } else if ((c == 's') && (strncmp("sunvideo", devtype, len) == 0)) {
        objPtr->devicePtr = GetSunVideoJpegDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("sunvideo")+1);
	strcpy(objPtr->devtype, "sunvideo");
    } else if ((c == 'c') && (strncmp("cosmo", devtype, len) == 0)) {
        objPtr->devicePtr = GetCosmoDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("cosmo")+1);
	strcpy(objPtr->devtype, "cosmo");
    } else {
	objPtr->devicePtr = GetDefaultJpegDevice(objPtr);
	objPtr->devtype = ckalloc(strlen(DEFAULT_JPEG_PLAY_DEV_TYPE)+1);
	strcpy(objPtr->devtype, DEFAULT_JPEG_PLAY_DEV_TYPE);
    }
}

/************************************************************
 *
 * CreateNewMJPEGResource
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static MJPEGResource *
CreateNewMJPEGResource(Tcl_Interp *interp, char *name)
{
    MJPEGResource *new;
    char *acceptName;
    char **event_list, **subst_list;
    CM_BindProc **subst_procs;
    ClientData *subst_cds;
    int i;

    new = (MJPEGResource *) ckalloc(sizeof(MJPEGResource));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);

    new->width = 0;
    new->height = 0;
    new->xpos = -1;
    new->ypos = -1;
    new->squeeze = 0;
    new->devtype = NULL;
    new->devicePtr = NULL;
    new->xid = 0;
    new->lts = NULL;
    new->ltsTriggerNum = -1;
    new->ltsName = NULL;
    new->maxFrameRate = 0;
    new->statStartTime = 0.0;
    new->missed = 0;
    new->late = 0;
    new->played = 0;
    new->bytes = 0;
    new->ready = 0;
    new->inited = 0;
    new->inDecode = 0;
    new->queue = NULL;
    new->cntrlEveryHandle = NULL;
    new->destroyFlag = 0;
    new->at_tokens = NULL;
    new->num_at_tokens = 0;

    /*
     * Set up bindings with binding mechanism.
     */

    event_list = (char **) ckalloc(5*sizeof(char *));
    event_list[0] = (char *) ckalloc(strlen("<FRAME_DROPPED>")+1);
    strcpy(event_list[0], "<FRAME_DROPPED>");
    event_list[1] = (char *) ckalloc(strlen("<FRAME_RECEIVED>")+1);
    strcpy(event_list[1], "<FRAME_RECEIVED>");
    event_list[2] = (char *) ckalloc(strlen("<FRAME_PLAYED>")+1);
    strcpy(event_list[2], "<FRAME_PLAYED>");
    event_list[3] = (char *) ckalloc(strlen("<FRAME_DECODE_BEGIN>")+1);
    strcpy(event_list[3], "<FRAME_DECODE_BEGIN>");
    event_list[4] = (char *) ckalloc(strlen("<FRAME_DECODE_DONE>")+1);
    strcpy(event_list[4], "<FRAME_DECODE_DONE>");

    subst_list = (char **) ckalloc(6*sizeof(char *));
    subst_procs = (CM_BindProc **) ckalloc(6*sizeof(CM_BindProc *));
    subst_cds = (ClientData *) ckalloc(6*sizeof(ClientData));

    subst_list[0] = (char *) ckalloc(2);
    (subst_list[0])[0] = 's';
    (subst_list[0])[1] = '\0';
    subst_procs[0] = GetLogStart;
    subst_cds[0] = (ClientData) new;

    subst_list[1] = (char *) ckalloc(2);
    (subst_list[1])[0] = 'e';
    (subst_list[1])[1] = '\0';
    subst_procs[1] = GetLogEnd;
    subst_cds[1] = (ClientData) new;

    subst_list[2] = (char *) ckalloc(2);
    (subst_list[2])[0] = 'l';
    (subst_list[2])[1] = '\0';
    subst_procs[2] = GetDataLen;
    subst_cds[2] = (ClientData) new;

    subst_list[3] = (char *) ckalloc(2);
    (subst_list[3])[0] = 'w';
    (subst_list[3])[1] = '\0';
    subst_procs[3] = GetWidth;
    subst_cds[3] = (ClientData) new;

    subst_list[4] = (char *) ckalloc(2);
    (subst_list[4])[0] = 'h';
    (subst_list[4])[1] = '\0';
    subst_procs[4] = GetHeight;
    subst_cds[4] = (ClientData) new;

    subst_list[5] = (char *) ckalloc(2);
    (subst_list[5])[0] = 'b';
    (subst_list[5])[1] = '\0';
    subst_procs[5] = GetCumulLenDecoded;
    subst_cds[5] = (ClientData) new;

    new->bindToken = CreateCMBinding(interp,
				     new->name,
				     5, event_list,
				     6, subst_list, subst_procs, subst_cds);

    for (i=0; i<5; i++) {
      ckfree(event_list[i]);
    }
    ckfree((char *) event_list);

    for (i=0; i<6; i++) {
      ckfree(subst_list[i]);
    }
    ckfree((char *) subst_list);
    ckfree((char *) subst_procs);
    ckfree((char *) subst_cds);

    ChangeDeviceType (new, DEFAULT_JPEG_PLAY_DEV_TYPE);

    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+9);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) MJPEG_RSRC_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    return new;
}


/************************************************************
 *
 * DestroyMJPEGResource
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

static int
DestroyMJPEGResource(MJPEGResource *objPtr)
{
    int result = TCL_OK;
    int i, devReturn;

    /*
     * If not being called by DecodeDone first delete Tcl command.
     */
    if (objPtr->destroyFlag != 1) {
	Tcl_DeleteCommand(objPtr->interp, objPtr->name);
    }

    /*
     * Make unready, cancel any DoDecode when idle calls.
     */
    objPtr->ready = 0;
    Tcm_CancelIdleCall("video", (Tk_IdleProc *) MJPEG_RSRC_DoDecode, objPtr);

    /*
     * If in decode, set destroy flag and return. Let DecodeDone do destroy.
     */
    if (objPtr->inDecode) {
	objPtr->destroyFlag = 1;
	return TCL_OK;
    }

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }

    if (objPtr->queue != NULL) {
	MJPEGFrame *framePtr, *nextPtr;

	framePtr = objPtr->queue;

	while (framePtr != NULL) {
	    nextPtr = framePtr->next;
	    MJPEG_RSRC_FreeFrame(framePtr);
	    framePtr = nextPtr;
	}
	objPtr->queue = NULL;
    }

    /*
     * Shut down the device
     */
    if (objPtr->devicePtr != NULL) {
	devReturn = objPtr->devicePtr->shutdownProc(objPtr);
	if (devReturn != 0) {
	    result = TCL_ERROR;
	    if (objPtr->devicePtr->errorProc != NULL) {
		Tcl_AppendResult(objPtr->interp,
			objPtr->devicePtr->errorProc(objPtr, devReturn),
			(char *) NULL);
	    }
	}
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
    }

    /*
     * Clean up lts
     */
    if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
        Tk_Release(objPtr->lts);
        objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    /*
     * Clean up previously scheduled display requests
     */
    if (objPtr->at_tokens != NULL) {
	for (i=0; i<objPtr->num_at_tokens; i++) {
	    if (objPtr->at_tokens[i] > 0) {
		AtEvent *at_event;
		int token;

		token = objPtr->at_tokens[i];
		at_event = Tcm_FindAtEvent(token);
		if (token && (at_event != NULL)) {
		    MJPEG_RSRC_FreeFrame((MJPEGFrame *) at_event->fireClientData);
		    Tcm_DeleteAtHandler(token);
		    objPtr->at_tokens[i] = 0;
		}
	    }
	}
	ckfree((char *) objPtr->at_tokens);
    }

    /*
     * Unregister with bind mechanism.
     */

    DestroyCMBinding(objPtr->interp, objPtr->bindToken);

    ckfree((char *) objPtr);
    return result;
}


/************************************************************
 *
 * ConfigureMJPEGResource
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

static int
ConfigureMJPEGResource(MJPEGResource * objPtr, int argc, char **argv)
{

    Tcl_Interp *interp;
    int cntr;
    char str[50];

    interp = objPtr->interp;

    /*
     * Case 1: <obj> configure
     */
    if (argc == 0) {
	sprintf(str, "{%d}", objPtr->width);
	Tcl_AppendResult(interp, "{", "-width ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->height);
	Tcl_AppendResult(interp, "{", "-height ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->xpos);
	Tcl_AppendResult(interp, "{", "-xpos ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->ypos);
	Tcl_AppendResult(interp, "{", "-ypos ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->squeeze);
	Tcl_AppendResult(interp, "{", "-squeeze ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->xid);
	Tcl_AppendResult(interp, "{", "-xid ", "0 ", str, "} ",
			 (char *)NULL);
	Tcl_AppendResult(interp, "{", "-device ", "{", DEFAULT_JPEG_PLAY_DEV_TYPE, "} ",
			 objPtr->devtype, "} ", (char *)NULL);

	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{}", "} ",
			 (char *)NULL);
	} else {
	  Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{",
			   objPtr->ltsName, "}", "} ", (char *) NULL);
	}
	sprintf(str, "{%d}", objPtr->maxFrameRate);
	Tcl_AppendResult(interp, "{", "-maxFrameRate ", "0 ", str, "} ",
			 (char *) NULL);

	/*
	 * Give the device a chance to add any options of its own.
	 */

	if (objPtr->devicePtr != NULL) {
	    if (objPtr->devicePtr->configProc != NULL) {
		objPtr->devicePtr->configProc(objPtr, NULL, NULL);
	    }
	}
	return TCL_OK;
    }

    /*
     * Case 2: <obj> configure <-slot>
     */
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 'w') && (strncmp("-width", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->width);
	    Tcl_AppendResult(interp, "-width ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'h') && (strncmp("-height", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->height);
	    Tcl_AppendResult(interp, "-height ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'x') && (strncmp("-xpos", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->xpos);
	    Tcl_AppendResult(interp, "-xpos ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'y') && (strncmp("-ypos", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->ypos);
	    Tcl_AppendResult(interp, "-ypos ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 's') && (strncmp("-squeeze", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->squeeze);
	    Tcl_AppendResult(interp, "-squeeze ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'x') && (strncmp("-xid", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->xid);
	    Tcl_AppendResult(interp, "-xid ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-device", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-device ", "{", DEFAULT_JPEG_PLAY_DEV_TYPE, "} ", 
			     objPtr->devtype,
			     (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	  if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "-lts ", "{} ", "{}", (char *) NULL);
	  } else {
	    Tcl_AppendResult(interp, "-lts ", "{} ", "{", objPtr->ltsName, "}", (char *) NULL);
	  }
	    return TCL_OK;
	} else if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->maxFrameRate);
	    Tcl_AppendResult(interp, "-maxFrameRate ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((objPtr->devicePtr != NULL) && 
		   (objPtr->devicePtr->configProc != NULL)) {
	    if (objPtr->devicePtr->configProc(objPtr, argv[0], NULL) != 0) 
	      goto bad_option;
	    return TCL_OK;
	}
      bad_option:
	Tcl_AppendResult(interp, "bad configuration option: ", argv[0], (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: <obj> configure <-slot> <newValue>
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;

	/*
	 * Make resource unready since something is changing.
	 */

	MJPEG_RSRC_Unready(objPtr);

	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	if ((c == 'w') && (strncmp("-width", argv[cntr], len) == 0)) {
	    int newWidth;

	    cntr++;
	    if (cntr == argc) {
noValueError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newWidth) != TCL_OK) {
badValueError:
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->width = newWidth;
	} else if ((c == 'h') && (strncmp("-height", argv[cntr], len) == 0)) {
	    int newHeight;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newHeight) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->height = newHeight;
	} else if ((c == 'x') && (strncmp("-xpos", argv[cntr], len) == 0)) {
	    int newXpos;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newXpos) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->xpos = newXpos;
	} else if ((c == 'y') && (strncmp("-ypos", argv[cntr], len) == 0)) {
	    int newYpos;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newYpos) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->ypos = newYpos;
	} else if ((c == 's') && (strncmp("-squeeze", argv[cntr], len) == 0)) {
	    int newSqueeze;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newSqueeze) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->squeeze = newSqueeze;
	} else if ((c == 'x') && (strncmp("-xid", argv[cntr], len) == 0)) {
	    int newXid;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newXid) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->xid = newXid;
	} else if ((c == 'd') && (strncmp("-device", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    ChangeDeviceType(objPtr, argv[cntr]);
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (strlen(argv[cntr]) == 0) {
	        newLts = NULL;
	    }
	    else if ((newLts = Tcm_FindAtom(interp, argv[cntr])) == NULL) {
		goto badValueError;
	    }
	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }

            /*
             * Release our lock on the old lts...
             */
            if (objPtr->lts != NULL) {
	        if (objPtr->ltsTriggerNum != -1) {
		    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
		    objPtr->ltsTriggerNum = -1;
	        }
                Tk_Release(objPtr->lts);
                objPtr->lts = NULL;
            }

	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
	        objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts, 
					 LTS_SPEED_FLAG | LTS_VALUE_FLAG |
					 LTS_EARLY_PHASE, 
					 MJPEG_RSRC_LTSChanged,
					 (ClientData) objPtr);
	        objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
	        strcpy(objPtr->ltsName, argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }
	} else if ((c == 'm') &&
		   (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
	    int newMaxFrameRate;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->maxFrameRate = newMaxFrameRate;
	} else if ((objPtr->devicePtr != NULL) && 
		   (objPtr->devicePtr->configProc != NULL)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]) != 0) {
		return TCL_ERROR;
	    }
	}
	cntr++;
    }
    return TCL_OK;
}


/************************************************************
 *
 * MJPEG_RSRC_Ready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MJPEG_RSRC_Ready(MJPEGResource * objPtr)
{
    int init_res;

    /*
     * Can only be made ready if
     *  1. We have a device.
     *  2. We have an lts
     *  3. We have a window
     */
    if (objPtr->devicePtr == NULL) {
        Tcl_AppendResult(objPtr->interp, "Can't make MJPEG resource \"",
                objPtr->name, "\" ready: device type is not set.", (char *)NULL);
        return TCL_ERROR;
    }
    if (objPtr->lts == NULL) {
        Tcl_AppendResult(objPtr->interp, "Can't make MJPEG resource \"",
                objPtr->name, "\" ready: no lts has been attached.", (char *)NULL);
        return TCL_ERROR;
    }
    if (objPtr->xid == 0) {
        Tcl_AppendResult(objPtr->interp, "Can't make MJPEG resource \"",
                objPtr->name, "\" ready: output window not specified.",
		(char *)NULL);
        return TCL_ERROR;
    }


    /*
     * If not inited, do initProc, catch result.
     */
    init_res = 0;
    if (!objPtr->inited) {
      if (objPtr->devicePtr->initProc != NULL) {
	init_res = objPtr->devicePtr->initProc(objPtr);
      }
      if (init_res == 0) 
	objPtr->inited = 1;
    }
    else {
      /*
       * Otherwise, do reinitProc, catch result.
       */
      if (objPtr->devicePtr->reinitProc != NULL) {
	init_res = objPtr->devicePtr->reinitProc(objPtr);
      }
    }
      
    /*
     * If bad result, get error string, return error.
     */
    if (init_res != 0) {
      char *errStr;

      if (objPtr->devicePtr->errorProc != NULL) {
	  errStr = objPtr->devicePtr->errorProc(objPtr, init_res);
      } else {
	  errStr = "Error with device initialization.";
      }

      Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
      return TCL_ERROR;
    }

    objPtr->ready = 1;

    /*
     * If queue is not empty, start a decode.
     */

    if (objPtr->queue != NULL) {
      MJPEG_RSRC_DoDecode(objPtr);
    }

    /*
     * Reset stat stuff.
     */

    objPtr->statStartTime = Tcm_ReadSysClock();
    objPtr->missed = objPtr->played = objPtr->late = objPtr->bytes = 0;

    return TCL_OK;
}


/************************************************************
 *
 * MJPEG_RSRC_LTSChanged
 *
 * Parameters:
 *     clientdata - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *      Called when the LTS changes value.  Dequeues any previously
 *      queued play requests, frees any memory we're holding, and
 *      returns.
 *
 ************************************************************/
static void
MJPEG_RSRC_LTSChanged(ClientData clientData)
{
    MJPEGResource *objPtr = (MJPEGResource *) clientData;
    int i;

    /*
     * If not ready, do nothing.
     */
    if (!objPtr->ready) {
	return;
    }

    /*
     * Blow away "at" queue.
     */
    for (i=0; i<objPtr->num_at_tokens; i++) {
	if (objPtr->at_tokens[i] > 0) {
	    AtEvent *at_event;
	    int token;

	    token = objPtr->at_tokens[i];
	    at_event = Tcm_FindAtEvent(token);
	    if (token && (at_event != NULL)) {
		MJPEG_RSRC_FreeFrame((MJPEGFrame *) at_event->fireClientData);
		Tcm_DeleteAtHandler(token);
		objPtr->at_tokens[i] = 0;
	    }
	}
    }

    /*
     * Blow away my frame queue.
     */
    if (objPtr->queue != NULL) {
	MJPEGFrame *framePtr, *nextPtr;

	framePtr = objPtr->queue;

	while (framePtr != NULL) {
	    nextPtr = framePtr->next;
	    MJPEG_RSRC_FreeFrame(framePtr);
	    framePtr = nextPtr;
	}
	objPtr->queue = NULL;
    }

    /*
     * Reset all the statistics stuff.
     */
    objPtr->statStartTime = Tcm_ReadSysClock();
    objPtr->missed = objPtr->late = objPtr->played = objPtr->bytes = 0;
}

/************************************************************
 *
 * MJPEG_RSRC_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
MJPEG_RSRC_Accept(MJPEGResource *objPtr, SB_List *sbl)
{
    int i;
    MJPEGFrame *newFrame, *framePtr, *prevPtr, *nextPtr;
    double speed, now, eTime, sTime;
    signed char *sb_data;
    Buffer *bufPtr;

    if (!objPtr->ready) {
	return TCL_OK;
    }

    speed = LTS_Speed(objPtr->lts);
    now = LTS_Value(objPtr->lts);

    if (sbl == NULL) return TCL_OK;
    if (sbl->count < 1) return TCL_OK;

    /*
     * Increment reference count on buffers passed in
     */
    AttachToAllBuffers(objPtr->interp, sbl,
			"tcmMJPEGResource:MJPEG_RSRC_Accept");

    /*
     * For each frame in sb list, create frame structure,
     * fill in info and add to queue.
     */

    for (i=0; i<sbl->count; i++) {
	newFrame = (MJPEGFrame *) ckalloc(sizeof(MJPEGFrame));

	sb_data = GetFromSB(sbl->sbs[i], CM_MJPEG_BUF_HDR, &bufPtr, NULL);

	if (sb_data == NULL) {
	    ckfree((char *) newFrame);
	    continue;
	}
	memcpy(&(newFrame->frameHeader), sb_data-8, sizeof(MJPEGBufferHeader));
	newFrame->frameHeader.hdrType = ntohl(newFrame->frameHeader.hdrType);
	newFrame->frameHeader.length = ntohl(newFrame->frameHeader.length);
	newFrame->frameHeader.width = ntohl(newFrame->frameHeader.width);
	newFrame->frameHeader.height = ntohl(newFrame->frameHeader.height);
	newFrame->frameHeader.device = ntohl(newFrame->frameHeader.device);
	newFrame->frameHeader.subformat = ntohl(newFrame->frameHeader.subformat);
	newFrame->frameHeader.qfactor = ntohl(newFrame->frameHeader.qfactor);
	newFrame->frameHeader.logEnd_s = ntohl(newFrame->frameHeader.logEnd_s);
	newFrame->frameHeader.logEnd_ns = ntohl(newFrame->frameHeader.logEnd_ns);
	newFrame->frameHeader.logStart_s = ntohl(newFrame->frameHeader.logStart_s);
	newFrame->frameHeader.logStart_ns = ntohl(newFrame->frameHeader.logStart_ns);
	
	eTime = ((double )newFrame->frameHeader.logEnd_s) + 
	  (((double) newFrame->frameHeader.logEnd_ns) / ONE_BILLION);
	sTime = ((double )newFrame->frameHeader.logStart_s) + 
	  (((double) newFrame->frameHeader.logStart_ns) / ONE_BILLION);
	
	newFrame->data = GetFromSB(sbl->sbs[i], CM_DATA_SEGMENT, &bufPtr, 
				   &newFrame->dataLength);
	if (newFrame->data == NULL) {
	    ckfree((char *) newFrame);
	    continue;
	}

	newFrame->deviceData = NULL;
	newFrame->at_token = -1;
	newFrame->objPtr = objPtr;

	newFrame->dataBuffer = bufPtr;
	newFrame->logStart = sTime;
	newFrame->logEnd = eTime;
	newFrame->sb = CopyScatterBuffer(sbl->sbs[i]);

#if 0
	{
	  char logStr[100];
	  sprintf(logStr, "MJPEGFrame AddingToQueue %x %x", newFrame,
		  newFrame->dataBuffer->data);
	  Tcm_Log(logStr);
	}
#endif

	/*
	 * Generate FRAME_RECEIVED event.
	 */
	
	objPtr->bindFramePtr = newFrame;
	ExecuteCMBinding(objPtr->interp, objPtr->bindToken, "<FRAME_RECEIVED>");

	if (objPtr->queue == NULL) {
	    objPtr->queue = newFrame;
	    newFrame->next = newFrame->prev = NULL;
	} else if (speed == 0.0) {
	    /* Remove all old frames */
	    framePtr = objPtr->queue;
	    while (framePtr != NULL) {
		nextPtr = framePtr->next;
		MJPEG_RSRC_FreeFrame(framePtr);
		framePtr = nextPtr;
	    }
	    objPtr->queue = newFrame;
	    newFrame->next = newFrame->prev = NULL;
	} else {
	    framePtr = objPtr->queue;
	    prevPtr = NULL;
	    while (framePtr != NULL) {
		if (speed < 0.0) {
		    if (framePtr->logEnd > newFrame->logEnd) {
			prevPtr = framePtr;
			framePtr = framePtr->next;
		    } else {
			break;
		    }
		} else {
		    if (framePtr->logEnd < newFrame->logEnd) {
			prevPtr = framePtr;
			framePtr = framePtr->next;
		    } else {
			break;
		    }
		}
	    }

	    if (framePtr == NULL) {
		prevPtr->next = newFrame;
		newFrame->prev = prevPtr;
		newFrame->next = NULL;
	    } else if (prevPtr == NULL) {
		newFrame->next = framePtr;
		framePtr->prev = newFrame;
		objPtr->queue = newFrame;
		newFrame->prev = NULL;
	    } else {
		prevPtr->next = newFrame;
		newFrame->prev = prevPtr;
		framePtr->prev = newFrame;
		newFrame->next = framePtr;
	    }
	}
    }
#if 0
    {
    int count;
    MJPEGFrame *f;
    char logStr[100];

    count = 0;
    f = objPtr->queue;
    while (f != NULL) {
	count++;
	f=f->next;
    }
    sprintf(logStr, "MJPEGAccept: queue length = %d", count);
    Tcm_Log(logStr);
    }
#endif

    /*
     * If no decode in process, start a decode
     */
    if (!objPtr->inDecode) {
	MJPEG_RSRC_DoDecode(objPtr);
    }

    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_RSRC_Unready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MJPEG_RSRC_Unready(MJPEGResource * objPtr)
{
  int i;

  objPtr->ready = 0;

  /*
   * Blow away "at" queue.
   */
  for (i =0; i<objPtr->num_at_tokens; i++) {
      if (objPtr->at_tokens[i] > 0) {
	  AtEvent *at_event;
	  int token;

	  token = objPtr->at_tokens[i];
	  at_event = Tcm_FindAtEvent(token);
	  if (token && (at_event != NULL)) {
	      MJPEG_RSRC_FreeFrame((MJPEGFrame *) at_event->fireClientData);
	      Tcm_DeleteAtHandler(token);
	      objPtr->at_tokens[i] = 0;
	  }
      }
  }

  /*
   * Blow away my frame queue.
   */
  if (objPtr->queue != NULL) {
      MJPEGFrame *framePtr, *nextPtr;

      framePtr = objPtr->queue;
      
      while (framePtr != NULL) {
	  nextPtr = framePtr->next;
	  MJPEG_RSRC_FreeFrame(framePtr);
	  framePtr = nextPtr;
      }
      objPtr->queue = NULL;
  }
  
  return TCL_OK;
}

/************************************************************
 *
 * MJPEG_RSRC_Stats
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MJPEG_RSRC_Stats(MJPEGResource *objPtr)
{
    sprintf (objPtr->interp->result, "%d %d %d %d",
 	     objPtr->missed, objPtr->late, objPtr->played, objPtr->bytes);

    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_RSRC_DoDisplay
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MJPEG_RSRC_DoDisplay(MJPEGFrame *framePtr)
{
  MJPEGResource *objPtr;
  double now, speed;

  objPtr = framePtr->objPtr;

  now = LTS_Value(objPtr->lts);
  speed = LTS_Speed(objPtr->lts);

#if 0
  {
    char logStr[100];
    sprintf(logStr, "MJPEGFrame DoDisplay %x %x", framePtr, framePtr->dataBuffer->data);
    Tcm_Log(logStr);
  }
#endif

  /*
   * Collect stats.
   */
  objPtr->played++;
      
  /*
   * Call device's display procedure.
   */
  if (objPtr->devicePtr != NULL) {
    if (objPtr->devicePtr->displayProc != NULL) {
      objPtr->devicePtr->displayProc(framePtr);
    }
  }

  /*
   * Generate FRAME_PLAYED event.
   */

  objPtr->bindFramePtr = framePtr;
  ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		   "<FRAME_PLAYED>");

  MJPEG_RSRC_FreeFrame(framePtr);

  return TCL_OK;
}


/************************************************************
 *
 * MJPEG_RSRC_FreeFrame
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static void
MJPEG_RSRC_FreeFrame(MJPEGFrame *framePtr)
{
  int i;

  /*
   * Call device frame freeer.
   */
  if (framePtr->objPtr->devicePtr->freeFrameProc != NULL) {
      framePtr->objPtr->devicePtr->freeFrameProc(framePtr);
  }

#if 0
  {
    char logStr[100];
    sprintf(logStr, "MJPEGFrame FreeingFrame %x %x", framePtr, framePtr->dataBuffer->data);
    Tcm_Log(logStr);
  }
#endif


  /*
   * Detach from buffer.
   */

  for (i=0; i<framePtr->sb->count; i++) {
      DetachBuffer(framePtr->objPtr->interp, framePtr->sb->buffers[i],
		   "tcmMJPEGResource:MJPEG_RSRC_FreeFrame");
  }

  FreeScatterBuffer(framePtr->sb);
  framePtr->sb = NULL;

  if (framePtr->at_token >= 0) {
      framePtr->objPtr->at_tokens[framePtr->at_token] = 0;
  }

  ckfree((char *) framePtr);

  return;
}

/************************************************************
 *
 * MJPEG_RSRC_GetEligibleFrame
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Acgtion:
 *
 ************************************************************/

static MJPEGFrame *
MJPEG_RSRC_GetEligibleFrame(MJPEGResource *objPtr)
{
  double now, speed;
  MJPEGFrame *chosenFrame, *framePtr, *nextPtr;

  if (objPtr->queue ==  NULL) return NULL;

  now = LTS_Value(objPtr->lts);
  speed = LTS_Speed(objPtr->lts);

  if (objPtr->devicePtr != NULL) {
      if (objPtr->devicePtr->nextFrameProc != NULL) {
	  if (objPtr->devicePtr->nextFrameProc(objPtr, objPtr->queue, &chosenFrame) 
	      != 0) {
	      return NULL;
	  }
      } else {
	  chosenFrame = objPtr->queue;
	  while (chosenFrame->next != NULL) {
	      if (speed < 0.0) {
		  if (chosenFrame->logEnd < now) break;
	      } else {
		  if (chosenFrame->logEnd > now) break;
	      }
	      chosenFrame = chosenFrame->next;
	  }
      }
  }

  if (chosenFrame == NULL)
    return NULL;

#if 0
  {
    char logStr[100];
    sprintf(logStr, "MJPEGFrame Chosen %x %x", chosenFrame, chosenFrame->dataBuffer->data);
    Tcm_Log(logStr);
  }
#endif

  if (chosenFrame->prev == NULL) {
      objPtr->queue = chosenFrame->next;
      chosenFrame->next = NULL;
      if (objPtr->queue != NULL) 
	objPtr->queue->prev = NULL;
  } else if (chosenFrame->next == NULL) {
      chosenFrame->prev->next = NULL;
      chosenFrame->prev = NULL;
  } else {
      chosenFrame->prev->next = chosenFrame->next;
      chosenFrame->next->prev = chosenFrame->prev;
      chosenFrame->next = chosenFrame->prev = NULL;
  }

  /*
   * Clean out queue of late frames.
   */

  framePtr = objPtr->queue;
  nextPtr = NULL;

  while (framePtr != NULL && framePtr->next != NULL) {
      if (speed < 0.0) {
	  if (framePtr->logEnd > now) {
	      nextPtr = framePtr->next;
	      if (framePtr->prev == NULL) {
		  objPtr->queue = nextPtr;
		  if (objPtr->queue != NULL) 
		      objPtr->queue->prev = NULL;
	      } else if (nextPtr == NULL) {
		  framePtr->prev->next = NULL;
	      } else {
		  framePtr->prev->next = nextPtr;
		  nextPtr->prev = framePtr->prev;
	      }
	      objPtr->late++;

	      /*
	       * Generate FRAME_DROPPED event.
	       */

	      objPtr->bindFramePtr = framePtr;
	      ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
			       "<FRAME_DROPPED>");

#if 0
	      {
		char logStr[100];
		sprintf(logStr, "MJPEGFrame LatePurge %x %x", framePtr, framePtr->dataBuffer->data);
		Tcm_Log(logStr);
	      }
#endif

	      MJPEG_RSRC_FreeFrame(framePtr);


	  } else {
	      break;
	  }
      } else {
	  if (framePtr->logEnd < now) {
	      nextPtr = framePtr->next;
	      if (framePtr->prev == NULL) {
		  objPtr->queue = nextPtr;
		  if (objPtr->queue != NULL) 
		      objPtr->queue->prev = NULL;
	      } else if (nextPtr == NULL) {
		  framePtr->prev->next = NULL;
	      } else {
		  framePtr->prev->next = nextPtr;
		  nextPtr->prev = framePtr->prev;
	      }
	      objPtr->late++;

	      /*
	       * Generate FRAME_DROPPED event.
	       */
	      
	      objPtr->bindFramePtr = framePtr;
	      ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
			       "<FRAME_DROPPED>");
	      
#if 0
	      {
		char logStr[100];
		sprintf(logStr, "MJPEGFrame LatePurge %x %x", framePtr, framePtr->dataBuffer->data);
		Tcm_Log(logStr);
	      }
#endif

	      MJPEG_RSRC_FreeFrame(framePtr);
	  } else {
	      break;
	  }
      }
      framePtr = nextPtr;
  }

  objPtr->queue = framePtr;
  if (framePtr && framePtr->next != NULL)
    framePtr->prev = NULL;

  return chosenFrame;
}

/************************************************************
 *
 * MJPEG_RSRC_Refresh
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MJPEG_RSRC_Refresh(MJPEGResource * objPtr)
{
  int res;
  char *errStr;

  /*
   * If not ready, do nothing.
   */

  if (!objPtr->ready) return TCL_OK;

  /*
   * If device has refresh, call it.
   */

  if (objPtr->devicePtr != NULL) {
    if (objPtr->devicePtr->refreshProc != NULL) {
      res = objPtr->devicePtr->refreshProc(objPtr);
      if (res != 0) {
	errStr = "Error refreshing picture.";
	if (objPtr->devicePtr->errorProc != NULL) {
	  errStr = objPtr->devicePtr->errorProc(objPtr, res);
	}
	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
      }
    }
  }

  return TCL_OK;
}

/************************************************************
 *
 * MJPEG_RSRC_DecodeDone
 *
 * Parameters:
 *     framePtr - pointer to frame just finished with decoding.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
int
MJPEG_RSRC_DecodeDone(MJPEGFrame *framePtr)
{
    int i, new_num;
    MJPEGResource *objPtr;

    objPtr = framePtr->objPtr;

    /*
     * Reset in decode
     */

    objPtr->inDecode = 0;

    /*
     * If destroy flag set, free frame, call destroy and return.
     */
    if (framePtr->objPtr->destroyFlag) {
	MJPEG_RSRC_FreeFrame(framePtr);
	DestroyMJPEGResource(objPtr);
	return TCL_OK;
    }


    /*
     * If not ready, free frame and return.
     */
    if (!objPtr->ready) {
	MJPEG_RSRC_FreeFrame(framePtr);
	return TCL_OK;
    }

    /*
     * Gather stats.
     */
    objPtr->bytes += framePtr->dataLength;

    /*
     * Generate FRAME_DECODE_DONE event.
     */

    objPtr->bindFramePtr = framePtr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		     "<FRAME_DECODE_DONE>");

    /*
     * Schedule display.  Search for a free at_token.  If we can't find
     * one, to increase size of at_token array
     */
    if (objPtr->num_at_tokens == 0) {
	objPtr->num_at_tokens = 30;
	objPtr->at_tokens = (int *) ckalloc(sizeof(int) * 30);
	for (i=0; i<objPtr->num_at_tokens; i++) {
	    objPtr->at_tokens[i] = 0;
	}
    } 
    for (i=0; i<objPtr->num_at_tokens; i++) {
	if (objPtr->at_tokens[i] == 0) {
	    break;
	}
    }
    if (i == objPtr->num_at_tokens) {
	new_num = objPtr->num_at_tokens + 10;
	objPtr->at_tokens = (int *) realloc((char *) objPtr->at_tokens,
						      sizeof(int) * new_num);
	for (i=objPtr->num_at_tokens; i < new_num; i++) {
	    objPtr->at_tokens[i] = 0;
	}
	i = objPtr->num_at_tokens;
	objPtr->num_at_tokens = new_num;
    }

    /*
     * If speed's zero and we're in the window, throw up the current
     * frame.  Otherwise, create an atEvent for it.
     */
    if (LTS_Speed(objPtr->lts) == 0.0) {
	framePtr->at_token = -1; 
	MJPEG_RSRC_DoDisplay(framePtr);
    } else {
	double t1, t2;
	t1 = LTS_SystemTimeOf(objPtr->lts, framePtr->logStart);
	t2 = LTS_SystemTimeOf(objPtr->lts, framePtr->logEnd);
	objPtr->at_tokens[i] = 
		Tcm_CreateAtHandler(t1<t2?t1:t2, HUGE_VAL,
			    "video", (Tk_TimerProc *) MJPEG_RSRC_DoDisplay, 
			    (ClientData) framePtr, NULL,
			    NULL, (Tk_IdleProc *) NULL,
			    (ClientData) NULL, NULL);
	framePtr->at_token = i;
    }
		
    /*
     * Arrange to decode next frame when we're idle
     */
    if (objPtr->queue != NULL) {
	Tcm_DoWhenIdle("video", (Tk_IdleProc *) MJPEG_RSRC_DoDecode, objPtr);
    }

    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_RSRC_DoDecode
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
int
MJPEG_RSRC_DoDecode(MJPEGResource *objPtr)
{
    MJPEGFrame *frame_ptr;

    /* 
     * If not ready or alraady in decode, do nothing.
     */
    if ((!objPtr->ready) || (objPtr->inDecode)) {
	return TCL_OK;
    }

    /*
     * Get next eligible frame.
     */
    frame_ptr = MJPEG_RSRC_GetEligibleFrame(objPtr);
    if (frame_ptr == NULL) {
	return TCL_OK;
    }

    /*
     * Record the fact that we're decoding a frame.
     */
    objPtr->inDecode = 1;

    /*
     * Generate FRAME_DECODE_BEGIN event.
     */

    objPtr->bindFramePtr = frame_ptr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		     "<FRAME_DECODE_BEGIN>");

    /*
     * Start the decoding loop
     */
    MJPEG_RSRC_ContinueDecode (frame_ptr);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * MJPEG_RSRC_ContinueDecode --
 *
 *	Continues an asynchronous decode until it's finished.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	May schedule a frame for display if the decoding is finished,
 *	otherwise, schedules itself to be called again when processor
 *	is idle.
 *
 *--------------------------------------------------------------
 */
static void
MJPEG_RSRC_ContinueDecode (MJPEGFrame *framePtr)
{
    MJPEGResource *objPtr;

    objPtr = framePtr->objPtr;

    /* 
     * If not ready or alraady in decode, do nothing.
     */
    if ((!objPtr->ready) || (!objPtr->inDecode)) {
	return;
    }

    /*
     * Call device decode procedure.
     */
    if (objPtr->devicePtr != NULL) {
        if (objPtr->devicePtr->decodeProc != NULL) {
	    objPtr->devicePtr->decodeProc(framePtr);
        }
    }

    /*
     * If decoding is not finished, schedule another call.  Otherwise
     * we're done.
     */
    if (objPtr->devicePtr->asyncDecodeFlag) {
	if (objPtr->devicePtr->decodeProc != NULL) {
	    Tcm_DoWhenIdle("video", (Tk_IdleProc *) MJPEG_RSRC_ContinueDecode,
			    framePtr);
	}
	return;
    }

    MJPEG_RSRC_DecodeDone(framePtr);
    return;
}

/************************************************************
 *
 * MJPEGResourceModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
MJPEGResourceModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mjpegPlay", MJPEGResourceCreateCmd,
			NULL, (void (*)()) NULL);
    Tcl_GlobalEval(interp, "cminfo add media {video}");
    Tcl_GlobalEval(interp, "cminfo add formats video {jpeg}");
    Tcl_GlobalEval(interp, "cminfo add objects video jpeg mjpegPlay");
    Tcl_VarEval(interp, "cminfo add object_info mjpegPlay {play",
#ifdef J300
			" j300",
#endif
#ifdef PARALLAX
			" parallax",
#endif
#ifdef JPEG_SW
			" software",
#endif
			"}", NULL);
    return TCL_OK;
}

/******************************************************************
 *
 * MJPEG_RSRC_DeviceList
 *
 * Parameters:
 *     objPtr - pointer to mpeg play object.
 *
 * Results:
 *     Standard Tcl Result.
 *
 * Actions:
 *     Sets result to list of devices supported. Returns TCL_OK;
 *
 ******************************************************************/
static int
MJPEG_RSRC_DeviceList(MJPEGResource *objPtr)
{
    Tcl_Interp *interp = objPtr->interp;
    Tcl_DString resStr;

    Tcl_DStringInit(&resStr);

#ifdef J300
    Tcl_DStringAppendElement(&resStr, "j300");
#endif
#ifdef PARALLAX
    Tcl_DStringAppendElement(&resStr, "parallax");
#endif
#ifdef JPEG_SW
    Tcl_DStringAppendElement(&resStr, "software");
#endif
#ifdef SUN_VIDEO
    Tcl_DStringAppendElement(&resStr, "sunvideo");
#endif
#ifdef COSMO
    Tcl_DStringAppendElement(&resStr, "cosmo");
#endif

    Tcl_SetResult(interp, Tcl_DStringValue(&resStr), TCL_VOLATILE);
    Tcl_DStringFree(&resStr);

    return TCL_OK;
}

static char *
GetLogStart(Tcl_Interp *interp, ClientData clientdata)
{
  MJPEGResource *objPtr;
  char *resStr;

  objPtr = (MJPEGResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%f", objPtr->bindFramePtr->logStart);

  return resStr;
}

static char *
GetLogEnd(Tcl_Interp *interp, ClientData clientdata)
{
  MJPEGResource *objPtr;
  char *resStr;

  objPtr = (MJPEGResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%f", objPtr->bindFramePtr->logEnd);

  return resStr;
}

static char *
GetCumulLenDecoded(Tcl_Interp *interp, ClientData clientdata)
{
  MJPEGResource *objPtr;
  char *resStr;

  objPtr = (MJPEGResource *) clientdata;

  resStr = (char *) ckalloc(50);
  sprintf(resStr, "%d", objPtr->bytes);
  objPtr->bytes = 0;
  return resStr;
}

static char *
GetDataLen(Tcl_Interp *interp, ClientData clientdata)
{
  MJPEGResource *objPtr;
  char *resStr;

  objPtr = (MJPEGResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->dataLength);

  return resStr;
}

static char *
GetWidth(Tcl_Interp *interp, ClientData clientdata)
{
  MJPEGResource *objPtr;
  char *resStr;

  objPtr = (MJPEGResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.width);

  return resStr;
}

static char *
GetHeight(Tcl_Interp *interp, ClientData clientdata)
{
  MJPEGResource *objPtr;
  char *resStr;

  objPtr = (MJPEGResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.height);

  return resStr;
}
