/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMJPEGPriority.c,v 1.23 1996/02/06 22:41:47 kpatel Exp $
 *
 *  A simple CMT object (created using the mjpegPriority command)
 *  that acts as a filter between the MPEGClipfile object and the
 *  pktSource object.  Its sole purpose is to reorder the buffers
 *  passed in into a priority order for the pktSource.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include "cmAtom.h"
#include "cmEvery.h"
#include "cmBufferlist.h"

/*
 * One of these structures is created for each mjpegPriority object.
 */
typedef struct MJPEG_Priority {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    SB_List *sbl;     /* Scatter buffer list of buffers to be reordered. */
    char *outCmd;		/* Command to send buffers to. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;		/* ClientData for outCCmd. */
    char *inCmd;                /* Command to get buffers from. */
    OutFunc inCCmd;             /* C pointer to inCmd. */
    ClientData inCD;            /* ClientData for inCCmd. */
    double cycleTime;           /* Cycle time for calling in and out commands. */
    int ready;                  /* Flag to indicate readiness. */
    char *everyHandle;          /* Handle for every event created if cycleTime is 
				   nonzero. */
    char *acceptHandle;
    char *sendHandle;
} MJPEG_Priority;

static int nextMJPEG_PriorityNum = 0;	/* Variable for next unique object
					 * name. */
/*
 * Procedures defined in this module
 */
MJPEG_Priority *CreateNewMJPEG_Priority	(Tcl_Interp *interp, char *name);
int DestroyMJPEG_Priority		(MJPEG_Priority *objPtr);
int ConfigureMJPEG_Priority		(MJPEG_Priority *objPtr,
					 int argc, char **argv);
int MJPEG_PriorityCreateCmd		(ClientData clientData,
					 Tcl_Interp *interp,
					 int argc, char **argv);
int MJPEG_PriorityObjCmd		(ClientData clientData,
					 Tcl_Interp *interp,
					 int argc, char **argv);
int MJPEG_PriorityModuleInit		(Tcl_Interp *interp);
static int MJPEG_Priority_Accept	(MJPEG_Priority *objPtr, SB_List *sbl);
static int MJPEG_Priority_Send          (MJPEG_Priority *objPtr, SB_List **sblPtr);
static int MJPEG_Priority_ChangeOutCmd	(MJPEG_Priority *objPtr, char *newOut);
static int ReadyMJPEG_Priority          (MJPEG_Priority *objPtr);
static int UnreadyMJPEG_Priority        (MJPEG_Priority *objPtr);
static void MJPEG_PrioritizeSBL         (SB_List *sbl);
static int MJPEG_Priority_ChangeInCmd   (MJPEG_Priority *objPtr, char *newIn);
static void DoCycleMJPEG_Priority        (ClientData clientData);
extern char *Tcm_CreateEvery();

/************************************************************
 *
 * CreateNewMJPEG_Priority
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

MJPEG_Priority *
CreateNewMJPEG_Priority(Tcl_Interp *interp, char *name)
{

    MJPEG_Priority *new;
    char *acceptName;
    char *sendName;

    new = (MJPEG_Priority *) ckalloc(sizeof(MJPEG_Priority));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy (new->name, name);
    new->sbl = NULL;
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->inCmd = NULL;
    new->inCCmd = NULL;
    new->inCD = NULL;
    new->ready = 0;
    new->cycleTime = 0.0;
    new->everyHandle = NULL;

    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+9);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) MJPEG_Priority_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    /*
     * Register send as an atom.
     */

    sendName = (char *) ckalloc(strlen(name)+9);
    sprintf(sendName, "%s.%s", name, "send");
    Tutil_RegisterCommand(interp, sendName, (ClientData) MJPEG_Priority_Send,
			  (ClientData) new);
    new->sendHandle = sendName;

    return new;
}

/************************************************************
 *
 * DestroyMJPEG_Priority
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

int
DestroyMJPEG_Priority(MJPEG_Priority *objPtr)
{
    /* 
     * Unready the object if readied.
     */

    if (objPtr->ready) {
	UnreadyMJPEG_Priority(objPtr);
    }

    /*
     * Remove command name
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }

    /*
     * Unregister the send atom.
     */
    if (objPtr->sendHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->sendHandle);
	ckfree(objPtr->sendHandle);
	objPtr->sendHandle = NULL;
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    if (objPtr->inCmd != NULL) {
	ckfree(objPtr->inCmd);
    }

    if (objPtr->everyHandle != NULL) {
	Tcm_DeleteEvery(objPtr->interp, objPtr->everyHandle, NULL);
	objPtr->everyHandle = NULL;
    }

    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureMJPEG_Priority
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

int
ConfigureMJPEG_Priority(MJPEG_Priority * objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    
    interp = objPtr->interp;
  
    if (objPtr->ready) {
	UnreadyMJPEG_Priority(objPtr);
    }

    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}
	if (objPtr->inCmd == NULL) {
	    Tcl_AppendResult(interp, "{-inCmd {} {}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-inCmd {} {", objPtr->inCmd,
			     "}} ", (char *)NULL);
	}
	{
	    char cycleTimeStr[64];
	    
	    sprintf(cycleTimeStr, "%f", objPtr->cycleTime);
	    Tcl_AppendResult(interp, "{-cycleTime 0.0 ", cycleTimeStr, "}", 
			     (char *) NULL);
	}
	
	return TCL_OK;
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	else if ((c == 'i') && (strncmp("-inCmd", argv[0], len) == 0)) {
	    if (objPtr->inCmd == NULL) {
		Tcl_AppendResult(interp, "-inCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-inCmd {} {", objPtr->inCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	else if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
	    char cycleTimeStr[64];

	    sprintf(cycleTimeStr, "%g", objPtr->cycleTime);
	    Tcl_AppendResult(interp, "-cycleTime 0.0 ", cycleTimeStr, (char *) NULL);
	    return TCL_OK;
	}
	Tcl_AppendResult(interp, "bad configuration option: ",
			  argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (MJPEG_Priority_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	} else if ((c == 'i') && (strncmp("-inCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (MJPEG_Priority_ChangeInCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	} else if ((c == 'c') && (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	    double newCycleTime;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if ((Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) || 
		(newCycleTime < 0.0)) {
		Tcl_AppendResult(interp, "Bad value specified trying to set ",
				 "configuration option ", argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->cycleTime = newCycleTime;
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			  argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_PriorityCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the mjpegResource command.
 *     A new object of type MJPEG_Priority is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
MJPEG_PriorityCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    MJPEG_Priority *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextMJPEG_PriorityNum++;
    sprintf(new_name, "%smjpegPriority%d", argv[1], nextMJPEG_PriorityNum);

    new = CreateNewMJPEG_Priority(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MJPEG_PriorityObjCmd,
			(ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureMJPEG_Priority(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_PriorityObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     MJPEG_Priority object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
MJPEG_PriorityObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    MJPEG_Priority *objPtr = (MJPEG_Priority *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureMJPEG_Priority(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyMJPEG_Priority(objPtr);
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return ReadyMJPEG_Priority(objPtr);
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return UnreadyMJPEG_Priority(objPtr);
    } else if ((c == 's') && (strncmp("send", argv[1], len) == 0)) {
	SB_List *sb_list;
	char *sbStr;

	if (!objPtr->ready) {
	    sprintf(objPtr->interp->result, "{}");
	    return TCL_OK;
	}

	if (MJPEG_Priority_Send(objPtr, &sb_list) != TCL_OK) {
	    Tcl_AppendResult(objPtr->interp, "Error in ", argv[0], " send.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	
	sbStr = MakeSBListString(sb_list);
	
	Tcl_SetResult(objPtr->interp, sbStr, TCL_VOLATILE); 
	ckfree(sbStr);
	ckfree((char *) sb_list);
	return TCL_OK;
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl;
	int rv;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}

	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	  ckfree((char *) sbl);
	  Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
		(char *) NULL);
	  return TCL_ERROR;
	}

	rv = MJPEG_Priority_Accept(objPtr, sbl);
	FreeSBList(sbl);
	return rv;
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ",
		argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
		"  ready\n", 
	        "  unready\n",
		"  send\n",
                "  accept <ScatterBufferList>\n",
                (char *) NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * MJPEG_PriorityModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
MJPEG_PriorityModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mjpegPriority", MJPEG_PriorityCreateCmd,
			NULL, (void (*)()) NULL);
    return TCL_OK;
}


/************************************************************
 * 
 * ReadyMJPEG_Priority
 *
 * Parameters:
 *     clientData - pointer to MJPEG Priority object.
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Readies the priority object if unready. Entails setting up
 *     cycle time callback.
 *
 ************************************************************/

static int
ReadyMJPEG_Priority(MJPEG_Priority *objPtr)
{
    if (objPtr->ready) return TCL_OK;

    /*
     * Blow away old every handle if it exists.
     */

    if (objPtr->everyHandle != NULL) {
	Tcm_DeleteEvery(objPtr->interp, objPtr->everyHandle, NULL);
	objPtr->everyHandle = NULL;
    }

    /*
     * Set up new every handle if needed.
     */

    if ((objPtr->outCmd != NULL) || (objPtr->inCmd != NULL)) {
	if (objPtr->cycleTime > 0.0) {
	    objPtr->everyHandle =
	      Tcm_CreateEvery((int)(objPtr->cycleTime * 1000.0), "video",
			      DoCycleMJPEG_Priority, (ClientData) objPtr);
	}
    }

    objPtr->ready = 1;

    return TCL_OK;
}


/************************************************************
 * 
 * UneadyMJPEG_Priority
 *
 * Parameters:
 *     clientData - pointer to MJPEG Priority object.
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Unreadies the priority object if unready. Entails killing
 *     cycle time callback if it exists.
 *
 ************************************************************/

static int
UnreadyMJPEG_Priority(MJPEG_Priority *objPtr)
{
    if (!objPtr->ready) return TCL_OK;

    if (objPtr->everyHandle != NULL) {
	Tcm_DeleteEvery(objPtr->interp, objPtr->everyHandle, NULL);
	objPtr->everyHandle = NULL;
    }
    
    objPtr->ready = 0;
    return TCL_OK;
}



/*
 *--------------------------------------------------------------
 *
 * MJPEG_Prioritize --
 *
 *	A prioritization algorithm (by Daishi) that computes the
 *	ordering for the MJPEG frames.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters (these are my best guesses -- bsmith)
 *	int n		Number of frames to prioritize
 *	int m		Starting frame number.
 *	int k		Starting priority number
 *	int *pl		Output priority array
 *
 *--------------------------------------------------------------
 */
void
MJPEG_Prioritize (int n, int m, int k, int *pl)
{
    int s,t;

    if (n==1) {
	*pl = m;
    } else {
	s = n/2;
	t = n-s;
	MJPEG_Prioritize(t,m,k*2,pl);
	MJPEG_Prioritize(s,m+t,k*2,pl+k);
    }
}

/************************************************************
 *
 * MJPEG_Priority_Send
 *
 * Paramters: 
 *     objPtr - pointer to object;
 *     sblPtr - pointer to scatter buffer list
 *
 * Results:
 *     Standard Tcl return value.
 *
 * Action:
 *     Prioritizes scatter buffer lists object has received since
 *     last send. Sets sblPtr to prioritized sbl.
 *
 ************************************************************/

static int
MJPEG_Priority_Send (MJPEG_Priority *objPtr, SB_List **sblPtr)
{
    
    /*
     * If object is not ready, return empty SB List. 
     */

    if (!objPtr->ready) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;

	return TCL_OK;
    }

    /*
     * If -inCmd set, execute now.
     */

    if (objPtr->inCmd != NULL) {
	SB_List *inSBL = NULL;

	if (objPtr->inCCmd != NULL) {
	    /*
	     * Need to check existance since object may have been destroyed.
	     */
	    
	    if (Tutil_FindAtom(objPtr->interp, objPtr->inCmd+1) != NULL) {
		objPtr->inCCmd(objPtr->inCD, &inSBL);
		if (inSBL != NULL) {
		    if (inSBL->count > 0) {

			AttachToAllBuffers(objPtr->interp, inSBL, NULL);

			if (objPtr->sbl == NULL) {
			    objPtr->sbl = inSBL;
			} else {
			    MergeSBLists(objPtr->sbl, inSBL);
			    FreeSBList(inSBL);
			}
		    } else {
			FreeSBList(inSBL);
		    }
		}
	    } else {
		/*
		 * In C command has unexpectedly disappeard.
		 */
		objPtr->inCCmd = NULL;
		ckfree(objPtr->inCmd);
		objPtr->inCmd = NULL;
	    }
	} else {
	    /*
	     * In command is tcl.
	     */
	    if (Tcl_VarEval(objPtr->interp, objPtr->inCmd, (char *) NULL) == TCL_OK) {

		inSBL = (SB_List *) ckalloc(sizeof(SB_List));
		inSBL->count = 0;

		if (ParseSBList(objPtr->interp, objPtr->interp->result, 
				inSBL) == TCL_OK) {
		    if (inSBL->count > 0) {

			AttachToAllBuffers(objPtr->interp, inSBL, NULL);

			if (objPtr->sbl == NULL) {
			    objPtr->sbl = inSBL;
			} else {
			    MergeSBLists(objPtr->sbl, inSBL);
			    FreeSBList(inSBL);
			} 
		    } else {
			FreeSBList(inSBL);
		    }
		}
	    }
	}
    }
	
    if (objPtr->sbl == NULL) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;

	return TCL_OK;
    }

    MJPEG_PrioritizeSBL(objPtr->sbl);

    /* 
     * Detach to all buffers. Since detach is always put on the 
     * idle queue this is safe to do since recieving object will
     * attach before these detach's actually occur.
     */

    DetachToAllBuffers(objPtr->interp, objPtr->sbl, NULL);
    *sblPtr = objPtr->sbl;
    objPtr->sbl = NULL;
    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_Priority_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -	scatter buffer list containing frame data
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
MJPEG_Priority_Accept(MJPEG_Priority *objPtr, SB_List *sbl)
{
    SB_List *local_sbl_copy;

    /*
     * Return if not ready.
     */

    if (!objPtr->ready) return TCL_OK;
   
    if ((sbl == NULL) || (sbl->count < 1)) return TCL_OK;

    AttachToAllBuffers(objPtr->interp, sbl, NULL);

    /*
     * Merge incoming buffers.
     * If cycle time is 0.0: 
     *    If -outCmd is set:
     *        Prioritize buffers.
     *        Send out through -outCmd
     *    If not: 
     *        Return.
     * If cycleTime > 0.0:
     *    Return.
     */

    local_sbl_copy = CopySBList(sbl);
    if (objPtr->sbl == NULL) {
	objPtr->sbl = local_sbl_copy;
    } else {
	MergeSBLists(objPtr->sbl, local_sbl_copy);
	FreeSBList(local_sbl_copy);
    }

    if (objPtr->cycleTime == 0.0) {

	if (objPtr->outCmd != NULL) {
	    
	    MJPEG_PrioritizeSBL(objPtr->sbl);

	    /* 
	     * Pass on data to outCmd
	     */
	    if (objPtr->outCCmd != NULL) {
		/*
		 * Need to check if it still exists since object may have
		 * been destroyed.
		 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		    objPtr->outCCmd(objPtr->outCD, objPtr->sbl);
		} else {
		    /*
		     * Out C command has unexpectedly disappeared.
		     */
		    objPtr->outCCmd = NULL;
		    ckfree(objPtr->outCmd);
		    objPtr->outCmd = NULL;
		}
	    } else {
		/*
		 * Out command is tcl, make sb list into a string and 
		 * call using Tcl_Eval. 
		 */
		char *sbStr = MakeSBListString(objPtr->sbl);
		Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
				sbStr, (char *) NULL); 
		ckfree(sbStr);
	    }
	    DetachToAllBuffers(objPtr->interp, objPtr->sbl, NULL);
	    if (objPtr->sbl != NULL) 
	      FreeSBList(objPtr->sbl);
	    objPtr->sbl = NULL;
	}
    }

    return TCL_OK;
}

/************************************************************
 *
 * MJPEG_PrioritizeSBL
 *
 * Parameters:
 *     sbl - scatter buffer list to be prioritized.
 *   
 * Results:
 *     None.
 *
 * Actions:
 *     Reorders sbl according to priority algorithm.
 *
 ************************************************************/

static void
MJPEG_PrioritizeSBL(SB_List *sbl)
{
    int numSBs, i;
    ScatterBuffer **tempSBs;
    int *orderVec = NULL;

    if (sbl == NULL) return;
    if (sbl->count == 0) return;

    /*
     * Re-order sbl in priority order.
     */
    numSBs = sbl->count;
    
    if (numSBs == 0) return;
	
    orderVec = (int *)ckalloc(sizeof(int)*numSBs);
    tempSBs = (ScatterBuffer **)ckalloc(sizeof(ScatterBuffer *)*numSBs);
    MJPEG_Prioritize(numSBs, 0, 1, orderVec);
    for (i=0; i<numSBs; i++) {
		tempSBs[i] = sbl->sbs[orderVec[i]];
    }
    for (i=0; i<numSBs; i++) {
		sbl->sbs[i] = tempSBs[i];
    }
    ckfree ((char *)tempSBs);
}

/************************************************************
 *
 * MJPEG_Priority_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
MJPEG_Priority_ChangeOutCmd(MJPEG_Priority *objPtr, char *newOut) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    objPtr->outCmd = NULL;
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

#ifdef SECURE_CMT
	if (newOut[0] != '@') {
	    Tcl_SetResult(objPtr->interp, "Cannot set Tcl outCmd because this program was compiled with -DSECURE_CMT.\nRecompile without security to use arbitrary Tcl outCmd's\n", TCL_STATIC);
	    return TCL_ERROR;
	}
#endif
    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
				       &(objPtr->outCD), 
				       (ClientData *) &(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    

/************************************************************
 *
 * MJPEG_Priority_ChangeInCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *     newIn - string representation of new inCmd.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes in command.
 ************************************************************/

static int
MJPEG_Priority_ChangeInCmd(MJPEG_Priority *objPtr, char *newIn) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->inCmd != NULL) {
	ckfree(objPtr->inCmd);
    }
    objPtr->inCmd = NULL;
    objPtr->inCCmd = NULL;
    objPtr->inCD = NULL;

    /*
     * If new inCmd is empty, return.
     */
    if (strlen(newIn) == 0) {
	return TCL_OK;
    }

#ifdef SECURE_CMT
	if (newIn[0] != '@') {
	    Tcl_SetResult(objPtr->interp, "Cannot set Tcl inCmd because this program was compiled with -DSECURE_CMT.\nRecompile without security to use arbitrary Tcl inCmd's\n", TCL_STATIC);
	    return TCL_ERROR;
	}
#endif


    /*
     * If new inCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newIn[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newIn,
				       &(objPtr->inCD), 
				       (ClientData *) &(objPtr->inCCmd)) != TCL_OK) {
	    objPtr->inCD = NULL;
	    objPtr->inCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newIn,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new in command into inCmd string.
     */
    objPtr->inCmd = (char *) ckalloc(strlen(newIn)+1);
    strcpy(objPtr->inCmd,newIn);
    return TCL_OK;
}    

/****************************************************************8
 *
 * DoCycleMJPEG_Priority
 *
 * Parameters:
 *     clientData - pointer to MJPEG Priority object.
 *
 * Results:
 *     None
 *
 * Actions:
 *     Called by every handler when time expires. Does -inCmd and
 *     -outCmd if they exist.
 *
 ****************************************************************/

static void
DoCycleMJPEG_Priority(ClientData clientData)
{
    MJPEG_Priority *objPtr = (MJPEG_Priority *) clientData;
    
    if (!objPtr->ready) {
	return;
    }

    /*
     * If -inCmd set, execute now.
     */

    if (objPtr->inCmd != NULL) {
	SB_List *inSBL = NULL;

	if (objPtr->inCCmd != NULL) {
	    /*
	     * Need to check existance since object may have been destroyed.
	     */
	    
	    if (Tutil_FindAtom(objPtr->interp, objPtr->inCmd+1) != NULL) {
		objPtr->inCCmd(objPtr->inCD, &inSBL);
		if (inSBL != NULL) {
		    if (inSBL->count > 0) {

			AttachToAllBuffers(objPtr->interp, inSBL, NULL);

			if (objPtr->sbl == NULL) {
			    objPtr->sbl = inSBL;
			} else {
			    MergeSBLists(objPtr->sbl, inSBL);
			    FreeSBList(inSBL);
			}
		    } else {
			FreeSBList(inSBL);
		    }
		}
	    } else {
		/*
		 * In C command has unexpectedly disappeard.
		 */
		objPtr->inCCmd = NULL;
		ckfree(objPtr->inCmd);
		objPtr->inCmd = NULL;
	    }
	} else {
	    /*
	     * In command is tcl.
	     */
	    if (Tcl_VarEval(objPtr->interp, objPtr->inCmd, (char *) NULL) == TCL_OK) {

		inSBL = (SB_List *) ckalloc(sizeof(SB_List));
		inSBL->count = 0;

		if (ParseSBList(objPtr->interp, objPtr->interp->result, 
				inSBL) == TCL_OK) {
		    if (inSBL->count > 0) {

			AttachToAllBuffers(objPtr->interp, inSBL, NULL);

			if (objPtr->sbl == NULL) {
			    objPtr->sbl = inSBL;
			}
			else {
			    MergeSBLists(objPtr->sbl, inSBL);
			    FreeSBList(inSBL);
			}
		    } else {
			FreeSBList(inSBL);
		    }
		}
	    }
	}
    }
    
    if (objPtr->outCmd != NULL) {
	    
	MJPEG_PrioritizeSBL(objPtr->sbl);

	/* 
	 * Pass on data to outCmd
	 */
	if (objPtr->outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, objPtr->sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree(objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and 
	     * call using Tcl_Eval. 
	     */
	    char *sbStr = MakeSBListString(objPtr->sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree(sbStr);
	}
	DetachToAllBuffers(objPtr->interp, objPtr->sbl, NULL);
	if (objPtr->sbl != NULL)
	  FreeSBList(objPtr->sbl);
	objPtr->sbl = NULL;
    }
}
