/*
 * COPYRIGHT NOTICE 
 * 
 * Copyright (c) Digital Equipment Corporation, 1993, 1994.
 * All Rights reserved. Unpublished rights reserved under the 
 * copyright laws of the United States. Copyright is claimed in 
 * the computer program and user interface thereof. 
 * 
 * The software contained on this media is proprietary to and 
 * embodies the confidential technology of Digital Equipment 
 * Corporation. Possession, use, duplication or dissemination of 
 * the software and media is authorized only pursuant to a valid
 * written license from Digital Equipment Corporation. 
 * 
 * The name of Digital Equipment Corporation may not be used to endorse or 
 * promote products derived from this software without specific prior 
 * written permission. All other rights reserved.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, IMPLIED WARRANTIES OF
 * NON-INFRINGEMENT, MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 
 * Digital assumes no responsibility AT ALL for the use or reliability of 
 * this software.
 * 
 * 
 * +---------------------------------------------------------------------+
 * | USE, DUPLICATION OR DISCLOSURE BY THE U.S. GOVERNMENT IS SUBJECT TO | 
 * | RESTRICTIONS AS SET FORTH IN SUBPARAGRAPH (c)                       |
 * | DFARS 252.227-7013, OR IN FAR 52.227-19 OR IN 52.227-14 ALT III,    |
 * | AS APPLICABLE                                                       |
 * +---------------------------------------------------------------------+
 * 
 * 
 * 
 * 
 * 
 * 
 */
/*
 * HISTORY
 */
/*
**++
** FACILITY:	Digital Equipment Corporation
**		Multimedia Services for Alpha AXP
**		Development Kit Sample Code
**
** FILE NAME:	color_utils.c
**
** PURPOSE:	Common routines used to implement colormap strategies.
**
** NOTES:	Public interface definition in color_utils.h.
**--
*/
#include <stdio.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include <mme/mme_api.h>
#include "color_utils.h"

#define DORED         0
#define DOGREEN       1
#define DOBLUE        2
#define DOALL         3

/*
 * Global definitions.
 */
static Atom _XA_MMS_RGB_25_MAP;
static Atom _XA_MMS_RGB_50_MAP;
static Atom _XA_MMS_RGB_75_MAP;
#define SIZOFLONG 65535

typedef unsigned long XPixel_t;


/*
**++
**  FUNCTIONAL_NAME: mmcGetRgbFromColormap 
**
**  FUNCTIONAL_DESCRIPTION:
**      set up the colors in the RGBQUAD structure from the
**      colormap passed in.
**
**  FORMAL PARAMETERS:
**           display     - pointer to the display
**           colormap    - colormap to use
**           num         - number of free color cells in the colormap
**           rgb         - RGB structure to fill up 
**           freeCells   - Flag to free color cells after allocation 
**
**  RETURN VALUE:
**
**           True        - if this call is successful
**           False       - if this call failed
**
**  COMMENTS:
**
**  DESIGN:
**
**
*/
int mmcGetRgbFromColormap(
           Display * display,
           Colormap colormap,
           int * num,
           RGBQUAD * rgb,
           int freeCells)
{
   int i;
   int alloc_colors = 0;
   int fixed_colors = 0;
   XPixel_t    pixels[MAX_COLORS];               
   XColor xcolors[MAX_COLORS];

   /*
    * Copy as many colors as possible.
    * initialize all returned arrays
    */
   for (i = 0; i < MAX_COLORS; i++)
   {
      rgb[i].rgbReserved = 1;
      rgb[i].rgbRed = rgb[i].rgbGreen = rgb[i].rgbBlue = 0;
   }

   /*
    * allocate as many color cells as possible to see how many
    * cells are free.
    */
   for (alloc_colors = MAX_COLORS; alloc_colors; alloc_colors--)
   {
      if (XAllocColorCells (display,
                            colormap,
                            False, NULL, 0,
                            pixels, alloc_colors))
              break;
   }
   /*
    * setup the fixed flag for each color cell not allocated
    */
   for (i = 0; i < alloc_colors; i++)
      rgb[pixels[i]].rgbReserved = 0;

   /*
    * prepare an array to receive values for all fixed colors
    */
   for (i = 0; i < MAX_COLORS; i++)
      if (rgb[i].rgbReserved == 1)
         xcolors[fixed_colors++].pixel = i;
   /*
    * ask for the values of all fixed colors
    */
   if (XQueryColors (display, colormap, xcolors, fixed_colors))
   {
      return False;
   }

   /*
    * store the values of all fixed colors in the output array
    */
   for (i = 0; i < fixed_colors; i++)
   {
      rgb[xcolors[i].pixel].rgbRed = xcolors[i].red;
      rgb[xcolors[i].pixel].rgbGreen = xcolors[i].green;
      rgb[xcolors[i].pixel].rgbBlue = xcolors[i].blue;
   }

   /*
    * Free all the color cells alloced above by XAllocColorCells
    * if we cmap is the default colormap so that other applications
    * may use these free cells in the default colormap.
    *
    * If the colormap is a private colormap, leave the color cells
    * allocated.  We will follow this call to mmcGetRgbFromColormap
    * with a call to mmcSetColorsInColormap and copy these colors 
    * into the private colormap.
    */
   if (freeCells)
   { 
      XFreeColors(display, colormap, pixels, alloc_colors, NULL);
   }
   /* 
    * return the number of free color cells 
    */
   *num = alloc_colors;

   return True;
}

/*
**++
**  FUNCTIONAL_NAME: mmcSetColorsInColormap
**
**  FUNCTIONAL_DESCRIPTION:
**      set up the colors in the private colormap from the RGBQUAD
**      structures, merge the color specifications in the structure
**      old and rgb before storing the colors in the colormap.
**
**  FORMAL PARAMETERS:
**           display        - pointer to the display
**           colormap       - colormap to fill
**           rgb            - RGB color structure to use
**           old            - RGB color structure to use
**           CmapStrategy   - colormap strategy
**           AllocColors    - Flag - allocate colors before storing
**
**  RETURN VALUE:
**           True           - If call was successful
**           False          - If call failed
**  COMMENTS:
**
**  DESIGN:
**
**
*/
int mmcSetColorsInColormap(
                    Display * display,
                    Colormap colormap,
		    RGBQUAD * rgb,
		    RGBQUAD * old,
                    int CmapStrategy,
                    int AllocColors)
{
    int i;
    int modified_colors = 0;
    XPixel_t pixels[MAX_COLORS];
    XColor xcolors[MAX_COLORS];
    unsigned char cr,cg,cb;

  for (i = 0; i < MAX_COLORS; i++)
  {
     if ( old != NULL )
     {
        if ( old[i].rgbReserved == 0 )
        {
           if ( rgb[i].rgbReserved != 0 )
           {
              cr = rgb[i].rgbRed   & 0xff;
              cg = rgb[i].rgbGreen & 0xff;
              cb = rgb[i].rgbBlue  & 0xff;

              xcolors[modified_colors].red   = cr << 8;
              xcolors[modified_colors].green = cg << 8;
              xcolors[modified_colors].blue  = cb << 8;

              xcolors[modified_colors].pixel = i;
              xcolors[modified_colors].flags = DoRed | DoGreen | DoBlue;
              modified_colors++;
           }
        }
        else
        {
           if ( rgb[i].rgbReserved != 0 )
           {
              cr = old[i].rgbRed   & 0xff;
              cg = old[i].rgbGreen & 0xff;
              cb = old[i].rgbBlue  & 0xff;

              xcolors[modified_colors].red   = cr << 8;
              xcolors[modified_colors].green = cg << 8;
              xcolors[modified_colors].blue  = cb << 8;

              xcolors[modified_colors].pixel = i;
              xcolors[modified_colors].flags = DoRed | DoGreen | DoBlue;
              modified_colors++;
           }
        }
     }   
     else /* if old == NULL */
     {
        if ( rgb[i].rgbReserved != 0 )
        {
           cr = rgb[i].rgbRed   & 0xff;
           cg = rgb[i].rgbGreen & 0xff;
           cb = rgb[i].rgbBlue  & 0xff;

           xcolors[modified_colors].red   = cr << 8;
           xcolors[modified_colors].green = cg << 8;
           xcolors[modified_colors].blue  = cb << 8;

           xcolors[modified_colors].pixel = i;
           xcolors[modified_colors].flags = DoRed | DoGreen | DoBlue;

           modified_colors++;
        }
     } 
  }  /* for */

  /*
   * tell X the values of all colors set by the dithering code
   */
  if (modified_colors)
  {
     if ( CmapStrategy == DEFAULT_CMAP )
     {
        for ( i = 0; i < modified_colors; i++ )
           XAllocColor(display, colormap, &xcolors[i]);
     }
     else 
     {
        /*
         * We may not need to allocate the colormap cells here since we
         * may have passed in a false in the call to mmcGetRgbFromColormap.
         * This leaves the free color cells allocated so go ahead and 
         * store the colors.
         */
        if ( AllocColors )
        {
           /*
            * Allocate the color cells and store them in the
            * colormap.
            * NOTE: Need 3 extra colors for the window border!!!
            */
           if (XAllocColorCells(display,
                      colormap,
                      False, NULL, 0,
                      pixels,  modified_colors+3 ) == 0 )
           {
              fprintf(stderr, "Alloc color cells failed\n"); 
              return False;
           }
        }

        XStoreColors(display, colormap, xcolors, modified_colors);
     } 
  }
  return True;
}

/*
**++
**  FUNCTIONAL_NAME: mmcUpdateColormap 
**
**  FUNCTIONAL_DESCRIPTION:
**      Update the colormap with colors defined in the RGBQUAD structure
**      and colors ncolors created based on the settings in localStdCmap.
**
**  FORMAL PARAMETERS:
**           dpy          - pointer to the display 
**           map_entries  - Number of cells in colormap  
**           colorcube    - The colormap to update 
**           rgb          - use these colors first 
**           ncolors      - The number of color requested 
**           flag         - which plane to set 
**           cmapType     - colormap type 
**
**  RETURN VALUE:
**
**           True         - If call was successful
**           False        - If call failed
**
**  COMMENTS:
**
**  DESIGN:
**
**
*/
int mmcUpdateColormap(
                    Display *dpy, 
                    int map_entries,     
                    XStandardColormap *colorcube,  
                    RGBQUAD * rgb,       
                    int ncolors, 
                    int flag,
                    int cmapType)     
{
    int i, j;
    int colorIndex = 1;
    int reservedColors=0;
    int r;
    int g;
    int gg;
    XColor colors[MAX_COLORS];
    unsigned char cr,cg,cb;
    unsigned long pixels[MAX_COLORS];

    if ( rgb != NULL )
    {
       j = 0;
       colorIndex = 1; /* Don't need black so start at 1 */
       r = colorcube->red_mult;
       g = colorcube->green_mult;
       gg = colorcube->green_max + 1;

       for (i = 0; i < MAX_COLORS; i++)
       {
          if ( rgb[i].rgbReserved == 1 )
          {
             cr = rgb[i].rgbRed   & 0xff;
             cg = rgb[i].rgbGreen & 0xff;
             cb = rgb[i].rgbBlue  & 0xff;

             colors[j].red   = cr << 8;
             colors[j].green = cg << 8;
             colors[j].blue  = cb << 8;

             colors[j].pixel = i;
             switch (flag)
             {
                case DOALL:
                     colors[j].flags = DoRed | DoGreen | DoBlue;
                     break;
                case DORED:
                     colors[j].flags = DoRed;
                     break;
                case DOGREEN:
                     colors[j].flags =  DoGreen;
                     break;
                case DOBLUE:
                     colors[j].flags = DoBlue;
                     break;
             }
             j++;
          }
          else /* rgb[i].rgbReserved == 0 */
          {
             if ( colorIndex < ncolors )
             {
                switch (flag)
                {
                   case DOALL:
                       colors[j].flags =  DoRed | DoGreen | DoBlue;
                       colors[j].red = (unsigned short)
                               (((colorIndex/r) * SIZOFLONG) /
                                  colorcube->red_max);
                       colors[j].green = (unsigned short)
                               ((((colorIndex/g)%gg) * SIZOFLONG) /
                                   colorcube->green_max);
                       colors[j].blue = (unsigned short)
                               (((colorIndex%g) * SIZOFLONG) /
                                  colorcube->blue_max);
                       break;
                   case DORED:
                       colors[j].flags =  DoRed;
                       colors[j].red = (unsigned short)
                               (((colorIndex/r) * SIZOFLONG) /
                                  colorcube->red_max);
                       break;
                   case DOGREEN:
                       colors[j].flags =  DoGreen;
                       colors[j].green = (unsigned short)
                               ((((colorIndex/g)%gg) * SIZOFLONG) /
                                   colorcube->green_max);
                       break;
                   case DOBLUE:
                       colors[j].flags =  DoBlue;
                       colors[j].blue = (unsigned short)
                               (((colorIndex%g) * SIZOFLONG) /
                                  colorcube->blue_max);
                       break;
                }
                colors[j].pixel = (unsigned long) i;
                colorIndex++;
                j++;
             }
          }
       }
       reservedColors = j;
    }

    else /* rgb == NULL */
    {
       r = colorcube->red_mult;
       g = colorcube->green_mult;
       gg = colorcube->green_max + 1;

       for ( i=0; i < ncolors; i++)
       {
          switch (flag)
          {
             case DOALL:
                colors[i].flags =  DoRed | DoGreen | DoBlue;
                colors[i].red = (unsigned short)
                              (((i/r) * SIZOFLONG) / colorcube->red_max);
                colors[i].green = (unsigned short)
                              ((((i/g)%gg) * SIZOFLONG) / colorcube->green_max);
                colors[i].blue = (unsigned short)
                              (((i%g) * SIZOFLONG) / colorcube->blue_max);
                break;
             case DORED:
                colors[i].flags =  DoRed;
                colors[i].red = (unsigned short)
                              (((i/r) * SIZOFLONG) / colorcube->red_max);
                break;
             case DOGREEN:
                colors[i].flags =  DoGreen;
                colors[i].green = (unsigned short)
                              ((((i/g)%gg) * SIZOFLONG) / colorcube->green_max);
                break;
             case DOBLUE:
                colors[i].flags =  DoBlue;
                colors[i].blue = (unsigned short)
                              (((i%g) * SIZOFLONG) / colorcube->blue_max);
                break;
          }
          colors[i].pixel = (unsigned long) i; 
       }
       reservedColors = ncolors;
    }

    if ( cmapType == PRIVATE_CMAP)
    {
       XAllocColorCells(dpy, colorcube->colormap,
                   False, NULL, 0,
                   pixels, reservedColors);
       XStoreColors(dpy, colorcube->colormap, colors, reservedColors);
    }
    else if ( (cmapType == STD25_CMAP) ||
              (cmapType == STD50_CMAP) ||
              (cmapType == STD75_CMAP) )
    {
       for ( i = 0; i < reservedColors; i++)
       {
          XAllocColor(dpy, colorcube->colormap, &colors[i]);
       }
    }
    return True;
}

/*
**++
**  FUNCTIONAL_NAME: mmcCreateStdMap 
**
**  FUNCTIONAL_DESCRIPTION:
**      set up the colors in the private colormap from the RGBQUAD
**      structure.
**
**  FORMAL PARAMETERS:
**           dpy        - Display id 
**           visual     - Visual id, currenly 8 bit PseudoColor 
**           property   - Property to create 
**           ncolors    - Number of colors in the std colormap 
**
**  RETURN VALUE:
**
**           True       - If call was successful
**           False      - If call failed
**
**  COMMENTS:
**
**  DESIGN:
**
**
*/
int mmcCreateStdMap(
                 Display *dpy,     
                 Visual *visual,   
                 Atom property,    
                 int ncolors       
                 )
{

    Display *mydpy;
    Window  root;
    Screen  *newScreen;
    int i;
    int screenNum;
    int success = False;
    char *atomName;
    XStandardColormap localStdCmap;
    RGBQUAD rgb[MAX_COLORS];
    int num;
    Visual *default_vis;
    int default_dep;

    newScreen = XDefaultScreenOfDisplay(dpy);
    root      = RootWindowOfScreen(newScreen);
    screenNum = XScreenNumberOfScreen(newScreen);

    mydpy = dpy;
    dpy = XOpenDisplay(DisplayString(mydpy));

    localStdCmap.colormap = XCreateColormap(dpy, 
                                            root, 
                                            visual, 
                                            AllocNone);
    /*
     * Get the allocated colors from the 
     * default map.
     */
    default_vis = DefaultVisual(dpy, screenNum);
    default_dep = DefaultDepth(dpy, screenNum);
    if ( (default_vis->class == PseudoColor) && (default_dep == 8) )
       if (!mmcGetRgbFromColormap( dpy, 
                                  DefaultColormap(dpy, screenNum),
                                  &num,
                                  rgb, True))
          fprintf(stderr, "Could not get default colormap entries\n");
          

    localStdCmap.red_max = 3;
    localStdCmap.red_mult = 16;
    localStdCmap.green_max = 3;
    localStdCmap.green_mult = 4;
    localStdCmap.blue_max = 3;
    localStdCmap.blue_mult = 1;

    /*
     * Update the colormap with colors from default
     * colormap and colors ncolors created based on
     * the settings in localStdCmap.
     */
    if (mmcUpdateColormap( dpy, MAX_COLORS, &localStdCmap, rgb,
                      ncolors, DOALL, PRIVATE_CMAP ))
       success = True;

    if (success)
    {
       localStdCmap.visualid = newScreen->root_visual->visualid;
       localStdCmap.killid = XCreatePixmap(dpy,root,1,1,1);
    }


    /*
     * Now Create and store this property
     */

    XChangeProperty(
                     dpy, root, property,
                     XA_RGB_COLOR_MAP, 32,
                     PropModeReplace,
                     (unsigned char *)&localStdCmap,
                     sizeof( XStandardColormap)
                   );
    /* 
     *  grab the server, and call XSetCloseDownMode
     *  to retain the property just created until
     *  the end of this session.  Then ungrab the
     *  server and close the connection.
     */ 
    if (property)
    {
       XGrabServer(dpy);  /* pose as another client and grab the server */
       XSetCloseDownMode(dpy,RetainPermanent);
       XUngrabServer(dpy);
       XCloseDisplay(dpy);
    }


    return True;
}

/*
**++
**  FUNCTIONAL_NAME: mmcCreateProperty 
**
**  FUNCTIONAL_DESCRIPTION:
**      set up the colors in the private colormap from the RGBQUAD
**      structure.
**
**  FORMAL PARAMETERS:
**           dpy        - Display id 
**           visual     - Visual id, currenly 8 bit PseudoColor 
**
**  RETURN VALUE:
**           True       - If call was successful
**           False      - If call failed
**
**  COMMENTS:
**
**  DESIGN:
**
**
*/
int mmcCreateProperty( Display *dpy, 
                       Visual *visual, 
                       Bool verbose )
{
    int i;
    Atom property;

    mmcCreateStdMap( dpy, visual, _XA_MMS_RGB_25_MAP, 25 );
    if ( verbose )
    {
       if ( XInternAtom( dpy, "_MMS_RGB_25_MAP", True ) )
          fprintf(stderr, "\nCreated Atom: _MMS_RGB_25_MAP\n");
    }

    mmcCreateStdMap( dpy, visual, _XA_MMS_RGB_50_MAP, 50 );
    if ( verbose )
    {
       if ( XInternAtom( dpy, "_MMS_RGB_50_MAP", True ) )
          fprintf(stderr, "Created Atom: _MMS_RGB_50_MAP\n");
    }

    mmcCreateStdMap( dpy, visual, _XA_MMS_RGB_75_MAP, 75 );
    if ( verbose )
    {
       if ( XInternAtom( dpy, "_MMS_RGB_75_MAP", True ) )
          fprintf(stderr, "Created Atom: _MMS_RGB_75_MAP\n\n");
    }

    return True;
}

/*
**++
**  FUNCTIONAL_NAME: mmcListAtoms 
**
**  FUNCTIONAL_DESCRIPTION:
**      set up the colors in the private colormap from the RGBQUAD
**      structure.
**
**  FORMAL PARAMETERS:
**           dpy        - Display id 
**
**  RETURN VALUE:
**           True       - If call was successful
**           False      - If call failed
**
**  COMMENTS:
**
**  DESIGN:
**
**
*/
int mmcListAtoms( Display *dpy )
{
    Bool doesNotExist = True;

    fprintf(stderr, "\nStandard colormap properties currently defined are:\n");

    if ( XInternAtom( dpy, "_MMS_RGB_25_MAP", True ) )
    {
       fprintf(stderr, "   _XA_MMS_RGB_25_MAP\n");
       doesNotExist = False;
    }

    if ( XInternAtom( dpy, "_MMS_RGB_50_MAP", True ) )
    {
       fprintf(stderr, "   _XA_MMS_RGB_50_MAP\n");
       doesNotExist = False;
    }

    if ( XInternAtom( dpy, "_MMS_RGB_75_MAP", True ) )
    {
       fprintf(stderr, "   _XA_MMS_RGB_75_MAP\n");
       doesNotExist = False;
    }
    fprintf(stderr, "\n");

    if ( doesNotExist )
    {
       fprintf(stderr, "MME Standard colormap properties not defined \n");
    }

    return True;
}

/*
**++
**  FUNCTIONAL_NAME: mmcCreateAtoms 
**
**  FUNCTIONAL_DESCRIPTION:
**      Create the standard colormap atoms if they do not already
**      exist.  If they exist then assume that the property also
**      exists so do not create the the atom again.
**
**  FORMAL PARAMETERS:
**           dpy        - Display id 
**
**  RETURN VALUE:
**           True       - If all the standard colormap atoms are successfully
**                        created.  
**           False      - If any one of the atoms already exists.
**
**  COMMENTS:
**
**  DESIGN:
**
**
*/
int mmcCreateAtoms( Display *dpy, 
                    Bool verbose )
{
    int i;
    int status=0;

    _XA_MMS_RGB_25_MAP = NULL;
    _XA_MMS_RGB_50_MAP = NULL;
    _XA_MMS_RGB_75_MAP = NULL;

    if ( XInternAtom( dpy, "_MMS_RGB_25_MAP", True) == NULL )
    {
       _XA_MMS_RGB_25_MAP    = XInternAtom( dpy, "_MMS_RGB_25_MAP", False);
       status++;
    }

    if ( XInternAtom( dpy, "_MMS_RGB_50_MAP", True) == NULL )
    {
       _XA_MMS_RGB_50_MAP    = XInternAtom( dpy, "_MMS_RGB_50_MAP", False);
       status++;
    }

    if ( XInternAtom( dpy, "_MMS_RGB_75_MAP", True) == NULL )
    {
       _XA_MMS_RGB_75_MAP    = XInternAtom( dpy, "_MMS_RGB_75_MAP", False);
       status++;
    }

    if ( verbose )
       mmcListAtoms( dpy );

    return(status) ;
}
/*
**++
**  FUNCTIONAL_NAME: mmcInitColormap 
**
**  FUNCTIONAL_DESCRIPTION:
**      set up the colors in the private colormap from the RGBQUAD
**      structure.
**
**  FORMAL PARAMETERS:
**           dpy          - Display id 
**           visual       - Visual id, currenly 8 bit PseudoColor 
**           screenNum    - Display screen number
**           cmap         - colormap id.
**           rgb          - structure to hold color values
**           CmapStrategy - colormap strategy used
**
**  RETURN VALUE:
**           None
**
**  COMMENTS:
**
**  DESIGN:
**
**
*/
mmcInitColormap(Display *dpy, 
             Visual *visual,
             int screenNum,
             Colormap *cmap,
             RGBQUAD * rgb,
             int CmapStrategy,
             Bool verbose )
{
    int     num;
    Window  root;
    int     cmapCount;
    Screen  *newScreen;
    XStandardColormap *stdcmap;
    Visual *default_vis;
    int default_dep;
  

    if ( CmapStrategy == PRIVATE_CMAP )
    {
       if ( verbose )
          fprintf(stderr, "\n Using a private colormap \n\n");
       /*
        * Create a private colormap.
        */
       *cmap = XCreateColormap(dpy,
                        RootWindow(dpy, screenNum),
                        visual,
                        AllocNone);
       /*
        * If the default visual is a 
        * 8 bit Pseudocolor then get 
        * colors already allocated.
        */
       default_vis = DefaultVisual(dpy, screenNum);
       default_dep = DefaultDepth(dpy, screenNum);

       if ( (default_vis->class == PseudoColor) && (default_dep == 8) )
       {
          /*
           * Set default colormap colors
           * in the rgb structure.
           */
          if (!mmcGetRgbFromColormap( dpy,
                                      DefaultColormap(dpy, screenNum),
                                      &num,
                                      rgb, True))
             fprintf(stderr, "Could not get default colormap entries\n");
          /*
           * Set the colors from the default colormap
           * in the cmap just created.
           */
          if (!mmcSetColorsInColormap( dpy, *cmap,
                                       rgb, NULL,
                                       CmapStrategy,
                                       True ) )
          {
             fprintf(stderr, "Could not set colors from the default\n"); 
             fprintf(stderr, "colormap into the private colormap just\n"); 
             fprintf(stderr, "created\n");
          } 
       }
    }
    else if ( (CmapStrategy == STD25_CMAP) ||
              (CmapStrategy == STD50_CMAP) ||
              (CmapStrategy == STD75_CMAP) )
    {
       if ( verbose )
          fprintf(stderr, "\n Creating standard colormaps\n\n");
       /*
        * Create atoms that define the standard colormaps then 
        * set the property corresponding to each atom.
        */
       if ( mmcCreateAtoms( dpy, verbose ) > 0 )
          mmcCreateProperty( dpy, visual, verbose );

       newScreen = DefaultScreenOfDisplay(dpy);
       root = RootWindowOfScreen(newScreen);

       switch ( CmapStrategy )
       {
          case STD25_CMAP :
               _XA_MMS_RGB_25_MAP = XInternAtom( dpy, "_MMS_RGB_25_MAP", True );
               if (XGetRGBColormaps(dpy, root,
                                    &stdcmap, &cmapCount,
                                   _XA_MMS_RGB_25_MAP))
               {
                  if ( verbose )
                     fprintf(stderr,
                          "\nUsing Standard Colormap: _XA_MMS_RGB_25_MAP\n\n");
                  *cmap = stdcmap->colormap;
                }
                else
                {
                   if ( verbose )
                      fprintf(stderr,
                          "Could not find _XA_MMS_RGB_25_MAP colormap\n\n");
                }
                break;
          case STD50_CMAP :
               _XA_MMS_RGB_50_MAP = XInternAtom( dpy, "_MMS_RGB_50_MAP", True );
               if (XGetRGBColormaps(dpy, root,
                                    &stdcmap, &cmapCount,
                                   _XA_MMS_RGB_50_MAP))
               {
                  if ( verbose )
                     fprintf(stderr,
                           "\nUsing Standard Colormap: _XA_MMS_RGB_50_MAP\n\n");
                  *cmap = stdcmap->colormap;
                }
                else
                {
                   if ( verbose )
                      fprintf(stderr,
                          "Could not find _XA_MMS_RGB_50_MAP colormap\n\n");
                }
                break;
          case STD75_CMAP :
               _XA_MMS_RGB_75_MAP = XInternAtom( dpy, "_MMS_RGB_75_MAP", True );
               if (XGetRGBColormaps(dpy, root,
                                    &stdcmap, &cmapCount,
                                   _XA_MMS_RGB_75_MAP))
               {
                  if ( verbose )
                     fprintf(stderr,
                          "\nUsing Standard Colormap: _XA_MMS_RGB_75_MAP\n\n");
                  *cmap = stdcmap->colormap;
                }
                else
                {
                   if ( verbose )
                      fprintf(stderr,
                           "Could not find _XA_MMS_RGB_75_MAP colormap\n\n");
                }
                break;
       }
    }
    else if ( CmapStrategy == DEFAULT_CMAP)
    {
       if ( verbose )
          fprintf(stderr,"\n Using default colormap \n\n");

       *cmap =  DefaultColormap(dpy, screenNum);
       /*
        * Go out and get the colors from the default
        * colormap and explicitly set them in the
        * copy of the default colormap.  This should 
        * not be necessary, but if not done colormaps
        * looked bad when the app. was first run.
        */
       if (!mmcGetRgbFromColormap( dpy,
                                   DefaultColormap(dpy, screenNum),
                                   &num,
                                   rgb, True))
          fprintf(stderr, "Could not get default colormap entries\n");
       /*
        * Set the colors from the default colormap
        * in the cmap just created.
        */
       if (!mmcSetColorsInColormap( dpy, *cmap,
                                    rgb, NULL,
                                    CmapStrategy,
                                    True ) )
       {
          fprintf(stderr, "Could not set colors from the default\n");
          fprintf(stderr, "colormap into the private colormap just\n");
          fprintf(stderr, "created\n");
       }

    }
}
