/*
 * COPYRIGHT NOTICE 
 * 
 * Copyright (c) Digital Equipment Corporation, 1993, 1994.
 * All Rights reserved. Unpublished rights reserved under the 
 * copyright laws of the United States. Copyright is claimed in 
 * the computer program and user interface thereof. 
 * 
 * The software contained on this media is proprietary to and 
 * embodies the confidential technology of Digital Equipment 
 * Corporation. Possession, use, duplication or dissemination of 
 * the software and media is authorized only pursuant to a valid
 * written license from Digital Equipment Corporation. 
 * 
 * The name of Digital Equipment Corporation may not be used to endorse or 
 * promote products derived from this software without specific prior 
 * written permission. All other rights reserved.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, IMPLIED WARRANTIES OF
 * NON-INFRINGEMENT, MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 
 * Digital assumes no responsibility AT ALL for the use or reliability of 
 * this software.
 * 
 * 
 * +---------------------------------------------------------------------+
 * | USE, DUPLICATION OR DISCLOSURE BY THE U.S. GOVERNMENT IS SUBJECT TO | 
 * | RESTRICTIONS AS SET FORTH IN SUBPARAGRAPH (c)                       |
 * | DFARS 252.227-7013, OR IN FAR 52.227-19 OR IN 52.227-14 ALT III,    |
 * | AS APPLICABLE                                                       |
 * +---------------------------------------------------------------------+
 * 
 * 
 * 
 * 
 * 
 * 
 */
/*
 * HISTORY
 */
/*
**++
** FACILITY:	Digital Equipment Corporation
**		Multimedia Services for Alpha AXP
**		Development Kit Sample Code
**
** FILE NAME:	common.c
**
** PURPOSE:	Common routines.
**
** NOTES:	Public interface definition in common.h.
**--
*/
#include <stdio.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include <mme/mme_api.h>
#include "common.h"

/*
 * Global definitions.
 */
typedef unsigned long XPixel_t;


/*
**++
**  FUNCTIONAL_NAME: mmcSetVisual 
**
**  FUNCTIONAL_DESCRIPTION:
**      set up the colors in the private colormap from the RGBQUAD
**      structure.
**
**  FORMAL PARAMETERS:
**           dpy        - Display id 
**           depth      - Screen depth
**           visual     - Visual id, currenly 8 bit PseudoColor 
**           screenNum  - Display screen number
**
**  RETURN VALUE:
**           None
**
**  COMMENTS:
**
**  DESIGN:
**
**
*/
mmcSetVisual(Display *dpy, 
          unsigned int *depth, 
          Visual **visual,
          int screenNum, 
          int CmapStrategy,
	  int desiredDepth,
	  int desiredClass)
{
    XVisualInfo *visualList ;
    XVisualInfo info ;
    int match ;
    Visual *default_vis;
    int matchesDefault=False;
    int defaultEntry = 0;
    int i;

    
    /*
     * Get the default Visual type
     */
    default_vis = DefaultVisual(dpy, screenNum);
    
    /*
     * See if this display supports 8 bit pseudocolor
     * visual; exit if it does not.
     */
    info.screen = screenNum;
    info.depth  = desiredDepth;
    visualList = XGetVisualInfo(dpy,
                                VisualClassMask | VisualDepthMask
                                | VisualScreenMask,
                                &info, &match);
    if (match)
    {
       *visual = visualList[0].visual;
       *depth  = visualList[0].depth;
       for (i = 0 ; i < match ; i++)
       {
          if (visualList[i].visual == default_vis &&
	      visualList[i].screen == screenNum )
          {
             matchesDefault = True;
             *visual = visualList[i].visual;
             *depth  = visualList[i].depth;
             break;
          }
          else
          {
             matchesDefault = False;
          }
       }
    }
    else
    {
       fprintf(stderr, "\nCannot find an %d bit %s visual.\n",
	   desiredDepth, 
	   desiredClass == StaticGray ? "StaticGray" :
	   desiredClass == GrayScale ? "GrayScale" :
	   desiredClass == PseudoColor ? "PseudoColor" :
	   desiredClass == TrueColor ? "TrueColor" :
	   desiredClass == DirectColor ? "DirectColor" :
	   "unknown class");
       exit(0);
    }
 
    if ( !matchesDefault && CmapStrategy !=  PRIVATE_CMAP )
    {
       fprintf(stderr, "\nThe default visual must match the desired\n");
       fprintf(stderr, 
	   "visual to support a colormap strategy other than private.\n");
       fprintf(stderr, 
	  "This screen does not have a default visual supporting %d bit %s.\n",
	  desiredDepth,
	   desiredClass == StaticGray ? "StaticGray" :
	   desiredClass == GrayScale ? "GrayScale" :
	   desiredClass == PseudoColor ? "PseudoColor" :
	   desiredClass == TrueColor ? "TrueColor" :
	   desiredClass == DirectColor ? "DirectColor" :
	   "unknown class");
       fprintf(stderr,"You should try the private colormap strategy\n");
	   
       exit(0);
    }
    XFree (visualList);
}

/*
 *  Shared memory is used only if possible 
 */
static BOOL noxshm = FALSE;
static int error_handler (Display *display, XErrorEvent *error)
{
    noxshm = TRUE;
}

BOOL mmcAttach(Display *dpy,
               XShmSegmentInfo *shminfo)
{
    static int (*default_handle) ();
    default_handle = XSetErrorHandler(error_handler);

    if (!XShmQueryExtension(dpy))
        return (TRUE);

    if ( !XShmAttach(dpy, shminfo) )
    {
        fprintf(stderr, "Can't do the xshm attach\n");
    }
    XSync(dpy, False);
    XSetErrorHandler(default_handle);
    return noxshm;
}
              

int mmcInitDisplayWindow(Display *dpy,
   		         Visual *visual,
		         GC *gc,
		         Window *window,
                         Colormap colormap,
                         int cmapstrategy,
                         unsigned int scrnum,
                         int depth,
		         int width,
		         int height,
		         char *name,
			 Bool forceAspect,
			 int horizAspect,
			 int vertAspect)
{
    int fcolors ; 
    XSetWindowAttributes xswa ;
    unsigned  xswa_mask ;	      /* Fields to be looked at */
    XSizeHints hints ;
    int x, y;
    
    /*
     * Initialize window attributes structure and mask for images which
     * require colormap
     */
    xswa_mask =  CWEventMask | CWBackPixel | CWBorderPixel;
    xswa.event_mask = ExposureMask | ButtonPressMask | 
	StructureNotifyMask | ConfigureNotify;
    xswa.background_pixel = BlackPixel(dpy, scrnum);
    xswa.border_pixel = BlackPixel(dpy, scrnum);

    if ( cmapstrategy != DEFAULT_CMAP )
    {
       xswa_mask |= CWColormap;
       xswa.colormap = colormap;
    }
    
    /*
     * Create a window and set its attributes
     */
    x = 0 ;
    y = 0 ;
    *window = XCreateWindow(dpy,         /* Display ID               */
			   RootWindow (dpy, scrnum),/* Parent         */
			   x,	 	/* X coordinate             */
			   y, 		/* Y coordinate             */
			   width,	/* Width of window          */
			   height,	/* Height of window         */
			   5,           /* Border width             */
			   depth,	 /* Depth                    */
			   InputOutput,	/* Class                    */
			   visual,           /* Visual type              */
			   xswa_mask,	 /* Relevant fields to look  */
			   &xswa) ;           /* Window structure         */
    if (!*window)
    {
	fprintf(stderr, "Unable to Create Window.\n");
	return FALSE;
    }
    
    XStoreName(dpy, *window, name) ; /* window name */

    XSetIconName(dpy, *window, name) ; /* icon name */

    /*
     * Assign colormap to the Window
     */
    if ( cmapstrategy != DEFAULT_CMAP )
    {
	XSetWindowColormap(dpy, *window, colormap) ;
    }

    /*
     * graphics Context for root window
     */
    *gc = XCreateGC(dpy, *window, 0L, NULL) ;
    XSetForeground(dpy, *gc, BlackPixel(dpy,0)) ;
    XSetBackground(dpy, *gc, WhitePixel(dpy,0)) ;
    XSetLineAttributes(dpy, *gc, 0, LineOnOffDash, CapButt, JoinMiter) ;

    /*
     * Normally, fix the window's aspect ratio 
     */
    hints.flags = PSize | PAspect;
    hints.width  = hints.min_aspect.x = hints.max_aspect.x = width ;
    hints.height = hints.min_aspect.y = hints.max_aspect.y = height ;
    hints.x = 0;
    hints.y = 0;
    if ( forceAspect ) {
        hints.flags |= PPosition | USPosition | USSize | PResizeInc ;
        hints.width_inc = hints.min_aspect.x = hints.max_aspect.x = horizAspect;
        hints.height_inc = hints.min_aspect.y = hints.max_aspect.y = vertAspect;
    }
    XSetNormalHints(dpy, *window, &hints) ;
    
    /*
     * Dpy window
     */
    XMapWindow(dpy, *window) ;
    XSync(dpy, 0) ;			/* Flush output buffer */
    
    return TRUE;
}


void mmcMapImage (Display *dpy,
	          Visual *visual,
	          GC gc,
	          Window window,
	          int Lines,
	          int PixelsPerLine,
		  int depth,
	          XShmSegmentInfo *shminfo,
	          char * address,
                  int noxshm)
{
    XImage  *ximage;
#define DIV_ROUND_UP_(n,d) (((n) + (d) - 1) / (d))
    int    BytesPerLine = DIV_ROUND_UP_(PixelsPerLine, 4) * 4 * 
	(depth == 24 ? 4 : 1);

    /*
     * Set up input data structure for the image
     */
    ximage =
	XCreateImage(dpy,		/* X display identifier    */
		     visual,		/* Visual type             */
		     depth,		/* Depth of image          */
		     ZPixmap,		/* X Image format          */
		     0,			/* Initial offset in data  */
		     address, 		/* Address of data         */
		     PixelsPerLine, 	/* Width in pels           */
		     Lines,		/* Number of scanlines     */
		     XBitmapPad(dpy), 	/* Bitmap pad?             */
		     BytesPerLine); 	/* Bytes per scanline      */

    if ( !noxshm )
    {
       ximage->obdata = (char *)shminfo;

       /*
        * Display image
        */
       XShmPutImage(dpy, 
		    window, 
		    gc,		/* XDefaultGC(display,0),  */
		    ximage,                       
		    0,                           
		    0,                           
		    0,                               
		    0,                               
		    PixelsPerLine,
		    Lines,
		    False);
        ximage->obdata = NULL;
        ximage->data = NULL;
    }
    else
    {
       /*
        * Display image
        */
       XPutImage(dpy, 
	         window, 
	         gc,		/* XDefaultGC(display,0),  */
	         ximage,                       
	         0,                           
	         0,                           
	         0,                               
	         0,                               
	         PixelsPerLine,
	         Lines);
    }

    XFlush (dpy);
    XDestroyImage (ximage);
    XSync (dpy, False);
}				/* end of mapImage() */

int jvColorSetup(Display * display, 
		 Colormap * MyColormap, 
		 int * num, 
		 RGBQUAD * rgb )
{
  int i;
  int alloc_colors;
  int fixed_colors = 0;
  XPixel_t    pixels[MAX_COLORS];               /* Available Color Cells   */
  Colormap colormap;
  XColor xcolors[MAX_COLORS];

  *MyColormap = DefaultColormap (display, DefaultScreen(display));
  /*
   * initialize all returned arrays
   */
  for (i = 0; i < MAX_COLORS; i++)
    {
      rgb[i].rgbReserved = 1;
      rgb[i].rgbRed = rgb[i].rgbGreen = rgb[i].rgbBlue = 0;
    }

  /*
   * allocate as many color cells as possible
   */
  for (alloc_colors = MAX_COLORS; alloc_colors; alloc_colors--)
    if (XAllocColorCells (display, *MyColormap, False, NULL, 0,
			  pixels, alloc_colors))
      break;

  /*
   * clear the fixed flag for each color cell allocated
   */
  for (i = 0; i < alloc_colors; i++)
    rgb[pixels[i]].rgbReserved = 0;

  /*
   * prepare an array to receive values for all fixed colors
   */
  for (i = 0; i < MAX_COLORS; i++)
    if (rgb[i].rgbReserved == 1)
      xcolors[fixed_colors++].pixel = i;

  /*
   * ask for the values of all fixed colors
   */
  if (XQueryColors (display, *MyColormap, xcolors, fixed_colors))
    return -2;

  /*
   * store the values of all fixed colors in the output array
   */
  for (i = 0; i < fixed_colors; i++)
    {
      rgb[xcolors[i].pixel].rgbRed = xcolors[i].red;
      rgb[xcolors[i].pixel].rgbGreen = xcolors[i].green;
      rgb[xcolors[i].pixel].rgbBlue = xcolors[i].blue;
    }

  /*
   * return the number of allocated colors, if requested
   */
  *num = alloc_colors;

  return 0;
}

int jvColorCleanup (Display * display,
		    Colormap MyColormap,
		    RGBQUAD * rgb,
		    RGBQUAD * old)
{
    int i;
    int unused_colors = 0;
    int modified_colors = 0;
    XPixel_t pixels[MAX_COLORS];
    XColor xcolors[MAX_COLORS];
    unsigned char cr,cg,cb;
    int error;

    for (i = 0; i < MAX_COLORS; i++) {
	if ( old[i].rgbReserved == 0 ) {
      	    if ( rgb[i].rgbReserved == 0 ) {
		pixels[unused_colors++] = i;
	    }
       	    else {
	      	cr = rgb[i].rgbRed   & 0xff;
	      	cg = rgb[i].rgbGreen & 0xff;
	      	cb = rgb[i].rgbBlue  & 0xff;
	
	      	xcolors[modified_colors].red   = cr << 8;
	      	xcolors[modified_colors].green = cg << 8;
	      	xcolors[modified_colors].blue  = cb << 8;
	
	      	xcolors[modified_colors].pixel = i;
	      	xcolors[modified_colors].flags = DoRed | DoGreen | DoBlue;
	      	modified_colors++;
	    }
	}
	else {
	    /*
	     * For colors that were already in use, try to 
	     * allocate them read only so that the X server has a
	     * reference count bumped. This means the colors don't
	     * get put on the free list when the application that
	     * initially allocated them exits, which would leave
	     * this application with funny colors.
	     */
		 
		 if ( rgb[i].rgbReserved == 1 ) {
		     cr = old[i].rgbRed   & 0xff;
		     cg = old[i].rgbGreen & 0xff;
		     cb = old[i].rgbBlue  & 0xff;
	
		     xcolors[MAX_COLORS].red   = cr << 8;
		     xcolors[MAX_COLORS].green = cg << 8;
		     xcolors[MAX_COLORS].blue  = cb << 8;
	
		     xcolors[MAX_COLORS].pixel = i;
		     xcolors[MAX_COLORS].flags = DoRed | DoGreen | DoBlue;
		     if ( error = 
			 XAllocColor(display, MyColormap, 
				     &xcolors[MAX_COLORS]) > 1 ) 
			 fprintf(stderr,
				 "XAllocColor returned bad %d on color %d\n",
				 error,i);
		 }
	}

    }

    /*
     * free the unused colors
     */
    if (unused_colors &&
	(error = XFreeColors (display, MyColormap, pixels, unused_colors, 0L)))
	return -1;

    /*
     * tell X the values of all colors set by the dithering code
     */
    if (modified_colors) {
	for ( i = 0; i < modified_colors; i++ )
	    pixels[i] = xcolors[i].pixel;
	if ((error = 
	     XFreeColors (display, MyColormap, pixels, modified_colors, 0L)))
	    fprintf(stderr,"Could not free colors\n");
	for ( i = 0; i < modified_colors; i++ ) 
	    if ( error = XAllocColor(display, MyColormap, &xcolors[i]) > 1) 
		fprintf(stderr,"Could not store color %d\n",i);
    }

    if (error) return -2;

    for (i = 0 ; i < MAX_COLORS; i++)
	xcolors[i].pixel = i;
    XQueryColors (display, MyColormap, xcolors, MAX_COLORS);

    return 0;
}


/*
**++
**  FUNCTIONAL_NAME: printVideoError
**
**  FUNCTIONAL_DESCRIPTION:
**	prints the video capture error
**
**  FORMAL PARAMETERS:
**	video capture handle
**      status message
**
**  RETURN VALUE:
**      none
**
*/
void
printVideoError(
    LPHVIDEO lphvideo,
    MMRESULT status
)
{
    static char * error = NULL;
    if ( error == NULL ) error = (char *)mmcAllocMem(1024);
    videoGetErrorText(*lphvideo, status, error, 1024);
    fprintf(stderr,"%s\n", error);
}

