/*
 * COPYRIGHT NOTICE 
 * 
 * Copyright (c) Digital Equipment Corporation, 1993, 1994.
 * All Rights reserved. Unpublished rights reserved under the 
 * copyright laws of the United States. Copyright is claimed in 
 * the computer program and user interface thereof. 
 * 
 * The software contained on this media is proprietary to and 
 * embodies the confidential technology of Digital Equipment 
 * Corporation. Possession, use, duplication or dissemination of 
 * the software and media is authorized only pursuant to a valid
 * written license from Digital Equipment Corporation. 
 * 
 * The name of Digital Equipment Corporation may not be used to endorse or 
 * promote products derived from this software without specific prior 
 * written permission. All other rights reserved.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, IMPLIED WARRANTIES OF
 * NON-INFRINGEMENT, MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. 
 * Digital assumes no responsibility AT ALL for the use or reliability of 
 * this software.
 * 
 * 
 * +---------------------------------------------------------------------+
 * | USE, DUPLICATION OR DISCLOSURE BY THE U.S. GOVERNMENT IS SUBJECT TO | 
 * | RESTRICTIONS AS SET FORTH IN SUBPARAGRAPH (c)                       |
 * | DFARS 252.227-7013, OR IN FAR 52.227-19 OR IN 52.227-14 ALT III,    |
 * | AS APPLICABLE                                                       |
 * +---------------------------------------------------------------------+
 * 
 * 
 * 
 * 
 * 
 * 
 */
/*
 * HISTORY
 */
/*
**++
** FACILITY:	Digital Equipment Corporation
**		Multimedia Services for Alpha AXP
**		Development Kit Sample Code
**
** FILE NAME:	mmc_memory.c
**
** PURPOSE:	Shared memory manager
**
** NOTES:	Public interface definition in mmc_memory.h.
**--
*/

#include <mme/mme_api.h>
#include <stdio.h>
#include "mmc_memory.h"

#define DivRoundUp_(n, d) (((n) + (d) - 1) / (d))
#define RoundUp_(n, pad) (DivRoundUp_(n, pad) * (pad)) 
#define MEM_PAD 8

static mmcSharedMemList_p SharedMemoryList = NULL;


/*
** ++
** FUNCTION:  mmcAllocMem
**
** PURPOSE:  Allocate WAVEFORMAT data structure to be used in mme calls.
**
** INPUT:
**	size = size of data structure to be allocated
**
** OUTPUT: none.
**
** RETURN VALUE:
** 	NULL if memory allocation failed or
**      Address of the WAVEHDR data structure
**
** SIDE EFFECTS: none.
**
** NOTES:  none.
**--
*/
char * mmcAllocMem (int size)
     
{
    mmcMemType_e type;
    
    type = MEM_TYPE_DATA;
    return(mmcAlloc(type,size));
    
}



/*
** ++
** FUNCTION:  mmcAllocBuffer
**
** PURPOSE:  Allocate data buffer to be used in mme calls.
**
** INPUT:
**	size = size of data buffer to be allocated
**
** OUTPUT: none.
**
** RETURN VALUE:
** 	NULL if memory allocation failed or
**      Address of the data buffer
**
** SIDE EFFECTS: none.
**
** NOTES:  none.
**--
*/
char * mmcAllocBuffer (int size)
     
{
    mmcMemType_e type;
    
    type = MEM_TYPE_BUFFER;
    return(mmcAlloc(type,size));
    
}


/*
** ++
** FUNCTION:  mmcAlloc
**
** PURPOSE:  Allocate shared memory by the page(s). Keep track of actual
**           shared memory used and allow multiple functions to 'share'
**           the shared memory since shared memory is very expensive.
**
** INPUT:
**	size = size of share memory (in bytes) to be allocated
**
** OUTPUT:
**  SharedMemoryList = points to first shared memory page
**  SharedMemoryList->detail = pointer to linked list of detail breakup
**                             of an allocated shared memory page(s)
**  SharedMemoryList->next = pointer to linked list of allocated 
**                           shared memory pages
**
**
** RETURN VALUE:
**	NULL if memory allocation failed or
**      address of shared memory section
**
** SIDE EFFECTS:
**
** NOTES:  none.
**--
*/
char * mmcAlloc (mmcMemType_e type, int size)
     
{
    mmcSharedMemList_p prevList;
    mmcSharedMemList_p list;
    mmcSharedMemDetail_p detail;
    mmcSharedMemDetail_p prevDetail;
    char * sharedMemory;
    int pageSize;
    int byteSize;

    size = RoundUp_(size,MEM_PAD); 
    prevList = NULL;
    list = SharedMemoryList;
    sharedMemory = NULL;
    
    /*
     * If shared memory has already been allocated, look through link list
     * of shared memory looking for chunk of memory large enough for current 
     * request. Note, only the shared memory link list (not its detail 
     * linked list) is searched. This means that if a previously allocated
     * piece of memory is freed up and is not the last chunk for that shared
     * memory, it will not be reused. Only the last chunk of shared memory
     * is checked. Reason is that this is a temporary fix, plan to remove 
     * shared memory restrictions in the near future.
     */
    if (list != NULL)
    {
	while (list != NULL) 
	{
	    /* Found a chunk of memory to satisfy current request */
	    if ((list->currentMemory + size) < list->endMemory)
	    {
		/* go to end of shared memory's detail list */
		detail = list->detail;
		while (detail != NULL)
		{
		    prevDetail = detail;
		    detail = prevDetail->next;
		}
		/* update shared memory's memory information */
		sharedMemory = list->currentMemory;
		list->currentMemory += size;
		
		/* set up detail information */
		detail = 
		  (mmcSharedMemDetail_p)malloc(sizeof(mmcSharedMemDetail_t));
		detail->memory = sharedMemory;
		detail->size = size;
		detail->type = type;
		detail->next = NULL;

		/* add to detail linked list*/
		prevDetail->next = detail;
		break;
	    }
	    /* get ne4xt shared memory page in linked list */
	    prevList = list;
	    list = list->next;
	} /* end while */
    }
    /* No shared memory allocated with enough size for current request */
    if (list == NULL)
    {
	/* calculate the number of pages need to meet the request */
	pageSize = getpagesize();
	byteSize = DivRoundUp_(size, pageSize);
	byteSize = byteSize * pageSize;
	if (type == MEM_TYPE_DATA)
	  sharedMemory = mmeAllocMem(byteSize);
	else
	  sharedMemory = mmeAllocBuffer(byteSize);

	/* if shared memory allocated, then ... */
	if (sharedMemory != NULL)
	{
	    /* set up shared memory list information */
	    list = 
	      (mmcSharedMemList_p) malloc(sizeof(mmcSharedMemList_t)); 
	    list->startMemory = sharedMemory;
	    list->endMemory = sharedMemory + byteSize;
	    list->currentMemory = sharedMemory + size;
	    list->next = NULL;
	    if (prevList != NULL)
	      prevList->next = list;

	    /* set up detail information */
	    detail = 
	      (mmcSharedMemDetail_p) malloc(sizeof(mmcSharedMemDetail_t));
	    detail->memory = sharedMemory;
	    detail->size = size;
	    detail->type = type;
	    detail->next = NULL;
	    
	    /* add shared memory list to linked list */
	    list->detail = detail;
	    if (SharedMemoryList == NULL)
	      SharedMemoryList = list;
	}
    } /* (list == NULL) */	
    
    if (!sharedMemory)
    {
	mmcErrorDisplay("Can't allocate shared memory",0,NULL);
	mmcErrorDisplay("Server might be down?",0,NULL);
    }
    return (sharedMemory);
}


/*
** ++
** FUNCTION:  mmcFreeMem
**
** PURPOSE:  Free shared memory allocated via mmcAllocMem
**
** INPUT:
**	sharedMemory = pointer to chunk of shared memory to be freed
**
** OUTPUT: none.
**
** RETURN VALUE:
** 	status code:
**		TRUE if successful
**		FALSE if memory pointer is not valid
**
** SIDE EFFECTS: none.
**
** NOTES:  none.
**--
*/
int mmcFreeMem (char * sharedMemory)
     
{
    mmcMemType_e type;

    type = MEM_TYPE_DATA;

    return(mmcFree(type, sharedMemory));
}


/*
** ++
** FUNCTION:  mmcFreeBuffer
**
** PURPOSE:  Free shared memory allocated via mmcAllocBuffer
**
** INPUT:
**	sharedMemory = pointer to chunk of shared memory to be freed
**
** OUTPUT: none.
**
** RETURN VALUE:
** 	status code:
**		TRUE if successful
**		FALSE if memory pointer is not valid
**
** SIDE EFFECTS: none.
**
** NOTES:  none.
**--
*/
int mmcFreeBuffer (char * sharedMemory)
     
{
    mmcMemType_e type;

    type = MEM_TYPE_BUFFER;

    return(mmcFree(type, sharedMemory));
}


/*
** ++
** FUNCTION:  mmcFree
**
** PURPOSE:  Free chunk of shared memory
**
** INPUT:
**      type = MEM_TYPE_DATA indicated memory allocated via mmcAllocMem;
**             MEM_TYPE_BUFFER indicates memory allocated via mmcAllocBuffer
**	sharedMemory = pointer to shared memory to be freed
**
** OUTPUT: 
**      SharedMemList = points to first allocated shared memory page
**      SharedMemoryList->next = if freed shared memory chunk is the only
**                               chunk allocated from shared memory page, then
**                               remove entry from linked list
**      SharedMemoryList->detail = remove detail breakup of freed shared memory
**                                 chunk from linked list
**
** RETURN VALUE:
**	TRUE = success
**	FALSE = memory pointed to is not valid
**
** SIDE EFFECTS: none.
**
** NOTES:  none.
**--
*/
int mmcFree (mmcMemType_e type, char * sharedMemory)
     
{
    mmcSharedMemList_p prevList;
    mmcSharedMemList_p list;
    mmcSharedMemDetail_p prevDetail;
    mmcSharedMemDetail_p detail;
    int pageSize;
    int status = 1;

    prevList = NULL;
    list = SharedMemoryList;
    
    if (list == NULL)
      return (0);
    
    while (list != NULL) 
    {
	if ((sharedMemory >= list->startMemory) && 
	    (sharedMemory <= list->endMemory))
	{
	    prevDetail = NULL;
	    detail = list->detail;
	    while (detail != NULL)
	    {
		if ((detail->memory == sharedMemory) && 
		    (detail->type == type))
		{
		    /* sharedMemory found, update detail information */
		    if (prevDetail == NULL)
		      list->detail = detail->next;
		    else
		    {
			prevDetail->next = detail->next;
			/* if last chunk, then update list->currentMemory */
			if (detail->next == NULL)
			{
			    list->currentMemory = prevDetail->memory +
			      prevDetail->size;
			}
		    }
		    free(detail);
		    break;
		}
		/* Get next detail */
		prevDetail = detail;
		detail = detail->next;
	    } /* (detail != NULL) */

	    if (detail != NULL)
	    {
		/* if sharedMemory found, then update list info if necessary */
		if (list->detail == NULL)
		{ 
		    if (prevList == NULL)
		      SharedMemoryList = list->next;
		    else
		      prevList->next = list->next;
		    
		    if (type == MEM_TYPE_DATA)
		      status = mmeFreeMem(list->startMemory);
		    else
		      status = mmeFreeBuffer(list->startMemory);
		    free(list);
		}
		break;
	    } /* (detail != NULL) */
	} /* ((sharedMemory >= list->startMemory) && (...)) */

	/* Get next list */
	prevList = list;
	list = list->next;
    } /* (list != NULL) */
    return (status);
}



/*
** ++
** FUNCTION:  mmcFreeAll
**
** PURPOSE:  Free all shared memory
**
** INPUT: none.
**
** OUTPUT: none.
**
** RETURN VALUE:
** 	status code:
**		MMSYSERR_NOERROR = no error.
**		MMSYSERR_NOMEM = memory allocation failed.
**
** SIDE EFFECTS: none.
**
** NOTES:  none.
**--
*/
void mmcFreeAll ()
     
{
    mmcSharedMemList_p list;
    mmcMemType_e type;
    char * sharedMemory;
    int status;

    while (SharedMemoryList != NULL)
    {
	list = SharedMemoryList;
	sharedMemory = list->detail->memory;
	type = list->detail->type;
	status = mmcFree(type, sharedMemory);
	if (!status)
	  mmcErrorDisplay("Error freeing shared memory %d of type %d",
			  sharedMemory, type);
    }
}


/*
** ++
** FUNCTION:  mmcDisplayMem
**
** PURPOSE:  Debug routine that displays shared memory information
**
** INPUT: none.
**
** OUTPUT: none.
**
** RETURN VALUE: none.
**
** SIDE EFFECTS: 
**      Display in stderr shared memory information
**
** NOTES:  none.
**--
*/
static MMRESULT mmcDisplayMem ()
     
{
    mmcSharedMemList_p prevList;
    mmcSharedMemList_p list;
    mmcSharedMemDetail_p prevDetail;
    mmcSharedMemDetail_p detail;
    int pageSize;

    prevList = NULL;
    list = SharedMemoryList;
    
    if (list == NULL)
      return (0);

    while (list != NULL) 
    {
	mmcStatusDisplay("Shared Memory %x %x %x ---  %x  ---- %x",list->startMemory, list->endMemory, list->currentMemory, list->detail, list->next);
	prevDetail = NULL;
	detail = list->detail;
	while (detail != NULL)
	{
	    mmcStatusDisplay("...Detail %x --- %x (%d) %d %d %x",detail, detail->memory,detail->memory,detail->size, detail->type, detail->next);
	    detail=detail->next;
	}
	list = list->next;
    }
    mmcStatusDisplay("---------------------------------------------\n");

    return (0);
}

