/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include "cmCookies.h"
#include "cmBufferlist.h"
#include "cmChunk.h"
#include "cmLTS.h"
#include "cmAtom.h"

#ifndef O_BINARY
#    define O_BINARY 0
#endif

#define ONE_BILLION 1000000000.0
#define CF_READ_BUFFER_SIZE 512

#define MPEG_I_FRAME 1
#define MPEG_P_FRAME 2
#define MPEG_B_FRAME 3
#define MPEG_SYS_HDR_START_CODE 0x000001b3
#define MPEG_END_SEQ_CODE       0x000001b7
#define MPEG_PIC_START_CODE     0x00000100

#define MPEG_BUF_HDR_SIZE (sizeof(MPEGBufferHeader)-8)

typedef struct RefEntry {
  int type;
  int ref1;
  int ref2;
  int headerRef;   /* -1 = default matrices. */
  
  /* The following fields are not stored in file but only in memory. */

  int indexNum;
  int marker;
} RefEntry;

typedef struct HdrEntry {
    int intra_matrix_flag;
    int intra_matrix[64];
    int non_intra_matrix_flag;
    int non_intra_matrix[64];
} HdrEntry;

/*
 * The following macro is the size of a ref entry on disk.
 */

#define MPEG_REF_ENTRY_SIZE 16

/* 
 * Structure kept for each segment in segment list.
 */
typedef struct Segment {
    double clipStart;
    double clipEnd;
    double logicalStart;
    double logicalEnd;
    struct Segment *next;
    struct Segment *prev;
    int *FOT;
    double *FTT;
    RefEntry *refTable;
    int *indexTable;
    int numFrames;
    int maxSize;
    int segId;
} Segment;

/*
 * Structure used when picking next set of frames to be sent.
 */

typedef struct {
    Segment *segPtr;
    int index;
    int offset;
    int size;
    double sTime;
    double eTime;
    int sendFlag;
    int displayFlag;
} FramePick;

/*
 * Structure kept for each clipfile object. 
 */

typedef struct MPEG_CF {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    ClientData lts;		/* LTS for object to use. */
    char *ltsName;
    int ltsTriggerNum;             /* LTS trigger for changed lts. */
    double sendAhead;		/* Send ahead time. */
    int bandwidth;		/* Bandwidth available */
    double cycleTime;		/* Cycle period. */
    char *filename;		/* Filename of clipfile. */
    tChunkFileState *cfState;   /* Pointer to chunkfile structure. */
    int *rawFOT;                /* Pointer to FOT if file is raw data. */
    double *rawFTT;             /* Pointer to FTT if file is raw data. */
    RefEntry *rawRefTable;      /* Pointer to RefTable if file is raw data. */
    int rawfd;                  /* File descriptor for raw file. -1 if not open. */
    char *outCmd;		/* Command to send buffers to. */
    double maxFrameRate;	/* Maximum frame rate. */
    double extraFrameCounter;   /* Used to adjust number of frames sent per cycleTime
				   to match with maxFrameRate */
    int extraBWCounter;         /* Used to adjust number of bytes send per cycleTime
				   to mactch with bandwidth restrictions. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;           /* ClientData for outCCmd. */
    int nextSegId;		/* Next segment id. */
    int segmentListChangeFlag;  /* Indicates if anything has happened to segment list. 
				   Will cause ready command to check for overlaps. */
    Segment *segments;		/* List of segment structures. */
    Segment *lastSeg;           /* Segment from which data was sent last. */
    int lastFrameNum;           /* Frame number of last frame sent. */
    MPEGHeader *cfInfo;	        /* MPEG Header from clipfile. */
    int numFrames;		/* Number of frames in clipfile. */
    double clipLength;		/* Length of clipfile. */
    int dataStart;              /* Offset to start of data. */
    int ready;                  /* Ready flag. */
    int at_token;               /* Token for at event. */
    int lastMarker;             /* Used to keep track of what has been sent. */
    int numHeaders;
    HdrEntry *headerTable;      /* Table of offset and lengths for system and GOP
				   headers. */
} MPEG_CF;

/* Globals used by this object module. */

static int nextMPEG_CFNum = 0;	/* Variable for next unique object name. */

/* Forward declarations. */

int MPEG_CFObjCmd(ClientData clientData, Tcl_Interp * interp, int argc, char **argv);
static int MPEG_CF_AddSegment(MPEG_CF * objPtr, double clipStart, double clipEnd, 
				  double logStart, double logEnd, 
				  int logStartFlag, int logEndFlag);
static int MPEG_CF_RemoveSegment(MPEG_CF *objPtr, int segId);
static int MPEG_CF_ClearSegments(MPEG_CF *objPtr);
static int MPEG_CF_SegmentInfo(MPEG_CF *objPtr, int segId);
static int MPEG_CF_Info(MPEG_CF *objPtr);
static int MPEG_CF_Ready(MPEG_CF *objPtr);
static int MPEG_CF_Unready(MPEG_CF *objPtr);
static int MPEG_CF_ChangeOutCmd(MPEG_CF *objPtr, char *newOut);
static int MPEG_CF_OpenClipfile(MPEG_CF *objPtr, char *filename);
void MPEG_CF_LTSChanged(ClientData clientdata);
void MPEG_CF_SendData(ClientData clientData);
static void MPEG_CF_ResetPickList ();
static int MPEG_CF_AddToPickList(MPEG_CF *objPtr, Segment *segPtr, int index,
				 int speedFlag, int recurseFlag);
static SB_List *MPEG_CF_MakeSBListFromPickList(MPEG_CF *objPtr);
static int MPEG_CF_FrameInfo(MPEG_CF *objPtr, int frameOffset);
static int MPEG_CF_ConfigureSegment(MPEG_CF *objPtr, int segmentId, double clipStart, 
				    double clipEnd, double logStart, double logEnd, 
				    int logStartFlag, int logEndFlag);
static int MPEG_CF_SegmentList(MPEG_CF *objPtr);


/* Static pick list variables. */
static int pickListCounter = 0;
static int pickListLength = 0;
static FramePick *pickList = NULL;

/************************************************************
 *
 * CreateNewMPEG_CF
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

MPEG_CF *
CreateNewMPEG_CF(Tcl_Interp *interp,  /* Pointer to interpreter. */
		 char *name           /* Name of new object. */)
{

    MPEG_CF *new;

    new = (MPEG_CF *) ckalloc(sizeof(MPEG_CF));

    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);
    new->lts = NULL;
    new->ltsName = NULL;
    new->ltsTriggerNum = -1;
    new->sendAhead = 0.2;
    new->bandwidth = 0;
    new->cycleTime = 0.0;
    new->filename = NULL;
    new->cfState = NULL;
    new->rawFOT = NULL;
    new->rawFTT = NULL;
    new->rawRefTable = NULL;
    new->rawfd = -1;
    new->outCmd = NULL;
    new->maxFrameRate = 0.0;
    new->extraFrameCounter = 0.0;
    new->extraBWCounter = 0;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->nextSegId = 0;
    new->segmentListChangeFlag = 0;
    new->segments = NULL;
    new->lastSeg = NULL;
    new->lastFrameNum = -1;
    new->cfInfo = NULL;
    new->numFrames = 0;
    new->clipLength = 0.0;
    new->dataStart = 0;
    new->ready = 0;
    new->at_token = 0;
    new->lastMarker = 0;
    new->numHeaders = 0;
    new->headerTable = NULL;
    return new;
}

/************************************************************
 *
 * DestroyMPEG_CF
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

int
DestroyMPEG_CF(MPEG_CF *objPtr       /* Pointer to clipfile object. */)
{

    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
	objPtr->at_token = 0;
    }

    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    if (objPtr->ltsName != NULL) {
	ckfree((char *) objPtr->ltsName);
    }

    if (objPtr->filename != NULL) {
	free(objPtr->filename);
	objPtr->filename = NULL;
    }

    if (objPtr->cfState != NULL) {
	Chunk_CloseFile(objPtr->cfState);
	objPtr->cfState = NULL;
    }

    if (objPtr->rawfd > -1) {
	close(objPtr->rawfd);
	objPtr->rawfd = -1;
    }

    if (objPtr->rawFTT != NULL) {
	ckfree((char *) objPtr->rawFTT);
	objPtr->rawFTT = NULL;
    }

    if (objPtr->rawFOT != NULL) {
	ckfree((char *) objPtr->rawFOT);
	objPtr->rawFOT = NULL;
    }

    if (objPtr->rawRefTable != NULL) {
	ckfree((char *) objPtr->rawRefTable);
	objPtr->rawRefTable = NULL;
    }

    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);;
    }

    if (objPtr->headerTable != NULL) {
	ckfree((char *)objPtr->headerTable);
	objPtr->headerTable = NULL;
    }

    if (objPtr->outCmd != NULL)
	ckfree((char *) objPtr->outCmd);

    /*
     * Free all segments.
     */
    MPEG_CF_ClearSegments(objPtr);

    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureMPEG_CF
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

int
ConfigureMPEG_CF(MPEG_CF * objPtr, int argc, char **argv)
{
  
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
  
    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {} {}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName, "}} ",
		    (char *) NULL);
	}
	Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	Tcl_AppendResult(interp, "{-sendAhead {0.2} ", str, "} ", (char *)NULL);

	sprintf(str, "{%d}", objPtr->bandwidth);
	Tcl_AppendResult(interp, "{-bandwidth {0} ", str, "} ", (char *)NULL);

	Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	Tcl_AppendResult(interp, "{-cycleTime {0.0} ", str, "} ", (char *)NULL);

	if (objPtr->filename == NULL) {
	    Tcl_AppendResult(interp, "{-filename {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-filename {} {", objPtr->filename,
			    "}} ",(char *)NULL);
	}
	
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}

	Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	Tcl_AppendResult(interp, "{-maxFrameRate {0.0} ", str,
			  "} ", (char *)NULL);
	return TCL_OK;
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->lts == NULL) {
		Tcl_AppendResult(interp, "-lts {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName, "} ",
				  (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	    Tcl_AppendResult(interp, "-sendAhead {0.2} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'b') && (strncmp("-bandwidth", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->bandwidth);
	    Tcl_AppendResult(interp, "-bandwidth {0} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	    Tcl_AppendResult(interp, "-cycleTime {0.0} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'f') && (strncmp("-filename", argv[0], len) == 0)) {
	    if (objPtr->filename == NULL) {
		Tcl_AppendResult(interp, "-filename {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-filename {} {",
			  objPtr->filename, "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	    Tcl_AppendResult(interp, "-maxFrameRate {0.0} ", str, (char *)NULL);
	    return TCL_OK;
	}
	Tcl_AppendResult(interp, "bad configuration option: ",
			  argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	/* ------------------------- LTS ------------------------- */
	if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newlts;
	    
	    /*
	     * Check args
	     */
	    cntr++;
	    if (cntr == argc) {
argcError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr], (char *)NULL);
		return TCL_ERROR;
	    }

	    /*
	     * If resetting to null value, just free up old lts.
	     */
	    if (strlen(argv[cntr]) == 0) {
		if (objPtr->lts != NULL) {
		    if (objPtr->ltsTriggerNum != -1) {
			LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
			objPtr->ltsTriggerNum = -1;
		    }
		    Tk_Release(objPtr->lts);
		    objPtr->lts = NULL;
		}
		if (objPtr->ltsName != NULL) {
		    ckfree((char *) objPtr->ltsName);
		    objPtr->ltsName = NULL;
		}
	    } else {
		newlts = (ClientData)Tcm_FindAtom(interp, argv[cntr]);
		if (newlts == NULL) {
valueError:	    
		    Tcl_AppendResult(interp, "bad value \"", argv[cntr],
				  "\" given for configuration option ",
				  argv[cntr-1], (char *)NULL);
		    return TCL_ERROR;
		}
		
		/*
		 * Blow away old callback.
		 */
		if (objPtr->lts != NULL) {
		    if (objPtr->ltsTriggerNum != -1) {
			LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
			objPtr->ltsTriggerNum = -1;
		    }
		    Tk_Release(objPtr->lts);
		    objPtr->lts = NULL;
		}
		objPtr->lts = newlts;
		Tk_Preserve (objPtr->lts);

		/*
		 * Copy in new lts name
		 */
		if (objPtr->ltsName != NULL) {
		    ckfree((char *) objPtr->ltsName);
		}
		objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
		strcpy(objPtr->ltsName, argv[cntr]);
		
		/*
		 * Set up new lts trigger
		 */
		objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts,
				LTS_SPEED_FLAG | LTS_OFFSET_FLAG | LTS_VALUE_FLAG,
			        MPEG_CF_LTSChanged, (ClientData) objPtr);
	    }

	    MPEG_CF_Unready(objPtr);

	/* ------------------------- SEND_AHEAD ------------------------- */
	} else if ((c=='s') && (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	    double newSendAhead;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->sendAhead = newSendAhead;

	/* ------------------------- BANDWIDTH -------------------------- */
	} else if ((c=='b') && (strncmp("-bandwidth", argv[cntr], len) == 0)) {
	    int newBandwidth;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newBandwidth) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->bandwidth = newBandwidth;
	    objPtr->extraBWCounter = 0;

	/* ------------------------- CYCLE_TIME ------------------------- */
	} else if ((c=='c') && (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	    double newCycleTime;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->cycleTime = newCycleTime;

	/* ------------------------- FILE_NAME -------------------------- */
	} else if ((c=='f') && (strncmp("-filename", argv[cntr], len) == 0)) {
	    Tcl_DString fname_dstring;
	    char *fname;

	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (objPtr->filename != NULL) {
		Tcl_AppendResult(interp,
				"Can not set clipfile name more than once.",
				(char *)NULL);
		return TCL_ERROR;
	    }

	    fname = Tcl_TildeSubst(interp, argv[cntr], &fname_dstring);

	    if (MPEG_CF_OpenClipfile(objPtr, fname) != TCL_OK) {
		Tcl_DStringFree(&fname_dstring);
		return TCL_ERROR;
	    }
	    
	    objPtr->filename = (char *) ckalloc(strlen(fname) + 1);
	    strcpy(objPtr->filename, fname);
	    Tcl_DStringFree(&fname_dstring);

	/* ------------------------- OUT_CMD ---------------------------- */
	} else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    
	    if (MPEG_CF_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }

	    MPEG_CF_Unready(objPtr);
	    
	/* ------------------------- MAX_FRAME_RATE --------------------- */
	} else if ((c=='m') &&
		   (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
	    double newMaxFrameRate;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->maxFrameRate = newMaxFrameRate;
	    objPtr->extraFrameCounter = 0.0;

	/* ------------------------- UNKNOWN OPTION ----------------------*/
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			  argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CFCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the mpegFile command.
 *     A new object of type MPEG_CF is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
MPEG_CFCreateCmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char **argv)
{
    char new_name[256];
    MPEG_CF *new;
    
    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextMPEG_CFNum++;
    sprintf(new_name, "%smpgc%d", argv[1], nextMPEG_CFNum);
    
    new = CreateNewMPEG_CF(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MPEG_CFObjCmd,
		      (ClientData) new, (void (*)()) NULL);
    
    if (argc > 2) {
	ConfigureMPEG_CF(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CFObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     MPEG_CF object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
MPEG_CFObjCmd(ClientData clientData, Tcl_Interp * interp,
		int argc, char **argv)
{
    char c;
    int len, i;
    MPEG_CF *objPtr = (MPEG_CF *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    /*------------------------- CONFIGURE ------------------------------*/
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureMPEG_CF(objPtr, argc - 2, argv + 2);

    /*------------------------- DESTROY --------------------------------*/
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyMPEG_CF(objPtr);

    /*------------------------- ADD_SEGMENT -------------------------*/
    } else if ((c == 'a') && (strncmp("addSegment", argv[1], len) == 0)) {
	double clipStart;
	double clipEnd;
	double logStart = 0.0;
	double logEnd = 0.0;
	int logStartFlag = 0;
	int logEndFlag = 0;

	if (argc < 4) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			" addSegment <clipStart> <clipEnd> ",
			"?-start <logicalStart>? ?-end <logicalEnd>? ",
			(char *) NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetDouble(interp, argv[2], &clipStart) != TCL_OK) {
	    Tcl_AppendResult(interp,
		"error found while parsing clip start time ", (char *)NULL);
	    return TCL_ERROR;
	}

	if (strcmp(argv[3], "end") == 0) {
	    clipEnd = HUGE_VAL;
	} else if (Tcl_GetDouble(interp, argv[3], &clipEnd) != TCL_OK) {
	    Tcl_AppendResult(interp,
		  "error found while parsing clip end time ", (char *)NULL);
	    return TCL_ERROR;
	}

	for (i = 4; i < argc; i++) {
	    char ac;
	    int alen;

	    ac = argv[i][1];
	    alen = strlen(argv[i]);

	    if ((ac == 's') && (strncmp("-start", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Must specify parameter value ",
			      "after optional argument flag ", argv[i],
			      " passed to ", argv[0], " ", argv[1],
			      (char *) NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logStart) != TCL_OK) {
		    Tcl_AppendResult(interp,
				"error found while logical start end time ",
				(char *)NULL);
		    return TCL_ERROR;
		}
		logStartFlag = 1;
	    } else if ((ac == 'e') && (strncmp("-end", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Must specify parameter value ",
			      "after optional argument flag ", argv[i],
			      " passed to ", argv[0], " ", argv[1],
			      (char *) NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logEnd) != TCL_OK) {
		    Tcl_AppendResult(interp,
				"error found while logical end end time ",
				(char *)NULL);
		    return TCL_ERROR;
		}
		logEndFlag = 1;
	    }
	  }

	return MPEG_CF_AddSegment(objPtr, clipStart, clipEnd,
				  logStart, logEnd, logStartFlag, logEndFlag);

    /*------------------------- CONFIGURE_SEGMENT -------------------------*/
    } else if ((c == 'c') && (strncmp("configureSegment", argv[1], len) == 0)) {
	double clipStart;
	double clipEnd;
	double logStart = 0.0;
	double logEnd = 0.0;
	int logStartFlag = 0;
	int logEndFlag = 0;
	int segmentId;

	if (argc < 5) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			" configureSegment <segmentId> <clipStart> <clipEnd> ",
			"?-start <logicalStart>? ?-end <logicalEnd>? ",
			(char *) NULL);
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[2], &segmentId) != TCL_OK) {
	  Tcl_AppendResult(interp, "error found while parsing segment id ",
			   (char *) NULL);
	  return TCL_ERROR;
	}

	if (Tcl_GetDouble(interp, argv[3], &clipStart) != TCL_OK) {
	    Tcl_AppendResult(interp,
		"error found while parsing clip start time ", (char *)NULL);
	    return TCL_ERROR;
	}

	if (strcmp(argv[4], "end") == 0) {
	    clipEnd = HUGE_VAL;
	} else if (Tcl_GetDouble(interp, argv[4], &clipEnd) != TCL_OK) {
	    Tcl_AppendResult(interp,
		  "error found while parsing clip end time ", (char *)NULL);
	    return TCL_ERROR;
	}

	for (i = 5; i < argc; i++) {
	    char ac;
	    int alen;

	    ac = argv[i][1];
	    alen = strlen(argv[i]);

	    if ((ac == 's') && (strncmp("-start", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Must specify parameter value ",
			      "after optional argument flag ", argv[i],
			      " passed to ", argv[0], " ", argv[1],
			      (char *) NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logStart) != TCL_OK) {
		    Tcl_AppendResult(interp,
				"error found while logical start end time ",
				(char *)NULL);
		    return TCL_ERROR;
		}
		logStartFlag = 1;
	    } else if ((ac == 'e') && (strncmp("-end", argv[i], alen) == 0)) {
		i++;
		if (i == argc) {
		    Tcl_AppendResult(interp, "Must specify parameter value ",
			      "after optional argument flag ", argv[i],
			      " passed to ", argv[0], " ", argv[1],
			      (char *) NULL);
		    return TCL_ERROR;
		}
		if (Tcl_GetDouble(interp, argv[i], &logEnd) != TCL_OK) {
		    Tcl_AppendResult(interp,
				"error found while logical end end time ",
				(char *)NULL);
		    return TCL_ERROR;
		}
		logEndFlag = 1;
	    }
	  }

	return MPEG_CF_ConfigureSegment(objPtr, segmentId, clipStart, clipEnd,
				  logStart, logEnd, logStartFlag, logEndFlag);

    /*------------------------- REMOVE_SEGMENT -------------------------*/
    } else if ((c == 'r') && (strncmp("removeSegment", argv[1], len) == 0)) {
	int segId;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			argv[0], " removeSegment <segmentId> ", (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &segId) != TCL_OK) {
	    Tcl_AppendResult(interp, "Bad integer parameter value ",
			argv[2], " given to: ", argv[0], " ", argv[1],
			(char *) NULL);
	    return TCL_ERROR;
	}
	return MPEG_CF_RemoveSegment(objPtr, segId);

    /*------------------------ SEGMENT LIST ----------------------------*/
    } else if ((c == 's') && (strncmp("segmentList", argv[1], len) == 0)) {
	return MPEG_CF_SegmentList(objPtr);
    
    /*------------------------- CLEAR_SEGMENTS -------------------------*/
    } else if ((c == 'c') && (strncmp("clearSegments", argv[1], len) == 0)) {
	return MPEG_CF_ClearSegments(objPtr);

    /*------------------------- SEGMENT_INFO ---------------------------*/
    } else if ((c == 's') && (strncmp("segmentInfo", argv[1], len) == 0)) {
	int segId;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			argv[0], " segmentInfo <segmentId> ", (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &segId) != TCL_OK) {
	    Tcl_AppendResult(interp, "Bad integer parameter value ", argv[2],
			" given to: ", argv[0], " ", argv[1], (char *) NULL);
	    return TCL_ERROR;
	}
	return MPEG_CF_SegmentInfo(objPtr, segId);


    /*------------------------- INFO -----------------------------------*/
    } else if ((c == 'i') && (strncmp("info", argv[1], len) == 0)) {
	return MPEG_CF_Info(objPtr);

    /*------------------------- READY ----------------------------------*/
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return MPEG_CF_Ready(objPtr);

    /*------------------------- UNREADY --------------------------------*/
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return MPEG_CF_Unready(objPtr);

    /*------------------------ FRAME INFO ------------------------------*/
    } else if ((c == 'f') && (strncmp("frameInfo", argv[1], len) == 0)) {
	int frame_offset;

	if (argc == 3) {
	    if (Tcl_GetInt(interp, argv[2], &frame_offset) != TCL_OK) {
		Tcl_AppendResult(interp, "Bad frame offset value ", argv[2],
				 " gvien to: ", argv[0], " ", argv[1], 
				 (char *) NULL);
		return TCL_ERROR;
	    }
	} else {
	    frame_offset = 0;
	}

	return MPEG_CF_FrameInfo(objPtr, frame_offset);
    }
    Tcl_AppendResult(interp, "Bad command option: ",
		argv[0], argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  addSegment <clipStart> <clipEnd>",
			" ?-start <logicalStart>? ?-end <logicalEnd>?\n",
                "  removeSegment <segmentId>\n",
                "  clearSegments\n",
                "  configureSegment <segmentId> <clipStart> <clipEnd>",
			" ?-start <logicalStart>? ?-end <logicalEnd>?\n",
		"  segmentList\n",
                "  segmentInfo <segmentId>\n",
                "  ready\n",
                "  unready\n",
                "  info\n",
		"  frameInfo\n",
                (char *) NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * MPEG_CFModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
MPEG_CFModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mpegFile", MPEG_CFCreateCmd, NULL,
		(void (*)()) NULL);
    return Tcl_GlobalEval(interp, "cminfo add object_info mpegFile  {mpeg}");
}

static double
GetTimeFromFile(int fd)
{
    int sec, nsec;
    double res;

    read(fd, &sec, 4);
    read(fd, &nsec, 4);
    
    sec = ntohl(sec);
    nsec = ntohl(nsec);

    res = ((double) sec) + (((double) nsec) / ONE_BILLION);
    
    return res;
}

static void
GetOffsetFromFile(int fd, int *start, int *length)
{
    int tstart, tlength;

    read(fd, &tstart, 4);
    read(fd, &tlength, 4);
    
    (*start) = ntohl(tstart);
    (*length) = ntohl(tlength);

    return;
}

static void
GetRefEntryFromFile(int fd, RefEntry *refEntryPtr) 
{
    read(fd, &(refEntryPtr->type), 4);
    read(fd, &(refEntryPtr->ref1), 4);
    read(fd, &(refEntryPtr->ref2), 4);
    read(fd, &(refEntryPtr->headerRef), 4);

    refEntryPtr->type = ntohl(refEntryPtr->type);
    refEntryPtr->ref1 = ntohl(refEntryPtr->ref1);
    refEntryPtr->ref2 = ntohl(refEntryPtr->ref2);
    refEntryPtr->headerRef = ntohl(refEntryPtr->headerRef);

    return;
}

static void 
MPEG_CF_GetHeaderEntryFromFile(int fd, HdrEntry *hdrEntryPtr)
{
    unsigned char readBuffer[130];
    int idx = 0, i;

    read(fd, readBuffer, 130);

    hdrEntryPtr->intra_matrix_flag = readBuffer[idx++];
    if (hdrEntryPtr->intra_matrix_flag) {
	for (i=0; i<64; i++) {
	    hdrEntryPtr->intra_matrix[i] = readBuffer[idx++];
	}
    }
    hdrEntryPtr->non_intra_matrix_flag = readBuffer[idx++];
    if (hdrEntryPtr->non_intra_matrix_flag) {
	for (i=0; i<64; i++) {
	    hdrEntryPtr->non_intra_matrix[i] = readBuffer[idx++];
	}
    }

    return;
}

 
static void
CopyRefEntry(RefEntry *r1, RefEntry *r2)
{
    memcpy((char *) r1, (char *) r2, sizeof(RefEntry));
    return;
}

    
/************************************************************
 *
 * MPEG_CF_AddSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     clipStart -
 *     clipEnd -
 *     logStart -
 *     logEnd -
 *     logStartFlag - indicates whetehr logStart value set or not.
 *     logEndFlag - indicates whether logEnd value set or not.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_CF_AddSegment(MPEG_CF *objPtr, double clipStart, double clipEnd, 
		       double logStart, double logEnd, 
		       int logStartFlag, int logEndFlag)
{
  Segment *newSeg;
  char retStr[32];
  double clip_dur;
  Segment *nextSeg, *prevSeg;
  tChunkHeader chunk_hdr;
  int beg_frame, end_frame;
  double this_time, next_time, beg_time, end_time;
  int count, i, maxSize;
  double scale;
  RefEntry refEntry, tmpRefEntry;
  RefEntry *refEntryPtr;
  int ref_table_start;
  int orig_beg_frame, orig_end_frame;
  int *indexTable;
  int front_counter, end_counter;
  int tmpIndex, tmpOffset, tmpLength;

  MPEG_CF_Unready(objPtr);

  if (clipStart > clipEnd) {
    Tcl_AppendResult(objPtr->interp, "Clip start can not be after clip end.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  if (objPtr->filename == NULL) {
    Tcl_AppendResult(objPtr->interp, "Must specify clipfile first.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  if (clipStart < 0.0) clipStart = 0.0;
  if (clipStart > objPtr->clipLength) {
    Tcl_AppendResult(objPtr->interp, "Clip start too large.", (char *) NULL);
    return TCL_ERROR;
  }
  if (clipEnd > objPtr->clipLength) clipEnd = objPtr->clipLength;
  if (clipEnd < 0.0) {
    Tcl_AppendResult(objPtr->interp, "Clip end too small.", (char *) NULL);
    return TCL_ERROR;
  }
  
  newSeg = (Segment *) ckalloc(sizeof(Segment));
  newSeg->FOT = NULL;
  newSeg->FTT = NULL;
  newSeg->indexTable = NULL;
  newSeg->refTable = NULL;

  newSeg->clipStart = clipStart;
  newSeg->clipEnd = clipEnd;
  
  clip_dur = clipEnd - clipStart;
  
  if (clip_dur <= 0.0) {
    Tcl_AppendResult(objPtr->interp, "Clip segment duration too small.",
		     (char *) NULL);
    goto error;
  }
  
  
  /*
   * Do both start and end specified.
   */
  if (logStartFlag && logEndFlag) {
    /*
     * Set logStart and logEnd.
     */
    if (logStart > logEnd) {
      newSeg->logicalStart = logEnd;
      newSeg->logicalEnd = logStart;
    } else {
      newSeg->logicalStart = logStart;
      newSeg->logicalEnd = logEnd;
    }
  } else if (logStartFlag) {
    newSeg->logicalStart = logStart;
    logEnd = newSeg->logicalEnd = logStart+clip_dur;
  } else if (logEndFlag) {
    newSeg->logicalEnd = logEnd;
    logStart = newSeg->logicalStart = logEnd - clip_dur;
  } else {
    logStart = newSeg->logicalStart = 0.0;
    logEnd = newSeg->logicalEnd = clip_dur;
  }
  
  
  /*
   * Check for zero durations.
   */
  if (newSeg->logicalStart - newSeg->logicalEnd == 0.0) {
    Tcl_AppendResult(objPtr->interp, "Zero length logical time.",
		     (char *) NULL);
    goto error;
  }
  
  if (newSeg->clipStart - newSeg->clipEnd == 0.0) {
    Tcl_AppendResult(objPtr->interp, "Zero length clip time.",
		     (char *) NULL);
    goto error;
  }

  /* 
   * Find position in list by start times.
   */

  nextSeg = objPtr->segments;
  prevSeg = NULL;
  
  while (nextSeg != NULL) {
    if (nextSeg->logicalStart < newSeg->logicalStart) {
      prevSeg = nextSeg;
      nextSeg = nextSeg->next;
    } else {
      break;
    }
  }
  
  /* 
   * Create reference, offset, and time tables for this segment.
   * This means skipping frames until the correct beginning and
   * then finding correct end. There may be reference frames preceding
   * the trueStart that are needed as well as reference frames after
   * the true end. 
   * Generate segment FOT and FTT for these frames with time scaled properly. 
   * 
   * If rawFOT is not NULL, use rawFOT, rawFTT, and rawRefTable instead of
   * opening chunk files (i.e. data is in raw format and not clipfile format).
   */
  
  if (objPtr->rawFOT == NULL) {
    if (objPtr->cfState != NULL) {
      Chunk_ReadRewind(objPtr->cfState);
    } else {
      objPtr->cfState = Chunk_OpenFile(objPtr->filename, "r");
    }

    /* 
     * First find frame index number of starting and ending frames.
     */

    if (Chunk_FindType(objPtr->cfState, CM_FTT, &chunk_hdr)) {
      Tcl_AppendResult(objPtr->interp,
		       "Problem reading clipfile frame time table.", (char *) NULL);
      goto error;
    }
    
    beg_frame = 0;
    this_time = GetTimeFromFile(objPtr->cfState->fd);
    next_time = GetTimeFromFile(objPtr->cfState->fd);
    count = 0;
    while ((this_time < newSeg->clipStart) &&
	   (next_time <= newSeg->clipStart)) {
      this_time = next_time;
      next_time = GetTimeFromFile(objPtr->cfState->fd);
      count++;
      if (count == objPtr->numFrames) {
	Tcl_AppendResult(objPtr->interp,
			 "Error constructing segment frame time table.", (char *) NULL);
	goto error;
      }
    }
    
    beg_frame = count;
    
    while ((this_time < newSeg->clipEnd) && (next_time < newSeg->clipEnd)) {
      this_time = next_time;
      next_time = GetTimeFromFile(objPtr->cfState->fd);
      count++;
      if (count == objPtr->numFrames) {
	Tcl_AppendResult(objPtr->interp,
			 "Error constructing segment frame time table.", (char *) NULL);
	goto error;
      }
    }
    
    end_frame = count;
    newSeg->numFrames = end_frame - beg_frame + 1;

    /*
     * Now find REAL start and end (i.e. must start with I frame and
     * must end on P or I frame).
     */

    Chunk_ReadRewind(objPtr->cfState);
    if (Chunk_FindType(objPtr->cfState, CM_MPEG_REF_TABLE, &chunk_hdr)) {
      Tcl_AppendResult(objPtr->interp,
		       "Error constructing segment reference table.", (char *) NULL);
      goto error;
    }
    ref_table_start = lseek(objPtr->cfState->fd, 0L, SEEK_CUR);

    lseek(objPtr->cfState->fd, beg_frame * MPEG_REF_ENTRY_SIZE, SEEK_CUR);

    GetRefEntryFromFile(objPtr->cfState->fd, &refEntry);
    
    orig_beg_frame = beg_frame;

    while (refEntry.type != MPEG_I_FRAME) {
	beg_frame--;
	if (beg_frame < 0) {
	    break;
	}

	lseek(objPtr->cfState->fd, 
	      ref_table_start + (beg_frame * MPEG_REF_ENTRY_SIZE), SEEK_SET);

	GetRefEntryFromFile(objPtr->cfState->fd, &refEntry);
    }

    if (beg_frame < 0) {
	beg_frame = orig_beg_frame;

	lseek(objPtr->cfState->fd, 
	      ref_table_start + (beg_frame * MPEG_REF_ENTRY_SIZE), SEEK_SET);

	GetRefEntryFromFile(objPtr->cfState->fd, &refEntry);
    
	while (refEntry.type != MPEG_I_FRAME) {
	    beg_frame++;
	    if (beg_frame >= objPtr->numFrames) {
		Tcl_AppendResult(objPtr->interp, 
				 "No I Frame found in file. Very bizarre.",
				 (char *) NULL);
		goto error;
	    }

	    lseek(objPtr->cfState->fd,
		  ref_table_start + (beg_frame * MPEG_REF_ENTRY_SIZE), SEEK_SET);
	    
	    GetRefEntryFromFile(objPtr->cfState->fd, &refEntry);

	}
    }
    
    lseek(objPtr->cfState->fd, 
	  ref_table_start + (end_frame * MPEG_REF_ENTRY_SIZE), SEEK_SET);

    GetRefEntryFromFile(objPtr->cfState->fd, &refEntry);

    orig_end_frame = end_frame;

    while (refEntry.type == MPEG_B_FRAME) {
	end_frame++;
	if (end_frame >= objPtr->numFrames) {
	    break;
	}

	lseek(objPtr->cfState->fd, 
	      ref_table_start + (end_frame * MPEG_REF_ENTRY_SIZE), SEEK_SET);
	
	GetRefEntryFromFile(objPtr->cfState->fd, &refEntry);
    }

    if (end_frame >= objPtr->numFrames) {
	end_frame = orig_end_frame;

	lseek(objPtr->cfState->fd, 
	      ref_table_start + (end_frame * MPEG_REF_ENTRY_SIZE), SEEK_SET);
	
	GetRefEntryFromFile(objPtr->cfState->fd, &refEntry);

	while (refEntry.type == MPEG_B_FRAME) {
	    end_frame--;
	    if (end_frame < 0) {
		Tcl_AppendResult(objPtr->interp, 
				 "No I or P Frame found in file. Very bizarre.",
				 (char *) NULL);
		goto error;
	    }

	    lseek(objPtr->cfState->fd, 
		  ref_table_start + (end_frame * MPEG_REF_ENTRY_SIZE), SEEK_SET);
	    
	    GetRefEntryFromFile(objPtr->cfState->fd, &refEntry);
	}
    }

    if (beg_frame > end_frame) {
	Tcl_AppendResult(objPtr->interp, "Ended up with empty clip.", (char *) NULL);
	goto error;
    }

    newSeg->numFrames = end_frame - beg_frame + 1;


    /* 
     * Now construct reference table with type and reference info. 
     * Here is where we find any necessary reference frame that actually
     * precede the true start of the segement stored in beg_frame.
     * Be aware that mpeg's may start invalidly (e.g., with a B frame that
     * doesn't have a preceding I or P frame) and deal with that accordingly.
     */

    indexTable = (int *) ckalloc(sizeof(int)*newSeg->numFrames);
    newSeg->indexTable = indexTable;

    newSeg->refTable = (RefEntry *) ckalloc(sizeof(RefEntry)*newSeg->numFrames);

    lseek(objPtr->cfState->fd,
	  ref_table_start + (beg_frame * MPEG_REF_ENTRY_SIZE), SEEK_SET);

    for (i=0; i<newSeg->numFrames; i++) {
	indexTable[i] = beg_frame+i;
	GetRefEntryFromFile(objPtr->cfState->fd, 
			    &(newSeg->refTable[i]));
	newSeg->refTable[i].marker = 0;
    }

    /*
     * Read current FTT values.
     */

    newSeg->FTT = (double *) ckalloc((newSeg->numFrames+1)*sizeof(double));
    
    Chunk_ReadRewind(objPtr->cfState);
    if (Chunk_FindType(objPtr->cfState, CM_FTT, &chunk_hdr)) {
      Tcl_AppendResult(objPtr->interp,
		       "Error constructing segment frame time table.", (char *) NULL);
      goto error;
    }

    lseek(objPtr->cfState->fd, beg_frame * 8, SEEK_CUR);
    
    for (i=0; i <= newSeg->numFrames; i++) {
	newSeg->FTT[i] = GetTimeFromFile(objPtr->cfState->fd);
    }

    newSeg->clipStart = newSeg->FTT[0];
    newSeg->clipEnd = newSeg->FTT[newSeg->numFrames];

    scale = (newSeg->logicalEnd - newSeg->logicalStart) / 
      (newSeg->clipEnd - newSeg->clipStart);
    
    newSeg->FTT[0] = newSeg->logicalStart;
    
    for (i=1; i<newSeg->numFrames; i++) {
      newSeg->FTT[i] = (scale * (newSeg->FTT[i] - newSeg->clipStart)) +
	newSeg->logicalStart;
    }
    newSeg->FTT[i] = newSeg->logicalEnd;
    
    /*
     * Construct segment frame offset table.
     */
    Chunk_ReadRewind(objPtr->cfState);
    if (Chunk_FindType(objPtr->cfState, CM_FOT, &chunk_hdr)) {
      Tcl_AppendResult(objPtr->interp,
		       "Error constructing segment frame offset table.", (char *)NULL);
      goto error;
    }
    
    newSeg->FOT = (int *) ckalloc(sizeof(int)*2*newSeg->numFrames);

    lseek(objPtr->cfState->fd, beg_frame * 8, SEEK_CUR);

    for (i=0; i<newSeg->numFrames; i++) {
	GetOffsetFromFile(objPtr->cfState->fd, &(newSeg->FOT[i*2]),
			  &(newSeg->FOT[(i*2)+1]));
    }
    
  } else {
      int ftt_index;
      
      /* 
       * Construct segment FOT and FTT from rawFOT and rawFTT. 
       * First find out which frames are needed.
       */
      
      beg_frame = 0;
      ftt_index = 0;
      
      this_time = objPtr->rawFTT[ftt_index++];
      next_time = objPtr->rawFTT[ftt_index++];
      
      count = 0;
      
      while ((this_time < newSeg->clipStart) &&
	     (next_time <= newSeg->clipStart)) {
	  this_time = next_time;
	  next_time = objPtr->rawFTT[ftt_index++];
	  count++;
	  if (count == objPtr->numFrames) {
	      Tcl_AppendResult(objPtr->interp,
			       "Error constructing segment frame time table.", 
			       (char *) NULL);
	      goto error;
	  }
      }
      
      beg_frame = count;
      
      while ((this_time < newSeg->clipEnd) && (next_time < newSeg->clipEnd)) {
	  this_time = next_time;
	  next_time = objPtr->rawFTT[ftt_index++];
	  count++;
	  if (count == objPtr->numFrames) {
	      Tcl_AppendResult(objPtr->interp,
			       "Error constructing segment frame time table.", 
			       (char *) NULL);
	      goto error;
	  }
      }
      
      end_frame = count;
      newSeg->numFrames = end_frame - beg_frame + 1;
      
      /*
       * Find real beginning and end.
       */
      
      orig_beg_frame = beg_frame;
      
      refEntryPtr = &(objPtr->rawRefTable[beg_frame]);
      
      while (refEntryPtr->type != MPEG_I_FRAME) {
	  beg_frame--;
	  if (beg_frame < 0) break;
	  
	  refEntryPtr = &(objPtr->rawRefTable[beg_frame]);
      }
      
      if (beg_frame < 0) {
	  beg_frame = orig_beg_frame;
	  
	  refEntryPtr = &(objPtr->rawRefTable[beg_frame]);
	  
	  while (refEntryPtr->type != MPEG_I_FRAME) {
	      beg_frame++;
	      if (beg_frame >= objPtr->numFrames) {
		  Tcl_AppendResult(objPtr->interp, 
				   "No I Frame found in file. Very bizarre.",
				   (char *) NULL);
		  goto error;
	      }
	      
	      refEntryPtr = &(objPtr->rawRefTable[beg_frame]);
	}
    }

    refEntryPtr = &(objPtr->rawRefTable[end_frame]);

    orig_end_frame = end_frame;
      
    while (refEntryPtr->type == MPEG_B_FRAME) {
	end_frame++;
	if (end_frame >= objPtr->numFrames) break;
	refEntryPtr = &(objPtr->rawRefTable[end_frame]);
    }

    if (end_frame >= objPtr->numFrames) {
	end_frame = orig_end_frame;
	refEntryPtr = &(objPtr->rawRefTable[end_frame]);
	while (refEntryPtr->type == MPEG_B_FRAME) {
	    end_frame--;
	    if (end_frame < 0) {
		Tcl_AppendResult(objPtr->interp, 
				 "No I or P Frame found in file. Very bizarre.",
				 (char *) NULL);
		goto error;
	    }
	    refEntryPtr = &(objPtr->rawRefTable[beg_frame]);
	}
    }
      
      
      if (beg_frame > end_frame) {
	  Tcl_AppendResult(objPtr->interp, "Ended up with empty clip.", (char *) NULL);
	  goto error;
      }
    
      newSeg->numFrames = end_frame - beg_frame + 1;
      
      
      /* 
       * Now construct reference table with type and reference info. 
       * Here is where we find any necessary reference frame that actually
       * precede the true start of the segement stored in beg_frame.
       * Be aware that mpeg's may start invalidly (e.g., with a B frame that
       * doesn't have a preceding I or P frame) and deal with that accordingly.
       * Here we use rawRefTable.
       */
      
      indexTable = (int *) ckalloc(sizeof(int)*newSeg->numFrames);
      newSeg->indexTable = indexTable;
      
      newSeg->refTable = (RefEntry *) ckalloc(sizeof(RefEntry)*newSeg->numFrames);
      
      for (i=0; i<newSeg->numFrames; i++) {
	  indexTable[i] = beg_frame+i;
	  CopyRefEntry(&(newSeg->refTable[i]),
		       &(objPtr->rawRefTable[beg_frame+i]));
	  newSeg->refTable[i].marker = 0;
      }
      
      /* 
       * Construct FTT for new segment from rawFTT. 
       */
      
      newSeg->FTT = (double *) ckalloc((newSeg->numFrames+1)*sizeof(double));
      
      for (i=0; i<=newSeg->numFrames; i++) {
	  newSeg->FTT[i] = objPtr->rawFTT[beg_frame+i];
      }
      
      newSeg->clipStart = newSeg->FTT[0];
      newSeg->clipEnd = newSeg->FTT[newSeg->numFrames];
      
      
      scale = (newSeg->logicalEnd - newSeg->logicalStart) / 
	(newSeg->clipEnd - newSeg->clipStart);
      
      newSeg->FTT[0] = newSeg->logicalStart;
      
      for (i=1; i<newSeg->numFrames; i++) {
	  newSeg->FTT[i] = (scale * (newSeg->FTT[i] - newSeg->clipStart)) +
	    newSeg->logicalStart;
      }
      newSeg->FTT[i] = newSeg->logicalEnd;
      
      /* 
       * Construct seg. FOT.
       */
      
      newSeg->FOT = (int *) ckalloc(sizeof(int)*2*newSeg->numFrames);
      
      for (i=0; i<newSeg->numFrames; i++) {
	  newSeg->FOT[i*2] = objPtr->rawFOT[(beg_frame+i)*2];
	  newSeg->FOT[(i*2)+1] = objPtr->rawFOT[((beg_frame+i)*2)+1];
      }
  }
  
  
  /*
   * Resolve references.
   */
  
  for (i=0; i<newSeg->numFrames; i++) {
      if (newSeg->refTable[i].ref1 < beg_frame) {
	  newSeg->refTable[i].ref1 = -1;
      } else if (newSeg->refTable[i].ref1 > end_frame) {
	  newSeg->refTable[i].ref1 = -1;
      } else {
	  newSeg->refTable[i].ref1 -= beg_frame;
      }

      if (newSeg->refTable[i].ref2 < beg_frame) {
	  newSeg->refTable[i].ref2 = -1;
      } else if (newSeg->refTable[i].ref2 > end_frame) {
	  newSeg->refTable[i].ref2 = -1;
      } else {
	  newSeg->refTable[i].ref2 -= beg_frame;
      }
  }

  /*
   * Now deal with logStart > logEnd (i.e., clip being played backwards).
   * Flip offset table, refTable, and indexTable.
   * Mathematically flip time table.
   */

  if (logStart > logEnd) {
      front_counter = 0;
      end_counter = newSeg->numFrames-1;
      
      while (front_counter < end_counter) {
	  tmpIndex = indexTable[front_counter];
	  indexTable[front_counter] = indexTable[end_counter];
	  indexTable[end_counter] = tmpIndex;
	  
	  CopyRefEntry(&tmpRefEntry, &(newSeg->refTable[front_counter]));
	  CopyRefEntry(&(newSeg->refTable[front_counter]), 
		       &(newSeg->refTable[end_counter]));
	  CopyRefEntry( &(newSeg->refTable[end_counter]), &tmpRefEntry);
	  
	  tmpOffset = newSeg->FOT[front_counter*2];
	  tmpLength = newSeg->FOT[(front_counter*2)+1];
	  newSeg->FOT[front_counter*2] = newSeg->FOT[end_counter*2];
	  newSeg->FOT[(front_counter*2)+1] = newSeg->FOT[(end_counter*2)+1];
	  newSeg->FOT[end_counter*2] = tmpOffset;
	  newSeg->FOT[(end_counter*2)+1] = tmpLength;
	  
	  front_counter++;
	  end_counter--;
      }

      beg_time = newSeg->FTT[0];
      end_time = newSeg->FTT[newSeg->numFrames];
      
      for (i=0; i<=newSeg->numFrames; i++) {
	  newSeg->FTT[i] = (((newSeg->FTT[i] - beg_time) * (-1.0)) + end_time);
      }

      /*
       * Reset references.
       */

      for (i=0; i<newSeg->numFrames; i++) {
	  if (newSeg->refTable[i].ref1 != -1) {
	      newSeg->refTable[i].ref1 = newSeg->numFrames - 1 - newSeg->refTable[i].ref1;
	  }
	  if (newSeg->refTable[i].ref2 != -1) {
	      newSeg->refTable[i].ref2 = newSeg->numFrames - 1 - newSeg->refTable[i].ref2;
	  }
      }
  }

  /*
   * Tag the frames with index numbers.
   */

  for (i=0; i<newSeg->numFrames; i++) {
    newSeg->refTable[i].indexNum = (int) (((newSeg->FTT[i] + newSeg->FTT[i+1])/2.0) 
					  * 1000.0);
  }

  /*
   * Set segment id.
   */
  newSeg->segId = objPtr->nextSegId++;
  
  /*
   * Find max frame size.
   */
  maxSize = -1;
  for (i=0; i<newSeg->numFrames; i++) {
      if (newSeg->FOT[(i*2)+1] > maxSize) {
	  maxSize = newSeg->FOT[(i*2)+1];
      }
  }
  newSeg->maxSize = maxSize;
  
  /*
   * Set up next and prev pointers and return.
   */
  newSeg->next = nextSeg;
  newSeg->prev = prevSeg;
  if (prevSeg == NULL) {
      objPtr->segments = newSeg;
  } else {
      prevSeg->next = newSeg;
  }
  if (nextSeg != NULL) {
      nextSeg->prev = newSeg;
  }
  
  sprintf(retStr, "%d ", newSeg->segId);
  Tcl_SetResult(objPtr->interp, retStr, TCL_VOLATILE);

  objPtr->segmentListChangeFlag = 1;

  return TCL_OK;

 error:
  if (newSeg != NULL) {
      if (newSeg->FOT != NULL) {
	  ckfree((char *) newSeg->FOT);
      }
      if (newSeg->FTT != NULL) {
	  ckfree((char *) newSeg->FTT);
      }
      if (newSeg->refTable != NULL) {
	  ckfree((char *) newSeg->refTable);
      }
      if (newSeg->indexTable != NULL) {
	  ckfree((char *) newSeg->indexTable);
      }
      ckfree((char *) newSeg);
  }
  return TCL_ERROR;
}

/************************************************************
 *
 * MPEG_CF_RemoveSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segId -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_CF_RemoveSegment(MPEG_CF * objPtr, int segId)
{
  Segment *segPtr;
  
  objPtr->lastSeg = NULL;
  objPtr->lastFrameNum = -1;
  
  segPtr = objPtr->segments;
  
  while (segPtr != NULL) {
      if (segPtr->segId == segId) break;
      segPtr = segPtr->next;
  }
  
  if (segPtr == NULL) return TCL_OK;
  
  if (segPtr->prev == NULL) {
      objPtr->segments = segPtr->next;
  } else {
      segPtr->prev->next = segPtr->next;
  }
  
  if (segPtr->next != NULL) {
      segPtr->next->prev = segPtr->prev;
  }
  
  if (segPtr->FOT != NULL)  {
      ckfree((char *) segPtr->FOT);
  }
  if (segPtr->FTT != NULL) {
      ckfree((char *) segPtr->FTT);
  }
  if (segPtr->refTable != NULL) {
      ckfree((char *) segPtr->refTable);
  }
  if (segPtr->indexTable != NULL) {
      ckfree((char *) segPtr->indexTable);
  }

  ckfree((char *) segPtr);
  
  return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_ClearSegments
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_CF_ClearSegments(MPEG_CF * objPtr)
{
    while (objPtr->segments != NULL) {
	MPEG_CF_RemoveSegment(objPtr, objPtr->segments->segId);
    }
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_SegmentInfo
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segId -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sets the Tcl result string to {clipStart clipEnd logStart logEnd 
 *                                    numFrames maxSize} 
 *
 ************************************************************/

static int
MPEG_CF_SegmentInfo(MPEG_CF * objPtr, int segId)
{
    Segment *segPtr;
    char retStr[256];

    segPtr = objPtr->segments;

    while (segPtr != NULL) {
	if (segPtr->segId == segId) {
	    break;
	}
	segPtr = segPtr->next;
    }

    if (segPtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "Segment doesn't exist.",
		(char *) NULL);
	return TCL_ERROR;
    }

    sprintf(retStr, "%g %g %g %g %d %d", segPtr->clipStart, segPtr->clipEnd,
	    segPtr->logicalStart, segPtr->logicalEnd,
	    segPtr->numFrames, segPtr->maxSize);

    Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_Info
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sets return string to {numFrames clipLength maxSize width height}
 *     where maxSize is maximum size of segments so far.
 *
 ************************************************************/
static int
MPEG_CF_Info(MPEG_CF *objPtr)
{
    Segment *segPtr;
    int maxSize;
    char retStr[256];

    segPtr = objPtr->segments;
    maxSize = 0;

    while(segPtr != NULL) {
	if (segPtr->maxSize > maxSize)
	    maxSize = segPtr->maxSize;
	segPtr = segPtr->next;
    }

    if (objPtr->cfInfo != NULL) {
	sprintf(retStr, "%d %g %d %d %d", objPtr->numFrames,
		objPtr->clipLength, maxSize,
		objPtr->cfInfo->width, objPtr->cfInfo->height);
    } else {
	sprintf(retStr, "%d %g %d 0 0", objPtr->numFrames,
		  objPtr->clipLength, maxSize);
    }

    Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_Ready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Gets the object ready and starts sending data.
 ************************************************************/

static int
MPEG_CF_Ready(MPEG_CF *objPtr)
{

    /*
     * Blow away the old stuff.
     */
    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
    }
    objPtr->at_token = 0;

    /*
     * Check for overlaps if necessary. 
     */

    if (objPtr->segmentListChangeFlag) {
      Segment *seg_ptr;

      seg_ptr = objPtr->segments;

      while (seg_ptr != NULL) {
	if (seg_ptr->next != NULL) {
	  if (seg_ptr->next->logicalStart < seg_ptr->logicalEnd) {
	    
	    /* Overlap detected. Definitely not ready. Return error. */

	    objPtr->ready = 0;
	    Tcl_AppendResult(objPtr->interp, "Segment overlap detected.", (char *) NULL);
	    return TCL_ERROR;
	  }
	}
	seg_ptr = seg_ptr->next;
      }

      objPtr->segmentListChangeFlag = 0;
    }

    /*
     * Set the ready flag.
     */
    objPtr->ready = 1;

    objPtr->lastMarker++;

    if (objPtr->lts != NULL) {
	MPEG_CF_SendData((ClientData) objPtr);
    }
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_Unready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Stops sending data and puts the object in the unready state.
 ************************************************************/

static int
MPEG_CF_Unready(MPEG_CF * objPtr)
{
    /*
     * Blow away the old stuff.
     */
    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
    }
    objPtr->at_token = 0;
    
    objPtr->lastSeg = NULL;
    objPtr->lastFrameNum = -1;
    objPtr->ready = 0;
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_LTSChanged
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Procedure called whenever the LTS changes.
 ************************************************************/

void
MPEG_CF_LTSChanged(ClientData clientdata)
{
    MPEG_CF *objPtr = (MPEG_CF *) clientdata;

    objPtr->extraFrameCounter = 0.0;
    objPtr->extraBWCounter = 0;

    if (objPtr->ready == 1) {
	MPEG_CF_Ready(objPtr);
    }
    return;
}

/************************************************************
 *
 * NextStartCode
 *
 * Parameters:
 *     fd - open file descriptor
 *
 * Return:
 *     Next start code found or -1 if EOF
 *
 * Action:
 *     Reads bytes from open file descriptor looking for next start
 *     code. Once found, it returns the start code and lseeks
 *     the file descriptor to be set to the first byte of the start
 *     code. 
 *     NOTE: only looks for and finds picture, seq. end, and seq. start
 *           codes. It ignores all others.
 ************************************************************/

static int
NextStartCode(int fd) 
{
    unsigned char readBuffer[512];
    int ret_value, cur_pos, idx, num_read;
    int state;

    cur_pos = lseek(fd, 0L, SEEK_CUR);
    num_read = read(fd, readBuffer, 512);

    if (num_read <= 0) return -1;

    idx = 0;
    state = 0;
    ret_value = -1;

    while (1) {

	if (state == 0) {
	    if (readBuffer[idx] == 0) {
		state = 1;
	    }
	    idx++;
	} else if (state == 1) {
	    if (readBuffer[idx] == 0) {
		state = 2;
	    } else {
		state = 0;
	    }
	    idx++;
	} else if (state == 2) {
	    if (readBuffer[idx] == 1) {
		state = 3;
	    } else if (readBuffer[idx] != 0) {
		state = 0;
	    }
	    idx++;
	} else if (state == 3) {
	    if (readBuffer[idx] == 0) {
		ret_value = MPEG_PIC_START_CODE;
		break;
	    } else if (readBuffer[idx] == 179) {
		ret_value = MPEG_SYS_HDR_START_CODE;
		break;
	    } else if (readBuffer[idx] == 183) {
		ret_value = MPEG_END_SEQ_CODE;
		break;
	    } else {
		state = 0;
	    }
	    idx++;
	}
   
	if (idx >= num_read) {
	    cur_pos = lseek(fd, 0L, SEEK_CUR);
	    num_read = read(fd, readBuffer, 512);
	    idx = 0;

	    if (num_read <= 0) return -1;
	}
    }

    idx -= 3;
    lseek(fd, cur_pos+idx, SEEK_SET);
    return ret_value;
}
    

/************************************************************
 *
 * FirstStartCode
 *
 * Parameters:
 *     fd - open file descriptor
 *
 * Return:
 *     First start code found or -1 if not found in first 1024 bytes
 *
 * Action:
 *     Reads bytes from open file descriptor looking for next start
 *     code. Once found, it returns the start code and lseeks
 *     the file descriptor to be set to the first byte of the start
 *     code. 
 *     NOTE: only looks for and finds seq. start
 *           codes. It ignores all others. It only looks in the
 *           first 1024 bytes. This is a heuristic for determining
 *           whether the file is an mpeg at all.
 ************************************************************/

static int
FirstStartCode(int fd) 
{
    unsigned char readBuffer[1024];
    int ret_value, cur_pos, idx, num_read;
    int state;

    cur_pos = lseek(fd, 0L, SEEK_CUR);
    num_read = read(fd, readBuffer, 1024);

    if (num_read <= 0) return -1;

    idx = 0;
    state = 0;
    ret_value = -1;

    while (1) {

	if (state == 0) {
	    if (readBuffer[idx] == 0) {
		state = 1;
	    }
	    idx++;
	} else if (state == 1) {
	    if (readBuffer[idx] == 0) {
		state = 2;
	    } else {
		state = 0;
	    }
	    idx++;
	} else if (state == 2) {
	    if (readBuffer[idx] == 1) {
		state = 3;
	    } else if (readBuffer[idx] != 0) {
		state = 0;
	    }
	    idx++;
	} else if (state == 3) {
	    if (readBuffer[idx] == 179) {
		ret_value = MPEG_SYS_HDR_START_CODE;
		break;
	    } else {
		state = 0;
	    }
	    idx++;
	}
   
	if (idx >= num_read) {
	  return -1;
	}
    }

    idx -= 3;
    lseek(fd, cur_pos+idx, SEEK_SET);
    return ret_value;
}
    

/************************************************************
 *
 * MPEG_CF_OpenRawfile
 *
 * Parameters:
 *     objPtr - pointer to object
 *     filename - filename
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Opens the file as a raw data file, constructs  basic info data.
 *     Closes file when done.
 ************************************************************/

static int
MPEG_CF_OpenRawfile(MPEG_CF *objPtr,         /* Pointer to mpeg clipfile object. */
		     char *filename            /* File to be opened as raw data. */)
{
    int hdrRef;
    int next_start_code;
    unsigned char readBuffer[200];
    int i, next_value;
    double sec_per_frame;
    int frameNum, ref1, ref2, lastI, nextI;
    int numFrameAlloced;
    int offset, picture_type, end_offset;
    int length;
    RefEntry tmpRefEntry;
    int tmpLength, tmpOffset;

    objPtr->rawfd = open(filename, O_RDONLY|O_BINARY);
    
    if (objPtr->rawfd < 0) {
	objPtr->rawfd = -1;
	return TCL_ERROR;
    }
    
    /* 
     * Basic algorithm is this:
     *
     * set hdrRef = -1
     * Find first system header.
     *   Parse header and set cfInfo.
     *   If intra or non_intra matrices given
     *     Parse tables. Append to headerTable (should be empty).
     *     Increment hdrRef.
     * Info in first header will give you sec/frame for calculating FTT values.
     * frameNum = 0, ref1 = -1, ref2 = -1
     * While next start code is not end of sequence
     *   If start code is system header
     *     If intra or non_intr matrices given.
     *     Parse tables. Append to headerTable, increment hdrRef.
     *   Else if picture start code.
     *     Parse type.
     *     Append rawFTT, rawFOT, rawRefTable
     *     Update current ref1 and ref2 based on type.
     *     Incr frameNum;
     *
     * Set numFrames, clipLength, dataStart, other misc. fields of objPtr.
     * Done.
     */
    
    hdrRef = -1;
    
    next_start_code = FirstStartCode(objPtr->rawfd);
    
    
    while ((next_start_code != MPEG_SYS_HDR_START_CODE) &&
	   (next_start_code != MPEG_END_SEQ_CODE) &&
	   (next_start_code != -1)) {
	lseek(objPtr->rawfd, 4, SEEK_CUR);
	next_start_code = NextStartCode(objPtr->rawfd);
    }
    
    if ((next_start_code == -1) || (next_start_code == MPEG_END_SEQ_CODE)) {
	/*
	 * Reached end of file without finding sys start code.
	 */	
	goto error;
    }

    /*
     * Read in system header, set up cfInfo.
     */

    read(objPtr->rawfd, readBuffer, 140);

    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);
	objPtr->cfInfo = NULL;
    }
    objPtr->cfInfo = (MPEGHeader *) ckalloc(sizeof(MPEGHeader));
    
    objPtr->cfInfo->width = (readBuffer[4]*16 + (readBuffer[5] >> 4));
    objPtr->cfInfo->height = (((readBuffer[5] & 0x0f) * 256) + readBuffer[6]);
    objPtr->cfInfo->aspectRatioCode = (readBuffer[7] >> 4);
    objPtr->cfInfo->frameRateCode = (readBuffer[7] & 0x0f);
    objPtr->cfInfo->bitRate = (((readBuffer[8] * 256*4) + (readBuffer[9] * 4) +
			       (readBuffer[10] >> 6)));
    objPtr->cfInfo->vbvBufferSize = ((((readBuffer[10] & 0x1f)*32) +
				     (readBuffer[11] >> 3)));
    if ((readBuffer[11] & 0x04)) 
      objPtr->cfInfo->constrainedParamFlag = 1;

    if (objPtr->headerTable != NULL) ckfree((char *) objPtr->headerTable);
    objPtr->numHeaders = 0;

    if ((readBuffer[11] & 0x02)) {
	hdrRef++;
	
	objPtr->numHeaders = 1;
	objPtr->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*1);
    
	objPtr->headerTable[0].intra_matrix_flag = 1;
	
	for(i=0; i<64; i++) {
	    next_value = ((readBuffer[11+i] & 0x01 * 128) + 
			  (readBuffer[12+i] >> 1));
	    objPtr->headerTable[0].intra_matrix[i] = next_value;
	}

	if ((readBuffer[75] & 0x01)) {
	    objPtr->headerTable[0].non_intra_matrix_flag = 1;
	    for (i=0; i<64; i++) {
		objPtr->headerTable[0].non_intra_matrix[i] = readBuffer[76+i];
	    }
	} else {
	    objPtr->headerTable[0].non_intra_matrix_flag = 0;
	    lseek(objPtr->rawfd, -1*(140-76), SEEK_CUR);
	}
    } else if ((readBuffer[11] & 0x01)) {
	hdrRef++;
	
	objPtr->numHeaders = 1;
	objPtr->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*1);
    
	objPtr->headerTable[0].intra_matrix_flag = 0;
	objPtr->headerTable[0].non_intra_matrix_flag = 1;
	for (i=0; i<64; i++) {
	    objPtr->headerTable[0].non_intra_matrix[i] = readBuffer[12+i];
	} 

	lseek (objPtr->rawfd, -1*(140-76), SEEK_CUR);
    } else {
	objPtr->numHeaders = 0;
	objPtr->headerTable = NULL;
	lseek (objPtr->rawfd, -1*(140-12), SEEK_CUR);
    }


    switch (objPtr->cfInfo->frameRateCode) {
      case 1:
	sec_per_frame = 1.0 / 23.976;
	break;
      case 2:
	sec_per_frame = 1.0 / 24.0;
	break;
      case 3:
	sec_per_frame = 1.0 / 25.0;
	break;
      case 4:
	sec_per_frame = 1.0 / 29.97;
	break;
      case 5:
	sec_per_frame = 1.0 / 30.0;
	break;
      case 6:
	sec_per_frame = 1.0 / 50.0;
	break;
      case 7:
	sec_per_frame = 1.0 / 59.94;
	break;
      case 8:
	sec_per_frame = 1.0 / 60.0;
	break;
      default:
	sec_per_frame = 1.0 / 30.0;
	break;
    }

    frameNum = 0; 
    ref1 = -1;
    ref2 = -1;
    numFrameAlloced = 100;

    if (objPtr->rawFOT != NULL) {
	ckfree((char *) objPtr->rawFOT);
    }
    if (objPtr->rawFTT != NULL) {
	ckfree((char *) objPtr->rawFTT);
    }
    if (objPtr->rawRefTable != NULL) {
	ckfree((char *) objPtr->rawRefTable);
    }	

    objPtr->rawFOT = (int *) ckalloc(sizeof(int) * 2 * numFrameAlloced);
    objPtr->rawFTT = (double *) ckalloc(sizeof(double) * (numFrameAlloced+1));
    objPtr->rawRefTable = (RefEntry *) ckalloc(sizeof(RefEntry) * numFrameAlloced);
    
    next_start_code = NextStartCode(objPtr->rawfd);

    while ((next_start_code != MPEG_END_SEQ_CODE) &&
	   (next_start_code != -1)) {
	
	if (next_start_code == MPEG_SYS_HDR_START_CODE) {
	    read(objPtr->rawfd, readBuffer, 12);

	    if ((readBuffer[11] & 0x02)) {
		hdrRef++;

		objPtr->numHeaders++;

		if (objPtr->headerTable == NULL) {
		    objPtr->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*
							       objPtr->numHeaders);
		} else {
		    objPtr->headerTable = (HdrEntry *) ckrealloc((char *) objPtr->headerTable,
							       sizeof(HdrEntry)*
							       objPtr->numHeaders);
		}

		read(objPtr->rawfd, &(readBuffer[12]), 64);
		
		objPtr->headerTable[hdrRef].intra_matrix_flag = 1;
	      
		for(i=0; i<64; i++) {
		    next_value = ((readBuffer[11+i] & 0x01 * 128) + 
				  (readBuffer[12+i] >> 1));
		    objPtr->headerTable[hdrRef].intra_matrix[i] = next_value;
		}

		if ((readBuffer[75] & 0x01)) {
		    objPtr->headerTable[hdrRef].non_intra_matrix_flag = 1;

		    read(objPtr->rawfd, &(readBuffer[76]), 64);

		    for (i=0; i<64; i++) {
			objPtr->headerTable[hdrRef].non_intra_matrix[i] = 
			  readBuffer[76+i];
		    }
		} else {
		    objPtr->headerTable[hdrRef].non_intra_matrix_flag = 0;
		}
	    } else if ((readBuffer[11] & 0x01)) {
		hdrRef++;

		objPtr->numHeaders++;

		if (objPtr->headerTable == NULL) {
		    objPtr->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*
							       objPtr->numHeaders);
		} else {
		    objPtr->headerTable = (HdrEntry *) ckrealloc((char *) objPtr->headerTable,
							       sizeof(HdrEntry)*
							       objPtr->numHeaders);
		}
	
		objPtr->headerTable[hdrRef].intra_matrix_flag = 0;
		objPtr->headerTable[hdrRef].non_intra_matrix_flag = 1;
		
		read(objPtr->rawfd, &(readBuffer[12]), 64);

		for (i=0; i<64; i++) {
		    objPtr->headerTable[hdrRef].non_intra_matrix[i] = readBuffer[12+i];
		} 
	    }

	    next_start_code = NextStartCode(objPtr->rawfd);

	} else if (next_start_code == MPEG_PIC_START_CODE) {
	    
	    offset = lseek(objPtr->rawfd, 0L, SEEK_CUR);

	    read(objPtr->rawfd, readBuffer, 6);

	    picture_type = (int)(readBuffer[5] & 0x38) >> 3;
	    
	    next_start_code = NextStartCode(objPtr->rawfd);
	    if (next_start_code == -1) {
		break;
	    }
	    end_offset = lseek(objPtr->rawfd, 0L, SEEK_CUR);

	    length = end_offset - offset;

	    if (frameNum >= numFrameAlloced) {
		numFrameAlloced += 100;

		objPtr->rawFOT = (int *) ckrealloc((char *) objPtr->rawFOT,
						 sizeof(int) * 2 * numFrameAlloced);
		objPtr->rawFTT = (double *) ckrealloc((char *) objPtr->rawFTT, 
						    sizeof(double) * 
						    (numFrameAlloced+1));
		objPtr->rawRefTable = (RefEntry *) ckrealloc((char *) objPtr->rawRefTable,
							   sizeof(RefEntry) * 
							   numFrameAlloced);
	    }
	    
	    objPtr->rawFOT[frameNum*2] = offset;
	    objPtr->rawFOT[(frameNum*2)+1] = length;
	    objPtr->rawFTT[frameNum] = ((double) frameNum) * sec_per_frame;
	    objPtr->rawRefTable[frameNum].type = picture_type;

	    if (picture_type != MPEG_B_FRAME) {

		/*
		 * We need to save the old reference 1 because we only want to do the
		 * flipping starting with the second and third reference frames and
		 * not the first and second. Thus, old_ref1 will be -1 for the two
		 * cases we want to avoid.
		 */

		ref1 = ref2;
		ref2 = frameNum;

		if (ref1 > -1) {
		    /*
		     * Here's the really tricky part.
		     * Now that we've got that next I or P after some
		     * sequence of B's, we need to adjust the index numbers
		     * to reflect display order instead of decode order.
		     */
		    
		    CopyRefEntry(&tmpRefEntry, &(objPtr->rawRefTable[ref1]));
		    tmpOffset = objPtr->rawFOT[ref1*2];
		    tmpLength = objPtr->rawFOT[(ref1*2)+1];
		    
		    for (i=ref1; i<ref2-1; i++) {
			CopyRefEntry(&(objPtr->rawRefTable[i]), 
				     &(objPtr->rawRefTable[i+1]));
			objPtr->rawFOT[i*2] = objPtr->rawFOT[(i+1)*2];
			objPtr->rawFOT[(i*2)+1] = objPtr->rawFOT[((i+1)*2)+1];
		    }
		    
		    CopyRefEntry(&(objPtr->rawRefTable[ref2-1]), &tmpRefEntry);
		    objPtr->rawFOT[(ref2-1)*2] = tmpOffset;
		    objPtr->rawFOT[((ref2-1)*2)+1] = tmpLength;
				    
		    for (i=ref1; i<ref2; i++) {
			objPtr->rawRefTable[i].ref2 = ref2-1;
		    }

		    ref1 = ref2-1;
		}
	    }

	    objPtr->rawRefTable[frameNum].ref1 = ref1;
	    objPtr->rawRefTable[frameNum].ref2 = ref2;
	    
	    objPtr->rawRefTable[frameNum].headerRef = hdrRef;

	    frameNum++;
	} else {
	    lseek(objPtr->rawfd, 4, SEEK_CUR);
	    next_start_code = NextStartCode(objPtr->rawfd);
	}
    }
    
    if (frameNum <= 0) goto error;

    /* 
     * Do the final flipping for the last set of B frames. 
     */

    ref1 = ref2;
    ref2 = frameNum;

    if (ref1 > -1) {
	CopyRefEntry(&tmpRefEntry, &(objPtr->rawRefTable[ref1]));
	tmpOffset = objPtr->rawFOT[ref1*2];
	tmpLength = objPtr->rawFOT[(ref1*2)+1];
		    
	for (i=ref1; i<ref2-1; i++) {
	    CopyRefEntry(&(objPtr->rawRefTable[i]), 
			 &(objPtr->rawRefTable[i+1]));
	    objPtr->rawFOT[i*2] = objPtr->rawFOT[(i+1)*2];
	    objPtr->rawFOT[(i*2)+1] = objPtr->rawFOT[((i+1)*2)+1];
	}
		    
	CopyRefEntry(&(objPtr->rawRefTable[ref2-1]), &tmpRefEntry);
	objPtr->rawFOT[(ref2-1)*2] = tmpOffset;
	objPtr->rawFOT[((ref2-1)*2)+1] = tmpLength;
				    
	for (i=ref1; i<ref2; i++) {
	    objPtr->rawRefTable[i].ref2 = ref2-1;
	}

	ref1 = ref2-1;
    }
    
    objPtr->rawRefTable = (RefEntry *) ckrealloc((char *) objPtr->rawRefTable,
					       sizeof(RefEntry) * frameNum);
    objPtr->rawFTT = (double *) ckrealloc((char *) objPtr->rawFTT,
					sizeof(double) * (frameNum+1));
    objPtr->rawFOT = (int *) ckrealloc((char *) objPtr->rawFOT,
				     sizeof(int) * 2 * frameNum);

    objPtr->rawFTT[frameNum] = ((double) frameNum) * sec_per_frame;

    /*
     * Resolve I and P references now. 
     */

    lastI = -1;
    for (i=0; i<frameNum; i++) {
	if (objPtr->rawRefTable[i].type == MPEG_I_FRAME) {
	    objPtr->rawRefTable[i].ref1 = lastI;
	    lastI = i;
	}
    }

    nextI = -1;

    for (i=frameNum-1; i>=0; i--) {
	if (objPtr->rawRefTable[i].type == MPEG_I_FRAME) {
	    objPtr->rawRefTable[i].ref2 = nextI;
	    nextI = i;
	}
	else if (objPtr->rawRefTable[i].type == MPEG_P_FRAME) {
	    objPtr->rawRefTable[i].ref2 = nextI;
	}
    }
	
    objPtr->numFrames = frameNum;
    objPtr->clipLength = objPtr->rawFTT[frameNum];
    objPtr->dataStart = 0;
    
    return TCL_OK;
    
  error:

    if (objPtr->rawFOT != NULL) {
	ckfree((char *) objPtr->rawFOT);
	objPtr->rawFOT = NULL;
    }
    if (objPtr->rawFTT != NULL) {
	ckfree((char *) objPtr->rawFTT);
	objPtr->rawFTT = NULL;
    }
    if (objPtr->rawRefTable != NULL) {
	ckfree((char *) objPtr->rawRefTable);
	objPtr->rawRefTable = NULL;
    }
    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);
	objPtr->cfInfo = NULL;
    }

    if (objPtr->rawfd > -1) {
	close(objPtr->rawfd);
	objPtr->rawfd = -1;
    }
    
    return TCL_ERROR;
}
  

/************************************************************
 *
 * MPEG_CF_OpenClipfile
 *
 * Parameters:
 *     objPtr - pointer to object
 *     filename - file name
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Opens the clipfile, reads basic info data.
 *     Closes file when done.
 ************************************************************/

static int
MPEG_CF_OpenClipfile(MPEG_CF *objPtr, char *filename)
{
    tChunkHeader chunk_hdr;
    MediaIndepHeader mi_hdr;
    int i;

    if (objPtr->cfState != NULL) {
	Tcl_AppendResult(objPtr->interp, "Problem: chunk file already open", 
			 (char *) NULL);
	return TCL_ERROR;
    }

    /* 
     * Attempt to open the file specified as a chunk file (i.e. pre formatted
     * clipfile). If this fails, attempt to open the file as a raw data file.
     */

    objPtr->cfState = Chunk_OpenFile(filename, "r");
    if (objPtr->cfState == NULL) {

      if (MPEG_CF_OpenRawfile(objPtr, filename) != TCL_OK) {
	Tcl_AppendResult(objPtr->interp, "Couldn't open file: ", filename,
			 " as either mpeg clipfile or raw mpeg data file.", 
			 (char *) NULL);
	return TCL_ERROR;
      }
      return TCL_OK;
    }

    /*
     * Get information from media independent header. 
     */

    if (Chunk_FindType(objPtr->cfState, CM_MEDIA_INDEP_CF, &chunk_hdr) != 0) {
	Tcl_AppendResult(objPtr->interp,
		      "Couldn't find media independent header.", (char *) NULL);
	return TCL_ERROR;
    }

    read(objPtr->cfState->fd, &mi_hdr, sizeof(mi_hdr));

    mi_hdr.media = ntohl(mi_hdr.media);
    mi_hdr.format = ntohl(mi_hdr.format);
    mi_hdr.device = ntohl(mi_hdr.device);
    mi_hdr.version = ntohl(mi_hdr.version);
    mi_hdr.num_frames = ntohl(mi_hdr.num_frames);
    mi_hdr.length_s = ntohl(mi_hdr.length_s);
    mi_hdr.length_ns = ntohl(mi_hdr.length_ns);

    /* Sanity check for type. */

    if ((mi_hdr.media != VIDEO_MEDIA_CODE) || 
	(mi_hdr.format != MPEG_FORMAT_CODE)) {
      Tcl_AppendResult(objPtr->interp,
		       "File is not a mpeg video chunkfile.", (char *) NULL);
      return TCL_ERROR;
    }

    objPtr->numFrames = mi_hdr.num_frames;
    objPtr->clipLength = mi_hdr.length_s + mi_hdr.length_ns/1000000000.0;

    /*
     * Get information from mpeg clipfile header.
     */

    if (Chunk_FindType(objPtr->cfState, CM_MPEG_CF, &chunk_hdr) != 0) {
	Chunk_ReadRewind(objPtr->cfState);
	if (Chunk_FindType(objPtr->cfState, CM_MPEG_CF, &chunk_hdr) != 0) {
	    Tcl_AppendResult(objPtr->interp,
			"Couldn't find mpeg header.", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);;
    }
    objPtr->cfInfo = (MPEGHeader *) ckalloc(sizeof(MPEGHeader));

    read(objPtr->cfState->fd, objPtr->cfInfo, sizeof(MPEGHeader));

    objPtr->cfInfo->width = ntohl(objPtr->cfInfo->width);
    objPtr->cfInfo->height = ntohl(objPtr->cfInfo->height);
    objPtr->cfInfo->aspectRatioCode = ntohl(objPtr->cfInfo->aspectRatioCode);
    objPtr->cfInfo->frameRateCode = ntohl(objPtr->cfInfo->frameRateCode);
    objPtr->cfInfo->bitRate = ntohl(objPtr->cfInfo->bitRate);
    objPtr->cfInfo->vbvBufferSize = ntohl(objPtr->cfInfo->vbvBufferSize);
    objPtr->cfInfo->constrainedParamFlag = ntohl(objPtr->cfInfo->constrainedParamFlag);

    /*
     * Find data segment and store location in file.
     */

    if (Chunk_FindType(objPtr->cfState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
	Chunk_ReadRewind(objPtr->cfState);
	if (Chunk_FindType(objPtr->cfState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
	    Tcl_AppendResult(objPtr->interp, "Couldn't find data segment.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
    }

    objPtr->dataStart = lseek(objPtr->cfState->fd, 0, SEEK_CUR);

    /*
     * Find header table and store table in memory. 
     */

    if (Chunk_FindType(objPtr->cfState, CM_MPEG_HDR_TABLE, &chunk_hdr) != 0) {
	Chunk_ReadRewind(objPtr->cfState);
	if (Chunk_FindType(objPtr->cfState, CM_MPEG_HDR_TABLE, &chunk_hdr) != 0) {
	    /* 
	     * No header table means that standard quant matrices are used.
	     */
	    goto done;
	}
    }

    read(objPtr->cfState->fd, &(objPtr->numHeaders), 4);
    objPtr->numHeaders = ntohl(objPtr->numHeaders);
    
    objPtr->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*objPtr->numHeaders);

    for (i=0; i<objPtr->numHeaders; i++) {
	MPEG_CF_GetHeaderEntryFromFile(objPtr->cfState->fd, &(objPtr->headerTable[i]));
    }

  done:

    Chunk_CloseFile(objPtr->cfState);
    objPtr->cfState = NULL;
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
MPEG_CF_ChangeOutCmd(MPEG_CF *objPtr, char *newOut) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
		    &(objPtr->outCD), (void *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    

/************************************************************
 *
 * MPEG_CF_ResetPickList
 *
 * Parameters:
 *     none.
 *
 * Return:
 *     none
 *
 * Action:
 *     Resets pick list variables and structures.
 *
 *************************************************************/

static void
MPEG_CF_ResetPickList ()
{
  pickListCounter = 0;

  if (pickListLength == 0) {
    pickListLength += 30;
    pickList = (FramePick *) ckalloc(sizeof(FramePick)*pickListLength);
  }

  return;
}

/************************************************************
 *
 * MPEG_CF_AddToPickList
 *
 * Parameters:
 *     objPtr         - pointer to MPEG_CF object
 *     segPtr         - pointer to segment of frame to be added
 *     index          - index of frame
 *     speedFlag      - -1 for speed < 0, 0 for speed = 0, 1 for speed > 0
 *     recurseFlag    - flag to indicate what to do for reference frames
 *                          0 = send only one frame, picking the most elgible for decode
 *                          1 = pick all frames needed but mark reference frames for
 *                              non-display
 *                          2 = pick all frames needed marking all frames for display
 *
 * Return:
 *    Standard tcl return value indicating success or not.
 *
 * Action:
 *    Given segment pointer and index of frame that we desire to send, this function
 *    recursively adds frames to the frame pick list according to the recurseFlag
 *    taking appropriate action for reference frames. 
 *
 *************************************************************/

static int
MPEG_CF_AddToPickList(MPEG_CF *objPtr,             /*Pointer to MPEG File object. */
		      Segment *segPtr,             /*Pointer to segment of frame. */
		      int index,                   /*Index number of frame. */
		      int speedFlag,               /* Indicates sign of speed. */
		      int recurseFlag             /*Flag indicating recursion action. */)
{
    int type;
    int ref1, ref2;
    
  /* 
   * Check validity of index. 
   */

  if ((index < 0) || (index >= segPtr->numFrames)) return TCL_ERROR;

  /*
   * Based on frame type do one of the following:
   * 
   * I_FRAME : fill out next pick structure and return.
   * P_FRAME : check ref1, if not sent then recurse.
   *           If recurse:
   *                Upon return, if recurseFlag = 0, return.
   *                             if recurseFlag = 1, unset last frames display flag
   *           Add this frame to pick list and return.
   * B_FRAME : Same a P Frame, but repeat ref procedure with ref2 after doing ref1
   *           and then add this frame and return.
   */

  type = segPtr->refTable[index].type;

  if (type != MPEG_I_FRAME) {
    /*
     * Must be either P or B. Do ref1 now.
     */
    
    ref1 = segPtr->refTable[index].ref1;
    if (ref1 < 0) return TCL_ERROR;
    if (segPtr->refTable[ref1].marker != objPtr->lastMarker) {
      if (MPEG_CF_AddToPickList(objPtr, segPtr, ref1, speedFlag, recurseFlag) 
	  != TCL_OK) {
	return TCL_ERROR;
      }
      
      if (!recurseFlag) return TCL_OK;
      else if (recurseFlag == 1) {
	pickList[pickListCounter-1].displayFlag = 0;
      }
    }
  }

  if (type == MPEG_B_FRAME) {
    ref2 = segPtr->refTable[index].ref2;
    if (ref2 < 0) return TCL_ERROR;
    if (segPtr->refTable[ref2].marker != objPtr->lastMarker) {
      if (MPEG_CF_AddToPickList(objPtr, segPtr, ref2, speedFlag, recurseFlag) 
	  != TCL_OK) {
	return TCL_ERROR;
      }
      
      if (!recurseFlag) return TCL_OK;
      else if (recurseFlag == 1) {
	pickList[pickListCounter-1].displayFlag = 0;
      }
    }
  }

  /*
   * Check marker for this frame. If already sent, don't pick it again.
   */

  if (segPtr->refTable[index].marker == objPtr->lastMarker) return TCL_OK;

  if (pickListCounter >= pickListLength) {
    pickListLength += 5;
    pickList = (FramePick *) ckrealloc((char *) pickList, sizeof(FramePick) * pickListLength);
  }
  
  pickList[pickListCounter].segPtr = segPtr;
  pickList[pickListCounter].index = index;
  pickList[pickListCounter].offset = segPtr->FOT[index*2];
  pickList[pickListCounter].size = segPtr->FOT[(index*2)+1];
  pickList[pickListCounter].sendFlag = 1;
  pickList[pickListCounter].displayFlag = 1;
    if (speedFlag < 0) {
	if (type != MPEG_B_FRAME) {
	    ref1 = segPtr->refTable[index].ref1;
	    if ((ref1 >= 0) && (ref1 < segPtr->numFrames)) {
		pickList[pickListCounter].eTime = segPtr->FTT[ref1+1];
	    } else {
		pickList[pickListCounter].eTime = segPtr->FTT[index];
	    }
	} else {
	    pickList[pickListCounter].eTime = segPtr->FTT[index];
	}
	pickList[pickListCounter].sTime = segPtr->FTT[index+1];
    } else {
	if (type != MPEG_B_FRAME) {
	    ref2 = segPtr->refTable[index].ref2;
	    if ((ref2 >= 0) && (ref2 < segPtr->numFrames)) {
		pickList[pickListCounter].eTime = segPtr->FTT[ref2];
	    } else {
		pickList[pickListCounter].eTime = segPtr->FTT[index+1];
	    }
	} else {
	    pickList[pickListCounter].eTime = segPtr->FTT[index+1];
	}
	pickList[pickListCounter].sTime = segPtr->FTT[index];
    }

    segPtr->refTable[index].marker = objPtr->lastMarker;

    pickListCounter++;
    return TCL_OK;
}
    
/************************************************************
 *
 * MPEG_CF_MakeSBListFromPickList
 *
 * Parameters:
 *     objPtr - pointer to MPEG File object.
 * 
 * Return:
 *     Pointer to Scatter Buffer List
 *
 * Action:
 *     Creates SB List filled with frames selected in pickList
 *     array. Ignores frames that are marked not to send.
 *     Includes system header frames as needed.
 *
 ************************************************************/
static SB_List *
MPEG_CF_MakeSBListFromPickList(MPEG_CF *objPtr)
{
    SB_List *sbl = NULL;
    Buffer *intra_buf = NULL;
    int intra_off = 0;
    int intra_len = 0;
    Buffer *non_intra_buf = NULL;
    int non_intra_off = 0;
    int non_intra_len = 0;
    Buffer *frame_buf = NULL;
    int frame_off = 0;
    int frame_len = 0;
    int num_to_send, i, sb_cntr, j;
    Segment *segPtr;
    int index;
    MPEGBufferHeader mpegBufHdr;
    int ref1, ref2, hdrRef, intra_matrix_type;
    int intra_matrix_length, next_value, non_intra_matrix_type;
    int non_intra_matrix_length;
    int data_seg_type, data_seg_length, data_fd;
    ScatterBuffer *sb;
    
    if (pickListCounter < 1) return NULL;

    /*
     * For each frame selected with sendFlag not set to 0:
     *     Check if system header needed. If so, 
     *     create required system header frame and add to SB List.
     *    
     *     Create frame add to SB List.
     */

    num_to_send = 0;
    for (i=0; i<pickListCounter; i++) {
	if (pickList[i].sendFlag)
	  num_to_send++;
    }

    if (num_to_send == 0) return NULL;

    sbl = (SB_List *) ckalloc(sizeof(SB_List));
    sbl->count = num_to_send;
    sbl->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *) * num_to_send);
    
    sb_cntr = 0;

    for (i=0; i<pickListCounter; i++) {
	if (pickList[i].sendFlag) {

	    segPtr = pickList[i].segPtr;
	    index = pickList[i].index;

	    /* Set up MPEGBuffferHeader */

	    mpegBufHdr.hdrType = CM_MPEG_BUF_HDR;
	    mpegBufHdr.length = MPEG_BUF_HDR_SIZE;
	    mpegBufHdr.width = objPtr->cfInfo->width;
	    mpegBufHdr.height = objPtr->cfInfo->height;
	    mpegBufHdr.type = segPtr->refTable[index].type;
	    
	    ref1 = segPtr->refTable[index].ref1;

	    if ((ref1 >= 0) && (mpegBufHdr.type != MPEG_I_FRAME)) {
		mpegBufHdr.ref1 = segPtr->refTable[ref1].indexNum;
	    }
	    else {
		mpegBufHdr.ref1 = -1;
	    }

	    ref2 = segPtr->refTable[index].ref2;

	    if ((ref2 >= 0) && (mpegBufHdr.type == MPEG_B_FRAME)) {
		mpegBufHdr.ref2 = segPtr->refTable[ref2].indexNum;
	    }
	    else {
		mpegBufHdr.ref2 = -1;
	    }

	    mpegBufHdr.index = segPtr->refTable[index].indexNum;
	    mpegBufHdr.logStart_s = floor(pickList[i].sTime);
	    mpegBufHdr.logStart_ns = floor((pickList[i].sTime -
					    floor(pickList[i].sTime)) * ONE_BILLION);
	    mpegBufHdr.logEnd_s = floor(pickList[i].eTime);
	    mpegBufHdr.logEnd_ns = floor((pickList[i].eTime -
					    floor(pickList[i].eTime)) * ONE_BILLION);
	    mpegBufHdr.displayFlag = pickList[i].displayFlag;

	    hdrRef = segPtr->refTable[index].headerRef;
	    if ((hdrRef >= 0) && (hdrRef < objPtr->numHeaders) && 
		(objPtr->headerTable != NULL)) {

		/*
		 * Need to send intra and/or non-intra quant tables.
		 */
		
		if (objPtr->headerTable[hdrRef].intra_matrix_flag) {
		    
		    /*
		     * Send intra matrix.
		     */

		    mpegBufHdr.intra_matrix_flag = 1;

		    intra_buf = MakeBuffer(objPtr->interp, 72, "Intra matrix");

		    if (intra_buf == NULL) {
			/* 
			 * Couldn't get buffer. Return nothing.
			 */

			goto error;
		    }

		    intra_matrix_type = htonl(CM_MPEG_INTRA_MATRIX);
		    intra_matrix_length = htonl(64);
		    
		    memcpy(intra_buf->data, &intra_matrix_type, 4);
		    memcpy(intra_buf->data+4, &intra_matrix_length, 4);
		    for (j=0; j<64; j++) {
			next_value = htonl(objPtr->headerTable[hdrRef].intra_matrix[j]);
			memcpy(intra_buf->data+8+(4*j), &next_value, 4);
		    }
		    intra_off = 0;
		    intra_len = 72;
		} else {
		    
		    mpegBufHdr.intra_matrix_flag = 0;
		}

		if (objPtr->headerTable[hdrRef].non_intra_matrix_flag) {
		    
		    /*
		     * Send non_intra matrix.
		     */

		    mpegBufHdr.non_intra_matrix_flag = 1;

		    non_intra_buf = MakeBuffer(objPtr->interp, 72, "Non_Intra matrix");
		    
		    if (non_intra_buf == NULL) {
			/* 
			 * Couldn't get buffer. Return nothing.
			 */
			
			goto error;
		    }
		    
		    non_intra_matrix_type = htonl(CM_MPEG_NON_INTRA_MATRIX);
		    non_intra_matrix_length = htonl(64);
		    
		    memcpy(non_intra_buf->data, &non_intra_matrix_type, 4);
		    memcpy(non_intra_buf->data+4, &non_intra_matrix_length, 4);
		    for (j=0; j<64; j++) {
			next_value = 
			  htonl(objPtr->headerTable[hdrRef].non_intra_matrix[j]);
			memcpy(non_intra_buf->data+8+(4*j), &next_value, 4);
		    }
		    non_intra_off = 0;
		    non_intra_len = 72;
		} else {
		    
		    mpegBufHdr.non_intra_matrix_flag = 0;
		}	
	    } else {
		mpegBufHdr.intra_matrix_flag = 0;
		mpegBufHdr.non_intra_matrix_flag = 0;
	    }

	    frame_buf = MakeBuffer(objPtr->interp, 8 + MPEG_BUF_HDR_SIZE + 8 + 
				   pickList[i].size, "Frame Buffer");

	    if (frame_buf == NULL) goto error;

	    mpegBufHdr.hdrType = htonl(mpegBufHdr.hdrType);
	    mpegBufHdr.length = htonl(mpegBufHdr.length);
	    mpegBufHdr.width = htonl(mpegBufHdr.width);
	    mpegBufHdr.height = htonl(mpegBufHdr.height);
	    mpegBufHdr.type = htonl(mpegBufHdr.type);
	    mpegBufHdr.ref1 = htonl(mpegBufHdr.ref1);
	    mpegBufHdr.ref2 = htonl(mpegBufHdr.ref2);
	    mpegBufHdr.intra_matrix_flag = htonl(mpegBufHdr.intra_matrix_flag);
	    mpegBufHdr.non_intra_matrix_flag = htonl(mpegBufHdr.non_intra_matrix_flag);
	    mpegBufHdr.index = htonl(mpegBufHdr.index);
	    mpegBufHdr.logStart_s = htonl(mpegBufHdr.logStart_s);
	    mpegBufHdr.logStart_ns = htonl(mpegBufHdr.logStart_ns);
	    mpegBufHdr.logEnd_s = htonl(mpegBufHdr.logEnd_s);
	    mpegBufHdr.logEnd_ns = htonl(mpegBufHdr.logEnd_ns);
	    mpegBufHdr.displayFlag = htonl(mpegBufHdr.displayFlag);

	    memcpy(frame_buf->data, &(mpegBufHdr), 8+MPEG_BUF_HDR_SIZE);

	    data_seg_type = htonl(CM_DATA_SEGMENT);
	    data_seg_length = htonl(pickList[i].size);

	    memcpy(frame_buf->data+8+MPEG_BUF_HDR_SIZE, &(data_seg_type), 4);
	    memcpy(frame_buf->data+12+MPEG_BUF_HDR_SIZE, &(data_seg_length), 4);
	    
	    /*
	     * Open file if necessary. 
	     */

	    if (objPtr->rawFOT != NULL) {
		if (objPtr->rawfd < 0) {
		    objPtr->rawfd = open(objPtr->filename, O_RDONLY|O_BINARY);
		}
		data_fd = objPtr->rawfd;
	    } else {
		if (objPtr->cfState == NULL) {
		    objPtr->cfState = Chunk_OpenFile(objPtr->filename, "r");
		} 
		data_fd = objPtr->cfState->fd;
	    }

	    lseek(data_fd, objPtr->dataStart+pickList[i].offset, SEEK_SET);

	    read(data_fd, frame_buf->data+16+MPEG_BUF_HDR_SIZE, pickList[i].size);

	    frame_off = 0;
	    frame_len = 16+MPEG_BUF_HDR_SIZE+pickList[i].size;

	    /* Pack it all up into a ScatterBuffer. */

	    sb = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));
	    
	    if (intra_buf != NULL) {
		if (non_intra_buf != NULL) {
		    sb->count = 3;
		    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *) * 3);
		    sb->buffers[0] = frame_buf;
		    sb->buffers[1] = intra_buf;
		    sb->buffers[2] = non_intra_buf;
		    sb->offsets = (int *) ckalloc(sizeof(int) * 3);
		    sb->offsets[0] = frame_off;
		    sb->offsets[1] = intra_off;
		    sb->offsets[2] = non_intra_off;
		    sb->lens = (int *) ckalloc(sizeof(int) * 3);
		    sb->lens[0] = frame_len;
		    sb->lens[1] = intra_len;
		    sb->lens[2] = non_intra_len;
		} else {
		    sb->count = 2;
		    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *) * 2);
		    sb->buffers[0] = frame_buf;
		    sb->buffers[1] = intra_buf;
		    sb->offsets = (int *) ckalloc(sizeof(int) * 2);
		    sb->offsets[0] = frame_off;
		    sb->offsets[1] = intra_off;
		    sb->lens = (int *) ckalloc(sizeof(int) * 2);
		    sb->lens[0] = frame_len;
		    sb->lens[1] = intra_len;
		}
	    } else {
		if (non_intra_buf != NULL) {
		    sb->count = 2;
		    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *) * 2);
		    sb->buffers[0] = frame_buf;
		    sb->buffers[1] = non_intra_buf;
		    sb->offsets = (int *) ckalloc(sizeof(int) * 2);
		    sb->offsets[0] = frame_off;
		    sb->offsets[1] = non_intra_off;
		    sb->lens = (int *) ckalloc(sizeof(int) * 2);
		    sb->lens[0] = frame_len;
		    sb->lens[1] = non_intra_len;	
		} else {
		    sb->count = 1;
		    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *) * 1);
		    sb->buffers[0] = frame_buf;
		    sb->offsets = (int *) ckalloc(sizeof(int) * 1);
		    sb->offsets[0] = frame_off;
		    sb->lens = (int *) ckalloc(sizeof(int) * 1);
		    sb->lens[0] = frame_len;
		}
	    }

	    sbl->sbs[sb_cntr++] = sb;

	    frame_buf = NULL;
	    intra_buf = NULL;
	    non_intra_buf = NULL;
	}
    }

    return sbl;

  error:
    
    if (frame_buf != NULL) {
	DetachBuffer(objPtr->interp, frame_buf, "Error in Make SBL from PickList");
	frame_buf = NULL;
    }
    if (intra_buf != NULL) {
	DetachBuffer(objPtr->interp, intra_buf, "Error in Make SBL from PickList");
	intra_buf = NULL;
    }
     if (non_intra_buf != NULL) {
	DetachBuffer(objPtr->interp, non_intra_buf, "Error in Make SBL from PickList");
	non_intra_buf = NULL;
    }

    if (sbl != NULL) {
	for (i=0; i<sb_cntr; i++) {
	    sb = sbl->sbs[i];
	    for (j=0; j<sb->count; j++) {
		DetachBuffer(objPtr->interp, sb->buffers[j],
			     "Error in Make SBL from PickList");
	    }
	    ckfree((char *) sb);
	}
	ckfree((char *) sbl);
    }

    return NULL;
}


/*
 *--------------------------------------------------------------
 *
 * MPEG_Prioritize --
 *
 *	A prioritization algorithm (by Daishi Harada) that computes the
 *	ordering for the MPEG frames.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters (these are my best guesses -- bsmith)
 *	int n		Number of frames to prioritize
 *	int m		Starting frame number.
 *	int k		Starting priority number
 *	int *pl		Output priority array
 *
 *--------------------------------------------------------------
 */
void
MPEG_Prioritize (int n, int m, int k, int *pl)
{
    int s,t;

    if (n==0) return;

    if (n==1) {
	*pl = m;
    } else {
	s = n/2;
	t = n-s;
	MPEG_Prioritize(t,m,k*2,pl);
	MPEG_Prioritize(s,m+t,k*2,pl+k);
    }
}

/************************************************************
 *
 * MPEG_CF_PrioritizePicklist
 *
 * Parameters:
 *     orderVec - array of n integers where n is number of
 *                frames to be prioritized.
 *   
 * Results:
 *     None.
 *
 * Actions:
 *     Sets value of i element in orderVec to be the index of
 *     the i th most important frame.
 *
 ************************************************************/

static void
MPEG_CF_PrioritizePickList (int *orderVec)
{
    int i, j, idx, numB, type, b_priority;

    /*
     * Make I and P frames most important.
     */

    idx = 0;

    for (i=0; i<pickListCounter; i++) {
	type = pickList[i].segPtr->refTable[pickList[i].index].type;
	if (type == MPEG_I_FRAME) {
	    orderVec[idx++] = i;
	}
    }

    for (i=0; i<pickListCounter; i++) {
	type = pickList[i].segPtr->refTable[pickList[i].index].type;
	if (type == MPEG_P_FRAME) {
	    orderVec[idx++] = i;
	}
    }

    numB = pickListCounter - idx;

    MPEG_Prioritize (numB, 0, 1, &(orderVec[idx]));

    b_priority = 0;

    for (i=0; i<pickListCounter; i++) {
	type = pickList[i].segPtr->refTable[pickList[i].index].type;
	if (type == MPEG_B_FRAME) {
	    for (j=idx; j<pickListCounter; j++) {
		if (orderVec[j] == b_priority) {
		    orderVec[j] = -1 * i;
		    break;
		}
	    }
	    b_priority++;
	}
    }

    for (j=idx; j<pickListCounter; j++) {
      if (orderVec[j] < 0) {
	orderVec[j] *= -1;
      }
    }

    return;
}

/************************************************************
 *
 * Mpeg_CF_SendData
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sends data according to current value of LTS, queues
 *     itself on the at queue for next send time.
 ************************************************************/

void
MPEG_CF_SendData(ClientData clientData) 
{
    MPEG_CF *objPtr = (MPEG_CF *)clientData;
    double now, speed, when;
    double eTime, sTime, lastETime;
    Segment *segPtr;
    int i, curBw;
    SB_List *sbl;
    char *sbStr;
    int numToSend, bwToSend, origNumToSend;
    int frameNum, curIndex;
    int data_fd;
    int speedFlag, recurseFlag;
    int want_to_send;
    double frac_want_to_send;
    int bw_want_to_send;

    /*
     * Return if not ready, no LTS, or no out command.
     */
    if ((!objPtr->ready) ||
	(objPtr->lts == NULL) ||
	(objPtr->outCmd == NULL)) {
	goto done;
    }

    /*
     * Find segment we are in.  Two cases: forward and reverse play.
     */
    now = LTS_Value(objPtr->lts);
    speed = LTS_Speed(objPtr->lts);
    now += objPtr->sendAhead * speed;
    if (objPtr->lastSeg == NULL) 
      segPtr = objPtr->segments;
    else segPtr = objPtr->lastSeg;

    if (speed >= 0.0) {
	while (segPtr != NULL) {
	    if (segPtr->logicalEnd > now) {
		if (segPtr->logicalStart <= now)
		  break;
		if (segPtr->prev == NULL) 
		  break;
		segPtr = segPtr->prev;
	    } else {
		if (segPtr->next != NULL) {
		    if (segPtr->next->logicalStart > now) {
			segPtr = segPtr->next;
			break;
		    }
		}
		segPtr = segPtr->next;
	    }
	}
    } else {
	/*
	 * Speed is backwards so now should be less or equal to segment end.
	 * and strictly greater than beginning. Almost identical to above
	 * code but with different boundary conditions.
	 */

	while (segPtr != NULL) {
	    if (segPtr->logicalEnd >= now) {
		if (segPtr->logicalStart < now)
		  break;
		segPtr = segPtr->prev;
	    } else {
		if (segPtr->next != NULL) {
		    if (segPtr->next->logicalStart >= now) {
			break;
		    }
		} else break;
		segPtr = segPtr->next;
	    }
	}
    }

    /*
     * If segment pointer is NULL, means we will never be
     * needed (i.e. time past all segments or going backwards
     * before all segments).
     */

    if (segPtr == NULL) {
	objPtr->lastSeg = NULL;
	objPtr->lastFrameNum = -1;
	goto done;
    }

    /*
     * If we are in the same segment we were in previously, lastFrameNum is
     * probably still good. If not, blow it away.
     */

    if (segPtr != objPtr->lastSeg) {
	objPtr->lastFrameNum = -1;
    }
    objPtr->lastSeg = segPtr;

    /*
     * If not in a segment, calculate when we will be and set up at queue.
     */
    if (speed > 0.0) {
	if (now < segPtr->logicalStart) {
	    when = LTS_SystemTimeOf(objPtr->lts,
			(segPtr->logicalStart-(objPtr->sendAhead*speed)));
	    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				       "video", MPEG_CF_SendData, objPtr, NULL,
				       NULL, NULL, NULL, NULL);
	    return;
	}
    } else if (speed < 0.0) {
	if (now > segPtr->logicalEnd) {
	    when = LTS_SystemTimeOf(objPtr->lts,
			segPtr->logicalEnd-(objPtr->sendAhead*speed));
	    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				     "video", MPEG_CF_SendData, objPtr, NULL,
				     NULL, NULL, NULL, NULL);
	  return;
	}
    } else {
	if ((now < segPtr->logicalStart) || (now >= segPtr->logicalEnd)) {
	    return;
	}
    }

    /*
     * If chunk file not open, open it now. Set data_fd to chunk file descriptor.
     * If raw file, open raw file and set data_fd.
     */
    
    if (objPtr->rawFOT != NULL) {
      if (objPtr->rawfd < 0) {
	objPtr->rawfd = open(objPtr->filename, O_RDONLY|O_BINARY);
      }
      data_fd = objPtr->rawfd;
    } else {      
      if (objPtr->cfState == NULL) {
	objPtr->cfState = Chunk_OpenFile(objPtr->filename, "r");
      }
      data_fd = objPtr->cfState->fd;
    }

    /*
     * If only sending one frame because speed = 0.0 or
     * cycle time = 0.0, do this now. 
     */
    if ((objPtr->cycleTime == 0.0) || (speed == 0.0)) {
	
	/*
	 * Find frame to send.
	 */
	if (speed > 0.0) {
	    for (i=objPtr->lastFrameNum+1; i<segPtr->numFrames; i++)  {
		if (segPtr->FTT[i+1] > now) 
		  break;
	    }
	    for (; i>=0; i--) {
		if (segPtr->FTT[i] <= now) 
		  break;
	    }
	} else if (speed < 0.0) {
	    if (objPtr->lastFrameNum == -1) 
	      i = segPtr->numFrames;
	    else i = objPtr->lastFrameNum;

	    for (; i>0; i--) {
		if (segPtr->FTT[i-1] < now) 
		    break;
	    }
	    for (; i<=objPtr->lastFrameNum; i++) {
		if (segPtr->FTT[i] >= now)
		  break;
	    }
	} else {
	    for (i=0; i<segPtr->numFrames; i++) {
		if ((segPtr->FTT[i] <= now) && (segPtr->FTT[i+1] > now)) {
		    break;
		}
	    }
	}
	
	if (speed > 0.0) {
	    if (i < 0) i = 0;
	    if (i >= segPtr->numFrames) {
		i = segPtr->numFrames-1;
	    }

	    sTime = segPtr->FTT[i];
	    eTime = segPtr->FTT[i+1];
	    frameNum = i;
	} else if (speed < 0.0) {
	    if (i <= 0) i = 1;
	    if (i > segPtr->numFrames) {
		i = segPtr->numFrames;
	    }

	    sTime = segPtr->FTT[i];
	    eTime = segPtr->FTT[i-1];
	    frameNum = i-1;
	} 
	else {
	  if (i < 0) i = 0;
	  if (i >= segPtr->numFrames) {
	    i = segPtr->numFrames-1;
	  }

	  sTime = eTime = now;
	  frameNum = i;
	}

	objPtr->lastFrameNum = frameNum;

	MPEG_CF_ResetPickList();
	
	if (speed < 0.0) speedFlag = -1;
	else if (speed > 0.0) speedFlag = 1;
	else speedFlag = 0;

	if (!speedFlag) recurseFlag = 1;
	else recurseFlag = 0;

	if (MPEG_CF_AddToPickList(objPtr, segPtr, frameNum, speedFlag, recurseFlag) ==
	    TCL_OK) {
	    
	    if (!speedFlag) {
		for (i=0; i<pickListCounter; i++) {
		    pickList[i].sTime = pickList[i].eTime = now;
		}
	    }
	    
	    sbl = MPEG_CF_MakeSBListFromPickList(objPtr);

	    if (sbl != NULL) {
		/*
		 * If out command is C, create scatter buffer list and call.
		 */
		if (objPtr->outCCmd != NULL) {
		    /*
		     * Need to check if it still exists since object may have
		     * been destroyed.
		     */
		    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
			objPtr->outCCmd(objPtr->outCD, sbl);	       
		    } else {
			/*
			 * Out C command has unexpectedly disappeared.
			 */
			objPtr->outCCmd = NULL;
			ckfree((char *) objPtr->outCmd);
			objPtr->outCmd = NULL;
		    }
		} else {
		    /*
		     * Out command is tcl, make sb list into a string and 
		     * call using Tcl_Eval. 
		     */
		    sbStr = MakeSBListString(sbl);
		    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
				sbStr, (char *) NULL); 
		    ckfree((char *) sbStr);
		}
		DetachToAllBuffers(objPtr->interp, sbl, 
				   "tcmMPEGFile.c:MPEG_CF_SendData");
		FreeSBList(sbl);
	    }
	}
	/*
	 * Figure out when we need to send again, queue it up and return.
	 */

	if (speed == 0.0) {
	    return;
	}

	when = LTS_SystemTimeOf(objPtr->lts,
		(eTime-(objPtr->sendAhead*speed)));
	objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				     "video", MPEG_CF_SendData, objPtr, NULL,
				     NULL, NULL, NULL, NULL);
	return;
    } 
    
    /*
     * Cycle time is not 0.0, so send multiple frames.
     * Find frames to be sent.
     */
    numToSend = 0;
    bwToSend = 0;

    /*
     * Find frame to send.
     */
    if (speed > 0.0) {
	for (i=objPtr->lastFrameNum+1; i<segPtr->numFrames; i++)  {
	    if (segPtr->FTT[i+1] > now) 
	      break;
	}
	for (; i>=0; i--) {
	    if (segPtr->FTT[i] <= now) 
	      break;
	}
    } else if (speed < 0.0) {
	if (objPtr->lastFrameNum == -1) 
	  i = segPtr->numFrames;
	else i = objPtr->lastFrameNum;
	
	for (; i>0; i--) {
	    if (segPtr->FTT[i-1] < now) 
	      break;
	}
	for (; i<=objPtr->lastFrameNum; i++) {
	    if (segPtr->FTT[i] >= now)
	      break;
	}
    }

    if (speed >= 0.0) {
	if (i < 0) {
	   i = 0;
	} else if (i >= segPtr->numFrames) {
	    i = segPtr->numFrames-1;
	}
	frameNum = i;
    } else {
	if (i <= 0) {
	   i = 1;
	} else if (i > segPtr->numFrames) {
	    i = segPtr->numFrames;
	}
	frameNum = i-1;
    }

    if (speed < 0.0) speedFlag = -1;
    else speedFlag = 1;

    MPEG_CF_ResetPickList();

    MPEG_CF_AddToPickList(objPtr, segPtr, frameNum, speedFlag, 2);

    curIndex = frameNum;
    objPtr->lastSeg = segPtr;
    objPtr->lastFrameNum = curIndex;

    if (speedFlag == 1) {
	while (1) {
	    curIndex++;
	    if (curIndex >= segPtr->numFrames) {
		segPtr = segPtr->next;
		if (segPtr == NULL) {
		    break;
		}
		curIndex = 0;
	    }
	    if (segPtr->FTT[curIndex] < now+objPtr->cycleTime*speed) {
		objPtr->lastSeg = segPtr;
		objPtr->lastFrameNum = curIndex;
		MPEG_CF_AddToPickList(objPtr, segPtr, curIndex, speedFlag, 2);
	    } else {
		break;
	    }
	}
    } else {
	while (1) {
	    curIndex--;
	    if (curIndex < 0) {
		segPtr = segPtr->prev;
		if (segPtr == NULL) {
		    break;
		}
		curIndex = segPtr->numFrames-1;
	    }
	    if (segPtr->FTT[curIndex+1] > now+objPtr->cycleTime*speed) {
		objPtr->lastSeg = segPtr;
		objPtr->lastFrameNum = curIndex;
		MPEG_CF_AddToPickList(objPtr, segPtr, curIndex, speedFlag, 2);
	    } else {
		break;
	    }
	}
    }
    
    bwToSend = 0;
    for (i=0; i<pickListCounter; i++) {
	bwToSend += pickList[i].size;
    }
    numToSend = pickListCounter;

    if (numToSend == 0) {
      /* 
       * This condition occurs when we got called to send frames
       * and we really shouldn't have (i.e., we've sent the last frame
       * of the segment that we were going to send, now we're trying to
       * send it again but the markers tell us we've already sent it, so it
       * didn't really get put on the pickList, thus numToSend is 0). 
       * Now we need to decide whether or not we need to be scheduled for the 
       * at queue. The best way to do this is to call ourselves again since 
       * by now we should be past the point where we are getting confused and
       * determine if we should sleep indefinitely or until the start time
       * of the next segment.
       */

      /*
       * If segment pointer is NULL, we've fallen off the end and won't be needed
       * unless the LTS changes in which case LTSChanged will do the right thing.
       */

      if (segPtr == NULL) {
	return;
      } 

      if (speed < 0.0) {
	when = LTS_SystemTimeOf(objPtr->lts,
				(segPtr->FTT[curIndex]-(objPtr->sendAhead*speed)));
      } else {
	when = LTS_SystemTimeOf(objPtr->lts,
				(segPtr->FTT[curIndex+1]-(objPtr->sendAhead*speed)));
      }
      objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
					     "video", MPEG_CF_SendData, objPtr, NULL,
					     NULL, NULL, NULL, NULL);
      return;
    }

    if (speed < 0.0) {
      lastETime = objPtr->lastSeg->FTT[pickList[pickListCounter-1].index];
    } else {
      lastETime = objPtr->lastSeg->FTT[pickList[pickListCounter-1].index+1];
    }

   /*
    * If number of frames to be sent greater than the number of
    * frames we can send (due to bandwidth or max frame rate),
    * mark low priority frames so we don't bother reading them.
    */

    frac_want_to_send = (objPtr->maxFrameRate * objPtr->cycleTime) - 
      objPtr->extraFrameCounter;
    want_to_send = (int) (ceil(frac_want_to_send));
    
    if (objPtr->maxFrameRate != 0.0) {
      objPtr->extraFrameCounter = ((double) want_to_send) - frac_want_to_send;
    }

    bw_want_to_send = objPtr->bandwidth - ((int) ((((double) objPtr->extraBWCounter) / 
						   objPtr->cycleTime)));

    curBw = (int) (((double) bwToSend) / objPtr->cycleTime);
    
    origNumToSend = numToSend;
    if ((numToSend > 0) &&
	(((numToSend > want_to_send) && objPtr->maxFrameRate) ||
	 ((curBw  > bw_want_to_send)    && objPtr->bandwidth && 
	  ((numToSend > 1) || (objPtr->extraBWCounter > 0))))) {
	int *orderVec;

	orderVec = (int *)ckalloc(numToSend*sizeof(int));
	MPEG_CF_PrioritizePickList (orderVec);
	while ((numToSend > 0) &&
	       ((objPtr->maxFrameRate && (numToSend > want_to_send)) ||
	        (objPtr->bandwidth && (curBw > bw_want_to_send) && 
		 ((numToSend > 1) || (objPtr->extraBWCounter > 0))))) {
	    int pick_idx;
	    int seg_idx;

	    numToSend--;

	    pick_idx = orderVec[numToSend];
	    seg_idx = pickList[pick_idx].index;

	    pickList[pick_idx].sendFlag = 0;
	    
	    /* 
	     * Invalidate marker. Since we are no longer sending this frame, 
	     * must change marker so that it does not appear to have been sent.
	     */

	    (pickList[pick_idx].segPtr)->refTable[seg_idx].marker = -1;

	    bwToSend -= pickList[pick_idx].size;
	    curBw  = (int) (((double) bwToSend) / objPtr->cycleTime);
	}
	ckfree ((char *)orderVec);
    }

    /*
     * Adjust bandwidth counter.
     */
    
    if (objPtr->bandwidth > 0) {
      objPtr->extraBWCounter += bwToSend - ((int) (objPtr->bandwidth * 
						   objPtr->cycleTime));
    }

    /*
     * Create scatte buffer list to send.
     */

    sbl = MPEG_CF_MakeSBListFromPickList(objPtr);

    if (sbl != NULL) {
	/*
	 * If out command is C, create scatter buffer list and call.
	 */
	if (objPtr->outCCmd != NULL) {
	    
	    
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree((char *) objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and
	     * call using Tcl_Eval. 
	     */
	    sbStr = MakeSBListString(sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree((char *) sbStr);
	}

	DetachToAllBuffers(objPtr->interp, sbl, 
			  "tcmMPEGFile.c:MPEG_CF_SendData");
	FreeSBList(sbl);
    }

    /*
     * Find next wake up time and set up atqueue.
     */
    when = LTS_SystemTimeOf(objPtr->lts,
		    (lastETime-(objPtr->sendAhead*speed)));
    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
			       "video", MPEG_CF_SendData, objPtr, NULL,
			       NULL, NULL, NULL, NULL);
    return;

done:
    if (objPtr->cfState != NULL) {
	Chunk_CloseFile(objPtr->cfState);
	objPtr->cfState = NULL;
    } 
    if (objPtr->rawfd > -1) {
      close(objPtr->rawfd);
      objPtr->rawfd = -1;
    }
    
    return;
}

/************************************************************
 *
 * MPEG_CF_FrameInfo
 *
 * Parameters:
 *     objPtr - pointer to object
 *     frameOffset - offset from current position
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Returns information about frame at current lts value
 *     plus/minus frameOffset.
 *     If the object is not ready, sets result to "" and returns
 *     TCL_OK. If current value of lts is not within a segment
 *     also sets result to "".
 ************************************************************/

static int
MPEG_CF_FrameInfo(MPEG_CF *objPtr, int frameOffset) 
{
    char type_char = '\0';
    char resString[256];
    int i;
    double now;
    Segment *segPtr;

    if ((!objPtr->ready) || (objPtr->lts == NULL)) {
	strcpy(objPtr->interp->result, "");
	return TCL_OK;
    }
    
    now = LTS_Value(objPtr->lts);
    
    if (objPtr->lastSeg == NULL)
      segPtr = objPtr->segments;
    else segPtr = objPtr->lastSeg;

    while (segPtr != NULL) {
	if (segPtr->logicalEnd > now) {
	    if (segPtr->logicalStart <= now)
	      break;
	    if (segPtr->prev == NULL) 
	      break;
	    segPtr = segPtr->prev;
	} else  {
	    if (segPtr->next != NULL) {
		if (segPtr->next->logicalStart > now) {
		    segPtr = segPtr->next;
		    break;
		}
	    }
	    segPtr = segPtr->next;
	}
    }

    if ((segPtr == NULL) || (now < segPtr->logicalStart)) {
	strcpy(objPtr->interp->result, "");
	return TCL_OK;
    }

    if (segPtr != objPtr->lastSeg) {
	objPtr->lastFrameNum = -1;
    }

    objPtr->lastSeg = segPtr;

    /*
     * Find current frame.
     */

    for (i=objPtr->lastFrameNum+1; i<segPtr->numFrames; i++)  {
	if (segPtr->FTT[i+1] > now) 
	  break;
    }
    for (; i>=0; i--) {
	if (segPtr->FTT[i] <= now) 
	  break;
    }

    if (i < 0) i = 0;
    if (i >= segPtr->numFrames) {
	i = segPtr->numFrames-1;
    }

    i += frameOffset;

    if ((i < 0) || (i >= segPtr->numFrames)) {
	strcpy(objPtr->interp->result, "");
	return TCL_OK;
    }

    if (segPtr->refTable[i].type == 1)
      type_char = 'I';
    else if (segPtr->refTable[i].type == 2)
      type_char = 'P';
    else if (segPtr->refTable[i].type == 3)
      type_char = 'B';

    sprintf(resString, "%d %g %g %d %c %d %d", i, segPtr->FTT[i], segPtr->FTT[i+1],
	    segPtr->FOT[(i*2)+1], type_char, segPtr->refTable[i].ref1,
	    segPtr->refTable[i].ref2);

    Tcl_SetResult(objPtr->interp, resString, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_SegmentList
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Returns list of current segment id's. If no segments,
 *     list is empty.
 ************************************************************/

static int
MPEG_CF_SegmentList(MPEG_CF *objPtr)
{
    Tcl_DString res_list;
    Segment *seg_ptr;
    char temp_str[100];

    seg_ptr = objPtr->segments;

    Tcl_DStringInit(&res_list);

    while (seg_ptr != NULL) {
	sprintf(temp_str, "%d", seg_ptr->segId);
    
	Tcl_DStringAppendElement(&res_list, temp_str);
	
	seg_ptr = seg_ptr->next;
    }

    Tcl_SetResult(objPtr->interp, Tcl_DStringValue(&res_list), TCL_VOLATILE);
    
    Tcl_DStringFree(&res_list);

    return TCL_OK;
}

/************************************************************
 *
 * MPEG_CF_ConfigureSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segmentId - id of segment to be reconfigured
 *     clipStart - new clip start.
 *     clipEnd - new clip end.
 *     logStart - new logical start.
 *     logEnd - new logical end.
 *     logStartFlag - indicates whetehr logStart value set or not.
 *     logEndFlag - indicates whether logEnd value set or not.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Reconfigures a previously created segment. Allows all
 *     aspects of the segment to be changed (i.e. clip start/end
 *     and logical start/end). Comparisons made with current
 *     segment info to decide whether change requires rebuilding
 *     tables or if current tables can be slightly altered.
 *     Segment removed from list and re-inserted to ensure proper
 *     ordering.
 *
 ************************************************************/

static int
MPEG_CF_ConfigureSegment(MPEG_CF *objPtr, int segmentId,
			 double clipStart, double clipEnd, 
			 double logStart, double logEnd, 
			 int logStartFlag, int logEndFlag)
{
  int new_seg_id;
  int result;
  Segment *segment_ptr;

  /*
   * This function could be optimized to recognize cases where
   * the clip only moves in logical time or the new clip is a 
   * proper subset of the original. In interest of getting this
   * function working as soon as possible, however, this code
   * currently just calls MPEG_CF_AddSegment to create a new
   * segment with the new clip parameters, deletes the old
   * segment and changes the segmentId of the new segment to be
   * the same as the old segment.
   */

  /* Unready the object. */
  MPEG_CF_Unready(objPtr);

  /* 
   * Find the original segment to be sure that it
   * existed.
   */

  segment_ptr = objPtr->segments;

  while (segment_ptr != NULL) {
    if (segment_ptr->segId == segmentId) 
      break;
    segment_ptr = segment_ptr->next;
  }

  if (segment_ptr == NULL) {
    char tempStr[40];

    sprintf(tempStr, "%d", segmentId);
    Tcl_AppendResult(objPtr->interp, "Segment ", tempStr, " does not exist.",
		     (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Remove the original segment.
   */

  MPEG_CF_RemoveSegment(objPtr, segmentId);

  /* 
   * Create a segment with new parameters.
   */

  result = MPEG_CF_AddSegment(objPtr, clipStart, clipEnd, logStart,
			      logEnd, logStartFlag, logEndFlag);

  /*
   * Change segment id of new segment to be segment id of original segment.
   */

  if (result == TCL_OK) {
    char resStr[40];

    if (Tcl_GetInt(objPtr->interp, objPtr->interp->result, &new_seg_id) != TCL_OK) {
      Tcl_AppendResult(objPtr->interp, "Couldn't reconfigure segment.", (char *) NULL);
      return TCL_ERROR;
    }

    segment_ptr = objPtr->segments;
    while (segment_ptr != NULL) {
      if (segment_ptr->segId == new_seg_id) {
	break;
      }
      segment_ptr = segment_ptr->next;
    }

    if (segment_ptr == NULL) {
      Tcl_AppendResult(objPtr->interp, "Couldn't reconfigure segment.", (char *) NULL);
      return TCL_ERROR;
    }

    segment_ptr->segId = segmentId;

    sprintf(resStr, "%d", segmentId);
    Tcl_SetResult(objPtr->interp, resStr, TCL_VOLATILE);

    /* 
     * Set segment list change flag so that next ready will check for overlaps.
     */

    objPtr->segmentListChangeFlag = 1;

    return TCL_OK;
  }

  return TCL_ERROR;
}



