/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id
 *
 * This file contains the code for the MPEG File Interface module
 * used by the MPEGSegment object.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <tcl.h>
#include <math.h>

#include "cmMPEGFileInterface.h"

#ifndef O_BINARY
#    define O_BINARY 0
#endif

/*
 * Misc. macros. 
 */

#define ONE_BILLION 1000000000.0


static int MPEG_FIP_FrameFromTime(MPEGFileInterface *fip, double time, double speed);
static int MPEG_FIP_PostRef(MPEGFileInterface *fip, int frameNum, int *refPtr);
static int MPEG_FIP_PreRefListRecurse(MPEGFileInterface *fip, int frameNum, int **reflistPtr, int level);
static int MPEG_FIP_PreRefList(MPEGFileInterface *fip, int frameNum, int **reflistPtr);
static int MPEG_FIP_ParseInfo(MPEGDataFile *dataFile);
static int MPEG_FIP_ParseInfoRaw (MPEGDataFile *dataFile);
static int NextStartCode(int fd);
static int FirstStartCode(int fd);
static void CopyRefEntry(RefEntry *r1, RefEntry *r2);
static void MPEG_FIP_FreeBitArray (DynamicBitArray *barray);
static DynamicBitArray * MPEG_FIP_NewBitArray (int size);



/* 
 * Initialization flag indicates whether hash table has been set up.
 */

static moduleInitFlag = 0;
static Tcl_HashTable fileHashTable;


/*********************************************************************
 * MPEG_FIP_ModuleInit
 *
 * Parameters:
 *    None
 *
 * Results:
 *    None
 * 
 * Action:
 *    Initializes MPEG File Interface module by setting up hash
 *    table and setting init flag.
 *********************************************************************/
static void
MPEG_FIP_ModuleInit()
{
    if (moduleInitFlag) return;

    moduleInitFlag = 1;

    Tcl_InitHashTable(&fileHashTable, TCL_STRING_KEYS);
}


/********************************************************************
 * MPEG_FIP_Open
 *
 * Parameters:
 *    fname - filename for desired file.
 *
 * Results:
 *    Pointer to MPEGFileInterface structure.
 * 
 * Action:
 *    Finds underlying file data structure in hash table, creating
 *    it if not found. If the file data structure had to be created
 *    table parsing is started. Creates a file interface structure 
 *    for the segment object and returns it.
 *********************************************************************/

MPEGFileInterface *
MPEG_FIP_Open(char *fname)
{
    MPEGFileInterface *new_fip;
    MPEGDataFile *dataFile;
    Tcl_HashEntry *entry_ptr;
    int newFile;

    /* 
     *  Find data file structure if it already exists. 
     */

    if (!moduleInitFlag) {
	MPEG_FIP_ModuleInit();
    }

    entry_ptr = Tcl_CreateHashEntry(&fileHashTable, fname, &newFile);

    if (newFile) {

	/* 
         * Didn't find one, so create one. 
	 */

	dataFile = (MPEGDataFile *) ckalloc(sizeof(MPEGDataFile));
	dataFile->filename = (char *) ckalloc(strlen(fname)+1);
	strcpy(dataFile->filename, fname);
	dataFile->cfState = NULL;
	dataFile->FOT = NULL;
	dataFile->FTT = NULL;
	dataFile->refTable = NULL;
	dataFile->numFrames = 0;
	dataFile->clipLength = 0.0;
	dataFile->dataStart = 0;
	dataFile->numHeaders = 0;
	dataFile->headerTable = NULL;
	dataFile->raw_flag = 0;
	dataFile->fd = -1;
	dataFile->cfInfo = NULL;
	dataFile->file_refCount = 1;
	dataFile->open_refCount = 0;

	/* 
	 * Get basic info (i.e., width, height, etc.)
	 * and parse info about each file.
	 */

	if (MPEG_FIP_ParseInfo(dataFile) != TCL_OK) {
	    Tcl_DeleteHashEntry(entry_ptr);
	    ckfree(dataFile->filename);
	    ckfree((char *) dataFile);
	    return NULL;
	}
	Tcl_SetHashValue(entry_ptr, (ClientData) dataFile);
    } else {
	/* 
	 * Found one that already existed.
	 */
	dataFile = (MPEGDataFile *) Tcl_GetHashValue(entry_ptr);
	dataFile->file_refCount++;
    }

    /* 
     * Create file interface structure for segment, fill and return. 
     */

    new_fip = (MPEGFileInterface *) ckalloc(sizeof(MPEGFileInterface));
    
    new_fip->data_file = dataFile;
    new_fip->is_open = 0;
    new_fip->bit_array = MPEG_FIP_NewBitArray(300);

    return new_fip;
}


/********************************************************************
 * MPEG_FIP_FDOpen
 *
 * Parameters:
 *    fip - Pointer to file interface structure.
 *
 * Results:
 *    None
 * 
 * Action:
 *    Used by segment object to indicate that it would like the
 *    file to have been "opened". File may or may not already be
 *    open. Reference count needs to be maintained.
 *********************************************************************/

void
MPEG_FIP_FDOpen(MPEGFileInterface *fip)
{
    MPEGDataFile *dataFile;

    /* 
     * If file is already open in context to this segment object,
     * do nothing. This prevents the reference count for increasing
     * if a segment object does more than one open call.
     */

    if (fip->is_open == 1) {
	return;
    }

    /*
     * Set open flag in file interface structure and advance
     * reference count.
     */

    fip->is_open = 1;

    dataFile = fip->data_file;
    dataFile->open_refCount++;

    /*
     * Make sure file is actually open.
     */

    if (dataFile->raw_flag) {
	if (dataFile->fd < 0) {
	    dataFile->fd = open(dataFile->filename, O_RDONLY|O_BINARY);
	}
    } else {
	if (dataFile->cfState == NULL) {
	    dataFile->cfState = Chunk_OpenFile(dataFile->filename, "r");
	    dataFile->fd = dataFile->cfState->fd;
	}
    }
}


/********************************************************************
 * MPEG_FIP_FDClose
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    None
 * 
 * Action:
 *    Used by segment object to close file. Doesn't actually happen
 *    unless all segment objects that share this file close.
 *********************************************************************/
void
MPEG_FIP_FDClose(MPEGFileInterface *fip) 
{
    MPEGDataFile *dataFile;

    /*
     * Check for open flag. Maintain reference count.
     */

    if (fip->is_open == 0) {
	return;
    } 

    fip->is_open = 0;
    dataFile = fip->data_file;
    dataFile->open_refCount--;

    /*
     * If reference count falls to zero, actually do the close.
     */

    if (dataFile->open_refCount <= 0) {
	if (dataFile->raw_flag) {
	    if (dataFile->fd > -1) {
		close(dataFile->fd);
	    }
	} else {
	    if (dataFile->cfState != NULL) {
		Chunk_CloseFile(dataFile->cfState);
		dataFile->cfState = NULL;
	    }
	}
	dataFile->fd = -1;
    }
}


/********************************************************************
 * MPEG_FIP_Close
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    None
 * 
 * Action:
 *    Called by segment object to free file interface structure.
 *    Indicates that segment object is now totally divorced from
 *    this file.
 *********************************************************************/

void
MPEG_FIP_Close(MPEGFileInterface *fip) 
{
    Tcl_HashEntry *entry_ptr;
    MPEGDataFile *dataFile;

    /* 
     * Close the file.
     */

    MPEG_FIP_FDClose(fip);

    dataFile = fip->data_file;

    /*
     * Decrement file structure reference count, free if falls to 0.
     */

    dataFile->file_refCount--;

    if (fip->bit_array != NULL) {
	MPEG_FIP_FreeBitArray(fip->bit_array);
    }
    ckfree((char *) fip);

    if (dataFile->file_refCount <= 0) {
	entry_ptr = Tcl_FindHashEntry(&fileHashTable, dataFile->filename);

	if (entry_ptr != NULL) {
	    Tcl_DeleteHashEntry(entry_ptr);
	}

	if (dataFile->filename != NULL) {
	    ckfree(dataFile->filename);
	}
	if (dataFile->FOT != NULL) {
	    ckfree((char *) dataFile->FOT);
	}
	if (dataFile->FTT != NULL) {
	    ckfree((char *) dataFile->FTT);
	}
	if (dataFile->refTable != NULL) {
	    ckfree((char *) dataFile->refTable);
	}
	if (dataFile->headerTable != NULL) {
	    ckfree((char *) dataFile->headerTable);
	}
	if (dataFile->cfInfo != NULL) {
	    ckfree((char *) dataFile->cfInfo);
	}
	ckfree((char *) dataFile);
    }
}


/********************************************************************
 * MPEG_FIP_Width
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer width value.
 * 
 * Action:
 *    Used by segment object to access width information from file.
 *********************************************************************/	

int
MPEG_FIP_Width(MPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->width;
}


/********************************************************************
 * MPEG_FIP_Height
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer height value.
 * 
 * Action:
 *    Used by segment object to access height information from file.
 *********************************************************************/	

int
MPEG_FIP_Height(MPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->height;
}


/********************************************************************
 * MPEG_FIP_AspectRatioCode
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer aspect ratio code value.
 * 
 * Action:
 *    Used by segment object to access aspect ratio code information from file.
 *********************************************************************/	

int
MPEG_FIP_AspectRatioCode(MPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->aspectRatioCode;
}


/********************************************************************
 * MPEG_FIP_VBVBufferSize
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer vbv buffer size value.
 * 
 * Action:
 *    Used by segment object to access vbv buffer size information from file.
 *********************************************************************/	

int
MPEG_FIP_VBVBufferSize(MPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->vbvBufferSize;
}


/********************************************************************
 * MPEG_FIP_BitRate
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer bit rate value.
 * 
 * Action:
 *    Used by segment object to access bit rate information from file.
 *********************************************************************/	

int
MPEG_FIP_BitRate(MPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->bitRate;
}


/********************************************************************
 * MPEG_FIP_FrameRateCode
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer frame rate code value.
 * 
 * Action:
 *    Used by segment object to access frame rate code information from file.
 *********************************************************************/	

int
MPEG_FIP_FrameRateCode(MPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->frameRateCode;
}


/********************************************************************
 * MPEG_FIP_ConstrainedParamFlag
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Integer constrained parameter flag value.
 * 
 * Action:
 *    Used by segment object to access constrained parameter flag information from file.
 *********************************************************************/	

int
MPEG_FIP_ConstrainedParamFlag(MPEGFileInterface *fip) 
{
    return fip->data_file->cfInfo->constrainedParamFlag;
}



/********************************************************************
 * MPEG_FIP_ClipLength
 *
 * Parameters:
 *    fip - file interface pointer
 *
 * Results:
 *    Double clip length value.
 * 
 * Action:
 *    Used by segment object to access cliplength information from file.
 *********************************************************************/	

double
MPEG_FIP_ClipLength(MPEGFileInterface *fip) {
    MPEGDataFile *dataFile;

    dataFile = fip->data_file;

    return dataFile->clipLength;
}


/********************************************************************
 * MPEG_FIP_Info
 *
 * Parameters:
 *    fip - file interface pointer
 *    start - beginning of interval specified in clip time.
 *    end - end of interval specified in clip time.
 *    numFramesPtr - pointer to intege set to number of frames
 *                   in interval.
 *    maxSizePtr - pointer to integer set to max size of any frame in
 *                 interval.
 *    numIPtr - pointer to integer set to number of I frames in interval.
 *    numPPtr - pointer to integer set to number of P frames in interval.
 *    numBPtr - pointer to integer set to number of B frames in interval.
 *
 * Results:
 *    None.
 * 
 * Action:
 *    Returns info about frames in clip during interval defined by start and
 *    end.
 *********************************************************************/	

void
MPEG_FIP_Info(MPEGFileInterface *fip, 
	      double start, 
	      double end, 
	      int *numFramesPtr, 
	      int *maxSizePtr, 
	      int *numIPtr, 
	      int *numPPtr, 
	      int *numBPtr)
{
    int *framelist;
    MPEGDataFile *dataFile;
    int numFrames, numI, numP, numB, maxSize;
    int i, frame_num;
    double tmp, dummy;
    int type, size;

    dataFile = fip->data_file;

    if (start > end) {
	tmp = start;
	start = end;
	end = tmp;
    }

    if (start < 0.0) {
	start = 0.0;
    } else if (start >= dataFile->clipLength) {
	start = dataFile->clipLength;
    }

    if (end < 0.0) {
	end = 0.0;
    } else if (end >= dataFile->clipLength) {
	end = dataFile->clipLength;
    }

    numFrames = MPEG_FIP_FrameListFromTime(fip, start, end, 0, &framelist, &dummy);

    numI = 0;
    numP = 0;
    numB = 0;
    maxSize = 0;

    for (i=0; i<numFrames; i++) {
	frame_num = framelist[i];

	type = dataFile->refTable[frame_num].type;
	if (type == MPEG_I_FRAME) {
	    numI++;
	} else if (type == MPEG_P_FRAME) {
	    numP++;
	} else {
	    numB++;
	}

	size = dataFile->FOT[(frame_num*2)+1];
	if (size > maxSize) {
	    maxSize = size;
	}
    }

    ckfree((char *) framelist);

    *numFramesPtr = numFrames;
    *maxSizePtr = maxSize;
    *numIPtr = numI;
    *numPPtr = numP;
    *numBPtr = numB;

    return;
}
    

/********************************************************************
 * MPEG_FIP_GetAllFrameInfo
 *
 * Parameters:
 *    fip - file interface pointer
 *    frameNum - frame number of frame to get info about.
 *    sizePtr - pointer to integer set to size of frame.
 *    startPtr - pointer to double set to start time of frame.
 *    endPtr - pointer to double set to end time of frame.
 *    ref1Ptr - pointer to integer set to frame number of ref 1.
 *    ref2Ptr - pointer to integer set to frame number of ref 2.
 *    typePtr - pointer to integer set to type of frame.
 *
 * Results:
 *    Standard Tcl result code.
 * 
 * Action:
 *    Returns info about frame specified by frameNum.
 *********************************************************************/	

int MPEG_FIP_GetAllFrameInfo (MPEGFileInterface *fip, 
			      int frameNum, 
			      int *sizePtr, 
			      double *startPtr, 
			      double *endPtr, 
			      int *ref1Ptr, 
			      int *ref2Ptr, 
			      int *typePtr)
{
    MPEGDataFile *dataFile;

    dataFile = fip->data_file;

    if ((frameNum < 0) || (frameNum >= dataFile->numFrames)) {
	return TCL_ERROR;
    }

    *sizePtr = dataFile->FOT[frameNum*2+1];
    *startPtr = dataFile->FTT[frameNum];
    *endPtr = dataFile->FTT[frameNum+1];
    *ref1Ptr = dataFile->refTable[frameNum].ref1;
    *ref2Ptr = dataFile->refTable[frameNum].ref2;
    *typePtr = dataFile->refTable[frameNum].type;

    return TCL_OK;
}



/********************************************************************
 * MPEG_FIP_TimeFromFrame
 *
 * Parameters:
 *    fip - file interface pointer
 *    frameNum - frame number of frame to get info about.
 *    startPtr - pointer to double set to start time of frame.
 *    endPtr - pointer to double set to end time of frame.
 *
 * Results:
 *    Standard Tcl error code.
 * 
 * Action:
 *    Returns timing info about frame specified by frameNum.
 *********************************************************************/	

int
MPEG_FIP_TimeFromFrame(MPEGFileInterface *fip, 
		       int frameNum, 
		       double *startPtr, 
		       double *endPtr)
{
    MPEGDataFile *dataFile;
    
    dataFile = fip->data_file;

    if ((frameNum < 0) || (frameNum >= dataFile->numFrames)) {
	return TCL_ERROR;
    }

    *startPtr = dataFile->FTT[frameNum];
    *endPtr = dataFile->FTT[frameNum+1];
    return TCL_OK;
}


/********************************************************************
 * MPEG_FIP_ReadFrameFromFile
 *
 * Parameters:
 *    fip - file interface pointer
 *    frameNum - frame number of frame to get info about.
 *    buffer - pointer to buffer where frame data should go.
 *
 * Results:
 *    Standard Tcl error code.
 * 
 * Action:
 *    Reads frame data from file into buffer. Buffer is assumed to be
 *    large enough.
 *********************************************************************/	

int 
MPEG_FIP_ReadFrameFromFile (MPEGFileInterface *fip, int frameNum, char *buffer)
{
    MPEGDataFile *dataFile;
    int offset, length;

    if (!fip->is_open) {
	MPEG_FIP_FDOpen(fip);
    }

    dataFile = fip->data_file;

    offset = dataFile->FOT[frameNum*2];
    length = dataFile->FOT[(frameNum*2)+1];

    lseek(dataFile->fd, dataFile->dataStart+offset, SEEK_SET);
    read(dataFile->fd, buffer, length);

    return TCL_OK;
}


/********************************************************************
 * MPEG_FIP_FrameListFromTime
 *
 * Parameters:
 *    fip - file interface pointer
 *    start - double value for start of interval.
 *    end - double value for end of interval.
 *    framelistPtr - pointer to integer array to be set to array of frame
 *                   numbers.
 *    etimePtr - pointer to double set to end time of last frame in list
 *               valid during interval.
 *
 * Results:
 *    Number of frames in framelist.
 * 
 * Action:
 *    Returns a list of frame numbers corresponding to frames required in
 *    the interval defined by start and end.
 *********************************************************************/	

int 
MPEG_FIP_FrameListFromTime (MPEGFileInterface *fip, 
			    double start, 
			    double end,  
			    int reverseFlag,
			    int **framelistPtr,
			    double *etimePtr)
{
    int start_frame, end_frame;
    int index, cntr;
    int *pre_ref_list;
    int post_ref;
    int num_pre_refs;
    int post_ref_flag;
    double dummy;

    start_frame = MPEG_FIP_FrameFromTime(fip, start, 0.0);
    end_frame = MPEG_FIP_FrameFromTime(fip, end, 0.0);

    if ((start_frame < end_frame) || 
	((start_frame == end_frame) &&
	 (!reverseFlag)))
      {
	  num_pre_refs = MPEG_FIP_PreRefList(fip, start_frame, &pre_ref_list);
	  post_ref_flag = MPEG_FIP_PostRef(fip, end_frame, &post_ref);
	  MPEG_FIP_TimeFromFrame(fip, end_frame, &dummy, etimePtr);

    } else {
	num_pre_refs = MPEG_FIP_PreRefList(fip, end_frame, &pre_ref_list);
	post_ref_flag = MPEG_FIP_PostRef(fip, start_frame, &post_ref);
	MPEG_FIP_TimeFromFrame(fip, start_frame, etimePtr, &dummy);

    }


    index = 0;

    if (start_frame == end_frame) {
	*framelistPtr = (int *) ckalloc(sizeof(int)*(1+num_pre_refs+post_ref_flag));
	
	for (cntr=0; cntr<num_pre_refs; cntr++) {
	    (*framelistPtr)[index++] = pre_ref_list[cntr];
	}
	if (post_ref_flag) {
	    (*framelistPtr)[index++] = post_ref;
	}
	(*framelistPtr)[index++] = start_frame;
	
    } else if (end_frame < start_frame) {
	
	*framelistPtr = (int *) ckalloc(sizeof(int)*(start_frame-end_frame+num_pre_refs+
						     post_ref_flag));
	for (cntr=0; cntr<num_pre_refs; cntr++) {
	    (*framelistPtr)[index++] = pre_ref_list[cntr];
	}
	if (post_ref_flag) {
	    (*framelistPtr)[index++] = post_ref;
	}
	for (cntr=start_frame-1; cntr >= end_frame; cntr--) {
	    (*framelistPtr)[index++] = cntr;
	}
	
    } else {
	*framelistPtr = (int *) ckalloc(sizeof(int)*(end_frame-start_frame+num_pre_refs+
						     post_ref_flag));
	for (cntr=0; cntr<num_pre_refs; cntr++) {
	    (*framelistPtr)[index++] = pre_ref_list[cntr];
	}
	if (post_ref_flag) {
	    (*framelistPtr)[index++] = post_ref;
	}
	for (cntr=start_frame; cntr < end_frame; cntr++) {
	    (*framelistPtr)[index++] = cntr;
	}
    }

    if (num_pre_refs != 0) {
	ckfree((char *) pre_ref_list);
    }

    return index;
}


/********************************************************************
 * MPEG_FIP_CreateIntraMatrixBuffer
 *
 * Parameters:
 *    interp - pointer to Tcl interpreter.
 *    fip - file interface pointer
 *    frameNum - frame number.
 *    buflenPtr - pointer to integer set to buffer length.
 *
 * Results:
 *    Pointer to buffer manager allocated buffer structure if intra
 *    matrix needs to be sent for frame number. NULL otherwise.
 * 
 * Action:
 *    Determines if intra matrix buffer needs to be sent. If so,
 *    creates a buffer structure with the proper buffer headers, etc.
 *    If not, returns NULL. buflenPtr used to pass back size of buffer.
 *********************************************************************/	

Buffer *
MPEG_FIP_CreateIntraMatrixBuffer (Tcl_Interp *interp,
				  MPEGFileInterface *fip, 
				  int frameNum, 
				  int *buflenPtr)
{
    MPEGDataFile *dataFile;
    Buffer *ret_buf;
    int j, intra_matrix_type, intra_matrix_length;
    int headerRef;
    int next_value;
    
    dataFile = fip->data_file;

    if ((frameNum < 0) || (frameNum >= dataFile->numFrames)) {
	return NULL;
    }

    headerRef = dataFile->refTable[frameNum].headerRef;
    if ( headerRef == -1) {
	return NULL;
    }

    if (!dataFile->headerTable[headerRef].intra_matrix_flag) {
	return NULL;
    }

    ret_buf= MakeBuffer(interp, 72, "MPEG intra matrix");

    if (ret_buf == NULL) return NULL;

    intra_matrix_type = htonl(CM_MPEG_INTRA_MATRIX);
    intra_matrix_length = htonl(64);
    
    memcpy(ret_buf->data, &intra_matrix_type, 4);
    memcpy(ret_buf->data+4, &intra_matrix_length, 4);

    for (j=0; j<64; j++) {
	next_value = htonl(dataFile->headerTable[headerRef].intra_matrix[j]);
	memcpy(ret_buf->data+8+(4*j), &next_value, 4);
    }

    *buflenPtr = 72;

    return ret_buf;
}

/********************************************************************
 * MPEG_FIP_CreateNonIntraMatrixBuffer
 *
 * Parameters:
 *    interp - pointer to Tcl interpreter.
 *    fip - file interface pointer
 *    frameNum - frame number.
 *    buflenPtr - pointer to integer set to buffer length.
 *
 * Results:
 *    Pointer to buffer manager allocated buffer structure if non-intra
 *    matrix needs to be sent for frame number. NULL otherwise.
 * 
 * Action:
 *    Determines if non-intra matrix buffer needs to be sent. If so,
 *    creates a buffer structure with the proper buffer headers, etc.
 *    If not, returns NULL. buflenPtr used to pass back size of buffer.
 *********************************************************************/	

Buffer *
MPEG_FIP_CreateNonIntraMatrixBuffer (Tcl_Interp *interp,
				     MPEGFileInterface *fip, 
				     int frameNum, 
				     int *buflenPtr)
{
    MPEGDataFile *dataFile;
    Buffer *ret_buf;
    int j, non_intra_matrix_type, non_intra_matrix_length;
    int headerRef, next_value;
    
    dataFile = fip->data_file;

    if ((frameNum < 0) || (frameNum >= dataFile->numFrames)) {
	return NULL;
    }

    headerRef = dataFile->refTable[frameNum].headerRef;
    if ( headerRef == -1) {
	return NULL;
    }

    if (!dataFile->headerTable[headerRef].non_intra_matrix_flag) {
	return NULL;
    }

    ret_buf= MakeBuffer(interp, 72, "MPEG non intra matrix");

    if (ret_buf == NULL) return NULL;

    non_intra_matrix_type = htonl(CM_MPEG_NON_INTRA_MATRIX);
    non_intra_matrix_length = htonl(64);
    
    memcpy(ret_buf->data, &non_intra_matrix_type, 4);
    memcpy(ret_buf->data+4, &non_intra_matrix_length, 4);

    for (j=0; j<64; j++) {
	next_value = htonl(dataFile->headerTable[headerRef].non_intra_matrix[j]);
	memcpy(ret_buf->data+8+(4*j), &next_value, 4);
    }

    *buflenPtr = 72;

    return ret_buf;
}


/********************************************************************
 * MPEG_FIP_FrameFromTime
 *
 * Parameters:
 *    fip - file interface pointer
 *    time - clip time value.
 *    speed - lts speed value (affects boundary conditions)
 *
 * Results:
 *    Integer frame number.
 * 
 * Action:
 *    Maps time to frame number.
 *********************************************************************/	

static int
MPEG_FIP_FrameFromTime(MPEGFileInterface *fip, double time, double speed) 
{
    MPEGDataFile *dataFile;
    int low_frame, high_frame, mid_frame;

    dataFile = fip->data_file;
    
    if (time < 0.0) {
	return 0;
    }

    if (time > dataFile->clipLength) {
	return dataFile->numFrames-1;
    }

    high_frame = dataFile->numFrames;
    
    low_frame = 0;
    
    if (time <= dataFile->FTT[low_frame]) {
	return low_frame;
    }

    if (time >= dataFile->FTT[high_frame]) {
	return high_frame-1;
    }

    while (high_frame-low_frame > 1) {
	mid_frame = (high_frame+low_frame) / 2;
	
	if (time < dataFile->FTT[mid_frame]) {
	    high_frame = mid_frame;
	} else if (time > dataFile->FTT[mid_frame]) {
	    low_frame = mid_frame;
	} else {
	    if (speed >= 0.0) {
		return mid_frame;
	    } else {
		return mid_frame-1;
	    }
	}
    }
    return low_frame;
}



/********************************************************************
 * MPEG_FIP_PreRefList
 *
 * Parameters:
 *    fip - file interface pointer
 *    frameNum - frame number
 *    reflistPtr - pointer to integer array to be set to array of 
 *                 pre reference frames.
 *
 * Results:
 *    Number of frames returned in list.
 * 
 * Action:
 *    Creates a list of pre reference (i.e. ref1) frames needed to
 *    decode frame.
 *********************************************************************/	

static int
MPEG_FIP_PreRefList(MPEGFileInterface *fip, int frameNum, int **reflistPtr)
{
    MPEGDataFile *dataFile;

    dataFile = fip->data_file;

    if ((frameNum < 0) || (frameNum >= dataFile->numFrames)) {
	return 0;
    }

    return MPEG_FIP_PreRefListRecurse(fip, frameNum, reflistPtr, 0);
}

/********************************************************************
 * MPEG_FIP_PreRefListRecurse
 *
 * Parameters:
 *    fip - file interface pointer.
 *    frameNum - frame number
 *    reflistPtr - pointer to integer array to be set to array of 
 *                 pre reference frames.
 *    level - indicates level of iteration.
 *
 *
 * Results:
 *    Length of reflist created.
 * 
 * Action:
 *    Checks type of frame. If I frame, returns 0. If P or B frame,
 *    recurses on ref1 field and inserts itself in list unless it
 *    was the very top level.
 *********************************************************************/	

static int
MPEG_FIP_PreRefListRecurse(MPEGFileInterface *fip, int frameNum, int **reflistPtr, 
			   int level)
{
    MPEGDataFile *dataFile;
    int type;
    int ref1;
    int list_length;

    dataFile = fip->data_file;

    type = dataFile->refTable[frameNum].type;
    
    if (type == MPEG_I_FRAME) {
	return 0;
    }

    ref1 = dataFile->refTable[frameNum].ref1;

    if (ref1 == -1) {
	return 0;
    }

    list_length = MPEG_FIP_PreRefListRecurse(fip, ref1, reflistPtr, level+1);
    
    if (list_length == 0) {
	*reflistPtr = (int *) ckalloc(sizeof(int)*(level+1));

	(*reflistPtr)[0] = ref1;

	list_length = level+1;
    }

    if (level != 0) {
	(*reflistPtr)[list_length-level] = frameNum;
    }

    return list_length;
}

	

/********************************************************************
 * MPEG_FIP_PostRef
 *
 * Parameters:
 *    fip - file interface pointer
 *    frameNum - frame number
 *    refPtr - pointer to integer to be set to 
 *             post reference frame.
 *
 * Results:
 *    1 if post ref frame exists, 0 otherwise.
 * 
 * Action:
 *    Returns ref2 in refPtr if needed with return value of 1 if
 *    needed and 0 if not.
 *********************************************************************/	

static int
MPEG_FIP_PostRef(MPEGFileInterface *fip, int frameNum, int *refPtr)
{
    MPEGDataFile *dataFile;
    int type;

    dataFile = fip->data_file;

    if ((frameNum < 0) || (frameNum >= dataFile->numFrames)) {
	return 0;
    }

    type = dataFile->refTable[frameNum].type;

    if (type != MPEG_B_FRAME) {
	return 0;
    }

    *refPtr = dataFile->refTable[frameNum].ref2;
    return 1;
}

    
/********************************************************************
 * MPEG_FIP_ParseInfo
 *
 * Parameters:
 *    dataFile - data file pointer
 *
 * Results:
 *    Standard Tcl error code.
 * 
 * Action:
 *    Gets basic info (i.e., width, height, etc.) from
 *    file and parses reference and header info about each file.
 *    Allocates space for and fills the various fields of the
 *    dataFile structure passed in.
 *********************************************************************/	

static int
MPEG_FIP_ParseInfo(MPEGDataFile *dataFile)
{
    MediaIndepHeader mi_hdr;
    tChunkHeader chunk_hdr;
    int *int_ftt_array;
    int i;

    /* 
     * Attempt to open the file specified as a chunk file (i.e. pre formatted
     * clipfile). If this fails, attempt to open the file as a raw data file.
     */
    
    dataFile->cfState = Chunk_OpenFile(dataFile->filename, "r");
    if (dataFile->cfState == NULL) {
	return MPEG_FIP_ParseInfoRaw(dataFile);
    }

    /*
     * Get information from media independent header. 
     */

    if (Chunk_FindType(dataFile->cfState, CM_MEDIA_INDEP_CF, &chunk_hdr) != 0) {
	goto error;
    }

    read(dataFile->cfState->fd, &mi_hdr, sizeof(mi_hdr));

    mi_hdr.media = ntohl(mi_hdr.media);
    mi_hdr.format = ntohl(mi_hdr.format);
    mi_hdr.device = ntohl(mi_hdr.device);
    mi_hdr.version = ntohl(mi_hdr.version);
    mi_hdr.num_frames = ntohl(mi_hdr.num_frames);
    mi_hdr.length_s = ntohl(mi_hdr.length_s);
    mi_hdr.length_ns = ntohl(mi_hdr.length_ns);
    
    /* Sanity check for type. */
    
    if ((mi_hdr.media != VIDEO_MEDIA_CODE) || 
	(mi_hdr.format != MPEG_FORMAT_CODE)) {
	goto error;
    }

    dataFile->numFrames = mi_hdr.num_frames;
    dataFile->clipLength = mi_hdr.length_s + mi_hdr.length_ns/1000000000.0;

    /*
     * Get information from mpeg clipfile header.
     */

    if (Chunk_FindType(dataFile->cfState, CM_MPEG_CF, &chunk_hdr) != 0) {
	Chunk_ReadRewind(dataFile->cfState);
	if (Chunk_FindType(dataFile->cfState, CM_MPEG_CF, &chunk_hdr) != 0) {
	    goto error;
	}
    }

    dataFile->cfInfo = (MPEGHeader *) ckalloc(sizeof(MPEGHeader));

    read(dataFile->cfState->fd, dataFile->cfInfo, sizeof(MPEGHeader));

    dataFile->cfInfo->width = ntohl(dataFile->cfInfo->width);
    dataFile->cfInfo->height = ntohl(dataFile->cfInfo->height);
    dataFile->cfInfo->aspectRatioCode = ntohl(dataFile->cfInfo->aspectRatioCode);
    dataFile->cfInfo->frameRateCode = ntohl(dataFile->cfInfo->frameRateCode);
    dataFile->cfInfo->bitRate = ntohl(dataFile->cfInfo->bitRate);
    dataFile->cfInfo->vbvBufferSize = ntohl(dataFile->cfInfo->vbvBufferSize);
    dataFile->cfInfo->constrainedParamFlag = ntohl(dataFile->cfInfo->constrainedParamFlag);

    /*
     * Find data segment and store location in file.
     */

    if (Chunk_FindType(dataFile->cfState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
	Chunk_ReadRewind(dataFile->cfState);
	if (Chunk_FindType(dataFile->cfState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
	    goto error;
	}
    }

    dataFile->dataStart = lseek(dataFile->cfState->fd, 0, SEEK_CUR);

    /*
     * Find header table and store table in memory. 
     */

    Chunk_ReadRewind(dataFile->cfState);
    if (Chunk_FindType(dataFile->cfState, CM_MPEG_HDR_TABLE, &chunk_hdr) == 0) {

	read(dataFile->cfState->fd, &(dataFile->numHeaders), 4);
	dataFile->numHeaders = ntohl(dataFile->numHeaders);
    
	dataFile->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*
						     dataFile->numHeaders);

	for (i=0; i<dataFile->numHeaders; i++) {
	    int readValueArray[64];
	    int idx = 0;
	    int j;

	    read(dataFile->cfState->fd, &(dataFile->headerTable[i].intra_matrix_flag), 4);
	    dataFile->headerTable[i].intra_matrix_flag = ntohl(dataFile->headerTable[i].intra_matrix_flag);

	    if (dataFile->headerTable[i].intra_matrix_flag) {

		read(dataFile->cfState->fd, readValueArray, 64*4);
		for (j=0; j<64; j++) {
		    dataFile->headerTable[i].intra_matrix[j] = ntohl(readValueArray[j]);
		}
	    }

	    read(dataFile->cfState->fd, &(dataFile->headerTable[i].non_intra_matrix_flag), 4);
	    dataFile->headerTable[i].non_intra_matrix_flag = 
	      ntohl(dataFile->headerTable[i].non_intra_matrix_flag);

	    if (dataFile->headerTable[i].non_intra_matrix_flag) {
		read(dataFile->cfState->fd, readValueArray, 64*4);
		for (j=0; j<64; j++) {
		    dataFile->headerTable[i].non_intra_matrix[j] = ntohl(readValueArray[j]);
		}
	    }
	}
    }

    /* Read in FOT. */

    Chunk_ReadRewind(dataFile->cfState);
    if (Chunk_FindType(dataFile->cfState, CM_FOT, &chunk_hdr) != 0) {
	goto error;
    }

    dataFile->FOT = (int *) ckalloc(sizeof(int)*dataFile->numFrames*2);
    
    read(dataFile->cfState->fd, dataFile->FOT, 4*dataFile->numFrames*2);
    for (i=0; i<dataFile->numFrames; i++) {
	dataFile->FOT[i] = ntohl(dataFile->FOT[i]);
    }

    /* Read in FTT. */

    Chunk_ReadRewind(dataFile->cfState);
    if (Chunk_FindType(dataFile->cfState, CM_FTT, &chunk_hdr) != 0) {
	goto error;
    }

    dataFile->FTT = (double *) ckalloc(sizeof(double)*(dataFile->numFrames+1));
    int_ftt_array = (int *) ckalloc(sizeof(int)*(dataFile->numFrames+1)*2);

    read(dataFile->cfState->fd, int_ftt_array, 4*(dataFile->numFrames+1)*2);
    
    for(i=0; i<=dataFile->numFrames; i++) {
	dataFile->FTT[i] = ((double) ntohl(int_ftt_array[i*2])) + 
	  (((double) ntohl(int_ftt_array[i*2+1])) / ONE_BILLION);
    }

    ckfree((char *) int_ftt_array);

    /* Read in Ref. Table. */

    Chunk_ReadRewind(dataFile->cfState);
    if (Chunk_FindType(dataFile->cfState, CM_MPEG_REF_TABLE, &chunk_hdr) != 0) {
	goto error;
    }

    dataFile->refTable = (RefEntry *) ckalloc(sizeof(RefEntry)*dataFile->numFrames);

    read(dataFile->cfState->fd, dataFile->refTable, 
	 sizeof(RefEntry)*dataFile->numFrames);
    
    for (i=0; i<dataFile->numFrames; i++) {
	dataFile->refTable[i].type = ntohl(dataFile->refTable[i].type);
	dataFile->refTable[i].ref1 = ntohl(dataFile->refTable[i].ref1);
	dataFile->refTable[i].ref2 = ntohl(dataFile->refTable[i].ref2);
	dataFile->refTable[i].headerRef = ntohl(dataFile->refTable[i].headerRef);
    }
	
    dataFile->raw_flag = 0;

    Chunk_CloseFile(dataFile->cfState);
    dataFile->cfState = NULL;
    dataFile->fd = -1;
    
    return TCL_OK;

  error:
    Chunk_CloseFile(dataFile->cfState);
    dataFile->cfState = NULL;

    if (dataFile->FOT != NULL) {
	ckfree((char *) dataFile->FOT);
	dataFile->FOT = NULL;
    }
    if (dataFile->FTT != NULL) {
	ckfree((char *) dataFile->FTT);
	dataFile->FTT = NULL;
    }
    if (dataFile->refTable != NULL) {
	ckfree((char *) dataFile->refTable);
	dataFile->refTable = NULL;
    }
    if (dataFile->headerTable != NULL) {
	ckfree((char *) dataFile->headerTable);
	dataFile->headerTable = NULL;
    }
    if (dataFile->cfInfo != NULL) {
	ckfree((char *) dataFile->cfInfo);
	dataFile->cfInfo = NULL;
    }

    return TCL_ERROR;

}


/********************************************************************
 * MPEG_FIP_ParseInfoRaw
 *
 * Parameters:
 *    dataFile - data file pointer
 *
 * Results:
 *    Standard Tcl error code.
 * 
 * Action:
 *    Gets basic info (i.e., width, height, etc.) from
 *    file and parses reference and header info about each frame.
 *    Allocates space for and fills the various fields of the
 *    dataFile structure passed in. This procedure used when the
 *    file is not a chunk file but is a raw MPEG-I video stream.
 *********************************************************************/

static int
MPEG_FIP_ParseInfoRaw (MPEGDataFile *dataFile)
{
    int hdrRef;
    int next_start_code;
    unsigned char readBuffer[200];
    int i, next_value;
    double sec_per_frame;
    int frameNum, ref1, ref2, lastI, nextI;
    int numFrameAlloced;
    int offset, picture_type, end_offset;
    int length;
    RefEntry tmpRefEntry;
    int tmpLength, tmpOffset;


    dataFile->fd = open(dataFile->filename, O_RDONLY|O_BINARY);

    if (dataFile->fd < 0) {
	return TCL_ERROR;
    }
    
    /* 
     * Basic algorithm is this:
     *
     * set hdrRef = -1
     * Find first system header.
     *   Parse header and set cfInfo.
     *   If intra or non_intra matrices given
     *     Parse tables. Append to headerTable (should be empty).
     *     Increment hdrRef.
     * Info in first header will give you sec/frame for calculating FTT values.
     * frameNum = 0, ref1 = -1, ref2 = -1
     * While next start code is not end of sequence
     *   If start code is system header
     *     If intra or non_intr matrices given.
     *     Parse tables. Append to headerTable, increment hdrRef.
     *   Else if picture start code.
     *     Parse type.
     *     Append rawFTT, rawFOT, rawRefTable
     *     Update current ref1 and ref2 based on type.
     *     Incr frameNum;
     *
     * Set numFrames, clipLength, dataStart, other misc. fields of dataFile.
     * Done.
     */
    
    hdrRef = -1;
    
    next_start_code = FirstStartCode(dataFile->fd);
    
    
    while ((next_start_code != MPEG_SYS_HDR_START_CODE) &&
	   (next_start_code != MPEG_END_SEQ_CODE) &&
	   (next_start_code != -1)) {
	lseek(dataFile->fd, 4, SEEK_CUR);
	next_start_code = NextStartCode(dataFile->fd);
    }
    
    if ((next_start_code == -1) || (next_start_code == MPEG_END_SEQ_CODE)) {
	/*
	 * Reached end of file without finding sys start code.
	 */	
	goto error;
    }

    /*
     * Read in system header, set up cfInfo.
     */

    read(dataFile->fd, readBuffer, 140);

    dataFile->cfInfo = (MPEGHeader *) ckalloc(sizeof(MPEGHeader));
    
    dataFile->cfInfo->width = (readBuffer[4]*16 + (readBuffer[5] >> 4));
    dataFile->cfInfo->height = (((readBuffer[5] & 0x0f) * 256) + readBuffer[6]);
    dataFile->cfInfo->aspectRatioCode = (readBuffer[7] >> 4);
    dataFile->cfInfo->frameRateCode = (readBuffer[7] & 0x0f);
    dataFile->cfInfo->bitRate = (((readBuffer[8] * 256*4) + (readBuffer[9] * 4) +
			       (readBuffer[10] >> 6)));
    dataFile->cfInfo->vbvBufferSize = ((((readBuffer[10] & 0x1f)*32) +
				     (readBuffer[11] >> 3)));
    if ((readBuffer[11] & 0x04)) 
      dataFile->cfInfo->constrainedParamFlag = 1;

    dataFile->numHeaders = 0;

    if ((readBuffer[11] & 0x02)) {
	hdrRef++;
	
	dataFile->numHeaders = 1;
	dataFile->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*1);
    
	dataFile->headerTable[0].intra_matrix_flag = 1;
	
	for(i=0; i<64; i++) {
	    next_value = ((readBuffer[11+i] & 0x01 * 128) + 
			  (readBuffer[12+i] >> 1));
	    dataFile->headerTable[0].intra_matrix[i] = next_value;
	}

	if ((readBuffer[75] & 0x01)) {
	    dataFile->headerTable[0].non_intra_matrix_flag = 1;
	    for (i=0; i<64; i++) {
		dataFile->headerTable[0].non_intra_matrix[i] = readBuffer[76+i];
	    }
	} else {
	    dataFile->headerTable[0].non_intra_matrix_flag = 0;
	    lseek(dataFile->fd, -1*(140-76), SEEK_CUR);
	}
    } else if ((readBuffer[11] & 0x01)) {
	hdrRef++;
	
	dataFile->numHeaders = 1;
	dataFile->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*1);
    
	dataFile->headerTable[0].intra_matrix_flag = 0;
	dataFile->headerTable[0].non_intra_matrix_flag = 1;
	for (i=0; i<64; i++) {
	    dataFile->headerTable[0].non_intra_matrix[i] = readBuffer[12+i];
	} 

	lseek (dataFile->fd, -1*(140-76), SEEK_CUR);
    } else {
	dataFile->numHeaders = 0;
	dataFile->headerTable = NULL;
	lseek (dataFile->fd, -1*(140-12), SEEK_CUR);
    }


    switch (dataFile->cfInfo->frameRateCode) {
      case 1:
	sec_per_frame = 1.0 / 23.976;
	break;
      case 2:
	sec_per_frame = 1.0 / 24.0;
	break;
      case 3:
	sec_per_frame = 1.0 / 25.0;
	break;
      case 4:
	sec_per_frame = 1.0 / 29.97;
	break;
      case 5:
	sec_per_frame = 1.0 / 30.0;
	break;
      case 6:
	sec_per_frame = 1.0 / 50.0;
	break;
      case 7:
	sec_per_frame = 1.0 / 59.94;
	break;
      case 8:
	sec_per_frame = 1.0 / 60.0;
	break;
      default:
	sec_per_frame = 1.0 / 30.0;
	break;
    }

    frameNum = 0; 
    ref1 = -1;
    ref2 = -1;
    numFrameAlloced = 100;

    dataFile->FOT = (int *) ckalloc(sizeof(int) * 2 * numFrameAlloced);
    dataFile->FTT = (double *) ckalloc(sizeof(double) * (numFrameAlloced+1));
    dataFile->refTable = (RefEntry *) ckalloc(sizeof(RefEntry) * numFrameAlloced);
    
    next_start_code = NextStartCode(dataFile->fd);

    while ((next_start_code != MPEG_END_SEQ_CODE) &&
	   (next_start_code != -1)) {
	
	if (next_start_code == MPEG_SYS_HDR_START_CODE) {
	    read(dataFile->fd, readBuffer, 12);

	    if ((readBuffer[11] & 0x02)) {
		hdrRef++;

		dataFile->numHeaders++;

		if (dataFile->headerTable == NULL) {
		    dataFile->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*
							       dataFile->numHeaders);
		} else {
		    dataFile->headerTable = (HdrEntry *) ckrealloc((char *) dataFile->headerTable,
							       sizeof(HdrEntry)*
							       dataFile->numHeaders);
		}

		read(dataFile->fd, &(readBuffer[12]), 64);
		
		dataFile->headerTable[hdrRef].intra_matrix_flag = 1;
	      
		for(i=0; i<64; i++) {
		    next_value = ((readBuffer[11+i] & 0x01 * 128) + 
				  (readBuffer[12+i] >> 1));
		    dataFile->headerTable[hdrRef].intra_matrix[i] = next_value;
		}

		if ((readBuffer[75] & 0x01)) {
		    dataFile->headerTable[hdrRef].non_intra_matrix_flag = 1;

		    read(dataFile->fd, &(readBuffer[76]), 64);

		    for (i=0; i<64; i++) {
			dataFile->headerTable[hdrRef].non_intra_matrix[i] = 
			  readBuffer[76+i];
		    }
		} else {
		    dataFile->headerTable[hdrRef].non_intra_matrix_flag = 0;
		}
	    } else if ((readBuffer[11] & 0x01)) {
		hdrRef++;

		dataFile->numHeaders++;

		if (dataFile->headerTable == NULL) {
		    dataFile->headerTable = (HdrEntry *) ckalloc(sizeof(HdrEntry)*
							       dataFile->numHeaders);
		} else {
		    dataFile->headerTable = (HdrEntry *) ckrealloc((char *) dataFile->headerTable,
							       sizeof(HdrEntry)*
							       dataFile->numHeaders);
		}
	
		dataFile->headerTable[hdrRef].intra_matrix_flag = 0;
		dataFile->headerTable[hdrRef].non_intra_matrix_flag = 1;
		
		read(dataFile->fd, &(readBuffer[12]), 64);

		for (i=0; i<64; i++) {
		    dataFile->headerTable[hdrRef].non_intra_matrix[i] = readBuffer[12+i];
		} 
	    }

	    next_start_code = NextStartCode(dataFile->fd);

	} else if (next_start_code == MPEG_PIC_START_CODE) {
	    
	    offset = lseek(dataFile->fd, 0L, SEEK_CUR);

	    read(dataFile->fd, readBuffer, 6);

	    picture_type = (int)(readBuffer[5] & 0x38) >> 3;
	    
	    next_start_code = NextStartCode(dataFile->fd);
	    if (next_start_code == -1) {
		break;
	    }
	    end_offset = lseek(dataFile->fd, 0L, SEEK_CUR);

	    length = end_offset - offset;

	    if (frameNum >= numFrameAlloced) {
		numFrameAlloced += 100;

		dataFile->FOT = (int *) ckrealloc((char *) dataFile->FOT,
						 sizeof(int) * 2 * numFrameAlloced);
		dataFile->FTT = (double *) ckrealloc((char *) dataFile->FTT, 
						    sizeof(double) * 
						    (numFrameAlloced+1));
		dataFile->refTable = (RefEntry *) ckrealloc((char *) dataFile->refTable,
							   sizeof(RefEntry) * 
							   numFrameAlloced);
	    }
	    
	    dataFile->FOT[frameNum*2] = offset;
	    dataFile->FOT[(frameNum*2)+1] = length;
	    dataFile->FTT[frameNum] = ((double) frameNum) * sec_per_frame;
	    dataFile->refTable[frameNum].type = picture_type;

	    if (picture_type != MPEG_B_FRAME) {

		/*
		 * We need to save the old reference 1 because we only want to do the
		 * flipping starting with the second and third reference frames and
		 * not the first and second. Thus, old_ref1 will be -1 for the two
		 * cases we want to avoid.
		 */

		ref1 = ref2;
		ref2 = frameNum;

		if (ref1 > -1) {
		    /*
		     * Here's the really tricky part.
		     * Now that we've got that next I or P after some
		     * sequence of B's, we need to adjust the index numbers
		     * to reflect display order instead of decode order.
		     */
		    
		    CopyRefEntry(&tmpRefEntry, &(dataFile->refTable[ref1]));
		    tmpOffset = dataFile->FOT[ref1*2];
		    tmpLength = dataFile->FOT[(ref1*2)+1];
		    
		    for (i=ref1; i<ref2-1; i++) {
			CopyRefEntry(&(dataFile->refTable[i]), 
				     &(dataFile->refTable[i+1]));
			dataFile->FOT[i*2] = dataFile->FOT[(i+1)*2];
			dataFile->FOT[(i*2)+1] = dataFile->FOT[((i+1)*2)+1];
		    }
		    
		    CopyRefEntry(&(dataFile->refTable[ref2-1]), &tmpRefEntry);
		    dataFile->FOT[(ref2-1)*2] = tmpOffset;
		    dataFile->FOT[((ref2-1)*2)+1] = tmpLength;
				    
		    for (i=ref1; i<ref2; i++) {
			dataFile->refTable[i].ref2 = ref2-1;
		    }

		    ref1 = ref2-1;
		}
	    }

	    dataFile->refTable[frameNum].ref1 = ref1;
	    dataFile->refTable[frameNum].ref2 = ref2;
	    
	    dataFile->refTable[frameNum].headerRef = hdrRef;

	    frameNum++;
	} else {
	    lseek(dataFile->fd, 4, SEEK_CUR);
	    next_start_code = NextStartCode(dataFile->fd);
	}
    }
    
    if (frameNum <= 0) goto error;

    /* 
     * Do the final flipping for the last set of B frames. 
     */

    ref1 = ref2;
    ref2 = frameNum;

    if (ref1 > -1) {
	CopyRefEntry(&tmpRefEntry, &(dataFile->refTable[ref1]));
	tmpOffset = dataFile->FOT[ref1*2];
	tmpLength = dataFile->FOT[(ref1*2)+1];
		    
	for (i=ref1; i<ref2-1; i++) {
	    CopyRefEntry(&(dataFile->refTable[i]), 
			 &(dataFile->refTable[i+1]));
	    dataFile->FOT[i*2] = dataFile->FOT[(i+1)*2];
	    dataFile->FOT[(i*2)+1] = dataFile->FOT[((i+1)*2)+1];
	}
		    
	CopyRefEntry(&(dataFile->refTable[ref2-1]), &tmpRefEntry);
	dataFile->FOT[(ref2-1)*2] = tmpOffset;
	dataFile->FOT[((ref2-1)*2)+1] = tmpLength;
				    
	for (i=ref1; i<ref2; i++) {
	    dataFile->refTable[i].ref2 = ref2-1;
	}

	ref1 = ref2-1;
    }
    
    dataFile->refTable = (RefEntry *) ckrealloc((char *) dataFile->refTable,
					       sizeof(RefEntry) * frameNum);
    dataFile->FTT = (double *) ckrealloc((char *) dataFile->FTT,
					sizeof(double) * (frameNum+1));
    dataFile->FOT = (int *) ckrealloc((char *) dataFile->FOT,
				     sizeof(int) * 2 * frameNum);

    dataFile->FTT[frameNum] = ((double) frameNum) * sec_per_frame;

    /*
     * Resolve I and P references now. 
     */

    lastI = -1;
    for (i=0; i<frameNum; i++) {
	if (dataFile->refTable[i].type == MPEG_I_FRAME) {
	    dataFile->refTable[i].ref1 = lastI;
	    lastI = i;
	}
    }

    nextI = -1;

    for (i=frameNum-1; i>=0; i--) {
	if (dataFile->refTable[i].type == MPEG_I_FRAME) {
	    dataFile->refTable[i].ref2 = nextI;
	    nextI = i;
	}
	else if (dataFile->refTable[i].type == MPEG_P_FRAME) {
	    dataFile->refTable[i].ref2 = nextI;
	}
    }

    dataFile->numFrames = frameNum;
    dataFile->clipLength = dataFile->FTT[frameNum];
    dataFile->dataStart = 0;

    dataFile->raw_flag = 1;
    dataFile->cfState = NULL;

    if (dataFile->fd > -1) {
	close(dataFile->fd);
	dataFile->fd = -1;
    }

    return TCL_OK;
    
  error:

    if (dataFile->FOT != NULL) {
	ckfree((char *) dataFile->FOT);
	dataFile->FOT = NULL;
    }
    if (dataFile->FTT != NULL) {
	ckfree((char *) dataFile->FTT);
	dataFile->FTT = NULL;
    }
    if (dataFile->refTable != NULL) {
	ckfree((char *) dataFile->refTable);
	dataFile->refTable = NULL;
    }
    if (dataFile->cfInfo != NULL) {
	ckfree((char *) dataFile->cfInfo);
	dataFile->cfInfo = NULL;
    }
    if (dataFile->headerTable != NULL) {
	ckfree((char *) dataFile->headerTable);
	dataFile->headerTable = NULL;
    }

    if (dataFile->fd > -1) {
	close(dataFile->fd);
	dataFile->fd = -1;
    }
    
    return TCL_ERROR;
}

/************************************************************
 *
 * NextStartCode
 *
 * Parameters:
 *     fd - open file descriptor
 *
 * Return:
 *     Next start code found or -1 if EOF
 *
 * Action:
 *     Reads bytes from open file descriptor looking for next start
 *     code. Once found, it returns the start code and lseeks
 *     the file descriptor to be set to the first byte of the start
 *     code. 
 *     NOTE: only looks for and finds picture, seq. end, and seq. start
 *           codes. It ignores all others.
 ************************************************************/

static int
NextStartCode(int fd) 
{
    unsigned char readBuffer[512];
    int ret_value, cur_pos, idx, num_read;
    int state;

    cur_pos = lseek(fd, 0L, SEEK_CUR);
    num_read = read(fd, readBuffer, 512);

    if (num_read <= 0) return -1;

    idx = 0;
    state = 0;
    ret_value = -1;

    while (1) {

	if (state == 0) {
	    if (readBuffer[idx] == 0) {
		state = 1;
	    }
	    idx++;
	} else if (state == 1) {
	    if (readBuffer[idx] == 0) {
		state = 2;
	    } else {
		state = 0;
	    }
	    idx++;
	} else if (state == 2) {
	    if (readBuffer[idx] == 1) {
		state = 3;
	    } else if (readBuffer[idx] != 0) {
		state = 0;
	    }
	    idx++;
	} else if (state == 3) {
	    if (readBuffer[idx] == 0) {
		ret_value = MPEG_PIC_START_CODE;
		break;
	    } else if (readBuffer[idx] == 179) {
		ret_value = MPEG_SYS_HDR_START_CODE;
		break;
	    } else if (readBuffer[idx] == 183) {
		ret_value = MPEG_END_SEQ_CODE;
		break;
	    } else {
		state = 0;
	    }
	    idx++;
	}
   
	if (idx >= num_read) {
	    cur_pos = lseek(fd, 0L, SEEK_CUR);
	    num_read = read(fd, readBuffer, 512);
	    idx = 0;

	    if (num_read <= 0) return -1;
	}
    }

    idx -= 3;
    lseek(fd, cur_pos+idx, SEEK_SET);
    return ret_value;
}
    

/************************************************************
 *
 * FirstStartCode
 *
 * Parameters:
 *     fd - open file descriptor
 *
 * Return:
 *     First start code found or -1 if not found in first 1024 bytes
 *
 * Action:
 *     Reads bytes from open file descriptor looking for next start
 *     code. Once found, it returns the start code and lseeks
 *     the file descriptor to be set to the first byte of the start
 *     code. 
 *     NOTE: only looks for and finds seq. start
 *           codes. It ignores all others. It only looks in the
 *           first 1024 bytes. This is a heuristic for determining
 *           whether the file is an mpeg at all.
 ************************************************************/

static int
FirstStartCode(int fd) 
{
    unsigned char readBuffer[1024];
    int ret_value, cur_pos, idx, num_read;
    int state;

    cur_pos = lseek(fd, 0L, SEEK_CUR);
    num_read = read(fd, readBuffer, 1024);

    if (num_read <= 0) return -1;

    idx = 0;
    state = 0;
    ret_value = -1;

    while (1) {

	if (state == 0) {
	    if (readBuffer[idx] == 0) {
		state = 1;
	    }
	    idx++;
	} else if (state == 1) {
	    if (readBuffer[idx] == 0) {
		state = 2;
	    } else {
		state = 0;
	    }
	    idx++;
	} else if (state == 2) {
	    if (readBuffer[idx] == 1) {
		state = 3;
	    } else if (readBuffer[idx] != 0) {
		state = 0;
	    }
	    idx++;
	} else if (state == 3) {
	    if (readBuffer[idx] == 179) {
		ret_value = MPEG_SYS_HDR_START_CODE;
		break;
	    } else {
		state = 0;
	    }
	    idx++;
	}
   
	if (idx >= num_read) {
	  return -1;
	}
    }

    idx -= 3;
    lseek(fd, cur_pos+idx, SEEK_SET);
    return ret_value;
}

/************************************************************
 *
 * CopyRefEntry
 *
 * Parameters:
 *     r1 - Ref Entry destination.
 *     r2 - Ref Entry source.
 *
 * Return:
 *     None.
 *
 * Action:
 *     Copies RefEntry from r2 to r1.
 ************************************************************/

static void
CopyRefEntry(RefEntry *r1, RefEntry *r2)
{
    memcpy((char *) r1, (char *) r2, sizeof(RefEntry));
    return;
}



/**********************************************************
 *
 * MPEG_FIP_NewBitArray
 *
 * Parameters:
 *     size - size of bit array cache.
 *
 * Results:
 *     Pointer to new bit array structure.
 *
 * Action:
 *     Creates and initializes new dynamic bit array.
 *
 ***********************************************************/

static DynamicBitArray *
MPEG_FIP_NewBitArray (int size)
{
    DynamicBitArray *new;

    new = (DynamicBitArray *) ckalloc(sizeof(DynamicBitArray));
    
    new->offset = 0;
    new->bit_array = (char *) ckalloc(size);
    memset(new->bit_array, 0, size);

    new->size = size;

    return new;
}

/**********************************************************
 *
 * MPEG_FIP_FreeBitArray
 *
 * Parameters:
 *     barray - pointer to dynamic bit array to be freed.
 *
 * Results:
 *     None.
 *
 * Action:
 *     Frees dynamic bit array.
 *
 ***********************************************************/

static void
MPEG_FIP_FreeBitArray (DynamicBitArray *barray)
{
    if (barray->bit_array != NULL) {
	ckfree((char *) barray->bit_array);
    }
    ckfree((char *) barray);
    return;
}

/**********************************************************
 *
 * MPEG_FIP_BitArraySet
 *
 * Parameters:
 *     barray - pointer to dynamic bit array to be freed.
 *     pos - integer bit position to be set.
 *
 * Results:
 *     None.
 *
 * Action:
 *     Sets bit pos in bit array sliding bit array window if 
 *     necessary.
 *
 ***********************************************************/

void
MPEG_FIP_BitArraySet(DynamicBitArray *barray, int pos)
{
    int idx, half;

    idx = pos-barray->offset;

    if ((idx < barray->size) && (idx >= 0)) {
	barray->bit_array[idx] = 1;
	return;
    }


    half = barray->size / 2;

    if (idx < 0) {
	if (idx < -half) {
	    memset(barray->bit_array, 0, barray->size);
	    idx = half;
	    barray->offset = pos - half;
	    barray->bit_array[idx] = 1;
	} else {
	    memmove(barray->bit_array+half, barray->bit_array, 
		    barray->size-half);
	    memset(barray->bit_array, 0, half);
	    barray->offset -= half;
	    idx += half;
	    barray->bit_array[idx] = 1;
	}
    } else {
	if (idx >= barray->size+half) {
	    memset(barray->bit_array, 0, barray->size);
	    idx = half;
	    barray->offset = pos - half;
	    barray->bit_array[idx] = 1;
	} else {
	    memmove(barray->bit_array, barray->bit_array+half, 
		    barray->size-half);
	    memset(barray->bit_array+barray->size-half, 0, half);
	    barray->offset += half;
	    idx -= half;
	    barray->bit_array[idx] = 1;
	}
    }
}

/**********************************************************
 *
 * MPEG_FIP_BitArrayGet
 *
 * Parameters:
 *     barray - pointer to dynamic bit array to be freed.
 *     pos - integer bit position to get.
 *
 * Results:
 *     Integer value (1 if set, 0 if not).
 *
 * Action:
 *     Returns value of pos bit in bitarray.
 *
 ***********************************************************/

int
MPEG_FIP_BitArrayGet(DynamicBitArray *barray, int pos)
{
    int idx;

    idx = pos-barray->offset;

    if (idx < 0) {
	return 0;
    } else if (idx >= barray->size) {
	return 0;
    } else {
	return barray->bit_array[idx];
    }
}

/**********************************************************
 *
 * MPEG_FIP_BitArrayClear
 *
 * Parameters:
 *     barray - pointer to dynamic bit array to be freed.
 *
 * Results:
 *     None.
 *
 * Action:
 *     Clears the bit array.
 *
 ***********************************************************/

void
MPEG_FIP_BitArrayClear(DynamicBitArray *barray)
{
    memset(barray->bit_array, 0, barray->size);
}
