/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMPEGFileSegment.c,v 1.8 1996/02/15 06:25:06 aswan Exp $
 *
 * Implementation of the MPEG file segment object, which reads
 * files containing MPEG video streams and produces cm mpeg
 * video streams.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>
#include "cmCookies.h"
#include "cmBufferlist.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmMPEGFileInterface.h"

#ifndef O_BINARY
#    define O_BINARY 0
#endif

#define ONE_BILLION 1000000000.0
#define CF_READ_BUFFER_SIZE 512
#define MPEG_BUF_HDR_SIZE (sizeof(MPEGBufferHeader)-8)
#define MPEG_I_FRAME 1
#define MPEG_P_FRAME 2
#define MPEG_B_FRAME 3



/*
 * Structure used when picking next set of frames to be sent.
 */

typedef struct {
    Buffer *bufPtr;
    int index;
    int frame_num;
    int size;
    double sTime;
    double eTime;
    int ref1;
    int ref2;
    int type;
    int sendFlag;
} FramePick;

/*
 * Structure kept for each clipfile object. 
 */

typedef struct MPEGSegment {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */

    ClientData lts;		/* LTS for object to use. */
    char *ltsName;
    int ltsTriggerNum;          /* LTS trigger for changed lts. */

    double sendAhead;		/* Send ahead time. */
    double cycleTime;		/* Cycle period. */

    int bandwidth;		/* Bandwidth available */
    int extraBWCounter;         /* Used to adjust number of bytes send per cycleTime
				   to mactch with bandwidth restrictions. */
    double maxFrameRate;	/* Maximum frame rate. */
    double extraFrameCounter;   /* Used to adjust number of frames sent per cycleTime
				   to match with maxFrameRate */

    char *filename;		/* Filename of clipfile. */
    MPEGFileInterface *fip;     /* File interface pointer. */
    MPEGHeader *cfInfo;	        /* MPEG Header from clipfile. */

    char *outCmd;		/* Command to send buffers to. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;           /* ClientData for outCCmd. */

    int lastFrameNum;           /* Frame number of last frame sent. */

    int ready;                  /* Ready flag. */

    int at_token;               /* Token for at event. */

    double clipStart;           /* Start time relative to clip beginning = 0.*/
    double clipEnd;             /* End time relative to clip beginning = 0. */
    double logicalStart;        /* Start time relative to lts. */
    double logicalEnd;          /* End time relative to lts. */
} MPEGSegment;

/* Globals used by this object module. */

static int nextMPEGSegmentNum = 0;	/* 
					 *  Variable for next unique 
					 *  object name. 
					 */

/* Macros for converting between object clip time and object logical time. */

#define LOG_TIME_FROM_CLIP_TIME(objPtr,time) ((((time - objPtr->clipStart) *          \
						(objPtr->logicalEnd -                 \
						 objPtr->logicalStart)) /             \
					       (objPtr->clipEnd - objPtr->clipStart)) \
					      + objPtr->logicalStart)

#define CLIP_TIME_FROM_LOG_TIME(objPtr,time) ((((time - objPtr->logicalStart) *    \
						(objPtr->clipEnd -                 \
						 objPtr->clipStart)) /             \
					       (objPtr->logicalEnd -               \
						objPtr->logicalStart))             \
					      + objPtr->clipStart)

/* Forward declarations. */

int MPEGSegmentObjCmd(ClientData clientData, Tcl_Interp * interp, int argc, char **argv);
static int MPEGSegmentInfo(MPEGSegment *objPtr, double start, double end);
static int MPEGSegmentReady(MPEGSegment *objPtr);
static int MPEGSegmentUnready(MPEGSegment *objPtr);
static int MPEGSegmentChangeOutCmd(MPEGSegment *objPtr, char *newOut);
static MPEGFileInterface *MPEGSegmentOpenClipfile(MPEGSegment *objPtr, char *filename);
void MPEGSegmentLTSChanged(ClientData clientdata);
void MPEGSegmentSendData(ClientData clientdata);
static SB_List *MPEGSegmentMakeSBLFromFPickList(MPEGSegment *objPtr, double start, double speed, int reverseFlag, FramePick *fPickList, int origNumFrames, int numFrames, int *orderVec);
static int MPEGSegmentBuildFPickList (MPEGSegment *objPtr, int *framelist, FramePick *fPickList, int numFrames);
static void MPEGSegmentPrioritize (int numFrames, FramePick *fPickList, int *orderVec);
static void MPEG_Prioritize (int n, int m, int k, int *pl);

/************************************************************
 *
 * CreateNewMPEGSegment
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

MPEGSegment *
CreateNewMPEGSegment(Tcl_Interp *interp,  /* Pointer to interpreter. */
		     char *name           /* Name of new object. */)
{

    MPEGSegment *new;

    new = (MPEGSegment *) ckalloc(sizeof(MPEGSegment));

    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);

    new->lts = NULL;
    new->ltsName = NULL;
    new->ltsTriggerNum = -1;

    new->sendAhead = 0.2;
    new->cycleTime = 0.0;

    new->bandwidth = 0;
    new->extraBWCounter = 0;
    new->maxFrameRate = 0.0;
    new->extraFrameCounter = 0.0;

    new->filename = NULL;
    new->fip = NULL;

    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;

    new->lastFrameNum = -1;

    new->cfInfo = NULL;

    new->ready = 0;

    new->at_token = 0;

    new->clipStart = 0.0;
    new->clipEnd = 0.0;
    new->logicalStart = 0.0;
    new->logicalEnd = 0.0;

    return new;
}

/************************************************************
 *
 * DestroyMPEGSegment
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

int
DestroyMPEGSegment(MPEGSegment *objPtr       /* Pointer to seg. object. */)
{

    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
	objPtr->at_token = 0;
    }

    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    if (objPtr->filename != NULL) {
	ckfree(objPtr->filename);
	objPtr->filename = NULL;
    }

    if (objPtr->fip != NULL) {
	MPEG_FIP_Close(objPtr->fip);
	objPtr->fip = NULL;
    }

    if (objPtr->outCmd != NULL)
	ckfree((char *) objPtr->outCmd);

    if (objPtr->cfInfo != NULL) {
	ckfree((char *) objPtr->cfInfo);
    }

    if (objPtr->ltsName != NULL) {
	ckfree((char *) objPtr->ltsName);
    }

    ckfree((char *) objPtr);

    return TCL_OK;
}

/************************************************************
 *
 * ConfigureMPEGSegment
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

int
ConfigureMPEGSegment(MPEGSegment *objPtr, int argc, char **argv)
{
  
    Tcl_Interp *interp;
    int cntr;
    char str[50];
    
    interp = objPtr->interp;
  
    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {} {}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName, "}} ",
		    (char *) NULL);
	}
	Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	Tcl_AppendResult(interp, "{-sendAhead {0.2} ", str, "} ", (char *)NULL);

	sprintf(str, "{%d}", objPtr->bandwidth);
	Tcl_AppendResult(interp, "{-bandwidth {0} ", str, "} ", (char *)NULL);

	Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	Tcl_AppendResult(interp, "{-cycleTime {0.0} ", str, "} ", (char *)NULL);

	if (objPtr->filename == NULL) {
	    Tcl_AppendResult(interp, "{-filename {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-filename {} {", objPtr->filename,
			    "}} ",(char *)NULL);
	}
	
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}

	Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	Tcl_AppendResult(interp, "{-maxFrameRate {0.0} ", str,
			  "} ", (char *)NULL);

	Tcl_PrintDouble(interp, objPtr->clipStart, str);
	Tcl_AppendResult(interp, "{-clipStart 0.0 ", str, "} ", (char *) NULL);

	Tcl_PrintDouble(interp, objPtr->clipEnd, str);
	Tcl_AppendResult(interp, "{-clipEnd 0.0 ", str, "} ", (char *) NULL);

	Tcl_PrintDouble(interp, objPtr->logicalStart, str);
	Tcl_AppendResult(interp, "{-logicalStart 0.0 ", str, "} ", (char *) NULL);

	Tcl_PrintDouble(interp, objPtr->logicalEnd, str);
	Tcl_AppendResult(interp, "{-logicalEnd 0.0 ", str, "} ", (char *) NULL);
	
	return TCL_OK;
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->lts == NULL) {
		Tcl_AppendResult(interp, "-lts {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName, "} ",
				  (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->sendAhead, str);
	    Tcl_AppendResult(interp, "-sendAhead {0.2} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'b') && (strncmp("-bandwidth", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->bandwidth);
	    Tcl_AppendResult(interp, "-bandwidth {0} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->cycleTime, str);
	    Tcl_AppendResult(interp, "-cycleTime {0.0} ", str, (char *)NULL);
	    return TCL_OK;
	} 
	if ((c == 'f') && (strncmp("-filename", argv[0], len) == 0)) {
	    if (objPtr->filename == NULL) {
		Tcl_AppendResult(interp, "-filename {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-filename {} {",
			  objPtr->filename, "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	} 
	if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
	    Tcl_AppendResult(interp, "-maxFrameRate {0.0} ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'c') && (strncmp("-clipStart", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->clipStart, str);
	    Tcl_AppendResult(interp, "-clipStart 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'c') && (strncmp("-clipEnd", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->clipEnd, str);
	    Tcl_AppendResult(interp, "-clipEnd 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'l') && (strncmp("-logicalStart", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->logicalStart, str);
	    Tcl_AppendResult(interp, "-logicalStart 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'l') && (strncmp("-logicalEnd", argv[0], len) == 0)) {
	    Tcl_PrintDouble(interp, objPtr->logicalEnd, str);
	    Tcl_AppendResult(interp, "-logicalEnd 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}


	Tcl_AppendResult(interp, "bad configuration option: ",
			  argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	/* ------------------------- LTS ------------------------- */
	if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newlts;
	    
	    /*
	     * Check args
	     */
	    cntr++;
	    if (cntr == argc) {
argcError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }

	    /*
	     * If resetting to null value, just free up old lts.
	     */
	    if (strlen(argv[cntr]) == 0) {
		if (objPtr->lts != NULL) {
		    if (objPtr->ltsTriggerNum != -1) {
			LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
			objPtr->ltsTriggerNum = -1;
		    }
		    Tk_Release(objPtr->lts);
		    objPtr->lts = NULL;
		}
		if (objPtr->ltsName != NULL) {
		    ckfree((char *) objPtr->ltsName);
		    objPtr->ltsName = NULL;
		}
	    } else {
		newlts = (ClientData)Tcm_FindAtom(interp, argv[cntr]);
		if (newlts == NULL) {
valueError:	    
		    Tcl_AppendResult(interp, "bad value \"", argv[cntr],
				  "\" given for configuration option ",
				  argv[cntr-1], (char *)NULL);
		    return TCL_ERROR;
		}
		
		/*
		 * Blow away old callback.
		 */
		if (objPtr->lts != NULL) {
		    if (objPtr->ltsTriggerNum != -1) {
			LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
			objPtr->ltsTriggerNum = -1;
		    }
		    Tk_Release(objPtr->lts);
		    objPtr->lts = NULL;
		}
		objPtr->lts = newlts;
		Tk_Preserve (objPtr->lts);

		/*
		 * Copy in new lts name
		 */
		if (objPtr->ltsName != NULL) {
		    ckfree((char *) objPtr->ltsName);
		}
		objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
		strcpy(objPtr->ltsName, argv[cntr]);
		
		/*
		 * Set up new lts trigger
		 */
		objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts,
				LTS_SPEED_FLAG | LTS_OFFSET_FLAG | LTS_VALUE_FLAG,
			        MPEGSegmentLTSChanged, (ClientData) objPtr);
	    }
	    MPEGSegmentUnready(objPtr);

	/* ------------------------- SEND_AHEAD ------------------------- */
	} else if ((c=='s') && (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	    double newSendAhead;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->sendAhead = newSendAhead;

	/* ------------------------- BANDWIDTH -------------------------- */
	} else if ((c=='b') && (strncmp("-bandwidth", argv[cntr], len) == 0)) {
	    int newBandwidth;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newBandwidth) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->bandwidth = newBandwidth;
	    objPtr->extraBWCounter = 0;

	/* ------------------------- CYCLE_TIME ------------------------- */
	} else if ((c=='c') && (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	    double newCycleTime;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->cycleTime = newCycleTime;

	/* ------------------------- FILE_NAME -------------------------- */
	} else if ((c=='f') && (strncmp("-filename", argv[cntr], len) == 0)) {
	    Tcl_DString fname_dstring;
	    char *fname;

	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (objPtr->filename != NULL) {
		Tcl_AppendResult(interp,
				"Can not set clipfile name more than once.",
				(char *)NULL);
		return TCL_ERROR;
	    }

	    fname = Tcl_TildeSubst(objPtr->interp, argv[cntr], &fname_dstring);

	    objPtr->fip = MPEGSegmentOpenClipfile(objPtr, fname);

	    if (objPtr->fip == NULL) {
		Tcl_DStringFree(&fname_dstring);
		Tcl_AppendResult(interp, "Couldn't open file \"",
				 fname, "\" as mpeg file",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    objPtr->filename = (char *) ckalloc(strlen(fname) + 1);
	    strcpy(objPtr->filename, fname);
	    Tcl_DStringFree(&fname_dstring);

	/* ------------------------- OUT_CMD ---------------------------- */
	} else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    
	    if (MPEGSegmentChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	    MPEGSegmentUnready(objPtr);
	    
	/* ------------------------- MAX_FRAME_RATE --------------------- */
	} else if ((c=='m') &&
		   (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
	    double newMaxFrameRate;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
		goto valueError;
	    }
	    objPtr->maxFrameRate = newMaxFrameRate;
	    objPtr->extraFrameCounter = 0.0;

	/* ------------------------- CLIP_START --------------------- */
	} else if ((c=='c') &&
		   (strncmp("-clipStart", argv[cntr], len) == 0)) {
	    double newClipStart;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (argv[cntr][0] == 'e') {
		if (objPtr->fip != NULL) {
		    objPtr->clipStart = MPEG_FIP_ClipLength(objPtr->fip);
		}
	    } else if (argv[cntr][0] == '?') {
		objPtr->clipStart = objPtr->clipEnd - 
		  (objPtr->logicalEnd - objPtr->logicalStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newClipStart) != TCL_OK) {
		    goto valueError;
		}
		objPtr->clipStart = newClipStart;
	    }

	    MPEGSegmentUnready(objPtr);

	/* ------------------------- CLIP_END --------------------- */
	} else if ((c=='c') &&
		   (strncmp("-clipEnd", argv[cntr], len) == 0)) {
	    double newClipEnd;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (argv[cntr][0] == 'e') {
		if (objPtr->fip != NULL) {
		    objPtr->clipEnd = MPEG_FIP_ClipLength(objPtr->fip);
		}
	    } else if (argv[cntr][0] == '?') {
		objPtr->clipEnd = objPtr->clipStart + 
		  (objPtr->logicalEnd - objPtr->logicalStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newClipEnd) != TCL_OK) {
		    goto valueError;
		}
		objPtr->clipEnd = newClipEnd;
	    }

	    MPEGSegmentUnready(objPtr);

	/* ------------------------- LOGICAL_START --------------------- */
	} else if ((c=='l') &&
		   (strncmp("-logicalStart", argv[cntr], len) == 0)) {
	    double newLogicalStart;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    if (argv[cntr][0] == '?') {
		objPtr->logicalStart = objPtr->logicalEnd - 
		  fabs(objPtr->clipEnd - objPtr->clipStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newLogicalStart) != TCL_OK) {
		    goto valueError;
		}
		objPtr->logicalStart = newLogicalStart;
	    }

	    MPEGSegmentUnready(objPtr);

	/* ------------------------- LOGICAL_END --------------------- */
	} else if ((c=='l') &&
		   (strncmp("-logicalEnd", argv[cntr], len) == 0)) {
	    double newLogicalEnd;
	    
	    cntr++;
	    if (cntr == argc) {
		goto argcError;
	    }
	    
	    if (argv[cntr][0] == '?') {
		objPtr->logicalEnd = objPtr->logicalStart +
		  fabs(objPtr->clipEnd - objPtr->clipStart);
	    } else {
		if (Tcl_GetDouble(interp, argv[cntr], &newLogicalEnd) != TCL_OK) {
		    goto valueError;
		}
		objPtr->logicalEnd = newLogicalEnd;
	    }

	    MPEGSegmentUnready(objPtr);

	/* ------------------------- UNKNOWN OPTION ----------------------*/
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			  argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}

/************************************************************
 *
 * MPEGSegmentCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the mpegSegment command.
 *     A new object of type MPEGSegment is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
MPEGSegmentCreateCmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char **argv)
{
    char new_name[256];
    MPEGSegment *new;
    
    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextMPEGSegmentNum++;
    sprintf(new_name, "%smpegSeg%d", argv[1], nextMPEGSegmentNum);
    
    new = CreateNewMPEGSegment(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MPEGSegmentObjCmd,
		      (ClientData) new, (void (*)()) NULL);
    
    if (argc > 2) {
	ConfigureMPEGSegment(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MPEGSegmentObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     MPEGSegment object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
MPEGSegmentObjCmd(ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv)
{
    char c;
    int len, i;
    MPEGSegment *objPtr = (MPEGSegment *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    /*------------------------- CONFIGURE ------------------------------*/
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureMPEGSegment(objPtr, argc - 2, argv + 2);

    /*------------------------- DESTROY --------------------------------*/
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyMPEGSegment(objPtr);

    /*------------------------- INFO -----------------------------------*/
    } else if ((c == 'i') && (strncmp("info", argv[1], len) == 0)) {
	double start_param, end_param;
	
	start_param = 0.0;
	end_param = 0.0;
	
	if (argc > 2) {
	    if (Tcl_GetDouble(interp, argv[2], &start_param) != TCL_OK) {
		Tcl_AppendResult(interp, "Bad value: \"", argv[2], "\" given",
				 " as argument to: ", argv[0], " ", argv[1],
				 (char *) NULL);
	    }
	    if (argc > 3) {
		if (Tcl_GetDouble(interp, argv[3], &end_param) != TCL_OK) {
		    Tcl_AppendResult(interp, "Bad value: \"", argv[3], 
				     "\" given as argument to: ", 
				     argv[0], " ", argv[1],
				     (char *) NULL);
		}
	    } else {
		end_param = start_param;
	    }
	} else {
	    start_param = objPtr->clipStart;
	    end_param = objPtr->clipEnd;
	}
	return MPEGSegmentInfo(objPtr, start_param, end_param);
	
    /*------------------------- READY ----------------------------------*/
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return MPEGSegmentReady(objPtr);

    /*------------------------- UNREADY --------------------------------*/
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return MPEGSegmentUnready(objPtr);
    /*------------------------- WIDTH ----------------------------------*/
    } else if ((c == 'w') && (strncmp("width", argv[1], len) == 0)) {
	if (objPtr->fip == NULL) {
	    Tcl_SetResult(interp, "0", TCL_VOLATILE);
	    return TCL_OK;
	} else {
	    char retStr[100];
	    sprintf(retStr, "%d", MPEG_FIP_Width(objPtr->fip));
	    Tcl_SetResult(interp, retStr, TCL_VOLATILE);
	    return TCL_OK;
	}
    /*------------------------- HEIGHT ----------------------------------*/
    } else if ((c == 'h') && (strncmp("height", argv[1], len) == 0)) {
	if (objPtr->fip == NULL) {
	    Tcl_SetResult(interp, "0", TCL_VOLATILE);
	    return TCL_OK;
	} else {
	    char retStr[100];
	    sprintf(retStr, "%d", MPEG_FIP_Height(objPtr->fip));
	    Tcl_SetResult(interp, retStr, TCL_VOLATILE);
	    return TCL_OK;
	}
    }

    Tcl_AppendResult(interp, "Bad command option: ",
		argv[0], " ", argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  ready\n",
                "  unready\n",
                "  info ?clip_time ?clip_time??\n",
		"  width\n",
		"  height\n",
                (char *) NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * MPEGSegmentModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
MPEGSegmentModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mpegSegment", MPEGSegmentCreateCmd, NULL,
		(void (*)()) NULL);
    Tcl_GlobalEval(interp, "cminfo add object_info mpegSegment mpeg mpeg-1");
    return TCL_OK;
}


/************************************************************
 *
 * MPEGSegmentReady
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Gets the object ready and starts sending data.
 ************************************************************/

static int
MPEGSegmentReady(MPEGSegment * objPtr)
{
    double clip_length;

    if (objPtr->ready) {
	return TCL_OK;
    }

    if (objPtr->fip == NULL) {
	Tcl_AppendResult(objPtr->interp, "Must specify a filename before readying this",
			 " object", (char *) NULL);
	return TCL_ERROR;
    }


    MPEG_FIP_BitArrayClear(objPtr->fip->bit_array);

    /*
     * Blow away the old stuff.
     */

    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
	objPtr->at_token = 0;
    }

    /*
     * Check for clip start, end and logical start, end 
     * consistency, etc.
     */

    if (objPtr->logicalEnd - objPtr->logicalStart <= 0.0) {
	Tcl_AppendResult(objPtr->interp, "Logical end time must be greater than ",
			 "logical start time.", (char *) NULL);
	return TCL_ERROR;
    }

    if (objPtr->clipStart < 0.0) {
	objPtr->clipStart = 0.0;
    }
    if (objPtr->clipEnd < 0.0) {
	objPtr->clipEnd = 0.0;
    }
    
    clip_length = MPEG_FIP_ClipLength(objPtr->fip);

    if (objPtr->clipStart > clip_length) {
	objPtr->clipStart = clip_length;
    }
    if (objPtr->clipEnd > clip_length) {
	objPtr->clipEnd = clip_length;
    }

    if (objPtr->clipEnd - objPtr->clipStart == 0.0) {
	Tcl_AppendResult(objPtr->interp, "Clip start and end times specified create ",
			 "empty segment.", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Set the ready flag.
     */
    objPtr->ready = 1;

    if (objPtr->lts != NULL) {
	MPEGSegmentSendData((ClientData) objPtr);
    }
    return TCL_OK;
}

/************************************************************
 *
 * MPEGSegmentUnready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Stops sending data and puts the object in the unready state.
 ************************************************************/

static int
MPEGSegmentUnready(MPEGSegment * objPtr)
{
    if (!objPtr->ready) return;

    if (objPtr->at_token != 0) {
	Tcm_DeleteAtHandler(objPtr->at_token);
	objPtr->at_token = 0;
    }
    objPtr->lastFrameNum = -1;

    objPtr->ready = 0;

    return TCL_OK;
}

/************************************************************
 *
 * MPEGSegmentLTSChanged
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Procedure called whenever the LTS changes.
 ************************************************************/

void
MPEGSegmentLTSChanged(ClientData clientdata)
{
    MPEGSegment *objPtr = (MPEGSegment *) clientdata;

    objPtr->extraFrameCounter = 0.0;
    objPtr->extraBWCounter = 0;

    if (objPtr->ready == 1) {
	MPEGSegmentUnready(objPtr);
	MPEGSegmentReady(objPtr);
    }
    return;
}


/************************************************************
 *
 * MPEGSegmentChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
MPEGSegmentChangeOutCmd(MPEGSegment *objPtr, char *newOut) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree((char *) objPtr->outCmd);
    }
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

#ifdef SECURE_CMT
    if (newOut[0] != '@') {
	Tcl_SetResult(objPtr->interp, "Cannot set Tcl outCmd because this program was compiled with -DSECURE_CMT.\nRecompile without security to use arbitrary Tcl outCmd's\n", TCL_STATIC);
	return TCL_ERROR;
    }
#endif

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
		&(objPtr->outCD), (ClientData *)&(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    

/************************************************************
 *
 * MPEGSegmentSendData
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sends data according to current value of LTS, queues
 *     itself on the at queue for next send time.
 ************************************************************/

void
MPEGSegmentSendData(ClientData clientData) 
{
    MPEGSegment *objPtr = (MPEGSegment *) clientData;
    double now, speed, when;
    double eTime;
    double clip_now, clip_cycle_end; 
    int i, curBw;
    double frac_want_to_send; 
    int want_to_send; 
    int bw_want_to_send; 
    SB_List *sbl; 
    char *sbStr; 
    int numToSend, bwToSend, origNumToSend;
    int *framelist; 
    static FramePick *fPickList; 
    static int fPickNum = 0; 
    int *orderVec; 
    int pick_idx; 
    int reverse_flag;

    /*
     * Return if not ready, no LTS, or no out command.
     */
    if ((!objPtr->ready) ||
	(objPtr->lts == NULL) ||
	(objPtr->outCmd == NULL)) {
	goto done;
    }

    /* 
     * Create some fPick structures to begin with. 
     */
    
    if (fPickNum == 0) {
	fPickList = (FramePick *) ckalloc(sizeof(FramePick)*100);
	fPickNum = 100;
    }	

    /*
     * Find out if we should be sending data or going to sleep.
     * If going to sleep, for how long.
     */

    now = LTS_Value(objPtr->lts);
    speed = LTS_Speed(objPtr->lts);
    now += objPtr->sendAhead * speed;

    /* Slightly different for forwards and backwards. */

    if (speed > 0.0) {
	if (objPtr->logicalEnd <= now) {
	    /* Sleep forever case. */
	    return;
	} else if (objPtr->logicalStart > now) {
	    /* Sleep until expected to be here. */

	    when = LTS_SystemTimeOf(objPtr->lts,
			(objPtr->logicalStart-(objPtr->sendAhead*speed)));
	    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				       "video", MPEGSegmentSendData, objPtr, NULL,
				       NULL, NULL, NULL, NULL);
	    return;
	}
    } else if (speed < 0.0) {
	if (objPtr->logicalStart >= now) {
	    /* Sleep forever case. */
	    return;
	} else if (objPtr->logicalEnd < now) {
	    /* Sleep until expected to be here. */
	    
	    when = LTS_SystemTimeOf(objPtr->lts,
				    objPtr->logicalEnd+(objPtr->sendAhead*speed));
	    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
				     "video", MPEGSegmentSendData, objPtr, NULL,
				     NULL, NULL, NULL, NULL);
	  return;

	}
    } else {
	if (!((objPtr->logicalStart <= now) && 
	      (objPtr->logicalEnd > now)))  {
	    /* Sleep forever. */
	    MPEG_FIP_FDClose(objPtr->fip);
	    return;
	}
    }

    /* 
     * Currently active, need to send data. 
     * Tell file interface to be open (i.e., 
     */

    MPEG_FIP_FDOpen(objPtr->fip);


    /*
     * Calculate revers flag. We are playing in reverse if the segment is
     * mapped forward but speed is backward, or if segment is mapped backward
     * and speed is forward.
     */

    if (((objPtr->clipStart < objPtr->clipEnd) && (speed < 0.0)) ||
	((objPtr->clipStart > objPtr->clipEnd) && (speed > 0.0))) {
	reverse_flag = 1;
    } else {
	reverse_flag = 0;
    }

    /*
     * Calculate time in terms of clip time.
     */

    clip_now = CLIP_TIME_FROM_LOG_TIME(objPtr, now);

    /*
     * If only sending one frame because speed = 0.0 or
     * cycle time = 0.0, do this now. 
     */
    if ((objPtr->cycleTime == 0.0) || (speed == 0.0)) {
	
	/*
	 * Find frames to send.
	 * Could be more than one frame since reference frames may be needed.
	 */

	numToSend = MPEG_FIP_FrameListFromTime(objPtr->fip, clip_now, clip_now,
					       reverse_flag,
					       &framelist, &eTime);

	if (numToSend > fPickNum) {
	    fPickNum = numToSend+1;
	    fPickList = (FramePick *) ckrealloc((char *) fPickList,
						sizeof(FramePick)*fPickNum);
	}

	numToSend = MPEGSegmentBuildFPickList(objPtr, framelist, fPickList, numToSend);
	ckfree((char *) framelist);

	if (numToSend != 0) {
	    orderVec = (int *)  ckalloc(sizeof(int) * numToSend);

	    for (i=0; i<numToSend; i++) {
		orderVec[i] = i;
	    }

	    sbl = MPEGSegmentMakeSBLFromFPickList(objPtr,
						  clip_now, speed, reverse_flag,
						  fPickList, numToSend, 
						  numToSend, orderVec);

	    ckfree((char *) orderVec);

	    /*
	     * If out command is C, create scatter buffer list and call.
	     */
	    if (objPtr->outCCmd != NULL) {
		/*
		 * Need to check if it still exists since object may have
		 * been destroyed.
		 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		    objPtr->outCCmd(objPtr->outCD, sbl);	       
		} else {
		    /*
		     * Out C command has unexpectedly disappeared.
		     */
		    objPtr->outCCmd = NULL;
		    ckfree((char *) objPtr->outCmd);
		    objPtr->outCmd = NULL;
		}
	    } else {
		/*
		 * Out command is tcl, make sb list into a string and 
		 * call using Tcl_Eval. 
		 */
		sbStr = MakeSBListString(sbl);
		Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			    sbStr, (char *) NULL); 
		ckfree((char *) sbStr);
	    }
	    DetachToAllBuffers(objPtr->interp, sbl, 
			       "tcmMPEGFileSegment:MPEGSegmentSendData");
	    FreeSBList(sbl);
	}

	/*
	 * Figure out when we need to send again, queue it up and return.
	 */
	if (speed == 0.0) {
	    return;
	}
	
	eTime = LOG_TIME_FROM_CLIP_TIME(objPtr, eTime);
	when = LTS_SystemTimeOf(objPtr->lts,
				(eTime-(objPtr->sendAhead*speed)));

	objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
					       "video", 
					       MPEGSegmentSendData, objPtr, 
					       NULL, NULL, NULL, NULL, NULL);
	return;
    } 
    
    /*
     * Cycle time is not 0.0, so send multiple frames.
     * Find frames to be sent.
     */
    numToSend = 0;
    bwToSend = 0;

    if (speed > 0.0) {
      if (now+objPtr->cycleTime*speed > objPtr->logicalEnd) {
	clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, objPtr->logicalEnd);
      } else {
	clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, now+objPtr->cycleTime*speed);
      }
    } else {
      if (now+objPtr->cycleTime*speed < objPtr->logicalStart) {
	clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, objPtr->logicalStart);
      } else {
	clip_cycle_end = CLIP_TIME_FROM_LOG_TIME(objPtr, now+objPtr->cycleTime*speed);
      }
    }

    numToSend = MPEG_FIP_FrameListFromTime(objPtr->fip, clip_now, 
					    clip_cycle_end, reverse_flag,
					   &framelist, &eTime);

    if (numToSend > fPickNum) {
	fPickNum = numToSend+1;
	fPickList = (FramePick *) ckrealloc((char *) fPickList,
					    sizeof(FramePick)*fPickNum);
    }

    numToSend = MPEGSegmentBuildFPickList(objPtr, framelist, fPickList, numToSend);
    ckfree((char *) framelist);

    if (numToSend != 0) {
	/*
	 * If number of frames to be sent greater than the number of
	 * frames we can send (due to bandwidth or max frame rate),
	 * mark low priority frames so we don't bother reading them.
	 */
	frac_want_to_send = (objPtr->maxFrameRate * objPtr->cycleTime) -
	  objPtr->extraFrameCounter;
	want_to_send = (int) (ceil(frac_want_to_send));

	if (objPtr->maxFrameRate != 0.0) {
	    objPtr->extraFrameCounter = ((double) want_to_send) - frac_want_to_send;
	}

	bw_want_to_send = objPtr->bandwidth - ((int) ((((double) objPtr->extraBWCounter) /
						       objPtr->cycleTime)));
	curBw = (int) (((double) bwToSend) / objPtr->cycleTime);

	origNumToSend = numToSend;

	orderVec = (int *) ckalloc(sizeof(int)*numToSend);
	MPEGSegmentPrioritize (numToSend, fPickList, orderVec);

	if ((numToSend > 0) &&
	    (((numToSend > want_to_send) && objPtr->maxFrameRate) ||
	     ((curBw  > bw_want_to_send)    && objPtr->bandwidth &&
	      ((numToSend > 1) || (objPtr->extraBWCounter > 0))))) {
	    
	    while ((numToSend > 0) &&
		   (((numToSend > want_to_send) && objPtr->maxFrameRate) ||
		    ((curBw  > bw_want_to_send)    && objPtr->bandwidth &&
		     ((numToSend > 1) || (objPtr->extraBWCounter > 0))))) {
		numToSend--;
		pick_idx = orderVec[numToSend];
		
		fPickList[pick_idx].sendFlag = 0;
		bwToSend -= fPickList[pick_idx].size;
		curBw  = (int) (((double) bwToSend) / objPtr->cycleTime);
	    }
	}
	
	/*
	 * Adjust bandwidth counter.
	 */
	
	if (objPtr->bandwidth > 0) {
	    objPtr->extraBWCounter += bwToSend - ((int) (objPtr->bandwidth * 
							 objPtr->cycleTime));
	}
    }

    if (numToSend > 0) {

	/*
	 * Create SBL for data going out.
	 */
	
	sbl = MPEGSegmentMakeSBLFromFPickList(objPtr, clip_now, speed, reverse_flag,
					      fPickList, origNumToSend, numToSend, 
					      orderVec);
	
	/*
	 * If out command is C, create scatter buffer list and call.
	 */
	if (objPtr->outCCmd != NULL) {
	    
	    
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree((char *) objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and
	     * call using Tcl_Eval. 
	     */
	    sbStr = MakeSBListString(sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree((char *) sbStr);
	}
	DetachToAllBuffers(objPtr->interp, sbl, 
			   "cmMPEGFileSegment:MPEGSegementSendData");
	FreeSBList (sbl);
    }
    
    /*
     * Find next wake up time and set up atqueue.
     */
    
    eTime = LOG_TIME_FROM_CLIP_TIME(objPtr, eTime);
    when = LTS_SystemTimeOf(objPtr->lts,
		    (eTime-(objPtr->sendAhead*speed)));
    objPtr->at_token = Tcm_CreateAtHandler(when, HUGE_VAL,
			       "video", MPEGSegmentSendData, objPtr, NULL,
			       NULL, NULL, NULL, NULL);
    return;

done:
    MPEG_FIP_FDClose(objPtr->fip);
    return;
}


/**************************************************************
 * MPEGSegmentOpenClipfile
 *
 * Parameters:
 *    objPtr - pointer to Segment object.
 *    filename - character string file name.
 * 
 * Results:
 *    Returns a MPEGFileInterface pointer valid for filename on success,
 *    NULL on failure.
 * 
 * Actions:
 *    Interfaces with MPEG File Interface module to get access to file.
 *    Queries the module in order to fill the cfInfo structure in the object.
 *    
 ******************************************************************/

static MPEGFileInterface *
MPEGSegmentOpenClipfile(MPEGSegment *objPtr, char *filename)
{
    MPEGFileInterface *fip;

    fip = MPEG_FIP_Open(filename);

    if (fip == NULL) 
      return NULL;

    objPtr->cfInfo = (MPEGHeader *) ckalloc(sizeof(MPEGHeader));

    objPtr->cfInfo->width = MPEG_FIP_Width(fip);
    objPtr->cfInfo->height = MPEG_FIP_Height(fip);
    objPtr->cfInfo->aspectRatioCode = MPEG_FIP_AspectRatioCode(fip);
    objPtr->cfInfo->frameRateCode = MPEG_FIP_FrameRateCode(fip);
    objPtr->cfInfo->bitRate = MPEG_FIP_BitRate(fip);
    objPtr->cfInfo->vbvBufferSize = MPEG_FIP_VBVBufferSize(fip);
    objPtr->cfInfo->constrainedParamFlag = MPEG_FIP_ConstrainedParamFlag(fip);

    return fip;
}  

/*************************************************************
 * MPEGSegmentInfo
 * 
 * Parameters:
 *     objPtr - pointer to MPEG Segment object.
 *     start - start time of segment that info is needed.
 *     end - end time of segment that info is needed.
 * 
 * Result:
 *     Standard tcl result.
 *
 * Action:
 *     Returns in result a list of five numbers: numFrames in segment, 
 *     maxSize of those frames, number of I frames, number of P frames,
 *     and number of B frames.
 *
 *************************************************************/

static int 
MPEGSegmentInfo(MPEGSegment *objPtr, double start, double end)
{
    int numFrames;
    int maxSize;
    char retStr[500];
    int numI, numP, numB;

    if (objPtr->fip == NULL) {
	sprintf(retStr, "%d %d %d %d %d", 0, 0, 0, 0, 0);
	Tcl_SetResult(objPtr->interp, retStr, TCL_VOLATILE);
	return TCL_OK;
    }

    MPEG_FIP_Info(objPtr->fip, start, end, &numFrames, &maxSize, &numI, &numP, &numB);

    sprintf(retStr, "%d %d %d %d %d", numFrames, maxSize, numI, numP, numB);
    Tcl_SetResult(objPtr->interp, retStr, TCL_VOLATILE);
    return TCL_OK;
}

/*************************************************************
 * MPEGSegmentBuildFPickList
 * 
 * Parameters:
 *     objPtr - pointer to MPEG Segment object.
 *     framelist - integer array of frame numbers to build FramePick
 *                 structures for.
 *     fPickList - array of FramePick structures to fill.
 *     numFrames - number of frames in framelist.
 * 
 * Result:
 *     Number of frames in fPickList.
 *
 * Action:
 *     Fills FrameList structure for each frame specified in framelist.
 *     Length of fPickList array is assumed to be equal to or greater
 *     than numFrames. Checks to make sure that frame should be sent
 *     by checking dynamic bit array. It is possible that all the frames
 *     in the list have already been sent, so the caller should be able
 *     to deal with a 0 return value indicating that the list is empty.
 *
 *************************************************************/

static int
MPEGSegmentBuildFPickList (MPEGSegment *objPtr, 
			   int *framelist,
			   FramePick *fPickList,
			   int numFrames)
{
    int i;
    int frame_num;
    int size, ref1, ref2, type;
    double sTime, eTime;
    int listLength = 0;

    
    for (i=0; i<numFrames; i++) {
	
	frame_num = framelist[i];

	if (!MPEG_FIP_BitArrayGet(objPtr->fip->bit_array, frame_num)) {
	    MPEG_FIP_GetAllFrameInfo(objPtr->fip, frame_num, &size, 
				     &sTime, &eTime, 
				     &ref1, &ref2, &type);
	
	    fPickList[listLength].bufPtr = 0;
	    fPickList[listLength].frame_num = frame_num;
	    fPickList[listLength].index = (int) ((LOG_TIME_FROM_CLIP_TIME(objPtr, 
								((sTime + eTime)
								/ 2.0)))
						 * 1000.0);
	    fPickList[listLength].size = size;
	    fPickList[listLength].sTime = sTime;
	    fPickList[listLength].eTime = eTime;
	    fPickList[listLength].ref1 = ref1;
	    fPickList[listLength].ref2 = ref2;
	    fPickList[listLength].type = type;
	    fPickList[listLength].sendFlag = 1;

	    listLength++;
	}
    }

    return listLength;
}
   
/*************************************************************
 * MPEGSegmentMakeSBLFromFPickList
 * 
 * Parameters:
 *     objPtr - pointer to MPEG Segment object.
 *     start - clip start time of interval represented by these frames.
 *     end - clip end time of interval represented by these frames.
 *     speed - speed of lts.
 *     fPickList - array of FramePick structures.
 *     origNumFrames - number of frames total in fPickList array.
 *     numFrames - number of frames in fPickList array to be sent.
 *     orderVec - prioritization vector.
 *     etimePtr - pointer to double. Used to return last end time of
 *                frames in array. Used to calculate next wake up.
 * 
 * Result:
 *     Pointer to scatter buffer list.
 *
 * Action:
 *     Creates a properly formatted SBL that can be used to send
 *     frame data with all the correct headers on to the next
 *     object. Also detects the last end time. This is a bit trickier
 *     than it might seem.
 *
 *************************************************************/

static SB_List *
MPEGSegmentMakeSBLFromFPickList(MPEGSegment *objPtr,
				double start,
				double speed,
				int reverseFlag,
				FramePick *fPickList,
				int origNumFrames,
				int numFrames,
				int *orderVec)
{
    SB_List *sbl;
    ScatterBuffer *sb;
    int i;
    MPEGBufferHeader mpegBufHdr;
    Buffer *frame_buf, *intra_buf, *non_intra_buf;
    int intra_len, non_intra_len;
    int data_seg_type, data_seg_length;
    int frame_len;
    double rsTime, reTime;
    int sb_cntr, j;
    int pick;

    sbl = (SB_List *) ckalloc(sizeof(SB_List));

    sbl->count = numFrames;
    sbl->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *)*numFrames);
    
    sb_cntr = 0;

    for (i=0; i<origNumFrames; i++) {

	pick = orderVec[i];

	if (fPickList[pick].sendFlag) {

	    MPEG_FIP_BitArraySet(objPtr->fip->bit_array, 
				 fPickList[pick].frame_num);
	    
	    mpegBufHdr.hdrType = CM_MPEG_BUF_HDR;
	    mpegBufHdr.length = MPEG_BUF_HDR_SIZE;
	    mpegBufHdr.width = objPtr->cfInfo->width;
	    mpegBufHdr.height = objPtr->cfInfo->height;
	    mpegBufHdr.type = fPickList[pick].type;
	    if (mpegBufHdr.type != MPEG_I_FRAME) {
		double rs, re;

		MPEG_FIP_TimeFromFrame(objPtr->fip, fPickList[pick].ref1, &rs, &re);
		mpegBufHdr.ref1 = (int) (LOG_TIME_FROM_CLIP_TIME(objPtr, ((rs+re) / 2.0)) * 1000.0);
	    } else {
		mpegBufHdr.ref1 = -1;
	    }
	    if (mpegBufHdr.type == MPEG_B_FRAME) {
		double rs, re;
		MPEG_FIP_TimeFromFrame(objPtr->fip, fPickList[pick].ref2, &rs, &re);
		mpegBufHdr.ref2 = (int) (LOG_TIME_FROM_CLIP_TIME(objPtr, ((rs+re) / 2.0)) * 1000.0);
	    } else {
		mpegBufHdr.ref2 = -1;
	    }
	    mpegBufHdr.index = fPickList[pick].index;
	    
	    mpegBufHdr.displayFlag = 1;

	    if (speed == 0.0) {
		fPickList[pick].sTime = fPickList[pick].eTime = start;
		if (i < numFrames-1) 
		  mpegBufHdr.displayFlag = 0;
	    } else {
		
		if (reverseFlag) {
		    double tmp_time;
		    
		    tmp_time = fPickList[pick].sTime;
		    fPickList[pick].sTime = fPickList[pick].eTime;
		    fPickList[pick].eTime = tmp_time;
		} 
		
		if (mpegBufHdr.type != MPEG_B_FRAME) {
		    
		    if (!reverseFlag) {
			MPEG_FIP_TimeFromFrame(objPtr->fip, fPickList[pick].ref2, &rsTime, &reTime);
			fPickList[pick].eTime = rsTime;
		    } else  {
			MPEG_FIP_TimeFromFrame(objPtr->fip, fPickList[pick].ref1, &rsTime, &reTime);
			fPickList[pick].eTime = reTime;
		    }
		}
	    }

	    fPickList[pick].sTime = LOG_TIME_FROM_CLIP_TIME(objPtr, fPickList[pick].sTime);
	    fPickList[pick].eTime = LOG_TIME_FROM_CLIP_TIME(objPtr, fPickList[pick].eTime);

	    if (fPickList[pick].sTime < fPickList[pick].eTime) {
	      if (fPickList[pick].sTime < objPtr->logicalStart) {
		if (fPickList[pick].eTime <= objPtr->logicalStart) {
		  mpegBufHdr.displayFlag = 0;
		} else {
		  fPickList[pick].sTime = objPtr->logicalStart;
		}
	      }
	      if (fPickList[pick].eTime > objPtr->logicalEnd) {
		if (fPickList[pick].sTime >= objPtr->logicalEnd) {
		  mpegBufHdr.displayFlag = 0;
		} else {
		  fPickList[pick].eTime = objPtr->logicalEnd;
		}
	      }
	    } else {
	      if (fPickList[pick].sTime > objPtr->logicalEnd) {
		if (fPickList[pick].eTime >= objPtr->logicalEnd) {
		  mpegBufHdr.displayFlag = 0;
		} else {
		  fPickList[pick].sTime = objPtr->logicalEnd;
		}
	      }
	      if (fPickList[pick].eTime < objPtr->logicalStart) {
		if (fPickList[pick].sTime <= objPtr->logicalStart) {
		  mpegBufHdr.displayFlag = 0;
		} else {
		  fPickList[pick].eTime = objPtr->logicalStart;
		}
	      }
	    }
	  
	    
	    mpegBufHdr.logStart_s = floor(fPickList[pick].sTime);
	    mpegBufHdr.logStart_ns = floor((fPickList[pick].sTime - 
					    floor(fPickList[pick].sTime)) * ONE_BILLION);
	    
	    mpegBufHdr.logEnd_s = floor(fPickList[pick].eTime);
	    mpegBufHdr.logEnd_ns = floor((fPickList[pick].eTime - 
					  floor(fPickList[pick].eTime)) * ONE_BILLION);
	    
	    intra_buf = MPEG_FIP_CreateIntraMatrixBuffer(objPtr->interp,
							 objPtr->fip,
							fPickList[pick].frame_num,
							 &intra_len);
	    if (intra_buf == NULL) {
		mpegBufHdr.intra_matrix_flag = 0;
	    } else {
		mpegBufHdr.intra_matrix_flag = 1;
	    }
	    
	    non_intra_buf = MPEG_FIP_CreateNonIntraMatrixBuffer(objPtr->interp,
								objPtr->fip,
								fPickList[pick].frame_num,
								&non_intra_len);
	    if (non_intra_buf == NULL) {
		mpegBufHdr.non_intra_matrix_flag = 0;
	    } else {
		mpegBufHdr.non_intra_matrix_flag = 1;
	    }
	    
	    frame_buf = MakeBuffer(objPtr->interp, 8 + MPEG_BUF_HDR_SIZE + 8 + 
				   fPickList[pick].size, "FrameBuffer");
	
	    if (frame_buf == NULL) goto error;
	    
	    mpegBufHdr.hdrType = htonl(mpegBufHdr.hdrType);
	    mpegBufHdr.length = htonl(mpegBufHdr.length);
	    mpegBufHdr.width = htonl(mpegBufHdr.width);
	    mpegBufHdr.height = htonl(mpegBufHdr.height);
	    mpegBufHdr.type = htonl(mpegBufHdr.type);
	    mpegBufHdr.ref1 = htonl(mpegBufHdr.ref1);
	    mpegBufHdr.ref2 = htonl(mpegBufHdr.ref2);
	    mpegBufHdr.intra_matrix_flag = htonl(mpegBufHdr.intra_matrix_flag);
	    mpegBufHdr.non_intra_matrix_flag = htonl(mpegBufHdr.non_intra_matrix_flag);
	    mpegBufHdr.index = htonl(mpegBufHdr.index);
	    mpegBufHdr.logStart_s = htonl(mpegBufHdr.logStart_s);
	    mpegBufHdr.logStart_ns = htonl(mpegBufHdr.logStart_ns);
	    mpegBufHdr.logEnd_s = htonl(mpegBufHdr.logEnd_s);
	    mpegBufHdr.logEnd_ns = htonl(mpegBufHdr.logEnd_ns);
	    mpegBufHdr.displayFlag = htonl(mpegBufHdr.displayFlag);
	    
	    memcpy(frame_buf->data, &(mpegBufHdr), 8+MPEG_BUF_HDR_SIZE);
	    
	    data_seg_type = htonl(CM_DATA_SEGMENT);
	    data_seg_length = htonl(fPickList[pick].size);
	    
	    memcpy(frame_buf->data+8+MPEG_BUF_HDR_SIZE, &(data_seg_type), 4);
	    memcpy(frame_buf->data+12+MPEG_BUF_HDR_SIZE, &(data_seg_length), 4);
	    
	    MPEG_FIP_ReadFrameFromFile(objPtr->fip, fPickList[pick].frame_num, 
				   frame_buf->data+16+MPEG_BUF_HDR_SIZE);
	    
	    frame_len = 16+MPEG_BUF_HDR_SIZE+fPickList[pick].size;
	    
	    sb = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));
	    
	    if (intra_buf != NULL) {
		if (non_intra_buf != NULL) {
		    sb->count = 3;
		    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *) * 3);
		    sb->buffers[0] = frame_buf;
		    sb->buffers[1] = intra_buf;
		    sb->buffers[2] = non_intra_buf;
		    sb->offsets = (int *) ckalloc(sizeof(int) * 3);
		    sb->offsets[0] = 0;
		    sb->offsets[1] = 0;
		    sb->offsets[2] = 0;
		    sb->lens = (int *) ckalloc(sizeof(int) * 3);
		    sb->lens[0] = frame_len;
		    sb->lens[1] = intra_len;
		    sb->lens[2] = non_intra_len;
		} else {
		    sb->count = 2;
		    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *) * 2);
		    sb->buffers[0] = frame_buf;
		    sb->buffers[1] = intra_buf;
		    sb->offsets = (int *) ckalloc(sizeof(int) * 2);
		    sb->offsets[0] = 0;
		    sb->offsets[1] = 0;
		    sb->lens = (int *) ckalloc(sizeof(int) * 2);
		    sb->lens[0] = frame_len;
		    sb->lens[1] = intra_len;
		}
	    } else {
		if (non_intra_buf != NULL) {
		    sb->count = 2;
		    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *) * 2);
		    sb->buffers[0] = frame_buf;
		    sb->buffers[1] = non_intra_buf;
		    sb->offsets = (int *) ckalloc(sizeof(int) * 2);
		    sb->offsets[0] = 0;
		    sb->offsets[1] = 0;
		    sb->lens = (int *) ckalloc(sizeof(int) * 2);
		    sb->lens[0] = frame_len;
		    sb->lens[1] = non_intra_len;	
		} else {
		    sb->count = 1;
		    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *) * 1);
		    sb->buffers[0] = frame_buf;
		    sb->offsets = (int *) ckalloc(sizeof(int) * 1);
		    sb->offsets[0] = 0;
		    sb->lens = (int *) ckalloc(sizeof(int) * 1);
		    sb->lens[0] = frame_len;
		}
	    }
	
	    sbl->sbs[i] = sb;
	
	    sb_cntr++;

	    frame_buf = NULL;
	    intra_buf = NULL;
	    non_intra_buf = NULL;
	}
    }

    return sbl;


  error:
    if (frame_buf != NULL) {
	DetachBuffer(objPtr->interp, frame_buf, "Error in Make SBL from PickList");
	frame_buf = NULL;
    }
    if (intra_buf != NULL) {
	DetachBuffer(objPtr->interp, intra_buf, "Error in Make SBL from PickList");
	intra_buf = NULL;
    }
    if (non_intra_buf != NULL) {
	DetachBuffer(objPtr->interp, non_intra_buf, "Error in Make SBL from PickList");
	non_intra_buf = NULL;
    }
    
    if (sbl != NULL) {
	for (i=0; i<sb_cntr; i++) {
	    sb = sbl->sbs[i];
	    for (j=0; j<sb->count; j++) {
		DetachBuffer(objPtr->interp, sb->buffers[j],
			     "Error in Make SBL from PickList");
	    }
	    ckfree((char *) sb);
	}
	ckfree((char *) sbl);
    }
    
    return NULL;
}

/************************************************************
 *
 * MPEGSegmentPrioritize
 *
 * Parameters:
 *     numFrames - length of fPickList array to be prioritized.
 *     fPickList - array of frame pick structures.
 *     orderVec - array of n integers where n is number of
 *                frames to be prioritized. Used to pass back
 *                results of priority algorithm.
 *   
 * Results:
 *     None.
 *
 * Actions:
 *     Sets value of i element in orderVec to be the index of
 *     the i th most important frame.
 *
 ************************************************************/

static void
MPEGSegmentPrioritize (int numFrames, 
		       FramePick *fPickList, 
		       int *orderVec)
{
    int i, j, idx, numB, type, b_priority;

    /*
     * Make I and P frames most important.
     */

    idx = 0;

    for (i=0; i<numFrames; i++) {
	type = fPickList[i].type;
	if (type == MPEG_I_FRAME) {
	    orderVec[idx++] = i;
	}
    }

    for (i=0; i<numFrames; i++) {
	type = fPickList[i].type;
	if (type == MPEG_P_FRAME) {
	    orderVec[idx++] = i;
	}
    }

    numB = numFrames - idx;

    MPEG_Prioritize (numB, 0, 1, &(orderVec[idx]));

    b_priority = 0;

    for (i=0; i<numFrames; i++) {
	type = fPickList[i].type;
	if (type == MPEG_B_FRAME) {
	    for (j=idx; j<numFrames; j++) {
		if (orderVec[j] == b_priority) {
		    orderVec[j] = -1 * i;
		    break;
		}
	    }
	    b_priority++;
	}
    }

    for (j=idx; j<numFrames; j++) {
      if (orderVec[j] < 0) {
	orderVec[j] *= -1;
      }
    }

    return;
}

/*
 *--------------------------------------------------------------
 *
 * MPEG_Prioritize --
 *
 *	A prioritization algorithm (by Daishi Harada) that computes the
 *	ordering for the MPEG frames.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters (these are my best guesses -- bsmith)
 *	int n		Number of frames to prioritize
 *	int m		Starting frame number.
 *	int k		Starting priority number
 *	int *pl		Output priority array
 *
 *--------------------------------------------------------------
 */
static void
MPEG_Prioritize (int n, int m, int k, int *pl)
{
    int s,t;

    if (n==0) return;

    if (n==1) {
	*pl = m;
    } else {
	s = n/2;
	t = n-s;
	MPEG_Prioritize(t,m,k*2,pl);
	MPEG_Prioritize(s,m+t,k*2,pl+k);
    }
}
