/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * tcmMPEGOptiVideoDevice.c
 *
 * Device to work with the CM MpegPlay object to decode
 * a motion MPEG stream on an Optivsion board
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <assert.h>
#include <tcl.h>
#include <tk.h>
#include <dp.h>
#include <errno.h>

#include <sys/types.h>
#include <netinet/in.h>

#include "cmMPEGPlay.h"

#ifdef MPEG_OPTIVIDEO_DEVICE

#include <sys/ioctl.h>
#include "cmLTS.h"

#ifndef _WINDOWS
#include "mpeg_ioctl.h"
#endif

#define MPEG_I_FRAME 1
#define MPEG_P_FRAME 2
#define MPEG_B_FRAME 3

/* Errors */
#define CANT_OPEN_DEVICE_ERROR  100001


static int MPEG_OptiDev_Decode(MPEGFrame *framePtr);
static int MPEG_OptiDev_Refresh(MPEGPlay *objPtr);
static int MPEG_OptiDev_Display(MPEGFrame *framePtr);
static int MPEG_OptiDev_Init(MPEGPlay  *objPtr);
static int MPEG_OptiDev_Reinit(MPEGPlay *objPtr);
static int MPEG_OptiDev_Shutdown(MPEGPlay *objPtr);
static int MPEG_OptiDev_FreeFrame(MPEGFrame *framePtr);
static int MPEG_OptiDev_Config(MPEGPlay *objPtr, char *option, char *value);
static int MPEG_OptiDev_NextFrame(MPEGPlay *objPtr, MPEGFrame *frameList, 
				  MPEGFrame **chosenFrame);
static char *MPEG_OptiDev_Error(MPEGPlay *objPtr, int errNum);

extern Display *Tcm_Display();

typedef struct {
    int wasDisplayed;
} MPEGOptiVideoFrameData;

typedef struct {
#ifndef _WINDOWS
    int device;			/* File descriptor for open device */
#else
    void *device;
#endif
    int seq_header_sent;
    int last_width;
    int last_height;
} MPEGOptiVideoDevice;

#ifdef _WINDOWS
typedef struct _DRAWBUFFER {
    unsigned char *lpData;	/* frame data to be drawn */
    unsigned long length;	/* Length of frame to be drawn */
    unsigned long ulWidth;	/* width of frame in pixels */
    unsigned long ulHeight;	/* height of frame in pixels */
    unsigned long Format;	/* h/w specific data format code */
    RECT	rcSource;	/* write only this rect to the device */
} DRAWBUFFER, *PDRAWBUFFER;

typedef void *VCUSER_HANDLE;
typedef unsigned short * PWCHAR;

VCUSER_HANDLE __stdcall VC_OpenDevice(PWCHAR pDriverName, int DeviceIndex);
extern void __stdcall VC_CloseDevice(VCUSER_HANDLE vh);
extern int __stdcall VC_DrawFrame(VCUSER_HANDLE vh, PDRAWBUFFER pBuffer);
#endif

/*****************************************************************
 *
 * GetOptiVideoMpegDevice
 *
 * Parameters:
 *     objPtr - pointer to mpeg play object.
 *
 * Return:
 *     Pointer to new OptiVideo device.
 *
 * Action:
 *     Creates device structure and fills it.
 *
 *****************************************************************/
MPEGDevice *
GetOptiVideoMpegDevice (MPEGPlay *objPtr) 
{
    MPEGDevice *new_dev;

    new_dev = (MPEGDevice *) ckalloc(sizeof(MPEGDevice));
    if (new_dev == NULL) {
	return NULL;
    }

    new_dev->deviceData = NULL;
    new_dev->asyncDecodeFlag = 0;
    new_dev->decodeProc = MPEG_OptiDev_Decode;
    new_dev->refreshProc = MPEG_OptiDev_Refresh;
    new_dev->displayProc = MPEG_OptiDev_Display;
    new_dev->initProc = MPEG_OptiDev_Init;
    new_dev->reinitProc = MPEG_OptiDev_Reinit;
    new_dev->shutdownProc = MPEG_OptiDev_Shutdown;
    new_dev->freeFrameProc = MPEG_OptiDev_FreeFrame;
    new_dev->configProc = MPEG_OptiDev_Config;
    new_dev->nextFrameProc = MPEG_OptiDev_NextFrame;
    new_dev->errorProc = MPEG_OptiDev_Error;

    return new_dev;
}

/*********************************************************************
 *
 * MPEG_OptiDev_Decode
 *
 * Parameters:
 *     framePtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_OptiDev_Decode(MPEGFrame *framePtr)
{
    MPEGPlay *objPtr;
    MPEGDevice *devicePtr;
    MPEGOptiVideoDevice *optiDevPtr;
    MPEGOptiVideoFrameData *frameDevData;

    if (framePtr == NULL) return 0;
    
    objPtr = framePtr->objPtr;
    if (objPtr == NULL) return 0;
    
    devicePtr = objPtr->devicePtr;
    if (devicePtr == NULL) return 0;

    optiDevPtr = (MPEGOptiVideoDevice *) devicePtr->deviceData;
    if (optiDevPtr == NULL) return 0;

    /*
     * Check device client data in frame ptr.
     * NULL indicates new frame not in progress.
     *      Allocate structures and fill them.
     *      Parse off picture header.
     *      Set device's asynchDecode flag to 1.
     * Calculate how many macroblocks to be decoded.
     * Decode them.
     * If frame is completely decoded:
     *    Dither the picture.
     *    Set device's asynchDecode flag to 0.
     */
    
    if (framePtr->deviceData == NULL) {
	framePtr->deviceData = (ClientData) ckalloc(sizeof(MPEGOptiVideoFrameData));
	frameDevData = (MPEGOptiVideoFrameData *) framePtr->deviceData;
	frameDevData->wasDisplayed = 0;

	/* Display even reference frames on OptiVideo */
	framePtr->frameHeader.displayFlag = 1;

	devicePtr->asyncDecodeFlag = 0;
    }
    return 0;
}
/*********************************************************************
 *
 * MPEG_OptiDev_Refresh
 *
 * Parameters:
 *     objPtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_OptiDev_Refresh(MPEGPlay *objPtr)
{
    return 0;
}


/*********************************************************************
 *
 * MPEG_OptiDev_Display_Util
 *
 * Parameters:
 *     framePtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int 
MPEG_OptiDev_Display_Util(MPEGFrame *framePtr, double speed, int isRef)
{
    MPEGOptiVideoDevice *optiDevPtr;
    MPEGDevice *devicePtr;
    MPEGOptiVideoFrameData *frameDevData, *refDevData;
    MPEGPlay *objPtr;
    int n, type;
#ifdef _WINDOWS
    DRAWBUFFER Draw;
#endif
    static unsigned char headerSeqAndGOP[] =
	{0x00, 0x00, 0x01, 0xb3, 0x0a, 0x00, 0x78, 0x13,
#if 0
	 /*0xff*/ 0x0f, 0xff, 0xe0, /* 0x84 */ 0xa0,
#else
	 0xff, 0xff, 0xe0, 0x84,
#endif
	 0x00, 0x00, 0x01, 0xb8, 0x00, 0x00, 0x00, 0x00};

#ifdef LOG_IT
    static int fd = -1;

    if (fd == -1) {
	fd = open("/tmp/write.mpg", O_WRONLY|O_TRUNC|O_CREAT, 0666);
    }
#endif

    frameDevData = (MPEGOptiVideoFrameData *) framePtr->deviceData;
    if (frameDevData->wasDisplayed == 1) {
	return 0;
    }

    type = framePtr->frameHeader.type;
    if (type != MPEG_I_FRAME) {
	if (framePtr->ref1) {
	    refDevData = (MPEGOptiVideoFrameData *) framePtr->ref1->deviceData;
	    if (refDevData->wasDisplayed == 0) {
		MPEG_OptiDev_Display_Util(framePtr->ref1, speed, 1);
	    }
	}
    }

    if (type == MPEG_B_FRAME) {
	if (framePtr->ref2) {
	    refDevData = (MPEGOptiVideoFrameData *) framePtr->ref2->deviceData;
	    if (refDevData->wasDisplayed == 0) {
		MPEG_OptiDev_Display_Util(framePtr->ref2, speed, 1);
	    }
	}
    }

    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    optiDevPtr = (MPEGOptiVideoDevice *) devicePtr->deviceData;

#if 0
    {
      char logStr[256];
      double now;

      now = LTS_Value(objPtr->lts);
      sprintf(logStr, "Displaying: start: %f now: %f end: %f index: %d type: %d", framePtr->logStart, 
	      now, framePtr->logEnd, framePtr->frameHeader.index, type);
      Tcm_Log(logStr);
    }
#endif

    if (framePtr->frameHeader.width <= 352 && framePtr->frameHeader.height <= 288) {
	if (framePtr->frameHeader.width != optiDevPtr->last_width ||
	    framePtr->frameHeader.height != optiDevPtr->last_height)
	{
	    optiDevPtr->last_width = framePtr->frameHeader.width;
	    optiDevPtr->last_height = framePtr->frameHeader.height;
	    optiDevPtr->seq_header_sent = 0;
	}


	if (!optiDevPtr->seq_header_sent) {
	    headerSeqAndGOP[4] = ((framePtr->frameHeader.width + 15) / 16);
	    headerSeqAndGOP[5] = (framePtr->frameHeader.height > 256) ? 1 : 0;
	    headerSeqAndGOP[6] = (unsigned char)
		(framePtr->frameHeader.height - 256 * headerSeqAndGOP[5]);

#ifdef _WINDOWS
	    Draw.lpData = headerSeqAndGOP;
	    Draw.length = sizeof(headerSeqAndGOP);

	    VC_DrawFrame(optiDevPtr->device, &Draw);
#else
	    write(optiDevPtr->device, headerSeqAndGOP, sizeof(headerSeqAndGOP));
#endif
#ifdef LOG_IT
	    write(fd, headerSeqAndGOP, sizeof(headerSeqAndGOP));
#endif
	    optiDevPtr->seq_header_sent = 1;
	}
	    
#if 0
	printf("To board: %08x, Index %6d, %c\n", 
	       framePtr,
	       framePtr->frameHeader.index,
	       (type == 1) ? 'I' : ((type == 2) ? 'P' : 'B'));
#endif

#ifdef _WINDOWS
	Draw.lpData = framePtr->data;
	Draw.length = framePtr->dataLength;

	if (speed < 0.0) {
	    if (type == MPEG_I_FRAME) {
		VC_DrawFrame(optiDevPtr->device, &Draw);
	    }
	} else {
	    VC_DrawFrame(optiDevPtr->device, &Draw);
	    if ((speed == 0.0) && (! isRef) &&
		(type != MPEG_P_FRAME))
	    {
		VC_DrawFrame(optiDevPtr->device, &Draw);
		VC_DrawFrame(optiDevPtr->device, &Draw);
	    }
	}
#else
	if (speed < 0.0) {
	    if (type == MPEG_I_FRAME) {
		n = write(optiDevPtr->device, framePtr->data,
			  framePtr->dataLength);
#ifdef LOG_IT
		write(fd, framePtr->data, framePtr->dataLength);
#endif
	    }
	} else {
	    n = write(optiDevPtr->device, framePtr->data,
		      framePtr->dataLength);
#ifdef LOG_IT
	    write(fd, framePtr->data, framePtr->dataLength);
#endif
	    if ((speed == 0.0) && (! isRef) &&
		(type != MPEG_P_FRAME))
	    {
		n = write(optiDevPtr->device, framePtr->data,
			  framePtr->dataLength);
		n = write(optiDevPtr->device, framePtr->data,
			  framePtr->dataLength);
#ifdef LOG_IT
		write(fd, framePtr->data, framePtr->dataLength);
		write(fd, framePtr->data, framePtr->dataLength);
#endif
	    }
	}
#endif
	/* write(fd, framePtr->data, framePtr->dataLength); */
	frameDevData->wasDisplayed = 1;
    } else {
	printf("Frame too big: %d x %d\n",
	       framePtr->frameHeader.width,
	       framePtr->frameHeader.height);
    }
    return 0;
}

/*********************************************************************
 *
 * MPEG_OptiDev_Display
 *
 * Parameters:
 *     framePtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int 
MPEG_OptiDev_Display(MPEGFrame *framePtr)
{
    MPEGOptiVideoFrameData *frameDevData;
    MPEGPlay *objPtr;
    double speed;

    frameDevData = (MPEGOptiVideoFrameData *) framePtr->deviceData;
    if (frameDevData == NULL) {
        return 0;
    }

    if (frameDevData->wasDisplayed == 1) {
	return 0;
    }

    objPtr = framePtr->objPtr;
    speed = LTS_Speed(objPtr->lts);

    return MPEG_OptiDev_Display_Util(framePtr, speed, 0);
}

/*********************************************************************
 *
 * MPEG_OptiDev_Init
 *
 * Parameters:
 *     objPtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_OptiDev_Init(MPEGPlay *objPtr)
{
    MPEGDevice *devicePtr;
    MPEGOptiVideoDevice *optiDevPtr;
    if (objPtr == NULL) return 0;
   
    devicePtr = objPtr->devicePtr;
    if (devicePtr == NULL) return 0;

    optiDevPtr = (MPEGOptiVideoDevice *) devicePtr->deviceData;
    if (optiDevPtr == NULL) {
	optiDevPtr = (MPEGOptiVideoDevice *) ckalloc(sizeof(MPEGOptiVideoDevice));
	devicePtr->deviceData = (ClientData) optiDevPtr;
    }

#ifndef _WINDOWS
    optiDevPtr->device = open("/dev/mpegdecv0", O_WRONLY|O_NONBLOCK);

    if (optiDevPtr->device < 0)
    {
	ckfree((char *) optiDevPtr);
	devicePtr->deviceData = NULL;
	return -errno;
    }
#else
    optiDevPtr->device = VC_OpenDevice(NULL, 0);
    if (optiDevPtr->device == NULL)
    {
	ckfree((char *) optiDevPtr);
	devicePtr->deviceData = NULL;
	return CANT_OPEN_DEVICE_ERROR;
    }
#endif

    optiDevPtr->seq_header_sent = 0;
    optiDevPtr->last_width = optiDevPtr->last_height = 0;
    return 0;
}


/*********************************************************************
 *
 * MPEG_OptiDev_Reinit
 *
 * Parameters:
 *     objPtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_OptiDev_Reinit(MPEGPlay *objPtr)
{
    MPEGDevice *devicePtr;
    MPEGOptiVideoDevice *optiDevPtr;

    devicePtr = objPtr->devicePtr;
    if (devicePtr == NULL) return 0;

    optiDevPtr = (MPEGOptiVideoDevice *) devicePtr->deviceData;
    if (optiDevPtr == NULL) return 0;

    return 0;
}


/*********************************************************************
 *
 * MPEG_OptiDev_Shutdown
 *
 * Parameters:
 *     objPtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_OptiDev_Shutdown(MPEGPlay *objPtr) 
{
    MPEGOptiVideoDevice *optiDevPtr;

    if (objPtr == NULL) return 0;
    if (objPtr->devicePtr == NULL) return 0;

    optiDevPtr = (MPEGOptiVideoDevice *) objPtr->devicePtr->deviceData;

    if (optiDevPtr != NULL) {
#ifndef _WINDOWS
	if (optiDevPtr->device != -1) {
	    close(optiDevPtr->device);
	}
#else
	VC_CloseDevice(optiDevPtr->device);
#endif
	ckfree((char *) optiDevPtr);
    }
    objPtr->devicePtr->deviceData = NULL;
    return 0;
}


/*********************************************************************
 *
 * MPEG_OptiDev_FreeFrame
 *
 * Parameters:
 *     framePtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_OptiDev_FreeFrame(MPEGFrame *framePtr)
{
    MPEGPlay *objPtr;
    MPEGOptiVideoFrameData *frameDevData;

#if 0
    printf("Freeing %c frame\n",
	   (framePtr->frameHeader.type == 1) ? 'I' :
	   ((framePtr->frameHeader.type == 2) ?
	    'P' : 'B'));
#endif
    objPtr = framePtr->objPtr;
    frameDevData = (MPEGOptiVideoFrameData *) framePtr->deviceData;

    if (framePtr->deviceData != NULL) {
	ckfree((char *) frameDevData);
	framePtr->deviceData = NULL;
    }

    return 0;
}


/*********************************************************************
 *
 * MPEG_OptiDev_Config
 *
 * Parameters:
 *     objPtr
 *     option
 *     value
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int 
MPEG_OptiDev_Config(MPEGPlay *objPtr, char *option, char *value)
{
    return 0;
}

/*********************************************************************
 *
 * MPEG_OptiDev_NextFrame
 *
 * Parameters:
 *     objPtr
 *     frameList, 
 *     chosenFrame
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_OptiDev_NextFrame(MPEGPlay *objPtr, MPEGFrame *frameList, MPEGFrame **chosenFrame)
{
    MPEGFrame *framePtr, *latestFrame;
    double speed, now, end;

    if (frameList == NULL) {
	*chosenFrame = NULL;
	return 1;
    }

    speed = LTS_Speed(objPtr->lts);
    now = LTS_Value(objPtr->lts);

    latestFrame = frameList;

    for (framePtr = frameList; framePtr != NULL; framePtr = framePtr->next) {
	end = (speed < 0.0) ? framePtr->logStart : framePtr->logEnd;

	if (now <= end) {
	    *chosenFrame = framePtr;
	    return 0;
	}

	if (speed < 0.0) {
	    if (framePtr->logEnd < latestFrame->logEnd) {
		latestFrame = framePtr;
	    }
	} else {
	    if (framePtr->logEnd > latestFrame->logEnd) {
		latestFrame = framePtr;
	    }
	}
    }

    *chosenFrame = latestFrame;
    return 0;
}


/*********************************************************************
 *
 * MPEG_OptiDev_Error
 *
 * Parameters:
 *     objPtr
 *     errNum
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static char *dummyStr="OptiVideo Dev dummy string";

static char *
MPEG_OptiDev_Error(MPEGPlay *objPtr, int errNum) 
{
    /* Take care of file type errors first */
    if (errNum < 0) {
	return strerror(-errNum);
    }
    switch (errNum) {
      case CANT_OPEN_DEVICE_ERROR:
	return "Unable to open OptiVideo device";
    }

    printf("In OptiVideo Dev error: %d\n", errNum);
    return dummyStr;
}

#else /* MPEG_SW_DEVICE */

/*****************************************************************
 *
 * GetOptiVideoMpegDevice
 *
 * Parameters:
 *     objPtr - pointer to mpeg play object.
 *
 * Return:
 *     Pointer to new OptiVideo device.
 *
 * Action:
 *     Stub for compiling without OptiVideo device.
 *
 *****************************************************************/
MPEGDevice *
GetOptiVideoMpegDevice (MPEGPlay *objPtr) 
{
    return NULL;
}

#endif /* MPEG_OPTIVIDEO_DEVICE */
