/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMPEGPlay.c,v 1.30 1996/02/06 22:42:00 kpatel Exp $
 *
 * Implements the "mpegPlay" object.  Takes incoming MPEG frames,
 * manages of queue of decodable/playable frames, then uses a
 * mpeg decoding device to decode and display frames
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <math.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "cmMPEGPlay.h"
#include "tcmEvent.h"
#include "dp.h"
#include "cmEvery.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmEvery.h"
#include "cmBind.h"
#include "cmUtil.h"


#define ONE_BILLION 1000000000.0
#define MPEG_I_FRAME 1
#define MPEG_P_FRAME 2
#define MPEG_B_FRAME 3

static int nextMPEGPlayNum = 0;	/* Variable for next unique object
					 * name. */

/*
 * Prototypes for routines used internally
 */
static MPEGPlay *	CreateNewMPEGPlay(Tcl_Interp *interp, char *name);
static void		ChangeDeviceType(MPEGPlay *objPtr, char *devtype);
static int		DestroyMPEGPlay(MPEGPlay *objPtr);
static int		MPEGPlayObjCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		MPEG_Play_Accept(MPEGPlay *objPtr, SB_List *sbl);
static int		MPEG_Play_Ready(MPEGPlay *objPtr);
static int		MPEG_Play_Unready(MPEGPlay *objPtr);
static int		MPEG_Play_Stats(MPEGPlay *objPtr);
static int		MPEG_Play_Refresh(MPEGPlay *objPtr);
static void		MPEG_Play_LTSChanged(ClientData clientData);
static int		MPEG_Play_DoDecode(MPEGPlay *objPtr);
static void		MPEG_Play_FreeFrame(MPEGFrame *framePtr);
static int		MPEGPlayCreateCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static void		MPEG_Play_ContinueDecode(MPEGFrame *framePtr);
static int		ConfigureMPEGPlay(MPEGPlay *objPtr,
			    int argc, char **argv);
static int		MPEG_Play_MissDisplay(MPEGFrame *framePtr);
static void		MPEG_Play_FreeFrameNow(MPEGFrame *framePtr);
static void		FreeMPEGFrameList(MPEGFrame *framePtr);
static void		MPEG_Play_InsertIntoDecodeQueue(MPEGPlay *objPtr,
			    MPEGFrame *framePtr);
static int		MPEG_Play_DeviceList(MPEGPlay *objPtr);
static char *		GetIndex(Tcl_Interp *interp, ClientData clientdata);
static char *		GetRef2Index(Tcl_Interp *interp,
			    ClientData clientdata);
static char *		GetRef1Index(Tcl_Interp *interp,
			    ClientData clientdata);
static char *		GetType(Tcl_Interp *interp, ClientData clientdata);
static char *		GetHeight(Tcl_Interp *interp, ClientData clientdata);
static char *		GetWidth(Tcl_Interp *interp, ClientData clientdata);
static char *		GetDataLen(Tcl_Interp *interp, ClientData clientdata);
static char *		GetLogEnd(Tcl_Interp *interp, ClientData clientdata);
static char *		GetLogStart(Tcl_Interp *interp, ClientData clientdata);
static char *		GetCumulLenDecoded(Tcl_Interp *interp,
			    ClientData clientdata);


/************************************************************
 *
 * MPEGPlayCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the "mpegPlay" Tcl command.
 *     A new object of type MPEGPlay is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
MPEGPlayCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    MPEGPlay *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextMPEGPlayNum++;
    sprintf(new_name, "%smpegPlay%d", argv[1], nextMPEGPlayNum);

    new = CreateNewMPEGPlay(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MPEGPlayObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureMPEGPlay(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MPEGPlayObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     MPEGPlay object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
MPEGPlayObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    MPEGPlay *objPtr = (MPEGPlay *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureMPEGPlay(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyMPEGPlay(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl = NULL;
	int result;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}

	/*
	 * Parse the ScatterBufferList
	 */
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	  ckfree((char *) sbl);
	  Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
			   argv[2],  (char *)NULL);
	  return TCL_ERROR;
	}

	result = MPEG_Play_Accept(objPtr, sbl);
	if (sbl != NULL) {
	    FreeSBList(sbl);
	}
	return result;

    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return MPEG_Play_Ready(objPtr);

    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return MPEG_Play_Unready(objPtr);

    } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
	return MPEG_Play_Stats(objPtr);

    } else if ((c == 'r') && (strncmp("refresh", argv[1], len) == 0)) {
	return MPEG_Play_Refresh(objPtr);

    } else if ((c == 'd') && (strncmp("deviceList", argv[1], len) == 0)) {
	return MPEG_Play_DeviceList(objPtr);
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ",
		argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
		"\n<command> can be one of:\n",
		"  configure ?-slot ?value??\n",
		"  destroy\n",
		"  accept <ScatterBufferList>\n",
		"  ready\n",
		"  unready\n",
		"  stats\n",
		"  refresh\n",
		"  deviceList\n",
		(char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ChangeDeviceType --
 *
 *	Sets the device type for the MPEGReseource.  Called whenever
 *	the "devtype" slot is changed, and on initialization of the
 *	Play.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The old device is shut down, if appropriate, and the new
 *	device is created.
 *
 * Parameters:
 *	MPEGPlay *objPtr	Pointer to Play object.
 *	char *devtype		String giving name of new device
 *
 *--------------------------------------------------------------
 */
static void
ChangeDeviceType (MPEGPlay *objPtr, char *devtype)
{
    int len;
    char c;

    len = strlen(devtype);
    c = devtype[0];

    /*
     * If there's no change, return
     */
    if ((objPtr->devtype != NULL) &&
	(strncmp (objPtr->devtype, devtype, len) == 0)) {
	return;
    }

    /*
     * Shut down old device, if necessary, ignoring any errors it
     * returns.
     */
    MPEG_Play_Unready(objPtr);
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
	objPtr->devtype = NULL;;
    }
    if (objPtr->devicePtr != NULL) {
	if (objPtr->devicePtr->shutdownProc != NULL)
	  objPtr->devicePtr->shutdownProc(objPtr);
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
	objPtr->inited = 0;
    }

    /*
     * Create new device.
     */
    if ((c == 's') && (strncmp("software", devtype, len) == 0)) {
	objPtr->devicePtr = GetSoftwareMpegDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("software")+1);
	strcpy(objPtr->devtype, "software");
    } else if ((c == 'o') && (strncmp("optivideo", devtype, len) == 0)) {
	objPtr->devicePtr = GetOptiVideoMpegDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("optivideo")+1);
	strcpy(objPtr->devtype, "optivideo");
    } else {
	devtype = "default";
	objPtr->devicePtr = GetDefaultMpegDevice(objPtr);
	objPtr->devtype = ckalloc(strlen(DEFAULT_MPEG_PLAY_DEV_TYPE)+1);
	strcpy(objPtr->devtype, DEFAULT_MPEG_PLAY_DEV_TYPE);
    }
}

/************************************************************
 *
 * CreateNewMPEGPlay
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static MPEGPlay *
CreateNewMPEGPlay(Tcl_Interp *interp, char *name)
{
    MPEGPlay *new;
    char *acceptName;
    char **event_list, **subst_list;
    CM_BindProc **subst_procs;
    ClientData *subst_cds;
    int i;

    new = (MPEGPlay *) ckalloc(sizeof(MPEGPlay));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);

    new->width = 0;
    new->height = 0;
    new->xpos = 0;
    new->ypos = 0;
    new->squeeze = 0;
    new->devtype = NULL;
    new->devicePtr = NULL;
    new->xid = 0;
    new->lts = NULL;
    new->ltsTriggerNum = -1;
    new->ltsName = NULL;
    new->maxFrameRate = 0;
    new->statStartTime = 0.0;
    new->missed = 0;
    new->late = 0;
    new->played = 0;
    new->bytes = 0;
    new->ready = 0;
    new->inited = 0;
    new->inDecodeFrame = NULL;
    new->queue = NULL;
    new->refQueue = NULL;
    new->waitQueue = NULL;
    new->cntrlEveryHandle = NULL;
    new->at_tokens = NULL;
    new->num_at_tokens = 0;

    /*
     * Set up bindings with binding mechanism.
     */

    event_list = (char **) ckalloc(5*sizeof(char *));
    event_list[0] = (char *) ckalloc(strlen("<FRAME_DROPPED>")+1);
    strcpy(event_list[0], "<FRAME_DROPPED>");
    event_list[1] = (char *) ckalloc(strlen("<FRAME_RECEIVED>")+1);
    strcpy(event_list[1], "<FRAME_RECEIVED>");
    event_list[2] = (char *) ckalloc(strlen("<FRAME_PLAYED>")+1);
    strcpy(event_list[2], "<FRAME_PLAYED>");
    event_list[3] = (char *) ckalloc(strlen("<FRAME_DECODE_BEGIN>")+1);
    strcpy(event_list[3], "<FRAME_DECODE_BEGIN>");
    event_list[4] = (char *) ckalloc(strlen("<FRAME_DECODE_DONE>")+1);
    strcpy(event_list[4], "<FRAME_DECODE_DONE>");

    subst_list = (char **) ckalloc(10*sizeof(char *));
    subst_procs = (CM_BindProc **) ckalloc(10*sizeof(CM_BindProc *));
    subst_cds = (ClientData *) ckalloc(10*sizeof(ClientData));

    subst_list[0] = (char *) ckalloc(2);
    (subst_list[0])[0] = 's';
    (subst_list[0])[1] = '\0';
    subst_procs[0] = GetLogStart;
    subst_cds[0] = (ClientData) new;

    subst_list[1] = (char *) ckalloc(2);
    (subst_list[1])[0] = 'e';
    (subst_list[1])[1] = '\0';
    subst_procs[1] = GetLogEnd;
    subst_cds[1] = (ClientData) new;

    subst_list[2] = (char *) ckalloc(2);
    (subst_list[2])[0] = 'l';
    (subst_list[2])[1] = '\0';
    subst_procs[2] = GetDataLen;
    subst_cds[2] = (ClientData) new;

    subst_list[3] = (char *) ckalloc(2);
    (subst_list[3])[0] = 'w';
    (subst_list[3])[1] = '\0';
    subst_procs[3] = GetWidth;
    subst_cds[3] = (ClientData) new;

    subst_list[4] = (char *) ckalloc(2);
    (subst_list[4])[0] = 'h';
    (subst_list[4])[1] = '\0';
    subst_procs[4] = GetHeight;
    subst_cds[4] = (ClientData) new;

    subst_list[5] = (char *) ckalloc(2);
    (subst_list[5])[0] = 't';
    (subst_list[5])[1] = '\0';
    subst_procs[5] = GetType;
    subst_cds[5] = (ClientData) new;

    subst_list[6] = (char *) ckalloc(2);
    (subst_list[6])[0] = 'p';
    (subst_list[6])[1] = '\0';
    subst_procs[6] = GetRef1Index;
    subst_cds[6] = (ClientData) new;

    subst_list[7] = (char *) ckalloc(2);
    (subst_list[7])[0] = 'n';
    (subst_list[7])[1] = '\0';
    subst_procs[7] = GetRef2Index;
    subst_cds[7] = (ClientData) new;

    subst_list[8] = (char *) ckalloc(2);
    (subst_list[8])[0] = 'i';
    (subst_list[8])[1] = '\0';
    subst_procs[8] = GetIndex;
    subst_cds[8] = (ClientData) new;

    subst_list[9] = (char *) ckalloc(2);
    (subst_list[9])[0] = 'b';
    (subst_list[9])[1] = '\0';
    subst_procs[9] = GetCumulLenDecoded;
    subst_cds[9] = (ClientData) new;


    new->bindToken = CreateCMBinding(interp,
				     new->name,
				     5, event_list,
				     10, subst_list, subst_procs, subst_cds);

    for (i=0; i<5; i++) {
      ckfree(event_list[i]);
    }
    ckfree((char *) event_list);

    for (i=0; i<10; i++) {
      ckfree(subst_list[i]);
    }
    ckfree((char *) subst_list);
    ckfree((char *) subst_procs);
    ckfree((char *) subst_cds);
	   

    ChangeDeviceType (new, "default");

    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+9);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) MPEG_Play_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    return new;
}


/************************************************************
 *
 * DestroyMPEGPlay
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

static int
DestroyMPEGPlay(MPEGPlay *objPtr)
{
    int result = TCL_OK;
    int i, devReturn;

    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Make unready, cancel any DoDecode when idle calls.
     */
    objPtr->ready = 0;
    Tcm_CancelIdleCall("video", (Tk_IdleProc *) MPEG_Play_DoDecode, objPtr);

    /*
     * If in decode, set destroy flag and return. Let DecodeDone do destroy.
     */
    if (objPtr->inDecodeFrame != NULL) {
	Tcm_CancelIdleCall("video", (Tk_IdleProc *) MPEG_Play_ContinueDecode, 
			   objPtr->inDecodeFrame);
	MPEG_Play_FreeFrameNow(objPtr->inDecodeFrame);
	objPtr->inDecodeFrame = NULL;
    }

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }

    /*
     * Shut down the device
     */
    if (objPtr->devicePtr != NULL) {
	if (objPtr->devicePtr->shutdownProc)
	  devReturn = objPtr->devicePtr->shutdownProc(objPtr);
	else devReturn = 0;
	if (devReturn != 0) {
	    result = TCL_ERROR;
	    if (objPtr->devicePtr->errorProc != NULL) {
		Tcl_AppendResult(objPtr->interp,
			objPtr->devicePtr->errorProc(objPtr, devReturn),
			(char *) NULL);
	    }
	}
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
    }

    /*
     * Clean up lts
     */
    if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
        Tk_Release(objPtr->lts);
        objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs, frame queues, and such.
     */
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }

    /*
     * Clean up previously scheduled display requests
     * Must use FreeFrameNow instead of FreeFrame since
     * some of the at tokens may have actually been FreeFrames
     * for ref frames that have been displayed but were still useful.
     */
    if (objPtr->at_tokens != NULL) {
	for (i=0; i<objPtr->num_at_tokens; i++) {
	    if (objPtr->at_tokens[i] > 0) {
		AtEvent *at_event;
		int token;

		token = objPtr->at_tokens[i];
		at_event = Tcm_FindAtEvent(token);
		if (token && (at_event != NULL)) {
		    MPEG_Play_FreeFrameNow((MPEGFrame *) at_event->fireClientData);
		    Tcm_DeleteAtHandler(token);
		    objPtr->at_tokens[i] = 0;
		}
	    }
	}
	ckfree((char *) objPtr->at_tokens);
    }

    /*
     * Free queue's. This must be done after killing all the at tokens
     * since free'ing the at tokens may have freed ref frames that lived
     * on the refQueue.
     */

    if (objPtr->queue != NULL) {
	FreeMPEGFrameList (objPtr->queue);
	objPtr->queue = NULL;
    }
    if (objPtr->waitQueue != NULL) {
	FreeMPEGFrameList (objPtr->waitQueue);
	objPtr->waitQueue = NULL;
    }
    if (objPtr->refQueue != NULL) {
	FreeMPEGFrameList (objPtr->refQueue);
	objPtr->refQueue = NULL;
    }

    /*
     * Unregister with bind mechanism.
     */

    DestroyCMBinding(objPtr->interp, objPtr->bindToken);

    ckfree((char *) objPtr);
    return result;
}


/************************************************************
 *
 * ConfigureMPEGPlay
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing. Sends any configuration
 *     options it doesn't understand to the device if present.
 *
 ************************************************************/

static int
ConfigureMPEGPlay(MPEGPlay * objPtr, int argc, char **argv)
{

    Tcl_Interp *interp;
    int cntr;
    char str[50];

    interp = objPtr->interp;

    /*
     * Case 1: <obj> configure
     */
    if (argc == 0) {
	sprintf(str, "{%d}", objPtr->width);
	Tcl_AppendResult(interp, "{", "-width ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->height);
	Tcl_AppendResult(interp, "{", "-height ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->xpos);
	Tcl_AppendResult(interp, "{", "-xpos ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->ypos);
	Tcl_AppendResult(interp, "{", "-ypos ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->squeeze);
	Tcl_AppendResult(interp, "{", "-squeeze ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->xid);
	Tcl_AppendResult(interp, "{", "-xid ", "0 ", str, "} ",
			 (char *)NULL);
	Tcl_AppendResult(interp, "{", "-device ", "{", DEFAULT_MPEG_PLAY_DEV_TYPE, "} ",
			 objPtr->devtype, "} ", (char *)NULL);

	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{}", "} ",
			 (char *)NULL);
	} else {
	  Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{",
			   objPtr->ltsName, "}", "} ", (char *) NULL);
	}
	sprintf(str, "{%d}", objPtr->maxFrameRate);
	Tcl_AppendResult(interp, "{", "-maxFrameRate ", "0 ", str, "} ",
			 (char *) NULL);

	/* Now let the device add some of its own. */

	if (objPtr->devicePtr != NULL) {
	    if (objPtr->devicePtr->configProc != NULL) {
		objPtr->devicePtr->configProc(objPtr, NULL, NULL);
	    }
	}

	return TCL_OK;
    }

    /*
     * Case 2: <obj> configure <-slot>
     */
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 'w') && (strncmp("-width", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->width);
	    Tcl_AppendResult(interp, "-width ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'h') && (strncmp("-height", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->height);
	    Tcl_AppendResult(interp, "-height ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'x') && (strncmp("-xpos", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->xpos);
	    Tcl_AppendResult(interp, "-xpos ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'y') && (strncmp("-ypos", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->ypos);
	    Tcl_AppendResult(interp, "-ypos ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 's') && (strncmp("-squeeze", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->squeeze);
	    Tcl_AppendResult(interp, "-squeeze ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'x') && (strncmp("-xid", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->xid);
	    Tcl_AppendResult(interp, "-xid ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-device", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-device ", "{", DEFAULT_MPEG_PLAY_DEV_TYPE, "} ",
			     objPtr->devtype, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	  if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "-lts ", "{} ", "{}", (char *) NULL);
	  } else {
	    Tcl_AppendResult(interp, "-lts ", "{} ", "{", objPtr->ltsName, "}", (char *) NULL);
	  }
	    return TCL_OK;
	} else if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->maxFrameRate);
	    Tcl_AppendResult(interp, "-maxFrameRate ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else {
	    /* Send to device and see if it understands it. */

	    if (objPtr->devicePtr != NULL) {
		if (objPtr->devicePtr->configProc != NULL) {
		    int err_code;

		    err_code = objPtr->devicePtr->configProc(objPtr, argv[0], NULL);
		    if (!err_code) return TCL_OK;
		}
	    }

	    Tcl_AppendResult(interp, "bad configuration option: ", 
			     argv[0], (char *) NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Case 3: <obj> configure <-slot> <newValue>
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;

	/*
	 * Make Play unready since something is changing.
	 */

	MPEG_Play_Unready(objPtr);

	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	if ((c == 'w') && (strncmp("-width", argv[cntr], len) == 0)) {
	    int newWidth;

	    cntr++;
	    if (cntr == argc) {
noValueError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newWidth) != TCL_OK) {
badValueError:
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->width = newWidth;

	    /*
	     * Send to device as well if configProc exists.
	     */

	    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->configProc != NULL)) {
		objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    }

	} else if ((c == 'h') && (strncmp("-height", argv[cntr], len) == 0)) {
	    int newHeight;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newHeight) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->height = newHeight;

	    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->configProc != NULL)) {
		objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    }

	} else if ((c == 'x') && (strncmp("-xpos", argv[cntr], len) == 0)) {
	    int newXpos;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newXpos) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->xpos = newXpos;

	    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->configProc != NULL)) {
		objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    }

	} else if ((c == 'y') && (strncmp("-ypos", argv[cntr], len) == 0)) {
	    int newYpos;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newYpos) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->ypos = newYpos;

	    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->configProc != NULL)) {
		objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    }

	} else if ((c == 's') && (strncmp("-squeeze", argv[cntr], len) == 0)) {
	    int newSqueeze;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newSqueeze) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->squeeze = newSqueeze;

	    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->configProc != NULL)) {
		objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    }

	} else if ((c == 'x') && (strncmp("-xid", argv[cntr], len) == 0)) {
	    int newXid;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newXid) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->xid = newXid;

	    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->configProc != NULL)) {
		objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    }

	} else if ((c == 'd') && (strncmp("-device", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    ChangeDeviceType(objPtr, argv[cntr]);
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (strlen(argv[cntr]) == 0) {
	        newLts = NULL;
	    }
	    else if ((newLts = Tcm_FindAtom(interp, argv[cntr])) == NULL) {
		goto badValueError;
	    }
	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }

            /*
             * Release our lock on the old lts...
             */
            if (objPtr->lts != NULL) {
	        if (objPtr->ltsTriggerNum != -1) {
		    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
		    objPtr->ltsTriggerNum = -1;
	        }
                Tk_Release(objPtr->lts);
                objPtr->lts = NULL;
            }

	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
	        objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts, 
					 LTS_SPEED_FLAG | LTS_VALUE_FLAG |
					 LTS_EARLY_PHASE, 
					 MPEG_Play_LTSChanged,
					 (ClientData) objPtr);
	        objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
	        strcpy(objPtr->ltsName, argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }

	    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->configProc != NULL)) {
		objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    }

	} else if ((c == 'm') &&
		   (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
	    int newMaxFrameRate;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->maxFrameRate = newMaxFrameRate;

	    if ((objPtr->devicePtr != NULL) && (objPtr->devicePtr->configProc != NULL)) {
		objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    }

	} else if ((objPtr->devicePtr != NULL) && 
		   (objPtr->devicePtr->configProc != NULL)){
	    int err_code;

	    cntr++;

	    if (cntr == argc) {
		goto noValueError;
	    }
	    
	    err_code = objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]);
	    if (err_code) {
		if (objPtr->devicePtr->errorProc != NULL) {
		    Tcl_AppendResult(objPtr->interp,
				     objPtr->devicePtr->errorProc(objPtr, err_code),
				     (char *) NULL);
		    return TCL_ERROR;
		} else {
		    Tcl_AppendResult(objPtr->interp, "Device configuration error on ",
				     "option \"", argv[cntr-1], argv[cntr], "\"",
				     (char *) NULL);
		    return TCL_ERROR;
		}
	    }
	}
	cntr++;
    }

    return TCL_OK;
}


/************************************************************
 *
 * MPEG_Play_Ready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_Play_Ready(MPEGPlay * objPtr)
{
    int init_res;

    /*
     * Can only be made ready if
     *  1. We have a device.
     *  2. We have an lts
     *  3. We have a window
     */
    if (objPtr->devicePtr == NULL) {
        Tcl_AppendResult(objPtr->interp, "Can't make MPEG Play \"",
                objPtr->name, "\" ready: device type is not set.", (char *)NULL);
        return TCL_ERROR;
    }
    if (objPtr->lts == NULL) {
        Tcl_AppendResult(objPtr->interp, "Can't make MPEG Play \"",
                objPtr->name, "\" ready: no lts has been attached.", (char *)NULL);
        return TCL_ERROR;
    }
    if (objPtr->xid == 0) {
        Tcl_AppendResult(objPtr->interp, "Can't make MPEG Play \"",
                objPtr->name, "\" ready: output window not specified.",
		(char *)NULL);
        return TCL_ERROR;
    }


    /*
     * If not inited, do initProc, catch result.
     */
    init_res = 0;
    if (!objPtr->inited) {
      if (objPtr->devicePtr->initProc != NULL) {
	init_res = objPtr->devicePtr->initProc(objPtr);
      }
      if (init_res == 0) 
	objPtr->inited = 1;
    }
    else {
      /*
       * Otherwise, do reinitProc, catch result.
       */
      if (objPtr->devicePtr->reinitProc != NULL) {
	init_res = objPtr->devicePtr->reinitProc(objPtr);
      }
    }
      
    /*
     * If bad result, get error string, return error.
     */
    if (init_res != 0) {
      char *errStr;

      if (objPtr->devicePtr->errorProc != NULL) {
	  errStr = objPtr->devicePtr->errorProc(objPtr, init_res);
      } else {
	  errStr = "Error with device initialization.";
      }

      Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
      return TCL_ERROR;
    }

    objPtr->ready = 1;

    /*
     * If queue is not empty, start a decode.
     */

    if (objPtr->queue != NULL) {
      MPEG_Play_DoDecode(objPtr);
    }

    /*
     * Reset stat stuff.
     */

    objPtr->statStartTime = Tcm_ReadSysClock();
    objPtr->missed = objPtr->played = objPtr->late = objPtr->bytes = 0;

    return TCL_OK;
}


/************************************************************
 *
 * MPEG_Play_LTSChanged
 *
 * Parameters:
 *     clientdata - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *      Called when the LTS changes value.  Dequeues any previously
 *      queued Play requests, frees any memory we're holding, and
 *      returns.
 *
 ************************************************************/
static void
MPEG_Play_LTSChanged(ClientData clientData)
{
    MPEGPlay *objPtr = (MPEGPlay *) clientData;
    int i;

    /*
     * If not ready, do nothing.
     */
    if (!objPtr->ready) {
	return;
    }

    /*
     * Blow away "at" queue.
     */
    for (i=0; i<objPtr->num_at_tokens; i++) {
	if (objPtr->at_tokens[i] > 0) {
	    AtEvent *at_event;
	    int token;

	    token = objPtr->at_tokens[i];
	    at_event = Tcm_FindAtEvent(token);
	    if (token && (at_event != NULL)) {
		MPEG_Play_FreeFrameNow((MPEGFrame *) at_event->fireClientData);
		Tcm_DeleteAtHandler(token);
		objPtr->at_tokens[i] = 0;
	    }
	}
    }

    /*
     * Blow away my frame queues.
     */
    if (objPtr->queue != NULL) {
	FreeMPEGFrameList (objPtr->queue);
	objPtr->queue = NULL;
    }
    if (objPtr->refQueue != NULL) {
	FreeMPEGFrameList (objPtr->refQueue);
	objPtr->refQueue = NULL;
    }
    if (objPtr->waitQueue != NULL) {
	FreeMPEGFrameList (objPtr->waitQueue);
	objPtr->waitQueue = NULL;
    }

    if (objPtr->inDecodeFrame != NULL) {
	Tcm_CancelIdleCall("video", (Tk_IdleProc *) MPEG_Play_ContinueDecode, 
			   objPtr->inDecodeFrame);
	MPEG_Play_FreeFrameNow(objPtr->inDecodeFrame);
	objPtr->inDecodeFrame = NULL;
    }

    /*
     * Reset all the statistics stuff.
     */
    objPtr->statStartTime = Tcm_ReadSysClock();
    objPtr->missed = objPtr->late = objPtr->played = objPtr->bytes = 0;
}

/************************************************************
 *
 * MPEG_Play_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
MPEG_Play_Accept(MPEGPlay *objPtr, SB_List *sbl)
{
    int i, j;
    MPEGFrame *newFrame, *framePtr, *tmpFramePtr;
    ScatterBuffer *sbPtr;
    char *dataPtr;
    Buffer *dummyBufferPtr;
    
    if (!objPtr->ready) {
	return TCL_OK;
    }

    if (sbl == NULL) return TCL_OK;
    if (sbl->count < 1) return TCL_OK;

    /*
     * Increment reference count on buffers passed in
     */
    AttachToAllBuffers(objPtr->interp, sbl,
			"tcmMPEGPlay:MPEG_Play_Accept");

    /*
     * For each frame in sb list, create frame structure, fill in
     * info and add to queue.
     */

    for (i=0; i<sbl->count; i++) {
	newFrame = (MPEGFrame *) ckalloc(sizeof(MPEGFrame));

	sbPtr = sbl->sbs[i];

	dataPtr = GetFromSB(sbPtr, CM_MPEG_BUF_HDR, NULL, NULL);
	if (dataPtr == NULL) {
	    ckfree((char *) newFrame);
	} else {
	    memcpy(&(newFrame->frameHeader), dataPtr-8, sizeof(MPEGBufferHeader));
	    newFrame->frameHeader.hdrType = 
	      ntohl(newFrame->frameHeader.hdrType);
	    newFrame->frameHeader.length = 
	      ntohl(newFrame->frameHeader.length);
	    newFrame->frameHeader.width = 
	      ntohl(newFrame->frameHeader.width);
	    newFrame->frameHeader.height = 
	      ntohl(newFrame->frameHeader.height);
	    newFrame->frameHeader.type = 
	      ntohl(newFrame->frameHeader.type);
	    newFrame->frameHeader.ref1 = 
	      ntohl(newFrame->frameHeader.ref1);
	    newFrame->frameHeader.ref2 = 
	      ntohl(newFrame->frameHeader.ref2);
	    newFrame->frameHeader.intra_matrix_flag = 
	      ntohl(newFrame->frameHeader.intra_matrix_flag);
	    newFrame->frameHeader.non_intra_matrix_flag = 
	      ntohl(newFrame->frameHeader.non_intra_matrix_flag);
	    newFrame->frameHeader.index = 
	      ntohl(newFrame->frameHeader.index);
	    newFrame->frameHeader.logStart_s = 
	      ntohl(newFrame->frameHeader.logStart_s);
	    newFrame->frameHeader.logStart_ns = 
	      ntohl(newFrame->frameHeader.logStart_ns);
	    newFrame->frameHeader.logEnd_s = 
	      ntohl(newFrame->frameHeader.logEnd_s);
	    newFrame->frameHeader.logEnd_ns = 
	      ntohl(newFrame->frameHeader.logEnd_ns);
	    newFrame->frameHeader.displayFlag = 
	      ntohl(newFrame->frameHeader.displayFlag);
	    
	    newFrame->logStart = ((double) newFrame->frameHeader.logStart_s) +
	      ((double) newFrame->frameHeader.logStart_ns) / ONE_BILLION;
	    newFrame->logEnd = ((double) newFrame->frameHeader.logEnd_s) +
	      ((double) newFrame->frameHeader.logEnd_ns) / ONE_BILLION;

	    dataPtr = GetFromSB(sbPtr, CM_DATA_SEGMENT, &(newFrame->dataBuffer),
				&newFrame->dataLength);    
	    if (dataPtr == NULL) {
		ckfree((char *) newFrame);
	    } else {
		newFrame->sb = CopyScatterBuffer(sbPtr);
		newFrame->data = dataPtr;

		if (newFrame->frameHeader.intra_matrix_flag) {
		    dataPtr = GetFromSB(sbPtr, CM_MPEG_INTRA_MATRIX, NULL, NULL);
		    if (dataPtr == NULL) {
			newFrame->intra_matrix = NULL;
		    } else {
			newFrame->intra_matrix = (int *) ckalloc(sizeof(int)*64);
			for (j=0; j<64; j++) {
			    newFrame->intra_matrix[j] =
				ntohl(*((unsigned long *) dataPtr));
			    dataPtr += 4;
			}
		    }
		} else {
		    newFrame->intra_matrix = NULL;
		}
		
		if (newFrame->frameHeader.non_intra_matrix_flag) {
		    dataPtr = GetFromSB(sbPtr, CM_MPEG_NON_INTRA_MATRIX, NULL, NULL);
		    if (dataPtr == NULL) {
			newFrame->non_intra_matrix = NULL;
		    } else {
			newFrame->non_intra_matrix = (int *) ckalloc(sizeof(int)*64);
			for (j=0; j<64; j++) {
			    newFrame->non_intra_matrix[j] =
				ntohl(*((unsigned long *) dataPtr));
			    dataPtr += 4;
			}
		    }
		} else {
		    newFrame->non_intra_matrix = NULL;
		}

		/* 
		 * See if necessary reference frames are here and set
		 * pointers accordingly.
		 */

		newFrame->ref1 = newFrame->ref2 = NULL;

		if (newFrame->frameHeader.type != MPEG_I_FRAME) {
		    /*
		     * Needs ref1 at least.
		     */

		    framePtr = objPtr->refQueue;
		    while (framePtr != NULL) {
			if (newFrame->frameHeader.ref1 == framePtr->frameHeader.index) {
			    break;
			}
			framePtr = framePtr->next;
		    }
		    newFrame->ref1 = framePtr;
		    if (framePtr != NULL) 
		      framePtr->refCount++;

		    if (newFrame->frameHeader.type == MPEG_B_FRAME) {
			/*
			 * Needs ref2 as well.
			 */
			
			framePtr = objPtr->refQueue;
			while (framePtr != NULL) { 
			    if (newFrame->frameHeader.ref2 == 
				framePtr->frameHeader.index) {
				break;
			    }
			    framePtr = framePtr->next;
			}
			newFrame->ref2 = framePtr;
			if (framePtr != NULL) 
			  framePtr->refCount++;
		    }
		}

		newFrame->deviceData = NULL;
		newFrame->at_token = -1;
		newFrame->objPtr = objPtr;
		newFrame->refCount = 0;
		newFrame->onRefQueue = 0;
		newFrame->freeFlag = 0;

		/*
		 * Generate FRAME_RECEIVED event.
		 */

		objPtr->bindFramePtr = newFrame;
		ExecuteCMBinding(objPtr->interp, objPtr->bindToken, "<FRAME_RECEIVED>");

		/* 
		 * Now put on appropriate queue.
		 */

		if (((newFrame->frameHeader.type != MPEG_I_FRAME) &&
		    (newFrame->ref1 == NULL)) ||
		    ((newFrame->frameHeader.type == MPEG_B_FRAME) &&
		     (newFrame->ref2 == NULL))) {
		    newFrame->next = objPtr->waitQueue;
		    newFrame->prev = NULL;
		    if (objPtr->waitQueue != NULL) {
		      objPtr->waitQueue->prev = newFrame;
		    }
		    objPtr->waitQueue = newFrame;

#if 0
		    {
		      char logStr[100];
		      sprintf(logStr, "Play Frame Accept WaitQueue %x %g %g", newFrame, newFrame->logStart, newFrame->logEnd);
		      Tcm_Log(logStr);
		    }
#endif

		} else {
		    double now, speed;

		    now = LTS_Value(objPtr->lts);
		    speed = LTS_Speed(objPtr->lts);

		    
		    /*
		     * Clean out decode queue of old stuff.
		     */
		    
		    framePtr = objPtr->queue;
		    if (speed < 0.0) {
			while ((framePtr != NULL) && (framePtr->logEnd > now)) {
			    tmpFramePtr = framePtr->next;
			    framePtr->prev = framePtr->next = NULL;
			    MPEG_Play_MissDisplay(framePtr);
			    framePtr = tmpFramePtr;
			}
		    } else if (speed > 0.0) {
			while ((framePtr != NULL) && (framePtr->logEnd < now)) {
			    tmpFramePtr = framePtr->next;
			    framePtr->prev = framePtr->next = NULL;
			    MPEG_Play_MissDisplay(framePtr);
			    framePtr = tmpFramePtr;
			}
		    }

		    objPtr->queue = framePtr;
		    if (objPtr->queue != NULL) 
		      objPtr->queue->prev = NULL;
		    
		    /*
		     * Now do insertion.
		     */
		    
		    MPEG_Play_InsertIntoDecodeQueue(objPtr, newFrame);		   

#if 0
		    {
		      char logStr[100];
		      sprintf(logStr, "Play Frame Accept DecodeQueue %x %g %g", newFrame, newFrame->logStart, newFrame->logEnd);
		      Tcm_Log(logStr);
		    }
#endif

		}   
	    }
	}
    }
    
    /*
     * If no decode in process, start a decode
     */
    if (objPtr->inDecodeFrame == NULL) {
	MPEG_Play_DoDecode(objPtr);
    }

    return TCL_OK;
}

/************************************************************
 *
 * MPEG_Play_Unready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_Play_Unready(MPEGPlay * objPtr)
{
  int i;

  objPtr->ready = 0;

  /*
   * Blow away "at" queue.
   */
  for (i =0; i<objPtr->num_at_tokens; i++) {
      if (objPtr->at_tokens[i] > 0) {
	  AtEvent *at_event;
	  int token;

	  token = objPtr->at_tokens[i];
	  at_event = Tcm_FindAtEvent(token);
	  if (token && (at_event != NULL)) {
	      MPEG_Play_FreeFrameNow((MPEGFrame *) at_event->fireClientData);
	      Tcm_DeleteAtHandler(token);
	      objPtr->at_tokens[i] = 0;
	  }
      }
  }
  
  /*
   * Blow away my frame queues.
   */
  if (objPtr->queue != NULL) {
      FreeMPEGFrameList (objPtr->queue);
      objPtr->queue = NULL;
  }
  if (objPtr->refQueue != NULL) {
      FreeMPEGFrameList (objPtr->refQueue);
      objPtr->refQueue = NULL;
  }
  if (objPtr->waitQueue != NULL) {
      FreeMPEGFrameList (objPtr->waitQueue);
      objPtr->waitQueue = NULL;
  }

  if (objPtr->inDecodeFrame != NULL) {
      Tcm_CancelIdleCall("video", (Tk_IdleProc *) MPEG_Play_ContinueDecode, 
			 objPtr->inDecodeFrame);
      MPEG_Play_FreeFrameNow(objPtr->inDecodeFrame);
      objPtr->inDecodeFrame = NULL;
  }  

  return TCL_OK;
}

/************************************************************
 *
 * MPEG_Play_Stats
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_Play_Stats(MPEGPlay * objPtr)
{
  return TCL_OK;
}

/************************************************************
 *
 * MPEG_Play_MissDisplay
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_Play_MissDisplay(MPEGFrame * framePtr)
{
  MPEGPlay *objPtr;

  objPtr = framePtr->objPtr;

  /*
   * Collect stats.
   */

  objPtr->missed++;

  /*
   * Generate FRAME_DROPPED event.
   */

  objPtr->bindFramePtr = framePtr;
  ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		   "<FRAME_DROPPED>");

#if 0
  {
    char logStr[100];
    sprintf(logStr, "Play Frame Missed Missed %x %g %g", framePtr, 
	    framePtr->logStart, framePtr->logEnd);
    Tcm_Log(logStr);
  }
#endif

  MPEG_Play_FreeFrame(framePtr);

  return TCL_OK;
}

/************************************************************
 *
 * MPEG_Play_DoDisplay
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_Play_DoDisplay(MPEGFrame *framePtr)
{
  MPEGPlay *objPtr;
  static MPEGFrame *delayedFramePtr = NULL;

  objPtr = framePtr->objPtr;

  /*
   * Collect stats.
   */
  objPtr->played++;

  /* Check to see if what we're decoding needs to be
     displayed before what we're about to display.
   */

  if (objPtr->inDecodeFrame != NULL) {

      if (framePtr->logStart < framePtr->logEnd) {
	  if (objPtr->inDecodeFrame->logStart < framePtr->logStart) {
	      /* Problem detected here. */

	      if (delayedFramePtr != NULL) {
		  /* Frame already waiting. Drop it and wait on this one. */

		  MPEG_Play_MissDisplay(delayedFramePtr);
	      }
	      delayedFramePtr = framePtr;
	      return TCL_OK;
	  } 
      } else {
	  if (objPtr->inDecodeFrame->logStart > framePtr->logStart) {
	      /* Problem detected here. */
	      if (delayedFramePtr != NULL) {
		  /* Frame already waiting. Drop it and wait on this one. */
		  
		  MPEG_Play_MissDisplay(delayedFramePtr);
	      }
	      delayedFramePtr = framePtr;
	      return TCL_OK;
	  }
      }
  }

      
  /*
   * Call device's display procedure.
   */
  if (objPtr->devicePtr != NULL) {
      if (objPtr->devicePtr->displayProc != NULL) {
	  objPtr->devicePtr->displayProc(framePtr);
      }
  } 

  /*
   * Generate FRAME_PLAYED event.
   */

  objPtr->bindFramePtr = framePtr;
  ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		   "<FRAME_PLAYED>");

#if 0
  {
    char logStr[100];
    sprintf(logStr, "Play Frame Display Display %x %g %g", framePtr, 
	    framePtr->logStart, framePtr->logEnd);
    Tcm_Log(logStr);
  }
#endif

  MPEG_Play_FreeFrame(framePtr);


  /* Play delayed frame if it exists. */

  if (delayedFramePtr != NULL) {
      if (objPtr->devicePtr != NULL) {
	  if (objPtr->devicePtr->displayProc != NULL) {
	      objPtr->devicePtr->displayProc(delayedFramePtr);
	  }
      } 

      /*
       * Generate FRAME_PLAYED event.
       */
      
      objPtr->bindFramePtr = delayedFramePtr;
      ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		       "<FRAME_PLAYED>");

      MPEG_Play_FreeFrame(delayedFramePtr);

      delayedFramePtr = NULL;
  }

  return TCL_OK;
}


/************************************************************
 *
 * MPEG_Play_FreeFrameNow
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Does the actual freeing of frames irregardless of 
 *     anything. Does maintain ref queue if onRefQueue flag
 *     is set.
 *
 ************************************************************/
static void
MPEG_Play_FreeFrameNow(MPEGFrame *framePtr) {
    int i;

    /*
     * Call device frame freeer.
     */
    if (framePtr->objPtr->devicePtr->freeFrameProc != NULL) {
	framePtr->objPtr->devicePtr->freeFrameProc(framePtr);
    }

#if 0
    {
      char logStr[100];
      sprintf(logStr, "Play Frame Free Now %x %g %g", framePtr, 
	      framePtr->logStart, framePtr->logEnd);
      Tcm_Log(logStr);
    }
#endif
    
    /*
     * Detach from buffer.
     */
    
    for (i=0; i<framePtr->sb->count; i++) {
	DetachBuffer(framePtr->objPtr->interp, framePtr->sb->buffers[i],
		     "tcmMPEGPlay:MPEG_Play_FreeFrame");
    }
	
    FreeScatterBuffer(framePtr->sb);
    framePtr->sb = NULL;
    
    if (framePtr->intra_matrix != NULL) {
	ckfree((char *) framePtr->intra_matrix);
    }
    if (framePtr->non_intra_matrix != NULL) {
	ckfree((char *) framePtr->non_intra_matrix);
    }
    
    if (framePtr->at_token >= 0) {
	framePtr->objPtr->at_tokens[framePtr->at_token] = 0;
    }

    if (framePtr->onRefQueue) {
	if (framePtr->prev == NULL) {
	    framePtr->objPtr->refQueue = framePtr->next;
	    if (framePtr->next != NULL) {
		framePtr->next->prev = NULL;
	    }
	} else {
	    framePtr->prev->next = framePtr->next;
	    if (framePtr->next != NULL) {
		framePtr->next->prev = framePtr->prev;
	    }
	}
    }

    ckfree((char *) framePtr);
}
    
/************************************************************
 *
 * MPEG_Play_FreeFrame
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static void
MPEG_Play_FreeFrame(MPEGFrame *framePtr)
{
    int i, new_num;
    double speed, now;
    
    /*
     * Free immediately only if its a B frame.
     */

    speed = LTS_Speed(framePtr->objPtr->lts);
    now = LTS_Value(framePtr->objPtr->lts);

#if 0
    {
      char logStr[100];
      sprintf(logStr, "Play Frame Free First %x %g %g", framePtr, 
	      framePtr->logStart, framePtr->logEnd);
      Tcm_Log(logStr);
    }
#endif

    if (framePtr->frameHeader.type == MPEG_B_FRAME) {
	
	/*
	 * Reduce ref counts of ref. frames.
	 */

	if (framePtr->ref1 != NULL) {
	    framePtr->ref1->refCount--;
	    if ((framePtr->ref1->refCount <= 0) && (framePtr->ref1->freeFlag)) {
		MPEG_Play_FreeFrameNow(framePtr->ref1);
	    }
	}

	if (framePtr->ref2 != NULL) {
	    framePtr->ref2->refCount--;
	    if ((framePtr->ref2->refCount <= 0) && (framePtr->ref2->freeFlag)) {
		MPEG_Play_FreeFrameNow(framePtr->ref2);
	    }
	}
	framePtr->ref1 = framePtr->ref2 = NULL;

	MPEG_Play_FreeFrameNow(framePtr);

    } else {
	if (framePtr->frameHeader.type == MPEG_P_FRAME) {
	    if (framePtr->ref1 != NULL) {
		framePtr->ref1->refCount--;
		if ((framePtr->ref1->refCount <= 0) && (framePtr->ref1->freeFlag)) {
		    MPEG_Play_FreeFrameNow(framePtr->ref1);
		}
	    }
	    framePtr->ref1 = NULL;
	}

	if (((speed < 0.0) && (framePtr->logEnd < now)) ||
	    ((speed > 0.0) && (framePtr->logEnd > now))) {
	    double t1, t2;
	    MPEGPlay *objPtr = framePtr->objPtr;


	    /* 
	     * Must be a reference frame and thus on some queue.
	     * Schedule free to occur only after system endTime has passed.
	     */

	    if (objPtr->num_at_tokens == 0) {
		objPtr->num_at_tokens = 30;
		objPtr->at_tokens = (int *) ckalloc(sizeof(int) * 30);
		for (i=0; i<objPtr->num_at_tokens; i++) {
		    objPtr->at_tokens[i] = 0;
		}
	    } 
	    for (i=0; i<objPtr->num_at_tokens; i++) {
		if (objPtr->at_tokens[i] == 0) {
		    break;
		}
	    }
	    if (i == objPtr->num_at_tokens) {
		new_num = objPtr->num_at_tokens + 10;
		objPtr->at_tokens = (int *) realloc((char *) objPtr->at_tokens,
						sizeof(int) * new_num);
		for (i=objPtr->num_at_tokens; i < new_num; i++) {
		    objPtr->at_tokens[i] = 0;
		}
		i = objPtr->num_at_tokens;
		objPtr->num_at_tokens = new_num;
	    }
	
	    t1 = LTS_SystemTimeOf(objPtr->lts, framePtr->logStart);
	    t2 = LTS_SystemTimeOf(objPtr->lts, framePtr->logEnd);
	    objPtr->at_tokens[i] = 
	      Tcm_CreateAtHandler(t1>t2?t1:t2, HUGE_VAL,
				  "video",
				  (Tk_TimerProc *) MPEG_Play_FreeFrame, 
				  (ClientData) framePtr, NULL,
				  NULL, (Tk_IdleProc *) NULL,
				  (ClientData) NULL, NULL);
	    framePtr->at_token = i;
	} else if (framePtr->refCount > 0) {
	    framePtr->freeFlag = 1;
	} else {
	    MPEG_Play_FreeFrameNow(framePtr);
	}
    }

    return;
}

/************************************************************
 *
 * MPEG_Play_GetEligibleFrame
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Acgtion:
 *
 ************************************************************/

static MPEGFrame *
MPEG_Play_GetEligibleFrame(MPEGPlay *objPtr, MPEGFrame *queue)
{
    MPEGFrame *framePtr;
    double now, speed;

    if (queue == NULL) return NULL;

    now = LTS_Value(objPtr->lts);
    speed = LTS_Speed(objPtr->lts);

    framePtr = queue;

    while (framePtr != NULL) {
	
	if (speed < 0.0) {
	    if (framePtr->logEnd < now) break;
	} else if (speed > 0.0) {
	    if (framePtr->logEnd > now) break;
	} else {
	    if (framePtr->logEnd == now) break;
	}

	framePtr = framePtr->next;
    }

    /*
     * Gather stats.
     */

    return framePtr;
}

/************************************************************
 *
 * MPEG_Play_Refresh
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
MPEG_Play_Refresh(MPEGPlay * objPtr)
{
  int res;
  char *errStr;

  /*
   * If not ready, do nothing.
   */

  if (!objPtr->ready) return TCL_OK;

  /*
   * If device has refresh, call it.
   */

  if (objPtr->devicePtr != NULL) {
    if (objPtr->devicePtr->refreshProc != NULL) {
      res = objPtr->devicePtr->refreshProc(objPtr);
      if (res != 0) {
	errStr = "Error refreshing picture.";
	if (objPtr->devicePtr->errorProc != NULL) {
	  errStr = objPtr->devicePtr->errorProc(objPtr, res);
	}
	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
      }
    }
  }

  return TCL_OK;
}

/************************************************************
 *
 * MPEG_Play_DecodeDone
 *
 * Parameters:
 *     framePtr - pointer to frame just finished with decoding.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
int
MPEG_Play_DecodeDone(MPEGFrame *framePtr)
{
    int i, new_num;
    MPEGPlay *objPtr;
    MPEGFrame *tmpFramePtr, *tmpNextFramePtr;

    objPtr = framePtr->objPtr;

    /*
     * Reset in decode
     */
    objPtr->inDecodeFrame = NULL;

    /*
     * If not ready, free frame and return.
     */
    if (!objPtr->ready) {
	MPEG_Play_FreeFrameNow(framePtr);
	return TCL_OK;
    }

    /*
     * Gather stats.
     */
    objPtr->bytes += framePtr->dataLength;

    /*
     * Generate FRAME_DECODE_DONE event.
     */

    objPtr->bindFramePtr = framePtr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		     "<FRAME_DECODE_DONE>");

#if 0
    {
      char logStr[100];
      sprintf(logStr, "Play Frame Decode Done %x %g %g", framePtr, framePtr->logStart, framePtr->logEnd);
      Tcm_Log(logStr);
    }
#endif

    /*
     * If frame is an I or P frame, put it on the refQueue.
     */

    if (framePtr->frameHeader.type != MPEG_B_FRAME) {
	double now, speed;

	now = LTS_Value(objPtr->lts);
	speed = LTS_Speed(objPtr->lts);

	framePtr->next = objPtr->refQueue;
	framePtr->prev = NULL;
	if (objPtr->refQueue != NULL) {
	  objPtr->refQueue->prev = framePtr;
	}
	objPtr->refQueue = framePtr;
	framePtr->onRefQueue = 1;


	/*
	 * Scan the wait queue and transfer things to the decode queue
	 * if this reference frame was the one being waited for.
	 * Take this opportunity to throw away anything old in wait queue.
	 */

	tmpFramePtr = objPtr->waitQueue;
	while (tmpFramePtr != NULL) {
	    if ((speed < 0.0) && (tmpFramePtr->logEnd > now)) {
		if (tmpFramePtr->prev == NULL) {
		    objPtr->waitQueue = tmpFramePtr->next;
		    if (tmpFramePtr->next != NULL) {
			tmpFramePtr->next->prev = NULL;
		    }
		    MPEG_Play_MissDisplay(tmpFramePtr);
		    tmpFramePtr = objPtr->waitQueue;
		} else {
		    tmpFramePtr->prev->next = tmpFramePtr->next;
		    if (tmpFramePtr->next != NULL) {
			tmpFramePtr->next->prev = tmpFramePtr->prev;
		    }
		    tmpNextFramePtr = tmpFramePtr->next;
		    MPEG_Play_MissDisplay(tmpFramePtr);
		    tmpFramePtr = tmpNextFramePtr;
		}
	    } else if ((speed > 0.0) && (tmpFramePtr->logEnd < now)) {
		if (tmpFramePtr->prev == NULL) {
		    objPtr->waitQueue = tmpFramePtr->next;
		    if (tmpFramePtr->next != NULL) {
			tmpFramePtr->next->prev = NULL;
		    }
		    MPEG_Play_MissDisplay(tmpFramePtr);
		    tmpFramePtr = objPtr->waitQueue;
		} else {
		    tmpFramePtr->prev->next = tmpFramePtr->next;
		    if (tmpFramePtr->next != NULL) {
			tmpFramePtr->next->prev = tmpFramePtr->prev;
		    }
		    tmpNextFramePtr = tmpFramePtr->next;
		    MPEG_Play_MissDisplay(tmpFramePtr);
		    tmpFramePtr = tmpNextFramePtr;
		}
	    } else {
		if (tmpFramePtr->frameHeader.ref1 == framePtr->frameHeader.index) {
		    tmpFramePtr->ref1 = framePtr;
		    framePtr->refCount++;
		}
		if (tmpFramePtr->frameHeader.ref2 == framePtr->frameHeader.index) {
		    tmpFramePtr->ref2 = framePtr;
		    framePtr->refCount++;
		}

		tmpNextFramePtr = tmpFramePtr->next;
		if ((tmpFramePtr->frameHeader.type == MPEG_P_FRAME) &&
		    (tmpFramePtr->ref1 != NULL)) {
		    
		    if (tmpFramePtr->prev == NULL) {
			objPtr->waitQueue = tmpFramePtr->next;
		    } else {
			tmpFramePtr->prev->next = tmpFramePtr->next;
		    }
		    if (tmpFramePtr->next != NULL) {
			tmpFramePtr->next->prev = tmpFramePtr->prev;
		    }
		    tmpFramePtr->next = tmpFramePtr->prev = NULL;

		    MPEG_Play_InsertIntoDecodeQueue(objPtr, tmpFramePtr);
#if 0
		    {
		      char logStr[100];
		      sprintf(logStr, "Play Frame Waiting DecodeQueue %x %g %g", tmpFramePtr, 
			      tmpFramePtr->logStart, tmpFramePtr->logEnd);
		      Tcm_Log(logStr);
		    }
#endif

		} else if ((tmpFramePtr->frameHeader.type == MPEG_B_FRAME) &&
			   (tmpFramePtr->ref1 != NULL) &&
			   (tmpFramePtr->ref2 != NULL)) {

		    if (tmpFramePtr->prev == NULL) {
			objPtr->waitQueue = tmpFramePtr->next;
		    } else {
			tmpFramePtr->prev->next = tmpFramePtr->next;
		    }
		    if (tmpFramePtr->next != NULL) {
			tmpFramePtr->next->prev = tmpFramePtr->prev;
		    }
		    tmpFramePtr->next = tmpFramePtr->prev = NULL;

		    MPEG_Play_InsertIntoDecodeQueue(objPtr, tmpFramePtr);
#if 0
		    {
		      char logStr[100];
		      sprintf(logStr, "Play Frame Waiting DecodeQueue %x %g %g", tmpFramePtr, 
			      tmpFramePtr->logStart, tmpFramePtr->logEnd);
		      Tcm_Log(logStr);
		    }
#endif

		} 

		tmpFramePtr = tmpNextFramePtr;
	    }
	}						 
    }


    /*
     * Check display flag.
     */

    if (framePtr->frameHeader.displayFlag) {
	/*
	 * If speed's zero and we're in the window, throw up the current
	 * frame.  Otherwise, create an atEvent for it.
	 */
	if (LTS_Speed(objPtr->lts) == 0.0) {
	    MPEG_Play_DoDisplay(framePtr);
	} else {
	    double t1, t2;
	    
	    /*
	     * Schedule display.  Search for a free at_token.  If we can't find
	     * one, to increase size of at_token array
	     */
	    if (objPtr->num_at_tokens == 0) {
		objPtr->num_at_tokens = 30;
		objPtr->at_tokens = (int *) ckalloc(sizeof(int) * 30);
		for (i=0; i<objPtr->num_at_tokens; i++) {
		    objPtr->at_tokens[i] = 0;
		}
	    } 
	    for (i=0; i<objPtr->num_at_tokens; i++) {
		if (objPtr->at_tokens[i] == 0) {
		    break;
		}
	    }
	    if (i == objPtr->num_at_tokens) {
		new_num = objPtr->num_at_tokens + 10;
		objPtr->at_tokens = (int *) realloc((char *) objPtr->at_tokens,
						    sizeof(int) * new_num);
		for (i=objPtr->num_at_tokens; i < new_num; i++) {
		    objPtr->at_tokens[i] = 0;
		}
		i = objPtr->num_at_tokens;
		objPtr->num_at_tokens = new_num;
	    }
	    
	    t1 = LTS_SystemTimeOf(objPtr->lts, framePtr->logStart);
	    t2 = LTS_SystemTimeOf(objPtr->lts, framePtr->logEnd);
	    objPtr->at_tokens[i] = 
	      Tcm_CreateAtHandler(t1<t2?t1:t2, HUGE_VAL,
				  "video", (Tk_TimerProc *) MPEG_Play_DoDisplay, 
				  (ClientData) framePtr, NULL,
				  NULL, (Tk_IdleProc *) NULL,
				  (ClientData) NULL, NULL);
	    framePtr->at_token = i;
	}
    } else {
	MPEG_Play_FreeFrame(framePtr);
    }

    /*
     * Arrange to decode next frame when we're idle
     */
    if (objPtr->queue != NULL) {
	Tcm_DoWhenIdle("video", (Tk_IdleProc *) MPEG_Play_DoDecode, objPtr); 
    }

    return TCL_OK;
}

/************************************************************
 *
 * MPEG_Play_DoDecode
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
MPEG_Play_DoDecode(MPEGPlay *objPtr)
{
    MPEGFrame *frame_ptr;

    /* 
     * If not ready or alraady in decode, do nothing.
     */
    if ((!objPtr->ready) || (objPtr->inDecodeFrame != NULL)) {
	return TCL_OK;
    }

    /*
     * Get next elgible frame. Call device proc. if available.
     */

    frame_ptr = NULL;

    if (objPtr->devicePtr->nextFrameProc != NULL) {
	objPtr->devicePtr->nextFrameProc(objPtr, objPtr->queue, &frame_ptr);
    } else {
	frame_ptr = MPEG_Play_GetEligibleFrame(objPtr, objPtr->queue);
    }

    if (frame_ptr == NULL) {
	return TCL_OK;
    }

#if 0
    {
      char logStr[100];
      sprintf(logStr, "Play Frame Decode Chosen %x %g %g", frame_ptr, frame_ptr->logStart, frame_ptr->logEnd);
      Tcm_Log(logStr);
    }
#endif 

    if (frame_ptr->prev == NULL) {
	if (frame_ptr->next == NULL) {
	    objPtr->queue = NULL;
	} else {
	    frame_ptr->next->prev = NULL;
	    objPtr->queue = frame_ptr->next;
	}
    } else {
	frame_ptr->prev->next = frame_ptr->next;
	if (frame_ptr->next != NULL) {
	    frame_ptr->next->prev = frame_ptr->prev;
	}
    }

    frame_ptr->prev = frame_ptr->next = NULL;

    /*
     * Record the fact that we're decoding a frame.
     */
    objPtr->inDecodeFrame = frame_ptr;

    /*
     * Generate FRAME_DECODE_BEGIN event.
     */

    objPtr->bindFramePtr = frame_ptr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		     "<FRAME_DECODE_BEGIN>");

    /*
     * Start the decoding loop
     */
    MPEG_Play_ContinueDecode (frame_ptr);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * MPEG_Play_ContinueDecode --
 *
 *	Continues an asynchronous decode until it's finished.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	May schedule a frame for disPlay if the decoding is finished,
 *	otherwise, schedules itself to be called again when processor
 *	is idle.
 *
 *--------------------------------------------------------------
 */
static void
MPEG_Play_ContinueDecode (MPEGFrame *framePtr)
{
    MPEGPlay *objPtr;

    objPtr = framePtr->objPtr;

    /* 
     * If not ready or alraady in decode, free the frame and do nothing.
     */
    if ((!objPtr->ready) || (objPtr->inDecodeFrame == NULL)) {
	MPEG_Play_FreeFrameNow(framePtr);
	objPtr->inDecodeFrame = NULL;
	return;
    }

    /*
     * Call device decode procedure.
     */
    if (objPtr->devicePtr != NULL) {
        if (objPtr->devicePtr->decodeProc != NULL) {
	    objPtr->devicePtr->decodeProc(framePtr);
        }
    }

    /*
     * If decoding is not finished, schedule another call.  Otherwise
     * we're done.
     */
    if (objPtr->devicePtr->asyncDecodeFlag) {
	if (objPtr->devicePtr->decodeProc != NULL) {
	    Tcm_DoWhenIdle("video", (Tk_IdleProc *) MPEG_Play_ContinueDecode,
			    framePtr);
	}
	return;
    }

    MPEG_Play_DecodeDone(framePtr);
    return;
}


/************************************************************
 *
 * MPEGPlayModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
MPEGPlayModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mpegPlay", MPEGPlayCreateCmd,
			NULL, (void (*)()) NULL);
    Tcl_GlobalEval(interp, "cminfo add media {video}");
    Tcl_GlobalEval(interp, "cminfo add formats video {mpeg}");
    Tcl_GlobalEval(interp, "cminfo add objects video mpeg mpegPlay");
    Tcl_VarEval(interp, "cminfo add object_info mpegPlay {play",
#ifdef MPEG_OPTIVIDEO
			" optivideo",
#endif
#ifdef MPEG_SW
			" software",
#endif
			"}", NULL);
    return TCL_OK;
}

/**********************************************************************
 *
 * MPEG_Play_InsertIntoDecodeQueue
 *
 * Parameters:
 *     objPtr - pointer to MPEG Play object.
 *     framePtr - pointer to frame to be inserted in decode queue.
 *
 * Results:
 *     None.
 * 
 * Actions:
 *     framePtr is inserted into decode queue sorted first by start time and
 *     then by end time.
 *
 **********************************************************************/ 

static void
MPEG_Play_InsertIntoDecodeQueue(MPEGPlay *objPtr, MPEGFrame *framePtr)
{
    MPEGFrame *curFramePtr, *prevFramePtr;
    double speed;
    
    speed = LTS_Speed(objPtr->lts);
    
    if (objPtr->queue == NULL) {
	objPtr->queue = framePtr;
	framePtr->next = framePtr->prev = NULL;
    } else {
	prevFramePtr = NULL;
	curFramePtr = objPtr->queue;
	if (speed < 0.0) {
	    while (curFramePtr != NULL) {
		if (framePtr->logStart >= curFramePtr->logStart) {
		    if (framePtr->logEnd <= curFramePtr->logEnd) {
			break;
		    }
		}
		prevFramePtr = curFramePtr;
		curFramePtr = curFramePtr->next;
	    }
	} else {
	    while (curFramePtr != NULL) {
		if (framePtr->logStart <= curFramePtr->logStart) {
		    if (framePtr->logEnd >= curFramePtr->logEnd) {
			break;
		    }
		}
		prevFramePtr = curFramePtr;
		curFramePtr = curFramePtr->next;
	    }
	}
	if (curFramePtr == NULL) {
	    prevFramePtr->next = framePtr;
	    framePtr->prev = prevFramePtr;
	    framePtr->next = NULL;
	} else if (prevFramePtr == NULL) {
	    objPtr->queue = framePtr;
	    framePtr->prev = NULL;
	    framePtr->next = curFramePtr;
	    curFramePtr->prev = framePtr;
	} else {
	    curFramePtr->prev = framePtr;
	    framePtr->next = curFramePtr;
	    prevFramePtr->next = framePtr;
	    framePtr->prev = prevFramePtr;
	}
    } 
}

/******************************************************************
 *
 * FreeMPEGFrameList
 *
 * Parameters:
 *     framePtr - pointer to list to be freed.
 *
 * Results:
 *    None.
 *
 * Actions:
 *    Calls MPEG_Play_FreeFrameNow on each frame in list.
 *
 ******************************************************************/

static void
FreeMPEGFrameList (MPEGFrame *framePtr)
{
    MPEGFrame *curFramePtr;
    MPEGFrame *tmpFramePtr;

    curFramePtr = framePtr;

    while (curFramePtr != NULL) {
	tmpFramePtr = curFramePtr->next;
	MPEG_Play_FreeFrameNow(curFramePtr);
	curFramePtr = tmpFramePtr;
    }

    return;
}


/******************************************************************
 *
 * MPEG_Play_DeviceList
 *
 * Parameters:
 *     objPtr - pointer to mpeg play object.
 *
 * Results:
 *     Standard Tcl Result.
 *
 * Actions:
 *     Sets result to list of devices supported. Returns TCL_OK;
 *
 ******************************************************************/
static int
MPEG_Play_DeviceList(MPEGPlay *objPtr)
{
    Tcl_Interp *interp = objPtr->interp;
    Tcl_DString resStr;

    Tcl_DStringInit(&resStr);

#ifdef MPEG_SW_DEVICE
    Tcl_DStringAppendElement(&resStr, "software");
#endif
#ifdef MPEG_OPTIVIDEO_DEVICE
    Tcl_DStringAppendElement(&resStr, "optivideo");
#endif

    Tcl_SetResult(interp, Tcl_DStringValue(&resStr), TCL_VOLATILE);
    Tcl_DStringFree(&resStr);

    return TCL_OK;
}

static char *
GetLogStart(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%f", objPtr->bindFramePtr->logStart);

  return resStr;
}

static char *
GetLogEnd(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%f", objPtr->bindFramePtr->logEnd);

  return resStr;
}

static char *
GetDataLen(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->dataLength);

  return resStr;
}

static char *
GetCumulLenDecoded(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);
  sprintf(resStr, "%d", objPtr->bytes);
  objPtr->bytes = 0;
  return resStr;
}

static char *
GetWidth(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.width);

  return resStr;
}

static char *
GetHeight(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.height);

  return resStr;
}

static char *
GetType(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.type);

  return resStr;
}

static char *
GetRef1Index(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.ref1);

  return resStr;
}

static char *
GetRef2Index(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.ref2);

  return resStr;
}

static char *
GetIndex(Tcl_Interp *interp, ClientData clientdata)
{
  MPEGPlay *objPtr;
  char *resStr;

  objPtr = (MPEGPlay *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.index);

  return resStr;
}
