/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMPEGPriority.c,v 1.20 1996/02/06 22:42:03 kpatel Exp $
 *
 *  A simple CMT object (created using the mpegPriority command)
 *  that acts as a filter between the MPEGClipfile object and the
 *  pktSource object.  Its sole purpose is to reorder the buffers
 *  passed in into a priority order for the pktSource.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <sys/types.h>
#include <netinet/in.h>
#include "cmAtom.h"
#include "cmEvery.h"
#include "cmBufferlist.h"
#include "cmCookies.h"

#define MPEG_PRIORITY_GENERAL 0
#define MPEG_PRIORITY_I_OUT_CMD 1
#define MPEG_PRIORITY_P_OUT_CMD 2
#define MPEG_PRIORITY_B_OUT_CMD 3
#define MPEG_I_FRAME 1
#define MPEG_P_FRAME 2
#define MPEG_B_FRAME 3


/*
 * One of these structures is created for each mpegPriority object.
 */
typedef struct MPEG_Priority {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    SB_List *i_sbl;             /* SB List for I frames. */
    SB_List *p_sbl;             /* SB List for P frames. */
    SB_List *b_sbl;             /* SB List for B frames. */
    char *outCmd;		/* Command to send buffers to generally. */
    char *i_outCmd;             /* Command to send I frames to. */
    char *p_outCmd;             /* Command to send P frames to. */
    char *b_outCmd;             /* Command to send B frames to. */
    InFunc outCCmd;		/* C pointer to out command */
    InFunc i_outCCmd;		/* C pointer to I out command */
    InFunc p_outCCmd;		/* C pointer to P out command */
    InFunc b_outCCmd;		/* C pointer to B out command */
    ClientData outCD;		/* ClientData for outCCmd. */
    ClientData i_outCD;		/* ClientData for I outCCmd. */
    ClientData p_outCD;		/* ClientData for P outCCmd. */
    ClientData b_outCD;		/* ClientData for B outCCmd. */
    char *inCmd;                /* Command to get buffers from. */
    OutFunc inCCmd;             /* C pointer to inCmd. */
    ClientData inCD;            /* ClientData for inCCmd. */
    double cycleTime;           /* Cycle time for calling in and out commands. */
    int ready;                  /* Flag to indicate readiness. */
    char *everyHandle;          /* Handle for every event created if cycleTime is 
				   nonzero. */
    char *acceptHandle;
    char *sendHandle;
    char *i_sendHandle;
    char *p_sendHandle;
    char *b_sendHandle;
} MPEG_Priority;

static int nextMPEG_PriorityNum = 0;	/* Variable for next unique object
					 * name. */
/*
 * Procedures defined in this module
 */
int			MPEG_PriorityModuleInit(Tcl_Interp *interp);

static MPEG_Priority *	CreateNewMPEG_Priority(Tcl_Interp *interp,
			    char *name);
static int		DestroyMPEG_Priority(MPEG_Priority *objPtr);
static int		ConfigureMPEG_Priority(MPEG_Priority *objPtr,
			    int argc, char **argv);
static int		MPEG_PriorityCreateCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		MPEG_PriorityObjCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		MPEG_Priority_Accept(MPEG_Priority *objPtr,
			    SB_List *sbl);
static int		MPEG_Priority_InternalAccept(MPEG_Priority *objPtr,
			    SB_List *sbl);
static int		MPEG_Priority_Send(MPEG_Priority *objPtr,
			    SB_List **sblPtr);
static int		MPEG_Priority_I_Send(MPEG_Priority *objPtr,
			    SB_List **sblPtr);
static int		MPEG_Priority_P_Send(MPEG_Priority *objPtr,
			    SB_List **sblPtr);
static int		MPEG_Priority_B_Send(MPEG_Priority *objPtr,
			    SB_List **sblPtr);
static int		MPEG_Priority_ChangeOutCmd(MPEG_Priority *objPtr,
			    char *newOut, int typeFlag);
static int		ReadyMPEG_Priority(MPEG_Priority *objPtr);
static int		UnreadyMPEG_Priority(MPEG_Priority *objPtr);
static int		MPEG_Priority_ChangeInCmd(MPEG_Priority *objPtr,
			    char *newIn);
static void		DoCycleMPEG_Priority(ClientData clientData);

/************************************************************
 *
 * CreateNewMPEG_Priority
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

MPEG_Priority *
CreateNewMPEG_Priority(Tcl_Interp *interp, char *name)
{

    MPEG_Priority *new;
    char *acceptName;
    char *sendName;

    new = (MPEG_Priority *) ckalloc(sizeof(MPEG_Priority));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy (new->name, name);
    new->i_sbl = NULL;
    new->p_sbl = NULL;
    new->b_sbl = NULL;
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->i_outCmd = NULL;
    new->i_outCCmd = NULL;
    new->i_outCD = NULL;
    new->p_outCmd = NULL;
    new->p_outCCmd = NULL;
    new->p_outCD = NULL;
    new->b_outCmd = NULL;
    new->b_outCCmd = NULL;
    new->b_outCD = NULL;
    new->inCmd = NULL;
    new->inCCmd = NULL;
    new->inCD = NULL;
    new->ready = 0;
    new->cycleTime = 0.0;
    new->everyHandle = NULL;
    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+15);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) MPEG_Priority_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    /*
     * Register general and type specific sends as an atom.
     */

    sendName = (char *) ckalloc(strlen(name)+15);
    sprintf(sendName, "%s.%s", name, "send");
    Tutil_RegisterCommand(interp, sendName, (ClientData) MPEG_Priority_Send,
			  (ClientData) new);
    new->sendHandle = sendName;

    sendName = (char *) ckalloc(strlen(name)+15);
    sprintf(sendName, "%s.%s", name, "i_send");
    Tutil_RegisterCommand(interp, sendName, (ClientData) MPEG_Priority_I_Send,
			  (ClientData) new);
    new->i_sendHandle = sendName;

    sendName = (char *) ckalloc(strlen(name)+15);
    sprintf(sendName, "%s.%s", name, "p_send");
    Tutil_RegisterCommand(interp, sendName, (ClientData) MPEG_Priority_P_Send,
			  (ClientData) new);
    new->p_sendHandle = sendName;

    sendName = (char *) ckalloc(strlen(name)+15);
    sprintf(sendName, "%s.%s", name, "b_send");
    Tutil_RegisterCommand(interp, sendName, (ClientData) MPEG_Priority_B_Send,
			  (ClientData) new);
    new->b_sendHandle = sendName;

    return new;
}

/************************************************************
 *
 * DestroyMPEG_Priority
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

int
DestroyMPEG_Priority(MPEG_Priority *objPtr)
{
    /* 
     * Unready the object if readied.
     */

    if (objPtr->ready) {
	UnreadyMPEG_Priority(objPtr);
    }

    /*
     * Remove command name
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }

    /*
     * Unregister the send atoms.
     */
    if (objPtr->sendHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->sendHandle);
	ckfree(objPtr->sendHandle);
	objPtr->sendHandle = NULL;
    }
    if (objPtr->i_sendHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->i_sendHandle);
	ckfree(objPtr->i_sendHandle);
	objPtr->i_sendHandle = NULL;
    }
    if (objPtr->p_sendHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->p_sendHandle);
	ckfree(objPtr->p_sendHandle);
	objPtr->p_sendHandle = NULL;
    }
    if (objPtr->b_sendHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->b_sendHandle);
	ckfree(objPtr->b_sendHandle);
	objPtr->b_sendHandle = NULL;
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    if (objPtr->i_outCmd != NULL) {
	ckfree(objPtr->i_outCmd);
    }
    if (objPtr->p_outCmd != NULL) {
	ckfree(objPtr->p_outCmd);
    }
    if (objPtr->b_outCmd != NULL) {
	ckfree(objPtr->b_outCmd);
    }
    if (objPtr->inCmd != NULL) {
	ckfree(objPtr->inCmd);
    }

    if (objPtr->everyHandle != NULL) {
	Tcm_DeleteEvery(objPtr->interp, objPtr->everyHandle, NULL);
	objPtr->everyHandle = NULL;
    }

    ckfree((char *) objPtr);
    return TCL_OK;
}

/************************************************************
 *
 * ConfigureMPEG_Priority
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

int
ConfigureMPEG_Priority(MPEG_Priority * objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    
    interp = objPtr->interp;

    /*
     * Case 1: "<obj> configure"
     */
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}
	if (objPtr->i_outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-i_outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-i_outCmd {} {", objPtr->i_outCmd,
			     "}} ", (char *)NULL);
	}
	if (objPtr->p_outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-p_outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-p_outCmd {} {", objPtr->p_outCmd,
			     "}} ", (char *)NULL);
	}
	if (objPtr->b_outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-b_outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-b_outCmd {} {", objPtr->b_outCmd,
			     "}} ", (char *)NULL);
	}
	if (objPtr->inCmd == NULL) {
	    Tcl_AppendResult(interp, "{-inCmd {} {}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-inCmd {} {", objPtr->inCmd,
			     "}} ", (char *)NULL);
	}
	{
	    char cycleTimeStr[64];
	    
	    sprintf(cycleTimeStr, "%f", objPtr->cycleTime);
	    Tcl_AppendResult(interp, "{-cycleTime 0.0 ", cycleTimeStr, "}", 
			     (char *) NULL);
	}
	
	return TCL_OK;
    }

    /*
     * Case 2: "<obj> configure -slot"
     */
    if (argc == 1) {
	char c;
	int len;
	
	len = strlen(argv[0]);
	c = argv[0][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	else if ((c == 'i') && (strncmp("-i_outCmd", argv[0], len) == 0)) {
	    if (objPtr->i_outCmd == NULL) {
		Tcl_AppendResult(interp, "-i_outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-i_outCmd {} {", objPtr->i_outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	else if ((c == 'p') && (strncmp("-p_outCmd", argv[0], len) == 0)) {
	    if (objPtr->p_outCmd == NULL) {
		Tcl_AppendResult(interp, "-p_outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-p_outCmd {} {", objPtr->p_outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	else if ((c == 'b') && (strncmp("-b_outCmd", argv[0], len) == 0)) {
	    if (objPtr->b_outCmd == NULL) {
		Tcl_AppendResult(interp, "-b_outCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-b_outCmd {} {", objPtr->b_outCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	else if ((c == 'i') && (strncmp("-inCmd", argv[0], len) == 0)) {
	    if (objPtr->inCmd == NULL) {
		Tcl_AppendResult(interp, "-inCmd {} {}", (char *)NULL);
	    } else {
	      Tcl_AppendResult(interp, "-inCmd {} {", objPtr->inCmd,
			  "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	else if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
	    char cycleTimeStr[64];

	    sprintf(cycleTimeStr, "%g", objPtr->cycleTime);
	    Tcl_AppendResult(interp, "-cycleTime 0.0 ", cycleTimeStr, (char *) NULL);
	    return TCL_OK;
	}
	Tcl_AppendResult(interp, "bad configuration option: ",
			  argv[0], (char *)NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: "<obj> configure -slot value"
     * Since values may change, lots of side effects are buried in here.
     */

  
    if (objPtr->ready) {
	UnreadyMPEG_Priority(objPtr);
    }

    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (MPEG_Priority_ChangeOutCmd(objPtr, argv[cntr], MPEG_PRIORITY_GENERAL) 
		== TCL_ERROR) {
		return TCL_ERROR;
	    }
	} else if ((c == 'i') && (strncmp("-i_outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (MPEG_Priority_ChangeOutCmd(objPtr, argv[cntr], MPEG_PRIORITY_I_OUT_CMD) 
		== TCL_ERROR) {
		return TCL_ERROR;
	    }
	} else if ((c == 'p') && (strncmp("-p_outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (MPEG_Priority_ChangeOutCmd(objPtr, argv[cntr], MPEG_PRIORITY_P_OUT_CMD) 
		== TCL_ERROR) {
		return TCL_ERROR;
	    }
	} else if ((c == 'b') && (strncmp("-b_outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (MPEG_Priority_ChangeOutCmd(objPtr, argv[cntr], MPEG_PRIORITY_B_OUT_CMD) 
		== TCL_ERROR) {
		return TCL_ERROR;
	    }
	} else if ((c == 'i') && (strncmp("-inCmd", argv[cntr], len) == 0)) {
	    cntr++;

	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (MPEG_Priority_ChangeInCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	} else if ((c == 'c') && (strncmp("-cycleTime", argv[cntr], len) == 0)) {
	    double newCycleTime;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if ((Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) || 
		(newCycleTime < 0.0)) {
		Tcl_AppendResult(interp, "Bad value specified trying to set ",
				 "configuration option ", argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->cycleTime = newCycleTime;
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			  argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_PriorityCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the mpegResource command.
 *     A new object of type MPEG_Priority is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
MPEG_PriorityCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    MPEG_Priority *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextMPEG_PriorityNum++;
    sprintf(new_name, "%smpegPriority%d", argv[1], nextMPEG_PriorityNum);

    new = CreateNewMPEG_Priority(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, MPEG_PriorityObjCmd,
			(ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureMPEG_Priority(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * MPEG_PriorityObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     MPEG_Priority object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
MPEG_PriorityObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    MPEG_Priority *objPtr = (MPEG_Priority *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureMPEG_Priority(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyMPEG_Priority(objPtr);
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return ReadyMPEG_Priority(objPtr);
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return UnreadyMPEG_Priority(objPtr);
    } else if ((c == 's') && (strncmp("send", argv[1], len) == 0)) {
	SB_List *sb_list;
	char *sbStr;

	if (!objPtr->ready) {
	    sprintf(objPtr->interp->result, "{}");
	    return TCL_OK;
	}

	if (MPEG_Priority_Send(objPtr, &sb_list) != TCL_OK) {
	    Tcl_AppendResult(objPtr->interp, "Error in ", argv[0], " send.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	
	sbStr = MakeSBListString(sb_list);
	
	Tcl_SetResult(objPtr->interp, sbStr, TCL_VOLATILE); 
	ckfree(sbStr);
	ckfree((char *) sb_list);
	return TCL_OK;
    } else if ((c == 'i') && (strncmp("i_send", argv[1], len) == 0)) {
	SB_List *sb_list;
	char *sbStr;

	if (!objPtr->ready) {
	    sprintf(objPtr->interp->result, "{}");
	    return TCL_OK;
	}

	if (MPEG_Priority_I_Send(objPtr, &sb_list) != TCL_OK) {
	    Tcl_AppendResult(objPtr->interp, "Error in ", argv[0], " i_send.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	
	sbStr = MakeSBListString(sb_list);
	
	Tcl_SetResult(objPtr->interp, sbStr, TCL_VOLATILE); 
	ckfree(sbStr);
	ckfree((char *) sb_list);
	return TCL_OK;
    } else if ((c == 'p') && (strncmp("p_send", argv[1], len) == 0)) {
	SB_List *sb_list;
	char *sbStr;

	if (!objPtr->ready) {
	    sprintf(objPtr->interp->result, "{}");
	    return TCL_OK;
	}

	if (MPEG_Priority_P_Send(objPtr, &sb_list) != TCL_OK) {
	    Tcl_AppendResult(objPtr->interp, "Error in ", argv[0], " p_send.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	
	sbStr = MakeSBListString(sb_list);
	
	Tcl_SetResult(objPtr->interp, sbStr, TCL_VOLATILE); 
	ckfree(sbStr);
	ckfree((char *) sb_list);
	return TCL_OK;
    } else if ((c == 'b') && (strncmp("b_send", argv[1], len) == 0)) {
	SB_List *sb_list;
	char *sbStr;

	if (!objPtr->ready) {
	    sprintf(objPtr->interp->result, "{}");
	    return TCL_OK;
	}

	if (MPEG_Priority_B_Send(objPtr, &sb_list) != TCL_OK) {
	    Tcl_AppendResult(objPtr->interp, "Error in ", argv[0], " b_send.",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	
	sbStr = MakeSBListString(sb_list);
	
	Tcl_SetResult(objPtr->interp, sbStr, TCL_VOLATILE); 
	ckfree(sbStr);
	ckfree((char *) sb_list);
	return TCL_OK;
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl;
	int ret_val;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}

	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	  ckfree((char *) sbl);
	  Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
		(char *) NULL);
	  return TCL_ERROR;
	}

        ret_val = MPEG_Priority_Accept(objPtr, sbl);
	FreeSBList(sbl);
	return ret_val;
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ",
		argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
		"  ready\n",
		"  unready\n",
		"  send\n",
		"  i_send\n",
		"  p_send\n",
		"  b_send\n",
                "  accept <ScatterBufferList>\n",
                (char *) NULL);

    return TCL_ERROR;
}

/************************************************************
 *
 * MPEG_PriorityModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
MPEG_PriorityModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "mpegPriority", MPEG_PriorityCreateCmd,
			NULL, (void (*)()) NULL);
    return TCL_OK;
}


/************************************************************
 * 
 * ReadyMPEG_Priority
 *
 * Parameters:
 *     clientData - pointer to MPEG Priority object.
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Readies the priority object if unready. Entails setting up
 *     cycle time callback.
 *
 ************************************************************/

static int
ReadyMPEG_Priority(MPEG_Priority *objPtr)
{
    if (objPtr->ready) return TCL_OK;

    /*
     * Blow away old every handle if it exists.
     */

    if (objPtr->everyHandle != NULL) {
	Tcm_DeleteEvery(objPtr->interp, objPtr->everyHandle, NULL);
	objPtr->everyHandle = NULL;
    }

    /*
     * Set up new every handle if needed.
     */

    if ((objPtr->outCmd != NULL) || (objPtr->inCmd != NULL) ||
	(objPtr->i_outCmd != NULL) || (objPtr->p_outCmd != NULL) ||
	(objPtr->b_outCmd != NULL)) {
	if (objPtr->cycleTime > 0.0) {
	    objPtr->everyHandle =
	      Tcm_CreateEvery((int)(objPtr->cycleTime * 1000.0), "video",
			      DoCycleMPEG_Priority, (ClientData) objPtr);
	}
    }

    objPtr->ready = 1;

    return TCL_OK;
}


/************************************************************
 * 
 * UneadyMPEG_Priority
 *
 * Parameters:
 *     clientData - pointer to MPEG Priority object.
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Unreadies the priority object if unready. Entails killing
 *     cycle time callback if it exists.
 *
 ************************************************************/

static int
UnreadyMPEG_Priority(MPEG_Priority *objPtr)
{
    if (!objPtr->ready) return TCL_OK;

    if (objPtr->everyHandle != NULL) {
	Tcm_DeleteEvery(objPtr->interp, objPtr->everyHandle, NULL);
	objPtr->everyHandle = NULL;
    }

    /*
     * Blow away everything in sbl queues.
     */

    if (objPtr->i_sbl != NULL) {
      DetachToAllBuffers(objPtr->interp, objPtr->i_sbl, "Unreadying MPEG Priority");
      FreeSBList(objPtr->i_sbl);
    }
    objPtr->i_sbl = NULL;

    if (objPtr->p_sbl != NULL) {
      DetachToAllBuffers(objPtr->interp, objPtr->p_sbl, "Unreadying MPEG Priority");
      FreeSBList(objPtr->p_sbl);
    }
    objPtr->p_sbl = NULL;

    if (objPtr->b_sbl != NULL) {
      DetachToAllBuffers(objPtr->interp, objPtr->b_sbl, "Unreadying MPEG Priority");
      FreeSBList(objPtr->b_sbl);
    }
    objPtr->b_sbl = NULL;

    objPtr->ready = 0;
    return TCL_OK;
}


/************************************************************
 *
 * MPEG_Priority_Send
 *
 * Paramters: 
 *     objPtr - pointer to object;
 *     sblPtr - pointer to scatter buffer list
 *
 * Results:
 *     Standard Tcl return value.
 *
 * Action:
 *     Combined i, p, and b, sb lists into one and returns it
 *     in sblPtr.
 *
 ************************************************************/

static int
MPEG_Priority_Send (MPEG_Priority *objPtr, SB_List **sblPtr)
{
    
    /*
     * If object is not ready, return empty SB List. 
     */

    if (!objPtr->ready) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;

	return TCL_OK;
    }
    
    /*
     * If -inCmd set, execute now.
     */

    if (objPtr->inCmd != NULL) {
	SB_List *inSBL = NULL;

	if (objPtr->inCCmd != NULL) {
	    /*
	     * Need to check existance since object may have been destroyed.
	     */
	    
	    if (Tutil_FindAtom(objPtr->interp, objPtr->inCmd+1) != NULL) {
		objPtr->inCCmd(objPtr->inCD, &inSBL);
		if (inSBL != NULL) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    } else {
		/*
		 * In C command has unexpectedly disappeard.
		 */
		objPtr->inCCmd = NULL;
		ckfree(objPtr->inCmd);
		objPtr->inCmd = NULL;
	    }
	} else {
	    /*
	     * In command is tcl.
	     */
	    if (Tcl_VarEval(objPtr->interp, objPtr->inCmd, (char *) NULL) == TCL_OK) {
		
		inSBL = (SB_List *) ckalloc(sizeof(SB_List));
		inSBL->count = 0;
		
		if (ParseSBList(objPtr->interp, objPtr->interp->result, 
				inSBL) == TCL_OK) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    }
	}
    }
    
    if ((objPtr->i_sbl == NULL) && (objPtr->p_sbl == NULL) && (objPtr->b_sbl == NULL)) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;
	
	return TCL_OK;
    }

    /*
     * Merge i_sbl, p_sbl, and b_sbl into one list and return it through sblPtr.
     */

    if (objPtr->i_sbl != NULL) {
	if (objPtr->p_sbl != NULL) {
	    MergeSBLists(objPtr->i_sbl, objPtr->p_sbl);
	    FreeSBList(objPtr->p_sbl);
	    objPtr->p_sbl = NULL;
	}
	if (objPtr->b_sbl != NULL) {
	    MergeSBLists(objPtr->i_sbl, objPtr->b_sbl);
	    FreeSBList(objPtr->b_sbl);
	    objPtr->b_sbl = NULL;
	}

	*sblPtr = objPtr->i_sbl;
	objPtr->i_sbl = NULL;
    } else {
	if (objPtr->p_sbl != NULL) {
	    if (objPtr->b_sbl != NULL) {
		MergeSBLists(objPtr->p_sbl, objPtr->b_sbl);
		FreeSBList(objPtr->b_sbl);
		objPtr->b_sbl = NULL;
	    }
	    *sblPtr = objPtr->p_sbl;
	    objPtr->p_sbl = NULL;
	} else {
	    *sblPtr = objPtr->b_sbl;
	    objPtr->b_sbl = NULL;
	}
    }

    if ((*sblPtr)->count > 0) {
	DetachToAllBuffers(objPtr->interp, *sblPtr, NULL);
    }

    return TCL_OK;
}


/************************************************************
 *
 * MPEG_Priority_I_Send
 *
 * Paramters: 
 *     objPtr - pointer to object;
 *     sblPtr - pointer to scatter buffer list
 *
 * Results:
 *     Standard Tcl return value.
 *
 * Action:
 *     Returns i sb list in sblPtr.
 *
 ************************************************************/

static int
MPEG_Priority_I_Send (MPEG_Priority *objPtr, SB_List **sblPtr)
{
    
    /*
     * If object is not ready, return empty SB List. 
     */

    if (!objPtr->ready) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;

	return TCL_OK;
    }
    
    /*
     * If -inCmd set, execute now.
     */

    if (objPtr->inCmd != NULL) {
	SB_List *inSBL = NULL;

	if (objPtr->inCCmd != NULL) {
	    /*
	     * Need to check existance since object may have been destroyed.
	     */
	    
	    if (Tutil_FindAtom(objPtr->interp, objPtr->inCmd+1) != NULL) {
		objPtr->inCCmd(objPtr->inCD, &inSBL);
		if (inSBL != NULL) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    } else {
		/*
		 * In C command has unexpectedly disappeard.
		 */
		objPtr->inCCmd = NULL;
		ckfree(objPtr->inCmd);
		objPtr->inCmd = NULL;
	    }
	} else {
	    /*
	     * In command is tcl.
	     */
	    if (Tcl_VarEval(objPtr->interp, objPtr->inCmd, (char *) NULL) == TCL_OK) {
		
		inSBL = (SB_List *) ckalloc(sizeof(SB_List));
		inSBL->count = 0;
		
		if (ParseSBList(objPtr->interp, objPtr->interp->result, 
				inSBL) == TCL_OK) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    }
	}
    }
    
    if (objPtr->i_sbl == NULL) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;
	
	return TCL_OK;
    }

    /*
     * Return i_sbl through sblPtr.
     */

    *sblPtr = objPtr->i_sbl;
    objPtr->i_sbl = NULL;

    if ((*sblPtr)->count > 0) {
	DetachToAllBuffers(objPtr->interp, *sblPtr, NULL);
    }

    return TCL_OK;
}

/************************************************************
 *
 * MPEG_Priority_P_Send
 *
 * Paramters: 
 *     objPtr - pointer to object;
 *     sblPtr - pointer to scatter buffer list
 *
 * Results:
 *     Standard Tcl return value.
 *
 * Action:
 *     Returns i sb list in sblPtr.
 *
 ************************************************************/

static int
MPEG_Priority_P_Send (MPEG_Priority *objPtr, SB_List **sblPtr)
{
    
    /*
     * If object is not ready, return empty SB List. 
     */

    if (!objPtr->ready) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;

	return TCL_OK;
    }
    
    /*
     * If -inCmd set, execute now.
     */

    if (objPtr->inCmd != NULL) {
	SB_List *inSBL = NULL;

	if (objPtr->inCCmd != NULL) {
	    /*
	     * Need to check existance since object may have been destroyed.
	     */
	    
	    if (Tutil_FindAtom(objPtr->interp, objPtr->inCmd+1) != NULL) {
		objPtr->inCCmd(objPtr->inCD, &inSBL);
		if (inSBL != NULL) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    } else {
		/*
		 * In C command has unexpectedly disappeard.
		 */
		objPtr->inCCmd = NULL;
		ckfree(objPtr->inCmd);
		objPtr->inCmd = NULL;
	    }
	} else {
	    /*
	     * In command is tcl.
	     */
	    if (Tcl_VarEval(objPtr->interp, objPtr->inCmd, (char *) NULL) == TCL_OK) {
		
		inSBL = (SB_List *) ckalloc(sizeof(SB_List));
		inSBL->count = 0;
		
		if (ParseSBList(objPtr->interp, objPtr->interp->result, 
				inSBL) == TCL_OK) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */

 			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    }
	}
    }
    
    if (objPtr->p_sbl == NULL) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;
	
	return TCL_OK;
    }

    /*
     * Return p_sbl through sblPtr.
     */

    *sblPtr = objPtr->p_sbl;
    objPtr->p_sbl = NULL;

    if ((*sblPtr)->count > 0) {
	DetachToAllBuffers(objPtr->interp, *sblPtr, NULL);
    }

    return TCL_OK;
}

/************************************************************
 *
 * MPEG_Priority_B_Send
 *
 * Paramters: 
 *     objPtr - pointer to object;
 *     sblPtr - pointer to scatter buffer list
 *
 * Results:
 *     Standard Tcl return value.
 *
 * Action:
 *     Returns i sb list in sblPtr.
 *
 ************************************************************/

static int
MPEG_Priority_B_Send (MPEG_Priority *objPtr, SB_List **sblPtr)
{
    
    /*
     * If object is not ready, return empty SB List. 
     */

    if (!objPtr->ready) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;

	return TCL_OK;
    }
    
    /*
     * If -inCmd set, execute now.
     */

    if (objPtr->inCmd != NULL) {
	SB_List *inSBL = NULL;

	if (objPtr->inCCmd != NULL) {
	    /*
	     * Need to check existance since object may have been destroyed.
	     */
	    
	    if (Tutil_FindAtom(objPtr->interp, objPtr->inCmd+1) != NULL) {
		objPtr->inCCmd(objPtr->inCD, &inSBL);
		if (inSBL != NULL) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */ 

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    } else {
		/*
		 * In C command has unexpectedly disappeard.
		 */
		objPtr->inCCmd = NULL;
		ckfree(objPtr->inCmd);
		objPtr->inCmd = NULL;
	    }
	} else {
	    /*
	     * In command is tcl.
	     */
	    if (Tcl_VarEval(objPtr->interp, objPtr->inCmd, (char *) NULL) == TCL_OK) {
		
		inSBL = (SB_List *) ckalloc(sizeof(SB_List));
		inSBL->count = 0;
		
		if (ParseSBList(objPtr->interp, objPtr->interp->result, 
				inSBL) == TCL_OK) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */ 

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    }
	}
    }
    
    if (objPtr->b_sbl == NULL) {
	*sblPtr = (SB_List *) ckalloc(sizeof(SB_List));
	(*sblPtr)->count = 0;
	(*sblPtr)->sbs = NULL;
	
	return TCL_OK;
    }

    /*
     * Return b_sbl through sblPtr.
     */

    *sblPtr = objPtr->b_sbl;
    objPtr->b_sbl = NULL;

    if ((*sblPtr)->count > 0) {
	DetachToAllBuffers(objPtr->interp, *sblPtr, NULL);
    }

    return TCL_OK;
}


/************************************************************
 *
 * MPEG_Priority_InternalAccept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -	scatter buffer list containing frame data
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sorts all the SB's in the SB List by mpeg frame type
 *     and appends them to the appropriate object sb list.
 *
 ************************************************************/

static int
MPEG_Priority_InternalAccept (MPEG_Priority *objPtr, SB_List *sbl) 
{
    int i;
    char *dataPtr;
    ScatterBuffer *sbPtr;
    MPEGBufferHeader mpegBufHdr;

    if (!objPtr->ready) {
	return TCL_OK;
    }

    if (sbl == NULL) {
	return TCL_OK;
    }

    if (sbl->count <= 0) {
	return TCL_OK;
    }

    AttachToAllBuffers(objPtr->interp, sbl, NULL);

    for (i=0; i<sbl->count; i++) {
	sbPtr = sbl->sbs[i];
	sbl->sbs[i] = NULL;

	dataPtr = GetFromSB(sbPtr, CM_MPEG_BUF_HDR, NULL, NULL);

	if (dataPtr != NULL) {
	    memcpy(&mpegBufHdr, dataPtr-8, sizeof(MPEGBufferHeader));
	    mpegBufHdr.type = ntohl(mpegBufHdr.type);

	    if (mpegBufHdr.type == MPEG_I_FRAME) {
		if (objPtr->i_sbl == NULL) {
		    objPtr->i_sbl = (SB_List *) ckalloc(sizeof(SB_List));
		    objPtr->i_sbl->count = 1;
		    objPtr->i_sbl->sbs = (ScatterBuffer **) 
		      ckalloc(sizeof(ScatterBuffer *) * 1);
		    objPtr->i_sbl->sbs[0] = sbPtr;
		} else {
		    objPtr->i_sbl->count++;
		    objPtr->i_sbl->sbs = (ScatterBuffer **)
		      ckrealloc((char *) objPtr->i_sbl->sbs, (sizeof(ScatterBuffer *) *
							      objPtr->i_sbl->count));
		    objPtr->i_sbl->sbs[objPtr->i_sbl->count-1] = sbPtr;
		}
	    } else if (mpegBufHdr.type == MPEG_P_FRAME) {
		if (objPtr->p_sbl == NULL) {
		    objPtr->p_sbl = (SB_List *) ckalloc(sizeof(SB_List));
		    objPtr->p_sbl->count = 1;
		    objPtr->p_sbl->sbs = (ScatterBuffer **) 
		      ckalloc(sizeof(ScatterBuffer *) * 1);
		    objPtr->p_sbl->sbs[0] = sbPtr;
		} else {
		    objPtr->p_sbl->count++;
		    objPtr->p_sbl->sbs = (ScatterBuffer **)
		      ckrealloc((char *) objPtr->p_sbl->sbs, (sizeof(ScatterBuffer *) *
							      objPtr->p_sbl->count));
		    objPtr->p_sbl->sbs[objPtr->p_sbl->count-1] = sbPtr;
		}
	    } else if (mpegBufHdr.type == MPEG_B_FRAME) { 
		if (objPtr->b_sbl == NULL) {
		    objPtr->b_sbl = (SB_List *) ckalloc(sizeof(SB_List));
		    objPtr->b_sbl->count = 1;
		    objPtr->b_sbl->sbs = (ScatterBuffer **) 
		      ckalloc(sizeof(ScatterBuffer *) * 1);
		    objPtr->b_sbl->sbs[0] = sbPtr;
		} else {
		    objPtr->b_sbl->count++;
		    objPtr->b_sbl->sbs = (ScatterBuffer **)
		      ckrealloc((char *) objPtr->b_sbl->sbs, (sizeof(ScatterBuffer *) *
							      objPtr->b_sbl->count));
		    objPtr->b_sbl->sbs[objPtr->b_sbl->count-1] = sbPtr;
		}
	    } else {
		DetachScatterBuffer(objPtr->interp, sbPtr, "Couldn't find frame type.");
		FreeScatterBuffer(sbPtr);
	    }
	} else {
	    DetachScatterBuffer(objPtr->interp, sbPtr, "Counldn't find mpeg hdr.");
	    FreeScatterBuffer(sbPtr);
	}
    }

    ckfree((char *) sbl->sbs);
    sbl->sbs = NULL;
    sbl->count = 0;
    
    return TCL_OK;
}

    


/************************************************************
 *
 * MPEG_Priority_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -	scatter buffer list containing frame data
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
MPEG_Priority_Accept(MPEG_Priority *objPtr, SB_List *sbl)
{
    SB_List *priv_sbl;

    /*
     * Return if not ready.
     */

    if (!objPtr->ready) return TCL_OK;
   
    if ((sbl == NULL) || (sbl->count < 1)) return TCL_OK;

    /* 
     * We don't attach to the buffers here since that is always
     * done in InternalAccept.
     */

    
    priv_sbl = CopySBList(sbl);

    /*
     * Merge incoming buffers.
     * If cycle time is 0.0: 
     *    Send buffers through i, p, or b out cmd if set.
     *    If any left and -outCmd is set, send
     *       buffers through outCmd.
     *    If not: 
     *        Return.
     * If cycleTime > 0.0:
     *    Return.
     */

    /* Internal Accept does an attach, so don't do it here. */ 

    MPEG_Priority_InternalAccept(objPtr, priv_sbl);
    FreeSBList(priv_sbl);

    if (objPtr->cycleTime == 0.0) {

	/*
	 * Do I Out Cmd.
	 */

	if ((objPtr->i_outCmd != NULL) && (objPtr->i_sbl != NULL)) {
	    
	    /* 
	     * Pass on data to i_outCmd
	     */
	    if (objPtr->i_outCCmd != NULL) {
		/*
		 * Need to check if it still exists since object may have
		 * been destroyed.
		 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->i_outCmd+1) != NULL) {
		    objPtr->i_outCCmd(objPtr->i_outCD, objPtr->i_sbl);
		} else {
		    /*
		     * Out C command has unexpectedly disappeared.
		     */
		    objPtr->i_outCCmd = NULL;
		    ckfree(objPtr->i_outCmd);
		    objPtr->i_outCmd = NULL;
		}
	    } else {
		/*
		 * Out command is tcl, make sb list into a string and 
		 * call using Tcl_Eval. 
		 */
		char *sbStr = MakeSBListString(objPtr->i_sbl);
		Tcl_VarEval(objPtr->interp, objPtr->i_outCmd, " ",
				sbStr, (char *) NULL); 
		ckfree(sbStr);
	    }
	    DetachToAllBuffers(objPtr->interp, objPtr->i_sbl, NULL);
	    FreeSBList(objPtr->i_sbl);
	    objPtr->i_sbl = NULL;
	}

	/*
	 * Do P Out Cmd.
	 */

	if ((objPtr->p_outCmd != NULL) && (objPtr->p_sbl != NULL)) {
	    
	    /* 
	     * Pass on data to p_outCmd
	     */
	    if (objPtr->p_outCCmd != NULL) {
		/*
		 * Need to check if it still exists since object may have
		 * been destroyed.
		 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->p_outCmd+1) != NULL) {
		    objPtr->p_outCCmd(objPtr->p_outCD, objPtr->p_sbl);
		} else {
		    /*
		     * Out C command has unexpectedly disappeared.
		     */
		    objPtr->p_outCCmd = NULL;
		    ckfree(objPtr->p_outCmd);
		    objPtr->p_outCmd = NULL;
		}
	    } else {
		/*
		 * Out command is tcl, make sb list into a string and 
		 * call using Tcl_Eval. 
		 */
		char *sbStr = MakeSBListString(objPtr->p_sbl);
		Tcl_VarEval(objPtr->interp, objPtr->p_outCmd, " ",
				sbStr, (char *) NULL); 
		ckfree(sbStr);
	    }
	    DetachToAllBuffers(objPtr->interp, objPtr->p_sbl, NULL);
	    FreeSBList(objPtr->p_sbl);
	    objPtr->p_sbl = NULL;
	}

	/*
	 * Do B Out Cmd.
	 */

	if ((objPtr->b_outCmd != NULL) && (objPtr->b_sbl != NULL)) {
	    
	    /* 
	     * Pass on data to b_outCmd
	     */
	    if (objPtr->b_outCCmd != NULL) {
		/*
		 * Need to check if it still exists since object may have
		 * been destroyed.
		 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->b_outCmd+1) != NULL) {
		    objPtr->b_outCCmd(objPtr->b_outCD, objPtr->b_sbl);
		} else {
		    /*
		     * Out C command has unexpectedly disappeared.
		     */
		    objPtr->b_outCCmd = NULL;
		    ckfree(objPtr->b_outCmd);
		    objPtr->b_outCmd = NULL;
		}
	    } else {
		/*
		 * Out command is tcl, make sb list into a string and 
		 * call using Tcl_Eval. 
		 */
		char *sbStr = MakeSBListString(objPtr->b_sbl);
		Tcl_VarEval(objPtr->interp, objPtr->b_outCmd, " ",
				sbStr, (char *) NULL); 
		ckfree(sbStr);
	    }
	    DetachToAllBuffers(objPtr->interp, objPtr->b_sbl, NULL);
	    FreeSBList(objPtr->b_sbl);
	    objPtr->b_sbl = NULL;
	}

	/*
         * Do general out cmd if anything left.
	 */

	if ((objPtr->outCmd != NULL) && ((objPtr->i_sbl != NULL) ||
					 (objPtr->p_sbl != NULL) ||
					 (objPtr->b_sbl != NULL))) {
	    SB_List *outSBL;

	    /*
	     * Merge i_sbl, p_sbl, and b_sbl into one list.
	     */

	    if (objPtr->i_sbl != NULL) {
		if (objPtr->p_sbl != NULL) {
		    MergeSBLists(objPtr->i_sbl, objPtr->p_sbl);
		    FreeSBList(objPtr->p_sbl);
		    objPtr->p_sbl = NULL;
		}
		if (objPtr->b_sbl != NULL) {
		    MergeSBLists(objPtr->i_sbl, objPtr->b_sbl);
		    FreeSBList(objPtr->b_sbl);
		    objPtr->b_sbl = NULL;
		}
		outSBL = objPtr->i_sbl;
		objPtr->i_sbl = NULL;
	    } else {
		if (objPtr->p_sbl != NULL) {
		    if (objPtr->b_sbl != NULL) {
			MergeSBLists(objPtr->p_sbl, objPtr->b_sbl);
			FreeSBList(objPtr->b_sbl);
			objPtr->b_sbl = NULL;
		    }
		    outSBL = objPtr->p_sbl;
		    objPtr->p_sbl = NULL;
		} else {
		    outSBL = objPtr->b_sbl;
		    objPtr->b_sbl = NULL;
		}
	    }

	    /* 
	     * Pass on data to outCmd
	     */
	    if (objPtr->outCCmd != NULL) {
		/*
		 * Need to check if it still exists since object may have
		 * been destroyed.
		 */
		if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		    objPtr->outCCmd(objPtr->outCD, outSBL);
		} else {
		    /*
		     * Out C command has unexpectedly disappeared.
		     */
		    objPtr->outCCmd = NULL;
		    ckfree(objPtr->outCmd);
		    objPtr->outCmd = NULL;
		}
	    } else {
		/*
		 * Out command is tcl, make sb list into a string and 
		 * call using Tcl_Eval. 
		 */
		char *sbStr = MakeSBListString(outSBL);
		Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
				sbStr, (char *) NULL); 
		ckfree(sbStr);
	    }
	    DetachToAllBuffers(objPtr->interp, outSBL, NULL);
	    FreeSBList(outSBL);
	    outSBL = NULL;
	}
    }

    return TCL_OK;
}

/************************************************************
 *
 * MPEG_Priority_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *     newOut - string of new out command.
 *     typeFlag - indicates which out command to changed I, P,
 *                B, or general.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
MPEG_Priority_ChangeOutCmd(MPEG_Priority *objPtr, char *newOut, int typeFlag) 
{
#ifdef SECURE_CMT
	if ((newOut != NULL) && (newOut[0] != '@')) {
	    Tcl_SetResult(objPtr->interp, "Cannot set Tcl outCmd because this program was compiled with -DSECURE_CMT.\nRecompile without security to use arbitrary Tcl outCmd's\n", TCL_STATIC);
	    return TCL_ERROR;
	}
#endif
    
    if (typeFlag == MPEG_PRIORITY_GENERAL) {
	/*
	 * Clear the previous entry.
	 */
	if (objPtr->outCmd != NULL) {
	    ckfree(objPtr->outCmd);
	    objPtr->outCmd = NULL;
	}
	objPtr->outCCmd = NULL;
	objPtr->outCD = NULL;
	
	/*
	 * If new outCmd is empty, return.
	 */
	if (strlen(newOut) == 0) {
	    return TCL_OK;
	}
	
	/*
	 * If new outCmd is C proc, get them as atoms and set everything
	 * accordingly. 
	 */
	if (newOut[0] == '@') {
	    if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
					   &(objPtr->outCD), 
					   (ClientData *) &(objPtr->outCCmd)) 
		!= TCL_OK) {

		objPtr->outCD = NULL;
		objPtr->outCCmd = NULL;
		Tcl_AppendResult(objPtr->interp,
				 "Could find registered C procedure: ", newOut,
				 (char *)NULL);
		return TCL_ERROR;
	    }
	}
	
	/*
	 * Copy new out command into outCmd string.
	 */
	objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
	strcpy(objPtr->outCmd,newOut);

    } else if (typeFlag == MPEG_PRIORITY_I_OUT_CMD) {
	/*
	 * Clear the previous entry.
	 */
	if (objPtr->i_outCmd != NULL) {
	    ckfree(objPtr->i_outCmd);
	    objPtr->i_outCmd = NULL;
	}
	objPtr->i_outCCmd = NULL;
	objPtr->i_outCD = NULL;
	
	/*
	 * If new i_outCmd is empty, return.
	 */
	if (strlen(newOut) == 0) {
	    return TCL_OK;
	}
	
	/*
	 * If new i_outCmd is C proc, get them as atoms and set everything
	 * accordingly. 
	 */
	if (newOut[0] == '@') {
	    if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
					   &(objPtr->i_outCD), 
					   (ClientData *) &(objPtr->i_outCCmd)) 
		!= TCL_OK) {

		objPtr->i_outCD = NULL;
		objPtr->i_outCCmd = NULL;
		Tcl_AppendResult(objPtr->interp,
				 "Could find registered C procedure: ", newOut,
				 (char *)NULL);
		return TCL_ERROR;
	    }
	}
	
	/*
	 * Copy new out command into i_outCmd string.
	 */
	objPtr->i_outCmd = (char *) ckalloc(strlen(newOut)+1);
	strcpy(objPtr->i_outCmd,newOut);

    } else if (typeFlag == MPEG_PRIORITY_P_OUT_CMD) {
	/*
	 * Clear the previous entry.
	 */
	if (objPtr->p_outCmd != NULL) {
	    ckfree(objPtr->p_outCmd);
	    objPtr->p_outCmd = NULL;
	}
	objPtr->p_outCCmd = NULL;
	objPtr->p_outCD = NULL;
	
	/*
	 * If new p_outCmd is empty, return.
	 */
	if (strlen(newOut) == 0) {
	    return TCL_OK;
	}
	
	/*
	 * If new p_outCmd is C proc, get them as atoms and set everything
	 * accordingly. 
	 */
	if (newOut[0] == '@') {
	    if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
					   &(objPtr->p_outCD), 
					   (ClientData *) &(objPtr->p_outCCmd)) 
		!= TCL_OK) {

		objPtr->p_outCD = NULL;
		objPtr->p_outCCmd = NULL;
		Tcl_AppendResult(objPtr->interp,
				 "Could find registered C procedure: ", newOut,
				 (char *)NULL);
		return TCL_ERROR;
	    }
	}
	
	/*
	 * Copy new out command into p_outCmd string.
	 */
	objPtr->p_outCmd = (char *) ckalloc(strlen(newOut)+1);
	strcpy(objPtr->p_outCmd,newOut);

    } else if (typeFlag == MPEG_PRIORITY_B_OUT_CMD) {
	/*
	 * Clear the previous entry.
	 */
	if (objPtr->b_outCmd != NULL) {
	    ckfree(objPtr->b_outCmd);
	    objPtr->b_outCmd = NULL;
	}
	objPtr->b_outCCmd = NULL;
	objPtr->b_outCD = NULL;
	
	/*
	 * If new b_outCmd is empty, return.
	 */
	if (strlen(newOut) == 0) {
	    return TCL_OK;
	}
	
	/*
	 * If new b_outCmd is C proc, get them as atoms and set everything
	 * accordingly. 
	 */
	if (newOut[0] == '@') {
	    if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
					   &(objPtr->b_outCD), 
					   (ClientData *) &(objPtr->b_outCCmd)) 
		!= TCL_OK) {

		objPtr->b_outCD = NULL;
		objPtr->b_outCCmd = NULL;
		Tcl_AppendResult(objPtr->interp,
				 "Could find registered C procedure: ", newOut,
				 (char *)NULL);
		return TCL_ERROR;
	    }
	}
	
	/*
	 * Copy new out command into b_outCmd string.
	 */
	objPtr->b_outCmd = (char *) ckalloc(strlen(newOut)+1);
	strcpy(objPtr->b_outCmd,newOut);
    }
    return TCL_OK;
}    

/************************************************************
 *
 * MPEG_Priority_ChangeInCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *     newIn - string representation of new inCmd.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes in command.
 ************************************************************/

static int
MPEG_Priority_ChangeInCmd(MPEG_Priority *objPtr, char *newIn) 
{
    /*
     * Clear the previous entry.
     */
    if (objPtr->inCmd != NULL) {
	ckfree(objPtr->inCmd);
    }

    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new inCmd is empty, return.
     */
    if (strlen(newIn) == 0) {
	return TCL_OK;
    }
   
#ifdef SECURE_CMT
    if (newIn[0] != '@') {
	Tcl_SetResult(objPtr->interp, "Cannot set Tcl inCmd because this program was compiled with -DSECURE_CMT.\nRecompile without security to use arbitrary Tcl outCmd's\n", TCL_STATIC);
        return TCL_ERROR;
    }
#endif

    /*
     * If new inCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newIn[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newIn,
				       &(objPtr->inCD), 
				       (ClientData *) &(objPtr->inCCmd)) != TCL_OK) {
	    objPtr->inCD = NULL;
	    objPtr->inCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Could find registered C procedure: ", newIn,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new in command into inCmd string.
     */
    objPtr->inCmd = (char *) ckalloc(strlen(newIn)+1);
    strcpy(objPtr->inCmd,newIn);
    return TCL_OK;
}    

/****************************************************************8
 *
 * DoCycleMPEG_Priority
 *
 * Parameters:
 *     clientData - pointer to MPEG Priority object.
 *
 * Results:
 *     None
 *
 * Actions:
 *     Called by every handler when time expires. Does -inCmd and all
 *     -outCmd's if they exist.
 *
 ****************************************************************/

static void
DoCycleMPEG_Priority(ClientData clientData)
{
    MPEG_Priority *objPtr = (MPEG_Priority *) clientData;
    
    if (!objPtr->ready) {
	return;
    }

    /*
     * If -inCmd set, execute now.
     */

    if (objPtr->inCmd != NULL) {
	SB_List *inSBL = NULL;

	if (objPtr->inCCmd != NULL) {
	    /*
	     * Need to check existance since object may have been destroyed.
	     */
	    
	    if (Tutil_FindAtom(objPtr->interp, objPtr->inCmd+1) != NULL) {
		objPtr->inCCmd(objPtr->inCD, &inSBL);
		if (inSBL != NULL) {
		    if (inSBL->count > 0) {

			/* Internal Accept does an attach, so don't do it here. */

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    } else {
		/*
		 * In C command has unexpectedly disappeard.
		 */
		objPtr->inCCmd = NULL;
		ckfree(objPtr->inCmd);
		objPtr->inCmd = NULL;
	    }
	} else {
	    /*
	     * In command is tcl.
	     */
	    if (Tcl_VarEval(objPtr->interp, objPtr->inCmd, (char *) NULL) == TCL_OK) {
		
		inSBL = (SB_List *) ckalloc(sizeof(SB_List));
		inSBL->count = 0;
		
		if (ParseSBList(objPtr->interp, objPtr->interp->result, 
				inSBL) == TCL_OK) {
		    if (inSBL->count > 0) {
			/* Internal Accept does an attach, so don't do it here. */

			MPEG_Priority_InternalAccept(objPtr, inSBL);
		    }
		    FreeSBList(inSBL);
		}
	    }
	}
    }

    /*
     * Now do out commands.
     */
    
    /*
     * Do I Out Cmd.
     */
    
    if ((objPtr->i_outCmd != NULL) && (objPtr->i_sbl != NULL)) {
	
	/* 
	 * Pass on data to i_outCmd
	 */
	if (objPtr->i_outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->i_outCmd+1) != NULL) {
		objPtr->i_outCCmd(objPtr->i_outCD, objPtr->i_sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->i_outCCmd = NULL;
		ckfree(objPtr->i_outCmd);
		objPtr->i_outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and 
	     * call using Tcl_Eval. 
	     */
	    char *sbStr = MakeSBListString(objPtr->i_sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->i_outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree(sbStr);
	}
	DetachToAllBuffers(objPtr->interp, objPtr->i_sbl, NULL);
	FreeSBList(objPtr->i_sbl);
	objPtr->i_sbl = NULL;
    }
    
    /*
     * Do P Out Cmd.
     */
    
    if ((objPtr->p_outCmd != NULL) && (objPtr->p_sbl != NULL)) {
	
	/* 
	 * Pass on data to p_outCmd
	 */
	if (objPtr->p_outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->p_outCmd+1) != NULL) {
		objPtr->p_outCCmd(objPtr->p_outCD, objPtr->p_sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->p_outCCmd = NULL;
		ckfree(objPtr->p_outCmd);
		objPtr->p_outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and 
	     * call using Tcl_Eval. 
	     */
	    char *sbStr = MakeSBListString(objPtr->p_sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->p_outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree(sbStr);
	}
	DetachToAllBuffers(objPtr->interp, objPtr->p_sbl, NULL);
	FreeSBList(objPtr->p_sbl);
	objPtr->p_sbl = NULL;
    }
    
    /*
     * Do B Out Cmd.
     */
    
    if ((objPtr->b_outCmd != NULL) && (objPtr->b_sbl != NULL)) {
	
	/* 
	 * Pass on data to b_outCmd
	 */
	if (objPtr->b_outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->b_outCmd+1) != NULL) {
		objPtr->b_outCCmd(objPtr->b_outCD, objPtr->b_sbl);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->b_outCCmd = NULL;
		ckfree(objPtr->b_outCmd);
		objPtr->b_outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and 
	     * call using Tcl_Eval. 
	     */
	    char *sbStr = MakeSBListString(objPtr->b_sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->b_outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree(sbStr);
	}
	DetachToAllBuffers(objPtr->interp, objPtr->b_sbl, NULL);
	FreeSBList(objPtr->b_sbl);
	objPtr->b_sbl = NULL;
    }
    
    /*
     * Do general out cmd if anything left.
     */
    
    if ((objPtr->outCmd != NULL) && ((objPtr->i_sbl != NULL) ||
				     (objPtr->p_sbl != NULL) ||
				     (objPtr->b_sbl != NULL))) {
	SB_List *outSBL;
	
	/*
	 * Merge i_sbl, p_sbl, and b_sbl into one list.
	 */

	if (objPtr->i_sbl != NULL) {
	    if (objPtr->p_sbl != NULL) {
		MergeSBLists(objPtr->i_sbl, objPtr->p_sbl);
		FreeSBList(objPtr->p_sbl);
		objPtr->p_sbl = NULL;
		}
	    if (objPtr->b_sbl != NULL) {
		MergeSBLists(objPtr->i_sbl, objPtr->b_sbl);
		FreeSBList(objPtr->b_sbl);
		objPtr->b_sbl = NULL;
	    }
	    outSBL = objPtr->i_sbl;
	    objPtr->i_sbl = NULL;
	} else {
	    if (objPtr->p_sbl != NULL) {
		    if (objPtr->b_sbl != NULL) {
			MergeSBLists(objPtr->p_sbl, objPtr->b_sbl);
			FreeSBList(objPtr->b_sbl);
			objPtr->b_sbl = NULL;
		    }
		    outSBL = objPtr->p_sbl;
		    objPtr->p_sbl = NULL;
	    } else {
		outSBL = objPtr->b_sbl;
		objPtr->b_sbl = NULL;
	    }
	}
	
	/* 
	 * Pass on data to outCmd
	 */
	if (objPtr->outCCmd != NULL) {
	    /*
	     * Need to check if it still exists since object may have
	     * been destroyed.
	     */
	    if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
		objPtr->outCCmd(objPtr->outCD, outSBL);
	    } else {
		/*
		 * Out C command has unexpectedly disappeared.
		 */
		objPtr->outCCmd = NULL;
		ckfree(objPtr->outCmd);
		objPtr->outCmd = NULL;
	    }
	} else {
	    /*
	     * Out command is tcl, make sb list into a string and 
	     * call using Tcl_Eval. 
	     */
	    char *sbStr = MakeSBListString(outSBL);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
			sbStr, (char *) NULL); 
	    ckfree(sbStr);
	}
	DetachToAllBuffers(objPtr->interp, outSBL, NULL);
	FreeSBList(outSBL);
	outSBL = NULL;
    }
    
    return;
}
