/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * tcmObjTags.c --
 *       Manages the object tagging facilities. Basically a double hash
 *       table that goes from object names to tags and tag to object names.
 *       No checking is done on object types or even if the object name is
 *       is really an object or not.
 *
 *       This module also generates and uses the CID seen by the client applications.
 *       In reality the CID is nothing more than a part of the object management.
 *  
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#   include "compat/stdlib.h"
#else
#   include <stdlib.h>
#endif
#ifdef NO_STRING_H
#   include "compat/string.h"
#else
#   include <string.h>
#endif
#include "tk.h"
#include "tkInt.h"


typedef struct TcmTag {
  int CID;
  Tk_Uid tag;

  struct TcmTag *next;
} TcmTag;

typedef struct TcmTagList {
  TcmTag *tags;

  Tcl_HashTable *tablePtr;
  Tcl_HashEntry *hashPtr;
} TcmTagList;

typedef struct ObjName {
  int CID;
  Tk_Uid name;

  struct ObjName *next;
} ObjName;

typedef struct ObjNameList {
  ObjName *names;

  Tcl_HashTable *tablePtr;
  Tcl_HashEntry *hashPtr;
} ObjNameList;

static Tcl_HashTable objectTable;
static Tcl_HashTable tagTable;

static int initialized = 0;

static void
ObjTagInit()
{
  initialized = 1;

  Tcl_InitHashTable(&objectTable, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&tagTable, TCL_ONE_WORD_KEYS);
}

static int
ObjectAddTag(Tcl_Interp *interp,
	     int CID,
	     char *objName,
	     char *tag) 
{
  Tk_Uid nameId;
  Tk_Uid tagId;
  Tcl_HashEntry *hash_entry;
  int new;

  /* Construct object name key. */

  nameId = Tk_GetUid(objName);
  tagId = Tk_GetUid(tag);

  /* Get hash entry. */

  hash_entry = Tcl_CreateHashEntry(&objectTable, (char *) nameId, &new);

  /* If hash entry does not exist, create new tag list, enter tag, and set hash
     entry value. 
  */

  if (new) {
    TcmTagList *tlist;

    tlist = (TcmTagList *) ckalloc(sizeof(TcmTagList));
    tlist->tags = (TcmTag *) ckalloc(sizeof(TcmTag));
    tlist->tags->CID = CID;
    tlist->tags->tag = tagId;
    tlist->tags->next = NULL;
    tlist->tablePtr = &objectTable;
    tlist->hashPtr = hash_entry;
    Tcl_SetHashValue(hash_entry, tlist);
  } else {

    /* If not new, then search current list for tag to prevent duplicates.
       If not in current list, add tag to list.
    */

    TcmTagList *tlist;
    TcmTag *tagPtr, *newTagPtr;

    tlist = (TcmTagList *) Tcl_GetHashValue(hash_entry);

    tagPtr = tlist->tags;

    while (tagPtr != NULL) {
      if ((tagPtr->tag == tagId) && (tagPtr->CID == CID)) {
	return TCL_OK;
      }
      tagPtr = tagPtr->next;
    }

    newTagPtr = (TcmTag *) ckalloc(sizeof(TcmTag));
    newTagPtr->CID = CID; 
    newTagPtr->tag = tagId;
    newTagPtr->next = tlist->tags;
    tlist->tags = newTagPtr;
  }

  /* Now put CID and objectName into tag hash table. */

  /* Get hash entry. */

  hash_entry = Tcl_CreateHashEntry(&tagTable, (char *) tagId, &new);

  /* If hash entry does not exist, create new object list, enter object, and set hash
     entry value. 
  */

  if (new) {
    ObjNameList *nlist;

    nlist = (ObjNameList *) ckalloc(sizeof(ObjNameList));
    nlist->names = (ObjName *) ckalloc(sizeof(ObjName));
    nlist->names->CID = CID;
    nlist->names->name = nameId;
    nlist->names->next = NULL;
    nlist->tablePtr = &tagTable;
    nlist->hashPtr = hash_entry;
    Tcl_SetHashValue(hash_entry, nlist);
  } else {
    ObjName *newObjNamePtr;

    /* If not new, add object to current list of objects. Don't have to search
       for duplicates because should have already caught that case when adding
       tag to object table.
    */

    ObjNameList *nlist;

    nlist = (ObjNameList *) Tcl_GetHashValue(hash_entry);

    newObjNamePtr = (ObjName *) ckalloc(sizeof(ObjName));
    newObjNamePtr->CID = CID;
    newObjNamePtr->name = nameId;
    newObjNamePtr->next = nlist->names;
    nlist->names = newObjNamePtr;
  }

  return TCL_OK;
}

static int
ObjectDelTag (Tcl_Interp *interp,
	      int CID,
	      char *objName,
	      char *tag)
{

  Tk_Uid nameId, tagId;
  Tcl_HashEntry *hash_entry;
  ObjNameList *nlist;
  TcmTagList *tlist;
  int new;
  TcmTag *tag_ptr, *prev_tag_ptr;
  ObjName *obj_name_ptr, *prev_obj_name_ptr;

  nameId = Tk_GetUid(objName);
  tagId = Tk_GetUid(tag);

  /* Look up tag in object table. */

  hash_entry = Tcl_CreateHashEntry(&objectTable, (char *) nameId, &new);

  /* If entry does not exist, object was never tagged to begin with, return
     TCL_OK. 
  */

  if (new) {
    Tcl_DeleteHashEntry(hash_entry);
    return TCL_OK;
  }

  tlist = (TcmTagList *) Tcl_GetHashValue(hash_entry);

  /* Search for tag. If not in list, obj. never tagged with it, return TCL_OK,
     otherwise proceed with deletion.
  */

  tag_ptr = tlist->tags;
  prev_tag_ptr = NULL;

  while (tag_ptr != NULL) {
    if ((tag_ptr->tag == tagId) && (tag_ptr->CID == CID))
      break;

    prev_tag_ptr = tag_ptr;
    tag_ptr = tag_ptr->next;
  }

  if (tag_ptr == NULL)
    return TCL_OK;

  
  if (prev_tag_ptr == NULL) {
    tlist->tags = tag_ptr->next;
  }
  else {
    prev_tag_ptr->next = tag_ptr->next;
  } 
  ckfree((char *) tag_ptr);
 
  /* Check to see if you've removed the last tag. If so, remove the hash entry. */

  if (tlist->tags == NULL) {
    Tcl_DeleteHashEntry(hash_entry);
    ckfree((char *) tlist);
  }

  /* Now remove object from tags list in tag table. */

  hash_entry = Tcl_CreateHashEntry(&tagTable, (char *) tagId, &new);

  /* If entry doesn't exist, there's a problem. */

  if (new) {
    Tcl_DeleteHashEntry(hash_entry);
    Tcl_AppendResult(interp, "Internal object tag inconsistency", (char *) NULL);
    return TCL_ERROR;
  }

  nlist = (ObjNameList *) Tcl_GetHashValue(hash_entry);

  /* Search for object. If not in list, there is a problem. */

  obj_name_ptr = nlist->names;
  prev_obj_name_ptr = NULL;

  while (obj_name_ptr != NULL) {
    if ((obj_name_ptr->CID == CID) && (obj_name_ptr->name == nameId)) 
      break;

    prev_obj_name_ptr = obj_name_ptr;
    obj_name_ptr = obj_name_ptr->next;
  }

  if (obj_name_ptr == NULL) {
    Tcl_AppendResult(interp, "Internal object tag inconsistency", (char *) NULL);
    return TCL_ERROR;
  }
  
  if (prev_obj_name_ptr == NULL) {
    nlist->names = obj_name_ptr->next;
  }
  else {
    prev_obj_name_ptr->next = obj_name_ptr->next;
  }
  ckfree((char *) obj_name_ptr);

  /* Check to see if you've removed to last object from this tag. If
     so, remove the hash entry. 
  */

  if (nlist->names == NULL) {
    Tcl_DeleteHashEntry(hash_entry);
    ckfree((char *) nlist);
  }

  return TCL_OK;
}

static int
ObjectDeleteAllTags (Tcl_Interp *interp,
		     int CID,
		     char *objName)
{

  Tk_Uid nameId;
  Tcl_HashEntry *hash_entry;
  TcmTagList *tlist;
  int new;
  TcmTag *tag_ptr;
  TcmTag *temp_tag_ptr;

  nameId = Tk_GetUid(objName);

  /* Look up tags in object table. */

  hash_entry = Tcl_CreateHashEntry(&objectTable, (char *) nameId, &new);

  /* If entry does not exist, object was never tagged to begin with, return
     TCL_OK. 
  */

  if (new) {
    Tcl_DeleteHashEntry(hash_entry);
    return TCL_OK;
  }

  tlist = (TcmTagList *) Tcl_GetHashValue(hash_entry);

  tag_ptr = tlist->tags;

  /* If no tags, you're done, destroy structures. */

  if (tag_ptr == NULL) {
    Tcl_DeleteHashEntry(hash_entry);
    ckfree((char *) tlist);
    return TCL_OK;
  }

  while (tag_ptr != NULL) {
      Tcl_HashEntry *n_hash_entry;
      ObjNameList *nlist;
      ObjName *obj_name_ptr, *prev_obj_name_ptr;

      n_hash_entry = Tcl_FindHashEntry(&tagTable, (char *) tag_ptr->tag);

      if (n_hash_entry != NULL) {
	  nlist = (ObjNameList *) Tcl_GetHashValue(n_hash_entry);
      
	  obj_name_ptr = nlist->names;
	  prev_obj_name_ptr = NULL;

	  while (obj_name_ptr != NULL) {
	      if ((obj_name_ptr->CID == CID) && (obj_name_ptr->name == nameId))
		break;

	      prev_obj_name_ptr = obj_name_ptr;
	      obj_name_ptr = obj_name_ptr->next;
	  }

	  if (obj_name_ptr == NULL) {
	      Tcl_AppendResult(interp, "Internal object tag inconsistency", 
			       (char *) NULL);
	      return TCL_ERROR;
	  }

	  if (prev_obj_name_ptr == NULL) {
	      nlist->names = obj_name_ptr->next;
	  } else {
	      prev_obj_name_ptr->next = obj_name_ptr->next;
	  }
	  ckfree((char *) obj_name_ptr);
	  
	  /* Check to see if you've removed to last object from this tag. If
	   * so, remove the hash entry. 
	   */
	  
	  if (nlist->names == NULL) {
	      Tcl_DeleteHashEntry(n_hash_entry);
	      ckfree((char *) nlist);
	  }
      }

      temp_tag_ptr = tag_ptr;
      tag_ptr = tag_ptr->next;
      ckfree((char *) temp_tag_ptr);
  }

  Tcl_DeleteHashEntry(hash_entry);
  ckfree((char *) tlist);
  
  return TCL_OK;
}

static int
ObjectGetTags(Tcl_Interp *interp,
	      int CID,
	      char *objName)
{
  Tk_Uid nameId;
  TcmTagList *tlist;
  TcmTag *tag_ptr;
  int targc; 
  Tcl_HashEntry *hash_entry;
  int new;
  char *targv[256];

  nameId = Tk_GetUid(objName);

  hash_entry = Tcl_CreateHashEntry(&objectTable, (char *) nameId, &new);

  /* If no hash entry, never tagged, return empty list. */

  if (new) {
    Tcl_DeleteHashEntry(hash_entry);
    interp->result[0] = '\0';
    return TCL_OK;
  }

  tlist = (TcmTagList *) Tcl_GetHashValue(hash_entry);

  /* If there are no tags, there is a problem. */

  if (tlist->tags == NULL) {
    Tcl_AppendResult(interp, "Inconsistency in object tag tables", (char *) NULL);
    return TCL_ERROR;
  }

  tag_ptr = tlist->tags;
  targc = 0;

  while (tag_ptr != NULL) {
    if (tag_ptr->CID == CID) {
      targv[targc] = tag_ptr->tag;
      targc++;
    }
    tag_ptr = tag_ptr->next;
  }

  /* If no tags with right CID, return empty string. */

  if (targc == 0) {
    interp->result[0] = '\0';
    return TCL_OK;
  }
    
  Tcl_SetResult(interp, Tcl_Merge(targc, targv), (Tcl_FreeProc *) free);

  return TCL_OK;
}

  
static int 
ObjectGetAllTags(Tcl_Interp *interp,
		     char *objName)
{
  char *argv[256];
  Tcl_DString dvec[256];
  int argc;
  int cid_vector[256];
  TcmTagList *tlist;
  TcmTag *tagPtr;
  int i, cntr;
  Tcl_HashEntry *hash_entry;
  int new;
  Tk_Uid nameId;

  nameId = Tk_GetUid(objName);

  hash_entry = Tcl_CreateHashEntry(&objectTable, (char *) nameId, &new);

  /* If hash entry doesn't exist, object never tagged, return empty string. */

  if (new) {
    Tcl_DeleteHashEntry(hash_entry);
    interp->result[0] = '\0';
    return TCL_OK;
  }

  tlist = (TcmTagList *) Tcl_GetHashValue(hash_entry);

  if (tlist->tags == NULL) { 
    interp->result[0] = '\0';
    return TCL_OK;
  }

  tagPtr = tlist->tags;
  argc = 0;

  while (tagPtr != NULL) {
    cntr = 0;

    while (cntr != argc) {
      if (tagPtr->CID == cid_vector[cntr]) {
	argv[cntr] = Tcl_DStringAppendElement(&(dvec[cntr]), tagPtr->tag);
	break;
      }
      cntr++;
    }

    if (cntr == argc) {
      char cidstr[256];

      cid_vector[cntr] = tagPtr->CID;
      argc++;

      Tcl_DStringInit(&(dvec[cntr]));
      sprintf(cidstr, "%d", tagPtr->CID);
      Tcl_DStringAppendElement(&(dvec[cntr]), cidstr);
      Tcl_DStringStartSublist(&(dvec[cntr]));
      Tcl_DStringAppendElement(&(dvec[cntr]), tagPtr->tag);
    }
    tagPtr = tagPtr->next;
  }

  for (i=0; i<argc; i++) {
    Tcl_DStringEndSublist(&(dvec[i]));
    argv[i] = dvec[i].string;
  }
  
  Tcl_SetResult(interp, Tcl_Merge(argc, argv), (Tcl_FreeProc *) free);

  for (i=0; i<argc; i++) {
    Tcl_DStringFree(&(dvec[i]));
  }

  return TCL_OK;
}

static int 
ObjectGetObjects(Tcl_Interp *interp,
		int CID,
		char *tag)
{

  Tk_Uid tagId;
  Tcl_HashEntry *hash_entry;
  int new;
  char *nargv[256];
  int nargc;
  ObjNameList *nlist;
  ObjName *namePtr;

  tagId = Tk_GetUid(tag);
  
  hash_entry = Tcl_CreateHashEntry(&tagTable, (char *) tagId, &new);

  /* If new, tag doesn't exist. Return empty string. */

  if (new) {
    Tcl_DeleteHashEntry(hash_entry);
    interp->result[0] = '\0';
    return TCL_OK;
  }

  nlist = (ObjNameList *) Tcl_GetHashValue(hash_entry);

  namePtr = nlist->names;

  if (namePtr == NULL) {
    interp->result[0] = '\0';
    return TCL_OK;
  }

  nargc = 0;

  while (namePtr != NULL) {
    if (namePtr->CID == CID) {
      nargv[nargc] = namePtr->name;
      nargc++;
    }

    namePtr = namePtr->next;
  }

  if (nargc == 0) {
    interp->result[0] = '\0';
    return TCL_OK;
  }

  Tcl_SetResult(interp, Tcl_Merge(nargc, nargv), (Tcl_FreeProc *) free);

  return TCL_OK;
}    

static int 
ObjectGetAllObjects(Tcl_Interp *interp,
			char *tag)
{
  char *argv[256];
  Tcl_DString dvec[256];
  int argc;
  int cid_vector[256];
  ObjNameList *nlist;
  ObjName *namePtr;
  int i, cntr;
  Tcl_HashEntry *hash_entry;
  int new;
  Tk_Uid tagId;

  tagId = Tk_GetUid(tag);
  
  hash_entry = Tcl_CreateHashEntry(&tagTable, (char *) tagId, &new);

  /* If hash entry doesn't exist, object never tagged, return empty string. */

  if (new) {
    Tcl_DeleteHashEntry(hash_entry);
    interp->result[0] = '\0';
    return TCL_OK;
  }

  nlist = (ObjNameList *) Tcl_GetHashValue(hash_entry);

  if (nlist->names == NULL) { 
    interp->result[0] = '\0';
    return TCL_OK;
  }

  namePtr = nlist->names;
  argc = 0;

  while (namePtr != NULL) {
    cntr = 0;

    while (cntr != argc) {
      if (namePtr->CID == cid_vector[cntr]) {
	argv[cntr] = Tcl_DStringAppendElement(&(dvec[cntr]), namePtr->name);
	break;
      }
      cntr++;
    }

    if (cntr == argc) {
      char cidstr[256];

      cid_vector[cntr] = namePtr->CID;
      argc++;

      Tcl_DStringInit(&(dvec[cntr]));
      sprintf(cidstr, "%d", namePtr->CID);
      Tcl_DStringAppendElement(&(dvec[cntr]), cidstr);
      Tcl_DStringStartSublist(&(dvec[cntr]));
      Tcl_DStringAppendElement(&(dvec[cntr]), namePtr->name);
    }
    namePtr = namePtr->next;
  }

  for (i=0; i<argc; i++) {
    Tcl_DStringEndSublist(&(dvec[i]));
    argv[i] = dvec[i].string;
  }
  
  Tcl_SetResult(interp, Tcl_Merge(argc, argv), (Tcl_FreeProc *) free);

  for (i=0; i<argc; i++) {
    Tcl_DStringFree(&(dvec[i]));
  }

  return TCL_OK;
}    
    


int
Tcm_ObjectTagCmd(ClientData clientData, 
		 Tcl_Interp *interp, 
		 int argc, 
		 char **argv)
{
  char c;
  int length;
  int CID;

  if (argc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"",
		      argv[0], " command ?arguments?\"", (char *) NULL);
    return TCL_ERROR;
  }

  /* Initialize tables if necessary. */

  if (!initialized) 
    ObjTagInit();

  c = argv[1][0];
  length = strlen(argv[1]);

  if ((c == 'a') && (strncmp(argv[1], "addTag", length) == 0)) {
    
    if (argc < 5) {
      Tcl_AppendResult (interp, "wrong # args: should be\"",
			argv[0], "addTag CID objName tag\"", (char *) NULL);
      return  TCL_ERROR;
    }

    if (Tcl_GetInt(interp, argv[2], &CID) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad CID: ", argv[2], (char *) NULL);
      return TCL_ERROR;
    }

    return ObjectAddTag(interp, CID, argv[3], argv[4]);

  } else if ((c == 'd') && (strncmp(argv[1], "delTag", length) == 0)) {

    if (argc < 5) {
      Tcl_AppendResult (interp, "wrong # args: should be\"",
			argv[0], "delTag CID objName tag\"", (char *) NULL);
      return  TCL_ERROR;
    }

    if (Tcl_GetInt(interp, argv[2], &CID) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad CID: ", argv[2], (char *) NULL);
      return TCL_ERROR;
    }
    
    return ObjectDelTag(interp, CID, argv[3], argv[4]);
  } else if ((c == 'g') && (strncmp(argv[1], "getTags", length) == 0)) {
    
    if (argc < 4) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
		       argv[0], "getTags CID objName\"", (char *) NULL);
      return TCL_ERROR;
    }

    if (Tcl_GetInt(interp, argv[2], &CID) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad CID: ", argv[2], (char *) NULL);
      return TCL_ERROR;
    }

    return ObjectGetTags(interp, CID, argv[3]);

  } else if  ((c == 'g') && (strncmp(argv[1], "getAllTags", length) == 0)) {
    
    if (argc < 3) {
      Tcl_AppendResult(interp, "wrong # args: should be\"",
		       argv[0], "getAllTags objName\"", (char *) NULL);
      return TCL_ERROR;
    }

    return ObjectGetAllTags(interp, argv[2]);

  } else if ((c == 'g') && (strncmp(argv[1], "getObjs", length) == 0)) {
    
    if (argc < 4) {
      Tcl_AppendResult(interp, "wrong # args: should be\"",
		       argv[0], "getObjs CID tag\"", (char *) NULL);
      return TCL_ERROR;
    }

    if (Tcl_GetInt(interp, argv[2], &CID) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad CID: ", argv[2], (char *) NULL);
      return TCL_ERROR;
    }

    return ObjectGetObjects(interp, CID, argv[3]);

  } else if ((c == 'g') && (strncmp(argv[1], "getAllObjs", length) == 0)) {
    
    if (argc < 3) {
      Tcl_AppendResult(interp, "wrong # args: should be\"",
		       argv[0], "getAllObjs tag\"", (char *) NULL);
      return TCL_ERROR;
    }

    return ObjectGetAllObjects(interp, argv[2]);

  } else if ((c == 'd') && (strncmp(argv[1], "delAllTags", length) == 0)) {
    if (argc < 4) {
      Tcl_AppendResult(interp, "wrong # args: should be\"",
		       argv[0], "delAllTags CID objName\"", (char *) NULL);
      return TCL_ERROR;
    }

    if (Tcl_GetInt(interp, argv[2], &CID) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad CID: ", argv[2], (char *) NULL);
      return TCL_ERROR;
    }

    return ObjectDeleteAllTags(interp, CID, argv[3]);
  }
  else {
    Tcl_AppendResult(interp, "bad objectTag command \"", argv[1], "\" should be: ",
		     "addTag, delTag, getTags, getAllTags, getObjs, delAllTags, or getAllObjs",
		     (char *) NULL);
    return TCL_ERROR;
  }
}


