/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifndef PACKET_H
#define PACKET_H

/*
 * Actual size in bytes of a packet
 */
#define PACKET_SIZE 8192
#define PACKET_DATA_LEN (PACKET_SIZE-sizeof(PacketHeader))

/*
 * The following structure is part of the packet for cyclic-UDP
 * connections
 */
typedef struct PacketHeader {
    int epoch;			/* Epoch # for packet */
    int epochBytes;		/* # bytes sent in the prev epoch */
    int epochDuration;		/* Duration of the prev epoch (msec) */
    int sendTimeSec;		/* time packet was sent (seconds) */
    int sendTimeUSec;		/* time packet was sent (useconds) */
    int cycleId;		/* Increases by at least one each cycle */
    int bufferSize;		/* Size of buffer */
    int bufferId;		/* The id of the buffer (for reassembly) */
    int packetSize;		/* Size of this packet */
    int packetId;		/* Unique id for this packet */
    int packetOfs;		/* Offset of this packet's data in its buffer */
    int packetNum;		/* Which packet this is within its buffer */
    int burstId;		/* Burst number when packet was sent */
    int ppc;			/* Packets in this cycle (pkts/cycle) */
    int bpc;			/* Buffers in this cycle (buffers/cycle) */
    int ppb;			/* Packets in this buffer (pkts/buffer) */
    int EOC;			/* non zero means this packet marks the end
    				 * of cycle */
} PacketHeader;

/*
 * A ResendRequest requests the source to resend one or more packets
 * from a cycle that are presumed dead.  Each request specifies 
 * the cycle number and set of packets to resend (represented
 * as an array of pktNum and burstNum).  The resend request is sent
 * in network byte order.
 *
 * ResendRequests are sent out periodically over a UDP socket from the
 * dest to the source.
 *
 */
typedef struct ResendRequest {
    int magic;			/* Magic Number telling pktSrc what this is */
    int cycleId;		/* Cycle ID of packets */
    int numReq;			/* Number of packets requested for resend */
    int burstNum;		/* Burst number when detected */
    int *pktNum;		/* Array of packet numbers to be resent. */
} ResendRequest;

/*
 * A feedback unit is a small UDP packet sent from the Dest to the source
 * telling it about the state of the connection.
 */
typedef struct FeedbackUnit {
    int magic;			/* Magic Number telling pktSrc what this is */
    int bandwidth;		/* bandwidth during this epoch */
    int lossRate;		/* % of bytes lost */
    int delay;			/* Average packet delay */
    int delayDev;		/* Mean deviation of delay */
    int epoch;			/* Which epoch this FU corresponds to */
    int skewMsec;		/* Skew between clocks (milliseconds).  This
				 * value should be added to the local system
				 * clock to get the destinations system
				 * clock time */

} FeedbackUnit;

/*
 * A finished marker is a packet sent from the Dest to the source
 * telling it we've completely received the current cycle.
 */
typedef struct FinishedMarker {
    int magic;			/* Magic Number telling pktSrc what this is */
    int cycleId;		/* bandwidth during this epoch */
    int epoch;			/* Which epoch this FU corresponds to */
} FinishedMarker;

#define FEEDBACK_UNIT_MAGIC	1
#define RESEND_REQUEST_MAGIC	2
#define FINISHED_MARKER_MAGIC	3

/*
 * The following structure is part of the packet for simple-UDP
 * connections
 */
typedef struct UdpHeader {
    int size;			/* Size of this buffer */
} UdpHeader;

#endif
