/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmPacketDest.c,v 1.35 1996/02/07 20:33:48 kpatel Exp $
 *
 *	This module receives packets from the PacketSrc,
 *	re-assembles them in to frames and passes them up
 *	to the media dest modules.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <unistd.h>
#include <limits.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include "tcl.h"
#include "dp.h"
#include "tk.h"
#include "cmBufferlist.h"
#include "cmPacket.h"
#include "cmAtom.h"
#include "cmUtil.h"

#ifdef NO_SENDRECVMSG
#    include "compat/sendrecvmsg.h"
#endif

#ifndef _WINDOWS
#    ifndef abs
#        define abs(a)	((a)>0?(a):-(a))
#    endif
#    define min(a,b)	((a)>(b)?(b):(a))
#    define max(a,b)	((a)>(b)?(a):(b))
#endif

/*
 * Maximum number of reads we do at one time
 */
#define MAX_SEQ_READS 5

struct CMBundle;

/*
 * An actual PacketDest object
 */
typedef struct PacketDest {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */

    char *outCmd;		/* The command to push data */
    InFunc outCCmd;		/* C pointer to out command */
    void *outObj;		/* The object to send to */

    DP_SOCKET udp;		/* File descriptor for UDP data socket */
    int port;			/* Port number for UDP data socket */
    int shared;			/* Should buffers be in shared memory */
    int xAttach;		/* Should buffers be attached
				 * to the X server */
    char *address;		/* String with internet address and port */
    char *pg;			/* Priority group */
    struct CMBundle *bundles;	/* Linked list of current sources
				 * we are receiving from keyed by srcAddr */
} PacketDest;

/*
 * One of these structs is kept for each source
 * that we are receiving packets from
 */
typedef struct CMBundle {
    struct CMBundle *next;	/* Linked list pointer */
    
    struct sockaddr_in srcAddr;	/* Where packets are coming from */

    int *packetMask[2];		/* Mask of packets received in each cycle */
    int *callbackMask[2];	/* mask of callbacks made */
    int buffers[2];		/* How many bufferss in the last 2 cycles */

    int skewSec;		/* skew this cycle */
    int skewUsec;
    int prevSkewSec;		/* skew last cycle */
    int prevSkewUsec;

    int meanDelay;		/* The weighted mean delay so far */
    int delayDev;		/* Deviation in the delay */
    
    int cycle;			/* The cycleId of the last packet */
    int epoch;			/* The current epoch */

    int pktsRecvd;		/* packets received this cycle */
    int bytesRecvd;		/* total bytes received */
    
    int lastPacket;		/* Id of the last packet received */
    
    PacketDest *pdPtr;		/* The Packet Dest this bundle is part of */

} CMBundle;

/* Globals used by this object module. */

static int nextPacketDestNum = 0;   /* Variable for next unique object name. */

/*
 * Prototypes for routines in the file
 */
int			PacketDestModuleInit _ANSI_ARGS_((Tcl_Interp *interp));

static PacketDest *	CreateNewPacketDest _ANSI_ARGS_((Tcl_Interp *interp,
			    char *name));
static int		ConfigurePacketDest _ANSI_ARGS_((PacketDest *objPtr,
			    int argc, char **argv));
static int		DestroyPacketDest _ANSI_ARGS_((PacketDest *objPtr));
static int		PacketDestCreateCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
static int		PacketDestObjCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
static void		HandleIncomingPacket _ANSI_ARGS_((ClientData clientData,
			    int mask));
static CMBundle *	AddSource _ANSI_ARGS_((PacketDest *pdPtr,
			    struct sockaddr_in *srcAddr));
static Buffer *		FindPacketBuffer _ANSI_ARGS_((PacketDest *pdPtr,
			    PacketHeader *phtPtr));
static void		NewCycle _ANSI_ARGS_((PacketDest *pdPtr,
			    CMBundle *bundle, int newCycle, int packets,
			    int buffers));
static void		FinishedBuffer _ANSI_ARGS_((Buffer *bufPtr,
			    CMBundle *bundle, PacketDest *pdPtr,
			    PacketHeader *phPtr));

/************************************************************
 *
 * CreateNewPacketDest
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static PacketDest *
CreateNewPacketDest (Tcl_Interp *interp, char *name)
{
    PacketDest *new;
    DP_SOCKET udp;
    int port, len, res;
    struct sockaddr_in sockaddr;
    char str[256];
    int optname, optval, optlen;
    int ll, ul;

    udp = Tdp_inet_connect ("", 0, 0, 1, 0, 0);
    if (udp == DP_INVALID_SOCKET) {
	Tcl_AppendResult (interp, "Couldn't open udp socket : ",
			  Tcl_PosixError(interp), (char *)NULL);
	return NULL;
    }

    /*
     * Figure out port number
     */
    len = sizeof (sockaddr);
    res = getsockname (udp, (struct sockaddr *) &sockaddr, &len);
    if (res < 0) {
	Tcl_AppendResult (interp, "Error getting port number of udp socket : ",
			  Tdp_Error(interp), (char *)NULL);
	closesocket(udp);
	return NULL;
	
    }
    port = ntohs(sockaddr.sin_port);

    /*
     * Set the receive buffer to a large value to try to prevent
     * overflows.  Since this varies from OS to OS, we just search for
     * the max size buffer less than 100k
     */
    optname = SO_RCVBUF;
    optlen = sizeof(int);
    res = getsockopt(udp, SOL_SOCKET, optname, (char *)&optval, &optlen);
    if (res != -1) {

	/*
	 * Phase 1: Double size of buffer until we get an error or
	 * buffer size is 50k
	 */
	if (optval < 1) {
	    optval = 1024;
	}
	while ((res != -1) && (optval < 1000000)) {
	    optval <<= 1;
	    res = setsockopt(udp, SOL_SOCKET, optname, (char *)&optval, optlen);
	}

	/*
	 * Phase 2: Do binary search on max buffer size.
	 */
	ul = optval;
	ll = optval >> 1;
	optval = (ll+ul)/2;
	while ((optval < 1000000) && (ul-ll > 1000)) {
	    res = setsockopt(udp, SOL_SOCKET, optname, (char *)&optval, optlen);
	    if (res == -1) {
		ul = optval;
	    } else {
		ll = optval;
	    }
	    optval = (ll+ul)/2;
	}
    }

    /*
     * Create the new object!
     */
    new = (PacketDest *) ckalloc(sizeof(PacketDest));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outObj = NULL;

    new->udp = udp;
    new->port = port;
    new->shared = 0;
    new->xAttach = 0;
    new->pg = ckalloc(3);
    memcpy(new->pg, "tk", 3);
    new->bundles = NULL;

    sprintf (str, "%s %d", Tcm_GetLocalAddress(interp), port);
    new->address = ckalloc(strlen(str)+1);
    strcpy(new->address, str);

    /*
     * Set up a handler on the socket
     */
    Tcm_CreateFileHandler(new->pg, udp, TK_READABLE, HandleIncomingPacket,
			  (ClientData)new);
    
    return new;
}


/************************************************************
 *
 * DestroyPacketDest
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 ************************************************************/

static int 
DestroyPacketDest (PacketDest *objPtr)
{
    CMBundle *bundle, *next;

    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Delete the bundles held by this PacketDest
     */
    for (bundle = objPtr->bundles; bundle != NULL; bundle = next) {
	next = bundle->next;
	
	if (bundle->packetMask[0] != NULL) {
	    ckfree((char *) bundle->packetMask[0]);
	}
	if (bundle->packetMask[1] != NULL) {
	    ckfree((char *) bundle->packetMask[1]);
	}
    
	if (bundle->callbackMask[0] != NULL) {
	    ckfree((char *) bundle->callbackMask[0]);
	}
	if (bundle->callbackMask[1] != NULL) {
	    ckfree((char *) bundle->callbackMask[1]);
	}
	
	ckfree((char *) bundle);
    }

    /*
     * Close the socket, free up memory.
     */
    Tk_DeleteFileHandler(objPtr->udp);
    closesocket (objPtr->udp);
    if (objPtr->name != NULL) {
	ckfree((char *) objPtr->name);
    }
    if (objPtr->address != NULL) {
	ckfree((char *) objPtr->address);
    }

    ckfree((char *) objPtr->pg);
    ckfree((char *) objPtr);
    return TCL_OK;
}


/************************************************************
 *
 * PacketDestCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the pktDest command.
 *     A new object of type PacketDest is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

static int
PacketDestCreateCmd (ClientData clientData,
		     Tcl_Interp *interp, int argc, char **argv)
{
    char new_name[256];
    PacketDest *new;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent", (char *) NULL);
	return TCL_ERROR;
    }

    nextPacketDestNum++;
    sprintf(new_name, "%spd%d", argv[1], nextPacketDestNum);

    new = CreateNewPacketDest(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, new_name, PacketDestObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * PacketDestObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     PacketDest object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

static int
PacketDestObjCmd (ClientData clientData, Tcl_Interp *interp,
		  int argc, char **argv)
{

    char c;
    int len;
    PacketDest *objPtr = (PacketDest *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 "<command> ?arguments?", (char *) NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyPacketDest(objPtr);
    } else if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigurePacketDest(objPtr, argc-2, argv+2);
    } else if ((c == 'a') && (strncmp("address", argv[1], len) == 0)) {
	Tcl_SetResult (interp, objPtr->address, TCL_VOLATILE);
	return TCL_OK;
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ",
			argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  address\n",
                (char *) NULL);
    return TCL_ERROR;
}


/************************************************************
 *
 * ConfigurePacketDest
 *
 * Parameters: 
 *	objPtr - the object to configure
 *	argc - argument count
 *	argv - argument values
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *	Configures the packet dest object
 *
 ************************************************************/

static int
ConfigurePacketDest (PacketDest *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    
    interp = objPtr->interp;
    
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", NULL);
	} else {
	    Tcl_AppendResult(interp,
			     "{-outCmd {} {", objPtr->outCmd, "}} ",
			     (char *) NULL);
	}
	Tcl_AppendResult(interp, "{-shared {false} {",
			 (objPtr->shared ? "true" : "false"),
			 "}} ", NULL);
	Tcl_AppendResult(interp, "{-attach {false} {",
			 (objPtr->xAttach ? "true" : "false"),
			 "}} ", NULL);
	Tcl_AppendResult(interp, "{-address {", objPtr->address, "} {",
			 objPtr->address, "}} ", NULL);
	Tcl_AppendResult(interp, "{-priorityGroup {tk} {",
			 objPtr->pg, "}} ", NULL);
	return TCL_OK;
    } 
    
    if (argc == 1) {
	int len;
	
	len = strlen(argv[0]);
	if (strncmp("-outCmd", argv[0], len) == 0) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	    } else {
		Tcl_AppendResult(interp,
				 "-outCmd {} {", objPtr->outCmd, "}",
				 (char *) NULL);
	    }
	    return TCL_OK;
	} else if (strncmp("-shared", argv[0], len) == 0) {
	    Tcl_AppendResult(interp, "-shared {false} {",
			     (objPtr->shared ? "true" : "false"),
			     "} ", NULL);
	    return TCL_OK;
	} else if (strncmp("-attach", argv[0], len) == 0) {
	    Tcl_AppendResult(interp, "-attach {false} {",
			     (objPtr->xAttach ? "true" : "false"),
			     "} ", NULL);
	    return TCL_OK;
	} else if (strncmp("-address", argv[0], len) == 0) {
	    Tcl_AppendResult(interp, "-address {", objPtr->address, "} {",
			     objPtr->address, "} ", NULL);
	    return TCL_OK;
	} else if (strncmp("-priorityGroup", argv[0], len) == 0) {
	    Tcl_AppendResult(interp, "-priorityGroup {tk} {",
			     objPtr->pg, "}", NULL);
	    return TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "bad configuration option: ", NULL);
	    return TCL_ERROR;
	}
    }
    
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
  
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    Tcl_CmdInfo cmdInfo;
	    char *objName, *division;
    
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -outCmd",
				 NULL);
		return TCL_ERROR;
	    }
    
#ifdef SECURE_CMT
	    if (argv[cntr][0] != '@') {
		Tcl_SetResult(objPtr->interp, "Cannot set Tcl outCmd because this program was compiled with -DSECURE_CMT.\nRecompile without security to use arbitrary Tcl outCmd's\n", TCL_STATIC);
		return TCL_ERROR;
	    }
#endif

	    if (objPtr->outCmd != NULL) {
		ckfree((char *) objPtr->outCmd);
	    }
    
	    objPtr->outCmd = ckalloc(strlen(argv[cntr]) + 1);
	    strcpy(objPtr->outCmd, argv[cntr]);
	    
	    if (objPtr->outCmd[0] == '@') {
		if (Tutil_GetRegisteredCommand(interp, objPtr->outCmd,
					       (ClientData *) &(objPtr->outObj),
					       (ClientData *) &(objPtr->outCCmd)) != TCL_OK) {
		    Tcl_AppendResult(interp,
				     "Could not find register procedure ",
				     argv[cntr] + 1, NULL);
		    ckfree(objPtr->outCmd);
		    objPtr->outCmd = NULL;
		    objPtr->outObj = NULL;
		    objPtr->outCCmd = NULL;
		    return TCL_ERROR;
		}
	    } else {
		objPtr->outObj = NULL;
		objPtr->outCCmd = NULL;
	    }
	} else if ((c == 's') && (strncmp("-shared", argv[cntr], len) == 0)) {
	    int newshared;
	    
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -shared",
				 NULL);
		return TCL_ERROR;
	    }
	    
	    if (Tcl_GetBoolean(interp, argv[cntr], &newshared) != TCL_OK) {
		return TCL_ERROR;
	    }
	    objPtr->shared = newshared;
	} else if ((c == 'a') && (strncmp("-attach", argv[cntr], len) == 0)) {
	    int newattach;
	    
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -attach",
				 NULL);
		return TCL_ERROR;
	    }
	    
	    if (Tcl_GetBoolean(interp, argv[cntr], &newattach) != TCL_OK) {
		return TCL_ERROR;
	    }
#ifndef _WINDOWS
#ifndef NO_X11
	    if (cmXShmQueryExtension(Tcm_Display())) {
		objPtr->xAttach = newattach;
	    }
#endif
#endif
	} else if ((c == 'a') && (strncmp("-address", argv[cntr], len) == 0)) {
	    Tcl_AppendResult(interp,
			     "Can't set address slot of a pktDest", NULL);
	    return TCL_ERROR;
	} else if ((c == 'p') && (strncmp("-priorityGroup", argv[cntr], len) == 0)){
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -priorityGroup",
				 NULL);
		return TCL_ERROR;
	    }
	    if (Tcm_FindPg(argv[cntr]) == NULL) {
		Tcl_AppendResult(interp, "unknown priority group \"", argv[cntr],
				 "\"", NULL);
		return TCL_ERROR;
	    }
	    ckfree((char *) objPtr->pg);
	    objPtr->pg = (char *) ckalloc(strlen(argv[cntr]) + 1);
	    strcpy(objPtr->pg, argv[cntr]);
	    Tcm_DeleteFileHandler(objPtr->udp);
	    Tcm_CreateFileHandler(objPtr->pg, objPtr->udp, TK_READABLE,
				  HandleIncomingPacket, (ClientData) objPtr);
	} else {
	    Tcl_AppendResult(interp,
			     "Invalid configuration parameter ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}
	    
	cntr++;
    }
    return TCL_OK;
}



/************************************************************
 *
 * PacketDestModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     keywords as well as creating the Tcl command
 *     for object type creation
 *
 ************************************************************/

int
PacketDestModuleInit (Tcl_Interp *interp)
{
    /*
     * Force Tcm_GetLocalAddress to cache the local host ip address
     */
    Tcm_GetLocalAddress(interp);
    Tcl_CreateCommand(interp, "pktDest", PacketDestCreateCmd,
		      NULL, (void (*)()) NULL);
    Tcl_VarEval(interp, "cminfo set object_info ", "pktDest ",
		"{unicast network dest}", (char *) NULL);

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * AddSource
 *
 *	Adds another CMBundle representing a new source that
 *	this packetDest will receive from
 *
 * Results:
 *	The bundle that was added
 *
 * Side effects:
 *	A new CMBundle is allocated, initialized, and added
 *	to the hash table in the packetDest
 *
 *--------------------------------------------------------------
 */
static CMBundle *
AddSource(pdPtr, srcAddr)
    PacketDest *pdPtr;			/* PacketDest to add this bundle to */
    struct sockaddr_in *srcAddr;	/* Address where packets come from */
{
    CMBundle *bundle;

    /*
     * Create the CMBundle
     */
    bundle = (CMBundle *) ckalloc(sizeof(CMBundle));

    bundle->srcAddr.sin_port = srcAddr->sin_port;
    bundle->srcAddr.sin_addr.s_addr = srcAddr->sin_addr.s_addr;

    bundle->packetMask[0] = NULL;
    bundle->packetMask[1] = NULL;
    bundle->callbackMask[0] = NULL;
    bundle->callbackMask[1] = NULL;

    bundle->buffers[0] = 0;
    bundle->buffers[1] = 0;

    bundle->skewSec = INT_MAX;
    bundle->skewUsec = INT_MAX;
    bundle->prevSkewSec = INT_MAX;
    bundle->prevSkewUsec = INT_MAX;

    bundle->meanDelay = 0;
    bundle->delayDev = 0;
    
    bundle->cycle = -1;
    bundle->epoch = -1;
    
    bundle->pktsRecvd = 0;
    bundle->bytesRecvd = 0;
    
    bundle->lastPacket = -1;
	
    bundle->pdPtr = pdPtr;
    
    /*
     * Add it to the list
     */
    bundle->next = pdPtr->bundles;
    pdPtr->bundles = bundle;
    
    return bundle;
}


/*
 *--------------------------------------------------------------
 *
 * HandleIncomingPacket --
 *
 *	Called whenever a packet is available to be read
 *	from the socket.  Retrieves the packet and figures
 *	out how to handle it.
 *
 * Results:
 *	no return value
 *
 * Side effects:
 *	Calls various procdures to store the contents of the
 *	packet in the appropriate place and make a callback
 *	if necessary when a frame is completely received
 *
 *--------------------------------------------------------------
 */
static void
HandleIncomingPacket(clientData, mask)
    ClientData clientData;	/* Actually pointer to a PacketDest */
    int mask;		  	/* Unused */
{
    PacketDest *pdPtr;
    int count;
    struct sockaddr_in addr;
    struct iovec vec[2];
    struct msghdr msghdr;
    int addrLen;
    PacketHeader ph, dummyPh;
    Buffer *bufPtr;
    struct timeval recvTime, startTime, endTime;
    int delaySec, delayUsec, currDelay, prevDelay;
    int delay, err;
    CMBundle *bundle;
    FeedbackUnit fu;
    ResendRequest rr;
    int *resends, resendCount;
    FinishedMarker fm;
    int packet, pktMask, done, packetIndex, numBytes;

    pdPtr = (PacketDest *) clientData;
    bufPtr = NULL;

    /*
     * Limit the amount of work we do at one time
     */
    count = MAX_SEQ_READS;
    
    while (count-- && (Tdp_SocketIsReady(pdPtr->udp) & TCL_FILE_READABLE)) {
	gettimeofday(&startTime, NULL);
	
	/*
	 * Peek at the packet header
	 */
	addrLen = sizeof(addr);
	memset((char *)&addr, 0, addrLen);
	
	if (recvfrom(pdPtr->udp, (char *) &ph, sizeof(PacketHeader),
			MSG_PEEK, (struct sockaddr *)&addr, &addrLen) == -1) {
	    int error;

	    error = DP_GetLastError();
	    if (error != DP_EMSGSIZE) {
		Tcl_AppendResult(pdPtr->interp,
				 "Error reading socket",
				 Tdp_Error(pdPtr->interp),
				 NULL);
		return;
	    }
	}
	
	/*
	 * Note the time we got the packet
	 */
	gettimeofday(&recvTime, NULL);
	
	/*
	 * Convert the fields of the packet header
	 */
	ph.sendTimeSec = ntohl(ph.sendTimeSec);
	ph.sendTimeUSec = ntohl(ph.sendTimeUSec);
	ph.cycleId = ntohl(ph.cycleId);
	ph.epoch = ntohl(ph.epoch);
	ph.epochDuration = ntohl(ph.epochDuration);
	ph.epochBytes = ntohl(ph.epochBytes);
	ph.bufferSize = ntohl(ph.bufferSize);
	ph.bufferId = ntohl(ph.bufferId);
	ph.packetSize = ntohl(ph.packetSize);
	ph.packetId = ntohl(ph.packetId);
	ph.packetOfs = ntohl(ph.packetOfs);
	ph.packetNum = ntohl(ph.packetNum);
	ph.burstId = ntohl(ph.burstId);
	ph.ppc = ntohl(ph.ppc);
	ph.bpc = ntohl(ph.bpc);
	ph.ppb = ntohl(ph.ppb);
	ph.EOC = ntohl(ph.EOC);
	packetIndex = ph.cycleId % 2;

	
/*	printf("Got packet %d:\n\tcycle %d, size %d, buffer %d, burst %d\n",
	       ph.packetId, ph.cycleId, ph.bufferSize, ph.bufferId,
	       ph.burstId);
	printf("\t%d ppc, %d bpc, %d ppb, EOC = %d\n", ph.ppc, ph.bpc,
	       ph.ppb, ph.EOC);
 */	

	/*
	 * Find the appropriate CMBundle
	 */
	for (bundle = pdPtr->bundles; bundle != NULL; bundle = bundle->next) {
	    if ((bundle->srcAddr.sin_port != addr.sin_port) ||
		(bundle->srcAddr.sin_addr.s_addr != addr.sin_addr.s_addr)) {
		continue;
	    }
	    break;
	}
	
	if (bundle == NULL) {
	    bundle = AddSource(pdPtr, &addr);
	    /* printf("New source was created\n"); */
	}

	/*
	 * Check for a new cycle id
	 */
	if (ph.cycleId > bundle->cycle ) {
	    NewCycle(pdPtr, bundle, ph.cycleId, ph.ppc, ph.bpc);
	}

	if ((ph.cycleId < (bundle->cycle - 1))
	    || (bundle->packetMask[packetIndex][ph.packetId / 32]
		 & (1 << (ph.packetId % 32)))) {
	    /*
	     * Worthless packet -- just read it and ignore it
	     */
	    memset((char *)&addr, 0, addrLen);
	    addrLen = sizeof(addr);
	    recvfrom(pdPtr->udp, (char *)&dummyPh, sizeof(PacketHeader),
		     0, (struct sockaddr *)&addr, &addrLen);
	    continue;
	} else if (ph.EOC) {
	    /*
	     * Grab the data but continue on
	     */
	    memset((char *)&addr, 0, addrLen);
	    addrLen = sizeof(addr);
	    numBytes = recvfrom(pdPtr->udp, (char *)&dummyPh,
				sizeof(PacketHeader), 0,
				(struct sockaddr *)&addr, &addrLen);
	} else {
	    /*
	     * Find the buffer where this packet belongs
	     */
	    bufPtr = FindPacketBuffer(pdPtr, &ph);

	    /*
	     * Grab the actual data
	     */
	    vec[0].iov_base = (char *) &dummyPh;
	    vec[0].iov_len = sizeof(PacketHeader);
	    vec[1].iov_base = bufPtr->data + ph.packetOfs;
	    vec[1].iov_len = ph.packetSize;
	    
	    memset((char *)&addr, 0, addrLen);
	    msghdr.msg_namelen = sizeof(addr);
	    msghdr.msg_name = (char *) &addr;
	    msghdr.msg_iov = vec;
	    msghdr.msg_iovlen = 2;
#if !defined(_SCO_DS)
	    msghdr.msg_accrights = NULL;
	    msghdr.msg_accrightslen = 0;
#endif 
	    numBytes = recvmsg(pdPtr->udp, &msghdr, 0);


/*
printf("After:\tPDPtr: %x Cycle %d, Packet Id %d, Packet Num %d, PBuffer Id %d, PPB %d, PPC %d\n",
       pdPtr, ph.cycleId, ph.packetId, ph.packetNum, ph.bufferId, ph.ppb, ph.ppc);
printf("      \tBuffer %d, first 4 bytes: %x num_read: %d\n", bufPtr->id, *(int*)bufPtr->data, numBytes);
*/


	    /*
	     * Mark this packet as received
	     */
	    if ((bundle->packetMask[packetIndex][ph.packetId / 32]
		 & (1 << (ph.packetId % 32))) == 0) {
		bundle->packetMask[packetIndex][ph.packetId / 32] |=
		    1 << (ph.packetId % 32);
		
		bundle->pktsRecvd++;
		
		if (bundle->pktsRecvd == ph.ppc) {
		    /*
		     * Send back a finished marker if we've
		     * finished the cycle
		     */
		    fm.magic = htonl(FINISHED_MARKER_MAGIC);
		    fm.cycleId = htonl(ph.cycleId);
		    fm.epoch = htonl(ph.epoch);
		    
		    sendto(pdPtr->udp, (char *)&fm, sizeof(fm), 0,
			   (struct sockaddr *)&addr, sizeof(addr));
		}
	    }
	}

	/*
	 * Update total bytes received
	 */
	bundle->bytesRecvd += numBytes;

	/*
	 * Do the bookkeeping for delay based bandwidth estimate
	 * First, calculate the real delay in milliseconds.
	 */
	delaySec = recvTime.tv_sec - ph.sendTimeSec;
	delayUsec = recvTime.tv_usec - ph.sendTimeUSec;

	if (((delaySec == bundle->skewSec) && (delayUsec < bundle->skewUsec))
	    || (delaySec < bundle->skewSec)) {
	    bundle->skewSec = delaySec;
	    bundle->skewUsec = delayUsec;
	    delaySec = delayUsec = 0;
	    currDelay = 0;
	} else {
	    currDelay = (delaySec - bundle->skewSec)*1000
		+ (delayUsec - bundle->skewUsec)/1000;
	}
	prevDelay = (delaySec - bundle->prevSkewSec)*1000
	    + (delayUsec - bundle->prevSkewUsec)/1000;
	delay = min(currDelay, prevDelay);
	delay = max(0,delay);

	/*
	 * Update weighted meanDelay and mean delay deviation
	 */
	err = delay - bundle->meanDelay;
	bundle->meanDelay += err >> 3;
	bundle->delayDev += (abs(err) - bundle->delayDev) >> 2;

	/*
	 * Check for a new epoch
	 */
	if (ph.epoch > bundle->epoch) {
	    /*
	     * Calculate stats
	     */
	    fu.magic = htonl(FEEDBACK_UNIT_MAGIC);
	    fu.bandwidth = htonl(1000 * bundle->bytesRecvd / ph.epochDuration);
	    if (ph.epochBytes != 0) {
		fu.lossRate = htonl(256 -
				    ((bundle->bytesRecvd<<8) / ph.epochBytes));
	    } else {
		fu.lossRate = 0;
	    }
	    fu.delay = htonl(bundle->meanDelay);
	    fu.delayDev = htonl(bundle->delayDev);
	    fu.epoch = htonl(ph.epoch - 1);
	    fu.skewMsec = htonl(bundle->skewSec*1000 + bundle->skewUsec/1000);
	    
	    /*
	     * Send them back -- socket on the other end is
	     * still in addr
	     */
	    sendto(pdPtr->udp, (char *)&fu, sizeof(fu), 0, 
		   (struct sockaddr *)&addr, sizeof(addr));
	    
	    bundle->epoch = ph.epoch;
	    bundle->bytesRecvd = 0;
	}
	
	/*
	 * Ask for resends if necessary
	 */
	resends = NULL;
	if ((ph.cycleId == bundle->cycle)
	    && (ph.packetId == (bundle->lastPacket + 1))) {

	    resendCount = 0;

	    for (packet = 0; packet < ph.packetId; packet++) {
		if ((bundle->packetMask[packetIndex][packet / 32] &
		     (1 << (packet % 32))) == 0) {
		    if (resendCount == 0) {
			resends =  (int *) ckalloc(ph.ppc * sizeof(int));
		    }
		    
		    resends[resendCount++] = htonl(packet);
		}
	    }

	    /*
	     * If we found anything, send off the request
	     */
	    if (resends != NULL) {

		/*
		 * Send it off
		 */
		rr.magic = htonl(RESEND_REQUEST_MAGIC);
		rr.cycleId = htonl(bundle->cycle);
		rr.numReq = htonl(resendCount);
		rr.burstNum = htonl(ph.burstId);
		vec[0].iov_base = (char *) &rr;
		vec[0].iov_len = 4 * sizeof(int);
		vec[1].iov_base = (caddr_t) resends;
		vec[1].iov_len = resendCount * sizeof(int);
	    
		msghdr.msg_iov = vec;
		msghdr.msg_iovlen = 2;
#if !defined(_SCO_DS)
		msghdr.msg_accrights = NULL;
		msghdr.msg_accrightslen = 0;
#endif	
		sendmsg(pdPtr->udp, &msghdr, 0);

		/*
		 * Free allocated memory
		 */
		ckfree((char *) resends);
	    }
	}

	if (ph.EOC == 0) {
	    /*
	     * Make the callback if the buffer was completed, but
	     * only make the callback once in case we start getting
	     * EOC packets or repeat packets
	     */
	    done = 1;
	    for (packet = 0; packet < ph.ppb; packet++) {
		pktMask = packet + ph.packetId - ph.packetNum;
		if ((bundle->packetMask[packetIndex][pktMask / 32] &
		     (1 << (pktMask % 32))) == 0) {
		    done = 0;
		}
	    }

	    if ((done) &&
		((bundle->callbackMask[packetIndex][ph.bufferId / 32] &
		  (1 << (ph.bufferId % 32)))) == 0) {

		FinishedBuffer(bufPtr, bundle, pdPtr, &ph);
	    }
	}
	
	bundle->lastPacket = ph.packetId;
	
	gettimeofday(&endTime, NULL);
/*	
 *	fprintf(stderr, "Time elapsed: %d seconds, %d microseconds\n",
 *		(endTime.tv_sec - startTime.tv_sec),
 *		(endTime.tv_usec - startTime.tv_usec));
 */
    }
}

/*
 *--------------------------------------------------------------
 *
 * FindPacketBuffer --
 *
 *	Finds the buffer where the data in this packet will go
 *
 * Results:
 *	A pointer to a buffer structure
 *
 * Side effects:
 *	If the buffer cannot be found (e.g., this is the first
 *	packet in this buffer) then a buffer may be taken from
 *	the free list or allocated if necessary
 *
 *--------------------------------------------------------------
 */
static Buffer *
FindPacketBuffer(pdPtr, phPtr)
    PacketDest *pdPtr;		/* The packet dest we are working with */
    PacketHeader *phPtr;	/* The header from this packet */
{
    FindQuery query;
    ConfigureInfo info;
    Buffer *bufPtr;
    char tagName[64];
    int id;
    int *hits;
    
    /*
     * Set up a query so we can find the proper buffer
     */
    query.minsize = phPtr->bufferSize;
    query.maxsize = BUFFER_DONT_CARE;
    query.unique = 1;
    query.state = "filling";
    query.shared = BUFFER_DONT_CARE;
    query.xAttach = BUFFER_DONT_CARE;
    sprintf(tagName, "pd%x c%d f%d", (unsigned int)pdPtr, phPtr->cycleId,
	    phPtr->bufferId);
    query.tag = tagName;
    
    /*
     * Look for the appropriate buffer
     */
    hits = FindBuffer(pdPtr->interp, &query);
    if (hits[0] > 0) {
      return GetBufferFromID(hits[1]);
    }

    /*
     * We didn't find it, so first we look for a buffer
     * that has been allocated but isn't in use right now
     */
    query.minsize = phPtr->bufferSize;
    query.maxsize = BUFFER_DONT_CARE;
    query.shared = pdPtr->shared;
    query.xAttach = pdPtr->xAttach;
    query.tag = NULL;
    
    id = FindOrCreateBuffer(pdPtr->interp, &query);

    if (id < 0) {
      /*
       * Problem finding or creating buffer. We could be out of
       * shared memory. If pdPtr->shared and pdPtr->xAttach is
       * set, unset them and try again, otherwise return NULL;
       */

      if ((query.shared) || (query.xAttach)) {
	query.shared = query.xAttach = 0;
	id = FindOrCreateBuffer(pdPtr->interp, &query);
      }

      if (id < 0) {
	/*
	 * Still no luck, return null.
	 */
	return NULL;
      }
    }

    bufPtr = GetBufferFromID(id);
    if (bufPtr->refCount != 0) {
	printf ("pktDest:FindPacketBuffer buffer state = %s\n",
		bufPtr->state);
    }
    AttachBuffer (pdPtr->interp, bufPtr, "pktDest:FindPacketBuffer");

    /*
     * Make sure the buffer is properly configured
     */
    info.size = BUFFER_DONT_MODIFY;
    info.shared = BUFFER_DONT_MODIFY;
    info.xAttach = BUFFER_DONT_MODIFY;
    info.state = "filling";
    info.callback = NULL;
    
    if (ConfigureBuffer(pdPtr->interp, bufPtr, &info) != TCL_OK) {
	DestroyBuffer(pdPtr->interp, bufPtr);
	return NULL;
    }
	
    /*
     * Add the cycle/frame tag
     */
    if (AddTag(pdPtr->interp, bufPtr, tagName) != TCL_OK) {
	DestroyBuffer(pdPtr->interp, bufPtr);
	return NULL;
    }
    
    return bufPtr;
}

/*
 *--------------------------------------------------------------
 *
 * NewCycle --
 *
 *	Called whenever the cycleId in an incoming packet is
 *	higher than anything we have already seen.  This
 *	function gets rid of everything from 2 cycles ago
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Old buffers are marked free
 *
 *--------------------------------------------------------------
 */
static void
NewCycle(pdPtr, bundle, newCycle, packets, buffers)
    PacketDest *pdPtr;		/* The current PacketDest */
    CMBundle *bundle;		/* The current CM Bundle */
    int newCycle;		/* The cycleId of the new cycle */
    int packets;		/* Number of packets in the new cycle */
    int buffers;		/* Number of buffers in the new cycle */
{
    FindQuery query;
    char tagName[64];
    int i, *hits, cycleIndex;
    
    /*
     * First, take all filling buffers from 2 cycles back
     * and mark them as free
     */
    query.minsize = BUFFER_DONT_CARE;
    query.maxsize = BUFFER_DONT_CARE;
    query.unique = 1;
    query.state = "filling";
    query.shared = BUFFER_DONT_CARE;
    query.xAttach = BUFFER_DONT_CARE;
    query.tag = tagName;
    
    cycleIndex = newCycle % 2;
    
    for (i = 0; i < bundle->buffers[cycleIndex]; i++) {
	sprintf(tagName, "pd%x c%d f%d", (unsigned int)pdPtr, newCycle-2, i);
	hits = FindBuffer(pdPtr->interp, &query);
    
	if (hits[0] > 0) {
	    Buffer *bufPtr;
	    
	    bufPtr = GetBufferFromID(hits[1]);
	    DetachBuffer(pdPtr->interp, bufPtr, "pktDest:NewCycle");
	    DeleteTag(bufPtr, tagName);
	}
    }
    
    /*
     * Reset the callback mask, mask and last cycle
     */
    if (bundle->callbackMask[cycleIndex] != NULL) {
	ckfree((char *) bundle->callbackMask[cycleIndex]);
    }

    bundle->callbackMask[cycleIndex] =
	(int *) ckalloc(((buffers/32) + 1) * sizeof(int));
    for (i=0; i < ((buffers/32)+1); i++) {
	bundle->callbackMask[cycleIndex][i] = 0;
    }
	   
    if (bundle->packetMask[cycleIndex] != NULL) {
	ckfree((char *) bundle->packetMask[cycleIndex]);
    }

    bundle->packetMask[cycleIndex] =
	(int *) ckalloc (((packets/32) + 1) * sizeof(int));
    for (i=0; i< ((packets/32)+1); i++) {
	bundle->packetMask[cycleIndex][i] = 0;
    }

    bundle->buffers[cycleIndex] = buffers;
    
    bundle->pktsRecvd = 0;
    
    bundle->prevSkewSec = bundle->skewSec;
    bundle->prevSkewUsec = bundle->skewUsec;
    bundle->skewSec = INT_MAX;
    bundle->skewUsec = INT_MAX;
    
    bundle->cycle = newCycle;
}

/*
 *--------------------------------------------------------------
 *
 * FinishedBuffer --
 *
 *	This routine is called whenever we have finished
 *	assembling a buffer.  It puts together the argument
 *	vector and makes the callback as well as noting the
 *	fact that the callback has been made so that it isn't
 *	made again on an EOC packet.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The associated callback for the completed buffer
 *	is called.
 *
 * Parameters:
 *	Buffer *bufPtr		The completed buffer
 *	CMBundle *bundle	The bundle it came from
 *	PacketDest *pdPtr	The packet dest it was received on
 *	PacketHeader *phPtr	Header from one of the packets
 *
 *--------------------------------------------------------------
 */
static void
FinishedBuffer(bufPtr, bundle, pdPtr, phPtr)
    Buffer *bufPtr;
    CMBundle *bundle;
    PacketDest *pdPtr;
    PacketHeader *phPtr;
{
    int i, offset;
    char tagName[64];
    ScatterBuffer *sb;
    SB_List *sbl;
    char *sbStr;

    /*
     * Remove the tag from the buffer
     */
    sprintf(tagName, "pd%x c%d f%d", (unsigned int)pdPtr, phPtr->cycleId,
	   phPtr->bufferId);
    DeleteTag(bufPtr, tagName);

    /*
     * Set up an SB_List structure for the callback
     */
    sb = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));
    sb->count = ntohl(*(int *)bufPtr->data);
    sb->buffers = (Buffer **) ckalloc(sb->count * sizeof(Buffer *));
    sb->lens = (int *) ckalloc(sb->count * sizeof(int));
    sb->offsets = (int *) ckalloc(sb->count * sizeof(int));

    offset = (sb->count + 1) * sizeof(int);
    for (i = 0; i < sb->count; i++) {
	sb->buffers[i] = bufPtr;
	sb->lens[i] = ntohl(((int *) bufPtr->data)[i+1]);
	sb->offsets[i] = offset;
	offset += sb->lens[i];
    }

    sbl = (SB_List *) ckalloc(sizeof(SB_List));
    sbl->count = 1;
    sbl->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *));
    sbl->sbs[0] = sb;
    
    /*
     * Make the callback
     */

    if (pdPtr->outCmd != NULL) {

        if (pdPtr->outCmd[0] == '@') {
            (pdPtr->outCCmd)(pdPtr->outObj, sbl);
        } else {
            sbStr = MakeSBListString(sbl);
            Tcl_VarEval(pdPtr->interp, pdPtr->outCmd, " ", sbStr, NULL);
            ckfree((char *) sbStr);
        }
    }

    /*
     * Detach the buffer and free up the memory allocated above.
     */
    DetachBuffer(pdPtr->interp, bufPtr, "pktDest:FinishedBuffer");
    FreeSBList (sbl);

    /*
     * Mark the fact that the callback has been made
     */
    bundle->callbackMask[phPtr->cycleId%2][phPtr->bufferId / 32] |=
	1 << (phPtr->bufferId % 32);
}


