/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmPacketSource.c,v 1.27 1996/02/09 01:14:52 chaffee Exp $
 *
 *  Introduction -- This file implements the packet source module.
 *	The packet source and packet dest module make a pair.
 *
 *	The basic idea of the packet source/dest is to provide a
 *	prioritized, timely delivery mechanism for a group of buffers.
 *	That is, a client of the module will periodically give
 *	the module a list of buffers to deliver to the packet dest.
 *	and the module will do the best job it can to get them there
 *	as fast as possible.  High priority buffers are early in the
 *	list, low priority buffers are later in the list.
 *
 *  References -- To understand this module, you should take a look at
 *	packet.h, since this defines the headers attached to packets
 *	for delivery across the network.  Also, you probably want
 *	to read that last chapter of Brian Smith's PhD dissertation,
 *	since it describes the flow control strategies used in
 *	this implementation.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <unistd.h>
#include <math.h>
#include <tcl.h>
#include <tk.h>
#include <dp.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include "cmAtom.h"
#include "cmPacket.h"
#include "cmBuffer.h"
#include "cmBufferlist.h"
#include "cmBind.h"
#include "cmUtil.h"

#ifdef NO_SENDRECVMSG
#  include "compat/sendrecvmsg.h"
#endif

/*
 * The following information is kept about each packet that is queued
 * for sending.  ComputePktInfo() builds up this information.
 */
typedef struct PktInfo {
    char sent;			/* Is this packet mark as sent? (0 or 1) */
    short burstNum;		/* Burst number when we sent it */
    unsigned short bufferNum;	/* The buffer number (for reassembly) */
    unsigned short packetNum;	/* The packet number within the buffer (for
				 * reassembly) */
    int bufSize;		/* Size of this whole buffer */
    int packetOfs;		/* Packet offset */
    unsigned short ppb;		/* # of packets in the buffer (pkt/buf) */

    /* Information about data to send: The next 3 fields work together */
    ScatterBuffer *buf;		/* Buffer that contains data to send */
    int startBuf;		/* Data segment starts in buf->argv[startBuf] */
    int ofs;			/* Data begins at argv[] + ofs */
    int pktSize;		/* Specifies size of packet payload */
} PktInfo;

/*
 * For each attached packet source, there is one of the following objects
 * maintained.
 */
typedef struct PacketSource {
    /*
     * Section 1. General information
     */
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    int attached;		/* Is it PacketSource attached yet? */
    DP_SOCKET udp;		/* File descriptor for UDP data socket */
    char *pg;			/* Priority group */ 
    struct sockaddr destAddr;	/* Data address for UDP data socket */
    char *addrInfo;		/* String with <host port> */
    int state;			/* State of the source.  See STATE_* below */

    /*
     * Section 2. Information about current buffers to send
     */
    ScatterBuffer ** buffer;	/* Array of buffers we're currently sending */
    int numBuf;			/* Number of active items in the array */
    int maxBuf;			/* Number of allocated items in the array */
    PktInfo *pktQueue;		/* Array of packets to be sent */
    int pktQueueSize;		/* Size of the array */
    int numPackets;		/* Number of packets to send this round */
    
    /*
     * Section 3. Stuff needed by SendBurst()
     */
    int cycleId;		/* Current Cycle ID */
    int burstNum;		/* Current burst number */
    int pktSize;		/* Size of packets */
    int burstSize;		/* Ideal burst size (in bytes) */
    int maxBurstSize;		/* Maximum burst size (in bytes) */
    double lastSendTime;	/* Time last SendBurst occurred */
    Tk_TimerToken sendTimer;	/* Token for SendBurst callback */
    
    /*
     * Section 4. User-supplied targets determining connection behavior
     */
    int targetDelayMsec;	/* Target Packet Delay (in msec) */
    int xpctdDrpRate;		/* Expected drop rate (0-255 = 0-100%) */
    int maxBW;			/* Maximum allowable bandwidth */
    int minBW;			/* Minimum allowable bandwidth */
    FILE *logfile;		/* Log file for debugging */
    
    /*
     * Section 5. Feedback information from the packet dest
     */
    int pktDestEpoch;		/* Epoch the following data corresponds to */
    int loss;			/* Current loss (n/256 -- from Packet Dest) */
    int delay;			/* Current delay (from Packet Dest) */
    int delayDev;		/* Durrent delary deviation (from PD) */
    int skewMsec;		/* Skew between clocks (in msec) */
    
    /*
     * Section 6. Derived information about connection
     */
    int ltBW;			/* long term bandwidth estimate */
    int alpha;			/* factor to determine weights of new
				 * measurements when computing ltBW */
    int epoch;			/* Current epoch */
    int epochDuration;		/* Duration of the epochs (in msec) */
    double nextEpochTime;	/* Time next epoch starts */
    int currEpochBytes;		/* Bytes we've sent this epoch */
    int prevEpochBytes;		/* Bytes we' sent in the previous epoch */
    int estBW;			/* Estimated bandwidth */

    /*
     * Section 7. Internally maintained statistics and cmBind token.
     */
    int bytesReceived;          /* Total length of all buffers received.
				 * Accounted for in AppendBuffers.
				 */
    int bindToken;              /* Token id registered with cm bind mechanism. */

				  
} PacketSource;

/*
 * Legal states of the Packet Source
 */
#define STATE_IDLE	0
#define STATE_SENDING	1
#define STATE_EOC	2
#define STATE_PARTITION	3

static char *state_strings[] = { "idle", "sending", "eoc", "partition" };

#define NO_TOKEN	((Tk_TimerToken) -1)
#define ALPHA_MAX	4

#ifndef _WINDOWS
#    define max(a,b)	((a)>(b)?(a):(b))
#    define min(a,b)	((a)<(b)?(a):(b))
#    ifndef abs
#	define abs(a)	((a)>0?(a):-(a))
#    endif
#endif

/* Globals used by this object module. */
static int nextPacketSourceNum = 0;    /* Variable for unique object names. */

/*
 * Prototypes
 */
int PacketSourceModuleInit(Tcl_Interp *interp);
int PacketSourceCreateCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char **argv);
PacketSource *CreateNewPacketSource(Tcl_Interp *interp, char *name);
int PacketSourceObjCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char **argv);
int PacketSourceInFunc (PacketSource *objPtr, SB_List *data);
int DestroyPacketSource(PacketSource *objPtr);
int ConfigurePacketSource(PacketSource *objPtr, int argc, char **argv);
static int AppendBuffers(PacketSource *objPtr, SB_List *sbl);
int StartCycle(PacketSource *objPtr);
int NewCycle(PacketSource *objPtr);
static int PacketSourceAttach(PacketSource *objPtr, struct sockaddr *addrPtr);
static void ComputePktInfo(PacketSource *objPtr);
static void NewEpoch(double currTime, PacketSource *pktSrc);
static void FreeBufferList(PacketSource *objPtr);
static void UdpReady(ClientData clientData, int mask);
static void ProcessFeedback(PacketSource *pktSrc);
static void ProcessFinished(PacketSource *pktSrc);
static void ProcessResend(PacketSource *pktSrc);
static void SendBurst(ClientData clientData);
static char *GetBytesReceived(Tcl_Interp *interp, ClientData clientdata);


/*
 *--------------------------------------------------------------
 *
 * PacketSourceModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 *--------------------------------------------------------------
 */

int
PacketSourceModuleInit (Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp,
		      "pktSrc", PacketSourceCreateCmd,
		      NULL, (void (*)()) NULL);
    Tcl_VarEval(interp,
		"cminfo add object_info pktSrc {unicast}",
		(char *) NULL);
   
    return TCL_OK; 
}


/*
 *--------------------------------------------------------------
 *
 * PacketSourceCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the pktSrc command.
 *     A new object of type PacketSource is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 *--------------------------------------------------------------
 */

int
PacketSourceCreateCmd (ClientData clientData, Tcl_Interp *interp,
		       int argc, char **argv)
{
    char *new_name;
    PacketSource *new;

    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ",
			 argv[0], " parent ?options?",
			 (char *) NULL);
	return TCL_ERROR;
    }

    new_name = ckalloc(strlen(argv[1]) + 16);
    nextPacketSourceNum++;
    sprintf(new_name, "%sps%d", argv[1], nextPacketSourceNum);

    new = CreateNewPacketSource(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, new_name, PacketSourceObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigurePacketSource(new, argc-2, argv+2);
    }

    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * CreateNewPacketSource
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 *--------------------------------------------------------------
 */

PacketSource *
CreateNewPacketSource (Tcl_Interp *interp, char *name)
{
    PacketSource *new;
    char *atomName, *suffix = "accept";
    DP_SOCKET udp;

    udp = Tdp_inet_connect ("", 0, 0, 1, 0, 0);
    if (udp == DP_INVALID_SOCKET) {
	Tcl_AppendResult (interp, "Couldn't open udp socket : ",
			  Tdp_Error(interp), (char *)NULL);
	return NULL;
    }

    new = (PacketSource *) ckalloc(sizeof(PacketSource));

    new->interp = interp;
    new->attached = 0;
    new->udp = udp;
    new->addrInfo = NULL;
    new->pg = ckalloc(3);
    memcpy(new->pg, "tk", 3);

    memset (&new->destAddr, 0, sizeof(new->destAddr));
    new->state = STATE_IDLE;

    new->buffer = NULL;
    new->numBuf = 0;
    new->maxBuf = 0;
    new->pktQueue = NULL;
    new->pktQueueSize = 0;
    new->numPackets = 0;

    new->cycleId = 0;
    new->burstNum = 0;
    new->pktSize = 8000;
    new->burstSize  = 4000;
    new->maxBurstSize  = 12000;
    new->lastSendTime = Tcm_ReadSysClock();
    new->sendTimer = NO_TOKEN;

    new->targetDelayMsec = 100;
    new->xpctdDrpRate = 5;
    new->maxBW = 1000000;
    new->minBW = 10000;
    new->logfile = NULL;

    new->pktDestEpoch = -1;
    new->loss = 0;
    new->delay = 0;
    new->delayDev = 0;
    new->skewMsec = 0;

    new->ltBW = 360000;
    new->alpha = 1;
    new->epoch = 0;
    new->epochDuration = 200;
    new->nextEpochTime = Tcm_ReadSysClock();
    new->currEpochBytes = 0;
    new->prevEpochBytes = 0;
    new->estBW = 360000;

    new->bytesReceived = 0;

    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    atomName = ckalloc(strlen(name) + strlen(suffix) + 3);
    sprintf(atomName, "%s.%s", name, suffix);
    Tutil_RegisterCommand(interp, atomName, (ClientData) PacketSourceInFunc,
			  (ClientData) new);
    
    /*
     * Register CM Bind events.
     */
    {
	char **event_list, **subst_list;
	ClientData *subst_cds;
	CM_BindProc **subst_procs;
	int i;

	event_list = (char **) ckalloc(1*sizeof(char *));
	event_list[0] = (char *) ckalloc(strlen("<SBL_RECEIVED>")+1);
	strcpy(event_list[0], "<SBL_RECEIVED>");
	
	subst_list = (char **) ckalloc(1*sizeof(char *));
	subst_procs = (CM_BindProc **) ckalloc(1*sizeof(CM_BindProc *));
	subst_cds = (ClientData *) ckalloc(1*sizeof(ClientData));
	
	subst_list[0] = (char *) ckalloc(2);
	(subst_list[0])[0] = 'b';
	(subst_list[0])[1] = '\0';
	subst_procs[0] = GetBytesReceived;
	subst_cds[0] = (ClientData) new;
	
	new->bindToken = CreateCMBinding(interp,
					 new->name,
					 1, event_list,
					 1, subst_list, subst_procs, subst_cds);
	
	for (i=0; i<1; i++) {
	    ckfree(event_list[i]);
	}
	ckfree((char *) event_list);
	
	for (i=0; i<1; i++) {
	    ckfree(subst_list[i]);
	}
	ckfree((char *) subst_list);
	ckfree((char *) subst_procs);
	ckfree((char *) subst_cds);
    }

    return new;
}


/*
 *--------------------------------------------------------------
 *
 * PacketSourceObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     PacketSource object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 *--------------------------------------------------------------
 */

int
PacketSourceObjCmd (ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
    char c;
    int len;
    PacketSource *objPtr = (PacketSource *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " <command> ?arguments?",
			 (char *) NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    switch (c) {
      case 'a':
	/* --------------------- ADD BUFFER --------------------- */
	if (strncmp("addBuffers", argv[1], len) == 0) {
	    SB_List sbl;
	    int res;
	    
	    if (argc != 3) {
		Tcl_AppendResult (interp,
				  "wrong # of args, should be: ",
				  argv[0], " addBuffers <scatter buffer list>",
				  (char *) NULL);
		return TCL_ERROR;
	    }
	    
	    if (ParseSBList(interp, argv[2], &sbl) != TCL_OK) {
		return TCL_ERROR;
	    }
	    
	    res = AppendBuffers(objPtr, &sbl);
	    ckfree((char *) sbl.sbs);
	    return res;
	}
	break;

      case 'c':
	/* ---------------------- CONFIGURE ---------------------- */
	if (strncmp("configure", argv[1], len) == 0) {
	    return ConfigurePacketSource(objPtr, argc-2, argv+2);
	}
	break;
	
      case 'd':
	/* ----------------------- DESTROY ----------------------- */
	if (strncmp("destroy", argv[1], len) == 0) {
	    return DestroyPacketSource(objPtr);
	}
	break;

      case 'n':
	/* ---------------------- NEW CYCLE ---------------------- */
	if (strncmp("newCycle", argv[1], len) == 0) {
	    return NewCycle(objPtr);
	}
	break;

      case 's':
	/* --------------------- START CYCLE --------------------- */
	if (strncmp("startCycle", argv[1], len) == 0) {
	    return StartCycle(objPtr);
	}
	break;
     }
	
usage:
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ", argv[1],
		     (char *) NULL);
    Tcl_AppendResult(interp,
		"\n<command> can be one of:\n",
		"  configure ?-slot ?value??\n",
		"  destroy\n",
		"  addBuffers <ScatterBufferList>\n",
		"  newCycle\n",
		"  startCycle\n",
		(char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * PacketSourceInFunc
 *
 *	The routine made available to other objects via
 *	the Atom registered as the PacketSource
 *	InFunction.
 *
 *	Takes a list of buffers and creates a new cycle
 *	consisting of those buffers.
 *
 * Parameters: 
 *	PacketSource *objPtr	The packet source receiving data
 *	SB_List  data	The buffers containing the data
 * 
 * Return: 
 *	Standard Tcl result
 * 
 * Action: 
 *	Calls the routines NewCycle, AppendBuffers, and
 *	StartCycle.
 *
 *--------------------------------------------------------------
 */

int 
PacketSourceInFunc (PacketSource *objPtr, SB_List *data)
{
    if ((data == NULL) || (data->count < 1)) return TCL_OK;

    NewCycle(objPtr);
    AppendBuffers(objPtr, data);
    StartCycle(objPtr);
    
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * DestroyPacketSource
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 *--------------------------------------------------------------
 */

int 
DestroyPacketSource (PacketSource *objPtr)
{
    /* First delete Tcl command. */

    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->numBuf != 0) {
	FreeBufferList(objPtr);
	ckfree((char *) objPtr->buffer);
    }

    if (objPtr->pktQueue != NULL) {
	ckfree ((char *) objPtr->pktQueue);
    }

    Tcm_DeleteFileHandler(objPtr->udp);
    closesocket(objPtr->udp);
    if (objPtr->sendTimer != NO_TOKEN) {
	Tcm_DeleteTimerHandler(objPtr->sendTimer);
	objPtr->sendTimer = NO_TOKEN;
    }

    if (objPtr->addrInfo != NULL) {
	ckfree (objPtr->addrInfo);
    }
    if (objPtr->name != NULL) {
	ckfree((char *) objPtr->name);
    }

    /*
     * Unregister with bind mechanism.
     */
    
    DestroyCMBinding(objPtr->interp, objPtr->bindToken);

    ckfree((char *) objPtr->pg);
    ckfree((char *) objPtr);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * ConfigurePacketSource
 *
 * Parameters: 
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 * 
 * Return: 
 *     Standard TCL return value
 * 
 * Action: 
 *     Does the usual configuration thing
 *
 *--------------------------------------------------------------
 */

int
ConfigurePacketSource (PacketSource *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;

    interp = objPtr->interp;

    if (argc == 0) {
	char tmp[50];
	
	Tcl_AppendResult(interp,
			 "{-state {} {",
			 state_strings[objPtr->state], "}} ", NULL);
	sprintf(tmp, "%d", objPtr->cycleId);
	Tcl_AppendResult(interp,
			 "{-cycleId {0} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->pktSize);
	Tcl_AppendResult(interp,
			 "{-pktSize {3000} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->burstSize);
	Tcl_AppendResult(interp,
			 "{-burstSize {3000} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->maxBurstSize);
	Tcl_AppendResult(interp,
			 "{-maxBurstSize {35000} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->targetDelayMsec);
	Tcl_AppendResult(interp,
			 "{-targetDelayMsec {100} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->xpctdDrpRate);
	Tcl_AppendResult(interp,
			 "{-xpctdDrpRate {51} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->maxBW);
	Tcl_AppendResult(interp,
			 "{-maxBW {1000000} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->minBW);
	Tcl_AppendResult(interp,
			 "{-minBW {1} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->skewMsec);
	Tcl_AppendResult(interp,
			 "{-skewMsec {0} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->ltBW);
	Tcl_AppendResult(interp,
			 "{-ltBW {0} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->epochDuration);
	Tcl_AppendResult(interp,
			 "{-epochDuration {100} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->estBW);
	Tcl_AppendResult(interp,
			 "{-estBW {360000} {", tmp, "}} ", NULL);

	if (objPtr->addrInfo != NULL) {
	    Tcl_AppendResult(interp, "{-dest ", "{} {", objPtr->addrInfo,
			     "}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-dest ", "{} {}} ", (char *) NULL);
	}

	Tcl_AppendResult(interp, "{-priorityGroup {tk} {", objPtr->pg, "}} ",
			 (char *) NULL);
	return TCL_OK;
    }
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 's') && (strncmp("-state", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-state {} {",
			     state_strings[objPtr->state], "}",
			     NULL);
	    return TCL_OK;
	} else if ((c == 'c') && (strncmp("-cycleId", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->cycleId);
	    Tcl_AppendResult(interp, "-cycleId ", "{0} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'p') && (strncmp("-pktSize", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->pktSize);
	    Tcl_AppendResult(interp,
			     "-pktSize ", "{3000} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'b') && (strncmp("-burstSize", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->burstSize);
	    Tcl_AppendResult(interp,
			     "-burstSize ", "{3000} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'm') && (strncmp("-maxBurstSize",argv[0],len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->maxBurstSize);
	    Tcl_AppendResult(interp,
			     "-maxBurstSize ", "{35000} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 't') &&
		   (strncmp("-targetDelayMsec", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->targetDelayMsec);
	    Tcl_AppendResult(interp, "-targetDelayMsec ", "{100} ", str,
			     (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'x') &&
		   (strncmp("-xpctdDrpRate", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->xpctdDrpRate);
	    Tcl_AppendResult(interp, "-xpctdDrpRate ", "{51} ", str,
			     (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'm') && (strncmp("-maxBW", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->maxBW);
	    Tcl_AppendResult(interp, "-maxBW ", "{1000000} ", str,
			     (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'm') && (strncmp("-minBW", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->minBW);
	    Tcl_AppendResult(interp, "-minBW ", "{1} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 's') && (strncmp("-skewMsec", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->skewMsec);
	    Tcl_AppendResult(interp, "-skewMsec ", "{0} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-ltBW", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->ltBW);
	    Tcl_AppendResult(interp, "-ltBW ", "{0} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'e') &&
		   (strncmp("-epochDuration", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->epochDuration);
	    Tcl_AppendResult(interp, "-epochDuration ", "{100} ", str,
			     (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'e') && (strncmp("-estBW", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->estBW);
	    Tcl_AppendResult(interp,
			     "-estBW ", "{360000} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-dest", argv[0], len) == 0)) {
	    if (objPtr->addrInfo != NULL) {
		Tcl_AppendResult(interp, "-dest ", "{} {", objPtr->addrInfo,
				 "}", (char *) NULL);
	    } else {
		Tcl_AppendResult(interp, "-dest ", "{} {}", (char *) NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'p') && (strncmp("-priorityGroup", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-priorityGroup {tk} {",
			     objPtr->pg, "}", NULL);
	    return TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "bad configuration option: ",
			     argv[0], (char *) NULL);
	    return TCL_ERROR;
	}

    }
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;

	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	if ((c == 's') && (strncmp("-state", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing parameter for -state",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (strncmp(argv[cntr], "idle", 4) == 0) {
		objPtr->state = STATE_IDLE;
	    } else if (strncmp(argv[cntr], "sending", 7) == 0) {
		objPtr->state = STATE_SENDING;
	    } else if (strncmp(argv[cntr], "eoc", 3) == 0) {
		objPtr->state = STATE_EOC;
	    } else if (strncmp(argv[cntr], "partition", 9) == 0) {
		objPtr->state = STATE_PARTITION;
	    } else {
		Tcl_AppendResult(interp,
				 argv[cntr], " is  not a valid state",
				 NULL);
		return TCL_ERROR;
	    }
	} else if ((c == 'c') && (strncmp("-cycleId", argv[cntr], len) == 0)) {
	    int newcycleId;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -cycleId",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newcycleId) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->cycleId = newcycleId;
	} else if ((c == 'p') && (strncmp("-pktSize", argv[cntr], len) == 0)) {
	    int newpktSize;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -pktSize",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newpktSize) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->pktSize = newpktSize;
	} else if ((c == 'b') &&
		   (strncmp("-burstSize", argv[cntr], len) == 0)) {
	    int newburstSize;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -burstSize",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newburstSize) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->burstSize = newburstSize;
	} else if ((c == 'm') &&
		   (strncmp("-maxBurstSize", argv[cntr], len) == 0)) {
	    int newmaxBurstSize;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -maxBustSize",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newmaxBurstSize) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->maxBurstSize = newmaxBurstSize;
	} else if ((c == 't') &&
		   (strncmp("-targetDelayMsec", argv[cntr], len) == 0)) {
	    int newtargetDelayMsec;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -targetDelayMsec",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp,
			   argv[cntr], &newtargetDelayMsec) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->targetDelayMsec = newtargetDelayMsec;
	} else if ((c == 'x') &&
		   (strncmp("-xpctdDrpRate", argv[cntr], len) == 0)) {
	    int newxpctdDrpRate;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -xpctdDrpRate",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newxpctdDrpRate) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->xpctdDrpRate = newxpctdDrpRate;
	} else if ((c == 'm') && (strncmp("-maxBW", argv[cntr], len) == 0)) {
	    int newmaxBW;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -maxBW",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newmaxBW) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->maxBW = newmaxBW;
	} else if ((c == 'm') && (strncmp("-minBW", argv[cntr], len) == 0)) {
	    int newminBW;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -minBW",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newminBW) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->minBW = newminBW;
	} else if ((c == 'l') && (strncmp("-logfile", argv[cntr], len) == 0)) {
	    FILE *newLogfile;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -logfile",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetOpenFile(interp, argv[cntr],
				0, 0, &newLogfile) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->logfile = newLogfile;
	} else if ((c == 's') && (strncmp("-skewMsec", argv[cntr], len) == 0)) {
	    int newskewMsec;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -skewMsec",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newskewMsec) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->skewMsec = newskewMsec;
	} else if ((c == 'l') && (strncmp("-ltBW", argv[cntr], len) == 0)) {
	    int newltBW;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -ltBW",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newltBW) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->ltBW = newltBW;
	} else if ((c == 'e') &&
		   (strncmp("-epochDuration", argv[cntr], len) == 0)) {
	    int newepochDuration;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -epochDuration",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newepochDuration) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->epochDuration = newepochDuration;
	} else if ((c == 'd') && (strncmp("-dest", argv[cntr], len) == 0)) {
	    struct sockaddr *addrPtr;
	    int res;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing parameter for -dest",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    /*
	     * Get new address
	     */
	    res = Tcl_VarEval (interp, "eval dp_address create ", argv[cntr],
				NULL);
	    if (res != TCL_OK) {
		Tcl_ResetResult (interp);
		Tcl_AppendResult (interp, "Error: invalid destination \"",
			argv[cntr], "\"", NULL);
		return TCL_ERROR;
	    }
	    addrPtr = Tdp_FindAddr (interp->result);
	    PacketSourceAttach (objPtr, addrPtr);
	    Tcl_VarEval (interp, "dp_address delete ", interp->result, NULL);
	    if (objPtr->addrInfo != NULL) {
		ckfree (objPtr->addrInfo);
		objPtr->addrInfo = NULL;
	    }
	    objPtr->addrInfo = ckalloc(strlen(argv[cntr])+1);
	    strcpy (objPtr->addrInfo, argv[cntr]);

	} else if ((c == 'e') && (strncmp("-estBW", argv[cntr], len) == 0)) {
	    int newestBW;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing parameter for -estBW",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newestBW) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->estBW = newestBW;
	} else if ((c == 'p') && (strncmp("-priorityGroup", argv[cntr], len) == 0)){
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -priorityGroup",
				 NULL);
		return TCL_ERROR;
	    }
	    if (Tcm_FindPg(argv[cntr]) == NULL) {
		Tcl_AppendResult(interp, "unknown priority group \"", argv[cntr],
				 "\"", NULL);
		return TCL_ERROR;
	    }
	    ckfree((char *) objPtr->pg);
	    objPtr->pg = (char *) ckalloc(strlen(argv[cntr]) + 1);
	    strcpy(objPtr->pg, argv[cntr]);
	    Tcm_DeleteFileHandler(objPtr->udp);
	    Tcm_CreateFileHandler(objPtr->pg, objPtr->udp, TK_READABLE,
				  UdpReady, (ClientData) objPtr);
	} else {
	    Tcl_AppendResult(interp,
			     "Invalid configuration parameter ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * PacketSourceAttach --
 *
 *	
 *      This function provide a C language binding to the "attach"
 *	pktSrc function.  It attaches a new PacketSource object to
 *	the socket associated with the PacketSource.
 *
 * Results:
 *	1 for success, 0 on failure.
 *
 * Side effects:
 *	An entry is made in the pktSrcs table and a file handler
 *	is created for the udp socket to handle resend requests.
 *
 * Parameters:
 *	Tcl_Interp *interp		Interp associated with CMC
 *	struct sockaddr *addrPtr	Destination address to use
 *
 *--------------------------------------------------------------
 */
static int
PacketSourceAttach (objPtr, addrPtr)
    PacketSource *objPtr;
    struct sockaddr *addrPtr;
{
    /*
     * Section 1. General information
     */
    memcpy (&objPtr->destAddr, (char *)addrPtr, sizeof(struct sockaddr));

    if (!objPtr->attached) {
	Tcm_CreateFileHandler(objPtr->pg, objPtr->udp, TK_READABLE, UdpReady,
			      (ClientData) objPtr);
    }

    objPtr->attached = 1;

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * AppendBuffers --
 *
 *	Given a scatter buffer list, append all its component
 *	scatter buffers to the list for the current cycle
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The PacketSource will attempt to transmit the data passed in
 *	to the destination until the next call to NewCycle.
 *
 * Parameters:
 *	PacketSource *objPtr	The source to add buffers to
 *	SB_List *sbl;		The list of scatter buffer
 *	int copy;		Should we copy the vectors?  If not,
 *				they become property of the PacketSource 
 *				and it will free them when it no longer
 *				needs them.
 *
 *--------------------------------------------------------------
 */
static int
AppendBuffers (PacketSource *objPtr, SB_List *sbl)
{
    int t, buf;
    ScatterBuffer *buffer;

    /*
     * Initialize room for our cache of buffers, if necessary
     */
    if (objPtr->maxBuf == 0) {
	objPtr->maxBuf = 32;
	objPtr->buffer = (ScatterBuffer **)ckalloc(objPtr->maxBuf*
						   sizeof(ScatterBuffer *));
    }
    
    /*
     * Make a copy of the scatter buffer
     */
    for (t=0; t<sbl->count; t++) {
	if (objPtr->numBuf == objPtr->maxBuf) {
	    objPtr->maxBuf += 32;
	    objPtr->buffer = (ScatterBuffer **) ckrealloc((char *) objPtr->buffer,
				  objPtr->maxBuf*sizeof(ScatterBuffer *));
	}

	buffer = CopyScatterBuffer(sbl->sbs[t]);
	objPtr->buffer[objPtr->numBuf++] = buffer;
	for (buf=0; buf<buffer->count; buf++) {
	    AttachBuffer(objPtr->interp, buffer->buffers[buf],
			 "pktSrc:AppendBuffers");
	    objPtr->bytesReceived += buffer->lens[buf];
	}
    }
    
    /*
     * Invoke cm bindings for SBL_RECEIVED event.
     */

    ExecuteCMBinding(objPtr->interp, objPtr->bindToken, "<SBL_RECEIVED>");

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * StartCycle --
 *
 *	Starts deliverying data from the packet source passed
 *	in.  If the source is already delivering, this function
 *	has no effect.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Data begins to flow out on the connection.
 *
 * Parameters:
 *	PacketSource *objPtr;	Pointer to PacketSource
 *
 *--------------------------------------------------------------
 */
int
StartCycle(objPtr)
    PacketSource *objPtr;
{
    ComputePktInfo(objPtr);
#ifdef DEBUG
    if (objPtr->logfile) {
		int i;
		for (i=0; i<objPtr->numPackets; i++) {
			fprintf (objPtr->logfile, "%4d", i);
		}
		fprintf (objPtr->logfile, "\n");
    }
#endif
	
    if (objPtr->sendTimer == NO_TOKEN) {
		SendBurst ((ClientData) objPtr);
    }
    
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * NewCycle --
 *
 *	This function provide a C language binding to the "newCycle"
 *	pktSrc function.  It makes a new cycle using the supplied
 *	array of scatter/gather vectors to represent buffers to
 *	transmit.  If the array is a NULL pointer, or its length
 *	is zero, then the current cycle is simply flushed and the
 *	PktSrc is placed in an idle state.  The array is property of
 *	the PktSrc until the next call to Tdp_PktSrcNewCycle().
 *
 *	A user will usually want to call Tdp_PktSrcBufferList() to
 *	get the current buffer list owned by the PktSrc, call
 *	Tdp_PktSrcNewCycle() to replace that buffer list, and then
 *	free or otherwise recycle the old buffers.  Otherwise, they
 *	could be lost.
 *
 * Results:
 *	An non-zero integer indentifier for the new Cycle, or 0 if
 *	an error is detected (e.g., bad parameters).
 *
 * Side effects:
 *	From now on, the buffers passed in will be transmitted to the
 *	packet dest.
 *
 * Parameters:
 *	PacketSource *objPtr	Pointer to PacketSource
 *
 *--------------------------------------------------------------
 */
int
NewCycle (objPtr)
    PacketSource *objPtr;
{
    /*
     * Stop the send bursts
     */
    if (objPtr->sendTimer != NO_TOKEN) {
	Tcm_DeleteTimerHandler(objPtr->sendTimer);
	objPtr->sendTimer = NO_TOKEN;
    }

    /*
     * Free data from previous cycle.
     */
    if (objPtr->numBuf != 0) {
	FreeBufferList(objPtr);
	objPtr->numBuf = 0;
    }
    objPtr->cycleId++;
    objPtr->burstNum = 0;
#ifdef DEBUG
    if (objPtr->logfile) {
	fprintf (objPtr->logfile, "New Cycle %d\n", objPtr->cycleId);
    }
#endif

    return objPtr->cycleId;
}


/*
 *--------------------------------------------------------------
 *
 * ComputePktInfo --
 *
 *	Recompute the pktQueue part of the packet source, after
 *	a new cycle is register or the packet size is changed.
 *
 *	Build up the packet information needed for sending.  This is
 *	a bit ugly, since we want to recast the array of scatter/gather
 *	buffers in terms of fixed length packets without actually
 *	copying any data.  At the end, we'll have an array of packets,
 *	each with sent,burst number, etc, and the appropriate pointers
 *	into the linked list of scatter gather vectors.  The latter
 *	information is held in the buf, ofs, and pktSize fields of the
 *	pktQueue for this packet.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	PacketSource *pktSrc;		The packet source to work on
 *
 *--------------------------------------------------------------
 */
static void
ComputePktInfo (pktSrc)
    PacketSource *pktSrc;
{
    int numPackets, pktSize;
    int *bufSize;
    int i, j, p, pkt;
    int bytesLeft, ppb;
    ScatterBuffer *buf;
    ScatterBuffer **buffer;
    int numBuf;
    int startBuf;
    int argc;
    int *argl;
    int offsetTblSize;

    if (pktSrc->numBuf == 0) {
	return;
    }

    /*
     * Calculate how many packets we'll need for this cycle
     */
    pktSize = pktSrc->pktSize;
    numBuf = pktSrc->numBuf;
    buffer = pktSrc->buffer;
    numPackets = 0;

    bufSize = (int *)ckalloc(numBuf*sizeof(int));
    for (i=0; i<numBuf; i++) {
	/*
	 * Reserve (1+buf->count)*sizeof(int) extra bytes at the head of
	 * the first packet for the length table, indicating the length
	 * of each segment of the scatter buffer
	 */
	buf = buffer[i];
	argc = buf->count;
	bufSize[i] = (1+argc)*sizeof(int);
	for (j=0; j<argc; j++) {
	    bufSize[i] += buf->lens[j];
	}
	numPackets += (bufSize[i] + pktSize - 1)/pktSize;
    }

    if (numPackets == 0) {
	ckfree ((char *) bufSize);
	return;
    }

    /*
     * Expand pktQueue if necessary
     */
    if (pktSrc->pktQueue == NULL) {
	pktSrc->pktQueue = (PktInfo *)ckalloc(numPackets*sizeof(PktInfo));
    } else if (numPackets > pktSrc->pktQueueSize) {
	pktSrc->pktQueue = (PktInfo *)ckrealloc((char *) pktSrc->pktQueue,
				    numPackets*sizeof(PktInfo));
    }
    pktSrc->pktQueueSize = numPackets;
    pktSrc->numPackets = numPackets;

    /*
     * Compute the information for each packet.
     */
    p = 0;
    for (i=0; i<numBuf; i++) {
	ppb = (bufSize[i] + pktSize - 1)/pktSize;
	buf = buffer[i];
	argc = buf->count;
	argl = buf->lens;
	offsetTblSize = (1+argc)*sizeof(int);
	for (pkt=0; pkt<ppb; pkt++) {
	    pktSrc->pktQueue[p].sent = 0;
	    pktSrc->pktQueue[p].burstNum = 0;
	    pktSrc->pktQueue[p].bufferNum = i;
	    pktSrc->pktQueue[p].packetNum = pkt;
	    pktSrc->pktQueue[p].ppb = ppb;
	    pktSrc->pktQueue[p].bufSize = bufSize[i];
	    pktSrc->pktQueue[p].packetOfs = pktSize*pkt;

	    /*
	     * Find correct starting buffer and offset for this packet.
	     * The values we're computing here are:
	     *	o pktSrc->pktQueue[p].buf
	     *	o pktSrc->pktQueue[p].startBuf
	     *	o pktSrc->pktQueue[p].ofs
	     *	o pktSrc->pktQueue[p].pktSize
	     *
	     * NOTE: this code assumes (1+argc)*sizeof(int) < pktSize
	     */
	    pktSrc->pktQueue[p].buf = buf;
	    if (pkt == 0) {
		pktSrc->pktQueue[p].startBuf = 0;
		pktSrc->pktQueue[p].ofs = 0;
	    } else {
		bytesLeft = pktSrc->pktQueue[p].packetOfs - offsetTblSize;
		for (startBuf=0; startBuf < argc; startBuf++) {
		    if (bytesLeft < argl[startBuf]) {
			break;
		    }
		    bytesLeft -= argl[startBuf];
		}
		pktSrc->pktQueue[p].startBuf = startBuf;
		pktSrc->pktQueue[p].ofs = bytesLeft;
	    }
	    if (pkt == ppb-1) {
		pktSrc->pktQueue[p].pktSize = bufSize[i] - pktSize*pkt;
	    } else {
		pktSrc->pktQueue[p].pktSize = pktSize;
	    }
	    p++;
	}
    }
    ckfree((char *) bufSize);
}


/*
 *--------------------------------------------------------------
 *
 * NewEpoch --
 *
 *	This function is called whenever a new epoch is begun.
 *	It basically increments the epoch # and check that we've
 *	heard from the pktDest recently.  If we havn't gotten a
 *	feedback unit from the dest in a while, the network might
 *	be partitioned, so we should back off exponentially.  We
 *	back off from the ltBW by an exponential factor proportional
 *	to the length of time we haven't heard anything.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The estBw will drop if we haven't received a feedback unit
 *	from the dest recently.
 *
 * Parameters:
 *	double currTime;	Current value on system clock.
 *	PacketSource *pktSrc;	Pointer to packetSrc.
 *
 *--------------------------------------------------------------
 */
static void
NewEpoch (currTime, pktSrc)
    double currTime;
    PacketSource *pktSrc;
{
#define EPOCH_THRESHOLD	4
    int maxBandwidth, minBandwidth;

    pktSrc->epoch++;
    pktSrc->prevEpochBytes = pktSrc->currEpochBytes;
    pktSrc->currEpochBytes = 0;
    pktSrc->nextEpochTime = currTime + pktSrc->epochDuration/1000.0;

    /*
     * The pktDestEpoch can get messed up if we pick up an old feedback
     * unit from the udp socket.  The following line will correct the problem.
     */
    if (pktSrc->pktDestEpoch > pktSrc->epoch) {
	pktSrc->pktDestEpoch = pktSrc->epoch-1;
    }

    /*
     * If we haven't gotten a feedback unit in EPOCH_THRESHOLD
     * epochs, then we may be killing the net.  Back off
     * exponentially
     */
    if ((pktSrc->epoch - pktSrc->pktDestEpoch) > EPOCH_THRESHOLD) {
	int diff = (pktSrc->epoch - pktSrc->pktDestEpoch - EPOCH_THRESHOLD)/2;
	pktSrc->estBW = pktSrc->ltBW >> diff;
	pktSrc->state = STATE_PARTITION;
	pktSrc->alpha = 1;

	/*
	 * Make sure new bandwidth ends up in a reasonable range
	 */
	maxBandwidth = pktSrc->maxBW;
	minBandwidth = pktSrc->minBW;
	if (pktSrc->estBW > maxBandwidth) {
	    pktSrc->estBW = maxBandwidth;
	}
	if (pktSrc->estBW < minBandwidth) {
	    pktSrc->estBW = minBandwidth;
	}
	if (pktSrc->logfile) {
	    fprintf (pktSrc->logfile, "%d\t", pktSrc->epoch);
	    fprintf (pktSrc->logfile, "newBW = 0\t");
	    fprintf (pktSrc->logfile, "ltBW = ??\t");
	    fprintf (pktSrc->logfile, "delay = ??\t");
	    fprintf (pktSrc->logfile, "loss = 100\t");
	    fprintf (pktSrc->logfile, "estBW = %d\n", pktSrc->estBW);
	}
    } else {
	pktSrc->state = STATE_SENDING;
    }
}


/*
 *--------------------------------------------------------------
 *
 * FreeBufferList --
 *
 *	Frees the buffers in the buffer list associated
 *	with a given PacketSource object.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	You should never reference the buffer chain again
 *
 * Parameters:
 *	PacketSource *objPtr	The packet source object
 *
 *--------------------------------------------------------------
 */
static void
FreeBufferList (PacketSource *objPtr)
{
    int i, j, numBuf;
    ScatterBuffer *buf, **buffers;
    
    numBuf = objPtr->numBuf;
    buffers = objPtr->buffer;

    for (i=0; i<numBuf; i++) {
	buf = buffers[i];
	
	for (j=0; j<buf->count; j++) {
	    DetachBuffer(objPtr->interp, buf->buffers[j],
			 "pktSrc:FreeBufferList");
	}
	
	FreeScatterBuffer (buf);
	buffers[i] = NULL;
    }
}


/*
 *--------------------------------------------------------------
 *
 * UdpReady --
 *
 *	This function is called from as a file handler when the udp
 *	socket associated with a source becomes readable.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Dispatches to ProcessResend of ProcessFeddback
 *
 * Parameters:
 *	ClientData clientData	Pointer to the packet source.
 *	int mask		An integer mask indicating the state of
 *				the socket (ignored)
 *
 *--------------------------------------------------------------
 */
static void
UdpReady (clientData, mask)
    ClientData clientData;
    int mask;
{
    PacketSource *pktSrc = (PacketSource *)clientData;
    DP_SOCKET udp;
    int magic;
    int len;
    struct sockaddr_in addr;
    int addrLen;

    udp = pktSrc->udp;

    /*
     * Peek at the header and decode it.  If it's not what we expect,
     * then toss the message if necessary.
     */
    addrLen = sizeof(addr);
    len = recvfrom(udp, (char *)&magic, sizeof(magic), MSG_PEEK,
		   (struct sockaddr *)&addr, &addrLen);
    if (len != sizeof(magic)) {
	char str[32];
	int error;
  
	error = DP_GetLastError();
	if (error != DP_EMSGSIZE) {
	readError:
	    sprintf (str, "file%d", udp);
	    Tcl_AppendResult(pktSrc->interp,
			     "Unexpected message recvd on udp socket for packet source ",
			     str, NULL);
	    recvfrom(udp, (char *)&magic, sizeof(magic), 0,
		     (struct sockaddr *)&addr, &addrLen);
	    Tk_BackgroundError(pktSrc->interp);
	    return;
	}
    }

    magic = ntohl(magic);
    if (magic == FEEDBACK_UNIT_MAGIC) {
	ProcessFeedback(pktSrc);
    } else if (magic == RESEND_REQUEST_MAGIC) {
	ProcessResend(pktSrc);
    } else if (magic == FINISHED_MARKER_MAGIC) {
	ProcessFinished(pktSrc);
    } else {
	goto readError;
    }
}


/*
 *--------------------------------------------------------------
 *
 * ProcessFeedback --
 *
 *	Process a feedback unit sent from the source.  This
 *	is called from UdpReady() above.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Will probably change the estimated bandwidth for the channel.
 *
 * Parameters:
 *	PacketSource *pktSrc	Pointer to packetSource
 *
 *--------------------------------------------------------------
 */
static void
ProcessFeedback (pktSrc)
    PacketSource *pktSrc;
{
    FeedbackUnit fu;
    DP_SOCKET udp;
    int newBW;
    int err;
    int L;
    int X;
    int Dt, maxBandwidth, minBandwidth;
    struct sockaddr_in addr;
    int addrLen;

    udp = pktSrc->udp;

    /*
     * Peek at the header and decode it.  If it's not what we expect,
     * then toss the message if necessary.
     */
    addrLen = sizeof(addr);
    if (recvfrom(udp, (char *)&fu, sizeof(fu), 0,
	     (struct sockaddr *)&addr, &addrLen) != sizeof(fu)) {
	char str[32];
	int error;

	error = DP_GetLastError();
	if (error != DP_EMSGSIZE) {
	    sprintf (str, "file%d", udp);
	    Tcl_AppendResult(pktSrc->interp,
		 "Short FeedbackUnit recvd on udp socket for packet source ",
		 str, NULL);
	    Tk_BackgroundError(pktSrc->interp);
	    return;
	}
    }

    /*
     * If this feedback unit is old, ignore
     */
    fu.epoch = ntohl(fu.epoch);
    if (fu.epoch <= pktSrc->pktDestEpoch) {
	return;
    }

    /*
     * Update values in packetSource.  Don't update bandwidth if
     * we got no loss and we were under-driving the circuit, since
     * that doesn't give us any "real" information.
     */
    pktSrc->loss = ntohl(fu.lossRate);
    newBW = ntohl(fu.bandwidth);
    if ((pktSrc->alpha < ALPHA_MAX) ||
	(pktSrc->loss != 0) ||
	(newBW > pktSrc->ltBW)) {
	err = newBW - pktSrc->ltBW;
	pktSrc->ltBW += err>>pktSrc->alpha;
	if (pktSrc->alpha < ALPHA_MAX) {
	    pktSrc->alpha++;
	}
    }
    pktSrc->delay = ntohl(fu.delay);
    pktSrc->delayDev = ntohl(fu.delayDev);
    pktSrc->skewMsec = ntohl(fu.skewMsec);
    pktSrc->pktDestEpoch = fu.epoch;

    /*
     * If loss or delay are high, lower estBW.  Otherwise, raise it.
     * The calculation uses fixed-point arithmetic.  The guard statement
     * "(pktSrc->ltBW > (1<<22))" checks if we might overflow using
     * this method.  If so, we use an alternate order of calculation
     * to avoid integer overflow.
     */
    L = pktSrc->loss;
    X = pktSrc->xpctdDrpRate;
    if (L > X) {
        pktSrc->estBW = (256.0-L)*pktSrc->ltBW/(256.0-X);
    } else if (pktSrc->delay > pktSrc->targetDelayMsec) {
        Dt = pktSrc->targetDelayMsec;
        pktSrc->estBW = Dt*(1.0+X/256.0)*pktSrc->ltBW/pktSrc->delay;
    } else {
        pktSrc->estBW = (1.0+X/256.0)*pktSrc->ltBW;
    }

    /*
     * Make sure new bandwidth ends up in a reasonable range
     */
    maxBandwidth = pktSrc->maxBW;
    minBandwidth = pktSrc->minBW;
    if (pktSrc->estBW > maxBandwidth) {
	pktSrc->estBW = maxBandwidth;
    }
    if (pktSrc->estBW < minBandwidth) {
	pktSrc->estBW = minBandwidth;
    }
    if (pktSrc->logfile) {
	fprintf (pktSrc->logfile, "%d\t", pktSrc->epoch);
        fprintf (pktSrc->logfile, "newBW = %ld\t", ntohl(fu.bandwidth));
        fprintf (pktSrc->logfile, "ltBW = %d\t", pktSrc->ltBW);
        fprintf (pktSrc->logfile, "delay = %d\t", pktSrc->delay);
        fprintf (pktSrc->logfile, "loss = %d\t", L);
        fprintf (pktSrc->logfile, "estBW = %d\n", pktSrc->estBW);
    }
}


/*
 *--------------------------------------------------------------
 *
 * ProcessFinished --
 *
 *	Process a finished marker sent from the source.  This
 *	is called from UdpReady() above.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Will cause the source to go into an idle state.
 *
 * Parameters:
 *	PacketSource *pktSrc	Pointer to packetSource
 *
 *--------------------------------------------------------------
 */
static void
ProcessFinished (pktSrc)
    PacketSource *pktSrc;
{
    FinishedMarker fm;
    DP_SOCKET udp;
    struct sockaddr_in addr;
    int addrLen;

    udp = pktSrc->udp;

    /*
     * Peek at the header and decode it.  If it's not what we expect,
     * then toss the message if necessary.
     */
    addrLen = sizeof(addr);
    if (recvfrom(udp, (char *)&fm, sizeof(fm), 0,
	     (struct sockaddr *)&addr, &addrLen) != sizeof(fm)) {
	char str[32];
	int error;

	error = DP_GetLastError();
	if (error != DP_EMSGSIZE) {
	    sprintf (str, "file%d", udp);
	    Tcl_AppendResult(pktSrc->interp,
		"Short FinishedMarker recvd on udp socket for packet source ",
		str, NULL);
	    Tk_BackgroundError(pktSrc->interp);
	    Tk_BackgroundError(pktSrc->interp);
	    return;
	}
    }

    /*
     * If this finished marker is old, ignore
     */
    fm.cycleId = ntohl(fm.cycleId);
    if (fm.cycleId < pktSrc->cycleId) {
	return;
    }

    /*
     * Delete send timer -- we'll then be in an idle state.
     */
    if (pktSrc->sendTimer != NO_TOKEN) {
	Tcm_DeleteTimerHandler(pktSrc->sendTimer);
	pktSrc->sendTimer = NO_TOKEN;
    }
    pktSrc->state = STATE_IDLE;
}


/*
 *--------------------------------------------------------------
 *
 * ProcessResend --
 *
 *	Process a resend request sent from the source.  This
 *	is called from as a file handler event when the udp
 *	socket associated with a source becomes readable.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Some of the packets in the pktQueue may get marked as
 *	unsent causing them to be resent on the next burst.
 *
 * Parameters:
 *	PacketSource *pktSrc	Pointer to packetSource
 *
 *--------------------------------------------------------------
 */
static void
ProcessResend (pktSrc)
    PacketSource *pktSrc;
{
    PktInfo *pktQueue;
    ResendRequest req;
    int udp;
    int *buf;
    int buflen;
    int hdr[4];
    int messageRead;
    int fastBuf[256];
    int i, p;
    int lastP;
    struct sockaddr_in addr;
    int addrLen;

    udp = pktSrc->udp;
    messageRead = 0;
    buf = fastBuf;
    pktQueue = pktSrc->pktQueue;

    /*
     * Peek at the header and decode it.  If it's not what we expect,
     * then toss the message if necessary.
     */
    addrLen = sizeof(addr);
    if (recvfrom(udp, (char *)hdr, sizeof(hdr), MSG_PEEK,
	     (struct sockaddr *)&addr, &addrLen) != sizeof(hdr)) {
	char str[32];
	int error;

	error = DP_GetLastError();
	if (error != DP_EMSGSIZE) {
	readError:
	    sprintf (str, "file%d", udp);
	    Tcl_AppendResult(pktSrc->interp,
		"Short resend message recvd on udp socket for packet source ",
		 str, NULL);
	    if (!messageRead) {
		recvfrom(udp, (char *)hdr, sizeof(hdr), 0,
		     (struct sockaddr *)&addr, &addrLen);
	    }
	    Tk_BackgroundError(pktSrc->interp);
	    return;
	}
    }
    req.cycleId    = ntohl(hdr[1]);
    req.numReq     = ntohl(hdr[2]);
    req.burstNum   = ntohl(hdr[3]);

    /*
     * Request is for a previous cycle -- ignore
     */
    if (req.cycleId != pktSrc->cycleId) {
	recvfrom(udp, (char *)hdr, sizeof(hdr), 0,
	     (struct sockaddr *)&addr, &addrLen);
	return;
    }

    /*
     * Allocate space for the whole message, read it and decode it.
     */
    buflen = sizeof(hdr) + req.numReq*sizeof(int);
    if (buflen > sizeof(fastBuf)) {
	buf = (int *) ckalloc(buflen);
    }
    if (recvfrom(udp, (char *)buf, buflen, 0, 
	     (struct sockaddr *)&addr, &addrLen) != buflen) {
	int error = DP_GetLastError();
	if (error != DP_EMSGSIZE) {
	    messageRead = 1;
	    if (buf != fastBuf) {
		ckfree ((char *) buf);
	    }
	    goto readError;
	}
    }
    req.pktNum = buf+sizeof(hdr)/sizeof(int);

    /*
     * Mark packets in pktQueue and return
     */
    lastP = -1;
    for (i=0; i<req.numReq; i++) {
	p = ntohl(req.pktNum[i]);
#ifdef DEBUG
	if (pktSrc->logfile) {
		int j;

	    for (j=lastP; j<p-1; j++) {
		fprintf (pktSrc->logfile, "%4s", "");
	    }
	    fprintf (pktSrc->logfile, "%3dR", req.burstNum);
	    lastP = p;
	}
#endif
	if (pktQueue[p].burstNum <= req.burstNum && pktQueue[p].sent) {
	    pktQueue[p].sent = 0;
	}
    }
#ifdef DEBUG
    if (pktSrc->logfile) { fprintf (pktSrc->logfile, "\n"); }
#endif
    if (buf != fastBuf) {
	ckfree ((char *) buf);
    }
    if ((pktSrc->state == STATE_EOC) && (pktSrc->sendTimer != NO_TOKEN)) {
	Tcm_DeleteTimerHandler(pktSrc->sendTimer);
	pktSrc->sendTimer = NO_TOKEN;
    }
    if (pktSrc->sendTimer == NO_TOKEN) {
	SendBurst ((ClientData)pktSrc);
    }
    pktSrc->state = STATE_SENDING;
}


/*
 *--------------------------------------------------------------
 *
 * SendBurst --
 *
 *	This function is called in response to a timer event.
 *	It sends a burst of packets out on the udp socket, then
 *	schedules another timer event to call this function again
 *	for the next burst.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Some of the packets in the pktQueue may get marked as
 *	sent.
 *
 * Parameters:
 *	ClientData clientData	Pointer to the packet source.
 *
 *--------------------------------------------------------------
 */
static void
SendBurst (clientData)
    ClientData clientData;
{
    PacketSource *pktSrc = (PacketSource *)clientData;
    PktInfo *pktQueue = pktSrc->pktQueue;
    PacketHeader pktHdr;
    ScatterBuffer *buf;
    int i, j;
    int bytesToSend;
    int bytesSent;
    int bytesLeft;
    int ofs, v;
    int burstSize;
    DP_SOCKET udp;
    int estBandwidth;
    int delay;
    struct msghdr msg;
    struct iovec iov[256];
    struct timeval tv;
    double currTime;
    int n;
    int argn;
    int argc;
    int *argl;

    argl = NULL;
    pktSrc->sendTimer = NO_TOKEN;
    pktSrc->burstNum++;
    udp = pktSrc->udp;
    bytesSent = 0;
    burstSize = pktSrc->burstSize;
    estBandwidth = pktSrc->estBW;

    if (!pktSrc->attached) {
	return;
    }

    /*
     * Record timing info and calculate how many bytes to send.  We need
     * to do this because we may not get called at exactly the right time
     * by the timer handlers due to the granularity of the clock or other
     * event handling.
     */
    (void) gettimeofday(&tv, NULL);
    currTime = tv.tv_sec + tv.tv_usec / 1000000.0;
    bytesToSend = (currTime - pktSrc->lastSendTime)*estBandwidth;

    /*
     * Start new epoch if appropriate.
     */
    if (currTime > pktSrc->nextEpochTime) {
	NewEpoch (currTime, pktSrc);
    }

    /*
     * Put a cap on the size on any single burst so we don't overflow the
     * receiver's buffers.
     */
    if (bytesToSend > pktSrc->maxBurstSize) {
	bytesToSend = pktSrc->maxBurstSize;
    }

    /*
     * Fill in constants of packet header
     */
    pktHdr.epoch = htonl(pktSrc->epoch);
    pktHdr.epochBytes = htonl(pktSrc->prevEpochBytes);
    pktHdr.epochDuration = htonl(pktSrc->epochDuration);
    pktHdr.sendTimeSec = htonl(tv.tv_sec);
    pktHdr.sendTimeUSec = htonl(tv.tv_usec);
    pktHdr.cycleId = htonl(pktSrc->cycleId);
    pktHdr.burstId = htonl(pktSrc->burstNum);
    pktHdr.ppc = htonl(pktSrc->pktQueueSize);
    pktHdr.bpc = htonl(pktSrc->numBuf);
    pktHdr.EOC = htonl(0);

    /*
     * Init base fields of msg for sendmsg
     */
    msg.msg_name = (caddr_t)&pktSrc->destAddr;
    msg.msg_namelen = sizeof(pktSrc->destAddr);
    msg.msg_iov = iov;
    msg.msg_iovlen = 0;
#if !defined(_SCO_DS)
    msg.msg_accrights = NULL;
    msg.msg_accrightslen = 0;
#endif
    iov[0].iov_base = (caddr_t)&pktHdr;
    iov[0].iov_len = sizeof(PacketHeader);
    msg.msg_iovlen = 1;

    /*
     * Go through all the packets in the send buffer, sending
     * each unsent packet, up to burstSize bytes.
     * On a sparc 10, using gcc -O and 1000 byte packets, this loop
     * takes about 500 microseconds/iteration
     */
    for (i=0; i<pktSrc->pktQueueSize && bytesSent < bytesToSend; i++) {

	/*
	 * Check if we should skip the packet (already sent).
	 */
	if (pktQueue[i].sent) {
#ifdef DEBUG
	    if (pktSrc->logfile) { fprintf (pktSrc->logfile, "%4s", ""); }
#endif
	    continue;
	}
#ifdef DEBUG
	if (pktSrc->logfile) {
	    fprintf (pktSrc->logfile, "%4d", pktSrc->burstNum);
	}
#endif

	/*
	 * Fill in packet header
	 */
	pktHdr.bufferSize = htonl(pktQueue[i].bufSize);
	pktHdr.bufferId = htonl(pktQueue[i].bufferNum);
	pktHdr.packetNum = htonl(pktQueue[i].packetNum);
	pktHdr.packetSize = htonl(pktQueue[i].pktSize);
	pktHdr.packetId = htonl(i);
	pktHdr.ppb = htonl(pktQueue[i].ppb);
	pktHdr.packetOfs = htonl(pktQueue[i].packetOfs);

	/*
	 * Record bookkeeping information
	 */
	pktQueue[i].burstNum = pktSrc->burstNum;
	pktQueue[i].sent = 1;

	/*
	 * Set up the iov structure
	 */
	bytesLeft = pktQueue[i].pktSize;
	buf = pktQueue[i].buf;
	ofs = pktQueue[i].ofs;
	v = 1;

	/*
	 * If this is the first packet, pad the front with
	 *	argc and the argl vector.
	 */
	if (pktQueue[i].packetNum == 0) {
	    int bufferNum = pktQueue[i].bufferNum;
	    int len, orig_argc;

	    orig_argc = pktSrc->buffer[bufferNum]->count;
	    argc = htonl(orig_argc);
	    iov[v].iov_base = (char *)&argc;
	    iov[v].iov_len = sizeof(int);
	    bytesLeft -= sizeof(int);
	    len = orig_argc*sizeof(int);
	    argl = (int *) ckalloc(len);
	    memcpy (argl, pktSrc->buffer[bufferNum]->lens, len);
	    for (j=0; j<orig_argc; j++) {
		argl[j] = htonl(argl[j]);
	    }
	    iov[v+1].iov_base = (char *) argl;
	    iov[v+1].iov_len = len;
	    bytesLeft -= iov[v+1].iov_len;
	    v += 2;
	}

	/*
	 * Build the rest from the buffers
	 */
	argn = pktQueue[i].startBuf;
	ofs = pktQueue[i].ofs;
	while (bytesLeft) {
	    iov[v].iov_base = buf->buffers[argn]->data + ofs +
			      buf->offsets[argn];
	    iov[v].iov_len = min(buf->lens[argn] - ofs, bytesLeft);
	    bytesLeft -= iov[v].iov_len;
	    v++;
	    argn++;
	    ofs = 0;
	}
	msg.msg_iovlen = v;
	n = sendmsg(udp, &msg, 0);
	if (argl != NULL) {
	    ckfree((char *) argl);
	    argl = NULL;
	}
	bytesSent += n;
	if (pktSrc->state != STATE_PARTITION) {
	    pktSrc->state = STATE_SENDING;
	}
    }
    pktSrc->currEpochBytes += bytesSent;

    /*
     * Send EOC packet if necessary
     */
    delay = 1000*(burstSize + bytesSent - bytesToSend)/estBandwidth;
    if (i == pktSrc->pktQueueSize) {
	/*
	 * Fill in packet header
	 */
	pktHdr.bufferSize = htonl(0);
	pktHdr.bufferId = htonl(-1);
	pktHdr.packetSize = htonl(sizeof(PacketHeader));
	pktHdr.packetId = htonl(pktSrc->pktQueueSize);
	pktHdr.ppb = htonl(1);
	pktHdr.EOC = htonl(1);
#ifdef DEBUG
	if (pktSrc->logfile) {
	    fprintf (pktSrc->logfile, " EOC");
	}
#endif

	/*
	 * Just send the header.
	 */
	msg.msg_iovlen = 1;
	n = sendmsg(udp, &msg, 0);
	pktSrc->currEpochBytes += n;
	bytesSent += n;
	delay = 100;
	if (pktSrc->state != STATE_PARTITION) {
	    pktSrc->state = STATE_EOC;
	}
    }
#ifdef DEBUG
    if (pktSrc->logfile) {
	fprintf (pktSrc->logfile, "\n");
    }
#endif

    pktSrc->lastSendTime = currTime;

    /*
     * Schedule next send burst
     */
    pktSrc->sendTimer = Tcm_CreateTimerHandler(pktSrc->pg, delay,
					       SendBurst, clientData);
}

static char *
GetBytesReceived(Tcl_Interp *interp, ClientData clientdata)
{
  PacketSource *objPtr;
  char *resStr;

  objPtr = (PacketSource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bytesReceived);
  objPtr->bytesReceived = 0;

  return resStr;
}
