/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>

#if PARALLAX

#include "XPlxExt.h"
#include "cmMJPEGCapture.h"
#include "cmUtil.h"
#include "cmBuffer.h"
#include "cmBufferlist.h"

/* 
 * The following data structure is maintained per parallax device.
 * At this time, each new MJPEG Capture object that uses the parallax device
 * gets an independent device structure as if there were multiple devices
 * on the machine even though there is only one board. In the future, we will
 * want to change this so that multiple capture objects share one device
 * structure.
 */

typedef struct ParallaxCapture {
    int ShmEvent;
    Display *dpy;
    XPlxCImage *rec_image;
    int waiting;
    char *qt;
    int qtSize;
    int tag;
    int qfactor;
    Window win;
    GC gc;
    plx_IO *plx_io;
    int channel;
    int standard;
    int format;
    int type;
    int display_x;
    int display_y;
    int display_width;
    int display_height;
    int dx, dy, dw, dh;
    int vx, vy, vw, vh;
    int canXAttach;
    int initFlag;
} ParallaxCapture;

/* 
 * Error codes defined here.
 */

#define NO_WINDOW_ERROR 1
#define NO_HARDWARE_ERROR 2
#define NO_SYNCH_ERROR 3
#define BAD_XID_ERROR 4
#define BAD_ATTRIBUTE_ERROR 5
#define BAD_DW_ERROR 6
#define BAD_DH_ERROR 7
#define BAD_DX_ERROR 8
#define BAD_DY_ERROR 9
#define VX_OUT_OF_SIGNAL_ERROR 10
#define VY_OUT_OF_SIGNAL_ERROR 11
#define FRAME_PTR_NULL_ERROR 12
#define PARALLAX_NOSHM_CAPTURE_ERROR 13

/*
 * Interface routines
 */

MJPEGCaptureDevice *GetParallaxCaptureDevice(MJPEGCapture *captureObj);
static int Init(MJPEGCapture *captureObj);
static int Shutdown(MJPEGCapture *captureObj);
static int Capture(MJPEGFrame *framePtr);
static int Refresh(MJPEGCapture *captureObj);
static int FreeFrame(MJPEGFrame *framePtr);
static int Configure(MJPEGCapture *captureObj, char *option, char *value);
static char *ErrorString(MJPEGCapture *captureObj, int errCode);


/*-------------------------------------------------------------
 * 
 * GetParallaxCaptureDevice
 *     Constructs device data structure if necessary and returns pointer
 *     to it.
 *
 * Parameters:
 *     captureObj - pointer to MJPEG Capture object.
 *
 * Results:
 *     Pointer to parallax device.
 *
 * Action:
 *     Determines if parallax device exists or not. Creates it if
 *     necessary. Returns a pointer to it.
 *
 *-------------------------------------------------------------*/

MJPEGCaptureDevice *
GetParallaxCaptureDevice(MJPEGCapture *captureObj)
{
    MJPEGCaptureDevice *new_dev;
    ParallaxCapture *devicePtr;

    new_dev = (MJPEGCaptureDevice *) ckalloc(sizeof(MJPEGCaptureDevice));

    devicePtr = (ParallaxCapture *) ckalloc(sizeof(ParallaxCapture));
	
    devicePtr->initFlag = 0;
    devicePtr->dpy = Tcm_Display();
    devicePtr->canXAttach = cmXShmQueryExtension(devicePtr->dpy);

    if (devicePtr->canXAttach) {
	devicePtr->ShmEvent = XShmGetEventBase(devicePtr->dpy) + ShmCompletion;
    }

    devicePtr->rec_image = NULL;
    devicePtr->waiting = 0;
    devicePtr->qt = NULL;
    devicePtr->qtSize = 0;
    devicePtr->qfactor = -1;
    devicePtr->win = 0;
    devicePtr->gc = None;
    devicePtr->plx_io = NULL;
    devicePtr->channel = PLX_INPUT_0;
    devicePtr->standard = PLX_NTSC;
    devicePtr->format = PLX_COMP;
    devicePtr->type = PLX_RGB24;
    devicePtr->display_x = -1;
    devicePtr->display_y = -1;
    devicePtr->display_height = 0;
    devicePtr->display_width = 0;

    new_dev->deviceData = (ClientData) devicePtr;
    new_dev->asyncEncodeFlag = 0;
    new_dev->captureProc = Capture;
    new_dev->encodeProc = NULL;
    new_dev->refreshProc = Refresh;
    new_dev->initProc = Init;
    new_dev->reinitProc = Init;
    new_dev->shutdownProc = Shutdown;
    new_dev->freeFrameProc = FreeFrame;
    new_dev->configProc = Configure;
    new_dev->errorProc = ErrorString;

    return new_dev;
}


/*---------------------------------------------------------------
 *
 * Shutdown
 *     Shuts down device for capture object.
 *     
 * Parameters:
 *     captureObj - MJPEG Capture object.
 * 
 * Results:
 *     Integer error code.
 *
 * Actions:
 *     Decrements reference count to device. Removes any
 *     xids in xid list associated with the capture object.
 *     If ref count falls to zero, destroys device completely.
 *
 *--------------------------------------------------------------*/
static int
Shutdown(MJPEGCapture *captureObj)
{
    ParallaxCapture *devicePtr;

    devicePtr = (ParallaxCapture *) captureObj->devicePtr->deviceData;

    if (devicePtr->initFlag) {
	if (devicePtr->qt != NULL) {
	    ckfree((char *) devicePtr->qt);
	}
	
	if (devicePtr->gc != NULL) {
	    XFreeGC(devicePtr->dpy, devicePtr->gc);
	}

    }

    ckfree((char *) devicePtr);
    return 0;
}


/*------------------------------------------------------------------
 *
 * CalculateDisplayParameters
 *     Calculates parameters fro XPlxVideoSqueezeLive and 
 *     XPlxVideoLive calls.
 *
 * Parameters:
 *     captureObj - pointer to MJPEG Capture object.
 *     devicePtr- pointer to Parallax Capture device.
 *     signal_w - signal width
 *     signal_h - signal height
 *     vx_ptr, vy_ptr, vh_ptr, vw_ptr, etc. - pointer to where parameters
 *                                            should be returned.
 *
 * Results:
 *     Integer error code. 0 on success.
 *
 * Action:
 *     Straightforward calculation based on signal dimensions and user
 *     attributes.
 *
 *--------------------------------------------------------------------*/

static int
CalculateDisplayParameters(MJPEGCapture *captureObj,
			   ParallaxCapture *devicePtr,
			   int signal_w, int signal_h)
{
    int vx, vy, vh, vw;
    int dx, dy, dh, dw;
    int diff;

    /* 
     * Calculate vx and vw from capture object attributes and signal 
     * dimensions.
     */

    if (captureObj->xpos < 0) {
	if (captureObj->width == 0) {
	    vx = 0;
	    vw = signal_w;
	} else {
	    if (captureObj->width >= signal_w) {
		vx = 0;
		vw = signal_w;
	    } else {
		diff = signal_w - captureObj->width;
		vx = diff/2;
		vw = captureObj->width;
	    }
	}
    } else {
	vx = captureObj->xpos;

	if (vx > signal_w) {
	    return VX_OUT_OF_SIGNAL_ERROR;
	}

	if ((vx + captureObj->width > signal_w) || 
	    (captureObj->width == 0)) {
	    vw = signal_w - vx;
	} else {
	    vw = captureObj->width;
	}
    }

    /* 
     * Calculate vy and vh from capture object attributes and signal 
     * dimensions.
     */

    if (captureObj->ypos < 0) {
	if (captureObj->height == 0) {
	    vy = 0;
	    vh = signal_h;
	} else {
	    if (captureObj->height >= signal_h) {
		vy = 0;
		vh = signal_h;
	    } else {
		diff = signal_h - captureObj->height;
		vy = diff/2;
		vh = captureObj->height;
	    }
	}
    } else {
	vy = captureObj->ypos;

	if (vy > signal_h) {
	    return VY_OUT_OF_SIGNAL_ERROR;
	}

	if ((vy + captureObj->height > signal_h) || 
	    (captureObj->height == 0)) {
	    vh = signal_h - vy;
	} else {
	    vh = captureObj->height;
	}
    }

    /*
     * Calculate dx and dw. 
     * Dw is only calculated if squeeze flag is set.
     * Calculation of dx also depends a little on squeeze flag.
     */
    
    if (captureObj->squeeze) {
	/* With squeeze on, ignore display_x < 0 option. */

	if (devicePtr->display_x < 0) {
	    dx = 0;
	} else {
	    dx = devicePtr->display_x;
	}

	if ((devicePtr->display_width == 0) ||
	    (devicePtr->display_width > vw)) {
	    dw = vw;
	} else {
	    dw = devicePtr->display_width;
	}
    } else {
	/*
	 * Squeeze not on, do centering if needed. Dw not calculated.
	 */

	if (devicePtr->display_x < 0) {
	    if (devicePtr->display_width > vw) {
		diff = devicePtr->display_width - vw;
		dx = diff/2;
	    } else {
		dx = 0;
	    }
	} else {
	    dx = devicePtr->display_x;
	}
    }

    /*
     * Calculate dy and dh. 
     * Dh is only calculated if squeeze flag is set.
     * Calculation of dy also depends a little on squeeze flag.
     */
    
    if (captureObj->squeeze) {
	/* With squeeze on, ignore display_y < 0 option. */

	if (devicePtr->display_y < 0) {
	    dy = 0;
	} else {
	    dy = devicePtr->display_y;
	}

	if ((devicePtr->display_height == 0) ||
	    (devicePtr->display_height > vh)) {
	    dh = vh;
	} else {
	    dh = devicePtr->display_height;
	}
    } else {
	/*
	 * Squeeze not on, do centering if needed. Dh not calculated.
	 */

	if (devicePtr->display_y < 0) {
	    if (devicePtr->display_height > vh) {
		diff = devicePtr->display_height - vh;
		dy = diff/2;
	    } else {
		dy = 0;
	    }
	} else {
	    dy = devicePtr->display_y;
	}
    }

    devicePtr->dx = dx;
    devicePtr->dy = dy;
    devicePtr->dw = dw;
    devicePtr->dh = dh;
    devicePtr->vx = vx;
    devicePtr->vy = vy;
    devicePtr->vw = vw;
    devicePtr->vh = vh;

    return 0;
}


/*------------------------------------------------------------
 * 
 * Init
 *      Called to initialize the parallax device.
 *
 * Parameters:
 *      captureObj - pointer to MJPEG capture object.
 * 
 * Results:
 *      Integer error code. 0 if success.
 *
 * Actions:
 *      If device not previously initialized, initializes all
 *      necessary slots. Checks that it knows about at least
 *      one window.
 *-----------------------------------------------------------*/

static int
Init(MJPEGCapture *captureObj)
{
    ParallaxCapture *devicePtr;
    plx_signal *signal_ptr;
    int res_code;

    devicePtr = (ParallaxCapture *) captureObj->devicePtr->deviceData;

    if (!cmParallaxQueryExtension(devicePtr->dpy)) {
	return NO_HARDWARE_ERROR;
    }

    if (devicePtr->win == 0) {
	return NO_WINDOW_ERROR;
    }

    if (devicePtr->gc != None) {
	XFreeGC(devicePtr->dpy, devicePtr->gc);
	devicePtr->gc = None;
    };
    devicePtr->gc = XCreateGC(devicePtr->dpy, devicePtr->win, 0, 0);

    devicePtr->qfactor = -1;
    if (devicePtr->qt != NULL) {
	free(devicePtr->qt);
	devicePtr->qt = NULL;
    }
    devicePtr->qtSize = 0;

    if (devicePtr->plx_io == NULL) {
	devicePtr->plx_io = (plx_IO *) XPlxQueryConfig(devicePtr->dpy, 
					    devicePtr->win,
					    devicePtr->gc);
    }

    
    XPlxVideoInputSelect(devicePtr->dpy, devicePtr->win,
			 devicePtr->gc, devicePtr->channel,
			 devicePtr->standard, devicePtr->format,
			 devicePtr->type);

    
    signal_ptr = (plx_signal *) XPlxQueryVideo(devicePtr->dpy, devicePtr->win, 
					       devicePtr->gc);

    if (signal_ptr == NULL) {
	return NO_SYNCH_ERROR;
    }

    res_code = CalculateDisplayParameters(captureObj, devicePtr,
					  signal_ptr->w, signal_ptr->h);


    if (res_code != 0) {
	return res_code;
    }

    if (captureObj->squeeze) {
	XPlxVideoSqueezeLive(devicePtr->dpy, devicePtr->win,
			     devicePtr->gc, 
			     devicePtr->vx, devicePtr->vy, 
			     devicePtr->vw, devicePtr->vh,
			     devicePtr->dx, devicePtr->dy, 
			     devicePtr->dw, devicePtr->dh);
    } else {
	XPlxVideoLive(devicePtr->dpy, devicePtr->win, devicePtr->gc,
		      devicePtr->vx, devicePtr->vy, 
		      devicePtr->dx, devicePtr->dy, 
		      devicePtr->vw, devicePtr->vh);
    }

    devicePtr->initFlag = 1;
    return 0;
}




/*-------------------------------------------------------------------
 * 
 * Configure
 *    Configure procedure for parallax device.
 *
 * Parameters:
 *    captureObj - MJPEG capture object.
 *    option - attribute being configured.
 *    value - value for attribute.
 *  
 * Result:
 *    Integer error code.
 *
 * Actions:
 *    If option and value are both null, append to interp's result
 *    option-value tuples for device.
 *    If only value is null, append to interp's result, option-value
 *    tuple for option specified.
 *    If both option and value are specified, set option to value.
 *    If option or value not valid for desired function, return error code,
 *    otherwise return 0.
 *------------------------------------------------------------------*/

static int
Configure(MJPEGCapture *captureObj, 
	  char *option,
	  char *value)
{
    ParallaxCapture *devicePtr;
    int len;
    char c;
    
    devicePtr = (ParallaxCapture *) captureObj->devicePtr->deviceData;

    if (option == NULL) {
	{
	    char xid_str[20];

	    sprintf(xid_str, "0x%.8x", devicePtr->win);

	    Tcl_AppendResult(captureObj->interp, " {-xid 0 ",
			     xid_str, " } ", (char *) NULL);
	}
	{
	    char dh_str[20];
	    
	    sprintf(dh_str, "%d", devicePtr->display_height);

	    Tcl_AppendResult(captureObj->interp, " {-display_height 0 ",
			     dh_str, " } ", (char *) NULL);
	}
	{
	    char dw_str[20];
	    
	    sprintf(dw_str, "%d", devicePtr->display_width);

	    Tcl_AppendResult(captureObj->interp, " {-display_width 0 ",
			     dw_str, " } ", (char *) NULL);
	}
	{
	    char dx_str[20];
	    
	    sprintf(dx_str, "%d", devicePtr->display_x);

	    Tcl_AppendResult(captureObj->interp, " {-display_xpos 0 ",
			     dx_str, " } ", (char *) NULL);
	}
	{
	    char dy_str[20];
	    
	    sprintf(dy_str, "%d", devicePtr->display_y);

	    Tcl_AppendResult(captureObj->interp, " {-display_ypos 0 ",
			     dy_str, " } ", (char *) NULL);
	}

	return 0;
    }
    
    len = strlen(option);
    c = option[1];

    if ((c == 'x') && (strncmp(option, "-xid", len) == 0)) {
	if (value == NULL) {
	    char xid_str[100];
	    
	    sprintf(xid_str, "0x%.8x", devicePtr->win);
	    Tcl_ResetResult(captureObj->interp);
	    Tcl_AppendResult(captureObj->interp, "-xid 0 ",
			     xid_str, (char *) NULL);
	    return 0;
	} else {
	    /* Setting xid attribute here. */
	    int new_xid;
	    
	    if (Tcl_GetInt(captureObj->interp, value, &new_xid) != TCL_OK) {
		return BAD_XID_ERROR;
	    }

	    if (new_xid != devicePtr->win) {
		if (devicePtr->initFlag) {
		    XPlxVideoStop(devicePtr->dpy, devicePtr->win, devicePtr->gc);
		    devicePtr->initFlag = 0;
		}
	    }

	    devicePtr->win = new_xid;
	    return 0;
	}
    } else if ((c == 'd') && (strncmp(option, "-display_height", len) == 0)) {
	if (value == NULL) {
	    char dh_str[20];

	    sprintf(dh_str, "%d", devicePtr->display_height);
	    Tcl_ResetResult(captureObj->interp);
	    Tcl_AppendResult(captureObj->interp, " -display_height 0 ",
			     dh_str, (char *) NULL);

	} else {
	    int new_dh;

	    if (Tcl_GetInt(captureObj->interp, value, &new_dh) != TCL_OK) {
		return BAD_DH_ERROR;
	    } 
	    devicePtr->display_height = new_dh;
	}
	return 0;
    } else if ((c == 'd') && (strncmp(option, "-display_width", len) == 0)) {
	if (value == NULL) {
	    char dw_str[20];

	    sprintf(dw_str, "%d", devicePtr->display_width);
	    Tcl_ResetResult(captureObj->interp);
	    Tcl_AppendResult(captureObj->interp, " -display_width 0 ",
			     dw_str, (char *) NULL);

	} else {
	    int new_dw;

	    if (Tcl_GetInt(captureObj->interp, value, &new_dw) != TCL_OK) {
		return BAD_DW_ERROR;
	    } 

	    devicePtr->display_width = new_dw;
	}
	return 0;
    } else if ((c == 'd') && (strncmp(option, "-display_xpos", len) == 0)) {
	if (value == NULL) {
	    char dx_str[20];

	    sprintf(dx_str, "%d", devicePtr->display_x);
	    Tcl_ResetResult(captureObj->interp);
	    Tcl_AppendResult(captureObj->interp, " -display_xpos 0 ",
			     dx_str, (char *) NULL);

	} else {
	    int new_dx;

	    if (Tcl_GetInt(captureObj->interp, value, &new_dx) != TCL_OK) {
		return BAD_DX_ERROR;
	    } 
	    devicePtr->display_x = new_dx;
	}
	return 0;
    } else if ((c == 'd') && (strncmp(option, "-display_ypos", len) == 0)) {
	if (value == NULL) {
	    char dy_str[20];

	    sprintf(dy_str, "%d", devicePtr->display_y);
	    Tcl_ResetResult(captureObj->interp);
	    Tcl_AppendResult(captureObj->interp, " -display_ypos 0 ",
			     dy_str, (char *) NULL);

	} else {
	    int new_dy;

	    if (Tcl_GetInt(captureObj->interp, value, &new_dy) != TCL_OK) {
		return BAD_DY_ERROR;
	    } 
	    devicePtr->display_y = new_dy;
	}
	return 0;
    } else {
	return BAD_ATTRIBUTE_ERROR;
    }
}


/*--------------------------------------------------------------------------
 * Refresh
 *    Used to refresh the live video display after expose events.
 *
 * Parameters:
 *    captureObj - pointer to MJPEG Capture object.
 *
 * Results:
 *    Integer error code. 0 on success.
 *
 * Actions:
 *    Goes through same process as Init procedure and makes Parallax call
 *    to reset the live video window. The important difference is that the
 *    current GC is still used (i.e., not free'd and realloc'd) and qtables
 *    not touched. 
 *-------------------------------------------------------------------------*/
 
static int
Refresh(MJPEGCapture *captureObj)
{
    ParallaxCapture *devicePtr;
    plx_signal *signal_ptr;

    devicePtr = (ParallaxCapture *) captureObj->devicePtr->deviceData;

    if (devicePtr->initFlag == 0) {
	return 0;
    }

    if (captureObj->squeeze) {
	XPlxVideoSqueezeLive(devicePtr->dpy, devicePtr->win,
			     devicePtr->gc, 
			     devicePtr->vx, devicePtr->vy, 
			     devicePtr->vw, devicePtr->vh,
			     devicePtr->dx, devicePtr->dy, 
			     devicePtr->dw, devicePtr->dh);
    } else {
	XPlxVideoLive(devicePtr->dpy, devicePtr->win, devicePtr->gc,
		      devicePtr->vx, devicePtr->vy, 
		      devicePtr->dx, devicePtr->dy, 
		      devicePtr->vw, devicePtr->vh);
    }

    return 0;
}


/*--------------------------------------------------------------------------
 * Capture
 *    Procedure called to actually capture a frame.
 *
 * Parameters:
 *    framePtr - pointer to frame data structure.
 * 
 * Results:
 *    Integer error code. 0 on succes.
 *
 * Actions:
 *    Various fields of framePtr are filled. Most importantly, scatter buffer
 *    properly formated with an MJPEG Frame is created. Q factor changes
 *    are dealt with by reinstalling q tables into device if q factor field
 *    of object does not match last q factor used. 
 *------------------------------------------------------------------------*/

static int
Capture(MJPEGFrame *framePtr) 
{
    MJPEGCapture *captureObj;
    ParallaxCapture *devicePtr;
    XPlxCImage *image;
    MJPEGBufferHeader tmp_frame_hdr;
    int idx;
    int tmpInt;
    Buffer *headerBuffer;

    if (framePtr == NULL)
      return FRAME_PTR_NULL_ERROR;
    

    captureObj = framePtr->objPtr;
    devicePtr=captureObj->devicePtr->deviceData;

    /* Check for and deal with q factor change. */

    if (devicePtr->qfactor != captureObj->qfactor) {
	if (devicePtr->qt != NULL) {
	    free(devicePtr->qt);
	    devicePtr->qt = NULL;
	}
	devicePtr->qtSize = MakeQTables(captureObj->qfactor, &(devicePtr->qt));
	devicePtr->qfactor = captureObj->qfactor;

	XPlxPutTable(devicePtr->dpy, devicePtr->win, devicePtr->gc, devicePtr->qt,
		     devicePtr->qtSize, 0);
	XFlush(devicePtr->dpy);
    }

#ifdef HAVE_SHARED_MEMORY
    if (devicePtr->canXAttach) {

	/* Use shared memory if possible. */
	
	framePtr->dataBuffer = MakeSharedBuffer(captureObj->interp,
						devicePtr->dw*devicePtr->dh, NULL);

	if (framePtr->dataBuffer == NULL) {
	    goto shared_failure;
	}

	image = (XPlxCImage *) ckalloc(sizeof(XPlxCImage));
	image->width = devicePtr->dw;
	image->height = devicePtr->dh;
	image->size = framePtr->dataBuffer->size;
	image->data = framePtr->dataBuffer->data;
	image->info = (char *) &(framePtr->dataBuffer->shminfo);
	
	XFlush(devicePtr->dpy);
	XPlxShmGetCImage(devicePtr->dpy, devicePtr->win,
			 devicePtr->gc, image, 
			 devicePtr->dx, devicePtr->dy,
			 devicePtr->dw, devicePtr->dh);
	
	/* Fill in frame info. */
	
	framePtr->data = framePtr->dataBuffer->data;
	framePtr->frameHeader.width = image->width;
	framePtr->frameHeader.height = image->height;
	framePtr->frameHeader.device = PARALLAX_DEV_CODE;
	framePtr->frameHeader.subformat = RAW_JPEG_NO_HEADERS_CODE;
	framePtr->frameHeader.qfactor = devicePtr->qfactor;
	framePtr->dataLength = image->size;
	framePtr->deviceData = NULL;
	
	/* HTONL frame header into a temporary frameHeader. */
	
	tmp_frame_hdr.hdrType = htonl(framePtr->frameHeader.hdrType);
	tmp_frame_hdr.length = htonl(framePtr->frameHeader.length);
	tmp_frame_hdr.width = htonl(framePtr->frameHeader.width);
	tmp_frame_hdr.height = htonl(framePtr->frameHeader.height);
	tmp_frame_hdr.device = htonl(framePtr->frameHeader.device);
	tmp_frame_hdr.subformat = htonl(framePtr->frameHeader.subformat);
	tmp_frame_hdr.qfactor = htonl(framePtr->frameHeader.qfactor);
	tmp_frame_hdr.logStart_s = htonl(framePtr->frameHeader.logStart_s);
	tmp_frame_hdr.logStart_ns = htonl(framePtr->frameHeader.logStart_ns);
	tmp_frame_hdr.logEnd_s = htonl(framePtr->frameHeader.logEnd_s);
	tmp_frame_hdr.logEnd_ns = htonl(framePtr->frameHeader.logEnd_ns);
	
	
	/* Create header buffer (also included chunk type and length for data) */
	
	headerBuffer = MakeBuffer(captureObj->interp, sizeof(MJPEGBufferHeader)+8, NULL);
	
	/* Write frame header into buffer. */
	
	memcpy(headerBuffer->data, &tmp_frame_hdr, sizeof(MJPEGBufferHeader));
	idx = sizeof(MJPEGBufferHeader);
	
	/* Write CM DATA SEGMENT buffer chunk type and size info. */
	
	tmpInt = htonl(CM_DATA_SEGMENT);
	memcpy(headerBuffer->data+idx, &tmpInt, 4);
	idx += 4;
	
	tmpInt = htonl(framePtr->dataLength);
	memcpy(headerBuffer->data+idx, &tmpInt, 4);
	idx += 4;
	
	/* Destroy image structure now that its been co-opted. */
	
	ckfree((char *) image);
	
	/* Create scatter buffer. */
	
	framePtr->sb = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));
	framePtr->sb->count = 2;
	
	framePtr->sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *)*2);
	framePtr->sb->offsets = (int *) ckalloc(sizeof(int)*2);
	framePtr->sb->lens = (int *) ckalloc(sizeof(int)*2);
	
	framePtr->sb->buffers[0] = headerBuffer;
	framePtr->sb->offsets[0] = 0;
	framePtr->sb->lens[0] = idx;
	
	framePtr->sb->buffers[1] = framePtr->dataBuffer;
	framePtr->sb->offsets[1] = 0;
	framePtr->sb->lens[1] = framePtr->dataLength;
	
	return 0;
    }

  shared_failure:
#endif

    image = XPlxGetCImage(devicePtr->dpy, devicePtr->win,
			  devicePtr->gc, devicePtr->dx,
			  devicePtr->dy, devicePtr->dw,
			  devicePtr->dh, devicePtr->dw, 
			  devicePtr->dh);


    if (image == NULL) {
	return PARALLAX_NOSHM_CAPTURE_ERROR;
    }

    /* Fill in frame info. */

    framePtr->frameHeader.width = image->width;
    framePtr->frameHeader.height = image->height;
    framePtr->frameHeader.device = PARALLAX_DEV_CODE;
    framePtr->frameHeader.subformat = RAW_JPEG_NO_HEADERS_CODE;
    framePtr->frameHeader.qfactor = devicePtr->qfactor;
    framePtr->dataLength = image->size;
    framePtr->deviceData = NULL;
    
    /* Create Buffer out of image memory. */

    framePtr->dataBuffer = MakeExistingBuffer(captureObj->interp, NULL,
					      image->data, image->size);
    AttachBuffer(captureObj->interp, framePtr->dataBuffer, NULL);
    framePtr->data = framePtr->dataBuffer->data;
    image->data = NULL;

    /* HTONL frame header into a temporary frameHeader. */

    tmp_frame_hdr.hdrType = htonl(framePtr->frameHeader.hdrType);
    tmp_frame_hdr.length = htonl(framePtr->frameHeader.length);
    tmp_frame_hdr.width = htonl(framePtr->frameHeader.width);
    tmp_frame_hdr.height = htonl(framePtr->frameHeader.height);
    tmp_frame_hdr.device = htonl(framePtr->frameHeader.device);
    tmp_frame_hdr.subformat = htonl(framePtr->frameHeader.subformat);
    tmp_frame_hdr.qfactor = htonl(framePtr->frameHeader.qfactor);
    tmp_frame_hdr.logStart_s = htonl(framePtr->frameHeader.logStart_s);
    tmp_frame_hdr.logStart_ns = htonl(framePtr->frameHeader.logStart_ns);
    tmp_frame_hdr.logEnd_s = htonl(framePtr->frameHeader.logEnd_s);
    tmp_frame_hdr.logEnd_ns = htonl(framePtr->frameHeader.logEnd_ns);


    /* Create header buffer (also included chunk type and length for data) */

    headerBuffer = MakeBuffer(captureObj->interp, sizeof(MJPEGBufferHeader)+8, NULL);

    /* Write frame header into buffer. */
    
    memcpy(headerBuffer->data, &tmp_frame_hdr, sizeof(MJPEGBufferHeader));
    idx = sizeof(MJPEGBufferHeader);

    /* Write CM DATA SEGMENT buffer chunk type and size info. */

    tmpInt = htonl(CM_DATA_SEGMENT);
    memcpy(headerBuffer->data+idx, &tmpInt, 4);
    idx += 4;

    tmpInt = htonl(framePtr->dataLength);
    memcpy(headerBuffer->data+idx, &tmpInt, 4);
    idx += 4;

    /* Destroy image structure now that its been co-opted. */

    XPlxDestroyCImage(image);

    /* Create scatter buffer. */

    framePtr->sb = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));
    framePtr->sb->count = 2;

    framePtr->sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *)*2);
    framePtr->sb->offsets = (int *) ckalloc(sizeof(int)*2);
    framePtr->sb->lens = (int *) ckalloc(sizeof(int)*2);

    framePtr->sb->buffers[0] = headerBuffer;
    framePtr->sb->offsets[0] = 0;
    framePtr->sb->lens[0] = idx;

    framePtr->sb->buffers[1] = framePtr->dataBuffer;
    framePtr->sb->offsets[1] = 0;
    framePtr->sb->lens[1] = framePtr->dataLength;

    return 0;
}


/*--------------------------------------------------------------------------
 * ErrorString
 *    Procedure called to convert error code into text message.
 *
 * Parameters:
 *    objPtr - pointer to MJPEG Capture object.
 *    errCode - error code to translate.
 * 
 * Results:
 *    String pointer to text message.
 *
 * Actions:
 *    Not much, just a big switch statement.
 *------------------------------------------------------------------------*/

static char *
ErrorString(MJPEGCapture *objPtr, int errCode) {

    switch (errCode) {
      case NO_WINDOW_ERROR:
	return "Xid of capture/live window not set. Required by Parallax.";
      case NO_HARDWARE_ERROR:
	return "Couldn't find Parallax board.";
      case NO_SYNCH_ERROR:
	return "Video input synch not detected.";
      case BAD_XID_ERROR:
	return "Xid of capture/live window not valid.";
      case BAD_ATTRIBUTE_ERROR:
	return "Unknown attribute.";
      case BAD_DW_ERROR:
	return "Illegal value for display width.";
      case BAD_DH_ERROR:
	return "Illegal value for display height.";
      case BAD_DX_ERROR:
	return "Illegal value for display x offset.";
      case BAD_DY_ERROR:
	return "Illegal value for display y offset.";
      case VX_OUT_OF_SIGNAL_ERROR:
	return "Video x offset not within signal bounds.";
      case VY_OUT_OF_SIGNAL_ERROR:
	return "Video y offset not within signal bounds.";
      case FRAME_PTR_NULL_ERROR:
	return "Frame pointer structure was NULL";
      case PARALLAX_NOSHM_CAPTURE_ERROR:
	return "Shared memory failure.";
      default:
	return "Parallax device error.";
    }
}


/*--------------------------------------------------------------------------
 * FreeFrame
 *    Procedure called to free any memory associated with the frame.
 *
 * Parameters:
 *    framePtr - pointer to MJPEG Frame.
 * 
 * Results:
 *    Integer error code, 0 means success.
 *
 * Actions:
 *    Nothing. 
 *------------------------------------------------------------------------*/

static int
FreeFrame(MJPEGFrame *framePtr)
{
    /* 
     * Don't need to do anything, but may need to in the future so 
     * I went a head a provided a routine that just returns.
     */

    return 0;
}
    

#else /* PARALLAX */

#endif


