/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmParallaxDevice.c,v 1.26 1996/02/06 22:42:24 kpatel Exp $
 *
 * Implements the parallax device, used with the mjpegPlay
 * object to do hardware assisted jpeg decoding
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>
#include "cmMJPEGPlay.h"

#if PARALLAX

#include "XPlxExt.h"
#include "cmUtil.h"

typedef struct ParallaxDevice {
    int ShmEvent;
    Window win;
    Display *dpy;
    GC gc;
    XPlxCImage cimage;
    int waiting;
    char *qt;
    int qtSize;
    int tag;
    int qfactor;
    Buffer *lastDataBuffer;
    char *lastPicData;
    int lastPicDataLength;
    int lastPicSubformat;
    int lastPicWidth;
    int lastPicHeight;
} ParallaxDevice;

#define NO_WINDOW_ERROR		1
#define NO_HARDWARE_ERROR	2

static int canXAttach = -1;


/*
 *------------------------------------------------------------
 *
 * cmParallaxQueryExtension
 *
 *    Determine if this machine has a Parallax board.
 *
 * Results:
 *	Always returns 0 so the program doesn't exit
 *
 * Side Effects:
 *	Causes ParallaxInit to fail
 *
 *------------------------------------------------------------
 */

int
cmParallaxQueryExtension(Display *displayPtr)
{
    Tk_ErrorHandler handler;
    static int inited = 0;
    static int cache;
    int handlerResult = 1;
    int ignore;

    if (inited) {
	return cache;
    }
    inited++;

    /*
     * Just because the Parallax extension is compiled in
     * doesn't mean we can really use it; the display
     * may be on a remote machine that doesn't have it.
     */
    handlerResult = XQueryExtension(displayPtr, "ParallaxVideo",
				    &ignore, &ignore, &ignore);
    if (!handlerResult) {
	cache = 0;
	return cache;
    }

    cache = 1;
    return cache;
}

/*
 *--------------------------------------------------------------
 *
 * LookForShmEvent --
 *
 *	This little helper funciton is used in combination with
 *	the XIfEvent call below to look for ShmCompletion events.
 *
 * Results:
 *	1 if event is a ShmCompletion event, 0 otherwise.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *
 *--------------------------------------------------------------
 */
static int
LookForShmEvent (Display *display, XEvent *e, char *arg)
{
    ParallaxDevice *pdevice;
    XAnyEvent *eventPtr = (XAnyEvent *)e;

    pdevice = (ParallaxDevice *) arg;
    return ((eventPtr->type == pdevice->ShmEvent) && 
	    (eventPtr->window == pdevice->win));
}


/*
 *--------------------------------------------------------------
 *
 * WaitForCompletion --
 *
 *	Blocks waiting for a XShmCompletion event from the X server.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	May block
 *
 *--------------------------------------------------------------
 */
static void
WaitForCompletion (ParallaxDevice *pdevice)
{
    XEvent shmEvent;

    if (!pdevice->win) {
	pdevice->waiting = 0;
	return;
    }
    if (pdevice->waiting) {
	XIfEvent(pdevice->dpy, &shmEvent, LookForShmEvent, (char *) pdevice);
	pdevice->waiting = 0;
    }
}


/*
 *--------------------------------------------------------------
 *
 * ProcessShmEvent --
 *
 *	Called when a shmCompletion event is received.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static int
ProcessShmEvent (ClientData clientData, XEvent *e)
{
    ParallaxDevice *pdevice = (ParallaxDevice *)clientData;
    XAnyEvent *shmEventPtr = (XAnyEvent *)e;

    if (!pdevice->waiting) {
	return 0;
    } else if (!pdevice->win) {
	pdevice->waiting = 0;
    } else if ((pdevice->waiting) &&
        (shmEventPtr->type == pdevice->ShmEvent) && 
	(shmEventPtr->window == pdevice->win)) {
	pdevice->waiting = 0;
	return 1;
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * ParallaxInit --
 *
 *	Initialize the parallax device
 *
 * Results:
 *	0 for successful initilization, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
ParallaxInit(MJPEGResource *rsrcPtr)
{
    ParallaxDevice *pdevice;

    pdevice = (ParallaxDevice *) rsrcPtr->devicePtr->deviceData;

    if (rsrcPtr->xid == 0) {
	return NO_WINDOW_ERROR;
    }

    pdevice->dpy = Tcm_Display();
    pdevice->gc = XCreateGC(pdevice->dpy, rsrcPtr->xid, 0, 0);
    if (!cmParallaxQueryExtension(pdevice->dpy)) {
	XFreeGC(pdevice->dpy, pdevice->gc);
	return NO_HARDWARE_ERROR;
    }
    XSetForeground(pdevice->dpy, pdevice->gc, 0);

    pdevice->ShmEvent = XShmGetEventBase(pdevice->dpy)+ShmCompletion;
    Tk_CreateGenericHandler (ProcessShmEvent, (ClientData)pdevice);

    pdevice->win = rsrcPtr->xid;

    pdevice->lastDataBuffer = NULL;
    pdevice->lastPicData = NULL;
    pdevice->lastPicSubformat = 0;
    pdevice->lastPicDataLength = 0;
    pdevice->lastPicWidth = pdevice->lastPicHeight = 0;

    pdevice->qt = NULL;
    pdevice->qtSize = 0;
    pdevice->qfactor = -1;
    pdevice->waiting = 0;

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * ParallaxReinit --
 *
 *	Reinitialize the parallax device (after paraameter change).
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
int
ParallaxReinit(MJPEGResource *rsrcPtr)
{
    ParallaxDevice *pdevice;

    pdevice = (ParallaxDevice *) rsrcPtr->devicePtr->deviceData;

    if (rsrcPtr->xid != pdevice->win) {
	pdevice->win = rsrcPtr->xid;

	WaitForCompletion(pdevice);
	if (pdevice->qt)  {
	    free(pdevice->qt);
	}
	pdevice->qt = NULL;
	
	if (pdevice->gc) {
	    XFreeGC(pdevice->dpy, pdevice->gc);
	}
	pdevice->gc = XCreateGC(pdevice->dpy, rsrcPtr->xid, 0, 0);
	XSetForeground(pdevice->dpy, pdevice->gc, 0);
	
	pdevice->qtSize = 0;
	pdevice->qfactor = -1;
	pdevice->waiting = 0;    
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * ParallaxDecode --
 *
 *	Stub for parallax decoder function.  Since display routine
 *	handles this, do nothing
 *
 * Results:
 *	0 for success
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
ParallaxDecode (MJPEGFrame *framePtr) 
{
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * ParallaxDisplay --
 *
 *	Display a video frame in the window associated with
 *	the this device
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Window is redrawn
 *
 *--------------------------------------------------------------
 */
int
ParallaxDisplay (MJPEGFrame *framePtr)
{
    ParallaxDevice *pdevice;
    int sx, sy, sh, sw, dx, dy, dh, dw;

    pdevice = (ParallaxDevice *) framePtr->objPtr->devicePtr->deviceData;

    /*
     * Do qfactor change here.
     */
    if (framePtr->frameHeader.qfactor != pdevice->qfactor) {
	pdevice->qfactor = framePtr->frameHeader.qfactor;
	if (pdevice->qt) {
	    free (pdevice->qt);
	}
	pdevice->qt = NULL;
	pdevice->qtSize = MakeQTables (pdevice->qfactor, &(pdevice->qt));
	XPlxPutTable (pdevice->dpy, pdevice->win, pdevice->gc, 
		      pdevice->qt, pdevice->qtSize, 0);
	XPlxPutTable (pdevice->dpy, pdevice->win, pdevice->gc, 
		      pdevice->qt, pdevice->qtSize, 1);
	XPlxVideoTag (pdevice->dpy, pdevice->win, pdevice->gc, PLX_VIDEO);
	pdevice->tag = PLX_VIDEO;
    }

    /*
     * Handle different subformats
     */
    if (framePtr->frameHeader.subformat == VID_FRAME_CODE) {
	memcpy(&(pdevice->cimage), framePtr->data, sizeof(XPlxCImage));
	if (framePtr->dataBuffer->shared) {
#ifdef HAVE_SHARED_MEMORY	  
	    pdevice->cimage.info = &(framePtr->dataBuffer->shminfo);
#endif
	} else {
	    pdevice->cimage.info = NULL;
	}
	pdevice->cimage.data = framePtr->data + sizeof(XPlxCImage);
    } else if (framePtr->frameHeader.subformat == JFIF_NO_HUFFMAN_FRAME_CODE) {
	pdevice->cimage.width = framePtr->frameHeader.width;
	pdevice->cimage.height = framePtr->frameHeader.height;
	pdevice->cimage.size = framePtr->dataLength - 195;
	pdevice->cimage.data = framePtr->data + 195;
	/*
	 * If data is not in VID_FRAME_CODE format, can't use shared memory
	 * since begining of data is not at beginning of shared memory segment.
	 */
	pdevice->cimage.info = NULL;
    } else if (framePtr->frameHeader.subformat == JFIF_FRAME_CODE) {
	pdevice->cimage.width = framePtr->frameHeader.width;
	pdevice->cimage.height = framePtr->frameHeader.height;
	pdevice->cimage.size = framePtr->dataLength - 609;
	pdevice->cimage.data = framePtr->data + 609;
	/*
	 * If data is not in VID_FRAME_CODE format, can't use shared memory
	 * since begining of data is not at beginning of shared memory segment.
	 */
	pdevice->cimage.info = NULL;
    } else if (framePtr->frameHeader.subformat == RAW_JPEG_NO_HEADERS_CODE) {
	pdevice->cimage.width = framePtr->frameHeader.width;
	pdevice->cimage.height = framePtr->frameHeader.height;
	pdevice->cimage.size = framePtr->dataLength;
	pdevice->cimage.data = framePtr->data;
	pdevice->cimage.info = NULL;
    } else {
	return 0;
    }

    WaitForCompletion(pdevice);

    /*
     * Calculate source and dest image rectangles. 
     */

    sx = 0;
    sw = framePtr->frameHeader.width;
    dx = framePtr->objPtr->xpos;
    dw = framePtr->objPtr->width;

    if (dw == 0) {
	dw = sw;
	if (dx < 0) dx = 0;
    } else if (dw > sw) {
	if (!framePtr->objPtr->squeeze) {
	    if (dx == -1) {
		dx = (dw - sw - 1) / 2;
		dw = sw;
	    } else {
		if ((dw - dx) < sw) {
		    dw -= dx;
		} else {
		    dw = sw;
		}
	    }
	} else {
	    if (dx == -1) {
		dx = 0;
	    } else {
		dw -= dx;
	    }
	}
    } else {
	if (!framePtr->objPtr->squeeze) {
	    if (dx == -1) {
		sx =  (sw - dw - 1) / 2;
		sw = dw;
	    } else {
		dw -= dx;
	    }
	} else {
	    if (dx == -1) {
		dx = 0;
	    } else {
		dw -= dx;
	    }
	}
    }

    sy = 0;
    sh = framePtr->frameHeader.height;
    dy = framePtr->objPtr->ypos;
    dh = framePtr->objPtr->height;
    
    if (dh == 0) {
	dh = sh;
	if (dy < 0) dy = 0;
    } else if (dh > sh) {
	if (!framePtr->objPtr->squeeze) {
	    if (dy == -1) {
		dy = (dh - sh - 1) / 2;
		dh = sh;
	    } else {
		if ((dh - dy) < sh) {
		    dh -= dy;
		} else {
		    dh = sh;
		}
	    }
	} else {
	    if (dy == -1) {
		dy = 0;
	    } else {
		dh -= dy;
	    }
	}
    } else {
	if (!framePtr->objPtr->squeeze) {
	    if (dy == -1) {
		sy =  (sh - dh - 1) / 2;
		sh = dh;
	    } else {
		dh -= dy;
	    }
	} else {
	    if (dy == -1) {
		dy = 0;
	    } else {
		dh -= dy;
	    }
	}
    }

    if (canXAttach == -1) {
	canXAttach = cmXShmQueryExtension(pdevice->dpy);
    }

    if (pdevice->cimage.info != NULL && canXAttach) {
	XPlxShmPutCImage(pdevice->dpy, pdevice->win, pdevice->gc,
			 &(pdevice->cimage), sx, sy, sw, sh,
			 dx, dy, dw, dh, framePtr->objPtr->squeeze, 1);
	pdevice->waiting = 1;
    } else {
	XPlxPutCImage(pdevice->dpy, pdevice->win, pdevice->gc,
		      &(pdevice->cimage), 
		      sx, sy, sw, sh, dx, dy, dw, dh,
		      framePtr->objPtr->squeeze);
	pdevice->waiting = 0;
    }

    XFlush (pdevice->dpy);

    if (pdevice->lastDataBuffer) {
	DetachBuffer(framePtr->objPtr->interp, pdevice->lastDataBuffer,
		     "parallax:ParallaxDisplay"); 
	pdevice->lastDataBuffer = NULL;
	pdevice->lastPicData = NULL;
	pdevice->lastPicSubformat = 0;
	pdevice->lastPicDataLength = 0;
	pdevice->lastPicWidth = pdevice->lastPicHeight = 0;

    }

    pdevice->lastPicData = framePtr->data;
    pdevice->lastPicSubformat = framePtr->frameHeader.subformat;
    pdevice->lastPicDataLength = framePtr->dataLength;
    pdevice->lastPicWidth = framePtr->frameHeader.width;
    pdevice->lastPicHeight = framePtr->frameHeader.height;
    pdevice->lastDataBuffer = framePtr->dataBuffer;
    AttachBuffer(framePtr->objPtr->interp, pdevice->lastDataBuffer,
		 "parallax:ParallaxDisplay"); 
    
    if (pdevice->tag == PLX_VIDEO) {
	XPlxVideoTag(pdevice->dpy, pdevice->win, pdevice->gc, PLX_VIDEO_OVR);
	pdevice->tag = PLX_VIDEO_OVR;
    }
}

/*
 *--------------------------------------------------------------
 *
 * ParallaxRefresh --
 *
 *	Redraw the video frame.  Called typically from the reception
 *	of an expose event.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Window is redrawn
 *
 *--------------------------------------------------------------
 */
int
ParallaxRefresh(MJPEGResource *rsrcPtr)
{
    ParallaxDevice *pdevice;
    int sx, sy, dx, dy, sh, sw, dh, dw;
    double speed;

    pdevice = (ParallaxDevice *) rsrcPtr->devicePtr->deviceData;

    /*
     * If nothing to redraw, return
     *!/
    if (pdevice->lastDataBuffer == NULL) {
	return 0;
    }

    /*
     * Change the tag to force the screen update
     */
    XPlxVideoTag (pdevice->dpy, pdevice->win, pdevice->gc, PLX_VIDEO);
    pdevice->tag = PLX_VIDEO;
    
    /*
     * Branch according to data format
     */
    if (pdevice->lastPicSubformat == VID_FRAME_CODE) {
	memcpy(&(pdevice->cimage), pdevice->lastPicData, sizeof(XPlxCImage));
	if (pdevice->lastDataBuffer->shared) {
#ifdef HAVE_SHARED_MEMORY
	    pdevice->cimage.info = &(pdevice->lastDataBuffer->shminfo);
#endif
	} else {
	    pdevice->cimage.info = NULL;
	}
	pdevice->cimage.data = pdevice->lastPicData + sizeof(XPlxCImage);
    } else if (pdevice->lastPicSubformat == JFIF_NO_HUFFMAN_FRAME_CODE) {
	pdevice->cimage.width = pdevice->lastPicWidth;
	pdevice->cimage.height = pdevice->lastPicHeight;
	pdevice->cimage.size = pdevice->lastPicDataLength - 195;
	pdevice->cimage.data = pdevice->lastPicData + 195;
	/*
	 * If data is not in VID_FRAME_CODE format, can't use shared memory
	 * since begining of data is not at beginning of shared memory segment.
	 */
	pdevice->cimage.info = NULL;
    } else if (pdevice->lastPicSubformat == JFIF_FRAME_CODE) {
	pdevice->cimage.width = pdevice->lastPicWidth;
	pdevice->cimage.height = pdevice->lastPicHeight;
	pdevice->cimage.size = pdevice->lastPicDataLength - 609;
	pdevice->cimage.data = pdevice->lastPicData + 609;
	/*
	 * If data is not in VID_FRAME_CODE format, can't use shared memory
	 * since begining of data is not at beginning of shared memory segment.
	 */
	pdevice->cimage.info = NULL;
    } else if (pdevice->lastPicSubformat == RAW_JPEG_NO_HEADERS_CODE) {
	pdevice->cimage.width = pdevice->lastPicWidth;
	pdevice->cimage.height = pdevice->lastPicHeight;
	pdevice->cimage.size = pdevice->lastPicDataLength;
	pdevice->cimage.data = pdevice->lastPicData;
	/*
	 * If data is not in VID_FRAME_CODE format, can't use shared memory
	 * since begining of data is not at beginning of shared memory segment.
	 */
	pdevice->cimage.info = NULL;
    } else {
	return 0;
    }

    WaitForCompletion(pdevice);

    /*
     * Calculate source and dest image rectangles. 
     */

    sx = 0;
    sw = pdevice->cimage.width;
    dx = rsrcPtr->xpos;
    dw = rsrcPtr->width;

    if (dw == 0) {
	dw = sw;
	if (dx < 0) dx = 0;
    } else if (dw > sw) {
	if (!rsrcPtr->squeeze) {
	    if (dx == -1) {
		dx = (dw - sw - 1) / 2;
		dw = sw;
	    } else {
		if ((dw - dx) < sw) {
		    dw -= dx;
		} else {
		    dw = sw;
		}
	    }
	} else {
	    if (dx == -1) {
		dx = 0;
	    } else {
		dw -= dx;
	    }
	}
    } else {
	if (!rsrcPtr->squeeze) {
	    if (dx == -1) {
		sx =  (sw - dw - 1) / 2;
		sw = dw;
	    } else {
		dw -= dx;
	    }
	} else {
	    if (dx == -1) {
		dx = 0;
	    } else {
		dw -= dx;
	    }
	}
    }

    sy = 0;
    sh = pdevice->cimage.height;
    dy = rsrcPtr->ypos;
    dh = rsrcPtr->height;
    
    if (dh == 0) {
	dh = sh;
	if (dy < 0) dy = 0;
    } else if (dh > sh) {
	if (!rsrcPtr->squeeze) {
	    if (dy == -1) {
		dy = (dh - sh - 1) / 2;
		dh = sh;
	    } else {
		if ((dh - dy) < sh) {
		    dh -= dy;
		} else {
		    dh = sh;
		}
	    }
	} else {
	    if (dy == -1) {
		dy = 0;
	    } else {
		dh -= dy;
	    }
	}
    } else {
	if (!rsrcPtr->squeeze) {
	    if (dy == -1) {
		sy =  (sh - dh - 1) / 2;
		sh = dh;
	    } else {
		dh -= dy;
	    }
	} else {
	    if (dy == -1) {
		dy = 0;
	    } else {
		dh -= dy;
	    }
	}
    }

    /*
     * Dump the image
     */
    if (canXAttach == -1) {
	canXAttach = cmXShmQueryExtension(pdevice->dpy);
    }

    if (pdevice->cimage.info != NULL && canXAttach) {
	XPlxShmPutCImage(pdevice->dpy, pdevice->win, pdevice->gc,
			 &(pdevice->cimage), 
			 sx, sy, sw, sh, dx, dy, dw, dh,
			 rsrcPtr->squeeze, 1);
	pdevice->waiting = 1;
    } else {
	XPlxPutCImage(pdevice->dpy, pdevice->win, pdevice->gc,
		      &(pdevice->cimage), 
		      sx, sy, sw, sh, dx, dy, dw, dh,
		      rsrcPtr->squeeze);
	pdevice->waiting = 0;
    }

    /*
     * Reset the tag
     */
    if (pdevice->tag == PLX_VIDEO) {
	XPlxVideoTag(pdevice->dpy, pdevice->win, pdevice->gc, PLX_VIDEO_OVR);
	pdevice->tag = PLX_VIDEO_OVR;
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * ParallaxShutdown --
 *
 *	Called to destroy device
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	The device must not be referenced until re-initialized
 *
 *--------------------------------------------------------------
 */
int
ParallaxShutdown(MJPEGResource *rsrcPtr) 
{
    ParallaxDevice *pdevice = (ParallaxDevice *) rsrcPtr->devicePtr->deviceData;

    if (!rsrcPtr->inited) {
	ckfree((char *) pdevice);
	rsrcPtr->devicePtr->deviceData = NULL;
	return 0;
    }

    WaitForCompletion(pdevice);

    if (pdevice->lastDataBuffer) {
	DetachBuffer(rsrcPtr->interp, pdevice->lastDataBuffer,
		     "parallax:ParallaxShutdown"); 
	pdevice->lastDataBuffer = NULL;
	pdevice->lastPicData = NULL;
    }

    Tk_DeleteGenericHandler (ProcessShmEvent, (ClientData)pdevice);
    ckfree((char *) pdevice);
    rsrcPtr->devicePtr->deviceData = NULL;

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * ParallaxFreeFrame --
 *
 *	Called to free whatever device data the parallax may
 *	have created to display the frame.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
int
ParallaxFreeFrame(MJPEGFrame *framePtr)
{
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * ParallaxError --
 *
 *	Map an error code (an integer returned by one of the functions
 *	above) into a human readable string.
 *
 * Results:
 *	Constant string.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
char *
ParallaxError(MJPEGResource *rsrcPtr, int errNum)
{
    if (errNum == NO_WINDOW_ERROR) {
	return "Attempt to initialize parallax device with no window set";
    }
    if (errNum == NO_HARDWARE_ERROR) {
	return "Parallax hardware is not present on display";
    }
    return "Parallax error.";
}


/*
 *--------------------------------------------------------------
 *
 * GetParallaxDevice --
 *
 *	Called when the code is compiled to use the parallax device.
 *
 * Results:
 *	Pointer to new device struct.
 *
 * Side effects:
 *	A new device is allocated.
 *
 *--------------------------------------------------------------
 */
MJPEGDevice *
GetParallaxDevice (MJPEGResource *objPtr)
{
    MJPEGDevice *new_dev;

    new_dev = (MJPEGDevice *) ckalloc(sizeof(MJPEGDevice));

    new_dev->initProc = ParallaxInit;
    new_dev->reinitProc = ParallaxReinit;
    new_dev->shutdownProc = ParallaxShutdown;

    new_dev->asyncDecodeFlag = 0;
    new_dev->decodeProc = ParallaxDecode;
    new_dev->displayProc = ParallaxDisplay;
    new_dev->refreshProc = ParallaxRefresh;
    new_dev->freeFrameProc = ParallaxFreeFrame;
    new_dev->configProc = NULL;
    new_dev->nextFrameProc = NULL;

    new_dev->errorProc = ParallaxError;

    new_dev->deviceData = (ParallaxDevice *) ckalloc(sizeof(ParallaxDevice));

    return new_dev;
}


#else	/* PARALLAX */

/*
 *--------------------------------------------------------------
 *
 * GetParallaxDevice --
 *
 *	Stub routine for machines without parallax boards
 *
 * Results:
 *	NULL
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
MJPEGDevice *
GetParallaxDevice (MJPEGResource *objPtr)
{
    return NULL;
}

#endif
