/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * --------------------------------------------------
 *
 * cmRTPH261Defrag.c
 *
 * Implementation of the "rtpH261Defrag" object.
 * Takes RTP packets from an rtpDest object,
 * reassembles them in to H261 frames, and sends
 * them out as a CM H261 stream.
 *
 * $Id: cmRTPH261Defrag.c,v 1.4 1996/02/07 20:33:51 kpatel Exp $
 *
 * --------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>

#include "tcl.h"

#include "cmAtom.h"
#include "cmBuffer.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmRTP.h"

typedef struct RTPH261Defrag {
    Tcl_Interp *interp;		/* Tcl interpreter that created object */
    char *name;			/* Tcl string name of this object */

    char *outCmd;		/* Out Command */
    InFunc outCCmd;
    void *outObj;
} RTPH261Defrag;

/* Globals used by this object module. */
static int nextRTPH261DefragNum = 0;
static int H261_BUFFER_SIZE = (1024 * 32);

/*
 * Prototypes for static routines
 */
static int RTPH261DefragCreateCmd (ClientData clientData, Tcl_Interp *interp,
				   int argc, char **argv);
static RTPH261Defrag *CreateNewRTPH261Defrag(Tcl_Interp *interp, char *name);
static int RTPH261DefragObjCmd(ClientData clientData, Tcl_Interp *interp,
			       int argc, char **argv);
static int DestroyRTPH261Defrag(RTPH261Defrag *objPtr);
static int ConfigureRTPH261Defrag(RTPH261Defrag *objPtr,
				  int argc, char **argv);
static int RTPH261DefragInFunc(RTPH261Defrag *objPtr, SB_List *data);


/*
 * --------------------------------------------------
 *
 * RTPH261DefragModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     keywords as well as creating the Tcl command
 *     for object type creation
 *
 * --------------------------------------------------
 */

int
RTPH261DefragModuleInit(Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "rtpH261Defrag", RTPH261DefragCreateCmd,
		      NULL, (void (*)()) NULL);

    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPH261DefragCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the rtpH261Defrag command.
 *     A new object of type RTPH261Defrag is created and
 *     initialized.  Naming of the new object follows the
 *     proper object naming scheme for all cm objects. A
 *     Tcl command is bound to the object name to implement
 *     the objects methods.
 *
 * --------------------------------------------------
 */

static int
RTPH261DefragCreateCmd (ClientData clientData, Tcl_Interp *interp,
		  int argc, char **argv)
{
    char new_name[256];
    RTPH261Defrag *new;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent", (char *) NULL);
	return TCL_ERROR;
    }

    nextRTPH261DefragNum++;
    sprintf(new_name, "%srtpH261defrag%d", argv[1], nextRTPH261DefragNum);

    new = CreateNewRTPH261Defrag(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, new_name, RTPH261DefragObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * CreateNewRTPH261Defrag
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 * --------------------------------------------------
 */

static RTPH261Defrag *
CreateNewRTPH261Defrag(Tcl_Interp *interp, char *name)
{
    RTPH261Defrag *new;
    char *atomName, *suffix = "accept";

    /*
     * Create the new object
     */
    new = (RTPH261Defrag *) ckalloc(sizeof(RTPH261Defrag));
    new->interp = interp;

    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outObj = NULL;
    
    atomName = ckalloc(strlen(name) + strlen(suffix) + 3);
    sprintf(atomName, "%s.%s", name, suffix);
    Tutil_RegisterCommand(interp, atomName, (ClientData) RTPH261DefragInFunc,
			  (ClientData) new);
    
    return new;
}


/*
 * --------------------------------------------------
 *
 * RTPH261DefragObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     RTPH261Defrag object. It is passed a pointer to
 *     the actual data structure of the particular
 *     object instance.
 *
 * --------------------------------------------------
 */
static int
RTPH261DefragObjCmd(ClientData clientData, Tcl_Interp *interp,
	      int argc, char **argv)
{
    char c;
    int len;
    RTPH261Defrag *objPtr = (RTPH261Defrag *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 "<command> ?arguments?", (char *) NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyRTPH261Defrag(clientData);
    } else if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureRTPH261Defrag(clientData, argc-2, argv+2);
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0],
			argv[1], (char *) NULL);
  usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                (char *) NULL);
    return TCL_ERROR;
}


/*
 * --------------------------------------------------
 *
 * DestroyRTPH261Defrag
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 * --------------------------------------------------
 */
static int 
DestroyRTPH261Defrag(RTPH261Defrag *objPtr)
{
    Tcl_HashEntry *entry;
    Tcl_HashSearch search;

    /*
     * Delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);
    
    ckfree((char *) objPtr);
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * ConfigureRTPH261Defrag
 *
 * Parameters: 
 *	objPtr - the object to configure
 *	argc - argument count
 *	argv - argument values
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *	Configures the packet dest object
 *
 * --------------------------------------------------
 */

static int
ConfigureRTPH261Defrag(RTPH261Defrag *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr, len;
    char c;
    
    interp = objPtr->interp;
    
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	} else {
	    Tcl_AppendResult(interp, "-outCmd {} {",
			     objPtr->outCmd, "}", NULL);
	}

	return TCL_OK;
    } 
    
    if (argc == 1) {
	len = strlen(argv[0]);

	if (strncmp("-outCmd", argv[0], len) == 0) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {",
				 objPtr->outCmd, "}", NULL);
	    }
	    return TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "bad configuration option: ",
			     argv[0], NULL);
	    return TCL_ERROR;
	}
    }
    
    cntr = 0;
    while (cntr < argc) {
	if ((cntr + 1) >= argc) {
	    Tcl_AppendResult(interp,
			     "Missing value for option: ", argv[cntr], NULL);
	    return TCL_ERROR;
	}
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    Tcl_CmdInfo cmdInfo;
	    char *objName, *division;

            cntr++;
            if (cntr == argc) {
                Tcl_AppendResult(interp, "Missing paramter for -outCmd",
                                 NULL);
                return TCL_ERROR;
            }

            if (objPtr->outCmd != NULL) {
                ckfree((char *) objPtr->outCmd);
            }

            objPtr->outCmd = ckalloc(strlen(argv[cntr]) + 1);
            strcpy(objPtr->outCmd, argv[cntr]);

            if (objPtr->outCmd[0] == '@') {
		if (Tutil_GetRegisteredCommand(interp, objPtr->outCmd,
					       (ClientData *) &(objPtr->outObj),
					       (ClientData *) &(objPtr->outCCmd)) != TCL_OK) {
                    Tcl_AppendResult(interp,
                                     "Could not find register procedure ",
                                     argv[cntr] + 1, NULL);
		    ckfree(objPtr->outCmd);
		    objPtr->outCmd = NULL;
		    objPtr->outObj = NULL;
		    objPtr->outCCmd = NULL;
                    return TCL_ERROR;
                }
            } else {
                objPtr->outObj = NULL;
                objPtr->outCCmd = NULL;
            }
	} else {
	    Tcl_AppendResult(interp, "Invalid configuration parameter: ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}
	
	cntr++;
    }
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPH261DefragInFunc
 *
 * ...
 *
 * --------------------------------------------------
 */
static int
RTPH261DefragInFunc(RTPH261Defrag *objPtr, SB_List *data)
{
    int i, len, hdrsize;
    ScatterBuffer *sbPtr;
    unsigned char *dataPtr, *bufPtr;
    RTPHeader *rh;
    Buffer *bufs[2], *tmpBuffer, *b;
    H261BufferHeader *hdr;
    ScatterBuffer sb;
    int offsets[2], lens[2];
    SB_List sbl;
    char *sbStr;
    
    if ((data == NULL) || (data->count <  1)) {
	return TCL_OK;
    }

    for (i=0; i<data->count; i++) {
	sbPtr = data->sbs[i];
	
	dataPtr = GetFromSB(sbPtr, CM_RTP_PACKET, &tmpBuffer, &len);
	if (dataPtr == NULL) {
	    continue;
	}
	
	rh = (RTPHeader *) dataPtr;
	
	if (rtpext(*rh)) {
	    /* TODO step over header */
	} else {
	    hdrsize = sizeof(*rh) + 4*rtpcc(*rh);
	    dataPtr += hdrsize;
	    len -= hdrsize;
	}

	b = MakeBuffer(objPtr->interp, sizeof(H261BufferHeader) + 16,
		       "RTPH261Defrag:InFunc");
	if (b == NULL) {
	    return TCL_OK;
	}
	
	bufPtr = b->data;
	((int *) bufPtr)[0] = CM_H261_BUF_HDR;
	((int *) bufPtr)[1] = sizeof(H261BufferHeader);
	bufPtr += 8;
	
	hdr = (H261BufferHeader *) bufPtr;
	/* XXX */
	hdr->logStart_s = 0;
	hdr->logStart_ns = 0;
	hdr->logEnd_s = 2000000000;
	hdr->logEnd_ns = 0;
	memcpy(&hdr->options, dataPtr, 4);
	dataPtr += 4;
	len -= 4;
	bufPtr += sizeof(H261BufferHeader);
	
	((int *) bufPtr)[0] = CM_DATA_SEGMENT;
	((int *) bufPtr)[1] = len;
	
	sb.count = 2;
	sb.buffers = bufs;
	sb.offsets = offsets;
	sb.lens = lens;

	bufs[0] = b;
	bufs[1] = tmpBuffer;
	offsets[0] = 0;
	offsets[1] = (dataPtr - (unsigned char *)tmpBuffer->data);
	lens[0] = sizeof(H261BufferHeader) + 16;
	lens[1] = len;
	
	sbl.count = 1;
	sbl.sbs = &sbPtr;
	sbPtr = &sb;
	    
	if (objPtr->outCmd[0] == '@') {
	    (objPtr->outCCmd)(objPtr->outObj, &sbl);
	} else {
	    sbStr = MakeSBListString(&sbl);
	    Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ", sbStr, NULL);
	    ckfree(sbStr);
	}
	
	DetachBuffer(objPtr->interp, b, "RTPH261Defrag:InFunc");
    }
}
