/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * --------------------------------------------------
 *
 * cmRTPJpegDefrag.c
 *
 * Implementation of the "rtpJpegDefrag" object.
 * Takes RTP packets from an rtpDest object,
 * reassembles them in to jpeg frames, and sends
 * them out as a CM jpeg stream.
 *
 * $Id: cmRTPJpegDefrag.c,v 1.6 1996/02/07 20:33:52 kpatel Exp $
 *
 * --------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>

#include "tcl.h"

#include "cmAtom.h"
#include "cmBuffer.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmRTP.h"
#include "cmRTPJpeg.h"

#ifndef abs
#  define abs(a)	((a)>0?(a):-(a))
#endif
#define min(a,b)	((a)>(b)?(b):(a))
#define max(a,b)	((a)>(b)?(a):(b))

typedef struct RTPJpegDefrag {
    Tcl_Interp *interp;		/* Tcl interpreter that created object */
    char *name;			/* Tcl string name of this object */

    unsigned long timestamp;	/* timestamp for frame being assembled */
    Buffer *b;			/* buffer holding the current frame */
    int offset;			/* offset of last packed received */

    char *outCmd;
    InFunc outCCmd;
    void *outObj;
    
    Buffer *header;		/* Buffer to hold header */
} RTPJpegDefrag;

/* Globals used by this object module. */
static int nextRTPJpegDefragNum = 0;   /* Variable for next unique object name. */
static int JPEG_BUFFER_SIZE = (1024 * 32);

/*
 * Prototypes for static routines
 */
static int RTPJpegDefragCreateCmd (ClientData clientData, Tcl_Interp *interp,
				   int argc, char **argv);
static RTPJpegDefrag *CreateNewRTPJpegDefrag(Tcl_Interp *interp, char *name);
static int RTPJpegDefragObjCmd(ClientData clientData, Tcl_Interp *interp,
			       int argc, char **argv);
static int DestroyRTPJpegDefrag(RTPJpegDefrag *objPtr);
static int ConfigureRTPJpegDefrag(RTPJpegDefrag *objPtr,
				  int argc, char **argv);
static int RTPJpegDefragInFunc(RTPJpegDefrag *objPtr, SB_List *data);


/*
 * --------------------------------------------------
 *
 * RTPJpegDefragModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     keywords as well as creating the Tcl command
 *     for object type creation
 *
 * --------------------------------------------------
 */

int
RTPJpegDefragModuleInit(Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "rtpJpegDefrag", RTPJpegDefragCreateCmd,
		      NULL, (void (*)()) NULL);

    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPJpegDefragCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the rtpJpegDefrag command.
 *     A new object of type RTPJpegDefrag is created and
 *     initialized.  Naming of the new object follows the
 *     proper object naming scheme for all cm objects. A
 *     Tcl command is bound to the object name to implement
 *     the objects methods.
 *
 * --------------------------------------------------
 */

static int
RTPJpegDefragCreateCmd (ClientData clientData, Tcl_Interp *interp,
		  int argc, char **argv)
{
    char new_name[256];
    RTPJpegDefrag *new;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent", (char *) NULL);
	return TCL_ERROR;
    }

    nextRTPJpegDefragNum++;
    sprintf(new_name, "%srtpjpegdefrag%d", argv[1], nextRTPJpegDefragNum);

    new = CreateNewRTPJpegDefrag(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, new_name, RTPJpegDefragObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * CreateNewRTPJpegDefrag
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 * --------------------------------------------------
 */

static RTPJpegDefrag *
CreateNewRTPJpegDefrag(Tcl_Interp *interp, char *name)
{
    RTPJpegDefrag *new;
    char *atomName, *suffix = "accept";
    MJPEGBufferHeader *bufHdr;

    /*
     * Create the new object
     */
    new = (RTPJpegDefrag *) ckalloc(sizeof(RTPJpegDefrag));
    new->interp = interp;

    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    new->timestamp = 0.0;
    new->b = NULL;
    new->offset = 0;
    
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outObj = NULL;
    
    new->header = MakeBuffer(interp, 1024,
			     "RTPJpegDefrag:CreateNewRTPJpegDefrag");
    bufHdr = (MJPEGBufferHeader *) new->header->data;
    bufHdr->hdrType = ntohl(CM_MJPEG_BUF_HDR);
    bufHdr->length = ntohl(sizeof(MJPEGBufferHeader) - 8);
    bufHdr->device = ntohl(OTHER_DEV_CODE);
    bufHdr->subformat = ntohl(RAW_JPEG_NO_HEADERS_CODE);

    atomName = ckalloc(strlen(name) + strlen(suffix) + 3);
    sprintf(atomName, "%s.%s", name, suffix);
    Tutil_RegisterCommand(interp, atomName, (ClientData) RTPJpegDefragInFunc,
			  (ClientData) new);
    return new;
}


/*
 * --------------------------------------------------
 *
 * RTPJpegDefragObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     RTPJpegDefrag object. It is passed a pointer to
 *     the actual data structure of the particular
 *     object instance.
 *
 * --------------------------------------------------
 */
static int
RTPJpegDefragObjCmd(ClientData clientData, Tcl_Interp *interp,
	      int argc, char **argv)
{
    char c;
    int len;
    RTPJpegDefrag *objPtr = (RTPJpegDefrag *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 "<command> ?arguments?", (char *) NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyRTPJpegDefrag(clientData);
    } else if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureRTPJpegDefrag(clientData, argc-2, argv+2);
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0],
			argv[1], (char *) NULL);
  usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                (char *) NULL);
    return TCL_ERROR;
}


/*
 * --------------------------------------------------
 *
 * DestroyRTPJpegDefrag
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 * --------------------------------------------------
 */
static int 
DestroyRTPJpegDefrag(RTPJpegDefrag *objPtr)
{
    Tcl_HashEntry *entry;
    Tcl_HashSearch search;

    /*
     * Delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);
    
    DetachBuffer(objPtr->interp, objPtr->b,
		 "RTPJpegDefrag:DestroyRTPJpegDefrag");
    
    ckfree((char *) objPtr);
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * ConfigureRTPJpegDefrag
 *
 * Parameters: 
 *	objPtr - the object to configure
 *	argc - argument count
 *	argv - argument values
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *	Configures the packet dest object
 *
 * --------------------------------------------------
 */

static int
ConfigureRTPJpegDefrag(RTPJpegDefrag *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr, len;
    char c;
    
    interp = objPtr->interp;
    
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	} else {
	    Tcl_AppendResult(interp, "-outCmd {} {",
			     objPtr->outCmd, "}", NULL);
	}

	return TCL_OK;
    } 
    
    if (argc == 1) {
	len = strlen(argv[0]);

	if (strncmp("-outCmd", argv[0], len) == 0) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {",
				 objPtr->outCmd, "}", NULL);
	    }
	    return TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "bad configuration option: ",
			     argv[0], NULL);
	    return TCL_ERROR;
	}
    }
    
    cntr = 0;
    while (cntr < argc) {
	if ((cntr + 1) >= argc) {
	    Tcl_AppendResult(interp,
			     "Missing value for option: ", argv[cntr], NULL);
	    return TCL_ERROR;
	}
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    Tcl_CmdInfo cmdInfo;
	    char *objName, *division;

            cntr++;
            if (cntr == argc) {
                Tcl_AppendResult(interp, "Missing paramter for -outCmd",
                                 NULL);
                return TCL_ERROR;
            }

            if (objPtr->outCmd != NULL) {
                ckfree((char *) objPtr->outCmd);
            }

            objPtr->outCmd = ckalloc(strlen(argv[cntr]) + 1);
            strcpy(objPtr->outCmd, argv[cntr]);

            if (objPtr->outCmd[0] == '@') {
		if (Tutil_GetRegisteredCommand(interp, objPtr->outCmd,
					       (ClientData *) &(objPtr->outObj),
					       (ClientData *) &(objPtr->outCCmd)) != TCL_OK) {
		    Tcl_AppendResult(interp,
				     "Could not find register procedure ",
				     argv[cntr] + 1, NULL);
		    ckfree(objPtr->outCmd);
		    objPtr->outCmd = NULL;
		    objPtr->outObj = NULL;
		    objPtr->outCCmd = NULL;
		    return TCL_ERROR;
		}
            } else {
                objPtr->outObj = NULL;
                objPtr->outCCmd = NULL;
            }
	} else {
	    Tcl_AppendResult(interp, "Invalid configuration parameter: ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}
	
	cntr++;
    }
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPJpegDefragInFunc
 *
 * ...
 *
 * --------------------------------------------------
 */
static int
RTPJpegDefragInFunc(RTPJpegDefrag *objPtr, SB_List *data)
{
    int i;
    ScatterBuffer *sb;
    unsigned char *dataPtr;
    int len;
    RTPHeader *rh;
    RTPJpegHdr *jh;
    int offset;
    
    if ((data == NULL) || (data->count <  1)) {
	return TCL_OK;
    }

    for (i=0; i<data->count; i++) {
	sb = data->sbs[i];
	
	dataPtr = GetFromSB(sb, CM_RTP_PACKET, NULL, &len);
	if (dataPtr == NULL) {
	    continue;
	}
	
	len -= 8;
	
	rh = (RTPHeader *) dataPtr;
    
	if (rtpext(*rh)) {
	    /* TODO step over the header */
	} else {
	    dataPtr += sizeof(*rh) + (4 * rtpcc(*rh));
	}
	
	jh = (RTPJpegHdr *) dataPtr;
	dataPtr += sizeof(*jh);
	
	len -= sizeof(*rh) + (4*rtpcc(*rh)) + sizeof(*jh);
	
	if (objPtr->b == NULL) {
	    /* XXX shared? */
	    objPtr->b = MakeBuffer(objPtr->interp, JPEG_BUFFER_SIZE,
				   "RTPJpegDefrag:RTPJpegDefragInFunc");
	    if (objPtr->b == NULL) {
		return TCL_ERROR;
	    }

	    ((int *) objPtr->b->data)[0] = ntohl(CM_DATA_SEGMENT);
	    objPtr->timestamp = rh->timestamp;
	    objPtr->offset = 0;
	}
	
	if (rh->timestamp > objPtr->timestamp) {
	    objPtr->timestamp = rh->timestamp;
	    objPtr->offset = 0;
	}
	
	offset = jh->offset & 0x0ffffff;
	if (offset != objPtr->offset) {
	    return TCL_OK;
	}
	
	memcpy(objPtr->b->data + offset + 8, dataPtr, len);
	objPtr->offset += len;
	
	if (rtpmark(*rh)) {
	    int q;
	    ScatterBuffer sbuf, *sbufs[1];
	    SB_List sbl;
	    Buffer *bufs[2];
	    int offsets[2], lens[2];
	    MJPEGBufferHeader *bh;
	    char *sbStr;
	
	    /* set q factor */
	    q = jh->q;
	    if (q < 1) {
		q = 5000;
	    } else if (q < 50) {
		q = 5000 / q;
	    } else if (q < 100) {
		q = 200 - q * 2;
	    } else {
		q = 1;
	    }

	    sbl.count = 1;
	    sbl.sbs = sbufs;
	    sbufs[0] = &sbuf;
	
	    sbuf.count = 2;
	    sbuf.buffers = bufs;
	    sbuf.offsets = offsets;
	    sbuf.lens = lens;

	    bufs[0] = objPtr->header;
	    bufs[1] = objPtr->b;

	    offsets[0] = offsets[1] = 0;

	    lens[0] = sizeof(MJPEGBufferHeader);
	    lens[1] = (objPtr->offset & 0x0ffffff) + 8;

	    ((int *) bufs[1]->data)[1] = ntohl(lens[1]);

	    bh = (MJPEGBufferHeader *) objPtr->header->data;
	    bh->width = ntohl(jh->width * 8);
	    bh->height = ntohl(jh->height * 8);
	    bh->qfactor = ntohl(q);
	    bh->logStart_s = ntohl(0);
	    bh->logStart_ns = ntohl(0);
	    bh->logEnd_s = ntohl(2000000000);
	    bh->logEnd_ns = ntohl(0);
	
	    if (objPtr->outCmd[0] == '@') {
		(objPtr->outCCmd)(objPtr->outObj, &sbl);
	    } else {
		sbStr = MakeSBListString(&sbl);
		Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ", sbStr, NULL);
		ckfree(sbStr);
	    }
	
	    DetachBuffer(objPtr->interp, objPtr->b,
			 "RTPJpegDefrag:RTPJpegDefragInFunc");
	    objPtr->b = MakeBuffer(objPtr->interp, JPEG_BUFFER_SIZE,
				   "RTPJpegDefrag:RTPJpegDefragInFunc");
	    ((int *) objPtr->b->data)[0] = ntohl(CM_DATA_SEGMENT);
	}
    }
}
