/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * --------------------------------------------------
 *
 * cmRTPJpegFrag.c
 *
 * Implementation of the "rtpJpegFrag" object.
 * Takes a CM mjpeg stream and breaks it up in to
 * RTP packets to be handed to an rtpSource object.
 *
 * $Id: cmRTPJpegFrag.c,v 1.6 1996/02/07 20:33:53 kpatel Exp $
 *
 * --------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>

#include "tcl.h"
#include "dp.h"

#include "cmAtom.h"
#include "cmBuffer.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmLTS.h"
#include "cmUtil.h"
#include "cmRTP.h"
#include "cmRTPJpeg.h"

#ifndef abs
#  define abs(a)	((a)>0?(a):-(a))
#endif
#define min(a,b)	((a)>(b)?(b):(a))
#define max(a,b)	((a)>(b)?(a):(b))

typedef struct RTPJpegFrag {
    Tcl_Interp *interp;		/* Tcl interpreter that created object */
    char *name;			/* Tcl string name of this object */
    
    ClientData lts;
    char *ltsName;
    double startTime;

    char *outCmd;
    InFunc outCCmd;
    void *outObj;
} RTPJpegFrag;

/* Globals used by this object module. */
static int nextRTPJpegFragNum = 0;   /* Variable for next unique object name. */

#define PACKET_SIZE 1024
#define PACKET_DATA (PACKET_SIZE - sizeof(RTPHeader) - sizeof(RTPJpegHdr))
#define ONE_BILLION 1E9

/*
 * Prototypes for static routines
 */
static int RTPJpegFragCreateCmd (ClientData clientData, Tcl_Interp *interp,
				 int argc, char **argv);
static RTPJpegFrag *CreateNewRTPJpegFrag(Tcl_Interp *interp, char *name);
static int RTPJpegFragObjCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char **argv);
static int DestroyRTPJpegFrag(RTPJpegFrag *objPtr);
static int ConfigureRTPJpegFrag(RTPJpegFrag *objPtr,
				int argc, char **argv);
static int RTPJpegFragInFunc(RTPJpegFrag *objPtr, SB_List *data);


/*
 * --------------------------------------------------
 *
 * RTPJpegFragModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     keywords as well as creating the Tcl command
 *     for object type creation
 *
 * --------------------------------------------------
 */

int
RTPJpegFragModuleInit(Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "rtpJpegFrag", RTPJpegFragCreateCmd,
		      NULL, (void (*)()) NULL);

    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPJpegFragCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the rtpJpegFrag command.
 *     A new object of type RTPJpegFrag is created and
 *     initialized.  Naming of the new object follows the
 *     proper object naming scheme for all cm objects. A
 *     Tcl command is bound to the object name to implement
 *     the objects methods.
 *
 * --------------------------------------------------
 */

static int
RTPJpegFragCreateCmd (ClientData clientData, Tcl_Interp *interp,
		  int argc, char **argv)
{
    char new_name[256];
    RTPJpegFrag *new;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent", (char *) NULL);
	return TCL_ERROR;
    }

    nextRTPJpegFragNum++;
    sprintf(new_name, "%srtpJpegFrag%d", argv[1], nextRTPJpegFragNum);

    new = CreateNewRTPJpegFrag(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, new_name, RTPJpegFragObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * CreateNewRTPJpegFrag
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 * --------------------------------------------------
 */

static RTPJpegFrag *
CreateNewRTPJpegFrag(Tcl_Interp *interp, char *name)
{
    RTPJpegFrag *new;
    char *atomName, *suffix = "accept";

    /*
     * Create the new object
     */
    new = (RTPJpegFrag *) ckalloc(sizeof(RTPJpegFrag));
    new->interp = interp;

    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outObj = NULL;

    new->lts = NULL;
    new->ltsName = NULL;
    new->startTime = Tcm_ReadSysClock();
    
    atomName = ckalloc(strlen(name) + strlen(suffix) + 3);
    sprintf(atomName, "%s.%s", name, suffix);
    Tutil_RegisterCommand(interp, atomName, (ClientData) RTPJpegFragInFunc,
			  (ClientData) new);
    
    return new;
}


/*
 * --------------------------------------------------
 *
 * RTPJpegFragObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     RTPJpegFrag object. It is passed a pointer to
 *     the actual data structure of the particular
 *     object instance.
 *
 * --------------------------------------------------
 */
static int
RTPJpegFragObjCmd(ClientData clientData, Tcl_Interp *interp,
	      int argc, char **argv)
{
    char c;
    int len;
    RTPJpegFrag *objPtr = (RTPJpegFrag *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 "<command> ?arguments?", (char *) NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyRTPJpegFrag(clientData);
    } else if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureRTPJpegFrag(clientData, argc-2, argv+2);
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0],
			argv[1], (char *) NULL);
  usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                (char *) NULL);
    return TCL_ERROR;
}


/*
 * --------------------------------------------------
 *
 * DestroyRTPJpegFrag
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 * --------------------------------------------------
 */
static int 
DestroyRTPJpegFrag(RTPJpegFrag *objPtr)
{
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);
    ckfree((char *) objPtr);

    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * ConfigureRTPJpegFrag
 *
 * Parameters: 
 *	objPtr - the object to configure
 *	argc - argument count
 *	argv - argument values
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *	Configures the packet dest object
 *
 * --------------------------------------------------
 */

static int
ConfigureRTPJpegFrag(RTPJpegFrag *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr, len;
    char c;
    
    interp = objPtr->interp;
    
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	} else {
	    Tcl_AppendResult(interp, "-outCmd {} {",
			     objPtr->outCmd, "}", NULL);
	}
	
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "-lts {} {}", NULL);
	} else {
	    Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName, "}", NULL);
	}

	return TCL_OK;
    } 
    
    if (argc == 1) {
	len = strlen(argv[0]);

	if (strncmp("-outCmd", argv[0], len) == 0) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {",
				 objPtr->outCmd, "}", NULL);
	    }
	    return TCL_OK;
	} else if (strncmp("-lts", argv[0], len) == 0) {
	    if (objPtr->ltsName == NULL) {
		Tcl_AppendResult(interp, "-lts {} {}", NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName, "}", NULL);
	    }
	} else {
	    Tcl_AppendResult(interp, "bad configuration option: ",
			     argv[0], NULL);
	    return TCL_ERROR;
	}
    }
    
    cntr = 0;
    while (cntr < argc) {
	if ((cntr + 1) >= argc) {
	    Tcl_AppendResult(interp,
			     "Missing value for option: ", argv[cntr], NULL);
	    return TCL_ERROR;
	}
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
	
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    Tcl_CmdInfo cmdInfo;
	    char *objName, *division;

            cntr++;
            if (cntr == argc) {
                Tcl_AppendResult(interp, "Missing paramter for -outCmd",
                                 NULL);
                return TCL_ERROR;
            }

            if (objPtr->outCmd != NULL) {
                ckfree((char *) objPtr->outCmd);
            }

            objPtr->outCmd = ckalloc(strlen(argv[cntr]) + 1);
            strcpy(objPtr->outCmd, argv[cntr]);

            if (objPtr->outCmd[0] == '@') {
		if (Tutil_GetRegisteredCommand(interp, objPtr->outCmd,
					       (ClientData *) &(objPtr->outObj),
					       (ClientData *) &(objPtr->outCCmd)) != TCL_OK) {
		    Tcl_AppendResult(interp,
				     "Could not find register procedure ",
				     argv[cntr] + 1, NULL);
		    ckfree(objPtr->outCmd);
		    objPtr->outCmd = NULL;
		    objPtr->outObj = NULL;
		    objPtr->outCCmd = NULL;
		    return TCL_ERROR;
		}
            } else {
                objPtr->outObj = NULL;
                objPtr->outCCmd = NULL;
            }
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;
	    
            cntr++;
            if (cntr == argc) {
                Tcl_AppendResult(interp, "Missing paramter for -lts",
                                 NULL);
                return TCL_ERROR;
            }
	    newLts = Tcm_FindAtom(interp, argv[cntr]);
	    if (newLts == NULL) {
		Tcl_AppendResult(interp, "Bad parameter ", argv[cntr],
				 " for -lts", NULL);
		return TCL_ERROR;
	    }
	    Tk_Preserve(objPtr->lts);
	    objPtr->lts = newLts;

	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
	    }
	    objPtr->ltsName = ckalloc(strlen(argv[cntr]) + 1);
	    strcpy(objPtr->ltsName, argv[cntr]);

	} else {
	    Tcl_AppendResult(interp, "Invalid configuration parameter: ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}
	
	cntr++;
    }
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPJpegFragInFunc
 *
 * ...
 *
 * --------------------------------------------------
 */
static int
RTPJpegFragInFunc(RTPJpegFrag *objPtr, SB_List *sbl)
{
    SB_List sblout;
    ScatterBuffer *sbs[1], sb;
    Buffer *buffers[1];
    int offsets[1], lens[1];
    unsigned char headers[PACKET_SIZE];
    int headerSize, *sizePtr;
    RTPHeader *rhdr;
    RTPJpegHdr *rjhdr;
    int i;
    MJPEGBufferHeader *jhdr;
    char *data;
    int size, offset, q, dataSize;
    double start, frameTime, ts;
    
    if ((sbl == NULL) || (sbl->count !=  1)
	|| (objPtr->outCmd == NULL) || (objPtr->lts == NULL)) {
	return TCL_OK;
    }
    
    AttachToAllBuffers(objPtr->interp, sbl, "RTPJpegFragInFunc");
    
    sblout.count = 1;
    sblout.sbs = sbs;
    sbs[0] = &sb;

    sb.count = 1;
    sb.buffers = buffers;
    sb.lens = lens;
    sb.offsets = offsets;

    offsets[0] = 0;
    
    ((int *) headers)[0] = CM_RTP_PACKET;
    sizePtr = &(((int *) headers)[1]);

    rhdr = (RTPHeader *) (headers + 8);
    rhdr->flags = RTP_JPEG_PT;
    rjhdr = (RTPJpegHdr *) (headers + 8 + sizeof(RTPHeader));
    headerSize = sizeof(RTPHeader) + sizeof(RTPJpegHdr) + 8;
    
    for (i=0; i<sbl->count; i++) {
	char *tmp;

	tmp = GetFromSB(sbl->sbs[i], CM_MJPEG_BUF_HDR, NULL, NULL);
	if (tmp == NULL) {
	    continue;
	}

	jhdr = (MJPEGBufferHeader *) (tmp-8);

	data = GetFromSB(sbl->sbs[i], CM_DATA_SEGMENT, NULL, &size);
	if (data == NULL) {
	    continue;
	}
	
	if (jhdr->subformat == VID_FRAME_CODE) {
	    /* skip over parallax header */
	    data += 20;
	} else if ((jhdr->subformat == JFIF_NO_HUFFMAN_FRAME_CODE)
		   || (jhdr->subformat == JFIF_FRAME_CODE)) {
	    /* TODO skip jfif header */
	}
	
	if (LTS_Speed(objPtr->lts) == 0.0) {
	    frameTime = Tcm_ReadSysClock();
	} else {
	    start = jhdr->logStart_s + ((double) jhdr->logStart_ns)/ONE_BILLION;
	    frameTime = LTS_SystemTimeOf(objPtr->lts, start);
	}
	ts = (frameTime - objPtr->startTime) * RTP_JPEG_CLOCK;
	rhdr->timestamp = (int) ts;

	rjhdr->type = RTP_JPEG_YUV422;
	if (jhdr->qfactor > 100) {
	    q = 5000/jhdr->qfactor;
	} else {
	    q = (200 - jhdr->qfactor) / 2;
	}
	if (q < 1) {
	    rjhdr->q = 1;
	} else if (q>99) {
	    rjhdr->q = 99;
	} else {
	    rjhdr->q = q;
	}

	rjhdr->width = jhdr->width >> 3;
	if (jhdr->width & 0x7) {
	    rjhdr->width++;
	}
	rjhdr->height = jhdr->height >> 3;
	if (jhdr->height & 0x7) {
	    rjhdr->height++;
	}

	offset = 0;
	while (size > 0) {
	    rjhdr->offset = offset;
	    
	    if (size > PACKET_DATA) {
		dataSize = PACKET_DATA;
	    } else {
		dataSize = size;
		rhdr->flags |= 0x80;	/* Set marker bit */
	    }

	    lens[0] = dataSize + headerSize;
	    *sizePtr = lens[0];
	    buffers[0] = MakeBuffer(objPtr->interp, PACKET_SIZE + 8,
				    "RTPJpegFragInFunc");
	    memcpy(buffers[0]->data, headers, headerSize);
	    memcpy(buffers[0]->data + headerSize, data+offset, dataSize);

	    if (objPtr->outCmd[0] == '@') {
		(objPtr->outCCmd)(objPtr->outObj, &sblout);
	    } else {
		char *sbStr;
		sbStr = MakeSBListString(&sblout);
		Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ", sbStr, NULL);
		ckfree(sbStr);
	    }
	    DetachBuffer(objPtr->interp, buffers[0], "RTPJpegFragInFunc");
	    
	    /*
	     * Clear marker bit in case it was set
	     */
	    rhdr->flags &= 0x0ff7f;
	    
	    size -= dataSize;
	    offset += dataSize;
	}
    }
    
    DetachToAllBuffers(objPtr->interp, sbl, "RTPJpegFragInFunc");
}
