/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * --------------------------------------------------
 *
 * cmRTPSession.c
 *
 * This file implements the rtpSession object, used to
 * send and receive data via the Real-Time Protocl (RTP)
 *
 * $Id: cmRTPSession.c,v 1.9 1996/02/07 20:33:55 kpatel Exp $
 *
 * --------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/uio.h>
#include <errno.h>
#include <sys/errno.h>
#include <pwd.h>
#include <math.h>

#ifdef NO_MULTICAST_H
#  include "ip_multi.h"
#endif

#include "tcl.h"
#include "tk.h"
#include "dp.h"

#include "cmAtom.h"
#include "cmBind.h"
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmUtil.h"
#include "cmRTP.h"

#ifndef abs
#  define abs(a)	((a)>0?(a):-(a))
#endif
#define min(a,b)	((a)>(b)?(b):(a))
#define max(a,b)	((a)>(b)?(a):(b))

#define MAX_PACKET 8192		/* Largest RTP packet we handle */
#define MAX_SDES 256	      	/* Largest SDES item we handle */

/*
 * One of these structures is kept for every unique
 * source ID from which we have received data or
 * rtcp reports
 */
typedef struct RTPSenderInfo {
    struct RTPSenderInfo *next;	/* List of all senders */
    int inited;

    unsigned int srcid;		/* The source id of the sender */
    unsigned char pt;		/* Payload Type of most recent pkt */
    char *cname;
    char *name;

    int sender;			/* Is this a sender or just a receiver */
    unsigned short firstseq;	/* Sequence no. of first packet received */
    unsigned short lastseq;	/* Sequence no. of last packet received */
    unsigned short seqCycles;	/* Number of times seq. #'s have cycled */
    
    double dataTs;		/* Time last data packet was received */
    double ctrlTs;		/* Time last rtcp packet was received */
    unsigned int ctrlNtp;	/* NTP timestamp from last rtcp packet */
    double jitter;		/* Interarival time jitter */
    
    unsigned int ackseq;    	/* Last sequence number acked */
    unsigned int ackpkts;     	/* Number of packets received at last ack */

    unsigned int pkts;		/* Total packets received from this source */
    unsigned int bytes;	/* Total bytes received from this source */

    char *outCmd;		/* The command to push data */
    InFunc outCCmd;		/* C pointer to out command */
    void *outObj;		/* The object to send to */
} RTPSenderInfo;

/*
 * One of these structures is kept for every client
 * attached to this session object
 */
typedef struct RTPSessionClient {
    struct RTPSessionClient *next;
    char *name;
    int cmBindingToken;
} RTPSessionClient;

/*
 * The RTPSession structure maintained for every
 * object that is instantiated.
 */
typedef struct RTPSession {
    int refcount;
    RTPSessionClient *clients;	/* Linked list of all clients */
    struct RTPSession *next;	/* Linked list of all session objects */
    Tcl_Interp *interp;

    struct in_addr address;	/* The multicast address */
    short port;			/* Port number for data stream */
    unsigned char ttl;		/* IP Time To Live (TTL) */
    
    int srcid;			/* The source ID we are using */
    unsigned int seqno;		/* Sequence number of next packet */
    unsigned int packets;	/* Number of packets sent */
    unsigned int bytes;		/* Number of bytes sent */
    
    char cname[MAX_SDES];	/* Our CNAME for RTCP packets */
    char name[MAX_SDES];	/* Our NAME for RTCP packets */
    int rtcpCt;			/* Number of RTCP packets we have sent */
    
    int dataFd;			/* The data sending socket */
    int ctrlFd;			/* The RTCP control sending socket */
    int dataRecvFd;		/* The data receving socket */
    int ctrlRecvFd;		/* The RTCP control receiving socket */
    
    char *pg;			/* Priority group for callbacks */
    int atToken;		/* At handler token for RTCP callback */
    
    Tcl_HashTable memberTab;	/* All the conference participants we
				 * have received RTCP packets from */
    struct RTPSenderInfo *memberList;	/* Linked list of all members */
    int numSenders;		/* Count of senders we have heard from */
    int numListeners;		/* Count of total participants in this
				 * conference we have heard from */

    unsigned char outEncoding;
    char *outCname;
    char *outName;		/* Used for cm bindings */

    /*
     * Statistics collected about the conference
     */
    double sessbw;		/* Bandwidth allocated to control data
				 * for this conference */
    int badversion;		/* Packets received with version != 2 */
    int runt;			/* Runt packets */
    int badseqno;		/* Packets with an apparently bad sequence # */
    int rtcpSz;			/* Average size of RTCP packets */
    /* others? */
} RTPSession;

static const int RTPVersion = 2;
static RTPSession *sessionList = NULL;

char *RTPSessionSourceSlots[] = { "-address", "-port", "-srcid", "-cname",
				      "-name", "-pg", "-ttl", NULL };
char *RTPSessionDestSlots[] = { "-address", "-port", "-srcid", "-cname",
				"-name", "-pg", NULL };

typedef enum { SEND, RECEIVE } RTPSessionSockType;

/*
 * Prototypes for internal routines
 */
static void RTPSessionGetName(char *cname, char *name);
static int RTPSessionOpenSocket(struct in_addr address, short port,
				RTPSessionSockType type);
static double RTPSessionCalculateRTCPInterval(RTPSession *objPtr);
static int RTPSessionLeaveGroup(RTPSession *objPtr);
static void RTPSessionHandleData(ClientData cd, int mask);
static RTPSenderInfo *RTPSessionFindMember(RTPSession *objPtr,
					   unsigned int srcid);
static RTPSenderInfo *RTPSessionFindMemberByCNAME(RTPSession *objPtr,
						  char *cname);
static void RTPSessionRemoveMember(RTPSession *objPtr, unsigned int srcid);
static void RTPSessionHandleRTCP(ClientData cd, int mask);
static void RTPSessionParseSDES(RTPSession *objPtr, unsigned char *data);
static void RTPSessionSendRTCP(ClientData cd);
static char *RTPSessionBuildRR(RTPSession *objPtr, char *buffer);
static char *RTPSessionCNAME(Tcl_Interp *interp, ClientData cd);
static char *RTPSessionNAME(Tcl_Interp *interp, ClientData cd);
static char *RTPSessionEncoding(Tcl_Interp *interp, ClientData cd);
static unsigned int GetNTPTime(void);


/*
 * --------------------------------------------------
 *
 * GetRTPSession
 *
 * Parameters: 
 *	interp - pointer to Tcl interpreter
 *	address - multicast address to join
 *	port - port number to listen on
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *	Looks for an existing RTPSesison object
 *	listening to the requested conference and
 *	returns a pointer to it.  If one cannot be
 *	found, creates a new RTPSession object,
 *	initializes it, and returns it.
 *
 * --------------------------------------------------
 */

RTPSession *
GetRTPSession(Tcl_Interp *interp, char *name,
	      struct in_addr *address, short port)
{
    RTPSession *new;
    int bufsize;
    struct timeval tv;
    RTPSessionClient *client;
    static char *events[] = {"<NEW_SENDER>", "<NEW_LISTENER>", "<NAME_CHANGED>", "<MEMBER_LEFT>", "<ENCODING_CHANGED>"};
    static char *escapes[] = {"c", "n", "e"};
    static CM_BindProc *procs[] = {RTPSessionCNAME, RTPSessionNAME, RTPSessionEncoding};
    static ClientData cds[3];
    double rtcpint, now;

    /*
     * Create a client structure
     */
    client = (RTPSessionClient *) ckalloc(sizeof(RTPSessionClient));
    client->name = ckalloc(strlen(name) + 1);
    strcpy(client->name, name);
    
    /*
     * Look for an existing RTPSession object in the list
     * of active sessions
     */
    for (new = sessionList; new != NULL; new = new->next) {
	if ((memcmp(&new->address, address, sizeof(struct in_addr)) == 0)
	    && (new->port == port)) {
	    
	    new->refcount++;
	    break;
	}
    }

    if (new == NULL) {
	/*
	 * Didn't find one, create a new object
	 */
	new = (RTPSession *) ckalloc(sizeof(RTPSession));
	if (new == NULL) {
	    return NULL;
	}
	
	new->refcount = 1;
	new->clients = NULL;
	new->next = sessionList;
	sessionList = new;
	
	new->interp = interp;
	memcpy(&new->address, address, sizeof(new->address) );
	new->port = port;
	new->ttl = 2;
	
	/*
	 * This algorithm for choosing the srcid is
	 * stolen from vic 
	 */
	gettimeofday(&tv, 0);
	new->srcid = tv.tv_sec + tv.tv_usec + (int) getuid()
	    + (int) getpid() + (int) address->s_addr;
	
	new->seqno = 1;
	new->packets = 0;
	new->bytes = 0;
	
	/*
	 * Create CNAME and NAME fields
	 */
	RTPSessionGetName(new->cname, new->name);
	
	new->rtcpCt = 0;
	new->memberList = NULL;
	new->numSenders = 0;
	new->numListeners = 0;
	
	new->outCname = NULL;
	new->outName = NULL;
	
	new->sessbw = 30000.;
	new->badversion = 0;
	new->runt = 0;
	new->badseqno = 0;
	new->rtcpSz = 0;
	
	/*
	 * XXX This should eventually be changed to use Tdp_multicast_connect
	 */
	new->dataFd = RTPSessionOpenSocket(new->address, new->port, SEND);
	if (new->dataFd == -1) {
	    ckfree(new);
	    return NULL;
	}
	
	new->ctrlFd = RTPSessionOpenSocket(new->address, new->port + 1, SEND);
	if (new->ctrlFd == -1) {
	    close(new->dataFd);
	    ckfree(new);
	    return NULL;
	}
	
	new->dataRecvFd = RTPSessionOpenSocket(new->address,
					       new->port, RECEIVE);
	if (new->dataRecvFd == -1) {
	    close(new->dataFd);
	    close(new->ctrlFd);
	    ckfree(new);
	    return NULL;
	}
	
	/*
	 * Expand the buffer used to receive incoming packets.  Try
	 * 64k first.  If that doesn't work, require at least 32k
	 */
	bufsize = 64 * 1024;
	if (setsockopt(new->dataRecvFd, SOL_SOCKET, SO_RCVBUF,
		       (char *) &bufsize, sizeof(bufsize)) < 0) {
	    bufsize = 32 * 1024;
	    if (setsockopt(new->dataRecvFd, SOL_SOCKET, SO_RCVBUF,
			   (char *) &bufsize, sizeof(bufsize)) < 0) {
		close(new->dataFd);
		close(new->ctrlFd);
		close(new->dataRecvFd);
		ckfree(new);
		return NULL;
	    }
	}
	
	new->ctrlRecvFd = RTPSessionOpenSocket(new->address,
					       new->port + 1, RECEIVE);
	if (new->ctrlRecvFd == -1) {
	    close(new->dataFd);
	    close(new->ctrlFd);
	    close(new->dataRecvFd);
	    ckfree(new);
	    return NULL;
	}
	
	new->pg = ckalloc(3);
	strcpy(new->pg, "tk");
	
	/*
	 * Set up callbacks when the sockets become readable
	 */
	Tcm_CreateFileHandler(new->pg, new->dataRecvFd, TK_READABLE,
			     RTPSessionHandleData, (ClientData) new);
	Tcm_CreateFileHandler(new->pg, new->ctrlRecvFd, TK_READABLE,
			     RTPSessionHandleRTCP, (ClientData) new);
	
	/*
	 * Set up at handler
	 */
	rtcpint = RTPSessionCalculateRTCPInterval(new);
	now = Tcm_ReadSysClock();

	new->atToken = Tcm_CreateAtHandler(now + rtcpint, HUGE_VAL, new->pg,
					   RTPSessionSendRTCP,
					   (ClientData) new, NULL, NULL,
					   NULL, NULL, NULL);

	Tcl_InitHashTable(&new->memberTab, TCL_ONE_WORD_KEYS);
    }

    client->next = new->clients;
    new->clients = client;
    cds[0] = cds[1] = cds[2] = new;
    client->cmBindingToken = CreateCMBinding(interp, name, 5, events,
					     3, escapes, procs, cds);
    return new;
}


/*
 * --------------------------------------------------
 *
 * RTPSessonGetName
 *
 * Creates the CNAME and NAME SDES items for this
 * conference.  CNAME is a string of the form
 * username@hostname and NAME is the real name of
 * the user.
 *
 * --------------------------------------------------
 */
static void
RTPSessionGetName(char *cname, char *name)
{
    struct passwd *pwent;
    char *username, *comma;
    int len;

    /*
     * First build CNAME
     */
    pwent = getpwuid(getuid());
    if (pwent != NULL) {
	strcpy(cname, pwent->pw_name);
    } else {
	username = getenv("USER");
	if (username == NULL) {
	    strcpy(cname, "unknown");
	} else {
	    strcpy(name, username);
	}
    }

    len = strlen(cname);
    cname[len++] = '@';
    gethostname(cname + len, (MAX_SDES - len - 1) );
    
    /*
     * If getpwuid() succeded, build NAME
     */
    if (pwent != NULL) {
	username = pwent->pw_gecos;
	comma = strchr(username, ',');
	if (comma == NULL) {
	    strcpy(name, username);
	} else {
	    len = comma - username;
	    strncpy(name, username, len);
	    name[len] = '\0';
	}
    } else {
	name[0] = '\0';
    }
}


/*
 * --------------------------------------------------
 *
 * RTPSessionOpenSocket
 *
 * Open a socket and bind it to a multicast address
 * Returns a file descriptor or -1 on failure
 *
 * --------------------------------------------------
 */

static int
RTPSessionOpenSocket(struct in_addr addr, short port, RTPSessionSockType type)
{
    int fd, flags, one = 1;
    struct sockaddr_in address;
    struct ip_mreq mr;
    unsigned char loop, ttl;
    
    if (!IN_MULTICAST(ntohl(addr.s_addr))) {
	/* XXX Should this handle rtp over unicast? */
	return -1;
    }
    
    /*
     * Create the socket and set non-blocking I/O
     */
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0) {
	return -1;
    }

    flags = fcntl(fd, F_GETFL, 0);
    flags |= O_NONBLOCK;
    if (fcntl(fd, F_SETFL, flags) < 0) {
	close(fd);
	return -1;
    }

    /*
     * Reuse the address and port if possible
     */
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
		   (char *) &one, sizeof(one) ) < 0) {
	close(fd);
	return -1;
    }

#ifdef SO_REUSEPORT    
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &one, sizeof(one) ) < 0) {
	close(fd);
	return -1;
    }
#endif
    
    if (type == SEND) {
	/*
	 * Connect to the remote address
	 */
	memset(&address, 0, sizeof(address) );
	address.sin_family = AF_INET;
	address.sin_port = htons(port);
	address.sin_addr = addr;
	if (connect(fd, (struct sockaddr *) &address, sizeof(address) ) < 0) {
	    close(fd);
	    return -1;
	}

	/*
	 * Turn of loopback
	 */
	loop = 0;
	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, 1) < 0) {
	    int err = errno;
	    close(fd);
	    return -1;
	}
	
	/*
	 * Set the TTL
	 */
	ttl = 2;
	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, 1) < 0) {
	    int err = errno;
	    close(fd);
	    return -1;
	}
    } else { /* type == RECEIVE */
	/*
	 * Join the multicast group.  It would be nice to be
	 * able to just bind to the class D address here.
	 */
	memset(&address, 0, sizeof(address) );
	address.sin_family = AF_INET;
	address.sin_port = htons(port);
	address.sin_addr.s_addr = INADDR_ANY;
	if (bind(fd, (struct sockaddr *) &address, sizeof(address) ) < 0) {
	    close(fd);
	    return -1;
	}
	
	mr.imr_multiaddr = addr;
	mr.imr_interface.s_addr = INADDR_ANY;
	if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		       (char *) &mr, sizeof(mr)) < 0) {
	    close(fd);
	    return -1;
	}
    }
    
    return fd;
}


/*
 * --------------------------------------------------
 *
 * ReleaseRTPSession
 *
 * Parameters: 
 *	objPtr - pointer to object.
 * 
 * Return: 
 *	Standard Tcl Result
 * 
 * Action: 
 *	Decrements the reference count and destroys
 *	the object if nobody else is using it
 *
 * --------------------------------------------------
 */

int 
ReleaseRTPSession (RTPSession *objPtr, char *name)
{
    RTPSession *session, *prev;
    RTPSenderInfo *sender, *next;
    RTPSessionClient *client, *prevClient;

    /*
     * Remove ourselves from the clients list
     */
    client = objPtr->clients;
    prevClient = NULL;
    while (client != NULL) {
	if (strcmp(client->name, name) == 0) {
	    if (prevClient == NULL) {
		objPtr->clients = client->next;
	    } else {
		prevClient->next = client->next;
	    }
	    break;
	}
	prevClient = client;
	client = client->next;
    }
    
    ckfree(client->name);
    DestroyCMBinding(objPtr->interp, client->cmBindingToken);
    ckfree(client);
    
    /*
     * Decrement the reference count and see if we were
     * the last ones using this object
     */
    objPtr->refcount--;
    if (objPtr->refcount > 0) {
	return;
    }
    
    /*
     * Need to destroy the object
     */
    RTPSessionLeaveGroup(objPtr);

    /*
     * Free the member list and hash table
     */
    Tcl_DeleteHashTable(&(objPtr->memberTab));

    sender = objPtr->memberList;
    while (sender != NULL) {
	next = sender->next;

	if (sender->cname != NULL) {
	    ckfree(sender->cname);
	}
	if (sender->name != NULL) {
	    ckfree(sender->name);
	}

	ckfree(sender);
	sender = next;
    }
    
    /*
     * Remove ourselves from the session list
     */
    session = sessionList;
    prev = NULL;
    while (session != NULL) {
	if (session == objPtr) {
	    if (prev == NULL) {
		sessionList = objPtr->next;
	    } else {
		prev->next = objPtr->next;
	    }
	    break;
	}
	prev = session;
	session = session->next;
    }
    
    ckfree((char *) objPtr);

    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPSessionGetSlot
 *
 * Parameters: 
 *	objPtr - the object to configure
 *	slot - the slot value to retrieve
 *	source - true if a source object is calling,
 *		false if a dest is calling
 * 
 * Return: 
 *	the value for the slot "slot"
 * 
 * Action: 
 *	looks up the given slot and returns its
 *	current value
 *
 * --------------------------------------------------
 */

char *
RTPSessionGetSlot(RTPSession *objPtr, char *slot, int source)
{
    int len;
    static char temp[256];

    len = strlen(slot);
    if (strncmp(slot, "-address", len) == 0) {
	sprintf(temp, "-address {0.0.0.0} {%s}", inet_ntoa(objPtr->address));
    } else if (strncmp(slot, "-port", len) == 0) {
	sprintf(temp, "-port {0} {%d}", objPtr->port);
    } else if (strncmp(slot, "-srcid", len) == 0) {
	sprintf(temp, "-srcid {0} {%dl}", objPtr->srcid);
    } else if (strncmp(slot, "-cname", len) == 0) {
	if (objPtr->cname == NULL) {
	    strcpy(temp, "-cname {} {}");
	} else {
	    sprintf(temp, "-cname {} {%s}", objPtr->cname);
	}
    } else if (strncmp(slot, "-name", len) == 0) {
	if (objPtr->name == NULL) {
	    strcpy(temp, "-name {} {}");
	} else {
	    sprintf(temp, "-name {} {%s}", objPtr->name);
	}
    } else if (strncmp(slot, "-pg", len) == 0) {
	sprintf(temp, "-pg {tk} {%s}", objPtr->pg);
    } else if (source && (strncmp(slot, "-ttl", len) == 0)) {
	sprintf(temp, "-ttl {2} {%d}", objPtr->ttl);
    } else if (!source && (strncmp(slot, "-outCmd", 7) == 0)) {
	char *cname, *end;
	int namelen;
	RTPSenderInfo *sender;
	
	if ((len < 8) || (slot[7] != '(')){
	    return NULL;
	}
	
	end = strchr(slot+8, ')');
	if (end == NULL) {
	    return NULL;
	}
	namelen = end - (slot+8);
	cname = ckalloc(namelen + 1);
	strncpy(cname, slot+8, namelen);
	cname[namelen] = '\0';
	
	sender = RTPSessionFindMemberByCNAME(objPtr, cname);
	ckfree(cname);
	if (sender == NULL) {
	    return NULL;
	}
	
	if (sender->outCmd == NULL) {
	    sprintf(temp, "%s {} {}", slot);
	} else {
	    sprintf(temp, "%s {} {%s}", slot, sender->outCmd);
	}
    } else {
	return NULL;
    }

    return temp;
}


/*
 * --------------------------------------------------
 *
 * RTPSessionSetSlot
 *
 * Parameters: 
 *	objPtr - the object to configure
 *	slot - the slot to set
 *	value - the value to put in the slot
 *	source - true if a source object is calling,
 *		false if a dest is calling
 * 
 * Return: 
 *	Standard Tcl result
 * 
 * Action: 
 *	Sets the value of "slot" to "value".  May
 *	do various other things depending on which slot
 *
 * --------------------------------------------------
 */

int
RTPSessionSetSlot(RTPSession *objPtr, char *slot, char *value, int source)
{
    int len;
    Tcl_Interp *interp = objPtr->interp;
    
    len = strlen(slot);

    if (strncmp(slot, "-cname", len) == 0) {
	if (value == NULL) {
	    objPtr->cname[0] = '\0';
	} else {
	    strcpy(objPtr->cname, value);
	}
    } else if (strncmp(slot, "-name", len) == 0) {
	if (value == NULL) {
	    objPtr->name[0] = '\0';
	} else {
	    strcpy(objPtr->name, value);
	}
    } else if (strncmp(slot, "-pg", len) == 0) {
	if (strcmp(objPtr->pg, value) != 0) {
	    ckfree(objPtr->pg);
	    objPtr->pg = ckalloc(strlen(value) + 1);
	    strcpy(objPtr->pg, value);

	    /*
	     * Change the priority groups of the existing file
	     * handlers.  Don't worry about the at event, as soon
	     * as the current one is called, all subsequent ones
	     * will be created in the new group.
	     */
	    Tcm_DeleteFileHandler(objPtr->dataRecvFd);
	    Tcm_CreateFileHandler(objPtr->pg, objPtr->dataRecvFd,
				  TK_READABLE, RTPSessionHandleData,
				  (ClientData) objPtr);

	    Tcm_DeleteFileHandler(objPtr->ctrlRecvFd);
	    Tcm_CreateFileHandler(objPtr->pg, objPtr->ctrlRecvFd,
				  TK_READABLE, RTPSessionHandleRTCP,
				  (ClientData) objPtr);
	}
    } else if (source && (strncmp(slot, "-ttl", len) == 0)) {
	int newttl;

	if ((Tcl_GetInt(interp, value, &newttl) != TCL_OK)
	    || (newttl < 0) || (newttl > 255)) {

	    return TCL_ERROR;
	}
	
	if (objPtr->ttl != newttl) {
	    objPtr->ttl = newttl;
	    if (setsockopt(objPtr->dataFd, IPPROTO_IP, IP_MULTICAST_TTL,
			   &objPtr->ttl, 1) < 0) {
		Tcl_AppendResult(interp, "Couldn't set ttl: ",
				 strerror(errno), NULL);
		return TCL_ERROR;
	    }
	    if (setsockopt(objPtr->ctrlFd, IPPROTO_IP, IP_MULTICAST_TTL,
			   &objPtr->ttl, 1) < 0) {
		Tcl_AppendResult(interp, "Couldn't set ttl: ",
				 strerror(errno), NULL);
		return TCL_ERROR;
	    }
	}
    } else if (!source && (strncmp(slot, "-outCmd", 7) == 0)) {
	char *cname, *paren;
	int namelen;
	RTPSenderInfo *sender;
	char *objName, *period, *old;
	Tcl_CmdInfo cmdInfo;
	
	if (slot[7] != '(') {
	    Tcl_AppendResult(interp, "Invalid slot: ", slot, NULL);
	    return TCL_ERROR;
	}
	
	paren = strchr(slot + 8, ')');
	if (paren == NULL) {
	    Tcl_AppendResult(interp, "Missing closing paren for: ",
			     slot, NULL);
	    return TCL_ERROR;
	}
	
	namelen = paren - (slot+8);
	cname = ckalloc(namelen + 1);
	strncpy(cname, slot+8, namelen);
	cname[namelen] = '\0';
	
	sender = RTPSessionFindMemberByCNAME(objPtr, cname);
	if (sender == NULL) {
	    Tcl_AppendResult(interp, "No sender with cname: ", cname, NULL);
	    return TCL_ERROR;
	}

	old = sender->outCmd;
	
	sender->outCmd = strdup(value);
	if (value[0] == '@') {
	    if (Tutil_GetRegisteredCommand(interp, sender->outCmd,
					   (ClientData *) &(sender->outObj),
					   (ClientData *) &(sender->outCCmd)) != TCL_OK) {

		Tcl_AppendResult(interp, "Could not find procedure ",
				 value + 1, NULL);
		ckfree(sender->outCmd);
		sender->outCmd = old;
		ckfree(cname);
		return TCL_ERROR;
	    }
	    ckfree(cname);
	    if (old != NULL) {
		ckfree(old);
	    }
	} else {
	    sender->outObj = NULL;
	    sender->outCCmd = NULL;
	}
    } else if ((strncmp(slot, "-address", len) == 0)
	       || (strncmp(slot, "-port", len) == 0)) {
	
	Tcl_AppendResult(interp, "The value for slot ", slot,
			 " cannot be changed", NULL);
	return TCL_ERROR;
    } else {
	Tcl_AppendResult(interp, "No such slot ", slot, NULL);
	return TCL_ERROR;
    }
    
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPSessionGetMembers
 *
 * Returns a list of the CNAMEs of all the listeners
 * on this multicast channel/port.  The list is a
 * string with the CNAMEs separated by spaces
 *
 * --------------------------------------------------
 */
char *
RTPSessionGetMembers(RTPSession *objPtr)
{
    Tcl_HashSearch search;
    Tcl_HashEntry *entry;
    static char *list = NULL;
    static int listlen = 0;
    RTPSenderInfo *sender;
    int len, namelen;
    
    if (list == NULL) {
	listlen = 4096;
	list = ckalloc(listlen);
    }
    
    len = 0;
    list[0] = '\0';

    for (entry = Tcl_FirstHashEntry(&objPtr->memberTab, &search);
	 entry != NULL; entry = Tcl_NextHashEntry(&search)) {
	
	sender = (RTPSenderInfo *) Tcl_GetHashValue(entry);
	namelen = strlen(sender->cname);
	while (len + namelen + 1 > listlen) {
	    listlen *= 2;
	    list = ckrealloc(list, listlen);
	}
	strcat(list + len, sender->cname);
	len += namelen;
	list[len++] = ' ';
	list[len] = '\0';
    }
    
    return list;
}

#ifdef OLD_ACCEPT


/*
 * --------------------------------------------------
 *
 * RTPSessionAccept
 *
 * Accepts incoming buffers and sends them out to
 * the multicast group
 *
 * --------------------------------------------------
 */
int
RTPSessionAccept(RTPSession *objPtr, SB_List *data)
{
    int i, j, count, length, index;
    ScatterBuffer *sb;
    struct iovec *iov;
    int retval = TCL_OK;
    
    if ((data == NULL) || (data->count < 1)) {
	return TCL_OK;
    }

    /*
     * Set up iovec array
     */
    count = 0;
    for (i=0; i<data->count; i++) {
	count += data->sbs[i]->count;
    }
    iov = (struct iovec *) ckalloc(count * sizeof(struct iovec) );

    index = length = 0;
    for (i=0; i<data->count; i++) {
	sb = data->sbs[i];
	for (j=0; j<sb->count; j++) {
	    length += sb->lens[j];
	    iov[index].iov_len = sb->lens[j];
	    iov[index].iov_base = sb->buffers[j]->data + sb->offsets[j];
	    index++;
	}
    }

    /*
     * Update statistics
     */
    objPtr->packets++;
    objPtr->bytes += length;
    
    /*
     * Actually send the packet
     */
    if (writev(objPtr->dataFd, iov, count) < 0) {
	retval = TCL_ERROR;
    }
    
    ckfree(iov);
    return retval;
}

#else


/*
 * --------------------------------------------------
 *
 * RTPSessionAccept
 *
 * Accepts incoming buffers and sends them out to
 * the multicast group
 *
 * --------------------------------------------------
 */
int
RTPSessionAccept(RTPSession *objPtr, SB_List *sbl)
{
    int i;
    ScatterBuffer *sb;
    unsigned char *packet;
    RTPHeader *hdr;
    int len;
    
    if ((sbl == NULL) || (sbl->count < 1)) {
	return TCL_OK;
    }

    for (i=0; i<sbl->count; i++) {
	sb = sbl->sbs[i];
	
	packet = GetFromSB(sb, CM_RTP_PACKET, NULL, &len);
	if (packet == NULL) {
	    continue;
	}
	
	hdr = (RTPHeader *) (packet);
	
	/*
	 * Set version, sequence number, srcid
	 */
	hdr->flags = ((hdr->flags & 0x03fff) | (RTPVersion << 14));
	hdr->seqno = objPtr->seqno++;
	hdr->ssrc = objPtr->srcid;

	/*
	 * Update statistics
	 */
	objPtr->packets++;
	objPtr->bytes += len-8;
	
	/*
	 * Send the packet
	 */
	if (write(objPtr->dataFd, hdr, len-8) < 0) {
	    return TCL_ERROR;
	}
    }

    return TCL_OK;
}

#endif


/*
 * --------------------------------------------------
 *
 * RTPSessionCalculateRTCPInterval
 *
 * Calculates the time, in seconds, until the next
 * RTCP report should be broadcast
 *
 * --------------------------------------------------
 */

static double
RTPSessionCalculateRTCPInterval(RTPSession *objPtr)
{
    double interval, bw;
    int others;
    
    /*
     * Calculate bandwidth allocated to receivers
     */
    bw = objPtr->sessbw * RTCP_RECV_BW_FRACTION;

    others = objPtr->numListeners - objPtr->numSenders;
    interval = (objPtr->rtcpSz * others) / bw;
    interval = max(interval, RTCP_MIN_INTERVAL);

    return (interval * (drand48() + 0.5));
}


/*
 * --------------------------------------------------
 *
 * RTPSessionLeaveGroup
 *
 * Leave the multicast group this object currently
 * belongs to
 *
 * --------------------------------------------------
 */

static int
RTPSessionLeaveGroup(RTPSession *objPtr)
{
    unsigned char packet[MAX_PACKET];
    RTCPHeader *hdr;
    int len;
    struct ip_mreq mr;
    
    Tcm_DeleteFileHandler(objPtr->dataRecvFd);
    Tcm_DeleteFileHandler(objPtr->ctrlRecvFd);
    Tcm_DeleteAtHandler(objPtr->atToken);
    
    /*
     * Send an SDES BYE packet
     */
    hdr = (RTCPHeader *) RTPSessionBuildRR(objPtr, packet);
    hdr->flags = (RTPVersion << 6) | 1;
    hdr->pt = RTCP_BYE;
    hdr->len = (sizeof(RTCPHeader) >> 2) - 1;
    hdr->ssrc = objPtr->srcid;
    
    len = ((unsigned char *) hdr) + sizeof(RTCPHeader) - packet;
    objPtr->rtcpSz += (len - objPtr->rtcpSz) >> 4;
    send(objPtr->ctrlFd, packet, len, 0);
    
    /*
     * Close the sockets
     */
    mr.imr_multiaddr = objPtr->address;
    mr.imr_interface.s_addr = INADDR_ANY;

    setsockopt(objPtr->dataRecvFd, IPPROTO_IP, IP_DROP_MEMBERSHIP,
	       (char *) &mr, sizeof(mr));
    close(objPtr->dataRecvFd);

    setsockopt(objPtr->ctrlRecvFd, IPPROTO_IP, IP_DROP_MEMBERSHIP,
	       (char *) &mr, sizeof(mr));
    close(objPtr->ctrlRecvFd);

    close(objPtr->dataFd);
    close(objPtr->ctrlFd);
    
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPSessionHandleData
 *
 * Registered as a Tk file handler, this routine is
 * called whenever the data socket becomes readable.
 * It reads the incoming packet and handles it.
 *
 * --------------------------------------------------
 */

static void
RTPSessionHandleData(ClientData cd, int mask)
{
    RTPSession *objPtr;
    static unsigned char scratch[MAX_PACKET], *packet;
    int packets;
    struct sockaddr_in from;
    int fromlen = sizeof(from);
    int len, version, delta;
    static const int MAX_MISORDER = 100;
    static const int MAX_DROPOUT = 3000;
    RTPHeader *rh;
    RTPSenderInfo *sender;
    RTPSessionClient *client;
    Buffer *b;
    ScatterBuffer *sb;
    SB_List *sbl;
    char *sbStr;

    objPtr = (RTPSession *) cd;
    
    /*
     * Don't handle more than 5 incoming packets at one time
     */
    packets = 5;
    
    while ((packets-- > 0)
	   && (Tdp_SocketIsReady(objPtr->dataRecvFd) & TCL_FILE_READABLE)) {

	b = MakeBuffer(objPtr->interp, MAX_PACKET, "RTPSessionHandleData");
	if (b == NULL) {
	    /* drop the packet */
	    len = recvfrom(objPtr->dataRecvFd, scratch, MAX_PACKET,
			   0, (struct sockaddr *) &from, &fromlen);
	    return;
	}
	    
	packet = b->data + 8;
	len = recvfrom(objPtr->dataRecvFd, packet, MAX_PACKET,
		       0, (struct sockaddr *) &from, &fromlen);
	if ((len == 0) && (errno != EWOULDBLOCK)) {
	    /*
	     * Something bad happened.  Perhaps it would be best
	     * to delete the file handler and try to signal an
	     * error somehow?
	     */
	    return;
	}
	
	rh = (RTPHeader *) packet;
	version = rtpver(*rh);
	
	if (version != RTPVersion) {
	    objPtr->badversion++;
	    return;
	}
	if (len < sizeof(RTPHeader)) {
	    objPtr->runt++;
	    return;
	}

	/*
	 * Except for the sequence number check, don't do any
	 * other checks at this point.  If the next object along
	 * wants to do more consistency checks, it receives the
	 * whole packet, so it can without any difficulty.
	 */
	
	/*
	 * Find the SenderInfo structure for this
	 * sender and update our stats
	 */
	sender = RTPSessionFindMember(objPtr, rh->ssrc);
	if (sender->inited == 0) {
	    sender->srcid = rh->ssrc;
	    sender->cname = sender->name = NULL;
	    sender->pt = rtppt(*rh);
	    sender->sender = 1;
	    sender->firstseq = sender->lastseq = rh->seqno;
	    sender->seqCycles = 0;
	    sender->dataTs = Tcm_ReadSysClock();
	    sender->ctrlTs = 0.0;
	    sender->ctrlNtp = 0;
	    sender->jitter = 0.0;
	    sender->ackseq = sender->ackpkts = 0;
	    sender->pkts = sender->bytes = 0;
	    sender->outCmd = NULL;
	    
	    sender->inited = 1;
	}
	
	if (sender->sender == 0) {
	    sender->sender = 1;
	    if (sender->cname != NULL) {
		objPtr->outCname = sender->cname;
		objPtr->outName = (sender->name == NULL)
		    ? sender->cname : sender->name;
		objPtr->outEncoding = sender->pt;
		for (client = objPtr->clients;
		     client != NULL; client = client->next) {
		    
		    ExecuteCMBinding(objPtr->interp, client->cmBindingToken,
				     "<NEW_SENDER>");
		}
	    }
	}

	sender->pkts++;
	sender->bytes += len;
	
	if (sender->pt != rtppt(*rh)) {
	    sender->pt = rtppt(*rh);
	    
	    objPtr->outCname = sender->cname;
	    objPtr->outName = (sender->name == NULL)
		? sender->cname : sender->name;
	    objPtr->outEncoding = sender->pt;

	    for (client = objPtr->clients;
		 client != NULL; client = client->next) {

		ExecuteCMBinding(objPtr->interp, client->cmBindingToken,
				 "<ENCODING_CHANGED>");
	    }
	}

	delta = rh->seqno - sender->lastseq;
	if ((delta < -MAX_MISORDER) || (delta > MAX_DROPOUT)) {
	    if (((1 << 16) + delta) < MAX_DROPOUT) {
		/* Sequence number wrapped */
		sender->seqCycles++;
	    } else {
		/*
		 * Sequence number changed a dramatically
		 *
		 * XXX should check to see if it is a restart
		 */
		objPtr->badseqno++;
		continue;
	    }
	    sender->lastseq = rh->seqno;
	} else {
	    /*
	     * Valid packet.  If delta > 0, update lastseq.
	     * If delta < 0, it is probably a duplicate
	     */
	    if (delta > 0) {
		sender->lastseq = rh->seqno;
	    }
	}
	
	if (sender->outCmd != NULL) {
	    /*
	     * Send the packet on
	     */
	    ((int *) b->data)[0] = CM_RTP_PACKET;
	    ((int *) b->data)[1] = len;
	    
	    sb = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));
	    sb->count = 1;
	    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *));
	    sb->buffers[0] = b;
	    sb->lens = (int *) ckalloc(sizeof(int));
	    sb->lens[0] = len+8;
	    sb->offsets = (int *) ckalloc(sizeof(int));
	    sb->offsets[0] = 0;
	    
	    sbl = (SB_List *) ckalloc(sizeof(SB_List));
	    sbl->count = 1;
	    sbl->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *));
	    sbl->sbs[0] = sb;

	    if (sender->outCmd[0] == '@') {
		(sender->outCCmd)(sender->outObj, sbl);
	    } else {
		sbStr = MakeSBListString(sbl);
		Tcl_VarEval(objPtr->interp, sender->outCmd, " ", sbStr, NULL);
		ckfree(sbStr);
	    }
	
	    DetachBuffer(objPtr->interp, b, "RTPSessionHandleData");
	    FreeSBList(sbl);
	}
    }
}


/*
 * --------------------------------------------------
 *
 * RTPSessionFindMember
 *
 * Looks up the sender of an RTP packet in the
 * members table.  If an entry does not already
 * exist, creates one
 *
 * --------------------------------------------------
 */
static RTPSenderInfo *
RTPSessionFindMember(RTPSession *objPtr, unsigned int srcid)
{
    Tcl_HashEntry *entry;
    int newFlag;
    RTPSenderInfo *s;
    
    entry = Tcl_FindHashEntry(&(objPtr->memberTab), (char *) srcid);
    if (entry == NULL) {
	/* XXX check for matching cname first? */
	
	/*
	 * Create a new SenderInfo structure
	 */
	s = (RTPSenderInfo *) ckalloc(sizeof(RTPSenderInfo));
	s->next = objPtr->memberList;
	objPtr->memberList = s;
	s->inited = 0;
	    
	entry = Tcl_CreateHashEntry(&objPtr->memberTab,
				    (char *) srcid, &newFlag);
	Tcl_SetHashValue(entry, (ClientData) s);
	objPtr->numListeners++;
    } else {
	s = (RTPSenderInfo *) Tcl_GetHashValue(entry);
    }
    
    return s;
}


/*
 * --------------------------------------------------
 *
 * RTPSessionFindMemberByCNAME
 *
 * Looks up the sender of an RTP packet in the
 * members table.  If an entry does not already
 * exist, creates one
 *
 * --------------------------------------------------
 */
static RTPSenderInfo *
RTPSessionFindMemberByCNAME(RTPSession *objPtr, char *cname)
{
    Tcl_HashSearch search;
    Tcl_HashEntry *entry;
    RTPSenderInfo *s;

    entry = Tcl_FirstHashEntry(&objPtr->memberTab, &search);
    while (entry != NULL) {
	s = (RTPSenderInfo *) Tcl_GetHashValue(entry);
	if ((s->cname != NULL) && (strcmp(s->cname, cname) == 0)) {
	    return s;
	}
	entry = Tcl_NextHashEntry(&search);
    }

    return NULL;
}


/*
 * --------------------------------------------------
 *
 * RTPSessionRemoveMember
 *
 * Removes a particular member from the list of
 * participants in this conference
 *
 * --------------------------------------------------
 */
static void
RTPSessionRemoveMember(RTPSession *objPtr, unsigned int srcid)
{
    Tcl_HashEntry *entry;
    RTPSenderInfo *sender, *s, *prev;
    RTPSessionClient *client;

    /*
     * Find the SenderInfo structure in the
     * members hash table
     */
    entry = Tcl_FindHashEntry(&objPtr->memberTab, (char *) srcid);
    if (entry == NULL) {
	return;
    }
    
    sender = (RTPSenderInfo *) Tcl_GetHashValue(entry);

    objPtr->outCname = sender->cname;
    objPtr->outName = (sender->name == NULL) ? sender->cname : sender->name;
    objPtr->outEncoding = sender->pt;
    
    for (client = objPtr->clients; client != NULL; client = client->next) {
	ExecuteCMBinding(objPtr->interp, client->cmBindingToken,
			 "<MEMBER_LEFT>");
    }

    objPtr->numListeners--;
    if (sender->sender) {
	objPtr->numSenders--;
    }

    /*
     * Remove the SenderInfo from the members linked list
     */
    for (prev = NULL, s = objPtr->memberList; s != NULL;
	 prev = s, s = s->next) {

	if (s == sender) {
	    if (prev == NULL) {
		objPtr->memberList = s->next;
	    } else {
		prev->next = s->next;
	    }
	    break;
	}
    }

    if (sender->cname != NULL) {
	ckfree(sender->cname);
    }
    if (sender->name != NULL) {
	ckfree(sender->name);
    }
    if (sender->outCmd != NULL) {
	ckfree(sender->outCmd);
    }
    ckfree(sender);

    Tcl_DeleteHashEntry(entry);
}


/*
 * --------------------------------------------------
 *
 * RTPSessionHandleRTCP
 *
 * Registered as a Tk file handler, this routine is
 * called whenever the rtcp socket becomes readable.
 * It reads the incoming rtcp packet and handles it.
 *
 * --------------------------------------------------
 */

static void
RTPSessionHandleRTCP(ClientData cd, int mask)
{
    RTPSession *objPtr;
    int packets, doread;
    unsigned char packet[MAX_PACKET];
    int len, packetlen, parsed = 0;
    struct sockaddr_in from;
    int fromlen = sizeof(from);
    RTCPHeader *hdr;
    RTPSenderInfo *sender;

    objPtr = (RTPSession *) cd;
    
    /*
     * Don't handle more than 5 incoming packets at a time
     */
    packets = 5;
    
    while ((packets-- > 0)
	   && (Tdp_SocketIsReady(objPtr->ctrlRecvFd) & TCL_FILE_READABLE)) {
	
	len = recvfrom(objPtr->ctrlRecvFd, packet, MAX_PACKET,
		       0, (char *) &from, &fromlen);
	if ((len == 0) && (errno != EWOULDBLOCK)) {
	    return;
	}
	
	if (len < sizeof(RTCPHeader)) {
	    objPtr->runt++;
	    return;
	}
	hdr = (RTCPHeader *) packet;
	if (rtcpver(*hdr) != RTPVersion) {
	    objPtr->badversion++;
	    return;
	}
	
	/*
	 * Update average size of RTCP packets we have seen
	 * using a low pass filter
	 */
	objPtr->rtcpSz += ((len - objPtr->rtcpSz) >> 4);
	sender = RTPSessionFindMember(objPtr, hdr->ssrc);
	if (sender->inited == 0) {
	    sender->srcid = hdr->ssrc;
	    sender->cname = sender->name = NULL;
	    sender->pt = 0xff;		/* Invalid PT */
	    sender->sender = 0;
	    sender->firstseq = sender->lastseq = 0;
	    sender->seqCycles = 0;
	    sender->dataTs = 0.0;
	    sender->ctrlTs = Tcm_ReadSysClock();
	    sender->ctrlNtp = GetNTPTime();
	    sender->jitter = 0.0;
	    sender->ackseq = sender->ackpkts = 0;
	    sender->pkts = sender->bytes = 0;
	    sender->outCmd = NULL;
	    
	    sender->inited = 1;
	}

	/*
	 * There may be multiple RTCP packets in this UDP
	 * datagram.  Look at them all
	 */
	while (parsed < len) {
	    packetlen = (hdr->len + 1) << 2;
	    if (len < (parsed + packetlen)) {
		objPtr->runt++;
		return;
	    }
	    
	    switch ((RTCPType) hdr->pt) {
	      case RTCP_SR: case OLD_RTCP_SR:
		/* XXX parse SR */
		break;
	    
	      case RTCP_RR: case OLD_RTCP_RR:
		/* XXX parse RR */
		break;
	    
	      case RTCP_SDES: case OLD_RTCP_SDES:
		RTPSessionParseSDES(objPtr, (unsigned char *) hdr);
		break;
	    
	      case RTCP_BYE: case OLD_RTCP_BYE:
		{
		    int i;
		    unsigned int *srcids;
		    
		    srcids = &(hdr->ssrc);
		    for (i = rtcpsc(*hdr); i > 0; i--) {
			RTPSessionRemoveMember(objPtr, *srcids);
			srcids++;
		    }
		}
		break;
	
	      case RTCP_APP: case OLD_RTCP_APP:
		/* ignore these */
		break;
	    }
	    
	    parsed += packetlen;
	    hdr = (RTCPHeader *) ((unsigned char *) hdr + packetlen);
	}
    }
}


/*
 * --------------------------------------------------
 *
 * RTPSessionParseSDES
 *
 * Handles an SDES packet received on the RTCP
 * control channel
 *
 * --------------------------------------------------
 */

static void
RTPSessionParseSDES(RTPSession *objPtr, unsigned char *data)
{
    RTCPHeader *hdr;
    int count, ssrc;
    char *dataPtr;
    RTCPSDESEntry *se;
    RTPSenderInfo *sender;
    RTCPSDES *s;
    RTPSessionClient *client;
		    
    /*
     * Position dataPtr at first ssrc
     */
    hdr = (RTCPHeader *) data;
    dataPtr = data + sizeof(*hdr) - 4;

    for (count = hdr->flags & 0x1f; count > 0; count--) {
	se = (RTCPSDESEntry *) dataPtr;
	ssrc = se->ssrc;
	sender = RTPSessionFindMember(objPtr, ssrc);

	if (sender->inited == 0) {
	    sender->srcid = ssrc;
	    sender->cname = sender->name = NULL;
	    sender->pt = 0xff;		/* Invalid PT */
	    sender->sender = 0;
	    sender->firstseq = sender->lastseq = 0;
	    sender->seqCycles = 0;
	    sender->dataTs = 0.0;
	    sender->ctrlTs = Tcm_ReadSysClock();
	    sender->ctrlNtp = GetNTPTime();
	    sender->jitter = 0.0;
	    sender->ackseq = sender->ackpkts = 0;
	    sender->pkts = sender->bytes = 0;
	    sender->outCmd = NULL;
			    
	    sender->inited = 1;
	}
			
	s = se->sdes;
	dataPtr += sizeof(se->ssrc);

	while (s->type != SDES_END) {
	    switch (s->type) {
	      case SDES_CNAME:
		/* 
		 * If CNAMEs are different, there is a srcid
		 * collision.  Ignore this packet -- the sender
		 * should choose a new srcid soon.
		 */
		if ((sender->cname != NULL)
		    && (strncmp(sender->cname, dataPtr+2, s->len) == 0)) {
		    break;
		}

		if (sender->cname != NULL) {
		    ckfree(sender->cname);
		}
		sender->cname = ckalloc(s->len + 1);
		strncpy(sender->cname, s->data, s->len);
		sender->cname[s->len] = '\0';

		/*
		 * Invoke the CM Binding
		 */
		objPtr->outCname = sender->cname;
		objPtr->outName = (sender->name == NULL)
		    ? sender->cname : sender->name;
		objPtr->outEncoding = sender->pt;
		for (client = objPtr->clients;
		     client != NULL; client = client->next) {
		    
		    if (sender->sender) {
			ExecuteCMBinding(objPtr->interp,
					 client->cmBindingToken,
					 "<NEW_SENDER>");
		    } else {
			ExecuteCMBinding(objPtr->interp,
					 client->cmBindingToken,
					 "<NEW_LISTENER>");
		    }
		}
		break;

	      case SDES_NAME:
		if ((sender->name != NULL)
		    && (strncmp(sender->name, dataPtr+2, s->len) == 0)) {
		    break;
		}

		if (sender->name != NULL) {
		    ckfree(sender->name);
		}
		sender->name = ckalloc(s->len + 1);
		strncpy(sender->name, s->data, s->len);
		sender->name[s->len] = '\0';
				
		/*
		 * Invoke the CM Binding
		 */
		objPtr->outCname = sender->cname;
		objPtr->outName = sender->name;
		objPtr->outEncoding = sender->pt;
		for (client = objPtr->clients;
		     client != NULL; client = client->next) {
				    
		    ExecuteCMBinding(objPtr->interp, client->cmBindingToken,
				     "<NAME_CHANGED>");
		}
		break;
				
	      case SDES_EMAIL:	case SDES_PHONE:
	      case SDES_LOC:	case SDES_TOOL:
	      case SDES_NOTE:	case SDES_PRIV:
		/* ignore */
	    }

	    dataPtr += s->len + 2;
	    s = (RTCPSDES *) dataPtr;
	}

	if ((int) dataPtr % 4 != 0) {
	    dataPtr += 4 - ((int) dataPtr % 4);
	}
    }
}



/*
 * --------------------------------------------------
 *
 * RTPSessionSendRTCP
 *
 * Registered as an at handler, this routine is
 * called periodically to send out rtcp reciever
 * report (RR) packets.
 *
 * --------------------------------------------------
 */

static void
RTPSessionSendRTCP(ClientData cd)
{
    RTPSession *objPtr;
    static unsigned char packet[MAX_PACKET], *pktPtr, *pktStart;
    RTCPHeader *hdr;
    RTCPSDES *sdes;
    int sdeslen, thislen, numSdes, packetlen;
    double interval, now;
    
    objPtr = (RTPSession *) cd;
    
    /*
     * First, build the RR packet
     */
    pktPtr = RTPSessionBuildRR(objPtr, packet);

    /*
     * Build an SDES packet to include
     * Start with the CNAME
     */
    hdr = (RTCPHeader *) pktPtr;
    
    pktPtr += sizeof(RTCPHeader);
    pktStart = pktPtr;
    sdes = (RTCPSDES *) pktPtr;
    sdes->type = (unsigned char) SDES_CNAME;
    sdes->len = strlen(objPtr->cname);
    strcpy(sdes->data, objPtr->cname);

    sdeslen = sdes->len + 2;
    numSdes = 1;
    pktPtr += sdeslen;

    objPtr->rtcpCt++;
    if (((objPtr->rtcpCt % 4) == 0) && (objPtr->name[0] != '\0')) {
	/*
	 * Every fourth packet, include the NAME
	 */
	sdes = (RTCPSDES *) pktPtr;
	sdes->type = (unsigned char) SDES_NAME;
	sdes->len = strlen(objPtr->name);
	strcpy(sdes->data, objPtr->name);

	thislen = sdes->len + 2;
	sdeslen += thislen;
	pktPtr += thislen;

	numSdes++;
    }


    /*
     * Pad to a word boundary, making sure we have
     * at least one zero byte
     */
    pktPtr += 4 - (sdeslen & 0x3);
    switch (sdeslen&0x3) {
      case 0:
	pktStart[sdeslen++] = '\0';
      case 1:
	pktStart[sdeslen++] = '\0';
      case 2:
	pktStart[sdeslen++] = '\0';
      case 3:
	pktStart[sdeslen++] = '\0';
    }
    
    sdeslen += sizeof(RTCPHeader);

    /* hdr->flags = (RTPVersion << 6) | numSdes; */
    hdr->flags = (RTPVersion << 6) | 1;
    hdr->pt = RTCP_SDES;
    hdr->len = (sdeslen / 4) - 1;
    hdr->ssrc = objPtr->srcid;
    
    /*
     * Update estimate of RTCP packet size
     */
    packetlen = pktPtr - packet;
    objPtr->rtcpSz += (packetlen - objPtr->rtcpSz) >> 4;
    
    /*
     * Actually send the packet
     */
    send(objPtr->ctrlFd, packet, packetlen, 0);
    
    /*
     * Schedule the next RTCP report
     */
    interval = RTPSessionCalculateRTCPInterval(objPtr);
    now = Tcm_ReadSysClock();
    objPtr->atToken = Tcm_CreateAtHandler(now + interval, HUGE_VAL,
					  objPtr->pg, RTPSessionSendRTCP,
					  (ClientData) objPtr, NULL,
					  NULL, NULL, NULL, NULL);
}


/*
 * --------------------------------------------------
 *
 * RTPSessionBuildRR
 *
 * Builds an RTCP RR packet
 *
 * --------------------------------------------------
 */

static char *
RTPSessionBuildRR(RTPSession *objPtr, char *buffer)
{
    RTPRRPkt *rr;
    RTCPHeader *hdr;
    int numRR;
    RTPSenderInfo *member;
    unsigned int sequence, dropped, dropperc;
    int len;

    rr = (RTPRRPkt *) buffer;
    hdr = &(rr->hdr);
    numRR = 0;
    for (member = objPtr->memberList; member != NULL; member = member->next) {
	if (member->sender == 0) {
	    continue;
	}
	
	sequence = (member->seqCycles << 16) | member->lastseq;
	dropped = sequence - member->ackseq + member->ackpkts - member->pkts;
	if (sequence == member->ackseq) {
	    dropperc = 0;
	} else {
	    dropperc = (dropped << 8) / (sequence - member->ackseq);
     	    if (dropperc > 255) {
		dropperc = 255;
	    } else if (dropperc <= 0) {
		dropperc = 0;
	    }
	}
	
	rr->rr[numRR].ssrc = htonl(member->srcid);
	rr->rr[numRR].lost = htonl(((unsigned char) dropperc) << 24 | (dropped & 0xfff));
	rr->rr[numRR].seqno = htonl(sequence);
	rr->rr[numRR].jitter = htonl((int) (member->jitter * 65536.));
	rr->rr[numRR].lsr = htonl(member->ctrlNtp);
	rr->rr[numRR].dlsr = htonl(65536.
				   * (Tcm_ReadSysClock() - member->ctrlTs));
	
	numRR++;
	if (numRR >= 31) {
	    break;
	}
	
	member->ackseq = sequence;
    }

    hdr->ssrc = objPtr->srcid;
    hdr->flags = (RTPVersion << 6) | numRR;
    hdr->pt = RTCP_RR;
    len = sizeof(RTCPHeader) + (numRR * sizeof(RTCPRecv));
    hdr->len = (len>>2) - 1;
    
    return (buffer + len);
}


/*
 * --------------------------------------------------
 *
 * RTPSessionCNAME, RTPSessionNAME
 *
 * Callbacks for cmbind mechanism for %c and %n escapes
 *
 * --------------------------------------------------
 */

static char *
RTPSessionCNAME(Tcl_Interp *interp, ClientData cd)
{
    RTPSession *objPtr;
    char *cname;
    
    objPtr = (RTPSession *) cd;
    cname = ckalloc(strlen(objPtr->outCname) + 1);
    strcpy(cname, objPtr->outCname);
    return cname;
}

static char *
RTPSessionNAME(Tcl_Interp *interp, ClientData cd)
{
    RTPSession *objPtr;
    char *name;
    
    objPtr = (RTPSession *) cd;
    name = ckalloc(strlen(objPtr->outName) + 1);
    strcpy(name, objPtr->outName);
    return name;
}

static char *
RTPSessionEncoding(Tcl_Interp *interp, ClientData cd)
{
    RTPSession *objPtr;
    char *encoding;
    int pt;

    objPtr = (RTPSession *) cd;
    encoding = ckalloc(20);
    pt = objPtr->outEncoding;
    sprintf(encoding, "%d", pt);

    return encoding;
}


/*
 * --------------------------------------------------
 *
 * GetNTPTime
 *
 * Returns the middle 32 bits of the current NTP
 * time.  See RTP spec, section 4
 *
 * --------------------------------------------------
 */
static unsigned int
GetNTPTime()
{
    struct timeval tv;
    unsigned int ntptime;
    
    gettimeofday(&tv, NULL);
    ntptime = (tv.tv_sec << 16) || (((long) tv.tv_usec) << 16 / 1000000);
    return ntptime;
}

