/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */



/*
 * --------------------------------------------------
 * 
 * cmRTPSource.c
 *
 * Defines the "rtpSource" object.  The rtpSource
 * object does very little by itself; it provides Tcl
 * level access to the sending capabilities of the
 * rtp session object defined in cmRTPSession.c.
 *
 * $Id: cmRTPSource.c,v 1.5 1996/02/06 22:42:34 kpatel Exp $
 *
 * --------------------------------------------------
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "tcl.h"

#include "cmBufferlist.h"
#include "cmRTPSession.h"

#ifndef abs
#  define abs(a)	((a)>0?(a):-(a))
#endif
#define min(a,b)	((a)>(b)?(b):(a))
#define max(a,b)	((a)>(b)?(a):(b))

/*
 * An actual RTPSource object
 */
typedef struct RTPSource {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */

    struct in_addr addr;	/* The multicast group we are listening to */
    short port;			/* Port number for data stream */

    RTPSession *session;	/* The RTPSession object */
} RTPSource;

/* Globals used by this object module. */
static int nextRTPSourceNum = 0;   /* Variable for next unique object name. */

/*
 * Prototypes for static routines
 */
static int RTPSourceCreateCmd (ClientData clientData, Tcl_Interp *interp,
			     int argc, char **argv);
static RTPSource *CreateNewRTPSource(Tcl_Interp *interp, char *name);
static int RTPSourceObjCmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char **argv);
static int DestroyRTPSource(RTPSource *objPtr);
static int ConfigureRTPSource(RTPSource *objPtr, int argc, char **argv);
static int RTPSourceInFunc(RTPSource *objPtr, SB_List *data);


/*
 * --------------------------------------------------
 *
 * RTPSourceModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     keywords as well as creating the Tcl command
 *     for object type creation
 *
 * --------------------------------------------------
 */

int
RTPSourceModuleInit(Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp, "rtpsource", RTPSourceCreateCmd,
		      NULL, (void (*)()) NULL);
    Tcl_VarEval(interp, "cminfo set object_info rtpSource ",
		"{multicast network source}", (char *) NULL);

    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPSourceCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the RTPSource command.  A new
 *     object of type RTPSource is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is
 *     bound to the object name to implement the objects
 *     methods.
 *
 * --------------------------------------------------
 */

static int
RTPSourceCreateCmd (ClientData clientData, Tcl_Interp *interp,
		  int argc, char **argv)
{
    char new_name[256];
    RTPSource *new;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent", (char *) NULL);
	return TCL_ERROR;
    }

    nextRTPSourceNum++;
    sprintf(new_name, "%srtpsrc%d", argv[1], nextRTPSourceNum);

    new = CreateNewRTPSource(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, new_name, RTPSourceObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * CreateNewRTPSource
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 * --------------------------------------------------
 */

static RTPSource *
CreateNewRTPSource(Tcl_Interp *interp, char *name)
{
    RTPSource *new;
    char *atomName, suffix[] = "accept";

    /*
     * Create the new object
     */
    new = (RTPSource *) ckalloc(sizeof(RTPSource));
    new->interp = interp;

    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    new->addr.s_addr = 0;
    new->port = 0;
    new->session = NULL;
    
    atomName = ckalloc(strlen(name) + strlen(suffix) + 3);
    sprintf(atomName, "%s.%s", name, suffix);
    Tutil_RegisterCommand(interp, atomName, (ClientData) RTPSourceInFunc,
			  (ClientData) new);
    
    return new;
}


/*
 * --------------------------------------------------
 *
 * RTPSourceObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     RTPSource object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 * --------------------------------------------------
 */
static int
RTPSourceObjCmd(ClientData clientData, Tcl_Interp *interp,
	      int argc, char **argv)
{
    char c;
    int len;
    RTPSource *objPtr = (RTPSource *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 "<command> ?arguments?", (char *) NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyRTPSource(clientData);
    } else if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureRTPSource(clientData, argc-2, argv+2);
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0],
			argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                (char *) NULL);
    return TCL_ERROR;
}


/*
 * --------------------------------------------------
 *
 * DestroyRTPSource
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 * --------------------------------------------------
 */
static int 
DestroyRTPSource(RTPSource *objPtr)
{
    /*
     * Delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);
    
    if (objPtr->session != NULL) {
	ReleaseRTPSession(objPtr->session, objPtr->name);
    }

    ckfree((char *) objPtr);
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * ConfigureRTPSource
 *
 * Parameters: 
 *	objPtr - the object to configure
 *	argc - argument count
 *	argv - argument values
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *	Configures the packet dest object
 *
 * --------------------------------------------------
 */

static int
ConfigureRTPSource(RTPSource *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr, len;
    char *value, temp[16];
    
    interp = objPtr->interp;
    
    if (argc == 0) {
	char **slot;
	if (objPtr->session == NULL) {
	    sprintf(temp, "%d", objPtr->port);
	    Tcl_AppendResult(interp, "{-address {0.0.0.0} {",
			     inet_ntoa(objPtr->addr), "}} -port {0} {",
			     temp, "}}", NULL);
	} else {
	    for (slot = RTPSessionSourceSlots; *slot != NULL; slot++) {
		value = RTPSessionGetSlot(objPtr->session, *slot, 1);
		Tcl_AppendResult(interp, "{", value, "} ", NULL);
	    }
	}
	
	return TCL_OK;
    } 
    
    if (argc == 1) {
	if (objPtr->session == NULL) {
	    len = strlen(argv[0]);
	    if (strncmp("-address", argv[0], len) == 0) {
		Tcl_AppendResult(interp, "-address {0.0.0.0} {",
				 inet_ntoa(objPtr->addr), "}", NULL);
		return TCL_OK;
	    } else if (strncmp("-port", argv[0], len) == 0) {
		sprintf(temp, "%d", objPtr->port);
		Tcl_AppendResult(interp, "-port {0} {", temp, "}", NULL);
		return TCL_OK;
	    } else {
		Tcl_AppendResult(interp, "Bad configuration option: ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }
	} else {
	    value = RTPSessionGetSlot(objPtr->session, argv[0], 1);

	    if (value != NULL) {
		Tcl_AppendResult(interp, value, NULL);
		return TCL_OK;
	    } else {
		Tcl_AppendResult(interp, "Bad configuration option: ",
				 argv[0], NULL);
		return TCL_ERROR;
	    }
	}
    }
    
    cntr = 0;
    while (cntr < argc) {
	len = strlen(argv[cntr]);
	
	if ((cntr + 1) >= argc) {
	    Tcl_AppendResult(interp,
			     "Missing value for option: ", argv[cntr], NULL);
	    return TCL_ERROR;
	}
	
	if (strncmp("-address", argv[cntr], len) == 0) {
	    long oldaddr = objPtr->addr.s_addr;

	    objPtr->addr.s_addr = inet_addr(argv[cntr + 1]);
	    if (objPtr->addr.s_addr == 0) {
		Tcl_AppendResult(objPtr->interp, "Bad value for -address: ",
				 argv[cntr+1], NULL);
		return TCL_ERROR;
	    }

	    if ((objPtr->port != 0) && (objPtr->addr.s_addr != oldaddr)) {
		if (objPtr->session != NULL) {
		    ReleaseRTPSession(objPtr->session, objPtr->name);
		    objPtr->session = NULL;
		}
		
		objPtr->session = GetRTPSession(objPtr->interp, objPtr->name,
						&objPtr->addr, objPtr->port);
		if (objPtr->session == NULL) {
		    objPtr->addr.s_addr = 0;
		    Tcl_AppendResult(interp,
				     "Error creating new session", NULL);
		    return TCL_ERROR;
		}
	    }
	} else if (strncmp("-port", argv[cntr], len) == 0) {
	    short oldport = objPtr->port;
	    if (sscanf(argv[cntr + 1], "%hd", &objPtr->port) != 1) {
		Tcl_AppendResult(objPtr->interp, "Bad value for -port: ",
				 argv[cntr + 1], NULL);
		return TCL_ERROR;
	    }
	    
	    if ((objPtr->addr.s_addr != 0) && (objPtr->port != oldport)) {
		if (objPtr->session != NULL) {
		    ReleaseRTPSession(objPtr->session, objPtr->name);
		    objPtr->session = NULL;
		}
		
		objPtr->session = GetRTPSession(objPtr->interp, objPtr->name,
						&objPtr->addr, objPtr->port);
		if (objPtr->session == NULL) {
		    objPtr->port = 0;
		    Tcl_AppendResult(interp,
				     "Error creating new session", NULL);
		    return TCL_ERROR;
		}
	    }
	} else if (objPtr->session == NULL) {
	    Tcl_AppendResult(interp, "Must set -address or -port first!", NULL);
	    return TCL_ERROR;
	} else {
	    if (RTPSessionSetSlot(objPtr->session, argv[cntr],
				  argv[cntr+1], 1) != TCL_OK) {
		return TCL_ERROR;
	    }
	}
	
	cntr += 2;
    }
    return TCL_OK;
}


/*
 * --------------------------------------------------
 *
 * RTPSourceInFunc
 *
 * Registered as the InFunc for every RTPSource
 * object.  Accepts incoming scatter buffer lists
 * and hands them off to the appropriate session
 * object
 *
 * --------------------------------------------------
 */

static int
RTPSourceInFunc(RTPSource *objPtr, SB_List *data)
{
    if (objPtr->session != NULL) {
	return RTPSessionAccept(objPtr->session, data);
    }
    
    return TCL_OK;
}
