/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * slot.c --
 *
 *	This file contains the Tcm_SlotValue procedure.
 *      It was hacked from tcmConfig.c.
 *
 */

#include "tkConfig.h"
#include "tk.h"

#include "cmConfig.h"

/*
 * Values for "flags" field of Tcm_ConfigSpec structures.  Be sure
 * to coordinate these values with those defined in tcmConfig.h.
 * There must not be overlap!
 *
 * INIT -		Non-zero means (char *) things have been
 *			converted to Tk_Uid's.
 */

#define INIT		0x20

/*
 * Forward declarations for procedures defined later in this file:
 */

static Tcm_ConfigSpec *FindSlotSpec	_ANSI_ARGS_((Tcl_Interp *interp,
						     Tcm_ConfigSpec *specs,
						     char *argvName,
						     int needFlags,
						     int hateFlags));

static char *FormatSlotValue		_ANSI_ARGS_((Tcm_ConfigSpec *specPtr,
						     char *objRec));

/*
 *--------------------------------------------------------------
 *
 * FindSlotSpec --
 *
 *	Search through a table of configuration specs, looking for
 *	one that matches a given argvName.
 *
 * Results:
 *	The return value is a pointer to the matching entry, or NULL
 *	if nothing matched.  In that case an error message is left
 *	in interp->result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static Tcm_ConfigSpec *
FindSlotSpec(Tcl_Interp *interp, Tcm_ConfigSpec *specs, char *slotName, int needFlags, int hateFlags)
                       		/* Used for reporting errors. */
                          	/* Pointer to table of configuration
				 * specifications for an object. */
                   		/* Name (suitable for use in a "config"
				 * command) identifying particular option. */
                  		/* Flags that must be present in matching
				 * entry. */
                  		/* Flags that must NOT be present in matching
				 * entry. */
{
    register Tcm_ConfigSpec *specPtr;
    register char c;		/* First character of current argument. */
    Tcm_ConfigSpec *matchPtr;	/* Matching spec, or NULL. */
    int length;

    char *argvName;

    argvName = (char *) ckalloc(strlen(slotName) + 2);

    /****************************************************************/
    /*
     * Prepend '-' to slotName if necessary.
     */

    if (slotName && (slotName[0] == '-'))
	strcpy(argvName, slotName);

    else {
	strcpy(argvName + 1, slotName);
	argvName[0] = '-';
    }

    /****************************************************************/

    c = argvName[1];
    length = strlen(argvName);
    matchPtr = NULL;
    for (specPtr = specs; specPtr->type != TCM_CONFIG_END; specPtr++) {
	if (specPtr->argvName == NULL) {
	    continue;
	}
	if ((specPtr->argvName[1] != c)
	    || (strncmp(specPtr->argvName, argvName, length) != 0)) {
	    continue;
	}
	if (((specPtr->specFlags & needFlags) != needFlags)
	    || (specPtr->specFlags & hateFlags)) {
	    continue;
	}
	if (specPtr->argvName[length] == 0) {
	    matchPtr = specPtr;
	    goto gotMatch;
	}
	if (matchPtr != NULL) {
	    Tcl_AppendResult(interp, "ambiguous option \"", argvName,
			     "\"", (char *) NULL);
	    ckfree(argvName);
	    return (Tcm_ConfigSpec *) NULL;
	}
	matchPtr = specPtr;
    }

    if (matchPtr == NULL) {
	Tcl_AppendResult(interp, "unknown option \"", argvName,
			 "\"", (char *) NULL);
	ckfree(argvName);
	return (Tcm_ConfigSpec *) NULL;
    }
    /*
     * Found a matching entry.  If it's a synonym, then find the entry that
     * it's a synonym for.
     */

gotMatch:
    specPtr = matchPtr;
    if (specPtr->type == TCM_CONFIG_SYNONYM) {
	for (specPtr = specs;; specPtr++) {
	    if (specPtr->type == TCM_CONFIG_END) {
		Tcl_AppendResult(interp,
				 "couldn't find synonym for option \"",
				 argvName, "\"", (char *) NULL);
		ckfree(argvName);
		return (Tcm_ConfigSpec *) NULL;
	    }
	    if ((specPtr->type != TCM_CONFIG_SYNONYM)
		&& ((specPtr->specFlags & needFlags) == needFlags)
		&& !(specPtr->specFlags & hateFlags)) {
		break;
	    }
	}
    }
    ckfree(argvName);
    return specPtr;
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_SlotValue --
 *
 *	Return the current value of a slot of an object.
 *
 * Results:
 *	Interp->result will be modified hold the slot value
 *	of the object, by matching objRec and specs.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

int
Tcm_SlotValue(Tcl_Interp *interp, Tcm_ConfigSpec *specs, char *objRec, char *slotName, int flags)
                       		/* Interpreter for error reporting. */
                          	/* Describes legal slots. */
                 		/* Record whose fields contain current values
				 * for slots. */
                   		/* The slot name. */
              			/* Used to specify additional flags that must
				 * be present in config specs for them to be
				 * considered. */
{
    register Tcm_ConfigSpec *specPtr;

    int needFlags, hateFlags;

    needFlags = flags & ~(TCM_CONFIG_USER_BIT - 1);
    hateFlags = 0;

    Tcl_SetResult(interp, (char *) NULL, TCL_STATIC);

    if (slotName) {
	specPtr = FindSlotSpec(interp, specs, slotName, needFlags,
			       hateFlags);
	if (!specPtr)
	    return TCL_ERROR;

	interp->result = FormatSlotValue(specPtr, objRec);
	interp->freeProc = (Tcl_FreeProc *) free;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * FormatSlotValue --
 *
 *	Create a valid Tcl value holding the slot value.
 *
 * Results:
 *	The slot's value, dynamically allocated.  The caller is
 *	expected to arrange for this value to be freed eventually.
 *
 * Side effects:
 *	Memory is allocated
 *
 *--------------------------------------------------------------
 */

static char *
FormatSlotValue(register Tcm_ConfigSpec *specPtr, char *objRec)
                                     	/* Pointer to information describing
					 * slot. */
                 		/* Pointer to record holding current values
				 * of info for object. */
{
    char *value, *ptr, *result;
    char buffer[200];

    ptr = objRec + specPtr->offset;
    value = "";

    switch (specPtr->type) {
    case TCM_CONFIG_BOOLEAN:
	if (*((int *) ptr) == 0) {
	    value = "false";
	} else {
	    value = "true";
	}
	break;
    case TCM_CONFIG_INT:
	sprintf(buffer, "%d", *((int *) ptr));
	value = buffer;
	break;
    case TCM_CONFIG_DOUBLE:
	sprintf(buffer, "%lf", *((double *) ptr));
	value = buffer;
	break;
    case TCM_CONFIG_STRING:
	value = (*(char **) ptr);
	break;
    case TCM_CONFIG_UID:{
	    Tk_Uid uid = *((Tk_Uid *) ptr);
	    if (uid != NULL) {
		value = uid;
	    }
	    break;
	}
    default:
	value = "?? unknown type ??";
    }

    result = NULL;

    if (value) {
	result = (char *) ckalloc(strlen(value) + 1);
	strcpy(result, value);
    } else {
	result = (char *) ckalloc(sizeof(char));
	result[0] = '\0';
    }

    return result;
}
