/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <math.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>
#include "cmAuPlay.h"

#ifdef SPARC_AU

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/filio.h>
#include <stropts.h>
#include "cmAuFile.h"
#include "libaudio.h"
#include "audio_device.h"
#include "audio_hdr.h"
#include "cmLTS.h"

static int fd = -1;	    	/* File descriptor for /dev/audio */
static Audio_hdr devHdr;	/* Stores current play state of audio device */

static int stopToken = 0;	/* Token for atEvent to silence the
				 * device when playback queue empties */
static int atToken = 0;		/* Token for atEvent for playback when
				 * audio device queue is empty */
static double lastStartTime=0;	/* Last time the SparcAuStartPlay()
				 * function was called. */
static double emptyTime;	/* System time when the data currently
				 * in the stream queue should flush */
static int errorCode;		/* Current error code */

/*
 * Interface routines
 */
static int SparcAuDeviceInit		(SpeakerRsrc *objPtr);
static int SparcAuDeviceReinit		(SpeakerRsrc *objPtr);
static int SparcAuDeviceShutdown	(SpeakerRsrc *objPtr);
static int SparcAuDeviceUnready		(SpeakerRsrc *objPtr);
static int SparcAuDeviceLTSChanged	(SpeakerRsrc *objPtr);
static int SparcAuDevicePlay		(CBInfo *cbPtr);
static char *SparcAuDeviceError		(SpeakerRsrc *objPtr, int errNum);
SpeakerDevice *GetSparcDevice		(SpeakerRsrc *objPtr);

/*
 * Internal Routines
 */
static void SparcAuSilence		(void);
static void SparcAuStartPlay 		(CBInfo *cbPtr);
static void SparcAuDeviceSetGain	(SpeakerRsrc *objPtr);
static void SparcAuDeviceSetOutput	(SpeakerRsrc *objPtr);
static int SparcAuDeviceSetPlayParams	(int sample_rate, int channels,
					 int encoding);

/* a safety factor when forcing restarts and syncing the clock */
#define SKEW		0.005

#define SYNC_TIME	10.0

#define IN_USE_ERROR	1
#define OPEN_ERROR	2
#define FORMAT_ERROR	3
#define TOO_EARLY_ERROR	4
#define NOT_OPEN_ERROR	5
#ifndef min
#  define min(a,b)	((a)<(b)?(a):(b))
#endif


/*
 * -------------------------------------------------------
 *
 * GetSparcDevice
 *
 *	Allocates a new SpeakerDevice structure, fills
 *	in to handle Sparc Au and return it.
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetSparcDevice(SpeakerRsrc *objPtr)
{
    SpeakerDevice *sdev;
    
    sdev = (SpeakerDevice *) ckalloc(sizeof(SpeakerDevice));
    sdev->deviceData = NULL;
    sdev->initProc = SparcAuDeviceInit;
    sdev->reinitProc = SparcAuDeviceReinit;
    sdev->shutdownProc = SparcAuDeviceShutdown;
    sdev->unreadyProc = SparcAuDeviceUnready;
    sdev->ltsProc = SparcAuDeviceLTSChanged;
    sdev->playProc = SparcAuDevicePlay;
    sdev->errorProc = SparcAuDeviceError;
    errorCode = 0;

    return sdev;
}

/*
 * -------------------------------------------------------
 *
 * SparcAuSilence
 *
 * Flush any pending output (i.e., silence the speaker)
 *
 * -------------------------------------------------------
 */

static void
SparcAuSilence()
{
    if (fd > 0) {
	audio_flush_play(fd);
    }
}


/*
 * -------------------------------------------------------
 *
 * SparcAuDeviceInit
 *
 * Initialize this device.
 *
 * -------------------------------------------------------
 */

static int
SparcAuDeviceInit(SpeakerRsrc *objPtr)
{
    int nonblock = 1;
    
    if (fd > 0) {
	return IN_USE_ERROR;
    }

    fd = open("/dev/audio", (O_WRONLY | O_NDELAY), 0);
    if (fd < 0) {
	fd = -1;
	return OPEN_ERROR;
    }
    /* Set to non-blocking I/O */
    ioctl(fd, FIONBIO, &nonblock); 

    audio_get_play_config (fd, &devHdr);
    SparcAuDeviceSetGain(objPtr);
    SparcAuDeviceSetOutput(objPtr);
    errorCode = 0;
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * StopTokenCallback --
 *
 *	Helper function for ResetStopToken
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
void
StopTokenCallback (ClientData clientData)
{
    SparcAuSilence();
    stopToken = 0;
}

/*
 *--------------------------------------------------------------
 *
 * ResetStopToken --
 *
 *	Called whenever the emptyTime variable changes.  Resets
 *	the at event associated with the queue becoming empty.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	When the system clock reaches the value specified by
 *	the global variable emptyTime, the speaker will be
 *      silenced.  This is necessary because the speaker otherwise
 *	may get left in a state where it's producing a high
 *	pitched whine.
 *
 *--------------------------------------------------------------
 */
static void
ResetStopToken()
{
    if (stopToken != 0) {
	Tcm_DeleteAtHandler(stopToken);
	stopToken = 0;
    }
    if (emptyTime != 0.0) {
	stopToken = Tcm_CreateAtHandler(emptyTime, HUGE_VAL,
			"audio", (Tk_TimerProc *)StopTokenCallback,
			(ClientData) NULL, NULL,
			NULL, (Tk_IdleProc *)NULL, (ClientData)NULL, NULL);
    }
}

/*
 * -------------------------------------------------------
 *
 * SparcAuDeviceShutdown
 *
 * Destroys the memory that has been allocated for a
 * given SparcAuDevice
 *
 * -------------------------------------------------------
 */

static int
SparcAuDeviceShutdown(SpeakerRsrc *objPtr)
{
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	emptyTime = 0;
	atToken = 0;
    }
    if (stopToken != 0) {
	Tcm_DeleteAtHandler(stopToken);
	stopToken = 0;
    }
    SparcAuSilence();
    if (fd > 0) {
	close(fd);
	fd = -1;
    }
    errorCode = 0;
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * SparcAuDeviceUnready
 *
 *	Stub routine
 *
 * -------------------------------------------------------
 */

static int
SparcAuDeviceUnready(SpeakerRsrc *objPtr)
{ 
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	emptyTime = 0;
	atToken = 0;
    }
    if (stopToken != 0) {
	Tcm_DeleteAtHandler(stopToken);
	stopToken = 0;
    }
    SparcAuSilence();
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * SparcAuDeviceReinit
 *
 *	Stub routine
 *
 * -------------------------------------------------------
 */
static int
SparcAuDeviceReinit(SpeakerRsrc *objPtr)
{ 

    SparcAuDeviceSetGain(objPtr);
    SparcAuDeviceSetOutput(objPtr);
    errorCode = 0;
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * SparcAuDeviceSetSampleRate
 *
 *	Set the sample rate for audio output.  Returns 1
 *	if successfully set, 0 otherwise
 *
 * Parameters:
 *	int sample_rate;         samples per second
 *	int channels;            # of interleaved channels
 *	int encoding;            data encoding format
 *
 *  NOTE: if any parameter is -1, the value is left unchanged.
 *
 * -------------------------------------------------------
 */
static int
SparcAuDeviceSetPlayParams (int sample_rate, int channels, int encoding)
{
    Audio_hdr hdr;
    int err;

    /*
     * If not initiatilzed, return
     */
    if (fd == -1) {
	return 1;
    }

    /*
     * Create structure that we're trying to set
     */
    memcpy (&hdr, &devHdr, sizeof(Audio_hdr));
    if (sample_rate != -1) {
	hdr.sample_rate = sample_rate;
    }
    if ((encoding == SUN_ULAW) || (encoding == SUN_LIN_8)) {
	hdr.bytes_per_unit = 1;
    } else if (encoding == SUN_LIN_16) {
	hdr.bytes_per_unit = 2;
    } else {
	return 0;
    }
    if (channels != -1) {
	hdr.channels = channels;
    }
    if (encoding != -1) {
	hdr.encoding = encoding;
    }

    /*
     * If no change, return
     */
    if (audio_cmp_hdr(&hdr, &devHdr) == 0) {
	return 1;
    }

    /*
     * Set it...
     */
    err = audio_set_play_config(fd, &hdr);

    /*
     * Update current state with acutal state of the device.
     */
    audio_get_play_config (fd, &devHdr);

    if (err == AUDIO_SUCCESS) {
	return 1;
    }
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * SparcAuDeviceSetGain
 *
 *	Set the gain of the audio device
 *
 * -------------------------------------------------------
 */

static void
SparcAuDeviceSetGain (SpeakerRsrc *objPtr)
{
    if ((objPtr != NULL) && (fd != -1)) {
	audio_set_play_gain(fd, &objPtr->gain);
    }
}


/*
 * -------------------------------------------------------
 *
 * SparcAuDeviceSetOutput
 *
 *	Direct the output of the audio device.
 *
 * -------------------------------------------------------
 */

static void
SparcAuDeviceSetOutput (SpeakerRsrc *objPtr)
{
    int port;

    if ((objPtr != NULL) && (fd != -1)) {
	if (objPtr->headphone) {
	    port = AUDIO_HEADPHONE;
	} else {
	    port = AUDIO_SPEAKER;
	}
	audio_set_play_port (fd, &port);
    }
}


/*
 * -------------------------------------------------------
 *
 * SparcAuDeviceLTSChanged
 *
 *	Notify the SparcAuDevice that the LTS has changed
 *
 * -------------------------------------------------------
 */

static int
SparcAuDeviceLTSChanged (SpeakerRsrc *objPtr)
{
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	emptyTime = 0;
	atToken = 0;
    }
    if (stopToken != 0) {
	Tcm_DeleteAtHandler(stopToken);
	stopToken = 0;
    }
    SparcAuSilence();
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * SparcAuDevicePlay
 *
 * Parameters: 
 *	cbPtr - Frame to be played
 * 
 * Return: 
 *	None
 * 
 * Action: 
 *	Check to make sure the AC is set up properly,
 *	then gives the data to AF to be played.
 *
 * -------------------------------------------------------
 */
static int
SparcAuDevicePlay(CBInfo *cbPtr)
{
    double now;
    int forceRestart;
    SpeakerRsrc *objPtr;
    
    if (fd < 0) {
	return NOT_OPEN_ERROR;
    }

    /*
     * Check for earlier failures that need special recovery (e.g.
     * unplayable format)
     */
    if (errorCode != 0) {
	return errorCode;
    }

    if (atToken != 0) {
	return TOO_EARLY_ERROR;
    }

    /*
     * The following code s tricky because it attempts to keep the
     * device queue full.  Since there's no way to directly query
     * the state of the device queue, we use the following strategy.
     * The Rsrc will normally call us a little while (sendAhead, about
     * 200 milliseconds) before we're ready to play.  There are 2 cases:
     *  1. if the device queue has data in it, just write the data
     *	   to the device.
     *	2. if the device queue is empty, wait until the real start
     *	   time to write the data.
     *  3. if the device parameters have changed (e.g., a different
     *	   data format), wait until the real start time to reset the
     *	   parameters and then write the data.
     *  4. If we've been writing data continuously in our attempts
     *	   to keep the queue full for a long time (SYNC_TIME), the
     *	   synchronization might have drifted.  So, wait until the
     *	   real start time, flush any remaining data, and then write
     *	   the data.
     *
     * Actions 2-4 are performed by the SparcAuStartPlay() funciton
     * below, since they are fired by an at event.  They care signaled
     * by setting the forceRestart variable to 1.
     *
     * In any case, the emptyTime variable is set to when the
     * queue will become empty again.
     */
    forceRestart = 0;

    /*
     * Check to see if the format has changed.  If so, wait for
     * device to drain and force restart.
     */
    if ((cbPtr->hdr.samplingRate != devHdr.sample_rate) ||
	(cbPtr->hdr.channelCount != devHdr.channels) ||
	(cbPtr->hdr.dataFormat != devHdr.encoding)) {
	forceRestart = 1;
    }

    /*
     * Check to see if it's been awhile (SYNC_TIME) since we did
     * the last restart.  If so, force a restart to keep synchronization
     * from drifting
     */
    objPtr = cbPtr->objPtr;
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));
    if (now > lastStartTime + SYNC_TIME) {
	forceRestart = 1;
    }

    /*
     * If we think the device queue is empty (now > emptyTime), force
     * a restart so we start as close to play time as possible.
     */
    if (now > emptyTime) {
	forceRestart = 1;
    }

    /*
     * Play it directly or wait.
     */
    if (forceRestart) {
	Tk_Preserve(cbPtr);
	AttachBuffer (objPtr->interp, cbPtr->buffer,
			"SparcAuDevice:SparcAuDevicePlay");
	
	atToken = Tcm_CreateAtHandler(cbPtr->start-SKEW, HUGE_VAL,
			    "audio", (Tk_TimerProc *)SparcAuStartPlay,
			    (ClientData) cbPtr, NULL,
			    NULL, (Tk_IdleProc *)NULL, (ClientData)NULL, NULL);

    } else {
	write(fd, cbPtr->data, cbPtr->len);
	emptyTime += (cbPtr->end - cbPtr->start);
	ResetStopToken();
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * SparcAuStartPlay --
 *
 *	Starts off the initial playback to the device.  Only
 *	called when teh device queu is thought to be empty.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Writes data to the device and updates emptyTime
 *
 *--------------------------------------------------------------
 */
static void
SparcAuStartPlay (CBInfo *cbPtr)
{
    double now, fraction;
    int offset, n, width;
    SpeakerRsrc *objPtr;
    
    atToken = 0;
    objPtr = cbPtr->objPtr;
    DetachBuffer (objPtr->interp, cbPtr->buffer,
		  "SparcAuDevice:SparcAuStartPlay");
    if (fd < 0) {
	Tk_Release(cbPtr);
	return;
    }

    /*
     * After either action, the emptyTime variable is set to when the
     * queue will become empty again.
     */
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));

    /*
     * Flush any data in the output queue
     */
    SparcAuSilence();

    /*
     * Set up the device parameters
     */
    if (!SparcAuDeviceSetPlayParams (cbPtr->hdr.samplingRate,
			cbPtr->hdr.channelCount, cbPtr->hdr.dataFormat)) {
	errorCode = FORMAT_ERROR;
	Tk_Release(cbPtr);
	return;
    }

    if (cbPtr->hdr.dataFormat == SUN_LIN_16) {
        width = 2;
    } else {
        width = 1;
    }

    /*
     * If callback is late, just play the appropriate fraction of the data
     */
    if (now > cbPtr->start) {
	fraction = (now - cbPtr->start) / (cbPtr->end - cbPtr->start);
	if (fraction > 1.0) {
	    Tk_Release(cbPtr);
	    return;
	}
	if (fraction < SKEW) {
	    fraction = 0.0;
	}
    } else {
	fraction = 0.0;
    }
    offset = fraction * cbPtr->len + 0.499;

    /*
     * Byte align the offset correctly
     */
    offset = width*(offset/width);
    n = write(fd, cbPtr->data + offset, cbPtr->len - offset);
    emptyTime = now + (cbPtr->end - cbPtr->start)*(1.0 - fraction);
    ResetStopToken();
    Tk_Release(cbPtr);
    lastStartTime = now;
    return;
}

/*
 *--------------------------------------------------------------
 *
 * SparcAuDeviceError --
 *
 *      Map an error code (an integer returned by one of the functions
 *      above) into a human readable string.
 *
 * Results:
 *      Constant string.
 *
 * Side effects:
 *      None
 *
 *--------------------------------------------------------------
 */
static char *
SparcAuDeviceError(SpeakerRsrc *rsrcPtr, int errNum)
{
    switch (errNum) {
      case IN_USE_ERROR:
	  return "Audio Device is already in use by another object";

      case NOT_OPEN_ERROR:
	  return "Attempt to play to unopen audio device";

      case OPEN_ERROR:
	  return "Error opening /dev/audio";

      case TOO_EARLY_ERROR:
	  return "Send-ahead time must be less than data length";

      case FORMAT_ERROR:
	  return "Audio data is of wrong format";

      default:
	  return "Unknown Sparc Audio Error";
    }
}


#else /* SPARC_AU */

/*
 * -------------------------------------------------------
 *
 * GetSparcDevice
 *
 *	Stub routine for non sparc machines
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetSparcDevice(SpeakerRsrc *objPtr)
{
    return (SpeakerDevice *)NULL;
}

#endif /* SPARC_AU */
