/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/* cmSparcAuRecord.c */
/* Author: Jon Logan, Cornell University  */
/*   adapted from SparcAuDevice, to support capturing audio input */

/*
 * HISTORY:
 *   5/25/95  Initial version
 *   6/20/95  Incorporated into Berkeley version of CMT 3.0
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <math.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>

#include "cmAuRecord.h"

#ifdef SPARC_AU

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/filio.h>
#include <stropts.h>
#include "cmAuFile.h"
#include "libaudio.h"
#include "audio_device.h"
#include "audio_hdr.h"
#include "cmLTS.h"

static int fd = -1;	    	/* File descriptor for /dev/audio */
static Audio_hdr devHdr;	/* Stores current state of audio device */

static int atToken = 0;		/* Token for atEvent for playback when
				 * audio device queue is empty */
static double emptyTime;	/* System time when the data currently
				 * in the stream queue should flush */
static int errorCode;		/* Current error code */

/*
 * Interface routines
 */
static int SparcAuRecordInit		(AudioInRsrc *objPtr);
static int SparcAuRecordReinit(AudioInRsrc *objPtr);
static int SparcAuRecordShutdown	(AudioInRsrc *objPtr);
static int SparcAuRecordUnready		(AudioInRsrc *objPtr);
static int SparcAuRecordLTSChanged	(AudioInRsrc *objPtr);
static int SparcAuRecord		    (AuInBufInfo *cbPtr);
static char *SparcAuRecordError		(AudioInRsrc *objPtr, int errNum);
AudioInDevice *GetSparcAuInDevice		(AudioInRsrc *objPtr);

/*
 * Internal Routines
 */
static void SparcAuRecordSetGain	(AudioInRsrc *objPtr);
static void SparcAuRecordSetInput	(AudioInRsrc *objPtr);
static int SparcAuRecordSetParams	(int sample_rate, int channels,
					 int encoding);

#define SKEW		0.005
#define SYNC_TIME	10.0

#define IN_USE_ERROR	1
#define OPEN_ERROR	2
#define FORMAT_ERROR	3
#define TOO_EARLY_ERROR	4
#define NOT_OPEN_ERROR	5
#define SET_PARAMS_ERR  6

#ifndef min
#  define min(a,b)	((a)<(b)?(a):(b))
#endif


/*
 * -------------------------------------------------------
 *
 * GetSparcAuInDevice
 *
 *	Allocates a new AudioInDevice structure, fills
 *	in to handle Sparc Au and return it.
 *
 * -------------------------------------------------------
 */

AudioInDevice *
GetSparcAuInDevice(AudioInRsrc *objPtr)
{
    AudioInDevice *sdev;
    
    sdev = (AudioInDevice *) ckalloc(sizeof(AudioInDevice));
    sdev->deviceData = NULL;
    sdev->initProc = SparcAuRecordInit;
    sdev->reinitProc = SparcAuRecordReinit;
    sdev->shutdownProc = SparcAuRecordShutdown;
    sdev->unreadyProc = SparcAuRecordUnready;
    sdev->ltsProc = SparcAuRecordLTSChanged;
    sdev->recordProc = SparcAuRecord;
    sdev->errorProc = SparcAuRecordError;
    errorCode = 0;

    return sdev;
}


/*
 * -------------------------------------------------------
 *
 * SparcAuRecordInit
 *
 * Initialize this device. Starts recording right away - so
 *   don't make the device ready until everything is humming along. 
 *
 * -------------------------------------------------------
 */

static int
SparcAuRecordInit(AudioInRsrc *objPtr)
{
    int nonblock = 1;

    /* comparing to a floating point number, usually unwise */
    if (LTS_Speed(objPtr->lts) != 1.0)
	return 0;

    if (fd > 0) {
	return IN_USE_ERROR;
    }

    fd = open("/dev/audio", (O_RDONLY | O_NDELAY), 0);
    if (fd < 0) {
	fd = -1;
	return OPEN_ERROR;
    }
    /* Set to non-blocking I/O */
    /*    ioctl(fd, FIONBIO, &nonblock);  */

    audio_get_record_config (fd, &devHdr);
    if (!SparcAuRecordSetParams(objPtr->sampleRate, objPtr->channels, 
				objPtr->format))
	return SET_PARAMS_ERR;

    SparcAuRecordSetGain(objPtr);
    SparcAuRecordSetInput(objPtr);
    errorCode = 0;
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * SparcAuRecordReinit
 *
 * Re-Initialize this device.
 *
 * -------------------------------------------------------
 */

static int
SparcAuRecordReinit(AudioInRsrc *objPtr)
{
    int nonblock = 1;

    if (LTS_Speed(objPtr->lts) != 1.0)
	return 0;

    if (fd < 0) {
	fd = open("/dev/audio", (O_RDONLY | O_NDELAY), 0);
	if (fd < 0) {
	    fd = -1;
	    return OPEN_ERROR;
	}	
	/* Set to non-blocking I/O */
	/*    ioctl(fd, FIONBIO, &nonblock);  */

    }

    audio_get_record_config (fd, &devHdr);
    if (!SparcAuRecordSetParams(objPtr->sampleRate, objPtr->channels, 
				objPtr->format))
	return SET_PARAMS_ERR;

    SparcAuRecordSetGain(objPtr);
    SparcAuRecordSetInput(objPtr);
    errorCode = 0;

    return 0;
}


/*
 * -------------------------------------------------------
 *
 * SparcAuRecordShutdown
 *
 * Destroys the memory that has been allocated for a
 * given SparcAuRecord
 *
 * -------------------------------------------------------
 */

static int
SparcAuRecordShutdown(AudioInRsrc *objPtr)
{
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	emptyTime = 0;
	atToken = 0;
    }

    if (fd > 0) {
	close(fd);
	fd = -1;
    }

    errorCode = 0;
    return 0;
}


/*
 * -------------------------------------------------------
 *
 * SparcAuRecordUnready
 *
 *	Stop recording, close audio device.
 *
 * -------------------------------------------------------
 */

static int
SparcAuRecordUnready(AudioInRsrc *objPtr)
{ 
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	emptyTime = 0;
	atToken = 0;
    }
    
    if (fd > 0) {
	close(fd);
	fd = -1;
    }

    return 0;
}

/*
 * -------------------------------------------------------
 *
 * SparcAuRecord
 *
 * Parameters:
 *    objPtr - audio input object
 * Action:
 *   Returns number of bytes actually recorded. Right now using blocking
 * reads so should be the same as requested...
 * Remarks:
 *   Need to extend this to circular buffer, etc. 
 * -------------------------------------------------------
 */
static int
SparcAuRecord(AuInBufInfo *bufPtr)
{

    int cnt;   /* number of bytes actually recorded */
    char str[1024];

    cnt = read(fd, ((char *)(bufPtr->bufPtr->data)) + bufPtr->offset, bufPtr->dlen);

    
    sprintf(str, "aurecord: read %i bytes into address %x", cnt,
	    (unsigned int)(((char *)(bufPtr->bufPtr->data)) + bufPtr->offset));
    Tcm_Log(str);

    return cnt;
}


/*
 * -------------------------------------------------------
 *
 * SparcAuRecordSetParams
 *
 *	Set the sample rate for audio input.  Returns 1
 *	if successfully set, 0 otherwise
 *
 * Parameters:
 *	int sample_rate;         samples per second
 *	int channels;            # of interleaved channels
 *	int format;            data encoding format
 *
 *  NOTE: if any parameter is -1, the value is left unchanged.
 *
 * -------------------------------------------------------
 */
static int
SparcAuRecordSetParams (int sample_rate, int channels, int format)
{
    Audio_hdr hdr;
    int err;

    /*
     * If not initiatilzed, return
     */
    if (fd == -1) {
	return 1;
    }

    /*
     * Create structure that we're trying to set
     */
    memcpy (&hdr, &devHdr, sizeof(Audio_hdr));
    if (sample_rate != -1) {
	hdr.sample_rate = sample_rate;
    }
    if ((format == SUN_ULAW) || (format == SUN_LIN_8)) {
	hdr.bytes_per_unit = 1;
    } else if (format == SUN_LIN_16) {
	hdr.bytes_per_unit = 2;
    } else {
	return 0;
    }
    if (channels != -1) {
	hdr.channels = channels;
    }
    if (format != -1) {
	if (format == SUN_ULAW)
	    hdr.encoding = AUDIO_ENCODING_ULAW;
	else if ((format = SUN_LIN_8 ) || (format = SUN_LIN_16))
	    hdr.encoding = AUDIO_ENCODING_LINEAR;
	else {
	    fprintf(stderr, "wierd format %i\n", format);
	    return SET_PARAMS_ERR;
	}
    }
    

    /*
     * If no change, return
     */
    if (audio_cmp_hdr(&hdr, &devHdr) == 0) {
	return 1;
    }

    /*
     * Set it...
     */
    err = audio_set_record_config(fd, &hdr);

    /*
     * Update current state with actual state of the device.
     */
    audio_get_record_config (fd, &devHdr);

    if (err == AUDIO_SUCCESS) {
	return 1;
    }
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * SparcAuRecordSetGain
 *
 *	Set the gain of the audio device
 *
 * -------------------------------------------------------
 */

static void
SparcAuRecordSetGain (AudioInRsrc *objPtr)
{
    if ((objPtr != NULL) && (fd != -1)) {
	audio_set_record_gain(fd, &objPtr->gain);
    }
}


/*
 * -------------------------------------------------------
 *
 * SparcAuRecordSetInput
 *
 *	Specify the input of the audio device.
 *
 * -------------------------------------------------------
 */

static void
SparcAuRecordSetInput (AudioInRsrc *objPtr)
{
    int port;

    if ((objPtr != NULL) && (fd != -1)) {
	if (objPtr->inputPort == MICROPHONE) {
	    port = AUDIO_MICROPHONE;
	} else {
	    port = AUDIO_LINE_IN;
	}
	audio_set_record_port (fd, &port);
    }
}


/*
 * -------------------------------------------------------
 *
 * SparcAuRecordLTSChanged
 *
 *	Notify the SparcAuRecord object that the LTS has changed
 *
 * -------------------------------------------------------
 */

static int
SparcAuRecordLTSChanged (AudioInRsrc *objPtr)
{
    if (atToken != 0) {
	Tcm_DeleteAtHandler(atToken);
	emptyTime = 0;
	atToken = 0;
    }

    if (LTS_Speed(objPtr->lts) != 1.0)
	SparcAuRecordUnready(objPtr);
    else if (objPtr->inited)
	return SparcAuRecordReinit(objPtr);
    else
	return SparcAuRecordInit(objPtr);

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * SparcAuRecordError --
 *
 *      Map an error code (an integer returned by one of the functions
 *      above) into a human readable string.
 *
 * Results:
 *      Constant string.
 *
 * Side effects:
 *      None
 *
 *--------------------------------------------------------------
 */
static char *
SparcAuRecordError(AudioInRsrc *rsrcPtr, int errNum)
{
    switch (errNum) {
    case IN_USE_ERROR:
	return "Audio Device is already in use by another object";
	
    case NOT_OPEN_ERROR:
	return "Attempt to record from unopen audio device";
	
    case OPEN_ERROR:
	return "Error opening /dev/audio";
	
    case TOO_EARLY_ERROR:
	return "Send-ahead time must be less than data length";
	
    case FORMAT_ERROR:
	return "Audio data is of wrong format";
	
    case SET_PARAMS_ERR:
	return "Unable to set recording parameters";
	
    default:
	return "Unknown Sparc Audio Error";
    }
}


#else /* SPARC_AU */

/*
 * -------------------------------------------------------
 *
 * GetSparcAuInDevice
 *
 *	Stub routine for non sparc machines
 *
 * -------------------------------------------------------
 */

AudioInDevice *
GetSparcAuInDevice(AudioInRsrc *objPtr)
{
    return (AudioInDevice *)NULL;
}

#endif /* SPARC_AU */


