/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmSunVideoDevice.c,v 1.10 1996/02/06 22:42:41 kpatel Exp $
 *
 * Implements the sunvideo device that works with the mjpegPlay
 * object to do hardware assisted jpeg decoding
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>
#include "cmMJPEGPlay.h"

#if SUN_VIDEO

#include <xil/xil.h>
#include "cmUtil.h"
#include "cmCookies.h"
#include "cmJPEGConvert.h"
#include "cmLTS.h"

#define MAX(x,y)	((x)>(y)?(x):(y))
#define MIN(x,y)	((x)<(y)?(x):(y))

typedef struct SunVideoDevice {
    Display *dpy;		/* X display and window */
    Window win;
    int inited;

    /*
     * Information for the sun decoder board that encapsulates the board's
     * state, and a compressed image sequence (cis) structure.  Compressed
     * images are placed in the cis for decoding.
     */
    XilSystemState state;
    XilCis cis;

    /*
     * We keep some timing information about how long decodes and
     * displays are taking.
     */
    float displayTime;
    int count;

    /*
     * The following variables are used for image decoding.  The
     * JPEG bytes sequence is decoded into imageYCC, with the width/height
     * indicated.
     */
    int imWidth, imHeight;
    XilImage imageYCC;

    /*
     * The following variables are used in image display. 
     * imageYCC is dithered into the displayimage structure, which
     * places it on the screen as a side effect.
     */
    XilImage displayimage;
    int winWidth, winHeight;

    /*
     * Stuff for dithering
     */
    int defaultVisual;		/* PSEUDOCOLOR or TRUECOLOR */
    XilLookup colorcube;
    float scale[3], offset[3];
    XilDitherMask dmask;
    Colormap cmap;
} SunVideoDevice;

#define JPEG 0
#define PSEUDOCOLOR 0
#define TRUECOLOR 1

/*
 * Errors generated by this module
 */
#define XIL_OPEN_FAIL                   1
#define XIL_ERROR_INSTALL_FAIL          2
#define XIL_CIS_CREATE_FAIL             3
#define XIL_CREATE_FROM_WINDOW_FAIL     4
#define XIL_BAD_VISUAL                  5
#define XIL_CREATE_FAIL                 6

static XilError currError;


/*
 *--------------------------------------------------------------
 *
 * XilErrorHandler --
 *
 *	Called when an error occurs in the X image library.
 *	Just set a global variable to indicate the error type.
 *
 *	All changes to the module made that decrease resource usage,
 *	but make the function harder to understand, modify, and debug.
 *
 * Results:
 *	Description of return values.
 *
 * Side effects:
 *	The global variable currError is overwritten.  The caller of
 *	any xil function should check this variable after the call
 *	to see if an error occured.
 *
 *--------------------------------------------------------------
 */
Xil_boolean
XilErrorHandler(XilError error)
{
    printf ("Internal error...\n");
    xil_call_next_error_handler(error);
    currError = error;
}


/*
 *--------------------------------------------------------------
 *
 * CreateColormap --
 *
 *	Create a colormap for dithering.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The color cells on the X server are allocated.
 *
 *--------------------------------------------------------------
 */
static void
CreateColormap (Display *display, Window window,
                int screen, Colormap *cmap, XilLookup yuvtorgb,
                XilLookup colorcube)
{
    unsigned long tmp_colors[256], pixels[256], mask;
    XColor cdefs[256];

    /*
     * colormap indices 255 and 254 
     */
    int top_colors = 2;
    int i, j, t;
    Xil_unsigned8 *data;
    XilLookup lut;
    int cmapsize;

    /*
     * Get colormap size 
     */
    cmapsize = xil_lookup_get_num_entries(yuvtorgb);

    /*
     * Create an X colormap for the cis. 
     */
    *cmap = XCreateColormap(display, window, DefaultVisual(display, screen),
			    AllocNone);

    /*
     * Allocate X Colormap cells
     * 
     * If we do not need the entire colormap, allocate `cmapsize' entries just
     * below the top of the colormap.  Here's how: Temporarily allocate some
     * entries at the front of the cmap.  Don't allocate the top_colors (the
     * top two color indices are often used by other applications). Allocate
     * the needed entries in the next cmap section Free the temporary
     * entries. This allows the X-Window manager to use them and reduces the
     * chances of of your window colormap flashing.
     */
    if (cmapsize < 256 - top_colors) {
	if (!XAllocColorCells(display, *cmap, 0, &mask, 0,
			      tmp_colors, 256 - cmapsize - top_colors)) {
	    fprintf(stderr, " XAllocColorCells for cmap_create failed(1)\n");
	    exit(1);
	}
    }
    if (!XAllocColorCells(display, *cmap, 0, &mask, 0, pixels, cmapsize)) {
	fprintf(stderr, " XAllocColorCells for cmap_create failed(2)\n");
	exit(1);
    }

    /*
     * The remaining code assumes that the values returned in pixels[0]
     * through pixels[cmapsize-1] are a contiguous range.
     *
     * Free the unused colors in the front 
     */
    if (cmapsize < 256 - top_colors) {
	XFreeColors(display, *cmap, tmp_colors, 256 - cmapsize - top_colors, 0);
    }

    /*
     * Allocate memory to hold colormap data. 
     */
    if ((data = (Xil_unsigned8 *)malloc(sizeof(Xil_unsigned8) *
					cmapsize * 3)) == NULL) {
	fprintf(stderr, "xilcis_color: out of memory for cmap data create\n");
	exit(1);
    }

    /*
     * Get the entries for the colormap. The method depends on the
     * compression type.  For JPEG, get the entries from the standard lookup
     * table yuv_to_rgb.
     */
    xil_lookup_get_values(yuvtorgb, xil_lookup_get_offset(yuvtorgb),
			  cmapsize, data);
    xil_lookup_set_offset(colorcube, (unsigned int) pixels[0]);
    for (i = 0, t = 0; i < cmapsize; i++, t += 3) {
	cdefs[i].pixel = pixels[i];
	cdefs[i].flags = DoRed | DoGreen | DoBlue;
	cdefs[i].blue = data[t] << 8;
	cdefs[i].green = data[t + 1] << 8;
	cdefs[i].red = data[t + 2] << 8;
    }
    
    XStoreColors(display, *cmap, cdefs, cmapsize);
    lut = yuvtorgb;

    /*
     * Install colormap before cursor is placed in window 
     */
    XInstallColormap(display, *cmap);
    XSync(display, False);
    XSetWindowColormap(display, window, *cmap);
    free(data);
}


/*
 *--------------------------------------------------------------
 *
 * SunVideoInit --
 *
 *	Initialize the device.
 *
 * Results:
 *	0 for successful initilization, non-zero error code otherwise
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */
int
SunVideoInit(MJPEGResource *rsrcPtr)
{
    Colormap xCmap;
    Visual *visual;
    VisualID visualid;
    XVisualInfo visualTemplate;
    int matchingVisuals;
    SunVideoDevice *svDev;
    int screen;
    float dx, dy;

    svDev = (SunVideoDevice *) rsrcPtr->devicePtr->deviceData;
    svDev->inited = 0;
    svDev->dpy = Tcm_Display();
    svDev->win = rsrcPtr->xid;
    screen = DefaultScreen(svDev->dpy);

    if ((svDev->state = xil_open()) == NULL) {
	return XIL_OPEN_FAIL;
    }

    /*
     * Install error handler
     */
    if (xil_install_error_handler(svDev->state, XilErrorHandler) ==
				XIL_FAILURE) {
        return XIL_ERROR_INSTALL_FAIL;
    }

    /*
     * Create the CIS
     */
    svDev->cis = xil_cis_create(svDev->state, "Jpeg");
    if (!svDev->cis) {
        return XIL_CIS_CREATE_FAIL;
    }

    /*
     * Create an XIL display image.  When we copy to/from this image
     * we actually read/write the device.
     */
    svDev->displayimage = xil_create_from_window(svDev->state, svDev->dpy,
						 svDev->win);
    if (!svDev->displayimage) {
        return XIL_CREATE_FROM_WINDOW_FAIL;
    }
    svDev->winWidth = xil_get_width (svDev->displayimage);
    svDev->winHeight = xil_get_height (svDev->displayimage);

    /*
     * Center the image
     */
    dx = (svDev->winWidth - svDev->imWidth)/2.0;
    dy = (svDev->winHeight - svDev->imHeight)/2.0;
    xil_set_origin (svDev->displayimage, dx, dy);

    /*
     * Figure out the right visual to use.  Try for a 24 bit one,
     * and if that doesn't work, use an 8 bit one.
     */
    visual = DefaultVisual(svDev->dpy, screen);
    visualid = XVisualIDFromVisual(visual);
    visualTemplate.visualid = visualid;
    visualTemplate.depth = 24;
    visualTemplate.class = TrueColor;
    XGetVisualInfo(svDev->dpy, VisualIDMask | VisualDepthMask | VisualClassMask,
		   &visualTemplate, &matchingVisuals);
    if (matchingVisuals == 1) {
	svDev->defaultVisual = TRUECOLOR;
    } else {
	visualTemplate.depth = 8;
	visualTemplate.class = PseudoColor;
	XGetVisualInfo(svDev->dpy, VisualIDMask|VisualDepthMask|VisualClassMask,
		       &visualTemplate, &matchingVisuals);
	if (matchingVisuals != 1) {
	    return XIL_BAD_VISUAL;
	}
	svDev->defaultVisual == PSEUDOCOLOR;
    }

    if (svDev->defaultVisual == PSEUDOCOLOR) {
	svDev->scale[0] = 255.0 / (235.0 - 16.0);
	svDev->scale[1] = 255.0 / (240.0 - 16.0);
	svDev->scale[2] = 255.0 / (240.0 - 16.0);
	svDev->offset[0] = -16.0 * svDev->scale[0];
	svDev->offset[1] = -16.0 * svDev->scale[1];
	svDev->offset[2] = -16.0 * svDev->scale[2];

	svDev->colorcube = xil_lookup_get_by_name(svDev->state, "cc855");
	svDev->dmask = xil_dithermask_get_by_name(svDev->state, "dm443");

	/*
	 * This routine is in the file: xilcis_color.c. Inside this routine
	 * the colormap is installed through X, and the color cube offset is
	 * set to the cmap offset.
	 */
	CreateColormap (svDev->dpy, svDev->win, screen, &svDev->cmap, 
			 xil_lookup_get_by_name(svDev->state, "yuv_to_rgb"),
			 svDev->colorcube);
    } else if (svDev->defaultVisual == TRUECOLOR) {
	XilColorspace rgb;
	rgb = xil_colorspace_get_by_name(svDev->state, "rgb709");
	xil_set_colorspace(svDev->displayimage, rgb);
    }

    svDev->inited = 1;
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * SunVideoReinit --
 *
 *	Reinitialize the sunVideo device.  Called when
 *	ready signal is sent, after device is configured.
 *	Things that might have changed:
 *		size (width, height)
 *		window (xid)
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
SunVideoReinit(MJPEGResource *rsrcPtr)
{
    SunVideoDevice *svDev;
    float dx, dy;

    svDev = (SunVideoDevice *) rsrcPtr->devicePtr->deviceData;
    if (!svDev->inited) {
	return 0;
    }
    xil_cis_reset(svDev->cis);

    /*
     * We have to "recreate" the image to get the xil to know the new
     * size.  There's probably a better way to do this, but it doesn't
     * happen too often, so it shouldn't be too much overhead.
     */
    if (svDev->displayimage) {
	xil_destroy(svDev->displayimage);
	svDev->displayimage = NULL;
    }
    svDev->win = rsrcPtr->xid;
    svDev->displayimage = xil_create_from_window(svDev->state,
					 svDev->dpy, svDev->win);
    if (!svDev->displayimage) {
	return XIL_CREATE_FROM_WINDOW_FAIL;
    }
    svDev->winWidth = xil_get_width (svDev->displayimage);
    svDev->winHeight = xil_get_height (svDev->displayimage);

    /*
     * Center the image
     */
    dx = (svDev->winWidth - svDev->imWidth)/2.0;
    dy = (svDev->winHeight - svDev->imHeight)/2.0;
    xil_set_origin (svDev->displayimage, dx, dy);

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * SunVideoDecode
 *
 *	Start the decoding process for a frame.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	This function will be periodcially called until frame
 *	decode is complete
 *
 *--------------------------------------------------------------
 */
int
SunVideoDecode (MJPEGFrame *framePtr) 

{
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * SunVideoDisplay --
 *
 *	Toss a decoded frame up on the screen.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Could block if previous shared memory put image hasn't returned.
 *
 *--------------------------------------------------------------
 */
int
SunVideoDisplay (MJPEGFrame *framePtr)
{
    MJPEGResource *objPtr;
    MJPEGDevice *devicePtr;
    SunVideoDevice *svDev;
    float dx, dy;
    double start, end;
    char *header, *data;
    int q, w, h, format, size;
    int hSize, dSize;
    double currTime, t1, t2;

    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    svDev = (SunVideoDevice *)devicePtr->deviceData;
    if (!svDev->inited) {
	return 0;
    }

    q = framePtr->frameHeader.qfactor;
    w = framePtr->frameHeader.width;
    h = framePtr->frameHeader.height;
    currTime = LTS_Value(objPtr->lts);
    t1 = framePtr->logStart;
    t2 = framePtr->logEnd;
    if (MAX(t1,t2) < currTime) {
	return 0;
    }

    if ((w != svDev->imWidth) || (h != svDev->imHeight)) {
	/*
	 * Destroy any previously allocated image
	 */
	if (svDev->imageYCC) {
	    xil_destroy(svDev->imageYCC);
	    svDev->imageYCC = NULL;
	}

	/*
	 * imageYCC is the XilImage that the cis will be decompressed into.
	 */
	svDev->imageYCC = xil_create(svDev->state, w, h, 3, XIL_BYTE);
	if (svDev->imageYCC == NULL) {
	    return XIL_CREATE_FAIL;
	}
	svDev->imWidth = w;
	svDev->imHeight = h;

	/*
	 * Center the image
	 */
	dx = (svDev->winWidth - svDev->imWidth)/2.0;
	dy = (svDev->winHeight - svDev->imHeight)/2.0;
	xil_set_origin (svDev->displayimage, dx, dy);


	if (svDev->defaultVisual == TRUECOLOR) {
	    XilColorspace ycc;
	    ycc = xil_colorspace_get_by_name(svDev->state, "ycc601");
	    xil_set_colorspace(svDev->imageYCC, ycc);
	}
    }

    format = framePtr->frameHeader.subformat;
    size = framePtr->dataLength;

    JpegConvert (framePtr->data, size, w, h, q, format, JFIF_FRAME_CODE,
		 &header, &hSize, &data, &dSize);

    /*
     * Step 1: Add the bits to the compressed image sequence (cis)
     */
    if (header) {
        xil_cis_put_bits_ptr(svDev->cis, hSize, 0, (void *)header, NULL);
        xil_cis_put_bits_ptr(svDev->cis, dSize, 0, (void *)data, NULL);
    } else if (data) {
        xil_cis_put_bits_ptr(svDev->cis, dSize, 0, (void *)data, NULL);
    }

    /*
     * Step 2: Decompress
     */
    xil_decompress(svDev->cis, svDev->imageYCC);

    /*
     * Step 3: Adjust the zero point for YCrCb->RGB mapping
     */
    if (svDev->defaultVisual == PSEUDOCOLOR) {
	xil_rescale (svDev->imageYCC, svDev->imageYCC, svDev->scale,
		     svDev->offset);
    }

    /*
     * Step 4: Force a reset every 20 frames or so.
     */
    svDev->count++;
    if ((svDev->count % 20) == 0) {
	xil_cis_reset(svDev->cis);
    }

    start = Tcm_ReadSysClock();

    /*
     * Step 5: Display (and time it!)
     */
    if (svDev->defaultVisual == PSEUDOCOLOR) {
	xil_ordered_dither(svDev->imageYCC, svDev->displayimage,
			   svDev->colorcube, svDev->dmask);
    } else if (svDev->defaultVisual == TRUECOLOR) {
	xil_color_convert(svDev->imageYCC, svDev->displayimage);
    }
    xil_sync(svDev->displayimage);

    end = Tcm_ReadSysClock();
    svDev->displayTime = (end-start + svDev->displayTime)/2.0;
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * SunVideoRefresh --
 *
 *	Called to refresh the display if part of it is exposed.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Display is repainted.  Could block if previous shared
 *	memory put image hasn't returned
 *
 *--------------------------------------------------------------
 */
int
SunVideoRefresh(MJPEGResource *rsrcPtr)
{
    SunVideoDevice *svDev;

    svDev = (SunVideoDevice *) rsrcPtr->devicePtr->deviceData;
    if (!svDev->inited) {
	return 0;
    }

    if (svDev->defaultVisual == PSEUDOCOLOR) {
	xil_ordered_dither(svDev->imageYCC, svDev->displayimage,
			   svDev->colorcube, svDev->dmask);
    } else if (svDev->defaultVisual == TRUECOLOR) {
	xil_color_convert(svDev->imageYCC, svDev->displayimage);
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * SunVideoShutdown --
 *
 *	Called to destroy device.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	The device must never be referenced again.
 *
 *--------------------------------------------------------------
 */
int
SunVideoShutdown(MJPEGResource *rsrcPtr) 
{
    SunVideoDevice *svDev;

    svDev = (SunVideoDevice *) rsrcPtr->devicePtr->deviceData;
    if (!svDev->inited) {
	return 0;
    }

    if (svDev->imageYCC) {
	xil_destroy(svDev->imageYCC);
	svDev->imageYCC = NULL;
    }
    if (svDev->displayimage) {
	xil_destroy(svDev->displayimage);
	svDev->displayimage = NULL;
    }
    xil_cis_destroy(svDev->cis);
    xil_close(svDev->state);
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * SunVideoFreeFrame --
 *
 *	The current frame is being taken away -- clean up any pointers
 *	we have to it.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
SunVideoFreeFrame(MJPEGFrame *framePtr)
{
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * SunVideoError --
 *
 *	Translates an error number (passed in) into a human readable
 *	string (returned).
 *
 * Results:
 *	Human readable string that indicates the error.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	MJPEGResource *rsrcPtr	The device
 *	int errNum		The error code
 *
 *--------------------------------------------------------------
 */
char *
SunVideoError(MJPEGResource *rsrcPtr, int errNum)
{
    return "SunVideo Decoder error.";
}


/*
 *--------------------------------------------------------------
 *
 * SunVideoNextFrame --
 *
 *	Figure out the next frame we should play.
 *
 * Results:
 *	0 for success, 1 for no frame to play
 *
 * Side effects:
 *	
 *
 *--------------------------------------------------------------
 */
static int
SunVideoNextFrame (MJPEGResource *objPtr, MJPEGFrame *frameList,
		   MJPEGFrame **chosenFrame)
{
    MJPEGFrame *framePtr;
    double speed, searchTime, dt;
    SunVideoDevice *svDev;

    *chosenFrame = NULL;
    svDev = (SunVideoDevice *) objPtr->devicePtr->deviceData;
    if (!svDev->inited) {
	return 1;
    }

    if (frameList == NULL) {
	return 1;
    }

    speed = LTS_Speed(objPtr->lts);
    dt = speed * (svDev->displayTime +0.025);
    searchTime = LTS_Value(objPtr->lts) + dt;

    for (framePtr = frameList; framePtr != NULL; framePtr = framePtr->next) {
	if ((framePtr->logStart <= searchTime) &&
	    (searchTime <= framePtr->logEnd)) {
	    *chosenFrame = framePtr;
	    break;
	}
    }

    /*
    if (*chosenFrame == NULL) {
	*chosenFrame = frameList;
    }
    */
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * GetSunVideoJpegDevice --
 *
 *	Called by resource to determine default device.  We should
 *	only be called if there's no jpeg hardware available.
 *
 * Results:
 *	Pointer to new MJPEGDevice
 *
 * Side effects:
 *	A couple of structures are allocated.
 *
 *--------------------------------------------------------------
 */
MJPEGDevice *
GetSunVideoJpegDevice (MJPEGResource *objPtr)
{
    MJPEGDevice *new_dev;
    SunVideoDevice *svDev;

    new_dev = (MJPEGDevice *) ckalloc(sizeof(MJPEGDevice));

    new_dev->asyncDecodeFlag = 0;
    new_dev->decodeProc = SunVideoDecode;
    new_dev->refreshProc = SunVideoRefresh;
    new_dev->displayProc = SunVideoDisplay;
    new_dev->initProc = SunVideoInit;
    new_dev->reinitProc = SunVideoReinit;
    new_dev->shutdownProc = SunVideoShutdown;
    new_dev->freeFrameProc = SunVideoFreeFrame;
    new_dev->nextFrameProc = NULL;
    new_dev->configProc = NULL;
    new_dev->errorProc = SunVideoError;
    svDev = (SunVideoDevice *) ckalloc(sizeof(SunVideoDevice));
    new_dev->deviceData = svDev;

    svDev->cis = NULL;
    svDev->count = 0;
    svDev->defaultVisual = PSEUDOCOLOR;
    svDev->displayTime = 0.100;
    svDev->displayimage = NULL;
    svDev->dpy = Tcm_Display();
    svDev->imWidth = svDev->imHeight = 0;
    svDev->imageYCC = NULL;
    svDev->inited = 0;
    svDev->offset[0] = svDev->offset[1] = svDev->offset[2] = 0;
    svDev->scale[0] = svDev->scale[1] = svDev->scale[2] = 0;
    svDev->state = NULL;
    svDev->win = NULL;
    svDev->winWidth = svDev->winHeight = 0;

    return new_dev;
}


#else /* SUN_VIDEO */

/*
 *--------------------------------------------------------------
 *
 * GetSunVideoJpegDevice --
 *
 *	Stub routine for machines without compiled in sunVideo
 *	JPEG decoder.
 *
 * Results:
 *	NULL
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
MJPEGDevice *
GetSunVideoJpegDevice (MJPEGResource *objPtr)
{
    return NULL;
}

#endif	/* SUN_VIDEO */
