/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * trace.c
 *
 *	to handle level tracing in the cm project.
 *
 */

/* Have to do this for Windows or else we get a name conflict with OpenFile */
#ifdef _WINDOWS
#    undef NEED_TCM_EVENT_H
#endif /* _WINDOWS */

#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include "default.h"
#include "tk.h"
#include "tclInt.h"


/* ------------------------------------------------------------- */

int Tcm_TraceCmd();
int Tcm_UntraceCmd();

/* ------------------------------------------------------------- */

typedef struct _TraceProc {
    int id;
    Tcl_Trace token;

    char *procedure;	/* Name or abbreviation of traced procedure */
    char *command;

    int proclength;	/* Length of the string in procedure */

    struct _TraceProc *next;

} TraceProc;

TraceProc *traceProcs = NULL;

/* ------------------------------------------------------------- */

/*
 *--------------------------------------------------------------
 *
 * Tcm_TraceProc --
 *
 *	This is the callback invoked by the low-level Tcl tracing
 *	facilities.  It evaluates a Tcl command.
 *
 * Results:
 *	Unused.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

 /* ARGSUSED */
void
Tcm_TraceProc(clientData, interp, level, command, cmdProc, cmdClientData,
	      argc, argv)
    ClientData clientData;
    Tcl_Interp *interp;
    int level;
    char *command;
    int (*cmdProc) ();
    ClientData cmdClientData;
    int argc;
    char **argv;
{
    static int inTrace = 0;  /* Used to prevent tracing the trace procedure */
    TraceProc *node = (TraceProc *) clientData;
    int length = node->proclength;

    char stringLevel[200];
    char stringId[200];

    char *args;

    if (inTrace) return;

    if ((argc > 0) && (strncmp(node->procedure, argv[0], length) == 0)) {
	args = Tcl_Merge(argc, argv);

	sprintf(stringLevel, "%d", level);
	sprintf(stringId, "%d", node->id);

	inTrace = 1;
	Tcl_VarEval(interp, node->command, " ",
		    stringId, " ",
		    stringLevel, " ",
		    args, (char *) NULL);
	inTrace = 0;

	ckfree(args);
    }
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_TraceInit --
 *
 *	This procedure is invoked to initialize the trace module;
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
Tcm_TraceInit(interp)
    Tcl_Interp *interp;		/* Current interpreter. */
{
    Tcl_CreateCommand(interp, "traceProc", Tcm_TraceCmd,
		      (ClientData) NULL, (void (*) ()) NULL);

    Tcl_CreateCommand(interp, "untraceProc", Tcm_UntraceCmd,
		      (ClientData) NULL, (void (*) ()) NULL);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_TraceCmd --
 *
 *	This procedure is invoked to process the "traceProc" TclTk
 *	command.  See the user documentation for details on what
 *	it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

 /* ARGSUSED */
int
Tcm_TraceCmd(notUsed, interp, argc, argv)
    ClientData notUsed;
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    static int id = 0;

    Tcl_Trace token;

    TraceProc *node;

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args : should be \"", argv[0],
			 " procedure command\"", (char *) NULL);
	return TCL_ERROR;
    }
    node = (TraceProc *) ckalloc(sizeof(struct _TraceProc));

    node->proclength = strlen(argv[1]);
    node->procedure = (char *) ckalloc(node->proclength + 1);
    node->command = (char *) ckalloc(strlen(argv[2]) + 1);

    strcpy(node->procedure, argv[1]);
    strcpy(node->command, argv[2]);

    node->id = id++;
    node->next = traceProcs;
    traceProcs = node;

    token = Tcl_CreateTrace(interp, MAX_NESTING_DEPTH,
			    Tcm_TraceProc, (ClientData) node);

    node->token = token;

    sprintf(interp->result, "%d", node->id);

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_UntraceCmd --
 *
 *	This procedure is invoked to process the "untraceProc" TclTk
 *	command.  See the user documentation for details on what
 *	it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

 /* ARGSUSED */
int
Tcm_UntraceCmd(notUsed, interp, argc, argv)
    ClientData notUsed;
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    int id;

    TraceProc *prev;
    TraceProc *curr;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args : should be \"", argv[0],
			 " traceId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[1], &id) != TCL_OK)
	return TCL_ERROR;

    prev = NULL;
    curr = traceProcs;

    while (curr && (curr->id != id)) {
	prev = curr;
	curr = curr->next;
    }

    if (curr) {
	if (prev)
	    prev->next = curr->next;
	else
	    traceProcs = curr->next;

	Tcl_DeleteTrace(interp, curr->token);

	ckfree((char *) curr->procedure);
	ckfree((char *) curr->command);
	ckfree((char *) curr);
    }
    return TCL_OK;
}
