/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * tcmUdpDest.c --
 *
 *	This module receives packets from the UdpSrc,
 *	re-assembles them in to frames and passes them up
 *	to the media dest modules.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#include "compat/stdlib.h"
#else
#include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <limits.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include "tcl.h"
#include "dp.h"
#include "tk.h"
#include "cmBufferlist.h"
#include "cmPacket.h"
#include "cmAtom.h"
#include "cmUtil.h"

#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef NO_SENDRECVMSG
#    include "compat/sendrecvmsg.h"
#endif

#ifndef _WINDOWS
#    ifndef abs
#        define abs(a)	((a)>0?(a):-(a))
#    endif
#    define min(a,b)	((a)>(b)?(b):(a))
#    define max(a,b)	((a)>(b)?(a):(b))
#endif

/*
 * Maximum number of reads we do at one time
 */
#define MAX_SEQ_READS 5

/*
 * An actual UdpDest object
 */
typedef struct UdpDest {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */

    char *outCmd;		/* The command to push data */
    InFunc outCCmd;		/* C pointer to out command */
    void *outObj;		/* The object to send to */

    DP_SOCKET udp;		/* File descriptor for UDP data socket */
    int port;			/* Port number for UDP data socket */
    int shared;			/* Should buffers be in shared memory */
    int xAttach;		/* Should buffers be attached
				 * to the X server */
    char *address;		/* String with internet address and port */
    char *pg;			/* Priority group */
} UdpDest;

/* Globals used by this object module. */

static int nextUdpDestNum = 0;   /* Variable for next unique object name. */

/*
 * Prototypes for routines in the file
 */
int UdpDestModuleInit (Tcl_Interp *interp);

static UdpDest *CreateNewUdpDest _ANSI_ARGS_((Tcl_Interp *interp,
		    char *name));
static int	DestroyUdpDest _ANSI_ARGS_((UdpDest *objPtr));
static int	UdpDestCreateCmd _ANSI_ARGS_((ClientData clientData,
		    Tcl_Interp *interp, int argc, char **argv));
static int	UdpDestObjCmd _ANSI_ARGS_((ClientData clientData,
		    Tcl_Interp *interp, int argc, char **argv));
static void	HandleIncomingUdp _ANSI_ARGS_((ClientData clientData,
		    int mask));
static Buffer  *FindUdpBuffer _ANSI_ARGS_((UdpDest *udpPtr, int size));
static void     FinishedBuffer _ANSI_ARGS_((Buffer *bufPtr, UdpDest *udpPtr,
		    int size));
static int	ConfigureUdpDest _ANSI_ARGS_((UdpDest *objPtr,
		    int argc, char **argv));

/************************************************************
 *
 * CreateNewUdpDest
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static UdpDest *
CreateNewUdpDest (Tcl_Interp *interp, char *name)
{
    UdpDest *new;
    DP_SOCKET udp;
    int port, len, res;
    struct sockaddr_in sockaddr;
    char str[256];
    int optname, optval, optlen;
    int ll, ul;

    udp = Tdp_inet_connect ("", 0, 0, 1, 0, 0);
    if (udp == DP_INVALID_SOCKET) {
	Tcl_AppendResult (interp, "Couldn't open udp socket : ",
			  Tdp_Error(interp), (char *)NULL);
	return NULL;
    }

    /*
     * Figure out port number
     */
    len = sizeof (sockaddr);
    res = getsockname (udp, (struct sockaddr *) &sockaddr, &len);
    if (res < 0) {
	Tcl_AppendResult (interp, "Error getting port number of udp socket : ",
			  Tcl_PosixError(interp), (char *)NULL);
	close(udp);
	return NULL;
	
    }
    port = ntohs(sockaddr.sin_port);

    /*
     * Set the receive buffer to a large value to try to prevent
     * overflows.  Since this varies from OS to OS, we just search for
     * the max size buffer less than 100k
     */
    optname = SO_RCVBUF;
    optlen = sizeof(int);
    res = getsockopt(udp, SOL_SOCKET, optname, (char *)&optval, &optlen);
    if (res != -1) {

	/*
	 * Phase 1: Double size of buffer until we get an error or
	 * buffer size is 50k
	 */
	if (optval < 1) {
	    optval = 1024;
	}
	while ((res != -1) && (optval < 100000)) {
	    optval <<= 1;
	    res = setsockopt(udp, SOL_SOCKET, optname, (char *)&optval, optlen);
	}

	/*
	 * Phase 2: Do binary search on max buffer size.
	 */
	ul = optval;
	ll = optval >> 1;
	optval = (ll+ul)/2;
	while ((optval < 100000) && (ul-ll > 1000)) {
	    res = setsockopt(udp, SOL_SOCKET, optname, (char *)&optval, optlen);
	    if (res == -1) {
		ul = optval;
	    } else {
		ll = optval;
	    }
	    optval = (ll+ul)/2;
	}
    }

    /*
     * Create the new object!
     */
    new = (UdpDest *) ckalloc(sizeof(UdpDest));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outObj = NULL;

    new->udp = udp;
    new->port = port;
    new->shared = 0;
    new->xAttach = 0;
    new->pg = ckalloc(3);
    memcpy(new->pg, "tk", 3);

    sprintf (str, "%s %d", Tcm_GetLocalAddress(interp), port);
    new->address = ckalloc(strlen(str)+1);
    strcpy(new->address, str);

    /*
     * Set up a handler on the socket
     */
    Tcm_CreateFileHandler(new->pg, udp, TK_READABLE, HandleIncomingUdp,
			  (ClientData)new);
    
    return new;
}


/************************************************************
 *
 * DestroyUdpDest
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 ************************************************************/

static int 
DestroyUdpDest (UdpDest *objPtr)
{
    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Close the socket, free up memory.
     */
    Tcm_DeleteFileHandler(objPtr->udp);
    closesocket(objPtr->udp);
    if (objPtr->name != NULL) {
	ckfree((char *) objPtr->name);
    }
    if (objPtr->address != NULL) {
	ckfree((char *) objPtr->address);
    }

    ckfree((char *) objPtr->pg);
    ckfree((char *) objPtr);
    return TCL_OK;
}


/************************************************************
 *
 * UdpDestCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the udpDest command.
 *     A new object of type UdpDest is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

static int
UdpDestCreateCmd (ClientData clientData,
		     Tcl_Interp *interp, int argc, char **argv)
{
    char new_name[256];
    UdpDest *new;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " parent", (char *) NULL);
	return TCL_ERROR;
    }

    nextUdpDestNum++;
    sprintf(new_name, "%spd%d", argv[1], nextUdpDestNum);

    new = CreateNewUdpDest(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, new_name, UdpDestObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * UdpDestObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     UdpDest object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

static int
UdpDestObjCmd (ClientData clientData, Tcl_Interp *interp,
		  int argc, char **argv)
{

    char c;
    int len;
    UdpDest *objPtr = (UdpDest *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ", argv[0],
			 "<command> ?arguments?", (char *) NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyUdpDest(objPtr);
    } else if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureUdpDest(objPtr, argc-2, argv+2);
    } else if ((c == 'a') && (strncmp("address", argv[1], len) == 0)) {
	Tcl_SetResult (interp, objPtr->address, TCL_VOLATILE);
	return TCL_OK;
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ",
			argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
                "\n<command> can be one of:\n",
                "  configure ?-slot ?value??\n",
                "  destroy\n",
                "  address\n",
                (char *) NULL);
    return TCL_ERROR;
}


/************************************************************
 *
 * ConfigureUdpDest
 *
 * Parameters: 
 *	objPtr - the object to configure
 *	argc - argument count
 *	argv - argument values
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *	Configures the packet dest object
 *
 ************************************************************/

static int
ConfigureUdpDest (UdpDest *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;
    
    interp = objPtr->interp;
    
    if (argc == 0) {
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", NULL);
	} else {
	    Tcl_AppendResult(interp,
			     "{-outCmd {} {", objPtr->outCmd, "}} ",
			     (char *) NULL);
	}
	Tcl_AppendResult(interp, "{-shared {false} {",
			 (objPtr->shared ? "true" : "false"),
			 "}} ", NULL);
	Tcl_AppendResult(interp, "{-attach {false} {",
			 (objPtr->xAttach ? "true" : "false"),
			 "}} ", NULL);
	Tcl_AppendResult(interp, "{-address {", objPtr->address, "} {",
			 objPtr->address, "}} ", NULL);
	Tcl_AppendResult(interp, "{-priorityGroup {tk} {",
			 objPtr->pg, "}} ", NULL);
	return TCL_OK;
    } 
    
    if (argc == 1) {
	int len;
	
	len = strlen(argv[0]);
	if (strncmp("-outCmd", argv[0], len) == 0) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", NULL);
	    } else {
		Tcl_AppendResult(interp,
				 "-outCmd {} {", objPtr->outCmd, "}",
				 (char *) NULL);
	    }
	    return TCL_OK;
	} else if (strncmp("-shared", argv[0], len) == 0) {
	    Tcl_AppendResult(interp, "-shared {false} {",
			     (objPtr->shared ? "true" : "false"),
			     "} ", NULL);
	    return TCL_OK;
	} else if (strncmp("-attach", argv[0], len) == 0) {
	    Tcl_AppendResult(interp, "-attach {false} {",
			     (objPtr->xAttach ? "true" : "false"),
			     "} ", NULL);
	    return TCL_OK;
	} else if (strncmp("-address", argv[0], len) == 0) {
	    Tcl_AppendResult(interp, "-address {", objPtr->address, "} {",
			     objPtr->address, "} ", NULL);
	    return TCL_OK;
	} else if (strncmp("-priorityGroup", argv[0], len) == 0) {
	    Tcl_AppendResult(interp, "-priorityGroup {tk} {",
			     objPtr->pg, "}", NULL);
	    return TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "bad configuration option: ", NULL);
	    return TCL_ERROR;
	}
    }
    
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];
  
	if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    Tcl_CmdInfo cmdInfo;
	    char *objName, *division;
    
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -outCmd",
				 NULL);
		return TCL_ERROR;
	    }
    
	    if (objPtr->outCmd != NULL) {
		ckfree((char *) objPtr->outCmd);
	    }
    
	    objPtr->outCmd = ckalloc(strlen(argv[cntr]) + 1);
	    strcpy(objPtr->outCmd, argv[cntr]);
	    
	    if (objPtr->outCmd[0] == '@') {
		if (Tutil_GetRegisteredCommand(interp, objPtr->outCmd,
					       (ClientData *) &(objPtr->outObj),
					       (ClientData *) &(objPtr->outCCmd)) != TCL_OK) {
		    Tcl_AppendResult(interp,
				     "Could not find register procedure ",
				     argv[cntr] + 1, NULL);
		    ckfree(objPtr->outCmd);
		    objPtr->outCmd = NULL;
		    objPtr->outObj = NULL;
		    objPtr->outCCmd = NULL;
		    return TCL_ERROR;
		}
	    } else {
		objPtr->outObj = NULL;
		objPtr->outCCmd = NULL;
	    }
	} else if ((c == 's') && (strncmp("-shared", argv[cntr], len) == 0)) {
	    int newshared;
	    
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -shared",
				 NULL);
		return TCL_ERROR;
	    }
	    
	    if (Tcl_GetBoolean(interp, argv[cntr], &newshared) != TCL_OK) {
		return TCL_ERROR;
	    }
	    objPtr->shared = newshared;
	} else if ((c == 'a') && (strncmp("-attach", argv[cntr], len) == 0)) {
	    int newattach;
	    
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -attach",
				 NULL);
		return TCL_ERROR;
	    }
	    
	    if (Tcl_GetBoolean(interp, argv[cntr], &newattach) != TCL_OK) {
		return TCL_ERROR;
	    }
	    objPtr->xAttach = newattach;
	} else if ((c == 'a') && (strncmp("-address", argv[cntr], len) == 0)) {
	    Tcl_AppendResult(interp,
			     "Can't set address slot of a udpDest", NULL);
	    return TCL_ERROR;
	} else if ((c == 'p') && (strncmp("-priorityGroup", argv[cntr], len) == 0)){
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -priorityGroup",
				 NULL);
		return TCL_ERROR;
	    }
	    if (Tcm_FindPg(argv[cntr]) == NULL) {
		Tcl_AppendResult(interp, "unknown priority group \"", argv[cntr],
				 "\"", NULL);
		return TCL_ERROR;
	    }
	    ckfree((char *) objPtr->pg);
	    objPtr->pg = (char *) ckalloc(strlen(argv[cntr]) + 1);
	    strcpy(objPtr->pg, argv[cntr]);
	    Tcm_DeleteFileHandler(objPtr->udp);
	    Tcm_CreateFileHandler(objPtr->pg, objPtr->udp, TK_READABLE,
				  HandleIncomingUdp, (ClientData) objPtr);
	} else {
	    Tcl_AppendResult(interp,
			     "Invalid configuration parameter ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}
	    
	cntr++;
    }
    return TCL_OK;
}



/************************************************************
 *
 * UdpDestModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     keywords as well as creating the Tcl command
 *     for object type creation
 *
 ************************************************************/

int
UdpDestModuleInit (Tcl_Interp *interp)
{
    /*
     * Force Tcm_GetLocalAddress to cache the local host ip address
     */
    Tcm_GetLocalAddress(interp);
    Tcl_CreateCommand(interp, "udpDest", UdpDestCreateCmd,
		      NULL, (void (*)()) NULL);
    Tcl_VarEval(interp, "cminfo set object_info ", "udpDest ",
		"{unicast network dest}", (char *) NULL);

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * HandleIncomingUdp --
 *
 *	Called whenever a packet is available to be read
 *	from the socket.  Retrieves the packet and figures
 *	out how to handle it.
 *
 * Results:
 *	no return value
 *
 * Side effects:
 *	Calls various procdures to store the contents of the
 *	packet in the appropriate place and make a callback
 *	if necessary when a frame is completely received
 *
 *--------------------------------------------------------------
 */
static void
HandleIncomingUdp(clientData, mask)
    ClientData clientData;	/* Actually pointer to a UdpDest */
    int mask;		  	/* Unused */
{
    UdpDest *udpPtr;
    int count;
    struct sockaddr_in addr;
    struct iovec vec[2];
    struct msghdr msghdr;
    int addrLen;
    UdpHeader ph, dummyPh;
    Buffer *bufPtr;

    udpPtr = (UdpDest *) clientData;

    /*
     * Limit the amount of work we do at one time
     */
    count = MAX_SEQ_READS;
    
    while (count-- && (Tdp_SocketIsReady(udpPtr->udp) & TCL_FILE_READABLE)) {
	/*
	 * Peek at the packet header
	 */
	addrLen = sizeof(addr);
	memset((char *)&addr, 0, addrLen);
	
	if (recvfrom(udpPtr->udp, (char *)&ph, sizeof(UdpHeader),
		     MSG_PEEK, (struct sockaddr *)&addr, &addrLen) <= 0) {
	    Tcl_AppendResult(udpPtr->interp, "Error reading socket", NULL);
	    return;
	}

	/*
	 * Convert the fields of the packet header
	 */
	ph.size = ntohl(ph.size);

	/*
	 * Get a buffer for this data.
	 */
	bufPtr = FindUdpBuffer(udpPtr, ph.size);

	/*
	 * Grab the actual data
	 */
	vec[0].iov_base = (char *) &dummyPh;
	vec[0].iov_len = sizeof(UdpHeader);
	vec[1].iov_base = bufPtr->data;
	vec[1].iov_len = ph.size;

	memset((char *)&addr, 0, addrLen);
	msghdr.msg_namelen = sizeof(addr);
	msghdr.msg_name = (char *) &addr;
	msghdr.msg_iov = vec;
	msghdr.msg_iovlen = 2;
#if !defined(_SCO_DS)
	msghdr.msg_accrights = NULL;
	msghdr.msg_accrightslen = 0;
#endif
	recvmsg(udpPtr->udp, &msghdr, 0);
	FinishedBuffer(bufPtr, udpPtr, ph.size);
    }
}

/*
 *--------------------------------------------------------------
 *
 * FindUdpBuffer --
 *
 *	Finds the buffer where the data in this packet will go
 *
 * Results:
 *	A pointer to a buffer structure
 *
 * Side effects:
 *	If the buffer cannot be found (e.g., this is the first
 *	packet in this buffer) then a buffer may be taken from
 *	the free list or allocated if necessary
 *
 *--------------------------------------------------------------
 */
static Buffer *
FindUdpBuffer(udpPtr, size)
    UdpDest *udpPtr;	/* The packet dest we are working with */
    int size;		/* The header from this packet */
{
    FindQuery query;
    ConfigureInfo info;
    Buffer *bufPtr;
    int id;
    
    /*
     * First we look for a buffer that has been allocated but isn't
     * in use right now
     */
    query.unique = 1;
    query.minsize = size;
    query.maxsize = BUFFER_DONT_CARE;
    query.shared = udpPtr->shared;
    query.xAttach = udpPtr->xAttach;
    query.tag = NULL;
    query.state = "free";
    
    id = FindOrCreateBuffer(udpPtr->interp, &query);
    bufPtr = GetBufferFromID(id);
    AttachBuffer (udpPtr->interp, bufPtr, "udpDest:FindUdpBuffer");

    /*
     * Make sure the buffer is properly configured
     */
    info.size = BUFFER_DONT_MODIFY;
    info.shared = BUFFER_DONT_MODIFY;
    info.xAttach = BUFFER_DONT_MODIFY;
    info.state = "filling";
    info.callback = NULL;
    
    if (ConfigureBuffer(udpPtr->interp, bufPtr, &info) != TCL_OK) {
	DestroyBuffer(udpPtr->interp, bufPtr);
	return NULL;
    }
	
    return bufPtr;
}

/*
 *--------------------------------------------------------------
 *
 * FinishedBuffer --
 *
 *	This routine is called whenever we have finished
 *	assembling a buffer.  It puts together the argument
 *	vector and makes the callback as well as noting the
 *	fact that the callback has been made so that it isn't
 *	made again on an EOC packet.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The associated callback for the completed buffer
 *	is called.
 *
 * Parameters:
 *	Buffer *bufPtr	The completed buffer
 *	UdpDest *udpPtr	The packet dest it was received on
 *	int size	Size of buffer
 *
 *--------------------------------------------------------------
 */
static void
FinishedBuffer(bufPtr, udpPtr, size)
    Buffer *bufPtr;
    UdpDest *udpPtr;
    int size;
{
    ScatterBuffer *sb;
    SB_List *sbl;
    char *sbStr;

    /*
     * Set up an SB_List structure for the callback
     */
    sb = (ScatterBuffer *) ckalloc(sizeof(ScatterBuffer));
    sb->count = 1;
    sb->buffers = (Buffer **) ckalloc(sizeof(Buffer *));
    sb->lens = (int *) ckalloc(sizeof(int));
    sb->offsets = (int *) ckalloc(sizeof(int));

    sb->buffers[0] = bufPtr;
    sb->lens[0] = size;
    sb->offsets[0] = 0;

    sbl = (SB_List *) ckalloc(sizeof(SB_List));
    sbl->count = 1;
    sbl->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *));
    sbl->sbs[0] = sb;
    
    /*
     * Make the callback
     */
    if (udpPtr->outCmd[0] == '@') {
	(udpPtr->outCCmd)(udpPtr->outObj, sbl);
    } else {
	sbStr = MakeSBListString(sbl);
	Tcl_VarEval(udpPtr->interp, udpPtr->outCmd, " ",
			sbStr, NULL);
	ckfree((char *) sbStr);
    }

    /*
     * Detach the buffer and free up the memory allocated above.
     */
    DetachBuffer(udpPtr->interp, bufPtr, "udpDest:FinishedBuffer");
    FreeSBList (sbl);
}
