/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * UdpSrc.c --
 *
 *  Introduction -- This file implements the udp source module.
 *	The udp source and udp dest module make a pair.
 *
 *	The basic idea of the udp source/dest is to provide a simple
 *	delivery mechanism for a group of buffers using udp.  Each
 *	buffer is sent as a single UDP datagram, with a header
 *	indicating the buffer size.
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif

#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <unistd.h>
#include <math.h>
#include <tcl.h>
#include <tk.h>
#include <dp.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "cmAtom.h"
#include "cmPacket.h"
#include "cmBuffer.h"
#include "cmBufferlist.h"
#include "cmUtil.h"

#ifdef NO_SENDRECVMSG
#    include "compat/sendrecvmsg.h"
#endif

/*
 * For each packet source, there is one of the following objects
 * maintained.
 */
typedef struct UdpSource {
    /*
     * Section 1. General information
     */
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    DP_SOCKET udp;		/* File descriptor for UDP data socket */
    char *pg;			/* Priority group */ 
    struct sockaddr destAddr;	/* Data address for UDP data socket */
    char *addrInfo;		/* String with <host port> */

    /*
     * Section 2. Information about current buffers to send
     */
    ScatterBuffer **buffer;	/* Array of buffers we're currently sending */
    int numBuf;			/* Number of active items in the array */
    int maxBuf;			/* Number of allocated items in the array */
    
    /*
     * Section 3. Flow control information
     */
    int burstSize;		/* Maximum burst size (in bytes) */
    int bandwidth;		/* Estimated bandwidth */
    double lastSendTime;	/* Time last SendBurst occurred */
    Tk_TimerToken sendTimer;	/* Token for SendBurst callback */
} UdpSource;

#define NO_TOKEN	((Tk_TimerToken) -1)

#ifndef _WINDOWS
#    define min(a,b)	((a)<(b)?(a):(b))
#endif

/* Globals used by this object module. */
static int nextUdpSourceNum = 0;    /* Variable for unique object names. */

/*
 * Prototypes
 */
int			UdpSourceModuleInit(Tcl_Interp *interp);

static int		UdpSourceCreateCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static UdpSource *	CreateNewUdpSource(Tcl_Interp *interp, char *name);
static int		UdpSourceObjCmd(ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv);
static int		UdpSourceInFunc(UdpSource *objPtr, SB_List *data);
static int		DestroyUdpSource(UdpSource *objPtr);
static int		ConfigureUdpSource(UdpSource *objPtr,
			    int argc, char **argv);
static int		AppendBuffers(UdpSource *objPtr, SB_List *sbl);
static int		UdpSourceAttach(UdpSource *objPtr,
			    struct sockaddr *addrPtr);
static void		FreeBufferList(UdpSource *objPtr);
static void		UdpSendBurst(ClientData clientData);


/*
 *--------------------------------------------------------------
 *
 * UdpSourceModuleInit
 *
 * Parameters: 
 *     interp - Tcl interpreter
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 *--------------------------------------------------------------
 */

int
UdpSourceModuleInit (Tcl_Interp *interp)
{
    Tcl_CreateCommand(interp,
		      "udpSrc", UdpSourceCreateCmd,
		      NULL, (void (*)()) NULL);
    Tcl_VarEval(interp,
		"cminfo add object_info udpSrc {unicast}",
		(char *) NULL);
   
    return TCL_OK; 
}


/*
 *--------------------------------------------------------------
 *
 * UdpSourceCreateCmd
 *
 * Parameters: 
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the udpSrc command.
 *     A new object of type UdpSource is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 *--------------------------------------------------------------
 */

static int
UdpSourceCreateCmd (ClientData clientData, Tcl_Interp *interp,
		       int argc, char **argv)
{
    char *new_name;
    UdpSource *new;

    if (argc < 2) {
	Tcl_AppendResult(interp,
			 "wrong # of args, should be: ",
			 argv[0], " parent ?options?",
			 (char *) NULL);
	return TCL_ERROR;
    }

    new_name = ckalloc(strlen(argv[1]) + 16);
    nextUdpSourceNum++;
    sprintf(new_name, "%sps%d", argv[1], nextUdpSourceNum);

    new = CreateNewUdpSource(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }

    Tcl_CreateCommand(interp, new_name, UdpSourceObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureUdpSource(new, argc-2, argv+2);
    }

    Tcl_SetResult (interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * CreateNewUdpSource
 *
 * Parameters: 
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 * 
 * Return: 
 *     Pointer to new object with fields
 *     initialized to default values.
 * 
 * Action: 
 *     Allocates memory for new object and
 *     initializes it.
 *
 *--------------------------------------------------------------
 */

static UdpSource *
CreateNewUdpSource (Tcl_Interp *interp, char *name)
{
    UdpSource *new;
    char *atomName, *suffix = "accept";
    DP_SOCKET udp;

    udp = Tdp_inet_connect ("", 0, 0, 1, 0, 0);
    if (udp == DP_INVALID_SOCKET) {
	Tcl_AppendResult (interp, "Couldn't open udp socket : ",
			  Tdp_Error(interp), (char *)NULL);
	return NULL;
    }

    new = (UdpSource *) ckalloc(sizeof(UdpSource));

    new->interp = interp;
    new->udp = udp;
    new->addrInfo = NULL;
    new->pg = ckalloc(3);
    memcpy(new->pg, "tk", 3);

    memset (&new->destAddr, 0, sizeof(new->destAddr));
    new->buffer = NULL;
    new->numBuf = 0;
    new->maxBuf = 0;

    new->burstSize  = 12000;
    new->lastSendTime = Tcm_ReadSysClock();
    new->sendTimer = NO_TOKEN;

    new->bandwidth = 360000;

    new->name = (char *) ckalloc(strlen(name)+1);
    strcpy(new->name,name);
    
    atomName = ckalloc(strlen(name) + strlen(suffix) + 3);
    sprintf(atomName, "%s.%s", name, suffix);
    Tutil_RegisterCommand(interp, atomName, (ClientData) UdpSourceInFunc,
			  (ClientData) new);
    
    return new;
}


/*
 *--------------------------------------------------------------
 *
 * UdpSourceObjCmd
 *
 * Parameters: 
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 * 
 * Return: 
 *     Standard Tcl return value.
 * 
 * Action: 
 *     This command implements the actual methods of
 *     UdpSource object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 *--------------------------------------------------------------
 */

static int
UdpSourceObjCmd (ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{
    char c;
    int len;
    UdpSource *objPtr = (UdpSource *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " <command> ?arguments?",
			 (char *) NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    switch (c) {
      case 'a':
	/* --------------------- ADD BUFFER --------------------- */
	if (strncmp("addBuffers", argv[1], len) == 0) {
	    SB_List sbl;
	    int res;
	    
	    if (argc != 3) {
		Tcl_AppendResult (interp,
				  "wrong # of args, should be: ",
				  argv[0], " addBuffers <scatter buffer list>",
				  (char *) NULL);
		return TCL_ERROR;
	    }
	    
	    if (ParseSBList(interp, argv[2], &sbl) != TCL_OK) {
		return TCL_ERROR;
	    }
	    
	    res = AppendBuffers(objPtr, &sbl);
	    ckfree((char *) sbl.sbs);
	    return res;
	}
	break;

      case 'c':
	/* ---------------------- CONFIGURE ---------------------- */
	if (strncmp("configure", argv[1], len) == 0) {
	    return ConfigureUdpSource(objPtr, argc-2, argv+2);
	}
	break;
	
      case 'd':
	/* ----------------------- DESTROY ----------------------- */
	if (strncmp("destroy", argv[1], len) == 0) {
	    return DestroyUdpSource(objPtr);
	}
	break;
    }

usage:
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ", argv[1],
		     (char *) NULL);
    Tcl_AppendResult(interp,
		"\n<command> can be one of:\n",
		"  configure ?-slot ?value??\n",
		"  destroy\n",
		"  addBuffers <ScatterBufferList>\n",
		(char *) NULL);
    return TCL_ERROR;
}


/*
 *--------------------------------------------------------------
 *
 * UdpSourceInFunc
 *
 *	The routine made available to other objects via
 *	the Atom registered as the UdpSource
 *	InFunction.
 *
 *	Takes a list of buffers and creates a new cycle
 *	consisting of those buffers.
 *
 * Parameters: 
 *	UdpSource *objPtr	The packet source receiving data
 *	BufferList buffers	The buffers containing the data
 * 
 * Return: 
 *	Standard Tcl result
 * 
 * Action: 
 *	Calls the routines AppendBuffers
 *
 *--------------------------------------------------------------
 */

static int 
UdpSourceInFunc (UdpSource *objPtr, SB_List *data)
{
    FreeBufferList (objPtr);
    AppendBuffers (objPtr, data);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * DestroyUdpSource
 *
 * Parameters: 
 *     objPtr - pointer to object.
 * 
 * Return: 
 *     Standard Tcl Result
 * 
 * Action: 
 *     Destroys previously created object.
 *
 *--------------------------------------------------------------
 */

static int 
DestroyUdpSource (UdpSource *objPtr)
{
    /* First delete Tcl command. */

    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->numBuf != 0) {
	FreeBufferList(objPtr);
	ckfree((char *) objPtr->buffer);
    }

    closesocket(objPtr->udp);
    if (objPtr->sendTimer != NO_TOKEN) {
	Tcm_DeleteTimerHandler(objPtr->sendTimer);
	objPtr->sendTimer = NO_TOKEN;
    }

    if (objPtr->addrInfo != NULL) {
	ckfree (objPtr->addrInfo);
    }
    if (objPtr->name != NULL) {
	ckfree((char *) objPtr->name);
    }

    ckfree((char *) objPtr->pg);
    ckfree((char *) objPtr);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * ConfigureUdpSource
 *
 * Parameters: 
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 * 
 * Return: 
 *     Standard TCL return value
 * 
 * Action: 
 *     Does the usual configuration thing
 *
 *--------------------------------------------------------------
 */

static int
ConfigureUdpSource (UdpSource *objPtr, int argc, char **argv)
{
    Tcl_Interp *interp;
    int cntr;

    interp = objPtr->interp;

    if (argc == 0) {
	char tmp[50];
	
	sprintf(tmp, "%d", objPtr->burstSize);
	Tcl_AppendResult(interp,
			 "{-burstSize {35000} {", tmp, "}} ", NULL);
	sprintf(tmp, "%d", objPtr->bandwidth);
	Tcl_AppendResult(interp,
			 "{-bandwidth {360000} {", tmp, "}} ", NULL);

	if (objPtr->addrInfo != NULL) {
	    Tcl_AppendResult(interp, "{-dest ", "{} {", objPtr->addrInfo,
			     "}} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{-dest ", "{} {}} ", (char *) NULL);
	}
	Tcl_AppendResult(interp, "{-priorityGroup {tk} {", objPtr->pg, "}} ",
			 (char *) NULL);
	return TCL_OK;
    }
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 'b') && (strncmp("-burstSize",argv[0],len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->burstSize);
	    Tcl_AppendResult(interp,
			     "-burstSize ", "{35000} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'b') && (strncmp("-bandwidth", argv[0], len) == 0)) {
	    char str[50];
	    sprintf(str, "{%d}", objPtr->bandwidth);
	    Tcl_AppendResult(interp,
			     "-bandwidth ", "{360000} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-dest", argv[0], len) == 0)) {
	    if (objPtr->addrInfo != NULL) {
		Tcl_AppendResult(interp, "-dest ", "{} {", objPtr->addrInfo,
				 "}", (char *) NULL);
	    } else {
		Tcl_AppendResult(interp, "-dest ", "{} {}", (char *) NULL);
	    }
	    return TCL_OK;
	} else if ((c == 'p') && (strncmp("-priorityGroup", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-priorityGroup {tk} {",
			     objPtr->pg, "}", NULL);
	    return TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "bad configuration option: ",
			     argv[0], (char *) NULL);
	    return TCL_ERROR;
	}

    }
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;

	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	if ((c == 'b') && (strncmp("-burstSize", argv[cntr], len) == 0)) {
	    int newBurstSize;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp,
				 "Missing parameter for -maxBustSize",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newBurstSize) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->burstSize = newBurstSize;
	} else if ((c == 'd') && (strncmp("-dest", argv[cntr], len) == 0)) {
	    struct sockaddr *addrPtr;
	    int res;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing parameter for -dest",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    /*
	     * Get new address
	     */
	    res = Tcl_VarEval (interp, "eval dp_address create ", argv[cntr],
				NULL);
	    if (res != TCL_OK) {
		Tcl_ResetResult (interp);
		Tcl_AppendResult (interp, "Error: invalid destination \"",
			argv[cntr], "\"", NULL);
		return TCL_ERROR;
	    }
	    addrPtr = Tdp_FindAddr (interp->result);
	    UdpSourceAttach (objPtr, addrPtr);
	    Tcl_VarEval (interp, "dp_address delete ", interp->result, NULL);
	    if (objPtr->addrInfo != NULL) {
		ckfree (objPtr->addrInfo);
		objPtr->addrInfo = NULL;
	    }
	    objPtr->addrInfo = ckalloc(strlen(argv[cntr])+1);
	    strcpy (objPtr->addrInfo, argv[cntr]);

	} else if ((c == 'b') &&
		   (strncmp("-bandwidth", argv[cntr], len) == 0)) {
	    int newBandwidth;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing parameter for -bandwidth",
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    if (Tcl_GetInt(interp, argv[cntr], &newBandwidth) != TCL_OK) {
		Tcl_AppendResult(interp,
				 "bad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->bandwidth = newBandwidth;
	} else if ((c == 'p') && (strncmp("-priorityGroup", argv[cntr], len) == 0)){
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Missing paramter for -priorityGroup",
				 NULL);
		return TCL_ERROR;
	    }
	    if (Tcm_FindPg(argv[cntr]) == NULL) {
		Tcl_AppendResult(interp, "unknown priority group \"", argv[cntr],
				 "\"", NULL);
		return TCL_ERROR;
	    }
	    ckfree((char *) objPtr->pg);
	    objPtr->pg = (char *) ckalloc(strlen(argv[cntr]) + 1);
	    strcpy(objPtr->pg, argv[cntr]);
	} else {
	    Tcl_AppendResult(interp,
			     "Invalid configuration parameter ",
			     argv[cntr], NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * UdpSourceAttach --
 *
 *	
 *      This function provide a C language binding to the "attach"
 *	udpSrc function.  It attaches a new UdpSource object to
 *	the socket associated with the UdpSource.
 *
 * Results:
 *	1 for success, 0 on failure.
 *
 * Side effects:
 *	An entry is made in the udpSrcs table and a file handler
 *	is created for the udp socket to handle resend requests.
 *
 * Parameters:
 *	Tcl_Interp *interp		Interp associated with CMC
 *	struct sockaddr *addrPtr	Destination address to use
 *
 *--------------------------------------------------------------
 */
static int
UdpSourceAttach (objPtr, addrPtr)
    UdpSource *objPtr;
    struct sockaddr *addrPtr;
{
    memcpy (&objPtr->destAddr, (char *)addrPtr, sizeof(struct sockaddr));
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * AppendBuffers --
 *
 *	Given a scatter buffer list, append all its component
 *	scatter buffers to the list for the current cycle
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The UdpSource will attempt to transmit the data passed in
 *	to the destination.
 *
 * Parameters:
 *	UdpSource *objPtr	The source to add buffers to
 *	SB_List *sbl;		The list of scatter buffer
 *	int copy;		Should we copy the vectors?  If not,
 *				they become property of the UdpSource 
 *				and it will free them when it no longer
 *				needs them.
 *
 *--------------------------------------------------------------
 */
static int
AppendBuffers (UdpSource *objPtr, SB_List *sbl)
{
    int t, buf;
    ScatterBuffer *buffer;

    /*
     * Initialize room for our cache of buffers, if necessary
     */
    if (objPtr->maxBuf == 0) {
	objPtr->maxBuf = 32;
	objPtr->buffer = (ScatterBuffer **)ckalloc(objPtr->maxBuf*
						   sizeof(ScatterBuffer *));
    }
    
    /*
     * Make a copy of the scatter buffer
     */
    for (t=0; t<sbl->count; t++) {
	if (objPtr->numBuf == objPtr->maxBuf) {
	    objPtr->maxBuf += 32;
	    objPtr->buffer = (ScatterBuffer **)ckrealloc((char *)objPtr->buffer,
				  objPtr->maxBuf*sizeof(ScatterBuffer *));
	}

	buffer = CopyScatterBuffer(sbl->sbs[t]);
	objPtr->buffer[objPtr->numBuf++] = buffer;
	for (buf=0; buf<buffer->count; buf++) {
	    AttachBuffer(objPtr->interp, buffer->buffers[buf],
			 "udpSrc:AppendBuffers");
	}
    }
    if (objPtr->sendTimer == NO_TOKEN) {
	UdpSendBurst((ClientData)objPtr);
    }
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * FreeBufferList --
 *
 *	Frees the buffers in the buffer list associated
 *	with a given UdpSource object.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	You should never reference the buffer chain again
 *
 * Parameters:
 *	UdpSource *objPtr	The packet source object
 *
 *--------------------------------------------------------------
 */
static void
FreeBufferList (UdpSource *objPtr)
{
    int i, j, numBuf;
    ScatterBuffer *buf, **buffers;
    
    numBuf = objPtr->numBuf;
    buffers = objPtr->buffer;

    for (i=0; i<numBuf; i++) {
	buf = buffers[i];
	
	if (buf != NULL) {
	    for (j=0; j<buf->count; j++) {
		DetachBuffer(objPtr->interp, buf->buffers[j],
			     "udpSrc:FreeBufferList");
	    }
	    FreeScatterBuffer (buf);
	    buffers[i] = NULL;
	}
    }
    objPtr->numBuf = 0;
}


/*
 *--------------------------------------------------------------
 *
 * UdpSendBurst --
 *
 *	This function is called in response to a timer event.
 *	It sends a burst of packets out on the udp socket, then
 *	schedules another timer event to call this function again
 *	for the next burst.
 *
 * Results:
 *	None
 *
 * Parameters:
 *	ClientData clientData	Pointer to the packet source.
 *
 *--------------------------------------------------------------
 */
static void
UdpSendBurst (clientData)
    ClientData clientData;
{
    UdpSource *udpSrc = (UdpSource *)clientData;
    UdpHeader udpHdr;
    ScatterBuffer *sbPtr;
    int i, j;
    int bytesToSend;
    int bytesSent;
    int v;
    int estBandwidth;
    int delay;
    struct msghdr msg;
    struct iovec iov[256];
    struct timeval tv;
    double currTime;
    int udpSize;
    DP_SOCKET udp;

    udpSrc->sendTimer = NO_TOKEN;
    udp = udpSrc->udp;
    bytesSent = 0;
    estBandwidth = udpSrc->bandwidth;

    /*
     * Record timing info and calculate how many bytes to send.  We need
     * to do this because we may not get called at exactly the right time
     * by the timer handlers due to the granularity of the clock or other
     * event handling.
     */
    (void) gettimeofday(&tv, NULL);
    currTime = tv.tv_sec + tv.tv_usec / 1000000.0;
    bytesToSend = (currTime - udpSrc->lastSendTime)*estBandwidth;

    /*
     * Put a cap on the size on any single burst so we don't overflow the
     * receiver's buffers.
     */
    if (bytesToSend > udpSrc->burstSize) {
	bytesToSend = udpSrc->burstSize;
    }

    /*
     * Init base fields of msg for sendmsg
     */
    msg.msg_name = (caddr_t)&udpSrc->destAddr;
    msg.msg_namelen = sizeof(udpSrc->destAddr);
    msg.msg_iov = iov;
    msg.msg_iovlen = 0;
#if !defined(_SCO_DS)
    msg.msg_accrights = NULL;
    msg.msg_accrightslen = 0;
#endif
    iov[0].iov_base = (caddr_t)&udpHdr;
    iov[0].iov_len = sizeof(UdpHeader);
    msg.msg_iovlen = 1;

    /*
     * Go through all the buffers in the send buffer, sending
     * each unsent buffers, up to burstSize bytes.
     */

    for (i=0; i<udpSrc->numBuf && bytesSent < bytesToSend; i++) {

	/*
	 * Check if we should skip the packet (already sent).
	 */
	if (udpSrc->buffer[i] == NULL) {
	    continue;
	}

	/*
	 * Set up the iov structure
	 */
	sbPtr = udpSrc->buffer[i];
	udpSize = 0;
	v = 1;

	/*
	 * Build up a scatter/gather vector for the scatter buffer.
	 */
	for (j=0; j<sbPtr->count; j++) {
	    Buffer *bufPtr;
	    int offset;

	    bufPtr = sbPtr->buffers[j];
	    offset = sbPtr->offsets[j];
	    iov[v].iov_base = bufPtr->data+offset;
	    iov[v].iov_len = sbPtr->lens[j];
	    udpSize += iov[v].iov_len;
	    v++;
	}

	/*
	 * Fill in packet header
	 */
	udpHdr.size = htonl(udpSize);

	msg.msg_iovlen = v;
	bytesSent += sendmsg(udp, &msg, 0);

	for (j=0; j<sbPtr->count; j++) {
	    DetachBuffer(udpSrc->interp, sbPtr->buffers[j],
			 "udpSrc:UdpSendBurst");
	}
	FreeScatterBuffer (sbPtr);
	udpSrc->buffer[i] = NULL;
    }
    udpSrc->lastSendTime = currTime;

    /*
     * Schedule next send burst
     */
    delay = 1000*(udpSrc->burstSize + bytesSent - bytesToSend)/estBandwidth;
    udpSrc->sendTimer = Tcm_CreateTimerHandler(udpSrc->pg, delay,
					       UdpSendBurst, clientData);
}
