/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    include <windows.h>
#    include <winsock.h>
#endif /* _WINDOWS */

#include <stdio.h>
#include <math.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#include <tcl.h>
#include <tk.h>
#include "cmAuPlay.h"

#ifdef WIN_AUDIO
#include "cmBufferlist.h"
#include "cmCookies.h"
#include "cmAuFile.h"
#include "cmLTS.h"
#include "cmAuUtil.h"


typedef struct WinAudioBuffer {
    HGLOBAL hBuf;		/* Handle to data */
    HGLOBAL hWaveHdr;		/* Handle to header */
    LPWAVEHDR lpWaveHdr;	/* The header itself */
    int	free_flag;		/* Should this buffer be freed? */
    struct WinAudioBuffer *next;
} WinAudioBuffer;

typedef struct WinAudioDevice {
    HWAVEOUT device;		/* Device handle */
    UINT deviceId;		/* Identifier of the device */
    WAVEOUTCAPS devCaps;	/* Device capabilities */
    PCMWAVEFORMAT pcmWaveFmt;	/* How device is currently open */
    unsigned int otherFmts;	/* What other formats are supported */
    SparcAuBufferHeader currFmt;/* The last format we changed to */
    int resampling;		/* Are we resampling to new rate? */
    int errorCode;		/* Current error code */
    int atToken;		/* Token for atEvent for playback when
				 * audio device queue is empty */
    double lastStartTime;	/* Last time the SparcAuStartPlay()
				 * function was called. */
    double emptyTime;		/* System time when the data currently
				 * in the stream queue should flush */
    WinAudioBuffer *auBufs;	/* List of audio buffers in use */
} WinAudioDevice;

/*
 * Interface routines
 */
static int WinAudioDeviceInit(SpeakerRsrc *objPtr);
static int WinAudioDeviceReinit(SpeakerRsrc *objPtr);
static int WinAudioDeviceShutdown(SpeakerRsrc *objPtr);
static int WinAudioDeviceUnready(SpeakerRsrc *objPtr);
static int WinAudioDeviceLTSChanged(SpeakerRsrc *objPtr);
static int WinAudioDevicePlay(CBInfo *cbPtr);
static char *WinAudioDeviceError(SpeakerRsrc *objPtr, int errNum);
static MMRESULT WinAudioDeviceSetPlayParams (WinAudioDevice *dev,
			     int sample_rate, int channels, int encoding);
static void WinAudioStartPlay (CBInfo *cbPtr);
    
/*
 * Internal routines
 */
static void WinAudioDeviceSetGain(WinAudioDevice *dev, double gain);


#define SKEW		0.005
#define SYNC_TIME	300.0


#define FMT_8000_1	0x01
#define FMT_8000_2	0x02
#define FMT_48000_1	0x04
#define FMT_48000_2	0x08

/*
 * Error codes: start at 100000 so we don't conflict with Windows errors
 */
#define CANT_OPEN_HANDLE_BUSY_ERROR	100000
#define CANT_OPEN_ERROR			100001
#define BAD_DEVICE_ERROR		100002
#define	BAD_SAMPLE_RATE_ERROR		100003
#define NO_AUDIO_DEVICES_ERROR		100004
#define NO_MATCHING_DEVICE_ERROR	100005
#define NOT_OPEN_ERROR			100006
#define TOO_EARLY_ERROR			100007
#define BAD_FORMAT_ERROR		100008
#define TOO_LAZY_ERROR			100009
#define NO_MEMORY_ERROR			100010
#define MEMORY_LOCK_ERROR		100011
#define UNSUPPORTED_FORMAT_ERROR	100012
#define FORMAT_ERROR			100013

/*
 * -------------------------------------------------------
 *
 * GetWinAudioDevice 
 *
 *	Allocates a new SpeakerDevice structure, filled
 *	in to handle WinAudio and return it.
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetWinAudioDevice(SpeakerRsrc *objPtr)
{
    SpeakerDevice *sdev;
    WinAudioDevice *dev;
    
    dev = (WinAudioDevice *) ckalloc(sizeof(WinAudioDevice));
    if (dev == NULL) {
	return NULL;
    }
    dev->device = NULL;
    dev->deviceId = (UINT) -1;
    dev->auBufs = NULL;
    memset(&dev->devCaps, 0, sizeof(dev->devCaps));
    
    sdev = (SpeakerDevice *) ckalloc(sizeof(SpeakerDevice));
    sdev->deviceData =	(ClientData) dev;
    sdev->initProc =	WinAudioDeviceInit;
    sdev->reinitProc =	WinAudioDeviceReinit;
    sdev->shutdownProc =WinAudioDeviceShutdown;
    sdev->unreadyProc =	WinAudioDeviceUnready;
    sdev->ltsProc =	WinAudioDeviceLTSChanged;
    sdev->playProc =	WinAudioDevicePlay;
    sdev->errorProc =	WinAudioDeviceError;
    
    return sdev;
}

/*
 *--------------------------------------------------------
 * WinAudioFreeBuffers --
 *
 *	Free all buffers whose free_flag is set.  Need to 
 *	unprepare the buffers before freeing them.
 *--------------------------------------------------------
 */
void
WinAudioFreeBuffers(WinAudioDevice *dev)
{
    WinAudioBuffer *buf, *prev, *next;

    prev = NULL;
    for (buf = dev->auBufs; buf != NULL; buf = next) {
	next = buf->next;
	if (buf->free_flag) {
	    waveOutUnprepareHeader(dev->device, buf->lpWaveHdr,
				   sizeof(WAVEHDR));
	    GlobalUnlock(buf->hWaveHdr);
	    GlobalFree(buf->hWaveHdr);
	    GlobalUnlock(buf->hBuf);
	    GlobalFree(buf->hBuf);

	    if (prev == NULL) {
		dev->auBufs = next;
	    } else {
		prev->next = next;
	    }
	    ckfree((char *) buf);
	} else {
	    prev = buf;
	}
    }
}

/*
 *--------------------------------------------------------
 * WinAudioCallback --
 *
 *	Callback from Windows when a buffer is completed,
 *	the device is opened, or the device is closed.
 *
 * Notes:
 *	This callback routine call can only make very limited
 *	calls.  The only Win32 functions that it can call are
 *	PostMessage(), timeGetSystemTime(), timeGetTime(),
 *	timeSetEvent(), timeKillEvent(), midiOutShortMsg(),
 *	midiOutLongMsg(), and OutputDebugStr().  It can't
 *	allocate or deallocate memory.  What we do is this:
 *	when the global buffer is allocated, we also allocate
 *	a header that we put on the free list, but we don't
 *	set the free flag.  This routine runs through the
 *	free list and sets the free flag for the buffer that
 *	can now be freed.  At shutdown time as well, reset time,
 *	and before we allocate a new buffer, we run through
 *	the list and free up anything with the free flag set.
 *
 * Parameters:
 *	device: 	Device Handle
 *	msg:		The message to handle
 *	dwInstance:	Parameter we passed in, WinAudioDevice *
 *	dwParam1:	The header pointer.
 *	dwParam2:	Unused.
 *--------------------------------------------------------
 */
void CALLBACK
WinAudioCallback(HWAVEOUT device, UINT msg, DWORD dwInstance,
		 DWORD dwParam1, DWORD dwParam2)
{
    WinAudioDevice *dev;
    LPWAVEHDR lpWaveHdr;
    WinAudioBuffer *buf;

    if (msg != MM_WOM_DONE) {
	return;
    }

    dev = (WinAudioDevice *) dwInstance;
    lpWaveHdr = (LPVOID) dwParam1;

    buf = (WinAudioBuffer *) lpWaveHdr->dwUser;
    buf->free_flag = 1;
}

/*
 * -------------------------------------------------------
 *
 * WinAudioDeviceInit
 *
 * Gets the WinAudioDevice ready to play sound
 *
 * -------------------------------------------------------
 */

static int
WinAudioDeviceInit(SpeakerRsrc *objPtr) {
    WinAudioDevice *dev;
    MMRESULT mmret;
    UINT nDevs, i;
    DWORD fmt;
    int channels;
    int rate;
    WORD sampleSize;
    PCMWAVEFORMAT pcmWaveFmt;
    
    dev = (WinAudioDevice *) objPtr->devicePtr->deviceData;
    
    if (dev->device != NULL) {
	mmret = WinAudioDeviceShutdown(objPtr);
	if (mmret) {
	    return mmret;
	}
    }

    nDevs = waveOutGetNumDevs();
    if (nDevs == 0) {
	return NO_AUDIO_DEVICES_ERROR;
    }

    dev->otherFmts = 0;
    /*
     * Since we don't yet know what the data format is going to be,
     * lets look and see if there is any decent format supported by
     * the driver/board combinations.  If there is, we can do whatever
     * format conversion is necessary later.  Check to see if the device
     * supports 8000 samples per second.  Since .au format typically comes
     * in this rate, it would be good if the device could support this
     * directly so we don't have to do any rate conversion unless necessary.
     */

    for (i = 0; i < nDevs; i++) {
	mmret = waveOutGetDevCaps(i, &dev->devCaps, sizeof(dev->devCaps));
	if (mmret != 0) {
	    continue;
	}
	fmt = dev->devCaps.dwFormats;

	channels = (((fmt & WAVE_FORMAT_1M08) | (fmt & WAVE_FORMAT_1M16) |
		     (fmt & WAVE_FORMAT_2M08) | (fmt & WAVE_FORMAT_2M16) |
		     (fmt & WAVE_FORMAT_4M08) | (fmt & WAVE_FORMAT_4M16))
		    != 0) ? 1 : 0;

	if (channels == 0) {
	    channels = (((fmt & WAVE_FORMAT_1S08) | (fmt & WAVE_FORMAT_1S16) |
			 (fmt & WAVE_FORMAT_2S08) | (fmt & WAVE_FORMAT_2S16) |
			 (fmt & WAVE_FORMAT_4S08) | (fmt & WAVE_FORMAT_4S16))
			!= 0) ? 2 : 0;
	}

	sampleSize = (((fmt & WAVE_FORMAT_1M08) | (fmt & WAVE_FORMAT_2M08) |
		       (fmt & WAVE_FORMAT_4M08) | (fmt & WAVE_FORMAT_1S08) |
		       (fmt & WAVE_FORMAT_2S08) | (fmt & WAVE_FORMAT_4S08))
		      != 0) ? 1: 0;
	if (!sampleSize) {
	    sampleSize = (((fmt & WAVE_FORMAT_1M16) |(fmt & WAVE_FORMAT_2M16) |
			   (fmt & WAVE_FORMAT_4M16) |(fmt & WAVE_FORMAT_1S16) |
			   (fmt & WAVE_FORMAT_2S16) |(fmt & WAVE_FORMAT_4S16))
			  != 0) ? 2 : 0;
	}

	dev->pcmWaveFmt.wf.wFormatTag = WAVE_FORMAT_PCM;
	dev->pcmWaveFmt.wf.nChannels = channels;
	if ((fmt & WAVE_FORMAT_1M08) || (fmt & WAVE_FORMAT_1S08)) {
	    rate = dev->pcmWaveFmt.wf.nSamplesPerSec = 11025;
	} else if ((fmt & WAVE_FORMAT_2M08) || (fmt & WAVE_FORMAT_2S08)) {
	    rate = dev->pcmWaveFmt.wf.nSamplesPerSec = 22050;
	} else if ((fmt & WAVE_FORMAT_4M08) || (fmt & WAVE_FORMAT_4S08)) {
	    rate = dev->pcmWaveFmt.wf.nSamplesPerSec = 44100;
	}
	dev->pcmWaveFmt.wf.nAvgBytesPerSec = rate * channels * sampleSize;
	dev->pcmWaveFmt.wf.nBlockAlign = sampleSize * channels;
	dev->pcmWaveFmt.wBitsPerSample = 8 * sampleSize;

	dev->deviceId = i;

	/* Check for 8000 Samples/sec, 8 bits per sample */
	pcmWaveFmt.wf.wFormatTag = WAVE_FORMAT_PCM;
	pcmWaveFmt.wf.nChannels = 1;
	pcmWaveFmt.wf.nSamplesPerSec = 8000;
	pcmWaveFmt.wf.nAvgBytesPerSec = 8000;
	pcmWaveFmt.wf.nBlockAlign = 1;
	pcmWaveFmt.wBitsPerSample = 8;

	mmret = waveOutOpen(NULL, i, &pcmWaveFmt.wf,
			    (DWORD) NULL, (DWORD) NULL, WAVE_FORMAT_QUERY);
	if (mmret == 0) {
	    dev->otherFmts |= FMT_8000_1;
	}

	/* Check for 8000 Samples/sec, 16 bits per sample */
	pcmWaveFmt.wf.wFormatTag = WAVE_FORMAT_PCM;
	pcmWaveFmt.wf.nChannels = 1;
	pcmWaveFmt.wf.nSamplesPerSec = 8000;
	pcmWaveFmt.wf.nAvgBytesPerSec = 16000;
	pcmWaveFmt.wf.nBlockAlign = 2;
	pcmWaveFmt.wBitsPerSample = 16;

	mmret = waveOutOpen(NULL, i, &pcmWaveFmt.wf,
			    (DWORD) NULL, (DWORD) NULL, WAVE_FORMAT_QUERY);
	if (mmret == 0) {
	    dev->otherFmts |= FMT_8000_2;
	}

	/* If we've got 8000 Hz, Mono, 8 bits per sample, use it */
	/* XXX: Try 16 bits per sample first when the rest of this
	 * module supports it
	 */
	if (dev->otherFmts & FMT_8000_1) {
	    dev->pcmWaveFmt.wf.wFormatTag = WAVE_FORMAT_PCM;
	    dev->pcmWaveFmt.wf.nChannels = 1;
	    dev->pcmWaveFmt.wf.nSamplesPerSec = 8000;
	    dev->pcmWaveFmt.wf.nAvgBytesPerSec = 8000;
	    dev->pcmWaveFmt.wf.nBlockAlign = 1;
	    dev->pcmWaveFmt.wBitsPerSample = 8;
	} 
	mmret = waveOutOpen(&dev->device, i, &dev->pcmWaveFmt.wf,
#if 0
			    (DWORD) WinAudioCallback, (DWORD) dev,
			    CALLBACK_FUNCTION
#else
			    (DWORD) NULL, (DWORD) NULL, 0
#endif
			    );
	if (mmret != 0) {
	    if (dev->otherFmts != 0) {
		/* XXX: If this supports 8K/s, we should use it */
	    }
	    dev->device = NULL;
	    dev->otherFmts = 0;
	    continue;
	}
	dev->currFmt.samplingRate = dev->pcmWaveFmt.wf.nSamplesPerSec;
	dev->currFmt.channelCount = dev->pcmWaveFmt.wf.nChannels;
	dev->currFmt.dataFormat = (sampleSize == 1) ? SUN_LIN_8 : SUN_LIN_16;
	break;
    }
    if (dev->device == NULL) {
	return NO_MATCHING_DEVICE_ERROR;
    }
    dev->errorCode = 0;
    dev->atToken = 0;
    dev->lastStartTime = 0.0;
    dev->emptyTime = 0.0;
    WinAudioDeviceSetGain(dev, objPtr->gain);
    
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * WinAudioDeviceReinit
 *
 * Re-initialize the device
 *
 * -------------------------------------------------------
 */

static int
WinAudioDeviceReinit(SpeakerRsrc *objPtr) {
    WinAudioDevice *dev;
    MMRESULT mmret;

    dev = (WinAudioDevice *) objPtr->devicePtr->deviceData;
    if (dev->device == NULL) {
	return 0;
    }

    WinAudioFreeBuffers(dev);

    mmret = waveOutReset(dev->device);
    if (mmret != 0) {
	if (mmret == MMSYSERR_HANDLEBUSY) {
	    return CANT_OPEN_HANDLE_BUSY_ERROR;
	} else {
	    return mmret;
	}
    }

    WinAudioDeviceSetGain((WinAudioDevice *) objPtr->devicePtr->deviceData,
			  objPtr->gain);
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * WinAudioDeviceShutdown
 *
 * Silences the speaker and closes the device.
 *
 * -------------------------------------------------------
 */

static int
WinAudioDeviceShutdown(SpeakerRsrc *objPtr)
{
    WinAudioDevice *dev;
    MMRESULT mmret;
    
    dev = (WinAudioDevice *) objPtr->devicePtr->deviceData;
    
    WinAudioFreeBuffers(dev);

    if (dev->device != NULL) {
	mmret = waveOutReset(dev->device);
	if (mmret != 0) {
	    return mmret;
	}

	mmret = waveOutClose(dev->device);
	if (mmret != 0) {
	    return mmret;
	}
	dev->device = NULL;
    }
    
    return 0;
}

/*
 * -------------------------------------------------------
 *
 * WinAudioDeviceUnready
 *
 * Silences the speaker when the device is unreadied
 *
 * -------------------------------------------------------
 */

static int
WinAudioDeviceUnready(SpeakerRsrc *objPtr)
{
    WinAudioDevice *dev;
    MMRESULT mmret;
    
    dev = (WinAudioDevice *) objPtr->devicePtr->deviceData;
    
    if (dev->device) {
	mmret = waveOutReset(dev->device);
	if (mmret != 0) {
	    return mmret;
	}
    }

    return 0;
}

/*
 * -------------------------------------------------------
 *
 * WinAudioDeviceLTSChanged
 *
 *	Notify the WinAudioDevice that the LTS has changed
 *
 * -------------------------------------------------------
 */

static int
WinAudioDeviceLTSChanged (SpeakerRsrc *objPtr)
{
    WinAudioDevice *dev;
    MMRESULT mmret;
    
    dev = (WinAudioDevice *) objPtr->devicePtr->deviceData;
    if (dev->atToken != 0) {
	Tcm_DeleteAtHandler(dev->atToken);
	dev->emptyTime = 0;
	dev->atToken = 0;
    }
#if 0
    if (dev->stopToken != 0) {
	Tcm_DeleteAtHandler(dev->stopToken);
	dev->stopToken = 0;
    }
#endif

    if (dev->device) {
	mmret = waveOutReset(dev->device);
	if (mmret != 0) {
	    return mmret;
	}
    }

    WinAudioFreeBuffers(dev);
    return 0;
}

/*
 *----------------------------------------------------------------------
 * WinAudioWriteData --
 *
 *	Data actually gets written to the device.  Need to follow all
 *	the rules for audio buffer handling.
 *----------------------------------------------------------------------
 */
static MMRESULT
WinAudioWriteData(WinAudioDevice *dev, unsigned char *data, int len)
{
    int inSampleSize, outSampleSize, i;
    unsigned short sample;
    unsigned char sampleByte;
    int allocSize;
    int fmt;
    HGLOBAL hBuf, hWaveHdr;
    LPVOID lpBuf;
    LPWAVEHDR lpWaveHdr;
    WinAudioBuffer *bufInfo;
    UINT wResult;

    WinAudioFreeBuffers(dev);

    outSampleSize = dev->pcmWaveFmt.wBitsPerSample / 8;

    fmt = dev->currFmt.dataFormat;
    inSampleSize = (fmt == SUN_ULAW || fmt == SUN_LIN_8) ? 1 : 2;
    allocSize = len * outSampleSize / inSampleSize;

    hBuf = GlobalAlloc(GMEM_MOVEABLE|GMEM_SHARE, allocSize);
    if (!hBuf) {
	return NO_MEMORY_ERROR;
    }

    lpBuf = GlobalLock(hBuf);
    if (!lpBuf) {
	GlobalFree(hBuf);
	return MEMORY_LOCK_ERROR;
    }

    hWaveHdr = GlobalAlloc(GMEM_MOVEABLE|GMEM_SHARE, (DWORD) sizeof(WAVEHDR));
    if (hWaveHdr == NULL) {
	GlobalUnlock(hBuf);
	GlobalFree(hBuf);
	return NO_MEMORY_ERROR;
    }

    lpWaveHdr = (LPWAVEHDR) GlobalLock(hWaveHdr);
    if (lpWaveHdr == NULL) {
	GlobalFree(hWaveHdr);
	GlobalUnlock(hBuf);
	GlobalFree(hBuf);
	return NO_MEMORY_ERROR;
    }

    lpWaveHdr->lpData = lpBuf;
    lpWaveHdr->dwBufferLength = allocSize;
    lpWaveHdr->dwFlags = 0L;
    lpWaveHdr->dwLoops = 0L;

    bufInfo = (WinAudioBuffer *) ckalloc(sizeof(WinAudioBuffer));
    bufInfo->hBuf = hBuf;
    bufInfo->hWaveHdr = hWaveHdr;
    bufInfo->lpWaveHdr = lpWaveHdr;
    bufInfo->free_flag = 0;
    bufInfo->next = dev->auBufs;
    dev->auBufs = bufInfo;

    lpWaveHdr->dwUser = (DWORD) bufInfo;

    /* Time to do any necessary conversion */
    if (dev->currFmt.dataFormat == SUN_ULAW) {
	if (outSampleSize == 1) {
	    unsigned char *ip, *op;

	    for (i = 0, op = lpBuf, ip = data; i < len; i++, ip++, op++) {
		/* XXX: Need conversion to 8 bits, 128 centered */
		/* Add this to auutil.c for increase speed */
		sample = tcm_UlawToLinearTable[*ip];
		sampleByte = (unsigned char) (sample >> 8);
		sampleByte ^= 128;
		*op = sampleByte;
	    }
	} else /* if (outSampleSize == 2) */ {
	    unsigned char *ip;
	    unsigned short *op;

	    for (i = 0, op = (unsigned short *) lpBuf, ip = data;
		 i < len; i++, ip++, op++)
	    {
		*op = tcm_UlawToLinearTable[*ip];
	    }
	}

    } else if (dev->currFmt.dataFormat == SUN_LIN_8) {
	if (outSampleSize == 1) {
	    memcpy(lpBuf, data, len);
	} else /* if (outSampleSize == 2) */ {
	    /* XXX: Write the conversion */
	    GlobalUnlock(hWaveHdr);
	    GlobalFree(hWaveHdr);
	    GlobalUnlock(hBuf);
	    GlobalFree(hBuf);
	    return UNSUPPORTED_FORMAT_ERROR;
	}

    } else if (dev->currFmt.dataFormat == SUN_LIN_16) {
	if (outSampleSize == 2) {
	    memcpy(lpBuf, data, len);
	} else /* if (outSampleSize == 1) */ {
	    /* XXX: Write the conversion */
	    GlobalUnlock(hWaveHdr);
	    GlobalFree(hWaveHdr);
	    GlobalUnlock(hBuf);
	    GlobalFree(hBuf);
	    return UNSUPPORTED_FORMAT_ERROR;
	}

    } else {
	GlobalUnlock(hWaveHdr);
	GlobalFree(hWaveHdr);
	GlobalUnlock(hBuf);
	GlobalFree(hBuf);
	return UNSUPPORTED_FORMAT_ERROR;
    }

    waveOutPrepareHeader(dev->device, lpWaveHdr, sizeof(WAVEHDR));
    wResult = waveOutWrite(dev->device, lpWaveHdr, sizeof(WAVEHDR));
    if (wResult != 0) {
	dev->auBufs->free_flag = 1;
    }
    return wResult;
}


/*
 * -------------------------------------------------------
 *
 * WinAudioDevicePlay
 *
 * Parameters: 
 *	cbPtr - Frame to be played
 * 
 * Return: 
 *	None
 * 
 * Action: 
 *
 * -------------------------------------------------------
 */
static int
WinAudioDevicePlay(CBInfo *cbPtr)
{
    WinAudioDevice *dev;
    double now;
    int forceRestart;
    SpeakerRsrc *objPtr;
    MMRESULT mmret;
    
    objPtr = cbPtr->objPtr;
    dev = (WinAudioDevice *) objPtr->devicePtr->deviceData;
    if (dev->device == NULL) {
	return NOT_OPEN_ERROR;
    }

    /*
     * Check for earlier failures that need special recovery (e.g.
     * unplayable format)
     */
    if (dev->errorCode != 0) {
	return dev->errorCode;
    }

    if (dev->atToken != 0) {
	return TOO_EARLY_ERROR;
    }

    /*
     * The following code is tricky because it attempts to keep the
     * device queue full.  Since there's no way to directly query
     * the state of the device queue, we use the following strategy.
     * (On Windows, there is a way to query the queue directly.)
     * The Rsrc will normally call us a little while (SKEW_TIME, about
     * 200 milliseconds) before we're ready to play.  There are 2 cases:
     *  1. if the device queue has data in it, just write the data
     *	   to the device.
     *	2. if the device queue is empty, wait until the real start
     *	   time to write the data.
     *  3. if the device parameters have changed (e.g., a different
     *	   data format), wait until the real start time to reset the
     *	   parameters and then write the data.
     *  4. If we've been writing data continuously in our attempts
     *	   to keep the queue full for a long time (SYNC_TIME), the
     *	   synchronization might have drifted.  So, wait until the
     *	   real start time, flush any remaining data, and then write
     *	   the data.
     *
     * Actions 2-4 are performed by the WinAudioStartPlay() funciton
     * below, since they are fired by an at event.  They are signaled
     * by setting the forceRestart variable to 1.
     *
     * In any case, the emptyTime variable is set to when the
     * queue will become empty again.
     */
    forceRestart = 0;

    /*
     * Check to see if the format has changed.  If so, wait for
     * device to drain and force restart.
     */
    if ((cbPtr->hdr.samplingRate != dev->currFmt.samplingRate) ||
	(cbPtr->hdr.channelCount != dev->currFmt.channelCount) ||
	(cbPtr->hdr.dataFormat != dev->currFmt.dataFormat)) {
	forceRestart = 1;
    }

    /*
     * Check to see if it's been awhile (SYNC_TIME) since we did
     * the last restart.  If so, force a restart to keep synchronization
     * from drifting
     */
    objPtr = cbPtr->objPtr;
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));
    if (now > dev->lastStartTime + SYNC_TIME) {
	forceRestart = 1;
    }

    /*
     * If we think the device queue is empty (now > dev->emptyTime), force
     * a restart so we start as close to play time as possible.
     */
    if (now > dev->emptyTime) {
	forceRestart = 1;
    }

    /*
     * Play it directly or wait.
     */
    if (forceRestart) {
	Tk_Preserve(cbPtr);
	AttachBuffer (objPtr->interp, cbPtr->buffer,
			"WinAudioDevice:WinAudioDevicePlay");
	
	dev->atToken = Tcm_CreateAtHandler(cbPtr->start-SKEW, HUGE_VAL,
			    "audio", (Tk_TimerProc *)WinAudioStartPlay,
			    (ClientData) cbPtr, NULL,
			    NULL, (Tk_IdleProc *)NULL, (ClientData)NULL, NULL);

    } else {
	mmret = WinAudioWriteData(dev, cbPtr->data, cbPtr->len);
	dev->emptyTime += (cbPtr->end - cbPtr->start);
	/* ResetStopToken(); */
    }
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * WinAudioStartPlay --
 *
 *	Starts off the initial playback to the device.  Only
 *	called when the device queue is thought to be empty.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	Writes data to the device and updates dev->emptyTime
 *
 *--------------------------------------------------------------
 */
static void
WinAudioStartPlay (CBInfo *cbPtr)
{
    double now, fraction;
    int offset, width;
    SpeakerRsrc *objPtr;
    WinAudioDevice *dev;
    MMRESULT mmret;

    objPtr = cbPtr->objPtr;
    dev = (WinAudioDevice *) objPtr->devicePtr->deviceData;

    if (dev->device == NULL) {
	Tk_Release(cbPtr);
	return;
    }

    dev->atToken = 0;
    DetachBuffer (objPtr->interp, cbPtr->buffer,
		  "WinAudioDevice:WinAudioStartPlay");
    if (dev->device == NULL) {
	Tk_Release(cbPtr);
	return;
    }

    /*
     * After either action, the dev->emptyTime variable is set to when the
     * queue will become empty again.
     */
    now = LTS_SystemTimeOf(objPtr->lts, LTS_Value(objPtr->lts));

    /*
     * Flush any data in the output queue
     */
    mmret = waveOutReset(dev->device);
    if (mmret != 0) {
	dev->errorCode = mmret;
	Tk_Release(cbPtr);
	return;
    }

    /*
     * Set up the device parameters
     */
    mmret = WinAudioDeviceSetPlayParams (dev, cbPtr->hdr.samplingRate,
					 cbPtr->hdr.channelCount,
					 cbPtr->hdr.dataFormat);
    if (mmret != 0) {
	dev->errorCode = mmret;
	Tk_Release(cbPtr);
	return;
    }

    if (cbPtr->hdr.dataFormat == SUN_LIN_16) {
        width = 2;
    } else {
        width = 1;
    }

    /*
     * If callback is late, just play the appropriate fraction of the data
     */
    if (now > cbPtr->start) {
	fraction = (now - cbPtr->start) / (cbPtr->end - cbPtr->start);
	if (fraction > 1.0) {
	    Tk_Release(cbPtr);
	    return;
	}
	if (fraction < SKEW) {
	    fraction = 0.0;
	}
    } else {
	fraction = 0.0;
    }
    offset = fraction * cbPtr->len + 0.499;

    /*
     * Byte align the offset correctly
     */
    offset = width*(offset/width);

    WinAudioWriteData(dev, cbPtr->data + offset, cbPtr->len - offset);
    dev->emptyTime = now + (cbPtr->end - cbPtr->start)*(1.0 - fraction);
    /* ResetStopToken(); */
    Tk_Release(cbPtr);
    dev->lastStartTime = now;
    return;
}


/*
 * -------------------------------------------------------
 *
 * WinAudioDeviceSetSampleRate
 *
 *	Set the sample rate for audio output.  Returns 0
 *	if successfully set, otherwise an error value
 *
 * Parameters:
 *	int sample_rate;         samples per second
 *	int channels;            # of interleaved channels
 *	int encoding;            data encoding format
 *
 *  NOTE: if any parameter is -1, the value is left unchanged.
 *
 * -------------------------------------------------------
 */
static MMRESULT
WinAudioDeviceSetPlayParams (WinAudioDevice *dev,
			     int sample_rate, int channels, int encoding)
{
    MMRESULT mmret;

    /*
     * If not initiatilzed, return
     */
    if (dev->device == NULL) {
	return 0;
    }

    /*
     * Create structure that we're trying to set
     */
    if ((encoding != SUN_ULAW) && (encoding != SUN_LIN_8) &&
	(encoding != SUN_LIN_16))
    {
	return BAD_FORMAT_ERROR;
    }
    dev->currFmt.dataFormat = encoding;

    if (channels != -1) {
	dev->currFmt.channelCount = channels;
    }

    if (sample_rate == -1) {
	return 0;
    }
    if (sample_rate == dev->currFmt.samplingRate) {
	return 0;
    }
    dev->currFmt.samplingRate = sample_rate;

    mmret = waveOutClose(dev->device);
    if (mmret != 0) {
	return mmret;
    }

    /* XXX: Need to reopen, but since my machine can handle 8000Hz and
     * I don't feel like reopening right now, write this section later
     */
    return TOO_LAZY_ERROR;
}


/*
 * -------------------------------------------------------
 *
 * WinAudioDeviceSetGain
 *
 * Sets the gain on the device
 *
 * -------------------------------------------------------
 */
static void
WinAudioDeviceSetGain(WinAudioDevice *dev, double gain)
{
    DWORD newVolume;

    if (dev->deviceId == (UINT) -1) {
	return;
    }

    if (gain > 1.0) {
	newVolume = 0xFFFF;
    } else if (gain < 0.0) {
	newVolume = 0x0000;
    } else {
	newVolume = (DWORD) (gain * 0xFFFF);
    }

    waveOutSetVolume(dev->deviceId, newVolume);
}


/*
 * -------------------------------------------------------
 *
 * WinAudioDeviceError
 *
 * Map error numbers to strings describing the error
 *
 * -------------------------------------------------------
 */

static char *
WinAudioDeviceError(SpeakerRsrc *objPtr, int errNum)
{
    switch (errNum) {
      case MMSYSERR_INVALHANDLE:
	return "Invalid audio handle";

      case MMSYSERR_HANDLEBUSY:
	return "Audio handle is busy";

      case CANT_OPEN_HANDLE_BUSY_ERROR:
	return
	    "Unable to properly close currently open device: handle is busy";

      case CANT_OPEN_ERROR:
	return "Unable to open device";

      case WAVERR_STILLPLAYING:
	return "Audio is still playing";

      case NO_MATCHING_DEVICE_ERROR:
	return "No useful PCM audio device was found";

      case BAD_DEVICE_ERROR:
	return "Invalid device";
	
      case BAD_SAMPLE_RATE_ERROR:
	return "Sampling rate incompatible with audio device";

      case BAD_FORMAT_ERROR:
	return "Audio format is incompatible with audio device";

      case UNSUPPORTED_FORMAT_ERROR:
	return "Audio format is currently unsupported";

      case NO_AUDIO_DEVICES_ERROR:
	return "No audio devices are available on this machine";
	
      case NOT_OPEN_ERROR:
	return "Audio device is not open";

      case TOO_EARLY_ERROR:
	return "Send-ahead time must be less than data length";

      case TOO_LAZY_ERROR:
	return "Gordon Chaffee is too lazy right now to set to the correct rate";

      case FORMAT_ERROR:
	  return "Audio data is of an unhandled format";

      case NO_MEMORY_ERROR:
	return "Out of memory";

      case MEMORY_LOCK_ERROR:
	return "Failed to lock memory for data chunk";

      default:
	return "Unknown WinAudio error";
    }
}

#else /* no WIN_AUDIO */

/*
 * -------------------------------------------------------
 *
 * GetWinAudioDevice
 *
 *      Stub routine for non sparc machines
 *
 * -------------------------------------------------------
 */

SpeakerDevice *
GetWinAudioDevice(SpeakerRsrc *objPtr)
{
    return (SpeakerDevice *)NULL;
}

#endif /* WIN_AUDIO */
