/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/* This file contains one function that determines whether shared
 * memory is supported or not. In addition to calling XShmQueryExtension,
 * this funtion determines if the X server is remote or not. The function
 * returns 0 if shared memory is not possible and 1 if it is possible.
 */ 

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <tcl.h>
#include <tk.h>

#ifdef HAVE_SHARED_MEMORY
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>


/*
 *------------------------------------------------------------
 *
 * NoXShm
 *
 *	Registered temporarily as a Tk error handler so that
 *	if we can't use the XShm extension (for example if the
 *	display is remote) the hasShm flag is unset.
 *
 * Results:
 *	Always returns 0 so the program doesn't exit
 *
 * Side Effects:
 *	Unsets the shared_mem_flag flag
 *
 *------------------------------------------------------------
 */
static int
NoXShm(ClientData cd, XErrorEvent *evt)
{
    int *handler_result_ptr = (int *) cd;

    *handler_result_ptr = 0;
    return 0;
}

int
cmXShmQueryExtension(Display *dpy)
{
    XShmSegmentInfo si;
    Tk_ErrorHandler handler;
    static int inited = 0;
    static int cache;
    int handler_result = 1;

    if (inited) {
	return cache;
    }

    inited++;
    if (!XShmQueryExtension(dpy)) {
	cache = 0;
	return cache;
    }

    /*
     * Just because the XShm extension is available
     * doesn't mean we can really use it; the display
     * may be on a remote machine.
     *
     * Determine if it is really available by setting
     * a temporary error handler and trying an XShm
     * operation.  If it fails, don't use XShm.
     */
    si.shmid = shmget(IPC_PRIVATE, 512, IPC_CREAT|0777);
    if (si.shmid < 0) {
	cache = 0;
	return cache;
    }

    si.readOnly = 1;
    XSync(dpy, 0);
    handler = Tk_CreateErrorHandler(dpy, -1, -1, -1,
				    NoXShm, (ClientData) &handler_result);
    XShmAttach(dpy, &si);
    XSync(dpy, 0);
    Tk_DeleteErrorHandler(handler);
    shmctl(si.shmid, IPC_RMID, 0);

    if (!handler_result) {
	cache = 0;
	return cache;
    }

    cache = 1;
    return cache;
}

#else /* HAVE_SHARED_MEMORY */

int
cmXShmQueryExtension(Display *dpy)
{
    return 0;
}

#endif /* HAVE_SHARED_MEMORY */
