/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */



#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <math.h>

#ifdef NO_STRING_H
#   include "compat/string.h"
#else
#   include <string.h>
#endif

#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <dp.h>
#include <tcl.h>
#include <tk.h>
#include <tkInt.h>
#include "tcmEvent.h"
#include "cmEvery.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmUtil.h"
#include "cmCookies.h"
#include "cmBufferlist.h"
#include "cmFilterUtils.h"
#include "cmFontUtils.h"

#define ONE_BILLION 1000000000.0
#define FREE(X) if ((X) != NULL) ckfree((X))

typedef struct YUVColor {
    unsigned char y, u, v;
} YUVColor;

typedef struct YUVFrame {
    YUVBufferHeader frameHeader;
    double logStart;
    double logEnd;
    ScatterBuffer *sb;
    Buffer *dataBuffer;
    char *data;
    int dataLength;
    int at_token;
    struct YUVFilter *objPtr;
    struct YUVFrame *next;
    struct YUVFrame *prev;
    ClientData deviceData;
    char name[30];
} YUVFrame;

static int nextYUVFilterNum = 0;	/* Variable for next unique object name. */
static int nextYUVFrameNum = 0;		 

typedef struct YUVFilter {
    Tcl_Interp *interp;		/* Tcl interpreter that created object. */
    char *name;			/* Tcl string name of this object. */
    int ready;			/* Ready flag. */
    ClientData lts;		/* LTS for object */
    int ltsTriggerNum;
    char *ltsName;
    
    int numInputs;		/* Can be either 2 or 1, implicitly set
				 * by filter type */

    YUVFrame *queueA;		/* Queue of frames to be filter */
    YUVFrame *queueB;		
    YUVFrame *lastA;		/* Ptr to last frame decoded */
    YUVFrame *lastB;
    YUVFrame *blackB;		/* A cached frame that is all black */

    int reqFrameRate;		/* requested frame rate (config option) */
    int frameRate;		/* actual frame rate to use */
    double sendAhead;
    double filterTime;		/* Average time to perform a filter action */

    char *outCmd;		/* Command to send buffers to. */
    InFunc outCCmd;		/* C pointer to out command */
    ClientData outCD;		/* ClientData for outCCmd. */

    int atToken;		/* Used for waking up to send out frames */
    double atWhen;		/* When that event was scheduled */

    char *acceptAHandle;
    char *acceptBHandle;

    /* Slots for different filters */
    
    int filter;			/* filter type */
    int xoffset;		/* image offsets (used by various filters) */
    int yoffset;

    /* Text overlay */
    char *textFont;
    char *textString;		/* Text string for overly. */
    short *textEncoded;		/* encoded version of string. */
    char *textColor;		/* Color specified by user */
    YUVColor textYUVColor;	/* A cached yuv conversion of above */
   
    /* Fade */
    double fadePercent;		/* Percent to fade from source A to B */

    /* Picture in Picture */
    int pinpScale;		/* factor to scale source B down to */

} YUVFilter;


void
printFrame (YUVFrame *frame )
{
    fprintf(stderr, "%s %f %f\n", frame->name, frame->logStart, frame->logEnd);
}

/* Filter types */
#define NONE 0
#define TEXT_OVERLAY 1
#define FADE 2
#define PIC_IN_PIC 3
#define MAX_WIDTH 640

/*
 * Prototypes for routines used internally
 */
int YUVFilterModuleInit		        (Tcl_Interp *interp);
static int YUVFilterCreateCmd		(ClientData clientData,
					 Tcl_Interp *interp,
					 int argc, char **argv);
static YUVFilter *CreateNewYUVFilter	(Tcl_Interp *interp,
					 char *name);
static int DestroyYUVFilter		(YUVFilter *objPtr);
static int YUVFilterObjCmd		(ClientData clientData,
					 Tcl_Interp *interp,
					 int argc, char **argv);
static int YUVFilter_Accept		(YUVFilter *objPtr,
					 SB_List * sbl, char c);
static int YUVFilter_AcceptA		(YUVFilter *objPtr,
					 SB_List * sbl);
static int YUVFilter_AcceptB		(YUVFilter *objPtr,
					 SB_List * sbl);
static int YUVFilter_Ready		(YUVFilter *objPtr);
static int YUVFilter_Unready		(YUVFilter *objPtr);
static void YUVFilter_LTSChanged	(ClientData clientData);
int YUVFilter_DoFilter                  (YUVFilter *objPtr, 
					 YUVFrame *frameA, 
					 YUVFrame *frameB,
					 double start,
					 double end);
static void YUVFilter_FreeFrame		(YUVFrame *framePtr);
static int ConfigureYUVFilter		(YUVFilter *objPtr,
					 int argc, char **argv);
static int YUVFilter_ChangeOutCmd	(YUVFilter *objPtr, char *newOut);

static void ResetFrameQueues		(YUVFilter *objPtr);
static int ChangeFilter                 (YUVFilter *objPtr, int filter);
static void RGBtoYUV                    (unsigned short r, unsigned short g, 
					 unsigned short b, YUVColor *yuv);
static void InsertFrame                 (YUVFrame *framePtr, 
					 YUVFrame **queue, double speed );
static YUVFrame *GetFrameByTime         (YUVFilter *objPtr, YUVFrame **queue, 
					 double start, double end);
static void RemoveFromQueue             (YUVFrame *frame, YUVFrame **queue);
static void PurgeLateFrames             (YUVFrame **queue, 
					 double speed, double now);
static void YUVFilterSendFrame          (ClientData clientData);
static YUVFrame *MakeBlackFrame         (YUVFrame *source);
static YUVFrame *CopyFrame              (YUVFrame *source);
static void YUVFilterMissed             (ClientData clientData);


/************************************************************
 *
 * YUVFilterModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int 
YUVFilterModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "yuvFilter", YUVFilterCreateCmd,
			NULL, (void (*)()) NULL);
    return TCL_OK;
}

/************************************************************
 *
 * YUVFilterCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the "YUVFilter" Tcl command.
 *     A new object of type YUVFilter is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
YUVFilterCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    YUVFilter *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextYUVFilterNum++;
    sprintf(new_name, "%syuvFilter%d", argv[1], nextYUVFilterNum);

    new = CreateNewYUVFilter(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, YUVFilterObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureYUVFilter(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}


/************************************************************
 *
 * CreateNewYUVFilter
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static YUVFilter *
CreateNewYUVFilter(Tcl_Interp *interp, char *name)
{
    YUVFilter *new;
    char *acceptName;
    char **event_list, **subst_list;
    int i;

    new = (YUVFilter *) ckalloc(sizeof(YUVFilter));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);

    new->ready = 0;
    new->lts = NULL;
    new->ltsTriggerNum = -1;
    new->ltsName = NULL;
    new->numInputs = 1;
    new->queueA = NULL;
    new->queueB = NULL;
    new->lastA = NULL;
    new->lastB = NULL;
    new->blackB = NULL;
    new->outCmd = NULL;
    new->outCCmd = NULL;
    new->outCD = NULL;
    new->atToken = -1;
    new->atWhen = HUGE_VAL;
    new->acceptAHandle = NULL;
    new->acceptBHandle = NULL;
    new->filter = NONE;
    new->xoffset = 0;
    new->yoffset = 0;
    new->textFont = NULL;
    new->textString = NULL;
    new->textEncoded = NULL;
    new->textColor = ckalloc(sizeof("white")+1);
    strcpy(new->textColor, "white");
    new->textYUVColor.y = 255;
    new->textYUVColor.u = 128;
    new->textYUVColor.v = 128;
    new->fadePercent = 0;
    new->pinpScale = 2;
    new->filterTime = .1;
    new->frameRate = 10;
    new->reqFrameRate = 10;
    new->sendAhead = .1;

    /*
     * Register accept commands as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+10);
    sprintf(acceptName, "%s.%s", name, "acceptA");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) YUVFilter_AcceptA,
			  (ClientData) new);
    new->acceptAHandle = acceptName;

    acceptName = (char *) ckalloc(strlen(name)+10);
    sprintf(acceptName, "%s.%s", name, "acceptB");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) YUVFilter_AcceptB,
			  (ClientData) new);
    new->acceptBHandle = acceptName;

    return new;
}


/************************************************************
 *
 * DestroyYUVFilter
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

static int
DestroyYUVFilter(YUVFilter *objPtr)
{
    int result = TCL_OK;
    int i, devReturn;

    /*
     * Remove command name
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    if (objPtr->atToken != -1) {
	Tcm_DeleteAtHandler(objPtr->atToken);
    }
    objPtr->atToken = -1;
    objPtr->atWhen = HUGE_VAL;

    /*
     * Unregister the accept atoms
     */
    if (objPtr->acceptAHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptAHandle);
	ckfree(objPtr->acceptAHandle);
	objPtr->acceptAHandle = NULL;
    }
    if (objPtr->acceptBHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptBHandle);
	ckfree(objPtr->acceptBHandle);
	objPtr->acceptBHandle = NULL;
    }

    ResetFrameQueues(objPtr);

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    FREE(objPtr->name);
    FREE(objPtr->ltsName);
    FREE(objPtr->outCmd);
    FREE(objPtr->acceptAHandle);
    FREE(objPtr->acceptBHandle);
    FREE(objPtr->textFont);
    FREE(objPtr->textString);
    FREE(objPtr->textEncoded);
    FREE(objPtr->textColor);
    ckfree((char *) objPtr);
    return result;
}


/************************************************************
 *
 * YUVFilterObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     YUVFilter object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
YUVFilterObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    YUVFilter *objPtr = (YUVFilter *) clientData;
    
    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
			 argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);
  
    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureYUVFilter(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyYUVFilter(objPtr);
    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return YUVFilter_Ready(objPtr);
    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return YUVFilter_Unready(objPtr);
    } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
	sprintf(interp->result, "%f %d", objPtr->filterTime, objPtr->frameRate);
	return TCL_OK;
    } else if ((c == 'a') && (strncmp("acceptA", argv[1], len) == 0)) {
	SB_List *sbl = NULL;
	int result;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			     argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	    ckfree((char *) sbl);
	    Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
			     argv[2],  (char *)NULL);
	    return TCL_ERROR;
	}

	result = YUVFilter_AcceptA(objPtr, sbl);
	if (sbl != NULL) {
	    FreeSBList(sbl);
	}
	return result;
    } else if ((c == 'a') && (strncmp("acceptB", argv[1], len) == 0)) {
	SB_List *sbl = NULL;
	int result;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
			     argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}

	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	    ckfree((char *) sbl);
	    Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
			     argv[2],  (char *)NULL);
	    return TCL_ERROR;
	}

	result = YUVFilter_AcceptB(objPtr, sbl);
	if (sbl != NULL) {
	    FreeSBList(sbl);
	}
	return result;
    } else if ((c == 'g') && (strncmp("getTextSize", argv[1], len) == 0)) {
        if (objPtr->filter == TEXT_OVERLAY) {
            if (objPtr->textEncoded == NULL) {
                if (objPtr->textFont != NULL) {
                    if (strlen(objPtr->textString) != 0) {
                        int fd = open(objPtr->textFont, O_RDONLY);
                        if (fd > 0) {
                            objPtr->textEncoded = bdftorlc(objPtr->textFont, 
                                                           NULL, NULL, 
                                                           objPtr->textString, 
                                                           FALSE);
                        } else {
                            sprintf(interp->result, "0 0");
                            return TCL_OK;
                        }
                        close(fd);
                    } else {
                        sprintf(interp->result, "0 0");
                        return TCL_OK;
                    }
                } else {
                    sprintf(interp->result, "0 0");
                    return TCL_OK;
                }
            }
            sprintf(interp->result, "%d %d", objPtr->textEncoded[2],
                    objPtr->textEncoded[1]);
            return TCL_OK;
        } else {
            sprintf(interp->result, "0 0");
            return TCL_OK;
        }
    }

        

  
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ",
		     argv[1], (char *) NULL);
  usage:
    Tcl_AppendResult(interp,
		     "\n<command> can be one of:\n",
		     "  configure ?-slot ?value??\n",
		     "  destroy\n",
		     "  ready\n",
		     "  unready\n",
		     "  stats\n",
		     "  acceptA <ScatterBufferList>\n",
		     "  acceptB <ScatterBufferList>\n",
		     "  getTextSize\n",
		     (char *) NULL);
    return TCL_ERROR;
}



/************************************************************
 *
 * ConfigureYUVFilter
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 * Slots:
 *     lts
 *     outCmd
 *     filter
 *     xoffset
 *     yoffset
 *     text
 *     color
 *     percent
 *     scale
 ************************************************************/

static int
ConfigureYUVFilter(YUVFilter * objPtr, int argc, char **argv)
{

    Tcl_Interp *interp;
    int cntr;
    char str[50];

    interp = objPtr->interp;

    /*
     * Case 1: <obj> configure
     */
    if (argc == 0) {
	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{-lts {} {}} ",
			     (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{ -lts {} {",
			     objPtr->ltsName, "}} ", (char *) NULL);
	}
	if (objPtr->outCmd == NULL) {
	    Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCmd,
			     "}} ", (char *)NULL);
	}	
	sprintf(str, "{%d}", objPtr->frameRate);
	Tcl_AppendResult(interp, "{-frameRate 0 ", str, "} ", (char *)NULL);
	sprintf(str, "{%f}", objPtr->sendAhead);
	Tcl_AppendResult(interp, "{-sendAhead 0.0 ", str, "} ", (char *)NULL);
	if (objPtr->filter == NONE) {
	    Tcl_AppendResult(interp, "{-filter {} {}} ", (char *)NULL);
	} else {
	    switch (objPtr->filter) {
	      case TEXT_OVERLAY:
		strcpy(str, "textOverlay");
		break;
	      case FADE:
		strcpy(str, "fade");
		break;
	      case PIC_IN_PIC:
		strcpy(str, "PictureInPicture");
		break;
	      default:
		strcpy(str, "unknown");
		break;
	    }
	    Tcl_AppendResult(interp, "{-filter {} {", str,
			     "}} ", (char *)NULL);
	}
	sprintf(str, "{%d}", objPtr->yoffset);
	Tcl_AppendResult(interp, "{", "-yoffset ", "0 ", str, "} ", (char *)NULL);
	sprintf(str, "{%d}", objPtr->xoffset);
	Tcl_AppendResult(interp, "{", "-xoffset ", "0 ", str, "} ", (char *)NULL);
	if (objPtr->textFont == NULL) {
	    Tcl_AppendResult(interp, "{-fontFile {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-fontFile {} {", objPtr->textFont,
			     "}} ", (char *)NULL);
	}	
	if (objPtr->textString == NULL) {
	    Tcl_AppendResult(interp, "{-text {} {}} ", (char *)NULL);
	} else {
	    Tcl_AppendResult(interp, "{-text {} {", objPtr->textString,
			     "}} ", (char *)NULL);
	}	
	Tcl_AppendResult(interp, "{-color white ", objPtr->textColor,"} ",
			 (char *)NULL);
	sprintf(str, "{%f}", objPtr->fadePercent);
	Tcl_AppendResult(interp, "{-percent 0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->pinpScale);
	Tcl_AppendResult(interp, "{-scale 2 ", str, "} ",
			 (char *)NULL);

	return TCL_OK;
    }

    /*
     * Case 2: <obj> configure <-slot>
     */
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	    if (objPtr->ltsName == NULL) {
		Tcl_AppendResult(interp, "-lts ", "{} ", "{}", (char *) NULL);
	    } else {
		Tcl_AppendResult(interp, "-lts ", "{} ", "{", objPtr->ltsName, "}", (char *) NULL);
	    }
	    return TCL_OK;
	}
	if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
	    if (objPtr->outCmd == NULL) {
		Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCmd,
				 "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	if ((c == 'f') && (strncmp("-frameRate", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->frameRate);
	    Tcl_AppendResult(interp, "-frameRate 0 ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
	    sprintf(str, "{%f}", objPtr->sendAhead);
	    Tcl_AppendResult(interp, "-sendAhead 0.0 ", str, (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'f') && (strncmp("-filter", argv[0], len) == 0)) {
	    if (objPtr->filter == NONE) {
		Tcl_AppendResult(interp, "-filter {} {}", (char *)NULL);
	    } else {
		switch (objPtr->filter) {
		  case TEXT_OVERLAY:
		    strcpy(str, "textOverlay");
		    break;
		  case FADE:
		    strcpy(str, "fade");
		    break;
		  case PIC_IN_PIC:
		    strcpy(str, "pictureInPicture");
		    break;
		  default:
		    strcpy(str, "unknown");
		    break;
		}
		Tcl_AppendResult(interp, "-filter {} {", str,
				 "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	if ((c == 'y') && (strncmp("-yoffset", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->yoffset);
	    Tcl_AppendResult(interp, "-yoffset ", "0 ", str,
			     (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'x') && (strncmp("-xoffset", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->xoffset);
	    Tcl_AppendResult(interp, "-xoffset ", "0 ", str,
			     (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 't') && (strncmp("-text", argv[0], len) == 0)) {
	    if (objPtr->textString == NULL) {
		Tcl_AppendResult(interp, "-text {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-text {} {", objPtr->textString,
				 "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	if ((c == 'f') && (strncmp("-fontFile", argv[0], len) == 0)) {
	    if (objPtr->textFont == NULL) {
		Tcl_AppendResult(interp, "-fontFile {} {}", (char *)NULL);
	    } else {
		Tcl_AppendResult(interp, "-fontFile {} {", objPtr->textFont,
				 "} ", (char *)NULL);
	    }
	    return TCL_OK;
	}
	if ((c == 'p') && (strncmp("-percent", argv[0], len) == 0)) {
	    sprintf(str, "{%f}", objPtr->fadePercent);
	    Tcl_AppendResult(interp, "-percent ", "0 ", str,
			     (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 's') && (strncmp("-scale", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->pinpScale);
	    Tcl_AppendResult(interp, "-scale ", "2 ", str,
			     (char *)NULL);
	    return TCL_OK;
	}
	if ((c == 'c') && (strncmp("-color", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-color white ", objPtr->textColor, (char *) NULL);
	    return TCL_OK;
	}
      bad_option:
	Tcl_AppendResult(interp, "bad configuration option: ", argv[0], (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: <obj> configure <-slot> <newValue>
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;
	int ready;
	
	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;

	    cntr++;
	    if (cntr == argc) 
		goto no_value_error;
	    else if (strlen(argv[cntr]) == 0) {
	        newLts = NULL;
	    }
	    if ((newLts = Tcm_FindAtom(interp, argv[cntr])) == NULL)
		goto bad_value_error;

	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }

            /*
             * Release our lock on the old lts...
             */
            if (objPtr->lts != NULL) {
	        if (objPtr->ltsTriggerNum != -1) {
		    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
		    objPtr->ltsTriggerNum = -1;
	        }
                Tk_Release(objPtr->lts);
                objPtr->lts = NULL;
            }

	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
	        objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts, 
						       LTS_SPEED_FLAG | LTS_VALUE_FLAG |
						       LTS_EARLY_PHASE, 
						       YUVFilter_LTSChanged,
						       (ClientData) objPtr);
	        objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
	        strcpy(objPtr->ltsName, argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }
	    ResetFrameQueues(objPtr);
	}
	else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (YUVFilter_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
		return TCL_ERROR;
	    }
	}
	else if ((c == 'f') && (strncmp("-frameRate", argv[cntr], len) == 0)) {
	    int new;

	    cntr++;
	    if (cntr == argc) 
		goto no_value_error;
	    if (Tcl_GetInt(interp, argv[cntr], &new) != TCL_OK)
		goto bad_value_error;
	    objPtr->frameRate = objPtr->frameRate = new;
	}
	else if ((c == 's') && (strncmp("-sendAhead", argv[cntr], len) == 0)) {
	    double new;

	    cntr++;
	    if (cntr == argc) 
		goto no_value_error;
	    if (Tcl_GetDouble(interp, argv[cntr], &new) != TCL_OK)
		goto bad_value_error;
	    objPtr->sendAhead = new;
	}
	else if  ((c == 'f') && (strncmp("-filter", argv[cntr], len) == 0)) {
	    int new, len;

	    cntr++;

	    len = strlen(argv[cntr]);

	    if (strlen(argv[cntr]) == 0) {
	        new = NONE;
		objPtr->numInputs = 1;
	    }
	    else if (strncmp("textOverlay", argv[cntr], len) == 0) {
	        new = TEXT_OVERLAY;
		objPtr->numInputs = 1;
	    }
	    else if (strncmp("fade", argv[cntr], len) == 0) {
	        new = FADE;
		objPtr->numInputs = 2;
	    }
	    else if (strncmp("pictureInPicture", argv[cntr], len) == 0) {
	        new = PIC_IN_PIC;
		objPtr->numInputs = 2;
	    }
	    else {
		Tcl_AppendResult(interp, "Filter must be one of {",
				 "textOverlay, fade, pictureInPicture, or {} }",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }	    
	    objPtr->filter = new;
	}
	else if ((c == 'y') && (strncmp("-yoffset", argv[cntr], len) == 0)) {
	    int new;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &new) != TCL_OK) {
		Tcl_AppendResult(interp, "Bad value specified trying to set ",
				 "configuration option ", argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->yoffset = new;
	}
	else if ((c == 'x') && (strncmp("-xoffset", argv[cntr], len) == 0)) {
	    int new;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &new) != TCL_OK) {
		Tcl_AppendResult(interp, "Bad value specified trying to set ",
				 "configuration option ", argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->xoffset = new;
	}
	else if ((c == 'f') && (strncmp("-fontFile", argv[cntr], len) == 0)) {
	    Tcl_DString fname_dstring;
	    char *fname;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }

	    fname = Tcl_TildeSubst(objPtr->interp, argv[cntr], &fname_dstring);
	    objPtr->textFont = (char *) ckalloc(strlen(fname) + 1);
            objPtr->textEncoded = NULL;
	    strcpy(objPtr->textFont, fname);
	    Tcl_DStringFree(&fname_dstring);
	}
	else if ((c == 't') && (strncmp("-text", argv[cntr], len) == 0)) {
	    int fd;

	    cntr++;

	    FREE(objPtr->textString);
	    FREE(objPtr->textEncoded);

	    if (strlen(argv[cntr]) == 0) {
		objPtr->textString = NULL;
	    } else {
		objPtr->textString = (char *) ckalloc(strlen(argv[cntr]) + 1);
		strcpy(objPtr->textString, argv[cntr]);
	    }
	    objPtr->textEncoded = NULL;

	    /* Encode new text string */
	    if (objPtr->textFont != NULL) {
		fd = open(objPtr->textFont, O_RDONLY);
		if (fd > 0) {
		    if (strlen(objPtr->textString) != 0) {
			objPtr->textEncoded = bdftorlc(objPtr->textFont, 
						       NULL, NULL, 
						       objPtr->textString, 
						       FALSE);
		    }
		    close(fd);
		}
	    }
	}
	else if ((c == 'c') && (strncmp("-color", argv[cntr], len) == 0)) {
	    XColor *temp;
	    TkWindow *win;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }

#ifndef NO_X11
	    win = tkMainWindowList[0].winPtr;
	    temp = Tk_GetColor(objPtr->interp, (Tk_Window)win, Tk_GetUid(argv[cntr]));
	    
	    if (temp == NULL) {
		Tcl_AppendResult(interp, "Bad color value specified when ",
				 "trying to set ",
				 "configuration option ", argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }

	    
	    FREE(objPtr->textColor);
	    objPtr->textColor = ckalloc(sizeof(argv[cntr])+1);
	    strcpy(objPtr->textColor, argv[cntr]);

	    RGBtoYUV(temp->red, temp->green, temp->blue, &(objPtr->textYUVColor));
#endif
	}
	else if ((c == 'p') && (strncmp("-percent", argv[cntr], len) == 0)) {
	    double new;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetDouble(interp, argv[cntr], &new) != TCL_OK || new < 0 || new > 1.0 ) {
		Tcl_AppendResult(interp, "Bad value specified trying to set ",
				 "configuration option ", argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->fadePercent = new;
	}
	else if ((c == 's') && (strncmp("-scale", argv[cntr], len) == 0)) {
	    int new;

	    cntr++;
	    if (cntr == argc) {
		Tcl_AppendResult(interp, "Must specify value when ",
				 "trying to set configuration option ",
				 argv[cntr-1], (char *)NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &new) != TCL_OK || new < 1 ) {
		Tcl_AppendResult(interp, "Bad value specified trying to set ",
				 "configuration option ", argv[cntr-1],
				 (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->pinpScale = new;
	} else {
	    Tcl_AppendResult(interp, "Error, configuration option: ",
			     argv[cntr], " unknown.", (char *)NULL);
	    return TCL_ERROR;
	}
	cntr++;
    }

	return TCL_OK;

  no_value_error:
    Tcl_AppendResult(interp, "Must specify value when ",
		     "trying to set configuration option ",
		     argv[cntr-1], (char *)NULL);
    return TCL_ERROR;
    
  bad_value_error:
    Tcl_AppendResult(interp, "Bad value specified while trying to set ",
		     "configuration option ", argv[cntr-1],
		     (char *) NULL);
    return TCL_ERROR;
    
}

/************************************************************
 *
 * YUVFilter_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
YUVFilter_ChangeOutCmd(YUVFilter *objPtr, char *newOut) 
{

    /*
     * Clear the previous entry.
     */
    if (objPtr->outCmd != NULL) {
	ckfree(objPtr->outCmd);
    }
    objPtr->outCmd = NULL;
    objPtr->outCCmd = NULL;
    objPtr->outCD = NULL;

    /*
     * If new outCmd is empty, return.
     */
    if (strlen(newOut) == 0) {
	return TCL_OK;
    }

    /*
     * If new outCmd is C proc, get them as atoms and set everything
     * accordingly. 
     */
    if (newOut[0] == '@') {
	if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
				       &(objPtr->outCD), 
				       (ClientData *) &(objPtr->outCCmd)) != TCL_OK) {
	    objPtr->outCD = NULL;
	    objPtr->outCCmd = NULL;
	    Tcl_AppendResult(objPtr->interp,
			  "Couldn't find registered C procedure: ", newOut,
			  (char *)NULL);
	    return TCL_ERROR;
	}
    }

    /*
     * Copy new out command into outCmd string.
     */
    objPtr->outCmd = (char *) ckalloc(strlen(newOut)+1);
    strcpy(objPtr->outCmd,newOut);
    return TCL_OK;
}    


/************************************************************
 *
 * YUVFilter_Ready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
YUVFilter_Ready(YUVFilter * objPtr)
{
    int init_res;
    YUVFrame *framePtr;
    double awaken, ftime;
    
    /*
     * Can only be made ready if we have an lts
     */

    if (objPtr->lts == NULL) {
        Tcl_AppendResult(objPtr->interp, "Can't make YUV filter \"",
                         objPtr->name, "\" ready: no lts has been attached.", (char *)NULL);
        return TCL_ERROR;
    }

    if (objPtr->atToken != -1) {
	Tcm_DeleteAtHandler(objPtr->atToken);
    }

    objPtr->ready = 1;
    
    ftime = (1.0)/(objPtr->frameRate);

    awaken = Tcm_ReadSysClock() + ftime;
    objPtr->atWhen = awaken;
    
    objPtr->atToken = Tcm_CreateAtHandler(awaken, HUGE_VAL, 
					  "video", YUVFilterSendFrame, objPtr,
                                          NULL, NULL, NULL, NULL, NULL);
    
    return TCL_OK;
}


/************************************************************
 *
 * ResetFrameQueues
 *
 ************************************************************/

static void
ResetFrameQueues(YUVFilter *objPtr)
{
    /*
     * Blow away my frame queues.
     */

    if (objPtr->queueA != NULL) {
	YUVFrame *framePtr, *nextPtr;

	framePtr = objPtr->queueA;

	while (framePtr != NULL) {
	    nextPtr = framePtr->next;
	    YUVFilter_FreeFrame(framePtr);
	    framePtr = nextPtr;
	}
	objPtr->queueA = NULL;
    }
    if (objPtr->queueB != NULL) {
	YUVFrame *framePtr, *nextPtr;

	framePtr = objPtr->queueB;

	while (framePtr != NULL) {
	    nextPtr = framePtr->next;
	    YUVFilter_FreeFrame(framePtr);
	    framePtr = nextPtr;
	}
	objPtr->queueB = NULL;
    }

    if(objPtr->lastA)
	YUVFilter_FreeFrame(objPtr->lastA);
    if(objPtr->lastB)
	YUVFilter_FreeFrame(objPtr->lastB);
    objPtr->lastA = NULL;
    objPtr->lastB = NULL;

}


/************************************************************
 *
 * YUVFilter_LTSChanged
 *
 * Parameters:
 *     clientdata - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *      Called when the LTS changes value.  Dequeues any previously
 *      queued play requests, frees any memory we're holding, and
 *      returns.
 *
 ************************************************************/

static void
YUVFilter_LTSChanged(ClientData clientData)
{
    YUVFilter *objPtr = (YUVFilter *) clientData;
    int i;

    /*
     * If not ready, do nothing.
     */
    if (!objPtr->ready) {
	return;
    }

    ResetFrameQueues(objPtr);
    YUVFilter_Ready(objPtr);
}

/************************************************************
 *
 * YUVFilter_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl 
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
YUVFilter_AcceptA(YUVFilter *objPtr, SB_List *sbl)
{
    return YUVFilter_Accept(objPtr, sbl, 'a');
}
YUVFilter_AcceptB(YUVFilter *objPtr, SB_List *sbl)
{
    return YUVFilter_Accept(objPtr, sbl, 'b');
}

static int
YUVFilter_Accept(YUVFilter *objPtr, SB_List *sbl, char which)
{
    int i;
    YUVFrame *newFrame;
    double speed, now, eTime, sTime;
    signed char *sb_data;
    Buffer *bufPtr;
    double  start, end;
    
    if (!objPtr->ready) {
	return TCL_OK;
    }

    speed = LTS_Speed(objPtr->lts);
    now = LTS_Value(objPtr->lts);

    if (sbl == NULL) return TCL_OK;
    if (sbl->count < 1) return TCL_OK;

    /*
     * Increment reference count on buffers passed in
     */
    AttachToAllBuffers(objPtr->interp, sbl,
			"tcmYUVFilter:YUVFilter_Accept");

    /*
     * For each frame in sb list, create frame structure,
     * fill in info and add to queue.
     */

    for (i=0; i<sbl->count; i++) {
	newFrame = (YUVFrame *) ckalloc(sizeof(YUVFrame));
        nextYUVFilterNum++; 
        sprintf( newFrame->name, "frame%d", nextYUVFilterNum );

        sb_data = GetFromSB(sbl->sbs[i], CM_YUV_BUF_HDR, &bufPtr, NULL);

	if (sb_data == NULL) {
	    ckfree((char *) newFrame);
	    continue;
	}
	memcpy(&(newFrame->frameHeader), sb_data-8, sizeof(MJPEGBufferHeader));
	newFrame->frameHeader.hdrType = CM_YUV_BUF_HDR;
	newFrame->frameHeader.length = sizeof(YUVBufferHeader)-8;
	newFrame->frameHeader.hdrType = ntohl(newFrame->frameHeader.hdrType);
	newFrame->frameHeader.length = ntohl(newFrame->frameHeader.length);
	newFrame->frameHeader.width = ntohl(newFrame->frameHeader.width);
	newFrame->frameHeader.height = ntohl(newFrame->frameHeader.height);
	newFrame->frameHeader.logEnd_s = ntohl(newFrame->frameHeader.logEnd_s);
	newFrame->frameHeader.logEnd_ns = ntohl(newFrame->frameHeader.logEnd_ns);
	newFrame->frameHeader.logStart_s = ntohl(newFrame->frameHeader.logStart_s);
	newFrame->frameHeader.logStart_ns = ntohl(newFrame->frameHeader.logStart_ns);
	
	eTime = ((double )newFrame->frameHeader.logEnd_s) + 
	  (((double) newFrame->frameHeader.logEnd_ns) / ONE_BILLION);
	sTime = ((double )newFrame->frameHeader.logStart_s) + 
	  (((double) newFrame->frameHeader.logStart_ns) / ONE_BILLION);
	
	newFrame->data = GetFromSB(sbl->sbs[i], CM_DATA_SEGMENT, &bufPtr, 
				   &newFrame->dataLength);
	if (newFrame->data == NULL) {
	    ckfree((char *) newFrame);
	    continue;
	}

	newFrame->at_token = -1;
	newFrame->objPtr = objPtr;

	newFrame->dataBuffer = bufPtr;
	newFrame->logStart = sTime;
	newFrame->logEnd = eTime;
	newFrame->sb = CopyScatterBuffer(sbl->sbs[i]);
	newFrame->prev = newFrame->next = NULL;

	if (which == 'a') {
#ifdef DEBUG
	    fprintf(stderr, "Accept A: "); printFrame(newFrame); 
#endif
	    InsertFrame( newFrame, &objPtr->queueA, speed );
	} else {
#ifdef DEBUG
	    fprintf(stderr, "Accept B: "); printFrame(newFrame);
#endif
	    InsertFrame( newFrame, &objPtr->queueB, speed );
	}
    }

    /* If framerate is zero, we are data driven, so we need
       to call SendFrame now */

    if (objPtr->frameRate == 0 || objPtr->atToken == -1) {
	YUVFilterSendFrame((ClientData) objPtr);

    }
    return TCL_OK;
}

/************************************************************
 *
 * Frame Utilities
 *
 * InsertFrame
 * GetFrameByTime
 * RemoveFramesBefore
 * RemoveFromQueue 
 * PurgeLateFrames
 *
 ************************************************************/


static void
InsertFrame( YUVFrame *newFrame, YUVFrame **queue, double speed )
{
    YUVFrame *prevPtr, *nextPtr, *framePtr;

    if (*queue == NULL) {
	*queue = newFrame;
	newFrame->next = newFrame->prev = NULL;
    } else if (speed == 0.0) {
	/* Remove all old frames */
	framePtr = *queue;
	while (framePtr != NULL) {
	    nextPtr = framePtr->next;
	    YUVFilter_FreeFrame(framePtr);
	    framePtr = nextPtr;
	}
	*queue = newFrame;
	newFrame->next = newFrame->prev = NULL;
    } else {
	framePtr = *queue;
	prevPtr = NULL;
	while (framePtr != NULL) {
	    if (speed > 0.0) {
		if (framePtr->logStart < newFrame->logStart) {
		    prevPtr = framePtr;
		    framePtr = framePtr->next;
		} else {
		    break;
		}
	    } else {
		if (framePtr->logEnd > newFrame->logEnd) {
		    prevPtr = framePtr;
		    framePtr = framePtr->next;
		} else {
		    break;
		}
	    }
	}
	
	if (framePtr == NULL) {
	    prevPtr->next = newFrame;
	    newFrame->prev = prevPtr;
	    newFrame->next = NULL;
	} else if (prevPtr == NULL) {
	    newFrame->prev = NULL;
	    newFrame->next = framePtr;
	    framePtr->prev = newFrame;
	    *queue = newFrame;
	} else {
	    prevPtr->next = newFrame;
	    newFrame->prev = prevPtr;
	    newFrame->next = framePtr;
	    framePtr->prev = newFrame;
	}
    }
}


/* GetFrameByTime

   Return the first frame whose logical start occurs AFTER specified start and
   BEFORE end time.
*/

static YUVFrame *
GetFrameByTime(YUVFilter *objPtr, YUVFrame **queue, double start, double end)
{
    double now, speed;
    YUVFrame *chosenFrame, *prev;

    speed = LTS_Speed(objPtr->lts);
    
    prev = NULL;
    chosenFrame = *queue;

    while (chosenFrame != NULL && chosenFrame->next != NULL) {
	if (speed > 0.0) {
	    if (chosenFrame->logStart >= start)	
		break;
	} else {
	    if (chosenFrame->logEnd <= end) 
		break;
	}
	prev = chosenFrame;
	chosenFrame = chosenFrame->next;
    }
    
    if (chosenFrame == NULL) {
	return NULL;
    }

    if (speed > 0.0) {
	if (chosenFrame->logStart > end)
	    return prev;
    } else {
	if (chosenFrame->logEnd < start)
	    return prev;
    }

    return chosenFrame;
}

static void
RemoveFromQueue(YUVFrame *frame, YUVFrame **queue)
{
#ifdef DEBUG
    fprintf(stderr, "Chosen: "); printFrame(frame);
#endif
    if (frame->prev == NULL) {
	*queue = frame->next;
	frame->next = NULL;
	if (*queue != NULL) 
	    (*queue)->prev = NULL;
    } else if (frame->next == NULL) {
	frame->prev->next = NULL;
	frame->prev = NULL;
    } else {
	frame->prev->next = frame->next;
	frame->next->prev = frame->prev;
	frame->next = frame->prev = NULL;
    }

}

static void
PurgeLateFrames(YUVFrame **queue, double speed, double now)
{
    YUVFrame *framePtr, *nextPtr;

    framePtr = *queue;
    nextPtr = NULL;

    while (framePtr != NULL && framePtr->next != NULL) {
	if (speed < 0.0) {
	    if (framePtr->logEnd > now) {
		nextPtr = framePtr->next;
		if (framePtr->prev == NULL) {
		    *queue = nextPtr;
		    if (*queue != NULL) 
			(*queue)->prev = NULL;
		} else if (nextPtr == NULL) {
		    framePtr->prev->next = NULL;
		} else {
		    framePtr->prev->next = nextPtr;
		    nextPtr->prev = framePtr->prev;
		}
		YUVFilter_FreeFrame(framePtr);

	    } else {
		break;
	    }

	/* speed >= 0.0 */
	} else {
	    if (framePtr->logEnd < now) {
 		nextPtr = framePtr->next;
 		if (framePtr->prev == NULL) {
 		    *queue = nextPtr;
 		    if (*queue != NULL) 
 			(*queue)->prev = NULL;
 		} else if (nextPtr == NULL) {
 		    framePtr->prev->next = NULL;
 		} else {
	 	    framePtr->prev->next = nextPtr;
 		    nextPtr->prev = framePtr->prev;
		}

#ifdef DEBUG
                fprintf(stderr, "Purging ");
#endif
 		YUVFilter_FreeFrame(framePtr);
            } else {
                break;
	    } 
	}  
 	framePtr = nextPtr;
    }

    *queue = framePtr;

    if (framePtr && framePtr->next != NULL)
	framePtr->prev = NULL;

}


/************************************************************
 *
 * YUVFilter_DoFilter
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/


static YUVFrame *
CopyFrame(YUVFrame *source)
{
    YUVFrame *new;
    char *data;


    new = ckalloc(sizeof(YUVFrame));
    nextYUVFilterNum++; 
    sprintf( new->name, "temp%d", nextYUVFilterNum );
    new->frameHeader = source->frameHeader;
    new->logStart = source->logStart;
    new->logEnd = source->logEnd;
    new->dataLength = source->dataLength;
    new->at_token = -1;
    new->objPtr = source->objPtr;
    new->next = NULL;
    new->prev = NULL;

    /* Create buffer for data, copy data */

    new->dataBuffer = MakeBuffer(source->objPtr->interp, 
				 new->dataLength,
				 "cmYUVFilter:CopyFrame");

    memcpy(new->dataBuffer->data, source->data, new->dataLength);
    new->data = new->dataBuffer->data;

    /* Create scatter buffer for data */
    new->sb = MakeScatterBuffer(new->dataBuffer, new->dataLength, NULL);

    return new;
}

static YUVFrame *
MakeBlackFrame(YUVFrame *source)
{
    YUVFrame *new;
    int ysize;

    ysize = source->dataLength / 2;

    new = ckalloc(sizeof(YUVFrame));
    nextYUVFilterNum++; 
    sprintf( new->name, "black%d", nextYUVFilterNum );
    new->frameHeader = source->frameHeader;
    new->logStart = source->logStart;
    new->logEnd = source->logEnd;
    new->dataLength = source->dataLength;
    new->at_token = -1;
    new->objPtr = source->objPtr;
    new->next = NULL;
    new->prev = NULL;

    /* Create buffer for data, initialize data */
    new->dataBuffer = MakeBuffer(source->objPtr->interp,
				 new->dataLength,
				 "cmYUVFilter:MakeBlackFrame");
    memset(new->dataBuffer->data,0,ysize);
    memset(new->dataBuffer->data+ysize,128,ysize);
    new->data = new->dataBuffer->data;

    /* Create scatter buffer for data */
    new->sb = MakeScatterBuffer(new->dataBuffer, new->dataLength, NULL);

    return new;
}

/*
 * a quick little hack here to keep from creating uneccesary black
 * frames.  Could be made cleaner, but don't want to complicate things
 * right now.
 * 
 * Also, I'm taking out code to check for different frame sizes... again,
 * could be put back in later.
 */

static YUVFrame *blackFrame;
static  int       blackFrameCreated = 0;

int
YUVFilter_DoFilter(YUVFilter *objPtr, YUVFrame *frameA, YUVFrame *frameB, double fstart,
		   double fend)
{
    YUVBufferHeader header;
    
    Buffer *hdrBuf, *chunkBuf, *dataBuffer;
    ScatterBuffer *sbPtr;
    SB_List *sbl;
    char *sbStr;
    int freeB = 0;
    int idx, dataLength, dataSegCode, width, height, ysize, uvsize;
    unsigned char *ya, *ua, *va, *yb, *ub, *vb, *result;
    double start, end;
    
    start = Tcm_ReadSysClock();
    

    /* Take out for now, we want to generate black output for the switcher */
    if (!objPtr->ready || objPtr->numInputs == 0) { 
 	return TCL_OK; 
    }
    
    if (frameA == NULL && (objPtr->numInputs == 1 || frameB == NULL)) { 
 	return TCL_OK; 
    }
    
    /* Sanity for null A and B frames */
    
    if (frameA == NULL) {
        if (!blackFrameCreated) {
            blackFrame = MakeBlackFrame(frameB);
            blackFrameCreated = 1;
        }
        frameA = blackFrame;
    }
    else if (frameB == NULL) {
        if (!blackFrameCreated) {
            blackFrame = MakeBlackFrame(frameA);
            blackFrameCreated = 1;
        }
        frameB = blackFrame;
    }

#if 0
    else {
        frameB = objPtr->blackB;
        if ((frameA->frameHeader.height != frameB->frameHeader.height) ||
            (frameA->frameHeader.width != frameB->frameHeader.width)) {
            YUVFilter_FreeFrame(frameB);
            objPtr->blackB = frameB = MakeBlackFrame(frameA);
        }
    }
    
    /*
     * If frame sizes don't match, set B to black
     */
    
    if ((frameA->frameHeader.height != frameB->frameHeader.height) ||
	(frameA->frameHeader.width != frameB->frameHeader.width)) {
	if (objPtr->blackB == NULL) {
	    objPtr->blackB = frameB = MakeBlackFrame(frameA);
	} else {
	    frameB = objPtr->blackB;
	    if ((frameA->frameHeader.height != frameB->frameHeader.height) ||
		(frameA->frameHeader.width != frameB->frameHeader.width)) {
		YUVFilter_FreeFrame(frameB);
		objPtr->blackB = frameB = MakeBlackFrame(frameA);
	    }
	}
    }
    
#endif

  START_FILTER:
    
    /* Since the data buffer for A is modified, we need to copy it */
    frameA = CopyFrame(frameA);

#ifdef DEBUG
    fprintf(stderr, "Filter: %s %s\n", frameA->name, frameB->name);
#endif

    /* Package Data to be sent out */
    /* First the header */

    header = frameA->frameHeader;

    width = header.width;
    height = header.height;
    ysize = width*height;
    uvsize = ysize/2;
    dataLength = 2*ysize;
    
    hdrBuf = MakeBuffer(objPtr->interp,
			sizeof(YUVBufferHeader),
			"cmYUVFilter:DoFilter");

    header.hdrType = htonl(CM_YUV_BUF_HDR);
    header.length = htonl(sizeof(YUVBufferHeader)-8);
    header.device = htonl(header.device);
    header.width = htonl(header.width);
    header.height = htonl(header.height);
    header.format = htonl(header.format);

    header.logStart_s = htonl(floor(fstart));
    header.logStart_ns = htonl(floor((fstart - floor(fstart))* ONE_BILLION));
    header.logEnd_s = htonl(floor(fend));
    header.logEnd_ns = htonl(floor((fend - floor(fend))* ONE_BILLION));

    idx=0;
    memcpy(hdrBuf->data+idx, &header, sizeof(YUVBufferHeader));
    idx+= sizeof(YUVBufferHeader);

    /* Now perform the effect on the data */
    
    ya = frameA->data;
    ua = ya+ysize;
    va = ua+uvsize;

    yb = frameB->data;
    ub = yb+ysize;
    vb = ub+uvsize;

    switch(objPtr->filter) {
    case TEXT_OVERLAY:
	if (objPtr->textFont == NULL || objPtr->textString == NULL) {
	    return TCL_OK;
	}
	
	/* Encode text string */
	if (objPtr->textEncoded == NULL) {
	    objPtr->textEncoded = bdftorlc(objPtr->textFont, 
					   NULL, NULL, 
					   objPtr->textString, 
					   FALSE);
	}
	if (objPtr->textEncoded == NULL) {
	    return TCL_OK;
	}
	result = TextOverlay (objPtr->textEncoded, ya, ua, va,
			      width, height,
			      objPtr->xoffset, objPtr->yoffset,
			      objPtr->textYUVColor.y,
			      objPtr->textYUVColor.u,
			      objPtr->textYUVColor.v,
			      NON_INTERLEAVED);
	if (result == NULL) {
	    return TCL_OK;
	}
	break;
    case FADE:
    {
	result = Fade (yb, ub, vb, ya, ua, va,
		       width, height,
		       objPtr->fadePercent, NON_INTERLEAVED);
	if (result == NULL) {
	    return TCL_OK;
	}
	break;
    }
    case PIC_IN_PIC:
    {
	result = PictInPict (yb, ub, vb, ya, ua, va,
			     width, height,
			     objPtr->xoffset, objPtr->yoffset,
			     objPtr->pinpScale, NON_INTERLEAVED);
	if (result == NULL) {
	    return TCL_OK;
	}
	break;
    }
    case NONE:
    default:
	break;
    }
    
    /* package up the data */

#ifdef DEBUG
    fprintf(stderr,"SendFrame: "); printFrame(frameA);
#endif
    chunkBuf = MakeBuffer(objPtr->interp, 
			  8, "cmYUVFilter:DoFilter");

    dataSegCode = htonl(CM_DATA_SEGMENT);
    dataLength = htonl(dataLength);

    memcpy(chunkBuf->data, &dataSegCode, 4);
    memcpy(chunkBuf->data+4, &dataLength, 4);

    dataBuffer = frameA->dataBuffer;

    /* 
     * Send buffer on its way
     */
    
    sbPtr = MakeScatterBuffer(hdrBuf, idx, 
			      chunkBuf, 8, 
			      dataBuffer, ntohl(dataLength),
			      NULL);

    sbl = MakeSBList(sbPtr, NULL);

    end = Tcm_ReadSysClock();

    if (objPtr->filterTime != 0.0) {
	objPtr->filterTime = ((0.80 * objPtr->filterTime) + 
			      (0.20 * (end - start)));
    } else {
	objPtr->filterTime = end - start;
    }


    /*
     * If out command is C, create scatter buffer list and call.
     */
    if (objPtr->outCCmd != NULL) {
	
	/*
	 * Need to check if it still exists since object may have
	 * been destroyed.
	 */
	if (Tutil_FindAtom(objPtr->interp, objPtr->outCmd+1) != NULL) {
	    objPtr->outCCmd(objPtr->outCD, sbl);
	} else {
	    /*
	     * Out C command has unexpectedly disappeared.
	     */
	    objPtr->outCCmd = NULL;
	    ckfree((char *) objPtr->outCmd);
	    objPtr->outCmd = NULL;
	}
    } else {
	/*
	 * Out command is tcl, make sb list into a string and
	 * call using Tcl_Eval. 
	 */
	char *sbStr;
	sbStr = MakeSBListString(sbl);
	Tcl_VarEval(objPtr->interp, objPtr->outCmd, " ",
		    sbStr, (char *) NULL); 
	ckfree((char *) sbStr);
    }

    /* Free sbl */
    FreeSBList(sbl);

    DetachBuffer(objPtr->interp, hdrBuf, 
		 "cmYUVFilter:DoFilter");

    DetachBuffer(objPtr->interp, chunkBuf, 
		 "cmYUVFilter:DoFilter");
	
    /* 
     * Free frameA since it is a copy
     */
	
    YUVFilter_FreeFrame(frameA);

    return TCL_OK;
}



/************************************************************
 *
 * YUVFilterSendFrame
 *
 * Parameters:
 *     ClientData clientdata
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static void
YUVFilterSendFrame( ClientData clientData)
{
    YUVFilter *objPtr = (YUVFilter *) clientData;
    double speed, now, when, awaken, purge, ftime;
    YUVFrame *aframe, *bframe;

#if 0
    /* increase frameRate to approach reqFrameRate */
    if (objPtr->atToken != -1) {
	if (objPtr->frameRate < objPtr->reqFrameRate) {
	    objPtr->frameRate++;
	}
    }
#endif


    if (!objPtr->ready)
	goto done;

    if (objPtr->lts == NULL)
	goto done;

    if (objPtr->outCmd == NULL)
	goto done;

    if (objPtr->atToken != -1) {
	/* when = objPtr->atWhen; */
	when = Tcm_ReadSysClock();
	now = LTS_LogicalTimeOf(objPtr->lts, when);
    } 
    else {
	when = Tcm_ReadSysClock();
	now = LTS_LogicalTimeOf(objPtr->lts, now);
    }

#ifdef DEBUG
    fprintf(stderr, "woken %f\n", now);
#endif
    
    objPtr->atToken = -1;
    speed = LTS_Speed(objPtr->lts);
    now += objPtr->sendAhead * speed;

    /* 
     *  Select A and B frames to filter
     */

    ftime = (1.0)/(objPtr->frameRate);

    aframe = GetFrameByTime( objPtr, &objPtr->queueA, 
			     now, now + ftime*speed);
    bframe = GetFrameByTime( objPtr, &objPtr->queueB, 
			     now, now + ftime*speed);

    /* 
     * Compare selected frame to previously filtered frame.
     * If either is NULL, pick the other.
     * Otherwise choose the frame with the greater end time.
     * 
     * If we take a frame from the queue we need to purge
     * the queue of earlier frames.
     *
     */


    if (aframe && objPtr->lastA) { 
	if (aframe->logStart > objPtr->lastA->logStart) {
	    purge = (speed > 0) ? (aframe->logStart) : (aframe->logEnd);
	    PurgeLateFrames( &objPtr->queueA, speed, purge);
	    YUVFilter_FreeFrame(objPtr->lastA);
	    objPtr->lastA = aframe;
	    RemoveFromQueue(aframe, &objPtr->queueA);
	} else {
	    aframe = objPtr->lastA;
	}	    
    } else if (objPtr->lastA) {
	aframe = objPtr->lastA;
    } else if (aframe) {
	purge = (speed > 0) ? (aframe->logStart) : (aframe->logEnd);
	PurgeLateFrames( &objPtr->queueA, speed, purge);
	objPtr->lastA = aframe;
	RemoveFromQueue(aframe, &objPtr->queueA);
    } 

    if (bframe && objPtr->lastB) { 
	if (bframe->logStart > objPtr->lastB->logStart) {
	    purge = (speed > 0) ? (bframe->logStart) : (bframe->logEnd);
	    PurgeLateFrames( &objPtr->queueB, speed, purge);
	    YUVFilter_FreeFrame(objPtr->lastB);
	    objPtr->lastB = bframe;
	    RemoveFromQueue(bframe, &objPtr->queueB);
	} else {
	    bframe = objPtr->lastB;
	}	    
    } else if (objPtr->lastB) {
	bframe = objPtr->lastB;
    } else if (bframe) {
	purge = (speed > 0) ? (bframe->logStart) : (bframe->logEnd);
	PurgeLateFrames( &objPtr->queueB, speed, purge);
	objPtr->lastB = bframe;
	RemoveFromQueue(bframe, &objPtr->queueB);
    } 

    /* perfrom filter operation */

    YUVFilter_DoFilter( objPtr, aframe, bframe, now, now + ftime*speed);
    
    /* 
     * Find next wake up time and set up 'at' event. 
     */
    if (speed == 0.0) {
	goto done;
    }
    if (objPtr->frameRate == 0) {
	goto done;
    }

    awaken = Tcm_ReadSysClock() + ftime;

    objPtr->atToken = Tcm_CreateAtHandler(awaken, HUGE_VAL, 
					  "video", YUVFilterSendFrame, objPtr, NULL, 
					  NULL, NULL, NULL, NULL);
    objPtr->atWhen = awaken;
    return;

  done:
#ifdef DEBUG
    fprintf(stderr, "sleeping...\n");
#endif

    objPtr->atToken = -1;
    return;
}

/* Called when the AtHandler misses the time window */

static void
YUVFilterMissed (ClientData clientData) 
{
    YUVFilter *objPtr = (YUVFilter *) clientData;

#ifdef DEBUG
    fprintf(stderr, "Missed window\n");
#endif

    /* decrement frameRate */
    if (objPtr->frameRate > 2) {
	objPtr->frameRate = objPtr->frameRate - 1;
    }
    
    objPtr->atToken = -1;

    /* Go Ahead and send some data */
    YUVFilterSendFrame(objPtr);
    return;

}    

/************************************************************
 *
 * YUVFilter_FreeFrame
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static void
YUVFilter_FreeFrame(YUVFrame *framePtr)
{
  int i;

  /*
   * Detach from buffer.
   */

#ifdef DEBUG
  fprintf(stderr, "Free: "); printFrame(framePtr);
#endif

  for (i=0; i<framePtr->sb->count; i++) {
      DetachBuffer(framePtr->objPtr->interp, framePtr->sb->buffers[i],
		   "cmYUVFilter:YUVFilter_FreeFrame");
  }

  FreeScatterBuffer(framePtr->sb);
  framePtr->sb = NULL;

  ckfree((char *) framePtr);

  return;
}


/************************************************************
 *
 * YUVFilter_Unready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
YUVFilter_Unready(YUVFilter * objPtr)
{
  int i;

  objPtr->ready = 0;

  if (objPtr->atToken != -1) {
      Tcm_DeleteAtHandler(objPtr->atToken);
  }
  objPtr->atToken = -1;
  objPtr->atWhen = HUGE_VAL;
  
  ResetFrameQueues(objPtr);

  return TCL_OK;
}



/************************************************************
 *
 * HELPER Functions
 *
 ************************************************************/

/************************************************************
 * ChangeFilter
 ************************************************************/

static int
ChangeFilter( YUVFilter *objPtr, int filter)
{
    int fd;

    /* Cleanup last filter */
    switch(filter) {
      case 0:
      case TEXT_OVERLAY:
      case FADE:
      case PIC_IN_PIC:      
      default:
	break;
    }

    /* Setup new filter */

    switch(filter) {
      case 0:
      case TEXT_OVERLAY:
      case FADE:
      case PIC_IN_PIC:      
      default:
	break;
    }
    return TCL_OK;
}


static void 
RGBtoYUV(unsigned short rr, unsigned short gg, unsigned short bb,
		     YUVColor *yuv)
{
    unsigned char y, u, v, r, g, b;

    r = rr / (unsigned short)256;
    g = gg / (unsigned short)256;
    b = bb / (unsigned short)256;

    y = .3*r + .6*g + .1*b;
    u = (((double)r - (double)y)/2.0) + 128.0;
    v = (((double)b - (double)y)/2.0) + 128.0;

    yuv->y = y;
    yuv->u = u;
    yuv->v = v;
}



