/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <math.h>

#ifdef NO_STRING_H
#   include "compat/string.h"
#else
#   include <string.h>
#endif

#include <sys/types.h>
#include <netinet/in.h>
#include <dp.h>
#include "cmEvery.h"
#include "cmYUVPlay.h"
#include "tcmEvent.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "cmBind.h"
#include "cmUtil.h"

#ifndef DEFAULT_YUV_PLAY_DEV_TYPE
#    define DEFAULT_YUV_PLAY_DEV_TYPE "software"
#endif

#define ONE_BILLION 1000000000.0

static int nextYUVResourceNum = 0;	/* Variable for next unique object
					 * name. */

/*
 * Prototypes for routines used internally
 */
static YUVResource *CreateNewYUVResource	(Tcl_Interp *interp,
						 char *name);
static void ChangeDeviceType			(YUVResource *objPtr,
						 char *devtype);
static int DestroyYUVResource			(YUVResource *objPtr);
static int YUVResourceObjCmd			(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc, char **argv);
static int YUV_Play_Accept			(YUVResource *objPtr,
						 SB_List * sbl);
static int YUV_Play_Ready			(YUVResource *objPtr);
static int YUV_Play_Unready			(YUVResource *objPtr);
static int YUV_Play_Stats			(YUVResource *objPtr);
static int YUV_Play_Refresh			(YUVResource *objPtr);
static void YUV_Play_LTSChanged		(ClientData clientData);
static void YUV_Play_FreeFrame		(YUVFrame *framePtr);
static int YUVResourceCreateCmd		(ClientData clientData,
						 Tcl_Interp *interp,
						 int argc, char **argv);
static int ConfigureYUVResource		(YUVResource *objPtr,
						 int argc, char **argv);
static int YUV_Play_DoDisplay		(YUVFrame *framePtr);
static int YUV_Play_DeviceList                (YUVResource *objPtr);
static char * GetHeight(Tcl_Interp *interp, ClientData clientdata);
static char * GetWidth(Tcl_Interp *interp, ClientData clientdata);
static char * GetDataLen(Tcl_Interp *interp, ClientData clientdata);
static char * GetLogEnd(Tcl_Interp *interp, ClientData clientdata);
static char * GetLogStart(Tcl_Interp *interp, ClientData clientdata);
static char * GetCumulLenDecoded(Tcl_Interp *interp, ClientData clientdata);
static void YUVInsertFrame(YUVFrame *newFrame,YUVFrame **queue, double speed);
static void YUVRemoveFrame(YUVFrame *frame, YUVFrame **queue);
static void YUVPurgeLateFrames(YUVFrame **queue, double speed, double now);
static YUVFrame *YUV_Play_GetNextFrame(YUVResource *objPtr);
int YUV_Play_DoDither(YUVResource *objPtr);
static void YUV_Play_ContinueDither (YUVFrame *framePtr);
int YUV_Play_DitherDone(YUVFrame *framePtr);



/************************************************************
 *
 * YUVResourceCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the "YUVResource" Tcl command.
 *     A new object of type YUVResource is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
YUVResourceCreateCmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char **argv)
{

    char new_name[256];
    YUVResource *new;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
			 " parent ?options?", (char *) NULL);
	return TCL_ERROR;
    }
    nextYUVResourceNum++;
    sprintf(new_name, "%syuvPlay%d", argv[1], nextYUVResourceNum);

    new = CreateNewYUVResource(interp, new_name);
    if (new == NULL) {
	Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_CreateCommand(interp, new_name, YUVResourceObjCmd,
		      (ClientData) new, (void (*)()) NULL);

    if (argc > 2) {
	ConfigureYUVResource(new, argc - 2, argv + 2);
    }
    Tcl_SetResult(interp, new_name, TCL_VOLATILE);
    return TCL_OK;
}

/************************************************************
 *
 * YUVResourceObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     YUVResource object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
YUVResourceObjCmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char **argv)
{
    char c;
    int len;
    YUVResource *objPtr = (YUVResource *) clientData;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <command> ?arguments?", (char *) NULL);
	goto usage;
    }
    c = argv[1][0];
    len = strlen(argv[1]);

    if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
	return ConfigureYUVResource(objPtr, argc - 2, argv + 2);
    } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
	return DestroyYUVResource(objPtr);
    } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
	SB_List *sbl = NULL;
	int result;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		argv[0], " <ScatterBufferList> ", (char *) NULL);
	    return TCL_ERROR;
	}

	/*
	 * Parse the ScatterBufferList
	 */
	sbl = (SB_List *) ckalloc(sizeof(SB_List));
	if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
	  ckfree((char *) sbl);
	  Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
			   argv[2],  (char *)NULL);
	  return TCL_ERROR;
	}

	result = YUV_Play_Accept(objPtr, sbl);
	if (sbl != NULL) {
	    FreeSBList(sbl);
	}
	return result;

    } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
	return YUV_Play_Ready(objPtr);

    } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
	return YUV_Play_Unready(objPtr);

    } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
	return YUV_Play_Stats(objPtr);

    } else if ((c == 'r') && (strncmp("refresh", argv[1], len) == 0)) {
	return YUV_Play_Refresh(objPtr);

    } else if ((c == 'd') && (strncmp("deviceList", argv[1], len) == 0)) {
	return YUV_Play_DeviceList(objPtr);
    }
    Tcl_AppendResult(interp, "Bad command option: ", argv[0], " ",
		argv[1], (char *) NULL);
usage:
    Tcl_AppendResult(interp,
		"\n<command> can be one of:\n",
		"  configure ?-slot ?value??\n",
		"  destroy\n",
		"  accept <ScatterBufferList>\n",
		"  ready\n",
		"  unready\n",
		"  stats\n",
		"  refresh\n",
		"  deviceList\n", 
		(char *) NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * ChangeDeviceType --
 *
 *	Sets the device type for the YUVReseource.  Called whenever
 *	the "devtype" slot is changed, and on initialization of the
 *	resource.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The old device is shut down, if appropriate, and the new
 *	device is created.
 *
 * Parameters:
 *	YUVResource *objPtr	Pointer to resource object.
 *	char *devtype		String giving name of new device
 *
 *--------------------------------------------------------------
 */
static void
ChangeDeviceType (YUVResource *objPtr, char *devtype)
{
    int len;
    char c;

    len = strlen(devtype);
    c = devtype[0];

    /*
     * If there's no change, return
     */
    if ((objPtr->devtype != NULL) &&
	(strncmp (objPtr->devtype, devtype, len) == 0)) {
	return;
    }

    /*
     * Shut down old device, if necessary, ignoring any errors it
     * returns.
     */
    YUV_Play_Unready(objPtr);
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
	objPtr->devtype = NULL;;
    }
    if (objPtr->devicePtr != NULL) {
	objPtr->devicePtr->shutdownProc(objPtr);
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
	objPtr->inited = 0;
    }

    /*
     * Create new device.
     */
    if ((c == 's') && (strncmp("software", devtype, len) == 0)) {
	objPtr->devicePtr = GetSoftwareYUVDevice(objPtr);
	objPtr->devtype = ckalloc(strlen("software")+1);
	strcpy(objPtr->devtype, "software");
    } else {
	objPtr->devicePtr = GetDefaultYUVDevice(objPtr);
	objPtr->devtype = ckalloc(strlen(DEFAULT_YUV_PLAY_DEV_TYPE)+1);
	strcpy(objPtr->devtype, DEFAULT_YUV_PLAY_DEV_TYPE);
    }
}

/************************************************************
 *
 * CreateNewYUVResource
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static YUVResource *
CreateNewYUVResource(Tcl_Interp *interp, char *name)
{
    YUVResource *new;
    char *acceptName;
    char **event_list, **subst_list;
    CM_BindProc **subst_procs;
    ClientData *subst_cds;
    int i;

    new = (YUVResource *) ckalloc(sizeof(YUVResource));
    new->interp = interp;
    new->name = (char *) ckalloc(strlen(name) + 1);
    strcpy(new->name, name);

    new->width = 0;
    new->height = 0;
    new->xpos = -1;
    new->ypos = -1;
    new->squeeze = 0;
    new->devtype = NULL;
    new->devicePtr = NULL;
    new->xid = 0;
    new->lts = NULL;
    new->ltsTriggerNum = -1;
    new->ltsName = NULL;
    new->maxFrameRate = 0;
    new->ctrlCallback = NULL;
    new->ctrlCycleTime = 0.0;
    new->statStartTime = 0.0;
    new->ditherTime = 0.0;
    new->missed = 0;
    new->late = 0;
    new->played = 0;
    new->bytes = 0;
    new->ready = 0;
    new->inited = 0;
    new->cntrlEveryHandle = NULL;
    new->destroyFlag = 0;
    new->at_tokens = NULL;
    new->num_at_tokens = 0;
    new->queue = NULL;
    new->indecode = 0;
    /*
     * Set up bindings with binding mechanism.
     */

    event_list = (char **) ckalloc(5*sizeof(char *));
    event_list[0] = (char *) ckalloc(strlen("<FRAME_DROPPED>")+1);
    strcpy(event_list[0], "<FRAME_DROPPED>");
    event_list[1] = (char *) ckalloc(strlen("<FRAME_RECEIVED>")+1);
    strcpy(event_list[1], "<FRAME_RECEIVED>");
    event_list[2] = (char *) ckalloc(strlen("<FRAME_PLAYED>")+1);
    strcpy(event_list[2], "<FRAME_PLAYED>");
    event_list[3] = (char *) ckalloc(strlen("<FRAME_DECODE_BEGIN>")+1);
    strcpy(event_list[3], "<FRAME_DECODE_BEGIN>");
    event_list[4] = (char *) ckalloc(strlen("<FRAME_DECODE_DONE>")+1);
    strcpy(event_list[4], "<FRAME_DECODE_DONE>");

    subst_list = (char **) ckalloc(6*sizeof(char *));
    subst_procs = (CM_BindProc **) ckalloc(6*sizeof(CM_BindProc *));
    subst_cds = (ClientData *) ckalloc(6*sizeof(ClientData));

    subst_list[0] = (char *) ckalloc(2);
    (subst_list[0])[0] = 's';
    (subst_list[0])[1] = '\0';
    subst_procs[0] = GetLogStart;
    subst_cds[0] = (ClientData) new;

    subst_list[1] = (char *) ckalloc(2);
    (subst_list[1])[0] = 'e';
    (subst_list[1])[1] = '\0';
    subst_procs[1] = GetLogEnd;
    subst_cds[1] = (ClientData) new;

    subst_list[2] = (char *) ckalloc(2);
    (subst_list[2])[0] = 'l';
    (subst_list[2])[1] = '\0';
    subst_procs[2] = GetDataLen;
    subst_cds[2] = (ClientData) new;

    subst_list[3] = (char *) ckalloc(2);
    (subst_list[3])[0] = 'w';
    (subst_list[3])[1] = '\0';
    subst_procs[3] = GetWidth;
    subst_cds[3] = (ClientData) new;

    subst_list[4] = (char *) ckalloc(2);
    (subst_list[4])[0] = 'h';
    (subst_list[4])[1] = '\0';
    subst_procs[4] = GetHeight;
    subst_cds[4] = (ClientData) new;

    subst_list[5] = (char *) ckalloc(2);
    (subst_list[5])[0] = 'b';
    (subst_list[5])[1] = '\0';
    subst_procs[5] = GetCumulLenDecoded;
    subst_cds[5] = (ClientData) new;

    new->bindToken = CreateCMBinding(interp,
				     new->name,
				     5, event_list,
				     6, subst_list, subst_procs, subst_cds);

    for (i=0; i<3; i++) {
      ckfree(event_list[i]);
    }
    ckfree((char *) event_list);

    for (i=0; i<6; i++) {
      ckfree(subst_list[i]);
    }
    ckfree((char *) subst_list);
    ckfree((char *) subst_procs);
    ckfree((char *) subst_cds);

    ChangeDeviceType (new, DEFAULT_YUV_PLAY_DEV_TYPE);

    /*
     * Register accept as an atom.
     */
    acceptName = (char *) ckalloc(strlen(name)+9);
    sprintf(acceptName, "%s.%s", name, "accept");
    Tutil_RegisterCommand(interp, acceptName, (ClientData) YUV_Play_Accept,
			  (ClientData) new);
    new->acceptHandle = acceptName;

    return new;
}


/************************************************************
 *
 * DestroyYUVResource
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

static int
DestroyYUVResource(YUVResource *objPtr)
{
    int result = TCL_OK;
    int i, devReturn;

    /*
     * First delete Tcl command.
     */
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);

    /*
     * Unregister the accept atom
     */
    if (objPtr->acceptHandle != NULL) {
	Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
	ckfree(objPtr->acceptHandle);
	objPtr->acceptHandle = NULL;
    }

    /*
     * Shut down the device
     */
    if (objPtr->devicePtr != NULL) {
	devReturn = objPtr->devicePtr->shutdownProc(objPtr);
	if (devReturn != 0) {
	    result = TCL_ERROR;
	    if (objPtr->devicePtr->errorProc != NULL) {
		Tcl_AppendResult(objPtr->interp,
			objPtr->devicePtr->errorProc(objPtr, devReturn),
			(char *) NULL);
	    }
	}
	ckfree((char *) objPtr->devicePtr);
	objPtr->devicePtr = NULL;
    }

    /*
     * Clean up lts
     */
    if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
    }

    /*
     * Release our lock on the old lts...
     */
    if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	}
        Tk_Release(objPtr->lts);
        objPtr->lts = NULL;
    }

    /*
     * Free up previous allocs, frame queue, and such.
     */
    if (objPtr->devtype != NULL) {
	ckfree(objPtr->devtype);
    }
    if (objPtr->name != NULL) {
	ckfree(objPtr->name);
    }
    if (objPtr->ctrlCallback != NULL) {
	free(objPtr->ctrlCallback);
    }
    if (objPtr->cntrlEveryHandle != NULL) {
	Tcm_DeleteEvery(objPtr->interp, objPtr->cntrlEveryHandle, NULL);
	objPtr->cntrlEveryHandle = NULL;
    }

    /*
     * Clean up previously scheduled display requests
     */
    if (objPtr->at_tokens != NULL) {
	for (i=0; i<objPtr->num_at_tokens; i++) {
	    if (objPtr->at_tokens[i] > 0) {
		AtEvent *at_event;
		int token;

		token = objPtr->at_tokens[i];
		at_event = Tcm_FindAtEvent(token);
		if (token && (at_event != NULL)) {
		    YUV_Play_FreeFrame((YUVFrame *) at_event->fireClientData);
		    Tcm_DeleteAtHandler(token);
		    objPtr->at_tokens[i] = 0;
		}
	    }
	}
	ckfree((char *) objPtr->at_tokens);
    }

    /*
     * Unregister with bind mechanism.
     */

    DestroyCMBinding(objPtr->interp, objPtr->bindToken);

    ckfree((char *) objPtr);
    return result;
}


/************************************************************
 *
 * ConfigureYUVResource
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

static int
ConfigureYUVResource(YUVResource * objPtr, int argc, char **argv)
{

    Tcl_Interp *interp;
    int cntr;
    char str[50];

    interp = objPtr->interp;

    /*
     * Case 1: <obj> configure
     */
    if (argc == 0) {
	sprintf(str, "{%d}", objPtr->width);
	Tcl_AppendResult(interp, "{", "-width ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->height);
	Tcl_AppendResult(interp, "{", "-height ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->xpos);
	Tcl_AppendResult(interp, "{", "-xpos ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->ypos);
	Tcl_AppendResult(interp, "{", "-ypos ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->squeeze);
	Tcl_AppendResult(interp, "{", "-squeeze ", "0 ", str, "} ",
			 (char *)NULL);
	sprintf(str, "{%d}", objPtr->xid);
	Tcl_AppendResult(interp, "{", "-xid ", "0 ", str, "} ",
			 (char *)NULL);
	Tcl_AppendResult(interp, "{", "-device ", "{", DEFAULT_YUV_PLAY_DEV_TYPE, "} ",
			 objPtr->devtype, "} ", (char *)NULL);

	if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{}", "} ",
			 (char *)NULL);
	} else {
	  Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{",
			   objPtr->ltsName, "}", "} ", (char *) NULL);
	}
	sprintf(str, "{%d}", objPtr->maxFrameRate);
	Tcl_AppendResult(interp, "{", "-maxFrameRate ", "0 ", str, "} ",
			 (char *) NULL);
	if (objPtr->ctrlCallback == NULL) {
	    Tcl_AppendResult(interp, "{", "-ctrlCallback ", "{} ", "{}",
			    "} ", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "{", "-ctrlCallback ", "{} ",
			    "{", objPtr->ctrlCallback, "}", "} ", (char *) NULL);
	}
	Tcl_PrintDouble(interp, objPtr->ctrlCycleTime, str);
	Tcl_AppendResult(interp, "{", "-ctrlCycleTime ", "0.0 ",
			str, "} ", (char *) NULL);

	/*
	 * Give the device a chance to add any options of its own.
	 */

	if (objPtr->devicePtr != NULL) {
	    if (objPtr->devicePtr->configProc != NULL) {
		objPtr->devicePtr->configProc(objPtr, NULL, NULL);
	    }
	}
	return TCL_OK;
    }

    /*
     * Case 2: <obj> configure <-slot>
     */
    if (argc == 1) {
	char c;
	int len;

	len = strlen(argv[0]);
	c = argv[0][1];

	if ((c == 'w') && (strncmp("-width", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->width);
	    Tcl_AppendResult(interp, "-width ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'h') && (strncmp("-height", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->height);
	    Tcl_AppendResult(interp, "-height ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'x') && (strncmp("-xpos", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->xpos);
	    Tcl_AppendResult(interp, "-xpos ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'y') && (strncmp("-ypos", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->ypos);
	    Tcl_AppendResult(interp, "-ypos ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 's') && (strncmp("-squeeze", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->squeeze);
	    Tcl_AppendResult(interp, "-squeeze ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'x') && (strncmp("-xid", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->xid);
	    Tcl_AppendResult(interp, "-xid ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'd') && (strncmp("-device", argv[0], len) == 0)) {
	    Tcl_AppendResult(interp, "-device ", "{", DEFAULT_YUV_PLAY_DEV_TYPE, "} ", 
			     objPtr->devtype,
			     (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
	  if (objPtr->ltsName == NULL) {
	    Tcl_AppendResult(interp, "-lts ", "{} ", "{}", (char *) NULL);
	  } else {
	    Tcl_AppendResult(interp, "-lts ", "{} ", "{", objPtr->ltsName, "}", (char *) NULL);
	  }
	    return TCL_OK;
	} else if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
	    sprintf(str, "{%d}", objPtr->maxFrameRate);
	    Tcl_AppendResult(interp, "-maxFrameRate ", "0 ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((c == 'c') && (strncmp("-ctrlCallback", argv[0], len) == 0)) {
	  if (objPtr->ctrlCallback == NULL) {
	    Tcl_AppendResult(interp, "-ctrlCallback ", "{} ", "{}", (char *) NULL);
	  } else {
	    Tcl_AppendResult(interp, "-ctrlCallback ", "{} ", "{", objPtr->ctrlCallback, "}", (char *) NULL);
	  }
	    return TCL_OK;
	} else if ((c == 'c') && (strncmp("-ctrlCycleTime", argv[0], len) == 0)) {
	    char str[TCL_DOUBLE_SPACE];

	    Tcl_PrintDouble(interp, objPtr->ctrlCycleTime, str);
	    Tcl_AppendResult(interp, "-ctrlCycleTime ", "{0.0} ", str, (char *) NULL);
	    return TCL_OK;
	} else if ((objPtr->devicePtr != NULL) && 
		   (objPtr->devicePtr->configProc != NULL)) {
	    if (objPtr->devicePtr->configProc(objPtr, argv[0], NULL) != 0) 
	      goto bad_option;
	    return TCL_OK;
	}
      bad_option:
	Tcl_AppendResult(interp, "bad configuration option: ", argv[0], (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Case 3: <obj> configure <-slot> <newValue>
     */
    cntr = 0;
    while (cntr < argc) {
	char c;
	int len;

	/*
	 * Make resource unready since something is changing.
	 */

	YUV_Play_Unready(objPtr);

	len = strlen(argv[cntr]);
	c = argv[cntr][1];

	if ((c == 'w') && (strncmp("-width", argv[cntr], len) == 0)) {
	    int newWidth;

	    cntr++;
	    if (cntr == argc) {
noValueError:
		Tcl_AppendResult(interp, "Must specify value when ",
			"trying to set configuration option ",
			argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newWidth) != TCL_OK) {
badValueError:
		Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
				 "\" given for configuration option ",
				 argv[cntr-1], (char *) NULL);
		return TCL_ERROR;
	    }
	    objPtr->width = newWidth;
	} else if ((c == 'h') && (strncmp("-height", argv[cntr], len) == 0)) {
	    int newHeight;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newHeight) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->height = newHeight;
	} else if ((c == 'x') && (strncmp("-xpos", argv[cntr], len) == 0)) {
	    int newXpos;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newXpos) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->xpos = newXpos;
	} else if ((c == 'y') && (strncmp("-ypos", argv[cntr], len) == 0)) {
	    int newYpos;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newYpos) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->ypos = newYpos;
	} else if ((c == 's') && (strncmp("-squeeze", argv[cntr], len) == 0)) {
	    int newSqueeze;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newSqueeze) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->squeeze = newSqueeze;
	} else if ((c == 'x') && (strncmp("-xid", argv[cntr], len) == 0)) {
	    int newXid;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newXid) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->xid = newXid;
	} else if ((c == 'd') && (strncmp("-device", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    ChangeDeviceType(objPtr, argv[cntr]);
	} else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
	    ClientData newLts;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (strlen(argv[cntr]) == 0) {
	        newLts = NULL;
	    }
	    else if ((newLts = Tcm_FindAtom(interp, argv[cntr])) == NULL) {
		goto badValueError;
	    }
	    if (objPtr->ltsName != NULL) {
		ckfree(objPtr->ltsName);
		objPtr->ltsName = NULL;
	    }

            /*
             * Release our lock on the old lts...
             */
            if (objPtr->lts != NULL) {
	        if (objPtr->ltsTriggerNum != -1) {
		    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
		    objPtr->ltsTriggerNum = -1;
	        }
                Tk_Release(objPtr->lts);
                objPtr->lts = NULL;
            }

	    objPtr->lts = newLts;
	    if (newLts != NULL) {
		Tk_Preserve (objPtr->lts);
	        objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts, 
					 LTS_SPEED_FLAG | LTS_VALUE_FLAG |
					 LTS_EARLY_PHASE, 
					 YUV_Play_LTSChanged,
					 (ClientData) objPtr);
	        objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
	        strcpy(objPtr->ltsName, argv[cntr]);
	    } else {
		objPtr->ltsName = NULL;
	    }
	} else if ((c == 'm') &&
		   (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
	    int newMaxFrameRate;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetInt(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->maxFrameRate = newMaxFrameRate;
	} else if ((c == 'c') && (strncmp("-ctrlCallback", argv[cntr], len) == 0)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (objPtr->ctrlCallback != NULL) {
		ckfree(objPtr->ctrlCallback);
	    }
	    objPtr->ctrlCallback = (char *) ckalloc(strlen(argv[cntr]) + 1);
	    strcpy(objPtr->ctrlCallback, argv[cntr]);
	} else if ((c == 'c') &&
		   (strncmp("-ctrlCycleTime", argv[cntr], len) == 0)) {
	    double newCtrlCycleTime;

	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (Tcl_GetDouble(interp,argv[cntr],&newCtrlCycleTime) != TCL_OK) {
		goto badValueError;
	    }
	    objPtr->ctrlCycleTime = newCtrlCycleTime;
	} else if ((objPtr->devicePtr != NULL) && 
		   (objPtr->devicePtr->configProc != NULL)) {
	    cntr++;
	    if (cntr == argc) {
		goto noValueError;
	    }
	    if (objPtr->devicePtr->configProc(objPtr, argv[cntr-1], argv[cntr]) != 0) {
		return TCL_ERROR;
	    }
	}
	cntr++;
    }
    return TCL_OK;
}


/************************************************************
 *
 * YUV_Play_Ready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
YUV_Play_Ready(YUVResource * objPtr)
{
    int init_res;

    /*
     * Can only be made ready if
     *  1. We have a device.
     *  2. We have an lts
     *  3. We have a window
     */
    if (objPtr->devicePtr == NULL) {
        Tcl_AppendResult(objPtr->interp, "Can't make YUV resource \"",
			 objPtr->name, "\" ready: device type is not set.", (char *)NULL);
        return TCL_ERROR;
    }
    if (objPtr->lts == NULL) {
        Tcl_AppendResult(objPtr->interp, "Can't make YUV resource \"",
                objPtr->name, "\" ready: no lts has been attached.", (char *)NULL);
        return TCL_ERROR;
    }

    if (objPtr->xid == 0) {
        Tcl_AppendResult(objPtr->interp, "Can't make YUV resource \"",
                objPtr->name, "\" ready: output window not specified.",
		(char *)NULL);
        return TCL_ERROR;
    }

    /*
     * If not inited, do initProc, catch result.
     */
    init_res = 0;
    if (!objPtr->inited) {
      if (objPtr->devicePtr->initProc != NULL) {
	init_res = objPtr->devicePtr->initProc(objPtr);
      }
      if (init_res == 0) 
	objPtr->inited = 1;
    }
    else {
      /*
       * Otherwise, do reinitProc, catch result.
       */
      if (objPtr->devicePtr->reinitProc != NULL) {
	init_res = objPtr->devicePtr->reinitProc(objPtr);
      }
    }
      
    /*
     * If bad result, get error string, return error.
     */
    if (init_res != 0) {
      char *errStr;

      if (objPtr->devicePtr->errorProc != NULL) {
	  errStr = objPtr->devicePtr->errorProc(objPtr, init_res);
      } else {
	  errStr = "Error with device initialization.";
      }

      Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
      return TCL_ERROR;
    }

    objPtr->ready = 1;

    /*
     * Reset stat stuff.
     */

    objPtr->statStartTime = Tcm_ReadSysClock();
    objPtr->missed = objPtr->played = objPtr->late = objPtr->bytes = 0;

    return TCL_OK;
}


/************************************************************
 *
 * YUV_Play_LTSChanged
 *
 * Parameters:
 *     clientdata - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *      Called when the LTS changes value.  Dequeues any previously
 *      queued play requests, frees any memory we're holding, and
 *      returns.
 *
 ************************************************************/
static void
YUV_Play_LTSChanged(ClientData clientData)
{
    YUVResource *objPtr = (YUVResource *) clientData;
    int i;

    /*
     * If not ready, do nothing.
     */
    if (!objPtr->ready) {
	return;
    }

    /*
     * Blow away "at" queue.
     */
    for (i=0; i<objPtr->num_at_tokens; i++) {
	if (objPtr->at_tokens[i] > 0) {
	    AtEvent *at_event;
	    int token;

	    token = objPtr->at_tokens[i];
	    at_event = Tcm_FindAtEvent(token);
	    if (token && (at_event != NULL)) {
		YUV_Play_FreeFrame((YUVFrame *) at_event->fireClientData);
		Tcm_DeleteAtHandler(token);
		objPtr->at_tokens[i] = 0;
	    }
	}
    }

    /*
     * Reset all the statistics stuff.
     */
    objPtr->statStartTime = Tcm_ReadSysClock();
    objPtr->missed = objPtr->late = objPtr->played = objPtr->bytes = 0;
}

/************************************************************
 *
 * YUV_Play_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
YUV_Play_Accept(YUVResource *objPtr, SB_List *sbl)
{
    int i, new_num;
    YUVFrame *newFrame, *framePtr, *prevPtr, *nextPtr;
    double speed, now, eTime, sTime;
    signed char *sb_data;
    Buffer *bufPtr;

    if (!objPtr->ready) {
	return TCL_OK;
    }

    speed = LTS_Speed(objPtr->lts);
    now = LTS_Value(objPtr->lts);

    if (sbl == NULL) return TCL_OK;
    if (sbl->count < 1) return TCL_OK;

    /*
     * Increment reference count on buffers passed in
     */
    AttachToAllBuffers(objPtr->interp, sbl,
		       "cmYUVPlay:YUV_Play_Accept");

    /*
     * For each frame in sb list, create frame structure,
     * fill in info and add to queue.
     */

    for (i=0; i<sbl->count; i++) {
	newFrame = (YUVFrame *) ckalloc(sizeof(YUVFrame));
	
	sb_data = GetFromSB(sbl->sbs[i], CM_YUV_BUF_HDR, &bufPtr, NULL);

	if (sb_data == NULL) {
	    ckfree((char *) newFrame);
	    continue;
	}
	memcpy(&(newFrame->frameHeader), sb_data-8, sizeof(YUVBufferHeader));
	newFrame->frameHeader.hdrType = ntohl(newFrame->frameHeader.hdrType);
	newFrame->frameHeader.length = ntohl(newFrame->frameHeader.length);
	newFrame->frameHeader.width = ntohl(newFrame->frameHeader.width);
	newFrame->frameHeader.height = ntohl(newFrame->frameHeader.height);
	newFrame->frameHeader.device = ntohl(newFrame->frameHeader.device);
	newFrame->frameHeader.logEnd_s = ntohl(newFrame->frameHeader.logEnd_s);
	newFrame->frameHeader.logEnd_ns = ntohl(newFrame->frameHeader.logEnd_ns);
	newFrame->frameHeader.logStart_s = ntohl(newFrame->frameHeader.logStart_s);
	newFrame->frameHeader.logStart_ns = ntohl(newFrame->frameHeader.logStart_ns);
	
	eTime = ((double )newFrame->frameHeader.logEnd_s) + 
	  (((double) newFrame->frameHeader.logEnd_ns) / ONE_BILLION);
	sTime = ((double )newFrame->frameHeader.logStart_s) + 
	  (((double) newFrame->frameHeader.logStart_ns) / ONE_BILLION);
	
	newFrame->data = GetFromSB(sbl->sbs[i], CM_DATA_SEGMENT, &bufPtr, 
				   &newFrame->dataLength);
	if (newFrame->data == NULL) {
	    ckfree((char *) newFrame);
	    continue;
	}

	newFrame->deviceData = NULL;
	newFrame->at_token = -1;
	newFrame->objPtr = objPtr;

	newFrame->dataBuffer = bufPtr;
	newFrame->logStart = sTime;
	newFrame->logEnd = eTime;
	newFrame->sb = CopyScatterBuffer(sbl->sbs[i]);

	/*
	 * Generate FRAME_RECEIVED event.
	 */
	
	objPtr->bindFramePtr = newFrame;
	ExecuteCMBinding(objPtr->interp, objPtr->bindToken,"<FRAME_RECEIVED>");

	/*
	 * If not ready, free frame and continue
	 */

	if (!objPtr->ready) {
	    YUV_Play_FreeFrame(newFrame);
	    continue;
	}

	/*
	 * Gather stats.
	 */
	objPtr->bytes += newFrame->dataLength;

#if 0        
	/* 
	 * Drop frame if it is too late 
	 */

	if (speed != 0.0 && newFrame->logEnd < now) {
	    YUV_Play_FreeFrame(newFrame);
	    continue;
	}
#endif
        
	/*
	 * Insert frame into our queue 
	 */
	
	YUVInsertFrame( newFrame, &objPtr->queue, speed );
    }

    /* Start a dither */
    if (objPtr->indecode == 0) {
	YUV_Play_DoDither(objPtr);
    }

    return TCL_OK;
}

/************************************************************
 *
 * YUV_Play_DoDither
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
int
YUV_Play_DoDither(YUVResource *objPtr)
{
    YUVFrame *frame_ptr;

    /* 
     * If not ready or already in decode, do nothing.
     */
    if (!objPtr->ready || objPtr->indecode) {
	return TCL_OK;
    }

    /*
     * Get next eligible frame.
     */
    frame_ptr = YUV_Play_GetNextFrame(objPtr);
    if (frame_ptr == NULL) {
	return TCL_OK;
    }

    /*
     * Generate FRAME_DECODE_BEGIN event.
     */
    objPtr->indecode = 1;

    objPtr->bindFramePtr = frame_ptr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		     "<FRAME_DECODE_BEGIN>");

    /*
     * Start the dithering process
     */

    Tcm_DoWhenIdle("video", (Tk_IdleProc *) YUV_Play_ContinueDither, frame_ptr);

    /* YUV_Play_ContinueDither (frame_ptr); */

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * YUV_Play_ContinueDither --
 *
 *	Continues an asynchronous dither until it's finished.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	May schedule a frame for display if the decoding is finished,
 *	otherwise, schedules itself to be called again when processor
 *	is idle.
 *
 *--------------------------------------------------------------
 */
static void
YUV_Play_ContinueDither (YUVFrame *framePtr)
{
    YUVResource *objPtr;

    objPtr = framePtr->objPtr;

    if (!objPtr->ready) {
	return;
    }

    /*
     * Call device dither procedure.
     */
    if (objPtr->devicePtr != NULL) {
        if (objPtr->devicePtr->ditherProc != NULL) {
	    objPtr->devicePtr->ditherProc(framePtr);
        }
    }

    YUV_Play_DitherDone(framePtr);
    return;
}

/************************************************************
 *
 * YUV_Play_DitherDone
 *
 * Parameters:
 *     framePtr - pointer to frame just finished with decoding.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
int
YUV_Play_DitherDone(YUVFrame *framePtr)
{
    int i, new_num;
    YUVResource *objPtr;

    objPtr = framePtr->objPtr;

    objPtr->indecode = 0;

    /*
     * If not ready, free frame and return.
     */
    if (!objPtr->ready) {
	YUV_Play_FreeFrame(framePtr);
	return TCL_OK;
    }

    /*
     * Gather stats.
     */
    objPtr->bytes += framePtr->dataLength;

    /*
     * Generate FRAME_DECODE_DONE event.
     */

    objPtr->bindFramePtr = framePtr;
    ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		     "<FRAME_DECODE_DONE>");

    /*
     * Schedule display.  Search for a free at_token.  If we can't find
     * one, we need to increase size of at_token array
     */
    if (objPtr->num_at_tokens == 0) {
	objPtr->num_at_tokens = 30;
	objPtr->at_tokens = (int *) ckalloc(sizeof(int) * 30);
	for (i=0; i<objPtr->num_at_tokens; i++) {
	    objPtr->at_tokens[i] = 0;
	}
    } 
    for (i=0; i<objPtr->num_at_tokens; i++) {
	if (objPtr->at_tokens[i] == 0) {
	    break;
	}
    }
    if (i == objPtr->num_at_tokens) {
	new_num = objPtr->num_at_tokens + 10;
	objPtr->at_tokens = (int *) realloc((char *) objPtr->at_tokens,
						      sizeof(int) * new_num);
	for (i=objPtr->num_at_tokens; i < new_num; i++) {
	    objPtr->at_tokens[i] = 0;
	}
	i = objPtr->num_at_tokens;
	objPtr->num_at_tokens = new_num;
    }

    /*
     * If speed is zero throw up the current
     * frame.  Otherwise, create an atEvent for it.
     */
    if (LTS_Speed(objPtr->lts) == 0.0) {
	framePtr->at_token = -1; 
	YUV_Play_DoDisplay(framePtr);
    } else {
	double t1, t2;
	t1 = LTS_SystemTimeOf(objPtr->lts, framePtr->logStart);
	t2 = LTS_SystemTimeOf(objPtr->lts, framePtr->logEnd);
	objPtr->at_tokens[i] = 
		Tcm_CreateAtHandler(t1<t2?t1:t2, HUGE_VAL,
			    "video", (Tk_TimerProc *) YUV_Play_DoDisplay, 
			    (ClientData) framePtr, NULL,
			    NULL, (Tk_IdleProc *) NULL,
			    (ClientData) NULL, NULL);
	framePtr->at_token = i;
    }
		
    /*
     * Arrange to dither next frame when we're idle
     */
    if (objPtr->queue != NULL) {
	Tcm_DoWhenIdle("video", (Tk_IdleProc *) YUV_Play_DoDither, objPtr);
    }

    return TCL_OK;
}

/************************************************************
 *
 * YUV_Play_GetNextFrame
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Acgtion:
 *
 ************************************************************/

static YUVFrame *
YUV_Play_GetNextFrame(YUVResource *objPtr)
{
  double now, speed;
  YUVFrame *chosenFrame, *framePtr, *nextPtr;

  if (objPtr->queue ==  NULL) return NULL;

  now = LTS_Value(objPtr->lts);
  speed = LTS_Speed(objPtr->lts);

  if (objPtr->devicePtr != NULL) {
      if (objPtr->devicePtr->nextFrameProc != NULL) {
	  if (objPtr->devicePtr->nextFrameProc(objPtr, objPtr->queue, 
					       &chosenFrame) != 0) {
	      return NULL;
	  }
      } else {
	  chosenFrame = objPtr->queue;
	  while (chosenFrame->next != NULL) {
	      if (speed < 0.0) {
		  if (chosenFrame->logEnd < now) break;
	      } else {
		  if (chosenFrame->logEnd > now) break;
	      }
	      chosenFrame = chosenFrame->next;
	  }
      }
  }

  if (chosenFrame == NULL)
      return NULL;

  YUVRemoveFrame( chosenFrame, &objPtr->queue );
  YUVPurgeLateFrames( &objPtr->queue, speed, now );

  return chosenFrame;
}

/************************************************************
 *
 * YUV_Play_Unready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
YUV_Play_Unready(YUVResource * objPtr)
{
  int i;

  objPtr->ready = 0;

  /*
   * Blow away "at" queue.
   */
  for (i =0; i<objPtr->num_at_tokens; i++) {
      if (objPtr->at_tokens[i] > 0) {
	  AtEvent *at_event;
	  int token;

	  token = objPtr->at_tokens[i];
	  at_event = Tcm_FindAtEvent(token);
	  if (token && (at_event != NULL)) {
	      YUV_Play_FreeFrame((YUVFrame *) at_event->fireClientData);
	      Tcm_DeleteAtHandler(token);
	      objPtr->at_tokens[i] = 0;
	  }
      }
  }

  return TCL_OK;
}

/************************************************************
 *
 * YUV_Play_Stats
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
YUV_Play_Stats(YUVResource *objPtr)
{
    sprintf (objPtr->interp->result, "%f %d %d %d %d",
	     objPtr->ditherTime,
	     objPtr->missed, objPtr->late, objPtr->played, objPtr->bytes);

    return TCL_OK;
}

/************************************************************
 *
 * YUV_Play_DoDisplay
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
YUV_Play_DoDisplay(YUVFrame *framePtr)
{
  YUVResource *objPtr;
  double now, speed;

  objPtr = framePtr->objPtr;

  now = LTS_Value(objPtr->lts);
  speed = LTS_Speed(objPtr->lts);

  /*
   * Collect stats.
   */
  objPtr->played++;
      
  /*
   * Call device's display procedure.
   */
  if (objPtr->devicePtr != NULL) {
    if (objPtr->devicePtr->displayProc != NULL) {
      objPtr->devicePtr->displayProc(framePtr);
    }
  }

  /*
   * Generate FRAME_PLAYED event.
   */

  objPtr->bindFramePtr = framePtr;
  ExecuteCMBinding(objPtr->interp, objPtr->bindToken,
		   "<FRAME_PLAYED>");

  YUV_Play_FreeFrame(framePtr);

  return TCL_OK;
}


/************************************************************
 *
 * YUV_Play_FreeFrame
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static void
YUV_Play_FreeFrame(YUVFrame *framePtr)
{
  int i;

  /*
   * Call device frame freeer.
   */
  if (framePtr->objPtr->devicePtr->freeFrameProc != NULL) {
      framePtr->objPtr->devicePtr->freeFrameProc(framePtr);
  }

  /*
   * Detach from buffer.
   */

  for (i=0; i<framePtr->sb->count; i++) {
      DetachBuffer(framePtr->objPtr->interp, framePtr->sb->buffers[i],
		   "cmYUVPlay:YUV_Play_FreeFrame");
  }

  FreeScatterBuffer(framePtr->sb);
  framePtr->sb = NULL;

  if (framePtr->at_token >= 0) {
      framePtr->objPtr->at_tokens[framePtr->at_token] = 0;
  }

  ckfree((char *) framePtr);

  return;
}


/************************************************************
 *
 * YUV_Play_Refresh
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
YUV_Play_Refresh(YUVResource * objPtr)
{
  int res;
  char *errStr;

  /*
   * If not ready, do nothing.
   */

  if (!objPtr->ready) return TCL_OK;

  /*
   * If device has refresh, call it.
   */

  if (objPtr->devicePtr != NULL) {
    if (objPtr->devicePtr->refreshProc != NULL) {
      res = objPtr->devicePtr->refreshProc(objPtr);
      if (res != 0) {
	errStr = "Error refreshing picture.";
	if (objPtr->devicePtr->errorProc != NULL) {
	  errStr = objPtr->devicePtr->errorProc(objPtr, res);
	}
	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
      }
    }
  }

  return TCL_OK;
}


/************************************************************
 *
 * YUVResourceModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
YUVResourceModuleInit(Tcl_Interp * interp)
{
    Tcl_CreateCommand(interp, "yuvPlay", YUVResourceCreateCmd,
			NULL, (void (*)()) NULL);
    Tcl_GlobalEval(interp, "cminfo add media {video}");
    Tcl_GlobalEval(interp, "cminfo add formats video {YUV}");
    Tcl_GlobalEval(interp, "cminfo add objects video YUV YUVPlay");
    Tcl_VarEval(interp, "cminfo add object_info YUVPlay {play",
			" software",
			"}", NULL);
    return TCL_OK;
}

/******************************************************************
 *
 * YUV_Play_DeviceList
 *
 * Parameters:
 *     objPtr - pointer to mpeg play object.
 *
 * Results:
 *     Standard Tcl Result.
 *
 * Actions:
 *     Sets result to list of devices supported. Returns TCL_OK;
 *
 ******************************************************************/
static int
YUV_Play_DeviceList(YUVResource *objPtr)
{
    Tcl_Interp *interp = objPtr->interp;
    Tcl_DString resStr;

    Tcl_DStringInit(&resStr);

    Tcl_DStringAppendElement(&resStr, "software");

    Tcl_SetResult(interp, Tcl_DStringValue(&resStr), TCL_VOLATILE);
    Tcl_DStringFree(&resStr);

    return TCL_OK;
}

static char *
GetLogStart(Tcl_Interp *interp, ClientData clientdata)
{
  YUVResource *objPtr;
  char *resStr;

  objPtr = (YUVResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%f", objPtr->bindFramePtr->logStart);

  return resStr;
}

static char *
GetLogEnd(Tcl_Interp *interp, ClientData clientdata)
{
  YUVResource *objPtr;
  char *resStr;

  objPtr = (YUVResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%f", objPtr->bindFramePtr->logEnd);

  return resStr;
}

static char *
GetCumulLenDecoded(Tcl_Interp *interp, ClientData clientdata)
{
  YUVResource *objPtr;
  char *resStr;

  objPtr = (YUVResource *) clientdata;

  resStr = (char *) ckalloc(50);
  sprintf(resStr, "%d", objPtr->bytes);
  objPtr->bytes = 0;
  return resStr;
}

static char *
GetDataLen(Tcl_Interp *interp, ClientData clientdata)
{
  YUVResource *objPtr;
  char *resStr;

  objPtr = (YUVResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->dataLength);

  return resStr;
}

static char *
GetWidth(Tcl_Interp *interp, ClientData clientdata)
{
  YUVResource *objPtr;
  char *resStr;

  objPtr = (YUVResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.width);

  return resStr;
}

static char *
GetHeight(Tcl_Interp *interp, ClientData clientdata)
{
  YUVResource *objPtr;
  char *resStr;

  objPtr = (YUVResource *) clientdata;

  resStr = (char *) ckalloc(50);

  sprintf(resStr, "%d", objPtr->bindFramePtr->frameHeader.height);

  return resStr;
}



static void
YUVInsertFrame( YUVFrame *newFrame, YUVFrame **queue, double speed )
{
    YUVFrame *prevPtr, *nextPtr, *framePtr;

    if (*queue == NULL) {
	*queue = newFrame;
	newFrame->next = newFrame->prev = NULL;
    } else if (speed == 0.0) {
	/* Remove all old frames */
	framePtr = *queue;
	while (framePtr != NULL) {
	    nextPtr = framePtr->next;
	    YUV_Play_FreeFrame(framePtr);
	    framePtr = nextPtr;
	}
	*queue = newFrame;
	newFrame->next = newFrame->prev = NULL;
    } else {
	framePtr = *queue;
	prevPtr = NULL;
	while (framePtr != NULL) {
	    if (speed < 0.0) {
		if (framePtr->logEnd > newFrame->logEnd) {
		    prevPtr = framePtr;
		    framePtr = framePtr->next;
		} else {
		    break;
		}
	    } else {
		if (framePtr->logEnd < newFrame->logEnd) {
		    prevPtr = framePtr;
		    framePtr = framePtr->next;
		} else {
		    break;
		}
	    }
	}
	
	if (framePtr == NULL) {
	    prevPtr->next = newFrame;
	    newFrame->prev = prevPtr;
	    newFrame->next = NULL;
	} else if (prevPtr == NULL) {
	    newFrame->next = framePtr;
	    framePtr->prev = newFrame;
	    *queue = newFrame;
	    newFrame->prev = NULL;
	} else {
	    prevPtr->next = newFrame;
	    newFrame->prev = prevPtr;
	    framePtr->prev = newFrame;
	    newFrame->next = framePtr;
	}
    }
}

static void
YUVRemoveFrame(YUVFrame *frame, YUVFrame **queue)
{
    if (frame->prev == NULL) {
	*queue = frame->next;
	frame->next = NULL;
	if (*queue != NULL) 
	    (*queue)->prev = NULL;
    } else if (frame->next == NULL) {
	frame->prev->next = NULL;
	frame->prev = NULL;
    } else {
	frame->prev->next = frame->next;
	frame->next->prev = frame->prev;
	frame->next = frame->prev = NULL;
    }

}

static void
YUVPurgeLateFrames(YUVFrame **queue, double speed, double now)
{
    YUVFrame *framePtr, *nextPtr;

    framePtr = *queue;
    nextPtr = NULL;

    while (framePtr != NULL && framePtr->next != NULL) {
	if (speed < 0.0) {
	    if (framePtr->logEnd > now) {
		nextPtr = framePtr->next;
		if (framePtr->prev == NULL) {
		    *queue = nextPtr;
		    if (*queue != NULL) 
			(*queue)->prev = NULL;
		} else if (nextPtr == NULL) {
		    framePtr->prev->next = NULL;
		} else {
		    framePtr->prev->next = nextPtr;
		    nextPtr->prev = framePtr->prev;
		}
		YUV_Play_FreeFrame(framePtr);

	    } else {
		break;
	    }

	/* speed >= 0.0 */
	} else {
	    if (framePtr->logEnd < now) {
		nextPtr = framePtr->next;
		if (framePtr->prev == NULL) {
		    *queue = nextPtr;
		    if (*queue != NULL) 
			(*queue)->prev = NULL;
		} else if (nextPtr == NULL) {
		    framePtr->prev->next = NULL;
		} else {
		    framePtr->prev->next = nextPtr;
		    nextPtr->prev = framePtr->prev;
		}
		YUV_Play_FreeFrame(framePtr);
	    } else {
		break;
	    }
	}
	framePtr = nextPtr;
    }

    *queue = framePtr;

    if (framePtr && framePtr->next != NULL)
	framePtr->prev = NULL;

}

