/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * YUVSw.c
 *
 * Device to work with the CM YUVPlay object to decode
 * a motion YUV stream in software.
 *
 */



#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#ifdef NO_STDLIB_H
#    include "compat/stdlib.h"
#else
#    include <stdlib.h>
#endif
#ifdef NO_STRING_H
#    include "compat/string.h"
#else
#    include <string.h>
#endif
#include <tcl.h>
#include <tk.h>
#include <dp.h>
#include "cmYUVPlay.h"
#include "cmBuffer.h"

#define YUV_SW
#ifdef YUV_SW

#include "cmUtil.h"
#include "cmDither.h"
#include "cmBufferlist.h"
#include "cmLTS.h"

#ifdef _WINDOWS
extern HDC tcm_DitherHDC;
#endif

#include <X11/Xlib.h>

#ifdef HAVE_SHARED_MEMORY
#  include <sys/ipc.h>
#  include <sys/shm.h>
#  include <X11/extensions/XShm.h>
#endif

/*
 * There is one of the following structs for every
 * unique software device created
 */
typedef struct YUV_SW_Device {
    /*
     * Stuff needed for display
     */
    Window win;
    Display *dpy;
    GC gc;
    int hasShm;
#ifdef HAVE_SHARED_MEMORY
    XShmSegmentInfo shminfo;
#endif
#ifdef _WINDOWS
    BITMAPINFO *image;
    unsigned char *image_data;
#else
    XImage *image;	/* The current frame */
#endif
    int waiting;	/* Waiting for shared memory completion event */
    int ShmEvent;	/* X Event number for shared memory events */

    DitherFunc ditherFunc;	/* Routine to do dithering */
    DitherRec *ditherRec;	/* Information for ditherFunc */

    Buffer *prevBuffer;		/* Information about previous frame */
    int prevWidth;
    int prevHeight;

} YUV_SW_Device;

/*
 * Prototypes for internal routines
 */
static int ProcessShmEvent (ClientData clientData, XEvent *e);
static int LookForShmEvent (Display *display, XAnyEvent *eventPtr, char *arg);
static void WaitForCompletion (YUV_SW_Device *swDevice);

typedef enum { INIT_DITHER_FAILED = 1, NEW_FRAME_FAILED, BAD_PRECISION,
		   BAD_GEOMETRY, BAD_DECIMATION, SHM_ERROR } YUV_SW_ErrorValue;


/*
 *------------------------------------------------------------
 *
 * NoXShm
 *
 *	Registered temporarily as a Tk error handler so that
 *	if we can't use the XShm extension (for example if the
 *	display is remote) the hasShm flag is unset.
 *
 * Results:
 *	Always returns 0 so the program doesn't exit
 *
 * Side Effects:
 *	Unsets the hasShm flag
 *
 *------------------------------------------------------------
 */
static int
NoXShm(ClientData cd, XErrorEvent *evt)
{
    YUV_SW_Device *swDevice = (YUV_SW_Device *) cd;
    
    swDevice->hasShm = 0;
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * ProcessShmEvent --
 *
 *	Registered as a Tk generic handler to unset the
 *	waiting flag on a shared memory event so that we
 *	don't miss shared memory events
 *
 * Results:
 *	1 if we handle the event, 0 otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static int
ProcessShmEvent (ClientData clientData, XEvent *e)
{
    YUVResource *objPtr;
    YUV_SW_Device *swDevice;
    XAnyEvent *shmEventPtr = (XAnyEvent *) e;

    objPtr = (YUVResource *) clientData;
    swDevice = (YUV_SW_Device *) objPtr->devicePtr->deviceData;
    
    if (swDevice->waiting == 0) {
	return 0;
    } else if (objPtr->xid == 0) {
	swDevice->waiting = 0;
	return 0;
    } else if ((swDevice->waiting)
	       && (shmEventPtr->type == swDevice->ShmEvent)
	       && (shmEventPtr->window = objPtr->xid)) {
	swDevice->waiting = 0;
	return 1;
    }
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * LookForShmEvent --
 *
 *	Helper function for XIfEvent call  that waits for the
 *	shared memory operation to complete.
 *
 * Results:
 *	True (1) if event is a ShmEvent for the right window, 
 *	False (0) otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static int
LookForShmEvent (Display *display, XAnyEvent *eventPtr, char *arg)
{
    YUV_SW_Device *swDevice;

    swDevice = (YUV_SW_Device *) arg;

    return ((eventPtr->type == swDevice->ShmEvent) && 
	    (eventPtr->window == swDevice->win));
}


/*
 *--------------------------------------------------------------
 *
 * WaitForCompletion --
 *
 *	If we've issued a shared memory put image call, block
 *	until X server sends us a message that it's complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Will block until appropriate shared memory event is
 *	returned from X server.
 *
 *--------------------------------------------------------------
 */
static void
WaitForCompletion (YUV_SW_Device *swDevice)
{
#ifdef HAVE_SHARED_MEMORY
    XEvent shmEvent;

    if (!swDevice->win) {
	swDevice->waiting = 0;
	return;
    }
    if (swDevice->waiting) {
	XIfEvent(swDevice->dpy, &shmEvent, LookForShmEvent, (char *) swDevice);
	swDevice->waiting = 0;
    }
#else
    swDevice->waiting = 0;
#endif
}


/*
 *--------------------------------------------------------------
 *
 * YUV_SW_Init --
 *
 *	Initialize the device.
 *
 * Results:
 *	0 for successful initilization, non-zero error code otherwise
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

#if 0
int ioError( Display *dpy, XErrorEvent *myerr) 
{
    char err[256];
    XGetErrorText(dpy, myerr->error_code, err, 256);
}
#endif

int
YUV_SW_Init(YUVResource *objPtr)
{
    YUV_SW_Device *swDevice;
#ifdef HAVE_SHARED_MEMORY
    XShmSegmentInfo si;
#endif
    Tk_ErrorHandler handler;
    int result = 0;
    DitherParam ditherInfo;

    swDevice = (YUV_SW_Device *) objPtr->devicePtr->deviceData;

    swDevice->win = objPtr->xid;
    swDevice->dpy = Tcm_Display();
#if 0
    XSync(swDevice->dpy, False);
    XSynchronize(swDevice->dpy, True);
    XSetIOErrorHandler( ioError );
    XSetErrorHandler( ioError );
#endif
#ifndef _WINDOWS
    swDevice->gc = XCreateGC(swDevice->dpy, objPtr->xid, 0, 0);
    XSetForeground(swDevice->dpy, swDevice->gc, 0);
    XSynchronize(swDevice->dpy,1);
#endif 

    swDevice->image = NULL;
    swDevice->waiting = 0;
/* #ifdef HAVE_SHARED_MEMORY */
#if 0
    if (cmXShmQueryExtension(Tcm_Display())) {
	swDevice->hasShm = 1;
	swDevice->ShmEvent = XShmGetEventBase(swDevice->dpy)+ShmCompletion;
	Tk_CreateGenericHandler(ProcessShmEvent, (ClientData) objPtr);
    } else {
	swDevice->hasShm = 0;
    }
#else
    swDevice->hasShm = 0;
#endif
    swDevice->prevBuffer = NULL;

    /*
     * Initialize the dithering module
     */
    ditherInfo.dType = CHOOSE_BEST;
    ditherInfo.minColors = 2;
    ditherInfo.maxColors = 256;
    ditherInfo.pFormat = YCRCB422;

    swDevice->ditherRec = InitDither(ditherInfo, swDevice->dpy,
				     swDevice->win,
				     &(swDevice->ditherFunc));
    if (swDevice->ditherRec == NULL) {
	result = INIT_DITHER_FAILED;
    }

    return result;
}


/*
 *--------------------------------------------------------------
 *
 * YUV_SW_Reinit --
 *
 *	Reinitialize the software device.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
int
YUV_SW_Reinit(YUVResource *objPtr)
{
    YUV_SW_Device *swDevice;

    swDevice = (YUV_SW_Device *) objPtr->devicePtr->deviceData;

    if (objPtr->xid != swDevice->win) {
	swDevice->win = objPtr->xid;

	WaitForCompletion(swDevice);
	
#ifndef _WINDOWS
	if (swDevice->gc)  {
	    XFreeGC(swDevice->dpy, swDevice->gc);
	}
	swDevice->gc = XCreateGC(swDevice->dpy, objPtr->xid, 0, 0);
	XSetForeground(swDevice->dpy, swDevice->gc, 0);
#endif
	
	swDevice->waiting = 0;    
    }
    return 0;
}



/*
 *--------------------------------------------------------------
 *
 * YUV_SW_Dither --
 *
 *	Dither a frame
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 *--------------------------------------------------------------
 */
int
YUV_SW_Dither (YUVFrame *framePtr)
{
    YUVResource *objPtr;
    YUV_SW_Device *swDevice;
    int imWidth, imHeight;
    int winWidth, winHeight;
    int xpos, ypos;
    int depth, finalsize;
    Display *dpy;
#ifdef _WINDOWS
    BITMAPINFO *image;
#else
    XImage *image = NULL;	/* Quiet gcc */
#endif
    char *dithered;
    Buffer *tempBuffer;

    objPtr = framePtr->objPtr;
    swDevice = (YUV_SW_Device *) framePtr->objPtr->devicePtr->deviceData;

    /* Create buffer */

    imWidth = framePtr->frameHeader.width;
    imHeight = framePtr->frameHeader.height;
    finalsize = imWidth * imHeight;

    if (swDevice->hasShm) {
	tempBuffer = MakeSharedBuffer(objPtr->interp, finalsize,
				      "yuvSW:YUV_SW_Dither");
	if (tempBuffer == NULL) {
	    tempBuffer = MakeBuffer(objPtr->interp, finalsize,
				    "yuvSW:YUV_SW_Dither");
	}
	dithered = tempBuffer->data;
    } else {
	tempBuffer = MakeBuffer(objPtr->interp, finalsize,
				"yuvSW:YUV_SW_Dither");
	dithered = tempBuffer->data;
    }

    /* Dither Frame */

    {
	int ysize, uvsize;
	char *y, *u, *v;
	double start, end;

	ysize = imHeight * imWidth;
	uvsize = ysize / 2;
	
	y = framePtr->data;
	u = (char *)(y + ysize);
	v = (char *)(u + uvsize);

	start = Tcm_ReadSysClock();

	swDevice->ditherFunc(swDevice->ditherRec,
			     0, 0, 
			     imWidth, imHeight,
			     y, u, v,
			     dithered,
			     imWidth, imWidth/2, imWidth/2, imWidth,
			     imWidth, imHeight);

	end = Tcm_ReadSysClock();

	if (objPtr->ditherTime != 0.0) {
	    objPtr->ditherTime = ((0.80 * objPtr->ditherTime) + 
				    (0.20 * (end - start)));
	} else {
	    objPtr->ditherTime = end - start;
	}

	if (swDevice->prevBuffer != NULL) {
	    DetachBuffer(framePtr->objPtr->interp, swDevice->prevBuffer,
			 "yuvSW:YUV_SW_Dither");
	}

	swDevice->prevBuffer = tempBuffer;
	AttachBuffer(objPtr->interp, swDevice->prevBuffer,
		     "yuvSW:YUV_SW_Dither");
    }

    /* Store dithered data in frame struct */
    (Buffer *)framePtr->deviceData = tempBuffer;
    
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * YUV_SW_Display --
 *
 *	Toss a decoded frame up on the screen.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Could block if previous shared memory put image hasn't returned.
 *
 *--------------------------------------------------------------
 */
int
YUV_SW_Display (YUVFrame *framePtr)
{
    YUVResource *objPtr;
    YUV_SW_Device *swDevice;
    int imWidth, imHeight;
    int winWidth, winHeight;
    int xpos, ypos;
    int depth, finalsize;
    Display *dpy;
#ifdef _WINDOWS
    BITMAPINFO *image;
#else
    XImage *image = NULL;	/* Quiet gcc */
#endif
    char *dithered;
    Buffer *tempBuffer;

    objPtr = framePtr->objPtr;
    swDevice = (YUV_SW_Device *) framePtr->objPtr->devicePtr->deviceData;

    /* Get image data from frame */
    tempBuffer = (Buffer *)framePtr->deviceData;
    dithered = tempBuffer->data;

    /* Display Image */

    dpy = swDevice->dpy;
    depth = swDevice->ditherRec->depth;
    imWidth = framePtr->frameHeader.width;
    imHeight = framePtr->frameHeader.height;
    winWidth = objPtr->width;
    winHeight = objPtr->height;

    xpos = objPtr->xpos;
    ypos = objPtr->ypos;

    if (xpos == -1) {
	xpos = (winWidth-imWidth)/2;
    }
    if (ypos == -1) {
	ypos = (winHeight-imHeight)/2;
    }

    WaitForCompletion(swDevice);

    /*
     * Create (or recreate) the XImage structure if necessary
     */
    if ((swDevice->prevWidth != imWidth) || (swDevice->prevHeight != imHeight)
	|| (swDevice->image == NULL)) {
	
	if (swDevice->image != NULL) {
#ifndef _WINDOWS
	    swDevice->image->data = NULL;
	    XDestroyImage(swDevice->image);
#else
	    free((char *) swDevice->image);
	    swDevice->image_data = NULL;
#endif
	}
	
	if (swDevice->hasShm) {
#ifdef HAVE_SHARED_MEMORY
	    image = XShmCreateImage(dpy, swDevice->ditherRec->visual,
				    depth,
				    ZPixmap, NULL, &(swDevice->shminfo),
				    imWidth, imHeight);
#endif
	} else {
#ifndef _WINDOWS
	    image = XCreateImage(dpy,swDevice->ditherRec->visual,
				 depth, ZPixmap, 0,
				 dithered, imWidth, imHeight, 32, 0);
#else
	    int i;
	    short *pPalIndex;
	    int allocSize;

	    if (dithered == NULL) {
		dithered = malloc(imWidth * imHeight * depth);
	    }
	    allocSize = sizeof(BITMAPINFO);
	    if (depth == 8) {
	        allocSize += 256 * sizeof(short);
	    }
	    image = (BITMAPINFO *) malloc(allocSize);

	    image->bmiHeader.biSize = (LONG)sizeof(BITMAPINFOHEADER);
	    image->bmiHeader.biWidth = imWidth;
	    image->bmiHeader.biHeight = imHeight;
	    image->bmiHeader.biPlanes = 1;
	    image->bmiHeader.biBitCount = depth;
	    image->bmiHeader.biCompression = 0L;
	    image->bmiHeader.biSizeImage = 0L;
	    image->bmiHeader.biXPelsPerMeter = 0L;
	    image->bmiHeader.biYPelsPerMeter = 0L;
	    image->bmiHeader.biClrUsed = 0L;
	    image->bmiHeader.biClrImportant = 0L;
	    if (depth == 8) {
		pPalIndex=(short *) image->bmiColors;
		for (i = 0; i < 256; i++) {
		    pPalIndex[i] = i;
		}
	    }
#endif
	}
	
	swDevice->prevWidth = imWidth;
	swDevice->prevHeight = imHeight;
    } else {
	image = swDevice->image;
    }

#ifndef _WINDOWS
    image->data = (signed char *)dithered;
#else
    swDevice->image_data = dithered;
#endif

    /*
     * Actually display the image
     */
    if ((swDevice->hasShm) && (swDevice->prevBuffer->shared)
	&& (swDevice->prevBuffer->xAttach)) {
#ifdef HAVE_SHARED_MEMORY
	memcpy( &(swDevice->shminfo), &(swDevice->prevBuffer->shminfo),
	       sizeof(XShmSegmentInfo) );

	fprintf(stderr, "XShmPutImage: dpy:%d win:%d gc:%d image:%d sx:%d sy:%d dx:%d dy:%d w:%d h:%d se:%d",
		dpy, swDevice->win, swDevice->gc,
		image, 0, 0, xpos, ypos, imWidth, imHeight, 1);
		
	XShmPutImage(dpy, swDevice->win, swDevice->gc,
		     image, 0, 0, xpos, ypos, imWidth, imHeight, 1);
	swDevice->waiting = 1;
#endif
    } else {
#ifndef _WINDOWS
	XPutImage(dpy, swDevice->win, swDevice->gc, image, 0, 0, xpos, ypos,
		  imWidth, imHeight);
#else
	{
	    HPALETTE pal, palPrev;
	    
	    pal = XGetWindowsPalette(dpy, 
				     XDefaultColormap(dpy, 
						      DefaultScreen(dpy)));
	    palPrev = SelectPalette(tcm_DitherHDC, pal, FALSE);
	    
	    SetDIBitsToDevice(tcm_DitherHDC, xpos, ypos,
			      imWidth, imHeight, 
			      0, 0, 0, imHeight,
			      dithered,
			      image,
			      ((depth == 8) ?  DIB_PAL_COLORS : DIB_RGB_COLORS));
	    SelectPalette(tcm_DitherHDC, palPrev, FALSE);
	}

#endif
	swDevice->waiting = 0;
    }
    
    DetachBuffer(objPtr->interp, tempBuffer,
		     "yuvSW:YUV_SW_Display");

    swDevice->image = image;
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * YUV_SW_NextFrame --
 *
 *	Chooses the next frame to decode
 *
 * Results:
 *	0 if successfull, non-zero error code otherwise
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	YUVResource *objPtr	The device
 *	YUVFrame *frameList	List of frames to choose from
 *	YUVFrame **chosen	Set to point to the chosen frame
 *
 *--------------------------------------------------------------
 */
static int
YUV_SW_NextFrame(YUVResource *objPtr, YUVFrame *frameList,
		  YUVFrame **chosen)
{
    YUVFrame *currentFrame, *longestFrame;
    YUV_SW_Device *swDevice;
    double speed, now, end, decode;
    
    swDevice = (YUV_SW_Device *) objPtr->devicePtr->deviceData;
    
    if (frameList == NULL) {
	*chosen = NULL;
	return 1;
    }
    
    speed = LTS_Speed((ClientData) objPtr->lts);
    now = LTS_Value((ClientData) objPtr->lts);

    /*
     * Go through the available frames, looking for one
     * we have time to dither
     */
    longestFrame = frameList;
    for (currentFrame = frameList; currentFrame != NULL;
	 currentFrame = currentFrame->next) {

	end = (speed < 0.0) ? currentFrame->logStart : currentFrame->logEnd;
	decode = objPtr->ditherTime;
	
	if ((now + decode*speed) <= end) {
	    *chosen = currentFrame;
	    return 0;
	}
	
	if (speed < 0.0) {
	    if (currentFrame->logEnd < longestFrame->logEnd) {
		longestFrame = currentFrame;
	    }
	} else {
	    if (currentFrame->logEnd > longestFrame->logEnd) {
		longestFrame = currentFrame;
	    }
	}
    }
    
    *chosen = longestFrame;
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * YUV_SW_Refresh --
 *
 *	Called to refresh the display if part of it is exposed.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	Display is repainted.  Could block if previous shared
 *	memory put image hasn't returned
 *
 *--------------------------------------------------------------
 */
int
YUV_SW_Refresh(YUVResource *objPtr)
{
    YUV_SW_Device *swDevice;
    double speed;
    int imWidth, imHeight;
    int winWidth, winHeight;
    int xpos, ypos;

    swDevice = (YUV_SW_Device *) objPtr->devicePtr->deviceData;

    if (swDevice->prevBuffer == NULL) {
	return 0;
    }

    speed = LTS_Speed(objPtr->lts);
    if (speed != 0.0) {
	return 0;
    }

    imWidth = swDevice->prevWidth;
    imHeight = swDevice->prevHeight;
    winWidth = objPtr->width;
    winHeight = objPtr->height;

    xpos = objPtr->xpos;
    ypos = objPtr->ypos;

    if (xpos == -1) {
	xpos = (winWidth-imWidth)/2;
    }
    if (ypos == -1) {
	ypos = (winHeight-imHeight)/2;
    }

    if ((swDevice->hasShm) && (swDevice->prevBuffer->shared)
	&& (swDevice->prevBuffer->xAttach)) {
#ifdef HAVE_SHARED_MEMORY

	fprintf(stderr, "XShmPutImage: dpy:%d win:%d gc:%d image:%d sx:%d sy:%d dx:%d dy:%d w:%d h:%d se:%d",
		swDevice->dpy, swDevice->win, swDevice->gc,
		swDevice->image, 0, 0, xpos, ypos,
		swDevice->prevWidth, swDevice->prevHeight, 1);

	XShmPutImage(swDevice->dpy, swDevice->win, swDevice->gc,
		     swDevice->image, 0, 0, xpos, ypos,
		     swDevice->prevWidth, swDevice->prevHeight, 1);
	swDevice->waiting = 1;
#endif
    } else {
#ifndef _WINDOWS
	XPutImage(swDevice->dpy, swDevice->win, swDevice->gc,
		  swDevice->image, 0, 0, xpos, ypos,
		  swDevice->prevWidth, swDevice->prevHeight);
#else
	{
	    HPALETTE pal, palPrev;
	    
	    pal = XGetWindowsPalette(swDevice->dpy, 
				     XDefaultColormap(swDevice->dpy, 
						      DefaultScreen(swDevice->dpy)));
	    palPrev = SelectPalette(tcm_DitherHDC, pal, FALSE);
	    
	    SetDIBitsToDevice(tcm_DitherHDC, xpos, ypos,
			      swDevice->prevWidth, swDevice->prevHeight,
			      0, 0, 0, swDevice->prevHeight,
			      swDevice->image_data,
			      swDevice->image,
			      ((swDevice->ditherRec->depth == 8) ?
			       DIB_PAL_COLORS : DIB_RGB_COLORS));
	    SelectPalette(tcm_DitherHDC, palPrev, FALSE);
	}
	
#endif
	swDevice->waiting = 0;
    }
    
    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * YUV_SW_Shutdown --
 *
 *	Called to destroy device.
 *
 * Results:
 *	0 if successful, non-zero error code otherwise
 *
 * Side effects:
 *	The device must never be referenced again.
 *
 *--------------------------------------------------------------
 */
int
YUV_SW_Shutdown(YUVResource *objPtr) 
{
    YUV_SW_Device *swDevice = (YUV_SW_Device *) objPtr->devicePtr->deviceData;
    
    if (!objPtr->inited) {
	ckfree((char *) swDevice);
	return 0;
    }

    WaitForCompletion(swDevice);

    if (swDevice->prevBuffer) {
	DetachBuffer(objPtr->interp, swDevice->prevBuffer,
		     "yuvSW:YUV_SW_Shutdown");
#ifndef _WINDOWS
	swDevice->image->data = NULL;	/* Prevent XDestroyImage
						 * from trying to free it */
	XDestroyImage(swDevice->image);
#else
	free((char *) swDevice->image);
	swDevice->image_data = NULL;
#endif
    }
    
#ifdef HAVE_SHARED_MEMORY
    Tk_DeleteGenericHandler(ProcessShmEvent, (ClientData) objPtr);
#endif
    if (swDevice->ditherRec != NULL ) {
	FreeDither(swDevice->ditherRec);
	swDevice->ditherRec = NULL;
    }
#ifndef _WINDOWS
    if (swDevice->gc) {
	XFreeGC(swDevice->dpy, swDevice->gc);
    }	
#endif

    ckfree((char *) swDevice);
    objPtr->devicePtr->deviceData = NULL;

    return 0;
}


/*
 *--------------------------------------------------------------
 *
 * YUV_SW_Error --
 *
 *	Translates an error number (passed in) into a human readable
 *	string (returned).
 *
 * Results:
 *	Human readable string that indicates the error.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *	YUVResource *objPtr	The device
 *	int errNum		The error code
 *
 *--------------------------------------------------------------
 */
char *
YUV_SW_Error(YUVResource *objPtr, int errNum)
{
    switch (errNum) {
      case INIT_DITHER_FAILED:
	return "Dither initialization failed.\n";
	
      case NEW_FRAME_FAILED:
	return "Initialization for new frame failed.\n";
	
      case BAD_PRECISION:
	return "YUV Stream is not at 8 bit precision.\n";
	
      case BAD_GEOMETRY:
	return "Image size in YUV header does not match CMT header.\n";
	
      case BAD_DECIMATION:
	return "Image is not 422.\n";
	
      case SHM_ERROR:
	return "YUV Shared Memory Error.\n";
	
      default:
	return "Unknown Software YUV Device Error.\n";
    }
}


/*
 *--------------------------------------------------------------
 *
 * GetSoftwareYUVDevice --
 *
 *	Called by resource to determine default device.  We should
 *	only be called if there's no YUV hardware available.
 *
 * Results:
 *	Pointer to new YUVDevice
 *
 * Side effects:
 *	A couple of structures are allocated.
 *
 *--------------------------------------------------------------
 */
YUVDevice *
GetSoftwareYUVDevice (YUVResource *objPtr)
{
    YUVDevice *new_dev;
    YUV_SW_Device *swDevice;

    new_dev = (YUVDevice *) ckalloc(sizeof(YUVDevice));

    new_dev->refreshProc = YUV_SW_Refresh;
    new_dev->ditherProc = YUV_SW_Dither;
    new_dev->displayProc = YUV_SW_Display;
    new_dev->initProc = YUV_SW_Init;
    new_dev->reinitProc = YUV_SW_Reinit;
    new_dev->shutdownProc = YUV_SW_Shutdown;
    new_dev->freeFrameProc = NULL;
    new_dev->configProc = NULL;
    new_dev->errorProc = YUV_SW_Error;
    new_dev->nextFrameProc = YUV_SW_NextFrame;
    new_dev->deviceData = NULL;
    new_dev->deviceData = (ClientData) ckalloc(sizeof(YUV_SW_Device));
    memset(new_dev->deviceData, 0, sizeof(YUV_SW_Device));

    /* Initialize any device specific configuration options */

    swDevice = (YUV_SW_Device *)new_dev->deviceData;
    return new_dev;
}


#endif	/* YUV_SW */

