# This file implements the cm object interface.
# A new object type "cmt" is defined which allows the user to connect
# to cm processes and services and create and control objects within
# the remote process.

#---------------------------------
#
# This pair of procedures is run on the server
# side by a client.  They're put into separate
# functions for security purposes, since we 
# don't want to allow the client to execute arbitrary
# dp_atclose commands (which can contain arbitrary
# scripts).
#
#---------------------------------

proc CMAppSetCleanUp {obj_name} {
    global dp_rpcFile
    dp_atclose $dp_rpcFile append "$obj_name destroy"
}

#---------------------------------
#
# cmt_CleanUp
#
# Parameter: Name of cmt connection
#
# Action:
#   This procedure is called when a connection
#   is closed.  It destroys any objects that
#   were created via the connection.
#
#----------------------------------

proc cmt_CleanUp {obj_name} {
    foreach i [info commands $obj_name*] {
	catch {$i destroy}
    }
}

#---------------------------------
#
# cmt_GetMostUniqueName
#
# Parameter: None
#
# Action:
#   This procedure returns a string identifier
#   that encapsulates the pid information
#   of the interpreter process and some random
#   information. This command can
#   be used to create a very unique parent for top
#   level cmt objects such that all objects created
#   remotely through them will be VERY likely to be
#   unique among other processes.
#
#----------------------------------

proc cmt_GetMostUniqueName {} {
    set my_pid [pid];

    set id "";

    while {$my_pid > 0} {
	set nchar [expr "$my_pid%52"];
	set my_pid [expr "$my_pid/52"];

	if {$nchar < 26} {
	    set nchar [expr "$nchar+65"];
	} else {
	    set nchar [expr "($nchar-26)+97"];
	}

	set id [format "%c%s" $nchar $id];
    }

    return $id;
}
    
    

#---------------------------------
#
# cmt_GetNextVarName
#
# Parameters: None
#
# Action:
#   This procedure simply returns the next unique variable name
#   in the form of cmtVar? where ? is replaced by a number.
#   The next number to use is stored in the global register
#   cmt_NextVarName which is initialized by cmtAPIInit.
#   This function is necessary for the creation of "semaphores"
#   for dp_RDO calls that do not want to block the process but
#   want to wait for the results before continuing.
#
#---------------------------------

proc cmt_GetNextVarName {} {
    global cmt_NextVarName;

    incr cmt_NextVarName;

    return "cmtVar$cmt_NextVarName";
}

#----------------------------------
#
# cmt_RDORetValSet
#
# Paramters:
#   varname : variable name to set.
#   varflag : variable flag to set.
#   res : result to set varname to.
#
# Action:
#   This procedure is used as a callback to dp_RDO in order
#   to place the result of the dp_RDO into the global variable
#   $varname. The global variable $varflag is set to 1 to signal
#   to any waiting dp_update loops that the dp_RDO has been 
#   completed.
#
#-----------------------------------------

proc cmt_RDORetValSet {varname varflag res} {
    
    global $varname $varflag;

    set $varflag 1;
    set $varname "$res";
}

#-----------------------------------------
#
# cmtAPIInit
#
# Parameters: None.
#
# Action:
#   Initializes the cmt object module.
#
#-----------------------------------------

proc cmtAPIInit {} {
    global cmt_ObjectList cmt_NextObjectName cmt_NextVarName

    set cmt_ObjectList "";
    set cmt_NextObjectName 0;
    set cmt_NextVarName 0;
}

#----------------------------------------
#
# cmt
#
# Parameters: 
#   parent : parent name of new cmt object
#   args : remaining arguments assumed to be configuration options.
#
# Actions:
#   This procedure is the cmt object creation procedure.
#   It creates a new cmt object with the name $parent{uniqueName}
#   where {uniqueName} is generated from some initialized counter.
#   All data structures for the object are set up, args are passed
#   along to the configuration method of the object, the object name
#   is mapped to a procedure to handle the object's methods.
#   The name of the new cmt object is returned.
#
#-----------------------------------------

proc cmt {parent args} {
    global cmt_ObjectList cmt_NextObjectName;
    global cmt_ObjectData env;

    # Check to see if API module already initialized. If not, 
    # initialize now.

    if {[catch {set cmt_ObjectList}]} {
	cmtAPIInit;
    }

    # Create object name.

    incr cmt_NextObjectName;

    set obj_name "${parent}cmt$cmt_NextObjectName";
    
    # Check to see if obj_name already exists.

    if {[lsearch -exact $cmt_ObjectList $obj_name] != -1} {
	error "$obj_name already exists";
    }

    # Add object name to cmt_ObjectList;

    lappend cmt_ObjectList $obj_name;

    # Create cmt object data structures.

    set cmt_ObjectData($obj_name,sub_object_list) "";
    set cmt_ObjectData($obj_name,CID) -1;
    set cmt_ObjectData($obj_name,rpc_file) "";

    # Check for environment variables CMT_NAMESERVER and CMT_NS_PORT
    # for initial values of these attributes.

    if {![catch {set env(CMT_NAMESERVER)}]} {
	set cmt_ObjectData($obj_name,nameserver) $env(CMT_NAMESERVER);
	set cmt_ObjectData($obj_name,nameserver_default) $env(CMT_NAMESERVER);
    } else {
	set cmt_ObjectData($obj_name,nameserver) "";
	set cmt_ObjectData($obj_name,nameserver_default) "";
    }

    if {![catch {set env(CMT_NS_PORT)}]} {
	set cmt_ObjectData($obj_name,ns_port) $env(CMT_NS_PORT);
	set cmt_ObjectData($obj_name,ns_port_default) $env(CMT_NS_PORT);
    } else {
	set cmt_ObjectData($obj_name,ns_port) "";
	set cmt_ObjectData($obj_name,ns_port_default) "";
    }

    # Set up callbacks.

    set cmt_ObjectData($obj_name,local_close_cmd) "";
    set cmt_ObjectData($obj_name,local_close_cmd_default) "";

    set cmt_ObjectData($obj_name,local_lost_conn_cmd) "";
    set cmt_ObjectData($obj_name,local_lost_conn_cmd_default) "";

    set cmt_ObjectData($obj_name,check_cmd) "cmapp_CheckCmd";

    set cmt_ObjectData($obj_name,remote_close_cmd) ""
    set cmt_ObjectData($obj_name,remote_close_cmd_default) ""

    set cmt_ObjectData($obj_name,remote_lost_conn_cmd) "";
    set cmt_ObjectData($obj_name,remote_lost_conn_cmd_default) "";

    # Create object procedure and map object name to it.

    proc $obj_name {args} "                              \
	return \[eval cmt_object_cmd $obj_name \$args\]; \
    "

    # If additional arguments, send to objects configuration command.

    if {[llength $args] != 0} {
	set firstarg [lindex $args 0];
	if {[string first "-" $firstarg] != 0} {
	    eval $obj_name $args;
	} else {
	    eval $obj_name configure $args;
	}
    }

    # Return object name.

    return $obj_name;
}    

#--------------------------------------------------------
#
# cmt_switch_expand
#
# Parameters:
#   str : string to expand.
#   str_list : list of possible expansions.
#
# Actions:
#   This procedure expands $str to one of the strings in
#   $str_list if $str is a proper prefix for any of the
#   strings and returns the expanded string. 
#   The string is expanded to the first string 
#   that matches. If none match, $str is returned.
#---------------------------------------------------------

proc cmt_switch_expand {str str_list} {
    foreach i $str_list {
	if {[string first $str $i] == 0} {
	    return $i
	}
    }
    return $str;
}

#---------------------------------------------------------
#
# cmt_object_configure
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the configure command for a cmt object
#
#-------------------------------------------------------------

proc cmt_object_configure {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];

    # If no arguments, return all current values.

    if {$argc == 0} {
	return  [list [list  "-nameserver" \
			     $cmt_ObjectData($obj_name,nameserver_default) \
			     $cmt_ObjectData($obj_name,nameserver) ]       \
		      [list "-nameserver_port" \
			    $cmt_ObjectData($obj_name,ns_port_default) \
			    $cmt_ObjectData($obj_name,ns_port) ] \
		      [list "-local_close_command" \
			    $cmt_ObjectData($obj_name,local_close_cmd_default) \
			    $cmt_ObjectData($obj_name,local_close_cmd) ] \
		      [list "-local_lost_connection_command" \
			    $cmt_ObjectData($obj_name,local_lost_conn_cmd_default) \
			    $cmt_ObjectData($obj_name,local_lost_conn_cmd) ] \
		      [list "-check_cmd" "cmapp_CheckCmd" \
			    $cmt_ObjectData($obj_name,check_cmd) ] \
		      [list "-remote_close_command" \
			    $cmt_ObjectData($obj_name,remote_close_cmd_default) \
			    $cmt_ObjectData($obj_name,remote_close_cmd) ] \
		      [list "-remote_lost_connection_command" \
			    $cmt_ObjectData($obj_name,remote_lost_conn_cmd_default) \
			    $cmt_ObjectData($obj_name,remote_lost_conn_cmd) ] \
		      [list "-CID" "" $cmt_ObjectData($obj_name,CID)] \
		      [list "-subobjects" "" $cmt_ObjectData($obj_name,sub_object_list)] \
		      [list "-rpc_file" "" $cmt_ObjectData($obj_name,rpc_file)]];
    }
    
    # If one argument, return value of requested attribute

    if {$argc == 1} {

	# Expand requested attribute name

	set switch_str [cmt_switch_expand [lindex $args 0] \
		{-nameserver -nameserver_port \
		-local_lost_connection_command -local_close_command \
		-remote_lost_connection_command -remote_close_command \
		-CID -rpc_file -check_cmd -subobjects}];

	switch [string range $switch_str 1 end] {
	    CID {
		return [list "-CID" "" $cmt_ObjectData($obj_name,CID)];
	    }
	    check_cmd {
		return [list "-check_cmd" "cmapp_CheckCmd" $cmt_ObjectData($obj_name,check_cmd)];
	    }
	    rpc_file {
		return [list "-rpc_file" "" $cmt_ObjectData($obj_name,rpc_file)];
	    }
	    subobjects {
		return [list "-subobjects" "" $cmt_ObjectData($obj_name,sub_object_list)];
	    }
	    nameserver {
		return [list  "-nameserver" \
			$cmt_ObjectData($obj_name,nameserver_default) \
			$cmt_ObjectData($obj_name,nameserver) ];
	    }
	    nameserver_port {
		return [list "-nameserver_port" \
			$cmt_ObjectData($obj_name,ns_port_default) \
			$cmt_ObjectData($obj_name,ns_port) ];
	    }
	    local_close_command {
		return [list "-local_close_command" \
			$cmt_ObjectData($obj_name,local_close_cmd_default) \
			$cmt_ObjectData($obj_name,local_close_cmd) ];
	    }
	    local_lost_connection_command {
		return [list "-local_lost_connection_command" \
			$cmt_ObjectData($obj_name,local_lost_conn_cmd_default) \
			$cmt_ObjectData($obj_name,local_lost_conn_cmd) ];
	    }
	    remote_close_command {
		return [list "-remote_close_command" \
			$cmt_ObjectData($obj_name,remote_close_cmd_default) \
			$cmt_ObjectData($obj_name,remote_close_cmd) ];
	    }
	    remote_lost_connection_command {
		return [list "-remote_lost_connection_command" \
			$cmt_ObjectData($obj_name,remote_lost_conn_cmd_default) \
			$cmt_ObjectData($obj_name,remote_lost_conn_cmd) ];
	    }
	    default {
		error "Bad configure option \"[lindex $args 0]\"";
	    }
	}
    }

    # If more than one argument, set each specified attribute to new value.

    # Set argument counter;

    set cntr 0;

    # Process while counter is less than number of arguments.

    while {$cntr < $argc} {

	# Expand attribute name

	set switch_str [cmt_switch_expand [lindex $args $cntr] \
		{-nameserver -nameserver_port  \
		-local_lost_connection_command -local_close_command \
		-remote_lost_connection_command -remote_close_command \
		-CID -subobjects -check_cmd -rpc_file}];
	

	switch [string range $switch_str 1 end] {
	    nameserver {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify configuration value after \"-nameserver\"";
		}
		set cmt_ObjectData($obj_name,nameserver) [lindex $args $cntr];
	    }
	    nameserver_port {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify configuration value after \"-nameserver_port\"";
		}
		set cmt_ObjectData($obj_name,ns_port) [lindex $args $cntr];
	    }
	    local_close_command {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify configuration value after \"-local_close_command\"";
		}
		set cmt_ObjectData($obj_name,local_close_cmd) [lindex $args $cntr];
	    }
	    local_lost_connection_command {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify configuration value after \"-local_lost_connection_command\"";
		}

		# If current command is set and CID != -1 use dp_atclose delete 
		# and append.

		set new_local_conn [lindex $args $cntr];

		if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
		    if {$cmt_ObjectData($obj_name,local_lost_conn_cmd) != ""} {
			dp_atclose $cmt_ObjectData($obj_name,rpc_file) delete $cmt_ObjectData($obj_name,local_lost_conn_cmd);
		    }
		    if {$new_local_conn != ""} {
			dp_atclose $cmt_ObjectData($obj_name,rpc_file) append $new_local_conn;
		    }
		}

		set cmt_ObjectData($obj_name,local_lost_conn_cmd) $new_local_conn;
		
	    }
	    check_cmd {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify configuration value after \"-check_cmd\"";
		}
		set cmt_ObjectData($obj_name,check_cmd) [lindex $args $cntr];
		if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
		    dp_SetCheckCmd $cmt_ObjectData($obj_name,rpc_file) \
				   $cmt_ObjectData($obj_name,check_cmd)
		}
	    }
	    remote_close_command {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify configuration value after \"-remote_close_command\"";
		}
		set cmt_ObjectData($obj_name,remote_close_cmd) [lindex $args $cntr];
	    }
	    remote_lost_connection_command {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify configuration value after \"-remote_lost_connection_command\"";
		}

		# If current command is set and CID != -1 use dp_atclose delete
		# and append in the remote procedure.

		set new_remote_conn [lindex $args $cntr];

		if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
		    if {$cmt_ObjectData($obj_name,remote_lost_conn_cmd) != ""} {
			dp_RDO $cmt_ObjectData($obj_name,rpc_file) eval dp_atclose \$dp_rpcFile delete [list $cmt_ObjectData($obj_name,remote_lost_conn_cmd)];
		    }
		    if {$new_remote_conn != ""} {
			dp_RDO $cmt_ObjectData($obj_name,rpc_file) eval dp_atclose \$dp_rpcFile append [list $new_remote_conn];
		    }
		}
		set cmt_ObjectData($obj_name,remote_lost_conn_cmd) $new_remote_conn;
	    }
	    default {
		switch [string range $switch_str 1 end] {
		    CID {
			error "-CID is a read only configuration option";
		    }
		    rpc_file {
			error "-rpc_file is read only configuration option";
		    }
		    subobjects {
			error "-subobjects is a read only configuration option";
		    }
		    default {
			error "Unknown configuration option \"[lindex $args $cntr]\"";
		    }
		}
	    }
	}
	
	incr cntr;
    }
}

#--------------------------------------------------------
#
# cmt_object_close
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the close command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_close {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];

    if {$cmt_ObjectData($obj_name,CID) == -1} {
	return;
    }

    # Kill all objects.


    foreach i $cmt_ObjectData($obj_name,sub_object_list) {
	$i destroy;
    }

    # Execute close command locally and remotely.

    # First remotely.

    if {$cmt_ObjectData($obj_name,remote_close_cmd) != ""} {
	if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
	    catch "dp_RDO $cmt_ObjectData($obj_name,rpc_file)  $cmt_ObjectData($obj_name,remote_close_cmd)"
	}
    }

    # Now locally, catching results.

    set close_cmd_ret "";
    set catch_flag 0;

    if {$cmt_ObjectData($obj_name,local_close_cmd) != ""} {
	set catch_flag [catch {eval $cmt_ObjectData($obj_name,local_close_cmd)} close_cmd_ret];
	global errorInfo
	set saveInfo $errorInfo
    }


    if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
	# Unregister the lost connection commands if any. 
	# First locally.

	if {$cmt_ObjectData($obj_name,local_lost_conn_cmd) != "" } {
	    dp_atclose $cmt_ObjectData($obj_name,rpc_file) delete $cmt_ObjectData($obj_name,local_lost_conn_cmd);
	}

	# Now remotely.

	if {$cmt_ObjectData($obj_name,remote_lost_conn_cmd) != "" } {
	    dp_RDO $cmt_ObjectData($obj_name,rpc_file) $cmt_ObjectData($obj_name,remote_lost_conn_cmd);
	}
       
	# Close RPC file.

	## eric14 hacked here - begin
	# catch "close $cmt_ObjectData($obj_name,rpc_file)";
	catch {CMApp_CloseHost $cmt_ObjectData($obj_name,rpc_file)};
	## eric14 hacked here - end

	set cmt_ObjectData($obj_name,rpc_file) "";
    }

    # Reset CID;
    set cmt_ObjectData($obj_name,CID) -1;

    # Return results of close command.

    if {$catch_flag == 0} {
	return "$close_cmd_ret";
    } else {
	error "$close_cmd_ret" $saveInfo
    }
}

#--------------------------------------------------------
#
# cmt_object_open
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the open command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_open {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];

    if {$cmt_ObjectData($obj_name,CID) != -1} {
	error "Already connected to a service. Use close first.";
    }

    if {$argc == 0} {
	error "Bad syntax, should be: \"$obj_name open -service <service_name>\" \nor \"$obj_name open -host <host_name> <port_num>\" \nor \"$obj_name open -local\"";
    }

    if {[string first [lindex $args 0] "-local"] != 0} {

	if {[string first [lindex $args 0] "-service"] == 0} {
	    # Do Connect to Service here.
	    
	    if {$argc != 2} {
		error "Wrong # args, should be: \"$obj_name open -service <service_name>\"";
	    }
	    
	    # Connect to service. Wait for Peter to write higher level API.
	    # Be sure to set rpc_file to correct connection for use later in
	    # the open command.
	    
	    if {[catch {set rpc_file [NS_GetServiceConn [lindex $args 1]]}]} {
		global errorInfo
		error "Couldn't open service [lindex $args 1]" $errorInfo;
	    }
	} else {
	    
	    if {[string first [lindex $args 0] "-host"] == -1} {
		error "Bad syntax, should be: \"$obj_name open -service <service_name>\" or \"$obj_name open -host <host_name> -port <port_num>\"";
	    }
	    
	    if {$argc != 3} {
		error "Wrong # args, should be: \"$obj_name open -host <host_name> <port_num>\"";
	    }

	    
	    set hostname [lindex $args 1];
	    set portnum [lindex $args 2];
	    
	    ## eric14 hacked here - begin
	    #    set rpc_file [dp_MakeRPCClient $hostname $portnum];
	    set rpc_file [CMApp_OpenHost $hostname $portnum \
			      $cmt_ObjectData($obj_name,check_cmd)];
	    ## eric14 hacked here - end
	    
	    if {$rpc_file == ""} {
		error "Unable to connect to $hostname at port $portnum";
	    }
	}
	if {[catch {set CID [dp_RPC $rpc_file CID_Init]}]} {
	    ## eric14 hacked here - begin
	    # dp_CloseRPC $rpc_file;
	    CMApp_CloseHost $rpc_file
	    ## eric14 hacked here - end
	    
	    global errorInfo
	    error "Service at $hostname $portnum not a CM compliant service" $errorInfo;
	}
    } else {
	set CID [CID_Init];
	set rpc_file "";
    }
    
    set cmt_ObjectData($obj_name,CID) $CID;
    set cmt_ObjectData($obj_name,rpc_file) $rpc_file;

    if {$rpc_file != ""} {
	# Register mandatory clean up callback.
	
	dp_atclose $rpc_file append "$obj_name destroy";
	
	# Set up local and remote lost connection callbacks.
	
	if {$cmt_ObjectData($obj_name,local_lost_conn_cmd) != ""} {
	    dp_atclose $rpc_file append $cmt_ObjectData($obj_name,local_lost_conn_cmd);
	}
	
	if {$cmt_ObjectData($obj_name,remote_lost_conn_cmd) != ""} {
	    dp_RDO $rpc_file $cmt_ObjectData($obj_name,remote_lost_conn_cmd);
	}
    }
}

#--------------------------------------------------------
#
# cmt_object_create_object
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the open command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_create_object {obj_name args} {
    global cmt_ObjectData cmt_subObjectData;
    set argc [llength $args];

    # Check to see if connected.

    if {$cmt_ObjectData($obj_name,CID) == -1} {
	error "Must be connected to a process or service to create remote objects";
    }

    if {$argc < 1} {
	error "Wrong # args, should be: \"$obj_name create_object <cm_object_type> ?options?\"";
    }
    
    set nobj_type [lindex $args 0];
    set nobj_args [lrange $args 1 end];

    #
    # Create the object in the remote process checking for errors.
    #
    set nobj_name "";
    set socket $cmt_ObjectData($obj_name,rpc_file)
    set head $obj_name-$cmt_ObjectData($obj_name,CID).
    if {$socket == ""} {
	if {[catch {eval $nobj_type $head $nobj_args} nobj_name]} {
	    global errorInfo
	    error "$nobj_name" $errorInfo;
	}
    } else {
	if {[catch {eval dp_RPC $socket $nobj_type $head $nobj_args} nobj_name]} {
	    global errorInfo
	    error "$nobj_name" $errorInfo;
	}
    }

    #
    # Register mandatory cleanup function.  Have to figure
    # out what the object's called on the other side...
    #

    if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
	dp_RDO $socket CMAppSetCleanUp $nobj_name

	#
	# If no errors, create local procedure wrapper
	#
	proc $nobj_name {args} "\
	    return \[eval cmt_remote_object_cmd $obj_name $nobj_name \$args\];\
	    "
    } else {
	
	#
	# New object is really local. Wrap around cmt_local_object_cmd.
	#

	set robj_name "${nobj_name}_hidden_[cmt_GetNextVarName]";
	rename $nobj_name $robj_name;

	proc $nobj_name {args} "\
            return \[eval cmt_local_object_cmd $obj_name $nobj_name $robj_name \$args\];\
        "
    }

    #
    # Add object name to sub_object list.
    #
    lappend cmt_ObjectData($obj_name,sub_object_list) $nobj_name;

    # Tag object in remote process with local object name and CID number.
    #$nobj_name addtag "CID $cmt_ObjectData($obj_name,CID)";

    #
    # Set up data structure for object.
    #
    set cmt_subObjectData($nobj_name,local_lost_command) "";
    set cmt_subObjectData($nobj_name,remote_lost_command) "";
    set cmt_subObjectData($nobj_name,local_destroy_command) "";
    set cmt_subObjectData($nobj_name,remote_destroy_command) "";
    set cmt_subObjectData($nobj_name,local_lost_command_default) "";
    set cmt_subObjectData($nobj_name,remote_lost_command_default) "";
    set cmt_subObjectData($nobj_name,local_destroy_command_default) "";
    set cmt_subObjectData($nobj_name,remote_destroy_command_default) "";

    #
    # If local and remote lost commands are not "", then register them.
    #
    if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
	if {$cmt_subObjectData($nobj_name,local_lost_command) != ""} {
	    dp_atclose $cmt_ObjectData($obj_name,rpc_file) append $cmt_subObjectData($nobj_name,local_lost_command);
	}

	if {$cmt_subObjectData($nobj_name,remote_lost_command) != ""} {
	    dp_RDO $cmt_ObjectData($obj_name,rpc_file) eval dp_atclose \$dp_rpcFile append [list $cmt_subObjectData($nobj_name,remote_lost_command)];
	}
    }

    #
    # Return local object name.
    #
    return $nobj_name;
}

#--------------------------------------------------------
#
# cmt_object_configure_object
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the open command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_configure_object {obj_name args} {
    global cmt_ObjectData cmt_subObjectData;
    set argc [llength $args];

    # If no connection, don't do it.
    
    if {$cmt_ObjectData($obj_name,CID) == -1} {
	error "No connection to service or process, thus no remote objects.";
    }

    # If no sub object, don't do it.

    set nobj_name [lindex $args 0];
    set config_args [lrange $args 1 end];
    incr argc -1;

    if {$argc == -1} {
	error "Wrong # args, should be: \"$obj_name configure_object <object_name> ?args?\"";
    }

    if {[catch {set cmt_subObjectData($nobj_name,local_destroy_command)}]} {
	global errorInfo
	error "The object \"$nobj_name\" does not exist." $errorInfo;
    }

    # Do it

    if {$argc == 0} {
	return  [list [list  "-local_lost_command" \
		$cmt_subObjectData($nobj_name,local_lost_command_default) \
		$cmt_subObjectData($nobj_name,local_lost_command) ]       \
		[list "-remote_lost_connection_command" \
		$cmt_subObjectData($nobj_name,remote_lost_command_default) \
		$cmt_subObjectData($nobj_name,remote_lost_command) ] \
		[list "-local_destroy_command" \
		$cmt_subObjectData($nobj_name,local_destroy_command_default) \
		$cmt_subObjectData($nobj_name,local_destroy_command) ]\
		[list "-remote_destroy_command" \
		$cmt_subObjectData($nobj_name,remote_destroy_command_default) \
		$cmt_subObjectData($nobj_name,remote_destroy_command) ]];
    }
    if {$argc == 1} {
	set switch_str [cmt_switch_expand [lindex $config_args 0] \
		{-local_lost_connection_command -remote_lost_connection_command \
		-local_destroy_command -remote_destroy_command}];
	
	switch [string range $switch_str 1 end] {
	    local_lost_connection_command {
		return  [list  "-local_lost_connection_command" \
		$cmt_subObjectData($nobj_name,local_lost_command_default) \
		$cmt_subObjectData($nobj_name,local_lost_command) ]
	    }
	    remote_lost_connection_command {
		return [list "-remote_lost_connection_command" \
		$cmt_subObjectData($nobj_name,remote_lost_command_default) \
		$cmt_subObjectData($nobj_name,remote_lost_command) ];
	    }
	    local_destroy_command {
		return [list "-local_destroy_command" \
		$cmt_subObjectData($nobj_name,local_destroy_command_default) \
		$cmt_subObjectData($nobj_name,local_destroy_command) ];
	    }
	    remote_destroy_command {
		return [list "-remote_destroy_command" \
		$cmt_subObjectData($nobj_name,remote_destroy_command_default) \
		$cmt_subObjectData($nobj_name,remote_destroy_command) ];
	    }
	    default {
		error "Bad object configuration option \"$switch_str\"";
	    }
	}
    }
    
    set cntr 0;
    while {$cntr < $argc} {
	set switch_str [cmt_switch_expand [lindex $config_args 0] \
		{-local_lost_connection_command -remote_lost_connection_command \
		-local_destroy_command -remote_destroy_command}];
    
	switch [string range $switch_str 1 end] {
	    local_lost_connection_command {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify object configuration value after \"-local_lost_connection_command\"";
		}
		# If current command is set and CID != -1 use dp_atclose delete 
		# and append.

		set new_local_conn [lindex $args $cntr];

		if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
		    if {$cmt_subObjectData($nobj_name,local_lost_command) != ""} {
			dp_atclose $cmt_ObjectData($obj_name,rpc_file) delete $cmt_subObjectData($nobj_name,local_lost_command);
		    }

		    if {$new_local_conn != ""} {
			dp_atclose $cmt_ObjectData($obj_name,rpc_file) append $new_local_conn;
		    }
		}

		set cmt_subObjectData($obj_name,local_lost_command) $new_local_conn;
	    }
	    remote_lost_connection_command {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify object configuration value after \"-remote_lost_connection_command\"";
		}
		# If current command is set and CID != -1 use dp_atclose delete 
		# and append.

		set new_remote_conn [lindex $args $cntr];

		if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
		    if {$cmt_subObjectData($nobj_name,remote_lost_command) != ""} {
			dp_RDO $cmt_ObjectData($obj_name,rpc_file) eval dp_atclose \$dp_rpcFile delete [list $cmt_subObjectData($nobj_name,remote_lost_command)];
		    }

		    if {$new_remote_conn != ""} {
			dp_RDO $cmt_ObjectData($obj_name,rpc_file) eval dp_atclose \$dp_rpcFile append [list $new_remote_conn];
		    }
		}

		set cmt_subObjectData($obj_name,remote_lost_command) new_remote_conn;
	    }
	    local_destroy_command {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify object configuration value after \"-local_destroy_command\"";
		}
		set cmt_subObjectData($nobj_name,local_destroy_command) [lindex $args $cntr];
	    }
	    remote_destroy_command {
		incr cntr;
		if {$cntr == $argc} {
		    error "Must specify object configuration value after \"-remote_destroy_command\"";
		}
		set cmt_subObjectData($nobj_name,remote_destroy_command) [lindex $args $cntr];
	    }
	    default {
		error "Unknown object configuration option \"[lindex $args $cntr]\"";
	    }
	}

	incr cntr;
    }
}

#--------------------------------------------------------
#
# cmt_object_info
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the info command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_info {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];
    if {$cmt_ObjectData($obj_name,CID) == -1} {
	error "Cannot query service info without establishing connection, use open";
    }

    if {$argc == 0} {
	error "Wrong # args, should be: \"$obj_name info <info_type> ?args?\"";
    }

    set info_type [cmt_switch_expand [lindex $args 0]
    {service version media formats sources resources destinations object_info}];

    set info_args [lrange $args 1 end];
    if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
	if [catch {dp_RPC $cmt_ObjectData($obj_name,rpc_file)
	    eval cminfo get $info_type $info_args} info_error] {
	    global errorInfo
	    error "$info_error" $errorInfo;
	}
    } else {
	if [catch {eval cminfo get $info_type $info_args} info_error] {
	    global errorInfo
	    error "$info_error" $errorInfo;
	}
    }

    return "$info_error";
}

#--------------------------------------------------------
#
# cmt_object_rpc
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the rpc command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_rpc {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];
    if {$cmt_ObjectData($obj_name,CID) == -1} {
	error "Cannot RPC without connecting to a process or service";
    }

    if {$cmt_ObjectData($obj_name,rpc_file) == ""} {
	error "Cannot RPC with cmt connected to local process. Must be connected to remote process.";
    }

    if {$argc == 0} {
	error "Wrong # args, should be: \"$obj_name rpc ?-events events? ?-timeout milliseconds? ?-timeoutReturn callback? command ?args?\"";
    }

    if [catch {eval dp_RPC $cmt_ObjectData($obj_name,rpc_file) $args} rpc_err] {
	global errorInfo
	error "$rpc_err" $errorInfo;
    }
    return "$rpc_err";
}

#--------------------------------------------------------
#
# cmt_object_rdo
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the rdo command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_rdo {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];
    if {$cmt_ObjectData($obj_name,CID) == -1} {
	error "Cannot RDO without connecting to a process or service";
    }

    if {$cmt_ObjectData($obj_name,rpc_file) == ""} {
	error "Cannot RDO while connected to local process. Must be connected to remote process.";
    }

    if {$argc == 0} {
	error "Wrong # args, should be: \"$obj_name rdo  ?-callback callback? ?-onerror callback? command ?args?\"";
    }

    eval dp_RDO $cmt_ObjectData($obj_name,rpc_file) $args;
}

#--------------------------------------------------------
#
# cmt_object_destroy
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the destroy command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_destroy {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];
    global cmt_ObjectList;

    #close the object
    $obj_name close;

    #destroy the object;
    #Take name out of cmt_ObjectList;
    set idx [lsearch $cmt_ObjectList $obj_name];
    set cmt_ObjectList [lreplace $cmt_ObjectList $idx $idx];
    unset cmt_ObjectData($obj_name,CID);
    unset cmt_ObjectData($obj_name,rpc_file);
    unset cmt_ObjectData($obj_name,nameserver);
    unset cmt_ObjectData($obj_name,nameserver_default);
    unset cmt_ObjectData($obj_name,ns_port);
    unset cmt_ObjectData($obj_name,ns_port_default);
    unset cmt_ObjectData($obj_name,local_close_cmd);
    unset cmt_ObjectData($obj_name,local_close_cmd_default);
    unset cmt_ObjectData($obj_name,check_cmd);
    unset cmt_ObjectData($obj_name,remote_close_cmd);
    unset cmt_ObjectData($obj_name,remote_close_cmd_default);
    unset cmt_ObjectData($obj_name,local_lost_conn_cmd);
    unset cmt_ObjectData($obj_name,local_lost_conn_cmd_default);
    unset cmt_ObjectData($obj_name,remote_lost_conn_cmd);
    unset cmt_ObjectData($obj_name,remote_lost_conn_cmd_default);
    unset cmt_ObjectData($obj_name,sub_object_list);
    rename $obj_name "";
}

#--------------------------------------------------------
#
# cmt_object_getobjs
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the getobjs command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_getobjs {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];
    if {$argc == 0} {
	return "$cmt_ObjectData($obj_name,sub_object_list)";
    }

    if {$cmt_ObjectData($obj_name,CID) == -1} {
	error "Cannot getobjs without connecting to a process or service";
    }
    
    set ret_val "";

    if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
	if {[catch {eval dp_RPC $cmt_ObjectData($obj_name,rpc_file) objectTag getObj $cmt_ObjectData($obj_name,CID) $args} ret_val]} {
	    global errorInfo
	    error "$ret_val" $errorInfo;
	}
    } else {
	if {[catch {eval objectTag getObj $cmt_ObjectData($obj_name,CID) $args} ret_val]} {
	    global errorInfo
	    error "$ret_val" $errorInfo;
	}
    }

    return $ret_val;
}

#--------------------------------------------------------
#
# cmt_object_getallobjs
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the getallobjs command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_getallobjs {obj_name args} {
    global cmt_ObjectData cmt_subObjectData;
    set argc [llength $args];
    if {$cmt_ObjectData($obj_name,CID) == -1}  {
	error "Cannot getallobjs without connecting to a process or service";
    }
    
    set ret_val "";

    if {$cmt_ObjectData($obj_name,rpc_file) != ""} {
	if {[catch {eval dp_RPC $cmt_ObjectData($obj_name,rpc_file) objectTag getAllObjs $args} ret_val]} {
	    global errorInfo
	    error "$ret_val" $errorInfo;
	}
    } else {
	if {[catch {eval objectTag getAllObjs $args} ret_val]} {
	    global errorInfo
	    error "$ret_val" $errorInfo;
	}
    }

    return $ret_val;
}

#--------------------------------------------------------
#
# cmt_object_create_proxie
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the create_proxie command for a cmt object
#
#---------------------------------------------------------

proc cmt_object_create_proxie {obj_name args} {
    global cmt_ObjectData
    set argc [llength $args];
    if {$argc < 1} {
	error "Wrong # args, should be: \"$obj_name create_proxie object_name\"";
    }
    if {$cmt_ObjectData($obj_name,CID) == -1} {
	error "Must be connected to cmt process first.";
    }
    if {$cmt_ObjectData($obj_name,rpc_file) == ""} {
	error "Can not create proxie objects through cmt connect to local process."
    }

    set rem_obj [lindex $args 0];

    dp_RPC $cmt_ObjectData($obj_name,rpc_file) cmtCreateProxie $rem_obj;
}

#---------------------------------------------------------
#
# cmt_object_cmd
#
# Parameters:
#   obj_name : name of object
#   cmd : object method being invoked
#   args : additional arguments.
#
# Actions:
#   This procedure is the procedure that cmt object names are
#   mapped to. The obj_name parameter is fixed at the time of the
#   mapping. This procedure implements all the cmt object's methods.
#
#-------------------------------------------------------------

proc cmt_object_cmd {obj_name {cmd none} args} {
    global cmt_ObjectData

    if {$cmd == "none"} {
	error "Wrong # of arguments, should be: \"$obj_name <command> ?args?\"";
    }

    # Expand cmd to proper value.
    set cmd [cmt_switch_expand $cmd {configure getCID open close info
		rpc rdo destroy create_object getobjs getallobjs
		configure_object create_proxie}];

    switch $cmd {
	configure { eval cmt_object_configure $obj_name $args }
	getCID { return $cmt_ObjectData($obj_name,CID); }
	open { eval cmt_object_open $obj_name $args }
	close { eval cmt_object_close $obj_name $args }
	create_object { eval cmt_object_create_object $obj_name $args }
	configure_object { eval cmt_object_configure_object $obj_name $args }
	info { eval cmt_object_info $obj_name $args }
	rpc { eval cmt_object_rpc $obj_name $args }
	rdo { eval cmt_object_rdo $obj_name $args }
	destroy { eval cmt_object_destroy $obj_name $args }
	getobjs { eval cmt_object_getobjs $obj_name $args }
	getallobjs { eval cmt_object_getallobjs $obj_name $args }
    	create_proxie { eval cmt_object_create_proxie $obj_name $args }
 	default { error "Unknown command \"$cmd\""; }
    }
}

#--------------------------------------------------------
#
# cmt_remote_object_default
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the default command for a remote cmt object
#   Send arguments on to remote object. Return result as error
#   if necessary.
#
#---------------------------------------------------------

proc cmt_remote_object_default {parent obj args} {
    global cmt_ObjectData

    set ret_val ""
    set socket $cmt_ObjectData($parent,rpc_file)

    if {[catch {dp_RPC $socket eval $obj $args} ret_val]} {
	global errorInfo
	error "$ret_val" $errorInfo;
    }
    return "$ret_val";
}

#--------------------------------------------------------
#
# cmt_remote_object_create_object
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the create_object command for a remote cmt object
#
# Notes:
#   This is necessary so that nested cmt's that can create objects downstream
#   do the right thing by setting up the local data structures and procedures
#   necessary.
#
#   The following was used to support cyclic cmt connections but this causes
#   problems so it was removed and replaced with non cyclic routines.
#
#   Pass the create_object downstream. 
#   Set callback to cmt_RDORetValSet var_name
#   Set error callback to cmt_RDORetValSet err_var_name
#
#---------------------------------------------------------

proc cmt_remote_object_create_object {parent obj args} {
    global cmt_ObjectData cmt_subObjectData
    set argc [llength $args];

    if {[catch {dp_RPC $cmt_ObjectData($parent,rpc_file) eval $obj $args} nobj_name]} {
	global errorInfo
	error $nob_name $errorInfo;
    }

    # Check to see if a procedure of the new object name already
    # exists. If so, then probably cyclic. Do not create wrapper
    # procedure, but instead hide it as a hidden remote object.
    # Be careful, though, if the object being hidden is a wrapper
    # function then don't create the datastructures, just use them
    # and be sure to set the wrapper flag for the hidden wrapper 
    # wrapper. Got that.
    
    global dp_rpcFile;

    if {[catch "set dp_rpcFile"]} {
	set gdp_rpcFile "none";
    } else {
	set gdp_rpcFile $dp_rpcFile;
    }

    set wrapper_flag 0;

    if {[lsearch [info commands] $nobj_name] != -1} {
	
	# Find out if its already a wrapper function

	if {![catch "set cmt_subObjectData($nobj_name,local_lost_command)"]} {
	    set wrapper_flag 1;
	} 

	# Rename current object and set up hidden wrapper.
	
	set robj_name "${nobj_name}_hidden_[cmt_GetNextVarName]";
	rename $nobj_name $robj_name;
	
	proc $nobj_name {args} " \
		return \[eval cmt_remote_hidden_object_cmd $parent $nobj_name $robj_name  $gdp_rpcFile $wrapper_flag \$args\]; \
		"
    } else {

	# If no error, return value represents new object. Go through same
	# object creation as before.

	proc $nobj_name {args} "                              \
		return \[eval cmt_remote_object_cmd $parent $nobj_name \$args\]; \
		"
    }

    if {!$wrapper_flag} {

	# Add object name to sub_object list.

	lappend cmt_ObjectData($parent,sub_object_list) $nobj_name;

	
	# Set up data structure for object.
	
	set cmt_subObjectData($nobj_name,local_lost_command) "";
	set cmt_subObjectData($nobj_name,remote_lost_command) "";
	set cmt_subObjectData($nobj_name,local_destroy_command) "";
	set cmt_subObjectData($nobj_name,remote_destroy_command) "";
	set cmt_subObjectData($nobj_name,local_lost_command_default) "";
	set cmt_subObjectData($nobj_name,remote_lost_command_default) "";
	set cmt_subObjectData($nobj_name,local_destroy_command_default) "";
	set cmt_subObjectData($nobj_name,remote_destroy_command_default) "";
	
	# If local and remote lost commands are not "", then register them.
	
	if {$cmt_subObjectData($nobj_name,local_lost_command) != ""} {
	    dp_atclose $cmt_ObjectData($parent,rpc_file) append $cmt_subObjectData($nobj_name,local_lost_command);
	}
	
	if {$cmt_subObjectData($nobj_name,remote_lost_command) != ""} {
	    dp_RDO $cmt_ObjectData($parent,rpc_file) eval dp_atclose \$dp_rpcFile append [list $cmt_subObjectData($nobj_name,remote_lost_command)];
	}
    }

    # Return local object name.

    return $nobj_name;
}

#--------------------------------------------------------
#
# cmt_remote_object_destroy
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the destroy command for a remote cmt object
#
#---------------------------------------------------------

proc cmt_remote_object_destroy {parent obj args} {
    global cmt_ObjectData cmt_subObjectData
    set argc [llength $args];

    #
    # Unregister lost connection commands.
    #
    if {$cmt_subObjectData($obj,local_lost_command) != ""} {
	dp_atclose $cmt_ObjectData($parent,rpc_file) delete $cmt_subObjectData($obj,local_lost_command);
    }

    if {$cmt_subObjectData($obj,remote_lost_command) != ""} {
	dp_RDO $cmt_ObjectData($parent,rpc_file) eval dp_atclose \$dp_rpcFile delete [list $cmt_subObjectData($obj,remote_lost_command)];
    }

    #
    # Execute destroy commands
    # First remotely
    #
    if {$cmt_subObjectData($obj,remote_destroy_command) != ""} {
	catch "dp_RDO $cmt_ObjectData($parent,rpc_file) $cmt_subObjectData($obj,remote_destroy_command)";
    }

    #
    # Now locally, catching results.
    #
    set destroy_cmd_ret "";
    set catch_flag 0;

    if {$cmt_subObjectData($obj,local_destroy_command) != ""} {
	set catch_flag [catch {eval $cmt_subObjectData($obj,local_destroy_cmd)} destroy_cmd_ret];
	global errorInfo
	set saveInfo $errorInfo
    }

    #
    # Destroy datastructures.
    #
    unset cmt_subObjectData($obj,local_lost_command_default);
    unset cmt_subObjectData($obj,local_lost_command);
    unset cmt_subObjectData($obj,remote_lost_command_default);
    unset cmt_subObjectData($obj,remote_lost_command);
    unset cmt_subObjectData($obj,local_destroy_command_default);
    unset cmt_subObjectData($obj,local_destroy_command);
    unset cmt_subObjectData($obj,remote_destroy_command_default);
    unset cmt_subObjectData($obj,remote_destroy_command);

    #
    # Remove from sub object list
    #
    set idx [lsearch $cmt_ObjectData($parent,sub_object_list) $obj];
    if {$idx != -1} {
	set cmt_ObjectData($parent,sub_object_list) [lreplace $cmt_ObjectData($parent,sub_object_list) $idx $idx];
    }

    #
    # Remove all tags from remote objectTag database.
    #
    # set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];
   
    catch "dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag delAllTags $cmt_ObjectData($parent,CID) $obj";
    
    #
    # Send destroy message.
    #
    catch "dp_RDO $cmt_ObjectData($parent,rpc_file) $obj destroy";

    #
    # Kill this procedure.
    #
    rename $obj "";

    if {$catch_flag == 0} {
	return "$destroy_cmd_ret";
    } else {
	error "$destroy_cmd_ret" $saveInfo;
    }
}

#--------------------------------------------------------
#
# cmt_remote_object_addtag
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the addtag command for a remote cmt object
#
#---------------------------------------------------------

proc cmt_remote_object_addtag {parent obj args} {
    global cmt_ObjectData 
    set argc [llength $args];
    if {$argc < 2} {
	error "Wrong # args, should be \"$obj addtag <tag>\"";
    }

    # set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];

    set tag_args [lrange $args 1 end];
    set ret_val "";

    if {[catch {eval dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag addTag $cmt_ObjectData($parent,CID) $obj $tag_args} ret_val]} {
	global errorInfo
	error "$ret_val" $errorInfo;
    }

    #
    # Send command down line if object has its own command. 
    # This is most necessary for cmt objects that create more objects.
    #
    dp_RDO $cmt_ObjectData($parent,rpc_file) -onerror none eval $obj $args

    return "$ret_val";
}

#--------------------------------------------------------
#
# cmt_remote_object_deltag
#
# Parameters:
#   obj_name : name of object
#   args : additional arguments.
#
# Actions:
#   Handle the deltag command for a remote cmt object
#
#---------------------------------------------------------

proc cmt_remote_object_deltag {parent obj args} {
    global cmt_ObjectData

    set argc [llength $args];
    if {$argc != 2} {
	error "Wrong # args, should be \"$obj deltag <tag>\"";
    }

    # set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];

    set ret_val "";
    if {[catch {eval dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag delTag $cmt_ObjectData($parent,CID) $obj [lindex $args 1]} ret_val]} {
	global errorInfo
	error "$ret_val" $errorInfo;
    }

    #
    # Send command down line if object has its own command. 
    # This is most necessary for cmt objects that create more objects.
    #
    dp_RDO $cmt_ObjectData($parent,rpc_file) -onerror none eval $obj $args;
    
    return "$ret_val";
}

#-------------------------------------------------------------
#
# cmt_remote_object_cmd
#
# Parameters:
#   parent : parent cmt object of object.
#   obj : object name
#   args : additional arguments;
#
# Actions:
#   This procedure is used to implement the local methods 
#   of remote objects created through a cmt object. The
#   remote object name is mapped to this procedure with
#   parent and obj parameters pre set. If the object method
#   invoked is one of the tags facility methods, action is
#   taken locally and sometimes passed to the remote object.
#   The cmt create_object method also has a local component
#   in order to deal with nested cmt objects. 
#   The destroy command also has a local action as well as
#   being passed downstream. 
#   All other commands are passed downstream.
#
#--------------------------------------------------------------

proc cmt_remote_object_cmd {parent obj args} {
    global cmt_ObjectData 

    # set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];

    # 2 ms here...
    set cmd [cmt_switch_expand [lindex $args 0] \
		{addtag deltag gettags getalltags create_object destroy}];

    # 8 ms here...
    switch $cmd {
	addtag {eval cmt_remote_object_addtag $parent $obj $args}
	deltag {eval cmt_remote_object_deltag $parent $obj $args}
	gettags {
	    set ret_val "";

	    if {[catch {eval dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag getTag $cmt_ObjectData($parent,CID) $obj} ret_val]} {
		global errorInfo
		error "$ret_val" $errorInfo;
	    }
	    return "$ret_val";
	}
	getalltags {
	    set ret_val "";

	    if {[catch {eval dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag getAllTags $obj} ret_val]} {
		global errorInfo
		error "$ret_val" $errorInfo;
	    }

	    return "$ret_val";
	}
	destroy {eval cmt_remote_object_destroy $parent $obj $args}
	create_object {eval cmt_remote_object_create_object $parent $obj $args}
	default {eval cmt_remote_object_default $parent $obj $args}
    }
}

#-------------------------------------------------------------
#
# cmt_remote_hidden_object_cmd
#
# Parameters:
#   parent : parent cmt object of object.
#   obj : object name
#   robj : the "real" object name;
#   good_dp_rpcFile : value of dp_rpcFile that should be set if
#                  execution is really at this level
#   wrapper_flag : flag that says hidden object is another wrapper.
#   args : additional arguments;
#
# Actions:
#   This procedure is used to implement the local methods 
#   of remote objects created through cyclical cmt objects.
#   It is very much like a normal local wrapper, but it
#   stops infinite cyclical execution and hides the fact
#   that things ended up in a cycle.
#
#----------------------------------------------------------------

proc cmt_remote_hidden_object_cmd {parent obj robj good_dp_rpcFile wrapper_flag args} {

    global cmt_ObjectData cmt_subObjectData dp_rpcFile

    if {[catch "set dp_rpcFile"]} {
	set gdp_rpcFile "none";
    } else {
	set gdp_rpcFile $dp_rpcFile;
    }
    
    # set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];
    
    set argc [llength $args];
    
    set cmd [lindex $args 0];
    
    set cmd [cmt_switch_expand $cmd {addtag deltag gettags getalltags create_object destroy}];
    
    switch $cmd {
	addtag {
	    
	    if {$gdp_rpcFile != "$good_dp_rpcFile"} {
		set ret_val [eval $robj $args];
		return $ret_val;
	    }
	    
	    if {$argc < 2} {
		error "Wrong # args, should be \"$obj addtag <tag>\"";
	    }
	    
	    set tag_args [lrange $args 1 end];
	    
	    set ret_val "";
	    
	    if {[catch {eval dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag addTag $cmt_ObjectData($parent,CID) $obj $tag_args} ret_val]} {
		global errorInfo
		error "$ret_val" $errorInfo;
	    }
	    
	    # Send command down line if object has its own command. 
	    # This is most necessary for cmt objects that create more objects.

	    dp_RDO $cmt_ObjectData($parent,rpc_file) -onerror none eval $obj $args;

	    return "$ret_val";
	}
	deltag {

	    if {$gdp_rpcFile != "$good_dp_rpcFile"} {
		set ret_val [eval $robj $args];
		return $ret_val;
	    }

	    if {$argc != 2} {
		error "Wrong # args, should be \"$obj deltag <tag>\"";
	    }

	    set ret_val "";

	    if {[catch {eval dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag delTag $cmt_ObjectData($parent,CID) $obj [lindex $args 1]} ret_val]} {
		global errorInfo
		error "$ret_val" $errorInfo;
	    }

	    # Send command down line if object has its own command. 
	    # This is most necessary for cmt objects that create more objects.

	    dp_RDO $cmt_ObjectData($parent,rpc_file) -onerror none eval $obj $args;
	    
	    return "$ret_val";
	}
	gettags {

	    if {$gdp_rpcFile != "$good_dp_rpcFile"} {
		set ret_val [eval $robj $args];
		return $ret_val;
	    }


	    set ret_val "";

	    if {[catch {eval dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag getTag $cmt_ObjectData($parent,CID) $obj} ret_val]} {
		global errorInfo
		error "$ret_val" $errorInfo;
	    }
	    return "$ret_val";
	}
	getalltags {

	    if {$gdp_rpcFile != "$good_dp_rpcFile"} {
		set ret_val [eval $robj $args];
		return $ret_val;
	    }

	    set ret_val "";

	    if {[catch {eval dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag getAllTags $obj} ret_val]} {
		global errorInfo
		error "$ret_val" $errorInfo;
	    }

	    return "$ret_val";
	}
	destroy {
	    
	    # Unregister lost connection commands.

	    if {$cmt_subObjectData($obj,local_lost_command) != ""} {
		dp_atclose $cmt_ObjectData($parent,rpc_file) delete $cmt_subObjectData($obj,local_lost_command);
	    }

	    if {$cmt_subObjectData($obj,remote_lost_command) != ""} {
		dp_RDO $cmt_ObjectData($parent,rpc_file) eval dp_atclose \$dp_rpcFile delete [list $cmt_subObjectData($obj,remote_lost_command)];
	    }

	    if {!$wrapper_flag} {
		# Destroy datastructures.

		unset cmt_subObjectData($obj,local_lost_command_default);
		unset cmt_subObjectData($obj,local_lost_command);
		unset cmt_subObjectData($obj,remote_lost_command_default);
		unset cmt_subObjectData($obj,remote_lost_command);
		unset cmt_subObjectData($obj,local_destroy_command_default);
		unset cmt_subObjectData($obj,local_destroy_command);
		unset cmt_subObjectData($obj,remote_destroy_command_default);
		unset cmt_subObjectData($obj,remote_destroy_command);

		# Remove from sub object list

		set idx [lsearch $cmt_ObjectData($parent,sub_object_list) $obj];
		if {$idx != -1} {
		    set cmt_ObjectData($parent,sub_object_list) [lreplace $cmt_ObjectData($parent,sub_object_list) $idx $idx];
		}
	    }

	    # Remove all tags from remote objectTag database.

	    set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];
	   
	    catch "dp_RPC $cmt_ObjectData($parent,rpc_file) objectTag delAllTags $cmt_ObjectData($parent,CID) $obj";
	    
	    # Send destroy message.

	    catch "dp_RDO $cmt_ObjectData($parent,rpc_file) $obj destroy";
	
	    # Kill this procedure.

	    rename $obj "";
	    rename $robj $obj;
	}
	create_object {

	    if {$gdp_rpcFile != "$good_dp_rpcFile"} {
		set ret_val [eval $robj $args];
		return $ret_val;
	    }

	    # This is necessary so that nested cmt's that can create objects downstream
	    # do the right thing by setting up the local data structures and procedures
	    # necessary.

	    # Pass the create_object downstream. 
	    # Set callback to cmt_RDORetValSet var_name
	    # Set error callback to cmt_RDORetValSet err_var_name
	    # 
	    # set var_name [cmt_GetNextVarName];
	    # set var_flag [cmt_GetNextVarName];
	    # set err_var_name [cmt_GetNextVarName];
	    # set err_var_flag [cmt_GetNextVarName];
	    # 
	    # global $var_name $err_var_name $err_var_flag $var_flag;
	    # 
	    # set $var_name "";
	    # set $err_var_name "";
	    # set $var_flag "";
	    # set $err_var_flag "";
	    # 
	    # 
	    # dp_RDO $cmt_ObjectData($parent,rpc_file) -callback "cmt_RDORetValSet $var_name $var_flag " -onerror "cmt_RDORetValSet $err_var_name $err_var_flag " eval $obj $args;
	    # 
	    # 
	    # while {1} {
		# if {[set $var_flag] != ""} {
		    # set is_error 0;
		    # break;
		# }
		# if {[set $err_var_flag] != ""} {
		    # set is_error 1;
		    # break;
		# }
		# dp_update;
	    # }
	    # 
	    # if {$is_error == 1} {
		# set err_msg [set $err_var_name];
		# unset $err_var_name;
		# unset $err_var_flag;
		# unset $var_name;
		# unset $var_flag;
		# 
		# error "$err_msg";
	    # }
	    # 
	    # set nobj_name [set $var_name];
	    # 
	    # unset $err_var_name;
	    # unset $err_var_flag;
	    # unset $var_name;
	    # unset $var_flag;

	    if {[catch {dp_RPC  $cmt_ObjectData($parent,rpc_file) eval $obj $args} nobj_name]} {
		global errorInfo
		error "$nobj_name" $errorInfo;
	    }


	    # Check to see if a procedure of the new object name already
	    # exists. If so, then probably cyclic. Do not create wrapper
            # procedure, but instead hide it as a hidden remote object.
	    # Be careful, though, if the object being hidden is a wrapper
	    # function then don't create the datastructures, just use them
	    # and be sure to set the wrapper flag for the hidden wrapper 
	    # wrapper. Got that.
	    
	    set wrapper_flag 0;

            if {[lsearch [info commands] $nobj_name] != -1} {
		
		# Find out if its already a wrapper function

		if {![catch "set cmt_subObjectData($nobj_name,local_lost_command)"]} {
		    set wrapper_flag 1;
		} 

		# Rename current object and set up hidden wrapper.
		
		set robj_name "${nobj_name}_hidden_[cmt_GetNextVarName]";
		rename $nobj_name $robj_name;
		
		proc $nobj_name {args} " \
			return \[eval  cmt_remote_hidden_object_cmd $parent $nobj_name $robj_name $gdp_rpcFile $wrapper_flag \$args\]; \
			"
	    } else {

		# If no error, return value represents new object. Go through same
		# object creation as before.

		proc $nobj_name {args} "                              \
			return \[eval cmt_remote_object_cmd $parent $nobj_name \$args\]; \
			"e
	    }

	    if {!$wrapper_flag} {

		# Add object name to sub_object list.

		lappend cmt_ObjectData($parent,sub_object_list) $nobj_name;

		
		# Set up data structure for object.
		
		set cmt_subObjectData($nobj_name,local_lost_command) "";
		set cmt_subObjectData($nobj_name,remote_lost_command) "";
		set cmt_subObjectData($nobj_name,local_destroy_command) "";
		set cmt_subObjectData($nobj_name,remote_destroy_command) "";
		set cmt_subObjectData($nobj_name,local_lost_command_default) "";
		set cmt_subObjectData($nobj_name,remote_lost_command_default) "";
		set cmt_subObjectData($nobj_name,local_destroy_command_default) "";
		set cmt_subObjectData($nobj_name,remote_destroy_command_default) "";
		
		# If local and remote lost commands are not "", then register them.
		
		if {$cmt_subObjectData($nobj_name,local_lost_command) != ""} {
		    dp_atclose $cmt_ObjectData($parent,rpc_file) append $cmt_subObjectData($nobj_name,local_lost_command);
		}
		
		if {$cmt_subObjectData($nobj_name,remote_lost_command) != ""} {
		    dp_RDO $cmt_ObjectData($parent,rpc_file) eval dp_atclose \$dp_rpcFile append [list $cmt_subObjectData($nobj_name,remote_lost_command)];
		}
	    }

	    # Return local object name.

	    return $nobj_name;
	}
	default {
	    eval $robj $args;
	}
    }
}

#------------------------------------------------------------------
#
# cmt_local_object_cmd
#
# Parameters:
#   parent : parent cmt object of object.
#   obj : object name
#   robj : "real" object name
#   args : additional arguments
#
# Actions:
#   This procedure is used to implement the methods of local objects
#   created through cmt objects have been opened with the "-local" flag.
#   The local object name is mapped to this procedure with
#   parent, obj, and robj parameters preset. If the object method
#   invoked is one of the tags facility methods, action is taken
#   locally and sometimes passed to the real object. The cmt create_object
#   method also has a local component in order to deal with nested
#   cmt objects. 
#   The destroy command also has a local action as well as being passed
#   downstream. 
#   All other commands are passed downstream.
#
#---------------------------------------------------------------------

proc cmt_local_object_cmd {parent obj robj args} {
    global cmt_ObjectData;

    set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];

    set cmd [cmt_switch_expand [lindex $args 0] \
		 {addtag deltag gettags getalltags create_object destroy}];

    switch $cmd {
	addtag {eval cmt_local_object_addtag $parent $obj $robj $args}
	deltag {eval cmt_local_object_deltag $parent $obj $robj $args}
	gettags {
	    set ret_val "";

	    if {[catch {eval objectTag getTag $cmt_ObjectData($parent,CID) $obj} ret_val]} {
		global errorInfo
		error "$ret_val" $errorInfo;
	    }
	    return "$ret_val";
	}
	getalltags {
	    set ret_val "";

	    if {[catch {eval objectTag getAllTags $obj} ret_val]} {
		global errorInfo
		error "$ret_val" $errorInfo;
	    }

	    return "$ret_val";
	}
	destroy {eval cmt_local_object_destroy $parent $obj $robj $args}
	create_object {eval cmt_local_object_create_object $parent $obj $robj $args}
	default {eval cmt_local_object_default $robj $args}
    }
}

#--------------------------------------------------------
#
# cmt_local_object_default
#
# Parameters:
#   parent : parent of object
#   obj : name of object
#   robj : real name of object
#   args : additional arguments.
#
# Actions:
#   Handle the default command for a local object
#   Return result as error if necessary.
#
#---------------------------------------------------------

proc cmt_local_object_default {robj args} {

    set ret_val ""

    if {[catch {eval $robj $args} ret_val]} {
	global errorInfo
	error "$ret_val" $errorInfo;
    }
    return "$ret_val";
}

#--------------------------------------------------------
#
# cmt_local_object_create_object
#
# Parameters:
#   parent : name of parent
#   obj : name of object
#   robj : real name of object
#   args : additional arguments.
#
# Actions:
#   Handle the create_object command for a local cmt object
#
# Notes:
#   This is necessary so that nested cmt's that can create objects downstream
#   do the right thing by setting up the local data structures and procedures
#   necessary. This is particularly tricky when the cmt parent of this object
#   is actually connected to the local process.
#
#
#---------------------------------------------------------

proc cmt_local_object_create_object {parent obj robj args} {
    global cmt_ObjectData cmt_subObjectData

    set argc [llength $args];

    if {[catch {eval $robj $args} nobj_name]} {
	global errorInfo
	error $nob_name $errorInfo;
    }

    # If no error, return value represents new object. Go through same
    # object creation as before, hidding real object and replacing with local
    # wrapper.

    set robj_name "${nobj_name}_hidden_[cmt_GetNextVarName]";
    rename $nobj_name $robj_name;

    proc $nobj_name {args} "\
        return \[eval cmt_local_object_cmd $parent $nobj_name $robj_name \$args\];";
    
    lappend cmt_ObjectData($parent,sub_object_list) $nobj_name;

    # Don't set up any data structures since that will have been done by the
    # local cmt child created through this cmt.

    return $nobj_name;
}

#--------------------------------------------------------
#
# cmt_local_object_destroy
#
# Parameters:
#   parent : name of parent
#   obj : name of object
#   robj : real name of object
#   args : additional arguments.
#
# Actions:
#   Handle the destroy command for a local object created
#   through a cmt that was connected with the "-local" flag.
#
#---------------------------------------------------------

proc cmt_local_object_destroy {parent obj robj args} {
    global cmt_ObjectData cmt_subObjectData
    set argc [llength $args];

    #
    # Execute local destroy command.
    #
    set destroy_cmd_ret "";
    set catch_flag 0;

    if {[info exists cmt_subObjectData($obj,local_destroy_command)]} {
	if {$cmt_subObjectData($obj,local_destroy_command) != ""} {
	    set catch_flag [catch {eval $cmt_subObjectData($obj,local_destroy_cmd)} destroy_cmd_ret];
	    global errorInfo
	    set saveInfo $errorInfo
	}

	#
	# Destroy datastructures.
	#
	unset cmt_subObjectData($obj,local_lost_command_default);
	unset cmt_subObjectData($obj,local_lost_command);
	unset cmt_subObjectData($obj,remote_lost_command_default);
	unset cmt_subObjectData($obj,remote_lost_command);
	unset cmt_subObjectData($obj,local_destroy_command_default);
	unset cmt_subObjectData($obj,local_destroy_command);
	unset cmt_subObjectData($obj,remote_destroy_command_default);
	unset cmt_subObjectData($obj,remote_destroy_command);
    }

    #
    # Remove from sub object list
    #
    set idx [lsearch $cmt_ObjectData($parent,sub_object_list) $obj];
    if {$idx != -1} {
	set cmt_ObjectData($parent,sub_object_list) [lreplace $cmt_ObjectData($parent,sub_object_list) $idx $idx];
    }

    #
    # Remove all tags from objectTag database.
    #
    set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];
   
    catch "objectTag delAllTags $cmt_ObjectData($parent,CID) $obj";
    
    #
    # Send destroy message to real object.
    #
    catch "$robj destroy";

    #
    # Kill this procedure.
    #
    catch {rename $obj ""};

    #
    # Kill hidden procedure.
    #
    catch {rename $robj ""};

    if {$catch_flag == 0} {
	return "$destroy_cmd_ret";
    } else {
	error "$destroy_cmd_ret" $saveInfo;
    }
}

#--------------------------------------------------------
#
# cmt_local_object_addtag
#
# Parameters:
#   parent : name of parent.
#   obj : name of object
#   robj : real name of object.
#   args : additional arguments.
#
# Actions:
#   Handle the addtag command for a local object made through
#   a cmt connected with the "-local" option.
#
#---------------------------------------------------------

proc cmt_local_object_addtag {parent obj robj args} {
    global cmt_ObjectData 

    set argc [llength $args];

    if {$argc < 2} {
	error "Wrong # args, should be \"$obj addtag <tag>\"";
    }

    set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];

    set tag_args [lrange $args 1 end];
    set ret_val "";

    if {[catch {eval objectTag addTag $cmt_ObjectData($parent,CID) $obj $tag_args} ret_val]} {
	global errorInfo
	error "$ret_val" $errorInfo;
    }

    #
    # Send command down line if object has its own command. 
    # This is most necessary for cmt objects that create more objects.
    #

    catch "eval $robj $args"

    return "$ret_val";
}

#--------------------------------------------------------
#
# cmt_local_object_deltag
#
# Parameters:
#   parent : name of parent
#   obj : name of object
#   robj : real name of object.
#   args : additional arguments.
#
# Actions:
#   Handle the deltag command for a local object made through
#   a cmt connected with the "-local" syntax.
#
#---------------------------------------------------------

proc cmt_local_object_deltag {parent obj robj args} {
    global cmt_ObjectData

    set argc [llength $args];

    if {$argc != 2} {
	error "Wrong # args, should be \"$obj deltag <tag>\"";
    }

    set obj_tail [string range $obj [expr "[string first $parent $obj]+[string length $parent]+3"] end];

    set ret_val "";
    if {[catch {eval objectTag delTag $cmt_ObjectData($parent,CID) $obj [lindex $args 1]} ret_val]} {
	global errorInfo
	error "$ret_val" $errorInfo;
    }

    #
    # Send command down line if object has its own command. 
    # This is most necessary for cmt objects that create more objects.
    #
    catch "eval $robj $args";
    
    return "$ret_val";
}

proc cmtCreateProxie {obj_name} {
    global dp_rpcFile;

    proc $obj_name {args} " \
	if {\[lindex \$args 0\] == \"destroy\"} { \
	    dp_RDO $dp_rpcFile eval $obj_name \$args; \
	    rename $obj_name \"\"; \
	    return; \
	};
	set ret_val \[dp_RPC $dp_rpcFile eval $obj_name \$args\]; \
	return \$ret_val; \
    "

    dp_atclose $dp_rpcFile append  "rename $obj_name \"\"";
    return;
}

