# ------------------------------------------------------------------
#
# {tcl2doc AUTHOR Eric Baldeschwieler, eric14@cs.berkeley.edu}
#
# Date: 03-02-95
#
# {tcl2doc DESCRIPTION}
# The CMAPP library is a collection of functions which are callable
# from any tcl/tk wish.  They provide a layer of helping functions
# between the raw CMT objects and user scripts.  This file supports
# some basic CMAPP commands.  Other commands can be found in:
#
#	cmapp-movie (simpleMovie.tcl)
#	cmapp-network (gen_net.tcl)
#	cmapp-tags (gen_net.tcl)
#
# ------------------------------------------------------------------

set gCMApp_DebugFile {}

# ------{tcl2doc FUNCTION}----------------------------------------
# If any of $args is not a number then an error is produced.
# Switches: -noerr cases 0 to be returned for non-number.
# Return: 1

proc CMApp_CkNum {args} {
    set noerr 0
    if {[lindex $args 0] == "-noerr"} {
	set noerr 1
	set args [lrange $args 1 end]
    }
    foreach I $args {
	if {[catch {expr "$I+1"}]} {
	    if {!$noErr} {
		error "not a number: $I"
	    }
	    return 0
	}
    }
    return 1
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Returns first non-empty non-comment line from open file 'fid' or
# {} for EOF.

proc CMApp_GetsClean {fid} {
    while {[gets $fid line] != -1} {
	set line [string trim $line]
	if {![string match "\#*" $line] && $line != {}} {
	    return $line
	}
    }
    return {}
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Used to determine if this process can invoke cmp commands directly
#     or if it needs to call them remotely.

proc CMApp_IsCMProccess {} {
    return [info exists kCMProcess]
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Like C's assert.  If script evaluates to 0 reports an error.  $args
#     are appended to the end of the error message.

proc ASSERT {scrpt args} {
    if {[catch {uplevel "expr $scrpt"} out]} {
	puts "CATCH OF ERROR - $out - {expr $scrpt}"
	set out 0
    }
    if {$out == 0} {
	return -code error "ASSERTION FAILURE: $scrpt  --  $args"
    }
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Simple debug output funtion.  If the variable gCMApp_DebugFile is
#     set to a file name (or stdout) then the string is sent to that
#     file, otherwise this function is a no-op.  Currently the
#     moduleTag is ignored, but this may later be used to allow finer
#     control of debug output.

proc CMApp_SetDebug {file} {
    global  gCMApp_DebugFile
    set gCMApp_DebugFile $file
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Simple wrapper for the common catch cases.  If an error occures in
#     $script, then a catch action is taken, based on $actType.
#     The $_errorTxt contains the text within the actScript and
#     $errorInfo contains the call chain info.  If present, the
#     $elseScript will be executed if no error occures.  If
#     $elseScript matches the string "as_above" the $actScript is
#     executed regardless of whether or not an error occures.
# rethrow - actScript is executed and then the error is rethrown,
#     using the possibly modified _errorTxt value.  Usefull for
#     clean-up actions.
# do - actScript is executed and the error is dropped
# report - As do, but the error is sent to stderr via puts.

proc CMApp_IfError {script actType actScript {else else} {elseScript {}}} {
    global  errorInfo
    uplevel global errorInfo
    upvar _errorTxt result
    set code [catch {uplevel $script} result]

    if {$code == 0} {
	if {$else != "else"} {error "bad keyword, expected 'else'"}
	if {$elseScript == "as_above"} {return [uplevel $actScript]}
	return [uplevel $elseScript]
    } elseif {$code != 1} {
	return -code $code $result
    }
    switch -- $actType {
	rethrow {
	    uplevel $actScript
	    error $result $errorInfo
	}
	do {
	    return [uplevel $actScript]
	}
	report {
	    puts stderr "** ERROR - $result \n$errorInfo"
	    return [uplevel $actScript]
	}
    }
    error "Unknown action type: $actType"
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Simple debug output funtion.  If the variable gCMApp_DebugFile is
#     set to a file name (or stdout) then the string is sent to that
#     file, otherwise this function is a no-op.  Currently the
#     moduleTag is ignored, but this may later be used to allow finer
#     control of debug output.

proc CMApp_Debug {moduleTag string} {
    global  gCMApp_DebugFile
    if {$gCMApp_DebugFile != {}} {
	puts $gCMApp_DebugFile "\# CMApp_DEBUG: $string"
    }
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Used to delay the execution of an event.  If several events are
# queued with the same tag, only the most recent is executed.  Proves
# very useful in UI construction.  Setting a tag to a script of {}
# clears any pending events with that tag.  Note that tags may not
# begin with an '@' character.

set gCMApp_IdleEvents(@Qed) 0

proc CMApp_DoWhenIdle {tag script} {
    global  gCMApp_IdleEvents

    set tag ".$tag"
    if {$script == {}} {
	if {[info exists gCMApp_IdleEvents($tag)]} {
	    set gCMApp_IdleEvents($tag) ""
	}
	return
    }
    if {!$gCMApp_IdleEvents(@Qed)} {
	after idle CMApp_DoWhenIdleUtil
	set gCMApp_IdleEvents(@Qed) 1
    }
    set gCMApp_IdleEvents($tag) $script
}

proc CMApp_DoWhenIdleUtil {} {
    # note that the order of events is VERY important here, because
    # this function is not atomic!
    global  gCMApp_IdleEvents

    CMApp_IfError {
	foreach E [array names gCMApp_IdleEvents {.*}] {
	    set gCMApp_IdleEvents(@E) $gCMApp_IdleEvents($E)
	    unset gCMApp_IdleEvents($E)
	    eval $gCMApp_IdleEvents(@E)
	}
    } do {
	puts stderr "ERROR - CMApp_DoWhenIdleUtil:"
	puts stderr "\t$_errorTxt\n\t$errorInfo"
    }

    if {[array names gCMApp_IdleEvents {.*}] != {}} {
	after idle CMApp_DoWhenIdleUtil
    } else {
	set gCMApp_IdleEvents(@Qed) 0
    }
}

# ------------------------------------------------------------------
# End of File
