# ------------------------------------------------------------------
#
# Author: Eric14@cs.berkeley.edu
#
# Date: 02-27-95
#
# Purpose: Generally useful list handling routines.
#
# ------------------------------------------------------------------

# ------{tcl2doc FUNCTION}----------------------------------------
# Maps a list of vals to a list of vars.  If the '-init' flag is
# present then, vars w/o corresponding vals are set to {}, else an
# error is returned.  Extra vals are always ignored.
# This is great for returning multipule values from a function!
#
#	CMApp_ListSet {?varName? ...} {?value? ...} ?-init? or ?-exact? \
#		?-excess?

proc CMApp_ListSet {varLst vals {flag -excess}} {
    set nVars [llength $varLst]
    set nVals [llength $vals]
    switch -- $flag {
	-init {}
	-excess {if {$nVals < $nVars} {
	    error "\# CMApp_ListSet - not enough values"}}
	-exact  {if {$nVals != $nVars} {
	    error "\# CMApp_ListSet - wrong number of values"}}
	default {error "unexpected final argument: $flag"}
    }
    set i 0
    foreach I $varLst {
	upvar $I v
	set v [lindex $vals $i]
	incr i
    }
}
    
# ------{tcl2doc FUNCTION}----------------------------------------
# Takes a list of switches (without the initial dash) and sets
# variables of the same names  to the values specified in the argList. 
#	(format switchL: "{name ?initVal? ?possibleValList?} ...")
#	(format argList: "-sw val -sw val...) 
# switchL entries can be either simple names or lists.  The 2nd arg is
# the initial value (if pressent) and the final arg is a list of legal
# values for the switch.
#
# If the final arg to the function is -init.  Then all switch
# variables are initalised to {}, otherwise they are unset.

proc CMApp_SwitchList {switchL argList {init {}} {upCnt 1}} {
    switch -- $init {
	-init {set init 1}
	{} {set init 0}
	default {error "CMApp_SwitchList: only -init ok as final arg"}
    }

    foreach I $switchL {
	set name [lindex $I 0]
	lappend sLst "-$name"
	upvar $upCnt $name $name
	if {[llength $I] > 1} {
	    set $name [lindex $I 1]
	} elseif {$init} {
	    set $name {}
	}
    }

    for {} {[llength $argList] > 1} {set argList [lrange $argList 2 end]} {
	set key [lindex $argList 0]
	if {[lsearch -exact $sLst $key] != -1} {
	    set [string range $key 1 end] [lindex $argList 1]
	} else {
	    error "Unknow option '$key', expecting one of: $sLst"
	}
    }
    if {[llength $argList] != 0} {
	error "Unexpected value or switch w/o value: $argList"
    }

    foreach I $switchL {
	if {[llength $I] > 2 && \
		[lsearch -exact [lindex $I 2] [set [lindex $I 0]]] == -1} {
	    error "[lindex $I 0] must be one of: [lindex $I 2]"
	}
    }
}

# ------{tcl2doc FUNCTION}----------------------------------------
# As CMApp_SwitchList, except that all arguments preceeding the
# first switch are identified and returned.

proc CMApp_ArgsAndSwitchList {switchL argList {init {}}} {
    set i [lsearch -glob $argList "-*"]
    if {$i == -1} {
	CMApp_SwitchList $switchL {} $init 2
	return $argList
    }
    CMApp_SwitchList $switchL [lrange $argList $i end] $init 2
    return [lrange $argList 0 [expr $i-1]]
}
    
# ------------------------------------------------------------------
# End of File
