# ------------------------------------------------------------------
#
#	simpleMovie.tcl
#
# {tcl2doc AUTHOR Eric Baldeschwieler, eric14@cs.berkeley.edu}
#
# date:	11-29-94
#
# {tcl2doc DESCRIPTION}
# Simple movie model functions.  These functions all deal with
# cmt-movies (scripts).  They have two representations, a file format
# and a playList format.  Playlists are a simple tcl internal
# representation (a list of lists) which can be stored in simple tcl
# variables.  When read from a script all of the media files
# (clip-files) represented in the playList are stored as URLs and many
# fields may be unspecified (BAD).  CMApp_ResolveMovie does the necisary
# network queries and file investigations to produce a list ready for
# playback (it replaces all BAD fields and adds an address tupple to the
# end of the list.
#
# File format:
#
# Following is an example movie file:
#
# 	# CMT-Movie-1.0
#
# 	movie video audio
#
# 	# empty lines, white space and comments allowed
# 	clip 0 ~eric14/raid1
# 	clip 0 cmtp://cms.berkeley.edu:666/n/vid6/meetings/chips.mpeg
# 	clip 0 /n/vid2/lost -ss 2 -se 3 -ls 5 -le 7 -type video
# 	clip 1 relativelyBad -ss 2 -se 3 -ls 5 -le 7 -type audio/mpeg2
#
#
# This file is intended to be scanned, not sourced, so 'clip' and
# 'movie' are just keywords, not tcl commands.  The movie line specifies
# the number and type of tracks in the file:
#
#     movie ?trackType? ... ?-tags {?{<tag> <value>}? ...}?
#
# So the example above specifies a movie with two tracks.  Track 0 is a
# video track and track 1 is an audio track.  The clip lines each
# specify a continuous segment of media from a single file.
#
#     clip <trackNum> <file> ?switch value? ...
#
# The file may be either a local file or a cmtp or vods URL.  All other
# fields are optional switches.  At read time any missing information
# about a clip will be derived by examining the source file via
# CMApp_SourceFileInfo which determines the TYPE, FORMAT and duration of
# any clip.  Given this information all of a clip's fields can be
# inteligently defaulted.  If the clip does not include a logical start
# time, then the clip is assumed to begin immediately after the previous
# clip in its track.
#
#     -ss    - source start time
#     -se    - source end time
#     -ls    - logical start time (output time)
#     -le    - logical end time
#     -type  - type/format info needed to choose CMT objects
#
#
# Internal PlayList format:
#
# The playList for the movie listed above (before CMApp_ResolveMovie) would be:
#
#    {video
#      {~eric14/raid1 BAD BAD BAD BAD BAD}
#      {cmtp://cms.berkeley.edu:666/n/vid6/meetings/chips.mpeg BAD BAD BAD BAD BAD}
#      {/n/vid2/lost video 2 3 5 7}}
#    {audio
#      {relativelyBad audio:mpeg2 2 3 5 7}}
#
# ----------------------------------------------------------------

set kCMMovie_Types {video audio}
set kCMMovie_Version   CMT-Movie-1.0
set kCMMovieHeader "\# $kCMMovie_Version"
set kCMCollectionHeader "## CMT Editor Collection file:"

# ------{tcl2doc FUNCTION}----------------------------------------
# Takes a playList (as returned from CMApp_ReadMovie) and translates
#   each media URL into a {host port file}-tupple which refferances
#   a runnning CMX.  Any BAD fields in a clip are also replaced with
#   valid entries.  This maybe expensive if media URLs point to
#   non-clip files or VODS urls...  The tupples are appended to the
#   end of each clip-list.

proc CMApp_ResolveMovie {playList} {
    lappend outList [lindex $playList 0]
    
    foreach I [lrange $playList 1 end] {
	set trkType [lindex $I 0]
	set outTrk $trkType
	
	set prevEnd 0
	foreach J [lrange $I 1 end] {
	    set J [CMApp_ResolveClip $J $prevEnd]
	    append  outTrk "\n   "
	    lappend outTrk $J
	    set prevEnd [lindex $J 3]
	}
	append outList "\n "
	lappend outList $outTrk
    }
    return $outList
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Takes a playList (as returned from CMApp_ReadMovie) and writes it to
#   the given file. If no file name or {} is given then the file is
#   returned, not writen.

proc CMApp_WriteMovie {playList {fileName {}}} {
    global  kCMMovieHeader kCMMovie_Types kCMMovie_Version

    set out "$kCMMovieHeader\n\n"
    
    set tags [lindex $playList 0]
    set playList [lrange $playList 1 end]

    set line "movie"
    foreach I $playList {
	lappend line [lindex $I 0]
    }
    if {$tags != {}} {
	lappend line -tags $tags
    }
    append out "$line\n"

    set trkCnt 0
    foreach I $playList {
	foreach J [lrange $I 1 end] {
	    CMApp_ListSet {ss se ls le file type tags} $J

	    lappend out  clip $trkCnt $file
	    if {$type != "BAD"} {lappend out  -type $type}
	    if {$ss != "BAD"}   {lappend out  -ss $ss}
	    if {$se != "BAD"}   {lappend out  -se $se}
	    if {$ls != "BAD"}   {lappend out  -ls $ls}
	    if {$le != "BAD"}   {lappend out  -le $le}
	    if {$tags != {}}    {lappend out  -tags $tags}
	    append out "\n"
	}
	incr trkCnt
    }

    if {$fileName != {}} {
	set fid [open $fileName w]
	puts $fid $out
	close $fid
    } else {
	return $out
    }
}
    
# ------{tcl2doc FUNCTION}----------------------------------------
# Opens the given file and tests the first line to see if it is a
#   proper CMT-Movie header.
# Returns: 0 or 1

proc CMApp_IsMovie {fileName} {
    global  kCMMovieHeader
    set out 1
    set bytes [string length $kCMMovieHeader]
    set fid [open $fileName r]
    if {![string match $kCMMovieHeader [read $fid $bytes]]} {
	set out 0
    }
    close $fid
    return $out
}
    
# ------{tcl2doc FUNCTION}----------------------------------------
# Opens the given file and tests the first line to see if it is a
#   proper CMT-Movie header.
# Returns: 0 or 1

proc CMApp_IsCollection {fileName} {
    global  kCMCollectionHeader
    set out 1
    set bytes [string length $kCMCollectionHeader]
    set fid [open $fileName r]
    set line [read $fid $bytes]
    if {![string match $kCMCollectionHeader $line]} {
	set out 0
    }
    close $fid
    return $out
}
    
# ------{tcl2doc FUNCTION}----------------------------------------
# This is the WONDER FUNCTION.  Takes any cmfile or URL and returns a
#   playList (as returned by CMApp_ReadMovie).  If the referanced
#   file is a Movie files, it is read using CMApp_ReadMovie,
#   otherwise, a simple script is generated to play the given media
#   file.

proc CMApp_ReadURL {url} {
    set tupple [CMApp_TranslateURL $url]
    CMApp_ListSet {type dur} [CMApp_SourceFileInfo $tupple]
    CMApp_ListSet {host port file} $tupple
    if {$type == "movie/cmt"} {
	set out [CMApp_RPC $host $port CMApp_ReadMovie $file]
	return $out
    }
    
    if {![regexp {^([^/]+)/} $type x cmType]} {
	error "URL <$url> has bad type: $type"
    }
    
    set clip [list 0 $dur 0 $dur $url $type {} $tupple]
    set playList [list {} [list $cmType $clip]]
}
    
# ------{tcl2doc FUNCTION}----------------------------------------
# Opens the given file and returns a playList.  See above for format.
# Returns: a list of the form:
#    {<tagList> ?<track>?...}
#
#    <track> := {<cmType> ?<clipInfo>?...}}
#    
#    <cmType> is one of {audio or video (may be extended)}
#    <clipInfo> := {<ss> <se> <ls> <le> <URL> <cmType>/<format> <tagList>}
#    <ss> := source start
#    <se> := source end
#    <ls> := LTS start
#    <le> := LTS end
#
#    <tagList> = {?{<tag> <value>}?...}

proc CMApp_ReadMovie {fileName} {
    global  kCMMovieHeader kCMMovie_Types kCMMovie_Version

    set fid [open $fileName r]
    set bytes [string length $kCMMovieHeader]
    if {![string match "$kCMMovieHeader" [read $fid $bytes]]} {
	error "Not a $kCMMovie_Version file: $fileName"
    }
    set line [CMApp_GetsClean $fid]
    set numTracks [llength $]
    if {[lindex $line 0] != "movie"} {
	error "First movie line not 'movie' command!"
    }
    set line [CMApp_ArgsAndSwitchList {{tags {}}} [lrange $line 1 end]]
    set numTracks [llength $line]

    set j 0
    foreach I $line {
	if {[lsearch -exact $kCMMovie_Types $I] == -1} {
	    error "Bad track type: $I in line: $line"
	}
	set tracks($j) $I
	# puts "tracks($j) = $tracks($j)"
	incr j
    }
    
    while {[set line [CMApp_GetsClean $fid]] != {}} {
	switch -- [lindex $line 0] {
	    clip {
		set trk [lindex $line 1]
		if {![CMApp_CkNum $trk] || $trk < 0 || $trk >= $numTracks} {
		    error "bad track in line: $line"
		}
		append tracks($trk) "\n   "
		lappend tracks($trk) [CMApp_ParseClipLine [lrange $line 2 end]]
	    }
	    default {
		error "Not a clip line: $line"
	    }
	}
    }
    
    lappend out $tags
    for {set j 0} {$j < $numTracks} {incr j} {
	append out "\n"
	lappend out $tracks($j)
    }
    
    close $fid
    return $out
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Helping function to scan a clip line

proc CMApp_ParseClipLine {line} {
    if {[llength $line] == 0} {
	error "No file given for clip"
    }
    set file [lindex $line 0]
    CMApp_SwitchList {
	{ss BAD} {se BAD} {ls BAD} {le BAD} {type BAD} 
	{tags {}} {tupple {}}
    } [lrange  $line 1 end]
    set out [list $ss $se $ls $le $file $type $tags]
    if {$tupple != {}} {
	lappend out $tupple
    }
    return $out
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Helping function to choose good default values for unknow fields of
#   a clip
# Returns: A news version of the clip without BAD fields. 

proc CMApp_ResolveClip {clip prevEnd} {
    CMApp_ListSet {ss se ls le file type tags tupple} $clip -init
    if {$tupple == {}} {
	set tupple [CMApp_TranslateURL $file]
    }
    if {($le=="BAD" && $se=="BAD") || ![match {*/*} $type]} {
	CMApp_ListSet {type dur} [CMApp_SourceFileInfo $tupple]
	if {$se == "BAD"} {set se $dur}
    }		

    if {$ls == "BAD"} {set ls $prevEnd}
    if {$ss == "BAD"} {set ss 0}

    if {$le=="BAD"} {
	set le [expr $ls+$se-$ss]
    } elseif {$se=="BAD"} {
	set se [expr $ss+$le-$ls]
    }
    
    list $ss $se $ls $le $file $type $tags $tupple
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Returns the type of the given resolved clip (a list).

proc CMApp_GetResolvedClipType {clip} {
    regsub {\/.*$} [lindex $clip 5] {} type
    return $type
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Returns the duration of the given resolved clip (a list).

proc CMApp_GetResolvedClipDuration {clip} {
    CMApp_ListSet {ss se ls le file type tags tupple} $clip -init
    return [expr $le-$ls]
}

# ------{tcl2doc FUNCTION}----------------------------------------
# Takes a resolved clip (a list) and adjusts the logical start and end
# times so that duration is unchanged and ls = $base.  The -maxdur
# flag causes the clip to be cropped if its duration exceeds the given
# value.  The -base flag sets the logical start of the cli.  It
# defaults to zero.

proc CMApp_NormalizeResolvedClip {clip args} {
    CMApp_ListSet {ss se ls le file type tags tupple} $clip -init
    CMApp_SwitchList {{maxdur {}} {base 0}} $args
    set le [expr $le-$ls]
    #set ls 0
    if {$maxdur != {} && $le > $maxdur} {
	ASSERT {$maxdur > 0}
	set le $maxdur
	set se [expr ($se-$ss)*($maxdur/$le)+$ss]
    }
    set ls $base
    set le [expr $base+$le]
    return [list $ss $se $ls $le $file $type $tags $tupple]

}

# ------------------------------------------------------------------

proc CMApp_ValidatePlayList {playlist} {
    set track 0
    #puts "VALIDATE- PL= <<$playlist>>"
    foreach TRK [lrange $playlist 1 end] {
	set clip 0
	#puts "TRK = <<$TRK>>"
	foreach CLP [lrange $TRK 1 end] {
	    #puts "CLP = <<$CLP>>"
	    CMApp_ListSet {ss se ls le file type tags tupple} $CLP -init
	    if {$ls >= $le} {
		error "INVALID: track,clip ($track,$clip) ls>le ($ls,$le)"
	    }
	    incr clip
	}
	incr track
    }
    
}
	    
# ------------------------------------------------------------------
# End of File
