/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
#define DEBUG_FRAMES
*/
#ifdef NO_STRING_H
#   include "compat/string.h"
#else
#   include <string.h>
#endif

#include <stdio.h>

#ifdef NO_STDLIB_H
#   include "compat/stdlib.h"
#else
#   include <stdlib.h>
#endif

#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <tcl.h>
#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#else
#   include "compat/unistd.h"
#endif
#include <tk.h>
#include <math.h>
#include "cmCookies.h"
#include "cmBufferlist.h"
#include "cmChunk.h"
#include "cmLTS.h"
#include "cmAtom.h"
#include "tcmAtQ.h"

#define ITERATOR(I,FORWARD,FIRST,LASTPLUSONE)				\
     (I)=(FORWARD)?(FIRST):((LASTPLUSONE)-1);				\
     (FORWARD)?((I)<(LASTPLUSONE)):((I)>=(LASTPLUSONE));		\
     (I)+=(FORWARD)*2-1

#define ONE_BILLION		1000000000.0
#define CF_READ_BUFFER_SIZE	512

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif

/* Description of a vq chunk */

typedef enum { VQ_FRAME, VQ_CODEBOOK } VQ_ChunkType;

typedef struct {
  VQ_ChunkType		type;
  int			size;
} VQ_ChunkHeader;

typedef struct {
  int			start;			/* in bytes, from the start of the data */
  int			length;			/* in bytes */
} frameOffset;

/* 
 * Structure kept for each segment in segment list.
 */

typedef struct SegmentStruct {
  int			segId;
  double		clipStart;
  double		clipEnd;
  double		logicalStart;
  double		logicalEnd;
  int			numFrames;
  frameOffset		*FOT;			/* n entries (replaces FOT) */
  double		*FTT;			/* Frame Time Table: n+1 start/end times */
  int			*FDT;			/* Frame Dependency Table: n Codebook indices */
  int			maxSize;
  struct SegmentStruct	*next;
  struct SegmentStruct	*prev;
} Segment;

typedef struct {
  int	alreadySent;
  int	length;
  char	*data;
} Codebook;

typedef struct {
  Buffer		*bufPtr;
  Segment		*segPtr;
  int			index;
  int			size;
  double		sTime;
  double		eTime;
  int			sendFlag;
  int			offset;			/* start of frame, from FOT in Segment */
} FramePick;

typedef struct {
  Tcl_Interp		*interp;		/* Tcl interpreter that created object. */
  char			*name;			/* Tcl string name of this object. */
  ClientData		lts;
  char			*ltsName;
  int			ltsTriggerNum;		/* LTS trigger for changed lts. */
  int			atToken;
  double		sendAhead;
  double		cycleTime;		/* in seconds */
  char			*fileName;
  tChunkFileState	*chunkFileState;
  VQHeader		*cfInfo;
  char			*outCommand;
  InFunc		outCCmd;		/* C pointer to out command */
  ClientData		outCD;			/* ClientData for outCCmd. */
  char			*codebookOutCmd;
  InFunc		codebookOutCCmd;	/* C pointer to codebook out command */
  ClientData		codebookOutCD;		/* ClientData for codebookOutCCmd. */
  int			isReady;
  double		fileLength;		/* in seconds (was clipLength) */
  Segment		*segments;		/* linked list of segments */
  int			numFrames;		/* Number of frames in clipfile. */
  int			dataStart;              /* Offset to start of data. */
  double		maxFrameRate;		/* Maximum frame rate. */
  int			bandwidth;		/* Bandwidth available */
  int			nextSegId;		/* Next segment id. */
  Codebook		*codebooks;		/* array of codebooks */
  char			*debugFileName;
  FILE			*debugFile;
} vq_CFObject;

/* Globals used by this object module. */

static int nextvq_CFNum = 0;	/* Variable for next unique object name. */

/************************************************************
 *
 * CreateNewvq_CF
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

vq_CFObject *CreateNewvq_CF(Tcl_Interp *interp, char *name)
{
  vq_CFObject *new;

  new = (vq_CFObject *) ckalloc(sizeof(vq_CFObject));
  if (new == NULL) {
    Tcl_AppendResult(interp, "Cannot allocate vq_CFObject named: ", name, NULL);
  }
  else {
    new->name = (char *) ckalloc(strlen(name) + 1);
    if (new->name == NULL) {
      Tcl_AppendResult(interp, "Cannot allocate vq_CFObject named: ", name, NULL);
      ckfree((char *) new);
      new = NULL;
    }
    else {
      new->interp	= interp;
      strcpy(new->name, name);
      new->lts		= NULL;
      new->ltsName	= NULL;
      new->ltsTriggerNum = -1;
      new->atToken	= 0;
      new->sendAhead	= 0.2;
      new->cycleTime	= 0.0;
      new->fileName	= NULL;
      new->chunkFileState = NULL;
      new->cfInfo	= NULL;
      new->outCommand	= NULL;
      new->outCCmd	= NULL;
      new->outCD	= NULL;
      new->codebookOutCmd	= NULL;
      new->codebookOutCCmd	= NULL;
      new->codebookOutCD	= NULL;
      new->isReady	= 0;
      new->fileLength	= 0.0;
      new->segments	= NULL;
      new->bandwidth	= 0;
      new->numFrames	= 0;
      new->dataStart	= 0;
      new->maxFrameRate	= 0.0;
      new->nextSegId	= 0;
      new->codebooks	= NULL;
    }
  }
  return new;
}

/************************************************************
 *
 * vq_CF_DeleteSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segId -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
vq_CF_DeleteSegment(vq_CFObject *objPtr, int segId)
{
  Segment *segPtr;
  
  segPtr = objPtr->segments;
  
  while ((segPtr != NULL) && (segPtr->segId != segId)) {
    segPtr = segPtr->next;
  }
  
  if (segPtr != NULL) {
    if (segPtr->prev == NULL) {
      objPtr->segments = segPtr->next;
    } else {
      segPtr->prev->next = segPtr->next;
    }
    
    if (segPtr->next != NULL) {
      segPtr->next->prev = segPtr->prev;
    }
    
    if (segPtr->FOT != NULL)  {
      ckfree((char *) segPtr->FOT);
    }
    if (segPtr->FTT != NULL) {
      ckfree((char *) segPtr->FTT);
    }
    ckfree((char *) segPtr);
  }
  
  return TCL_OK;
}

/************************************************************
 *
 * vq_CFObject_ClearSegments
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
vq_CF_ClearSegments(vq_CFObject * objPtr)
{
  while (objPtr->segments != NULL) {
    vq_CF_DeleteSegment(objPtr, objPtr->segments->segId);
  }
  return TCL_OK;
}

/************************************************************
 *
 * Destroyvq_CF
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

int
Destroyvq_CF(vq_CFObject *objPtr)
{

  /*
   * First delete Tcl command.
   */
  Tcl_DeleteCommand(objPtr->interp, objPtr->name);

  if (objPtr->lts != NULL) {
    if (objPtr->ltsTriggerNum != -1) {
      LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
      objPtr->ltsTriggerNum = -1;
    }
    Tk_Release(objPtr->lts);
    objPtr->lts = NULL;
    if (objPtr->ltsName != NULL) {
      ckfree(objPtr->ltsName);
    }
  }

  if (objPtr->atToken != 0) {
    Tcm_DeleteAtHandler(objPtr->atToken);
    objPtr->atToken = 0;
  }

  if (objPtr->name != NULL) {
    ckfree(objPtr->name);
  }

  if (objPtr->fileName != NULL) {
    ckfree(objPtr->fileName);
  }

  if (objPtr->chunkFileState != NULL) {
    Chunk_CloseFile(objPtr->chunkFileState);
    objPtr->chunkFileState = NULL;
  }

  if (objPtr->outCommand != NULL)
    ckfree(objPtr->outCommand);

  /*
   * Free all segments.
   */
  vq_CF_ClearSegments(objPtr);

  ckfree((char *) objPtr);
  
  return TCL_OK;
}

/************************************************************
 *
 * vq_FillInBufferHeader
 *
 * Parameters:
 *	bufferHeaderPtr	header to be filled in
 *	headerPtr	input
 *	startTime	logical start time
 *	endTime		logical end time
 *	codebookIndex	codebook this frame depends on
 *
 * Return:
 *     VQBufferHeader
 *
 * Action:
 *	Fills in bufferHeaderPtr from info in other parameters.  Converts
 *	values to network order.
 *     
 ************************************************************/
static void
vq_FillInBufferHeader(VQBufferHeader *bufferHeaderPtr,
		      VQHeader *headerPtr,
		      double startTime,
		      double endTime,
		      int codebookIndex)
{
  if ((bufferHeaderPtr == NULL) || (headerPtr == NULL)) {
    fprintf(stderr, "vq_FillInBufferHeader: high bogosity: I was passed a NULL pointer\n");
    return;
  }
    
  bufferHeaderPtr->hdrType	= htonl(CM_VQ_BUF_HDR);
  bufferHeaderPtr->length	= htonl(sizeof(VQBufferHeader)-2*sizeof(int));
  bufferHeaderPtr->width	= htonl(headerPtr->width);
  bufferHeaderPtr->height	= htonl(headerPtr->height);
  bufferHeaderPtr->device	= htonl(headerPtr->device);
  bufferHeaderPtr->subformat	= htonl(headerPtr->subformat);
  bufferHeaderPtr->logStart_s	= htonl(floor(startTime));
  bufferHeaderPtr->logStart_ns	= htonl(floor((startTime - floor(startTime))* ONE_BILLION));
  bufferHeaderPtr->logEnd_s	= htonl(floor(endTime));
  bufferHeaderPtr->logEnd_ns	= htonl(floor((endTime - floor(endTime))* ONE_BILLION));
  bufferHeaderPtr->codebookIndex	= htonl(codebookIndex);
}

/*
 *--------------------------------------------------------------
 *
 * vq_Prioritize -- (cut & pasted from MJPEG_Prioritize)
 *
 *	A prioritization algorithm (by Daishi) that computes the
 *	ordering for the VQ frames.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 * Parameters (these are my best guesses -- bsmith)
 *	int n		Number of frames to prioritize
 *	int m		Starting frame number.
 *	int k		Starting priority number
 *	int *pl		Output priority array
 *
 *--------------------------------------------------------------
 */
void
vq_Prioritize (int n, int m, int k, int *pl)
{
  int s,t;

  if (n==1) {
    *pl = m;
  } else {
    s = n/2;
    t = n-s;
    vq_Prioritize(t,m,k*2,pl);
    vq_Prioritize(s,m+t,k*2,pl+k);
  }
}

/************************************************************
 *
 * SendBuffer
 *
 * Parameters:
 *	interp		Tcl interpreter
 *	bufPtr		Buffer to send
 *	numBytes	# of bytes to send
 *	outCommandPtr	pointer to pointer to out command
 *	outCCommandPtr	pointer to C out function
 *	outCD		client data for out C function
 *
 * Return:
 *	None.
 *
 * Action:
 *	Send a buffer of data using the specified functions.
 *	If the C command doesn't exist, reset the pointers to NULL.
 *     
 ************************************************************/
static void
SendBuffer(Tcl_Interp	*interp,
	   Buffer	*bufPtr,
	   int		numBytes,
	   char		**outCommandPtr,
	   InFunc	*outCCommandPtr,
	   ClientData	outCD)
{
  SB_List	*sbList;
  ScatterBuffer	*sbPtr;
  char		*outCommand = *outCommandPtr;
  InFunc	outCCommand = *outCCommandPtr;
  char		*sbString;

  sbPtr = MakeScatterBuffer(bufPtr, numBytes, NULL);
  sbList = MakeSBList(sbPtr, NULL);

  /* use C function, if set */
  if (outCCommand != NULL) {
    /* make sure it still exists */
    if (Tutil_FindAtom(interp, outCommand+1) != NULL) {
      outCCommand(outCD, sbList);
    }
    else {
      /* The C command went away */
      *outCCommandPtr = NULL;
      ckfree(outCommand);
      *outCommandPtr = NULL;
    }
  }
  else {
    /* no C command, so use Tcl */
    sbString = MakeSBListString(sbList);
    Tcl_VarEval(interp, outCommand, sbString, (char *) NULL);
    ckfree(sbString);
  }
  FreeSBList(sbList);
  DetachBuffer(interp, bufPtr, "tcmvqFile:SendBuffer");
}
	   
/************************************************************
 *
 * vq_CF_ReadFrame
 *
 * Parameters:
 *	cfFd		clip file file descriptor
 *	frameBuf	pointer to the frame buffer
 *			(must already be allocated)
 *	hdrPtr		pointer to a (filled in) VQBufferHeader
 *	numBytes	number of bytes to be read
 *
 * Return:
 *	Number of bytes put in the buffer (including header info)
 *
 * Action:
 *	Fill in the frame with header info (provided in parameters) and
 *	frame data (read from the file.)
 *     
 ************************************************************/
static int
vq_CF_ReadFrame(int cfFd,
		char *frameBuf,
		const VQBufferHeader *hdrPtr,
		int numBytes)
{
  int idx;
  int dataSegCode, dataLength;

  idx = 0;
  
  memcpy(frameBuf+idx, hdrPtr, sizeof(VQBufferHeader));
  idx += sizeof(VQBufferHeader);
  
  dataSegCode = htonl(CM_DATA_SEGMENT);
  dataLength = htonl(numBytes);
  
  memcpy(frameBuf+idx, &dataSegCode, 4);
  idx += 4;
  
  memcpy(frameBuf+idx, &dataLength, 4);
  idx += 4;
  
  idx += read(cfFd, frameBuf+idx, numBytes);

  return idx;
}

/************************************************************
 *
 * SendCodebook
 *
 * Parameters:
 *	cfPtr		our VQ clipfile object	
 *	segPtr		segment containing frames of interest
 *	frameIndex	index of frame within segment
 *
 * Return:
 *	None.
 *
 * Action:
 *	If necessary, send codebook for indicated frame.  Update
 *	codebook data structure so it won't be sent again.
 *     
 ************************************************************/
static void
SendCodebook(vq_CFObject *cfPtr,
	     Segment *segPtr,
	     int frameIndex)
{
  Buffer	*bufPtr;
  int		bufIndex = 0;
  int		codebookIndex;
  Codebook	*codebookPtr;
  int		totalBytes;
  int		temp;

  codebookIndex = segPtr->FDT[frameIndex];
  codebookPtr = &cfPtr->codebooks[codebookIndex];

  if (!codebookPtr->alreadySent) {
#ifdef DEBUG_CB
    fprintf(stderr, "SendCodebook: sending cb %d for frame %d\n",
	    codebookIndex, frameIndex);
#endif
    
    totalBytes = codebookPtr->length + 16;
    bufPtr = MakeBuffer(cfPtr->interp, totalBytes,
			"tcmvqClipfile:SendCodebook.1");

    temp = htonl(CM_VQ_BUF_HDR);
    memcpy(bufPtr->data+bufIndex, &temp, 4);
    bufIndex+=4;
    
    temp = htonl(codebookPtr->length);
    memcpy(bufPtr->data+bufIndex, &temp, 4);
    bufIndex+=4;

    temp = htonl(CM_VQ_CODEBOOK_DATA);
    memcpy(bufPtr->data+bufIndex, &temp, 4);
    bufIndex+=4;

    temp = htonl(codebookIndex);
    memcpy(bufPtr->data+bufIndex, &temp, 4);
    bufIndex+=4;

    memcpy(bufPtr->data+bufIndex, codebookPtr->data, codebookPtr->length);
    
    SendBuffer(cfPtr->interp, bufPtr, totalBytes, &cfPtr->codebookOutCmd,
	       &cfPtr->codebookOutCCmd, cfPtr->codebookOutCD);

    codebookPtr->alreadySent = TRUE;
  }
}


/************************************************************
 *
 * vq_CF_SendData
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     none
 *
 * Action:
 *     Sends data according to current value of LTS, queues
 *     itself on the at queue for next send time.
 ************************************************************/

void
vq_CF_SendData(vq_CFObject *objPtr) 
{
  double now, speed, when;
  double eTime, sTime, lastETime;
  Segment *segPtr;
  int i, curBw;
  VQBufferHeader header;
  double curFPS;
  Buffer *bufPtr;
  int idx;
  ScatterBuffer *sbPtr;
  SB_List *sbl;
  char *sbStr;
  int numToSend, bytesToSend, origNumToSend;
  int frameNum, curIndex;
  int sbIdx;
  static FramePick *fPickList;
  static int fPickNum = 0;
  int data_fd;

#ifdef DEBUG_LTS
  fprintf(stderr, "vqFile:SendData entered\n");
#endif
  /*
   * Return if not ready, no LTS, or no out command.
   */
  if ((!objPtr->isReady) ||
      (objPtr->lts == NULL) ||
      (objPtr->outCommand == NULL)) {
    goto done;
  }

  /*
   * Find segment we are in.  Two cases: forward and reverse play.
   */
  now = LTS_Value(objPtr->lts);
  speed = LTS_Speed(objPtr->lts);
  now += objPtr->sendAhead * speed;
  segPtr = objPtr->segments;
  if (speed >= 0.0) {
    while (segPtr != NULL) {
      if (segPtr->logicalEnd > now) break;
      segPtr = segPtr->next;
    }
  } else {
    /*
     * Speed is backwards so go to end of segment list and search backwards.
     */
    if (segPtr != NULL) {
      while(segPtr->next != NULL) {
	segPtr = segPtr->next;
      }
	
      while (segPtr != NULL) {
	if (segPtr->logicalStart < now) break;
	segPtr = segPtr->prev;
      }
    }
  }

  /*
   * If segment pointer is NULL, means we will never be
   * needed (i.e. time past all segments or going backwards
   * before all segments).
   */

  if (segPtr == NULL) {
    goto done;
  }

#ifdef DEBUG_LTS
  fprintf(stderr, "vqFile:SendData:segment [ %f, \t%f ]\n",
	  segPtr->logicalStart, segPtr->logicalEnd);
#endif

  /*
   * If not in a segment, calculate when we will be and set up at queue.
   */
  if (speed > 0.0) {
    if (now < segPtr->logicalStart) {
      when = LTS_SystemTimeOf(objPtr->lts,
			      (segPtr->logicalStart-(objPtr->sendAhead*speed)));
      objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					     "tk", (Tk_TimerProc *)vq_CF_SendData, objPtr, NULL,
					     NULL, NULL, NULL, NULL);
      return;
    }
  }
  else if (speed < 0.0) {
    if (now > segPtr->logicalEnd) {
      when = LTS_SystemTimeOf(objPtr->lts,
			      segPtr->logicalEnd+(objPtr->sendAhead*speed));
      objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					     "tk", (Tk_TimerProc *) vq_CF_SendData, objPtr, NULL,
					     NULL, NULL, NULL, NULL);
      return;
    }
  }
  else {
    if ((now < segPtr->logicalStart) || (now >= segPtr->logicalEnd)) {
      return;
    }
  }

  /* We only read chunk files.  Open it now. */
  if (objPtr->chunkFileState == NULL) {
    objPtr->chunkFileState = Chunk_OpenFile(objPtr->fileName, "r");
  }
  data_fd = objPtr->chunkFileState->fd;

  /*
   * If only sending one frame because speed = 0.0 or
   * cycle time = 0.0, do this now. 
   */
  if ((objPtr->cycleTime == 0.0) || (speed == 0.0)) {
	
    /*
     * Find frame to send.
     */
    if (speed > 0.0) {
      for (i=0; i<segPtr->numFrames; i++) {
	if (segPtr->FTT[i+1] > now) {
	  break;
	}
      }
    }
    else if (speed < 0.0) {
      for (i=segPtr->numFrames; i>0; i--) {
	if (segPtr->FTT[i-1] < now) {
	  break;
	}
      }
    }
    else { /* speed == 0 */
      for (i=0; i<segPtr->numFrames; i++) {
	if ((segPtr->FTT[i] <= now) && (segPtr->FTT[i+1] > now)) {
	  break;
	}
      }
    }

    /* determine start and end times */
    
    if (speed > 0.0) {
      if (i < 0)
	i = 0;
      if (i >= segPtr->numFrames) {
	i = segPtr->numFrames-1;
      }

      sTime = segPtr->FTT[i];
      eTime = segPtr->FTT[i+1];
      frameNum = i;
    }
    else if (speed < 0.0) {
      if (i <= 0)
	i = 1;
      if (i > segPtr->numFrames) {
	i = segPtr->numFrames;
      }

      sTime = segPtr->FTT[i];
      eTime = segPtr->FTT[i-1];
      frameNum = i-1;
    }
    else {
      if (i < 0)
	i = 0;
      if (i >= segPtr->numFrames) {
	i = segPtr->numFrames-1;
      }

      sTime = eTime = now;
      frameNum = i;
    }

#if defined(DEBUG_LTS) || defined(DEBUG_FRAMES)
    fprintf(stderr, "vqFile:SendData: t %f\tf %d @ %d\n",
	    now, frameNum, segPtr->FOT[frameNum].start);
#endif

    /* Send codebook if necessary */
    SendCodebook(objPtr, segPtr, frameNum);
	
    vq_FillInBufferHeader(&header, objPtr->cfInfo, sTime, eTime,
			  segPtr->FDT[frameNum]);

    lseek(data_fd,
	  objPtr->dataStart+segPtr->FOT[frameNum].start,
	  SEEK_SET);

    bufPtr = MakeBuffer(objPtr->interp,
			sizeof(VQBufferHeader)+8+segPtr->FOT[frameNum].length,
			"tcmvqClipfile:vq_CF_SendData.1");

    idx = vq_CF_ReadFrame(data_fd, bufPtr->data, &header, segPtr->FOT[frameNum].length);
    
    /*
     * Create sbl.
     */
    sbPtr = MakeScatterBuffer(bufPtr, idx, NULL);
    sbl = MakeSBList(sbPtr, NULL);

    /*
     * If out command is C, create scatter buffer list and call.
     */
    if (objPtr->outCCmd != NULL) {
      /*
       * Need to check if it still exists since object may have
       * been destroyed.
       */
      if (Tutil_FindAtom(objPtr->interp, objPtr->outCommand+1) != NULL) {
	objPtr->outCCmd(objPtr->outCD, sbl);	       
      }
      else {
	/*
	 * Out C command has unexpectedly disappeared.
	 */
	objPtr->outCCmd = NULL;
	ckfree(objPtr->outCommand);
	objPtr->outCommand = NULL;
      }
    }
    else {
      /*
       * Out command is tcl, make sb list into a string and 
       * call using Tcl_Eval. 
       */
      sbStr = MakeSBListString(sbl);
      Tcl_VarEval(objPtr->interp, objPtr->outCommand, sbStr, (char *) NULL); 
      ckfree(sbStr);
    }
    FreeSBList(sbl);

    DetachBuffer(objPtr->interp, bufPtr, 
		 "tcmvqClipfile:vq_CF_SendData");

    /*
     * Figure out when we need to send again, queue it up and return.
     */
    if (speed == 0.0) {
      return;
    }

    when = LTS_SystemTimeOf(objPtr->lts,
			    (eTime-(objPtr->sendAhead*speed)));
    objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					  "tk", (Tk_TimerProc *) vq_CF_SendData, objPtr, NULL,
					  NULL, NULL, NULL, NULL);
    return;
  } 
    
  /*
   * Cycle time is not 0.0, so send multiple frames.
   * Find frames to be sent.
   */
  numToSend = 0;
  bytesToSend = 0;
  if (fPickNum == 0) {
    fPickList = (FramePick *) ckalloc(sizeof(FramePick)*100);
    if (fPickList == NULL) {
      fprintf(stderr, "error: couldn't alloc memory for fPickList in '%s' at line %d\n",
	      __FILE__, __LINE__);
      return;
    }
    fPickNum = 100;
  }

  if (speed > 0.0) {
    for (i=0; i<segPtr->numFrames; i++) {
      if (segPtr->FTT[i+1] > now)
	break;
    }
  }
  else {
    for (i=segPtr->numFrames; i>0; i--) {
      if (segPtr->FTT[i-1] < now) {
	break;
      }
    }
  }

  if (speed >= 0.0) {
    if (i < 0) {
      i = 0;
    }
    else if (i >= segPtr->numFrames) {
      i = segPtr->numFrames-1;
    }

    fPickList[0].sTime = segPtr->FTT[i];
    fPickList[0].eTime = segPtr->FTT[i+1];
    fPickList[0].index = i;
    fPickList[0].segPtr = segPtr;
    fPickList[0].offset = segPtr->FOT[i].start;
    fPickList[0].size = segPtr->FOT[i].length;
    fPickList[0].sendFlag = 1;
  } else {
    if (i <= 0) {
      i = 1;
    }
    else if (i > segPtr->numFrames) {
      i = segPtr->numFrames;
    }

    fPickList[0].sTime = segPtr->FTT[i];
    fPickList[0].eTime = segPtr->FTT[i-1];
    fPickList[0].index = i-1;
    fPickList[0].segPtr = segPtr;
    fPickList[0].offset = segPtr->FOT[i-1].start;
    fPickList[0].size = segPtr->FOT[i-1].length;
    fPickList[0].sendFlag = 1;
  }
  numToSend = 1;
  bytesToSend = fPickList[0].size;
  curIndex = fPickList[0].index;

  if (speed > 0.0) {
    while (1) {
      curIndex++;
      if (curIndex >= segPtr->numFrames) {
	segPtr = segPtr->next;
	if (segPtr == NULL)
	  break;
	curIndex = 0;
      }
      if (segPtr->FTT[curIndex] < now+objPtr->cycleTime*speed) {
	if (numToSend >= fPickNum) {
	  fPickNum += 100;
	  fPickList = (FramePick *) realloc((char *) fPickList,
					    sizeof(FramePick)*fPickNum);
	}

	fPickList[numToSend].sTime = segPtr->FTT[curIndex];
	fPickList[numToSend].eTime = segPtr->FTT[curIndex+1];
	fPickList[numToSend].index = curIndex;
	fPickList[numToSend].segPtr = segPtr;
	fPickList[numToSend].offset = segPtr->FOT[curIndex].start;
	fPickList[numToSend].size = segPtr->FOT[curIndex].length;
	fPickList[numToSend].sendFlag = 1;
	
	bytesToSend += fPickList[numToSend].size;
	numToSend++;
      }
      else {
	break;
      }
    }
  } else {
    while (1) {
      curIndex--;
      if (curIndex < 0) {
	segPtr = segPtr->prev;
	if (segPtr == NULL) break;
	curIndex = segPtr->numFrames-1;
      }
      if (segPtr->FTT[curIndex+1] > now+objPtr->cycleTime*speed) {
	if (numToSend >= fPickNum) {
	  fPickNum += 100;
	  fPickList = (FramePick *) realloc((char *) fPickList,
					    sizeof(FramePick)*fPickNum);
	}
		
	fPickList[numToSend].sTime = segPtr->FTT[curIndex+1];
	fPickList[numToSend].eTime = segPtr->FTT[curIndex];
	fPickList[numToSend].index = curIndex;
	fPickList[numToSend].segPtr = segPtr;
	fPickList[numToSend].offset = segPtr->FOT[curIndex].start;
	fPickList[numToSend].size = segPtr->FOT[curIndex].length;
	fPickList[numToSend].sendFlag = 1;
	
	bytesToSend += fPickList[numToSend].size;
	numToSend++;
      } else {
	break;
      }
    }
  }

  if (segPtr == NULL) {

  }
  /*
   * If number of frames to be sent greater than the number of
   * frames we can send (due to bandwidth or max frame rate),
   * mark low priority frames so we don't bother reading them.
   */
  curFPS = ((double) numToSend) / objPtr->cycleTime;
  curBw = (int) (((double) bytesToSend) / objPtr->cycleTime);
  origNumToSend = numToSend;
  if ((numToSend > 1) &&
      (((curFPS > objPtr->maxFrameRate) && objPtr->maxFrameRate) ||
       ((curBw  > objPtr->bandwidth)    && objPtr->bandwidth))) {
    int *orderVec;

    orderVec = (int *)ckalloc(numToSend*sizeof(int));
    if (orderVec == NULL) {
      fprintf(stderr, "error: couldn't alloc memory for order vector in '%s' at line %d\n",
	      __FILE__, __LINE__);
      return;
    }
    vq_Prioritize(numToSend, 0, 1, orderVec);
    while ((numToSend > 1) &&
	   ((objPtr->maxFrameRate && (curFPS > objPtr->maxFrameRate)) ||
	   (objPtr->bandwidth && (curBw > objPtr->bandwidth)))) {
      numToSend--;
      fPickList[orderVec[numToSend]].sendFlag = 0;
      bytesToSend -= fPickList[numToSend].size;
      curFPS = numToSend / objPtr->cycleTime;
      curBw  = bytesToSend / objPtr->cycleTime;
    }
    ckfree((char *)orderVec);
  }

  /*
   * For each frame chosen, create buffers with headers and data.
   */
  sbl = (SB_List *) ckalloc(sizeof(SB_List));
  if (sbl == NULL) {
    fprintf(stderr,
	    "error: couldn't alloc memory for scatter buffer list in '%s' at line %d\n",
	    __FILE__, __LINE__);
    return;
  }
  sbl->sbs = (ScatterBuffer **) ckalloc(sizeof(ScatterBuffer *) *
					numToSend);
  if (sbl->sbs == NULL) {
    fprintf(stderr,
	    "error: couldn't alloc memory for scatter buffer array in '%s' at line %d\n",
	    __FILE__, __LINE__);
    return;
  }
  sbl->count = numToSend;
  sbIdx = 0;

  for (i=0; i<origNumToSend; i++) {
    if (fPickList[i].sendFlag == 1) {
      segPtr = fPickList[i].segPtr;

      SendCodebook(objPtr, segPtr, i);
      
      vq_FillInBufferHeader(&header, objPtr->cfInfo,
			    fPickList[i].sTime, fPickList[i].eTime,
			    segPtr->FDT[i]);

      lseek(data_fd,
	    objPtr->dataStart+fPickList[i].offset,
	    SEEK_SET);

      bufPtr = MakeBuffer(objPtr->interp,
			  sizeof(VQBufferHeader)+8+fPickList[i].size,
			  "tcmvqClipfile:vq_CF_SendData.2");
      fPickList[i].bufPtr = bufPtr;

      idx = vq_CF_ReadFrame(data_fd, bufPtr->data, &header, fPickList[i].size);
#ifdef DEBUG_FRAMES
      fprintf(stderr, "vqFile:SendData: packing frame %d @ %f\t(%d bytes)\n",
	      fPickList[i].index, fPickList[i].sTime, fPickList[i].size);
#endif
      sbl->sbs[sbIdx] = MakeScatterBuffer(bufPtr, idx, NULL);
      sbIdx++;
    }
  }

  /*
   * If out command is C, create scatter buffer list and call.
   */
  if (objPtr->outCCmd != NULL) {


    /*
     * Need to check if it still exists since object may have
     * been destroyed.
     */
    if (Tutil_FindAtom(objPtr->interp, objPtr->outCommand+1) != NULL) {
      objPtr->outCCmd(objPtr->outCD, sbl);
    } else {
      /*
       * Out C command has unexpectedly disappeared.
       */
      objPtr->outCCmd = NULL;
      ckfree(objPtr->outCommand);
      objPtr->outCommand = NULL;
    }
  } else {
    /*
     * Out command is tcl, make sb list into a string and
     * call using Tcl_Eval. 
     */
    sbStr = MakeSBListString(sbl);
    Tcl_VarEval(objPtr->interp, objPtr->outCommand, sbStr, (char *) NULL); 
    ckfree(sbStr);
    FreeSBList(sbl);
  }

  for (i=0; i<origNumToSend; i++) {
    if (fPickList[i].sendFlag == 1) {
      lastETime = fPickList[i].eTime;
      DetachBuffer(objPtr->interp, fPickList[i].bufPtr,
		   "tcmvqClipfile:vq_CF_SendData");
    }
  }
  FreeSBList(sbl);
  
  /*
   * Find next wake up time and set up atqueue.
   */
  when = LTS_SystemTimeOf(objPtr->lts,
			  lastETime-(objPtr->sendAhead*speed));
  objPtr->atToken = Tcm_CreateAtHandler(when, HUGE_VAL,
					"tk", (Tk_TimerProc *) vq_CF_SendData, objPtr, NULL,
					NULL, NULL, NULL, NULL);
  return;

  done:
  if (objPtr->chunkFileState != NULL) {
    Chunk_CloseFile(objPtr->chunkFileState);
    objPtr->chunkFileState = NULL;
  } 
    
  return;
}

/************************************************************
 *
 * vq_CF_Ready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Gets the object ready and starts sending data.
 ************************************************************/

static int
vq_CF_Ready(vq_CFObject * objPtr)
{
  /*
   * Blow away the old stuff.
   */
  if (objPtr->atToken != 0) {
    Tcm_DeleteAtHandler(objPtr->atToken);
  }
  objPtr->atToken = 0;
    
  /*
   * Set the ready flag.
   */
  objPtr->isReady = 1;

  if (objPtr->lts != NULL) {
    vq_CF_SendData(objPtr);
  }
  return TCL_OK;
}

/************************************************************
 *
 * vq_CF_LTSChanged
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Procedure called whenever the LTS changes.
 ************************************************************/

void
vq_CF_LTSChanged(ClientData clientdata)
{
  vq_CFObject *objPtr = (vq_CFObject *) clientdata;

  if (objPtr->isReady == 1) {
    vq_CF_Ready(objPtr);
  }
  return;
}

/************************************************************
 *
 * Chunk_ReallyFindType
 *
 * Parameters:
 *	cf	Chunk file
 *	type	Type to search for
 *	hdr	header to put info into
 *
 * Return:
 *	Same as Chunk_FindType.
 *
 * Action:
 *	What most people really want from Chunk_FindType:
 *	Find the type wherever it is in the file.  As C_FT, it
 *	leaves the file pointer at the start of the data segment
 *	if successful, at the end of the file if not (I think).
 *
 ************************************************************/
static int
Chunk_ReallyFindType(tChunkFileState* cf, int32 type, tChunkHeader* hdr)
{
  int result;
  
  if ((result = Chunk_FindType(cf, type, hdr)) != 0) {
    /* didn't find it, so try from the beginning */
    Chunk_ReadRewind(cf);
    result = Chunk_FindType(cf, type, hdr);
  }
  return result;
}

/************************************************************
 *
 * ReadCodebooks
 *
 * Parameters:
 *
 *
 *
 * Return:
 *	A pointer to the codebooks, or NULL on error.
 *
 * Action:
 *
 *
 ************************************************************/
static Codebook *
ReadCodebooks(Tcl_Interp	*interp,
	      tChunkFileState	*chunkFile,
	      int		numCodebooks)
{
  tChunkHeader	chunkHeader;
  Codebook	*codebooks;
  int		i;
  
  if (Chunk_ReallyFindType(chunkFile, CM_VQ_CODEBOOKS, &chunkHeader) != 0) {
    Tcl_AppendResult(interp,
		     "Couldn't find VQ frame dependency header.", (char *) NULL);
    return NULL;
  }
  codebooks = (Codebook *) ckalloc(sizeof(Codebook) * numCodebooks);
  if (codebooks == NULL) {
    Tcl_AppendResult(interp,
		     "Couldn't allocate memory for codebook headers.", (char *) NULL);
    return NULL;
  }
  for (i = 0; i < numCodebooks; i++) {
    read(chunkFile->fd, &codebooks[i].length, sizeof(int));
    codebooks[i].data = ckalloc(codebooks[i].length);
    if (codebooks[i].data == NULL) {
      Tcl_AppendResult(interp,
		       "Couldn't allocate memory for codebook data.", (char *) NULL);
      /* Note: memory already allocaed is wasted, but since this is
         probably a fatal error, I don't care. */
      return NULL;
    }
    
    read(chunkFile->fd, codebooks[i].data, codebooks[i].length);
    codebooks[i].alreadySent = 0;
  }
  return codebooks;
}

/************************************************************
 *
 * vq_CF_OpenDebugfile
 *
 * Parameters:
 *     objPtr - pointer to object
 *     filename - file name
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Opens the clipfile, reads basic info data.
 *     Closes file when done.
 ************************************************************/

static int
vq_CF_OpenDebugfile(vq_CFObject *objPtr, char *filename)
{
  objPtr->debugFile = fopen(filename, "rb");
  if (objPtr->debugFile == NULL) {
    Tcl_AppendResult(objPtr->interp, "Error: couldn't open debug output file '",
		     filename, "'", (char *) NULL);
    return TCL_ERROR;
  }

  /* todo */
}

/************************************************************
 *
 * vq_CF_OpenClipfile
 *
 * Parameters:
 *     objPtr - pointer to object
 *     filename - file name
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Opens the clipfile, reads basic info data.
 *     Closes file when done.
 ************************************************************/

static int
vq_CF_OpenClipfile(vq_CFObject *objPtr, char *filename)
{
  tChunkHeader chunk_hdr;
  MediaIndepHeader mi_hdr;

  if (objPtr->chunkFileState != NULL) {
    Tcl_AppendResult(objPtr->interp, "Problem: chunk file already open", 
		     (char *) NULL);
    return TCL_ERROR;
  }

  /* 
   * Attempt to open the file specified as a chunk file (i.e. pre-formatted
   * clipfile).
   */

  objPtr->chunkFileState = Chunk_OpenFile(filename, "r");
  if (objPtr->chunkFileState == NULL) {
    Tcl_AppendResult(objPtr->interp, "Couldn't open file: ", filename,
		     " as a vq clipfile.", 
		     (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Get information from media independent header. 
   */

  if (Chunk_FindType(objPtr->chunkFileState, CM_MEDIA_INDEP_CF, &chunk_hdr) != 0) {
    Tcl_AppendResult(objPtr->interp,
		     "Couldn't find media independent header.", (char *) NULL);
    return TCL_ERROR;
  }

  read(objPtr->chunkFileState->fd, (char *) &mi_hdr, sizeof(mi_hdr));

  mi_hdr.media	    = ntohl(mi_hdr.media);
  mi_hdr.format	    = ntohl(mi_hdr.format);
  mi_hdr.device	    = ntohl(mi_hdr.device);
  mi_hdr.version    = ntohl(mi_hdr.version);
  mi_hdr.num_frames = ntohl(mi_hdr.num_frames);
  mi_hdr.length_s   = ntohl(mi_hdr.length_s);
  mi_hdr.length_ns  = ntohl(mi_hdr.length_ns);

  objPtr->numFrames = mi_hdr.num_frames;
  objPtr->fileLength = mi_hdr.length_s + mi_hdr.length_ns/ONE_BILLION;

  /*
   * Get information from vq clipfile header.
   */

  if (Chunk_FindType(objPtr->chunkFileState, CM_VQ_CF, &chunk_hdr) != 0) {
    Chunk_ReadRewind(objPtr->chunkFileState);
    if (Chunk_FindType(objPtr->chunkFileState, CM_VQ_CF, &chunk_hdr) != 0) {
      Tcl_AppendResult(objPtr->interp,
		       "Couldn't find VQ header.", (char *) NULL);
      return TCL_ERROR;
    }
  }

  if (objPtr->cfInfo != NULL) {
    ckfree((char *) objPtr->cfInfo);;
  }
  objPtr->cfInfo = (VQHeader *) ckalloc(sizeof(VQHeader));
  if (objPtr->cfInfo == NULL) {
    Tcl_AppendResult(objPtr->interp,
		     "Couldn't allocate clipfile info structure.", (char *) NULL);
    return TCL_ERROR;
  }

  read(objPtr->chunkFileState->fd, (char *) objPtr->cfInfo, sizeof(VQHeader));

  objPtr->cfInfo->width	       = ntohl(objPtr->cfInfo->width);
  objPtr->cfInfo->height       = ntohl(objPtr->cfInfo->height);
  objPtr->cfInfo->numCodebooks = ntohl(objPtr->cfInfo->numCodebooks);
  objPtr->cfInfo->device       = ntohl(objPtr->cfInfo->device);
  objPtr->cfInfo->subformat    = ntohl(objPtr->cfInfo->subformat);

  /*
   * Find data segment and store location in file in object.
   */

  if (Chunk_FindType(objPtr->chunkFileState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
    Chunk_ReadRewind(objPtr->chunkFileState);
    if (Chunk_FindType(objPtr->chunkFileState, CM_DATA_SEGMENT, &chunk_hdr) != 0) {
      Tcl_AppendResult(objPtr->interp, "Couldn't find data segment.",
		       (char *) NULL);
      return TCL_ERROR;
    }
  }

  objPtr->dataStart = lseek(objPtr->chunkFileState->fd, 0, SEEK_CUR);

  objPtr->codebooks = ReadCodebooks(objPtr->interp,
				    objPtr->chunkFileState,
				    objPtr->cfInfo->numCodebooks);
  if (objPtr->codebooks == NULL) {
      Tcl_AppendResult(objPtr->interp, "Error reading codebooks.",
		       (char *) NULL);
      return TCL_ERROR;
  }
    
  Chunk_CloseFile(objPtr->chunkFileState);
  objPtr->chunkFileState = NULL;
  return TCL_OK;
}

/************************************************************
 *
 * vq_CF_ChangeOutCmd
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes out command.
 ************************************************************/

static int
vq_CF_ChangeOutCmd(vq_CFObject *objPtr, char *newOut) 
{

  /*
   * Clear the previous entry.
   */
  if (objPtr->outCommand != NULL) {
    ckfree(objPtr->outCommand);
  }
  objPtr->outCCmd = NULL;
  objPtr->outCD = NULL;

  /*
   * If new outCmd is empty, return.
   */
  if (strlen(newOut) == 0) {
    return TCL_OK;
  }

  /*
   * If new outCmd is C proc, get them as atoms and set everything
   * accordingly. 
   */
  if (newOut[0] == '@') {
    if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
				   &(objPtr->outCD),
				   (ClientData) &(objPtr->outCCmd)) != TCL_OK) {
      objPtr->outCD = NULL;
      objPtr->outCCmd = NULL;
      Tcl_AppendResult(objPtr->interp,
		       "Couldn't find registered C procedure: ", newOut,
		       (char *)NULL);
      return TCL_ERROR;
    }
  }

  /*
   * Copy new out command into outCmd string.
   */
  objPtr->outCommand = (char *) ckalloc(strlen(newOut)+1);
  if (objPtr->outCommand == NULL) {
    Tcl_AppendResult(objPtr->interp,
		     "Couldn't alloc space for command name\n",
		     (char *)NULL);
    return TCL_ERROR;
  }
		     
  strcpy(objPtr->outCommand, newOut);
  return TCL_OK;
}    

/************************************************************
 *
 * vq_CF_ChangeCodebookOutCmd
 *
 * Parameters:
 *	objPtr	pointer to object
 *	newOut	new Codebook out command
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     changes codebook out command.
 ************************************************************/

static int
vq_CF_ChangeCodebookOutCmd(vq_CFObject *objPtr, char *newOut) 
{
  /*
   * Clear the previous entry.
   */
  if (objPtr->codebookOutCmd != NULL) {
    ckfree(objPtr->codebookOutCmd);
  }
  objPtr->codebookOutCCmd = NULL;
  objPtr->codebookOutCD = NULL;

  /*
   * If new outCmd is empty, return.
   */
  if (strlen(newOut) == 0) {
    return TCL_OK;
  }

  /*
   * If new outCmd is C proc, get them as atoms and set everything
   * accordingly. 
   */
  if (newOut[0] == '@') {
    if (Tutil_GetRegisteredCommand(objPtr->interp, newOut,
				   &(objPtr->codebookOutCD),
				   (ClientData) &(objPtr->codebookOutCCmd)) != TCL_OK) {
      objPtr->codebookOutCD = NULL;
      objPtr->codebookOutCCmd = NULL;
      Tcl_AppendResult(objPtr->interp,
		       "Couldn't find registered C procedure: ", newOut,
		       (char *)NULL);
      return TCL_ERROR;
    }
  }

  /*
   * Copy new out command into outCmd string.
   */
  objPtr->codebookOutCmd = (char *) ckalloc(strlen(newOut)+1);
  if (objPtr->codebookOutCmd == NULL) {
    Tcl_AppendResult(objPtr->interp,
		     "Couldn't alloc space for command name\n",
		     (char *)NULL);
    return TCL_ERROR;
  }
		     
  strcpy(objPtr->codebookOutCmd, newOut);
  return TCL_OK;
}

/************************************************************
 *
 * vq_CF_Unready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Stops sending data and puts the object in the unready state.
 ************************************************************/

static int
vq_CF_Unready(vq_CFObject * objPtr)
{
  objPtr->isReady = 0;
  return TCL_OK;
}

/************************************************************
 *
 * Configurevq_CF
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

int
Configurevq_CF(vq_CFObject * objPtr, int argc, char **argv)
{
  
  Tcl_Interp *interp;
  int cntr;
  char str[50];
    
  interp = objPtr->interp;
  
  /*
   * Case 1: "<obj> configure"
   */
  if (argc == 0) {
    if (objPtr->ltsName == NULL) {
      Tcl_AppendResult(interp, "{-lts {} {}} ", (char *) NULL);
    } else {
      Tcl_AppendResult(interp, "{-lts {} {", objPtr->ltsName, "}} ",
		       (char *) NULL);
    }
    Tcl_PrintDouble(interp, objPtr->sendAhead, str);
    Tcl_AppendResult(interp, "{-sendAhead {0.2} ", str, "} ", (char *)NULL);

    sprintf(str, "{%d}", objPtr->bandwidth);
    Tcl_AppendResult(interp, "{-bandwidth {0} ", str, "} ", (char *)NULL);
    
    Tcl_PrintDouble(interp, objPtr->cycleTime, str);
    Tcl_AppendResult(interp, "{-cycleTime {0.0} ", str, "} ", (char *)NULL);

    if (objPtr->fileName == NULL) {
      Tcl_AppendResult(interp, "{-filename {} {}} ", (char *)NULL);
    } else {
      Tcl_AppendResult(interp, "{-filename {} {", objPtr->fileName,
		       "}} ",(char *)NULL);
    }
	
    if (objPtr->outCommand == NULL) {
      Tcl_AppendResult(interp, "{-outCmd {} {}} ", (char *)NULL);
    } else {
      Tcl_AppendResult(interp, "{-outCmd {} {", objPtr->outCommand,
		       "}} ", (char *)NULL);
    }

    if (objPtr->codebookOutCmd == NULL) {
      Tcl_AppendResult(interp, "{-codebookOutCmd {} {}} ", (char *)NULL);
    } else {
      Tcl_AppendResult(interp, "{-codebookOutCmd {} {",
		       objPtr->codebookOutCmd, "}} ", (char *)NULL);
    }

    Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
    Tcl_AppendResult(interp, "{-maxFrameRate {0.0} ", str,
		     "} ", (char *)NULL);

    return TCL_OK;
  }

  /*
   * Case 2: "<obj> configure -slot"
   */
  if (argc == 1) {
    char c;
    int len;
	
    len = strlen(argv[0]);
    c = argv[0][1];
	
    if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
      if (objPtr->lts == NULL) {
	Tcl_AppendResult(interp, "-lts {} {}", (char *)NULL);
      } else {
	Tcl_AppendResult(interp, "-lts {} {", objPtr->ltsName, "} ",
			 (char *)NULL);
      }
      return TCL_OK;
    } 
    if ((c == 's') && (strncmp("-sendAhead", argv[0], len) == 0)) {
      Tcl_PrintDouble(interp, objPtr->sendAhead, str);
      Tcl_AppendResult(interp, "-sendAhead {0.2} ", str, (char *)NULL);
      return TCL_OK;
    } 
    if ((c == 'b') && (strncmp("-bandwidth", argv[0], len) == 0)) {
      sprintf(str, "{%d}", objPtr->bandwidth);
      Tcl_AppendResult(interp, "-bandwidth {0} ", str, (char *)NULL);
      return TCL_OK;
    } 
    if ((c == 'c') && (strncmp("-cycleTime", argv[0], len) == 0)) {
      Tcl_PrintDouble(interp, objPtr->cycleTime, str);
      Tcl_AppendResult(interp, "-cycleTime {0.0} ", str, (char *)NULL);
      return TCL_OK;
    } 
    if ((c == 'f') && (strncmp("-filename", argv[0], len) == 0)) {
      if (objPtr->fileName == NULL) {
	Tcl_AppendResult(interp, "-filename {} {}", (char *)NULL);
      } else {
	Tcl_AppendResult(interp, "-filename {} {",
			 objPtr->fileName, "} ", (char *)NULL);
      }
      return TCL_OK;
    } 
    if ((c == 'o') && (strncmp("-outCmd", argv[0], len) == 0)) {
      if (objPtr->outCommand == NULL) {
	Tcl_AppendResult(interp, "-outCmd {} {}", (char *)NULL);
      } else {
	Tcl_AppendResult(interp, "-outCmd {} {", objPtr->outCommand,
			 "} ", (char *)NULL);
      }
      return TCL_OK;
    } 
    if ((c == 'c') && (strncmp("-codebookOutCmd", argv[0], len) == 0)) {
      if (objPtr->outCommand == NULL) {
	Tcl_AppendResult(interp, "-codebookOutCmd {} {}", (char *)NULL);
      } else {
	Tcl_AppendResult(interp, "-codebookOutCmd {} {", objPtr->outCommand,
			 "} ", (char *)NULL);
      }
      return TCL_OK;
    } 
    if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
      Tcl_PrintDouble(interp, objPtr->maxFrameRate, str);
      Tcl_AppendResult(interp, "-maxFrameRate {0.0} ", str, (char *)NULL);
      return TCL_OK;
    }
    Tcl_AppendResult(interp, "bad configuration option: ",
		     argv[0], (char *)NULL);
    return TCL_ERROR;
  }

  /*
   * Case 3: "<obj> configure -slot value"
   * Since values may change, lots of side effects are buried in here.
   */
  cntr = 0;
  while (cntr < argc) {
    char c;
    int len;
	
    len = strlen(argv[cntr]);
    c = argv[cntr][1];
    /* ------------------------- LTS ------------------------- */
    if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
      ClientData newlts;
	    
      /*
       * Check args
       */
      cntr++;
      if (cntr == argc) {
      argcError:
	Tcl_AppendResult(interp, "Must specify value when ",
			 "trying to set configuration option ",
			 argv[cntr], (char *)NULL);
	return TCL_ERROR;
      }

      /*
       * If resetting to null value, just free up old lts.
       */
      if (strlen(argv[cntr]) == 0) {
	if (objPtr->lts != NULL) {
	  if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	  }
	  Tk_Release(objPtr->lts);
	  objPtr->lts = NULL;
	}
	if (objPtr->ltsName != NULL) {
	  ckfree(objPtr->ltsName);
	  objPtr->ltsName = NULL;
	}
      } else {
	newlts = (ClientData)Tcm_FindAtom(interp, argv[cntr]);
	if (newlts == NULL) {
	valueError:	    
	  Tcl_AppendResult(interp, "bad value \"", argv[cntr],
			   "\" given for configuration option ",
			   argv[cntr-1], (char *)NULL);
	  return TCL_ERROR;
	}
		
	/*
	 * Blow away old callback.
	 */
	if (objPtr->lts != NULL) {
	  if (objPtr->ltsTriggerNum != -1) {
	    LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	    objPtr->ltsTriggerNum = -1;
	  }
	  Tk_Release(objPtr->lts);
	  objPtr->lts = NULL;
	}
	objPtr->lts = newlts;
	Tk_Preserve (objPtr->lts);

	/*
	 * Copy in new lts name
	 */
	if (objPtr->ltsName != NULL) {
	  ckfree(objPtr->ltsName);
	}
	objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
	strcpy(objPtr->ltsName, argv[cntr]);
		
	/*
	 * Set up new lts trigger
	 */
	objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts,
					       LTS_SPEED_FLAG | LTS_VALUE_FLAG,
					       vq_CF_LTSChanged, objPtr);
      }
      
      /* ------------------------- SEND_AHEAD ------------------------- */
    } else if ((c=='s') && (strncmp("-sendAhead", argv[cntr], len) == 0)) {
      double newSendAhead;
      
      cntr++;
      if (cntr == argc) {
	goto argcError;
      }
      if (Tcl_GetDouble(interp, argv[cntr], &newSendAhead) != TCL_OK) {
	goto valueError;
      }
      objPtr->sendAhead = newSendAhead;
      
      /* ------------------------- BANDWIDTH -------------------------- */
    } else if ((c=='b') && (strncmp("-bandwidth", argv[cntr], len) == 0)) {
      int newBandwidth;
      
      cntr++;
      if (cntr == argc) {
	goto argcError;
      }
      if (Tcl_GetInt(interp, argv[cntr], &newBandwidth) != TCL_OK) {
	goto valueError;
      }
      objPtr->bandwidth = newBandwidth;
      
      /* ------------------------- CYCLE_TIME ------------------------- */
    } else if ((c=='c') && (strncmp("-cycleTime", argv[cntr], len) == 0)) {
      double newCycleTime;
	    
      cntr++;
      if (cntr == argc) {
	goto argcError;
      }
      if (Tcl_GetDouble(interp, argv[cntr], &newCycleTime) != TCL_OK) {
	goto valueError;
      }
      objPtr->cycleTime = newCycleTime;

      /* ------------------------- FILE_NAME -------------------------- */
    } else if ((c=='f') && (strncmp("-filename", argv[cntr], len) == 0)) {
      cntr++;
      if (cntr == argc) {
	goto argcError;
      }
      if (objPtr->fileName != NULL) {
	Tcl_AppendResult(interp, "Can not set clipfile name more than once.",
			 (char *)NULL);
	return TCL_ERROR;
      }
	    
      if (vq_CF_OpenClipfile(objPtr, argv[cntr]) != TCL_OK) {
	return TCL_ERROR;
      }
	    
      objPtr->fileName = (char *) ckalloc(strlen(argv[cntr]) + 1);
      strcpy(objPtr->fileName, argv[cntr]);

      /* ------------------------- OUT_CMD ---------------------------- */
    } else if ((c == 'o') && (strncmp("-outCmd", argv[cntr], len) == 0)) {
      cntr++;
      if (cntr == argc) {
	goto argcError;
      }
      
      if (vq_CF_ChangeOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
	return TCL_ERROR;
      }
      
      /* ------------------------- CODEBOOK_OUT_CMD ---------------------------- */
    } else if ((c == 'c') && (strncmp("-codebookOutCmd", argv[cntr], len) == 0)) {
      cntr++;
      if (cntr == argc) {
	goto argcError;
      }
      
      if (vq_CF_ChangeCodebookOutCmd(objPtr, argv[cntr]) == TCL_ERROR) {
	return TCL_ERROR;
      }
      
      /* ------------------------- MAX_FRAME_RATE --------------------- */
    } else if ((c=='m') &&
	       (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
      double newMaxFrameRate;
      
      cntr++;
      if (cntr == argc) {
	goto argcError;
      }
      if (Tcl_GetDouble(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
	goto valueError;
      }
      objPtr->maxFrameRate = newMaxFrameRate;
      
      /* ------------------------- Debug file -------------------------- */
    } else if ((c=='d') && (strncmp("-debugFile", argv[cntr], len) == 0)) {
      cntr++;
      if (cntr == argc) {
	goto argcError;
      }
      if (objPtr->debugFileName != NULL) {
	Tcl_AppendResult(interp, "Can not set debug file name more than once.",
			 (char *)NULL);
	return TCL_ERROR;
      }
	    
      if (vq_CF_OpenDebugfile(objPtr, argv[cntr]) != TCL_OK) {
	return TCL_ERROR;
      }
	    
      objPtr->debugFileName = (char *) ckalloc(strlen(argv[cntr]) + 1);
      strcpy(objPtr->debugFileName, argv[cntr]);

      /* ------------------------- UNKNOWN OPTION ----------------------*/
    } else {
      Tcl_AppendResult(interp, "Error, configuration option: ",
		       argv[cntr], " unknown.", (char *)NULL);
      return TCL_ERROR;
    }
    cntr++;
  }
  return vq_CF_Unready(objPtr);
}

static double
GetTimeFromFile(int fd)
{
  int sec, nsec;
  double res;

  read(fd, (char *) &sec, 4);
  read(fd, (char *) &nsec, 4);
    
  sec = ntohl(sec);
  nsec = ntohl(nsec);

  res = ((double) sec) + (((double) nsec) / ONE_BILLION);
    
  return res;
}

static frameOffset
GetOffsetFromFile(int fd)
{
  frameOffset result;
  int tstart, tlength;

  read(fd, (char *) &tstart, 4);
  read(fd, (char *) &tlength, 4);
    
  result.start = ntohl(tstart);
  result.length = ntohl(tlength);

  return result;
}

/************************************************************
 *
 * vq_CF_AppendSegment
 *
 * Parameters:
 *     objPtr - pointer to object
 *     clipStart -
 *     clipEnd -
 *     logStart -
 *     logEnd -
 *     logStartFlag - indicates whetehr logStart value set or not.
 *     logEndFlag - indicates whether logEnd value set or not.
 *	Note: make logStart > logEnd to play backwards
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
vq_CF_AppendSegment(vq_CFObject *objPtr, double clipStart, double clipEnd, 
		    double logStart, double logEnd, 
		    int logStartFlag, int logEndFlag)
{
  Segment	*newSeg;
  char		retStr[32];
  double	clipDuration;
  tChunkHeader	chunk_hdr;
  Segment	*nextSeg, *prevSeg;
  tChunkHeader	fttHeader;
  int		frameStart, frameEnd;
  double	this_time, next_time;
  int		count, i, maxSize;
  double	scale;
  
  if (clipStart >= clipEnd) {
    Tcl_AppendResult(objPtr->interp, "Clip start can not be after (or equal to) clip end.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  if (objPtr->fileName == NULL) {
    Tcl_AppendResult(objPtr->interp, "Must specify clipfile first.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  if (clipStart < 0.0)
    clipStart = 0.0;
  if (clipStart > objPtr->fileLength) {
    Tcl_AppendResult(objPtr->interp, "Clip start too large (past EOF).", (char *) NULL);
    return TCL_ERROR;
  }
  if (clipEnd > objPtr->fileLength)
    clipEnd = objPtr->fileLength;
  if (clipEnd < 0.0) {
    Tcl_AppendResult(objPtr->interp, "Clip end too small.", (char *) NULL);
    return TCL_ERROR;
  }
  
  newSeg = (Segment *) ckalloc(sizeof(Segment));
  if (newSeg == NULL) {
    Tcl_AppendResult(objPtr->interp, "Can't allocate new segment", (char *) NULL);
    return TCL_ERROR;
  }

  newSeg->FOT = NULL;
  newSeg->FTT = NULL;
  newSeg->FDT = NULL;
    
  newSeg->clipStart = clipStart;
  newSeg->clipEnd = clipEnd;
  
  clipDuration = clipEnd - clipStart;
  
  if (clipDuration <= 0.0) {
    ckfree((char *) newSeg);
    Tcl_AppendResult(objPtr->interp, "Clip segment duration too small.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  
  /*
   * Do both start and end specified.
   */
  if (logStartFlag && logEndFlag) {
    /*
     * Set logStart and logEnd.
     */
    if (logStart > logEnd) {
      newSeg->logicalStart = logEnd;
      newSeg->logicalEnd = logStart;
    }
    else {
      newSeg->logicalStart = logStart;
      newSeg->logicalEnd = logEnd;
    }
  }
  else if (logStartFlag) {
    newSeg->logicalStart = logStart;
    logEnd = newSeg->logicalEnd = logStart+clipDuration;
  }
  else if (logEndFlag) {
    newSeg->logicalEnd = logEnd;
    logStart = newSeg->logicalStart = logEnd - clipDuration;
  }
  else {
    logStart = newSeg->logicalStart = 0.0;
    logEnd = newSeg->logicalEnd = clipDuration;
  }
  
  
  /*
   * Check for zero durations.
   */
  if (newSeg->logicalStart - newSeg->logicalEnd == 0.0) {
    Tcl_AppendResult(objPtr->interp, "Zero length logical time.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  if (newSeg->clipStart - newSeg->clipEnd == 0.0) {
    Tcl_AppendResult(objPtr->interp, "Zero length clip time.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  /*
   * Check for overlap with other segments, setting nextSeg and
   * prevSeg. Return error if overlap.
   */
  nextSeg = objPtr->segments;
  prevSeg = NULL;
  
  while ((nextSeg != NULL) &&
	 (nextSeg->logicalStart < newSeg->logicalEnd)) {
    prevSeg = nextSeg;
    nextSeg = nextSeg->next;
  }
  
  if ((prevSeg != NULL) &&				/* not at the beginning */
      (prevSeg->logicalEnd > newSeg->logicalStart)) {
    Tcl_AppendResult(objPtr->interp, "Error: new segment overlaps with previous one.",
		     (char *) NULL);
    return TCL_ERROR;
  }
  
  
  /* 
   * Create frame offset and time tables for this segment.
   * This means skipping frames until the correct beginning and
   * then finding correct end. Then generating segment FOT and FTT
   * for these frames with time scaled properly. 
   */
  
  if (objPtr->chunkFileState != NULL) {
    Chunk_ReadRewind(objPtr->chunkFileState);
  } else {
    objPtr->chunkFileState = Chunk_OpenFile(objPtr->fileName, "r");
    if (objPtr->chunkFileState == NULL) {
      Tcl_AppendResult(objPtr->interp, "Couldn't open file '", objPtr->fileName,
		       "' for reading.", (char *) NULL);
      return TCL_ERROR;
    }
  }
    
  if (Chunk_FindType(objPtr->chunkFileState, CM_FTT, &fttHeader)) {
    ckfree((char *) newSeg);
    Tcl_AppendResult(objPtr->interp,
		     "Problem reading clipfile frame time table.", (char *) NULL);
    return TCL_ERROR;
  }
    
  frameStart = 0;
  this_time = GetTimeFromFile(objPtr->chunkFileState->fd);
  next_time = GetTimeFromFile(objPtr->chunkFileState->fd);
  count = 0;
  while ((this_time < newSeg->clipStart) &&
	 (next_time <= newSeg->clipStart)) {
    this_time = next_time;
    next_time = GetTimeFromFile(objPtr->chunkFileState->fd);
    count++;
    if (count == objPtr->numFrames) {
      Tcl_AppendResult(objPtr->interp,
		       "Error constructing segment frame time table.", (char *) NULL);
      return TCL_ERROR;
    }
  }
    
  frameStart = count;
    
  while ((this_time < newSeg->clipEnd) && (next_time < newSeg->clipEnd)) {
    this_time = next_time;
    next_time = GetTimeFromFile(objPtr->chunkFileState->fd);
    count++;
    if (count == objPtr->numFrames) {
      Tcl_AppendResult(objPtr->interp,
		       "Error constructing segment frame time table.", (char *) NULL);
      return TCL_ERROR;
    }
  }
    
  frameEnd = count;
  newSeg->numFrames = frameEnd - frameStart + 1;
    
  /*
     * Read current FTT values.
     */
  newSeg->FTT = (double *) ckalloc((newSeg->numFrames+1)*sizeof(double));
  if (newSeg->FTT == NULL) {
    Tcl_AppendResult(objPtr->interp,
		     "Couldn't allocate memory for frame time table.", (char *)NULL);
    goto error;
  }
    
  Chunk_ReadRewind(objPtr->chunkFileState);
  if (Chunk_FindType(objPtr->chunkFileState, CM_FTT, &chunk_hdr)) {
    if (newSeg->FTT != NULL) {
      ckfree((char *) newSeg->FTT);
    }
    ckfree((char *) newSeg);
    Tcl_AppendResult(objPtr->interp,
		     "Error constructing segment frame time table.", (char *) NULL);
    return TCL_ERROR;
  }
    
  if (frameStart != 0) {
    lseek(objPtr->chunkFileState->fd, frameStart * 8, SEEK_CUR);
  }
    
  for (i=0; i<newSeg->numFrames; i++) {
    newSeg->FTT[i] = GetTimeFromFile(objPtr->chunkFileState->fd);
  }
    
  scale = (newSeg->logicalEnd - newSeg->logicalStart) / 
    (newSeg->clipEnd - newSeg->clipStart);
    
  newSeg->FTT[0] = newSeg->logicalStart;
    
  for (i=1; i<newSeg->numFrames; i++) {
    newSeg->FTT[i] = scale * (newSeg->FTT[i]-newSeg->clipStart) +
      newSeg->logicalStart;
  }
  newSeg->FTT[i] = newSeg->logicalEnd;
    
  /*
   * Construct segment frame offset table.
   */
  Chunk_ReadRewind(objPtr->chunkFileState);
  if (Chunk_FindType(objPtr->chunkFileState, CM_FOT, &chunk_hdr)) {
    if (newSeg->FTT != NULL) {
      ckfree((char *) newSeg->FTT);
    }
    ckfree((char *) newSeg);
    Tcl_AppendResult(objPtr->interp,
		     "Error constructing segment frame offset table.", (char *)NULL);
    return TCL_ERROR;
  }
    
  newSeg->FOT = (frameOffset *) ckalloc(sizeof(frameOffset)*newSeg->numFrames);
  if (newSeg->FOT == NULL) {
    Tcl_AppendResult(objPtr->interp,
		     "Couldn't allocate memory for frame offset table.", (char *)NULL);
    goto error;
  }
    
  if (frameStart != 0) {
    lseek(objPtr->chunkFileState->fd, frameStart * sizeof(frameOffset), SEEK_CUR);
  }
    
  if (logStart < logEnd) {
    for (i=0; i<newSeg->numFrames; i++) {
      newSeg->FOT[i] = GetOffsetFromFile(objPtr->chunkFileState->fd);
    }
  }
  else {
    for (i=newSeg->numFrames-1; i>=0; i--) {
      newSeg->FOT[i] = GetOffsetFromFile(objPtr->chunkFileState->fd);
    }
  }

  /*
   *	Read frame dependency table for this segment
   */ 
  if (Chunk_ReallyFindType(objPtr->chunkFileState, CM_VQ_FDT, &chunk_hdr) != 0) {
    Tcl_AppendResult(objPtr->interp,
		     "Couldn't find segment frame dependency info.", (char *)NULL);
    goto error;
  }
  newSeg->FDT = (int *) ckalloc(sizeof(int) * newSeg->numFrames);
  if (newSeg->FDT == NULL) {
    Tcl_AppendResult(objPtr->interp,
		     "Couldn't allocate memory for frame dependency info.", (char *)NULL);
    goto error;
  }

  if (frameStart != 0) {
    lseek(objPtr->chunkFileState->fd, frameStart * sizeof(int), SEEK_CUR);
  }

  for (ITERATOR(i, logStart < logEnd, 0, newSeg->numFrames)) {
    read(objPtr->chunkFileState->fd, newSeg->FDT+i, sizeof(int));
  }

  /*
   * Set segment id.
   */
  newSeg->segId = objPtr->nextSegId++;
  
  /*
   * Find max frame size.
   */
  maxSize = -1;
  for (i=0; i<newSeg->numFrames; i++) {
    if (newSeg->FOT[i].length > maxSize) {
      maxSize = newSeg->FOT[i].length;
    }
  }
  newSeg->maxSize = maxSize;
  
  /*
   * Set up next and prev pointers and return.
   */
  newSeg->next = nextSeg;
  newSeg->prev = prevSeg;
  if (prevSeg == NULL) {
    objPtr->segments = newSeg;
  }
  else {
    prevSeg->next = newSeg;
  }
  if (nextSeg != NULL) {
    nextSeg->prev = newSeg;
  }
  
  sprintf(retStr, "%d", newSeg->segId);
  Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);
  
  return TCL_OK;

 error:
  if (newSeg != NULL) {
    if (newSeg->FTT != NULL) {
      ckfree((char *) newSeg->FTT);
    }
    if (newSeg->FOT != NULL) {
      ckfree((char *) newSeg->FOT);
    }
    if (newSeg->FDT != NULL) {
      ckfree((char *) newSeg->FDT);
    }
    ckfree((char *) newSeg);
  }

  return TCL_ERROR;
}

/************************************************************
 *
 * vq_CF_SegmentInfo
 *
 * Parameters:
 *     objPtr - pointer to object
 *     segId -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sets the Tcl result string to {clipStart clipEnd logStart logEnd 
 *                                    numFrames maxSize} 
 *
 ************************************************************/

static int
vq_CF_SegmentInfo(vq_CFObject * objPtr, int segId)
{
  Segment *segPtr;
  char retStr[256];

  segPtr = objPtr->segments;

  while (segPtr != NULL) {
    if (segPtr->segId == segId) {
      break;
    }
    segPtr = segPtr->next;
  }

  if (segPtr == NULL) {
    Tcl_AppendResult(objPtr->interp, "Segment doesn't exist.",
		     (char *) NULL);
    return TCL_ERROR;
  }

  sprintf(retStr, "%g %g %g %g %d %d", segPtr->clipStart, segPtr->clipEnd,
	  segPtr->logicalStart, segPtr->logicalEnd,
	  segPtr->numFrames, segPtr->maxSize);

  Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);
  return TCL_OK;
}

/************************************************************
 *
 * vq_CF_Info
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *     Sets return string to {numFrames clipLength maxSize width height}
 *     where maxSize is maximum size of segments so far.
 *
 ************************************************************/
static int
vq_CF_Info(vq_CFObject * objPtr)
{
  Segment *segPtr;
  int maxSize;
  char retStr[256];

  segPtr = objPtr->segments;
  maxSize = 0;

  while(segPtr != NULL) {
    if (segPtr->maxSize > maxSize)
      maxSize = segPtr->maxSize;
    segPtr = segPtr->next;
  }

  if (objPtr->cfInfo != NULL) {
    sprintf(retStr, "%d %g %d %d %d", objPtr->numFrames,
	    objPtr->fileLength, maxSize,
	    objPtr->cfInfo->width, objPtr->cfInfo->height);
  }
  else {
    sprintf(retStr, "%d %g %d 0 0", objPtr->numFrames,
	    objPtr->fileLength, maxSize);
  }

  Tcl_AppendResult(objPtr->interp, retStr, (char *) NULL);
  return TCL_OK;
}

/************************************************************
 *
 * vq_CFObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     vq_CF object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
vq_CFObjCmd(ClientData clientData, Tcl_Interp * interp,
	    int argc, char **argv)
{
  char c;
  int len, i;
  vq_CFObject *objPtr = (vq_CFObject *) clientData;

  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		     argv[0], " <command> ?arguments?", (char *) NULL);
    goto usage;
  }
  c = argv[1][0];
  len = strlen(argv[1]);

  /*------------------------- CONFIGURE ------------------------------*/
  if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
    return Configurevq_CF(clientData, argc - 2, argv + 2);

    /*------------------------- DESTROY --------------------------------*/
  } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
    return Destroyvq_CF(clientData);

    /*------------------------- ADD_SEGMENT -------------------------*/
  } else if ((c == 'a') && (strncmp("addSegment", argv[1], len) == 0)) {
    double clipStart;
    double clipEnd;
    double logStart = 0.0;
    double logEnd = 0.0;
    int logStartFlag = 0;
    int logEndFlag = 0;

    if (argc < 4) {
      Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
		       " addSegment <clipStart> <clipEnd> ",
		       "?-start <logicalStart>? ?-end <logicalEnd>? ",
		       (char *) NULL);
      return TCL_ERROR;
    }

    if (Tcl_GetDouble(interp, argv[2], &clipStart) != TCL_OK) {
      Tcl_AppendResult(interp,
		       "error found while parsing clip start time ", (char *)NULL);
      return TCL_ERROR;
    }

    if (strcmp(argv[3], "end") == 0) {
      clipEnd = HUGE_VAL;
    } else if (Tcl_GetDouble(interp, argv[3], &clipEnd) != TCL_OK) {
      Tcl_AppendResult(interp,
		       "error found while parsing clip end time ", (char *)NULL);
      return TCL_ERROR;
    }

    for (i = 4; i < argc; i++) {
      char ac;
      int alen;

      ac = argv[i][1];
      alen = strlen(argv[i]);

      if ((ac == 's') && (strncmp("-start", argv[i], alen) == 0)) {
	i++;
	if (i == argc) {
	  Tcl_AppendResult(interp, "Must specify parameter value ",
			   "after optional argument flag ", argv[i],
			   " passed to ", argv[0], " ", argv[1],
			   (char *) NULL);
	  return TCL_ERROR;
	}
	if (Tcl_GetDouble(interp, argv[i], &logStart) != TCL_OK) {
	  Tcl_AppendResult(interp,
			   "error found while logical start end time ",
			   (char *)NULL);
	  return TCL_ERROR;
	}
	logStartFlag = 1;
      } else if ((ac == 'e') && (strncmp("-end", argv[i], alen) == 0)) {
	i++;
	if (i == argc) {
	  Tcl_AppendResult(interp, "Must specify parameter value ",
			   "after optional argument flag ", argv[i],
			   " passed to ", argv[0], " ", argv[1],
			   (char *) NULL);
	  return TCL_ERROR;
	}
	if (Tcl_GetDouble(interp, argv[i], &logEnd) != TCL_OK) {
	  Tcl_AppendResult(interp,
			   "error found while logical end end time ",
			   (char *)NULL);
	  return TCL_ERROR;
	}
	logEndFlag = 1;
      }
    }
    return vq_CF_AppendSegment(objPtr, clipStart, clipEnd,
			       logStart, logEnd, logStartFlag, logEndFlag);

    /*------------------------- DELETE_SEGMENT -------------------------*/
  } else if ((c == 'd') && (strncmp("deleteSegment", argv[1], len) == 0)) {
    int segId;

    if (argc < 3) {
      Tcl_AppendResult(interp, "wrong # of args, should be: ",
		       argv[0], " deleteSegment <segmentId> ", (char *) NULL);
      return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[2], &segId) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad integer parameter value ",
		       argv[2], " given to: ", argv[0], " ", argv[1],
		       (char *) NULL);
      return TCL_ERROR;
    }
    return vq_CF_DeleteSegment(objPtr, segId);


    /*------------------------- CLEAR_SEGMENTS -------------------------*/
  } else if ((c == 'c') && (strncmp("clearSegments", argv[1], len) == 0)) {
    return vq_CF_ClearSegments(objPtr);

    /*------------------------- SEGMENT_INFO ---------------------------*/
  } else if ((c == 's') && (strncmp("segmentInfo", argv[1], len) == 0)) {
    int segId;

    if (argc < 3) {
      Tcl_AppendResult(interp, "wrong # of args, should be: ",
		       argv[0], " segmentInfo <segmentId> ", (char *) NULL);
      return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[2], &segId) != TCL_OK) {
      Tcl_AppendResult(interp, "Bad integer parameter value ", argv[2],
		       " given to: ", argv[0], " ", argv[1], (char *) NULL);
      return TCL_ERROR;
    }
    return vq_CF_SegmentInfo(objPtr, segId);


    /*------------------------- INFO -----------------------------------*/
  } else if ((c == 'i') && (strncmp("info", argv[1], len) == 0)) {
    return vq_CF_Info(objPtr);

    /*------------------------- READY ----------------------------------*/
  } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
    return vq_CF_Ready(objPtr);

    /*------------------------- UNREADY --------------------------------*/
  } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
    return vq_CF_Unready(objPtr);
  }

  Tcl_AppendResult(interp, "Bad command option: ",
		   argv[0], " ", argv[1], (char *) NULL);
 usage:
  Tcl_AppendResult(interp,
		   "\n<command> can be one of:\n",
		   "  configure ?-slot ?value??\n",
		   "  destroy\n",
		   "  addSegment <clipStart> <clipEnd>",
		   " ?-start <logicalStart>? ?-end <logicalEnd>?\n",
		   "  deleteSegment <segmentId>\n",
		   "  clearSegments\n",
		   "  segmentInfo <segmentId>\n",
		   "  ready\n",
		   "  unready\n",
		   "  info\n",
		   (char *) NULL);

  return TCL_ERROR;
}

/************************************************************
 *
 * vq_CFCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the vqCF command.
 *     A new object of type vq_CF is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the object's methods.
 *
 ************************************************************/

int
vq_CFCreateCmd(ClientData clientData, Tcl_Interp * interp,
	       int argc, char **argv)
{
  char new_name[256];
  vq_CFObject *new;
    
  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		     argv[0], " parent ?options?", (char *) NULL);
    return TCL_ERROR;
  }
  nextvq_CFNum++;
  sprintf(new_name, "%svqc%d", argv[1], nextvq_CFNum);
    
  new = CreateNewvq_CF(interp, new_name);
  if (new == NULL) {
    Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
    return TCL_ERROR;
  }
  Tcl_CreateCommand(interp, new_name, vq_CFObjCmd,
		    (ClientData) new, (void (*)()) NULL);
    
  if (argc > 2) {
    Configurevq_CF(new, argc - 2, argv + 2);
  }
  Tcl_SetResult(interp, new_name, TCL_VOLATILE);
  return TCL_OK;
}

/************************************************************
 *
 * vq_CFModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
vq_CFModuleInit(Tcl_Interp * interp)
{
  Tcl_CreateCommand(interp, "vqFile", vq_CFCreateCmd, NULL,
			(void (*)()) NULL);
  if (Tcl_GlobalEval(interp, "cminfo add object_info vqCF VQ")
      == TCL_OK) {
    return TCL_OK;
  }
  else {
    return TCL_ERROR;
  }
}
