/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
#define DEBUG_FRAMES
*/
#include <sys/types.h>
#include <netinet/in.h>
#include <math.h>
#include <string.h>
#include "dp.h"
#include "tcmEvent.h"
#include "vqPlay.h"
#include "cmLTS.h"
#include "cmAtom.h"

extern int Tcm_DeleteEvery(Tcl_Interp *interp, char *name, void (*freeProc)());

#define ONE_BILLION 1000000000.0

static int nextVQResourceNum = 0;	/* Variable for next unique object
					 * name. */

/************************************************************
 *
 * VQ_RSRC_FreeFrame
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static void
VQ_RSRC_FreeFrame(char *clientData)
{
  VQFrame *framePtr = (VQFrame *)clientData;
  int i;

  /*
   * Call device frame freeer.
   */
  if (framePtr->objPtr->devicePtr->freeFrameProc != NULL) {
    framePtr->objPtr->devicePtr->freeFrameProc(framePtr);
  }

  /*
   * Detach from buffer.
   */

  for (i=0; i<framePtr->sb->count; i++) {
    DetachBuffer(framePtr->objPtr->interp, framePtr->sb->buffers[i],
		 "tcmVQResource:VQ_RSRC_FreeFrame");
  }

  FreeScatterBuffer(framePtr->sb);
  framePtr->sb = NULL;

  if (framePtr->at_token >= 0) {
    framePtr->objPtr->at_tokens[framePtr->at_token] = 0;
  }

  ckfree((char *) framePtr);
}

/************************************************************
 *
 * VQ_RSRC_Unready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
VQ_RSRC_Unready(VQResource *objPtr)
{
  int i;

  objPtr->ready = 0;

  /*
   * Blow away "at" queue.
   */
  for (i =0; i<objPtr->num_at_tokens; i++) {
    if (objPtr->at_tokens[i] > 0) {
      AtEvent *at_event;
      int token;

      token = objPtr->at_tokens[i];
      at_event = Tcm_FindAtEvent(token);
      if (token && (at_event != NULL)) {
	VQ_RSRC_FreeFrame(at_event->fireClientData);
	Tcm_DeleteAtHandler(token);
	objPtr->at_tokens[i] = 0;
      }
    }
  }

  /*
   * Blow away frame queue.
   */
  if (objPtr->queue != NULL) {
    DetachToAllBuffers(objPtr->interp, objPtr->queue,
		       "tcmVQResource:VQ_RSRC_Unready");
    FreeSBList(objPtr->queue);
    objPtr->queue = NULL;
  }

  return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * ChangeDeviceType --
 *
 *	Sets the device type for the VQReseource.  Called whenever
 *	the "devtype" slot is changed, and on initialization of the
 *	resource.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	The old device is shut down, if appropriate, and the new
 *	device is created.
 *
 * Parameters:
 *	VQResource *objPtr	Pointer to resource object.
 *	char *devtype		String giving name of new device
 *
 *--------------------------------------------------------------
 */
static void
ChangeDeviceType(VQResource *objPtr, char *devtype)
{
  int len;
  char c;

  len = strlen(devtype);
  c = devtype[0];

  /*
   * If there's no change, return
   */
  if ((objPtr->devtype != NULL) &&
      (strncmp(objPtr->devtype, devtype, len) == 0)) {
    return;
  }

  /*
   * Shut down old device, if necessary, ignoring any errors it
   * returns.
   */
  VQ_RSRC_Unready(objPtr);
  if (objPtr->devtype != NULL) {
    ckfree(objPtr->devtype);
    objPtr->devtype = NULL;
  }
  if (objPtr->devicePtr != NULL) {
    objPtr->devicePtr->shutdownProc(objPtr);
    ckfree((char *) objPtr->devicePtr);
    objPtr->devicePtr = NULL;
    objPtr->inited = 0;
  }

  /*
   * Create new device.
   */
  devtype = "default";
  len = strlen(devtype);
  objPtr->devicePtr = GetDefaultVQDevice(objPtr);

  objPtr->devtype = ckalloc(len+1);
  if (objPtr->devtype == NULL) {
    fprintf(stderr, "ERROR: couldn't alloc memory for device type at %s:%d",
	    __FILE__, __LINE__);
    return;
  }
  strcpy(objPtr->devtype, devtype);
}

/************************************************************
 *
 * VQ_RSRC_GetEligibleFrame
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static VQFrame *
VQ_RSRC_GetEligibleFrame(VQResource *objPtr)
{
  char		*sb_data;
  double	sTime, eTime;
  double	now, speed;
  int		i;
  VQFrame	*newFrame;
  int		idx, new_count;
  Buffer	*bufPtr;

  if (objPtr->queue == NULL)
    return NULL;

  if (objPtr->queue->count <= 0) {
    FreeSBList(objPtr->queue);
    objPtr->queue = NULL;
    return NULL;
  }

  now = LTS_Value(objPtr->lts);
  speed = LTS_Speed(objPtr->lts);

  newFrame = (VQFrame *) ckalloc(sizeof(VQFrame));

  if (newFrame == NULL) {
    fprintf(stderr, "ERROR: couldn't alloc memory for new frame at %s:%d",
	    __FILE__, __LINE__);
    return NULL;
  }

  for (i=0; i<objPtr->queue->count; i++) {
    sb_data = GetFromSB(objPtr->queue->sbs[i], CM_VQ_BUF_HDR, NULL, NULL);
    if (sb_data != NULL) {
      memcpy(&(newFrame->frameHeader), sb_data-8, sizeof(VQBufferHeader));
      
      newFrame->frameHeader.hdrType	= ntohl(newFrame->frameHeader.hdrType);
      newFrame->frameHeader.length	= ntohl(newFrame->frameHeader.length);
      newFrame->frameHeader.width	= ntohl(newFrame->frameHeader.width);
      newFrame->frameHeader.height	= ntohl(newFrame->frameHeader.height);
      newFrame->frameHeader.device	= ntohl(newFrame->frameHeader.device);
      newFrame->frameHeader.subformat	= ntohl(newFrame->frameHeader.subformat);
      newFrame->frameHeader.logEnd_s	= ntohl(newFrame->frameHeader.logEnd_s);
      newFrame->frameHeader.logEnd_ns	= ntohl(newFrame->frameHeader.logEnd_ns);
      newFrame->frameHeader.logStart_s	= ntohl(newFrame->frameHeader.logStart_s);
      newFrame->frameHeader.logStart_ns	= ntohl(newFrame->frameHeader.logStart_ns);
      
      eTime = ((double) newFrame->frameHeader.logEnd_s) + 
	(((double) newFrame->frameHeader.logEnd_ns) / ONE_BILLION);
      sTime = ((double )newFrame->frameHeader.logStart_s) + 
	(((double) newFrame->frameHeader.logStart_ns) / ONE_BILLION);
      
      newFrame->data = GetFromSB(objPtr->queue->sbs[i], CM_DATA_SEGMENT, &bufPtr,
				 &newFrame-dataLength);
      if (newFrame->data != NULL) {
	newFrame->dataBuffer = bufPtr;
	if (((speed < 0) && (eTime < now)) ||
	    ((speed > 0) && (eTime > now)) ||
	    (eTime == now)) {
	  break;
	}
      }
    }
  }

  /*
   * Gather stats.
   */

  objPtr->late += i;

  if (i == objPtr->queue->count) {
    ckfree((char *) newFrame);
    DetachToAllBuffers(objPtr->interp, objPtr->queue,
		       "tcmVQResource:VQ_RSRC_GetEligibleFrame");
    FreeSBList(objPtr->queue);
    objPtr->queue = NULL;
    return NULL;
  }

  idx = i;

  newFrame->logStart = sTime;
  newFrame->logEnd = eTime;
  newFrame->sb = objPtr->queue->sbs[idx];
  
  newFrame->data += 4;
  memcpy(&newFrame->dataLength, newFrame->data, 4);
  newFrame->dataLength = ntohl(newFrame->dataLength);
#ifdef DEBUG_FRAMES
  fprintf(stderr, "vqPlay: GetElegibleFrame: length=%10d\n",
	  newFrame->dataLength);
#endif
  newFrame->data += 4;

  newFrame->deviceData = objPtr->devicePtr->deviceData;
  newFrame->at_token = -1;
  newFrame->objPtr = objPtr;

  new_count = objPtr->queue->count - (idx+1);

  /*
   * Free scatter buffer structures and detach to buffers of frames skipped.
   */

  for (i=0; i<idx; i++) {
    int j;
    for (j=0; j<objPtr->queue->sbs[i]->count; j++) {
      DetachBuffer(objPtr->interp, objPtr->queue->sbs[i]->buffers[j],
		   "tcmVQResource:VQ_RSRC_GetEligibleFrame");
    }
    FreeScatterBuffer(objPtr->queue->sbs[i]);
    objPtr->queue->sbs[i] = NULL;
  }

  objPtr->queue->count = new_count;
  if (new_count == 0) {
    FreeSBList(objPtr->queue);
    objPtr->queue = NULL;
  }
  else {
    memcpy(objPtr->queue->sbs, objPtr->queue->sbs+(idx+1),
	   new_count*sizeof(ScatterBuffer *));
    objPtr->queue->sbs = (ScatterBuffer **) realloc((char *) objPtr->queue->sbs,
						    new_count*sizeof(ScatterBuffer *));
  }
   
  return newFrame;
}


/*
 *--------------------------------------------------------------
 *
 * VQ_RSRC_ContinueDecode --
 *
 *	Continues an asynchronous decode until it's finished.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	May schedule a frame for display if the decoding is finished,
 *	otherwise, schedules itself to be called again when processor
 *	is idle.
 *
 *--------------------------------------------------------------
 */
static void
VQ_RSRC_ContinueDecode (VQFrame *framePtr)
{
  VQResource *objPtr;

  objPtr = framePtr->objPtr;

  /* 
   * If not ready or already in decode, do nothing.
   */
  if ((!objPtr->ready) || (!objPtr->inDecode)) {
    return;
  }

  /*
   * Call device decode procedure.
   */
  if (objPtr->devicePtr != NULL) {
    if (objPtr->devicePtr->decodeProc != NULL) {
      objPtr->devicePtr->decodeProc(framePtr);
    }
  }

  /*
   * If decoding is not finished, schedule another call.  Otherwise
   * we're done.
   */
  if (objPtr->devicePtr->asyncDecodeFlag) {
    if (objPtr->devicePtr->decodeProc != NULL) {
      Tcm_DoWhenIdle("tk", (Tk_IdleProc *) VQ_RSRC_ContinueDecode,
		     framePtr);
    }
    return;
  }
    
  VQ_RSRC_DecodeDone(framePtr);
  return;
}

/************************************************************
 *
 * VQ_RSRC_DoDecode
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
int
VQ_RSRC_DoDecode(VQResource *objPtr)
{
  VQFrame *frame_ptr;

  /* 
   * If not ready or alraady in decode, do nothing.
   */
  if ((!objPtr->ready) || (objPtr->inDecode)) {
    return TCL_OK;
  }

  /*
   * Get next eligible frame.
   */
  frame_ptr = VQ_RSRC_GetEligibleFrame(objPtr);
  if (frame_ptr == NULL) {
    return TCL_OK;
  }

  /*
   * Record the fact that we're decoding a frame.
   */
  objPtr->inDecode = 1;

  /*
   * Start the decoding loop
   */
  VQ_RSRC_ContinueDecode(frame_ptr);
  return TCL_OK;
}

/************************************************************
 *
 * VQ_RSRC_Accept
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
VQ_RSRC_Accept(VQResource *objPtr, SB_List *sbl)
{
#ifdef DEBUG_FRAMES
  int i;
  double sTime, eTime;
  VQBufferHeader	header;
  char			*sb_data;
#endif

  if (!objPtr->ready) {
    return TCL_OK;
  }

  if ((sbl == NULL) || (sbl->count < 1)) {
    return TCL_OK;
  }

  /*
   * Increment reference count on buffers passed in
   */
  AttachToAllBuffers(objPtr->interp, sbl,
		     "tcmVQResource:VQ_RSRC_Accept");

#ifdef DEBUG_FRAMES
  /*
   * Print out frame info
   */
  
  for (i=0; i<sbl->count; i++) {
    sb_data = GetFromSB(sbl->sbs[i], CM_VQ_BUF_HDR, NULL, NULL);
    if (sb_data != NULL) {
      memcpy(&(header), sb_data-8, sizeof(VQBufferHeader));
      
      header.hdrType	= ntohl(header.hdrType);
      header.length	= ntohl(header.length);
      header.width	= ntohl(header.width);
      header.height	= ntohl(header.height);
      header.device	= ntohl(header.device);
      header.subformat	= ntohl(header.subformat);
      header.logEnd_s	= ntohl(header.logEnd_s);
      header.logEnd_ns	= ntohl(header.logEnd_ns);
      header.logStart_s	= ntohl(header.logStart_s);
      header.logStart_ns	= ntohl(header.logStart_ns);
      
      eTime = ((double) header.logEnd_s) + 
	(((double) header.logEnd_ns) / ONE_BILLION);
      sTime = ((double )header.logStart_s) + 
	(((double) header.logStart_ns) / ONE_BILLION);

      fprintf(stderr, "vqPlay: got %10d bytes @ %4.4f\n",
	      header.length, sTime);
    }
  }

#endif
  
  /*
   * If queue is empty, copy sbl for queue, otherwise, merge. 
   */
  if (objPtr->queue == NULL) {
    objPtr->queue = CopySBList(sbl);
  }
  else {
    MergeSBLists(objPtr->queue, sbl);
  }

  /*
   * If no decode in process, start a decode
   */
  if (!objPtr->inDecode) {
    VQ_RSRC_DoDecode(objPtr);
  }

  return TCL_OK;
}


/************************************************************
 *
 * StoreCodebook
 *
 * Parameters:
 *	packetPtr	Packet containing the codebook
 *	dataLength	Returns length of codebook
 *	codebookIndex	Returns codebook's index
 *
 * Return:
 *	Standard TCL result.
 *
 * Action:
 *	Store the codebook in the resource.  Alloc space & copy data.
 *
 ************************************************************/
static int
StoreCodebook(VQResource *resPtr, char *dataPtr, int length, int index)
{
  VQCodebook	*newPtr;
  
  if (resPtr->codebookArraySize <= index) {
    int newsize = (index+1) * sizeof(VQCodebook);
    resPtr->codebookArray = (VQCodebook *) realloc(resPtr->codebookArray,
						   newsize);
    if (resPtr->codebookArray == NULL) {
      Tcl_AppendResult(resPtr->interp, "couldn't resize codebook array", (char *) NULL);
      return TCL_ERROR;
    }
    resPtr->codebookArraySize = index+1;
  }

  newPtr = resPtr->codebookArray+index;
  newPtr->data = (char *) ckalloc(length);
  if (newPtr->data == NULL) {
    Tcl_AppendResult(resPtr->interp, "couldn't allocate codebook data", (char *) NULL);
    return TCL_ERROR;
  }
  memcpy(newPtr->data, dataPtr, length);
  newPtr->length = length;

  fprintf(stderr, "StoreCodebook: got cb %d\n", index);
  
  return TCL_OK;
}

/************************************************************
 *
 * UnpackCodebook
 *
 * Parameters:
 *	packetPtr	Packet containing the codebook
 *	dataLength	Returns length of codebook
 *	codebookIndex	Returns codebook's index
 *
 * Return:
 *	Pointer to actual codebook data, or NULL if an error occurs.
 *
 * Action:
 *	Unpack the codebook.  The data is not copied.
 *
 ************************************************************/
static char
*UnpackCodebook(char *packetPtr, int *dataLength, int *codebookIndex)
{
  int index;
  int headerTag;
  int dataTag;
  
  index = 0;
  memcpy(&headerTag, packetPtr+index, 4);
  headerTag = ntohl(headerTag);
  index+=4;
  if (headerTag != CM_VQ_BUF_HDR) {
    fprintf(stderr, "vqPlay: Expected a CM_VQ_BUF_HDR(%d), but got %d\n",
	    CM_VQ_BUF_HDR,
	    headerTag);
    return NULL;
  }

  memcpy(dataLength, packetPtr+index, 4);
  *dataLength = ntohl(*dataLength);
  index+=4;
  
  memcpy(&dataTag, packetPtr+index, 4);
  dataTag = ntohl(dataTag);
  index+=4;
  if (dataTag != CM_VQ_CODEBOOK_DATA) {
    fprintf(stderr, "vqPlay: Expected a CM_VQ_CODEBOOK_DATA(%d), but got %d\n",
	    CM_VQ_CODEBOOK_DATA,
	    headerTag);
    return NULL;
  }

  memcpy(codebookIndex, packetPtr+index, 4);
  *codebookIndex = ntohl(*codebookIndex);
  index+=4;

  return packetPtr+index;
}
		  
/************************************************************
 *
 * VQ_RSRC_AcceptCodebook
 *
 * Parameters:
 *     objPtr - pointer to object
 *     sbl -
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
static int
VQ_RSRC_AcceptCodebook(VQResource *objPtr, SB_List *sbl)
{
  int		sbIndex;
  ScatterBuffer	*currentSBPtr;
  int		bufferIndex;
  Buffer	*currentBufferPtr;
  
  if (!objPtr->ready) {
    return TCL_OK;
  }

  if ((sbl == NULL) || (sbl->count < 1)) {
    return TCL_OK;
  }

  /*
   * Increment reference count on buffers passed in
   */
  AttachToAllBuffers(objPtr->interp, sbl,
		     "tcmVQResource:VQ_RSRC_AcceptCodebook");

  for (sbIndex = 0; sbIndex < sbl->count; sbIndex++) {
    currentSBPtr = sbl->sbs[sbIndex];
    for (bufferIndex = 0; bufferIndex < currentSBPtr->count; bufferIndex++) {
      int		codebookLength;
      int		codebookIndex;
      char		*codebookDataPtr;
      char		*packetPtr;
      
      currentBufferPtr = currentSBPtr->buffers[bufferIndex];

      packetPtr = currentBufferPtr->data + currentSBPtr->offsets[bufferIndex];
      codebookDataPtr = UnpackCodebook(packetPtr, &codebookLength,
				       &codebookIndex);
      if (codebookDataPtr == NULL) {
	Tcl_AppendResult(objPtr->interp, "couldn't unpack codebook", (char *) NULL);
	return TCL_ERROR;
      }
      
      if (StoreCodebook(objPtr, codebookDataPtr, codebookLength, codebookIndex)
	  == TCL_ERROR) {
	Tcl_AppendResult(objPtr->interp,
			 "Couldn't store codebook in VQ_RSRC_AcceptCodebook",
			 (char *) NULL);
	return TCL_ERROR;
      }
    }
  }
  
  return TCL_OK;
}

/************************************************************
 *
 * CreateNewVQResource
 *
 * Parameters:
 *     interp - pointer to Tcl interpreter
 *     name - Tcl name for new object
 *
 * Return:
 *     Pointer to new object with fields
 *     initialized to default values.
 *
 * Action:
 *     Allocates memory for new object and
 *     initializes it.
 *
 ************************************************************/

static VQResource *
CreateNewVQResource(Tcl_Interp *interp, char *name)
{
  VQResource *new;
  char *acceptName;

  new = (VQResource *) ckalloc(sizeof(VQResource));
  if (new == NULL) {
    fprintf(stderr, "ERROR: couldn't alloc memory for new resource at %s:%d",
	    __FILE__, __LINE__);
    return NULL;
  }

  new->interp = interp;
  new->name = (char *) ckalloc(strlen(name) + 1);
  if (new->name == NULL) {
    fprintf(stderr, "ERROR: couldn't alloc memory for new resource name at %s:%d",
	    __FILE__, __LINE__);
    return NULL;
  }
  strcpy(new->name, name);

  new->width		 = 0;
  new->height		 = 0;
  new->xpos		 = 0;
  new->ypos		 = 0;
  new->squeeze		 = 0;
  new->devtype		 = NULL;
  new->devicePtr	 = NULL;
  new->xid		 = 0;
  new->lts		 = NULL;
  new->ltsTriggerNum	 = -1;
  new->ltsName		 = NULL;
  new->maxFrameRate	 = 0;
  new->ctrlCallback	 = NULL;
  new->ctrlCycleTime	 = 0.0;
  new->statStartTime	 = 0.0;
  new->missed		 = 0;
  new->late		 = 0;
  new->played		 = 0;
  new->bytes		 = 0;
  new->ready		 = 0;
  new->inited		 = 0;
  new->inDecode		 = 0;
  new->queue		 = NULL;
  new->cntrlEveryHandle	 = NULL;
  new->destroyFlag	 = 0;
  new->at_tokens	 = NULL;
  new->num_at_tokens	 = 0;
  new->codebookArray	 = malloc(0);
  new->codebookArraySize = 0;

  ChangeDeviceType (new, "default");

  /*
   * Register accept as an atom.
   */
  acceptName = (char *) ckalloc(strlen(name)+8);
  sprintf(acceptName, "%s.%s", name, "accept");
  Tutil_RegisterCommand(interp, acceptName, (ClientData) VQ_RSRC_Accept,
			(ClientData) new);
  new->acceptHandle = acceptName;

  acceptName = (char *) ckalloc(strlen(name)+16);
  sprintf(acceptName, "%s.%s", name, "acceptCodebook");
  Tutil_RegisterCommand(interp, acceptName, (ClientData) VQ_RSRC_AcceptCodebook,
			(ClientData) new);
  new->acceptCodebookHandle = acceptName;

  return new;
}


/************************************************************
 *
 * VQ_RSRC_LTSChanged
 *
 * Parameters:
 *     clientdata - pointer to object.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *      Called when the LTS changes value.  Dequeues any previously
 *      queued play requests, frees any memory we're holding, and
 *      returns.
 *
 ************************************************************/
static void
VQ_RSRC_LTSChanged(ClientData clientData)
{
  VQResource *objPtr = (VQResource *) clientData;
  int i;
  
  /*
   * If not ready, do nothing.
   */
  if (!objPtr->ready) {
    return;
  }
  
  /*
   * Blow away "at" queue.
   */
  for (i=0; i<objPtr->num_at_tokens; i++) {
    if (objPtr->at_tokens[i] > 0) {
      AtEvent *at_event;
      int token;

      token = objPtr->at_tokens[i];
      at_event = Tcm_FindAtEvent(token);
      if (token && (at_event != NULL)) {
	VQ_RSRC_FreeFrame(at_event->fireClientData);
	Tcm_DeleteAtHandler(token);
	objPtr->at_tokens[i] = 0;
      }
    }
  }
  
  /*
   * Blow away my frame queue.
   */
  if (objPtr->queue != NULL) {
    DetachToAllBuffers(objPtr->interp, objPtr->queue,
		       "tcmVQResource:VQ_RSRC_LTSChanged");
    FreeSBList(objPtr->queue);
    objPtr->queue = NULL;
  }
  
  /*
   * Reset all the statistics stuff.
   */
  objPtr->statStartTime = Tcm_ReadSysClock();
  objPtr->missed = objPtr->late = objPtr->played = objPtr->bytes = 0;
}


/************************************************************
 *
 * ConfigureVQResource
 *
 * Parameters:
 *     objPtr - pointer to the object
 *     argc - num of config arguments
 *     argv - config arguments
 *
 * Return:
 *     Standard TCL return value
 *
 * Action:
 *     Does the usual configuration thing
 *
 ************************************************************/

static int
ConfigureVQResource(VQResource *objPtr, int argc, char **argv)
{

  Tcl_Interp *interp;
  int cntr;
  char str[50];

  interp = objPtr->interp;

  /*
   * Case 1: <obj> configure
   */
  if (argc == 0) {
    sprintf(str, "{%d}", objPtr->width);
    Tcl_AppendResult(interp, "{", "-width ", "0 ", str, "} ",
    (char *)NULL);
    sprintf(str, "{%d}", objPtr->height);
    Tcl_AppendResult(interp, "{", "-height ", "0 ", str, "} ",
    (char *)NULL);
    sprintf(str, "{%d}", objPtr->xpos);
    Tcl_AppendResult(interp, "{", "-xpos ", "0 ", str, "}",
    (char *)NULL);
    sprintf(str, "{%d}", objPtr->ypos);
    Tcl_AppendResult(interp, "{", "-ypos ", "0 ", str, "}",
    (char *)NULL);
    sprintf(str, "{%d}", objPtr->squeeze);
    Tcl_AppendResult(interp, "{", "-squeeze ", "0 ", str, "}",
    (char *)NULL);
    sprintf(str, "{%d}", objPtr->xid);
    Tcl_AppendResult(interp, "{", "-xid ", "0 ", str, "}",
    (char *)NULL);
    Tcl_AppendResult(interp, "{", "-device ", "default ",
    objPtr->devtype, "}", (char *)NULL);

    if (objPtr->ltsName == NULL) {
      Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{}", "}",
      (char *)NULL);
    }
    else {
      Tcl_AppendResult(interp, "{", "-lts ", "{} ", "{",
      objPtr->ltsName, "}", "}", (char *) NULL);
    }
    sprintf(str, "{%d}", objPtr->maxFrameRate);
    Tcl_AppendResult(interp, "{", "-maxFrameRate ", "0 ", str, "}",
    (char *) NULL);
    if (objPtr->ctrlCallback == NULL) {
      Tcl_AppendResult(interp, "{", "-ctrlCallback ", "{} ", "{}",
      "}", (char *) NULL);
    }
    else {
      Tcl_AppendResult(interp, "{", "-ctrlCallback ", "{} ",
      "{", objPtr->ctrlCallback, "}", "}", (char *) NULL);
    }
    Tcl_PrintDouble(interp, objPtr->ctrlCycleTime, str);
    Tcl_AppendResult(interp, "{", "-ctrlCycleTime ", "0.0 ",
    str, "}", (char *) NULL);
    return TCL_OK;
  }
    
  /*
   * Case 2: <obj> configure <-slot>
   */
  if (argc == 1) {
    char c;
    int len;

    len = strlen(argv[0]);
    c = argv[0][1];

    if ((c == 'w') && (strncmp("-width", argv[0], len) == 0)) {
      sprintf(str, "{%d}", objPtr->width);
      Tcl_AppendResult(interp, "-width ", "0 ", str, (char *) NULL);
      return TCL_OK;
    } else if ((c == 'h') && (strncmp("-height", argv[0], len) == 0)) {
      sprintf(str, "{%d}", objPtr->height);
      Tcl_AppendResult(interp, "-height ", "0 ", str, (char *) NULL);
      return TCL_OK;
    } else if ((c == 'x') && (strncmp("-xpos", argv[0], len) == 0)) {
      sprintf(str, "{%d}", objPtr->xpos);
      Tcl_AppendResult(interp, "-xpos ", "0 ", str, (char *) NULL);
      return TCL_OK;
    } else if ((c == 'y') && (strncmp("-ypos", argv[0], len) == 0)) {
      sprintf(str, "{%d}", objPtr->ypos);
      Tcl_AppendResult(interp, "-ypos ", "0 ", str, (char *) NULL);
      return TCL_OK;
    } else if ((c == 's') && (strncmp("-squeeze", argv[0], len) == 0)) {
      sprintf(str, "{%d}", objPtr->squeeze);
      Tcl_AppendResult(interp, "-squeeze ", "0 ", str, (char *) NULL);
      return TCL_OK;
    } else if ((c == 'x') && (strncmp("-xid", argv[0], len) == 0)) {
      sprintf(str, "{%d}", objPtr->xid);
      Tcl_AppendResult(interp, "-xid ", "0 ", str, (char *) NULL);
      return TCL_OK;
    } else if ((c == 'd') && (strncmp("-device", argv[0], len) == 0)) {
      Tcl_AppendResult(interp, "-device ", "default ", objPtr->devtype,
		       (char *) NULL);
      return TCL_OK;
    } else if ((c == 'l') && (strncmp("-lts", argv[0], len) == 0)) {
      if (objPtr->ltsName == NULL) {
	Tcl_AppendResult(interp, "-lts ", "{} ", "{}", (char *) NULL);
      } else {
	Tcl_AppendResult(interp, "-lts ", "{} ", "{", objPtr->ltsName, "}", (char *) NULL);
      }
      return TCL_OK;
    } else if ((c == 'm') && (strncmp("-maxFrameRate", argv[0], len) == 0)) {
      sprintf(str, "{%d}", objPtr->maxFrameRate);
      Tcl_AppendResult(interp, "-maxFrameRate ", "0 ", str, (char *) NULL);
      return TCL_OK;
    } else if ((c == 'c') && (strncmp("-ctrlCallback", argv[0], len) == 0)) {
      if (objPtr->ctrlCallback == NULL) {
	Tcl_AppendResult(interp, "-ctrlCallback ", "{} ", "{}", (char *) NULL);
      } else {
	Tcl_AppendResult(interp, "-ctrlCallback ", "{} ", "{", objPtr->ctrlCallback, "}", (char *) NULL);
      }
      return TCL_OK;
    } else if ((c == 'c') && (strncmp("-ctrlCycleTime", argv[0], len) == 0)) {
      char str[TCL_DOUBLE_SPACE];
	  
      Tcl_PrintDouble(interp, objPtr->ctrlCycleTime, str);
      Tcl_AppendResult(interp, "-ctrlCycleTime ", "{0.0} ", str, (char *) NULL);
      return TCL_OK;
    }
    Tcl_AppendResult(interp, "bad configuration option: ", argv[0], (char *) NULL);
    return TCL_ERROR;
  }
    
  /*
   * Case 3: <obj> configure <-slot> <newValue>
   */
  cntr = 0;
  while (cntr < argc) {
    char c;
    int len;

    /*
     * Make resource unready since something is changing.
     */

    VQ_RSRC_Unready(objPtr);

    len = strlen(argv[cntr]);
    c = argv[cntr][1];

    if ((c == 'w') && (strncmp("-width", argv[cntr], len) == 0)) {
      int newWidth;

      cntr++;
      if (cntr == argc) {
	noValueError:
	Tcl_AppendResult(interp, "Must specify value when ",
			 "trying to set configuration option ",
			 argv[cntr-1], (char *) NULL);
	return TCL_ERROR;
      }
      if (Tcl_GetInt(interp, argv[cntr], &newWidth) != TCL_OK) {
	badValueError:
	Tcl_AppendResult(interp, "\nbad value \"", argv[cntr],
			 "\" given for configuration option ",
			 argv[cntr-1], (char *) NULL);
	return TCL_ERROR;
      }
      objPtr->width = newWidth;
    } else if ((c == 'h') && (strncmp("-height", argv[cntr], len) == 0)) {
      int newHeight;

      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (Tcl_GetInt(interp, argv[cntr], &newHeight) != TCL_OK) {
	goto badValueError;
      }
      objPtr->height = newHeight;
    } else if ((c == 'x') && (strncmp("-xpos", argv[cntr], len) == 0)) {
      int newXpos;

      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (Tcl_GetInt(interp, argv[cntr], &newXpos) != TCL_OK) {
	goto badValueError;
      }
      objPtr->xpos = newXpos;
    } else if ((c == 'y') && (strncmp("-ypos", argv[cntr], len) == 0)) {
      int newYpos;

      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (Tcl_GetInt(interp, argv[cntr], &newYpos) != TCL_OK) {
	goto badValueError;
      }
      objPtr->ypos = newYpos;
    } else if ((c == 's') && (strncmp("-squeeze", argv[cntr], len) == 0)) {
      int newSqueeze;

      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (Tcl_GetInt(interp, argv[cntr], &newSqueeze) != TCL_OK) {
	goto badValueError;
      }
      objPtr->squeeze = newSqueeze;
    } else if ((c == 'x') && (strncmp("-xid", argv[cntr], len) == 0)) {
      int newXid;

      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (Tcl_GetInt(interp, argv[cntr], &newXid) != TCL_OK) {
	goto badValueError;
      }
      objPtr->xid = newXid;
    } else if ((c == 'd') && (strncmp("-device", argv[cntr], len) == 0)) {
      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      ChangeDeviceType(objPtr, argv[cntr]);
    } else if ((c == 'l') && (strncmp("-lts", argv[cntr], len) == 0)) {
      ClientData newLts;

      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (strlen(argv[cntr]) == 0) {
	newLts = NULL;
      }
      else if ((newLts = Tcm_FindAtom(interp, argv[cntr])) == NULL) {
	goto badValueError;
      }
      if (objPtr->ltsName != NULL) {
	ckfree(objPtr->ltsName);
	objPtr->ltsName = NULL;
      }

      /*
       * Release our lock on the old lts...
       */
      if (objPtr->lts != NULL) {
	if (objPtr->ltsTriggerNum != -1) {
	  LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
	  objPtr->ltsTriggerNum = -1;
	}
	Tk_Release(objPtr->lts);
	objPtr->lts = NULL;
      }

      objPtr->lts = newLts;
      if (newLts != NULL) {
	Tk_Preserve (objPtr->lts);
	objPtr->ltsTriggerNum = LTS_SetTrigger(objPtr->lts, 
					       LTS_SPEED_FLAG | LTS_VALUE_FLAG | LTS_EARLY_PHASE, 
					       VQ_RSRC_LTSChanged, objPtr);
	objPtr->ltsName = (char *) ckalloc(strlen(argv[cntr]) + 1);
	strcpy(objPtr->ltsName, argv[cntr]);
      } else {
	objPtr->ltsName = NULL;
      }
    } else if ((c == 'm') &&
	       (strncmp("-maxFrameRate", argv[cntr], len) == 0)) {
      int newMaxFrameRate;

      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (Tcl_GetInt(interp, argv[cntr], &newMaxFrameRate) != TCL_OK) {
	goto badValueError;
      }
      objPtr->maxFrameRate = newMaxFrameRate;
    } else if ((c == 'c') && (strncmp("-ctrlCallback", argv[cntr], len) == 0)) {
      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (objPtr->ctrlCallback != NULL) {
	ckfree(objPtr->ctrlCallback);
      }
      objPtr->ctrlCallback = (char *) ckalloc(strlen(argv[cntr]) + 1);
      strcpy(objPtr->ctrlCallback, argv[cntr]);
    } else if ((c == 'c') &&
	       (strncmp("-ctrlCycleTime", argv[cntr], len) == 0)) {
      double newCtrlCycleTime;

      cntr++;
      if (cntr == argc) {
	goto noValueError;
      }
      if (Tcl_GetDouble(interp,argv[cntr],&newCtrlCycleTime) != TCL_OK) {
	goto badValueError;
      }
      objPtr->ctrlCycleTime = newCtrlCycleTime;
    }
    cntr++;
  }
  return TCL_OK;
}


/************************************************************
 *
 * DestroyVQResource
 *
 * Parameters:
 *     objPtr - pointer to object.
 *
 * Return:
 *     Standard Tcl Result
 *
 * Action:
 *     Destroys previously created object.
 *
 ************************************************************/

static int
DestroyVQResource(VQResource *objPtr)
{
  int result = TCL_OK;
  int i, devReturn;

  /*
   * If not being called by DecodeDone first delete Tcl command.
   */
  if (objPtr->destroyFlag != 1) {
    Tcl_DeleteCommand(objPtr->interp, objPtr->name);
  }

  /*
   * Make unready, cancel any DoDecode when idle calls.
   */
  objPtr->ready = 0;
  Tcm_CancelIdleCall("tk", (Tk_IdleProc *) VQ_RSRC_DoDecode, objPtr);

  /*
   * If in decode, set destroy flag and return. Let DecodeDone do destroy.
   */
  if (objPtr->inDecode) {
    objPtr->destroyFlag = 1;
    return TCL_OK;
  }

  /*
   * Unregister the accept atom
   */
  if (objPtr->acceptHandle != NULL) {
    Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptHandle);
    ckfree(objPtr->acceptHandle);
    objPtr->acceptHandle = NULL;
  }

  if (objPtr->acceptCodebookHandle != NULL) {
    Tutil_UnregisterCommand(objPtr->interp, objPtr->acceptCodebookHandle);
    ckfree(objPtr->acceptCodebookHandle);
    objPtr->acceptCodebookHandle = NULL;
  }


  /*
   * Shut down the device
   */
  if (objPtr->devicePtr != NULL) {
    devReturn = objPtr->devicePtr->shutdownProc(objPtr);
    if (devReturn != 0) {
      result = TCL_ERROR;
      if (objPtr->devicePtr->errorProc != NULL) {
	Tcl_AppendResult(objPtr->interp,
			 objPtr->devicePtr->errorProc(objPtr, devReturn),
			 (char *) NULL);
      }
    }
    ckfree((char *) objPtr->devicePtr);
    objPtr->devicePtr = NULL;
  }

  /*
   * Clean up lts
   */
  if (objPtr->ltsName != NULL) {
    ckfree(objPtr->ltsName);
  }

  /*
   * Release our lock on the old lts...
   */
  if (objPtr->lts != NULL) {
    if (objPtr->ltsTriggerNum != -1) {
      LTS_UnsetTrigger(objPtr->lts, objPtr->ltsTriggerNum);
      objPtr->ltsTriggerNum = -1;
    }
    Tk_Release(objPtr->lts);
    objPtr->lts = NULL;
  }

  /*
   * Free up previous allocs, frame queue, and such.
   */
  if (objPtr->name != NULL) {
    ckfree(objPtr->name);
  }
  if (objPtr->ctrlCallback != NULL) {
    free(objPtr->ctrlCallback);
  }
  if (objPtr->queue != NULL) {
    DetachToAllBuffers (objPtr->interp, objPtr->queue,
			"tcmVQResource:DestroyVQResource");
    FreeSBList(objPtr->queue);
    objPtr->queue = NULL;
  }
  if (objPtr->cntrlEveryHandle != NULL) {
    Tcm_DeleteEvery(objPtr->interp, objPtr->cntrlEveryHandle, NULL);
    objPtr->cntrlEveryHandle = NULL;
  }

  /*
   * Clean up previously scheduled display requests
   */
  for (i=0; i<objPtr->num_at_tokens; i++) {
    if (objPtr->at_tokens[i] > 0) {
      AtEvent *at_event;
      int token;

      token = objPtr->at_tokens[i];
      at_event = Tcm_FindAtEvent(token);
      if (token && (at_event != NULL)) {
	VQ_RSRC_FreeFrame(at_event->fireClientData);
	Tcm_DeleteAtHandler(token);
	objPtr->at_tokens[i] = 0;
      }
    }
  }

  ckfree((char *) objPtr);
  return result;
}


/************************************************************
 *
 * VQ_RSRC_Ready
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
VQ_RSRC_Ready(VQResource *objPtr)
{
  int init_res;

  /*
   * Can only be made ready if
   *  1. We have a device.
   *  2. We have an lts
   *  3. We have a pad ID
   */
  if (objPtr->devicePtr == NULL) {
    Tcl_AppendResult(objPtr->interp, "Can't make VQ resource \"",
		     objPtr->name, "\" ready: device type is not set.", (char *)NULL);
    return TCL_ERROR;
  }
  if (objPtr->lts == NULL) {
    Tcl_AppendResult(objPtr->interp, "Can't make VQ resource \"",
		     objPtr->name, "\" ready: no lts has been attached.", (char *)NULL);
    return TCL_ERROR;
  }
  if (objPtr->xid == 0) {
    Tcl_AppendResult(objPtr->interp, "Can't make VQ resource \"",
		     objPtr->name, "\" ready: output InfoPad not specified.",
		     (char *)NULL);
    return TCL_ERROR;
  }


  /*
   * If not inited, do initProc, catch result.
   */
  init_res = 0;
  if (!objPtr->inited) {
    if (objPtr->devicePtr->initProc != NULL) {
      init_res = objPtr->devicePtr->initProc(objPtr);
    }
    if (init_res == 0) 
      objPtr->inited = 1;
  }
  else {
    /*
     * Otherwise, do reinitProc, catch result.
     */
    if (objPtr->devicePtr->reinitProc != NULL) {
      init_res = objPtr->devicePtr->reinitProc(objPtr);
    }
  }
      
  /*
   * If bad result, get error string, return error.
   */
  if (init_res != 0) {
    char *errStr;

    if (objPtr->devicePtr->errorProc != NULL) {
      errStr = objPtr->devicePtr->errorProc(objPtr, init_res);
    } else {
      errStr = "Error with device initialization.";
    }

    Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
    return TCL_ERROR;
  }

  objPtr->ready = 1;

  /*
   * If queue is not empty, start a decode.
   */

  if (objPtr->queue != NULL) {
    VQ_RSRC_DoDecode(objPtr);
  }

  /*
   * Reset stat stuff.
   */

  objPtr->statStartTime = Tcm_ReadSysClock();
  objPtr->missed = objPtr->played = objPtr->late = objPtr->bytes = 0;

  return TCL_OK;
}

/************************************************************
 *
 * VQ_RSRC_Stats
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
VQ_RSRC_Stats(VQResource * objPtr)
{
  return TCL_OK;
}

/************************************************************
 *
 * VQ_RSRC_Refresh
 *
 * Parameters:
 *     objPtr - pointer to object
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
VQ_RSRC_Refresh(VQResource * objPtr)
{
  int res;
  char *errStr;

  /*
   * If not ready, do nothing.
   */

  if (!objPtr->ready) return TCL_OK;

  /*
   * If device has refresh, call it.
   */

  if (objPtr->devicePtr != NULL) {
    if (objPtr->devicePtr->refreshProc != NULL) {
      res = objPtr->devicePtr->refreshProc(objPtr);
      if (res != 0) {
	errStr = "Error refreshing picture.";
	if (objPtr->devicePtr->errorProc != NULL) {
	  errStr = objPtr->devicePtr->errorProc(objPtr, res);
	}
	Tcl_AppendResult(objPtr->interp, errStr, (char *) NULL);
	return TCL_ERROR;
      }
    }
  }

  return TCL_OK;
}

/******************************************************************
 *
 * VQResourceDeviceList
 *
 * Parameters:
 *     objPtr - pointer to vq play object.
 *
 * Results:
 *     Standard Tcl Result.
 *
 * Actions:
 *     Sets result to list of devices supported. Returns TCL_OK;
 *
 ******************************************************************/
static int
VQResourceDeviceList(VQResource *objPtr)
{
    Tcl_Interp *interp = objPtr->interp;
    Tcl_DString resStr;

    Tcl_DStringInit(&resStr);

    Tcl_DStringAppendElement(&resStr, "infopad");

    Tcl_SetResult(interp, Tcl_DStringValue(&resStr), TCL_VOLATILE);
    Tcl_DStringFree(&resStr);

    return TCL_OK;
}


/************************************************************
 *
 * VQResourceObjCmd
 *
 * Parameters:
 *     clientData - pointer to object.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the actual methods of
 *     VQResource object. It is passed a pointer to the
 *     actual data structure of the particular object
 *     instance.
 *
 ************************************************************/

int
VQResourceObjCmd(ClientData clientData, Tcl_Interp * interp,
		 int argc, char **argv)
{
  char c;
  int len;
  VQResource *objPtr = (VQResource *) clientData;

  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # of args, should be: ",
		     argv[0], " <command> ?arguments?", (char *) NULL);
    goto usage;
  }
  c = argv[1][0];
  len = strlen(argv[1]);

  if ((c == 'c') && (strncmp("configure", argv[1], len) == 0)) {
    return ConfigureVQResource(clientData, argc - 2, argv + 2);
  } else if ((c == 'd') && (strncmp("destroy", argv[1], len) == 0)) {
    return DestroyVQResource(clientData);
  } else if ((c == 'a') && (strncmp("accept", argv[1], len) == 0)) {
    SB_List *sbl = NULL;
    int result;

    if (argc < 3) {
      Tcl_AppendResult(interp, "wrong # of args, should be: ",
		       argv[0], " <ScatterBufferList> ", (char *) NULL);
      return TCL_ERROR;
    }

    /*
     * Parse the ScatterBufferList
     */
    sbl = (SB_List *) ckalloc(sizeof(SB_List));
    if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
      ckfree((char *) sbl);
      Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
		       argv[2],  (char *)NULL);
      return TCL_ERROR;
    }

    result = VQ_RSRC_Accept(objPtr, sbl);
    if (sbl != NULL) {
      FreeSBList(sbl);
    }
    return result;

  } else if ((c == 'a') && (strncmp("acceptCodebook", argv[1], len) == 0)) {
    SB_List *sbl = NULL;
    int result;

    if (argc < 3) {
      Tcl_AppendResult(interp, "wrong # of args, should be: ",
		       argv[0], " <ScatterBufferList> ", (char *) NULL);
      return TCL_ERROR;
    }

    /*
     * Parse the ScatterBufferList
     */
    sbl = (SB_List *) ckalloc(sizeof(SB_List));
    if (ParseSBList(objPtr->interp, argv[2], sbl) != TCL_OK) {
      ckfree((char *) sbl);
      Tcl_AppendResult(objPtr->interp, "Bad scatterbuffer list.",
		       argv[2],  (char *)NULL);
      return TCL_ERROR;
    }

    result = VQ_RSRC_AcceptCodebook(objPtr, sbl);
    if (sbl != NULL) {
      FreeSBList(sbl);
    }
    return result;

  } else if ((c == 'r') && (strncmp("ready", argv[1], len) == 0)) {
    return VQ_RSRC_Ready(objPtr);

  } else if ((c == 'u') && (strncmp("unready", argv[1], len) == 0)) {
    return VQ_RSRC_Unready(objPtr);

  } else if ((c == 's') && (strncmp("stats", argv[1], len) == 0)) {
    return VQ_RSRC_Stats(objPtr);

  } else if ((c == 'r') && (strncmp("refresh", argv[1], len) == 0)) {
    return VQ_RSRC_Refresh(objPtr);

  } else if ((c == 'd') && (strncmp("deviceList", argv[1], len) == 0)) {
    return VQResourceDeviceList(objPtr);
  
  } else if ((c == 'r') && (strncmp("reinit", argv[1], len) == 0)) {
    return objPtr->devicePtr->reinitProc(objPtr);
  }

  Tcl_AppendResult(interp, "Bad command option: ", argv[0],
		   argv[1], (char *) NULL);
  usage:
  Tcl_AppendResult(interp,
		   "\n<command> can be one of:\n",
		   "  configure ?-slot ?value??\n",
		   "  destroy\n",
		   "  accept <ScatterBufferList>\n",
		   "  acceptCodebook <ScatterBufferList>\n",
		   "  ready\n",
		   "  unready\n",
		   "  stats\n",
		   "  refresh\n",
		   "  reinit\n",
		   (char *) NULL);
  return TCL_ERROR;
}

/************************************************************
 *
 * VQResourceCreateCmd
 *
 * Parameters:
 *     clientData - pointer to data when command was created.
 *                  Not used here.
 *     interp - pointer to Tcl interpreter
 *     argc - argument count
 *     argv - arguments
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     This command implements the "vqResource" Tcl command.
 *     A new object of type VQResource is created and initialized.
 *     Naming of the new object follows the proper object
 *     naming scheme for all cm objects. A Tcl command is bound
 *     to the object name to implement the objects methods.
 *
 ************************************************************/

int
VQResourceCreateCmd(ClientData clientData, Tcl_Interp *interp,
		    int argc, char **argv)
{

  char new_name[256];
  VQResource *new;

  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # of args, should be: ", argv[0],
		     " parent ?options?", (char *) NULL);
    return TCL_ERROR;
  }
  nextVQResourceNum++;
  sprintf(new_name, "%svqRSRC%d", argv[1], nextVQResourceNum);

  new = CreateNewVQResource(interp, new_name);
  if (new == NULL) {
    Tcl_AppendResult(interp, "couldn't create new object", (char *) NULL);
    return TCL_ERROR;
  }
  Tcl_CreateCommand(interp, new_name, VQResourceObjCmd,
		    (ClientData) new, (void (*)()) NULL);

  if (argc > 2) {
    ConfigureVQResource(new, argc - 2, argv + 2);
  }
  Tcl_SetResult(interp, new_name, TCL_VOLATILE);
  return TCL_OK;
}

/************************************************************
 *
 * VQ_RSRC_MissDisplay
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
VQ_RSRC_MissDisplay(VQFrame *framePtr)
{
  VQResource *objPtr;

  objPtr = framePtr->objPtr;

#ifdef DEBUG_FRAMES
  fprintf(stderr, "vqPlay:MissDisplay: missed @ %f\n", framePtr->logStart);
#endif

  /*
   * Collect stats.
   */

  objPtr->missed++;

  VQ_RSRC_FreeFrame((ClientData)framePtr);

  return TCL_OK;
}

/************************************************************
 *
 * VQ_RSRC_DoDisplay
 *
 * Parameters:
 *     framePtr - frame pointer.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/

static int
VQ_RSRC_DoDisplay(VQFrame *framePtr)
{
  VQResource *objPtr;
  double now, speed;

  objPtr = framePtr->objPtr;

  now = LTS_Value(objPtr->lts);
  speed = LTS_Speed(objPtr->lts);

  if (((speed >= 0) &&
       (framePtr->logStart <= now) && (framePtr->logEnd >= now)) ||
      ((speed < 0) &&
       (framePtr->logStart >= now) && (framePtr->logEnd <= now))) {

    /*
     * Collect stats.
     */
    objPtr->played++;
      
    /*
     * Call device's display procedure.
     */
    if (objPtr->devicePtr != NULL) {
      if (objPtr->devicePtr->displayProc != NULL) {
	objPtr->devicePtr->displayProc(framePtr);
      }
    }
    VQ_RSRC_FreeFrame((ClientData) framePtr);
  }
  else {
    VQ_RSRC_MissDisplay (framePtr);
  }
  return TCL_OK;
}


/************************************************************
 *
 * VQ_RSRC_DecodeDone
 *
 * Parameters:
 *     framePtr - pointer to frame just finished with decoding.
 *
 * Return:
 *     Standard Tcl return value
 *
 * Action:
 *
 ************************************************************/
int
VQ_RSRC_DecodeDone(VQFrame *framePtr)
{
  int i, new_num;
  VQResource *objPtr;

  objPtr = framePtr->objPtr;

  /*
   * Reset in decode
   */

  objPtr->inDecode = 0;

  /*
   * If destroy flag set, free frame, call destroy and return.
   */
  if (framePtr->objPtr->destroyFlag) {
    VQ_RSRC_FreeFrame((char *) framePtr);
    DestroyVQResource(objPtr);
    return TCL_OK;
  }


  /*
   * If not ready, free frame and return.
   */
  if (!objPtr->ready) {
    VQ_RSRC_FreeFrame((char *) framePtr);
    return TCL_OK;
  }

  /*
   * Gather stats.
   */
  objPtr->bytes += framePtr->dataLength;

  /*
   * Schedule display.  Search for a free at_token.  If we can't find
   * one, to increase size of at_token array
   */
  if (objPtr->num_at_tokens == 0) {
    objPtr->num_at_tokens = 30;
    objPtr->at_tokens = (int *) ckalloc(sizeof(int) * 30);
    for (i=0; i<objPtr->num_at_tokens; i++) {
      objPtr->at_tokens[i] = 0;
    }
  } 
  for (i=0; i<objPtr->num_at_tokens; i++) {
    if (objPtr->at_tokens[i] == 0) {
      break;
    }
  }
  if (i == objPtr->num_at_tokens) {
    new_num = objPtr->num_at_tokens + 10;
    objPtr->at_tokens = (int *) realloc((char *) objPtr->at_tokens,
					sizeof(int) * new_num);
    for (i=objPtr->num_at_tokens; i < new_num; i++) {
      objPtr->at_tokens[i] = 0;
    }
    i = objPtr->num_at_tokens;
    objPtr->num_at_tokens = new_num;
  }

  /*
   * If speed's zero and we're in the window, throw up the current
   * frame.  Otherwise, create an atEvent for it.
   */
  if (LTS_Speed(objPtr->lts) == 0.0) {
    framePtr->at_token = -1; 
    if ((framePtr->logStart <= LTS_Value(objPtr->lts)) &&
	(framePtr->logEnd >= LTS_Value(objPtr->lts))) {
      VQ_RSRC_DoDisplay(framePtr);
    }
    else {
      VQ_RSRC_MissDisplay(framePtr);
    }
  }
  else {
    double t1, t2;
    t1 = LTS_SystemTimeOf(objPtr->lts, framePtr->logStart);
    t2 = LTS_SystemTimeOf(objPtr->lts, framePtr->logEnd);
    objPtr->at_tokens[i] = 
      Tcm_CreateAtHandler(t1<t2?t1:t2, HUGE_VAL,
			  "tk", (Tk_TimerProc *) VQ_RSRC_DoDisplay, 
			  (ClientData) framePtr, NULL,
			  NULL, (Tk_IdleProc *) NULL,
			  (ClientData) NULL, NULL);
    framePtr->at_token = i;
  }
		
  /*
   * Arrange to decode next frame when we're idle
   */
  if (objPtr->queue != NULL) {
    Tcm_DoWhenIdle("tk", (Tk_IdleProc *) VQ_RSRC_DoDecode, objPtr);
  }

  return TCL_OK;
}


/************************************************************
 *
 * VQResourceModuleInit
 *
 * Parameters:
 *     interp - Tcl interpreter
 *
 * Return:
 *     Standard Tcl return value.
 *
 * Action:
 *     Initializes the object module, registering
 *     media, format, and object keywords as well
 *     as creating Tcl command for object type creation
 *
 ************************************************************/

int
VQResourceModuleInit(Tcl_Interp * interp)
{
  Tcl_CreateCommand(interp, "vqPlay", VQResourceCreateCmd,
		    NULL, (void (*)()) NULL);
  Tcl_GlobalEval(interp, "cminfo add media {video}");
  Tcl_GlobalEval(interp, "cminfo add formats video {vq}");
  Tcl_GlobalEval(interp, "cminfo add objects video vq vqResource");
  Tcl_VarEval(interp, "cminfo add object_info vqResource {resource",
#ifdef J300
	      " j300",
#endif
#ifdef PARALLAX
	      " parallax",
#endif
#ifdef JPEG_SW
	      " software",
#endif
	      "}", NULL);

  return TCL_OK;
}
