/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#include "vqRawFormat.h"
#include <stdio.h>
#ifdef NO_STDLIB_H
#   include "compat/stdlib.h"
#else
#   include <stdlib.h>
#endif
#include <sys/types.h>
#include <netinet/in.h>

VQRF_file *VQRF_Open(char *name, char *options, char **buffer)
{
  FILE *f;
  
  if ((f = fopen(name, options)) == NULL) {
    fprintf(stderr, "VQRF_Open: couldn't open file '%s'\n",
	    name);
    return NULL;
  }

  return VQRF_Reopen(f, buffer);
}

VQRF_file *VQRF_Reopen(FILE *f, char **buffer)
{
  VQRF_file	*result;
  int		bytesRead;

  result = (VQRF_file *) malloc(sizeof(*result));
  if (result == NULL) {
    fprintf(stderr, "VQRF_Open: couldn't allocate space for VQRF_file\n");
    return NULL;
  }
  result->file = f;

  bytesRead = fread(&result->header, 1, sizeof(result->header), result->file);
  if (bytesRead != sizeof(result->header)) {
    fprintf(stderr, "VQRF_Reopen: tried to read header (%ld bytes) but only got %d\n",
	    sizeof(VQRF_GetHeader(result)), bytesRead);
    free((char *) result);
    return NULL;
  }
  if ((result->header.majorVersion != 1) &&
      (result->header.minorVersion != 0)) {
    fprintf(stderr, "VQRF: wrong version number %d.%d (required 1.0)\n",
	    result->header.majorVersion, result->header.minorVersion);
  }
  
  if (buffer != NULL) {
    if (*buffer == NULL) {
      *buffer = (char *) malloc(result->header.extraHeaderSize);
      if (*buffer == NULL) {
	fprintf(stderr, "VQRF_Reopen: couldn't alloc %d bytes for extra header info\n",
		result->header.extraHeaderSize);
	free((char *) result);
	return NULL;
      }
    }
    bytesRead = fread(*buffer, 1, result->header.extraHeaderSize, result->file);
    if (bytesRead != result->header.extraHeaderSize) {
      fprintf(stderr, "VQRF_Reopen: tried to read extra info (%d bytes) but only got %d\n",
	      result->header.extraHeaderSize, bytesRead);
      free((char *) result);
      return NULL;
    }
  }
  return result;
}

void VQRF_Close(VQRF_file *f)
{
  fclose(f->file);
}

uint32 VQRF_ReadUint32(VQRF_file *f)
{
  uint32	result;
  int		bytesRead;
  int		before, after;

  if (feof(VQRF_GetFilePtr(f))) {
    fprintf(stderr, "VQRF_ReadUint32: tried to read when at EOF\n");
  }

  before = feof(VQRF_GetFilePtr(f));
  bytesRead = fread(&result, 1, sizeof(result), f->file);
  after = feof(VQRF_GetFilePtr(f));
  if (before != after)
    fprintf(stderr, "BOGOSITY!\n");
  if (bytesRead != sizeof(result)) {
    fprintf(stderr, "VQRF_ReadUnit32: unable to read uint32\n");
  }

  return ntohl(result);
}

VQRF_frameHunk VQRF_ReadFrameHunk(VQRF_file *f, char **frameData)
{
  VQRF_frameHunk	result;
  VQRF_codebookUpdate	cbUpdate;
  int			numUpdates;
  VQRF_codebookUpdate	*updateArray;

  /* read codebook updates, if any */

  updateArray = (VQRF_codebookUpdate	*) malloc(0);
  numUpdates = 0;
  
  cbUpdate = VQRF_ReadCodebookUpdate(f, NULL);
  while (cbUpdate.numEntries != 0) {
    numUpdates++;
    updateArray = realloc((char *) updateArray, numUpdates * sizeof(VQRF_codebookUpdate));
    updateArray[numUpdates-1] = cbUpdate;
    cbUpdate = VQRF_ReadCodebookUpdate(f, NULL);
  }
  result.numCodebookUpdates = numUpdates;
  result.codebookUpdates = updateArray;

  if (frameData == NULL) {
    frameData = &result.frameData;
    result.frameData = NULL;
  }
  result.frameSize = VQRF_ReadFrameData(f, frameData);
  result.frameData = *frameData;
  
  return result;
}

void VQRF_FreeFrameHunk(VQRF_frameHunk hunk)
{
  int i;

  for (i = 0; i < hunk.numCodebookUpdates; i++) {
    free((char *) hunk.codebookUpdates[i].data);
  }
  free((char *) hunk.codebookUpdates);
  free(hunk.frameData);
}

VQRF_codebookUpdate VQRF_ReadCodebookUpdate(VQRF_file *f, char **codebookData)
{
  VQRF_codebookUpdate	result;
  int			bytesToRead;
  int			bytesRead;

  result.numEntries = VQRF_ReadUint32(f);
  bytesToRead = result.numEntries * 16;
    
  if (result.numEntries > 0) {
    result.firstEntryIndex = VQRF_ReadUint32(f);
    if (codebookData == NULL) {
      result.data = (uint32 *) malloc(bytesToRead);
    }
    else if (*codebookData == NULL) {
      *codebookData = (char *) result.data = (uint32 *) malloc(bytesToRead);
    }
    else
      result.data = (uint32 *) codebookData;

    bytesRead = fread(result.data, 1, bytesToRead, VQRF_GetFilePtr(f));
  }
  else {
    result.firstEntryIndex = -1;
    result.data = NULL;
  }

  return result;
}

int VQRF_ReadFrameData(VQRF_file *f, char **frameData)
{
  int bytesToRead;
  int bytesRead;
  
  bytesToRead = VQRF_GetHeader(f).frameSize;

  if (frameData == NULL) {
    return 0;
  }
  if (*frameData == NULL) {
    *frameData = (char *) malloc(bytesToRead);
    if (*frameData == NULL) {
      fprintf(stderr, "VQRF_ReadFrameData: couldn't alloc frame data (%d bytes)\n",
	      bytesToRead);
      return -1;
    }
  }
  bytesRead = fread(*frameData, 1, bytesToRead, VQRF_GetFilePtr(f));
  if (bytesRead != bytesToRead) {
    fprintf(stderr, "VQRF_ReadFrameData: tried to read %d bytes, but only got %d bytes\n",
	    bytesToRead, bytesRead);
    return -1;
  }

  return bytesRead;
}
