/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*--------------------------------------------------------------*\
 *	edClip.c
 * author:	Eric14@cs.berkeley.edu
 * date:	12-02-94
 *
 * purpose:  <See edClip.h>
 *
\*--------------------------------------------------------------*/

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include "vwGeneral.h"
#include <ctype.h>
#include <tk.h>

#include "vwClip.h"

/*--------------------------------------------------------------*/

int
Clip_Copy (cTI* ti, tClip* self, tClip** retVal)
{
    Tcl_DString	clipInfo;
    tClip*	newClip;
    int		err;

    if (Clip_AsText(ti,self,kGen_BadTime,kGen_BadTime,kGen_BadTime) != TCL_OK) 
	return TCL_ERROR;
    Tcl_DStringInit(&clipInfo);
    Tcl_DStringGetResult(ti,&clipInfo);

    err= Clip_Create(ti, Tcl_DStringValue(&clipInfo), &newClip);
    Tcl_DStringFree(&clipInfo);
    if (!err)
    {
	ASSERT(retVal != NULL);
	*retVal = newClip;
    }
    return err;
}

/*--------------------------------------------------------------*/

int
Clip_AsText (cTI* ti, tClip* self, double start, double end, double base)
{
    Tcl_DString		out;
    double		ls, le, ss, se, ratio;
    
    if (Tcl_VarEval(ti,"CMApp_RemoveProp {", self->plist, "} _*", NULL)) {
	return TCL_ERROR;
    }
    ss = self->sStart, se = self->sEnd;
    ls = self->lStart, le = self->lEnd;
    ratio = (se-ss)/(le-ls);
    if (start != kGen_BadTime && start > ls) {
	ASSERT(le > start);
	ss += (start-ls)*ratio;
	ls = start;
    }
    if (end != kGen_BadTime && end < le) {
	ASSERT(ls < end);
	se += (end-le)*ratio;
	le = end;
    }
    if (base != kGen_BadTime) {
	le = base + le - ls;
	ls = base;
    }
	
    Tcl_DStringInit(&out);
    Gen_Format(&out, "%f %f %f %f%!s%!s%!s\n    %!s", 
	       ss, se, ls, le,
	       self->name, self->format,
	       ti->result, self->addrTupple);
    Tcl_DStringResult(ti,&out);
    return TCL_OK;
}

/*--------------------------------------------------------------*/

int
Clip_Type (cTI* ti, tClip* self)
{
    char*	p;

    Gen_Result(ti,"%s", self->format);
    p = strchr(ti->result,'/');
    if (p)  *p = '\0';
    return TCL_OK;
}

/*--------------------------------------------------------------*/

int
Clip_Trim (cTI* ti, tClip* self, double ls, double le)
{
    double  ratio = (self->sEnd-self->sStart)/(self->lEnd-self->lStart);

    ASSERT(ls < le);

    if (self->lStart != ls) {
	self->sStart += (ls-self->lStart)*ratio;
	self->lStart = ls;
    }
    if (self->lEnd != le) {
	self->sEnd += (le-self->lEnd)*ratio;
	self->lEnd = le;
    }
    self->flags = 0;

    /*DIAG("        -> <%f %f %f %f>"C_
	 self->sStart C_ self->sEnd C_ self->lStart C_ self->lEnd);*/
    return TCL_OK;
}

/*--------------------------------------------------------------*/

int
Clip_SourceShift (cTI* ti, tClip* self, double sDelta)
{
    self->sStart += sDelta;
    self->sEnd += sDelta;
    self->flags = 0;
    return TCL_OK;
}

/*--------------------------------------------------------------*/

int
Clip_SetProp (cTI* ti, tClip* self, char* property, char* value)
{
    Tcl_DString		cmd;
    int			err;

    Tcl_DStringInit(&cmd);
    Gen_Format(&cmd,"CMApp_SetProp %!s%!s%!s", self->plist, property, value);
    err = Tcl_Eval(ti,Tcl_DStringValue(&cmd));
    Tcl_DStringFree(&cmd);
    if (err)
	return TCL_ERROR;

    Gen_ForgetPtr(self->plist);
    self->plist = strdup(ti->result);
    ASSERT(self->plist);
    return TCL_OK;
}

/*--------------------------------------------------------------*/

int
Clip_GetProp (cTI* ti, tClip* self, char* property)
{
    Tcl_DString		cmd;
    int			err;

    Tcl_DStringInit(&cmd);
    Gen_Format(&cmd,"CMApp_GetProp %!s%!s", self->plist, property);
    err = Tcl_Eval(ti,Tcl_DStringValue(&cmd));
    Tcl_DStringFree(&cmd);
    return err;
}

/*--------------------------------------------------------------*/

int
Clip_Shift (cTI* ti, tClip* self, double shiftT)
{
    self->lStart += shiftT;
    self->lEnd += shiftT;
    self->flags = 0;
    return TCL_OK;
}

/*--------------------------------------------------------------*/

int
Clip_Destroy (cTI* ti, tClip* self)
{
    Gen_ForgetPtr(self->name);
    Gen_ForgetPtr(self->format);
    Gen_ForgetPtr(self->plist);
    Gen_ForgetPtr(self->addrTupple);
    free(self);
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Creates a clip from a fully resolved clip specification.  Returns
 * TCL_ERROR with a result string if error occures.  Resulting clips
 * stored in *out.
 */

int
Clip_Create (cTI* ti, char* inStr, tClip** out)
{
    double ss, se, ls, le;
    tClip* self = NULL;

    if (sscanf(inStr,"%lf %lf %lf %lf", &ss, &se, &ls, &le) != 4)
	return Gen_Result(ti,"#1Badly formated clip times");
    if (ls >= le)
	return Gen_Result(ti,"#1Logical start >= end: %f >= %f", ls, le);

    self = (tClip*) malloc(sizeof(tClip));
    Gen_CkError(!self,"Malloc failure, self");
    memset(self,0,sizeof(tClip));

    Tcl_VarEval(ti, "lindex {", inStr, "} 4", NULL);
    self->name = strdup(ti->result);
    Tcl_VarEval(ti, "lindex {", inStr, "} 5", NULL);
    self->format = strdup(ti->result);
    Tcl_VarEval(ti, "lindex {", inStr, "} 6", NULL);
    self->plist = strdup(ti->result);
    Tcl_VarEval(ti, "lindex {", inStr, "} 7", NULL);
    self->addrTupple = strdup(ti->result);
    if (!self->name || !self->format || !self->plist || !self->addrTupple ) {
	Gen_ForgetPtr(self->name);
	Gen_ForgetPtr(self->format);
	Gen_ForgetPtr(self->plist);
	Gen_ForgetPtr(self->addrTupple);
	free(self);
	return Gen_Result(ti, "#1Malloc failure, data");
    }

    self->sStart = ss;
    self->sEnd = se;
    self->lStart = ls;
    self->lEnd = le;
    self->flags = 0;

    ASSERT(out != NULL);
    *out = self;

    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* End of File */
