/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*--------------------------------------------------------------*\
 *	general.c
 * author:	Eric14@cs.berkeley.edu
 * date:	10-24-94
 *
 * purpose:  See "vwGeneral.h" - General helping routines.
 *
\*--------------------------------------------------------------*/

#include <tcl.h>
#include "vwGeneral.h"

typedef void (tOutProc) (ClientData dest, char* str);

char	gGen_ErrorStr[1000];

/*--------------------------------------------------------------*/
/* Mainly here so that diagnostics can be traced
 */

void
Gen_ScreamOut (void)
{
    printf("##SCREAM: %s\n", gGen_ErrorStr);
}

/*--------------------------------------------------------------*/
/* Mainly here so that diagnostics can be traced
 */

void
Gen_DieOut (void)
{
    printf("##SCREAM: %s\n", gGen_ErrorStr);
}

/*--------------------------------------------------------------*/

void
Gen_Assert (char* exp, char* file, int line)
{
    fprintf(stderr,"%s:%d: %s\n", file, line, exp);
    exit(1);
}

/*--------------------------------------------------------------*/
/* Mainly here so that diagnostics can be traced
 */

void
Gen_DiagOut (void)
{
    printf("##DEBUG: %s\n", gGen_ErrorStr);
}

/*--------------------------------------------------------------*/

char*
Gen_HashGet (Tcl_HashTable* tab, char* key)
{
    Tcl_HashEntry*	he = Tcl_FindHashEntry(tab,key);
    if (he == NULL)
	return NULL;
    return (char*) Tcl_GetHashValue(he);
}

/*--------------------------------------------------------------*/

int
Gen_HashSet (Tcl_HashTable* tab, char* key, void* val)
{
    int			okF;
    Tcl_HashEntry*	he = Tcl_CreateHashEntry(tab,key,&okF);
    if ( !he ) {
	SCREAM("Gen_HashSet: Badness");
	return TCL_ERROR;
    }
    Tcl_SetHashValue(he,val);
    return TCL_OK;
}

/*--------------------------------------------------------------*/

int
Gen_HashCopy (Tcl_HashTable* destT, Tcl_HashTable* srcT)
{
    Tcl_HashEntry*	he;
    Tcl_HashSearch	hs;

    for (he= Tcl_FirstHashEntry(srcT,&hs); he; he= Tcl_NextHashEntry(&hs))
    {
	char*	key = Tcl_GetHashKey(srcT,he);
	void*	val = Tcl_GetHashValue(he);

	if ( Gen_HashSet(destT,key,val) ) {
	    SCREAM("Gen_HashCopy: failure!");
	    return TCL_ERROR;
	}
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Given a tcl object command name, returns the client data for that
 *     command.  This is the object pointer in most cases.
 */

ClientData
Gen_GetCommandData(cTI* ti, char* cmdName)
{
    Tcl_CmdInfo	 infoPtr;

    if (Tcl_GetCommandInfo(ti,cmdName,&infoPtr) == 0)
	return 0;

    return infoPtr.clientData;
}

/*--------------------------------------------------------------*/
/* PrintF like function for building function results.
 * Assumes dstr is init'ed.  Appends to the end (see '%-')
 * %x! --> Apply format 'x' as add element, not append
 * %s --> string
 * %x --> hex int
 * %d --> decimal int
 * %f --> double - but kGen_BadTime --> "BAD_VALUE"
 */

static int
L_VAFormat (ClientData dest, tOutProc strProc, tOutProc elemProc,
	    char* format, va_list* ap)
{
    while ( *format != '\0' ) {
	int	elemFlag = FALSE;
	char	txt[40];
	char*	nxtStr = txt;
	char	op;

	for (txt[1] = '\0'; *format && *format!='%'; format++) {
	    txt[0] = *format;
	    strProc(dest,txt);
	}
	    
	if (*format != '%')
	    break;

	op = format[1], format += 2;
	if (op == '!')
	    op = *format, format++, elemFlag = TRUE;

	switch (op) {
/*	  case '-':
 *	    Tcl_DStringFree(dstr); break;
 *	  case '+':
 *	    nxtStr = ti->result;
 *	    break;
 */
	  case 's':
	    nxtStr = va_arg((*ap),char*); 
	    break;
	  case 'd':
	    sprintf(nxtStr,"%d", va_arg((*ap), int)); 
	    break;
	  case 'g': 
	  case 'f': {
	      double dval = va_arg((*ap), double );
	      if (dval == kGen_BadTime)
		  nxtStr = "BAD_VALUE";
	      else
		  sprintf(nxtStr,"%g", dval); 
	      break;
	  }
	  case 'x':
	    sprintf(nxtStr,"%x", va_arg((*ap), int)); 
	    break;
	  default:
	    DIE("Gen_VAFormat: Bad format '%c'"C_ op);
	}

	if (elemFlag)
	    elemProc(dest,nxtStr);
	else
	    strProc(dest,nxtStr);
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* See Gen_VAFormat 
 * Also any number of '#<c>' pairs (where <c> is a character) may be
 * prepended to the format string.  These are processed as commands.
 *  #0 - Causes the return to be zero (TCL_OK) (the default).
 *  #1 - cause return to be 1 (TCL_ERROR).
 *  #+ - Append the result of this command to existing value in format
 *       target.
 *  #- - don't append (the default).
 *  ## - Output a '#' ignore following '#'s.
 */


static void
L_AppStr2Res(ClientData dest, char* str)
{
    Tcl_AppendResult((cTI*) dest,str,NULL);
}

static void
L_AppElem2Res(ClientData dest, char* str)
{
    Tcl_AppendElement((cTI*) dest,str);
}

int
Gen_VAResult (cTI* ti, char* format, va_list* ap)
{
    int			out = TCL_OK, append= FALSE;

    while (*format == '#') {
	format++;
	switch (*format) {
	  case '0': out = TCL_OK; break;
	  case '1': out = TCL_ERROR; break;
	  case '+': append = TRUE; break;
	  case '-': append = FALSE; break;
	  default: goto EndLoop;
	}
	format++;
    }
  EndLoop:
    if (!append) {
	Tcl_ResetResult(ti);
    }

    L_VAFormat(ti,L_AppStr2Res,L_AppElem2Res,  format,ap);

    return out;
}

/*--------------------------------------------------------------*/
/* See Gen_VAResult
 */

int
Gen_Result (cTI* ti, char* format, ...)
{
    int			out;
    va_list		ap;

    va_start(ap,format);
    out = Gen_VAResult(ti,format,&ap);
    va_end(ap);

    return out;
}

/*--------------------------------------------------------------*/
/* See Gen_VAFormat
 */

static void
L_AppStr2DStr(ClientData dest, char* str)
{
    Tcl_DStringAppend((Tcl_DString*) dest,str,-1);
}

static void
L_AppElem2DStr(ClientData dest, char* str)
{
    Tcl_DStringAppendElement((Tcl_DString*) dest,str);
}


int
Gen_Format (Tcl_DString* dstr, char* format, ...)
{
    int			out = TCL_OK, append = FALSE;
    va_list		ap;

    while (*format == '#') {
	format++;
	switch (*format) {
	  case '0': out = TCL_OK; break;
	  case '1': out = TCL_ERROR; break;
	  case '+': append = TRUE; break;
	  case '-': append = FALSE; break;
	  default: goto EndLoop;
	}
	format++;
    }
  EndLoop:
    if (!append) {
	Tcl_DStringFree(dstr);
    }

    va_start(ap,format);
    L_VAFormat(dstr,L_AppStr2DStr,L_AppElem2DStr,  format,&ap);
    va_end(ap);

    return out;
}

/*--------------------------------------------------------------*/
/* End of File */
