/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*--------------------------------------------------------------*\
 *	vwStream.c
 * author:	Eric14@cs.berkeley.edu
 * date:	12-02-94
 *
 * purpose:  <See vwStream.h>
 *
\*--------------------------------------------------------------*/

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */


#include "vwGeneral.h"
#include <ctype.h>
#include <tk.h>
#include "otcl.h"

#include "darray.h"
#include "vwClip.h"

/*--------------------------------------------------------------*/

#define L_StreamGet(_s,_n)  	((tClip*) DA_Get(&((_s)->data->clips),(_n)))
#define L_StreamNum(_s)		(DA_Num(&((_s)->data->clips)))

/* 
 * Compensate for OTCL wierd argument padding
 */
#define L_ARG(_n)		(argv[(_n)+4])
#define L_ARGC()		(argc-4)
#define L_ARG_POP(_n) \
	    {int i= 4+(_n); for (argc--; i<argc; i++) argv[i]=argv[i+1];}

/* 
 * builds the full hook variable name from the hook name
 */
#define L_HOOK(_n)		    _n "_Hooks"
#define L_EXEC_HOOK(_t,_s,_h,_c)    L_StreamExecHook(_t,_s,_h,L_HOOK(_h),_c)

/* 
 * the stream class, save lookups.
 */
struct OTclClass* gStreamC;


typedef struct {
    tDArray	clips;
} tStreamData;

typedef struct {
    struct OTclClass*	class;
    tStreamData*	data;
    struct OTclObject*	self;
}  tStreamI;

/*--------------------------------------------------------------*/
/*--- Local Functions ------------------------------------------*/
/*--------------------------------------------------------------*/

static int
L_StreamExecHook (cTI* ti, tStreamI* sP, 
		  char* hookN, char* hookVarN, int clipNum)
{
    char*	hook;

    if ((hook= OTclGetInstVar(sP->self,ti,hookVarN,0))!= NULL &&
	hook[0] != '\0')
    {
	char*	name = OTclGetObjectName(sP->self,ti);
	if (clipNum >= 0)
	    Gen_Result(ti,"%s _exechook %s %s %d", name,hookN,name,clipNum);
	else
	    Gen_Result(ti,"%s _exechook %s %s", name,hookN,name);
	if (Tcl_VarEval(ti,ti->result,NULL))
	    return TCL_ERROR;
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Note this may be called during the destruction of a non-fully
 * instatiated object.
 */

static void
L_StreamMethPrep (cTI* ti, ClientData objID, tStreamI* sP)
{
    ClientData	data= NULL;

    sP->class = gStreamC;
    sP->self = OTclAsObject(ti,objID);
    OTclGetObjectData(sP->self, sP->class, &data);
	
    sP->data = (tStreamData*) data;
}

/*--------------------------------------------------------------*/

static double
L_StreamStart (tStreamI* sP)
{
    return (L_StreamNum(sP) > 0) ? 
	L_StreamGet(sP,0)->lStart : 0.0;
}   

/*--------------------------------------------------------------*/

static double
L_StreamEnd (tStreamI* sP)
{
    int		num = L_StreamNum(sP);
    return (num > 0) ? L_StreamGet(sP,num-1)->lEnd : 0.0;
}   

/*--------------------------------------------------------------*/
/* does bounds checking */

static tClip*
L_StreamGetCk (cTI* ti, tStreamI* sP, int idx)
{
    int		num = L_StreamNum(sP);

    if (idx < 0 || idx >= num)
    {
	if (num == 0) {
	    Gen_Result(ti,"clip index %d out of bounds (empty stream)");
	} else {
	    Gen_Result(ti,"clip index %d out of bounds (0 - %d)", 
		       idx, num-1);
	}
	return NULL;
    }
    return ((tClip*) DA_Get(&sP->data->clips,idx));
}

/*--------------------------------------------------------------*/

static int
L_StreamGetType (cTI* ti, tStreamI* sP, char** out)
{
    *out = OTclGetInstVar(sP->self,ti,"type", TCL_LEAVE_ERR_MSG);
    return (out == NULL) ? TCL_ERROR : TCL_OK;
}

/*--------------------------------------------------------------*/

static int
L_StreamDeleteClip (cTI* ti, tStreamI* sP, int idx)
{
    tClip*	clip;
    char*	hook;

    if ( (clip = L_StreamGetCk(ti,sP,idx)) == NULL)
	return TCL_ERROR;

    if (L_EXEC_HOOK(ti,sP,"deleteclip",idx))
	return TCL_ERROR;
    DA_Del(&sP->data->clips,idx);
    Clip_Destroy(ti,clip);
    return TCL_OK;
}

/*--------------------------------------------------------------*/

static int
L_StreamClear (cTI* ti, tStreamI* sP)
{
    int		i;

    for (i= L_StreamNum(sP)-1; i >= 0; i--)
	if (L_StreamDeleteClip(ti,sP,i))
	    return TCL_ERROR;
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* For binary search of clip list
 */

static int
L_ClipTimeComp (tClip* c, double* targT)
{
    if (c->lStart >= *targT) {
	return -1;
    }
    if (c->lEnd <= *targT) {
	return 1;
    }
    return 0;
}

/*--------------------------------------------------------------*/
/* returns nearest clip's index, outClip set to clip if time
 * intersects a clip.
 */

static int
L_StreamFind (cTI* ti, tStreamI* sP, double time, tClip** outClip)
{
    int		idx, res;
    
    if (outClip)
	*outClip = NULL;
    res= DA_BinSearch(&sP->data->clips,&time,
		      (tDA_CompFunc)L_ClipTimeComp, &idx);
    if (res == TCL_OK && outClip)
	*outClip = L_StreamGet(sP,idx);
    return idx;
}

/*--------------------------------------------------------------*/
/* retIdx is set to the index of the first clip after the deleted
 * range (iff non-NULL).  This may be out of range!
 */

static int
L_StreamClearRange (cTI* ti, tStreamI* sP,
	       double startT, double endT, int* retIdx)
{
    int		idx, err;
    tClip*	c;

    ASSERT(startT <= endT);
    /*DIAG("Stream_Clear %f %f"C_ startT C_ endT);*/

    idx  = L_StreamFind(ti, sP, startT, &c);
    
    if (c != NULL && c->lStart < startT && c->lEnd > endT) {
        /* Case of a single clip spanning whole delete range */
	tClip*  c2;
	idx ++;
	if (Clip_Copy(ti,c,&c2) != TCL_OK)
	    return TCL_ERROR;
	err= DA_Add(&sP->data->clips,idx,c2);
	ASSERT(!err);
	Clip_Trim(ti,c,c->lStart,startT);
	Clip_Trim(ti,c2,endT,c2->lEnd);
	if (L_EXEC_HOOK(ti,sP,"copyclip",idx))
	    return TCL_ERROR;
    } else {
        /* Normal Case, 1st clip doesn't spanning whole range */
	if (c != NULL && c->lStart < startT) {
	    Clip_Trim(ti,c,c->lStart,startT);
	    idx ++;
	}
	
	while (idx < L_StreamNum(sP)) {
	    c = L_StreamGet(sP,idx);
	    if (c->lEnd > endT) {
		if (c->lStart < endT)
		    Clip_Trim(ti,c,endT,c->lEnd);
		break;
	    }
	    if (L_StreamDeleteClip(ti,sP,idx))
		return TCL_ERROR;
	}
    }
    if (retIdx != NULL)
	*retIdx = idx;
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Shifts all elements in stream after time by delta.  If delta is
 *   negative, this involves deleting clips.  Anything but shift 0
 *   splits the clip at time. A time of kGen_BadTime signals a shift
 *   of all clips.
 */

static int
L_StreamShift (cTI* ti, tStreamI* sP, double time, double deltaT)
{
    int		idx= 0;
    double	startT = (deltaT < 0) ? time+deltaT : time;

    if (deltaT == 0)
	return TCL_OK;
    if (time != kGen_BadTime)
    {
	if (L_StreamClearRange(ti, sP, startT, time, &idx))
	    return TCL_ERROR;
    }
    
    for (; idx < L_StreamNum(sP); idx++) {
	tClip*  c = L_StreamGet(sP,idx);
	Clip_Shift(ti,c,deltaT);
    }

    return TCL_OK;
}

/*--------------------------------------------------------------*/

static int
L_StreamAddResolvedClip (cTI* ti, tStreamI* sP, char* clipTxt, int* outPos)
{
    int		idx, err= TCL_OK;
    char* 	sType;
    tClip*	clip;

    if (Clip_Create(ti,clipTxt,&clip) != TCL_OK)
	return TCL_ERROR;

    err = Clip_Type(ti,clip);
    if (!err) {
	sType = OTclGetInstVar(sP->self,ti,"type", TCL_LEAVE_ERR_MSG);
	if (sType == NULL)  err = TCL_ERROR;
    }
    if (strcmp(ti->result,sType) != 0) {
	err = TCL_ERROR;
	Gen_Result(ti,"#1Clip has wrong type for trk");
    }
    Tcl_ResetResult(ti);
    if (!err && clip->lStart < L_StreamEnd(sP))
	err= L_StreamClearRange(ti,sP,clip->lStart,clip->lEnd,&idx);
    else
	idx= L_StreamNum(sP);
    if (!err)
	err= DA_Add(&sP->data->clips,idx,clip);
    
    if (err)  
	Clip_Destroy(ti,clip);
    else
	err = L_EXEC_HOOK(ti,sP,"addclip",idx);
    if (outPos != NULL)		*outPos = idx;
    return err;
}

/*--------------------------------------------------------------*/

static int
L_StreamGetClipList (cTI* ti, tStreamI* sP, double start, double end, 
		     double base)
{
    int		idx, cnt;
    Tcl_DString	out;
    tClip*	clip;

    Tcl_DStringInit(&out);

    if (start == kGen_BadTime)
	start = L_StreamStart(sP);
    if (base == kGen_BadTime)
	base = start;
    idx = L_StreamFind(ti,sP,start,NULL);

    for (cnt= 0; idx < L_StreamNum(sP); idx++, cnt++) {
	clip = L_StreamGet(sP,idx);
	if (end != kGen_BadTime && clip->lStart >= end) 
	    break;
	if (clip->lStart > start) {
	    base += clip->lStart - start;
	    start = clip->lStart;
	}

	Tcl_ResetResult(ti);
	Clip_AsText(ti, clip,start,end,base);
	if (cnt != 0) Tcl_DStringAppend(&out,"\n  ",-1);
	Tcl_DStringAppendElement(&out,ti->result);
    }
    Tcl_DStringResult(ti,&out);
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/*--- OTcl Stream Methods --------------------------------------*/
/*--------------------------------------------------------------*/

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc foreach {start end streamVar clipVar script}
 *
 * Executes $script for each clip in the given time range. Within the
 * script $streamVar is set to the name of the stream object currently
 * executing.  $clipVar is set to the clip index of the current clip.
 */

int
Stream_ForEachMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int		idx;
    char	num[20];
    double	start= kGen_BadTime, end;
    int		retIdx, err;
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=5,
        "usage: foreach <start> <end> <streamName> <indexName> <script>");
    if (strcmp(L_ARG(0),"start")!=0)
	Gen_GetDouble(&start, L_ARG(0));
    if (strcmp(L_ARG(1),"end")!=0) {
	Gen_GetDouble(&end, L_ARG(1));
    } else
	end = L_StreamEnd(&strmI);
    Tcl_SetVar(ti,L_ARG(2),argv[0],0);

    idx = (start == kGen_BadTime) ? 0 :  L_StreamFind(ti,&strmI,start,NULL);
    for (; idx < L_StreamNum(&strmI); idx++) {
	tClip*  c = L_StreamGet(&strmI,idx);

	if (c->lStart >= end)
	    break;
	sprintf(num,"%d", idx);
	Tcl_SetVar(ti,L_ARG(3),num,0);
	err = Tcl_Eval(ti, L_ARG(4));
	if (err == TCL_CONTINUE)
	    continue;
	if (err == TCL_RETURN)
	    return TCL_OK;
	if (err != TCL_OK) {
	    char msg[100];
	    sprintf(msg, "\n    (\"foreach\" body line %d)",
		    ti->errorLine);
	    Tcl_AddErrorInfo(ti, msg);
	    return TCL_ERROR;
	} 
    }
    return TCL_OK;
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc getstart {}
 *
 * Returns the logical start time of the first clip in the stream.
 * returns 0.0 if the stream is empty.
 */


int
Stream_GetStartMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    return Gen_Result(ti,"%f", L_StreamStart(&strmI));
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc getend {}
 *
 * Returns the logical end time of the last clip in the stream.
 * returns 0.0 if the stream is empty.
 */

int
Stream_GetEndMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    return Gen_Result(ti,"%f", L_StreamEnd(&strmI));
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc getnumclips {}
 *
 * Returns the number of clips in the stream.
 */

int
Stream_GetNumClipsMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    return Gen_Result(ti,"%d", L_StreamNum(&strmI));
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc deleteclip {clipNum}
 *
 * Destroys the given clip.
 */

int
Stream_DeleteClipMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int		idx;
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=1,"usage: deleteclip <index>");
    Gen_GetInt(&idx, L_ARG(0));
    
    return L_StreamDeleteClip(ti,&strmI,idx);
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc getclip {clipNum {base none}}
 *
 * returns a textual version of the requested clip.  This is in
 * resolved clip format.
 */

int
Stream_GetClipMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int		idx;
    tStreamI	strmI;
    tClip*	clip;
    double	base= kGen_BadTime;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=1 && L_ARGC()!=2,"usage: getclip <index> ?base?");
    Gen_GetInt(&idx, L_ARG(0));
    if (L_ARGC() > 1 && strcmp(L_ARG(1),"none")!=0)
	Gen_GetDouble(&base, L_ARG(1));
    if ((clip = L_StreamGetCk(ti,&strmI,idx)) == NULL)
	return TCL_ERROR;
    return Clip_AsText(ti,clip,kGen_BadTime,kGen_BadTime,base);
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc clipflag {clipNum flagNum {flagValue {}}}
 *
 * Gets or sets the requested change flag.  A value of 1 indicates
 * that the clip has been changed since last reset to zero (with this
 * call).
 */

int
Stream_ClipFlagMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int		idx, flagNum, value;
    tStreamI	strmI;
    tClip*	clip;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=2 && L_ARGC()!=3,
		"usage: clipflag <clipNum> <flagNum> ?Boolean?");
    Gen_GetInt(&idx, L_ARG(0));
    Gen_GetInt(&flagNum, L_ARG(1));
    Gen_CkError(flagNum>31 || flagNum<0, "<flagNumber> out of range");
    if ((clip = L_StreamGetCk(ti,&strmI,idx)) == NULL)
	return TCL_ERROR;
    if (L_ARGC() > 2) 
    {
	Gen_GetBoolean(&value,L_ARG(2));
	if (value)
	    clip->flags &= (~1L<<flagNum);
	else
	    clip->flags |= (1L<<flagNum);
    }
    Gen_Result(ti,"%d", (clip->flags&(1L<<flagNum)) == 0);
    return TCL_OK;
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc clipprop {clipNum propName {Value {}}}
 *
 * Gets or sets the given property of the given clip.
 */

int
Stream_ClipPropMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int		idx;
    tStreamI	strmI;
    tClip*	clip;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=2 && L_ARGC()!=3,
		"usage: clipprop <clipNum> <propName> ?Value?");
    Gen_GetInt(&idx, L_ARG(0));
    if ((clip = L_StreamGetCk(ti,&strmI,idx)) == NULL)
	return TCL_ERROR;
    if (L_ARGC() > 2) 
	return Clip_SetProp(ti,clip,L_ARG(1),L_ARG(2));
    else
	return Clip_GetProp(ti,clip,L_ARG(1));
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc snap {time dist}
 *
 * Takes <time> and <dist> and returns the time of the closest clip
 * boundy within <dist> of time.  If there is none, then time is
 * returned.  Dist is allowed to be negative (takes abs value).
 */

int
Stream_SnapMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    double	time, dist, out, tt;
    int		idx;
    tClip*	clip;
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=2,"usage: snap <time> <dist>");
    Gen_GetDouble(&time, L_ARG(0));
    Gen_GetDouble(&dist, L_ARG(1));

    out= time;
    if (dist < 0.0) dist = -dist;

    idx = L_StreamFind(ti,&strmI,time,&clip);
    if (clip != NULL) {
	if (time - clip->lStart < dist) 
	    out = clip->lStart, dist = time - clip->lStart;
	if (clip->lEnd - time < dist) {
	    out  = clip->lEnd, dist = clip->lEnd - time;
	}
    } else {
	if (idx > 0) {
	    tt = L_StreamGet(&strmI,idx-1)->lEnd;
	    if (time - tt < dist)
		out = tt, dist= time - tt;
	}
	if (idx < L_StreamNum(&strmI)) {
	    tt = L_StreamGet(&strmI,idx)->lStart;
	    if (tt - time < dist)
		out = tt, dist= tt - time;
	}
    }
	    
    Gen_Result(ti,"%f", out);
    return TCL_OK;
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc find {time}
 *
 * returns the index of the next clip at or after time and a boolean
 * true if time is in a clip.
 */

int
Stream_FindMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    double	start;
    int		idx;
    tClip*	clip;
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=1,"usage: find <time>");
    Gen_GetDouble(&start, L_ARG(0));
    
    idx = L_StreamFind(ti,&strmI,start,&clip);
    Gen_Result(ti,"%d %d", idx, clip!=NULL);
    return TCL_OK;
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc clear {start end}
 *
 * Deletes all clips within the given time range.  Clips which are
 * partially within the range are cropped.  If the start and end times
 * are identical, this function will split any clip that spans that
 * time.
 */

int
Stream_ClearMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    double	start, end;
    int		retIdx, err;
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=2,"usage: clear <start> <end>");
    if (strcmp(L_ARG(0),"start")!=0) {
	Gen_GetDouble(&start, L_ARG(0));
    } else
	start = L_StreamStart(&strmI);
    if (strcmp(L_ARG(1),"end")!=0) {
	Gen_GetDouble(&end, L_ARG(1));
    } else
	end = L_StreamEnd(&strmI);
    
    err = L_StreamClearRange(ti,&strmI,start,end,&retIdx);
    if (!err) 
    {
	Tcl_ResetResult(ti);
	sprintf(ti->result, "%d", retIdx);
    }
    return err;	
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc shift {start delta}
 *
 * Moves all clips on or after $start time by $delta time units
 * (e.g. adds delta to the logical start and end times).  If delta is
 * negative, clips in the region behind $start are deleted. Clips
 * which span $start are split in two.  $start may equal "start".
 */

int
Stream_ShiftMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    double	start, delta;
    tStreamI	strmI;

    L_StreamMethPrep(ti,objID,&strmI);
    Gen_CkError(L_ARGC()!=2,"usage: shift <start> <delta>");
    if (strcmp(L_ARG(0),"start")!=0) {
	Gen_GetDouble(&start, L_ARG(0));
    } else
	start = kGen_BadTime;
    Gen_GetDouble(&delta, L_ARG(1));
    
    return L_StreamShift(ti,&strmI,start,delta);
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc getcliplist {{start start} {end end} {base none}}
 *
 * Gets the cliplist of the stream in text format. Only the specified
 * time range from the clip list is returned.  If base is given, then
 * the times of the clips are shifted to start at the given time.
 */

int
Stream_GetClipListMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    double	start= kGen_BadTime, end= kGen_BadTime, base= kGen_BadTime;
    tStreamI	strmI;
    L_StreamMethPrep(ti,objID,&strmI);

    Gen_CkError(L_ARGC() > 3, "usage: getrangelist ?start? ?end? ?base?");
    if (L_ARGC() > 0 && strcmp(L_ARG(0),"start")!=0)
	Gen_GetDouble(&start, L_ARG(0));
    if (L_ARGC() > 1 && strcmp(L_ARG(1),"end")!=0)
	Gen_GetDouble(&end, L_ARG(1));
    if (L_ARGC() > 2 && strcmp(L_ARG(2),"none")!=0)
	Gen_GetDouble(&base, L_ARG(2));
    return L_StreamGetClipList(ti,&strmI,start,end,base);
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc setcliplist {clipList}
 *
 * Gets or sets the cliplist of the stream from/to text format.  Each
 * clip in the clip list is in fully resolved format.
 */

int
Stream_SetClipListMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int		i;
    int		topC;
    char**	topV;
    tStreamI	strmI;
    L_StreamMethPrep(ti,objID,&strmI);

    Gen_CkError(L_ARGC() != 1, "usage: setcliplist clipList");

    if (Tcl_SplitList(ti,L_ARG(0),&topC,&topV) != TCL_OK) {
	Gen_Result(ti,"#1Split of top list failed!");
	return TCL_ERROR;
    }

    L_StreamClear(ti,&strmI);
    for (i = 0; i < topC; i++) {
	if (L_StreamAddResolvedClip(ti,&strmI,topV[i],NULL) != TCL_OK) {
	    free(topV);
	    /* L_StreamClear(ti,&strmI); */
	    return TCL_ERROR;
	}
    }
    
    free(topV);
    return TCL_OK;
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc addresolvedclip {resolvedClip}
 *
 * As addclip, but only takes the resolved format.
 */


int
Stream_AddResolvedClipMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int		clipNum;
    tStreamI	strmI;
    L_StreamMethPrep(ti,objID,&strmI);

    Gen_CkError(L_ARGC() != 1, "usage: <resolvedClip>");
    if (L_StreamAddResolvedClip(ti,&strmI,L_ARG(0),&clipNum))
	return TCL_ERROR;
    return Gen_Result(ti,"%d", clipNum);
}

/*----{c2doc TCL_PROC}------------------------------------------*/
/* Stream instproc destroy {}
 * 
 * Destroys the stream.
 */

int
Stream_DestroyMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int		err;
    tStreamI	strmI;
    L_StreamMethPrep(ti,objID,&strmI);
    
    if (strmI.data)
    {
	if (L_EXEC_HOOK(ti,&strmI,"destroystream",-1))
	    return TCL_ERROR;
	if (L_StreamClear(ti,&strmI))
	    return TCL_ERROR;
    
	DA_Nullify(&strmI.data->clips);
	free(strmI.data);
    }
    return OTclNextMethod(strmI.self,ti,argc,argv);
}

/*--------------------------------------------------------------*/

int
Stream_InitMeth (ClientData objID, cTI* ti, int argc, char* argv[])
{
    int			i, ok, err;
    struct OTclObject*	self = OTclAsObject(ti,objID);
    tStreamData* 	myData = NULL;
    Tcl_DString		cmd;

    Gen_CkError(argc < 5, "usage: create <name> <type> ?switches?...");

    Tcl_DStringInit(&cmd);
    Gen_Format(&cmd,"[lsearch [Stream set kTypes] %s] != -1", L_ARG(0));
    err = Tcl_ExprBoolean(ti,Tcl_DStringValue(&cmd),&ok);
    Tcl_DStringFree(&cmd);
    
    if (err || !ok) {
	Gen_Result(ti,"Unknown stream type: %s", L_ARG(0));
	goto error;
    }
    if (OTclSetInstVar(self,ti,"type",L_ARG(0),TCL_LEAVE_ERR_MSG) == NULL)
	goto error;
    if (OTclSetInstVar(self,ti,"numFlags","0",TCL_LEAVE_ERR_MSG) == NULL)
	goto error;

    myData = malloc(sizeof(tStreamData));
    if (myData == NULL)
	goto error;

    DA_Init(&myData->clips);
    OTclSetObjectData(objID, gStreamC, myData);

    /* remove <type> from arg list and call next */
    L_ARG_POP(0);
    if (OTclNextMethod(self,ti,argc,argv))
	goto error;
    return TCL_OK;

  error:
    OTclDeleteObject(ti,self);
    return TCL_ERROR;
}

/*--------------------------------------------------------------*/
/*--- Initialization Stuff -------------------------------------*/
/*--------------------------------------------------------------*/

int
Stream_Init (cTI* ti)
{
    if (gStreamC != NULL) 
	return TCL_OK;

    /* auto-load hack */
    if (Tcl_Eval(ti,"Stream info heritage")) {
	ti->result = "Failed to load stream class";
	return TCL_ERROR;
    }

    if ((gStreamC = OTclGetClass(ti,"Stream")) == NULL)
    {
	ti->result = "Stream class creation failed!";
	return TCL_ERROR;
    }
    OTclAddIMethod(gStreamC, "init", Stream_InitMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "destroy", Stream_DestroyMeth, NULL, NULL);

    OTclAddIMethod(gStreamC, "setcliplist", Stream_SetClipListMeth, NULL,NULL);
    OTclAddIMethod(gStreamC, "getcliplist", Stream_GetClipListMeth, NULL,NULL);
    OTclAddIMethod(gStreamC, "shift", Stream_ShiftMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "clear", Stream_ClearMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "find", Stream_FindMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "snap", Stream_SnapMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "foreach", Stream_ForEachMeth, NULL, NULL);

    OTclAddIMethod(gStreamC, "getnumclips", Stream_GetNumClipsMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "getstart", Stream_GetStartMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "getend", Stream_GetEndMeth, NULL, NULL);

    OTclAddIMethod(gStreamC, "addresolvedclip", 
		   Stream_AddResolvedClipMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "getclip", Stream_GetClipMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "clipflag", Stream_ClipFlagMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "clipprop", Stream_ClipPropMeth, NULL, NULL);
    OTclAddIMethod(gStreamC, "deleteclip", Stream_DeleteClipMeth, NULL, NULL);
    return TCL_OK;
}

int
VideoWidget_Init (cTI* ti)
{
    if (Stream_Init(ti)) {
	fprintf(stderr,"##ERROR - %s\n", ti->result);
	fprintf(stderr,"  The Video Widget subsystem can not be init'ed.\n");
	fprintf(stderr,"  Perhaps CMT_ROOT is badly set or\n");
	fprintf(stderr,"  the tclIndex was not built with cmtTclIndex.\n");
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* End of File */
