/* 
 * stdio.c
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

/*
 * Only three routines in this file currently support sockets:
 * Tknt_Fputc, Tknt_Fputs, and Tknt_Fgetc.  These were the functions
 * that were absolutely necessary to get Tcl-DP to work.  The
 * other functions should also be made to work with sockets.
 */

#include <windows.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include "util.h"

#ifdef WIN32S
#undef stdin
#undef stdout
#undef stderr

#define stdin  (FILE *)0
#define stdout (FILE *)1
#define stderr (FILE *)2
#endif

#define BUFFER_LENGTH 1024

/*
 * List of file descriptors and their Tcl-DP types.
 * This lists is used to intercept f* type calls (fprintf, feof, etc)
 * and replace them with Winsock calls.  The file descriptor needs to
 * be registered as a socket before it is properly handled.
 */

#define set_ferror(_stream) ((_stream)->_flag |= _IOERR)
#define set_feof(_stream) ((_stream)->_flag |= _IOEOF)

#define MAX_OPEN_FILES 1024
static unsigned char fdtype[MAX_OPEN_FILES];
static SOCKET fdToSocket[MAX_OPEN_FILES];

#define SOCKTYPE 1

/*
 *----------------------------------------------------------------------
 * Function: winio_registerSocket
 *----------------------------------------------------------------------
 */
void
Tknt_RegisterSocket(int fd, SOCKET socket)
{
    fdtype[fd] = SOCKTYPE;
    fdToSocket[fd] = socket;
}

/*
 *----------------------------------------------------------------------
 * Function: winio_unregisterSocket
 *----------------------------------------------------------------------
 */
void
Tknt_UnregisterSocket(int fd)
{
    fdtype[fd] = 0;
}

#define TK_READABLE  1
#define TK_WRITABLE  2
#define TK_EXCEPTION 4

/*
 *--------------------------------------------------------------
 *
 * BlockSocket --
 *
 *	This function is needed under Windows NT to get proper
 *	blocking behavior before with calls to recv().  This
 *	is because the call to WSAAsyncSelect() in tkEvent.c
 *	causes the socket to become non-blocking.  dpsh.exe does
 *	not have this problem because it uses an event loop that
 *	does not need to call WSAAsyncSelect().  The call to
 *	select() will only occur if the socket is a blocking socket.
 *	if the socket is not blocking, this function just returns.
 *
 * Results: None
 *
 * Side Effects:
 *	This function will return when an event occurs on the
 *	socket
 *
 *--------------------------------------------------------------
 */
static __inline void
BlockSocket(socket, mask)
    SOCKET socket;
    int mask;
{
    /* This next piece of code gets around a problem that occurs with
     * WSAAsyncSelect(): the socket becomes non-blocking when this call
     * is used.  Wait for select() to tell us something is available
     */
    fd_set ready[3];
    fd_set *readPtr, *writePtr, *exceptPtr;

    FD_ZERO(&ready[0]);
    FD_ZERO(&ready[1]);
    FD_ZERO(&ready[2]);

    if (mask & TK_READABLE) {
	FD_SET(socket, &ready[0]);
	readPtr = &ready[0];
    } else {
	readPtr = NULL;
    }

    if (mask & TK_WRITABLE) {
	FD_SET(socket, &ready[1]);
	writePtr = &ready[1];
    } else {
	writePtr = NULL;
    }

    if (mask & TK_WRITABLE) {
	FD_SET(socket, &ready[2]);
	exceptPtr = &ready[2];
    } else {
	exceptPtr = NULL;
    }

    select(1, readPtr, writePtr, exceptPtr, NULL);
}

static void 
wputs(const char *s, HWND hwnd)
{
    for(; *s != '\0'; s++) {
    	if  (*s	== '\n') {
	    SendMessage(hwnd, WM_USER+20, (WPARAM)'\r', 0L);
	} else {
	    SendMessage(hwnd, WM_USER+20, (WPARAM)*s, 0L);
	}
    }
}

int
Tknt_Printf(char * fmt, ...)
{
    va_list	arg_ptr;
    int ret;

    va_start(arg_ptr, fmt);
    
    if (Tknt_GetHasConsole()) {
	ret = vfprintf(stdout, fmt, arg_ptr);
	fflush(stdout);
	
    } else {
	char buffer[BUFFER_LENGTH + 1];
	va_list arg_ptr;

	va_start(arg_ptr, fmt);
	if((ret = vsprintf(buffer, fmt, arg_ptr)) > 0){
	    buffer[BUFFER_LENGTH] = '\0';
	    wputs(buffer, Tknt_GetPostWindow());
	}
    }

    va_end(arg_ptr);
    return ret;
} 

int 
Tknt_Fprintf(FILE *f, char *fmt, ...)
{
    va_list	arg_ptr;
    int ret;

    va_start(arg_ptr, fmt);
    if (Tknt_GetHasConsole()) {
	ret = vfprintf(f, fmt, arg_ptr);
	if (f == stdout || f == stderr) {
	    fflush(f);
	}
    } else {
	char buffer[BUFFER_LENGTH + 1];

	if (f == stdin) {
	    ret = -1;
	} else if (f == stdout || f == stderr) {  
	    ret = vsprintf(buffer, fmt, arg_ptr);
	    wputs(buffer, Tknt_GetPostWindow());
	} else {
	    ret = vfprintf(f, fmt, arg_ptr);
	}
    }
    va_end(arg_ptr);
    return ret;
}

int
Tknt_Fputc(int ch, FILE *f)
{
    int ret, error, fd;
    char c;

    if (!Tknt_GetHasConsole()) {
	if (f == stdout || f == stderr) {
	    SendMessage(Tknt_GetPostWindow(), WM_USER+20, (WPARAM)ch, 0L);
	    return ch;
	}
    }
    fd = fileno(f);

    switch (fdtype[fd]) {
    case SOCKTYPE:
	BlockSocket(fdToSocket[fd], TK_WRITABLE|TK_EXCEPTION);
	c = (char) ch;
	ret = send(fdToSocket[fd], &c, 1, 0);
	if (ret == 1) {
	    return ch;
	} else {
	    error = WSAGetLastError();
	    if (error != WSAEINTR && error != WSAEWOULDBLOCK) {
		set_feof(f);
	    } else {
		set_ferror(f);
	    }
	    return EOF;
	}
    default:
	ret = fputc(ch, f);
	if (f == stdout || f == stderr) {
	    fflush(f);
	}
    }
}

int
Tknt_Fputs(const char *s, FILE *f)
{
    int ret, len;
    int fd;

    if (!Tknt_GetHasConsole()) {
	if (f == stdout || f == stderr) {
	    wputs(s, Tknt_GetPostWindow());
	    return 1;
	}
    }
    fd = fileno(f);

    switch (fdtype[fd]) {
    case SOCKTYPE:
	BlockSocket(fdToSocket[fd], TK_WRITABLE|TK_EXCEPTION);
	len = strlen(s) + 1;
	ret = send(fdToSocket[fd], s, len, 0);
	if (ret != len) {
	    set_ferror(f);
	    return EOF;
	}
	return ret;
    default:
	ret = fputs(s, f);
	if (f == stdout || f == stderr) {
	    fflush(f);
	}
	return ret;
    }
}

int
Tknt_Puts(const char *s)
{
    Tknt_Fputs(s, stdout);
    return Tknt_Fputs("\n", stdout);
}

void 
Tknt_Clearerr(FILE *fp)
{
    if (!Tknt_GetHasConsole()) {
	if (fp == stdin || fp == stdout || fp == stderr) {
	    return;
	}
    }
    clearerr(fp);
}

      
int 
Tknt_Fileno(FILE *fp)		 
{
    if (! Tknt_GetHasConsole()) {
	if (fp == stdin) {
	    return 0;
	} else if (fp == stdout) {
	    return 1;
	} else if (fp == stderr) {
	    return 2;
	}
    }
    return fileno(fp);
}

int 
Tknt_Ferror(FILE *fp)
{
    if (! Tknt_GetHasConsole() ) {
	if (fp == stdin || fp == stdout || fp == stderr) {
	    return 0;
	}
    }
    return ferror(fp);
}

int 
Tknt_Feof(FILE *fp)
{
    if (! Tknt_GetHasConsole()) {
	if (fp == stdin || fp == stdout || fp == stderr) {
	    return 0;
	}
    }
    return feof(fp);
}

int 
Tknt_Fflush(FILE *fp)
{
    if (! Tknt_GetHasConsole() ) {
	if (fp == stdin || fp == stdout || fp == stderr) {
	    return 0;
	}
    }
    return fflush(fp);
}

char *
Tknt_Fgets(char *buffer, int length, FILE *fp)
{
    if (! Tknt_GetHasConsole() && fp == stdin) {
    	MSG msg;
    	
	while (GetMessage(&msg, (HWND) NULL, 0, 0)) {
	    if (msg.message == WM_USER) {
	        HGLOBAL hGlobal;
		char * pGlobal;
	        hGlobal = (HGLOBAL) GetWindowLong(Tknt_GetPostWindow(), 0);
		pGlobal = GlobalLock(hGlobal);
		strncpy(buffer, pGlobal, length);
		*pGlobal = '\0';
		GlobalUnlock(hGlobal);
		return buffer;
	    } else {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	    }
	}
	return NULL;
    } else {
    	return fgets(buffer, length, fp);
    }
}

int
Tknt_Fgetc(FILE *f)
{
    int fd, ret;
    char c;
    int hasConsole;

    hasConsole = Tknt_GetHasConsole();

    if (!hasConsole) {
	if (f == stdin) {

	    HGLOBAL hGlobal;
	    char * pGlobal;
	    MSG msg;
    	
	return_ch:
	    hGlobal = (HGLOBAL) GetWindowLong(Tknt_GetPostWindow(), 0);
	    pGlobal = GlobalLock(hGlobal);
	    if (*pGlobal != '\0') {
		int ch;

		ch = *pGlobal;
		strcpy(pGlobal, pGlobal+1);
		GlobalUnlock(hGlobal);
		return ch;
	    }
	    GlobalUnlock(hGlobal);

	    while (GetMessage(&msg, (HWND) NULL, 0, 0)) {
		if (msg.message == WM_USER) {
		    goto return_ch;
		} else {
		    TranslateMessage(&msg);
		    DispatchMessage(&msg);
		}
	    }
	    return EOF;
	}
    }
    fd = fileno(f);

    switch (fdtype[fd]) {
    case SOCKTYPE:
	BlockSocket(fdToSocket[fd], TK_READABLE|TK_EXCEPTION);
	ret = recv(fdToSocket[fd], &c, 1, 0);
	if (ret == 0) {
	    set_feof(f);
	    return EOF;
	} else if (ret != 1) {
	    set_ferror(f);
	    return EOF;
	}
	return (int) c;
    default:
	return fgetc(f);
    }
}
