/*
 * chain.c --
 *
 *	This module manages the event queues.  There are three
 *	event queues that are used in this library: an X event
 *	queue, a timer event queue, and a file event queue that
 *	is mainly used for sockets.
 *	
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#include "xlibInt.h"

/*
 *----------------------------------------------------------------------
 *
 * Lx_PeekEvent --
 *
 *	Copies an X event, if available, without removing it from
 *	the X event queue.
 *
 * Results:
 *	NULL if no events are in the queue, a filled in 'event'
 *	if there was at least one event in the X event queue.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
XEvent *
Lx_PeekEvent(Display *display, _XQEvent *cPtr, XEvent *event)	
{
    if (cPtr == display->head) {
	return NULL;
    }

    memcpy(event, &cPtr->event, sizeof(XEvent));
    /* printf("K%d ", event->xany.type); */
    return event;
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_PeekFileEvent --
 *
 *	Copies a file event, if available, without removing it from
 *	the file event queue.
 *
 * Results:
 *	NULL if no events are in the queue, a filled in 'event'
 *	if there was at least one event in the queue.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
XEvent *
Lx_PeekFileEvent(Display *display, _XQEvent *cPtr, XEvent *event)	
{
    if (cPtr == display->filehead) {
	return NULL;
    }

    memcpy(event, &cPtr->event, sizeof(XEvent));
    return event;
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_PeekTimerEvent --
 *
 *	Copies a timer event, if available, without removing it from
 *	the timer event queue.
 *
 * Results:
 *	NULL if no events are in the queue, a filled in 'event'
 *	if there was at least one event in the queue.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
XEvent *
Lx_PeekTimerEvent(Display *display, _XQEvent *cPtr, XEvent *event)	
{
    if (cPtr == display->timerhead) {
	return NULL;
    }

    memcpy(event, &cPtr->event, sizeof(XEvent));
    return event;
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_DiscardEvent --
 *
 *	Deletes the first event from the X event queue
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
void
Lx_DiscardEvent(Display *display, _XQEvent *cPtr)
{
    cPtr->prev->next = cPtr->next;
    cPtr->next->prev = cPtr->prev;
    display->qlen--;

    ckfree(cPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_DiscardFileEvent --
 *
 *	Deletes the first event from the X event queue
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
void
Lx_DiscardFileEvent(Display *display, _XQEvent *cPtr)
{
    cPtr->prev->next = cPtr->next;
    cPtr->next->prev = cPtr->prev;
    display->fileqlen--;

    ckfree(cPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_DiscardTimerEvent --
 *
 *	Deletes the first event from the X event queue
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
void
Lx_DiscardTimerEvent(Display *display, _XQEvent *cPtr)
{
    cPtr->prev->next = cPtr->next;
    cPtr->next->prev = cPtr->prev;
    display->timerqlen--;

    ckfree(cPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_GetEvent --
 *
 *	Copies an X event, if available, and removes it from
 *	the X event queue.
 *
 * Results:
 *	NULL if no events are in the queue, a filled in 'event'
 *	if there was at least one event in the queue.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
XEvent *
Lx_GetEvent(Display *display, XEvent *event)	
{
    _XQEvent *cPtr;
    if (display->qlen == 0) {
	return NULL;
    }
    cPtr = display->head->next;
    memcpy(event, &cPtr->event, sizeof(XEvent));

    cPtr->prev->next = cPtr->next;
    cPtr->next->prev = cPtr->prev;
    display->qlen--;

    ckfree(cPtr);
    /* printf("N%d ", event->xany.type); */
    return event;
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_GetFileEvent --
 *
 *	Copies a file event, if available, and removes it from
 *	the file event queue.
 *
 * Results:
 *	NULL if no events are in the queue, a filled in 'event'
 *	if there was at least one event in the queue.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
XEvent *
Lx_GetFileEvent(Display *display, XEvent *event)	
{
    _XQEvent *cPtr;
    if (display->fileqlen == 0) {
	return NULL;
    }
    cPtr = display->filehead->next;
    memcpy(event, &cPtr->event, sizeof(XEvent));

    cPtr->prev->next = cPtr->next;
    cPtr->next->prev = cPtr->prev;
    display->fileqlen--;

    ckfree(cPtr);
    return event;
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_GetTimerEvent --
 *
 *	Copies a timer event, if available, and removes it from
 *	the timer event queue.
 *
 * Results:
 *	NULL if no events are in the queue, a filled in 'event'
 *	if there was at least one event in the queue.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
XEvent *
Lx_GetTimerEvent(Display *display, XEvent *event)	
{
    _XQEvent *cPtr;
    if (display->timerqlen == 0) {
	return NULL;
    }
    cPtr = display->timerhead->next;
    memcpy(event, &cPtr->event, sizeof(XEvent));

    cPtr->prev->next = cPtr->next;
    cPtr->next->prev = cPtr->prev;
    display->timerqlen--;

    ckfree(cPtr);
    return event;
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_PutEvent --
 *
 *	Creates a duplicate event of the one passed in, and appends this
 *	to the end of the X event queue.
 *
 * Results:
 *	The new event or NULL if the ckalloc failed.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
_XQEvent *
Lx_PutEvent(Display *display, XEvent *event)
{
    _XQEvent *cPtr, *head, *curr;
    Window w;

    cPtr = (_XQEvent *) ckalloc(sizeof(_XQEvent));
    if (cPtr == NULL) {
	return NULL;
    }

    /* printf("P%d ", event->xany.type); */
    display->qlen++;
    head = display->head;

    if (event->xany.type == Expose) {
	/*
	 * Run throught the current events, incrementing the count
	 * for expose events of the same window
	 */

	w = event->xexpose.window;

	for (curr = head->next; curr != head; curr = curr->next) {
	    if (curr->event.xany.type == Expose) {
		if (curr->event.xexpose.window == w) {
		    curr->event.xexpose.count++;
		}
	    }
	}
    }

    cPtr->next = head;
    cPtr->prev = head->prev;
    head->prev->next = cPtr;
    head->prev = cPtr;

    memcpy(&cPtr->event, event, sizeof(XEvent));

    return cPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_PutFileEvent --
 *
 *	Creates a duplicate event of the one passed in, and appends this
 *	to the end of the file event queue.
 *
 * Results:
 *	The new event or NULL if the ckalloc failed.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
_XQEvent *
Lx_PutFileEvent(Display *display, XEvent *event)
{
    _XQEvent *cPtr, *head;

    cPtr = (_XQEvent *) ckalloc(sizeof(_XQEvent));
    if (cPtr == NULL) {
	return NULL;
    }

    display->fileqlen++;
    head = display->filehead;
    cPtr->next = head;
    cPtr->prev = head->prev;
    head->prev->next = cPtr;
    head->prev = cPtr;

    memcpy(&cPtr->event, event, sizeof(XEvent));

    return cPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_PutTimerEvent --
 *
 *	Creates a duplicate event of the one passed in, and appends this
 *	to the end of the timer event queue.
 *
 * Results:
 *	The new event or NULL if the ckalloc failed.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
_XQEvent *
Lx_PutTimerEvent(Display *display, XEvent *event)
{
    _XQEvent *cPtr, *head;

    cPtr = (_XQEvent *) ckalloc(sizeof(_XQEvent));
    if (cPtr == NULL) {
	return NULL;
    }

    display->timerqlen++;
    head = display->timerhead;
    cPtr->next = head;
    cPtr->prev = head->prev;
    head->prev->next = cPtr;
    head->prev = cPtr;

    memcpy(&cPtr->event, event, sizeof(XEvent));

    return cPtr;
}


/*
 *----------------------------------------------------------------------
 *
 * Lx_PutbackEvent --
 *
 *	Creates a duplicate event of the one passed in, and inserts this
 *	to the at the beginning of the X event queue.
 *
 * Results:
 *	The new event or NULL if the ckalloc failed.
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
_XQEvent *
Lx_PutbackEvent(Display *display, XEvent *event)
{
    _XQEvent *cPtr, *head;

    cPtr = (_XQEvent *) ckalloc(sizeof(_XQEvent));
    if (cPtr == NULL) {
	return NULL;
    }

    /* printf("PB%d ", event->xany.type); */
    memcpy(&(cPtr->event), event, sizeof(XEvent));

    display->qlen++;
    head = display->head;
    cPtr->next = head->next;
    cPtr->prev = head;
    head->next->prev = cPtr;
    head->next = cPtr;

    return cPtr;
}
