/*
 * xdisplay.c --
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */
 
#include "xlibInt.h" 

int lx_ScreenDepth;	/* Depth for the screen */

/* 
 * XSetWindowAttributes
 */

XSetWindowAttributes lx_DefWindowAttrs = {
    None,		/* foregound pixmap */ 
    RGB(255, 255, 255),	/* background pixel */
    None, 		/* border pixmap */	
    0L, 		/* border pixel */
    ForgetGravity, 	/* bit gravity */
    NorthWestGravity, 	/* window gravity */
    NotUseful,		/* backing store */
    0, 			/* backing planes */
    0, 			/* backing pixel */
    False, 		/* save under */
    0L, 		/* even tmask */
    0L, 		/* do not propagate mask */
    False, 		/* override redirect */
    1,	 		/* colormap */
    None,		/* cursor */
    False,		/* clear_on_expose */
    False		/* native widget */
};

/* 
 * XWindowChanges
 */

XWindowChanges lx_DefWindowChanges = {
    0, 		/* x */
    0, 		/* y */
    1, 		/* width */
    1, 		/* height */
    1, 		/* border width */
    None, 	/* sibling */
    Above,	/* stack mode */
    False,	/* pos_changed */
};

static Display *displayList = NULL;

static Screen *mkScreen(Display *);
static Visual *mkVisual(Display *, int depth);


/*
 *----------------------------------------------------------------------
 * XOpenDisplay --
 *
 *	If NULL is passed as the screenName, then this returns the display.
 *	If the display has already been opened, it returns the open display.
 *	Otherwise, it allocates the display, a screen, and a visual
 *
 * Results:
 *	The open display
 *----------------------------------------------------------------------
 */

Display *
XOpenDisplay(const char *screenName)
{
    Display *dPtr;
    Screen *sPtr;

    if (displayList != NULL) {
	if (screenName == NULL) {
	    return displayList;
	} else {
	    if (strcmp(displayList->display_name, screenName) == 0) {
		return displayList;
	    } else {
		return NULL;
	    }
	}
    }
    
    dPtr = (Display *) ckalloc(sizeof(Display));
    memset(dPtr, 0, sizeof(Display));
    if (dPtr == NULL) {
	return NULL;
    }
    sPtr = mkScreen(dPtr);
    if (sPtr == NULL) {
    	ckfree(dPtr);
        return NULL;
    }

    dPtr->head = (_XQEvent *) ckalloc(sizeof(_XQEvent));
    dPtr->filehead = (_XQEvent *) ckalloc(sizeof(_XQEvent));
    dPtr->timerhead = (_XQEvent *) ckalloc(sizeof(_XQEvent));

    if (dPtr->head == NULL || dPtr->filehead == NULL || dPtr->timerhead == NULL) {
    error:
	ckfree(sPtr);
	ckfree(dPtr);
	return NULL;
    }

    dPtr->request = 0;
    dPtr->last_request_read = 0;

    dPtr->head->next = dPtr->head;
    dPtr->head->prev = dPtr->head;
    dPtr->qlen = 0;   

    dPtr->filehead->next = dPtr->filehead;
    dPtr->filehead->prev = dPtr->filehead;
    dPtr->fileqlen = 0;   

    dPtr->timerhead->next = dPtr->timerhead;
    dPtr->timerhead->prev = dPtr->timerhead;
    dPtr->timerqlen = 0;   

    if (screenName != NULL && *screenName != '\0') {
    	dPtr->display_name = (char *)ckalloc(strlen(screenName) + 1);
	if (dPtr->display_name == NULL) {    
	    goto error;
	}
	strcpy(dPtr->display_name, screenName);		
    } else {
    	dPtr->display_name = "";
    }

#define BOGUS_DISPLAY_FD (SOCKET) -2
    dPtr->fd = (int) BOGUS_DISPLAY_FD;
    /* dPtr->fd = (int) GetCurrentThread(); */

    dPtr->default_screen = 0;     	/* default screen for operations */
    dPtr->nscreens = 1;           	/* number of screens on this server*/
    dPtr->screens = sPtr;        	/* pointer to list of screens */
    dPtr->pixmapList = NULL;
    dPtr->cursorList = NULL;

    dPtr->next = displayList;
    dPtr->focusWin = None;
    displayList = dPtr;
    return dPtr;
}

/*
 *----------------------------------------------------------------------
 * mkScreen --
 *
 *	Creates a screen structure and fills it in
 *----------------------------------------------------------------------
 */
static Screen *
mkScreen(Display *display)
{
    Screen *sPtr;
    HDC hdc;
    int depth, actual_depth;

    if ((sPtr = ckalloc(sizeof(Screen))) == NULL)
	return NULL;

    if (display->screens == NULL) {
	display->screens = sPtr;
    }

    sPtr->display = display;				/* back pointer to display structure */
    sPtr->root = HWND_DESKTOP;           		/* Root window id. */
    sPtr->width = GetSystemMetrics(SM_CXSCREEN);	/* width and height of screen */
    sPtr->height = GetSystemMetrics(SM_CYSCREEN);      

    hdc = GetDC(HWND_DESKTOP);
    depth = GetDeviceCaps(hdc, BITSPIXEL);
    actual_depth = depth;
#if 1
    if (depth != 8) {
	depth = 32;
    }
#else
    depth = 32;
#endif

    lx_ScreenDepth = depth;

    if (depth == 8) {
	sPtr->black_pixel = 0;
	sPtr->white_pixel = 1;
	lx_DefWindowAttrs.background_pixel = 1;
    } else {
	sPtr->white_pixel = RGB(255, 255, 255);		/* White and Black pixel values */
	sPtr->black_pixel = RGB(0, 0, 0);
    }

    sPtr->root_depth = depth;
    sPtr->mwidth = GetDeviceCaps(hdc, HORZSIZE); 	/* width and height of  in millimeters */
    sPtr->mheight = GetDeviceCaps(hdc, VERTSIZE);    
    ReleaseDC(HWND_DESKTOP, hdc);	  
    sPtr->backing_store = False;			/* Never, WhenMapped, Always */
    sPtr->save_unders = False;
    sPtr->root_visual = mkVisual(display, depth);

    sPtr->cmap = XCreateColormap(display, GetDesktopWindow(),
				 sPtr->root_visual, 0);	/* default colormap */
    ColorMapInfo[sPtr->cmap]->actual_depth = actual_depth;

    return sPtr;
}

static VisualID visualid = 0;
static Visual *visualList[1];

/*
 *----------------------------------------------------------------------
 * mkVisual --
 *
 *	Create a visual.  Internal function.
 *----------------------------------------------------------------------
 */
static Visual *
mkVisual(Display *display, int depth)
{
    static int visualid = 0;
    static Visual *visual = NULL;

    if (visual != NULL) {
	return visual;
    }
    visual = (Visual *) ckalloc(sizeof(Visual));
    if (visual == NULL) {
	return NULL;
    }
    visualList[visualid] = visual;
    visual->visualid = visualid++;
    visual->ext_data = NULL;
    if (depth != 8) {
	visual->class = DirectColor;
	visual->red_mask   = 0x000000ff;
	visual->green_mask = 0x0000ff00;
	visual->blue_mask  = 0x00ff0000;
	visual->bits_per_rgb = 8;
	visual->map_entries = 256 * 256 * 256;
    } else {
	visual->class = PseudoColor;
	visual->bits_per_rgb = 8;
	visual->map_entries = 256;
    }

    return visual;
}

/*
 *----------------------------------------------------------------------
 * XVisualIDFromVisual --
 *
 *	Returns the visual ID for the specified visual structure
 *----------------------------------------------------------------------
 */
VisualID
XVisualIDFromVisual(Visual *visual)
{
    return visual->visualid;
}

/*
 *----------------------------------------------------------------------
 * XGetVisualInfo --
 *
 *	Find the visual information structures that match the specified
 *	template
 *
 * Notes:
 *	XXX: This is not a complete implementation.  It returns the only
 *	visual 
 *----------------------------------------------------------------------
 */
XVisualInfo *
XGetVisualInfo(Display *display, long vinfo_mask, XVisualInfo *vinfo_template,
	       int *nitems_return)
{
    Visual *visual;
    XVisualInfo *vinfo;

    display->request++;
    visual = visualList[0];

    *nitems_return = 0;
    if (vinfo_mask & VisualIDMask) {
	visual = visualList[vinfo_template->visualid];
    }

    if (vinfo_mask & VisualScreenMask) {
	if (vinfo_template->screen != 0) {
	    return NULL;
	}
    }

    if (vinfo_mask & VisualDepthMask) {
	if (vinfo_template->depth != 32 && vinfo_template->depth != 8 &&
	    vinfo_template->depth != 24) {
	    return NULL;
	}
    }

    if (vinfo_mask & VisualClassMask) {
	if (vinfo_template->depth != visual->class) {
	    return NULL;
	}
    }

    if (vinfo_mask & VisualRedMaskMask) {
	if (vinfo_template->red_mask != visual->red_mask) {
	    return NULL;
	}
    }

    if (vinfo_mask & VisualGreenMaskMask) {
	if (vinfo_template->green_mask != visual->green_mask) {
	    return NULL;
	}
    }

    if (vinfo_mask & VisualBlueMaskMask) {
	if (vinfo_template->blue_mask != visual->blue_mask) {
	    return NULL;
	}
    }

    vinfo = (XVisualInfo *) ckalloc(sizeof(XVisualInfo));

    vinfo->visual = visual;
    vinfo->visualid = visual->visualid;
    vinfo->screen = 0;
    vinfo->depth = lx_ScreenDepth;
    vinfo->class = visual->class;
    vinfo->red_mask = visual->red_mask;
    vinfo->green_mask = visual->green_mask;
    vinfo->blue_mask = visual->blue_mask;
    vinfo->colormap_size = visual->map_entries;
    vinfo->bits_per_rgb = visual->bits_per_rgb;

    *nitems_return = 1;
    return vinfo;
}

/*
 *----------------------------------------------------------------------
 * XCloseDisplay --
 *
 *	Close up the display and free up all allocated memory
 *----------------------------------------------------------------------
 */

void 
XCloseDisplay(Display *display)
{
    if (displayList != NULL && displayList == display) {
        ckfree(display->screens);
	ckfree(display);
	displayList = NULL;
    }
}


/*
 *----------------------------------------------------------------------
 * XRootWindow --
 *
 *	Return the root window for a given screen number
 *----------------------------------------------------------------------
 */
Window
XRootWindow(Display *display, int screen_number)
{
    if (displayList != NULL && displayList == display 
    	    && screen_number <= display->nscreens) {
	return (display->screens[screen_number]).root;
    }
    return None;
}

/*
 *----------------------------------------------------------------------
 * XBlackPixel
 *
 *	Return the black pixel for a given screen number
 *----------------------------------------------------------------------
 */
unsigned long
XBlackPixel(Display *display, int screen_number)
{
    if (displayList != NULL && displayList == display 
    	    && screen_number <= display->nscreens) {
	return (display->screens[screen_number]).black_pixel;
    }
    return None;
    
}

/*
 *----------------------------------------------------------------------
 * XWhitePixel
 *
 *	Return the white pixel for a given screen number
 *----------------------------------------------------------------------
 */
unsigned long
XWhitePixel(Display *display, int screen_number)
{
    if (displayList != NULL && displayList == display 
    	    && screen_number <= display->nscreens) {
	return (display->screens[screen_number]).white_pixel;
    }
    return None;
    
}

/*
 *----------------------------------------------------------------------
 * XFlush --
 *
 *	Flush the queued requests.  Under NT, this does nothing
 *----------------------------------------------------------------------
 */
void
XFlush(Display *display)
{
}

/*
 *----------------------------------------------------------------------
 * XSynchronize --
 *
 *	A noop under NT
 *----------------------------------------------------------------------
 */
int (*
XSynchronize(Display *display, Bool onoff)
)()
{
    return NULL;
}

/*
 *----------------------------------------------------------------------
 * XSync --
 *
 *	Synchronize the server and the client.  Under NT, they are
 *	the same, so no synching is necessary and this does nothing.
 *----------------------------------------------------------------------
 */
void
XSync(Display *display, Bool dummy2)
{
}
