/*
 * xgraphic.c --
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#include "xlibInt.h" 

#define PI 3.14159265358979

static POINT *new_points;
static int local_points;


HDC
Lx_GetDrawableDC(HANDLE drawable, Lx_DrawableDCState *state)
{
    HDC hdc, hdcMemory;
    HPALETTE hPal;
    int n;

    /* The IsWindow() function is returning true for some bitmaps
     * that we create.
     */
    hPal = Lx_GetDrawablePalette(drawable);
    state->hPal = hPal;
    state->pixInfo = Lx_FindPixmap(drawable);
    if (! state->pixInfo) {
	hdc = NULL;
	while (hdc == NULL) {
	    hdc = GetDC(drawable);
	    if (hdc == NULL) {
		MessageBox(NULL, "System is running low on memory.\n"
			   "Please shutdown some applications and continue.",
			   "Memory is Low",
			   MB_OK);
	    }
	}
	if (hPal) {
	    state->hPalPrev = SelectPalette(hdc, hPal, FALSE);
	    n = RealizePalette(hdc);
	    if (n > 0) {
		UpdateColors (hdc);
	    }
	}
	return hdc;
    } else {
	hdc = NULL;
	while (hdc == NULL) {
	    hdc = GetDC(NULL);
	    if (hdc == NULL) {
		MessageBox(NULL, "System is running low on memory.\n"
			   "Please close some applications and continue.",
			   "Memory Low",
			   MB_OK);
	    }
	}
    	hdcMemory = CreateCompatibleDC(hdc);
	ReleaseDC(NULL, hdc);
	state->hBmpPrev = SelectObject(hdcMemory, drawable);

	if (hPal) {
	    state->hPalPrev = SelectPalette(hdcMemory, hPal, FALSE);
	    n = RealizePalette(hdcMemory);
	    if (n > 0) {
		UpdateColors(hdcMemory);
	    }
	}
	return hdcMemory;
    }
}

void 
Lx_ReleaseDrawableDC(HANDLE drawable, HDC hdc, Lx_DrawableDCState *state)
{
    if (state->hPal) {
	SelectPalette(hdc, state->hPalPrev, FALSE);
    }

    if (! state->pixInfo) {
    	ReleaseDC(drawable, hdc);
    } else {
	SelectObject(hdc, state->hBmpPrev);
	DeleteDC(hdc);
    }
}

static POINT *
XPointsToMSPoints(XPoint *points, int npoints)
{
    register int i;

    if (npoints > local_points) {
        if (new_points != NULL) {
	    ckfree((char *) new_points);
	}
        new_points = (POINT *) ckalloc(sizeof(POINT) * npoints);
	if (new_points == NULL) {
	    local_points = 0;
	    return NULL;
	}
	local_points = npoints;
    }

    for (i = 0; i < npoints; i++) {
	new_points[i].x = points[i].x;
	new_points[i].y = points[i].y;
    }
    return new_points;
}

static POINT *
GetOriginalPoints(XPoint *points, int npoints)
{
    register int i;

    if (npoints > local_points) {
        if (new_points != NULL) {
	    ckfree((char *) new_points);
	}
        new_points = (POINT *) ckalloc(sizeof(POINT) * npoints);
	if (new_points == NULL) {
	    local_points = 0;
	    return NULL;
	}
	local_points = npoints;
    }

    new_points[0].x = points[0].x;
    new_points[0].y = points[0].y;
    for (i = 1 ; i < npoints ; i++) {
	new_points[i].x = new_points[i-1].x + points[i].x;
	new_points[i].y = new_points[i-1].y + points[i].y;
    }
    return new_points;
}

static void
GetBoxPoints(POINT *points, int n,
	     int *left, int *top, int *right, int *bottom)
{
    int i;

    *left = *right = points[0].x;
    *top = *bottom = points[0].y;
    for (i = 1; i < n; i++) {
    	if (*left > points[i].x) *left = points[i].x;
    	if (*right < points[i].x) *right = points[i].x;
    	if (*top > points[i].y) *top = points[i].y;
    	if (*bottom < points[i].y) *bottom = points[i].y;
    }
}

static void
ShiftPoints(POINT *points, int n, int x, int y)
{
    int i;

    for (i = 0; i < n; i++, points++) {
    	points->x -= x;
    	points->y -= y;
    }
}

/*
 * XDrawPoints
 */
void
XDrawPoints(Display *display, Drawable d, GC gc, XPoint *points, int npoints, int mode)
{
    HDC hdc;
    int point;
    int x, y;
    XGCValues *gcv;
    COLORREF cref;
    int prevx, prevy;
    Lx_DrawableDCState state;

    display->request++;

    gcv = &gc->values;
    hdc = Lx_GetDrawableDC(d, &state);
    SetROP2(hdc, gcv->function);
    if (gc->rects == True) {
	SelectObject(hdc, ((Region) gcv->clip_mask)->hrgn);
    }

    if (state.pixInfo && state.pixInfo->depth == 1) {
	cref = gcv->foreground ? RGB(0,0,0) : RGB(255,255,255);
    } else {
	cref = LX_CONVERTPIXEL(gcv->foreground);
    }

    for (prevx = prevy = 0, point = 0; point < npoints; point++) {
	if (mode == CoordModeOrigin) {
	    x = points[point].x;
	    y = points[point].y;
	} else {
	    x = points[point].x + prevx;
	    y = points[point].y + prevy;
	}

 	SetPixel(hdc, x, y, cref);
    }
    Lx_ReleaseDrawableDC(d, hdc, &state);
}


/*
 * XDrawSegments
 */
void
XDrawSegments(Display *display, Drawable d, GC gc, XSegment *segments, int nsegments)
{
    HDC hdc;
    HPEN hpen, hpenPrev;
    POINT oldpos;
    HDC hdcMem;
    HBITMAP hbmp, hbmpPrev;
    HBRUSH hbrPat, hbrPrev;
    int width, height;
    int seg;
    int x1, y1, x2, y2;
    XGCValues *gcv;
    HPALETTE hPal, hPalMemPrev;
    COLORREF fg;
    Lx_DrawableDCState state;

    display->request++;

    gcv = &gc->values;
    hdc = Lx_GetDrawableDC(d, &state);
    SetROP2(hdc, gcv->function);
    if (gc->rects == True) {
	SelectObject(hdc, ((Region) gcv->clip_mask)->hrgn);
    }

    if (state.pixInfo && state.pixInfo->depth == 1) {
	fg = gcv->foreground ? RGB(0,0,0) : RGB(255,255,255);
    } else {
	fg = LX_CONVERTPIXEL(gcv->foreground);
    }

    hpen = CreatePen(gcv->line_style, gcv->line_width, fg);

    if ((gcv->fill_style == FillStippled ||
	 gcv->fill_style == FillOpaqueStippled)
	&& gcv->stipple != None)
    {
	hdcMem = CreateCompatibleDC(hdc);
	hPal = Lx_GetDrawablePalette(d);
	if (hPal) {
	    hPalMemPrev = SelectPalette(hdcMem, hPal, FALSE);
	}
	hbrPat = CreatePatternBrush(gcv->stipple);
	SetBkMode(hdcMem, TRANSPARENT);
	SetBrushOrgEx(hdc, gcv->ts_x_origin, gcv->ts_y_origin, NULL);
	hbrPrev = SelectObject(hdc, hbrPat);

	hpenPrev = SelectObject(hdcMem, hpen);
	for (seg = 0; seg < nsegments; seg++) {
	    x1 = segments[seg].x1;
	    x2 = segments[seg].x2;
	    y1 = segments[seg].y1;
	    y2 = segments[seg].y2;

	    width = max(x1, x2) + 30;
	    height = max(y1, y2) + 30;

	    hbmp = CreateCompatibleBitmap(hdc, width, height);
	    hbmpPrev = SelectObject(hdcMem, hbmp);

	    PatBlt(hdcMem, 0, 0, width, height, BLACKNESS);
	    MoveToEx(hdcMem, x1, y1, &oldpos);
	    LineTo(hdcMem, x2, y2);

	    BitBlt(hdc, 0, 0, width, height, hdcMem, 0, 0, 0x00AE0B06);

	    PatBlt(hdcMem, 0, 0, width, height, WHITENESS);
	    MoveToEx(hdcMem, x1, y1, &oldpos);
	    LineTo(hdcMem, x2, y2);

	    BitBlt(hdc, 0, 0, width, height, hdcMem, 0, 0, 0x00A803A9);

	    SelectObject(hdcMem, hbmpPrev);
	    DeleteObject(hbmp);
	}
	SelectObject(hdcMem, hpenPrev);
	SelectObject(hdc, hbrPrev);
	DeleteObject(hbrPat);
	if (hPal) {
	    SelectPalette(hdcMem, hPalMemPrev, FALSE);
	}
	DeleteDC(hdcMem);
    } else {
	hpenPrev = SelectObject(hdc, hpen);
	for (seg = 0; seg < nsegments; seg++) {
	    x1 = segments[seg].x1;
	    x2 = segments[seg].x2;
	    y1 = segments[seg].y1;
	    y2 = segments[seg].y2;

	    MoveToEx(hdc, x1, y1, &oldpos);
	    LineTo(hdc, x2, y2);
	}
	SelectObject(hdc, hpenPrev);
    }
    DeleteObject(hpen);
    Lx_ReleaseDrawableDC(d, hdc, &state);
}

/*
 *  XDrawLine
 *
 */

void
XDrawLine(Display *display, Drawable d, GC gc, int x1, int y1, int x2, int y2)
{
    XSegment segment;

    segment.x1 = x1;
    segment.y1 = y1;
    segment.x2 = x2;
    segment.y2 = y2;
    XDrawSegments(display, d, gc, &segment, 1);
}

/*
 *  XDrawLines
 *
 */

void
XDrawLines(Display *display, Drawable d, GC gc, XPoint *points, int npoints,
	   int mode)
{
    HDC hdc;
    HPEN hpen, hpenPrev;
    POINT *xp;
    XGCValues *gcv;
    Lx_DrawableDCState state;
    COLORREF fg;

    display->request++;

    gcv = &gc->values;
    hdc = Lx_GetDrawableDC(d, &state);
    SetROP2(hdc, gcv->function);
    if (gc->rects == True) {
	SelectObject(hdc, ((Region) gcv->clip_mask)->hrgn);
    }

    if (state.pixInfo && state.pixInfo->depth == 1) {
	fg = gcv->foreground ? RGB(0,0,0) : RGB(255,255,255);
    } else {
	fg = LX_CONVERTPIXEL(gcv->foreground);
    }

    hpen = CreatePen(gcv->line_style, gcv->line_width, fg);
    xp = (mode == CoordModePrevious)?
	GetOriginalPoints(points, npoints): XPointsToMSPoints(points, npoints);
    if (xp == NULL) return;
    if ((gcv->fill_style == FillStippled || gcv->fill_style == FillOpaqueStippled)
	    && gcv->stipple != None)
    {
	HDC hdcMem;
	HBITMAP hbmp, hbmpPrev;
	HBRUSH hbrPat, hbrPrevPat;
	HPALETTE hPal, hPalMemPrev;
	int x, y, width, height;

	GetBoxPoints(xp, npoints, &x, &y, &width, &height);
	x -= gcv->line_width;
	y -= gcv->line_width;
	ShiftPoints(xp, npoints, x, y);
	width -= x-1 - gcv->line_width * 2;
	height -= y-1 - gcv->line_width * 2;

	hdcMem = CreateCompatibleDC(hdc);
	hPal = Lx_GetDrawablePalette(d);
	if (hPal) {
	    hPalMemPrev = SelectPalette(hdcMem, hPal, FALSE);
	}
	SetBkMode(hdcMem, TRANSPARENT);
	hbmp = CreateCompatibleBitmap(hdc, width, height);
	hbmpPrev = SelectObject(hdcMem, hbmp);
	hpenPrev = SelectObject(hdcMem, hpen);

	hbrPat = CreatePatternBrush(gcv->stipple);
	SetBrushOrgEx(hdc, gcv->ts_x_origin, gcv->ts_y_origin, NULL);
	hbrPrevPat = SelectObject(hdc, hbrPat);

	PatBlt(hdcMem, 0, 0, width, height, BLACKNESS);
	Polyline(hdcMem, xp, npoints);

	BitBlt(hdc, x, y, width, height, hdcMem, 0, 0, 0x00AE0B09);

	PatBlt(hdcMem, 0, 0, width, height, WHITENESS);
	Polyline(hdcMem, xp, npoints);
	BitBlt(hdc, x, y, width, height, hdcMem, 0, 0, 0x00A803A9);
	DeleteObject(SelectObject(hdc, hbrPrevPat));

	SelectObject(hdcMem, hbrPrevPat);
	SelectObject(hdcMem, hpenPrev);
	SelectObject(hdcMem, hbmpPrev);
	if (hPal) {
	    SelectPalette(hdcMem, hPalMemPrev, FALSE);
	}

	DeleteObject(hbrPat);
	DeleteObject(hbmp);
	DeleteDC(hdcMem);
    } else {
	hpenPrev = SelectObject(hdc, hpen);
	Polyline(hdc, xp, npoints);
	SelectObject(hdc, hpenPrev);
    }
    DeleteObject(hpen);
    Lx_ReleaseDrawableDC(d, hdc, &state);
}


static void
DrawOrFillArc(Display *display, Drawable d, GC gc, int x, int y,
	      unsigned int width, unsigned int height,
	      int angle1, int angle2, int fill)
{
    HDC hdc;
    HPEN hpen, hpenPrev;
    HBRUSH hbr, hbrPrev;
    int xr, yr, xstart, ystart, xend, yend;
    double radian_start, radian_end, radian_tmp;
    XGCValues *gcv;
    COLORREF fg;
    Lx_DrawableDCState state;

    gcv = &gc->values;
    hdc = Lx_GetDrawableDC(d, &state);

    if(width % 2 == 1)
	xr = width / 2;
    else
	xr = (width - 1) / 2;

    if (height % 2 == 1)
	yr = height / 2;
    else
	yr = (height - 1) / 2;

    radian_start = (double)angle1 / 64 * PI / 180;
    radian_end = (double)(angle1+angle2) / 64 * PI / 180;
    if (angle2 < 0) {
	radian_tmp = radian_start;
	radian_start = radian_end;
	radian_end = radian_tmp;
    }

    xstart = x + (int)( (double)xr * (1+cos(radian_start)) );
    ystart = y + (int)( (double)yr * (1-sin(radian_start)) );
    xend = x + (int)( (double)xr * (1+cos(radian_end)) );
    yend = y + (int)( (double)yr * (1-sin(radian_end)) );

    SetROP2(hdc, gcv->function);
    if (gc->rects == True) {
	SelectObject(hdc, ((Region) gcv->clip_mask)->hrgn);
    }

    if (state.pixInfo && state.pixInfo->depth == 1) {
	fg = gcv->foreground ? RGB(0,0,0) : RGB(255,255,255);
    } else {
	fg = LX_CONVERTPIXEL(gcv->foreground);
    }

    if (!fill) {
	hpen = CreatePen(gcv->line_style, gcv->line_width, fg);
	hpenPrev = SelectObject(hdc, hpen);
	Arc(hdc, x, y, x + width, y + height, xstart, ystart, xend, yend);
	SelectObject(hdc, hpenPrev);
	DeleteObject(hpen);
    } else {
	hpen = CreatePen(PS_SOLID, 1, fg);
	hbr = CreateSolidBrush(fg);

	if ((gcv->fill_style == FillStippled || gcv->fill_style == FillOpaqueStippled)
		&& gcv->stipple != None) {
	    HDC hdcMem;
	    HBITMAP hbmp, hbmpPrev;
	    HBRUSH hbrPat, hbrPrevPat;
	    HPALETTE hPal, hPalMemPrev;
	
	    hdcMem = CreateCompatibleDC(hdc);
	    hPal = Lx_GetDrawablePalette(d);
	    if (hPal) {
		hPalMemPrev = SelectPalette(hdcMem, hPal, FALSE);
	    }
	    SetBkMode(hdcMem, TRANSPARENT);
	    hbmp = CreateCompatibleBitmap(hdc, width, height);
	    hbmpPrev = SelectObject(hdcMem, hbmp);
	    hpenPrev = SelectObject(hdcMem, hpen);
	    hbrPrev = SelectObject(hdcMem, hbr);

	    PatBlt(hdcMem, 0, 0, width, height, BLACKNESS);
	    if ( gcv->arc_mode == ArcChord ) {
	    	Chord(hdcMem, 0, 0, width, height, xstart - x, ystart -y, xend-x, yend-y);
	    } else {
	    	Pie(hdcMem, 0, 0, width, height, xstart-x , ystart-y, xend-x, yend-y);
	    }

	    hbrPat = CreatePatternBrush(gcv->stipple);
	    SetBrushOrgEx(hdc, gcv->ts_x_origin, gcv->ts_y_origin, NULL);
	    hbrPrevPat = SelectObject(hdc, hbrPat);
	    BitBlt(hdc, x , y, width, height, hdcMem, 0, 0, 0x00AE0B06);

	    PatBlt(hdcMem, 0, 0, width, height, WHITENESS);
	    if ( gcv->arc_mode == ArcChord ) {
	    	Chord(hdcMem, 0, 0, width, height, xstart - x, ystart -y, xend-x, yend-y);
	    } else {
	    	Pie(hdcMem, 0, 0, width, height, xstart-x , ystart-y, xend-x, yend-y);
	    }
	    BitBlt(hdc, x , y, width, height, hdcMem, 0, 0, 0x00A803A9);

	    SelectObject(hdc, hbrPrevPat);
	    SelectObject(hdcMem, hpenPrev);
	    SelectObject(hdcMem, hbrPrev);
	    SelectObject(hdcMem, hbmpPrev);
	    if (hPal) {
		SelectPalette(hdcMem, hPalMemPrev, FALSE);
	    }

	    DeleteObject(hbrPat);
	    DeleteObject(hbmp);
	    DeleteDC(hdcMem);
	} else {
	    hpenPrev = SelectObject(hdc, hpen);
	    hbrPrev = SelectObject(hdc, hbr);
	    if (gcv->arc_mode == ArcChord) {
	        Chord(hdc, x, y, x + width, y + height, xstart, ystart, xend, yend);
	    } else if (gcv->arc_mode == ArcPieSlice) {
		Pie(hdc, x, y, x + width, y + height, xstart, ystart, xend, yend);
	    }
	    SelectObject(hdc, hpenPrev);
	    SelectObject(hdc, hbrPrev);
	}
	DeleteObject(hpen);
	DeleteObject(hbr);
    }
    Lx_ReleaseDrawableDC(d, hdc, &state);
}

void
XDrawArc(Display *display, Drawable d, GC gc, int x, int y,
	 unsigned int width, unsigned int height, int angle1, int angle2)
{
    display->request++;

    DrawOrFillArc(display, d, gc, x, y, width, height, angle1, angle2, 0);
}

void 
XFillArc(Display *display, Drawable d, GC gc, int x, int y,
	 unsigned int width, unsigned int height, int angle1, int angle2)
{
    display->request++;

    DrawOrFillArc(display, d, gc, x, y, width, height, angle1, angle2, 1);
}


void
XFillPolygon(Display *display, Drawable d, GC gc, XPoint *points, int n,
	     int shape, int mode)
{
    HDC hdc;
    HPEN hpen, hpenPrev;
    HBRUSH hbr, hbrPrev;
    POINT *xp;
    XGCValues *gcv;
    COLORREF fg;
    Lx_DrawableDCState state;

    display->request++;
    gcv = &gc->values;

    hdc = Lx_GetDrawableDC(d, &state);
    SetROP2(hdc, gcv->function);
    SetPolyFillMode(hdc, gcv->fill_rule);
    if (gc->rects == True) {
	SelectObject(hdc, ((Region) gcv->clip_mask)->hrgn);
    }

    if (state.pixInfo && state.pixInfo->depth == 1) {
	fg = gcv->foreground ? RGB(0,0,0) : RGB(255,255,255);
    } else {
	fg = LX_CONVERTPIXEL(gcv->foreground);
    }
    hpen = CreatePen(PS_NULL, 0, fg);
    hbr = CreateSolidBrush(fg);

    xp = (mode == CoordModePrevious)?
	GetOriginalPoints(points, n):XPointsToMSPoints(points, n);
    if (xp == NULL) return;
    if ((gcv->fill_style == FillStippled || gcv->fill_style == FillOpaqueStippled)
	    && gcv->stipple != None)
    {
	HDC hdcMem;
	HBITMAP hbmp, hbmpPrev;
	HBRUSH hbrPat, hbrpatPrev;
	HPALETTE hPal, hPalMemPrev;
	int x, y, width, height;

	GetBoxPoints(xp, n, &x, &y, &width, &height);
	ShiftPoints(xp, n, x, y);
	width -= x-1;
	height -= y-1;

	hdcMem = CreateCompatibleDC(hdc);
	hPal = Lx_GetDrawablePalette(d);
	if (hPal) {
	    hPalMemPrev = SelectPalette(hdcMem, hPal, FALSE);
	}
	SetBkMode(hdcMem, TRANSPARENT);
	hbmp = CreateCompatibleBitmap(hdc, width, height);
	hbmpPrev = SelectObject(hdcMem, hbmp);
	hpenPrev = SelectObject(hdcMem, hpen);
	hbrPrev = SelectObject(hdcMem, hbr);

	hbrPat = CreatePatternBrush(gcv->stipple);
	SetBrushOrgEx(hdc, gcv->ts_x_origin, gcv->ts_y_origin, NULL);
	hbrpatPrev = SelectObject(hdc, hbrPat);

	PatBlt(hdcMem, 0, 0, width, height, BLACKNESS);
	Polygon(hdcMem, xp, n);
	BitBlt(hdc, x, y, width, height, hdcMem, 0, 0, 0x00AE0B09);

	PatBlt(hdcMem, 0, 0, width, height, WHITENESS);
	Polygon(hdcMem, xp, n);
	BitBlt(hdc, x, y, width, height, hdcMem, 0, 0, 0x00A803A9);

	SelectObject(hdc, hbrpatPrev);
	SelectObject(hdcMem, hpenPrev);
	SelectObject(hdcMem, hbrPrev);
	SelectObject(hdcMem, hbmpPrev);
	if (hPal) {
	    SelectPalette(hdcMem, hPalMemPrev, FALSE);
	}

	DeleteObject(hbrPat);
	DeleteObject(hbmp);
	DeleteDC(hdcMem);
     } else {
	hpenPrev = SelectObject(hdc, hpen);
	hbrPrev = SelectObject(hdc, hbr);
	Polygon(hdc, xp, n);
	SelectObject(hdc, hpenPrev);
	SelectObject(hdc, hbrPrev);
    }
    DeleteObject(hbr);
    DeleteObject(hpen);
    Lx_ReleaseDrawableDC(d, hdc, &state);
}

void  
XDrawRectangle(Display *display, Drawable d, GC gc, int x, int y,
	       unsigned int width, unsigned int height)
{
    HDC hdc;
    HPEN hPen, hPrevPen;
    POINT points[5];
    XGCValues *gcv;
    COLORREF fg;
    Lx_DrawableDCState state;

    gcv = &gc->values;
    display->request++;

    points[0].x = points[1].x = points[4].x = x; 
    points[2].x = points[3].x = x + width;
    points[0].y = points[3].y = points[4].y = y;
    points[1].y = points[2].y = y + height;

    hdc = Lx_GetDrawableDC(d, &state);
    SetROP2(hdc, gcv->function);
    if (gc->rects == True) {
	SelectObject(hdc, ((Region) gcv->clip_mask)->hrgn);
    }
    if (state.pixInfo && state.pixInfo->depth == 1) {
	fg = gcv->foreground ? RGB(0,0,0) : RGB(255,255,255);
    } else {
	fg = LX_CONVERTPIXEL(gcv->foreground);
    }
    hPen = CreatePen(gcv->line_style, gcv->line_width, fg);
    hPrevPen = SelectObject(hdc, hPen);
    Polyline(hdc, points, 5);
    SelectObject(hdc, hPrevPen);
    DeleteObject(hPen);
    Lx_ReleaseDrawableDC(d, hdc, &state);
}

void
XFillRectangle(Display *display, Drawable d, GC gc, int x, int y, 
	       unsigned int width, unsigned int height)
{
    XRectangle rectangle;
    rectangle.x = x;
    rectangle.y = y;
    rectangle.width = width;
    rectangle.height = height;
    XFillRectangles(display, d, gc, &rectangle, 1);
}

void
XFillRectangles(Display *display, Drawable d, GC gc, XRectangle *rectangles,
		int nrectangles)
{
    HDC hdc;
    HBRUSH hbr, hbrPrev;
    HDC hdcMem;
    HBITMAP hbmp, hbmpPrev;
    HBRUSH hbrPat, hbrPrevPat;
    HPALETTE hPal, hPalMemPrev;
    int rectno;
    int width, height;
    int x, y;
    XGCValues *gcv;
    RECT rect;
    Lx_DrawableDCState state;
    COLORREF fg;

    gcv = &gc->values;
    display->request++;

    hdc = Lx_GetDrawableDC(d, &state);
    SetROP2(hdc, gcv->function);

    if (state.pixInfo && state.pixInfo->depth == 1) {
	fg = gcv->foreground ? RGB(0,0,0) : RGB(255,255,255);
    } else {
	fg = LX_CONVERTPIXEL(gcv->foreground);
    }
    hbr = CreateSolidBrush(fg);
    if ((gcv->fill_style == FillStippled ||
	 gcv->fill_style == FillOpaqueStippled)
	&& gcv->stipple != None)
    {
	hdcMem = CreateCompatibleDC(hdc);
	hPal = Lx_GetDrawablePalette(d);
	if (hPal) {
	    hPalMemPrev = SelectPalette(hdcMem, hPal, FALSE);
	}
	hbrPrev = SelectObject(hdcMem, hbr);
	SetBkMode(hdcMem, TRANSPARENT);
	hbrPat = CreatePatternBrush(gcv->stipple);
	SetBrushOrgEx(hdc, gcv->ts_x_origin, gcv->ts_y_origin, NULL);
	hbrPrevPat = SelectObject(hdc, hbrPat);
	for (rectno = 0; rectno < nrectangles; rectno++) {
	    width = rectangles[rectno].width;
	    height = rectangles[rectno].height;
	    x = rectangles[rectno].x;
	    y = rectangles[rectno].y;

	    hbmp = CreateCompatibleBitmap(hdc, width, height);
	    hbmpPrev = SelectObject(hdcMem, hbmp);

	    PatBlt(hdcMem, 0, 0, width, height, BLACKNESS);
	    Rectangle(hdcMem, 0, 0, width, height );
	    BitBlt(hdc, x, y, width, height, hdcMem, 0, 0, 0x00AE0B09);

	    PatBlt(hdcMem, 0, 0, width, height, WHITENESS);
	    Rectangle(hdcMem, 0, 0, width, height );
	    BitBlt(hdc, x, y, width, height, hdcMem, 0, 0, 0x00A803A9);

	    SelectObject(hdcMem, hbmpPrev);
	    DeleteObject(hbmp);
	}
	SelectObject(hdc, hbrPrevPat);
	DeleteObject(hbrPat);
	SelectObject(hdcMem, hbrPrev);
	if (hPal) {
	    SelectPalette(hdcMem, hPalMemPrev, FALSE);
	}
	DeleteDC(hdcMem);
    } else {

	for (rectno = 0; rectno < nrectangles; rectno++) {
	    rect.left = rectangles[rectno].x;
	    rect.top = rectangles[rectno].y;
	    rect.right = rectangles[rectno].x + rectangles[rectno].width;
	    rect.bottom = rectangles[rectno].y + rectangles[rectno].height;
	    FillRect(hdc, &rect, hbr);
	}
    }
	
    DeleteObject(hbr);
    Lx_ReleaseDrawableDC(d, hdc, &state);
}

/* XXX: This routine can be sped up by not drawing into an offscreen bitmap
 * unless there is a stipple active
 */
static void
DrawStringUtil(Display *display, Drawable d, GC gc, int x, int y,
	       const char *string, int length, int imageString)
{
    HDC hdc, hdcMem;
    HBITMAP hbmp, hbmpPrev;
    HBRUSH hbr, hbrPrev;
    HFONT hfontPrev;
    SIZE sz;
    TEXTMETRIC tm;
    XGCValues *gcv;
    COLORREF fg, bg;
    HPALETTE hPal, hPalMemPrev;
    Lx_DrawableDCState state;

    gcv = &gc->values;
    hdc = Lx_GetDrawableDC(d, &state);
    SetROP2(hdc, gcv->function);
    if (gc->rects == True) {
	SelectObject(hdc, ((Region) gcv->clip_mask)->hrgn);
    }

    if (state.pixInfo && state.pixInfo->depth == 1) {
	fg = gcv->foreground ? RGB(0,0,0) : RGB(255,255,255);
	bg = gcv->background ? RGB(0,0,0) : RGB(255,255,255);
    } else {
	fg = LX_CONVERTPIXEL(gcv->foreground);
	bg = LX_CONVERTPIXEL(gcv->background);
    }
    if ((gcv->fill_style == FillStippled
	 || gcv->fill_style == FillOpaqueStippled)
	&& gcv->stipple != None)
    {
	hdcMem = CreateCompatibleDC(hdc);
	hPal = Lx_GetDrawablePalette(d);
	if (hPal) {
	    hPalMemPrev = SelectPalette(hdcMem, hPal, FALSE);
	}

	if (gcv->font != None) {
	    hfontPrev = SelectObject(hdcMem, gcv->font);
	}
	GetTextExtentPoint(hdcMem, string, length, &sz);
	GetTextMetrics(hdcMem, &tm);
	hbmp = CreateCompatibleBitmap(hdc, sz.cx + 10, sz.cy + 10);
	hbmpPrev = SelectObject(hdcMem, hbmp);
	SetTextAlign(hdcMem, TA_LEFT | TA_TOP);
	SetTextColor(hdcMem, fg);

	hbr = CreatePatternBrush(gcv->stipple);
	SetBrushOrgEx(hdc, gcv->ts_x_origin, gcv->ts_y_origin, NULL);

	if (imageString) {
	    SetBkMode(hdcMem, OPAQUE);
	    SetBkColor(hdcMem, bg);
	} else {
	    SetBkMode(hdcMem, TRANSPARENT);
	    SetBkColor(hdcMem, RGB(0, 0, 0));
	}
	PatBlt(hdcMem, 0, 0, sz.cx+10, sz.cy+10, BLACKNESS);
	TextOut(hdcMem, 0, 0, string, length);

	hbrPrev = SelectObject(hdc, hbr);
	BitBlt(hdc, x, y - tm.tmAscent, sz.cx, sz.cy, hdcMem, 0, 0, 0x00AE0B06); 

	PatBlt(hdcMem, 0, 0, sz.cx+10, sz.cy+10, WHITENESS);
	SetBkColor(hdcMem, RGB(255, 255, 255));
	TextOut(hdcMem, 0, 0, string, length);

	BitBlt(hdc, x, y - tm.tmAscent, sz.cx, sz.cy, hdcMem, 0, 0, 0x00A803A9); 

	if (gcv->font != None) {
	    SelectObject(hdcMem, hfontPrev);
	}
	SelectObject(hdc, hbrPrev);
	SelectObject(hdcMem, hbmpPrev);
	if (hPal) {
	    SelectPalette(hdcMem, hPalMemPrev, FALSE);
	}

	DeleteObject(hbr);
	DeleteObject(hbmp);
	DeleteDC(hdcMem);
    } else {
	SetTextAlign(hdc, TA_LEFT | TA_TOP);
	SetTextColor(hdc, fg);
	if (imageString) {
	    SetBkMode(hdc, OPAQUE);
	    SetBkColor(hdc, bg);
	} else {
	    SetBkMode(hdc, TRANSPARENT);
	}
	if (gcv->font != None) {
	    hfontPrev = SelectObject(hdc, gcv->font);
	}

	GetTextMetrics(hdc, &tm);
	TextOut(hdc, x, y - tm.tmAscent, string, length);

	if (gcv->font != None) {
	    SelectObject(hdc, hfontPrev);
	}
    }
    Lx_ReleaseDrawableDC(d, hdc, &state);
}
    
void 
XDrawString(Display *display, Drawable d, GC gc, int x, int y,
	    const char *string, int length)
{
    display->request++;

    DrawStringUtil(display, d, gc, x, y, string, length, 0);
}

void 
XDrawImageString(Display *display, Drawable d, GC gc, int x, int y,
	    const char *string, int length)
{
    display->request++;

    DrawStringUtil(display, d, gc, x, y, string, length, 1);
}

#ifdef KANJI 
void 
XDrawString16(Display *display, Drawable d, GC gc, int x, int y,
	      const XChar2b *string, int length)
{
    char *bp;
    char *buffer;
    XChar2b *wp;
    XChar2b *cp;
    int cnt = 0, i;
    BYTE c1, c2;

    display->request++;

    cp = (XChar2b *) ckalloc(sizeof(XChar2b) * length);
    memcpy(cp, string, sizeof(XChar2b) * length); 
    for(i = 0, wp = cp;  (i < length) && (wp->byte2 != 0); wp++, i++) {
	if (wp->byte1) {
	    c2 = wp->byte2;
	    c1 = wp->byte1;
	    wp->byte1 = (c1 - 0x21) / 2 + ((c1 <= 0x5e) ? 0x81 : 0xc1);
	    if( c1 & 1 ) {	/* odd */
		wp->byte2 = c2 + ((c2 <= 0x5f) ? 0x1f : 0x20);
	    } else {
		wp->byte2 = c2 + 0x7e;
	    }
	    cnt += 2;
	} else {
	    cnt++;
	}
    }
    buffer = (char *) ckalloc(cnt + 1);
    for(i = 0, wp = cp, bp = buffer;(i < length) && (wp->byte2 != 0); wp++, i++) {
	if (_ismbblead(wp -> byte1)) {
	    *bp++ = wp -> byte1;
	    *bp++ = wp -> byte2;
	} else {
	    *bp++ = wp -> byte2;
	}
    }
    *bp = '\0';    	
    XDrawString(display, d, gc, x, y, buffer, cnt);
    ckfree(buffer);
    ckfree(cp);
}
#endif /* KANJI */ 
