/*
 * xpixmap.c --
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */
 
#include "xlibInt.h"
#include <x11/Xproto.h>

static char *ReadBitmapFile (const char *filename, unsigned int *width,
    unsigned int *height, int *x_hot, int *y_hot);


#define BMP_PLANES 1
#define BMP_PIXELS 1

static unsigned char
swap(unsigned char byte)
{
    register int i;
    register unsigned char bit = 0x80;
    register unsigned char result = 0;

    byte = ~byte;
    for (i = 0; i < 8; i++) {
   	if (byte & 0x01) {
   	    result |= bit;
   	}
   	byte >>= 1;
   	bit >>= 1;
    }
    return result;
}     	

/*
 *----------------------------------------------------------------------
 * AddPixmap --
 *
 *	Add a pixmap to our list of known pixmaps.  We need to
 *	do this to try to distinguish between pixmaps and windows
 *	when determining what type a drawable is.
 *----------------------------------------------------------------------
 */
static void
AddPixmap(Display *display, Drawable d, Pixmap pixmap,
	  unsigned int width, unsigned int height, unsigned int depth)
{
    Lx_PixElem *pixElem;

    pixElem = (Lx_PixElem *) ckalloc(sizeof(Lx_PixElem));
    pixElem->next = display->pixmapList;
    pixElem->drawable = d;
    pixElem->pixmap = pixmap;
    pixElem->width = width;
    pixElem->height = height;
    pixElem->depth = depth;
    display->pixmapList = pixElem;
}

/*
 *----------------------------------------------------------------------
 * RemovePixmap --
 *
 *	Removes a pixmap from our list of known pixmaps
 *----------------------------------------------------------------------
 */
static void
RemovePixmap(Display *display, Pixmap pixmap)
{
    Lx_PixElem *prev, *curr, *next;

    for (prev = NULL, curr = display->pixmapList; curr != NULL;
	 prev = curr, curr = next)
    {
	next = curr->next;
	if (curr->pixmap == pixmap) {
	    if (prev != NULL) {
		prev->next = curr->next;
	    } else {
		display->pixmapList = curr->next;
	    }
	    ckfree((char *) curr);
	    return;
	}
    }
}

/*
 *----------------------------------------------------------------------
 * Lx_IsPixmap --
 *
 *	Returns TRUE if the drawable is a pixmap.
 *----------------------------------------------------------------------
 */
int
Lx_IsPixmap(Drawable drawable)
{
    Lx_PixElem *curr;
    Display *display;

    display = XOpenDisplay(NULL);
    for (curr = display->pixmapList; curr != NULL; curr = curr->next) {
	if (curr->pixmap == drawable) {
	    return TRUE;
	}
    }
    return FALSE;
}

/*
 *----------------------------------------------------------------------
 * Lx_FindPixmap --
 *
 *	Attempt to find a matching pixmap
 *----------------------------------------------------------------------
 */
Lx_PixElem *
Lx_FindPixmap(Drawable drawable)
{
    Lx_PixElem *curr;
    Display *display;

    display = XOpenDisplay(NULL);
    for (curr = display->pixmapList; curr != NULL; curr = curr->next) {
	if (curr->pixmap == drawable) {
	    return curr;
	}
    }
    return NULL;
}
/*
 *----------------------------------------------------------------------
 * Lx_GetDrawableWindow --
 *
 *	Gets the window associated with this drawable.  This may be
 *	the drawable itself.
 *----------------------------------------------------------------------
 */
Window
Lx_GetDrawableWindow(Drawable drawable)
{
    Lx_PixElem *elem, *pixElem;

    elem = Lx_FindPixmap(drawable);
    if (! elem) {
	return NULL;
    }
    while (elem) {
	if (elem->drawable == 0) {
	    return 0;
	}
	pixElem = Lx_FindPixmap(elem->drawable);
	if (! pixElem) {
	    return elem->drawable;
	}
	elem = pixElem;
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 * AddCursor --
 *
 *	Add a cursor to the list of cursors we know about
 *----------------------------------------------------------------------
 */
static void
AddCursor(Display *display, Cursor cursor)
{
    Lx_CursorElem *cursorElem;

    cursorElem = (Lx_CursorElem *) ckalloc(sizeof(Lx_CursorElem));
    cursorElem->next = display->cursorList;
    cursorElem->cursor = cursor;
    display->cursorList = cursorElem;
}

/*
 *----------------------------------------------------------------------
 * RemoveCursors --
 *
 *	Remove a cursor from the list of known cursors.
 *
 * Result:
 *	TRUE if the cursor was found, FALSE if it wasn't
 *----------------------------------------------------------------------
 */
static int
RemoveCursor(Display *display, Cursor cursor)
{
    Lx_CursorElem *prev, *curr, *next;

    for (prev = NULL, curr = display->cursorList; curr != NULL;
	 prev = curr, curr = next)
    {
	next = curr->next;
	if (curr->cursor == cursor) {
	    if (prev != NULL) {
		prev->next = curr->next;
	    } else {
		display->cursorList = curr->next;
	    }
	    ckfree((char *) curr);
	    return TRUE;
	}
    }
    return FALSE;
}

/*
 *----------------------------------------------------------------------
 * XCreateBitmapFromData --
 *
 *	Create a pixmap of depth 1 from the data
 *----------------------------------------------------------------------
 */
Pixmap
XCreateBitmapFromData(Display *display, Drawable drawable, const char *data,
	unsigned int width, unsigned int height)
{
    int byteWidth, dataWidth;
    int i, j, adjust = 0;
    BYTE *bPtr;
    HBITMAP hbmp;

    display->request++;

    byteWidth = dataWidth = (width + 7) / 8;

    if (dataWidth % 2) {
	byteWidth += 1;
    	adjust = 1;
    }

    bPtr = (BYTE *) ckalloc(byteWidth * height);
    for (i = 0; (unsigned)i < height; i++) {
        for (j = 0; j < byteWidth; j++) {
	    if (j == (byteWidth - 1) && adjust) {
	    	bPtr[i * byteWidth + j] = '\0';
	    } else {
		bPtr[i * byteWidth + j] = swap((BYTE)data[i * dataWidth + j]); 
	    }
    	}
    }
    hbmp = CreateBitmap(width, height, BMP_PLANES, BMP_PIXELS, bPtr);
    ckfree(bPtr);

    AddPixmap(display, drawable, hbmp, width, height, 1);
    return hbmp ;
}

/*
 *----------------------------------------------------------------------
 * XScrollWindow --
 *
 *	Not an true X routine, but was going to be used to do scrolling
 *	that would handle windows that were partially covered or off
 *	the screen.
 *----------------------------------------------------------------------
 */
void 
XScrollWindow(Display *display, Drawable w, GC gc, 
	int src_x, int src_y, unsigned int width, unsigned int height, 
	int dest_x, int dest_y)
{
    HDC hdc;
    HRGN hrgn;
    RECT rect, invalid;
    DWORD dwCount;
    int ret;
    XEvent event;
    LPRGNDATA lpRgnData;
    LPRECT lpRect;
    HWND postWin;
    Lx_DrawableDCState state;

    hdc = Lx_GetDrawableDC(w, &state);
    hrgn = CreateRectRgn(0, 0, 0, 0);
    GetWindowRect(w, &rect);

    rect.left = dest_x;
    rect.right = dest_x + width;
    if (dest_y < src_y) {
	rect.top = dest_y;
	rect.bottom = src_y + height;
    } else {
	rect.top = src_y;
	rect.bottom = dest_y + height;
    }
    
    ret = ScrollWindowEx(w, 0, dest_y - src_y, &rect, NULL, hrgn, &invalid, 0);
    if (gc->values.graphics_exposures) {
	postWin = Tknt_GetPostWindow();
	if (ret == NULLREGION) {
	  noexpose:
	    event.xnoexpose.type = NoExpose;
	    event.xnoexpose.serial = display->request;
	    event.xnoexpose.send_event = False;
	    event.xnoexpose.display = display;
	    event.xnoexpose.drawable = w;
	    event.xnoexpose.major_code = X_CopyPlane;
	    Lx_PutEvent(display, &event);
	    PostMessage(postWin, EVENTADDED, 0, 0L);
	} else if (ret == SIMPLEREGION) {
	    event.xany.type = GraphicsExpose;
	    event.xgraphicsexpose.serial = display->request;
	    event.xgraphicsexpose.send_event = False;
	    event.xgraphicsexpose.display = display;
	    event.xgraphicsexpose.drawable = w;
	    event.xgraphicsexpose.x = invalid.left;
	    event.xgraphicsexpose.y = invalid.top;
	    event.xgraphicsexpose.width = invalid.right - invalid.left;
	    event.xgraphicsexpose.height = invalid.bottom - invalid.top;
	    event.xgraphicsexpose.count = 0;
	    event.xgraphicsexpose.major_code = X_CopyArea;
	    event.xgraphicsexpose.minor_code = 0;
	    Lx_PutEvent(display, &event);
	    PostMessage(postWin, EVENTADDED, 0, 0L);
	} else {
	    dwCount = GetRegionData(hrgn, 0, NULL);
	    lpRgnData = (LPRGNDATA) ckalloc(dwCount);
	    if (lpRgnData == NULL) {
		goto noexpose;
	    }
	    ret = GetRegionData(hrgn, dwCount, lpRgnData);
	    lpRect = (LPRECT) lpRgnData->Buffer;
	    dwCount = lpRgnData->rdh.nCount;
	    while (dwCount > 0) {
		dwCount--;
		event.xany.type = GraphicsExpose;
		event.xgraphicsexpose.serial = display->request;
		event.xgraphicsexpose.send_event = False;
		event.xgraphicsexpose.display = display;
		event.xgraphicsexpose.drawable = w;
		event.xgraphicsexpose.x = lpRect->left;
		event.xgraphicsexpose.y = lpRect->top;
		event.xgraphicsexpose.width = lpRect->right - lpRect->left;
		event.xgraphicsexpose.height = lpRect->bottom - lpRect->top;
		event.xgraphicsexpose.count = dwCount;
		event.xgraphicsexpose.major_code = X_CopyArea;
		event.xgraphicsexpose.minor_code = 0;
		Lx_PutEvent(display, &event);
		lpRect++;
	    }
	    PostMessage(postWin, EVENTADDED, 0, 0L);
	    ckfree(lpRgnData);
	}
    }
	
    DeleteObject(hrgn);
    Lx_ReleaseDrawableDC(hdc, w, &state);
}

/*
 *----------------------------------------------------------------------
 * XCopyArea --
 *
 *	Copy an area of a drawable
 *
 * Notes:
 *	XXX: This routine does not cause proper expose events to occur if
 *	part of the image is covered by another toplevel window or is off
 *	the screen.
 *----------------------------------------------------------------------
 */
void 
XCopyArea(Display *display, Drawable src, Drawable dest, GC gc, 
	int src_x, int src_y, unsigned int width, unsigned int height, 
	int dest_x, int dest_y)
{
    HDC hsrc, hdest;
    XEvent event;
    Lx_DrawableDCState dstate, sstate;

    /* Create a NoExpose event.  If we could figure out how to determine
     * which parts of the window are not visible under Windows, we could
     * create GraphicsExpose events
     */

    display->request++;

    if (gc->values.graphics_exposures) {
	event.xnoexpose.type = NoExpose;
	event.xnoexpose.serial = display->request;
	event.xnoexpose.send_event = False;
	event.xnoexpose.display = display;
	event.xnoexpose.drawable = dest;
	event.xnoexpose.major_code = X_CopyArea;
	Lx_PutEvent(display, &event);
	PostMessage(Tknt_GetPostWindow(), EVENTADDED, 0, 0L);
    }

    hdest = Lx_GetDrawableDC(dest, &dstate);
    hsrc = (dest == src) ? hdest : Lx_GetDrawableDC(src, &sstate);

    BitBlt(hdest, dest_x, dest_y, width, height,
	   hsrc,  src_x,  src_y, SRCCOPY);

    if (dest != src) {
	Lx_ReleaseDrawableDC(src, hsrc, &sstate);
    }
    Lx_ReleaseDrawableDC(dest, hdest, &dstate);
}

void
XCopyPlane(Display *display, Drawable src, Drawable dest, GC gc,
	int src_x, int src_y, unsigned int width, unsigned int height, 
	int dest_x, int dest_y, unsigned long dummy)
{
    HDC hsrc, hdest;
    HBRUSH fgBrush, bgBrush, hPrevBrush;
    XEvent event;
    Lx_DrawableDCState dstate, sstate;
    COLORREF bg, fg;

    /* Create a NoExpose event.  If we could figure out how to determine
     * which parts of the window are not visible under Windows, we could
     * create GraphicsExpose events
     */
    display->request++;

    if (gc->values.graphics_exposures) {
	event.xnoexpose.type = NoExpose;
	event.xnoexpose.serial = display->request;
	event.xnoexpose.send_event = False;
	event.xnoexpose.display = display;
	event.xnoexpose.drawable = dest;
	event.xnoexpose.major_code = X_CopyPlane;
	Lx_PutEvent(display, &event);
	PostMessage(Tknt_GetPostWindow(), EVENTADDED, 0, 0L);
    }

    hdest = Lx_GetDrawableDC(dest, &dstate);
    hsrc = (dest == src) ? hdest : Lx_GetDrawableDC(src, &sstate);

    bg = LX_CONVERTPIXEL(gc->values.background);
    fg = LX_CONVERTPIXEL(gc->values.foreground);
    bgBrush = CreateSolidBrush(bg);
    fgBrush = CreateSolidBrush(fg);

    if (gc->values.clip_mask != None) {
	if (gc->values.clip_mask != src) {
	    printf("XCopyPlane: Unsupported clipping mask feature");
	}
	hPrevBrush = SelectObject(hdest, fgBrush);
	BitBlt(hdest, dest_x, dest_y, width, height, hsrc, src_x, src_y,
	       0xB8074A);
	SelectObject(hdest, hPrevBrush);
    } else {
	hPrevBrush = SelectObject(hdest, bgBrush);
	BitBlt(hdest, dest_x, dest_y, width, height, hsrc, src_x, src_y,
	       0xF00021);

	SelectObject(hdest, fgBrush);
	BitBlt(hdest, dest_x, dest_y, width, height, hsrc, src_x, src_y,
	       0xB8074A);
	SelectObject(hdest, hPrevBrush);
    }

    DeleteObject(fgBrush);
    DeleteObject(bgBrush);

    if (dest != src) {
	Lx_ReleaseDrawableDC(src, hsrc, &sstate);
    }
    Lx_ReleaseDrawableDC(dest, hdest, &dstate);
}

/*
 *----------------------------------------------------------------------
 * XCreatePixmap --
 *
 *	Create a new pixmap.  Right now, only 8 and 24 bit pixmaps are
 *	created.
 *----------------------------------------------------------------------
 */
Pixmap
XCreatePixmap(Display *display, Drawable drawable,
	unsigned int width, unsigned int height, unsigned int depth)
{
    BITMAPINFOHEADER bmih;
    HBITMAP hBitmap;
    HDC hdc;

    display->request++;

    if (depth == 1) {
	hBitmap = CreateBitmap(width, height, 1, depth, NULL);
    } else {
	bmih.biSize = sizeof(BITMAPINFOHEADER);
	bmih.biWidth = (DWORD) width;
	bmih.biHeight = (DWORD) height;
	bmih.biPlanes = 1;
	bmih.biBitCount = depth;
	bmih.biCompression = 0;
	bmih.biSizeImage = 0;
	bmih.biXPelsPerMeter = 0;
	bmih.biYPelsPerMeter = 0;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;
	hdc = GetDC(NULL);
	hBitmap = CreateDIBitmap(hdc, &bmih, 0L, NULL, NULL, 0);
	ReleaseDC(NULL, hdc);
    }
    AddPixmap(display, drawable, hBitmap, width, height, depth);
    return (Pixmap) hBitmap;
}

/*
 *----------------------------------------------------------------------
 * XFreePixmap --
 *
 *	Free up an allocated pixmap
 *----------------------------------------------------------------------
 */
void
XFreePixmap(Display *display, Pixmap pixmap)
{
    display->request++;

    if (pixmap != NULL) {
	DeleteObject(pixmap);
	RemovePixmap(display, pixmap);
    }
}

/*
 *----------------------------------------------------------------------
 * XCreatePixmapCursor --
 *
 *	Create a cursor from a Pixmap.  The pixmap is scaled to form
 *	a cursor of the proper Windows size.
 *----------------------------------------------------------------------
 */
Cursor
XCreatePixmapCursor(Display *display, Pixmap source, Pixmap mask,
		    XColor *fg_color, XColor *bg_color,
		    unsigned int x, unsigned int y)
{
    HICON hCursor;
    ICONINFO iconinfo;
    HDC hdc;
    LPBITMAPINFO lpbi;
    DWORD dwNeeded;
    unsigned char *srcData, *colorData, *maskData;
    int h, w;
    LONG height, width;
    DWORD inBytesPerLine, outBytesPerLine;
    unsigned char *inLineP, *outLineP, *maskLineP;
    unsigned char *inP, *outP, *maskP;
    unsigned int bit, bitmask;
    HBITMAP hBitmap;
    int ret;
    BITMAP bm;
    BYTE bg_red, bg_blue, bg_green, fg_red, fg_blue, fg_green;

    display->request++;

    GetObject(source, sizeof(bm), (LPSTR)&bm);
    if (bm.bmPlanes * bm.bmBitsPixel != 1) {
	/* Error */
	return 0;
    }

    height = bm.bmHeight;
    width = bm.bmHeight;
    hdc = GetDC(NULL);
    lpbi = ckalloc(sizeof(BITMAPINFO) + 2 * sizeof(RGBQUAD));
    lpbi->bmiHeader.biSize               = sizeof(BITMAPINFOHEADER);
    lpbi->bmiHeader.biWidth              = width;
    lpbi->bmiHeader.biHeight             = height;
    lpbi->bmiHeader.biPlanes             = 1;
    lpbi->bmiHeader.biBitCount	         = 1;
    lpbi->bmiHeader.biCompression	 = 0;
    lpbi->bmiHeader.biSizeImage          = 0;
    lpbi->bmiHeader.biXPelsPerMeter      = 0;
    lpbi->bmiHeader.biYPelsPerMeter      = 0;
    lpbi->bmiHeader.biClrUsed            = 0;
    lpbi->bmiHeader.biClrImportant       = 0;

    ret = GetDIBits(hdc, source, 0, height, NULL, lpbi, DIB_RGB_COLORS);

    srcData = colorData = maskData = NULL;
    hCursor = 0;

    inBytesPerLine = ((((width + 7) / 8) + 3) / 4) * 4;
    outBytesPerLine = ((width * 3 + 3) / 4) * 4;
    dwNeeded = outBytesPerLine * height;
    srcData = ckalloc(lpbi->bmiHeader.biSizeImage);
    colorData = ckalloc(dwNeeded);
    maskData = ckalloc(lpbi->bmiHeader.biSizeImage * 2);

    if (srcData == NULL || colorData == NULL || maskData == NULL) {
	goto free;
    }

    if (mask) {
	GetObject(source, sizeof(bm), (LPSTR)&bm);

	if (bm.bmPlanes * bm.bmBitsPixel != 1) {
	    goto free;
	}
	if (lpbi->bmiHeader.biWidth != width ||
	    lpbi->bmiHeader.biHeight != height)
	{
	    goto free;
	}
	ret = GetDIBits(hdc, mask, 0, height, maskData, lpbi, DIB_RGB_COLORS);
    } else {
	ret = GetDIBits(hdc, source, 0, height, maskData, lpbi, DIB_RGB_COLORS);
    }

#if 0
    ret = GetDIBits(hdc, mask, 0, height, maskData, lpbi, DIB_RGB_COLORS);
    printf("-------------------------------------\n");
    for (h = 0, inLineP = maskData; h < height; h++, inLineP += inBytesPerLine)
    {
	for (bit = 7, w = 0, inP = inLineP; w < width; w++, bit--)
	{
	    if (bit > 7) {
		inP++;
		bit = 7;
	    }
	    bitmask = 1 << bit;
	    if (*inP & bitmask) {
		printf(" ");
	    } else {
		printf("X");
	    }
	}
	printf("\n");
    }
#endif

    ret = GetDIBits(hdc, source, 0, height, srcData, lpbi,
		    DIB_RGB_COLORS);

    bg_red = (BYTE) (bg_color->red >> 8);
    bg_green = (BYTE) (bg_color->green >> 8);
    bg_blue = (BYTE) (bg_color->blue >> 8);

    fg_red = (BYTE) (fg_color->red >> 8);
    fg_green = (BYTE) (fg_color->green >> 8);
    fg_blue = (BYTE) (fg_color->blue >> 8);

    for (h = 0, inLineP = srcData, outLineP = colorData, maskLineP = maskData;
	 h < height;
	 h++, inLineP += inBytesPerLine, outLineP += outBytesPerLine,
	 maskLineP += inBytesPerLine)
    {
	for (bit = 7, w = 0, inP = inLineP, outP = outLineP, maskP = maskLineP;
	     w < width; w++, bit--)
	{
	    if (bit > 7) {
		inP++;
		maskP++;
		bit = 7;
	    }
	    bitmask = 1 << bit;
	    if (! (*maskP & bitmask)) {
		if (*inP & bitmask) {
		    *outP++ = bg_blue;
		    *outP++ = bg_green;
		    *outP++ = bg_red;
		} else {
		    *outP++ = fg_blue;
		    *outP++ = fg_green;
		    *outP++ = fg_red;
		}
	    } else {
		*outP++ = 0x0;
		*outP++ = 0x0;
		*outP++ = 0x0;
	    }
	}
    }

    lpbi->bmiHeader.biSizeImage = 0;
    lpbi->bmiHeader.biBitCount = 24;
    hBitmap = CreateDIBitmap(hdc, &lpbi->bmiHeader, CBM_INIT, colorData, lpbi, 
			     DIB_RGB_COLORS);

    iconinfo.fIcon = FALSE;
    iconinfo.xHotspot = x;
    iconinfo.yHotspot = y;

    iconinfo.hbmMask = (HBITMAP) mask;
    iconinfo.hbmColor = (HBITMAP) hBitmap /* source */;

    hCursor = CreateIconIndirect(&iconinfo);
    AddCursor(display, (Cursor) hCursor);

    DeleteObject(hBitmap);

 free:
    ReleaseDC(NULL, hdc);
    if (srcData) {
	ckfree(srcData);
    }
    if (colorData) {
	ckfree(colorData);
    }
    if (maskData) {
	ckfree(maskData);
    }
    return (Cursor) hCursor;
}

/*
 *----------------------------------------------------------------------
 * XFreeCursor --
 *
 *	Frees up an allocated cursor
 *----------------------------------------------------------------------
 */
void
XFreeCursor(Display *display, Cursor cursor)
{
    display->request++;

    if (cursor != NULL) {
	if (RemoveCursor(display, cursor)) {
	    DestroyIcon(cursor);
	}
    }
}

static int
Lx_XDestroyImage(XImage *ximage)
{
    ckfree(ximage->data);
    ckfree(ximage);
    return Success;
}

static int
Lx_XPutPixel(XImage *ximage, int x, int y, unsigned long pixel)
{
    unsigned char *pc;
    unsigned short *ps;
    unsigned long *pl;

    pc = ximage->data + ximage->bytes_per_line * y;
    switch (ximage->bits_per_pixel) {
      case 1:
	pc += x / 8;
	if (pixel) {
	    *pc |= (1 << (7 - (x % 8)));
	} else {
	    *pc &= ~(1 << (7 - (x % 8)));
	}
	break;

      case 8:
	pc += x;
	*pc = (unsigned char) pixel;
	break;

      case 16:
	ps = (unsigned short *) pc;
	ps += x;
	*ps = (unsigned short) pixel;
	break;

      case 32:
	pl = (unsigned long *) pc;
	pl += x;
	*pl = pixel;
	break;
    }

    return Success;
}

static unsigned long
Lx_XGetPixel(XImage *ximage, int x, int y)
{
    unsigned char *pc;
    unsigned short *ps;
    unsigned long *pl;

    pc = ximage->data + ximage->bytes_per_line * y;
    switch (ximage->bits_per_pixel) {
      case 1:
	pc += x / 8;
	return (*pc & (1 << (7 - (x % 8)))) ? 1 : 0;

      case 8:
	pc += x;
	return *pc;

      case 16:
	ps = (unsigned short *) pc;
	ps += x;
	return *ps;

      case 32:
	pl = (unsigned long *) pc;
	pl += x;
	return *pl;

    default:
	return 0;
    }
}

/*
 *----------------------------------------------------------------------
 * XGetImage --
 *
 *	Place contents of a rectangle from drawable into an image
 *----------------------------------------------------------------------
 */
XImage *
XGetImage(Display *display, Drawable d, int x, int y, unsigned int width,
	  unsigned int height, unsigned long plm, int format)
{
    XImage *ret_image;
    unsigned char *smallBitData, *smallBitBase, *bigBitData;
    HDC hdc, hdcMem;
    HBITMAP hbmp, hbmpPrev;
    BITMAPINFO *bmInfo = NULL;
    HPALETTE hPal, hPalPrev1, hPalPrev2;
    unsigned int byte_width;
    unsigned int h, w, n;
    unsigned char plmr, plmg, plmb;
    unsigned char *data;
    int size;
    unsigned int depth;
    Lx_DrawableDCState state;
    BOOL ret;

    display->request++;

    if (format != ZPixmap) {
	/* XXX: Don't handle XYPixmap and XYBitmap for now */
	return NULL;
    }

    hdc = Lx_GetDrawableDC(d, &state);

    /* Need to do a Blt operation to copy into a new bitmap */
    hbmp = CreateCompatibleBitmap(hdc, width, height);
    hdcMem = CreateCompatibleDC(hdc);
    hbmpPrev = SelectObject(hdcMem, hbmp);

    hPal = Lx_GetDrawablePalette(d);
    if (hPal) {
        hPalPrev1 = SelectPalette(hdcMem, hPal, FALSE);
        n = RealizePalette(hdcMem);
        if (n > 0) {
            UpdateColors (hdcMem);
        }
	hPalPrev2 = SelectPalette(hdc, hPal, FALSE);
        n = RealizePalette(hdc);
        if (n > 0) {
            UpdateColors (hdc);
        }
    }

    ret = BitBlt(hdcMem, 0, 0, width, height, hdc, x, y, SRCCOPY);
    if (hPal) {
	SelectPalette(hdc, hPalPrev2, FALSE);
    }
    SelectObject(hdcMem, hbmpPrev);
    Lx_ReleaseDrawableDC(d, hdc, &state);
    if (ret == FALSE) {
	goto cleanup;
    }

    depth = (state.pixInfo) ? state.pixInfo->depth : lx_ScreenDepth;

    size = sizeof(BITMAPINFO);
    if (depth <= 8) {
	size += sizeof(unsigned short) * (1 << depth);
    }
    bmInfo = (BITMAPINFO *) ckalloc(size);

    bmInfo->bmiHeader.biSize               = sizeof(BITMAPINFOHEADER);
    bmInfo->bmiHeader.biWidth              = width;
    bmInfo->bmiHeader.biHeight             = -(int) height;
    bmInfo->bmiHeader.biPlanes             = 1;
    bmInfo->bmiHeader.biBitCount           = depth;
    bmInfo->bmiHeader.biCompression        = BI_RGB;
    bmInfo->bmiHeader.biSizeImage          = 0;
    bmInfo->bmiHeader.biXPelsPerMeter      = 0;
    bmInfo->bmiHeader.biYPelsPerMeter      = 0;
    bmInfo->bmiHeader.biClrUsed            = 0;
    bmInfo->bmiHeader.biClrImportant       = 0;

    if (depth == 1) {
	unsigned char *p, *pend;
	GetDIBits(hdcMem, hbmp, 0, height, NULL, bmInfo, DIB_PAL_COLORS);
	data = ckalloc(bmInfo->bmiHeader.biSizeImage);
	if (!data) {
	    /* printf("Failed to allocate data area for XImage.\n"); */
	    ret_image = NULL;
	    goto cleanup;
	}
	ret_image = XCreateImage(display, NULL, depth, ZPixmap, 0, data,
				 width, height, 32, ((width + 31) >> 3) & ~1);
	if (ret_image == NULL) {
	    ckfree(data);
	    goto cleanup;
	}

	/* Get the BITMAP info into the Image. */
	if (GetDIBits(hdcMem, hbmp, 0, height, data, bmInfo,
		      DIB_PAL_COLORS) == 0)
	{
	    ckfree(ret_image->data);
	    ckfree(ret_image);
	    ret_image = NULL;
	    goto cleanup;
	}
	p = data;
	pend = data + bmInfo->bmiHeader.biSizeImage;
	while (p < pend) {
	    *p = ~*p;
	    p++;
	}
#if 0
    {
	int i, j;
	for (i = 0; i < height; i++) {
	    for (j = 0; j < width; j++) {
		printf("%c", (*(data + i * ret_image->bytes_per_line +
				(j >> 3)) & (7 - (j % 8)) ? 'X' : '.'));
	    }
	    printf("\n");
	}
    }
#endif
	
    } else if (depth == 8) {
	unsigned short *palette;
	unsigned int i;
	unsigned char *p;

	GetDIBits(hdcMem, hbmp, 0, height, NULL, bmInfo, DIB_PAL_COLORS);
	data = ckalloc(bmInfo->bmiHeader.biSizeImage);
	if (!data) {
	    /* printf("Failed to allocate data area for XImage.\n"); */
	    ret_image = NULL;
	    goto cleanup;
	}
	ret_image = XCreateImage(display, NULL, 8, ZPixmap, 0, data,
				 width, height, 8, width);
	if (ret_image == NULL) {
	    ckfree(data);
	    goto cleanup;
	}

	/* Get the BITMAP info into the Image. */
	if (GetDIBits(hdcMem, hbmp, 0, height, data, bmInfo,
		      DIB_PAL_COLORS) == 0)
	{
	    ckfree(ret_image->data);
	    ckfree(ret_image);
	    ret_image = NULL;
	    goto cleanup;
	}
	p = data;
	palette = (unsigned short *) bmInfo->bmiColors;
	for (i = 0; i < bmInfo->bmiHeader.biSizeImage; i++, p++) {
	    *p = (unsigned char) palette[*p];
	}
#if 0
    {
	int i, j;
	for (i = 0; i < height; i++) {
	    for (j = 0; j < width; j++) {
		printf("%02x ", (*(data + i * ret_image->bytes_per_line + j)));
	    }
	    printf("\n");
	}
    }
#endif
    } else {
	GetDIBits(hdcMem, hbmp, 0, height, NULL, bmInfo, DIB_RGB_COLORS);
	data = ckalloc(width * height * 4);
	if (!data) {
	    /* printf("Failed to allocate data area for XImage.\n"); */
	    ret_image = NULL;
	    goto cleanup;
	}
	ret_image = XCreateImage(display, NULL, 32, ZPixmap, 0, data,
				 width, height, 0, width * 4);
	if (ret_image == NULL) {
	    ckfree(data);
	    goto cleanup;
	}

	byte_width = ((width * 3 + 3) & ~3);
	smallBitBase = ckalloc(byte_width * height);
	if (!smallBitBase) {
	    ckfree(ret_image->data);
	    ckfree(ret_image);
	    ret_image = NULL;
	    goto cleanup;
	}
	smallBitData = smallBitBase;

	/* Get the BITMAP info into the Image. */
	if (GetDIBits(hdcMem, hbmp, 0, height, smallBitData, bmInfo,
		      DIB_RGB_COLORS) == 0)
	{
	    ckfree(ret_image->data);
	    ckfree(ret_image);
	    ret_image = NULL;
	    goto cleanup;
	}

	plmr = (unsigned char) (plm & 0xff0000) >> 16;
	plmg = (unsigned char) (plm & 0x00ff00) >> 8;
	plmb = (unsigned char) plm & 0x0000ff;

	/* Copy the 24 Bit Pixmap to a 32-Bit one. */
	for (h = 0; h < height; h++) {
	    bigBitData = ret_image->data + h * ret_image->bytes_per_line;
	    smallBitData = smallBitBase + h * byte_width;

	    for (w = 0; w < width; w++) {
		*bigBitData++ = ((*smallBitData++)) /* & plmr */;
		*bigBitData++ = ((*smallBitData++)) /* & plmg */;
		*bigBitData++ = ((*smallBitData++)) /* & plmb */;
		*bigBitData++ = 0;
	    }
	}
	/* Free the Device contexts, and the Bitmap */
	ckfree(smallBitBase);
    }

  cleanup:
    if (bmInfo) {
	ckfree(bmInfo);
    }
    if (hPal) {
	SelectPalette(hdcMem, hPalPrev1, FALSE);
    }
    DeleteDC(hdcMem);
    DeleteObject(hbmp);

    return ret_image;
}

void
XPutImage(Display *display, Drawable d, GC gc, XImage *image,
	  int src_x, int src_y, int dest_x, int dest_y,
	  unsigned int width, unsigned int height)
{
    unsigned int h, w;
    unsigned int byte_width;
    HDC hdc;
    BITMAPINFO *bmInfo;
    unsigned char *bigBitData, *smallBitBase, *smallBitData;
    int size;
    unsigned short i;
    unsigned short *pindex;
    Lx_DrawableDCState state;
    int depth;

    if (image->format != ZPixmap) {
	/* XXX: Don't handle XYPixmap and XYBitmap for now */
	return /* BadImplementation */;
    }

    if (image->depth == 32) {
	depth = 24;
    } else if (image->depth == 8 || image->depth == 1) {
	depth = image->depth;
    } else {
	return /* BadImplementation */;
    }

    if (image->depth == 1) {
	unsigned char *in, *out, *end;
	int len;

	in = image->data;
	len = image->bytes_per_line * height;
	smallBitBase = out = ckalloc(len);
	end = out + len;
	while (out < end) {
	    *out = ~*in;
	    out++; in++;
	}
    } else if (image->depth > 8) {
	smallBitBase = ckalloc(width*height*(image->depth/8));
	if (!smallBitBase) {
	    /* printf("Failed to allocate data area output image.\n"); *
	    return /* BadAlloc */;
	}
    } else {
	smallBitBase = image->data;
    }

    if (image->depth == 32) {
	/* Copy the 32 Bit Pixmap to a 24 Bit one. */
	byte_width = ((width * 3 + 3) & ~3);

	for (h = 0; h < height; h++) {
	    bigBitData = image->data + h * image->bytes_per_line;
	    smallBitData = smallBitBase + h * byte_width;

	    for (w = 0; w < width; w++) {
		smallBitData[2] = ((*bigBitData++)) /* & plmr */;
		smallBitData[1] = ((*bigBitData++)) /* & plmg */;
		smallBitData[0] = ((*bigBitData++)) /* & plmb */;
		smallBitData += 3;
		bigBitData++;
	    }
	}
    }

    hdc = Lx_GetDrawableDC(d, &state);

    size = sizeof(BITMAPINFO);
    if (image->depth == 8) {
	size += sizeof(unsigned short) * (1 << depth);
    } else if (image->depth == 1) {
	size += sizeof(COLORREF) * 2;
    }

    bmInfo = (BITMAPINFO *) ckalloc(size);
    bmInfo->bmiHeader.biSize = (LONG)sizeof(BITMAPINFOHEADER);
    bmInfo->bmiHeader.biWidth = width;
    bmInfo->bmiHeader.biHeight = -(int) height;
    bmInfo->bmiHeader.biPlanes = 1;
    bmInfo->bmiHeader.biBitCount = depth;
    bmInfo->bmiHeader.biCompression = BI_RGB;
    bmInfo->bmiHeader.biSizeImage = 0L;
    bmInfo->bmiHeader.biXPelsPerMeter = 0L;
    bmInfo->bmiHeader.biYPelsPerMeter = 0L;
    bmInfo->bmiHeader.biClrUsed = 0L;
    bmInfo->bmiHeader.biClrImportant = 0L;

    if (image->depth == 1) {
	bmInfo->bmiColors[0].rgbRed = 0;
	bmInfo->bmiColors[0].rgbGreen = 0;
	bmInfo->bmiColors[0].rgbBlue = 0;
	bmInfo->bmiColors[0].rgbReserved = 0;
	bmInfo->bmiColors[1].rgbRed = 0xff;
	bmInfo->bmiColors[1].rgbGreen = 0xff;
	bmInfo->bmiColors[1].rgbBlue = 0xff;
	bmInfo->bmiColors[1].rgbReserved = 0xff;
    }
    if (image->depth == 8) {
	pindex = (unsigned short *) &bmInfo->bmiColors[0];
	for (i = 0; i < 256; i++) {
	    *pindex++ = i;
	}
    }

    SetDIBitsToDevice(hdc, dest_x, dest_y, width, height,
		      src_x, src_y, 0, height, smallBitBase,
		      bmInfo,
		      (image->depth == 8) ? DIB_PAL_COLORS : DIB_RGB_COLORS);

    ckfree((char *) bmInfo);
    Lx_ReleaseDrawableDC(d, hdc, &state);
    if (image->depth != 8) {
	ckfree(smallBitBase);
    }
}    

void
_XInitImageFuncPtrs(XImage *image)
{
}

int 
XReadBitmapFile(Display *display, Drawable drawable, const char *filename,
	unsigned int *width, unsigned int *height, Pixmap *bitmap,
	int *x_hot, int *y_hot)
{
    struct _stat st;
    char *buffer;
    FILE *fp;
    char *bits;
    HBITMAP hbm;
    BITMAPFILEHEADER *bmfh;
    BITMAPINFOHEADER *bmih;
    BITMAPINFO *bmi;
    HDC hdc;
    int result;

    /* First, check if it is a Windows bitmap file */
    if (_stat(filename, &st) != 0)				   
 	return BitmapOpenFailed;				   
    if ((fp = fopen(filename, "rb")) == NULL)	   
 	return BitmapOpenFailed;
    if ((buffer = (char *)ckalloc((size_t)st.st_size)) == NULL) {
 	result = BitmapNoMemory;
	goto close_end;
    }
    if (fread(buffer, (size_t)st.st_size, 1, fp) != 1) {
 	result = BitmapFileInvalid;
	goto  free_end;
    }
    bmfh = (BITMAPFILEHEADER *)buffer;
    bmih = (BITMAPINFOHEADER *)(buffer + sizeof(BITMAPFILEHEADER));
    bmi = (BITMAPINFO *)bmih;
    bits = buffer + bmfh->bfOffBits;		

    if (strncmp((char *)&bmfh->bfType, "BM", 2) != 0) {	 
 	result = BitmapFileInvalid;
	goto free_end;
    }
    hdc = GetDC(NULL);
    hbm = CreateDIBitmap(hdc, bmih, CBM_INIT, bits, bmi, DIB_RGB_COLORS); 
    ReleaseDC(NULL, hdc);
    if (hbm == NULL) {
 	result =  BitmapFileInvalid;
	goto free_end;
    }
    *bitmap = hbm;
    *width = (unsigned int)bmih->biWidth;
    *height = (unsigned int)bmih->biHeight;
    *x_hot = 0;
    *y_hot = 0;		
    result = BitmapSuccess;

free_end:
    ckfree(buffer);
    if (result == BitmapFileInvalid) {
	/* Now go to phase two: it may just be an ASCII file.  Try to read it
	 * as an X bitmap file
	 */
	bits = ReadBitmapFile(filename, width, height, x_hot, y_hot);
	if (bits != NULL) {
	    *bitmap = CreateBitmap(*width, *height, BMP_PLANES, BMP_PIXELS, bits);
	    if (*bitmap != NULL) {
		result = BitmapSuccess;
	    }
	    ckfree(bits);
	}
    }

close_end:
    if (result == BitmapSuccess) {
	AddPixmap(display, drawable, *bitmap, *width, *height, 1);
    }
    fclose(fp);
    return result;
}	


/* #ifdef TCL_MSDOS_PORT */
/* The code below is slightly modified from the X11R5 Xlib source file
 * XRdBitF.c
 */
/* $XFree86: mit/lib/X/XRdBitF.c,v 1.4 1993/03/27 09:14:21 dawes Exp $ */
/* $XConsortium: XRdBitF.c,v 1.15 91/02/01 16:34:46 gildea Exp $ */
/* Copyright, 1987, Massachusetts Institute of Technology */

/* TkWin version 0.1 -- Tk for Windows
 * Copyright (C) 1994 K. K. Kubota
 * Mathematical Sciences Computing Facility, University of Kentucky,
 * Lexington, KY 40506, USA (e-mail: ken@ms.uky.edu)

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

#define MAX_SIZE 255

/* shared data for the image read/parse logic */
static short hexTable[256];		/* conversion value */
static Bool Xinitialized = False;	/* easier to fill in at run time */


/*
 *	Table index for the hex values. Initialized once, first time.
 *	Used for translation value or delimiter significance lookup.
 */
static void initHexTable()
{
    hexTable['0'] = 0;	hexTable['1'] = 1;
    hexTable['2'] = 2;	hexTable['3'] = 3;
    hexTable['4'] = 4;	hexTable['5'] = 5;
    hexTable['6'] = 6;	hexTable['7'] = 7;
    hexTable['8'] = 8;	hexTable['9'] = 9;
    hexTable['A'] = 10;	hexTable['B'] = 11;
    hexTable['C'] = 12;	hexTable['D'] = 13;
    hexTable['E'] = 14;	hexTable['F'] = 15;
    hexTable['a'] = 10;	hexTable['b'] = 11;
    hexTable['c'] = 12;	hexTable['d'] = 13;
    hexTable['e'] = 14;	hexTable['f'] = 15;

    /* delimiters of significance are flagged w/ negative value */
    hexTable[' '] = -1;	hexTable[','] = -1;
    hexTable['}'] = -1;	hexTable['\n'] = -1;
    hexTable['\t'] = -1;
	
    Xinitialized = True;
}

/*
 *	read next hex value in the input stream, return -1 if EOF
 */
static long NextInt (fstream)
    FILE *fstream;
{
    int	ch;
    long value = 0;
    int gotone = 0;
    int done = 0;
    
    /* loop, accumulate hex value until find delimiter  */
    /* skip any initial delimiters found in read stream */

    while (!done) {
	ch = getc(fstream);
	if (ch == EOF) {
	    value	= -1L;
	    done++;
	} else {
	    /* trim high bits, check type and accumulate */
	    ch &= 0xff;
	    if (isascii(ch) && isxdigit(ch)) {
		value = (value << 4) + hexTable[ch];
		gotone++;
	    } else if ((hexTable[ch]) < 0 && gotone)
	      done++;
	}
    }
    return value;
}

/* $XConsortium: Xlib.h,v 11.215 91/07/22 15:42:38 rws Exp $ */
/* 
 * Copyright 1985, 1986, 1987, 1991 by the Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided 
 * that the above copyright notice appear in all copies and that both that 
 * copyright notice and this permission notice appear in supporting 
 * documentation, and that the name of M.I.T. not be used in advertising
 * or publicity pertaining to distribution of the software without specific, 
 * written prior permission. M.I.T. makes no representations about the 
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * X Window System is a Trademark of MIT.
 *
 */

/* read bitmap from X10 or X11 bitmap files and return
 * array of data bits in format appropriate for Microsoft Windows
 */
static char *ReadBitmapFile (filename, width, height, x_hot, y_hot)
    const char *filename;
    unsigned int *width, *height;       /* RETURNED */
    int *x_hot, *y_hot;                 /* RETURNED */
{
    FILE *fstream;			/* handle on file  */
    char *data = NULL;			/* working variable */
    char line[MAX_SIZE];		/* input line from file */
    int size;				/* number of bytes of data */
    char name_and_type[MAX_SIZE];	/* an input line */
    char *type;				/* for parsing */
    long value;				/* from an input line */
    int val;
    int version10p;			/* boolean, old format */
    int padding;			/* to handle alignment */
    int bytes_per_line;			/* per scanline of data */
    unsigned int ww = 0;		/* width */
    unsigned int hh = 0;		/* height */
    int hx = 0;				/* x hotspot */
    int hy = 0;				/* y hotspot */
    char *retval = NULL;

    /* first time initialization */
    if (Xinitialized == False) initHexTable();

    if ((fstream = fopen(filename, "rt")) == NULL) {
	return NULL;
    }

    while (fgets(line, MAX_SIZE, fstream)) {
	if (strlen(line) == MAX_SIZE-1) {
	    goto error;
	}
	if (sscanf(line,"#define %s %d",name_and_type,&val) == 2) {
	    if ((type = strrchr(name_and_type, '_')) == NULL) {
		type = name_and_type;
	    } else {
		type++;
	    }

	    if (!strcmp("width", type)) {
		ww = (unsigned int) val;
	    }
	    if (!strcmp("height", type)) {
		hh = (unsigned int) val;
	    }
	    if (!strcmp("hot", type)) {
		if (type-- == name_and_type || type-- == name_and_type)
		    continue;
		if (!strcmp("x_hot", type))
		    hx = val;
		if (!strcmp("y_hot", type))
		    hy = val;
	    }
	    continue;
	}
    
	if (sscanf(line, "static short %s = {", name_and_type) == 1)
	    version10p = 1;
	else if (sscanf(line,"static unsigned char %s = {",name_and_type) == 1)
	    version10p = 0;
	else if (sscanf(line, "static char %s = {", name_and_type) == 1)
	    version10p = 0;
	else
	    continue;

	if ((type = strrchr(name_and_type, '_')) == NULL)
	    type = name_and_type;
	else
	    type++;

	if (strcmp("bits[]", type))
	    continue;
    
	if (!ww || !hh) {
	    goto error;
	}

	if ((ww % 16) && ((ww % 16) < 9) && !version10p) {
	    padding = 1;
	} else {
	    padding = 0;
	}

	bytes_per_line = (ww+7)/8 + padding;

	size = bytes_per_line * hh;
/* xxxxxxxxxx added +hh */
	data = (char *) ckalloc((unsigned int) (size + hh));
	if (!data) {
	    goto error;
	}

	if (version10p) {
	    unsigned char *ptr;
	    int bytes;

	    for (bytes=0, ptr=data; bytes<size; (bytes += 2)) {
		value = NextInt(fstream);
		if (value < 0) {
		    goto error;
		}
		*(ptr++) = swap((unsigned char) (value & 0xff));
		*(ptr++) = swap((unsigned char) ((value >> 8) & 0xff));
	    }
	} else {
	    unsigned char *ptr;
	    int bytes;

	    for (bytes=0, ptr=data; bytes<size; bytes++, ptr++) {
		value = NextInt(fstream);
		if (value < 0) {
		    goto error;
		}
		*ptr= swap((unsigned char) (value & 0xff));
		if (padding && ((bytes+2) % bytes_per_line) == 0) {
		    ptr++;
		    *ptr = (unsigned char) 0x00;
		    bytes++;
		}
	    }
	}
    }					/* end while */


    *width = ww;
    *height = hh;
    if (x_hot) *x_hot = hx;
    if (y_hot) *y_hot = hy;

    retval = data;

 error:
    if (data && retval != data) {
	ckfree(data);
    }
    fclose (fstream);
    return retval;
}

/*
 *----------------------------------------------------------------------
 * XCreateImage --
 *
 *	Allocate the memory for an image
 *----------------------------------------------------------------------
 */
XImage *
XCreateImage(Display *display, Visual *visual,
	     unsigned int depth, int format, int offset, char *data,
	     unsigned int width, unsigned int height,
	     int bitmap_pad, int bytes_per_line)
{
    XImage *ret_image;

    ret_image = (XImage *) ckalloc(sizeof(XImage));
    if (ret_image == NULL) {
	return NULL;
    }

    ret_image->data = data;
    ret_image->width = width;
    ret_image->height = height;
    ret_image->xoffset = 0;
    ret_image->format = ZPixmap;
    ret_image->byte_order = LSBFirst;
    ret_image->bitmap_bit_order = MSBFirst;
    ret_image->f.create_image = NULL;
    ret_image->f.destroy_image = Lx_XDestroyImage;
    ret_image->f.get_pixel = Lx_XGetPixel;
    ret_image->f.put_pixel = Lx_XPutPixel;
    ret_image->f.sub_image = NULL;
    ret_image->f.add_pixel = NULL;

    if (depth == 1) {
	ret_image->bitmap_unit = 1; /* ??? */
	ret_image->bitmap_pad = 16;
	ret_image->depth = 1;
	ret_image->bits_per_pixel = 1;
    } else if (depth == 8) {
	ret_image->bitmap_unit = 8;
	ret_image->bitmap_pad = 8;
	ret_image->depth = 8;
	ret_image->bits_per_pixel = 8;
    } else {
	ret_image->bitmap_unit = 32;
	ret_image->bitmap_pad = 32;
	ret_image->depth = 32;
	ret_image->bits_per_pixel = 32;
    }
    ret_image->bytes_per_line =
	((ret_image->bits_per_pixel * width + 31) >> 3) & ~3;

    return ret_image;
}
