#include "xlibInt.h" 
#include <x11/xutil.h>
#include <x11/xutilw.h>

/*
 *----------------------------------------------------------------------
 * XCreateRegion --
 *
 *	Creates an undefined region.  We make this an empty region under
 *	Windows.
 *
 * Results:
 *	The undefined region is returned.
 *----------------------------------------------------------------------
 */
Region
XCreateRegion(void)
{
    Region r;

    r = (Region) ckalloc(sizeof(struct _XRegion));
    if (r == NULL) {
	return NULL;
    }
    r->hrgn = CreateRectRgn(0, 0, 0, 0);
    return r;
}

/*
 *----------------------------------------------------------------------
 * XDestroyRegion --
 *
 *	Not much to say.
 *----------------------------------------------------------------------
 */
int
XDestroyRegion(Region r)
{
    DeleteObject(r->hrgn);
    ckfree((char *) r);
    return Success;
}

/*
 *----------------------------------------------------------------------
 * XUnionRectWithRegion
 *
 *	Add a rectangle to a region.
 *----------------------------------------------------------------------
 */
int
XUnionRectWithRegion(XRectangle *rectangle, Region src, Region dest)
{
    DWORD count;
    LPRGNDATA rgnData;
    HRGN hRectRgn;

    if (src != dest) {
	DeleteObject(dest->hrgn);
	count = GetRegionData(src->hrgn, 0, NULL);
	rgnData = ckalloc(count);
	if (rgnData == NULL) {
	    return BadAlloc;
	}
	GetRegionData(src->hrgn, count, rgnData);
	dest->hrgn = ExtCreateRegion(NULL, count, rgnData);
	ckfree((char *) rgnData);
    }

    hRectRgn = CreateRectRgn(rectangle->x, rectangle->y,
			     rectangle->x + rectangle->width,
			     rectangle->y + rectangle->height);
    CombineRgn(dest->hrgn, dest->hrgn, hRectRgn, RGN_OR);
    DeleteObject(hRectRgn);

    return Success;
}

/*
 *----------------------------------------------------------------------
 * XIntersectRegion --
 *
 *	Compute the intersection of two regions
 *----------------------------------------------------------------------
 */
int
XIntersectRegion(Region sra, Region srb, Region	dr_return)
{
    CombineRgn(dr_return->hrgn, sra->hrgn, srb->hrgn, RGN_AND);
    return Success;
}


/*
 *----------------------------------------------------------------------
 * XClipBox --
 *
 *	Generate the smallest rectangle enclosing a region.
 *---------------------------------------------------------------------- 
 */
int
XClipBox(Region r, XRectangle *rect_return)
{
    RECT rect;

    GetRgnBox(r->hrgn, &rect);
    rect_return->x = (short) rect.left;
    rect_return->y = (short) rect.top;
    rect_return->width = (unsigned short) (rect.right - rect.left);
    rect_return->height = (unsigned short) (rect.bottom - rect.top);

    return Success;
}

/*
 *----------------------------------------------------------------------
 * Lx_CopyRegion --
 *
 *	Create a copy of a region.
 *
 * Results:
 *	A new region
 *----------------------------------------------------------------------
 */
Region
Lx_CopyRegion(Region src)
{
    Region dst;
    dst = XCreateRegion();
    CombineRgn(dst->hrgn, src->hrgn, NULL, RGN_COPY);
    return dst;
}

/*
 *----------------------------------------------------------------------
 * XSetRegion --
 *
 *	Set clip_mask fo the graphics context to the specified region
 *
 * Results:
 *	None
 *
 * Notes:
 *	Need to convert the region to a bitmap, then actually use this
 *	bitmap as a clipping mask.  PlgBlt() under NT and Chicago can
 *	be used for this clipping.
 *----------------------------------------------------------------------
 */
void
XSetRegion(Display *display, GC gc, Region region)
{
    DWORD count;
    LPRGNDATA rgnData;
    struct _XRegion r;

    if (gc->rects) {
	XDestroyRegion((Region) gc->values.clip_mask);
	gc->rects = False;
    }

    count = GetRegionData(region->hrgn, 0, NULL);
    rgnData = ckalloc(count);
    if (rgnData == NULL) {
	return;
    }
    GetRegionData(region->hrgn, count, rgnData);
    r.hrgn = ExtCreateRegion(NULL, count, rgnData);
    ckfree((char *) rgnData);

    gc->values.clip_mask = (Pixmap) Lx_CopyRegion(&r);
    gc->rects = True;
    DeleteObject(r.hrgn);
}

/*
 *----------------------------------------------------------------------
 * XSetClipRectangles --
 *
 *	Change clip_mask in a graphics context to a list of rectangles
 *
 * Results:
 *	None:
 *----------------------------------------------------------------------
 */
void
XSetClipRectangles(Display *display, GC gc,
		   int clip_x_origin, int clip_y_origin,
		   XRectangle *rectangles, int  n, int ordering)
{
    Region r;
    int i;
    XRectangle rect;

    r = XCreateRegion();
    for (i = 0; i < n; i++) {
	rect.x = rectangles[i].x + clip_x_origin;
	rect.y = rectangles[i].y + clip_y_origin;
	rect.width = rectangles[i].width;
	rect.height = rectangles[i].height;
	XUnionRectWithRegion(&rect, r, r);
    }
    XSetRegion(display, gc, r);
    XDestroyRegion(r);
}
