/*
 * xwindow.c --
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */
 
#include "xlibInt.h" 


extern HCURSOR hDefaultCursor;

/*
 * A hash table of window ids
 */
static int initialized = 0;
static Lx_HashTable windowTable;

/*
 * Forward declarations
 */
static void ReStack(HWND, HWND, int);

/*
 *----------------------------------------------------------------------
 * Lx_NewWindowHashEntry --
 *
 *	Creates a new hash hash entry for a window
 *
 * Results:
 *	The new window entry
 *----------------------------------------------------------------------
 */
Lx_WindowEntry *
Lx_NewWindowHashEntry(HWND hwnd)
{
    Lx_HashEntry *hPtr;
    Lx_WindowEntry *ePtr;
    int dummy;

    if (! initialized) {
	Lx_InitHashTable(&windowTable, LX_ONE_WORD_KEYS);
	initialized = 1;
    }
    ePtr = (Lx_WindowEntry *) ckalloc(sizeof(Lx_WindowEntry));
    memcpy(&ePtr->setWinAttrs, &lx_DefWindowAttrs, sizeof(XSetWindowAttributes));
    memcpy(&ePtr->winChanges, &lx_DefWindowChanges, sizeof(XWindowChanges));
    ePtr->cursor = 0;
    ePtr->prevCursor = 0;

    hPtr = Lx_CreateHashEntry(&windowTable, hwnd, &dummy);
    Lx_SetHashValue(hPtr, ePtr);
    return ePtr;
}

/*
 *----------------------------------------------------------------------
 * Lx_GetWindowHashEntry --
 *
 *	Looks up a window and returns its hash entry
 *----------------------------------------------------------------------
 */
Lx_WindowEntry *
Lx_GetWindowHashEntry(HWND hwnd)
{
    Lx_HashEntry *hPtr;
    Lx_WindowEntry *ePtr;

    if (! initialized) {
	return NULL;
    }

    hPtr = Lx_FindHashEntry(&windowTable, hwnd);
    if (hPtr == NULL) {
        return NULL;
    }

    ePtr = (Lx_WindowEntry *) Lx_GetHashValue(hPtr);
    return ePtr;
}

/*
 *----------------------------------------------------------------------
 * Lx_FreeWindowHashEntry --
 *
 *	Free a hash entry for a window
 *----------------------------------------------------------------------
 */
void
Lx_FreeWindowHashEntry(HWND hwnd)
{
    Lx_HashEntry *hPtr;
    Lx_WindowEntry *ePtr;

    hPtr = Lx_FindHashEntry(&windowTable, hwnd);
    if (hPtr == NULL) {
	return;
    }
    ePtr = (Lx_WindowEntry *) Lx_GetHashValue(hPtr);
    ckfree((char *) ePtr);
    Lx_DeleteHashEntry(hPtr);
}

/*
 *----------------------------------------------------------------------
 *
/*
 *----------------------------------------------------------------------
 * XCreateWindow --
 *
 *	Create a new X window.
 *----------------------------------------------------------------------
 */
Window
XCreateWindow(Display *display, Window parent, int x, int y, UINT width, 
	UINT height, UINT border_width, int depth, UINT class, 
	Visual *visual, unsigned long valuemask, 
	XSetWindowAttributes *attributes)
{
    HWND child;
    DWORD dwStyle, dwExStyle;
    XWindowChanges changes;
    BYTE byResult;
    BOOLEAN toplevel;
    Lx_WindowEntry *winEntry;

    display->request++;

    toplevel = (parent == NULL || attributes->override_redirect == True) ? 1:0;
    if (toplevel) {
	RECT rc;
	if (attributes->override_redirect == True) {
	    dwStyle = WS_POPUP|WS_BORDER|WS_CLIPCHILDREN|CS_DBLCLKS;
	    dwExStyle = WS_EX_TOPMOST;
	} else {
	    dwStyle = WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN|CS_DBLCLKS;
	    dwExStyle = 0;
	}
	child = CreateWindow("TopLevel", "", dwStyle, 
		    attributes->override_redirect?x:CW_USEDEFAULT, 
		    attributes->override_redirect?y:CW_USEDEFAULT, 
		    Lx_AdjustWidth(width, attributes->override_redirect),
		    Lx_AdjustHeight(height, attributes->override_redirect), 
		    parent, NULL, Tknt_GetAppInstance(), NULL);
	GetWindowRect(child, &rc);
	y = Lx_ReAdjustY(rc.top, attributes->override_redirect);
	x = Lx_ReAdjustX(rc.left, attributes->override_redirect);
    } else if ((valuemask & CWNativeWidget) && (attributes->native_widget)) {
	switch (attributes->native_widget) {
	  case NativeWidgetButton:
	    child = CreateWindow("BUTTON", "",
			WS_CHILD|WS_CLIPCHILDREN|BS_PUSHBUTTON,
			x, y, width, height, parent, NULL,
			Tknt_GetAppInstance(), NULL);
	    break;
	  case NativeWidgetXScrollbar:
	    child = CreateWindow("SCROLLBAR", "",
			WS_CHILD|WS_CLIPCHILDREN|SBS_HORZ,
			x, y, width, height, parent, NULL,
			Tknt_GetAppInstance(), NULL);
	    break;
	  case NativeWidgetYScrollbar:
	    child = CreateWindow("SCROLLBAR", "",
			WS_CHILD|WS_CLIPCHILDREN|SBS_VERT,
			x, y, width, height, parent, NULL,
			Tknt_GetAppInstance(), NULL);
	    break;
	  default:
	    MessageBox(0, "Unknown native widget", "Error Handler", 0);
	    child = NULL;
	}
	if (child) {
	    XEvent e;

	    winEntry = Lx_NewWindowHashEntry(child);
	    switch (attributes->native_widget) {
	      case NativeWidgetButton:
		break;
	      case NativeWidgetXScrollbar:
	      case NativeWidgetYScrollbar:
		winEntry->defaultWinProc = (LONG)
		    (SetWindowLong(child, GWL_WNDPROC, (LONG) ScrollbarProc));
		break;
	    }

	    e.type = CreateNotify;
	    e.xcreatewindow.parent = (Window) GetParent(child);
	    e.xcreatewindow.window = child;
	    e.xcreatewindow.x = x;
	    e.xcreatewindow.y = y;
	    e.xcreatewindow.width = width;
	    e.xcreatewindow.height = height;
	    e.xcreatewindow.border_width = 1;
	    e.xcreatewindow.override_redirect = False;
	    Lx_PutEvent(display, &e);
	    PostMessage(child, EVENTADDED, 0, 0L);
	}
    } else {
	dwStyle = WS_CHILD|WS_CLIPCHILDREN|WS_CLIPSIBLINGS;
	child = CreateWindow("Widget", "", dwStyle, x, y, width, height, 
			     parent, NULL, Tknt_GetAppInstance(), NULL);
    }

    if (child == NULL) {
	Lx_PostError(BadWindow, parent);
	return NULL;
    }

    SetWindowLong(child, TKNT_OFFSET, TKNT_ID);

    byResult = Lx_SetWindowAttributes(child, valuemask, attributes);
    if (byResult != Success) {
	goto destroy;
    }

    changes.x = x;
    changes.y = y;
    changes.width = width;
    changes.height = height;
    changes.border_width = border_width;
    byResult = Lx_SetWindowChanges(child, CWX|CWY|CWWidth|CWHeight|CWBorderWidth, &changes);
    if (byResult != Success) {
	destroy:
	DestroyWindow(child);
	Lx_PostError(byResult, parent);
	return NULL;
    }

    return child;
}

/*
 *----------------------------------------------------------------------
 * XDestroyWindow --
 *
 *	Destroy an X window
 *----------------------------------------------------------------------
 */
void  
XDestroyWindow(Display *display, Window hwnd)
{
    display->request++;

    if (DestroyWindow(hwnd) == 0) {
	Lx_PostError(BadWindow, hwnd);
    }
    Lx_EnterLeave(hwnd, 0, 0, 1 /* windowMoved */);
    Lx_CheckAndResetFocus(display, hwnd);
}


/*
 *----------------------------------------------------------------------
 * XMapWindow --
 *
 *	Tries to display a window
 *----------------------------------------------------------------------
 */
void	
XMapWindow(Display *display, Window w)
{
    Lx_WindowEntry *winEntry;
    XWindowChanges *pXWChanges;
    XSetWindowAttributes *pXSWAttrs;
    XWMHints *hPtr;
    char propName[8];
    char *ptr;
    XWindowChanges changes;
    HGLOBAL hGlobal;

    display->request++;

    sprintf(propName, "#0x%0.4x", XA_WM_HINTS);
    
    if ((hGlobal = GetProp(w, propName)) != NULL) {
	int state;
	long flags;

	ptr = ((char *) GlobalLock(hGlobal)) + sizeof(Atom);
	hPtr = (XWMHints *) ptr;
	flags = hPtr->flags;
	state = hPtr->initial_state;
	GlobalUnlock(hGlobal);
	if (flags & StateHint) {
	    switch (state) {
	    case NormalState:
		if (IsIconic(w)) {
		    winEntry = Lx_GetWindowHashEntry(w);
		    pXSWAttrs = &winEntry->setWinAttrs;
		    pXWChanges = &winEntry->winChanges;

		    /* If the window was moved while it was iconified,
		     * widthdraw it, reset its position, and then display
		     * it
		     */
		    if (pXWChanges->pos_changed) {
			changes.x = pXWChanges->x;
			changes.y = pXWChanges->y;
			changes.width = pXWChanges->width;
			changes.height = pXWChanges->height;
		    }
		    ShowWindow(w, SW_SHOWNOACTIVATE);

		    if (pXWChanges->pos_changed) {
			pXWChanges->pos_changed = FALSE;
			Lx_SetWindowChanges(w, CWX|CWY|CWWidth|CWHeight,
					    &changes);
		    }
		    PostMessage(w, MAPNOTIFY, 0, 0L);
		    return;
		}
		goto normalmap;
	    case IconicState:
		ShowWindow(w, SW_SHOWMINNOACTIVE); break;
	    case WithdrawnState:
		ShowWindow(w, SW_HIDE); break;
	    }
	} 
    } else {	    	
	static int firstWindow = 1;
      normalmap:
	if (firstWindow == 1) {
	    firstWindow = 0;
	    ShowWindow(w, SW_SHOWNORMAL);
	} else {
	    ShowWindow(w, SW_SHOWNA);
	}
	PostMessage(w, MAPNOTIFY, 0, 0L);
    }
}


/*
 *	XUnmapWindow
 *
 */

void  
XUnmapWindow(Display *display, Window hwnd)
{
    display->request++;

    ShowWindow(hwnd, SW_HIDE);
    PostMessage(hwnd, UNMAPNOTIFY, 0, 0L);
}

/*
 *	XIconifyWindow
 *
 */

Status
XIconifyWindow(Display *display, Window hwnd, int dummy2)
{
    display->request++;

    ShowWindow(hwnd, SW_SHOWMINNOACTIVE);
    PostMessage(hwnd, UNMAPNOTIFY, 0, 0L);
    return 1;
}

/*
 *	XWithdrawWindow
 *
 */

Status
XWithdrawWindow(Display *display, Window hwnd, int dummy2)
{
    display->request++;

    ShowWindow(hwnd, SW_HIDE);
    PostMessage(hwnd, UNMAPNOTIFY, 0, 0L);
    return 1;
}

Window
XWindowFromPoint(Display *display, int x, int y)
{
    POINT p;
    Window w;

    display->request++;

    p.x = x;
    p.y = y;

    w = (Window) WindowFromPoint(p);
    if (w) {
	ScreenToClient(w, &p);

	if (p.x >= 0 && p.y >= 0) {
	    return w;
	}
    }
    return HWND_DESKTOP;
}

void			 
XClearArea(Display *display, Window w, int x, int y, 
	unsigned width, unsigned height, Bool dummy2)
{
    Lx_WindowEntry *winEntry;
    XSetWindowAttributes *pXSWAttrs;
    HDC hdc;
    HBRUSH hbr;
    HRGN hrgn;
    Lx_DrawableDCState state;

    display->request++;

    hdc = Lx_GetDrawableDC(w, &state);
    winEntry = Lx_GetWindowHashEntry(w);
    pXSWAttrs = &winEntry->setWinAttrs;
    hbr = CreateSolidBrush(LX_CONVERTPIXEL(pXSWAttrs->background_pixel));
    hrgn = CreateRectRgn(x, y, x + width, y + height);

    FillRgn(hdc, hrgn, hbr);
    
    DeleteObject(hbr);
    DeleteObject(hrgn);
    Lx_ReleaseDrawableDC(w, hdc, &state);
}

void			
XClearWindow(Display *display, Window w)
{
    RECT rc;
    Lx_WindowEntry *winEntry;
    XSetWindowAttributes *pXSWAttrs;
    HDC hdc;
    HBRUSH hbr;
    Lx_DrawableDCState state;

    display->request++;

    hdc = Lx_GetDrawableDC(w, &state);
    winEntry = Lx_GetWindowHashEntry(w);
    pXSWAttrs = &winEntry->setWinAttrs;
    hbr = CreateSolidBrush(LX_CONVERTPIXEL(pXSWAttrs->background_pixel));

    GetWindowRect(w, &rc);
    rc.right = rc.right - rc.left;
    rc.bottom = rc.bottom - rc.top;
    rc.left = 0;
    rc.top = 0;
    FillRect(hdc, &rc, hbr);
    
    DeleteObject(hbr);
    Lx_ReleaseDrawableDC(w, hdc, &state);
}

void
XChangeWindowAttributes(Display *display, Window w, unsigned long valuemask, 
	XSetWindowAttributes *attributes)
{
    unsigned char result;

    display->request++;

    if ((result = Lx_SetWindowAttributes(w, valuemask, attributes)) != Success)
    	Lx_PostError(result, (HWND) w);
}

void
XSetWindowBackground(Display *display, Window w, unsigned long background_pixel)
{
    XSetWindowAttributes attrs;
    unsigned char result;
	
    display->request++;

    attrs.background_pixel = background_pixel;
    if ((result = Lx_SetWindowAttributes(w, CWBackPixel, &attrs)) != Success)
    	Lx_PostError(result, (HWND) w);
}

void
XSetWindowBackgroundPixmap(Display *display, Window w, Pixmap background_pixmap)
{
    XSetWindowAttributes attrs;
    unsigned char result;
	
    display->request++;

    attrs.background_pixmap = background_pixmap;
    if ((result = Lx_SetWindowAttributes(w, CWBackPixmap, &attrs)) != Success)
    	Lx_PostError(result, (HWND) w);
}

void
XSetWindowBorder(Display *display, Window w, unsigned long border_pixel)
{
    XSetWindowAttributes attrs;
    unsigned char result;
	
    display->request++;

    attrs.border_pixel = border_pixel;
    if ((result = Lx_SetWindowAttributes(w, CWBorderPixel, &attrs)) != Success)
    	Lx_PostError(result, (HWND) w);
}

void
XSetWindowBorderPixmap(Display *display, Window w, Pixmap border_pixmap)
{
    XSetWindowAttributes attrs;
    unsigned char result;
	
    display->request++;

    attrs.border_pixmap = border_pixmap;
    if ((result = Lx_SetWindowAttributes(w, CWBorderPixmap, &attrs)) != Success)
    	Lx_PostError(result, (HWND) w);
}

void
XSetWindowColormap(Display *display, Window w, Colormap colormap)
{
    XSetWindowAttributes attrs;
    unsigned char result;
	
    display->request++;

    attrs.colormap = colormap;
    if ((result = Lx_SetWindowAttributes(w, CWColormap, &attrs)) != Success)
    	Lx_PostError(result, (HWND) w);
}

void
XSelectInput(Display *display, Window w, long event_mask)
{
    XSetWindowAttributes attrs;

    display->request++;

    attrs.event_mask = event_mask;
    Lx_SetWindowAttributes(w, CWEventMask, &attrs);
}

void
XSetClassHint(Display *display, Window w, XClassHint *class_hints)
{
    /* Don't do anything, at least for now */
    display->request++;
}

Status
XGetGeometry(Display *display, Drawable drawable, Window *root_return,
	     int *x_return, int *y_return,
	     unsigned int *width_return, unsigned int *height_return,
	     unsigned int *border_width_return, unsigned int *depth_return)
{
    XWindowChanges *attrs;
    Lx_PixElem *pixElem;
    Lx_WindowEntry *winEntry;

    display->request++;

    pixElem = Lx_FindPixmap(drawable);
    if (pixElem) {
	*width_return = pixElem->width;
	*height_return = pixElem->height;
	*depth_return = pixElem->depth;
	*border_width_return = 0;

    } else {

	winEntry = Lx_GetWindowHashEntry(drawable);
	if (winEntry == NULL) {
	    return BadWindow;
	}
	attrs = &winEntry->winChanges;
	*width_return = attrs->width;
	*height_return = attrs->width;
	*depth_return = display->screens->root_depth;
    }
    return Success;
}

/*
 * XConfigureWindow --
 *
 */

void
XConfigureWindow(Display *display, Window w, unsigned mask, 
	XWindowChanges *values)
{  
    unsigned char result;
    
    display->request++;

    result = Lx_SetWindowChanges(w, mask, values); 
    if (result != Success) 
    	Lx_PostError(result, (HWND) w);
}

/*
 * XMoveWindow --
 *
 */

void
XMoveWindow(Display *display, Window w, int x, int y)
{
    XWindowChanges changes;
    unsigned char result;
     
    display->request++;

    changes.x = x;
    changes.y = y;
    result = Lx_SetWindowChanges(w, CWX|CWY,  &changes);
    if (result != Success) 
    	Lx_PostError(result, (HWND) w);
}

void
XResizeWindow(Display *display, Window w, unsigned width, unsigned height)
{
    XWindowChanges changes;
    unsigned char result;

    display->request++;

    changes.width = width;
    changes.height = height;
    result = Lx_SetWindowChanges(w, CWWidth|CWHeight, &changes);
    if (result != Success) 
    	Lx_PostError(result, (HWND) w);
}

void
XMoveResizeWindow(Display *display, Window w, int x, int y, 
	unsigned width, unsigned height)
{
    XWindowChanges changes;
    unsigned char result;

    display->request++;

    changes.x = x;
    changes.y = y;
    changes.width = width;
    changes.height = height;
    result = Lx_SetWindowChanges(w, CWX|CWY|CWWidth|CWHeight, &changes);
    if (result != Success) 
    	Lx_PostError(result, (HWND) w);
}

void 
XSetWindowBorderWidth(Display *display, Window w, unsigned int width)
{
    XWindowChanges changes;
    unsigned char result;

    display->request++;

    changes.border_width = width;
    result = Lx_SetWindowChanges(w, CWBorderWidth, &changes);
    if (result != Success) 
    	Lx_PostError(result, (HWND) w);
}

void
XRaiseWindow(Display *display, Window w)
{
    XWindowChanges changes;
    unsigned char result;

    display->request++;

    changes.sibling = None;
    changes.stack_mode = Above;
    result = Lx_SetWindowChanges(w, CWSibling|CWStackMode, &changes);
    if (result != Success) 
    	Lx_PostError(result, (HWND) w);
}

void
XLowerWindow(Display *display, Window w)
{
    XWindowChanges changes;
    unsigned char result;

    display->request++;

    changes.sibling = None;
    changes.stack_mode = Below;
    result = Lx_SetWindowChanges(w, CWSibling|CWStackMode, &changes);
    if (result != Success) 
    	Lx_PostError(result, (HWND) w);
}

/*
 *----------------------------------------------------------------------
 *
 * Lx_SetWindowChanges --
 *
 *	Modify the XWindowChanges structure that is associated with
 *	the window.  Only the fields that are specified by bits in
 *	'mask' are affected.  The window may be moved, resized, 
 *	repositioned, or restacked.
 *
 * Result:
 *	Success if OK, BadWindow if the window is invalid
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */

unsigned char
Lx_SetWindowChanges(Window w, unsigned long mask, XWindowChanges * changes)
{
    XWindowChanges *pXWChanges;
    XSetWindowAttributes *pXSWAttrs;
    Lx_WindowEntry *winEntry;
    UINT flag;

    winEntry = Lx_GetWindowHashEntry(w);
    if (winEntry == NULL) {
	return BadWindow;
    }
    pXSWAttrs = &winEntry->setWinAttrs;
    pXWChanges = &winEntry->winChanges;

    flag = SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOZORDER|SWP_NOSIZE;

    if (mask & CWX) {
	pXWChanges->x = changes->x;
	flag &= ~SWP_NOMOVE;
    }
    if (mask & CWY) {
	pXWChanges->y = changes->y;
	flag &= ~SWP_NOMOVE;
    }
    if (mask & CWWidth)	{
    	pXWChanges->width = changes->width;
	flag &= ~SWP_NOSIZE;
    }
    if (mask & CWHeight) {
    	pXWChanges->height = changes->height;
	flag &= ~SWP_NOSIZE;
    }
    if (mask & CWBorderWidth) {
	pXWChanges->border_width = changes->border_width;
    }
    if (mask & CWSibling) {
      	pXWChanges->sibling = changes->sibling;
    }
    if (mask & CWStackMode) {
    	pXWChanges->stack_mode = changes->stack_mode;
    }
 	
    if (GetParent(w) == HWND_DESKTOP || pXSWAttrs->override_redirect) {

	/*
	 * If the window is iconic, we need to update the window attributes
	 * here.
	 */
	 
	if (IsIconic(w)) {
	    if (Lx_CheckSendMessage(w, ConfigureNotify)) {
		Lx_AddConfigureNotify(w, NULL,
				      pXWChanges->x, pXWChanges->y,
				      pXWChanges->width, pXWChanges->height);
	    }
	} else {
	    MoveWindow(w, 
		Lx_AdjustX(pXWChanges->x, pXSWAttrs->override_redirect),
		Lx_AdjustY(pXWChanges->y, pXSWAttrs->override_redirect),
   		Lx_AdjustWidth(pXWChanges->width, pXSWAttrs->override_redirect), 
   	    	Lx_AdjustHeight(pXWChanges->height, pXSWAttrs->override_redirect),
		TRUE);
	    Lx_EnterLeave(w, 0, 0, 1 /* windowMoved */);
	}
    } else {
   	MoveWindow(w, pXWChanges->x, pXWChanges->y, 
   		pXWChanges->width, pXWChanges->height, TRUE);
	Lx_EnterLeave(w, 0, 0, 1 /* windowMoved */);
    }

    if (mask & CWStackMode) {
    	ReStack(w,
		(HWND) (mask&CWSibling?pXWChanges->sibling:None),
		pXWChanges->stack_mode);
    }
    return Success;
}

/*
 *----------------------------------------------------------------------
 *
 * ReStack --
 *
 *	Reposition a window in the window stack
 *
 * Result:
 *	None
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */
static void
ReStack(HWND hwnd, HWND hwndSibling, int mode)
{
    HWND hwndInsertAfter;

    if (hwndSibling == None) {
	hwndInsertAfter = (mode == Above) ? HWND_TOP : HWND_BOTTOM;
    } else {
        if (mode == Above) {
	    SetWindowPos(hwnd, hwndSibling, 0, 0, 0, 0, 
			 SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOSIZE);
	    hwndInsertAfter = hwnd;
	    hwnd = hwndSibling;
        } else {
	    hwndInsertAfter = hwndSibling;
	}
    }
    SetWindowPos(hwnd, hwndInsertAfter, 0, 0, 0, 0, 
		 SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOSIZE);
}


/*
 *----------------------------------------------------------------------
 *
 * Lx_SetWindowAttributes --
 *
 *	Modify the XSetWindowAttributes structure that is associated with
 *	the window.  Only the fields that are specified by bits in
 *	'mask' are affected.  No changes that directly affect the window's
 *	appearance on the screen are made here.
 *
 * Result:
 *	Success if OK, BadWindow if the window is invalid
 *
 * Side Effects:
 *	None
 *
 *----------------------------------------------------------------------
 */

unsigned char 
Lx_SetWindowAttributes(Window w, unsigned long mask,
		       XSetWindowAttributes *attrs)
{
    XSetWindowAttributes *pXSWAttrs;
    Lx_WindowEntry *winEntry;
    HWND flag;

    winEntry = Lx_GetWindowHashEntry(w);
    if (winEntry == NULL) {
	return BadWindow;
    }
    pXSWAttrs = &winEntry->setWinAttrs;

    if (mask & CWBackPixmap) {
	pXSWAttrs->background_pixmap = attrs->background_pixmap;
    }
    if (mask & CWBackPixel) {
	pXSWAttrs->background_pixel = attrs->background_pixel;
    }
    if (mask & CWBorderPixmap) {
	pXSWAttrs->border_pixmap = attrs->border_pixmap;
    }
    if (mask & CWBorderPixel) {
	pXSWAttrs->border_pixel = attrs->border_pixel;
    }
    if (mask & CWBitGravity) {
	pXSWAttrs->bit_gravity = attrs->bit_gravity;
    }
    if (mask & CWWinGravity) {
	pXSWAttrs->win_gravity = attrs->win_gravity;
    }
    if (mask & CWBackingStore) {
	pXSWAttrs->backing_store = attrs->backing_store;
    }
    if (mask & CWBackingPlanes) {
	pXSWAttrs->backing_planes = attrs->backing_planes;
    }
    if (mask & CWBackingPixel) {
	pXSWAttrs->backing_pixel = attrs->backing_pixel;
    }
    if (mask & CWOverrideRedirect) {
	/*
	 * Don't use the next line because be can't really change the
	 * type of window _after_ a window is created.  It only matters
	 * before the window exists.
	 */
	pXSWAttrs->override_redirect = attrs->override_redirect;
	flag = attrs->override_redirect ? HWND_TOPMOST : HWND_NOTOPMOST;
	SetWindowPos(w, flag, 0, 0, 0, 0,
		     SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOSIZE);
    }
    if (mask & CWSaveUnder) {
	pXSWAttrs->save_under = attrs->save_under;
    }
    if (mask & CWEventMask) {
	pXSWAttrs->event_mask = attrs->event_mask;
    }
    if (mask & CWDontPropagate) {
	pXSWAttrs->do_not_propagate_mask = attrs->do_not_propagate_mask;
    }
    if (mask & CWColormap) {
	pXSWAttrs->colormap = attrs->colormap;
    }
    if (mask & CWCursor) {
	winEntry->prevCursor = winEntry->cursor;
	winEntry->cursor = attrs->cursor;
	pXSWAttrs->cursor = attrs->cursor;
    }
    if (mask & CWClearOnExpose) {
	pXSWAttrs->clear_on_expose = attrs->clear_on_expose;
    }
    if (mask & CWNativeWidget) {
	pXSWAttrs->native_widget = attrs->native_widget;
    }

    return Success;
}

/*
 *----------------------------------------------------------------------
 * Lx_GetDrawablePalette --
 *
 *	Get the colormap for a drawable
 *
 * Results:
 *	The colormap id
 *
 * Side Effects:
 *	None
 *----------------------------------------------------------------------
 */
HPALETTE
Lx_GetDrawablePalette(HANDLE drawable)
{
    Lx_WindowEntry *winEntry;
    XSetWindowAttributes *pXSWAttrs;
    Colormap cmap;
    XCMapInfo *cmapInfo;

    if (!Lx_IsPixmap(drawable)) {
	winEntry = Lx_GetWindowHashEntry(drawable);
	if (winEntry == NULL) {
	    return NULL;
	}
	pXSWAttrs = &winEntry->setWinAttrs;
	cmap = pXSWAttrs->colormap;
    } else {
	/* Use the default colormap */
	cmap = 1;
    }
    cmapInfo = ColorMapInfo[cmap];
    if (! cmapInfo) {
	return NULL;
    } else {
	return cmapInfo->hPalette;
    }
}
