#include <stdio.h>

#ifdef NO_STDLIB_H
#   include "compat/stdlib.h"
#else
#   include <stdlib.h>
#endif

#include <tcl.h>
#include "heap.h"


/*
 *--------------------------------------------------------------
 *
 * UpHeap --
 *
 *	Move an out of order item up in the heap data structure,
 *	restoring the heap order.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Heap order is restored.
 *
 *--------------------------------------------------------------
 */
static void
UpHeap (heapPtr, k)
Heap *heapPtr;		/* The heap */
int k;			/* Index of item that is out of order. */
{
    void *v;
    void **item;

    item = heapPtr->item;
    v = item[k];
    while ((k > 1) && ((*heapPtr->cmpProc)(item[k/2], v) > 0)) {
	item[k] = item[k/2];
	k /= 2;
    }
    item[k] = v;
}

/*
 *--------------------------------------------------------------
 *
 * DownHeap --
 *
 *	Move an out of order item down in the heap data structure,
 *	restoring the heap order.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Heap order is restored.
 *
 *--------------------------------------------------------------
 */
static void
DownHeap (heapPtr, k)
Heap *heapPtr;		/* The heap */
int k;			/* Index of item that is out of order. */
{
    int larger;
    void *v;
    void **item;
    int numItems;

    item = heapPtr->item;
    numItems = heapPtr->numItems;
    v = item[k];
    while (k <= numItems/2) {
	larger = 2*k;
	if ((larger < numItems) &&
	    ((*heapPtr->cmpProc)(item[larger], item[larger+1]) > 0)) {
	    larger++;
	}
	if ((*heapPtr->cmpProc)(v, item[larger]) < 0) {
	    break;
	}
	item[k] = item[larger];
	k = larger;
    }
    item[k] = v;
}

/*
 *--------------------------------------------------------------
 *
 * Heap_Alloc --
 *
 *	Create and return a heap data structure.
 *
 * Results:
 *	A new, empty heap.  NULL if out of memory.
 *
 * Side effects:
 *	Memory is allocated.
 *
 *--------------------------------------------------------------
 */
Heap *
Heap_Alloc(cmpProc, resizeInc)
int (*cmpProc)();
int resizeInc;
{
    Heap *heapPtr;
    heapPtr = (Heap *)ckalloc(sizeof(Heap));
    heapPtr->numItems = heapPtr->maxItems = 0;
    heapPtr->item = NULL;
    heapPtr->cmpProc = cmpProc;
    heapPtr->resizeInc = resizeInc;
    return heapPtr;
}

/*
 *--------------------------------------------------------------
 *
 * Heap_Free --
 *
 *	Deallocate a heap structure
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory is freed.
 *
 *--------------------------------------------------------------
 */
void
Heap_Free(heapPtr)
Heap *heapPtr;
{
    if (heapPtr == NULL) {
	return;
    }
    if (heapPtr->item != NULL) {
	ckfree (heapPtr->item);
    }
    ckfree (heapPtr);
}

/*
 *--------------------------------------------------------------
 *
 * Heap_Remove --
 *
 *	Return the largest item from the heap.
 *
 * Results:
 *	The integer that is the largest item, or -1 if heap is empty.
 *
 * Side effects:
 *	The heap size is decremented, and the smallest item is removed
 *	from the heap.
 *
 *--------------------------------------------------------------
 */
void *
Heap_Remove (heapPtr)
Heap *heapPtr;
{
    void *rv;

    if ((heapPtr != NULL) && (heapPtr->numItems > 0)) {
	rv = heapPtr->item[1];
	heapPtr->item[1] = heapPtr->item[heapPtr->numItems--];
	DownHeap(heapPtr, 1);
	return rv;
    } else {
	return NULL;
    }
}

/*
 *--------------------------------------------------------------
 *
 * Heap_Insert --
 *
 *	Insert an item into a heap.
 *
 * Results:
 *	The new smallest value on the heap, or NULL if there is a problem.
 *
 * Side effects:
 *	The heap size is incremented.  this may cause memory to be alloc'd
 *
 *--------------------------------------------------------------
 */
void *
Heap_Insert (heapPtr, v)
Heap *heapPtr;		/* The heap */
void *v;			/* New value to insert */
{
    if (heapPtr == NULL) {
	return NULL;
    }

    /*
     * Increase size of heap, if necessary
     */
    if (heapPtr->item == NULL) {
	heapPtr->numItems = 0;
	heapPtr->maxItems = heapPtr->resizeInc;
	heapPtr->item = (void **)ckalloc(heapPtr->maxItems*sizeof(void *));
    } else if ((heapPtr->numItems+1) == heapPtr->maxItems) {
	heapPtr->maxItems += heapPtr->resizeInc;
	heapPtr->item = (void **)ckrealloc((char *)heapPtr->item,
					heapPtr->maxItems*sizeof(void *));
    }
    if (heapPtr->item == NULL) {
	heapPtr->numItems = heapPtr->maxItems = 0;
	return NULL;
    }

    /*
     * Throw the item in at the bottom, then let it migrate to the top.
     */
    heapPtr->item[++heapPtr->numItems] = v;
    UpHeap (heapPtr, heapPtr->numItems);
    return heapPtr->item[1];
}

/*
 *--------------------------------------------------------------
 *
 * Heap_Peek --
 *
 *	Return the largest item in the heap without deleting it.
 *
 * Results:
 *	Smallest item in the heap.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
void *
Heap_Peek (heapPtr)
Heap *heapPtr;
{
    if (heapPtr == NULL) {
	return NULL;
    }

    if (heapPtr->numItems > 0) {
	return heapPtr->item[1];
    } else {
	return NULL;
    }
}

/*
 *--------------------------------------------------------------
 *
 * Heap_Size --
 *
 *	Return the number of items in the heap.
 *
 * Results:
 *	Number of items in the heap.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
int
Heap_Size(heapPtr)
Heap *heapPtr;
{
    if (heapPtr == NULL) {
	return 0;
    }
    return heapPtr->numItems;
}

/*
 *--------------------------------------------------------------
 *
 * FindHeap --
 *
 *	Locate an element in a branch of a heap beginning at index k.
 *
 * Results:
 *	0 if the item is not in that branch, the index of that item
 *	otherwise.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
static int
FindHeap (heapPtr, k, v)
Heap *heapPtr;
int k;
void *v;

{
    int n, rv;

    /* Check if we're outside the tree */
    n = heapPtr->numItems;
    if (k > n) {
	return 0;
    }

    /* Check for exact match */
    if (heapPtr->item[k] == v) {
	return k;
    }

    /* Check if we can prune the search by knowledge of heap ordering */
    if ((*heapPtr->cmpProc)(heapPtr->item[k], v) > 0) {
	return 0;
    }
    k = k*2;
    if ((rv = FindHeap (heapPtr, k, v)) != 0) {
	return rv;
    }
    if ((rv = FindHeap (heapPtr, k+1, v)) != 0) {
	return rv;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * Heap_Find --
 *
 *	Locate an element in a heap.
 *
 * Results:
 *	0 if the item is not in the heap, the index of that item
 *	otherwise.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
int
Heap_Find (heapPtr, v)
Heap *heapPtr;
void *v;

{
    if (heapPtr == NULL) {
	return 0;
    }
    return (FindHeap (heapPtr, 1, v));
}

/*
 *--------------------------------------------------------------
 *
 * Heap_Delete --
 *
 *	Delete an element from a heap.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
void
Heap_Delete (heapPtr, v)
Heap *heapPtr;
void *v;

{
    int n, index;
    
    if (heapPtr == NULL) {
       return;
    }
    index = FindHeap (heapPtr, 1, v);
    /*
     * If we found it, exchange it with the last element in the
     * heap, decrement the number of elements, then use Upheap or
     * Downheap accordingly to restore heap order
     */
    if (index == 0) {
	return;
    }
    n = heapPtr->numItems;
    if (index == n) {
	heapPtr->numItems--;
	return;
    }
    heapPtr->item[index] = heapPtr->item[n];
    heapPtr->item[n] = v;
    heapPtr->numItems--;
    if (index == 1) {
	DownHeap (heapPtr, 1);
    } else {
	if ((*heapPtr->cmpProc)(heapPtr->item[index/2],
				heapPtr->item[index]) > 0) {
	    UpHeap (heapPtr, index);
	} else {
	    DownHeap (heapPtr, index);
	}
    }
}
