/*
 * tcmAtQ.c --
 *
 *	This file provides for creation, deletion and evaluation
 *	of "at" events.
 *
 * Copyright (c) 1990, 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#define NEED_TCM_EVENT_H

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include "wndproc.h"
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <errno.h>
#include <signal.h>
#include <math.h>
#include <assert.h>
#include <sys/time.h>
#include "tk.h"
#include "heap.h"

/*
 * The following hash table is used to map from at tokens to the data
 * structures representing them.
 */
static Tcl_HashTable atTable;
static int atTableInited = 0;

/*
 * Needed by Windows NT to properly handle timers
 */
AtEvent *tcm_MinAtEvent;


/*
 *--------------------------------------------------------------
 *
 * CompareAtEvents --
 *
 *	Compare function used by the heap routines that manage the atQueue.
 *
 * Results:
 *	< 0 if at1 should fire before at2;
 *	  0 if they should fire at the same time
 *	> 0 if at1 should fire after at2
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
CompareAtEvents(at1, at2)
    AtEvent *at1, *at2;		/* Events to compare */
{
    if (at1->startTime.tv_sec == at2->startTime.tv_sec) {
	if (at1->startTime.tv_usec == at2->startTime.tv_usec) {
	    /*
	     * Start times are the same, compare end times
	     */
	    if (at1->endTime.tv_sec == at2->endTime.tv_sec) {
		if (at1->endTime.tv_usec == at2->endTime.tv_usec) {
		    return 0;
		} else if (at1->endTime.tv_usec < at2->endTime.tv_usec) {
		    return -1;
		} else {
		    return 1;
		}
	    } else if (at1->endTime.tv_sec < at2->endTime.tv_sec) {
		return -1;
	    } else {
		return 1;
	    }
	} else if (at1->startTime.tv_usec < at2->startTime.tv_usec) {
	    return -1;
	} else {
	    return 1;
	}
    } else if (at1->startTime.tv_sec < at2->startTime.tv_sec) {
	return -1;
    } else {
	return 1;
    }
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_FindAtEvent --
 *
 *	Given a token for an at event, this procedure returns
 *	the AtEvent associated with that token.
 *
 * Results:
 *	Pointer to the AtEvent, or NULL.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
AtEvent *
Tcm_FindAtEvent (token)
    int token;
{
    Tcl_HashEntry *entryPtr;

    if (!atTableInited) {
	return NULL;
    }
    entryPtr = Tcl_FindHashEntry (&atTable, (char *)token);
    if (entryPtr == NULL) {
	return NULL;
    }
    return (AtEvent *)Tcl_GetHashValue(entryPtr);
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_CreateAtHandler --
 *
 *	Arrange for a given C procedure to be invoked within a particular
 *	time window in the future and also passed a ClientData.
 *	Clients are responsible for the memory that ClientData's may
 *	consume.
 *
 * Results:
 *	0 if miss or failure, and an at handler token on success.
 *
 * Side effects:
 *	Sometime between startTime and endTime, the fireProc will be called.
 *	Later, when the interpreter is idle, the doneProc will be
 *	called.  If, for some reason, the fireProc is not able to be
 *	executed within the specified window, the missProc is called
 *	instead.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
int
Tcm_CreateAtHandler(startTime, endTime,
		    firePg, fireProc, fireClientData, fireFreeProc,
		    missPg, missProc, missClientData, missFreeProc)
    double startTime;
    double endTime;
    char *firePg;
    Tk_TimerProc *fireProc;
    ClientData fireClientData;
    void (* fireFreeProc) _ANSI_ARGS_((char *ptr));
    char *missPg;
    Tk_IdleProc *missProc;
    ClientData missClientData;
    void (* missFreeProc) _ANSI_ARGS_((char *ptr));
{
    register AtEvent *atEventPtr;
    double currTime;
    struct timeval tv;
    PriorityGroup *pg;
    static unsigned int id=1;
    int dummy;
    Tcl_HashEntry *entryPtr;

    /*
     * Figure out if we've miss the event already!
     */

    (void) gettimeofday(&tv, (struct timezone *) NULL);
    TimevalToTime(tv, currTime);
    if (currTime >= endTime) {
	if (missProc) {
	    Tcm_DoWhenIdle(missPg, missProc, missClientData);
	}
	if (missFreeProc) {
	    Tcm_DoWhenIdle(missPg, (Tk_IdleProc *)missFreeProc, missClientData);
	}
	return 0;
    }

    pg = Tcm_FindPg (firePg);
    if (pg == NULL) {
	return 0;
    }
    /*
     * Fill in fields for the at event, and place in heap.
     */

    atEventPtr = (AtEvent *) ckalloc(sizeof(AtEvent));

    TimeToTimeval(startTime, atEventPtr->startTime);
    TimeToTimeval(endTime, atEventPtr->endTime);
    atEventPtr->token = id++;
    atEventPtr->set = 0;
    if (!atTableInited) {
	Tcl_InitHashTable(&atTable, TCL_ONE_WORD_KEYS);
	atTableInited++;
    }
    entryPtr = Tcl_CreateHashEntry(&atTable, (char *) atEventPtr->token, &dummy);
    assert (dummy==1);	/* We *better* have just created this one! */
    Tcl_SetHashValue(entryPtr, atEventPtr);

    if (firePg != NULL) {
	atEventPtr->firePg = firePg;
    } else {
	atEventPtr->firePg = currPg->name;
    }
    atEventPtr->fireProc = fireProc;
    atEventPtr->fireClientData = fireClientData;
    atEventPtr->fireFreeProc = fireFreeProc;

    if (missPg != NULL) {
	atEventPtr->missPg = missPg;
    } else {
	atEventPtr->missPg = currPg->name;
    }
    atEventPtr->missProc = missProc;
    atEventPtr->missClientData = missClientData;
    atEventPtr->missFreeProc = missFreeProc;

    if (pg->atQueue == NULL)
	pg->atQueue = Heap_Alloc(CompareAtEvents, 100);
    Heap_Insert(pg->atQueue, (void *) atEventPtr);
    pg->active = 1;

    return (atEventPtr->token);
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_FinalFreeAtHandler --
 *
 *	Free up the memory allocated by create at handler and
 *	make free callbacks.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory may be freed.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
static void
Tcm_FinalFreeAtHandler(atEventPtr)
    AtEvent *atEventPtr;
{
    /*
     * Clean up the fire procedure... 
     */
    if ((atEventPtr->fireFreeProc != NULL) && 
	(atEventPtr->fireClientData != NULL)) {
	(*atEventPtr->fireFreeProc)((char *)atEventPtr->fireClientData);
    }

    /*
     * Clean up the miss procedure... 
     */
    if ((atEventPtr->missFreeProc != NULL) && 
	(atEventPtr->missClientData != NULL)) {
	(*atEventPtr->missFreeProc)((char *)atEventPtr->missClientData);
    }

    ckfree((char *) atEventPtr);
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_FreeAtHandler --
 *
 *	Call Tk_EventuallyFree to free up the memory allocated by create
 *	at handler.  Also erases the has table entry.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory may be freed.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
void
Tcm_FreeAtHandler(atEventPtr)
    AtEvent *atEventPtr;
{
    int token = atEventPtr->token;
    Tcl_HashEntry *entryPtr;

    if (atEventPtr == tcm_MinAtEvent) {
	tcm_MinAtEvent = NULL;
    }
    entryPtr = Tcl_FindHashEntry (&atTable, (char *)token);
    if (entryPtr != NULL) {
	Tcl_DeleteHashEntry(entryPtr);

	Tk_EventuallyFree(atEventPtr, (Tk_FreeProc *) Tcm_FinalFreeAtHandler);
    }
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_DeleteAtHandler --
 *
 *	Deletes the an at handler from the queue given a token.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory may be freed.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
void
Tcm_DeleteAtHandler(token)
    int token;
{
    AtEvent *atEventPtr;
    PriorityGroup *pg;

    atEventPtr = Tcm_FindAtEvent (token);
    if (atEventPtr == NULL) {
	return;
    }

    /*
     * Clean up the fire procedure... 
     */
    pg = Tcm_FindPg(atEventPtr->firePg);
    if (pg != NULL) {
	Heap_Delete(pg->atQueue, atEventPtr);
    }

    /*
     * Clean up the miss procedure... 
     */
    pg = Tcm_FindPg(atEventPtr->missPg);
    if (pg != NULL) {
	Heap_Delete(pg->atQueue, atEventPtr);
    }
    Tcm_FreeAtHandler(atEventPtr);
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_AtCmd --
 *
 *	Arrange for a given command to be evaluated at a particular
 *	time in the future.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	Sometime between startTime and endTime, the cmd will be evaluated.
 *	Later, when the interpreter is idle, the done command will be
 *	evaluated.  If, for some reason, the command is not able to be
 *	executed within the specified window, the miss command is evaluated
 *	instead.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
int
Tcm_AtCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Not used */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    double startTime, endTime;
    char *missCmd, *fireCmd;
    int token;
    EvalRequest *fireRequestPtr;
    EvalRequest *missRequestPtr;

    if (currPg->atQueue == NULL) {
	currPg->atQueue = Heap_Alloc(CompareAtEvents, 100);
    }
    if (Tcl_GetDouble (interp, argv[1], &startTime) != TCL_OK) {
	return TCL_ERROR;
    }
    if (argc == 5) {
	if (Tcl_GetDouble (interp, argv[2], &endTime) != TCL_OK) {
	    return TCL_ERROR;
	}
	fireCmd = argv[3][0]?argv[3]:NULL;
	missCmd = argv[4][0]?argv[4]:NULL;
    } else if (argc == 4) {
	if (Tcl_GetDouble (interp, argv[2], &endTime) != TCL_OK) {
	    return TCL_ERROR;
	}
	fireCmd = argv[3][0]?argv[3]:NULL;
	missCmd = NULL;
    } else if (argc == 3) {
	endTime = HUGE_VAL;
	fireCmd = argv[2][0]?argv[2]:NULL;
	missCmd = NULL;
    } else {
	Tcl_AppendResult(interp, "wrong # args: should be one of ",
		     "\"", argv[0], " startTime endTime fireCmd ?missCmd?\"\n",
		     "\"", argv[0], " startTime fireCmd\"\n",
		     (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Queue the at event up, using Tcm_CreateAtHandler.
     */
    if (fireCmd != NULL) {
	fireRequestPtr = Tcm_MakeEvalRequest (interp, fireCmd);
    } else {
	fireRequestPtr = NULL;
    }

    if (missCmd != 0) {
	missRequestPtr = Tcm_MakeEvalRequest (interp, missCmd);
    } else {
	missRequestPtr = NULL;
    }

    token = Tcm_CreateAtHandler(startTime, endTime,
				currPg->name,
				(fireRequestPtr!=NULL)?Tcm_Evaluator:NULL,
				(ClientData) fireRequestPtr,
				NULL,
				currPg->name,
				(missRequestPtr!=NULL)?Tcm_Evaluator:NULL,
				(ClientData) missRequestPtr,
				NULL);

    if (token != 0)
	sprintf(interp->result, "at%d", token);

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_AtrmCmd --
 *
 *	Delete a previously-registered at queue entry.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	Destroy the at event callback identified by token,
 *	so that its associated procedure will not be called.
 *	If the at even has already fired, or if the given
 *	token doesn't exist, then nothing happens.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
int
Tcm_AtrmCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* ignored */
    Tcl_Interp *interp;		/* tcl interpreter */
    int argc;			/* Number of arguments */
    char **argv;		/* Arg list */
{
    int n;
    int token;

    for (n = 1; n < argc; n++) {
	if (strncmp("at", argv[n], 2) != 0) {
	    Tcl_AppendResult(interp, argv[0], ": invalid token \"",
			     argv[n], "\"", (char *) NULL);
	    return TCL_ERROR;
	}
	token = atoi(argv[n] + 2);
	Tcm_DeleteAtHandler(token);
    }

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_AtqCmd --
 *
 *	Return a list of all events in the atQueue
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	Memory is allocated.
 *
 *--------------------------------------------------------------
 */
 /* ARGSUSED */
int
Tcm_AtqCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* ignored */
    Tcl_Interp *interp;		/* tcl interpreter */
    int argc;			/* Number of arguments */
    char **argv;		/* Arg list */
{
    char str[32];
    register AtEvent *atEventPtr;
    int i;
    PriorityGroup *pg;

    if (argc != 1) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			 "\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_ResetResult(interp);
    for (pg = pList; pg != NULL; pg = pg->next) {
	if (pg->atQueue != NULL) {
	    for (i = 1; i <= pg->atQueue->numItems; i++) {
		atEventPtr = (AtEvent *) pg->atQueue->item[i];
		sprintf(str, "at%d", atEventPtr->token);
		Tcl_AppendElement(interp, str);
	    }
	}
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_CreateTimerHandler --
 *
 *	Arrange for a given procedure to be invoked at a particular
 *	time in the future.
 *
 * Results:
 *	The return value is a token for the timer event, which
 *	may be used to delete the event before it fires.
 *
 * Side effects:
 *	When milliseconds have elapsed, proc will be invoked
 *	exactly once.
 *
 *--------------------------------------------------------------
 */

Tk_TimerToken
Tcm_CreateTimerHandler(pg, milliseconds, proc, clientData)
    char *pg;			/* Priority group */ 
    int milliseconds;		/* How many milliseconds to wait before
				 * invoking proc. */
    Tk_TimerProc *proc;		/* Procedure to invoke. */
    ClientData clientData;	/* Arbitrary data to pass to proc. */
{

    struct timeval tv;
    double time;

    /*
     * Compute when the event should fire.
     */

    (void) gettimeofday(&tv, (struct timezone *) NULL);
    TimevalToTime (tv, time);
    time += milliseconds / 1000.0;

    return (Tk_TimerToken) Tcm_CreateAtHandler (time, HUGE_VAL,
		pg,
		proc,
		clientData,
		NULL,
		(char *)NULL,
		(Tk_TimerProc *)NULL,
		(ClientData)NULL,
		NULL);
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_DeleteTimerHandler --
 *
 *	Delete a previously-registered timer handler.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Destroy the timer callback identified by TimerToken,
 *	so that its associated procedure will not be called.
 *	If the callback has already fired, or if the given
 *	token doesn't exist, then nothing happens.
 *
 *--------------------------------------------------------------
 */

void
Tcm_DeleteTimerHandler(token)
    Tk_TimerToken token;	/* Result previously returned by
				 * Tk_DeleteTimerHandler. */
{
    Tcm_DeleteAtHandler((int)token);
}

/*
 *--------------------------------------------------------------
 *
 * Tk_CreateTimerHandler --
 *
 *	Arrange for a given procedure to be invoked at a particular
 *	time in the future.  This routine is provided for backwards
 *	compatability.
 *
 * Results:
 *	The return value is a token for the timer event, which
 *	may be used to delete the event before it fires.
 *
 * Side effects:
 *	When milliseconds have elapsed, proc will be invoked
 *	exactly once.
 *
 *--------------------------------------------------------------
 */

Tk_TimerToken
Tk_CreateTimerHandler(milliseconds, proc, clientData)
    int milliseconds;		/* How many milliseconds to wait before
				 * invoking proc. */
    Tk_TimerProc *proc;		/* Procedure to invoke. */
    ClientData clientData;	/* Arbitrary data to pass to proc. */
{

    struct timeval tv;
    double time;

    /*
     * Compute when the event should fire.
     */

    (void) gettimeofday(&tv, (struct timezone *) NULL);
    TimevalToTime (tv, time);
    time += milliseconds / 1000.0;

    return (Tk_TimerToken) Tcm_CreateAtHandler (time, HUGE_VAL,
		"tk",
		proc,
		clientData,
		NULL,
		(char *)NULL,
		(Tk_TimerProc *)NULL,
		(ClientData)NULL,
		NULL);
}


/*
 *--------------------------------------------------------------
 *
 * Tk_DeleteTimerHandler --
 *
 *	Delete a previously-registered timer handler.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Destroy the timer callback identified by TimerToken,
 *	so that its associated procedure will not be called.
 *	If the callback has already fired, or if the given
 *	token doesn't exist, then nothing happens.
 *
 *--------------------------------------------------------------
 */

void
Tk_DeleteTimerHandler(token)
    Tk_TimerToken token;	/* Result previously returned by
				 * Tk_DeleteTimerHandler. */
{
    Tcm_DeleteAtHandler((int)token);
}

/*
 *--------------------------------------------------------------
 *
 * Tk_SystimeCmd --
 *
 *	This function implements the tcl "systime" command.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
            /* ARGSUSED */
int
Tk_SystimeCmd (clientData, interp, argc, argv)
    ClientData clientData;         /* Often ignored */
    Tcl_Interp *interp;             /* tcl interpreter */
    int argc;                       /* Number of arguments */
    char **argv;                   /* Arg list */
{
    struct timeval tv;

    if (argc != 1) {
	Tcl_AppendResult(interp, argv[0], " doesn't take any args", NULL);
	return TCL_ERROR;
    }

    (void) gettimeofday(&tv, NULL);
    sprintf (interp->result, "%d.%06d", tv.tv_sec, tv.tv_usec);
    return TCL_OK;
}
