/*
 * tcmEvent.h --
 *
 *	This file conatins the type declarations shared across the
 *	event handler modules.
 *
 * Copyright (c) 1990, 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 */
#ifndef _TK_EVENT_
#define _TK_EVENT_

#include "tk.h"
#include "dp.h"

#include "heap.h"
#include <sys/time.h>

#include <fcntl.h>
#ifdef NO_STDLIB_H
#   include "compat/stdlib.h"
#else
#   include <stdlib.h>
#endif
#include <sys/types.h>
#include <sys/file.h>
#ifdef HAVE_SYS_SELECT_H
#   include <sys/select.h>
#endif
#include <sys/stat.h>
#include <sys/time.h>
#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#else
#   include "compat/unistd.h"
#endif

/*
 * Define OPEN_MAX if it isn't already defined for this system.
 */

#ifndef OPEN_MAX
#   define OPEN_MAX 256
#endif

/*
 * The following macro defines the type of the mask arguments to
 * select:
 */

#ifndef NO_FD_SET
#   define SELECT_MASK fd_set
#else
#   ifndef _AIX
	typedef long fd_mask;
#   endif
#   if defined(_IBMR2)
#	define SELECT_MASK void
#   else
#	define SELECT_MASK int
#   endif
#endif

/*
 * Define "NBBY" (number of bits per byte) if it's not already defined.
 */

#ifndef NBBY
#   define NBBY 8
#endif

/*
 * The following macro defines the number of fd_masks in an fd_set:
 */

#if !defined(howmany)
#   define howmany(x, y) (((x)+((y)-1))/(y))
#endif
#ifdef NFDBITS
#   define MASK_SIZE howmany(FD_SETSIZE, NFDBITS)
#else
#   define MASK_SIZE howmany(OPEN_MAX, NBBY*sizeof(fd_mask))
#endif


#define TimevalToTime(tv,t)	(t=(tv).tv_sec+(tv).tv_usec/1000000.0)
#define TimeToTimeval(t,tv)                                             \
{                                                                       \
    if (t >= (int)(0x7fffffff)) {                                       \
        (tv).tv_sec= 0x7fffffff;                                        \
        (tv).tv_usec=0;                                                 \
    } else {                                                            \
        (tv).tv_sec=(int)(t);                                           \
        (tv).tv_usec=1000000.0 * ((t)-((int)(t)));                      \
    }                                                                   \
}

/*
 * For each item in the at queue for a given priority, there is one
 * record of the following type, stored in a heap.
 */

typedef struct {
    struct timeval startTime;	/* Earliest event can fire. */
    struct timeval endTime;	/* Latest event can fire. */
    int id;			/* Integer identifier for command;  used to
				 * cancel it. */
    unsigned int token;		/* Identifies event so it can be deleted */
    int set;			/* For Window, tells if a timer has been
				 * create for this at event yet */
    char *firePg;		/* Priority group of the fire routine */
    void (* fireProc)	_ANSI_ARGS_((ClientData clientData));
				/* Procedure to call when event fires */
    ClientData fireClientData;	/* Data passed to fireProc */
    void (* fireFreeProc) _ANSI_ARGS_((char *ptr));
				/* Procedure to free fireClientData on atrm */

    char *missPg;		/* Priority group of the miss routine */
    void (* missProc)	_ANSI_ARGS_((ClientData clientData));
    				/* Procedure to call if we miss */
    ClientData missClientData;	/* Data passed to missProc */
    void (* missFreeProc) _ANSI_ARGS_((char *ptr));
    				/* Procedure to free missClientData on atrm */
} AtEvent;

/*
 * To provide an interface to the DoWhenIdle functions, a record
 * of the following type is created that is used as a parameter to the
 * Tcm_IdleEvaluator function.
 */
typedef struct {
    Tcl_Interp *interp;
    char *cmd;
} EvalRequest;

/*
 * For each file registered in a call to Tk_CreateFileHandler,
 * and for each display that's currently active, there is one
 * record of the following type.  All of these records are
 * chained together into a single list.
 */

typedef struct FileHandler {
    DP_SOCKET fd;		/* POSIX file descriptor for file under Unix,
				 * a socket on NT . */
    char *pg;			/* Name of associated priority group. */
#ifndef _WINDOWS
    fd_mask *readPtr;		/* Pointer to word in ready array
				 * for this file's read mask bit. */
    fd_mask *writePtr;		/* Same for write mask bit. */
    fd_mask *exceptPtr;		/* Same for except mask bit. */
    fd_mask *checkReadPtr;	/* Pointer to word in check array for
				 * this file's read mask bit. */
    fd_mask *checkWritePtr;	/* Same for write mask bit. */
    fd_mask *checkExceptPtr;	/* Same for except mask bit. */
    fd_mask bitSelect;		/* Value to AND with *readPtr etc. to
				 * select just this file's bit. */
#else
    int lEvent;
#endif
    int isDisplay;		/* Only necessary for NT, but used by both */
    int mask;			/* Mask of desired events: TK_READABLE, etc. */
    Tk_FileProc *proc;		/* Procedure to call, in the style of
				 * Tk_CreateFileHandler.  This is NULL
				 * if the handler was created by
				 * Tk_CreateFileHandler2. */
    Tk_FileProc2 *proc2;	/* Procedure to call, in the style of
				 * Tk_CreateFileHandler2.  NULL means that
				 * the handler was created by
				 * Tk_CreateFileHandler. */
    ClientData clientData;	/* Argument to pass to proc. */
    struct FileHandler *nextPtr;/* Next in list of all files we
				 * care about (NULL for end of
				 * list). */
} FileHandler;

/*
 * The information below is used to provide read, write, and
 * exception masks to select during calls to Tk_DoOneEvent.
 */

#ifdef NFDBITS
#   define MAX_FD FD_SETSIZE
#else
#   define MAX_FD OPEN_MAX
#endif

extern int readCount;		/* Number of files for which we */
extern int writeCount;		/* care about each event type. */
extern int exceptCount;
#ifndef _WINDOWS
extern fd_mask check[3 * MASK_SIZE];/* Integer array containing official copies
                                 * of the three sets of masks. */
extern fd_mask ready[3 * MASK_SIZE];/* Temporary copy of masks, passed to select
				 * and modified by kernel to indicate which
				 * files are actually ready. */
extern fd_mask *readPtr;	/* Pointers to the portions of */
extern fd_mask *writePtr;	/* *readyPtr for reading, writing, */
extern fd_mask *exceptPtr;	/* and excepting.  Will be NULL if
				 * corresponding count (e.g. readCount is
				 * zero. */
#else
extern fd_set check[3];		/* Integer array containing official copies
                                 * of the three sets of masks. */
extern fd_set ready[3];		/* Temporary copy of masks, passed to select
				 * and modified by kernel to indicate which
				 * files are actually ready. */
extern fd_set *readPtr;	/* Pointers to the portions of */
extern fd_set *writePtr;	/* *readyPtr for reading, writing, */
extern fd_set *exceptPtr;	/* and excepting.  Will be NULL if
				 * corresponding count (e.g. readCount is
				 * zero. */

#endif
extern int numFds;		/* Number of valid bits in mask arrays (this
				 * value is passed to select). */

/*
 * There is one of the following structures for each of the
 * handlers declared in a call to Tk_DoWhenIdle.  All of the
 * currently-active handlers are linked together into a list.
 */

typedef struct IdleHandler {
    void (*proc) _ANSI_ARGS_((ClientData clientData));
				/* Procedure to call. */
    ClientData clientData;	/* Value to pass to proc. */
    int generation;		/* Used to distinguish older handlers from
				 * recently-created ones. */
    struct IdleHandler *nextPtr;/* Next in list of active handlers. */
} IdleHandler;

/*
 * For each priority group, there is one of the following
 * structures.  Each one contains the at events, file handlers and idle
 * events for the priority group.
 */

typedef struct PriorityGroup {
    int active;			/* Is there anything that might be ready to
				 * go in this pg? */
    Tk_Uid name;		/* Name of this priority group. */
    Heap *atQueue;		/* Heap of at events. */
    FileHandler *fileList;	/* List of all file events. */
    FileHandler *nextFile;        /* Next file event pointer to be processed. */
#ifndef _WINDOWS
    int fileMask[3*MASK_SIZE];	/* Masks for bits returned by select() */
#endif
    IdleHandler *idleList;	/* First in list of idle handlers. */
    IdleHandler *lastIdlePtr;	/* Last in list (or NULL for empty list). */
    struct PriorityGroup *next;	/* Next in list of priority groups */
} PriorityGroup;

extern PriorityGroup *  pList;
extern PriorityGroup *  currPg;
extern AtEvent *	tcm_MinAtEvent;


extern PriorityGroup *	Tcm_FindPg _ANSI_ARGS_((char *groupName));
extern EvalRequest *	Tcm_MakeEvalRequest _ANSI_ARGS_((Tcl_Interp *interp,
			    char *cmd));
extern void		Tcm_Evaluator _ANSI_ARGS_((ClientData clientData));
extern int		Tcm_DoWhenIdle _ANSI_ARGS_((char *groupName,
			     Tk_IdleProc *proc, ClientData clientData));
extern Tk_TimerToken	Tcm_CreateTimerHandler _ANSI_ARGS_((char *pg,
			    int milliseconds, Tk_TimerProc *proc,
			    ClientData clientData));
extern void		Tcm_DeleteTimerHandler _ANSI_ARGS_((Tk_TimerToken token));
extern AtEvent *	Tcm_FindAtEvent _ANSI_ARGS_((int token));
extern int		Tcm_CreateAtHandler _ANSI_ARGS_((double startTime,
                            double endTime,
                            char *firePg, Tk_TimerProc *fireProc,
                            ClientData fireClientData,
                            void (*fireFreeProc ) _ANSI_ARGS_((char *ptr)),
                            char *missPg, Tk_IdleProc *missProc,
                            ClientData missClientData,
                            void (*missFreeProc ) _ANSI_ARGS_((char *ptr))));
extern void		Tcm_FreeAtHandler _ANSI_ARGS_((AtEvent *atEventPtr));
extern void		Tcm_DeleteAtHandler _ANSI_ARGS_((int token));
extern int		Tcm_AtCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
extern int		Tcm_AtrmCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
extern int		Tcm_AtqCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
extern int		Tk_SystimeCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
extern int		Tcm_CreateFileHandler _ANSI_ARGS_((char *groupName,
			    DP_SOCKET fd, int mask, Tk_FileProc *proc,
			    ClientData clientData));
extern int		Tcm_CreateFileHandler2 _ANSI_ARGS_((char *groupName,
			    DP_SOCKET fd, Tk_FileProc2 *proc, int mask,
			    ClientData clientData));
extern void		Tcm_DeleteFileHandler _ANSI_ARGS_((DP_SOCKET fd));
extern int		Tcm_DoWhenIdle _ANSI_ARGS_((char *groupName,
			    Tk_IdleProc *proc, ClientData clientData));
extern int		Tcm_CancelIdleCall _ANSI_ARGS_((char *groupName,
			    Tk_IdleProc *proc, ClientData clientData));
extern int		Tk_IdleCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
extern void		InitPriorityGroup _ANSI_ARGS_((Tcl_Interp *interp));
extern int		Tcm_CreatePriorityGroup _ANSI_ARGS_((char *groupName));

#endif
