/* 
 * tkEvent.c --
 *
 *	This file provides basic event-managing facilities, whereby
 *	procedure callbacks may be attached to certain events.  It
 *	also contains the command procedures for the commands "after"
 *	and "fileevent", plus abridged versions of "tkwait" and
 *	"update", for use with Tk_EventInit.
 *
 * Copyright (c) 1990-1994 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

static char sccsid[] = "@(#) tkEvent.c 1.97 95/06/21 15:16:33";

#define NEED_TCM_EVENT_H

#ifdef _WINDOWS
#  define WIN32_LEAN_AND_MEAN
#  include "wndproc.h"
#  include <winsock.h>
#  undef WIN32_LEAN_AND_MEAN
#  include "util.h"
#  undef IGNORE
#endif /* _WINDOWS */
#include "tkPort.h"
#include "tkInt.h"
#include <errno.h>

/*
 * The information below is used to provide read, write, and
 * exception masks to select during calls to Tk_DoOneEvent.
 */

int readCount;			/* Number of files for which we */
int writeCount;			/* care about each event type. */
int exceptCount;
fd_set ready[3];
				/* Masks passed to select and modified
				 * by kernel to indicate which files are
				 * actually ready. */
fd_set check[3];
				/* Temporary set of masks, built up during
				 * Tk_DoOneEvent, that reflects what files
				 * we should wait for in the next select
				 * (doesn't include things that we've been
				 * asked to ignore in this call). */
fd_set *readPtr;		/* Pointers to the portions of */
fd_set *writePtr;		/* *readyPtr for reading, writing, */
fd_set *exceptPtr;		/* and excepting.  Will be NULL if
				 * corresponding count (e.g. readCount is
				 * zero. */
int numFds = 0;			/* Number of valid bits in mask
				 * arrays (this value is passed
				 * to select). */
/*
 * The following procedure provides a secret hook for tkXEvent.c so that
 * it can handle delayed mouse motion events at the right time.
 */

void (*tkDelayedEventProc) _ANSI_ARGS_((void)) = NULL;

/*
 * One of the following structures exists for each file with a handler
 * created by the "fileevent" command.  Several of the fields are
 * two-element arrays, in which the first element is used for read
 * events and the second for write events.
 */

typedef struct FileEvent {
    FILE *f;				/* Stdio handle for file. */
    Tcl_Interp *interps[2];		/* Interpreters in which to execute
					 * scripts.  NULL means no handler
					 * for event. */
    char *scripts[2];			/* Scripts to evaluate in response to
					 * events (malloc'ed).  NULL means no
					 * handler for event. */
    struct FileEvent *nextPtr;		/* Next in list of all file events
					 * currently defined. */
} FileEvent;

static FileEvent *firstFileEventPtr = NULL;
					/* First in list of all existing
					 * file events. */

/*
 * The data structure below is used by the "after" command to remember
 * the command to be executed later.
 */

typedef struct AfterInfo {
    Tcl_Interp *interp;		/* Interpreter in which to execute command. */
    char *command;		/* Command to execute.  Malloc'ed, so must
				 * be freed when structure is deallocated. */
    int id;			/* Integer identifier for command;  used to
				 * cancel it. */
    Tk_TimerToken token;	/* Used to cancel the "after" command.  NULL
				 * means that the command is run as an
				 * idle handler rather than as a timer
				 * handler. */
    struct AfterInfo *nextPtr;	/* Next in list of all "after" commands for
				 * the application. */
} AfterInfo;

static AfterInfo *firstAfterPtr = NULL;
				/* First in list of all pending "after"
				 * commands. */

/*
 * The data structure below is used to report background errors.  One
 * such structure is allocated for each error;  it holds information
 * about the interpreter and the error until tkerror can be invoked
 * later as an idle handler.
 */

typedef struct BgError {
    Tcl_Interp *interp;		/* Interpreter in which error occurred.  NULL
				 * means this error report has been cancelled
				 * (a previous report generated a break). */
    char *errorMsg;		/* The error message (interp->result when
				 * the error occurred).  Malloc-ed. */
    char *errorInfo;		/* Value of the errorInfo variable
				 * (malloc-ed). */
    char *errorCode;		/* Value of the errorCode variable
				 * (malloc-ed). */
    struct BgError *nextPtr;	/* Next in list of all pending error
				 * reports. */
} BgError;

static BgError *firstBgPtr = NULL;
				/* First in list of all background errors
				 * waiting to be processed (NULL if none). */
static BgError *lastBgPtr = NULL;
				/* First in list of all background errors
				 * waiting to be processed (NULL if none). */

/*
 * Prototypes for procedures referenced only in this file:
 */

static void		AfterProc _ANSI_ARGS_((ClientData clientData));
static void		DeleteFileEvent _ANSI_ARGS_((FILE *f));
static int		FileEventProc _ANSI_ARGS_((ClientData clientData,
			    int mask, int flags));
static void		FreeAfterPtr _ANSI_ARGS_((AfterInfo *afterPtr));
static void		HandleBgErrors _ANSI_ARGS_((ClientData clientData));
static int		TkwaitCmd2 _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
static int		UpdateCmd2 _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));
static char *		WaitVariableProc2 _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, char *name1, char *name2,
			    int flags));

/*
 *--------------------------------------------------------------
 *
 * Tk_DoOneEvent --
 *
 *	Process a single event of some sort.  If there's no
 *	work to do, wait for an event to occur, then process
 *	it.
 *
 * Results:
 *	The return value is 1 if the procedure actually found
 *	an event to process.  If no event was found then 0 is
 *	returned.
 *
 * Side effects:
 *	May delay execution of process while waiting for an
 *	X event, X error, file-ready event, or timer event.
 *	The handling of the event could cause additional
 *	side effects.  Collapses sequences of mouse-motion
 *	events for the same window into a single event by
 *	delaying motion event processing.
 *
 * Notes:
 *	Tk_CreateFileHandler2 is handled differently than stock
 *	Tk 4.0.  The reason for this is that the overhead added
 *	to the event loop is significant.  Although the use of
 *	Tk_CreateFileHandler2 in Tk 4.0 would stop any busy waiting,
 *	the cost is too high for running in the Continuous Media
 *	Toolkit.
 *
 *--------------------------------------------------------------
 */
int
Tk_DoOneEvent(flags)
    int flags;		/* Miscellaneous flag values:  may be any
			 * combination of TK_DONT_WAIT, TK_X_EVENTS,
			 * TK_FILE_EVENTS, and TK_IDLE_EVENTS.
			 */
{
    register FileHandler *filePtr;
    AtEvent *minTimePtr;
    AtEvent *atEventPtr;
    struct timeval curTime, timeout, *timePtr;
    int numFound;
    PriorityGroup *pg;
    PriorityGroup *oldPG;
    FileHandler *begList, *displayFilePtr;
    int firstTime;
    int active;
    Display *display;
    int queues;
    XEvent event;
    int error;
    SOCKET fd;
    DWORD waitTime;
    int mask;

    if ((flags & TK_ALL_EVENTS) == 0) {
	flags |= TK_ALL_EVENTS;
    }

    while (1) {
    checkAll:
	if (tcl_AsyncReady) {
	    (void) Tcl_AsyncInvoke((Tcl_Interp *) NULL, 0);
	    return 1;
	}

	minTimePtr = NULL;

	/*
	 * Prime the pump: Get current ready file descriptors, and current
	 * time.
	 */
	memcpy((VOID *)ready, (VOID *)check, 3*sizeof(fd_set));
	timeout.tv_sec = timeout.tv_usec = 0;
	timePtr = &timeout;
	do {
	    numFound = select(numFds,
			      (SELECT_MASK *)readPtr,
			      (SELECT_MASK *)writePtr,
			      (SELECT_MASK *)exceptPtr,
			      timePtr);
	} while ((numFound == -1) && (errno == EINTR));
	(void) gettimeofday(&curTime, (struct timezone *) NULL);
	for (pg = pList; pg != NULL; pg = pg->next) {
	    /* Skip inactive priority groups */
	    if (!pg->active) {
		continue;
	    }
	    active = 0;

	    /*
	     * Phase One: handle any fired (or missed) at events.
	     */

	    active |= ((pg->atQueue != NULL) && (pg->atQueue->numItems != 0));
	    if (flags & TK_TIMER_EVENTS) {
		while ((pg->atQueue != NULL) && (pg->atQueue->numItems != 0))
		{
		    atEventPtr = (AtEvent *) Heap_Peek(pg->atQueue);
		    timePtr = &atEventPtr->startTime;
		    if ((timePtr->tv_sec < curTime.tv_sec)
			|| ((timePtr->tv_sec == curTime.tv_sec)
			    && (timePtr->tv_usec <= curTime.tv_usec))) {

			(void) Heap_Remove(pg->atQueue);
			timePtr = &atEventPtr->endTime;
			if ((curTime.tv_sec < timePtr->tv_sec)
			    || ((curTime.tv_sec == timePtr->tv_sec)
				&& (curTime.tv_usec <= timePtr->tv_usec))) {

			    oldPG = currPg;
			    currPg = pg;
			    if (atEventPtr->fireProc) {
				(*atEventPtr->fireProc)(atEventPtr->fireClientData);
			    }
			    Tcm_FreeAtHandler(atEventPtr);
			    currPg = oldPG;
			    return 1;
			} else {
			    if (atEventPtr->missProc)
				Tcm_DoWhenIdle(atEventPtr->missPg,
					       atEventPtr->missProc,
					       atEventPtr->missClientData);
			    if (atEventPtr->missFreeProc)
				Tcm_DoWhenIdle(atEventPtr->missPg,
					       (Tk_IdleProc *)Tcm_FreeAtHandler,
					       atEventPtr);
			}

		    } else {
			/* Keep track of nearest event to fire as we go... */
			timePtr = &minTimePtr->startTime;
			if ((minTimePtr == NULL) ||
			    ((atEventPtr->startTime.tv_sec < timePtr->tv_sec) ||
			     ((atEventPtr->startTime.tv_sec == timePtr->tv_sec) &&
			      (atEventPtr->startTime.tv_usec < timePtr->tv_usec)))) {
			    minTimePtr = atEventPtr;
			}
			break;
		    }
		}
	    }
		
	    /*
	     * Phase Two: see if there's either a file ready to be processed.
	     */

	    active |= (pg->fileList != NULL);
	    if (flags & TK_FILE_EVENTS) {

		if (pg->nextFile == NULL) {
		    pg->nextFile = pg->fileList;
		}

		begList = pg->nextFile;
		firstTime = 1;

		for (filePtr = pg->nextFile;
		     ((filePtr != NULL) &&
		      (((firstTime == 0) && (filePtr != begList)) || (firstTime == 1)));
		     filePtr = (filePtr->nextPtr == NULL) ? pg->fileList : filePtr->nextPtr) {
		    int mask;
		    
		    pg->nextFile = filePtr->nextPtr;
		    firstTime = 0;

		    if (filePtr->isDisplay) {
			continue;
		    }
		    /*
		     * Not a display:  if the file is ready, call the
		     * appropriate handler.
		     */
		    mask = 0;
		    if (FD_ISSET(filePtr->fd, &ready[0])) {
			mask |= TK_READABLE;
			FD_CLR(filePtr->fd, &ready[0]);
		    }
		    if (FD_ISSET(filePtr->fd, &ready[1])) {
			mask |= TK_WRITABLE;
			FD_CLR(filePtr->fd, &ready[1]);
		    }	    
		    if (FD_ISSET(filePtr->fd, &ready[2])) {
			mask |= TK_EXCEPTION;
			FD_CLR(filePtr->fd, &ready[2]);
		    }	    

		    if (filePtr->proc2 != NULL) {
			/*
			 * Handler created by Tk_CreateFileHandler2.
			 */

			oldPG = currPg;
			currPg = pg;
			mask = (*filePtr->proc2)(filePtr->clientData, mask, flags);
			if (mask == TK_FILE_HANDLED) {
			    currPg = oldPG;
			    return 1;
			}
			currPg = oldPG;
		    } else {
			if (mask == 0) {
			    continue;
			}
			oldPG = currPg;
			currPg = pg;
			(*filePtr->proc) (filePtr->clientData, mask);
			currPg = oldPG;
			return 1;
		    }
		}
	    }

	    /*
	     * Phase Three: Handle any X events
	     */
	    displayFilePtr = NULL;
	    display = XOpenDisplay(NULL);
	    for (filePtr = pg->fileList; filePtr != NULL; filePtr = filePtr->nextPtr) {
		if (filePtr->isDisplay) {
		    displayFilePtr = filePtr;
		    if (XFileEventsQueued(display, QueuedForceReading, X_Queue) != 0) {
			mask = (*displayFilePtr->proc2)(filePtr->clientData, TK_READABLE, flags);
			if (mask == TK_FILE_HANDLED) {
			    return 1;
			}
		    }
		    break;
		}
	    }
	    
	    /*
	     * Phase Four: if there is a delayed motion event then call
	     * a procedure to handle it.  Do it now, before calling any
	     * DoWhenIdle handlers, since the goal of idle handlers is
	     * to delay until after all pending events have been processed.
	     *
	     * The particular implementation of this (a procedure variable
	     * shared with tkXEvent.c) is a bit kludgy, but it allows this
	     * file to be used separately without any of the rest of Tk.
	     */

	    if ((tkDelayedEventProc != NULL) && (flags & TK_X_EVENTS)) {
		(*tkDelayedEventProc)();
		return 1;
	    }

	    /*
	     * Phase Five: If we have a pending DoWhenIdle request, process it.
	     */
	    active |= (pg->idleList != NULL);
	    if ((pg->idleList != NULL) && (flags & TK_IDLE_EVENTS)) {
		register IdleHandler *idlePtr;
		/*
		 * The code below is trickier than it may look, for
		 * the following reasons:
		 *
		 * 1. The handler can call Tk_DoOneEvent, so we have to remove
		 *    the hander from the list before calling it. Otherwise an
		 *    infinite loop could result.
		 * 2. Tk_CancelIdleCall can be called to remove an element from
		 *    the list while a handler is executing, so the list could
		 *    change structure during the call.
		 */

		idlePtr = pg->idleList;
		pg->idleList = idlePtr->nextPtr;
		if (pg->idleList == NULL) {
		    pg->lastIdlePtr = NULL;
		}
		oldPG = currPg;
		currPg = pg;
		(*idlePtr->proc)(idlePtr->clientData);
		ckfree((char *) idlePtr);
		currPg = pg;
		return 1;
	    }

	    pg->active = active;
	}

	/*
	 * If we're not supposed to block, now is the time to return.
	 */
	if ((flags & TK_DONT_WAIT) ||
	    (!(flags & (TK_TIMER_EVENTS|TK_FILE_EVENTS|TK_X_EVENTS)))) {
	    return 0;
	}

	queues = 0;
	if (flags & TK_X_EVENTS) {
	    queues |= X_Queue;
	}
	if (flags & TK_FILE_EVENTS) {
	    queues |= File_Queue;
	}
	if (flags & TK_TIMER_EVENTS) {
	    queues |= Timer_Queue;

	    if (minTimePtr != NULL) {
		long usec, sec;

		if (! minTimePtr->set) {
		    sec = minTimePtr->startTime.tv_sec;
		    if (sec < curTime.tv_sec) {
			goto checkAll;
		    }
		    sec -= curTime.tv_sec;
		    usec = minTimePtr->startTime.tv_usec;
		    if (usec < curTime.tv_usec) {
			sec--;
			if (sec == 0) {
			    goto checkAll;
			}
			usec += 1000000 - curTime.tv_usec;
		    } else {
			usec -= curTime.tv_usec;
		    }
		    waitTime = (sec * 1000) + ((usec + 999) / 1000);

		    if (!SetTimer(Tknt_GetPostWindow(), 1, (UINT) waitTime, NULL)) {
			return 0;
		    }
		    if (tcm_MinAtEvent) {
			tcm_MinAtEvent->set = 0;
		    }
		    minTimePtr->set = 1;
		    tcm_MinAtEvent = minTimePtr;
		}
	    }
	}

	XNextFileEvent(display, &event, queues);
	switch (event.type) {
	case TimerNotify:
	    if (tcm_MinAtEvent) {
		tcm_MinAtEvent->set = 0;
	    }
	    tcm_MinAtEvent = NULL;
	    break;

#if 0 /* In CMT, we always use sockets */
	case InputNotify:
	    /* Maintained for compatability when not using sockets */
	    for (filePtr = fileList; filePtr != NULL;
		 filePtr = filePtr->nextPtr)
	    {
		if (!filePtr->isDisplay) {
		    inFile = filePtr;
		}
	    }

	    if (!(flags & TK_FILE_EVENTS) || inFile == NULL) {
		return 0;
	    } else if (inFile->mask == TK_READABLE && inFile->proc != NULL) {
		(inFile->proc)(inFile->clientData, 0);
	    }
	    return 1;
#endif

	case SocketNotify:
	    fd = event.xsocket.socket;
	    for (pg = pList; pg != NULL; pg = pg->next) {
		for (filePtr = pg->fileList; filePtr != NULL;
		     filePtr = filePtr->nextPtr)
		{
		    if (filePtr->fd == fd) {
			/*--- Reenable the asynchronous callbacks ---*/
			WSAAsyncSelect(fd, Tknt_GetPostWindow(), 0, 0);
			if (WSAAsyncSelect(fd, Tknt_GetPostWindow(), SOCKETNOTIFY,
					   filePtr->lEvent))
			{
			    panic("Unable to do WSAAsyncSelect\n");
			}
			goto found;
		    }
		}
	    }
	found:
	    memcpy((VOID *) ready, (VOID *) check,
		   3*sizeof(fd_set));
	    timeout.tv_sec = timeout.tv_usec = 0;
	    numFound = select(numFds, readPtr,
			      writePtr, exceptPtr,
			      &timeout);
	    error = WSAGetLastError();
	    if (numFound == -1 || errno == WSAEINTR) {
		/*
		 * Some systems don't clear the masks after an error, so
		 * we have to do it here.
		 */

		FD_ZERO(&ready[0]);
		FD_ZERO(&ready[1]);
		FD_ZERO(&ready[2]);
	    }
	    break;

	default:		/* Must be an X event */
	    XPutBackEvent(display, &event);
	    break;
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tk_Sleep --
 *
 *	Delay execution for the specified number of milliseconds.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Time passes.
 *
 *----------------------------------------------------------------------
 */

void
Tk_Sleep(ms)
    int ms;			/* Number of milliseconds to sleep. */
{
#ifdef _WINDOWS
    Sleep(ms);
#else
    static struct timeval delay;

    delay.tv_sec = ms/1000;
    delay.tv_usec = (ms%1000)*1000;
    (void) select(0, (SELECT_MASK *) 0, (SELECT_MASK *) 0,
	    (SELECT_MASK *) 0, &delay);
#endif
}

/*
 *----------------------------------------------------------------------
 *
 * Tk_BackgroundError --
 *
 *	This procedure is invoked to handle errors that occur in Tcl
 *	commands that are invoked in "background" (e.g. from event or
 *	timer bindings).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The command "tkerror" is invoked later as an idle handler to
 *	process the error, passing it the error message.  If that fails,
 *	then an error message is output on stderr.
 *
 *----------------------------------------------------------------------
 */

void
Tk_BackgroundError(interp)
    Tcl_Interp *interp;		/* Interpreter in which an error has
				 * occurred. */
{
    BgError *errPtr;
    char *varValue;

    /*
     * The Tcl_AddErrorInfo call below (with an empty string) ensures that
     * errorInfo gets properly set.  It's needed in cases where the error
     * came from a utility procedure like Tcl_GetVar instead of Tcl_Eval;
     * in these cases errorInfo still won't have been set when this
     * procedure is called.
     */

    Tcl_AddErrorInfo(interp, "");
    errPtr = (BgError *) ckalloc(sizeof(BgError));
    errPtr->interp = interp;
    errPtr->errorMsg = (char *) ckalloc((unsigned) (strlen(interp->result)
	    + 1));
    strcpy(errPtr->errorMsg, interp->result);
    varValue = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if (varValue == NULL) {
	varValue = errPtr->errorMsg;
    }
    errPtr->errorInfo = (char *) ckalloc((unsigned) (strlen(varValue) + 1));
    strcpy(errPtr->errorInfo, varValue);
    varValue = Tcl_GetVar(interp, "errorCode", TCL_GLOBAL_ONLY);
    if (varValue == NULL) {
	varValue = "";
    }
    errPtr->errorCode = (char *) ckalloc((unsigned) (strlen(varValue) + 1));
    strcpy(errPtr->errorCode, varValue);
    errPtr->nextPtr = NULL;
    if (firstBgPtr == NULL) {
	firstBgPtr = errPtr;
	Tk_DoWhenIdle(HandleBgErrors, (ClientData) NULL);
    } else {
	lastBgPtr->nextPtr = errPtr;
    }
    lastBgPtr = errPtr;
    Tcl_ResetResult(interp);
}

/*
 *----------------------------------------------------------------------
 *
 * HandleBgErrors --
 *
 *	This procedure is invoked as an idle handler to process all of
 *	the accumulated background errors.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Depends on what actions "tkerror" takes for the errors.
 *
 *----------------------------------------------------------------------
 */

static void
HandleBgErrors(clientData)
    ClientData clientData;		/* Not used. */
{
    Tcl_Interp *interp;
    char *command;
    char *argv[2];
    int code;
    BgError *errPtr;

    while (firstBgPtr != NULL) {
	interp = firstBgPtr->interp;
	if (interp == NULL) {
	    goto doneWithReport;
	}

	/*
	 * Restore important state variables to what they were at
	 * the time the error occurred.
	 */

	Tcl_SetVar(interp, "errorInfo", firstBgPtr->errorInfo,
		TCL_GLOBAL_ONLY);
	Tcl_SetVar(interp, "errorCode", firstBgPtr->errorCode,
		TCL_GLOBAL_ONLY);

	/*
	 * Create and invoke the tkerror command.
	 */

	argv[0] = "tkerror";
	argv[1] = firstBgPtr->errorMsg;
	command = Tcl_Merge(2, argv);
	Tcl_AllowExceptions(interp);
	code = Tcl_GlobalEval(interp, command);
	ckfree(command);
	if (code == TCL_ERROR) {
	    if (strcmp(interp->result, "\"tkerror\" is an invalid command name or ambiguous abbreviation") == 0) {
		fprintf(stderr, "%s\n", firstBgPtr->errorInfo);
	    } else {
		fprintf(stderr, "tkerror failed to handle background error.\n");
		fprintf(stderr, "    Original error: %s\n",
			firstBgPtr->errorMsg);
		fprintf(stderr, "    Error in tkerror: %s\n", interp->result);
	    }
	} else if (code == TCL_BREAK) {
	    /*
	     * Break means cancel any remaining error reports for this
	     * interpreter.
	     */

	    for (errPtr = firstBgPtr; errPtr != NULL;
		    errPtr = errPtr->nextPtr) {
		if (errPtr->interp == interp) {
		    errPtr->interp = NULL;
		}
	    }
	}

	/*
	 * Discard the command and the information about the error report.
	 */

	doneWithReport:
	ckfree(firstBgPtr->errorMsg);
	ckfree(firstBgPtr->errorInfo);
	ckfree(firstBgPtr->errorCode);
	errPtr = firstBgPtr->nextPtr;
	ckfree((char *) firstBgPtr);
	firstBgPtr = errPtr;
    }
    lastBgPtr = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Tk_AfterCmd --
 *
 *	This procedure is invoked to process the "after" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tk_AfterCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter.  Not used.*/
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    /*
     * The variable below is used to generate unique identifiers for
     * after commands.  This id can wrap around, which can potentially
     * cause problems.  However, there are not likely to be problems
     * in practice, because after commands can only be requested to
     * about a month in the future, and wrap-around is unlikely to
     * occur in less than about 1-10 years.  Thus it's unlikely that
     * any old ids will still be around when wrap-around occurs.
     */

    static int nextId = 1;
    int ms, id;
    AfterInfo *afterPtr;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " milliseconds ?command? ?arg arg ...?\" or \"",
		argv[0], " cancel id|command\"", (char *) NULL);
	return TCL_ERROR;
    }

    if (isdigit(UCHAR(argv[1][0]))) {
	if (Tcl_GetInt(interp, argv[1], &ms) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (ms < 0) {
	    ms = 0;
	}
	if (argc == 2) {
	    Tk_Sleep(ms);
	    return TCL_OK;
	}
	afterPtr = (AfterInfo *) ckalloc((unsigned) (sizeof(AfterInfo)));
	afterPtr->interp = interp;
	if (argc == 3) {
	    afterPtr->command = (char *) ckalloc((unsigned)
		    (strlen(argv[2]) + 1));
	    strcpy(afterPtr->command, argv[2]);
	} else {
	    afterPtr->command = Tcl_Concat(argc-2, argv+2);
	}
	afterPtr->id = nextId;
	nextId += 1;
	afterPtr->token = Tk_CreateTimerHandler(ms, AfterProc,
		(ClientData) afterPtr);
	afterPtr->nextPtr = firstAfterPtr;
	firstAfterPtr = afterPtr;
	sprintf(interp->result, "after#%d", afterPtr->id);
    } else if (strncmp(argv[1], "cancel", strlen(argv[1])) == 0) {
	char *arg;

	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " cancel id|command\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (argc == 3) {
	    arg = argv[2];
	} else {
	    arg = Tcl_Concat(argc-2, argv+2);
	}
	if (strncmp(arg, "after#", 6) == 0) {
	    if (Tcl_GetInt(interp, arg+6, &id) != TCL_OK) {
		return TCL_ERROR;
	    }
	    for (afterPtr = firstAfterPtr; afterPtr != NULL;
		    afterPtr = afterPtr->nextPtr) {
		if (afterPtr->id == id) {
		    break;
		}
	    }
	} else {
	    for (afterPtr = firstAfterPtr; afterPtr != NULL;
		    afterPtr = afterPtr->nextPtr) {
		if (strcmp(afterPtr->command, arg) == 0) {
		    break;
		}
	    }
	}
	if (arg != argv[2]) {
	    ckfree(arg);
	}
	if (afterPtr != NULL) {
	    if (afterPtr->token != NULL) {
		Tk_DeleteTimerHandler(afterPtr->token);
	    } else {
		Tk_CancelIdleCall(AfterProc, (ClientData) afterPtr);
	    }
	    FreeAfterPtr(afterPtr);
	}
    } else if (strncmp(argv[1], "idle", strlen(argv[1])) == 0) {
	if (argc < 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " idle script script ...\"", (char *) NULL);
	    return TCL_ERROR;
	}
	afterPtr = (AfterInfo *) ckalloc((unsigned) (sizeof(AfterInfo)));
	afterPtr->interp = interp;
	if (argc == 3) {
	    afterPtr->command = (char *) ckalloc((unsigned)
		    (strlen(argv[2]) + 1));
	    strcpy(afterPtr->command, argv[2]);
	} else {
	    afterPtr->command = Tcl_Concat(argc-2, argv+2);
	}
	afterPtr->id = nextId;
	nextId += 1;
	afterPtr->token = NULL;
	afterPtr->nextPtr = firstAfterPtr;
	firstAfterPtr = afterPtr;
	Tk_DoWhenIdle(AfterProc, (ClientData) afterPtr);
	sprintf(interp->result, "after#%d", afterPtr->id);
    } else {
	Tcl_AppendResult(interp, "bad argument \"", argv[1],
		"\": must be cancel, idle, or a number", (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * AfterProc --
 *
 *	Timer callback to execute commands registered with the
 *	"after" command.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Executes whatever command was specified.  If the command
 *	returns an error, then the command "tkerror" is invoked
 *	to process the error;  if tkerror fails then information
 *	about the error is output on stderr.
 *
 *----------------------------------------------------------------------
 */

static void
AfterProc(clientData)
    ClientData clientData;	/* Describes command to execute. */
{
    AfterInfo *afterPtr = (AfterInfo *) clientData;
    AfterInfo *prevPtr;
    int result;

    /*
     * First remove the callback from our list of callbacks;  otherwise
     * someone could delete the callback while it's being executed, which
     * could cause a core dump.
     */

    if (firstAfterPtr == afterPtr) {
	firstAfterPtr = afterPtr->nextPtr;
    } else {
	for (prevPtr = firstAfterPtr; prevPtr->nextPtr != afterPtr;
		prevPtr = prevPtr->nextPtr) {
	    /* Empty loop body. */
	}
	prevPtr->nextPtr = afterPtr->nextPtr;
    }

    /*
     * Execute the callback.
     */

    result = Tcl_GlobalEval(afterPtr->interp, afterPtr->command);
    if (result != TCL_OK) {
	Tcl_AddErrorInfo(afterPtr->interp, "\n    (\"after\" script)");
	Tk_BackgroundError(afterPtr->interp);
    }

    /*
     * Free the memory for the callback.
     */

    ckfree(afterPtr->command);
    ckfree((char *) afterPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * FreeAfterPtr --
 *
 *	This procedure removes an "after" command from the list of
 *	those that are pending and frees its resources.  This procedure
 *	does *not* cancel the timer handler;  if that's needed, the
 *	caller must do it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The memory associated with afterPtr is released.
 *
 *----------------------------------------------------------------------
 */

static void
FreeAfterPtr(afterPtr)
    AfterInfo *afterPtr;		/* Command to be deleted. */
{
    AfterInfo *prevPtr;
    if (firstAfterPtr == afterPtr) {
	firstAfterPtr = afterPtr->nextPtr;
    } else {
	for (prevPtr = firstAfterPtr; prevPtr->nextPtr != afterPtr;
		prevPtr = prevPtr->nextPtr) {
	    /* Empty loop body. */
	}
	prevPtr->nextPtr = afterPtr->nextPtr;
    }
    ckfree(afterPtr->command);
    ckfree((char *) afterPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Tk_FileeventCmd --
 *
 *	This procedure is invoked to process the "fileevent" Tcl
 *	command. See the user documentation for details on what it does.
 *	This command is based on Mark Diekhans' "addinput" command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tk_FileeventCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with interpreter.
				 * Not used.*/
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    FILE *f;
    int index, fd, c;
    size_t length;
    FileEvent *fevPtr, *prevPtr;

    /*
     * Parse arguments.
     */

    if ((argc != 3) && (argc != 4)) {
	Tcl_AppendResult(interp, "wrong # args: must be \"", argv[0],
		" fileId event ?script?", (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[2][0];
    length = strlen(argv[2]);
    if ((c == 'r') && (strncmp(argv[2], "readable", length) == 0)) {
	index = 0;
    } else if ((c == 'w') && (strncmp(argv[2], "writable", length) == 0)) {
	index = 1;
    } else {
	Tcl_AppendResult(interp, "bad event name \"", argv[2],
		"\": must be readable or writable", (char *) NULL);
	return TCL_ERROR;
    }
    if (Tcl_GetOpenFile(interp, argv[1], index, 1, &f) != TCL_OK) {
	return TCL_ERROR;
    }
    fd = fileno(f);

    /*
     * Locate an existing file handler for this file, if one exists,
     * and make a new one if none currently exists.
     */

    for (fevPtr = firstFileEventPtr; ; fevPtr = fevPtr->nextPtr) {
	if (fevPtr == NULL) {
	    if ((argc == 3) || (argv[3][0] == 0)) {
		return TCL_OK;
	    }
	    fevPtr = (FileEvent *) ckalloc(sizeof(FileEvent));
	    fevPtr->f = f;
	    fevPtr->interps[0] = NULL;
	    fevPtr->interps[1] = NULL;
	    fevPtr->scripts[0] = NULL;
	    fevPtr->scripts[1] = NULL;
	    fevPtr->nextPtr = firstFileEventPtr;
	    firstFileEventPtr = fevPtr;
	    Tk_CreateFileHandler2(fileno(f), FileEventProc,
		    (ClientData) fevPtr);
	    tcl_FileCloseProc = DeleteFileEvent;
	    break;
	}
	if (fevPtr->f == f) {
	    break;
	}
    }

    /*
     * If we're just supposed to return the current script, do so.
     */

    if (argc == 3) {
	if (fevPtr->scripts[index] != NULL) {
	    interp->result = fevPtr->scripts[index];
	}
	return TCL_OK;
    }

    /*
     * If we're supposed to delete the event handler, do so.
     */

    if (argv[3][0] == 0) {
	if (fevPtr->scripts[index] != NULL) {
	    fevPtr->interps[index] = NULL;
	    ckfree(fevPtr->scripts[index]);
	    fevPtr->scripts[index] = NULL;
	}
	if ((fevPtr->scripts[0] == NULL) && (fevPtr->scripts[1] == NULL)) {
	    if (firstFileEventPtr == fevPtr) {
		firstFileEventPtr = fevPtr->nextPtr;
	    } else {
		for (prevPtr = firstFileEventPtr; prevPtr->nextPtr != fevPtr;
			prevPtr = prevPtr->nextPtr) {
		    /* Empty loop body. */
		}
		prevPtr->nextPtr = fevPtr->nextPtr;
	    }
	    Tk_DeleteFileHandler(fileno(fevPtr->f));
	    ckfree((char *) fevPtr);
	}
	return TCL_OK;
    }

    /*
     * This is a new handler being created.  Save its script.
     */

    fevPtr->interps[index] = interp;
    if (fevPtr->scripts[index] != NULL) {
	ckfree(fevPtr->scripts[index]);
    }
    fevPtr->scripts[index] = ckalloc((unsigned) (strlen(argv[3]) + 1));
    strcpy(fevPtr->scripts[index], argv[3]);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FileEventProc --
 *
 *	This procedure is invoked by Tk's event loop to deal with file
 *	event bindings created by the "fileevent" command.
 *
 * Results:
 *	The return value is TK_FILE_HANDLED if the file was ready and
 *	a script was invoked to handle it.  Otherwise an OR-ed combination
 *	of TK_READABLE and TK_WRITABLE is returned, indicating the events
 *	that should be checked in future calls to select.
 *
 * Side effects:
 *	Whatever the event script does.
 *
 *----------------------------------------------------------------------
 */

static int
FileEventProc(clientData, mask, flags)
    ClientData clientData;	/* Pointer to FileEvent structure for file. */
    int mask;			/* OR-ed combination of the bits TK_READABLE,
				 * TK_WRITABLE, and TK_EXCEPTION, indicating
				 * current state of file. */
    int flags;			/* Flag bits passed to Tk_DoOneEvent;
				 * contains bits such as TK_DONT_WAIT,
				 * TK_X_EVENTS, Tk_FILE_EVENTS, etc. */
{
    FileEvent *fevPtr = (FileEvent *) clientData;
    Tcl_DString script;
    Tcl_Interp *interp;
    FILE *f;
    int code, checkMask;

    if (!(flags & TK_FILE_EVENTS)) {
	return 0;
    }

    /*
     * The code here is a little tricky, because the script for an
     * event could delete the event handler.  Thus, after we call
     * Tcl_GlobalEval we can't use fevPtr anymore.  We also have to
     * copy the script to make sure that it doesn't get freed while
     * being evaluated.
     */

    checkMask = 0;
    f = fevPtr->f;
    if (fevPtr->scripts[1] != NULL) {
	if (mask & TK_WRITABLE) {
	    Tcl_DStringInit(&script);
	    Tcl_DStringAppend(&script, fevPtr->scripts[1], -1);
	    interp = fevPtr->interps[1];
	    code = Tcl_GlobalEval(interp, Tcl_DStringValue(&script));
	    Tcl_DStringFree(&script);
	    if (code != TCL_OK) {
		goto error;
	    }
	    return TK_FILE_HANDLED;
	} else {
	    checkMask |= TK_WRITABLE;
	}
    }
    if (fevPtr->scripts[0] != NULL) {
	if ((mask & TK_READABLE) || TK_READ_DATA_PENDING(f)) {
	    Tcl_DStringInit(&script);
	    Tcl_DStringAppend(&script, fevPtr->scripts[0], -1);
	    interp = fevPtr->interps[0];
	    code = Tcl_GlobalEval(interp, Tcl_DStringValue(&script));
	    Tcl_DStringFree(&script);
	    if (code != TCL_OK) {
		goto error;
	    }
	    return TK_FILE_HANDLED;
	} else {
	    checkMask |= TK_READABLE;
	}
    }
    return checkMask;

    /*
     * An error occurred in the script, so we have to call
     * Tk_BackgroundError.  However, it's possible that the file ready
     * condition didn't get cleared for the file, so we could end
     * up in an infinite loop if we're not careful.  To be safe,
     * delete the event handler.
     */

    error:
    DeleteFileEvent(f);
    Tcl_AddErrorInfo(interp,
	    "\n    (script bound to file event - binding deleted)");
    Tk_BackgroundError(interp);
    return TK_FILE_HANDLED;
}

/*
 *----------------------------------------------------------------------
 *
 * DeleteFileEvent --
 *
 *	This procedure is invoked to delete all file event handlers
 *	for a file.  For example, this is necessary if a file is closed,
 *	or if an error occurs in a handler for a file.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The file event handler is removed, so it will never be invoked
 *	again.
 *
 *----------------------------------------------------------------------
 */

static void
DeleteFileEvent(f)
    FILE *f;			/* Stdio structure describing open file. */
{
    register FileEvent *fevPtr;
    FileEvent *prevPtr;

    /*
     * See if there exists a file handler for the given file.
     */

    for (prevPtr = NULL, fevPtr = firstFileEventPtr; ;
	    prevPtr = fevPtr, fevPtr = fevPtr->nextPtr) {
	if (fevPtr == NULL) {
	    return;
	}
	if (fevPtr->f == f) {
	    break;
	}
    }

    /*
     * Unlink it from the list, then free it.
     */

    if (prevPtr == NULL) {
	firstFileEventPtr = fevPtr->nextPtr;
    } else {
	prevPtr->nextPtr = fevPtr->nextPtr;
    }
    Tk_DeleteFileHandler(fileno(fevPtr->f));
    if (fevPtr->scripts[0] != NULL) {
	ckfree(fevPtr->scripts[0]);
    }
    if (fevPtr->scripts[1] != NULL) {
	ckfree(fevPtr->scripts[1]);
    }
    ckfree((char *) fevPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * TkEventCleanupProc --
 *
 *	This procedure is invoked whenever an interpreter is deleted.
 *	It deletes any file events and after commands that refer to
 *	that interpreter.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	File event handlers and after commands are removed.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
void
TkEventCleanupProc(clientData, interp)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Interpreter that is being deleted. */
{
    FileEvent *fevPtr, *prevPtr, *nextPtr;
    AfterInfo *afterPtr, *prevAfterPtr, *nextAfterPtr;
    int i;

    prevPtr = NULL;
    fevPtr = firstFileEventPtr;
    while (fevPtr != NULL) {
	for (i = 0; i < 2; i++) {
	    if (fevPtr->interps[i] == interp) {
		fevPtr->interps[i] = NULL;
		ckfree((char *) fevPtr->scripts[i]);
		fevPtr->scripts[i] = NULL;
	    }
	}
	if ((fevPtr->scripts[0] != NULL) || (fevPtr->scripts[1] != NULL)) {
	    prevPtr = fevPtr;
	    fevPtr = fevPtr->nextPtr;
	    continue;
	}
	nextPtr = fevPtr->nextPtr;
	if (prevPtr == NULL) {
	    firstFileEventPtr = nextPtr;
	} else {
	    prevPtr->nextPtr = nextPtr;
	}
	Tk_DeleteFileHandler(fileno(fevPtr->f));
	ckfree((char *) fevPtr);
	fevPtr = nextPtr;
    }

    prevAfterPtr = NULL;
    afterPtr = firstAfterPtr;
    while (afterPtr != NULL) {
	if (afterPtr->interp != interp) {
	    prevAfterPtr = afterPtr;
	    afterPtr = afterPtr->nextPtr;
	    continue;
	}
	nextAfterPtr = afterPtr->nextPtr;
	if (prevAfterPtr == NULL) {
	    firstAfterPtr = nextAfterPtr;
	} else {
	    prevAfterPtr->nextPtr = nextAfterPtr;
	}
	if (afterPtr->token != NULL) {
	    Tk_DeleteTimerHandler(afterPtr->token);
	} else {
	    Tk_CancelIdleCall(AfterProc, (ClientData) afterPtr);
	}
	ckfree(afterPtr->command);
	ckfree((char *) afterPtr);
	afterPtr = nextAfterPtr;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkwaitCmd2 --
 *
 *	This procedure is invoked to process the "tkwait" Tcl command.
 *	See the user documentation for details on what it does.  This
 *	is a modified version of tkwait with only the "variable"
 *	option, suitable for use in stand-alone mode without the rest
 *	of Tk.  It's only used when Tk_EventInit has been called.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
TkwaitCmd2(clientData, interp, argc, argv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    int c, done;
    size_t length;

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " variable name\"", (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'v') && (strncmp(argv[1], "variable", length) == 0)
	    && (length >= 2)) {
	Tcl_TraceVar(interp, argv[2],
		TCL_GLOBAL_ONLY|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
		WaitVariableProc2, (ClientData) &done);
	done = 0;
	while (!done) {
	    Tk_DoOneEvent(0);
	}
	Tcl_UntraceVar(interp, argv[2],
		TCL_GLOBAL_ONLY|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
		WaitVariableProc2, (ClientData) &done);
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\": must be variable", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Clear out the interpreter's result, since it may have been set
     * by event handlers.
     */

    Tcl_ResetResult(interp);
    return TCL_OK;
}

	/* ARGSUSED */
static char *
WaitVariableProc2(clientData, interp, name1, name2, flags)
    ClientData clientData;	/* Pointer to integer to set to 1. */
    Tcl_Interp *interp;		/* Interpreter containing variable. */
    char *name1;		/* Name of variable. */
    char *name2;		/* Second part of variable name. */
    int flags;			/* Information about what happened. */
{
    int *donePtr = (int *) clientData;

    *donePtr = 1;
    return (char *) NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateCmd2 --
 *
 *	This procedure is invoked to process the "update" Tcl command.
 *	See the user documentation for details on what it does.  This
 *	is a modified version of the command that doesn't deal with
 *	windows, suitable for use in stand-alone mode without the rest
 *	of Tk.  It's only used when Tk_EventInit has been called.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
UpdateCmd2(clientData, interp, argc, argv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    int flags;

    if (argc == 1) {
	flags = TK_DONT_WAIT|TK_FILE_EVENTS|TK_TIMER_EVENTS|TK_IDLE_EVENTS;
    } else if (argc == 2) {
	if (strncmp(argv[1], "idletasks", strlen(argv[1])) != 0) {
	    Tcl_AppendResult(interp, "bad argument \"", argv[1],
		    "\": must be idletasks", (char *) NULL);
	    return TCL_ERROR;
	}
	flags = TK_IDLE_EVENTS;
    } else {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " ?idletasks?\"", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Handle all pending events.
     */

    while (Tk_DoOneEvent(flags) != 0) {
	/* Empty loop body */
    }

    /*
     * Must clear the interpreter's result because event handlers could
     * have executed commands.
     */

    Tcl_ResetResult(interp);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tk_EventInit --
 *
 *	This procedure is invoked from Tcl_AppInit if the Tk event stuff
 *	is being used by itself (without the rest of Tk) in an application.
 *	It creates the "after" and "fileevent" commands.
 *
 * Results:
 *	Always returns TCL_OK.
 *
 * Side effects:
 *	New commands get added to interp.
 *
 *----------------------------------------------------------------------
 */

int
Tk_EventInit(interp)
    Tcl_Interp *interp;		/* Interpreter in which to set up
				 * event-handling. */
{
    Tcl_CreateCommand(interp, "after", Tk_AfterCmd, (ClientData) NULL,
	    (void (*)()) NULL);
    Tcl_CreateCommand(interp, "fileevent", Tk_FileeventCmd, (ClientData) NULL,
	    (void (*)()) NULL);
    Tcl_CreateCommand(interp, "tkwait", TkwaitCmd2, (ClientData) NULL,
	    (void (*)()) NULL);
    Tcl_CreateCommand(interp, "update", UpdateCmd2, (ClientData) NULL,
	    (void (*)()) NULL);
    Tcl_CallWhenDeleted(interp, TkEventCleanupProc, (ClientData) NULL);
    return TCL_OK;
}
