/*
 * tcmFileHndlr.c --
 *
 *	This file provides the implementation of the file handler
 *	facilities in the event model.
 *
 * Copyright (c) 1990, 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#define NEED_TCM_EVENT_H

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include "wndproc.h"
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <errno.h>
#include <signal.h>
#include <math.h>
#include <assert.h>
#include "tkInt.h"
#include "heap.h"
#ifdef _WINDOWS
#   include "util.h"
#endif

/*
 * For Window NT, but should not affect Unix
 */
#define BOGUS_DISPLAY_FD (DP_SOCKET) -2

/*
 * The following table is used to map from file descriptors to the
 * file handler data structures representing them.
 */
static FileHandler *fdTable[MAX_FD];

/*
 *----------------------------------------------------------------------
 * SetFdFileHandler
 *
 *	Way to handle the fd/FileHandler table under Unix and NT.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	None
 *----------------------------------------------------------------------
 */
#ifndef _WINDOWS
#define SetFdFileHandler(fd,filePtr) fdTable[fd] = (filePtr)
#else
#define SetFdFileHandler(fd,filePtr)
#endif /* ifndef _WINDOWS */

/*
 *----------------------------------------------------------------------
 *
 * FdToFileHandler --
 *
 *	Utility to find the FileHandler for a fd.  Under Unix,
 *	this is just a lookup table.  Under NT, we run through
 *	all the groups and look for the socket.
 *
 * Results:
 *	A pointer to the file event if found, NULL if not found.
 *
 * Side Effects:
 *	None
 *----------------------------------------------------------------------
 */
#ifdef _WINDOWS
static FileHandler *
FdToFileHandler(fd)
    DP_SOCKET fd;
{
    PriorityGroup *pg;
    FileHandler *filePtr;

    for (pg = pList; pg != NULL; pg = pg->next) {
	for (filePtr=pg->fileList; filePtr != NULL; filePtr=filePtr->nextPtr) {
	    if (filePtr->fd == fd) {
		return filePtr;
	    }
	}
    }
    return NULL;
}
#else
#define FdToFileHandler(fd) fdTable[fd]
#endif /* ifdef _WINDOWS */


/*
 *--------------------------------------------------------------
 *
 * Tcm_CreateFileHandler --
 *
 *	Arrange for a given procedure to be invoked whenever
 *	a given file becomes readable or writable.
 *
 * Results:
 *	0 for failure (usually priority group doesn't exist)
 *	1 for success.
 *
 * Side effects:
 *	From now on, whenever the I/O channel given by fd becomes
 *	ready in the way indicated by mask, proc will be invoked.
 *	See the manual entry for details on the calling sequence
 *	to proc.  If fd is already registered then the old mask
 *	and proc and clientData values will be replaced with
 *	new ones.
 *
 *--------------------------------------------------------------
 */
int
Tcm_CreateFileHandler(groupName, fd, mask, proc, clientData)
    char *groupName;		/* Name of the priority group */
    DP_SOCKET fd;		/* Integer identifier for stream. */
    int mask;			/* OR'ed combination of TK_READABLE,
				 * TK_WRITABLE, and TK_EXCEPTION: indicates
				 * conditions under which proc should be
				 * called. */
    Tk_FileProc *proc;		/* Procedure to call for each selected event.
				 * NULL means that this is a display, and
				 * that clientData is the (Display *) for it,
				 * and that events should be handled
				 * automatically. */
    ClientData clientData;	/* Arbitrary data to pass to proc. */
{
    register FileHandler *filePtr;
    int index;
    PriorityGroup *pg;
    long lEvent = 0;

#ifndef _WINDOWS
    if (fd > MAX_FD) {
	panic("Tcm_CreatefileHandler can't handle file id %d", fd);
    }
#endif

    pg = Tcm_FindPg(groupName);
    if (pg == NULL) {
	return 0;
    }

    /*
     * Make sure the file isn't already registered.  If there is, delete it
     * before continuing.
     */
    Tcm_DeleteFileHandler(fd);

    /*
     * Create a new record.
     */
    filePtr = (FileHandler *) ckalloc(sizeof(FileHandler));
    assert (filePtr != NULL);
    filePtr->pg = pg->name;
    filePtr->fd = fd;
#ifndef _WINDOWS
    index = fd / (NBBY*sizeof(fd_mask));
    filePtr->readPtr = &ready[index];
    filePtr->writePtr = &ready[index + MASK_SIZE];
    filePtr->exceptPtr = &ready[index + 2 * MASK_SIZE];
    filePtr->checkReadPtr = &check[index];
    filePtr->checkWritePtr = &check[index+MASK_SIZE];
    filePtr->checkExceptPtr = &check[index+2*MASK_SIZE];
    filePtr->bitSelect = 1 << (fd%(NBBY*sizeof(fd_mask)));
#endif
    filePtr->nextPtr = pg->fileList;
    filePtr->mask = mask;
    filePtr->proc = proc;
    filePtr->proc2 = NULL;
    filePtr->clientData = clientData;

    pg->fileList = filePtr;
    if ((mask & TK_IS_DISPLAY) || (fd == BOGUS_DISPLAY_FD)) {
	filePtr->isDisplay = 1;
#ifdef _WINDOWS
	return 1;
#endif
    } else {
	filePtr->isDisplay = 0;
    }
    pg->active = 1;
    SetFdFileHandler(fd, filePtr);

    /*
     * Update global mask, priority group mask, and counts.
     */
#ifndef _WINDOWS
    if (mask & TK_READABLE) {
	check[index] |= filePtr->bitSelect;
	pg->fileMask[index] |= filePtr->bitSelect;
	readCount++;
    }
    readPtr = (readCount == 0 ? NULL : &ready[0]);

    if (mask & TK_WRITABLE) {
	check[index + MASK_SIZE] |= filePtr->bitSelect;
	pg->fileMask[index + MASK_SIZE] |= filePtr->bitSelect;
	writeCount++;
    }
    writePtr = (writeCount == 0 ? NULL : &ready[MASK_SIZE]);

    if (mask & TK_EXCEPTION) {
	check[index + 2 * MASK_SIZE] |= filePtr->bitSelect;
	pg->fileMask[index + 2 * MASK_SIZE] |= filePtr->bitSelect;
	exceptCount++;
    }
    exceptPtr = (exceptCount == 0 ? NULL : &ready[2 * MASK_SIZE]);

    if (numFds <= fd) {
	numFds = fd + 1;
    }
#else
    if (fd != DP_INVALID_SOCKET) {
	if (mask & TK_READABLE) {
	    FD_SET(fd, &check[0]);
	    readCount++;
	    lEvent |= FD_READ;
	    lEvent |= FD_ACCEPT;
	}
	readPtr = (readCount == 0) ? (fd_set *) NULL : &ready[0];

	if (mask & TK_WRITABLE) {
	    FD_SET(fd, &check[1]);
	    writeCount++;
	    lEvent |= FD_WRITE;
	}
	writePtr = (writeCount == 0) ? (fd_set *) NULL : &ready[1];

	if (mask & TK_EXCEPTION) {
	    FD_SET(fd, &check[2]);
	    exceptCount++;
	    lEvent |= FD_OOB;
	}
	exceptPtr = (exceptCount == 0) ? (fd_set *) NULL : &ready[2];

	filePtr->lEvent = lEvent;

	if (Tknt_GetUseSocket() && Tknt_GetHasWindows()) {
	    if (WSAAsyncSelect(fd, Tknt_GetPostWindow(), SOCKETNOTIFY, lEvent)) {
		panic("Unable to do WSAAsyncSelect\n");
	    }
	}
    }

    numFds++;
#endif
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_CreateFileHandler2 --
 *
 *	Arrange for a given procedure to be invoked during the
 *	event loop to handle a particular file.
 *
 * Results:
 *	1 if it success, 0 if it fails, usually because the priority
 *	group does not exist.
 *
 * Side effects:
 *	In each pass through Tk_DoOneEvent, proc will be invoked to
 *	decide whether fd is "ready" and take appropriate action if
 *	it is.  See the manual entry for details on the calling
 *	sequence to proc.  If a handler for fd has already been
 *	registered then it is superseded by the new one.
 *
 *--------------------------------------------------------------
 */

int
Tcm_CreateFileHandler2(groupName, fd, proc, mask, clientData)
    char *groupName;		/* Priority group */
    DP_SOCKET fd;		/* Integer identifier for stream. */
    Tk_FileProc2 *proc;		/* Procedure to call from the event
				 * dispatcher. */
    int mask;			/* Events to be called back on */
    ClientData clientData;	/* Arbitrary data to pass to proc. */
{
    register FileHandler *filePtr;
    PriorityGroup *pg;

    /*
     * Let Tk_CreateFileHandler do all of the work of setting up
     * the handler, then just modify things a bit after it returns.
     */

    if (Tcm_CreateFileHandler(groupName, fd, mask,
			      (Tk_FileProc *) NULL, clientData) == 0) {
	return 0;
    }

    pg = Tcm_FindPg(groupName);
    if (pg == NULL) {
	return 0;
    }

    for (filePtr = pg->fileList; filePtr->fd != fd;
	 filePtr = filePtr->nextPtr) {
	/* Empty loop body. */
    }
    filePtr->proc = NULL;
    filePtr->proc2 = proc;
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * Tk_CreateFileHandler2 --
 *
 *	Arrange for a given procedure to be invoked during the
 *	event loop to handle a particular file.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	In each pass through Tk_DoOneEvent, proc will be invoked to
 *	decide whether fd is "ready" and take appropriate action if
 *	it is.  See the manual entry for details on the calling
 *	sequence to proc.  If a handler for fd has already been
 *	registered then it is superseded by the new one.
 *
 *--------------------------------------------------------------
 */

void
Tk_CreateFileHandler2(fd, proc, clientData)
    DP_SOCKET fd;		/* Integer identifier for stream. */
    Tk_FileProc2 *proc;		/* Procedure to call from the event
				 * dispatcher. */
    ClientData clientData;	/* Arbitrary data to pass to proc. */
{
    Tcm_CreateFileHandler2("tk", fd, proc, 0, clientData);
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_DeleteFileHandler --
 *
 *	Cancel a previously-arranged callback arrangement for
 *	a file.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If a callback was previously registered on fd, remove it.
 *
 *--------------------------------------------------------------
 */

void
Tcm_DeleteFileHandler(fd)
    DP_SOCKET fd;		/* Stream id for which to remove callback
				 * procedure. */
{
    register FileHandler *filePtr;
    FileHandler *prevPtr;
    FileHandler *tmpPtr;
    int index;
    PriorityGroup *pg;

    /*
     * Find the entry for the given file (and return if there isn't one).
     */
    filePtr = FdToFileHandler(fd);
    if (filePtr == NULL) {
	return;
    }
    SetFdFileHandler(fd, NULL);
    pg = Tcm_FindPg(filePtr->pg);
    assert (pg != NULL);	/* Has to be an associated priority group */

    /*
     * Clean up information in the callback record.
     */
#ifndef _WINDOWS
    index = filePtr->fd / (NBBY*sizeof(fd_mask));
    if (check[index] & filePtr->bitSelect) {
	readCount--;
	*filePtr->readPtr &= ~filePtr->bitSelect;
	check[index] &= ~filePtr->bitSelect;
	pg->fileMask[index] &= ~filePtr->bitSelect;
    }
    if (check[index + MASK_SIZE] & filePtr->bitSelect) {
	writeCount--;
	*filePtr->writePtr &= ~filePtr->bitSelect;
	check[index + MASK_SIZE] &= ~filePtr->bitSelect;
	pg->fileMask[index + MASK_SIZE] &= ~filePtr->bitSelect;
    }
    if (check[index + 2 * MASK_SIZE] & filePtr->bitSelect) {
	exceptCount--;
	*filePtr->exceptPtr &= ~filePtr->bitSelect;
	check[index + 2 * MASK_SIZE] &= ~filePtr->bitSelect;
	pg->fileMask[index + 2 * MASK_SIZE] &= ~filePtr->bitSelect;
    }
#else
    if (fd != DP_INVALID_SOCKET) {
	if (FD_ISSET(fd, &check[0])) {
	    FD_CLR(fd, &check[0]);
	    readCount--;
	    readPtr = (readCount == 0) ? (fd_set *) NULL : &ready[0];
	}
	if (FD_ISSET(fd, &check[1])) {
	    FD_CLR(fd, &check[1]);
	    writeCount--;
	    writePtr = (writeCount == 0) ? (fd_set *) NULL : &ready[1];
	}
	if (FD_ISSET(fd, &check[2])) {
	    FD_CLR(fd, &check[2]);
	    exceptCount--;
	    exceptPtr = (exceptCount == 0) ? (fd_set *) NULL : &ready[2];
	}

	if (Tknt_GetUseSocket() && Tknt_GetHasWindows()) {
	    if (! filePtr->isDisplay) {
		WSAAsyncSelect(fd, Tknt_GetPostWindow(), 0, 0);
	    }
	}
    }
#endif
    for (prevPtr = NULL, tmpPtr = pg->fileList;
	 tmpPtr != filePtr;
	 prevPtr = tmpPtr, tmpPtr = tmpPtr->nextPtr) {
    }
    if (prevPtr == NULL) {
	pg->fileList = filePtr->nextPtr;
    } else {
	prevPtr->nextPtr = filePtr->nextPtr;
    }
    if (pg->nextFile == filePtr) {
	pg->nextFile = filePtr->nextPtr;
    }

    ckfree((char *) filePtr);

    /*
     * Recompute numFds.
     */
#ifndef _WINDOWS
    numFds = 0;
    for (pg = pList; pg != NULL; pg = pg->next) {
	for (filePtr=pg->fileList; filePtr != NULL; filePtr=filePtr->nextPtr) {
	    if (numFds <= filePtr->fd) {
		numFds = filePtr->fd + 1;
	    }
	}
    }
#else
    numFds--;
#endif
}

/*
 *--------------------------------------------------------------
 *
 * Tk_CreateFileHandler --
 *
 *	This routine creates a file handler in the current priority
 *	group.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See the entry for Tcm_CreateFileHandler
 *
 *--------------------------------------------------------------
 */

void
Tk_CreateFileHandler(fd, mask, proc, clientData)
    DP_SOCKET fd;		/* Integer identifier for stream. */
    int mask;			/* OR'ed combination of TK_READABLE,
				 * TK_WRITABLE, and TK_EXCEPTION: indicates
				 * conditions under which proc should be
				 * called. */
    Tk_FileProc *proc;		/* Procedure to call for each selected event.
				 * NULL means that this is a display, and
				 * that clientData is the (Display *) for it,
				 * and that events should be handled
				 * automatically. */
    ClientData clientData;	/* Arbitrary data to pass to proc. */
{
    Tcm_CreateFileHandler("tk", fd, mask, proc, clientData);
}

/*
 *--------------------------------------------------------------
 *
 * Tk_DeleteFileHandler --
 *
 *	Cancel a previously-arranged callback arrangement for
 *	a file.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If a callback was previously registered on fd, remove it.
 *
 *--------------------------------------------------------------
 */

void
Tk_DeleteFileHandler(fd)
    DP_SOCKET fd;	/* Stream id for which to remove callback procedure. */
{
    Tcm_DeleteFileHandler(fd);
}
