/*
 * tcmIdle.c --
 *
 *	This file provides the implementation of idle
 *	events in the event model.
 *
 * Copyright (c) 1990, 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#define NEED_TCM_EVENT_H

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include "wndproc.h"
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <errno.h>
#include <signal.h>
#include <math.h>
#include <assert.h>
#include "tk.h"
#include "heap.h"

static int idleGeneration = 0;	/* Used to fill in the "generation" fields
				 * of IdleHandler structures.  Increments
				 * each time Tk_DoOneEvent starts calling
				 * idle handlers, so that all old handlers
				 * can be called without calling any of the
				 * new ones created by old ones. */


/*
 *--------------------------------------------------------------
 *
 * Tcm_MakeEvalRequest --
 *
 *	Return an initialized structure appropriate for use
 *	as an argument to Tcm_Evaluator.
 *
 * Results:
 *	Pointer to nuewly created structure.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
EvalRequest *
Tcm_MakeEvalRequest (interp, cmd)
    Tcl_Interp *interp;
    char *cmd;
{
    EvalRequest *req;
    req = (EvalRequest *) ckalloc(sizeof(EvalRequest));
    req->interp = interp;
    req->cmd = ckalloc(strlen(cmd) + 1);
    strcpy(req->cmd, cmd);
    return req;
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_Evaluator --
 *
 *	Evaluate a string in an interpreter and then release it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory is freed.
 *
 *--------------------------------------------------------------
 */

void
Tcm_Evaluator(clientData)
    ClientData clientData;
{
    EvalRequest *req = (EvalRequest *) clientData;

    if (Tcl_Eval(req->interp, req->cmd) != TCL_OK) {
	Tcl_AppendResult (req->interp,
			  "\nError occured in callback evaluation of \"",
			  req->cmd, "\"", (char *)NULL);
#ifdef TK
	Tk_BackgroundError(req->interp);
#else
	(void)fprintf (stderr, "Background error %s\n", req->interp->result);
#endif
    }
    free(req->cmd);
    free((char *) req);
}


/*
 *--------------------------------------------------------------
 *
 * Tcm_DoWhenIdle --
 *
 *	Arrange for proc to be invoked the next time the
 *	system is idle (i.e., just before the next time
 *	that Tk_DoOneEvent would have to wait for something
 *	to happen).
 *
 * Results:
 *	1 for success, 0 for failure (priority group doesn't exist)
 *
 * Side effects:
 *	Proc will eventually be called, with clientData
 *	as argument.  See the manual entry for details.
 *
 *--------------------------------------------------------------
 */
int
Tcm_DoWhenIdle(groupName, proc, clientData)
    char *groupName;		/* Name of priority group */
    Tk_IdleProc *proc;		/* Procedure to invoke. */
    ClientData clientData;	/* Arbitrary value to pass to proc. */
{
    register IdleHandler *idlePtr;
    PriorityGroup *pg;

    pg = Tcm_FindPg (groupName);
    if (pg == NULL) {
	return 0;
    }
    if (proc == NULL) {
	return 1;
    }

    idlePtr = (IdleHandler *) ckalloc(sizeof(IdleHandler));
    idlePtr->proc = proc;
    idlePtr->clientData = clientData;
    idlePtr->generation = idleGeneration;
    idlePtr->nextPtr = NULL;
    if (pg->lastIdlePtr == NULL) {
	pg->idleList = idlePtr;
    } else {
	pg->lastIdlePtr->nextPtr = idlePtr;
    }
    pg->active = 1;
    pg->lastIdlePtr = idlePtr;
    return 1;
}


/*
 *----------------------------------------------------------------------
 *
 * Tcm_CancelIdleCall --
 *
 *	If there are any when-idle calls requested to a given procedure
 *	with given clientData in the given priority group, cancel all of them.
 *
 * Results:
 *	1 for success, 0 for failure (priority group doesn't exist)
 *
 * Side effects:
 *	If the proc/clientData combination were on the when-idle list,
 *	they are removed so that they will never be called.
 *
 *----------------------------------------------------------------------
 */

int
Tcm_CancelIdleCall(groupName, proc, clientData)
    char *groupName;
    Tk_IdleProc *proc;		/* Procedure that was previously registered. */
    ClientData clientData;	/* Arbitrary value to pass to proc. */
{
    register IdleHandler *idlePtr, *prevPtr;
    IdleHandler *nextPtr;
    PriorityGroup *pg;

    pg = Tcm_FindPg (groupName);
    if (pg == NULL) {
	return 0;
    }
    for (prevPtr = NULL, idlePtr = pg->idleList; idlePtr != NULL;
	 prevPtr = idlePtr, idlePtr = idlePtr->nextPtr) {
	while ((idlePtr->proc == proc) && (idlePtr->clientData == clientData)) {
	    nextPtr = idlePtr->nextPtr;
	    ckfree((char *) idlePtr);
	    idlePtr = nextPtr;
	    if (prevPtr == NULL) {
		pg->idleList = idlePtr;
	    } else {
		prevPtr->nextPtr = idlePtr;
	    }
	    if (idlePtr == NULL) {
		pg->lastIdlePtr = prevPtr;
		return 1;
	    }
	}
    }
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * Tk_DoWhenIdle --
 *
 *	Arrange for an idle procedure within the current priority
 *	group.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Same as Tcm_DoWhenIdle
 *
 *--------------------------------------------------------------
 */

void
Tk_DoWhenIdle(proc, clientData)
    Tk_IdleProc *proc;		/* Procedure to invoke. */
    ClientData clientData;	/* Arbitrary value to pass to proc. */
{
    Tcm_DoWhenIdle(currPg->name, proc, clientData);
}


/*
 *----------------------------------------------------------------------
 *
 * Tk_CancelIdleCall --
 *
 *	Cancel a do-when-idle call within the tk group.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Same as Tcm_CancelIdleCall
 *
 *----------------------------------------------------------------------
 */

void
Tk_CancelIdleCall(proc, clientData)
    Tk_IdleProc *proc;		/* Procedure that was previously registered. */
    ClientData clientData;	/* Arbitrary value to pass to proc. */
{
    Tcm_CancelIdleCall(currPg->name, proc, clientData);
}


/*
 *--------------------------------------------------------------
 *
 * Tk_IdleCmd --
 *
 *	This function is called when the "whenidle" command is
 *	eval'd in the interp.  It provides a tcl interface to the 
 *
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	Global variables touched.
 *	I/O operations performed.
 *	Delayed effects.
 *
 *--------------------------------------------------------------
 */
            /* ARGSUSED */
int
Tk_IdleCmd (clientData, interp, argc, argv)
    ClientData clientData;         /* ignored */
    Tcl_Interp *interp;             /* tcl interpreter */
    int argc;                       /* Number of arguments */
    char **argv;                    /* Arg list */
{
    EvalRequest *req;

    if (argc != 2) {
	Tcl_AppendResult(interp, "Wrong number of args", NULL);
	Tcl_AppendResult(interp, " Usage: \"", argv[0], " cmd\"");
	return TCL_ERROR;
    }
    if (argv[1][0] == 0) {
	return TCL_OK;
    }

    req = Tcm_MakeEvalRequest (interp, argv[1]);
    Tcm_DoWhenIdle ("tk", Tcm_Evaluator, (ClientData) req);
    return TCL_OK;
}
