/*
 * tcmPG.c --
 *
 *	This file provides the implementation for the priority
 *	group facilities of the event model.
 *
 * Copyright (c) 1990, 1992, 1993 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#define NEED_TCM_EVENT_H

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include "wndproc.h"
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <errno.h>
#include <signal.h>
#include <math.h>
#include <assert.h>
#include "tk.h"
#include "heap.h"

/*
 * The information below is used to provide read, write, and
 * exception masks to select during calls to Tk_DoOneEvent.
 */

PriorityGroup *pList = NULL;
PriorityGroup *currPg = NULL;

static int              Tk_PriorityCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int argc, char **argv));


/*
 *--------------------------------------------------------------
 *
 * Tcm_CreatePriorityGroup --
 *
 *	This function allocates a priority group with the name
 *	passed in and places in on the front of the priority
 *	list.
 *
 * Results:
 *	-1 if we run out of memory, 0 if the priority group
 *	already exists, 1 otherwise.
 *
 * Side effects:
 *	The global variable pList will reflect the addition of
 *	the new priority group.
 *
 *--------------------------------------------------------------
 */
int
Tcm_CreatePriorityGroup (groupName)
    char *groupName;
{
    Tk_Uid name;
    PriorityGroup *new;

    name = Tk_GetUid(groupName);

    if (Tcm_FindPg(groupName) != NULL) {
	return 0;
    }

    new = (PriorityGroup *)ckalloc(sizeof(PriorityGroup));
    if (new == NULL) {
	return -1;
    }
    new->name = name;
    new->atQueue = NULL;
    new->fileList = NULL;
    new->nextFile = NULL;
    new->idleList = NULL;
    new->lastIdlePtr = NULL;
    new->next = pList;
    new->active = 0;
#ifndef _WINDOWS
    memset(new->fileMask, 0, sizeof(new->fileMask));
#endif
    pList = new;
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * InitPriorityGroup --
 *
 *	This function insures that the default priority group "tk"
 *	has been created.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A new priority group is created.
 *
 *--------------------------------------------------------------
 */
void
InitPriorityGroup (interp)
    Tcl_Interp *interp;
{
    static int inited;
    if (!inited) {
	inited++;
	Tcm_CreatePriorityGroup ("tk");
	Tcl_CreateCommand(interp, "priority", Tk_PriorityCmd,
		(ClientData) NULL, (void (*)()) NULL);
	currPg = pList;
	Tcm_CreatePriorityGroup ("rpc");
    }
}

/*
 *--------------------------------------------------------------
 *
 *  FindPrevPG --
 *
 *	Find the pointer to the group just before the group passed in in
 *	the PG list.
 *
 * Results:
 *	NULL if no previous group, or group doesn't exist, ptr to previous
 *	group otherwise
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static PriorityGroup *
FindPrevPG (pg)
    PriorityGroup *pg;
{
    PriorityGroup *prev;
    PriorityGroup *curr;
    for (prev = NULL, curr = pList; ; prev = curr, curr = curr->next) {
	if (curr == NULL) return NULL;
	if (curr == pg) return prev;
    }
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_FindPg --
 *
 *	Return a pointer to the priority group whose name is passed in.
 *
 * Results:
 *	Pointer to the priority group, or NULL.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
PriorityGroup *
Tcm_FindPg (groupName)
    char *groupName;
{
    PriorityGroup *rv;

    if (groupName == NULL) {
	return NULL;
    }
    groupName = Tk_GetUid(groupName);
    for (rv=pList; rv!=NULL; rv=rv->next) {
	if (rv->name == groupName) {
	    return rv;
	}
    }
    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * Tcm_DeletePriorityGroup --
 *
 *	This function deletes the priority group whose name is
 *	passed in.  It also cleans up all associated event handlers.
 *
 * Results:
 *	A standard tcl result
 *
 * Side effects:
 *	The priority group, and all its associated event handlers
 *	is destroyed.
 *
 *--------------------------------------------------------------
 */
static int
Tcm_DeletePriorityGroup (interp, groupName)
    Tcl_Interp *interp;
    char *groupName;
{
    PriorityGroup *pgPtr;
    PriorityGroup *prevPtr;
    AtEvent *atEventPtr;
    FileHandler *filePtr;
    IdleHandler *idlePtr;

    if (strcmp(groupName, "tk") == 0) {
	Tcl_AppendResult(interp, "priority delete: Can't delete group \"tk\"",
		    (char *)NULL);
	return TCL_ERROR;
    }

    pgPtr = Tcm_FindPg (groupName);
    if (pgPtr == NULL) {
	Tcl_AppendResult(interp, "priority delete: Unknown group \"",
		    groupName, "\"", (char *)NULL);
	return TCL_ERROR;
    }

    if (currPg == pgPtr) {
	currPg = Tcm_FindPg ("tk");
    }

    /*
     * Delete all the at events, one at a time.
     */
    if (pgPtr->atQueue != NULL) {
	while (pgPtr->atQueue->numItems > 0) {
	    atEventPtr = Heap_Peek(pgPtr->atQueue);
	    Tcm_DeleteAtHandler (atEventPtr->token);
	}
    }

    /*
     * Destroy all the file handlers.
     */
    while (pgPtr->fileList != NULL) {
	filePtr = pgPtr->fileList;
	Tcm_DeleteFileHandler (filePtr->fd);
    }
    
    /*
     * Cancel all the idle calls.
     */
    while (pgPtr->idleList != NULL) {
	idlePtr = pgPtr->idleList;
	Tcm_CancelIdleCall(pgPtr->name, idlePtr->proc, idlePtr->clientData);
    }

    /* Unlink this group from the list */
    prevPtr = FindPrevPG (pgPtr);
    if (prevPtr == NULL) {
	pList = pgPtr->next;
    } else {
	prevPtr->next = pgPtr->next;
    }

    /*
     * Free up the priority group and return.
     */
    ckfree ((char *) pgPtr);
    return TCL_OK;
}
/*
 *--------------------------------------------------------------
 *
 * Tk_PriorityCmd --
 *
 *	This function implements the tcl "priority" command.
 *	The possible options are:
 *
 *		priority create name
 *		priority move newGroup before|after existingGroup
 *		priority delete group
 *		priority getlist
 *		priority getcurrent
 *		priority setcurrent group
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	
 *
 *--------------------------------------------------------------
 */
            /* ARGSUSED */
static int
Tk_PriorityCmd (clientData, interp, argc, argv)
    ClientData clientData;         /* Often ignored */
    Tcl_Interp *interp;             /* tcl interpreter */
    int argc;                       /* Number of arguments */
    char **argv;                    /* Arg list */
{
    int len;
    char c;

    if (argc < 2) {
	Tcl_AppendResult(interp, "Wrong number of args", NULL);
	goto usage;
    }

    c = argv[1][0];
    len = strlen(argv[1]);

    /* ------------------------ CREATE --------------------------------- */
    if ((c == 'c') && (strncmp(argv[1], "create", len) == 0)) {
	/* create group */
	int result;

	if (argc != 3) {
	    goto usage;
	}

	result = Tcm_CreatePriorityGroup (argv[2]);

	if (result == -1) {
	    Tcl_AppendResult(interp, argv[0],
			" create: Out of memory creating group \"",
			argv[2], "\"", NULL);
	    return TCL_ERROR;
	} else if (result == 0) {
	    Tcl_AppendResult(interp, argv[0], " create: group \"", argv[2],
			     "\" already exists.", NULL);
	    return TCL_ERROR;
	}

    /* ------------------------ MOVE ----------------------------------- */
    } else if ((c == 'm') && (strncmp(argv[1], "move", len) == 0)) {
	/* move newGroup before|after existingGroup */
	int before;
	PriorityGroup *newGroup, *existingGroup, *prev;

	if (argc != 5) {
	    goto usage;
	}

	newGroup = Tcm_FindPg (argv[2]);
	if (newGroup == NULL) {
	    Tcl_AppendResult(interp, argv[0],
			     " move: target group \"", argv[2],
			     "\" doesn't exist", NULL);
	    return TCL_ERROR;
	}

	if (argv[3][0] == 'b') {
	    before = 1;
	} else if (argv[3][0] == 'a') {
	    before = 0;
	} else {
	    goto usage;
	}

	existingGroup = Tcm_FindPg (argv[4]);
	if (existingGroup == NULL) {
	    Tcl_AppendResult(interp, argv[0],
			     " move: reference group \"", argv[4],
			     "\" doesn't exist", NULL);
	    return TCL_ERROR;
	}

	if (existingGroup == newGroup) {
	    Tcl_AppendResult(interp, argv[0],
			     " move: reference group \"", argv[4],
			     "\" is same as target group", NULL);
	    return TCL_ERROR;
	}

	/* Phase 1, unlink newGroup */
	prev = FindPrevPG(newGroup);
	if (prev == NULL) 
	    pList = newGroup->next;
	else
	    prev->next = newGroup->next;

	if (before) {
	    newGroup->next = existingGroup;
	    if (pList == existingGroup) {
		pList = newGroup;
	    } else {
		prev = FindPrevPG(existingGroup);
		prev->next = newGroup;
	    }
	} else {
	    newGroup->next = existingGroup->next;
	    existingGroup->next = newGroup;
	}

    /* ------------------------ GETLIST--------------------------------- */
    } else if ((c == 'g') && (strncmp(argv[1], "getlist", len) == 0)) {
	PriorityGroup *curr;
	for (curr = pList; curr != NULL; curr = curr->next) {
	    Tcl_AppendResult(interp, curr->name, " ", (char *) NULL);
	}

    /* ------------------------ GETCURRENT ----------------------------- */
    } else if ((c == 'g') && (strncmp(argv[1], "getcurrent", len) == 0)) {
	if (argc != 2) {
	    goto usage;
	}
	Tcl_AppendResult(interp, currPg->name, " ", (char *) NULL);

    /* ------------------------ SETCURRENT ----------------------------- */
    } else if ((c == 's') && (strncmp(argv[1], "setcurrent", len) == 0)) {
	PriorityGroup *pgPtr;

	if (argc != 3) {
	    goto usage;
	}
	pgPtr = Tcm_FindPg (argv[2]);
	if (pgPtr == NULL) {
	    Tcl_AppendResult(interp, argv[0],
			     " setcurrent: Unknown priority group \"",
			     argv[2], "\"", (char *) NULL);
	    return TCL_ERROR;
	}
	currPg = pgPtr;
	Tcl_AppendResult(interp, currPg->name, " ", (char *) NULL);

    /* ------------------------ DELETE --------------------------------- */
    } else if ((c == 'd') && (strncmp(argv[1], "delete", len) == 0)) {
	if (argc != 3) {
	    goto usage;
	}
	return Tcm_DeletePriorityGroup (interp, argv[2]);

    /* ------------------------ SYNTAX ERROR --------------------------- */
    } else {
	goto usage;
    }
    return TCL_OK;

usage:
    Tcl_AppendResult(interp, " Possible usages:\n",
	 "\"", argv[0], " create name\"\n",
	 "\"", argv[0], " move newGroup before|after existingGroup\"\n",
	 "\"", argv[0], " delete group\"\n",
	 "\"", argv[0], " getlist\"\n",
	 "\"", argv[0], " getcurrent\"\n",
	 "\"", argv[0], " setcurrent group\"\n",
	 NULL);
    return TCL_ERROR;
}
