/*
 * tkWinUtil.c
 *
 * Copyright (c) 1994 Software Research Associates, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

#ifdef _WINDOWS
#  define WIN32_LEAN_AND_MEAN
#  include <windows.h>
#  undef WIN32_LEAN_AND_MEAN
#  undef IGNORE
#  include "util.h"
#  include "wndproc.h"
#endif /* ifdef _WINDOWS */
#include "tk.h"
#include "tkInt.h"
#include "tkPort.h"

typedef struct _TclConsoleEventCommand {
    Tcl_Interp *interp;
    DWORD event;
    char *command;
    struct _TclConsoleEventCommand *next;
} TclConsoleEventCommand;

static TclConsoleEventCommand *ConsoleEventCommands = NULL;

/*
 * use maneagement selection 
 */

extern int	SelGetProc _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, char *portion));

extern void	TbSetClipboard _ANSI_ARGS_((Tcl_Interp *interp, Tk_Window tkwin));
extern int	TbGetClipboardData _ANSI_ARGS_((Tcl_Interp *interp, 
			    Tk_GetSelProc *proc, ClientData clientData));



static PHANDLER_ROUTINE consoleEventHandler = NULL;
static BOOL disableCtrlC = FALSE;


/*
 *----------------------------------------------------------------------
 *
 * Prompt --
 *
 *	Issue a prompt on standard output, or invoke a script
 *	to issue the prompt.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A prompt gets output, and a Tcl script may be evaluated
 *	in interp.
 *
 *----------------------------------------------------------------------
 */

static void
Prompt(interp, partial)
    Tcl_Interp *interp;			/* Interpreter to use for prompting. */
    int partial;			/* Non-zero means there already
					 * exists a partial command, so use
					 * the secondary prompt. */
{
    char *promptCmd;
    int code;

    promptCmd = Tcl_GetVar(interp,
	partial ? "tcl_prompt2" : "tcl_prompt1", TCL_GLOBAL_ONLY);
    if (promptCmd == NULL) {
	defaultPrompt:
	if (!partial) {
	    fputs("% ", stdout);
	}
    } else {
	code = Tcl_Eval(interp, promptCmd);
	if (code != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
		    "\n    (script that generates prompt)");
	    fprintf(stderr, "%s\n", interp->result);
	    goto defaultPrompt;
	}
    }
    fflush(stdout);
}

/*
 *----------------------------------------------------------------------
 *
 * ConsoleEventFunc --
 *
 *	This function is invoked when a console event occurs.
 *	It makes the proper callbacks when necessary.
 *
 * Results:
 *	None
 *
 *----------------------------------------------------------------------
 */
static BOOL WINAPI
ConsoleEventFunc(DWORD event)
{
    TclConsoleEventCommand *curr;
    int result;
    BOOL handled;

    handled = FALSE;
    for (curr = ConsoleEventCommands; curr != NULL; curr = curr->next) {
	if (curr->event == event) {
	    handled = TRUE;
	    result = Tcl_GlobalEval(curr->interp, curr->command);
	    if (result == TCL_ERROR) {
		Tk_BackgroundError(curr->interp);
	    }
	}
    }
    return handled;
}

/*
 *----------------------------------------------------------------------
 *
 * AddConsoleEventHandler --
 *
 *	Adds or replaces a console event handler with the current
 *	one.  A previous handler is replaced if its interpreter
 *	and the event are the same.  If the command length is 0,
 *	then the handler is removed and a default handler is used.
 *
 * Results:
 *	None
 *
 *----------------------------------------------------------------------
 */
static void
AddConsoleEventHandler(Tcl_Interp *interp, DWORD event, char *command)
{
    TclConsoleEventCommand *prev, *curr, *next;
    int len;

    if (consoleEventHandler == NULL) {
	SetConsoleCtrlHandler(ConsoleEventFunc, TRUE);
	consoleEventHandler = ConsoleEventFunc;
    }

    len = strlen(command);
    for (prev = NULL, curr = ConsoleEventCommands; curr != NULL;
	 prev = curr, curr = next) {
	next = curr->next;
	if (curr->interp == interp && curr->event == event) {
	    if (curr->command) {
		ckfree(curr->command);
	    }
	    if (len > 0) {
		curr->command = ckalloc(len + 1);
		memcpy(curr->command, command, len + 1);
	    } else {
		ckfree((char *) curr);
		if (prev) {
		    prev->next = next;
		} else {
		    ConsoleEventCommands = next;
		}
	    }
	    return;
	}
    }
    if (len > 0) {
	curr = ckalloc(sizeof(TclConsoleEventCommand));
	curr->interp = interp;
	curr->event = event;
	curr->command = ckalloc(len + 1);
	memcpy(curr->command, command, len + 1);
	curr->next = NULL;
	if (prev) {
	    prev->next = curr;
	} else {
	    ConsoleEventCommands = curr;
	}
    }
    return;
}

/*
 *----------------------------------------------------------------------
 *
 * Tk_ConsoleCmd --
 *
 *	This procedure is invoked to process the "console" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

int
Tk_ConsoleCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    int c;
    int len;
    int w, h;
    int pos;
    HANDLE hStdout;
    CONSOLE_SCREEN_BUFFER_INFO bufInfo;
    HMENU hmenu;
    LONG err;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }

    c = argv[1][0];
    len = strlen(argv[1]);
    if ((c == 'b') && (strncmp(argv[1], "buffersize", len) == 0)) {
	COORD bufSize;
	if (argc != 2 && argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " buffersize ?width height?\"",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	if (hStdout == NULL) {
	    return TCL_OK;
	}
	if (argc == 2) {
	    if (! GetConsoleScreenBufferInfo(hStdout, &bufInfo)) {
		return TCL_OK;
	    }
	    sprintf(interp->result, "%d %d",
		    bufInfo.dwSize.X, bufInfo.dwSize.Y);
	    return TCL_OK;
	}

	if (Tcl_GetInt(interp, argv[2], &w) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[3], &h) != TCL_OK) {
	    return TCL_ERROR;
	}
	bufSize.X = (SHORT) w;
	bufSize.Y = (SHORT) h;
	SetConsoleScreenBufferSize(hStdout, bufSize);

    } else if ((c == 'c') && (strncmp(argv[1], "closeevent", len) == 0)  &&
	       (len > 2)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " closeevent command\"",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	AddConsoleEventHandler(interp, CTRL_CLOSE_EVENT, argv[2]);

    } else if ((c == 'c') && (strncmp(argv[1], "ctrlcevent", len) == 0)  &&
	       (len > 2)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " ctrlcevent command\"",
			     (char *) NULL);
	    return TCL_ERROR;
	}
	AddConsoleEventHandler(interp, CTRL_C_EVENT, argv[2]);

    } else if ((c == 'c') && (strncmp(argv[1], "command", len) == 0)  &&
	       (len > 2)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " command command_string\"",
			     (char *) NULL);
	    return TCL_ERROR;
	}

	puts(argv[2]);
	Tcl_RecordAndEval(interp, argv[2], TCL_EVAL_GLOBAL);
	if (*interp->result != 0) {
	    /*
	     * The statement below used to call "printf", but that resulted
	     * in core dumps under Solaris 2.3 if the result was very long.
	     */

	    puts(interp->result);
	}
	Prompt(interp, 0);
	Tcl_ResetResult(interp);

    } else if ((c == 'd') && (strncmp(argv[1], "deiconify", len) == 0) &&
	       (len > 2)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " deiconify\"", (char *) NULL);
	    return TCL_ERROR;
	}
	ShowWindow(Tknt_GetConsoleWindow(), SW_SHOWNORMAL);

    } else if ((c == 'g') && (strncmp(argv[1], "getmenuitem", len) == 0)) {
	char str[100];

	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " getmenuitem number\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &pos) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	hmenu = GetSystemMenu(Tknt_GetConsoleWindow(), FALSE);
	if (hmenu == NULL) {
	    err = GetLastError();
	    Tcl_AppendResult(interp, "get system menu failed: ",
			     wsa_strerror(GetLastError()), NULL);
	    return TCL_ERROR;
	}
	GetMenuString(hmenu, pos, str, sizeof(str), MF_BYPOSITION);
	Tcl_AppendResult(interp, str, NULL);

    } else if ((c == 'i') && (strncmp(argv[1], "iconify", len) == 0)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " iconify\"", (char *) NULL);
	    return TCL_ERROR;
	}
	ShowWindow(Tknt_GetConsoleWindow(), SW_MINIMIZE);

    } else if ((c == 'k') && (strncmp(argv[1], "killmenuitem", len) == 0)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " killmenuitem number\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &pos) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	hmenu = GetSystemMenu(Tknt_GetConsoleWindow(), FALSE);
	if (hmenu == NULL) {
	    err = GetLastError();
	    Tcl_AppendResult(interp, "get system menu failed: ",
			     wsa_strerror(GetLastError()), NULL);
	    return TCL_ERROR;
	}
	if (RemoveMenu(hmenu, pos, MF_BYPOSITION) == FALSE) {
	    Tcl_AppendResult(interp, "remove menu failed: ",
			     wsa_strerror(GetLastError()), NULL);
	    return TCL_ERROR;
	}

    } else if ((c == 's') && (strncmp(argv[1], "size", len) == 0)) {
	CONSOLE_SCREEN_BUFFER_INFO bufInfo;

	if (argc != 4 && argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " size ?width height?\"", (char *) NULL);
	    return TCL_ERROR;
	}
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	if (hStdout == NULL) {
	    return TCL_OK;
	}
	if (! GetConsoleScreenBufferInfo(hStdout, &bufInfo)) {
	    return TCL_OK;
	}
	if (argc == 2) {
	    sprintf(interp->result, "%d %d",
		    bufInfo.srWindow.Right - bufInfo.srWindow.Left + 1,
		    bufInfo.srWindow.Bottom - bufInfo.srWindow.Top + 1);
	    return TCL_OK;
	}
	    
	if (Tcl_GetInt(interp, argv[2], &w) != TCL_OK) {
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[3], &h) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (w > bufInfo.dwMaximumWindowSize.X) {
	    w = bufInfo.dwMaximumWindowSize.X;
	}
	if (h > bufInfo.dwMaximumWindowSize.Y) {
	    h = bufInfo.dwMaximumWindowSize.Y;
	}
	if (w > bufInfo.dwSize.X || h > bufInfo.dwSize.Y) {
	    if (w > bufInfo.dwSize.X) {
		bufInfo.dwSize.X = w - 1;
	    }
	    if (h > bufInfo.dwSize.Y) {
		bufInfo.dwSize.Y = h - 1;
	    }
	    SetConsoleScreenBufferSize(hStdout, bufInfo.dwSize);
	}

	if (bufInfo.srWindow.Left + w > bufInfo.dwSize.X) {
	    bufInfo.srWindow.Left = bufInfo.dwSize.X - w;
	}
	bufInfo.srWindow.Right = bufInfo.srWindow.Left + w - 1;

	if (bufInfo.srWindow.Top + h > bufInfo.dwSize.Y) {
	    bufInfo.srWindow.Top = bufInfo.dwSize.Y - h;
	}
	bufInfo.srWindow.Bottom = bufInfo.srWindow.Top + h - 1;

	SetConsoleWindowInfo(hStdout, TRUE, &bufInfo.srWindow);

    } else if ((c == 'w') && (strncmp(argv[1], "withdraw", len) == 0)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " withdraw\"", (char *) NULL);
	    return TCL_ERROR;
	}
	ShowWindow(Tknt_GetConsoleWindow(), SW_HIDE);
    } else {
	Tcl_AppendResult(interp, "unknown or ambiguous option \"", argv[1],
			 "\": must be buffersize, closeevent, command,"
			 " ctrlcevent, deiconify, getmenuitem, iconify,"
			 " killmenuitem, size,"
			 " or withdraw\"", (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TkGetServerInfo --
 *
 *	Given a window, this procedure returns information about
 *	the window server for that window.  This procedure provides
 *	the guts of the "winfo server" command.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkGetServerInfo(interp, tkwin)
    Tcl_Interp *interp;		/* The server information is returned in
				 * this interpreter's result. */
    Tk_Window tkwin;		/* Token for window;  this selects a
				 * particular display and server. */
{
    char buffer[50];
    OSVERSIONINFO info;
    int index = 0;
    static char* os[] = {"Win32", "Win32s", "Win32c"};

#ifndef VER_PLATFORM_WIN32c
#  define VER_PLATFORM_WIN32c 1
#endif

    info.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&info);
    if (info.dwPlatformId == VER_PLATFORM_WIN32s) {
	index = 1;
    } else if (info.dwPlatformId == VER_PLATFORM_WIN32c) {
	index = 2;
    }
    sprintf(buffer, "Windows %d.%d %d ", info.dwMajorVersion,
	    info.dwMinorVersion, info.dwBuildNumber);
    Tcl_AppendResult(interp, buffer, os[index], (char *) NULL);
}


/*
 * Tb_SetClipboard --
 *
 */

void
TbSetClipboard(Tcl_Interp *interp, Tk_Window tkwin)
{
    HGLOBAL hGlobal;
    char *pGlobal;
    HWND hwnd = Tk_WindowId(tkwin);
    int result;
    Tcl_DString selBytes;

    Tcl_DStringInit(&selBytes);
    result = Tk_GetSelection(interp, tkwin, XA_PRIMARY, XA_STRING, SelGetProc,
	    (ClientData) &selBytes);
    if (result != TCL_OK) {
	Tcl_DStringFree(&selBytes);
	return;
    }
    if ((hGlobal = GlobalAlloc(GHND, selBytes.length + 1)) == None) {
    	return;
    }
    pGlobal = GlobalLock(hGlobal);
    strcpy(pGlobal, selBytes.string);
    GlobalUnlock(hGlobal); 

    if (GetClipboardOwner() == hwnd) { 
	OpenClipboard(hwnd);
	SendMessage(hwnd, TKW_SELECTIONSAVE, 0, 0L);
	EmptyClipboard();
  	SetClipboardData(CF_TEXT, hGlobal);
	CloseClipboard();
    }
    Tcl_DStringFree(&selBytes);
}

int
TbGetClipboardData(interp, proc, clientData)
    Tcl_Interp *interp;		/* Interpreter to use for reporting
				 * errors. */
    Tk_GetSelProc *proc;	/* Procedure to call to process the
				 * selection, once it has been retrieved. */
    ClientData clientData;	/* Arbitrary value to pass to proc. */
{
    HGLOBAL hGlobal;
    int result;
    char *pGlobal;

    if (OpenClipboard((HWND) NULL)) {
	if ((hGlobal = GetClipboardData(CF_TEXT)) == (HGLOBAL) NULL) {
	    CloseClipboard();
	    return TCL_ERROR;
	}
	pGlobal = GlobalLock(hGlobal);
	result = (*proc)(clientData, interp, pGlobal);
	GlobalUnlock(hGlobal);
	CloseClipboard();
	return result;
    }
    return TCL_ERROR;
}

#if defined ( TB ) && defined ( KANJI )

int
TbWSTextWidth(font, wstr, gmask, len)
     XFontStruct *font;
     wchar *wstr;
     int gmask;
     int len;
{
#define bufsize	256
	XChar2b			buf[bufsize];
	XChar2b			*cp;
	XChar2b			*cpend = buf + bufsize;
	wchar			*wstr1 = wstr + len;
	int			c;
	int			width = 0;
	int			gr_mapping;

	while( wstr < wstr1 ) {
	    gr_mapping = 0;

	    cp = buf;

	    if( font == NULL ) {
		while( wstr < wstr1 && (*wstr & 0x8080) == (unsigned) gmask ) wstr++;
		continue;
	    }

	    while( wstr < wstr1 && ((c = *wstr) & 0x8080) == (unsigned) gmask ) {
		if( cp >= cpend - 1) {
		    /* flush */
		    width += XTextWidth16(font, buf, cp - buf);
		    cp = buf;
		}
		cp->byte1 = (c >> 8) & 0x7f;
		cp->byte2 = c & 0x7f;
		if (gr_mapping) {
		    if (cp->byte1 != 0) cp->byte1 |= 0x80;
		    cp->byte2 |= 0x80;
		}
		cp++;
		wstr++;
	    }

	    if( cp == buf ) continue;

	    /* flush */
	    width += XTextWidth16(font, buf, cp - buf);
	}

	return width;
}
#endif /* TB */
