/* This file contains code to convert mpeg files to mpeg chunk clipfiles
   with the appropriate information preparse out into chunks.

   Usage:

   mpeg2mpgc -in <mpeg_file> -out <output_file> ?-fps <fps>?
 
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <memory.h>
#include <netinet/in.h>
#include <unistd.h>
#include <math.h>
#include "cmChunk.h"
#include "cmCookies.h"


#define ONE_BILLION 1000000000.0
#define CF_READ_BUFFER_SIZE 512

#define MPEG_I_FRAME 1
#define MPEG_P_FRAME 2
#define MPEG_B_FRAME 3
#define MPEG_SYS_HDR_START_CODE 0x000001b3
#define MPEG_END_SEQ_CODE       0x000001b7
#define MPEG_PIC_START_CODE     0x00000100

#define MPEG_BUF_HDR_SIZE (sizeof(MPEGBufferHeader)-8)

typedef struct RefEntry {
  int type;
  int ref1;
  int ref2;
  int headerRef;   /* -1 = default matrices. */
  
  /* The following fields are not stored in file but only in memory. */

  int indexNum;
  int marker;
} RefEntry;

typedef struct HdrEntry {
    int intra_matrix_flag;
    int intra_matrix[64];
    int non_intra_matrix_flag;
    int non_intra_matrix[64];
} HdrEntry;

/*
 * The following macro is the size of a ref entry on disk.
 */

#define MPEG_REF_ENTRY_SIZE 16

static int
NextStartCode(int fd) 
{
    unsigned char readBuffer[512];
    int ret_value, cur_pos, idx, num_read;
    int state;

    cur_pos = lseek(fd, 0L, SEEK_CUR);
    num_read = read(fd, readBuffer, 512);

    if (num_read <= 0) return -1;

    idx = 0;
    state = 0;
    ret_value = -1;

    while (1) {

	if (state == 0) {
	    if (readBuffer[idx] == 0) {
		state = 1;
	    }
	    idx++;
	} else if (state == 1) {
	    if (readBuffer[idx] == 0) {
		state = 2;
	    } else {
		state = 0;
	    }
	    idx++;
	} else if (state == 2) {
	    if (readBuffer[idx] == 1) {
		state = 3;
	    } else if (readBuffer[idx] != 0) {
		state = 0;
	    }
	    idx++;
	} else if (state == 3) {
	    if (readBuffer[idx] == 0) {
		ret_value = MPEG_PIC_START_CODE;
		break;
	    } else if (readBuffer[idx] == 179) {
		ret_value = MPEG_SYS_HDR_START_CODE;
		break;
	    } else if (readBuffer[idx] == 183) {
		ret_value = MPEG_END_SEQ_CODE;
		break;
	    } else {
		state = 0;
	    }
	    idx++;
	}
   
	if (idx >= num_read) {
	    cur_pos = lseek(fd, 0L, SEEK_CUR);
	    num_read = read(fd, readBuffer, 512);
	    idx = 0;

	    if (num_read <= 0) return -1;
	}
    }

    idx -= 3;
    lseek(fd, cur_pos+idx, SEEK_SET);
    return ret_value;
}

static void
CopyRefEntry(RefEntry *r1, RefEntry *r2)
{
    memcpy((char *) r1, (char *) r2, sizeof(RefEntry));
    return;
}

#define RD_BUF_SIZE 8096

void
main (int argc, char **argv) 
{
    int in_fd;
    double fps;
    int cntr;
    int hdrRef, next_start_code;
    unsigned char readBuffer[RD_BUF_SIZE];
    MPEGHeader *cfInfo;
    MediaIndepHeader mi_header;
    int numHeaders;
    HdrEntry *headerTable;
    int i, next_value;
    double sec_per_frame;
    int frameNum, ref1, ref2;
    int numFrameAlloced;
    int *rawFOT;
    double *rawFTT;
    RefEntry *rawRefTable;
    int offset, picture_type, end_offset;
    int length;
    RefEntry tmpRefEntry;
    int tmpLength, tmpOffset;
    int numFrames;
    double clipLength;
    int dataStart;
    int lastI, nextI, nextP;
    tChunkFileState *chunkState;
    int j, num_read;


    fps = 0.0;
    in_fd = -1;
    chunkState = NULL;

    if (argc < 4) goto usage;

    cntr = 0;
    while (cntr < argc) {
	if (strcmp(argv[cntr], "-in") == 0) {
	    cntr++;

	    if (cntr == argc) {
	      usage:
		fprintf(stderr, "Usage: mpeg2mpgc -in <mpeg_file> -out <output_file> ?-fps <fps>?\n");
		exit(1);
	    }

	    in_fd = open(argv[cntr], O_RDONLY);

	    if (in_fd < 0) {
		fprintf(stderr, "Couldn't open input file: %s\n", argv[cntr]);
		exit(1);
	    }
	} else if (strcmp(argv[cntr], "-out") == 0) {
	    cntr++;

	    if (cntr == argc) goto usage;

	    chunkState = Chunk_OpenFile(argv[cntr], "wc");
	    if (chunkState == NULL) {
		fprintf(stderr, "Couldn't open output file: %s\n", argv[cntr]);
		exit(1);
	    }
	} else if (strcmp(argv[cntr], "-fps") == 0) {
	    cntr++;

	    if (cntr == argc) goto usage;

	    if (sscanf(argv[cntr], "%lf", &(fps)) != 1) {
		fprintf(stderr, "Didn't understand fps value: %s\n", argv[cntr]);
		exit(1);
	    }
	}
	cntr++;
    }

    if (in_fd < 0) {
	fprintf(stderr, "Must specify an input file.\n");
	exit(1);
    }

    if (chunkState == NULL) {
	fprintf(stderr, "Must specify an output file.\n");
	exit(1);
    }

    hdrRef = -1;
    
    next_start_code = NextStartCode(in_fd);
    
    while ((next_start_code != MPEG_SYS_HDR_START_CODE) &&
	   (next_start_code != MPEG_END_SEQ_CODE) &&
	   (next_start_code != -1)) {
	lseek(in_fd, 4, SEEK_CUR);
	next_start_code = NextStartCode(in_fd);
    }
    
    if ((next_start_code == -1) || (next_start_code == MPEG_END_SEQ_CODE)) {
	/*
	 * Reached end of file without finding sys start code.
	 */	
	fprintf(stderr, "Couldn`t find mpeg seq. start code.\n");
	exit(1);
    }

    /*
     * Read in system header, set up cfInfo.
     */

    read(in_fd, readBuffer, 140);

    cfInfo = (MPEGHeader *) malloc(sizeof(MPEGHeader));
    
    cfInfo->width = (readBuffer[4]*16 + (readBuffer[5] >> 4));
    cfInfo->height = (((readBuffer[5] & 0x0f) * 256) + readBuffer[6]);
    cfInfo->aspectRatioCode = (readBuffer[7] >> 4);
    cfInfo->frameRateCode = (readBuffer[7] & 0x0f);
    cfInfo->bitRate = (((readBuffer[8] * 256*4) + (readBuffer[9] * 4) +
			       (readBuffer[10] >> 6)));
    cfInfo->vbvBufferSize = ((((readBuffer[10] & 0x1f)*32) +
				     (readBuffer[11] >> 3)));
    if ((readBuffer[11] & 0x04)) 
      cfInfo->constrainedParamFlag = 1;

    numHeaders = 0;

    if ((readBuffer[11] & 0x02)) {
	hdrRef++;
	
	numHeaders = 1;
	headerTable = (HdrEntry *) malloc(sizeof(HdrEntry)*1);
    
	headerTable[0].intra_matrix_flag = 1;
	
	for(i=0; i<64; i++) {
	    next_value = ((readBuffer[11+i] & 0x01 * 128) + 
			  (readBuffer[12+i] >> 1));
	    headerTable[0].intra_matrix[i] = next_value;
	}

	if ((readBuffer[75] & 0x01)) {
	    headerTable[0].non_intra_matrix_flag = 1;
	    for (i=0; i<64; i++) {
		headerTable[0].non_intra_matrix[i] = readBuffer[76+i];
	    }
	} else {
	    headerTable[0].non_intra_matrix_flag = 0;
	    lseek(in_fd, -1*(140-76), SEEK_CUR);
	}
    } else if ((readBuffer[11] & 0x01)) {
	hdrRef++;
	
	numHeaders = 1;
	headerTable = (HdrEntry *) malloc(sizeof(HdrEntry)*1);
    
	headerTable[0].intra_matrix_flag = 0;
	headerTable[0].non_intra_matrix_flag = 1;
	for (i=0; i<64; i++) {
	    headerTable[0].non_intra_matrix[i] = readBuffer[12+i];
	} 

	lseek (in_fd, -1*(140-76), SEEK_CUR);
    } else {
	numHeaders = 0;
	headerTable = NULL;
	lseek (in_fd, -1*(140-12), SEEK_CUR);
    }


    if (fps == 0.0) {
	switch (cfInfo->frameRateCode) {
	  case 1:
	    sec_per_frame = 1.0 / 23.976;
	    fps = 23.976;
	    break;
	  case 2:
	    sec_per_frame = 1.0 / 24.0;
	    fps = 24.0;
	    break;
	  case 3:
	    sec_per_frame = 1.0 / 25.0;
	    fps = 25.0;
	    break;
	  case 4:
	    sec_per_frame = 1.0 / 29.97;
	    fps = 29.97;
	    break;
	  case 5:
	    sec_per_frame = 1.0 / 30.0;
	    fps = 30.0;
	    break;
	  case 6:
	    sec_per_frame = 1.0 / 50.0;
	    fps = 50.0;
	    break;
	  case 7:
	    sec_per_frame = 1.0 / 59.94;
	    fps = 59.94;
	    break;
	  case 8:
	    sec_per_frame = 1.0 / 60.0;
	    fps = 60.0;
	    break;
	  default:
	    sec_per_frame = 1.0 / 30.0;
	    fps = 30.0;
	    break;
	}
    } else {
	sec_per_frame = 1.0 / fps;
    }

    frameNum = 0; 
    ref1 = -1;
    ref2 = -1;
    numFrameAlloced = 100;

    rawFOT = (int *) malloc(sizeof(int) * 2 * numFrameAlloced);
    rawFTT = (double *) malloc(sizeof(double) * (numFrameAlloced+1));
    rawRefTable = (RefEntry *) malloc(sizeof(RefEntry) * numFrameAlloced);
    
    next_start_code = NextStartCode(in_fd);

    while ((next_start_code != MPEG_END_SEQ_CODE) &&
	   (next_start_code != -1)) {
	
	if (next_start_code == MPEG_SYS_HDR_START_CODE) {
	    read(in_fd, readBuffer, 12);

	    if ((readBuffer[11] & 0x02)) {
		hdrRef++;

		numHeaders++;

		if (headerTable == NULL) {
		    headerTable = (HdrEntry *) malloc(sizeof(HdrEntry)*
						       numHeaders);
		} else {
		    headerTable = (HdrEntry *) realloc(headerTable,
						       sizeof(HdrEntry)*
						       numHeaders);
		}

		read(in_fd, &(readBuffer[12]), 64);
		
		headerTable[hdrRef].intra_matrix_flag = 1;
	      
		for(i=0; i<64; i++) {
		    next_value = ((readBuffer[11+i] & 0x01 * 128) + 
				  (readBuffer[12+i] >> 1));
		    headerTable[hdrRef].intra_matrix[i] = next_value;
		}

		if ((readBuffer[75] & 0x01)) {
		    headerTable[hdrRef].non_intra_matrix_flag = 1;

		    read(in_fd, &(readBuffer[76]), 64);

		    for (i=0; i<64; i++) {
			headerTable[hdrRef].non_intra_matrix[i] = 
			  readBuffer[76+i];
		    }
		} else {
		    headerTable[hdrRef].non_intra_matrix_flag = 0;
		}
	    } else if ((readBuffer[11] & 0x01)) {
		hdrRef++;

		numHeaders++;

		if (headerTable == NULL) {
		    headerTable = (HdrEntry *) malloc(sizeof(HdrEntry)*
						       numHeaders);
		} else {
		    headerTable = (HdrEntry *) realloc(headerTable,
						       sizeof(HdrEntry)*
						       numHeaders);
		}
	
		headerTable[hdrRef].intra_matrix_flag = 0;
		headerTable[hdrRef].non_intra_matrix_flag = 1;
		
		read(in_fd, &(readBuffer[12]), 64);

		for (i=0; i<64; i++) {
		    headerTable[hdrRef].non_intra_matrix[i] = readBuffer[12+i];
		} 
	    }

	    next_start_code = NextStartCode(in_fd);

	} else if (next_start_code == MPEG_PIC_START_CODE) {
	    
	    offset = lseek(in_fd, 0L, SEEK_CUR);

	    read(in_fd, readBuffer, 6);

	    picture_type = (readBuffer[5] & 0x38) >> 3;
	    
	    next_start_code = NextStartCode(in_fd);
	    if (next_start_code == -1) {
		break;
	    }
	    end_offset = lseek(in_fd, 0L, SEEK_CUR);

	    length = end_offset - offset;

	    if (frameNum >= numFrameAlloced) {
		numFrameAlloced += 100;

		rawFOT = (int *) realloc(rawFOT,
					 sizeof(int) * 2 * numFrameAlloced);
		rawFTT = (double *) realloc(rawFTT, 
					    sizeof(double) * 
					    (numFrameAlloced+1));
		rawRefTable = (RefEntry *) realloc(rawRefTable,
						   sizeof(RefEntry) * 
						   numFrameAlloced);
	    }
	    
	    rawFOT[frameNum*2] = offset;
	    rawFOT[(frameNum*2)+1] = length;
	    rawFTT[frameNum] = ((double) frameNum) * sec_per_frame;
	    rawRefTable[frameNum].type = picture_type;
	  
	    if (picture_type != MPEG_B_FRAME) {

		ref1 = ref2;
		ref2 = frameNum;

		if (ref1 > -1) {
		    /*
		     * Here's the really tricky part.
		     * Now that we've got that next I or P after some
		     * sequence of B's, we need to adjust the index numbers
		     * to reflect display order instead of decode order.
		     */
		    
		    CopyRefEntry(&tmpRefEntry, &(rawRefTable[ref1]));
		    tmpOffset = rawFOT[ref1*2];
		    tmpLength = rawFOT[(ref1*2)+1];
		    
		    for (i=ref1; i<ref2-1; i++) {
			CopyRefEntry(&(rawRefTable[i]), 
				     &(rawRefTable[i+1]));
			rawFOT[i*2] = rawFOT[(i+1)*2];
			rawFOT[(i*2)+1] = rawFOT[((i+1)*2)+1];
		    }
		    
		    CopyRefEntry(&(rawRefTable[ref2-1]), &tmpRefEntry);
		    rawFOT[(ref2-1)*2] = tmpOffset;
		    rawFOT[((ref2-1)*2)+1] = tmpLength;
				    
		    for (i=ref1; i<ref2; i++) {
			rawRefTable[i].ref2 = ref2-1;
		    }

		    ref1 = ref2-1;
		}
	    }

	    rawRefTable[frameNum].ref1 = ref1;
	    rawRefTable[frameNum].ref2 = ref2;
	    
	    rawRefTable[frameNum].headerRef = hdrRef;

	    frameNum++;
	} else {
	    lseek(in_fd, 4, SEEK_CUR);
	    next_start_code = NextStartCode(in_fd);
	}
    }
 
    if (frameNum <= 0) {
	fprintf(stderr, "Found 0 frames.\n");
	exit(1);
    }

    /* 
     * Do the final flipping for the last set of B frames. 
     */

    ref1 = ref2;
    ref2 = frameNum;

    if (ref1 > -1) {
	CopyRefEntry(&tmpRefEntry, &(rawRefTable[ref1]));
	tmpOffset = rawFOT[ref1*2];
	tmpLength = rawFOT[(ref1*2)+1];
		    
	for (i=ref1; i<ref2-1; i++) {
	    CopyRefEntry(&(rawRefTable[i]), 
			 &(rawRefTable[i+1]));
	    rawFOT[i*2] = rawFOT[(i+1)*2];
	    rawFOT[(i*2)+1] = rawFOT[((i+1)*2)+1];
	}
		    
	CopyRefEntry(&(rawRefTable[ref2-1]), &tmpRefEntry);
	rawFOT[(ref2-1)*2] = tmpOffset;
	rawFOT[((ref2-1)*2)+1] = tmpLength;
				    
	for (i=ref1; i<ref2; i++) {
	    rawRefTable[i].ref2 = ref2-1;
	}

	ref1 = ref2-1;
    }
    
    rawRefTable = (RefEntry *) realloc(rawRefTable,
					       sizeof(RefEntry) * frameNum);
    rawFTT = (double *) realloc(rawFTT,
					sizeof(double) * (frameNum+1));
    rawFOT = (int *) realloc(rawFOT,
				     sizeof(int) * 2 * frameNum);

    rawFTT[frameNum] = ((double) frameNum) * sec_per_frame;

    /*
     * Resolve I and P references now. 
     */

    lastI = -1;
    for (i=0; i<frameNum; i++) {
	if (rawRefTable[i].type == MPEG_I_FRAME) {
	    rawRefTable[i].ref1 = lastI;
	    lastI = i;
	}
    }

    nextI = -1;
    nextP = -1;

    for (i=frameNum-1; i>=0; i--) {
	if (rawRefTable[i].type == MPEG_I_FRAME) {
	    rawRefTable[i].ref2 = nextI;
	    nextI = i;
	    nextP = i;
	}
	else if (rawRefTable[i].type == MPEG_P_FRAME) {
	    rawRefTable[i].ref2 = nextP;
	    nextP = i;
	}
    }
	
    numFrames = frameNum;
    clipLength = rawFTT[frameNum];
    dataStart = 0;

    mi_header.media = htonl(VIDEO_MEDIA_CODE);
    mi_header.format = htonl(MPEG_FORMAT_CODE);
    mi_header.device = htonl(OTHER_DEV_CODE);
    mi_header.version = htonl(3);
    mi_header.num_frames = htonl(numFrames);
    mi_header.length_s = htonl(((int) (floor(clipLength))));
    mi_header.length_ns = htonl(((int) ((clipLength - floor(clipLength)) * 
					ONE_BILLION)));

    Chunk_WriteHeader(chunkState, CM_MEDIA_INDEP_CF, NULL);
    write(chunkState->fd, &mi_header, (sizeof(MediaIndepHeader)));

    cfInfo->width = htonl(cfInfo->width);
    cfInfo->height = htonl(cfInfo->height);
    cfInfo->aspectRatioCode = htonl(cfInfo->aspectRatioCode);
    cfInfo->frameRateCode = htonl(cfInfo->frameRateCode);
    cfInfo->bitRate = htonl(cfInfo->bitRate);
    cfInfo->vbvBufferSize = htonl(cfInfo->vbvBufferSize);
    cfInfo->constrainedParamFlag = htonl(cfInfo->constrainedParamFlag);

    Chunk_WriteHeader(chunkState, CM_MPEG_CF, NULL);
    write(chunkState->fd, cfInfo, (sizeof(MPEGHeader)));

    Chunk_WriteHeader(chunkState, CM_FTT, NULL);

    for (i=0; i<=numFrames; i++) {
	int time_s;
	int time_ns;

	time_s = htonl((int) (floor(rawFTT[i])));
	time_ns = htonl((int) ((rawFTT[i] - floor(rawFTT[i])) * ONE_BILLION));

	write(chunkState->fd, &time_s, 4);
	write(chunkState->fd, &time_ns, 4);
    }

    Chunk_WriteHeader(chunkState, CM_FOT, NULL);

    for (i=0; i<numFrames; i++) {
	int offset;
	int length;

	offset = htonl(rawFOT[i*2]);
	length = htonl(rawFOT[(i*2)+1]);

	write(chunkState->fd, &offset, 4);
	write(chunkState->fd, &length, 4);
    }

    if (numHeaders > 0) {
	
	Chunk_WriteHeader(chunkState, CM_MPEG_HDR_TABLE, NULL);

	for (i=0; i<numHeaders; i++) {
	    int intra_flag, non_intra_flag;
	    int next_value;

	    if(headerTable[i].intra_matrix_flag) {
		intra_flag = htonl(1);
		write(chunkState->fd, &intra_flag, 4);

		for (j=0; j<64; j++) {
		    next_value = htonl(headerTable[i].intra_matrix[j]);
		    write(chunkState->fd, &next_value, 4);
		}
	    } else {
		intra_flag = htonl(0);
		write(chunkState->fd, &intra_flag, 4);
	    }

	    if(headerTable[i].non_intra_matrix_flag) {
		non_intra_flag = htonl(1);
		write(chunkState->fd, &non_intra_flag, 4);

		for (j=0; j<64; j++) {
		    next_value = htonl(headerTable[i].non_intra_matrix[j]);
		    write(chunkState->fd, &next_value, 4);
		}
	    } else {
		non_intra_flag = htonl(0);
		write(chunkState->fd, &non_intra_flag, 4);
	    }
	}
    }

    Chunk_WriteHeader(chunkState, CM_MPEG_REF_TABLE, NULL);
    
    for (i=0; i<numFrames; i++) {
	RefEntry writeRefEntry;

	writeRefEntry.type = htonl(rawRefTable[i].type);
	writeRefEntry.ref1 = htonl(rawRefTable[i].ref1);
	writeRefEntry.ref2 = htonl(rawRefTable[i].ref2);
	writeRefEntry.headerRef = htonl(rawRefTable[i].headerRef);
	
	write(chunkState->fd, &writeRefEntry, 16);
    }

    Chunk_WriteHeader(chunkState, CM_DATA_SEGMENT, NULL);

    lseek(in_fd, 0, SEEK_SET);

    num_read = read(in_fd, readBuffer, RD_BUF_SIZE);

    while (num_read > 0) {
	write(chunkState->fd, readBuffer, num_read);
	num_read = read(in_fd, readBuffer, RD_BUF_SIZE);
    }

    Chunk_CloseFile(chunkState);   
}


