/*
 * Copyright (c) 1995 Cornell University
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for research and educational purposes, without fee, and
 * without written agreement is hereby granted, provided that the above
 * copyright notice and the following two paragraphs appear in all copies
 * of this software.
 *
 * IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF CORNELL UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 */

/*
	Netvideo version 3.3
	Written by Ron Frederick <frederick@parc.xerox.com>
*/

/*
 * Copyright (c) Xerox Corporation 1992. All rights reserved.
 *  
 * License is granted to copy, to use, and to make and to use derivative
 * works for research and evaluation purposes, provided that Xerox is
 * acknowledged in all documentation pertaining to any such copy or derivative
 * work. Xerox grants no other licenses expressed or implied. The Xerox trade
 * name should not be used in any advertising without its written permission.
 *  
 * XEROX CORPORATION MAKES NO REPRESENTATIONS CONCERNING EITHER THE
 * MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS SOFTWARE
 * FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" without
 * express or implied warranty of any kind.
 *  
 * These notices must be retained in any copies of any part of this software.
 */

#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <memory.h>
#include <math.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <X11/Xlib.h>
#include "cmChunk.h"
#include "cmCookies.h"
#include "sized_types.h"
#include "rtp.h"
#include "vid_image.h"
#include "vid_code.h"
#include "nv.h"

#define MAX_ICON_WIDTH	96
#define MAX_ICON_HEIGHT	72

#define MAX_LOSS	50

#define MAX_NAMELEN	64
#define MAX_PACKLEN	8192

#define MAX_SOURCES	32

#define GRAB_ERROR_TIME	100	/* msec */
#define IDLE_POLL_TIME	1000	/* msec */

#define MAX_FLUSH_TIME	1	/* sec */
#define NAME_XMIT_TIME	5	/* sec */
#define MAX_IDLE_TIME	15	/* sec */

extern int gettimeofday(struct timeval *tp, struct timezone *tzp);
extern int write(int fd, void *buf, int nbyte);
/*extern int fread(void *ptr, int size, int nitems, FILE *stream);*/
/*extern int printf(char *format, ...);*/
extern long tell(int fd);
extern int fseek(FILE *fp, long offset, int ptrname);

#ifdef NO_STDLIB_H
extern int atoi(char *str);
#endif

extern double floor(double x);
extern int fclose(FILE *fp);

int color_ok;

static struct sockaddr_in myaddr;
static uint32 last_namexmit;
static char myname[MAX_NAMELEN];
static uint8 chanid;

static char *rtpcont_name[] =
    { "", "", "", "", "", "", "", "", "", "", "", "", "",
      "", "", "", "", "", "", "", "", "", "", "", "",
      "CellB", "JPEG", "CU-SeeMe", "nv", "PicWin", "CPV", "H.261" };

typedef struct {
    struct sockaddr_in	srcaddr;
    uint16		ssrc;
    uint8		content;
    uint8		active;
    uint8		flushed;
    char		name[MAX_NAMELEN];
    struct timeval	lastrecv;
    uint32		lastflush;
    uint32		laststamp;
    vidimage_t		*image;
    uint16		startseq[8], maxseq[8];
    uint32		pkts[8], frames[8], shownframes[8], time[8], bytes[8];
    uint32		totseq, totpkts, totframes, totshown, tottime, totbytes;
    int			avgpos;
} source_t;

typedef struct {
    int size;
    int offset;
    NVFrameOffset *offsetTable;
} OffsetTableCtl;

typedef struct {
    int sec;
    int nsec;
} TimeTable;

typedef struct {
    int size;
    int firstSec;
    int firstNsec;
    TimeTable *timeTable;
} TimeTableCtl;

static int max_source=0;
static source_t source[MAX_SOURCES];

char *display=NULL;

static 
int MatchNetAddr(struct sockaddr_in *addr1, struct sockaddr_in *addr2)
{
    return ((addr1->sin_addr.s_addr == addr2->sin_addr.s_addr) &&
	    (addr1->sin_port == addr2->sin_port));
}

static 
void FillRTPHeader(struct rtphdr *hdrp, int chanid, int opts, int sync,
			  int content, uint16 seq, uint32 timestamp)
{
    hdrp->rh_vers = RTP_VERSION;
    hdrp->rh_chanid = chanid;
    hdrp->rh_opts = opts;
    hdrp->rh_sync = sync;
    hdrp->rh_content = content;
    hdrp->rh_seq = htons(seq++);
    hdrp->rh_ts = htonl(timestamp);
}

static 
int FillSDESCOption(struct rtcpsdeschdr *opt, int fin,
			   struct sockaddr_in *addr, char *name)
{
    char *p=(char *) (opt+1);
    int optlen=(sizeof(*opt)+strlen(name))/4+1;

    opt->rtsh_fin = fin;
    opt->rtsh_type = RTPOPT_SDESC;
    opt->rtsh_optlen = optlen;
    opt->rtsh_id = 0;
    strcpy(p, name);

    return optlen*4;
}

static 
int NewSource(struct sockaddr_in *addr, uint16 ssrc, int color,
		     int width, int height, int content)
{
    int i;
 
    if (MatchNetAddr(addr, &myaddr)) {
	i = 0;
    } else {
	for (i=1; i<=max_source; i++)
	    if (source[i].image == NULL) break;
    }
 
    if (i > max_source) max_source = i;
 
    memset((char *)&source[i], 0, sizeof(source[i]));
    source[i].srcaddr = *addr;
    source[i].ssrc = ssrc;
    source[i].content = content;
    sprintf(source[i].name, "%d/%d (%s)",
	addr->sin_port, ssrc, rtpcont_name[content]);

    source[i].image = VidImage_Create(color, width, height);

    return i;
}

static 
int ParseRTPOptions(uint8 *packet, uint16 *ssrcp, char **namep)
{
    int done;
    struct rtphdr *rtp=(struct rtphdr *)packet;
    struct rtpopthdr *opt=(struct rtpopthdr *)(rtp+1);
    struct rtpssrchdr *ssrcopt;
    struct rtcpsdeschdr *sopt;

    *ssrcp = 0;
    *namep = NULL;
    done = !rtp->rh_opts;
    while (!done) {
	if (opt->roh_optlen == 0) break;

	switch (opt->roh_type) {
	case RTPOPT_SSRC:
	    ssrcopt = (struct rtpssrchdr *) opt;
	    *ssrcp = ssrcopt->rsh_id;
	    break;
	case RTPOPT_SDESC:
	    sopt = (struct rtcpsdeschdr *) opt;
	    *namep = (char *) (sopt+1);
	    break;
	}

	done = opt->roh_fin;
	opt = (struct rtpopthdr *) (((uint32 *)opt)+opt->roh_optlen);
    }

    return ((uint8 *)opt)-packet;
}

static
void PrepVideoData(struct sockaddr_in *srcaddrp, uint8 *packet, int len)
{
    int optlen;
    struct rtphdr *rtp;
    uint16 ssrc;
    char *name;
    uint8 *data;
 
    rtp = (struct rtphdr *) packet;

    optlen = ParseRTPOptions(packet, &ssrc, &name);
    data = packet+optlen;
    len -= optlen;

    if (len > 0) (void) NV_Prep(source[1].image, data, len);

}

static
void RecvVideoData(struct sockaddr_in *srcaddrp, uint8 *packet, int len)
{
    int i, j, color, iscolor, width, height, content, avgpos;
    int newframe=0, shownframe=0, optlen;
    struct rtphdr *rtp;
    uint16 ssrc;
    char *name;
    uint8 *data;
    struct timeval t;
    int16 seqdiff;
    int32 tsdiff, tdiff;
    decodeproc_t *decode;
 
    rtp = (struct rtphdr *) packet;

    optlen = ParseRTPOptions(packet, &ssrc, &name);
    data = packet+optlen;
    len -= optlen;

    content = rtp->rh_content;
    switch (content) {
    case RTPCONT_NV:
	color = 1;
	width = NTSC_WIDTH;
	height = NTSC_HEIGHT;
	decode = NV_Decode;
	break;
    default:
	/* Unknown packet content type -- discard */
	return;
    }

    for (i=0, j=0; i<=max_source; i++) {
	if (source[i].image == NULL) continue;
	if ((MatchNetAddr(srcaddrp, &source[i].srcaddr)) &&
	    (ssrc == source[i].ssrc)) break;
	if ((j == 0)  &&
	    (srcaddrp->sin_addr.s_addr == source[i].srcaddr.sin_addr.s_addr) &&
	    (source[i].lastrecv.tv_sec == 0)) j = i;
    }

    rtp->rh_seq = ntohs(rtp->rh_seq);
    rtp->rh_ts = ntohl(rtp->rh_ts);

    if (i > max_source) {
	if (j != 0) {
	    i = j;
	    source[i].srcaddr = *srcaddrp;
	    source[i].ssrc = ssrc;
	} else {
	    i = NewSource(srcaddrp, ssrc, color, width, height,
		rtp->rh_content);
	    source[i].active = 0;
	}
	source[i].flushed = 1;
	source[i].lastflush = 0;
	source[i].laststamp = rtp->rh_ts;
    } else if (source[i].content != content) {
	source[i].content = content;
	iscolor = ((source[i].image->flags & VIDIMAGE_ISCOLOR) != 0);
	if (color != iscolor)
	    VidImage_SetColor(source[i].image, color,
			      source[i].image->flags & VIDIMAGE_WANTCOLOR);
    }

    gettimeofday(&t, NULL);
    tdiff = (t.tv_sec-source[i].lastrecv.tv_sec)*1000000 +
	t.tv_usec-source[i].lastrecv.tv_usec;
    source[i].lastrecv = t;

    if (source[i].image == NULL) return;

    tsdiff = rtp->rh_ts - source[i].laststamp;
    if (tsdiff < 0) {
	/*printf("Source %d: Old timestamp: expected %d, got %d\n", i,
	    source[i].laststamp, rtp->rh_ts);*/
	return;
    } else if (tsdiff > 0) {
	newframe = 1;
	source[i].laststamp = rtp->rh_ts;
	if (((uint32)(t.tv_sec-source[i].lastflush) >= MAX_FLUSH_TIME)  &&
	    !source[i].flushed) {
	    shownframe++;
	    source[i].flushed = 1;
	    source[i].lastflush = t.tv_sec;
	}
    }
    if (len > 0) (void) decode(source[i].image, data, len);

    if (rtp->rh_sync) {
	if ((uint32)(t.tv_sec-source[i].lastflush) >= MAX_FLUSH_TIME) {
	    shownframe++;
	    source[i].flushed = 1;
	    source[i].lastflush = t.tv_sec;
	}
    } else {
	source[i].flushed = 0;
    }

    avgpos = source[i].avgpos;
    seqdiff = rtp->rh_seq - source[i].startseq[avgpos];
    if ((seqdiff > MAX_LOSS) || (seqdiff < -MAX_LOSS)) {
	source[i].startseq[avgpos] = rtp->rh_seq;
	source[i].maxseq[avgpos] = rtp->rh_seq-1;
	source[i].pkts[avgpos] = 0;
	seqdiff = 0;
    }

    if (seqdiff >= 0) source[i].pkts[avgpos]++;
    if ((int16) (rtp->rh_seq - source[i].maxseq[avgpos]) > 0)
	source[i].maxseq[avgpos] = rtp->rh_seq;
    source[i].frames[avgpos] += newframe;
    source[i].shownframes[avgpos] += shownframe;
    source[i].time[avgpos] += tdiff;
    source[i].bytes[avgpos] += len;

    if (source[i].time[avgpos] >= 500000) {
	uint16 maxseq = source[i].maxseq[avgpos];

	source[i].totseq += maxseq-source[i].startseq[avgpos]+1;
	source[i].totpkts += source[i].pkts[avgpos];
	source[i].totframes += source[i].frames[avgpos];
	source[i].totshown += source[i].shownframes[avgpos];
	source[i].tottime += source[i].time[avgpos];
	source[i].totbytes += source[i].bytes[avgpos];
	source[i].avgpos = avgpos = (avgpos+1) % 8;

	if (source[i].time[avgpos] != 0) {
	    source[i].totseq -=
		(source[i].maxseq[avgpos]-source[i].startseq[avgpos]+1);
	    source[i].totpkts -= source[i].pkts[avgpos];
	    source[i].totframes -= source[i].frames[avgpos];
	    source[i].totshown -= source[i].shownframes[avgpos];
	    source[i].tottime -= source[i].time[avgpos];
	    source[i].totbytes -= source[i].bytes[avgpos];
	}

	source[i].startseq[avgpos] = maxseq+1;
	source[i].maxseq[avgpos] = maxseq;
	source[i].pkts[avgpos] = 0;
	source[i].frames[avgpos] = 0;
	source[i].shownframes[avgpos] = 0;
	source[i].time[avgpos] = 0;
	source[i].bytes[avgpos] = 0;
    }
}

static 
void SaveVideoData(int fd, int sync, int content, int timestamp,
			  uint8 *hdr, int hdrlen, uint8 *data, int datalen)
{
    int opts=0, optlen=0, len;
    static uint32 packet[MAX_PACKLEN/4];
    struct rtphdr *rtp;
    struct rtcpsdeschdr *opthdr;
    uint8 *p;
    struct timeval t;
    static uint16 seq=1;
    u_short slen;
    u_char sid;

    gettimeofday(&t, NULL);
    if (t.tv_sec-last_namexmit > NAME_XMIT_TIME) opts=1;

    rtp = (struct rtphdr *)packet;
    FillRTPHeader(rtp, chanid, opts, sync, content, seq++, timestamp);

    p = (uint8 *)(rtp+1);
    if (opts) {
	opthdr = (struct rtcpsdeschdr *)p;
        optlen = FillSDESCOption(opthdr, 1, &myaddr, myname);
	p += optlen;
	last_namexmit = t.tv_sec;
    }

    if (hdrlen > 0) memcpy(p, hdr, hdrlen);
    if (datalen > 0) memcpy(p+hdrlen, data, datalen);
    len = sizeof(struct rtphdr)+optlen+hdrlen+datalen;

    slen = len+1;
    sid = 0;
    write(fd,&slen,2);
    write(fd,&sid,1);
    write(fd,&sid,1);
    /* set quad alignment for read as integer */
    if (len%4) len = (len+3)/4*4;
    write(fd,packet,len);
}

static
int recvFromFile(FILE *fp, unsigned char *packet)
{
	u_short	len;
	u_char	sid=1;

	while (sid != 0) {
		if (0 >= (int)fread(&len, 1, 2, fp)) return 0;
		if (0 >= (int)fread(&sid, 1, 1, fp)) return 0;
		if (sid != 0) printf("sid = %d\n",sid);
		if (0 >= (int)fread(packet, len-1, 1, fp)) return 0;
	}
	return	len;
}

static
int saveToFile(int fd, unsigned char *packet, int len)
{
	u_short	slen=len;
	u_char	sid=0;

        /* set quad alignment for read as integer */
        if ((len-1)%4) len = ((len-1)/4 + 1)*4+1;
	write(fd, &slen, 2);
	write(fd, &sid, 1);
	write(fd, &sid, 1);
	return write(fd, packet, len-1);
}

static
void putOffsetTable(OffsetTableCtl *otc, int frameNum, NVFrameOffset *offset)
{
    if (otc->size <= frameNum) {
        otc->offsetTable = (NVFrameOffset *)realloc(otc->offsetTable, 
                                                    sizeof(NVFrameOffset)*otc->size*2);
        otc->size = otc->size *2;
    }

    otc->offsetTable[frameNum].iExist = offset->iExist;
    otc->offsetTable[frameNum].fdOffset = offset->fdOffset - otc->offset;
    otc->offsetTable[frameNum].fdLength = offset->fdLength - offset->fdOffset;
    otc->offsetTable[frameNum].fdPacketNum = offset->fdPacketNum;
    otc->offsetTable[frameNum].bdOffset = offset->bdOffset - otc->offset;
    otc->offsetTable[frameNum].bdLength = offset->bdLength - offset->bdOffset;
    otc->offsetTable[frameNum].bdPacketNum = offset->bdPacketNum;
    otc->offsetTable[frameNum].iOffset = offset->iOffset - otc->offset;
    otc->offsetTable[frameNum].iLength = offset->iLength - offset->iOffset;
    otc->offsetTable[frameNum].iPacketNum = offset->iPacketNum;
    otc->offsetTable[frameNum].direction = offset->direction;
    otc->offsetTable[frameNum].iFrameNum = offset->iFrameNum;
}

static
void putTimeTable(TimeTableCtl *ttc, int frameNum, int timeStamp)
{
    int  sec,nsec;

    if (ttc->size <= frameNum) {
        ttc->timeTable = (TimeTable *)realloc(ttc->timeTable, 
                                              sizeof(TimeTable)*ttc->size*2);
        ttc->size = ttc->size *2;
    }

    sec = timeStamp >> 16;
    nsec = (((timeStamp&0xffff)*15625) >> 10)* 1000;

    if (frameNum == 0) {
        ttc->firstSec = sec;
        ttc->firstNsec = nsec;
        ttc->timeTable[0].sec = 0;
        ttc->timeTable[0].nsec = 0;
    } else {
        if (nsec < ttc->firstNsec) {
            nsec = 1000000000 - ttc->firstNsec + nsec;
            sec --;
        } else {
            nsec = nsec - ttc->firstNsec;
        }
        sec = sec - ttc->firstSec;

        ttc->timeTable[frameNum].sec = sec;
        ttc->timeTable[frameNum].nsec = nsec;
    }
}

static
int writeDataSegment(FILE *fp, int fd, int interval, 
                     TimeTableCtl *ttc, OffsetTableCtl *otc)
{
	static	uint32	packet[MAX_PACKLEN/4];
	struct sockaddr_in srcaddr;
	int		len;
	struct rtphdr	*rtp;
	long		offset;
        int             frameCnt;
        int             iFlag;
        NVFrameOffset   nvfo;

	/* malloc xmit matrix */
	NV_Encode_Reconfig(1,NTSC_WIDTH,NTSC_HEIGHT);

        /* write a interval number */
        write(fd, &interval, sizeof(int));

        frameCnt = 0;

        while (1) {
            /* set whole image flag */
            if ((frameCnt % interval) == 0) iFlag = 1;
            else iFlag = 0;

            /* set iExist of nvfo */
            nvfo.iExist = iFlag;

            /* write whole image flag */
            write(fd, &iFlag, sizeof(int));

            /* set fdOffset of nvfo */
            nvfo.fdOffset = tell(fd);
            nvfo.fdPacketNum = 0;

            /* write a forward diff of 1 frame and update image */
            while ((len = recvFromFile(fp,(unsigned char *)packet))>0) {
                saveToFile(fd,(unsigned char *)packet,len);
                nvfo.fdPacketNum ++;
                RecvVideoData(&srcaddr,(uint8 *)packet, len);
                rtp = (struct rtphdr *)packet;
                putTimeTable(ttc, frameCnt, rtp->rh_ts);
                if (rtp->rh_sync) break;
            }
            if (len <= 0) break;

            /* set fdLength of nvfo */
            nvfo.fdLength = tell(fd);
            nvfo.bdOffset = tell(fd);
            nvfo.bdPacketNum = 0;

            /* check off macro blocks to be updated next frame */
            offset = ftell(fp);
            while ((len = recvFromFile(fp,(unsigned char *)packet))>0) {
                PrepVideoData(&srcaddr,(uint8 *)packet, len);
                rtp = (struct rtphdr *)packet;
                if (rtp->rh_sync) {
                    fseek(fp,offset,0);
                    break;
                }
            }
            /* trimming last frame */
            if (len <= 0) break;

            /* write a backward diff of 1 frame */
            NV_Save(fd, source[1].image, SaveVideoData);

            /* set bdLength of nvfo */
            nvfo.bdLength = tell(fd);
            nvfo.iOffset = tell(fd);
            nvfo.iPacketNum = 0;
            
            /* write a whole image if it's interval frame */
            if ((frameCnt % interval) == 0) {
                NV_SetXmitAll();
                NV_Save(fd, source[1].image, SaveVideoData);
            }

            /* set bdLength of nvfo */
            nvfo.iLength = tell(fd);

            /* put offset table */
            putOffsetTable(otc, frameCnt, &nvfo);

            frameCnt ++;
        }

        /* write a backward diff of last frame */
        NV_SetXmitAll();
        NV_Save(fd, source[1].image, SaveVideoData);

        /* set bdLength of nvfo */
        nvfo.bdLength = tell(fd);
        nvfo.iOffset = tell(fd);
        nvfo.iPacketNum = 0;

        /* write a whole image of last frame if it's interval frame */
        if ((frameCnt % interval) == 0) {
            NV_SetXmitAll();
            NV_Save(fd, source[1].image, SaveVideoData);
        }

        /* set bdLength of nvfo */
        nvfo.iLength = tell(fd);

        /* put offset table */
        putOffsetTable(otc, frameCnt, &nvfo);

        frameCnt ++;

        return frameCnt;
}

int main(int argc, char *argv[])
{
    FILE *fp;  /* nv data file */
    tChunkFileState *cfState;  /* chunk file state */
    OffsetTableCtl otc;  /* offset table control block */
    TimeTableCtl ttc;  /* time table control block */
    double length;
    int length_s, length_ns;
    int numFrames;
    MediaIndepHeader mi_header;
    NVHeader nv_header;
    int sec, nsec;
    int interval;
    int iBoundary;
    int ifnum;
    int i;

    if (argc < 4) exit(0);

    if (NULL == (fp = fopen(argv[1],"r"))) {
        printf("nv file open error %s\n",argv[1]);
        exit(1);
    }
    if (NULL == (cfState = Chunk_OpenFile(argv[2],"wc"))) {
        printf("chunk file open error %s\n",argv[2]);
        exit(1);
    }

    /* initialize time table */
    ttc.size = 16;
    if (NULL == (ttc.timeTable = (TimeTable *)malloc(sizeof(TimeTable)*16))) {
        printf("ERROR malloc time table\n");
        exit(1);
    }

    /* initialize offset table */
    otc.size = 16;
    if (NULL == (otc.offsetTable = 
                 (NVFrameOffset *)malloc(sizeof(NVFrameOffset)*16))) {
        printf("ERROR malloc offset table\n");
        exit(1);
    }

    /* write deta segment header */
    if (Chunk_WriteHeader(cfState, CM_DATA_SEGMENT, "") != 0) {
        printf("ERROR writing chunk header of data segment\n");
        exit(1);
    }

    /* initialize offset table */
    otc.offset = tell(cfState->fd);

    /* write data segment and set up time table */
    numFrames = writeDataSegment(fp,cfState->fd,atoi(argv[3]),&ttc,&otc);

    /* calc time length */
    length = (double)(ttc.timeTable[numFrames-1].nsec)/1000000000 
        + ttc.timeTable[numFrames-1].sec;
    length = length*(numFrames+1)/numFrames;

    length_s = floor(length);
    length_ns = (length - length_s)*1000000000;

    if (ttc.size <= numFrames) {
        ttc.timeTable = (TimeTable *)realloc(ttc.timeTable, 
                                             sizeof(TimeTable)*(ttc.size+1));
        ttc.size ++;
    }

    ttc.timeTable[numFrames].sec = length_s;
    ttc.timeTable[numFrames].nsec = length_ns;

    /* write media independent header */
    if (Chunk_WriteHeader(cfState, CM_MEDIA_INDEP_CF, "") != 0) {
        printf("ERROR writing chunk header of media independent header\n");
        exit(1);
    }

    mi_header.media = htonl(VIDEO_MEDIA_CODE);
    mi_header.format = htonl(NV_FORMAT_CODE);
    mi_header.device = htonl(0);
    mi_header.version = htonl(3);
    mi_header.num_frames = htonl(numFrames);
    mi_header.length_s = htonl(length_s);
    mi_header.length_ns = htonl(length_ns);

    write(cfState->fd, &mi_header, sizeof(MediaIndepHeader));

    /* write media dependent header */
    if (Chunk_WriteHeader(cfState, CM_NV_CF, "") != 0) {
        printf("ERROR writing chunk header of nv media dependent header\n");
        exit(1);
    }
    
    nv_header.width = htonl(320);
    nv_header.height = htonl(240);
    nv_header.device = htonl(0);
    nv_header.subformat = htonl(0);

    write(cfState->fd, &nv_header, sizeof(NVHeader)); 

    /* write frame time table */
    if (Chunk_WriteHeader(cfState, CM_FTT, "") != 0) {
        printf("ERROR writing chunk header of frame time table\n");
        exit(1);
    }

    for (i=0; i<= numFrames; i++) {
        sec = htonl(ttc.timeTable[i].sec);
        nsec = htonl(ttc.timeTable[i].nsec);
        write(cfState->fd, &sec, sizeof(int));
        write(cfState->fd, &nsec, sizeof(int));
    }

    /* calc whole image frame direction and frame number, offset, length */
    interval = atoi(argv[3]);
    iBoundary = interval/2;
    for (i=0; i < numFrames; i++) {
        ifnum = (i+iBoundary)/interval*interval;
        if (ifnum >= numFrames) ifnum -= interval;
        otc.offsetTable[i].direction = i-ifnum;
        otc.offsetTable[i].iFrameNum = ifnum;
        otc.offsetTable[i].iOffset = otc.offsetTable[ifnum].iOffset;
        otc.offsetTable[i].iLength = otc.offsetTable[ifnum].iLength;
        otc.offsetTable[i].iPacketNum = otc.offsetTable[ifnum].iPacketNum;
    }

    /* write frame offset table */
    if (Chunk_WriteHeader(cfState, CM_FOT, "") != 0) {
        printf("ERROR writing chunk header of frame offset table\n");
        exit(1);
    }
    
    for (i=0; i< numFrames; i++) {
        otc.offsetTable[i].iExist = htonl(otc.offsetTable[i].iExist);
        otc.offsetTable[i].fdOffset = htonl(otc.offsetTable[i].fdOffset);
        otc.offsetTable[i].fdLength = htonl(otc.offsetTable[i].fdLength);
        otc.offsetTable[i].fdPacketNum = htonl(otc.offsetTable[i].fdPacketNum);
        otc.offsetTable[i].bdOffset = htonl(otc.offsetTable[i].bdOffset);
        otc.offsetTable[i].bdLength = htonl(otc.offsetTable[i].bdLength);
        otc.offsetTable[i].bdPacketNum = htonl(otc.offsetTable[i].bdPacketNum);
        otc.offsetTable[i].iOffset = htonl(otc.offsetTable[i].iOffset);
        otc.offsetTable[i].iLength = htonl(otc.offsetTable[i].iLength);
        otc.offsetTable[i].iPacketNum = htonl(otc.offsetTable[i].iPacketNum);
        otc.offsetTable[i].direction = htonl(otc.offsetTable[i].direction);
        otc.offsetTable[i].iFrameNum = htonl(otc.offsetTable[i].iFrameNum);
        write(cfState->fd, &(otc.offsetTable[i]), sizeof(NVFrameOffset));
    }

    /* close files */
    fclose(fp);
    Chunk_CloseFile(cfState);

    /* free allocated memory */ 
    free(ttc.timeTable);
    free(otc.offsetTable);

    exit(0);
    /*NOTREACHED*/
}

