/*
 * Copyright (c) 1992 Anders Klemets <klemets@sics.se>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms, etc.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dirent.h>
#include "nv.h"

/* each packet on disk is preceeded by 16 bit unsigned integer.  The first
   12 bits contain the length of the packet that follows.  The remaining
   bits are defined below.
 */
#define	LEN_MASK	0x0fff
#define	LEN_TIMESTAMP	0x1000	/* packet is a timestamp in milliseconds */

#define timediffinms(tp1,tp2)	(((tp1)->tv_sec - (tp2)->tv_sec) * 1000 + ((tp1)->tv_usec - (tp2)->tv_usec) / 1000)

/* defaults */
char *fileprefix = "video";
char *directory = ".";
long playoffset = 0;

FILE *fp;
struct src_info {
  char *id;
  u_char used;
};
#define NSID_MAX	64
struct src_info *sidtable[NSID_MAX];	/* source id table */

main(argc, argv)
int argc;
char **argv;
{
  parsearguments(argc, argv);
  dispchosenparams();
  sendpackets();
}

parsearguments(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  int c;
  while ((c = getopt(argc, argv, "f:d:o:")) != -1) {
    switch (c) {
    case 'f':
      fileprefix = optarg;
      break;
    case 'd':
      directory = optarg;
      break;
    case 'o':
      playoffset = atoi(optarg) * 1000;
      break;
    default:
      fprintf(stderr,"Usage:\t[-d <directory>] [-f <filename prefix>] [-o <time offset>]\n");
      exit(1);
    }
  }
}

dispchosenparams()
{
  printf("Reading video from %s/%s.*\n", directory, fileprefix);
  if (playoffset)
    	printf("Skipping first %d seconds of recording.\n", playoffset/1000);
}

#define min(a,b)	((a) < (b) ? (a) : (b))

sendpackets()
{
  long timestp, startstp, delay, lastid;
  struct timeval tim, offset;
  u_short len;
  u_char sid;
  char buf[1500];
  int cnt = 0;
  nv_hdr_t *nv;

  offset.tv_sec = 0;
  while (open_file() > 0) {
    while (fread (&len, 1, 2, fp) > 0) {
      if(fread (buf, len & LEN_MASK, 1, fp) != 1)
	break;
      if (len & LEN_TIMESTAMP) {
	gettimeofday(&tim, NULL);
	timestp = *(long *) buf;
	if(timestp == 0 || offset.tv_sec == 0) { /* start of new recording */
	  startstp = timestp;
	  offset = tim;
	  lastid = 0;
	}
	continue;
      }
      if (timestp - startstp - playoffset < 0) 
	continue;			/* skip forward until desired time */
      if((sid = buf[0]) > NSID_MAX) {
	fprintf(stderr,"invalid sid in data\n");
	continue;	/* ignore too large a speaker id */
      }
      update_speaker(sid);
      nv = (nv_hdr_t *) (buf + 1);
      if (nv->type == NV_ID) {
	  if (sidtable[sid]->id != NULL && !strcmp(sidtable[sid]->id, nv + 1))
	    ;
	  else {
	    if (sidtable[sid]->id != NULL)
	      free(sidtable[sid]->id);
	    sidtable[sid]->id = (char *) strdup(nv + 1);
	    printf("session %d: %s\n",sid, sidtable[sid]->id);
	  }
	}
      ++cnt;
      /* garbage collect sockets every 30 seconds */
      if (timestp - startstp - playoffset - lastid >= 30000) {
	garbagesockets();
	lastid = timestp - startstp - playoffset;
      }
    }
    printf("read %d pkts\n",cnt);
    fclose(fp);
  }
}


int open_file()
{
  DIR *dirp;
  struct dirent *dp;
  char buf[512];
  unsigned int mint = ~0, minu;
  static unsigned int oldmint = 0, oldminu = 0;
  unsigned int m1, um1;

  if ((dirp = opendir(directory)) == NULL) {
    perror("opendir");
    exit(1);
  }
  while ((dp = readdir(dirp)) != NULL) {
    if(!strncmp(dp->d_name, fileprefix, strlen(fileprefix))) {
      sscanf(dp->d_name + strlen(fileprefix), ".%x.%x", &m1, &um1);
      if (m1 && (m1 < mint) && (m1 > oldmint)) {
	mint = m1;
	minu = um1;
      }
      if (m1 && m1 == mint)
	if ((um1 < minu) && (um1 > oldminu))
	  minu = um1;
    }
  }
  closedir(dirp);
  if (mint == ~0)
    return 0;
  sprintf(buf, "%s/%s.%x.%x", directory, fileprefix, mint, minu);
  if((fp = fopen(buf, "r")) == NULL) {
    perror("fopen");
    exit(1);
  }
  oldmint = mint;
  oldminu = minu;
  return 1;
}

update_speaker(sid)	/* mark source as active */
u_char sid;
{
  if (sidtable[sid] == NULL) {
    sidtable[sid] = (struct src_info *) malloc(sizeof(struct src_info));
    sidtable[sid]->id = NULL;
    sidtable[sid]->used = 0;
  }
  sidtable[sid]->used++;
}

garbagesockets()		/* garbage collect open sockets */
{
  int i;
  for (i=0; i < NSID_MAX; ++i)
    if(sidtable[i] != NULL && sidtable[i]->used == 0) {
      if(sidtable[i]->id)
	free(sidtable[i]->id);
      free(sidtable[i]);
      sidtable[i] = NULL;
    }
  for (i=0; i < NSID_MAX; ++i)
    if(sidtable[i] != NULL)
      sidtable[i]->used = 0;
}
