/*
 * Copyright (c) 1992 Anders Klemets <klemets@sics.se>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms, etc.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dirent.h>
#include "nv.h"

/* each packet on disk is preceeded by 16 bit unsigned integer.  The first
   12 bits contain the length of the packet that follows.  The remaining
   bits are defined below.
 */
#define	LEN_MASK	0x0fff
#define	LEN_TIMESTAMP	0x1000	/* packet is a timestamp in milliseconds */

#define timediffinms(tp1,tp2)	(((tp1)->tv_sec - (tp2)->tv_sec) * 1000 + ((tp1)->tv_usec - (tp2)->tv_usec) / 1000)

/* defaults */
char *fileprefix = "video";
char *outfileprefix = "video.id";
char *directory = ".";
char *outdirectory = ".";
int sessionid = 0;
long playoffset = 0;

FILE *fp, *outfp;
struct src_info {
  int ok;
  u_char used;
};
#define NSID_MAX	64
struct src_info sidtable[NSID_MAX];	/* source id table */

main(argc, argv)
int argc;
char **argv;
{
  parsearguments(argc, argv);
  dispchosenparams();
  sendpackets();
}

parsearguments(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  int c;
  while ((c = getopt(argc, argv, "i:f:g:d:e:o:")) != -1) {
    switch (c) {
    case 'i':
      sessionid = atoi(optarg);
      if (sessionid >= 0 && sessionid < NSID_MAX)
	sidtable[sessionid].ok = 1;
      else {
	fprintf(stderr,"session id out of range.\n");
	exit(1);
      }
      break;
    case 'f':
      fileprefix = optarg;
      break;
    case 'g':
      outfileprefix = optarg;
      break;
    case 'd':
      directory = optarg;
      break;
    case 'e':
      outdirectory = optarg;
      break;
    case 'o':
      playoffset = atoi(optarg) * 1000;
      break;
    default:
      fprintf(stderr,"Usage:\t[-i <session id>]\n\t[-d <in directory>] [-f <in filename prefix>] [-o <time offset>]\n\t[-e <out directory>] [-g <out filename prefix>]\n");
      exit(1);
    }
  }
  sidtable[sessionid].ok = 1;	/* default session is 0 */
}

dispchosenparams()
{
  int i;
  printf("Reading session #");
  for (i=0; i < NSID_MAX; ++i)
    if (sidtable[i].ok)
      printf("%d ",i);
  printf("video from %s/%s.*\n", directory, fileprefix);
  if (playoffset)
    	printf("Skipping first %d seconds of recording.\n", playoffset/1000);
  printf("Writing video to %s/%s%d.*\n", outdirectory, outfileprefix,
	 sessionid);
}

#define min(a,b)	((a) < (b) ? (a) : (b))

sendpackets()
{
  long timestp, startstp, delay;
  struct timeval tim, offset;
  u_short len;
  u_char sid;
  char buf[1500];
  int cnt = 0;
  nv_hdr_t *nv;

  offset.tv_sec = 0;
  while (open_file() > 0) {
    while (fread (&len, 1, 2, fp) > 0) {
      if(fread (buf, len & LEN_MASK, 1, fp) != 1)
	break;
      if (len & LEN_TIMESTAMP) {
	gettimeofday(&tim, NULL);
	timestp = *(long *) buf;
	if(timestp == 0 || offset.tv_sec == 0) { /* start of new recording */
	  startstp = timestp;
	  offset = tim;
	}
	if (timestp - startstp - playoffset >= 0) {
	  fwrite(&len, 1, 2, outfp);
	  fwrite(buf, 1, sizeof(long), outfp);
	}
	continue;
      }
      if (timestp - startstp - playoffset < 0) 
	continue;			/* skip forward until desired time */
      if((sid = buf[0]) > NSID_MAX) {
	fprintf(stderr,"invalid sid in data\n");
	continue;	/* ignore too large a speaker id */
      }
      nv = (nv_hdr_t *) (buf + 1);
      ++cnt;
      if (sidtable[sid].ok) {
	fwrite(&len, 1, 2, outfp);	/* write length of packet */
	fwrite(buf, len, 1, outfp);
      }
    }
    printf("read %d pkts\n",cnt);
    fclose(fp);
    fclose(outfp);
  }
}

int open_file()
{
  DIR *dirp;
  struct dirent *dp;
  char buf[512];
  unsigned int mint = ~0, minu;
  static unsigned int oldmint = 0, oldminu = 0;
  unsigned int m1, um1;

  if ((dirp = opendir(directory)) == NULL) {
    perror("opendir");
    exit(1);
  }
  while ((dp = readdir(dirp)) != NULL) {
    if(!strncmp(dp->d_name, fileprefix, strlen(fileprefix))) {
      sscanf(dp->d_name + strlen(fileprefix), ".%x.%x", &m1, &um1);
      if (m1 && (m1 < mint) && (m1 > oldmint)) {
	mint = m1;
	minu = um1;
      }
      if (m1 && m1 == mint)
	if ((um1 < minu) && (um1 > oldminu))
	  minu = um1;
    }
  }
  closedir(dirp);
  if (mint == ~0)
    return 0;
  sprintf(buf, "%s/%s.%x.%x", directory, fileprefix, mint, minu);
  if((fp = fopen(buf, "r")) == NULL) {
    perror("fopen");
    exit(1);
  }
  oldmint = mint;
  oldminu = minu;
  sprintf(buf, "%s/%s%d.%x.%x", outdirectory, outfileprefix, sessionid,
	  mint, minu);
  if((outfp = fopen(buf, "w")) == NULL) {
    perror("fopen");
    exit(1);
  }
  return 1;
}
