/*
 * Copyright (c) 1992 Anders Klemets <klemets@sics.se>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms, etc.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <dirent.h>
#include "rtp.h"

#define NV_MIN_FLOW	32

/* Each packet on disk is preceeded by 16 bit unsigned integer which
 * contains the length of the packet. No extra information is kept in
 * this integer any more. (The integer is followed by an 8 bit session id.)
 */

#define timediffinms(tp1,tp2)	(((tp1)->tv_sec - (tp2)->tv_sec) * 1000 + ((tp1)->tv_usec - (tp2)->tv_usec) / 1000)

/* defaults */
char *IPaddr = "224.2.1.1";
char ttl = 16;
char *fileprefix = "video";
char *sessionname = NULL;
u_short UDPport = 4444;
char *directory = ".";
int keep = 0;
long playoffset = 0;

FILE *fp;
struct src_info {
  int s;
  u_char used;
};
#define NSID_MAX	64
struct src_info *sidtable[NSID_MAX];	/* source id table */

void signoff();
int interrupted = 0;

main(argc, argv)
int argc;
char **argv;
{
  parsearguments(argc, argv);
  dispchosenparams();
  sendpackets();
}

parsearguments(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  int c;
  while ((c = getopt(argc, argv, "i:p:t:f:d:ko:")) != -1) {
    switch (c) {
    case 'i':
      IPaddr = optarg;
      break;
    case 'p':
      UDPport = atoi(optarg);
      break;
    case 't':
      ttl = atoi(optarg);
      break;
    case 'f':
      fileprefix = optarg;
      break;
    case 'd':
      directory = optarg;
      break;
    case 'k':
      keep = 1;
      break;
    case 'o':
      playoffset = atoi(optarg) * 1000;
      break;
    default:
      fprintf(stderr,"Usage:\t[-i <IP address>] [-p <UDP port>] [-t <ttl>]\n\t[-d <directory>] [-f <filename prefix>] [-o <time offset>] [-k]\n");
      exit(1);
    }
  }
}

dispchosenparams()
{
  printf("Reading video from %s/%s.*\n", directory, fileprefix);
  if (playoffset)
    	printf("Skipping first %d seconds of recording.\n", playoffset/1000);
  printf("Transmitting packets to %s:%d\n", IPaddr, UDPport);
}

#define min(a,b)	((a) < (b) ? (a) : (b))

sendpackets()
{
  long timestp, startstp, delay, lastid = 0, tdiff;
  struct timeval tim, offset, lasttim;
  u_short len;
  u_char sid;
  char buf[66000]; /* first three bytes are unused for alignment reasons */
  int cnt = 0, lastlength;
  struct rtphdr *rtp;

  offset.tv_sec = 0;
  while (!interrupted && open_file() > 0) {
    while (!interrupted && fread (&len, 1, 2, fp) > 0) {
      if(fread (&buf[3], len, 1, fp) != 1)
	break;
      /* Each video packet is preceeded by a 8 bit session id to identify
	 the stream.  The original IP source address is not retained. */
      if((sid = buf[3]) > NSID_MAX) {
	fprintf(stderr,"invalid sid in data\n");
	continue;	/* ignore too large a speaker id */
      }
      rtp = (struct rtphdr *) &buf[4];
      gettimeofday(&tim, NULL);
      timestp = ((u_long)ntohl(rtp->rh_ts) >> 16)*1000 +
	      (((ntohl(rtp->rh_ts)&0xffff)*15625) >> 10) / 1000;
      if (offset.tv_sec == 0) {  /* start of new recording */
	      startstp = timestp;
	      offset = tim;
	      lastid = 0;
	      lastlength = 0;
      }
      /* Use NV timestamp to see if we need to wait */
      delay = timestp - startstp - timediffinms(&tim,&offset) - playoffset;

      /* NV timestamps are from when the frame is grabbed.  One single
       * frame generates a sequence of packets, all with the same timestamp.
       * I would need a timestamp of when the packet was actually sent.
       * Since that is not available, make sure to bandwidth limit the
       * what we send to 1 Mbps, lest packets be lost in spite of an
       * apparently successful send call.
       */
      if (delay <= 0 && lastlength != 0) {
	      tdiff = timediffinms(&tim,&lasttim);
	      if (tdiff > 0 && lastlength*8000 / tdiff > 1024*1024)
		      delay = lastlength*8000/(1024*1024);
      }
      if (delay > 0)
	      if (keep)
		      usleep (delay * 1000);
	      else {
		      /* catch up, but never pause longer than 5 seconds */
		      usleep (min(5000, delay) * 1000);
		      if (delay > 5000)
			      startstp += delay - 5000;
	      }
      if (timestp - startstp - playoffset < 0)
	continue;			/* skip forward until desired time */
      update_speaker(sid);
      ++cnt;
      /* garbage collect sockets every 30 seconds */
      if (timestp - startstp - playoffset - lastid >= 30000) {
	garbagesockets();
	lastid = timestp - startstp - playoffset;
      }
      send(sidtable[sid]->s, &buf[4], len-1, 0);
      lastlength = len-1;
      lasttim = tim;
    }
    printf("read %d pkts\n",cnt);
    fclose(fp);
  }
  if(interrupted)
    printf("Interrupted %d seconds into recording.\n",
	   (timestp - startstp)/1000);
}

int setupsocket()
{
  struct ip_mreq mreq;
  struct hostent *h;
  struct sockaddr_in addr;
  int s;
  
  if((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
      perror("can't open socket");
      exit(1);
    }
  if ((addr.sin_addr.s_addr = inet_addr(IPaddr)) == -1) {
    if ((h = gethostbyname(IPaddr)) == NULL) {
      fprintf(stderr,"Invalid hostname: %s\n",IPaddr);
      exit (1);
    }
    addr.sin_addr.s_addr = *((int *) h->h_addr_list[0]);
  }
  if(IN_MULTICAST(ntohl(addr.sin_addr.s_addr))) {
    mreq.imr_multiaddr = addr.sin_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    if (setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
      perror("setsockopt");
    if (setsockopt(s, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0)
      perror("setsockopt2");
  }
  addr.sin_family = AF_INET;
  addr.sin_port = htons(UDPport);
  if(connect(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("connect");
    exit(1);
  }
  signal(SIGINT, signoff);
  return s;
}

int open_file()
{
  DIR *dirp;
  struct dirent *dp;
  char buf[512];
  unsigned int mint = ~0, minu;
  static unsigned int oldmint = 0, oldminu = 0;
  unsigned int m1, um1;

  if ((dirp = opendir(directory)) == NULL) {
    perror("opendir");
    exit(1);
  }
  while ((dp = readdir(dirp)) != NULL) {
    if(!strncmp(dp->d_name, fileprefix, strlen(fileprefix))) {
      sscanf(dp->d_name + strlen(fileprefix), ".%x.%x", &m1, &um1);
      if (m1 && (m1 < mint) && (m1 > oldmint)) {
	mint = m1;
	minu = um1;
      }
      if (m1 && m1 == mint)
	if ((um1 < minu) && (um1 > oldminu))
	  minu = um1;
    }
  }
  closedir(dirp);
  if (mint == ~0)
    return 0;
  sprintf(buf, "%s/%s.%x.%x", directory, fileprefix, mint, minu);
  if((fp = fopen(buf, "r")) == NULL) {
    perror("fopen");
    exit(1);
  }
  oldmint = mint;
  oldminu = minu;
  return 1;
}

update_speaker(sid)	/* mark source as active */
u_char sid;
{
  if (sidtable[sid] == NULL) {
    sidtable[sid] = (struct src_info *) malloc(sizeof(struct src_info));
    sidtable[sid]->s = setupsocket();
    sidtable[sid]->used = 0;
  }
  if(sidtable[sid]->s == -1) {
    sidtable[sid]->s = setupsocket();
    sidtable[sid]->used = 0;
  }
  sidtable[sid]->used++;
}

garbagesockets()		/* garbage collect open sockets */
{
  int i;
  for (i=0; i < NSID_MAX; ++i)
    if(sidtable[i] != NULL && sidtable[i]->used == 0) {
      close(sidtable[i]->s);
      sidtable[i]->s = -1;
    }
  for (i=0; i < NSID_MAX; ++i)
    if(sidtable[i] != NULL)
      sidtable[i]->used = 0;
}

void signoff()
{
  interrupted = 1;
}
