/*
 * Copyright (c) 1992 Anders Klemets <klemets@sics.se>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms, etc.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include "rtp.h"

#define NV_MIN_FLOW	32

/* each packet on disk is preceeded by 16 bit unsigned integer which
 * contains the length of the packet. No extra information is kept in
 * this integer any more. (The integer is followed by an 8 bit session id.)
 */

#define timediffinms(tp1,tp2)	(((tp1)->tv_sec - (tp2)->tv_sec) * 1000 + ((tp1)->tv_usec - (tp2)->tv_usec) / 1000)

/* defaults */
char *IPaddr = "224.2.1.1";
char *directory = ".";
char *fileprefix = "video";
u_short UDPport = 4444;
int secperfile = 180;
int mintorecord = 0;		/* Number of minutres to record; 0 = forever */
int maxbytes = 0;			/* Max number bytes to record; 0 = unlimited */

FILE *open_file();
int interrupted = 0;
void sighandler();
int sv;		/* nv socket */

#define NSID_MAX	64
struct src_info {
  struct sockaddr_in sin;
  struct timeval last;
};
struct src_info *sidtable[NSID_MAX];	/* source id table */

main(argc, argv)
int argc;
char **argv;
{
  parsearguments(argc, argv);
  check_access();
  setupsocket();
  dispchosenparams();
  getpackets();
}

parsearguments(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  int c, i;
  while ((c = getopt(argc, argv, "i:p:s:t:b:f:d:")) != -1) {
    switch (c) {
    case 'i':
      IPaddr = optarg;
      break;
    case 'p':
      UDPport = atoi(optarg);
      break;
    case 's':
      i = atoi(optarg);
      if (i > 0)
	secperfile = i;
      break;
    case 't':
      mintorecord = atoi(optarg);
      break;
    case 'b':
      maxbytes = atoi(optarg);
      break;
    case 'f':
      fileprefix = optarg;
      break;
    case 'd':
      directory = optarg;
      break;
    default:
      fprintf(stderr,"Usage:\t[-i <IP address>] [-p <UDP port>]\n\t[-s <seconds per file>] [-t <minutes to record>] [-b <max bytes>] \n\t[-d <directory>] [-f <filename prefix>]\n");
      exit(1);
    }
  }
}

dispchosenparams()
{
  printf("Receiving packets from %s:%d\n", IPaddr,
	 UDPport);
  printf("Storing video in %s/%s.* (%d seconds per file)\n",
	 directory, fileprefix, secperfile);
  if (mintorecord)
	 printf("Recording will terminate in %d minutes\n", mintorecord);
  if (maxbytes)
	 printf("Recording will terminate after %d bytes have been recorded\n",
		maxbytes);
}

getpackets()
{
  struct timeval tim, tfile, startime;
  fd_set fds;
  struct sockaddr_in sin;
  int slen;
  u_int len;
  u_short short_len;
  int bytesofar;
  u_char sid;
  char buf[66000];
  struct rtphdr *rtp;
  FILE *fp = NULL;

  gettimeofday(&startime, NULL);
  if (mintorecord)
    alarm(mintorecord*60 + 10);
  bytesofar = 0;
  tfile.tv_sec = 0;
  FD_ZERO(&fds);
  while(!interrupted) {
    FD_SET (sv,&fds);
    if (select (sv+1, &fds, NULL, NULL, NULL) > 0) {
      gettimeofday(&tim, NULL); /* get timestamp for this video packet */
	  if (mintorecord && ((tim.tv_sec - startime.tv_sec) >= (mintorecord * 60)))
		break;
      /* open a new video file for writing */
      if (!tfile.tv_sec || timediffinms(&tim,&tfile) >= 1000 * secperfile) {
	if (fp)
	  fclose(fp);
	fp = open_file();
	tfile = tim;
      }
      
      slen = sizeof(sin);
      len = recvfrom(sv, buf, sizeof (buf), 0, &sin, &slen);
      rtp = (struct rtphdr *) buf;
      if(rtp->rh_vers != RTP_VERSION)		/* check for wrong version */
	continue;
      if(rtp->rh_flow < NV_MIN_FLOW)
	continue;
      if(len >= 65535) {
	fprintf(stderr,"Received giant packet from %s\n",
		inet_ntoa(sin.sin_addr));
	continue;
      }
/*      rtp_parse(buf); */
      sid = register_speaker(&sin,&tim); /* update list of active speakers */
      ++len;			/* increase length to fit speaker id */
      short_len = len;
      fwrite(&short_len, 1, 2, fp);	/* write length of packet */
      fwrite(&sid, 1, 1, fp);		/* write speaker id */
      fwrite(buf, len-1, 1, fp);
      bytesofar += len + 2;
      if (maxbytes && (bytesofar >= maxbytes))
	break;
    }
  }
  if (fp)
    fclose(fp);
}

setupsocket()
{
  struct ip_mreq mreq;
  struct hostent *h;
  struct sockaddr_in addr;
  int one;
  unsigned char loop = 1;
  
  if((sv = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
      perror("can't open socket");
      exit(1);
    }
  if ((addr.sin_addr.s_addr = inet_addr(IPaddr)) == -1) {
    if ((h = gethostbyname(IPaddr)) == NULL) {
      fprintf(stderr,"Invalid hostname, %s\n",IPaddr);
      return 1;
    }
    addr.sin_addr.s_addr = *((int *) h->h_addr_list[0]);
  }
  if(IN_MULTICAST(ntohl(addr.sin_addr.s_addr))) {
    mreq.imr_multiaddr = addr.sin_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    if (setsockopt(sv, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
      perror("setsockopt");
  }
#ifdef notdef /* not useful anyway */
  one = 1;   /* enable reuse of address */
  if (setsockopt(sa, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) < 0)
    perror("setsockopt2");
  if (setsockopt(sa, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop)) < 0)
    perror("setsockopt4");
#endif notdef
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(UDPport);
  if(bind(sv, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("bind");
    exit(1);
  }
  signal(SIGINT, sighandler);
  signal(SIGTERM, sighandler);
  signal(SIGALRM, sighandler);
  return 0;
}

FILE *open_file()
{
  FILE *fp;
  char buf[512];
  struct timeval tim;
  gettimeofday(&tim, NULL);
  sprintf(buf, "%s/%s.%x.%x", directory, fileprefix,
	  tim.tv_sec, tim.tv_usec/1000);
  if((fp = fopen(buf, "w")) == NULL) {
    perror("fopen");
    exit(1);
  }
  return fp;
}

check_access()
{
  char buf[512];
  sprintf(buf, "%s/%s", directory, fileprefix);
  if(access(buf, W_OK) == 0) {
    fprintf(stderr,"cannot write %s\n",buf);
    exit(1);
  }
}
  
int register_speaker(sin, tp)
struct sockaddr_in *sin;
struct timeval *tp;
{
  int i, minuse, minsid;
  for (i=0; i < NSID_MAX; ++i)
    if (sidtable[i] != NULL &&
	sidtable[i]->sin.sin_addr.s_addr == sin->sin_addr.s_addr &&
	sidtable[i]->sin.sin_port == sin->sin_port) {
      sidtable[i]->last = *tp;
      return i;
    }
  for (i=0; i < NSID_MAX; ++i)	/* previously unknown source */
    if (sidtable[i] == NULL) {
      sidtable[i] = (struct src_info *) malloc(sizeof(struct src_info));
      sidtable[i]->sin.sin_addr.s_addr = sin->sin_addr.s_addr;
      sidtable[i]->sin.sin_port = sin->sin_port;
      sidtable[i]->last = *tp;
      return i;
    }
  /* table is full, purge old least used entry */
  minuse = ~0;
  for (i=0; i < NSID_MAX; ++i)
    if (sidtable[i]->last.tv_sec < minuse) {
      minuse = sidtable[i]->last.tv_sec;
      minsid = i;
    }
  sidtable[minsid]->sin.sin_addr.s_addr = sin->sin_addr.s_addr;
  sidtable[minsid]->sin.sin_port = sin->sin_port;
  sidtable[minsid]->last = *tp;
  return minsid;
}

void sighandler()
{
  interrupted = 1;
}

rtp_parse(buf)
char *buf;
{
	struct rtphdr *rtp;
	struct rtpopthdr *ro;
	struct rtpsrchdr *src;
	struct rtpbophdr *bop;
	struct rtcpcdeschdr *cdesc;
	struct rtcpsdeschdr *sdesc;
	struct rtcpbyehdr *bye;
	struct in_addr in;
	int nvpacket;
	rtp = (struct rtphdr *) buf;
	if (rtp->rh_content == RTPCONT_NV)
		nvpacket = 1;
	else
		nvpacket = 0;
	printf ("flow %d sync %d content ",rtp->rh_flow, rtp->rh_sync);
	if (nvpacket)
		printf("NV ");
	else
		printf("%d ",rtp->rh_content);
	printf("seq %d ts %u.%u\n",ntohs(rtp->rh_seq),
	       (u_long)ntohl(rtp->rh_ts) >> 16,
	       ((ntohl(rtp->rh_ts)&0xffff)*15625) >> 10);
	if(rtp->rh_opts) {
		ro = (struct rtpopthdr *) (rtp+1);
		while(1) {
			printf("\topt type %d len %d\n",ro->roh_type,ro->roh_optlen);
			switch (ro->roh_type) {
			      case RTPOPT_CSRC:
				src = (struct rtpsrchdr *) ro;
				in.s_addr = src->rsh_addr;
				printf("CSRC:  addr %s:%d\n",inet_ntoa(in),ntohs(src->rsh_uid));
				break;
			      case RTPOPT_SSRC:
				src = (struct rtpsrchdr *) ro;
				in.s_addr = src->rsh_addr;
				printf("SSRC:  addr %s:%d\n",inet_ntoa(in),ntohs(src->rsh_uid));
				break;
			      case RTPOPT_BOP:
				bop = (struct rtpbophdr *) ro;
				printf("BOP:  seq %d\n",ntohs(bop->rbh_seq));
				break;
			      case RTPOPT_CDESC:
				cdesc = (struct rtcpcdeschdr *) ro;
				printf("CDESC:  content %d rport %d cqual %d cdesc %d\n",
				       cdesc->rtch_content, ntohs(cdesc->rtch_rport),
				       cdesc->rtch_cqual, ntohl(cdesc->rtch_cdesc));
				break;
			      case RTPOPT_SDESC:
				sdesc = (struct rtcpsdeschdr *) ro;
				in.s_addr = sdesc->rtsh_addr;
				printf("SDESC:  uid %s:%d\n",inet_ntoa(in),ntohs(sdesc->rtsh_uid));
				printf("\t%s\n",(char *)ro + sizeof(*sdesc));
				break;
			      case RTPOPT_FDESC:
				printf("FDESC:  ??\n");
				break;
			      case RTPOPT_BYE:
				bye = (struct rtcpbyehdr *) ro;
				in.s_addr = bye->rtbh_addr;
				printf("BYE:  uid %s:%d\n",inet_ntoa(in),ntohs(bye->rtbh_uid));
				break;
			      case RTPOPT_QOS:
				printf("QOS:  ??\n");
				break;
			      case RTPOPT_RAD:
				printf("RAD:  ??\n");
				break;
			      default:
				printf("unknown option\n");
			}
			if (!ro->roh_fin)
				ro = (struct rtpopthdr *)((char *)ro + ro->roh_optlen*4);
			else break;
		}
	}
}

