/*
 * Copyright (c) 1992 Anders Klemets <klemets@sics.se>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms, etc.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <dirent.h>
#include "vat.h"

/* each packet on disk is preceeded by 16 bit unsigned integer.  The first
   12 bits contain the length of the packet that follows.  The remaining
   bits are defined below.
 */
#define	LEN_MASK	0x0fff
#define	LEN_TIMESTAMP	0x1000	/* packet is a timestamp in milliseconds */
#define LEN_SESSION	0x2000	/* packet is vat session announcement */

#define timediffinms(tp1,tp2)	(((tp1)->tv_sec - (tp2)->tv_sec) * 1000 + ((tp1)->tv_usec - (tp2)->tv_usec) / 1000)

/* defaults */
char *IPaddr = "224.2.0.1";
char ttl = 16;
char *fileprefix = "audio";
char *sessionname = NULL;
u_short UDPport = 7878 /* 3456 */;
int confid = 0;
int keep = 0;
long playoffset = 0;
char *directory = ".";

int sa;		/* vat socket */
int si;		/* vat session socket */
FILE *fp;
char *sessionbufp = NULL;
int sessionbuflen;
void signoff();
int interrupted = 0;

main(argc, argv)
int argc;
char **argv;
{
  parsearguments(argc, argv);
  setupsockets();
  dispchosenparams();
  sendpackets();
}

parsearguments(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  int c;
  while ((c = getopt(argc, argv, "i:p:c:t:f:d:ko:")) != -1) {
    switch (c) {
    case 'i':
      IPaddr = optarg;
      break;
    case 'p':
      UDPport = atoi(optarg);
      break;
    case 'c':
      confid = atoi(optarg);
      break;
    case 't':
      ttl = atoi(optarg);
      break;
    case 'f':
      fileprefix = optarg;
      break;
    case 'd':
      directory = optarg;
      break;
    case 'k':
      keep = 1;
      break;
    case 'o':
      playoffset = atoi(optarg) * 1000;
      break;
    default:
      fprintf(stderr,"Usage:\t[-i <IP address>] [-p <UDP port>] [-c <conference ID>]\n\t[-t <ttl>] [-d <directory>] [-f <filename prefix>] [-k] [-o <time offset>]\n");
      exit(1);
    }
  }
}

dispchosenparams()
{
  printf("Reading audio from %s/%s.*\n", directory, fileprefix);
  if (playoffset)
    	printf("Skipping first %d seconds of recording.\n", playoffset/1000);
  printf("Transmitting packets to %s:%d Conference ID: %d\n", IPaddr,
	 UDPport, confid);
}

#define min(a,b)	((a) < (b) ? (a) : (b))

sendpackets()
{
  long timestp, startstp, delay;
  struct timeval tim, offset, lastid;
  u_short len;
  char buf[1500];
  int cnt = 0;
  vat_hdr_t *vat;

  offset.tv_sec = 0;
  while (!interrupted && open_file() > 0) {
    while (!interrupted && fread (&len, 1, 2, fp) > 0) {
      if(fread (buf, len & LEN_MASK, 1, fp) != 1)
	break;
      if (len & LEN_TIMESTAMP) {
	gettimeofday(&tim, NULL);
	timestp = *(long *) buf;
	if(timestp == 0 || offset.tv_sec == 0) { /* start of new recording */
	  startstp = timestp;
	  offset = tim;
	  lastid = tim;
	  lastid.tv_sec -= 5;		/* make sure we send id soon */
	}
	delay = timestp - startstp - timediffinms(&tim,&offset) - playoffset;
	if (delay < 0)
	  continue;
	if (keep)
	  usleep (delay * 1000);
	else {
	  /* catch up, but never pause longer than 5 seconds */
	  usleep (min(5000, delay) * 1000);
	  if (delay > 5000)
	    startstp += delay - 5000;
	}
	continue;
      }
      if (len & LEN_SESSION) {
	sendsession(buf, len & LEN_MASK, timestp - startstp - playoffset >= 0);
	lastid = tim;
	continue;
      }
      if (timestp - startstp - playoffset < 0) 
	continue;			/* skip forward until desired time */
      ++cnt;
      /* send session packet every 5 seconds */
      if (timediffinms(&tim,&lastid) >= 5000) {
	sendsession(NULL, 0, 1);
	lastid = tim;
      }
      vat = (vat_hdr_t *) buf;
      vat->confid = htons(confid);
      if(send(sa, buf, len, 0) < 0)
	perror("send");
    }
    printf("read %d pkts\n",cnt);
    fclose(fp);
  }
  usleep(500000);
  signoff();
  if(interrupted)
    printf("Interrupted %d seconds into recording.\n",
	   (timestp - startstp)/1000);
}

setupsockets()
{
  struct ip_mreq mreq;
  struct hostent *h;
  struct sockaddr_in addr;
  
  if((sa = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
      perror("can't open socket");
      exit(1);
    }
  if ((addr.sin_addr.s_addr = inet_addr(IPaddr)) == -1) {
    if ((h = gethostbyname(IPaddr)) == NULL) {
      fprintf(stderr,"Invalid hostname: %s\n",IPaddr);
      exit (1);
    }
    addr.sin_addr.s_addr = *((int *) h->h_addr_list[0]);
  }
  if(IN_MULTICAST(ntohl(addr.sin_addr.s_addr))) {
    mreq.imr_multiaddr = addr.sin_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    if (setsockopt(sa, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
      perror("setsockopt");
    if (setsockopt(sa, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0)
      perror("setsockopt2");
  }
  addr.sin_family = AF_INET;
  addr.sin_port = htons(UDPport);
  if(connect(sa, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("connect");
    exit(1);
  }
  if((si = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
      perror("can't open socket");
      exit(1);
    }
  if (setsockopt(si, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0)
    perror("setsockopt1");
  addr.sin_port = htons(UDPport+1);
  if(connect(si, (struct sockaddr *) &addr, sizeof(addr)) < 0)
    perror("connect");
  signal(SIGINT, signoff);
  return 0;
}

int open_file()
{
  DIR *dirp;
  struct dirent *dp;
  char buf[512];
  unsigned int mint = ~0, minu;
  static unsigned int oldmint = 0, oldminu = 0;
  unsigned int m1, um1;

  if ((dirp = opendir(directory)) == NULL) {
    perror("opendir");
    exit(1);
  }
  while ((dp = readdir(dirp)) != NULL) {
    if(!strncmp(dp->d_name, fileprefix, strlen(fileprefix))) {
      sscanf(dp->d_name + strlen(fileprefix), ".%x.%x", &m1, &um1);
      if (m1 && (m1 < mint) && (m1 > oldmint)) {
	mint = m1;
	minu = um1;
      }
      if (m1 && m1 == mint)
	if ((um1 < minu) && (um1 > oldminu))
	  minu = um1;
    }
  }
  closedir(dirp);
  if (mint == ~0)
    return 0;
  sprintf(buf, "%s/%s.%x.%x", directory, fileprefix, mint, minu);
  if((fp = fopen(buf, "r")) == NULL) {
    perror("fopen");
    exit(1);
  }
  oldmint = mint;
  oldminu = minu;
  sprintf(buf, "Recorded %s", ctime(&mint));
  if (sessionname)
    free(sessionname);
  buf[strlen(buf)-1] = '\0';
  sessionname = (char *) strdup(buf);
  return 1;
}

sendsession(ptr, plen, sendnow)
char *ptr;
int plen;
int sendnow;
{
  vat_hdr_t *vat;
  if (ptr) {
    if (sessionbufp)
      free(sessionbufp);
    sessionbufp = (char *) malloc(plen);
    memcpy(sessionbufp, ptr, plen);
    sessionbuflen = plen;
    vat = (vat_hdr_t *) sessionbufp;
    vat->confid = htons(confid);
  }
  if (sendnow) {
    if (sessionbufp)
      send(si, sessionbufp, sessionbuflen, 0);
    sendownid();
  }
}

sendownid()
{
  char buf[512];
  struct CtrlMsgHdr *msg;
  msg = (struct CtrlMsgHdr *) buf;
  msg->flags = 0;
  msg->type = CTRL_TYPE_ID;
  msg->confid = htons(confid);
  sprintf(msg + 1, "%s", sessionname);
  send(si, buf, sizeof (*msg) + strlen(sessionname), 0);
}

void signoff()
{
  struct CtrlMsgHdr msg;
  msg.flags = 0;
  msg.type = CTRL_TYPE_DONE;
  msg.confid = htons(confid);
  send(si, &msg, sizeof (msg), 0);
  interrupted = 1;
}
