/*
 * Copyright (c) 1992 Anders Klemets <klemets@sics.se>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms, etc.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include "vat.h"


/* each packet on disk is preceeded by 16 bit unsigned integer.  The first
   12 bits contain the length of the packet that follows.  The remaining
   bits are defined below.
 */
#define NSID_MAX	64
#define	LEN_MASK	0x0fff
#define	LEN_TIMESTAMP	0x1000	/* packet is a timestamp in milliseconds */
#define LEN_SESSION	0x2000	/* packet is vat session announcement */

#define timediffinms(tp1,tp2)	(((tp1)->tv_sec - (tp2)->tv_sec) * 1000 + ((tp1)->tv_usec - (tp2)->tv_usec) / 1000)

/* defaults */
char *IPaddr = "224.2.0.1";
char *directory = ".";
char *fileprefix = "audio";
u_short UDPport = 3456;
int confid = 0;
int secperfile = 180;
int mintorecord = 0;	/* Number of minutes to record; 0 = forever */
int maxbytes = 0;		/* Max number bytes to record; 0 = unlimited */

FILE *open_file();
int interrupted = 0;
void sighandler();
int sa;		/* vat socket */
int si;		/* vat session socket */

struct src_info {
  struct in_addr addr;
  char *name;
};
struct src_info *sidtable[NSID_MAX];	/* source id table */

main(argc, argv)
int argc;
char **argv;
{
  parsearguments(argc, argv);
  check_access();
  setupsockets();
  dispchosenparams();
  getpackets();
}

parsearguments(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  int c, i;
  while ((c = getopt(argc, argv, "i:p:c:s:t:b:f:d:")) != -1) {
    switch (c) {
    case 'i':
      IPaddr = optarg;
      break;
    case 'p':
      UDPport = atoi(optarg);
      break;
    case 'c':
      confid = atoi(optarg);
      break;
    case 's':
      i = atoi(optarg);
      if (i > 0)
	secperfile = i;
      break;
    case 't':
      mintorecord = atoi(optarg);
      break;
    case 'b':
      maxbytes = atoi(optarg);
      break;
    case 'f':
      fileprefix = optarg;
      break;
    case 'd':
      directory = optarg;
      break;
    default:
      fprintf(stderr,"Usage:\t[-i <IP address>] [-p <UDP port>] [-c <conference ID>]\n\t[-s <seconds per file>] [-t <minutes to record>] [-b <max bytes>]\n\t[-d <directory>] [-f <filename prefix>]\n");
      exit(1);
    }
  }
}

dispchosenparams()
{
  printf("Receiving packets from %s:%d Conference ID: %d\n", IPaddr,
	 UDPport, confid);
  printf("Storing audio in %s/%s.* (%d seconds per file)\n",
	 directory, fileprefix, secperfile);
  if (mintorecord)
	 printf("Recording will terminate in %d minutes\n", mintorecord);
  if (maxbytes)
	 printf("Recording will terminate after %d bytes have been recorded\n", 
		maxbytes);
}

getpackets()
{
  struct timeval tim, offset, tfile, startime;
  long timestp, lastts;
  fd_set fds;
  struct sockaddr_in sin;
  int slen;
  int bytesofar;
  u_short len;
  char buf[1500];
  vat_hdr_t *vat;
  FILE *fp = NULL;

  gettimeofday(&startime, NULL);
  if (mintorecord)
	alarm(mintorecord*60 + 10);
  bytesofar = 0;
  offset.tv_sec = 0;
  tfile.tv_sec = 0;
  FD_ZERO(&fds);
  while(!interrupted) {
    FD_SET (sa,&fds);
    FD_SET (si,&fds);
    if (select (si+1, &fds, NULL, NULL, NULL) > 0) {
      if(FD_ISSET(si,&fds))
	getsession(fp);
      if (!FD_ISSET(sa,&fds))
	continue;
      
      gettimeofday(&tim, NULL); /* get timestamp for this audio packet */
	  if (mintorecord && ((tim.tv_sec - startime.tv_sec) >= (mintorecord * 60)))
		break;		/* Recorded specified amount of time */
      /* open a new audio file for writing */
      if (!tfile.tv_sec || timediffinms(&tim,&tfile) >= 1000 * secperfile) {
	if (fp)
	  fclose(fp);
	fp = open_file();
	tfile = tim;
      }
      if (offset.tv_sec == 0) {	 /* first packet */
	offset = tim;
	puttime(0,fp);	/* write initial timestamp */
	lastts = 0;
      }
      timestp = timediffinms(&tim,&offset);
      if (timestp - lastts >= 100) {
	puttime(timestp,fp);	/* write timestamp every 100 ms */
	lastts = timestp;
      }
      
      slen = sizeof(sin);
      len = recvfrom(sa, buf, 1500, 0, &sin, &slen);
      vat = (vat_hdr_t *) buf;
      if (ntohs(vat->confid) != confid)
	continue;	/* wrong conference id */
      register_speaker(sin.sin_addr);	/* update list of active speakers */
      
      vat->nsid++;
      len += 4;	/* increase length to keep extra speaker id */
      len &= LEN_MASK;
      fwrite(&len, 1, 2, fp);	/* write length of packet */
      fwrite(vat, 1, sizeof(vat_hdr_t), fp);	/* write vat header */
      fwrite(&sin.sin_addr.s_addr, 1, 4, fp);	/* write speaker id */
      fwrite(buf + sizeof(vat_hdr_t), len - sizeof(vat_hdr_t) - 4, 1, fp);
      bytesofar += len + 2;
      if (maxbytes && (bytesofar >= maxbytes))
	break;
    }
  }
  if (fp)
    fclose(fp);
  if (offset.tv_sec)
    printf("Recorded %d seconds\n", tim.tv_sec - offset.tv_sec);
  else
    printf("Recorded nothing.\n");
}

getsession(fp)
FILE *fp;
{
  char buf[1500], *ptr;
  struct CtrlMsgHdr *msg;
  struct IDMsgHdr *idmsg;
  struct sockaddr_in sin;
  int i, slen;
  u_short len;
  slen = sizeof(sin);
  len = recvfrom(si, buf, 1500, 0, &sin, &slen);
  msg = (struct CtrlMsgHdr *) buf;
  if (ntohs(msg->confid) != confid)
    return;
  switch (msg->type) {
  case CTRL_TYPE_ID:
    update_speaker(sin.sin_addr, msg + 1, fp);
    break;
  case CTRL_TYPE_DONE:
    update_speaker(sin.sin_addr, NULL, fp);
    break;
  case CTRL_TYPE_IDLIST:
    idmsg = (struct IDMsgHdr *) buf;
    ptr = (char *) (idmsg + 1);
    for (i=0; i < idmsg->nsids; ++i) {
      update_speaker (*(struct in_addr *) ptr, ptr + 4, fp);
      len = strlen(ptr + 4);
      ptr += 4 + 4 * (len / 4);
      if (len % 4)
	ptr += 4;
    }
  default:
    printf("unknown session id type\n");
    break;
  }
}
  
setupsockets()
{
  struct ip_mreq mreq;
  struct hostent *h;
  struct sockaddr_in addr;
  int one;
  unsigned char loop = 1;
  
  if((sa = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
      perror("can't open socket");
      exit(1);
    }
  if ((addr.sin_addr.s_addr = inet_addr(IPaddr)) == -1) {
    if ((h = gethostbyname(IPaddr)) == NULL) {
      fprintf(stderr,"Invalid hostname, %s\n",IPaddr);
      return 1;
    }
    addr.sin_addr.s_addr = *((int *) h->h_addr_list[0]);
  }
  if(IN_MULTICAST(ntohl(addr.sin_addr.s_addr))) {
    mreq.imr_multiaddr = addr.sin_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    if (setsockopt(sa, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
      perror("setsockopt");
  }
#ifdef notdef /* not useful anyway */
  one = 1;   /* enable reuse of address */
  if (setsockopt(sa, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) < 0)
    perror("setsockopt2");
  if (setsockopt(sa, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop)) < 0)
    perror("setsockopt4");
#endif notdef
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(UDPport);
  if(bind(sa, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("bind");
    exit(1);
  }
  if((si = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
      perror("can't open socket");
      exit(1);
    }
  addr.sin_port = htons(UDPport+1);
  if(bind(si, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("bind");
    exit(1);
  }
  signal(SIGINT, sighandler);
  signal(SIGTERM, sighandler);
  signal(SIGALRM, sighandler);
  return 0;
}

FILE *open_file()
{
  FILE *fp;
  char buf[512];
  struct timeval tim;
  gettimeofday(&tim, NULL);
  sprintf(buf, "%s/%s.%x.%x", directory, fileprefix,
	  tim.tv_sec, tim.tv_usec/1000);
  if((fp = fopen(buf, "w")) == NULL) {
    perror("fopen");
    exit(1);
  }
  return fp;
}

check_access()
{
  char buf[512];
  sprintf(buf, "%s/%s", directory, fileprefix);
  if(access(buf, W_OK) == 0) {
    fprintf(stderr,"cannot write %s\n",buf);
    exit(1);
  }
}
  

puttime(ts, fp)
long ts;
FILE *fp;
{
  u_short len;
  len = sizeof(ts) | LEN_TIMESTAMP;
  fwrite(&len, 1, 2, fp);	/* write length of packet */
  fwrite(&ts, 1, sizeof(ts), fp);
}

int register_speaker(addr)
struct in_addr addr;
{
  int i;
  for (i=0; i < NSID_MAX; ++i)
    if (sidtable[i] != NULL && sidtable[i]->addr.s_addr == addr.s_addr)
      return;
  for (i=1; i < NSID_MAX; ++i)
    if (sidtable[i] == NULL) {
      sidtable[i] = (struct src_info *) malloc(sizeof(struct src_info));
      sidtable[i]->addr.s_addr = addr.s_addr;
      sidtable[i]->name = NULL;
      return;
    }
}

update_speaker(addr, str, fp)
struct in_addr addr;
char *str;
FILE *fp;
{
  int i;
  for (i=0; i < NSID_MAX; ++i)
    if (sidtable[i] != NULL && sidtable[i]->addr.s_addr == addr.s_addr)
      break;
  if (i == NSID_MAX)	/* source has not spoken yet - ignore it */
    return;
  if (str == NULL) {	/* purge speaker id */
    if (sidtable[i]->name)
      free(sidtable[i]->name);
    free(sidtable[i]);
    sidtable[i] = NULL;
    writesession(fp);	/* send sign off message (doesn't really work) */
    return;
  }
  if (sidtable[i]->name == NULL) {
    sidtable[i]->name = (char *) strdup(str);
    writesession(fp);
  }
  else
    if (strcmp(sidtable[i]->name,str)) {
      free(sidtable[i]->name);
      sidtable[i]->name = (char *) strdup(str);
      writesession(fp);
    }
}

writesession(fp)
FILE *fp;
{
  char buf[1500], *ptr;
  struct IDMsgHdr idmsg;
  int i, j, cnt = 0;
  u_short len;
  ptr = buf;
  for (i=0; i < NSID_MAX; ++i)
    if (sidtable[i] != NULL && sidtable[i]->name != NULL) {
      *(struct in_addr *) ptr = sidtable[i]->addr;
      strcpy(ptr+4, sidtable[i]->name);
      len = strlen(sidtable[i]->name) + 1;
      ptr += 4 + len;
      if (len % 4)
	for (j=0 ; j < 4 - len % 4; ++j)
	  *ptr++ = '\0';
      cnt++;
    }
  idmsg.h.flags = 0;
  idmsg.h.type = CTRL_TYPE_IDLIST;
  idmsg.h.confid = htons(confid);
  idmsg.nsids = cnt;
  idmsg.ainfo = 0;
  idmsg.blksize = 0;
  len = sizeof(idmsg) + ptr - buf;
  len |= LEN_SESSION;
  fwrite(&len, 1, 2, fp);		/* write length of packet */
  fwrite(&idmsg, 1, sizeof(idmsg), fp);	/* write session header */
  fwrite(buf, ptr - buf, 1, fp);
}

void sighandler()
{
  interrupted = 1;
}
