/*
 * Copyright (c) 1992 Anders Klemets <klemets@sics.se>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms, etc.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include "wb.h"

/* each packet on disk is preceeded by 16 bit unsigned integer.  The first
   12 bits contain the length of the packet that follows.  The remaining
   bits are defined below.
 */
#define	LEN_MASK	0x0fff
#define	LEN_TIMESTAMP	0x1000	/* packet is a timestamp in milliseconds */

#define timediffinms(tp1,tp2)	(((tp1)->tv_sec - (tp2)->tv_sec) * 1000 + ((tp1)->tv_usec - (tp2)->tv_usec) / 1000)

/* defaults */
char *IPaddr = "224.2.0.23";
char *directory = ".";
char *fileprefix = "wb";
u_short UDPport = 4567;
int secperfile = 180;

FILE *open_file();
int interrupted = 0;
void sighandler();
int sw;		/* wb socket */

#define NSID_MAX	64
struct src_info {
  struct sockaddr_in sin;
  struct timeval last;
};
struct src_info *sidtable[NSID_MAX];	/* source id table */

main(argc, argv)
int argc;
char **argv;
{
  parsearguments(argc, argv);
  check_access();
  setupsocket();
  dispchosenparams();
  getpackets();
}

parsearguments(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  int c, i;
  while ((c = getopt(argc, argv, "i:p:s:f:d:")) != -1) {
    switch (c) {
    case 'i':
      IPaddr = optarg;
      break;
    case 'p':
      UDPport = atoi(optarg);
      break;
    case 's':
      i = atoi(optarg);
      if (i > 0)
	secperfile = i;
      break;
    case 'f':
      fileprefix = optarg;
      break;
    case 'd':
      directory = optarg;
      break;
    default:
      fprintf(stderr,"Usage:\t[-i <IP address>] [-p <UDP port>]\n\t[-s <seconds per file>] [-d <directory>] [-f <filename prefix>]\n");
      exit(1);
    }
  }
}

dispchosenparams()
{
  printf("Receiving packets from %s:%d\n", IPaddr,
	 UDPport);
  printf("Storing whiteboard data in %s/%s.* (%d seconds per file)\n",
	 directory, fileprefix, secperfile);
}

getpackets()
{
  struct timeval tim, offset, tfile;
  long timestp, lastts;
  fd_set fds;
  struct sockaddr_in sin;
  int slen, cnt = 0;
  u_short len;
  u_char sid;
  char buf[1500];
  wb_hdr_t *wb;
  FILE *fp = NULL;
  int i;

  offset.tv_sec = 0;
  tfile.tv_sec = 0;
  FD_ZERO(&fds);
  while(!interrupted) {
    FD_SET (sw,&fds);
    if (select (sw+1, &fds, NULL, NULL, NULL) > 0) {
      gettimeofday(&tim, NULL); /* get timestamp for this video packet */
      /* open a new video file for writing */
      if (!tfile.tv_sec || timediffinms(&tim,&tfile) >= 1000 * secperfile) {
	if (fp)
	  fclose(fp);
	fp = open_file();
	tfile = tim;
      }
      if (offset.tv_sec == 0) {	/* first packet */
	offset = tim;
	puttime(0,fp);	/* write initial timestamp */
	lastts = 0;
      }
      timestp = timediffinms(&tim,&offset);
      if (timestp - lastts >= 100) {
	puttime(timestp,fp);	/* write timestamp every 100 ms */
	lastts = timestp;
      }
      
      slen = sizeof(sin);
      len = recvfrom(sw, buf, 1500, 0, &sin, &slen);
      wb = (wb_hdr_t *) buf;
/*      for(i=0; i<16; ++i)
	printf("%u ",wb->dummy[i]);
      printf("\n");
      printf("src %s\n",inet_ntoa(wb->src)); */
/*      printf("id %d\n",wb->id);
      if(wb->id == 0)
	printf("%s\n",wb+1); */
#ifdef notdef
      if(nv->ver != NV_VERSION)		/* check for wrong version */
	continue;
#endif
      sid = register_speaker(&sin,&tim); /* update list of active speakers */
      ++len;			/* increase length to fit speaker id */
      len &= LEN_MASK;
      fwrite(&len, 1, 2, fp);	/* write length of packet */
      fwrite(&sid, 1, 1, fp);	/* write speaker id */
      fwrite(buf, len-1, 1, fp);
      ++cnt;
    }
  }
  if (fp)
    fclose(fp);
  if(cnt)
    printf("Recorded %d packets\n",cnt);
  else
    printf("Recorded nothing.\n");
}

setupsocket()
{
  struct ip_mreq mreq;
  struct hostent *h;
  struct sockaddr_in addr;
  int one;
  unsigned char loop = 1;
  
  if((sw = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
      perror("can't open socket");
      exit(1);
    }
  if ((addr.sin_addr.s_addr = inet_addr(IPaddr)) == -1) {
    if ((h = gethostbyname(IPaddr)) == NULL) {
      fprintf(stderr,"Invalid hostname, %s\n",IPaddr);
      return 1;
    }
    addr.sin_addr.s_addr = *((int *) h->h_addr_list[0]);
  }
  if(IN_MULTICAST(ntohl(addr.sin_addr.s_addr))) {
    mreq.imr_multiaddr = addr.sin_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    if (setsockopt(sw, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0)
      perror("setsockopt");
  }
#ifdef notdef /* not useful anyway */
  one = 1;   /* enable reuse of address */
  if (setsockopt(sa, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) < 0)
    perror("setsockopt2");
  if (setsockopt(sa, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop)) < 0)
    perror("setsockopt4");
#endif notdef
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(UDPport);
  if(bind(sw, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("bind");
    exit(1);
  }
  signal(SIGINT, sighandler);
  signal(SIGTERM, sighandler);
  return 0;
}

FILE *open_file()
{
  FILE *fp;
  char buf[512];
  struct timeval tim;
  gettimeofday(&tim, NULL);
  sprintf(buf, "%s/%s.%x.%x", directory, fileprefix,
	  tim.tv_sec, tim.tv_usec/1000);
  if((fp = fopen(buf, "w")) == NULL) {
    perror("fopen");
    exit(1);
  }
  return fp;
}

check_access()
{
  char buf[512];
  sprintf(buf, "%s/%s", directory, fileprefix);
  if(access(buf, W_OK) == 0) {
    fprintf(stderr,"cannot write %s\n",buf);
    exit(1);
  }
}
  

puttime(ts, fp)
long ts;
FILE *fp;
{
  u_short len;
  len = sizeof(ts) | LEN_TIMESTAMP;
  fwrite(&len, 1, 2, fp);	/* write length of packet */
  fwrite(&ts, 1, sizeof(ts), fp);
}

int register_speaker(sin, tp)
struct sockaddr_in *sin;
struct timeval *tp;
{
  int i, minuse, minsid;
  for (i=0; i < NSID_MAX; ++i)
    if (sidtable[i] != NULL &&
	sidtable[i]->sin.sin_addr.s_addr == sin->sin_addr.s_addr &&
	sidtable[i]->sin.sin_port == sin->sin_port) {
      sidtable[i]->last = *tp;
      return i;
    }
  for (i=0; i < NSID_MAX; ++i)	/* previously unknown source */
    if (sidtable[i] == NULL) {
      sidtable[i] = (struct src_info *) malloc(sizeof(struct src_info));
      sidtable[i]->sin.sin_addr.s_addr = sin->sin_addr.s_addr;
      sidtable[i]->sin.sin_port = sin->sin_port;
      sidtable[i]->last = *tp;
      return i;
    }
  /* table is full, purge old least used entry */
  minuse = ~0;
  for (i=0; i < NSID_MAX; ++i)
    if (sidtable[i]->last.tv_sec < minuse) {
      minuse = sidtable[i]->last.tv_sec;
      minsid = i;
    }
  sidtable[minsid]->sin.sin_addr.s_addr = sin->sin_addr.s_addr;
  sidtable[minsid]->sin.sin_port = sin->sin_port;
  sidtable[minsid]->last = *tp;
  return minsid;
}

void sighandler()
{
  interrupted = 1;
}
