#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <unistd.h>
#include <math.h>
#include <fcntl.h>
#include "cmChunk.h"
#include "cmCookies.h"

#define BUF_SIZE 300000
#define ONE_BILLION 1000000000.0

/*
 *--------------------------------------------------------------
 *
 * CopyFile --
 *
 *	Copy a file to the end of the output file
 *
 * Results:
 *	1 for success, 0 for error (with errno set)
 *
 * Side effects:
 *	None
 *
 * Parameters:
 *	char *fileName	in	Name of file to copy
 *	int outFD	in	File descriptor to copy to
 *
 *--------------------------------------------------------------
 */
static int
CopyFile (char *fileName, int outFD)
{
    FILE *inFile;
    char buffer[BUF_SIZE];
    int numRead;

    inFile = fopen (fileName, "r");
    if (inFile == NULL) {
        return 0;
    }

    while (1) {
        numRead = fread(buffer, 1, BUF_SIZE, inFile);
        if (numRead == 0) {
            break;
        }
        if (numRead < 0) {
            fclose (inFile);
            return -1;
        }

        write (outFD, buffer, numRead);
    }

    fclose (inFile);
    return 1;
}

/*
 *--------------------------------------------------------------
 *
 * ReadJPEG --
 *
 *	Read in one JPEG image at the current file position and
 *	return the parameters.  This one basically *has* to match
 *	the format of the JFIF.  NextChar is a helper function.
 *	NextShort, NextInt are helper functions.
 *	
 *
 * Results:
 *	1 for success, 0 for failure (not a JPEG SOI)
 *
 * Side effects:
 *	File point is advanced to start of next image
 *
 * Parameters:
 *	FILE *filePtr	in	File descriptor to read
 *	int *width	out	Returned width of image
 *	int *height	out	Returned height of image
 *	int *len	out	Returned length (in bytes) of data
 *
 *--------------------------------------------------------------
 */
static unsigned char
NextChar (FILE *filePtr)
{
    unsigned char rv;
    fread(&rv, sizeof(rv), 1, filePtr);
    return rv;
}

static unsigned short
NextShort (FILE *filePtr)
{
    unsigned short rv;
    fread(&rv, sizeof(rv), 1, filePtr);
    return ntohs(rv);
}

static int
NextInt (FILE *filePtr)
{
    int rv;
    fread(&rv, sizeof(rv), 1, filePtr);
    return ntohl(rv);
}

static void
FindMarker (FILE *filePtr, int marker)
{
    while (1) {
	if (NextChar(filePtr) == 0xff) {
	    if (NextChar(filePtr) == marker) {
		return;
	    }
	}
    }
}

static int
ReadJPEG (char *fileName, int *width, int *height, int *len)
{
    FILE *inFile;
    int i, numComp, value;

    inFile = fopen (fileName, "r");
    if (inFile == NULL) {
        return 0;
    }

    /* Verify 0xffd8 == SOI */
    if (NextShort(inFile) != 0xffd8) {
        fclose (inFile);
        return 0;
    }

    /* Scan for SOF */
    FindMarker (inFile, 0xc0);

    NextShort(inFile);	/* Skip length = 17 bytes */
    NextChar (inFile);
    *height = NextShort(inFile);
    *width = NextShort(inFile);
    numComp = NextChar(inFile); /* # components */
    if (numComp != 3) {
	fclose (inFile);
	fprintf (stderr, "File is not a full color file\n");
	return 0;
    }
    for (i=0; i< numComp; i++) {
	NextChar(inFile);		/* Ignore component id */
	value = NextChar(inFile);	/* Decimation of component */
	if (((i == 0) && (value != 0x21)) ||
	    ((i  > 0) && (value != 0x11))) {
	    fclose (inFile);
	    fprintf (stderr, "File has bad format: must be 4:2:2\n");
	    return 0;
	}
	NextChar(inFile);		/* Ignore table id */
    }

    if (feof(inFile)) {
        fclose (inFile);
        return 0;
    }

    fseek (inFile, 0, SEEK_END);
    *len = ftell(inFile);
    fclose (inFile);
    return 1;
}

double *
MakeFTT (int numFrames, double fps, char *fttFileName)
{
    FILE *inFile;
    double *ftt;
    double base;
    int i;
    char str[256];

    /*
     * Allocate a "base" ftt at a rate of 1/fps that we can
     * return in case of error.
     */
    ftt = (double *) malloc((numFrames+1)*sizeof(double));
    for (i=0; i<=numFrames; i++) {
	ftt[i] = (((double) i) * (1.0 / fps));
    }

    if (fttFileName == NULL) {
        return ftt;
    }

    inFile = fopen (fttFileName, "r");
    if (inFile == NULL) {
        return ftt;
    }

    /*
     * Read the entries of the ftt in the file.  Break if we get
     * an entry that is less than the previous.
     */
    for (i=0; i<=numFrames; i++) {
        if (fgets(str, sizeof(str), inFile) == NULL) {
	    fclose (inFile);
	    break;
        }
        ftt[i] = atof(str);
        if ((i != 0) && (ftt[i] <= ftt[i-1])) {
            break;
        }
    }

    if (i == 0) {
        return ftt;
    }

    /*
     * Fill in the rest at fps
     */
    i--;
    base = ftt[i] - (i/fps);
    i++;
    while (i<=numFrames) {
	ftt[i] = base + i/fps;
	i++;
    }
    return ftt;
}

void
main (int argc, char **argv)
{
    tChunkFileState *outFile;
    char *outFileName;
    int size, offset, numFrames, i, width, height;
    MJPEGHeader mjpegHeader;
    MediaIndepHeader mediaIndepHdr;
    int *fot;
    char *fttFileName;
    double *ftt;
    double fps, clipLength;
    int clipLengthSec, clipLengthNanoSecs;
    int sec, nanoSec;
    int firstFileIndex;

    if (argc < 4) {
usage:
	fprintf(stderr,
		"Usage: %s ?-fps value? ?-ftt fttFile? -o outputFile file1 [file2 ...]\n",
		argv[0]);
	exit(1);
    }

    /*
     * Parse arguments.
     */
    outFileName = NULL;
    fps = 30.0;
    fttFileName = NULL;
    for (i=1; i<argc; i++) {
	if (strcmp(argv[i], "-fps") == 0) {
	    i++;
	    if (i == argc) {
		fprintf(stderr, "Must specify value after -fps option.\n");
		exit(1);
	    }
	    fps = atof(argv[i]);
	} else if (strcmp(argv[i], "-o") == 0) {
	    i++;
	    if (i == argc) {
		fprintf(stderr, "Must specify value after -o option.\n");
		exit(1);
	    }
	    outFileName = argv[i];
	} else if (strcmp(argv[i], "-ftt") == 0) {
	    i++;
	    if (i == argc) {
		fprintf(stderr, "Must specify value after -ftt option.\n");
		exit(1);
	    }
	    fttFileName = argv[i];
	} else {
	    break;
	}
    }
    firstFileIndex = i;

    /*
     * At this point, argv[firstFileIndex] is first input file.
     * Check that they've specified some input files and that they've
     * specified an output file.
     */
    if (firstFileIndex == argc) {
	fprintf(stderr, "%s error: no input files.\n", argv[0]);
	goto usage;
    }
    if (outFileName == NULL) {
	fprintf(stderr, "%s error: no output file specified.\n", argv[0]);
	goto usage;
    }

    /*
     * Open output file as chunk file.
     */
    outFile = Chunk_OpenFile(outFileName, "cw");
    if (outFile == NULL) {
	fprintf(stderr, "Couldn't open output file: %s\n", outFileName);
	perror ("");
	exit(1);
    }

    /* 
     * Read the header of each file to construct the FOT.
     * The FOT is formatted s an array of (offset, size) pairs
     *
     * - The total number of frames in (argc-i+1).
     * - offset is the offset of the output file relative to the beginning
     *   of the data segment.
     */
    fot = (int *)malloc((argc-i+1)*sizeof(int)*2);
    offset = 0;
    numFrames = 0;
    for (i=firstFileIndex; i<argc; i++) {
	if (!ReadJPEG (argv[i], &width, &height, &size)) {
	    fprintf(stderr, "File %s does not appear to be a raw JPEG file.\n",
	    	    argv[i]);
	    exit(1);
	}

	if ((width <= 0) || (height <= 0) || (size <= 0)) {
	    fprintf(stderr, "Frame header in raw file made no sense.\n");
	    exit(1);
	}

	fot[numFrames*2] = offset;
	fot[(numFrames*2)+1] = size;
	offset += size;
	fprintf (stderr, "%d\r", ++numFrames);
    }

    /*
     * Now create FTT.
     */
    ftt = MakeFTT (numFrames, fps, fttFileName);

    /*
     * Write media independent header to chunkfile.
     */
    mediaIndepHdr.media = htonl(VIDEO_MEDIA_CODE);
    mediaIndepHdr.format = htonl(MJPEG_FORMAT_CODE);
    mediaIndepHdr.device = htonl(OTHER_DEV_CODE);
    mediaIndepHdr.version = htonl(3);
    mediaIndepHdr.num_frames = htonl(numFrames);
    clipLength = ftt[numFrames];
    clipLengthSec = floor(clipLength);
    clipLengthNanoSecs = ((clipLength - floor(clipLength)) * ONE_BILLION);
    mediaIndepHdr.length_s = htonl(clipLengthSec);
    mediaIndepHdr.length_ns = htonl(clipLengthNanoSecs);
    if (Chunk_WriteHeader(outFile, CM_MEDIA_INDEP_CF, "") != 0) {
        fprintf(stderr, "Error writing media indep. header\n");
        exit(1);
    }
    write(outFile->fd, &mediaIndepHdr, (sizeof(MediaIndepHeader)));


    /*
     * Write media dependent header to chunkfile.
     */
    mjpegHeader.width = htonl(width);
    mjpegHeader.height = htonl(height);
    mjpegHeader.device = htonl(OTHER_DEV_CODE);
    mjpegHeader.subformat = htonl(JFIF_FRAME_CODE);
    mjpegHeader.qfactor = htonl(50);
    if (Chunk_WriteHeader(outFile, CM_MJPEG_CF, "") != 0) {
	fprintf(stderr, "Error writing mjpeg chunk header.\n");
	exit(1);
    }
    write(outFile->fd, &mjpegHeader, (sizeof(MJPEGHeader)));


    /*
     * Write the video data.
     */
    if (Chunk_WriteHeader(outFile, CM_DATA_SEGMENT, "") != 0) {
	fprintf(stderr, "Error writing data chunk header.\n");
	exit(1);
    }
    for (i=firstFileIndex; i<argc; i++) {
        if (!CopyFile (argv[i], outFile->fd)) {
	    fprintf(stderr, "Error copying data to output.\n");
	    perror(argv[i]);
	    exit(1);
        }
    }

    /*
     * Write frame offset table.
     */
    if (Chunk_WriteHeader(outFile, CM_FOT, "") != 0) {
	fprintf(stderr, "Error writing frame offset table chunk header.\n");
	exit(1);
    }
    for (i=0; i<numFrames; i++) {
	int offset, length;
	offset = htonl(fot[i*2]);
	length = htonl(fot[(i*2)+1]);
	write(outFile->fd, &offset, sizeof(int));
	write(outFile->fd, &length, sizeof(int));
    }

    /*
     * Write frame time table.
     */
    if (Chunk_WriteHeader(outFile, CM_FTT, "") != 0) {
	fprintf(stderr, "Error writing frame time table chunk header.\n");
	exit(1);
    }
    for (i=0; i<=numFrames; i++) {
	sec = floor(ftt[i]);
	nanoSec = ((ftt[i] - floor(ftt[i]))*ONE_BILLION);
	
	sec = htonl(sec);
	nanoSec = htonl(nanoSec);
	
	write(outFile->fd, &sec, sizeof(int));
	write(outFile->fd, &nanoSec, sizeof(int));
    }

    Chunk_CloseFile(outFile);   
}
