#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <unistd.h>
#include <math.h>
#include "cmAuFile.h"
#include "cmChunk.h"
#include "cmCookies.h"

#define ONE_BILLION 1000000000.0
#define CM_TEXT_SEGMENT 0xdabadaba

/*
 *--------------------------------------------------------------
 *
 * ReadAudioHeader --
 *
 *	Read a sparc/next audio file header from the file descriptor
 *	passed in.  Convert it to native format and make sure it's
 *	valid.  If so, alloc a new RawSparcAuHeader structure, fill it
 *	in, and return it to the caller.  The structure is completely
 *	filled in, including the dataSize field (which is sometimes
 *	left empty in sparc audio files).  The new memory belongs to
 *	the caller, and the caller should free it using ckfree() when
 *	they are finished.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	Memory is allocated and handed to the caller.
 *	The file pointer is reset to the point directly after the
 *	file header.
 *
 * Parameters:
 *	interp	(in)	Used to report errors
 *	fd	(in)	File descriptor to read
 *	hdrPtr	(out)	Plaace to write filled in structure.
 *
 *--------------------------------------------------------------
 */
static int
ReadAudioHeader(int fd, RawSparcAuHeader *srcHdr)
{
    int n;
    char *null;
    char *format;

    lseek (fd, 0, SEEK_SET);
    n = read (fd, srcHdr, sizeof(RawSparcAuHeader));
    
    if (n != sizeof(RawSparcAuHeader)) {
      fprintf (stderr, "File is not a sparc audio file\n");
      return 0;
    }

    srcHdr->magic        = ntohl(srcHdr->magic);
    srcHdr->dataLocation = ntohl(srcHdr->dataLocation);
    srcHdr->dataSize     = ntohl(srcHdr->dataSize);
    srcHdr->dataFormat   = ntohl(srcHdr->dataFormat);
    srcHdr->samplingRate = ntohl(srcHdr->samplingRate);
    srcHdr->channelCount = ntohl(srcHdr->channelCount);

    if (srcHdr->dataFormat == 1) {
      format = "Ulaw";
    } else if (srcHdr->dataFormat == 2) {
      format = "8-bit unsigned linear";
    } else if (srcHdr->dataFormat == 3) {
      format = "16-bit unsigned linear";
    } else {
      format = "unknown";
    }

    if (srcHdr->magic != SUN_MAGIC) {
	fprintf(stderr, "File is not a sparc audio file");
	return 0;
    }

    printf("file is %d bytes, type %s, %d channels, at %d samples/sec\n",
	   srcHdr->dataSize, format, srcHdr->channelCount, 
	   srcHdr->samplingRate );

    /* Read past comments in header (to dataLocation) */
    null = malloc(srcHdr->dataLocation - n+1);
    read (fd, null, srcHdr->dataLocation - n);
    free(null);

    if (srcHdr->dataFormat != SUN_ULAW &&
	srcHdr->dataFormat != SUN_LIN_8 &&
	srcHdr->dataFormat != SUN_LIN_16) {
	fprintf(stderr, "Source is in an unsupported format.\n");
	return 0;
    }

    return 1;
}



void
SparcToAuFile( char *in, char *out )
{
  tChunkFileState *cfState;
  MediaIndepHeader mediaHdr;
  SparcAuHeader sparcHdr;
  RawSparcAuHeader rawHdr;
  int fd, sampleSize, secs, nsecs, zero;
  int numRead, bufSize, numFrames, size;
  double length;
  char *buffer, string[256], format[20];

  /* if in = "-", read from stdin */

  if (strcmp(in, "-") != 0) {
    fd  = open(in, O_RDONLY);
    if (fd == -1) {
      fprintf( stderr, "Problem reading file %s.\n", in );
      exit(-1);
    }
  } else {
    fd = 0;
  }
  
  if( ReadAudioHeader(fd, &rawHdr) != 1 ) {
    exit(-1);
  }

  if( rawHdr.dataFormat == SUN_LIN_16 ) {
    sampleSize = 2;
  } else {
    sampleSize = 1;
  }
  sampleSize = sampleSize * rawHdr.channelCount;
  bufSize = sampleSize*rawHdr.samplingRate;

  /* Open Output as chunkfile */

  cfState = Chunk_OpenFile(out, "cw");

  if (cfState == NULL) {
    fprintf( stderr, "Couldn't create output chunk file: %s\n", out );
    exit(-1);
  }

  /* Write the audio data. */

  if (Chunk_WriteHeader(cfState, CM_DATA_SEGMENT, "") != 0) {
    fprintf(stderr, "Error writing data chunk header.\n");
    exit(-1);
  }

  buffer = (char *) malloc(bufSize);
  if (buffer == NULL) {
    fprintf(stderr, "Error allocating transfer buffer.\n");
    exit(-1);
  }
  
  size = 0;
  numRead = read(fd, buffer, bufSize); 
  
  while (numRead != 0) {
    size += numRead;
    write(cfState->fd, buffer, numRead);
    numRead = read(fd, buffer, bufSize);
  }

  free(buffer);
  close(fd);

  if (rawHdr.dataSize == SUN_UNSPEC) {
    rawHdr.dataSize = size;
  }

  if (size != rawHdr.dataSize) {
    fprintf(stderr, "Warning: size in header doesn't match actual size.\n");
    fprintf(stderr, "         hdr=%d, actual=%d\n", rawHdr.dataSize, size);
  }

  rawHdr.dataSize = size;
  
  length = (double)rawHdr.dataSize / 
           (double)(rawHdr.samplingRate * sampleSize);

  /* numFrames = (int)ceil(length); */
  secs = (int)floor(length);
  nsecs = (int)((length - secs) * ONE_BILLION);
  
  mediaHdr.media = htonl(AUDIO_MEDIA_CODE);
  mediaHdr.format = htonl(SPARC_AU_FORMAT_CODE);
  mediaHdr.device = htonl(SPARC_AU_DEV_CODE);
  mediaHdr.version = htonl(3);
  mediaHdr.num_frames = htonl(1);
  mediaHdr.length_s = htonl(secs);
  mediaHdr.length_ns = htonl(nsecs);

  sparcHdr.dataFormat = htonl(rawHdr.dataFormat);
  sparcHdr.samplingRate = htonl(rawHdr.samplingRate);
  sparcHdr.channelCount = htonl(rawHdr.channelCount);

  /* Write Media Independent Header */

  if (Chunk_WriteHeader(cfState, CM_MEDIA_INDEP_CF, "") != 0) {
    fprintf(stderr, "Error writing media indep. header\n");
    exit(-1);
  }
  write(cfState->fd, &mediaHdr, (sizeof(MediaIndepHeader)));

  /* Write Au Header */

  if (Chunk_WriteHeader(cfState, CM_SPARC_AU_CF, "") != 0) {
    fprintf(stderr, "Error writing data chunk header.\n");
    exit(-1);
  }
  write(cfState->fd, &sparcHdr, (sizeof(SparcAuHeader)));

  /* Write frame offset table. */

  if (Chunk_WriteHeader(cfState, CM_FOT, "") != 0) {
    fprintf(stderr, "Error writing frame offset table chunk header.\n");
    exit(-1);
  }

  zero = 0;
  size = htonl(rawHdr.dataSize);

  write(cfState->fd, &zero, 4); 
  write(cfState->fd, &size, 4);

  /* Write frame time table. */

  if (Chunk_WriteHeader(cfState, CM_FTT, "") != 0) {
    fprintf(stderr, "Error writing frame time table chunk header.\n");
    exit(-1);
  }

  zero = 0;
  secs = htonl(secs);
  nsecs = htonl(nsecs);

  write(cfState->fd, &zero, 4); 
  write(cfState->fd, &zero, 4);
  write(cfState->fd, &secs, 4); 
  write(cfState->fd, &nsecs, 4);
  
  Chunk_CloseFile(cfState);
}


void
main(int argc, char **argv)
{
  char *out, *p;

  if (argc < 2 || argc > 3) {
    fprintf(stderr, "Usage: au2acf infile.au outfile.acf    \n");
    fprintf(stderr, "       au2acf infile.au                ");
    fprintf(stderr, "# output file defaults to infile.acf\n");
    fprintf(stderr, "       au2acf - outfile.acf            ");
    fprintf(stderr, "# input comes from stdin\n");
    fprintf(stderr, "       au2acf -                        ");
    fprintf(stderr, "# input comes from stdin,\n");
    fprintf(stderr, "                                         output file is 'out.acf'\n");
    exit(-1);
  }

  if (argc == 3) {
    out = argv[2];
  } 
  else if (strcmp(argv[1],"-") == 0) {
    out = "out.acf";
  } else {
    out = malloc(strlen(argv[1])+10);
    strcpy(out,argv[1]);
    p = strrchr(out,'.');
    if (p != NULL) {
      *p = '\0';
    }
    strcat(out,".acf");
  }

  SparcToAuFile( argv[1], out);
  exit(1);
}


