/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement
 * is hereby granted, provided that the above copyright notice and the
 * following two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
 * $Id: cmMPEGSoftwareDevice.c,v 1.29 1996/02/07 22:00:13 kpatel Exp $
 *
 * Implementation of the mpeg software device, which is used
 * by the cmt "mpegPlay" object to do software decoding and
 * display of mpeg video
 *
 */

#ifdef _WINDOWS
#    define WIN32_LEAN_AND_MEAN
#    include <windows.h>
#    include <winsock.h>
#    undef WIN32_LEAN_AND_MEAN
#endif /* _WINDOWS */

#include <stdio.h>
#include <assert.h>
#include <tcl.h>
#include <tk.h>
#include <dp.h>

#include <sys/types.h>
#include <netinet/in.h>

#include "cmMPEGPlay.h"

#ifdef MPEG_SW_DEVICE

#include "cmBitParse.h"
#include "cmMPEGSoftwareDevice.h"
#include "cmDither.h"
#include "cmLTS.h"
#include "cmUtil.h"

#ifdef _WINDOWS
extern HDC tcm_DitherHDC;
#endif

extern void j_rev_dct(DCTBLOCK data);
extern void j_rev_dct_sparse(DCTBLOCK data, int pos);

#define DECODE_ERROR 1
#define MB_DECODE_ERROR 2
#define DITHER_INIT_ERROR 3
#define PARSE_OK 0

#define SLICE_MIN_START_CODE 0x00000101
#define SLICE_MAX_START_CODE 0x000001af
#define MPEG_PICTURE_START_CODE 0x00000100
#define MPEG_I_FRAME 1
#define MPEG_P_FRAME 2
#define MPEG_B_FRAME 3
#define MPEG_D_FRAME 4
#define MPEG_EXTENSION_START_CODE 0x000001b5
#define MPEG_USER_START_CODE 0x000001b2

static int MPEG_SoftDev_Decode(MPEGFrame *framePtr);
static int MPEG_SoftDev_Refresh(MPEGPlay *objPtr);
static int MPEG_SoftDev_Display(MPEGFrame *framePtr);
static int MPEG_SoftDev_Init(MPEGPlay  *objPtr);
static int MPEG_SoftDev_Reinit(MPEGPlay *objPtr);
static int MPEG_SoftDev_Shutdown(MPEGPlay *objPtr);
static int MPEG_SoftDev_FreeFrame(MPEGFrame *framePtr);
static int MPEG_SoftDev_Config(MPEGPlay *objPtr, char *option, char *value);
static int MPEG_SoftDev_NextFrame(MPEGPlay *objPtr, MPEGFrame *frameList, 
				  MPEGFrame **chosenFrame);
static char *MPEG_SoftDev_Error(MPEGPlay *objPtr, int errNum);
static int ParsePicture(MPEGFrame *framePtr);
static int ParseSlice(MPEGFrame *framePtr);
static int ParseMacroBlock(MPEGFrame *framePtr);

extern Display *Tcm_Display();


typedef struct {
  int type;                              /* Frame type. */
  int full_pel_forw_vector;              /* Forw. vectors specified in full
					    pixel values flag.              */
  unsigned int forw_r_size;              /* Used for vector decoding.       */
  unsigned int forw_f;                   /* Used for vector decoding.       */
  int full_pel_back_vector;              /* Back vectors specified in full 
					    pixel values flag.              */
  unsigned int back_r_size;              /* Used in decoding.               */
  unsigned int back_f;                   /* Used in decoding.               */
} Pict;

typedef struct {
  unsigned int vert_pos;                 /* Vertical position of slice. */
  unsigned int quant_scale;              /* Quantization scale.         */
} Slice;

typedef struct {
  int mb_address;                        /* Macroblock address.              */
  int past_mb_addr;                      /* Previous mblock address.         */
  int motion_h_forw_code;                /* Forw. horiz. motion vector code. */
  unsigned int motion_h_forw_r;          /* Used in decoding vectors.        */
  int motion_v_forw_code;                /* Forw. vert. motion vector code.  */
  unsigned int motion_v_forw_r;          /* Used in decdoinge vectors.       */
  int motion_h_back_code;                /* Back horiz. motion vector code.  */
  unsigned int motion_h_back_r;          /* Used in decoding vectors.        */
  int motion_v_back_code;                /* Back vert. motion vector code.   */
  unsigned int motion_v_back_r;          /* Used in decoding vectors.        */
  unsigned int cbp;                      /* Coded block pattern.             */
  int mb_intra;                          /* Intracoded mblock flag.          */
  int bpict_past_forw;                   /* Past B frame forw. vector flag.  */
  int bpict_past_back;                   /* Past B frame back vector flag.   */
  int past_intra_addr;                   /* Addr of last intracoded mblock.  */
  int recon_right_for_prev;              /* Past right forw. vector.         */
  int recon_down_for_prev;               /* Past down forw. vector.          */
  int recon_right_back_prev;             /* Past right back vector.          */
  int recon_down_back_prev;              /* Past down back vector.           */
} Macroblock;

typedef struct {
  short int dct_recon[8][8];             /* Reconstructed dct coeff matrix. */
  short int dct_dc_y_past;               /* Past lum. dc dct coefficient.   */
  short int dct_dc_cr_past;              /* Past cr dc dct coefficient.     */
  short int dct_dc_cb_past;              /* Past cb dc dct coefficient.     */
} Block;

typedef struct {
    Pict picture;
    Slice slice;
    Macroblock mblock;
    Block block;
    int mb_width;
    int mb_height;
    int num_mblocks;
    int mblocks_remaining;
    Buffer *luminance;
    Buffer *cr;
    Buffer *cb;
    Buffer *display;
    unsigned char *curData;
    int offset;
    int errorFlag;
    double idle_start_time;
} MPEGSoftwareDecodeState;

typedef struct {
    double secPerMblock;
    double secPerDither;
    double secPerIdle;
    DitherRec *ditherRec;
    DitherFunc ditherProc;
    GC gc;
    Buffer *last_display;
#ifdef _WINDOWS
    BITMAPINFO *image;
    unsigned char *image_data;
#else
    XImage *image;
#endif
    int last_width;
    int last_height;
    
#ifdef HAVE_SHARED_MEMORY
    XShmSegmentInfo last_shminfo;
#endif
    int ShmEvent;
    int shared_mem_flag;
    int waiting;
    char ditherName[32];
    int ditherType;
} MPEGSoftwareDevice;

/* Decoding table for macroblock_address_increment */
static mb_addr_inc_entry     mb_addr_inc[2048];

/* Decoding table for macroblock_type in predictive-coded pictures */
static mb_type_entry         mb_type_P[64];

/* Decoding table for macroblock_type in bidirectionally-coded pictures */
static mb_type_entry         mb_type_B[64];

/* Decoding table for motion vectors */
static motion_vectors_entry  motion_vectors[2048];

/* Decoding table for coded_block_pattern */

static coded_block_pattern_entry coded_block_pattern[512] = 
{ {(unsigned int)MPG_ERROR, 0}, {(unsigned int)MPG_ERROR, 0}, {39, 9}, {27, 9}, {59, 9}, {55, 9}, {47, 9}, {31, 9},
    {58, 8}, {58, 8}, {54, 8}, {54, 8}, {46, 8}, {46, 8}, {30, 8}, {30, 8},
    {57, 8}, {57, 8}, {53, 8}, {53, 8}, {45, 8}, {45, 8}, {29, 8}, {29, 8},
    {38, 8}, {38, 8}, {26, 8}, {26, 8}, {37, 8}, {37, 8}, {25, 8}, {25, 8},
    {43, 8}, {43, 8}, {23, 8}, {23, 8}, {51, 8}, {51, 8}, {15, 8}, {15, 8},
    {42, 8}, {42, 8}, {22, 8}, {22, 8}, {50, 8}, {50, 8}, {14, 8}, {14, 8},
    {41, 8}, {41, 8}, {21, 8}, {21, 8}, {49, 8}, {49, 8}, {13, 8}, {13, 8},
    {35, 8}, {35, 8}, {19, 8}, {19, 8}, {11, 8}, {11, 8}, {7, 8}, {7, 8},
    {34, 7}, {34, 7}, {34, 7}, {34, 7}, {18, 7}, {18, 7}, {18, 7}, {18, 7},
    {10, 7}, {10, 7}, {10, 7}, {10, 7}, {6, 7}, {6, 7}, {6, 7}, {6, 7}, 
    {33, 7}, {33, 7}, {33, 7}, {33, 7}, {17, 7}, {17, 7}, {17, 7}, {17, 7}, 
    {9, 7}, {9, 7}, {9, 7}, {9, 7}, {5, 7}, {5, 7}, {5, 7}, {5, 7}, 
    {63, 6}, {63, 6}, {63, 6}, {63, 6}, {63, 6}, {63, 6}, {63, 6}, {63, 6}, 
    {3, 6}, {3, 6}, {3, 6}, {3, 6}, {3, 6}, {3, 6}, {3, 6}, {3, 6}, 
    {36, 6}, {36, 6}, {36, 6}, {36, 6}, {36, 6}, {36, 6}, {36, 6}, {36, 6}, 
    {24, 6}, {24, 6}, {24, 6}, {24, 6}, {24, 6}, {24, 6}, {24, 6}, {24, 6}, 
    {62, 5}, {62, 5}, {62, 5}, {62, 5}, {62, 5}, {62, 5}, {62, 5}, {62, 5},
    {62, 5}, {62, 5}, {62, 5}, {62, 5}, {62, 5}, {62, 5}, {62, 5}, {62, 5},
    {2, 5}, {2, 5}, {2, 5}, {2, 5}, {2, 5}, {2, 5}, {2, 5}, {2, 5}, 
    {2, 5}, {2, 5}, {2, 5}, {2, 5}, {2, 5}, {2, 5}, {2, 5}, {2, 5}, 
    {61, 5}, {61, 5}, {61, 5}, {61, 5}, {61, 5}, {61, 5}, {61, 5}, {61, 5}, 
    {61, 5}, {61, 5}, {61, 5}, {61, 5}, {61, 5}, {61, 5}, {61, 5}, {61, 5}, 
    {1, 5}, {1, 5}, {1, 5}, {1, 5}, {1, 5}, {1, 5}, {1, 5}, {1, 5}, 
    {1, 5}, {1, 5}, {1, 5}, {1, 5}, {1, 5}, {1, 5}, {1, 5}, {1, 5}, 
    {56, 5}, {56, 5}, {56, 5}, {56, 5}, {56, 5}, {56, 5}, {56, 5}, {56, 5}, 
    {56, 5}, {56, 5}, {56, 5}, {56, 5}, {56, 5}, {56, 5}, {56, 5}, {56, 5}, 
    {52, 5}, {52, 5}, {52, 5}, {52, 5}, {52, 5}, {52, 5}, {52, 5}, {52, 5}, 
    {52, 5}, {52, 5}, {52, 5}, {52, 5}, {52, 5}, {52, 5}, {52, 5}, {52, 5}, 
    {44, 5}, {44, 5}, {44, 5}, {44, 5}, {44, 5}, {44, 5}, {44, 5}, {44, 5}, 
    {44, 5}, {44, 5}, {44, 5}, {44, 5}, {44, 5}, {44, 5}, {44, 5}, {44, 5}, 
    {28, 5}, {28, 5}, {28, 5}, {28, 5}, {28, 5}, {28, 5}, {28, 5}, {28, 5}, 
    {28, 5}, {28, 5}, {28, 5}, {28, 5}, {28, 5}, {28, 5}, {28, 5}, {28, 5}, 
    {40, 5}, {40, 5}, {40, 5}, {40, 5}, {40, 5}, {40, 5}, {40, 5}, {40, 5}, 
    {40, 5}, {40, 5}, {40, 5}, {40, 5}, {40, 5}, {40, 5}, {40, 5}, {40, 5}, 
    {20, 5}, {20, 5}, {20, 5}, {20, 5}, {20, 5}, {20, 5}, {20, 5}, {20, 5}, 
    {20, 5}, {20, 5}, {20, 5}, {20, 5}, {20, 5}, {20, 5}, {20, 5}, {20, 5}, 
    {48, 5}, {48, 5}, {48, 5}, {48, 5}, {48, 5}, {48, 5}, {48, 5}, {48, 5}, 
    {48, 5}, {48, 5}, {48, 5}, {48, 5}, {48, 5}, {48, 5}, {48, 5}, {48, 5}, 
    {12, 5}, {12, 5}, {12, 5}, {12, 5}, {12, 5}, {12, 5}, {12, 5}, {12, 5}, 
    {12, 5}, {12, 5}, {12, 5}, {12, 5}, {12, 5}, {12, 5}, {12, 5}, {12, 5}, 
    {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, 
    {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, 
    {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, 
    {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, {32, 4}, 
    {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, 
    {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, 
    {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, 
    {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, {16, 4}, 
    {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, 
    {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, 
    {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, 
    {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4}, {8, 4},
    {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4},
    {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, 
    {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, 
    {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4}, {4, 4},
    {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, 
    {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, 
    {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, 
    {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, 
    {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, 
    {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, 
    {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, 
    {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}, {60, 3}
};

/* Decoding tables for dct_dc_size_luminance */
static dct_dc_size_entry dct_dc_size_luminance[32] =
{   {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, 
    {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, 
    {0, 3}, {0, 3}, {0, 3}, {0, 3}, {3, 3}, {3, 3}, {3, 3}, {3, 3}, 
    {4, 3}, {4, 3}, {4, 3}, {4, 3}, {5, 4}, {5, 4}, {6, 5}, {(unsigned int)MPG_ERROR, 0}
};

static dct_dc_size_entry dct_dc_size_luminance1[16] =
{   {7, 6}, {7, 6}, {7, 6}, {7, 6}, {7, 6}, {7, 6}, {7, 6}, {7, 6},
    {8, 7}, {8, 7}, {8, 7}, {8, 7}, {9, 8}, {9, 8}, {10, 9}, {11, 9}
};

/* Decoding table for dct_dc_size_chrominance */
static dct_dc_size_entry dct_dc_size_chrominance[32] =
{   {0, 2}, {0, 2}, {0, 2}, {0, 2}, {0, 2}, {0, 2}, {0, 2}, {0, 2}, 
    {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, 
    {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, 
    {3, 3}, {3, 3}, {3, 3}, {3, 3}, {4, 4}, {4, 4}, {5, 5}, {(unsigned int)MPG_ERROR, 0}
};

static dct_dc_size_entry dct_dc_size_chrominance1[32] =
{   {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, 
    {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, 
    {7, 7}, {7, 7}, {7, 7}, {7, 7}, {7, 7}, {7, 7}, {7, 7}, {7, 7}, 
    {8, 8}, {8, 8}, {8, 8}, {8, 8}, {9, 9}, {9, 9}, {10, 10}, {11, 10}
};

/* DCT coeff tables. */

static unsigned short int dct_coeff_tbl_0[256] =
{
0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 
0x052f, 0x051f, 0x050f, 0x04ff, 
0x183f, 0x402f, 0x3c2f, 0x382f, 
0x342f, 0x302f, 0x2c2f, 0x7c1f, 
0x781f, 0x741f, 0x701f, 0x6c1f, 
0x028e, 0x028e, 0x027e, 0x027e, 
0x026e, 0x026e, 0x025e, 0x025e, 
0x024e, 0x024e, 0x023e, 0x023e, 
0x022e, 0x022e, 0x021e, 0x021e, 
0x020e, 0x020e, 0x04ee, 0x04ee, 
0x04de, 0x04de, 0x04ce, 0x04ce, 
0x04be, 0x04be, 0x04ae, 0x04ae, 
0x049e, 0x049e, 0x048e, 0x048e, 
0x01fd, 0x01fd, 0x01fd, 0x01fd, 
0x01ed, 0x01ed, 0x01ed, 0x01ed, 
0x01dd, 0x01dd, 0x01dd, 0x01dd, 
0x01cd, 0x01cd, 0x01cd, 0x01cd, 
0x01bd, 0x01bd, 0x01bd, 0x01bd, 
0x01ad, 0x01ad, 0x01ad, 0x01ad, 
0x019d, 0x019d, 0x019d, 0x019d, 
0x018d, 0x018d, 0x018d, 0x018d, 
0x017d, 0x017d, 0x017d, 0x017d, 
0x016d, 0x016d, 0x016d, 0x016d, 
0x015d, 0x015d, 0x015d, 0x015d, 
0x014d, 0x014d, 0x014d, 0x014d, 
0x013d, 0x013d, 0x013d, 0x013d, 
0x012d, 0x012d, 0x012d, 0x012d, 
0x011d, 0x011d, 0x011d, 0x011d, 
0x010d, 0x010d, 0x010d, 0x010d, 
0x282c, 0x282c, 0x282c, 0x282c, 
0x282c, 0x282c, 0x282c, 0x282c, 
0x242c, 0x242c, 0x242c, 0x242c, 
0x242c, 0x242c, 0x242c, 0x242c, 
0x143c, 0x143c, 0x143c, 0x143c, 
0x143c, 0x143c, 0x143c, 0x143c, 
0x0c4c, 0x0c4c, 0x0c4c, 0x0c4c, 
0x0c4c, 0x0c4c, 0x0c4c, 0x0c4c, 
0x085c, 0x085c, 0x085c, 0x085c, 
0x085c, 0x085c, 0x085c, 0x085c, 
0x047c, 0x047c, 0x047c, 0x047c, 
0x047c, 0x047c, 0x047c, 0x047c, 
0x046c, 0x046c, 0x046c, 0x046c, 
0x046c, 0x046c, 0x046c, 0x046c, 
0x00fc, 0x00fc, 0x00fc, 0x00fc, 
0x00fc, 0x00fc, 0x00fc, 0x00fc, 
0x00ec, 0x00ec, 0x00ec, 0x00ec, 
0x00ec, 0x00ec, 0x00ec, 0x00ec, 
0x00dc, 0x00dc, 0x00dc, 0x00dc, 
0x00dc, 0x00dc, 0x00dc, 0x00dc, 
0x00cc, 0x00cc, 0x00cc, 0x00cc, 
0x00cc, 0x00cc, 0x00cc, 0x00cc, 
0x681c, 0x681c, 0x681c, 0x681c, 
0x681c, 0x681c, 0x681c, 0x681c, 
0x641c, 0x641c, 0x641c, 0x641c, 
0x641c, 0x641c, 0x641c, 0x641c, 
0x601c, 0x601c, 0x601c, 0x601c, 
0x601c, 0x601c, 0x601c, 0x601c, 
0x5c1c, 0x5c1c, 0x5c1c, 0x5c1c, 
0x5c1c, 0x5c1c, 0x5c1c, 0x5c1c, 
0x581c, 0x581c, 0x581c, 0x581c, 
0x581c, 0x581c, 0x581c, 0x581c, 
};

static unsigned short int dct_coeff_tbl_1[16] = 
{
0x00bb, 0x202b, 0x103b, 0x00ab, 
0x084b, 0x1c2b, 0x541b, 0x501b, 
0x009b, 0x4c1b, 0x481b, 0x045b, 
0x0c3b, 0x008b, 0x182b, 0x441b, 
};

static unsigned short int dct_coeff_tbl_2[4] =
{
0x4019, 0x1429, 0x0079, 0x0839, 
};

static unsigned short int dct_coeff_tbl_3[4] = 
{
0x0449, 0x3c19, 0x3819, 0x1029, 
};

static unsigned short int dct_coeff_next[256] = 
{
0xffff, 0xffff, 0xffff, 0xffff, 
0xf7d5, 0xf7d5, 0xf7d5, 0xf7d5, 
0x0826, 0x0826, 0x2416, 0x2416, 
0x0046, 0x0046, 0x2016, 0x2016, 
0x1c15, 0x1c15, 0x1c15, 0x1c15, 
0x1815, 0x1815, 0x1815, 0x1815, 
0x0425, 0x0425, 0x0425, 0x0425, 
0x1415, 0x1415, 0x1415, 0x1415, 
0x3417, 0x0067, 0x3017, 0x2c17, 
0x0c27, 0x0437, 0x0057, 0x2817, 
0x0034, 0x0034, 0x0034, 0x0034, 
0x0034, 0x0034, 0x0034, 0x0034, 
0x1014, 0x1014, 0x1014, 0x1014, 
0x1014, 0x1014, 0x1014, 0x1014, 
0x0c14, 0x0c14, 0x0c14, 0x0c14, 
0x0c14, 0x0c14, 0x0c14, 0x0c14, 
0x0023, 0x0023, 0x0023, 0x0023, 
0x0023, 0x0023, 0x0023, 0x0023, 
0x0023, 0x0023, 0x0023, 0x0023, 
0x0023, 0x0023, 0x0023, 0x0023, 
0x0813, 0x0813, 0x0813, 0x0813, 
0x0813, 0x0813, 0x0813, 0x0813, 
0x0813, 0x0813, 0x0813, 0x0813, 
0x0813, 0x0813, 0x0813, 0x0813, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0xfbe1, 0xfbe1, 0xfbe1, 0xfbe1, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
0x0011, 0x0011, 0x0011, 0x0011, 
};

static unsigned short int dct_coeff_first[256] = 
{
0xffff, 0xffff, 0xffff, 0xffff, 
0xf7d5, 0xf7d5, 0xf7d5, 0xf7d5, 
0x0826, 0x0826, 0x2416, 0x2416, 
0x0046, 0x0046, 0x2016, 0x2016, 
0x1c15, 0x1c15, 0x1c15, 0x1c15, 
0x1815, 0x1815, 0x1815, 0x1815, 
0x0425, 0x0425, 0x0425, 0x0425, 
0x1415, 0x1415, 0x1415, 0x1415, 
0x3417, 0x0067, 0x3017, 0x2c17, 
0x0c27, 0x0437, 0x0057, 0x2817, 
0x0034, 0x0034, 0x0034, 0x0034, 
0x0034, 0x0034, 0x0034, 0x0034, 
0x1014, 0x1014, 0x1014, 0x1014, 
0x1014, 0x1014, 0x1014, 0x1014, 
0x0c14, 0x0c14, 0x0c14, 0x0c14, 
0x0c14, 0x0c14, 0x0c14, 0x0c14, 
0x0023, 0x0023, 0x0023, 0x0023, 
0x0023, 0x0023, 0x0023, 0x0023, 
0x0023, 0x0023, 0x0023, 0x0023, 
0x0023, 0x0023, 0x0023, 0x0023, 
0x0813, 0x0813, 0x0813, 0x0813, 
0x0813, 0x0813, 0x0813, 0x0813, 
0x0813, 0x0813, 0x0813, 0x0813, 
0x0813, 0x0813, 0x0813, 0x0813, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0412, 0x0412, 0x0412, 0x0412, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
0x0010, 0x0010, 0x0010, 0x0010, 
};

/* Macro for filling up the decoding table for mb_addr_inc */
#define ASSIGN1(start, end, step, val, num) \
  for (i = start; i < end; i+= step) { \
    for (j = 0; j < step; j++) { \
      mb_addr_inc[i+j].value = val; \
      mb_addr_inc[i+j].num_bits = num; \
    } \
    val--; \
    }

static unsigned int default_intra_matrix[] = { 
    8, 16, 16, 19, 16, 19, 22, 22, 
    22, 22, 22, 22, 26, 24, 26, 27, 
    27, 27, 26, 26, 26, 26, 27, 27,
    27, 29, 29, 29, 34, 34, 34, 29,
    29, 29, 27, 27, 29, 29, 32, 32,
    34, 34, 37, 38, 37, 35, 35, 34,
    35, 38, 38, 40, 40, 40, 48, 48,
    46, 46, 56, 56, 58, 69, 69, 83};

static unsigned int default_non_intra_matrix[] = { 
    16, 16, 16, 16, 16, 16, 16, 16, 
    16, 16, 16, 16, 16, 16, 16, 16, 
    16, 16, 16, 16, 16, 16, 16, 16, 
    16, 16, 16, 16, 16, 16, 16, 16, 
    16, 16, 16, 16, 16, 16, 16, 16, 
    16, 16, 16, 16, 16, 16, 16, 16, 
    16, 16, 16, 16, 16, 16, 16, 16, 
    16, 16, 16, 16, 16, 16, 16, 16};
 
static int zigzag_direct[64] = {
    0, 1, 8, 16, 9, 2, 3, 10, 17, 24, 32, 25, 18, 11, 4, 5, 12,
    19, 26, 33, 40, 48, 41, 34, 27, 20, 13, 6, 7, 14, 21, 28, 35,
    42, 49, 56, 57, 50, 43, 36, 29, 22, 15, 23, 30, 37, 44, 51,
    58, 59, 52, 45, 38, 31, 39, 46, 53, 60, 61, 54, 47, 55, 62, 63};

static unsigned int bitMask[] = 
                         {  0xffffffff, 0x7fffffff, 0x3fffffff, 0x1fffffff, 
			    0x0fffffff, 0x07ffffff, 0x03ffffff, 0x01ffffff,
			    0x00ffffff, 0x007fffff, 0x003fffff, 0x001fffff,
			    0x000fffff, 0x0007ffff, 0x0003ffff, 0x0001ffff,
			    0x0000ffff, 0x00007fff, 0x00003fff, 0x00001fff,
			    0x00000fff, 0x000007ff, 0x000003ff, 0x000001ff,
			    0x000000ff, 0x0000007f, 0x0000003f, 0x0000001f,
			    0x0000000f, 0x00000007, 0x00000003, 0x00000001};

static unsigned int bitTest[] = {  
    0x80000000, 0x40000000, 0x20000000, 0x10000000, 
    0x08000000, 0x04000000, 0x02000000, 0x01000000,
    0x00800000, 0x00400000, 0x00200000, 0x00100000,
    0x00080000, 0x00040000, 0x00020000, 0x00010000,
    0x00008000, 0x00004000, 0x00002000, 0x00001000,
    0x00000800, 0x00000400, 0x00000200, 0x00000100,
    0x00000080, 0x00000040, 0x00000020, 0x00000010,
    0x00000008, 0x00000004, 0x00000002, 0x00000001};

static unsigned int rBitMask[] = { 
    0xffffffff, 0xfffffffe, 0xfffffffc, 0xfffffff8, 
    0xfffffff0, 0xffffffe0, 0xffffffc0, 0xffffff80, 
    0xffffff00, 0xfffffe00, 0xfffffc00, 0xfffff800, 
    0xfffff000, 0xffffe000, 0xffffc000, 0xffff8000, 
    0xffff0000, 0xfffe0000, 0xfffc0000, 0xfff80000, 
    0xfff00000, 0xffe00000, 0xffc00000, 0xff800000, 
    0xff000000, 0xfe000000, 0xfc000000, 0xf8000000, 
    0xf0000000, 0xe0000000, 0xc0000000, 0x80000000};

/*
 * We use a lookup table to make sure values stay in the 0..255 range.
 * Since this is cropping (ie, x = (x < 0)?0:(x>255)?255:x; ), wee call this
 * table the "crop table".
 * MAX_NEG_CROP is the maximum neg/pos value we can handle.
 */

#define MAX_NEG_CROP 2048
#define NUM_CROP_ENTRIES (2048+2*MAX_NEG_CROP)
static unsigned char cropTbl[NUM_CROP_ENTRIES];

#define NextStartCode()                        \
{                                              \
    unsigned int data;                         \
    int flush_num;                             \
                                               \
    flush_num = ((8-(bp_bitOffset%8))%8);      \
    bp_Flush(flush_num);                       \
    bp_Peek(24, &data);                        \
                                               \
    while(data != 1) {                         \
	bp_Flush(8);                           \
	bp_Peek(24, &data);                    \
    }                                          \
} 

#define ParseExtraBitInfo()                    \
{                                              \
    unsigned int data;                         \
                                               \
    bp_Get(1, &data);                          \
                                               \
    while(data == 1) {                         \
        bp_Flush(8);                           \
        bp_Get(1, &data);                      \
    }                                          \
}


/*
 *--------------------------------------------------------------
 *
 * init_mb_addr_inc --
 *
 *	Initialize the VLC decoding table for macro_block_address_increment
 *
 * Results:
 *	The decoding table for macro_block_address_increment will
 *      be filled; illegal values will be filled as MPG_ERROR.
 *
 * Side effects:
 *	The global array mb_addr_inc will be filled.
 *
 *--------------------------------------------------------------
 */
static void
init_mb_addr_inc()
{
  int i, j, val;

  for (i = 0; i < 8; i++) {
    mb_addr_inc[i].value = MPG_ERROR;
    mb_addr_inc[i].num_bits = 0;
  }

  mb_addr_inc[8].value = MACRO_BLOCK_ESCAPE;
  mb_addr_inc[8].num_bits = 11;

  for (i = 9; i < 15; i++) {
    mb_addr_inc[i].value = MPG_ERROR;
    mb_addr_inc[i].num_bits = 0;
  }

  mb_addr_inc[15].value = MACRO_BLOCK_STUFFING;
  mb_addr_inc[15].num_bits = 11;

  for (i = 16; i < 24; i++) {
    mb_addr_inc[i].value = MPG_ERROR;
    mb_addr_inc[i].num_bits = 0;
  }

  val = 33;

  ASSIGN1(24, 36, 1, val, 11);
  ASSIGN1(36, 48, 2, val, 10);
  ASSIGN1(48, 96, 8, val, 8);
  ASSIGN1(96, 128, 16, val, 7);
  ASSIGN1(128, 256, 64, val, 5);
  ASSIGN1(256, 512, 128, val, 4);
  ASSIGN1(512, 1024, 256, val, 3);
  ASSIGN1(1024, 2048, 1024, val, 1);
}


/* Macro for filling up the decoding table for mb_type */
#define ASSIGN2(start, end, quant, motion_forward, motion_backward, pattern, intra, num, mb_type) \
  for (i = start; i < end; i ++) { \
    mb_type[i].mb_quant = quant; \
    mb_type[i].mb_motion_forward = motion_forward; \
    mb_type[i].mb_motion_backward = motion_backward; \
    mb_type[i].mb_pattern = pattern; \
    mb_type[i].mb_intra = intra; \
    mb_type[i].num_bits = num; \
  }
	 


/*
 *--------------------------------------------------------------
 *
 * init_mb_type_P --
 *
 *	Initialize the VLC decoding table for macro_block_type in
 *      predictive-coded pictures.
 *
 * Results:
 *	The decoding table for macro_block_type in predictive-coded
 *      pictures will be filled; illegal values will be filled as MPG_ERROR.
 *
 * Side effects:
 *	The global array mb_type_P will be filled.
 *
 *--------------------------------------------------------------
 */
static void
init_mb_type_P()
{
  int i;

  mb_type_P[0].mb_quant = mb_type_P[0].mb_motion_forward 
    = mb_type_P[0].mb_motion_backward = mb_type_P[0].mb_pattern 
      = mb_type_P[0].mb_intra = MPG_ERROR;
  mb_type_P[0].num_bits = 0;

  ASSIGN2(1, 2, 1, 0, 0, 0, 1, 6, mb_type_P)
  ASSIGN2(2, 4, 1, 0, 0, 1, 0, 5, mb_type_P)
  ASSIGN2(4, 6, 1, 1, 0, 1, 0, 5, mb_type_P);
  ASSIGN2(6, 8, 0, 0, 0, 0, 1, 5, mb_type_P);
  ASSIGN2(8, 16, 0, 1, 0, 0, 0, 3, mb_type_P);
  ASSIGN2(16, 32, 0, 0, 0, 1, 0, 2, mb_type_P);
  ASSIGN2(32, 64, 0, 1, 0, 1, 0, 1, mb_type_P);
}




/*
 *--------------------------------------------------------------
 *
 * init_mb_type_B --
 *
 *	Initialize the VLC decoding table for macro_block_type in
 *      bidirectionally-coded pictures.
 *
 * Results:
 *	The decoding table for macro_block_type in bidirectionally-coded
 *      pictures will be filled; illegal values will be filled as MPG_ERROR.
 *
 * Side effects:
 *	The global array mb_type_B will be filled.
 *
 *--------------------------------------------------------------
 */
static void
init_mb_type_B()
{
  int i;

  mb_type_B[0].mb_quant = mb_type_B[0].mb_motion_forward 
    = mb_type_B[0].mb_motion_backward = mb_type_B[0].mb_pattern 
      = mb_type_B[0].mb_intra = MPG_ERROR;
  mb_type_B[0].num_bits = 0;

  ASSIGN2(1, 2, 1, 0, 0, 0, 1, 6, mb_type_B);
  ASSIGN2(2, 3, 1, 0, 1, 1, 0, 6, mb_type_B);
  ASSIGN2(3, 4, 1, 1, 0, 1, 0, 6, mb_type_B);
  ASSIGN2(4, 6, 1, 1, 1, 1, 0, 5, mb_type_B);
  ASSIGN2(6, 8, 0, 0, 0, 0, 1, 5, mb_type_B);
  ASSIGN2(8, 12, 0, 1, 0, 0, 0, 4, mb_type_B);
  ASSIGN2(12, 16, 0, 1, 0, 1, 0, 4, mb_type_B);
  ASSIGN2(16, 24, 0, 0, 1, 0, 0, 3, mb_type_B);
  ASSIGN2(24, 32, 0, 0, 1, 1, 0, 3, mb_type_B);
  ASSIGN2(32, 48, 0, 1, 1, 0, 0, 2, mb_type_B);
  ASSIGN2(48, 64, 0, 1, 1, 1, 0, 2, mb_type_B);
}


/* Macro for filling up the decoding tables for motion_vectors */
#define ASSIGN3(start, end, step, val, num) \
  for (i = start; i < end; i+= step) { \
    for (j = 0; j < step / 2; j++) { \
      motion_vectors[i+j].code = val; \
      motion_vectors[i+j].num_bits = num; \
    } \
    for (j = step / 2; j < step; j++) { \
      motion_vectors[i+j].code = -val; \
      motion_vectors[i+j].num_bits = num; \
    } \
    val--; \
  }



/*
 *--------------------------------------------------------------
 *
 * init_motion_vectors --
 *
 *	Initialize the VLC decoding table for the various motion
 *      vectors, including motion_horizontal_forward_code, 
 *      motion_vertical_forward_code, motion_horizontal_backward_code,
 *      and motion_vertical_backward_code.
 *
 * Results:
 *	The decoding table for the motion vectors will be filled;
 *      illegal values will be filled as MPG_ERROR.
 *
 * Side effects:
 *	The global array motion_vector will be filled.
 *
 *--------------------------------------------------------------
 */
static void
init_motion_vectors()
{
  int i, j, val = 16;

  for (i = 0; i < 24; i++) {
    motion_vectors[i].code = MPG_ERROR;
    motion_vectors[i].num_bits = 0;
  }

  ASSIGN3(24, 36, 2, val, 11);
  ASSIGN3(36, 48, 4, val, 10);
  ASSIGN3(48, 96, 16, val, 8);
  ASSIGN3(96, 128, 32, val, 7);
  ASSIGN3(128, 256, 128, val, 5);
  ASSIGN3(256, 512, 256, val, 4);
  ASSIGN3(512, 1024, 512, val, 3);
  ASSIGN3(1024, 2048, 1024, val, 1);
}




/*
 *--------------------------------------------------------------
 *
 * init_tables --
 *
 *	Initialize all the tables for VLC decoding; this must be
 *      called when the system is set up before any decoding can
 *      take place.
 *
 * Results:
 *	All the decoding tables will be filled accordingly.
 *
 * Side effects:
 *	The corresponding global array for each decoding table 
 *      will be filled.
 *
 *--------------------------------------------------------------
 */    
static void
init_tables()
{
  static int inited=0;
  extern void init_pre_idct();
  int i;

  if (!inited) {
    init_mb_addr_inc();
    init_mb_type_P();
    init_mb_type_B();
    init_motion_vectors();
    init_pre_idct();

    /* Initialize crop table. */
    
    for (i = (-MAX_NEG_CROP); i < NUM_CROP_ENTRIES - MAX_NEG_CROP; i++) {
      if (i <= 0) {
	cropTbl[i + MAX_NEG_CROP] = 0;
#ifdef TWELVE_BIT
      } else if (i >= 2047) {
	cropTbl[i + MAX_NEG_CROP] = 2047;
#endif
      } else if (i >= 255) {
	cropTbl[i + MAX_NEG_CROP] = 255;
      } else {
	cropTbl[i + MAX_NEG_CROP] = i;
      }
    }
    
    inited = 1;
  }
}

/*
 *--------------------------------------------------------------
 *
 * DecodeDCTDCSizeLum --
 *
 *	Huffman Decoder for dct_dc_size_luminance; location where
 *      the result of decoding will be placed is passed as argument.
 *      The decoded values are obtained by doing a table lookup on
 *      dct_dc_size_luminance.
 *
 * Results:
 *	The decoded value for dct_dc_size_luminance or MPG_ERROR for 
 *      unbound values will be placed in the location specified.
 *
 * Side effects:
 *	Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */        
void
decodeDCTDCSizeLum(value)
unsigned int *value;
{
  unsigned int index;

  bp_Peek(5, &index);
  
  if (index < 31) {
  	*value = dct_dc_size_luminance[index].value;
  	bp_Flush(dct_dc_size_luminance[index].num_bits);
  }
  else {
	bp_Peek(9, &index);
	index -= 0x1f0;
	*value = dct_dc_size_luminance1[index].value;
	bp_Flush(dct_dc_size_luminance1[index].num_bits);
  }
}




/*
 *--------------------------------------------------------------
 *
 * DecodeDCTDCSizeChrom --
 *
 *	Huffman Decoder for dct_dc_size_chrominance; location where
 *      the result of decoding will be placed is passed as argument.
 *      The decoded values are obtained by doing a table lookup on
 *      dct_dc_size_chrominance.
 *
 * Results:
 *	The decoded value for dct_dc_size_chrominance or MPG_ERROR for
 *      unbound values will be placed in the location specified.
 *
 * Side effects:
 *	Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */    
void    
decodeDCTDCSizeChrom(value)
unsigned int *value;
{
  unsigned int index;

  bp_Peek(5, &index);
  
  if (index < 31) {
  	*value = dct_dc_size_chrominance[index].value;
  	bp_Flush(dct_dc_size_chrominance[index].num_bits);
  }
  else {
	bp_Peek(10, &index);
	index -= 0x3e0;
	*value = dct_dc_size_chrominance1[index].value;
	bp_Flush(dct_dc_size_chrominance1[index].num_bits);
  }
}



/*
 *--------------------------------------------------------------
 *
 * decodeDCTCoeff --
 *
 *	Huffman Decoder for dct_coeff_first and dct_coeff_next;
 *      locations where the results of decoding: run and level, are to
 *      be placed and also the type of DCT coefficients, either
 *      dct_coeff_first or dct_coeff_next, are being passed as argument.
 *      
 *      The decoder first examines the next 8 bits in the input stream,
 *      and perform according to the following cases:
 *      
 *      '0000 0000' - examine 8 more bits (i.e. 16 bits total) and
 *                    perform a table lookup on dct_coeff_tbl_0.
 *                    One more bit is then examined to determine the sign
 *                    of level.
 *
 *      '0000 0001' - examine 4 more bits (i.e. 12 bits total) and 
 *                    perform a table lookup on dct_coeff_tbl_1.
 *                    One more bit is then examined to determine the sign
 *                    of level.
 *      
 *      '0000 0010' - examine 2 more bits (i.e. 10 bits total) and
 *                    perform a table lookup on dct_coeff_tbl_2.
 *                    One more bit is then examined to determine the sign
 *                    of level.
 *
 *      '0000 0011' - examine 2 more bits (i.e. 10 bits total) and 
 *                    perform a table lookup on dct_coeff_tbl_3.
 *                    One more bit is then examined to determine the sign
 *                    of level.
 *
 *      otherwise   - perform a table lookup on dct_coeff_tbl. If the
 *                    value of run is not ESCAPE, extract one more bit
 *                    to determine the sign of level; otherwise 6 more
 *                    bits will be extracted to obtain the actual value 
 *                    of run , and then 8 or 16 bits to get the value of level.
 *                    
 *      
 *
 * Results:
 *	The decoded values of run and level or MPG_ERROR for unbound values
 *      are placed in the locations specified.
 *
 * Side effects:
 *	Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

static void
decodeDCTCoeff(dct_coeff_tbl, run, level)
unsigned short int *dct_coeff_tbl;                                       
unsigned int *run;
int *level;
{
  unsigned int temp, index;
  unsigned int value, next32bits, flushed;

  /*
   * Grab the next 32 bits and use it to improve performance of
   * getting the bits to parse. Thus, calls are translated as:
   *
   *	show_bitsX  <-->   next32bits >> (32-X)
   *	get_bitsX   <-->   val = next32bits >> (32-flushed-X);
   *			   flushed += X;
   *			   next32bits &= bitMask[flushed];
   *	flush_bitsX <-->   flushed += X;
   *			   next32bits &= bitMask[flushed];
   *
   */
  bp_Peek(32, &next32bits);
  flushed = 0;

  index = next32bits >> 24;

  if (index > 3) {
    value = dct_coeff_tbl[index];
    *run = (value & RUN_MASK) >> RUN_SHIFT;
    if (*run == END_OF_BLOCK) {
      *level = END_OF_BLOCK;
    }
    else {
      flushed = (value & NUM_MASK) + 1;
      next32bits &= bitMask[flushed];
      if (*run != ESCAPE) {
         *level = (value & LEVEL_MASK) >> LEVEL_SHIFT;
	 /* if (value) *level = -*level; */
	 if (next32bits >> (31-flushed)) *level = -*level;
	 flushed++;
	 /* next32bits &= bitMask[flushed];  last op before update */
       }
       else {    /* *run == ESCAPE */
	 temp = next32bits >> (18-flushed);
	 flushed += 14;
	 next32bits &= bitMask[flushed];
	 *run = temp >> 8;
	 temp &= 0xff;
	 if (temp == 0) {
	    *level = next32bits >> (24-flushed);
	    flushed += 8;
	    /* next32bits &= bitMask[flushed];  last op before update */
 	    assert(*level >= 128);
	 } else if (temp != 128) {
	    /* Grab sign bit */
	    *level = ((int) (temp << 24)) >> 24;
	 } else {
	    *level = next32bits >> (24-flushed);
	    flushed += 8;
	    /* next32bits &= bitMask[flushed];  last op before update */
	    *level = *level - 256;
	    assert(*level <= -128 && *level >= -255);
	 }
       }
       /* Update bitstream... */
       bp_Flush(flushed);
    }
  }
  else {
    if (index == 2) { 
      index = next32bits >> 22;
      value = dct_coeff_tbl_2[index & 3];
    }
    else if (index == 3) { 
      index = next32bits >> 22;
      value = dct_coeff_tbl_3[index & 3];
    }
    else if (index) {	/* index == 1 */
      index = next32bits >> 20;
      value = dct_coeff_tbl_1[index & 15];
    }
    else {   /* index == 0 */
      index = next32bits >> 16;
      value = dct_coeff_tbl_0[index & 255];
    }
    *run = (value & RUN_MASK) >> RUN_SHIFT;
    *level = (value & LEVEL_MASK) >> LEVEL_SHIFT;

    /*
     * Fold these operations together to make it fast...
     */
    /* num_bits = (value & NUM_MASK) + 1; */
    /* if (value) *level = -*level; */

    flushed = (value & NUM_MASK) + 2;
    if ((next32bits >> (32-flushed)) & 0x1) *level = -*level;

    /* Update bitstream ... */
    bp_Flush(flushed);
  }
}


/*
 *--------------------------------------------------------------
 *
 * decodeDCTCoeffFirst --
 *
 *	Huffman Decoder for dct_coeff_first. Locations for the
 *      decoded results: run and level, are being passed as
 *      arguments. Actual work is being done by calling DecodeDCTCoeff,
 *      with the table dct_coeff_first.
 *
 * Results:
 *	The decoded values of run and level for dct_coeff_first or
 *      MPG_ERROR for unbound values are placed in the locations given.
 *
 * Side effects:
 *	Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */        
void
decodeDCTCoeffFirst(run, level)
unsigned int *run;
int *level;
{
  decodeDCTCoeff(dct_coeff_first, run, level);
}




/*
 *--------------------------------------------------------------
 *
 * decodeDCTCoeffNext --
 *
 *	Huffman Decoder for dct_coeff_first. Locations for the
 *      decoded results: run and level, are being passed as
 *      arguments. Actual work is being done by calling DecodeDCTCoeff,
 *      with the table dct_coeff_next.
 *
 * Results:
 *	The decoded values of run and level for dct_coeff_next or
 *      MPG_ERROR for unbound values are placed in the locations given.
 *
 * Side effects:
 *	Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */ 
void       
decodeDCTCoeffNext(run, level)
unsigned int *run;
int *level;
{
  decodeDCTCoeff(dct_coeff_next, run, level);
}


/*
 *--------------------------------------------------------------
 *
 * ReconIMBlock --
 *
 *	Reconstructs intra coded macroblock.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
#ifndef NDEBUG
/* If people really want to see such things, check 'em */
#define myassert(x,expression)\
  if (!(expression)) {\
  fprintf (stderr,"Bad crop value (%d) at line %d\n", x, __LINE__);\
  NextStartCode(); return;}
#define assertCrop(x)	myassert(x,((x) >= -MAX_NEG_CROP) && \
				 ((x) <= 2048+MAX_NEG_CROP))
#else
#define assertCrop(x)
#endif
static void
ReconIMBlock(MPEGFrame *framePtr, int bnum)
{ 
  int mb_row, mb_col, row, col, row_size, rr;
  unsigned char *dest;
  MPEGSoftwareDecodeState *decodeState = 
    (MPEGSoftwareDecodeState *) framePtr->deviceData;

  /* Calculate macroblock row and column from address. */

  mb_row = decodeState->mblock.mb_address / decodeState->mb_width;
  mb_col = decodeState->mblock.mb_address % decodeState->mb_width;


  /* If block is luminance block... */

  if (bnum < 4) {

    /* Calculate row and col values for upper left pixel of block. */

    row = mb_row * 16;
    col = mb_col * 16;
    if (bnum > 1)
      row += 8;
    if (bnum % 2)
      col += 8;

    /* Set dest to luminance plane of current pict image. */

    dest = (unsigned char *)decodeState->luminance->data;

    /* Establish row size. */

    row_size = decodeState->mb_width * 16;
  }
  /* Otherwise if block is Cr block... */
  /* Cr first because of the earlier mixup */

  else if (bnum == 5) {

    /* Set dest to Cr plane of current pict image. */

    dest = (unsigned char *)decodeState->cr->data;

    /* Establish row size. */

    row_size = decodeState->mb_width * 8;

    /* Calculate row,col for upper left pixel of block. */

    row = mb_row * 8;
    col = mb_col * 8;
  }
  /* Otherwise block is Cb block, and ... */

  else {

    /* Set dest to Cb plane of current pict image. */

    dest = (unsigned char *)decodeState->cb->data;

    /* Establish row size. */

    row_size = decodeState->mb_width * 8;

    /* Calculate row,col for upper left pixel value of block. */

    row = mb_row * 8;
    col = mb_col * 8;
  }

  /*
   * For each pixel in block, set to cropped reconstructed value from inverse
   * dct.
   */
  {
    short *sp = &decodeState->block.dct_recon[0][0];
    unsigned char *cm = cropTbl + MAX_NEG_CROP;
    dest += row * row_size + col;
    for (rr = 0; rr < 4; rr++, sp += 16, dest += row_size) {
      dest[0] = cm[sp[0]];
      assertCrop(sp[0]);
      dest[1] = cm[sp[1]];
      assertCrop(sp[1]);
      dest[2] = cm[sp[2]];
      assertCrop(sp[2]);
      dest[3] = cm[sp[3]];
      assertCrop(sp[3]);
      dest[4] = cm[sp[4]];
      assertCrop(sp[4]);
      dest[5] = cm[sp[5]];
      assertCrop(sp[5]);
      dest[6] = cm[sp[6]];
      assertCrop(sp[6]);
      dest[7] = cm[sp[7]];
      assertCrop(sp[7]);

      dest += row_size;
      dest[0] = cm[sp[8]];
      assertCrop(sp[8]);
      dest[1] = cm[sp[9]];
      assertCrop(sp[9]);
      dest[2] = cm[sp[10]];
      assertCrop(sp[10]);
      dest[3] = cm[sp[11]];
      assertCrop(sp[11]);
      dest[4] = cm[sp[12]];
      assertCrop(sp[12]);
      dest[5] = cm[sp[13]];
      assertCrop(sp[13]);
      dest[6] = cm[sp[14]];
      assertCrop(sp[14]);
      dest[7] = cm[sp[15]];
      assertCrop(sp[15]);
    }
  }
}
 

/*
 *--------------------------------------------------------------
 *
 * ReconPMBlock --
 *
 *	Reconstructs forward predicted macroblocks.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static void
ReconPMBlock(MPEGFrame *framePtr, 
	     int bnum, 
	     int recon_right_for, 
	     int recon_down_for, 
	     int zflag)
{
  int mb_row, mb_col, row, col, row_size, rr;
  unsigned char *dest, *past;
  static int right_for, down_for, right_half_for, down_half_for;
  unsigned char *rindex1, *rindex2;
  unsigned char *index;
  short int *blockvals;
  MPEGSoftwareDecodeState *decodeState;

  decodeState = (MPEGSoftwareDecodeState *) framePtr->deviceData;
  
#ifdef LOOSE_MPEG
  int maxx, maxy, cc;
  int illegalBlock = 0;
  int row_start, row_end, rfirst, rlast, col_start, col_end, cfirst, clast;
#endif
  
  /* Calculate macroblock row and column from address. */
  
  mb_row = decodeState->mblock.mb_address / decodeState->mb_width;
  mb_col = decodeState->mblock.mb_address % decodeState->mb_width;
  
  if (bnum < 4) {
    
    /* Calculate right_for, down_for motion vectors. */
    
    right_for = recon_right_for >> 1;
    down_for = recon_down_for >> 1;
    right_half_for = recon_right_for & 0x1;
    down_half_for = recon_down_for & 0x1;
    
    /* Set dest to luminance plane of current pict image. */
    
    dest = (unsigned char *)decodeState->luminance->data;
    
    if (framePtr->ref1 != NULL) {
        MPEGSoftwareDecodeState *state;
        state = (MPEGSoftwareDecodeState *)framePtr->ref1->deviceData;
        past = (unsigned char *)state->luminance->data;
    }
    
    /* Establish row size. */
    
    row_size = decodeState->mb_width << 4;
    
    /* Calculate row,col of upper left pixel in block. */
    
    row = mb_row << 4;
    col = mb_col << 4;
    if (bnum > 1)
      row += 8;
    if (bnum % 2)
      col += 8;
    
#ifdef LOOSE_MPEG
    /* Check for block illegality. */
    
    maxx = lmaxx; maxy = lmaxy;
    
    if (row + down_for + 7 > maxy) illegalBlock |= 0x4;
    else if (row + down_for < 0)  illegalBlock |= 0x1;
    
    if (col + right_for + 7 > maxx) illegalBlock |= 0x2;
    else if (col + right_for < 0) illegalBlock |= 0x8;
    
#endif
  }
  /* Otherwise, block is NOT luminance block, ... */
  
  else {

    /* Construct motion vectors. */
    
    recon_right_for /= 2;
    recon_down_for /= 2;
    right_for = recon_right_for >> 1;
    down_for = recon_down_for >> 1;
    right_half_for = recon_right_for & 0x1;
    down_half_for = recon_down_for & 0x1;
    
    /* Establish row size. */
    
    row_size = decodeState->mb_width << 3;
    
    /* Calculate row,col of upper left pixel in block. */
    
    row = mb_row << 3;
    col = mb_col << 3;
    
#ifdef LOOSE_MPEG
    /* Check for block illegality. */
    
    maxx = cmaxx; maxy = cmaxy;
    
    if (row + down_for  + 7 > maxy) illegalBlock |= 0x4;
    else if (row + down_for < 0) illegalBlock |= 0x1;
    
    if (col + right_for  + 7 > maxx) illegalBlock  |= 0x2;
    else if (col + right_for < 0) illegalBlock |= 0x8;
    
#endif
    
    /* If block is Cr block... */
    /* 5 first because order was mixed up in earlier versions */
    
    if (bnum == 5) {

      /* Set dest to Cr plane of current pict image. */
      
      dest = (unsigned char *)decodeState->cr->data;
      
      if (framePtr->ref1 != NULL) {
          MPEGSoftwareDecodeState *state;
          state = (MPEGSoftwareDecodeState *)framePtr->ref1->deviceData;
          past = (unsigned char *)state->cr->data;
      }
    }
    /* Otherwise, block is Cb block... */
    
    else {
      
      /* Set dest to Cb plane of current pict image. */
      
      dest = (unsigned char *)decodeState->cb->data;
      
      if (framePtr->ref1 != NULL) {
          MPEGSoftwareDecodeState *state;
          state = (MPEGSoftwareDecodeState *)framePtr->ref1->deviceData;
          past = (unsigned char *)state->cb->data;
      }
    }
  }
  
  /* For each pixel in block... */
  
#ifdef LOOSE_MPEG
  
  if (illegalBlock) {
    if (illegalBlock & 0x1) {
      row_start = 0;
      row_end = row+down_for+8;
      rfirst = rlast = 8 - row_end;
    }
    else if (illegalBlock & 0x4) {
      row_start = row + down_for;
      row_end = maxy+1;
      rlast = row_end - row_start - 1;
      rfirst = 0;
    }
    else {
      row_start = row+down_for;
      row_end = row_start+8;
      rfirst = 0;
    }
    
    if (illegalBlock & 0x8) {
      col_start = 0;
      col_end = col + right_for + 8;
      cfirst = clast = 8 - col_end;
    }
    else if (illegalBlock & 0x2) {
      col_start = col + right_for;
      col_end = maxx + 1;
      clast = col_end - col_start - 1;
      cfirst = 0;
    }
    else {
      col_start = col + right_for;
      col_end = col_start + 8;
      cfirst = 0;
    }

    for (rr = row_start; rr < row_end; rr++) {
      rindex1 = past + (rr * row_size) + col_start;
      index = dest + ((row + rfirst) * row_size) + col + cfirst;
      for (cc = col_start; cc < col_end; cc++) {
	*index++ = *rindex1++;
      }
    }

    if (illegalBlock & 0x1) {
      for (rr = rlast -1; rr >=0; rr--) {
	index = dest + ((row + rr) * row_size) + col;
	rindex1 = dest + ((row + rlast) * row_size) + col;
	for (cc = 0; cc < 8; cc ++) {
	  *index++ = *rindex1++;
	}
      }
    }
    else if (illegalBlock & 0x4) {
      for (rr = rlast+1; rr < 8; rr++) {
	index = dest + ((row + rr) * row_size) + col;
	rindex1 = dest + ((row + rlast) * row_size) + col;
	for (cc = 0; cc < 8; cc ++) {
	  *index++ = *rindex1++;
	}
      }
    }

    if (illegalBlock & 0x2) {
      for (cc = clast+1; cc < 8; cc++) {
	index = dest + (row * row_size) + (col + cc);
	rindex1 = dest + (row * row_size) + (col + clast);
	for (rr = 0; rr < 8; rr++) {
	  *index = *rindex1;
	  index += row_size;
	  rindex1 += row_size;
	}
      }
    }
    else if (illegalBlock & 0x8) {
      for (cc = clast-1; cc >= 0; cc--) {
	index = dest + (row * row_size) + (col + cc);
	rindex1 = dest + (row * row_size) + (col + clast);
	for (rr = 0; rr < 8; rr++) {
	  *index = *rindex1;
	  index += row_size;
	  rindex1 += row_size;
	}
      }
    }

    if (!zflag) {
      for (rr = 0; rr < 8; rr++) {
	index = dest + (row*row_size) + col;
	blockvals = &(vid_stream->block.dct_recon[rr][0]);
	index[0] += blockvals[0];
	index[1] += blockvals[1];
	index[2] += blockvals[2];
	index[3] += blockvals[3];
	index[4] += blockvals[4];
	index[5] += blockvals[5];
	index[6] += blockvals[6];
	index[7] += blockvals[7];
      }
    }
  }
  else {

#endif

    index = dest + (row * row_size) + col;
    rindex1 = past + (row + down_for) * row_size + col + right_for;
    
    blockvals = &(decodeState->block.dct_recon[0][0]);
    
    /*
     * Calculate predictive pixel value based on motion vectors and copy to
     * dest plane.
     */
    
    if ((!down_half_for) && (!right_half_for)) {
      unsigned char *cm = cropTbl + MAX_NEG_CROP;
      if (!zflag)
	for (rr = 0; rr < 4; rr++) {
	  index[0] = cm[(int) rindex1[0] + (int) blockvals[0]];
	  index[1] = cm[(int) rindex1[1] + (int) blockvals[1]];
	  index[2] = cm[(int) rindex1[2] + (int) blockvals[2]];
	  index[3] = cm[(int) rindex1[3] + (int) blockvals[3]];
	  index[4] = cm[(int) rindex1[4] + (int) blockvals[4]];
	  index[5] = cm[(int) rindex1[5] + (int) blockvals[5]];
	  index[6] = cm[(int) rindex1[6] + (int) blockvals[6]];
	  index[7] = cm[(int) rindex1[7] + (int) blockvals[7]];
	  index += row_size;
	  rindex1 += row_size;
	  
	  index[0] = cm[(int) rindex1[0] + (int) blockvals[8]];
	  index[1] = cm[(int) rindex1[1] + (int) blockvals[9]];
	  index[2] = cm[(int) rindex1[2] + (int) blockvals[10]];
	  index[3] = cm[(int) rindex1[3] + (int) blockvals[11]];
	  index[4] = cm[(int) rindex1[4] + (int) blockvals[12]];
	  index[5] = cm[(int) rindex1[5] + (int) blockvals[13]];
	  index[6] = cm[(int) rindex1[6] + (int) blockvals[14]];
	  index[7] = cm[(int) rindex1[7] + (int) blockvals[15]];
	  blockvals += 16;
	  index += row_size;
	  rindex1 += row_size;
	}
      else {
	if (right_for & 0x1) {
	  /* No alignment, use bye copy */
	  for (rr = 0; rr < 4; rr++) {
	    index[0] = rindex1[0];
	    index[1] = rindex1[1];
	    index[2] = rindex1[2];
	    index[3] = rindex1[3];
	    index[4] = rindex1[4];
	    index[5] = rindex1[5];
	    index[6] = rindex1[6];
	    index[7] = rindex1[7];
	    index += row_size;
	    rindex1 += row_size;
	    
	    index[0] = rindex1[0];
	    index[1] = rindex1[1];
	    index[2] = rindex1[2];
	    index[3] = rindex1[3];
	    index[4] = rindex1[4];
	    index[5] = rindex1[5];
	    index[6] = rindex1[6];
	    index[7] = rindex1[7];
	    index += row_size;
	    rindex1 += row_size;
	  }
	} else if (right_for & 0x2) {
	  /* Half-word bit aligned, use 16 bit copy */
	  short *src = (short *)rindex1;
	  short *dest = (short *)index;
	  row_size >>= 1;
	  for (rr = 0; rr < 4; rr++) {
	    dest[0] = src[0];
	    dest[1] = src[1];
	    dest[2] = src[2];
	    dest[3] = src[3];
	    dest += row_size;
	    src += row_size;
	    
	    dest[0] = src[0];
	    dest[1] = src[1];
	    dest[2] = src[2];
	    dest[3] = src[3];
	    dest += row_size;
	    src += row_size;
	  }
	} else {
	  /* Word aligned, use 32 bit copy */
	  int *src = (int *)rindex1;
	  int *dest = (int *)index;
	  row_size >>= 2;
	  for (rr = 0; rr < 4; rr++) {
	    dest[0] = src[0];
	    dest[1] = src[1];
	    dest += row_size;
	    src += row_size;
	    
	    dest[0] = src[0];
	    dest[1] = src[1];
	    dest += row_size;
	    src += row_size;
	  }
	}
      }
    } else {
      unsigned char *cm = cropTbl + MAX_NEG_CROP;
      rindex2 = rindex1 + right_half_for + (down_half_for * row_size);
      if (!zflag)
	for (rr = 0; rr < 4; rr++) {
	  index[0] = cm[((int) (rindex1[0] + rindex2[0]) >> 1) + blockvals[0]];
	  index[1] = cm[((int) (rindex1[1] + rindex2[1]) >> 1) + blockvals[1]];
	  index[2] = cm[((int) (rindex1[2] + rindex2[2]) >> 1) + blockvals[2]];
	  index[3] = cm[((int) (rindex1[3] + rindex2[3]) >> 1) + blockvals[3]];
	  index[4] = cm[((int) (rindex1[4] + rindex2[4]) >> 1) + blockvals[4]];
	  index[5] = cm[((int) (rindex1[5] + rindex2[5]) >> 1) + blockvals[5]];
	  index[6] = cm[((int) (rindex1[6] + rindex2[6]) >> 1) + blockvals[6]];
	  index[7] = cm[((int) (rindex1[7] + rindex2[7]) >> 1) + blockvals[7]];
	  index += row_size;
	  rindex1 += row_size;
	  rindex2 += row_size;
	  
	  index[0] = cm[((int) (rindex1[0] + rindex2[0]) >> 1) + blockvals[8]];
	  index[1] = cm[((int) (rindex1[1] + rindex2[1]) >> 1) + blockvals[9]];
	  index[2] = cm[((int) (rindex1[2] + rindex2[2]) >> 1) + blockvals[10]];
	  index[3] = cm[((int) (rindex1[3] + rindex2[3]) >> 1) + blockvals[11]];
	  index[4] = cm[((int) (rindex1[4] + rindex2[4]) >> 1) + blockvals[12]];
	  index[5] = cm[((int) (rindex1[5] + rindex2[5]) >> 1) + blockvals[13]];
	  index[6] = cm[((int) (rindex1[6] + rindex2[6]) >> 1) + blockvals[14]];
	  index[7] = cm[((int) (rindex1[7] + rindex2[7]) >> 1) + blockvals[15]];
	  blockvals += 16;
	  index += row_size;
	  rindex1 += row_size;
	  rindex2 += row_size;
	}
      else
	for (rr = 0; rr < 4; rr++) {
	  index[0] = (int) (rindex1[0] + rindex2[0]) >> 1;
	  index[1] = (int) (rindex1[1] + rindex2[1]) >> 1;
	  index[2] = (int) (rindex1[2] + rindex2[2]) >> 1;
	  index[3] = (int) (rindex1[3] + rindex2[3]) >> 1;
	  index[4] = (int) (rindex1[4] + rindex2[4]) >> 1;
	  index[5] = (int) (rindex1[5] + rindex2[5]) >> 1;
	  index[6] = (int) (rindex1[6] + rindex2[6]) >> 1;
	  index[7] = (int) (rindex1[7] + rindex2[7]) >> 1;
	  index += row_size;
	  rindex1 += row_size;
	  rindex2 += row_size;
	  
	  index[0] = (int) (rindex1[0] + rindex2[0]) >> 1;
	  index[1] = (int) (rindex1[1] + rindex2[1]) >> 1;
	  index[2] = (int) (rindex1[2] + rindex2[2]) >> 1;
	  index[3] = (int) (rindex1[3] + rindex2[3]) >> 1;
	  index[4] = (int) (rindex1[4] + rindex2[4]) >> 1;
	  index[5] = (int) (rindex1[5] + rindex2[5]) >> 1;
	  index[6] = (int) (rindex1[6] + rindex2[6]) >> 1;
	  index[7] = (int) (rindex1[7] + rindex2[7]) >> 1;
	  index += row_size;
	  rindex1 += row_size;
	  rindex2 += row_size;
	}
    }

#ifdef LOOSE_MPEG
  }
#endif
}


/*
 *--------------------------------------------------------------
 *
 * ReconBMBlock --
 *
 *	Reconstructs back predicted macroblocks.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static void
ReconBMBlock(MPEGFrame *framePtr, 
	     int bnum, 
	     int recon_right_back, 
	     int recon_down_back, 
	     int zflag)
{
    int mb_row, mb_col, row, col, row_size, rr;
    unsigned char *dest, *future;
    int right_back, down_back, right_half_back, down_half_back;
    unsigned char *rindex1, *rindex2;
    unsigned char *index;
    short int *blockvals;
    MPEGSoftwareDecodeState *decodeState =
      (MPEGSoftwareDecodeState *) framePtr->deviceData;
    
#ifdef LOOSE_MPEG
    int illegalBlock = 0;
    int maxx, maxy, cc;
    int row_start, row_end, rlast, rfirst, col_start, col_end, clast, cfirst;
#endif
    
    /* Calculate macroblock row and column from address. */
    
    mb_row = decodeState->mblock.mb_address / decodeState->mb_width;
    mb_col = decodeState->mblock.mb_address % decodeState->mb_width;
    
    /* If block is luminance block... */
    
    if (bnum < 4) {
	
	/* Calculate right_back, down_back motion vectors. */
	
	right_back = recon_right_back >> 1;
	down_back = recon_down_back >> 1;
	right_half_back = recon_right_back & 0x1;
	down_half_back = recon_down_back & 0x1;
	
	/* Set dest to luminance plane of current pict image. */
	
	dest = (unsigned char *)decodeState->luminance->data;
	
	/*
	 * If future frame exists, set future to luminance plane of future frame.
	 */
	
	if (framePtr->ref2 != NULL) {
            MPEGSoftwareDecodeState *state;
            state = (MPEGSoftwareDecodeState *)framePtr->ref2->deviceData;
            future = (unsigned char *)state->luminance->data;
	}
	
	/* Establish row size. */
	
	row_size = decodeState->mb_width << 4;
	
	/* Calculate row,col of upper left pixel in block. */
	
	row = mb_row << 4;
	col = mb_col << 4;
	if (bnum > 1)
	  row += 8;
	if (bnum % 2)
	  col += 8;
	
#ifdef LOOSE_MPEG
	
	/* Check for block illegality. */
	
	maxx = lmaxx; maxy = lmaxy;
	
	if (row + down_back + 7 > maxy) illegalBlock |= 0x4;
	else if (row + down_back < 0)  illegalBlock |= 0x1;
	
	if (col + right_back + 7 > maxx) illegalBlock |= 0x2;
	else if (col + right_back < 0) illegalBlock |= 0x8;
	
#endif
	
    }
    /* Otherwise, block is NOT luminance block, ... */
    
    else {
	
	/* Construct motion vectors. */
	
	recon_right_back /= 2;
	recon_down_back /= 2;
	right_back = recon_right_back >> 1;
	down_back = recon_down_back >> 1;
	right_half_back = recon_right_back & 0x1;
	down_half_back = recon_down_back & 0x1;
	
	/* Establish row size. */
	
	row_size = decodeState->mb_width << 3;
	
	/* Calculate row,col of upper left pixel in block. */
	
	row = mb_row << 3;
	col = mb_col << 3;
	
#ifdef LOOSE_MPEG
	
	/* Check for block illegality. */
	
	maxx = cmaxx; maxy = cmaxy;
	
	if (row + down_back + 7 > maxy) illegalBlock |= 0x4;
	else if (row + down_back < 0) illegalBlock |= 0x1;
	
	if (col + right_back + 7 > maxx) illegalBlock  |= 0x2;
	else if (col + right_back < 0) illegalBlock |= 0x8;
	
#endif
	
	/* If block is Cr block... */
	/* They were switched earlier, so 5 is first - eyhung */
	
	if (bnum == 5) {
	    
	    /* Set dest to Cr plane of current pict image. */
	    
	    dest = (unsigned char *)decodeState->cr->data;
	    
	    /*
	     * If future frame exists, set future to Cr plane of future image.
	     */
	    
	    if (framePtr->ref2 != NULL) {
		MPEGSoftwareDecodeState *state;
		state = (MPEGSoftwareDecodeState *)framePtr->ref2->deviceData;
		future = (unsigned char *)state->cr->data;
	    }
	}
	/* Otherwise, block is Cb block... */
	
	else {
	    
	    /* Set dest to Cb plane of current pict image. */
	    
	    dest = (unsigned char *)decodeState->cb->data;
	    
	    /*
	     * If future frame exists, set future to Cb plane of future frame.
	     */
	    
	    if (framePtr->ref2 != NULL) {
		MPEGSoftwareDecodeState *state;
		state = (MPEGSoftwareDecodeState *)framePtr->ref2->deviceData;
		future = (unsigned char *)state->cb->data;
	    }
	}
    }
    
    /* For each pixel in block do... */
    
#ifdef LOOSE_MPEG
    
    if (illegalBlock) {
	if (illegalBlock & 0x1) {
	    row_start = 0;
	    row_end = row+down_back+8;
	    rfirst = rlast = 8 - row_end;
	}
	else if (illegalBlock & 0x4) {
	    row_start = row + down_back;
	    row_end = maxy+1;
	    rlast = row_end - row_start - 1;
	    rfirst = 0;
	}
	else {
	    row_start = row+down_back;
	    row_end = row_start+8;
	    rfirst = 0;
	}
	
	if (illegalBlock & 0x8) {
	    col_start = 0;
	    col_end = col + right_back + 8;
	    cfirst = clast = 8 - col_end;
	}
	else if (illegalBlock & 0x2) {
	    col_start = col + right_back;
	    col_end = maxx + 1;
	    clast = col_end - col_start - 1;
	    cfirst = 0;
	}
	else {
	    col_start = col + right_back;
	    col_end = col_start + 8;
	    cfirst = 0;
	}
	
	for (rr = row_start; rr < row_end; rr++) {
	    rindex1 = future + (rr * row_size) + col_start;
	    index = dest + ((row + rfirst) * row_size) + col + cfirst;
	    for (cc = col_start; cc < col_end; cc++) {
		*index++ = *rindex1++;
	    }
	}
	
	if (illegalBlock & 0x1) {
	    for (rr = rlast -1; rr >=0; rr--) {
		index = dest + ((row + rr) * row_size) + col;
		rindex1 = dest + ((row + rlast) * row_size) + col;
		for (cc = 0; cc < 8; cc ++) {
		    *index++ = *rindex1++;
		}
	    }
	}
	else if (illegalBlock & 0x4) {
	    for (rr = rlast+1; rr < 8; rr++) {
		index = dest + ((row + rr) * row_size) + col;
		rindex1 = dest + ((row + rlast) * row_size) + col;
		for (cc = 0; cc < 8; cc ++) {
		    *index++ = *rindex1++;
		}
	    }
	}
	
	if (illegalBlock & 0x2) {
	    for (cc = clast+1; cc < 8; cc++) {
		index = dest + (row * row_size) + (col + cc);
		rindex1 = dest + (row * row_size) + (col + clast);
		for (rr = 0; rr < 8; rr++) {
		    *index = *rindex1;
		    index += row_size;
		    rindex1 += row_size;
		}
	    }
	}
	else if (illegalBlock & 0x8) {
	    for (cc = clast-1; cc >= 0; cc--) {
		index = dest + (row * row_size) + (col + cc);
		rindex1 = dest + (row * row_size) + (col + clast);
		for (rr = 0; rr < 8; rr++) {
		    *index = *rindex1;
		    index += row_size;
		    rindex1 += row_size;
		}
	    }
	}
	
	if (!zflag) {
	    for (rr = 0; rr < 8; rr++) {
		index = dest + (row*row_size) + col;
		blockvals = &(vid_stream->block.dct_recon[rr][0]);
		index[0] += blockvals[0];
		index[1] += blockvals[1];
		index[2] += blockvals[2];
		index[3] += blockvals[3];
		index[4] += blockvals[4];
		index[5] += blockvals[5];
		index[6] += blockvals[6];
		index[7] += blockvals[7];
	    }
	}
  }
    else {
	
#endif
	
	index = dest + (row * row_size) + col;
	rindex1 = future + (row + down_back) * row_size + col + right_back;
	
	blockvals = &(decodeState->block.dct_recon[0][0]);
	
	if ((!right_half_back) && (!down_half_back)) {
	    unsigned char *cm = cropTbl + MAX_NEG_CROP;
	    if (!zflag)
	      for (rr = 0; rr < 4; rr++) {
		  index[0] = cm[(int) rindex1[0] + (int) blockvals[0]];
		  index[1] = cm[(int) rindex1[1] + (int) blockvals[1]];
		  index[2] = cm[(int) rindex1[2] + (int) blockvals[2]];
		  index[3] = cm[(int) rindex1[3] + (int) blockvals[3]];
		  index[4] = cm[(int) rindex1[4] + (int) blockvals[4]];
		  index[5] = cm[(int) rindex1[5] + (int) blockvals[5]];
		  index[6] = cm[(int) rindex1[6] + (int) blockvals[6]];
		  index[7] = cm[(int) rindex1[7] + (int) blockvals[7]];
		  index += row_size;
		  rindex1 += row_size;
		  
		  index[0] = cm[(int) rindex1[0] + (int) blockvals[8]];
		  index[1] = cm[(int) rindex1[1] + (int) blockvals[9]];
		  index[2] = cm[(int) rindex1[2] + (int) blockvals[10]];
		  index[3] = cm[(int) rindex1[3] + (int) blockvals[11]];
		  index[4] = cm[(int) rindex1[4] + (int) blockvals[12]];
		  index[5] = cm[(int) rindex1[5] + (int) blockvals[13]];
		  index[6] = cm[(int) rindex1[6] + (int) blockvals[14]];
		  index[7] = cm[(int) rindex1[7] + (int) blockvals[15]];
		  blockvals += 16;
		  index += row_size;
		  rindex1 += row_size;
	      }
	    else {
	if (right_back & 0x1) {
	    /* No alignment, use bye copy */
	    for (rr = 0; rr < 4; rr++) {
		index[0] = rindex1[0];
		index[1] = rindex1[1];
		index[2] = rindex1[2];
		index[3] = rindex1[3];
		index[4] = rindex1[4];
		index[5] = rindex1[5];
		index[6] = rindex1[6];
		index[7] = rindex1[7];
		index += row_size;
		rindex1 += row_size;
		
		index[0] = rindex1[0];
		index[1] = rindex1[1];
		index[2] = rindex1[2];
		index[3] = rindex1[3];
		index[4] = rindex1[4];
		index[5] = rindex1[5];
		index[6] = rindex1[6];
		index[7] = rindex1[7];
		index += row_size;
		rindex1 += row_size;
	    }
	} else if (right_back & 0x2) {
	    /* Half-word bit aligned, use 16 bit copy */
	    short *src = (short *)rindex1;
	    short *dest = (short *)index;
	    row_size >>= 1;
	    for (rr = 0; rr < 4; rr++) {
		dest[0] = src[0];
		dest[1] = src[1];
		dest[2] = src[2];
		dest[3] = src[3];
		dest += row_size;
		src += row_size;
		
		dest[0] = src[0];
		dest[1] = src[1];
		dest[2] = src[2];
		dest[3] = src[3];
		dest += row_size;
		src += row_size;
	    }
	} else {
	    /* Word aligned, use 32 bit copy */
	    int *src = (int *)rindex1;
	    int *dest = (int *)index;
	    row_size >>= 2;
	    for (rr = 0; rr < 4; rr++) {
		dest[0] = src[0];
		dest[1] = src[1];
		dest += row_size;
		src += row_size;
		
		dest[0] = src[0];
		dest[1] = src[1];
		dest += row_size;
		src += row_size;
	    }
	}
    }
	} else {
	    unsigned char *cm = cropTbl + MAX_NEG_CROP;
	    rindex2 = rindex1 + right_half_back + (down_half_back * row_size);
	    if (!zflag)
	      for (rr = 0; rr < 4; rr++) {
		  index[0] = cm[((int) (rindex1[0] + rindex2[0]) >> 1) + blockvals[0]];
		  index[1] = cm[((int) (rindex1[1] + rindex2[1]) >> 1) + blockvals[1]];
		  index[2] = cm[((int) (rindex1[2] + rindex2[2]) >> 1) + blockvals[2]];
		  index[3] = cm[((int) (rindex1[3] + rindex2[3]) >> 1) + blockvals[3]];
		  index[4] = cm[((int) (rindex1[4] + rindex2[4]) >> 1) + blockvals[4]];
		  index[5] = cm[((int) (rindex1[5] + rindex2[5]) >> 1) + blockvals[5]];
		  index[6] = cm[((int) (rindex1[6] + rindex2[6]) >> 1) + blockvals[6]];
		  index[7] = cm[((int) (rindex1[7] + rindex2[7]) >> 1) + blockvals[7]];
		  index += row_size;
		  rindex1 += row_size;
		  rindex2 += row_size;
		  
		  index[0] = cm[((int) (rindex1[0] + rindex2[0]) >> 1) + blockvals[8]];
		  index[1] = cm[((int) (rindex1[1] + rindex2[1]) >> 1) + blockvals[9]];
		  index[2] = cm[((int) (rindex1[2] + rindex2[2]) >> 1) + blockvals[10]];
		  index[3] = cm[((int) (rindex1[3] + rindex2[3]) >> 1) + blockvals[11]];
		  index[4] = cm[((int) (rindex1[4] + rindex2[4]) >> 1) + blockvals[12]];
		  index[5] = cm[((int) (rindex1[5] + rindex2[5]) >> 1) + blockvals[13]];
		  index[6] = cm[((int) (rindex1[6] + rindex2[6]) >> 1) + blockvals[14]];
		  index[7] = cm[((int) (rindex1[7] + rindex2[7]) >> 1) + blockvals[15]];
		  blockvals += 16;
		  index += row_size;
		  rindex1 += row_size;
		  rindex2 += row_size;
	      }
      else
	for (rr = 0; rr < 4; rr++) {
	    index[0] = (int) (rindex1[0] + rindex2[0]) >> 1;
	    index[1] = (int) (rindex1[1] + rindex2[1]) >> 1;
	    index[2] = (int) (rindex1[2] + rindex2[2]) >> 1;
	    index[3] = (int) (rindex1[3] + rindex2[3]) >> 1;
	    index[4] = (int) (rindex1[4] + rindex2[4]) >> 1;
	    index[5] = (int) (rindex1[5] + rindex2[5]) >> 1;
	    index[6] = (int) (rindex1[6] + rindex2[6]) >> 1;
	    index[7] = (int) (rindex1[7] + rindex2[7]) >> 1;
	    index += row_size;
	    rindex1 += row_size;
	    rindex2 += row_size;
	    
	    index[0] = (int) (rindex1[0] + rindex2[0]) >> 1;
	    index[1] = (int) (rindex1[1] + rindex2[1]) >> 1;
	    index[2] = (int) (rindex1[2] + rindex2[2]) >> 1;
	    index[3] = (int) (rindex1[3] + rindex2[3]) >> 1;
	    index[4] = (int) (rindex1[4] + rindex2[4]) >> 1;
	    index[5] = (int) (rindex1[5] + rindex2[5]) >> 1;
	    index[6] = (int) (rindex1[6] + rindex2[6]) >> 1;
	    index[7] = (int) (rindex1[7] + rindex2[7]) >> 1;
	    index += row_size;
	    rindex1 += row_size;
	    rindex2 += row_size;
	}
	}
	
#ifdef LOOSE_MPEG
    }
#endif
    
}


/*
 *--------------------------------------------------------------
 *
 * ReconBiMBlock --
 *
 *	Reconstructs bidirectionally predicted macroblocks.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static void
ReconBiMBlock(MPEGFrame *framePtr,
	      int bnum, 
	      int recon_right_for, 
	      int recon_down_for,
	      int recon_right_back, 
	      int recon_down_back, 
	      int zflag)
{
  int mb_row, mb_col, row, col, row_size, rr;
  unsigned char *dest, *past, *future;
  int right_for, down_for, right_half_for, down_half_for;
  int right_back, down_back, right_half_back, down_half_back;
  unsigned char *index, *rindex1, *bindex1;
  short int *blockvals;
  int forw_row_start, back_row_start, forw_col_start, back_col_start;
  MPEGSoftwareDecodeState *decodeState =
    (MPEGSoftwareDecodeState *) framePtr->deviceData;

#ifdef LOOSE_MPEG
  int illegal_forw = 0;
  int illegal_back = 0;
#endif

  /* Calculate macroblock row and column from address. */

  mb_row = decodeState->mblock.mb_address / decodeState->mb_width;
  mb_col = decodeState->mblock.mb_address % decodeState->mb_width;

  /* If block is luminance block... */

  if (bnum < 4) {

    /*
     * Calculate right_for, down_for, right_half_for, down_half_for,
     * right_back, down_bakc, right_half_back, and down_half_back, motion
     * vectors.
     */

    right_for = recon_right_for >> 1;
    down_for = recon_down_for >> 1;
    right_half_for = recon_right_for & 0x1;
    down_half_for = recon_down_for & 0x1;

    right_back = recon_right_back >> 1;
    down_back = recon_down_back >> 1;
    right_half_back = recon_right_back & 0x1;
    down_half_back = recon_down_back & 0x1;

    /* Set dest to luminance plane of current pict image. */

    dest = (unsigned char *)decodeState->luminance->data;

    /* If past frame exists, set past to luminance plane of past frame. */

    if (framePtr->ref1 != NULL) {
        MPEGSoftwareDecodeState *state;
        state = (MPEGSoftwareDecodeState *)framePtr->ref1->deviceData;
        past = (unsigned char *)state->luminance->data;
    }

    /*
     * If future frame exists, set future to luminance plane of future frame.
     */

    if (framePtr->ref2 != NULL) {
        MPEGSoftwareDecodeState *state;
        state = (MPEGSoftwareDecodeState *)framePtr->ref2->deviceData;
        future = (unsigned char *)state->luminance->data;
    }

    /* Establish row size. */

    row_size = (decodeState->mb_width << 4);

    /* Calculate row,col of upper left pixel in block. */

    row = (mb_row << 4);
    col = (mb_col << 4);
    if (bnum > 1)
      row += 8;
    if (bnum & 0x01)
      col += 8;

    forw_col_start = col + right_for;
    forw_row_start = row + down_for;

    back_col_start = col + right_back;
    back_row_start = row + down_back;

#ifdef LOOSE_MPEG

    /* Check for illegal pred. blocks. */


    if (forw_col_start+7 > lmaxx) illegal_forw = 1;
    else if (forw_col_start < 0) illegal_forw = 1;

    if (forw_row_start+7 > lmaxy) illegal_forw = 1;
    else if (forw_row_start < 0) illegal_forw = 1;

    if (back_col_start+7 > lmaxx) illegal_back = 1;
    else if (back_col_start < 0) illegal_back = 1;

    if (back_row_start+7 > lmaxy) illegal_back = 1;
    else if (back_row_start < 0) illegal_back = 1;

#endif

  }
  /* Otherwise, block is NOT luminance block, ... */

  else {

    /* Construct motion vectors. */

    recon_right_for /= 2;
    recon_down_for /= 2;
    right_for = recon_right_for >> 1;
    down_for = recon_down_for >> 1;
    right_half_for = recon_right_for & 0x1;
    down_half_for = recon_down_for & 0x1;

    recon_right_back /= 2;
    recon_down_back /= 2;
    right_back = recon_right_back >> 1;
    down_back = recon_down_back >> 1;
    right_half_back = recon_right_back & 0x1;
    down_half_back = recon_down_back & 0x1;

    /* Establish row size. */

    row_size = (decodeState->mb_width << 3);

    /* Calculate row,col of upper left pixel in block. */

    row = (mb_row << 3);
    col = (mb_col << 3);

    forw_col_start = col + right_for;
    forw_row_start = row + down_for;

    back_col_start = col + right_back;
    back_row_start = row + down_back;

#ifdef LOOSE_MPEG

    /* Check for illegal pred. blocks. */

    if (forw_col_start+7 > cmaxx) illegal_forw = 1;
    else if (forw_col_start < 0) illegal_forw = 1;

    if (forw_row_start+7 > cmaxy) illegal_forw = 1;
    else if (forw_row_start < 0) illegal_forw = 1;

    if (back_col_start+7 > cmaxx) illegal_back = 1;
    else if (back_col_start < 0) illegal_back = 1;
    
    if (back_row_start+7 > cmaxy) illegal_back = 1;
    else if (back_row_start < 0) illegal_back = 1;

#endif
    
    /* If block is Cr block... */
	/* Switched earlier, so we test Cr first - eyhung */

    if (bnum == 5) {

      /* Set dest to Cr plane of current pict image. */

      dest = (unsigned char *)decodeState->cr->data;

      /* If past frame exists, set past to Cr plane of past image. */

      if (framePtr->ref1 != NULL) {
          MPEGSoftwareDecodeState *state;
          state = (MPEGSoftwareDecodeState *)framePtr->ref1->deviceData;
          past = (unsigned char *)state->cr->data;
      }

      /*
       * If future frame exists, set future to Cr plane of future image.
       */

      if (framePtr->ref2 != NULL) {
	  MPEGSoftwareDecodeState *state;
	  state = (MPEGSoftwareDecodeState *)framePtr->ref2->deviceData;
	  future = (unsigned char *)state->cr->data;
      }
    }
    /* Otherwise, block is Cb block... */

    else {

      /* Set dest to Cb plane of current pict image. */

      dest = (unsigned char *)decodeState->cb->data;

      /* If past frame exists, set past to Cb plane of past frame. */

      if (framePtr->ref1 != NULL) {
	  MPEGSoftwareDecodeState *state;
	  state = (MPEGSoftwareDecodeState *)framePtr->ref1->deviceData;
	  past = (unsigned char *)state->cb->data;
      }

      /*
       * If future frame exists, set future to Cb plane of future frame.
       */

      if (framePtr->ref2 != NULL) {
	  MPEGSoftwareDecodeState *state;
	  state = (MPEGSoftwareDecodeState *)framePtr->ref2->deviceData;
	  future = (unsigned char *)state->cb->data;
      }
    }
  }

  /* For each pixel in block... */

  index = dest + (row * row_size) + col;

#ifdef LOOSE_MPEG
  if (illegal_forw) 
    rindex1 = future + back_row_start * row_size + back_col_start;
  else 
#endif
    rindex1 = past + forw_row_start  * row_size + forw_col_start;

#ifdef LOOSE_MPEG
  if (illegal_back) 
    bindex1 = past + forw_row_start * row_size + forw_col_start;
  else 
#endif
    bindex1 = future + back_row_start * row_size + back_col_start;

  blockvals = (short int *) &(decodeState->block.dct_recon[0][0]);

  {
  unsigned char *cm = cropTbl + MAX_NEG_CROP;
  if (!zflag)
    for (rr = 0; rr < 4; rr++) {
      index[0] = cm[((int) (rindex1[0] + bindex1[0]) >> 1) + blockvals[0]];
      index[1] = cm[((int) (rindex1[1] + bindex1[1]) >> 1) + blockvals[1]];
      index[2] = cm[((int) (rindex1[2] + bindex1[2]) >> 1) + blockvals[2]];
      index[3] = cm[((int) (rindex1[3] + bindex1[3]) >> 1) + blockvals[3]];
      index[4] = cm[((int) (rindex1[4] + bindex1[4]) >> 1) + blockvals[4]];
      index[5] = cm[((int) (rindex1[5] + bindex1[5]) >> 1) + blockvals[5]];
      index[6] = cm[((int) (rindex1[6] + bindex1[6]) >> 1) + blockvals[6]];
      index[7] = cm[((int) (rindex1[7] + bindex1[7]) >> 1) + blockvals[7]];
      index += row_size;
      rindex1 += row_size;
      bindex1 += row_size;

      index[0] = cm[((int) (rindex1[0] + bindex1[0]) >> 1) + blockvals[8]];
      index[1] = cm[((int) (rindex1[1] + bindex1[1]) >> 1) + blockvals[9]];
      index[2] = cm[((int) (rindex1[2] + bindex1[2]) >> 1) + blockvals[10]];
      index[3] = cm[((int) (rindex1[3] + bindex1[3]) >> 1) + blockvals[11]];
      index[4] = cm[((int) (rindex1[4] + bindex1[4]) >> 1) + blockvals[12]];
      index[5] = cm[((int) (rindex1[5] + bindex1[5]) >> 1) + blockvals[13]];
      index[6] = cm[((int) (rindex1[6] + bindex1[6]) >> 1) + blockvals[14]];
      index[7] = cm[((int) (rindex1[7] + bindex1[7]) >> 1) + blockvals[15]];
      blockvals += 16;
      index += row_size;
      rindex1 += row_size;
      bindex1 += row_size;
    }

  else
    for (rr = 0; rr < 4; rr++) {
      index[0] = (int) (rindex1[0] + bindex1[0]) >> 1;
      index[1] = (int) (rindex1[1] + bindex1[1]) >> 1;
      index[2] = (int) (rindex1[2] + bindex1[2]) >> 1;
      index[3] = (int) (rindex1[3] + bindex1[3]) >> 1;
      index[4] = (int) (rindex1[4] + bindex1[4]) >> 1;
      index[5] = (int) (rindex1[5] + bindex1[5]) >> 1;
      index[6] = (int) (rindex1[6] + bindex1[6]) >> 1;
      index[7] = (int) (rindex1[7] + bindex1[7]) >> 1;
      index += row_size;
      rindex1 += row_size;
      bindex1 += row_size;

      index[0] = (int) (rindex1[0] + bindex1[0]) >> 1;
      index[1] = (int) (rindex1[1] + bindex1[1]) >> 1;
      index[2] = (int) (rindex1[2] + bindex1[2]) >> 1;
      index[3] = (int) (rindex1[3] + bindex1[3]) >> 1;
      index[4] = (int) (rindex1[4] + bindex1[4]) >> 1;
      index[5] = (int) (rindex1[5] + bindex1[5]) >> 1;
      index[6] = (int) (rindex1[6] + bindex1[6]) >> 1;
      index[7] = (int) (rindex1[7] + bindex1[7]) >> 1;
      index += row_size;
      rindex1 += row_size;
      bindex1 += row_size;
    }
  }
}


/*
 *--------------------------------------------------------------
 *
 * ComputeVector --
 *
 *	Computes motion vector given parameters previously parsed
 *      and reconstructed.
 *
 * Results:
 *      Reconstructed motion vector info is put into recon_* parameters
 *      passed to this function. Also updated previous motion vector
 *      information.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

#define ComputeVector(recon_right_ptr, recon_down_ptr, recon_right_prev, recon_down_prev, f, full_pel_vector, motion_h_code, motion_v_code, motion_h_r, motion_v_r)				\
									\
{									\
  int comp_h_r, comp_v_r;						\
  int right_little, right_big, down_little, down_big;			\
  int max, min, new_vector;						\
									\
  /* The following procedure for the reconstruction of motion vectors 	\
     is a direct and simple implementation of the instructions given	\
     in the mpeg December 1991 standard draft. 				\
  */									\
									\
  if (f == 1 || motion_h_code == 0)					\
    comp_h_r = 0;							\
  else 									\
    comp_h_r = f - 1 - motion_h_r;					\
									\
  if (f == 1 || motion_v_code == 0)					\
    comp_v_r = 0;							\
  else 									\
    comp_v_r = f - 1 - motion_v_r;					\
									\
  right_little = motion_h_code * f;					\
  if (right_little == 0)						\
    right_big = 0;							\
  else {								\
    if (right_little > 0) {						\
      right_little = right_little - comp_h_r;				\
      right_big = right_little - 32 * f;				\
    }									\
    else {								\
      right_little = right_little + comp_h_r;				\
      right_big = right_little + 32 * f;				\
    }									\
  }									\
									\
  down_little = motion_v_code * f;					\
  if (down_little == 0)							\
    down_big = 0;							\
  else {								\
    if (down_little > 0) {						\
      down_little = down_little - comp_v_r;				\
      down_big = down_little - 32 * f;					\
    }									\
    else {								\
      down_little = down_little + comp_v_r;				\
      down_big = down_little + 32 * f;					\
    }									\
  }									\
  									\
  max = 16 * f - 1;							\
  min = -16 * f;							\
									\
  new_vector = recon_right_prev + right_little;				\
									\
  if (new_vector <= max && new_vector >= min)				\
    *recon_right_ptr = recon_right_prev + right_little;			\
                      /* just new_vector */				\
  else									\
    *recon_right_ptr = recon_right_prev + right_big;			\
  recon_right_prev = *recon_right_ptr;					\
  if (full_pel_vector)							\
    *recon_right_ptr = *recon_right_ptr << 1;				\
									\
  new_vector = recon_down_prev + down_little;				\
  if (new_vector <= max && new_vector >= min)				\
    *recon_down_ptr = recon_down_prev + down_little;			\
                      /* just new_vector */				\
  else									\
    *recon_down_ptr = recon_down_prev + down_big;			\
  recon_down_prev = *recon_down_ptr;					\
  if (full_pel_vector)							\
    *recon_down_ptr = *recon_down_ptr << 1;				\
}


/*
 *--------------------------------------------------------------
 *
 * ComputeForwVector --
 *
 *	Computes forward motion vector by calling ComputeVector
 *      with appropriate parameters.
 *
 * Results:
 *	Reconstructed motion vector placed in recon_right_for_ptr and
 *      recon_down_for_ptr.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void 
ComputeForwVector(MPEGSoftwareDecodeState *decodeState,
		  int *recon_right_for_ptr, 
		  int *recon_down_for_ptr)
{

  Pict *picture;
  Macroblock *mblock;

  picture = &(decodeState->picture);
  mblock = &(decodeState->mblock);

  ComputeVector(recon_right_for_ptr, recon_down_for_ptr,
		mblock->recon_right_for_prev, 
		mblock->recon_down_for_prev,
		picture->forw_f, picture->full_pel_forw_vector,
		mblock->motion_h_forw_code, mblock->motion_v_forw_code,
		mblock->motion_h_forw_r, mblock->motion_v_forw_r); 
}


/*
 *--------------------------------------------------------------
 *
 * ComputeBackVector --
 *
 *	Computes backward motion vector by calling ComputeVector
 *      with appropriate parameters.
 *
 * Results:
 *	Reconstructed motion vector placed in recon_right_back_ptr and
 *      recon_down_back_ptr.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void 
ComputeBackVector(MPEGSoftwareDecodeState *decodeState,
		  int *recon_right_back_ptr, 
		  int *recon_down_back_ptr)
{
  Pict *picture;
  Macroblock *mblock;

  picture = &(decodeState->picture);
  mblock = &(decodeState->mblock);

  ComputeVector(recon_right_back_ptr, recon_down_back_ptr,
		mblock->recon_right_back_prev, 
		mblock->recon_down_back_prev,
		picture->back_f, picture->full_pel_back_vector,
		mblock->motion_h_back_code, mblock->motion_v_back_code,
		mblock->motion_h_back_r, mblock->motion_v_back_r); 

}

/* Macro for returning 1 if num is positive, -1 if negative, 0 if 0. */

#define Sign(num) ((num > 0) ? 1 : ((num == 0) ? 0 : -1))


/*
 *--------------------------------------------------------------
 *
 * ParseReconBlock --
 *
 *	Parse values for block structure from bitstream.
 *      n is an indication of the position of the block within
 *      the macroblock (i.e. 0-5) and indicates the type of 
 *      block (i.e. luminance or chrominance). Reconstructs
 *      coefficients from values parsed and puts in 
 *      block.dct_recon array in vid stream structure.
 *      sparseFlag is set when the block contains only one
 *      coeffictient and is used by the IDCT.
 *
 * Results:
 *	
 *
 * Side effects:
 *      Bit stream irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

#define DCT_recon blockPtr->dct_recon
#define DCT_dc_y_past blockPtr->dct_dc_y_past
#define DCT_dc_cr_past blockPtr->dct_dc_cr_past
#define DCT_dc_cb_past blockPtr->dct_dc_cb_past

#define DECODE_DCT_COEFF_FIRST DecodeDCTCoeffFirst
#define DECODE_DCT_COEFF_NEXT DecodeDCTCoeffNext

void
ParseReconBlock(MPEGFrame *framePtr,
		int n)
{
    MPEGSoftwareDecodeState *decodeState = 
      (MPEGSoftwareDecodeState *) framePtr->deviceData;
    Block *blockPtr = &(decodeState->block);
    int coeffCount;
    int diff;
    int size, level, i, run, pos, coeff;
    short int *reconptr;
    unsigned int *iqmatrixptr, *niqmatrixptr;
    int qscale;
    
    reconptr = DCT_recon[0];
    
    /* 
     * Hand coded version of memset that's a little faster...
     * Old call:
     *	memset((char *) DCT_recon, 0, 64*sizeof(short int));
     */
    {
	INT32 *p;
	p = (INT32 *) reconptr;
	
	p[0] = p[1] = p[2] = p[3] = p[4] = p[5] = p[6] = p[7] = p[8] = p[9] = 
	  p[10] = p[11] = p[12] = p[13] = p[14] = p[15] = p[16] = p[17] = p[18] =
	    p[19] = p[20] = p[21] = p[22] = p[23] = p[24] = p[25] = p[26] = p[27] =
	      p[28] = p[29] = p[30] = p[31] = 0;
	
    }
    
    if (decodeState->mblock.mb_intra) {
	
	if (n < 4) {
	    
	    /*
	     * Get the luminance bits.  This code has been hand optimized to
	     * get by the normal bit parsing routines.  We get some speedup
	     * by grabbing the next 16 bits and parsing things locally.
	     * Thus, calls are translated as:
	     *
	     *	show_bitsX  <-->   next16bits >> (16-X)
	     *	get_bitsX   <-->   val = next16bits >> (16-flushed-X);
	     *			   flushed += X;
	     *			   next16bits &= bitMask[flushed];
	     *	flush_bitsX <-->   flushed += X;
	     *			   next16bits &= bitMask[flushed];
	     *
	     * I've streamlined the code a lot, so that we don't have to mask
	     * out the low order bits and a few of the extra adds are removed.
	     *	bsmith
	     */
	    unsigned int next16bits, index, flushed;

	    bp_Peek(16, &next16bits);
	    index = next16bits >> (16-5);
	    if (index < 31) {
		size = dct_dc_size_luminance[index].value;
		flushed = dct_dc_size_luminance[index].num_bits;
	    }
	    else {
		index = next16bits >> (16-9);
		index -= 0x1f0;
		size = dct_dc_size_luminance1[index].value;
		flushed = dct_dc_size_luminance1[index].num_bits;
	    }
	    next16bits &= bitMask[16+flushed];
	    
	    if (size != 0) {
		flushed += size;
		diff = next16bits >> (16-flushed);
		if (!(diff & bitTest[32-size])) {
		    diff = rBitMask[size] | (diff + 1);
		}
	    } else {
		diff = 0;
	    }
	    bp_Flush(flushed);
	    
	    if (n == 0) {
		coeff = diff << 3;
		if (decodeState->mblock.mb_address -
		    decodeState->mblock.past_intra_addr > 1) 
		  coeff += 1024;
		else coeff += DCT_dc_y_past;
		DCT_dc_y_past = coeff;
	    } else {
		coeff = DCT_dc_y_past + (diff << 3);
		DCT_dc_y_past = coeff;
	    }
	} else {
	
	    /*
	     * Get the chrominance bits.  This code has been hand optimized to
	     * as described above
	     */
	    unsigned int next16bits, index, flushed;

	    bp_Peek(16, &next16bits);
	    index = next16bits >> (16-5);
	    if (index < 31) {
		size = dct_dc_size_chrominance[index].value;
		flushed = dct_dc_size_chrominance[index].num_bits;
	    }
	    else {
		index = next16bits >> (16-10);
		index -= 0x3e0;
		size = dct_dc_size_chrominance1[index].value;
		flushed = dct_dc_size_chrominance1[index].num_bits;
	    }
	    next16bits &= bitMask[16+flushed];
	
	    if (size != 0) {
		flushed += size;
		diff = next16bits >> (16-flushed);
		if (!(diff & bitTest[32-size])) {
		    diff = rBitMask[size] | (diff + 1);
		}
	    } else {
		diff = 0;
	    }
	    bp_Flush(flushed);
	    
	    /* We test 5 first; a result of the mixup of Cr and Cb */
	    if (n == 5) {
		coeff = diff << 3;

		if (decodeState->mblock.mb_address -
		    decodeState->mblock.past_intra_addr > 1) 
		  coeff += 1024;
		else coeff += DCT_dc_cr_past;
		DCT_dc_cr_past = coeff;
		
	    } else {
		coeff = diff << 3;
		if (decodeState->mblock.mb_address -
		    decodeState->mblock.past_intra_addr > 1) 
		  coeff += 1024;
		else coeff += DCT_dc_cb_past;
		DCT_dc_cb_past = coeff;
	    }
	}
      
	*reconptr = coeff;
	i = 0; pos = 0;
	coeffCount = (coeff != 0);
    
	if (decodeState->picture.type != 4) {
	    
	    qscale = decodeState->slice.quant_scale;
	    if (framePtr->intra_matrix != NULL) {
		iqmatrixptr = (unsigned int *)framePtr->intra_matrix;
	    } else {
		iqmatrixptr = default_intra_matrix;
	    }
	    
	    while(1) {
		
		DECODE_DCT_COEFF_NEXT(run, level);
		
		if (run == END_OF_BLOCK) break;
		
		i = i + run + 1;
		pos = zigzag_direct[i];
		
		/* quantizes and oddifies each coefficient */
		if (level < 0) {
		    coeff = ((level<<1) * qscale * 
			     ((int) (iqmatrixptr[i]))) / 16; 
		    coeff += (1 - (coeff & 1));
		} else {
		    coeff = ((level<<1) * qscale * 
			     ((int) (*(iqmatrixptr+i)))) >> 4; 
		    coeff -= (1 - (coeff & 1));
		}

		reconptr[pos] = coeff;
		if (coeff) {
		    coeffCount++;
		}
		
	    }
	    bp_Flush(2);
	    
	    goto end;
	}
    }
    
    else {
      
	if (framePtr->non_intra_matrix != NULL) {
	    niqmatrixptr = (unsigned int *)framePtr->non_intra_matrix;
	} else {
	    niqmatrixptr = default_non_intra_matrix;
	}

	qscale = decodeState->slice.quant_scale;
      
	DECODE_DCT_COEFF_FIRST(run, level);
	i = run;

	pos = zigzag_direct[i];

	/* quantizes and oddifies each coefficient */
	if (level < 0) {
	    coeff = (((level<<1) - 1) * qscale * 
		     ((int) (niqmatrixptr[i]))) / 16; 
	    coeff += (1 - (coeff & 1));
	} else {
	    coeff = (((level<<1) + 1) * qscale * 
		     ((int) (*(niqmatrixptr+i)))) >> 4; 
	    coeff -= (1 - (coeff & 1));
	}

	reconptr[pos] = coeff;
	if (coeff) {
	    coeffCount = 1;
	}

	if (decodeState->picture.type != 4) {
	    
	    while(1) {
	  
		DECODE_DCT_COEFF_NEXT(run, level);
		
		if (run == END_OF_BLOCK) break;

		i = i+run+1;
		pos = zigzag_direct[i];
		if (level < 0) {
		    coeff = (((level<<1) - 1) * qscale * 
			     ((int) (niqmatrixptr[i]))) / 16; 
		    coeff += (coeff & 1);
		} else {
		    coeff = (((level<<1) + 1) * qscale * 
			     ((int) (*(niqmatrixptr+i)))) >> 4; 
		    coeff -= (coeff & 1);
		}
		reconptr[pos] = coeff;
		if (coeff) {
		    coeffCount++;
		}
	    }

	    bp_Flush(2);
	    
	    goto end;
	}
    }
    
  end:

    if (coeffCount == 1) j_rev_dct_sparse (reconptr, pos);
    else j_rev_dct(reconptr);
}

#undef DCT_recon 
#undef DCT_dc_y_past 
#undef DCT_dc_cr_past 
#undef DCT_dc_cb_past 


/*
 *--------------------------------------------------------------
 *
 * ParseAwayBlock --
 *
 *	Parses off block values, throwing them away.
 *      Used with grayscale dithering.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void
ParseAwayBlock(MPEGFrame *framePtr,
	       int n)
{
  unsigned int diff;
  unsigned int size, run;
  int level;
  MPEGSoftwareDecodeState *decodeState = 
    (MPEGSoftwareDecodeState *) framePtr->deviceData;

  if (decodeState->mblock.mb_intra) {

      /* If the block is a luminance block... */

      if (n < 4) {

	  /* Parse and decode size of first coefficient. */

	  DecodeDCTDCSizeLum(size);

	  /* Parse first coefficient. */

	  if (size != 0) {
	      bp_Get(size, &diff);
	  }
      }

      /* Otherwise, block is chrominance block... */
      
      else {

	  /* Parse and decode size of first coefficient. */
	  
	  DecodeDCTDCSizeChrom(size);
	  
	  /* Parse first coefficient. */

	  if (size != 0) {
	      bp_Get(size, &diff);
	  }
      }
  }

  /* Otherwise, block is not intracoded... */
  
  else {
      
      /* Decode and set first coefficient. */
      
      DECODE_DCT_COEFF_FIRST(run, level);
  }
  
  /* If picture is not D type (i.e. I, P, or B)... */

  if (decodeState->picture.type != 4) {
      
      /* While end of macroblock has not been reached... */
      
      while (1) {

	  /* Get the dct_coeff_next */
	  
	  DECODE_DCT_COEFF_NEXT(run, level);

	  if (run == END_OF_BLOCK) break;
      }
      
      /* End_of_block */

      bp_Flush(2);
  }
}


/*
 *--------------------------------------------------------------
 *
 * ReconSkippedBlock --
 *
 *	Reconstructs predictive block for skipped macroblocks
 *      in B Frames.
 *
 * Results:
 *	No return values.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static void
ReconSkippedBlock(unsigned char *source, 
		  unsigned char *dest, 
		  int row, 
		  int col, 
		  int row_size,
		  int right, 
		  int down, 
		  int right_half, 
		  int down_half, 
		  int width)
{
    int rr;
    unsigned char *source2;
    
    source += ((row + down) * row_size) + col + right;
    
    if (width == 16) {
	if ((!right_half) && (!down_half)) {
	    if (right & 0x1) {
		/* No alignment, use bye copy */
		for (rr = 0; rr < 16; rr++) {
		    dest[0] = source[0];
		    dest[1] = source[1];
		    dest[2] = source[2];
		    dest[3] = source[3];
		    dest[4] = source[4];
		    dest[5] = source[5];
		    dest[6] = source[6];
		    dest[7] = source[7];
		    dest[8] = source[8];
		    dest[9] = source[9];
		    dest[10] = source[10];
		    dest[11] = source[11];
		    dest[12] = source[12];
		    dest[13] = source[13];
		    dest[14] = source[14];
		    dest[15] = source[15];
		    dest += 16;
		    source += row_size;
		}
	    } else if (right & 0x2) {
		/* Half-word bit aligned, use 16 bit copy */
		short *src = (short *)source;
		short *d = (short *)dest;
		row_size >>= 1;
		for (rr = 0; rr < 16; rr++) {
		    d[0] = src[0];
		    d[1] = src[1];
		    d[2] = src[2];
		    d[3] = src[3];
		    d[4] = src[4];
		    d[5] = src[5];
		    d[6] = src[6];
		    d[7] = src[7];
		    d += 8;
		    src += row_size;
		}
	    } else {
		/* Word aligned, use 32 bit copy */
		int *src = (int *)source;
		int *d = (int *)dest;
		row_size >>= 2;
		for (rr = 0; rr < 16; rr++) {
		    d[0] = src[0];
		    d[1] = src[1];
		    d[2] = src[2];
		    d[3] = src[3];
		    d += 4;
		    src += row_size;
		}
	    }
	} else {
	    source2 = source + right_half + (row_size * down_half);
	    for (rr = 0; rr < width; rr++) {
		dest[0] = (int) (source[0] + source2[0]) >> 1;
		dest[1] = (int) (source[1] + source2[1]) >> 1;
		dest[2] = (int) (source[2] + source2[2]) >> 1;
		dest[3] = (int) (source[3] + source2[3]) >> 1;
		dest[4] = (int) (source[4] + source2[4]) >> 1;
		dest[5] = (int) (source[5] + source2[5]) >> 1;
		dest[6] = (int) (source[6] + source2[6]) >> 1;
		dest[7] = (int) (source[7] + source2[7]) >> 1;
		dest[8] = (int) (source[8] + source2[8]) >> 1;
		dest[9] = (int) (source[9] + source2[9]) >> 1;
		dest[10] = (int) (source[10] + source2[10]) >> 1;
		dest[11] = (int) (source[11] + source2[11]) >> 1;
		dest[12] = (int) (source[12] + source2[12]) >> 1;
		dest[13] = (int) (source[13] + source2[13]) >> 1;
		dest[14] = (int) (source[14] + source2[14]) >> 1;
		dest[15] = (int) (source[15] + source2[15]) >> 1;
		dest += width;
		source += row_size;
		source2 += row_size;
	    }
	}
    } else {			/* (width == 8) */
	assert(width == 8);
	if ((!right_half) && (!down_half)) {
	    if (right & 0x1) {
		for (rr = 0; rr < width; rr++) {
		    dest[0] = source[0];
		    dest[1] = source[1];
		    dest[2] = source[2];
		    dest[3] = source[3];
		    dest[4] = source[4];
		    dest[5] = source[5];
		    dest[6] = source[6];
		    dest[7] = source[7];
		    dest += 8;
		    source += row_size;
		}
	    } else if (right & 0x02) {
		short *d = (short *)dest;
		short *src = (short *)source;
		row_size >>= 1;
		for (rr = 0; rr < width; rr++) {
		    d[0] = src[0];
		    d[1] = src[1];
		    d[2] = src[2];
		    d[3] = src[3];
		    d += 4;
		    src += row_size;
		}
	    } else {
		int *d = (int *)dest;
		int *src = (int *)source;
		row_size >>= 2;
		for (rr = 0; rr < width; rr++) {
		    d[0] = src[0];
		    d[1] = src[1];
		    d += 2;
		    src += row_size;
		}
	    }
	} else {
	    source2 = source + right_half + (row_size * down_half);
	    for (rr = 0; rr < width; rr++) {
		dest[0] = (int) (source[0] + source2[0]) >> 1;
		dest[1] = (int) (source[1] + source2[1]) >> 1;
		dest[2] = (int) (source[2] + source2[2]) >> 1;
		dest[3] = (int) (source[3] + source2[3]) >> 1;
		dest[4] = (int) (source[4] + source2[4]) >> 1;
		dest[5] = (int) (source[5] + source2[5]) >> 1;
		dest[6] = (int) (source[6] + source2[6]) >> 1;
		dest[7] = (int) (source[7] + source2[7]) >> 1;
		dest += width;
		source += row_size;
		source2 += row_size;
	    }
	}
    }
}


/*
 *--------------------------------------------------------------
 *
 * ProcessSkippedPFrameMBlocks --
 *
 *	Processes skipped macroblocks in P frames.
 *
 * Results:
 *	Calculates pixel values for luminance, Cr, and Cb planes
 *      in current pict image for skipped macroblocks.
 *
 * Side effects:
 *	Pixel values in pict image changed.
 *
 *--------------------------------------------------------------
 */

static void
ProcessSkippedPFrameMBlocks(MPEGFrame *framePtr)
{
  int row_size, half_row, mb_row, mb_col, row, col, rr;
  int addr, row_incr, half_row_incr, crow, ccol;
  int *dest, *src, *dest1, *src1;
  MPEGSoftwareDecodeState *decodeState =
    (MPEGSoftwareDecodeState *) framePtr->deviceData;

  /* Calculate row sizes for luminance and Cr/Cb macroblock areas. */

  row_size = decodeState->mb_width << 4;
  half_row = (row_size >> 1);
  row_incr = row_size >> 2;
  half_row_incr = half_row >> 2;

  /* For each skipped macroblock, do... */

  for (addr = decodeState->mblock.past_mb_addr + 1;
       addr < decodeState->mblock.mb_address; addr++) {

    /* Calculate macroblock row and col. */

    mb_row = addr / decodeState->mb_width;
    mb_col = addr % decodeState->mb_width;

    /* Calculate upper left pixel row,col for luminance plane. */

    row = mb_row << 4;
    col = mb_col << 4;


    /* For each row in macroblock luminance plane... */

    dest = (int *)(decodeState->luminance->data + (row * row_size) + col);
    src = (int *)(((MPEGSoftwareDecodeState *) 
		   framePtr->ref1->deviceData)->luminance->data 
		  + (row * row_size) + col);
    
    for (rr = 0; rr < 8; rr++) {
	
	/* Copy pixel values from last I or P picture. */
	
	dest[0] = src[0];
	dest[1] = src[1];
	dest[2] = src[2];
	dest[3] = src[3];
	dest += row_incr;
	src += row_incr;
	
	dest[0] = src[0];
	dest[1] = src[1];
	dest[2] = src[2];
	dest[3] = src[3];
	dest += row_incr;
	src += row_incr;
    }
    
    /*
     * Divide row,col to get upper left pixel of macroblock in Cr and Cb
     * planes.
     */
    
    crow = row >> 1;
    ccol = col >> 1;
    
    /* For each row in Cr, and Cb planes... */
    
    dest = (int *)(decodeState->cr->data + (crow * half_row) + ccol);
    src = (int *)(((MPEGSoftwareDecodeState *) framePtr->ref1->deviceData)->cr->data 
		  + (crow * half_row) + ccol);
    dest1 = (int *)(decodeState->cb->data + (crow * half_row) + ccol);
    src1 = (int *)(((MPEGSoftwareDecodeState *) framePtr->ref1->deviceData)->cb->data
		   + (crow * half_row) + ccol);
    
    for (rr = 0; rr < 4; rr++) {
	
	/* Copy pixel values from last I or P picture. */
	
	dest[0] = src[0];
	dest[1] = src[1];
	
	dest1[0] = src1[0];
	dest1[1] = src1[1];
	
	dest += half_row_incr;
	src += half_row_incr;
	dest1 += half_row_incr;
	src1 += half_row_incr;
	
	dest[0] = src[0];
	dest[1] = src[1];
	
	dest1[0] = src1[0];
	dest1[1] = src1[1];
	
	dest += half_row_incr;
	src += half_row_incr;
	dest1 += half_row_incr;
	src1 += half_row_incr;
    }
}

  decodeState->mblock.recon_right_for_prev = 0;
  decodeState->mblock.recon_down_for_prev = 0;
}





/*
 *--------------------------------------------------------------
 *
 * ProcessSkippedBFrameMBlocks --
 *
 *	Processes skipped macroblocks in B frames.
 *
 * Results:
 *	Calculates pixel values for luminance, Cr, and Cb planes
 *      in current pict image for skipped macroblocks.
 *
 * Side effects:
 *	Pixel values in pict image changed.
 *
 *--------------------------------------------------------------
 */

static void
ProcessSkippedBFrameMBlocks(MPEGFrame *framePtr)
{
    int row_size, half_row, mb_row, mb_col, row, col, rr;
    int right_half_for, down_half_for, c_right_half_for, c_down_half_for;
    int right_half_back, down_half_back, c_right_half_back, c_down_half_back;
    int addr, right_for, down_for;
    int recon_right_for, recon_down_for;
    int recon_right_back, recon_down_back;
    int right_back, down_back;
    int c_right_for, c_down_for;
    int c_right_back, c_down_back;
    unsigned char forw_lum[256];
    unsigned char forw_cr[64], forw_cb[64];
    unsigned char back_lum[256], back_cr[64], back_cb[64];
    int row_incr, half_row_incr;
    int ccol, crow;
    MPEGSoftwareDecodeState *decodeState =
      (MPEGSoftwareDecodeState *) framePtr->deviceData;
    
    /* Calculate row sizes for luminance and Cr/Cb macroblock areas. */
    
    row_size = decodeState->mb_width << 4;
    half_row = (row_size >> 1);
    row_incr = row_size >> 2;
    half_row_incr =  half_row >> 2;
    
    /* Establish motion vector codes based on full pixel flag. */
    
    if (decodeState->picture.full_pel_forw_vector) {
	recon_right_for = decodeState->mblock.recon_right_for_prev << 1;
	recon_down_for = decodeState->mblock.recon_down_for_prev << 1;
    } else {
	recon_right_for = decodeState->mblock.recon_right_for_prev;
	recon_down_for = decodeState->mblock.recon_down_for_prev;
    }

    if (decodeState->picture.full_pel_back_vector) {
	recon_right_back = decodeState->mblock.recon_right_back_prev << 1;
	recon_down_back = decodeState->mblock.recon_down_back_prev << 1;
    } else {
	recon_right_back = decodeState->mblock.recon_right_back_prev;
	recon_down_back = decodeState->mblock.recon_down_back_prev;
    }

    /* Calculate motion vectors. */
    
    if (decodeState->mblock.bpict_past_forw) {
	right_for = recon_right_for >> 1;
	down_for = recon_down_for >> 1;
	right_half_for = recon_right_for & 0x1;
	down_half_for = recon_down_for & 0x1;
	
	recon_right_for /= 2;
	recon_down_for /= 2;
	c_right_for = recon_right_for >> 1;
	c_down_for = recon_down_for >> 1;
	c_right_half_for = recon_right_for & 0x1;
	c_down_half_for = recon_down_for & 0x1;
	
    }
    if (decodeState->mblock.bpict_past_back) {
	right_back = recon_right_back >> 1;
	down_back = recon_down_back >> 1;
	right_half_back = recon_right_back & 0x1;
	down_half_back = recon_down_back & 0x1;
	
	recon_right_back /= 2;
	recon_down_back /= 2;
	c_right_back = recon_right_back >> 1;
	c_down_back = recon_down_back >> 1;
	c_right_half_back = recon_right_back & 0x1;
	c_down_half_back = recon_down_back & 0x1;
	
    }
    /* For each skipped macroblock, do... */
    
    for (addr = decodeState->mblock.past_mb_addr + 1;
	 addr < decodeState->mblock.mb_address; addr++) {
    
	/* Calculate macroblock row and col. */
	
	mb_row = addr / decodeState->mb_width;
	mb_col = addr % decodeState->mb_width;
	
	/* Calculate upper left pixel row,col for luminance plane. */
	
	row = mb_row << 4;
	col = mb_col << 4;
	crow = row / 2;
	ccol = col / 2;
	
	/* If forward predicted, calculate prediction values. */
	
	if (decodeState->mblock.bpict_past_forw) {
	    
	    ReconSkippedBlock((unsigned char *)((MPEGSoftwareDecodeState *) 
			       framePtr->ref1->deviceData)->luminance->data,
			      forw_lum,
			      row, col, row_size, right_for, down_for,
			      right_half_for, down_half_for, 16);
	    
	    ReconSkippedBlock((unsigned char *)((MPEGSoftwareDecodeState *) 
			       framePtr->ref1->deviceData)->cr->data,
			      forw_cr, crow, ccol, half_row,
			      c_right_for, c_down_for, c_right_half_for, 
			      c_down_half_for, 8);
	    
	    ReconSkippedBlock((unsigned char *)((MPEGSoftwareDecodeState *) 
			       framePtr->ref1->deviceData)->cb->data,
			      forw_cb, crow, ccol, half_row,
			      c_right_for, c_down_for, c_right_half_for, 
			      c_down_half_for, 8);
	}
	/* If back predicted, calculate prediction values. */
	
	if (decodeState->mblock.bpict_past_back) {
	    ReconSkippedBlock((unsigned char *)((MPEGSoftwareDecodeState *) 
			       framePtr->ref2->deviceData)->luminance->data,
			      back_lum,
			      row, col, row_size, right_back, down_back,
			      right_half_back, down_half_back, 16);
	    
	    ReconSkippedBlock((unsigned char *)((MPEGSoftwareDecodeState *) 
			       framePtr->ref2->deviceData)->cr->data,
			      back_cr, crow, ccol, half_row,
			      c_right_back, c_down_back,
			      c_right_half_back, c_down_half_back, 8);
	    ReconSkippedBlock((unsigned char *)((MPEGSoftwareDecodeState *) 
			       framePtr->ref2->deviceData)->cb->data,
			      back_cb, crow, ccol, half_row,
			      c_right_back, c_down_back,
			      c_right_half_back, c_down_half_back, 8);
	}
	if (decodeState->mblock.bpict_past_forw &&
	    !decodeState->mblock.bpict_past_back) {
	    
	    int *dest, *dest1;
	    int *src, *src1;
	    dest = (int *)(decodeState->luminance->data + (row * row_size) + col);
	    src = (int *)forw_lum;
	    
	    for (rr = 0; rr < 16; rr++) {
		
		/* memcpy(dest, forw_lum+(rr<<4), 16);  */
		dest[0] = src[0];
		dest[1] = src[1];
		dest[2] = src[2];
		dest[3] = src[3];
		dest += row_incr;
		src += 4;
	    }
	    
	    dest = (int *)(decodeState->cr->data + (crow * half_row) + ccol);
	    dest1 = (int *)(decodeState->cb->data + (crow * half_row) + ccol);
	    src = (int *)forw_cr;
	    src1 = (int *)forw_cb;
	  
	    for (rr = 0; rr < 8; rr++) {
		/*
		 * memcpy(dest, forw_cr+(rr<<3), 8); memcpy(dest1, forw_cb+(rr<<3),
		 * 8);
		 */
		
		dest[0] = src[0];
		dest[1] = src[1];
		
		dest1[0] = src1[0];
		dest1[1] = src1[1];
		
		dest += half_row_incr;
		dest1 += half_row_incr;
		src += 2;
		src1 += 2;
	    }
	} else if (decodeState->mblock.bpict_past_back &&
		   !decodeState->mblock.bpict_past_forw) {
      
	    int *src, *src1;
	    int *dest, *dest1;
	    dest = (int *)(decodeState->luminance->data + (row * row_size) + col);
	    src = (int *)back_lum;
      
	    for (rr = 0; rr < 16; rr++) {
		dest[0] = src[0];
		dest[1] = src[1];
		dest[2] = src[2];
		dest[3] = src[3];
		dest += row_incr;
		src += 4;
	    }
	    
	    
	    dest = (int *)(decodeState->cr->data + (crow * half_row) + ccol);
	    dest1 = (int *)(decodeState->cb->data + (crow * half_row) + ccol);
	    src = (int *)back_cr;
	    src1 = (int *)back_cb;
	    
	    for (rr = 0; rr < 8; rr++) {
		/*
		 * memcpy(dest, back_cr+(rr<<3), 8); memcpy(dest1, back_cb+(rr<<3),
		 * 8);
		 */
	      
		dest[0] = src[0];
		dest[1] = src[1];
	      
		dest1[0] = src1[0];
		dest1[1] = src1[1];
	      
		dest += half_row_incr;
		dest1 += half_row_incr;
		src += 2;
		src1 += 2;
	    }
	} else {
	    
	    unsigned char *src1, *src2, *src1a, *src2a;
	    unsigned char *dest, *dest1;
	    dest = (unsigned char *)decodeState->luminance->data +
		   (row * row_size) + col;
	    src1 = forw_lum;
	    src2 = back_lum;
	    
	    for (rr = 0; rr < 16; rr++) {
		dest[0] = (int) (src1[0] + src2[0]) >> 1;
		dest[1] = (int) (src1[1] + src2[1]) >> 1;
		dest[2] = (int) (src1[2] + src2[2]) >> 1;
		dest[3] = (int) (src1[3] + src2[3]) >> 1;
		dest[4] = (int) (src1[4] + src2[4]) >> 1;
		dest[5] = (int) (src1[5] + src2[5]) >> 1;
		dest[6] = (int) (src1[6] + src2[6]) >> 1;
		dest[7] = (int) (src1[7] + src2[7]) >> 1;
		dest[8] = (int) (src1[8] + src2[8]) >> 1;
		dest[9] = (int) (src1[9] + src2[9]) >> 1;
		dest[10] = (int) (src1[10] + src2[10]) >> 1;
		dest[11] = (int) (src1[11] + src2[11]) >> 1;
		dest[12] = (int) (src1[12] + src2[12]) >> 1;
		dest[13] = (int) (src1[13] + src2[13]) >> 1;
		dest[14] = (int) (src1[14] + src2[14]) >> 1;
		dest[15] = (int) (src1[15] + src2[15]) >> 1;
		dest += row_size;
		src1 += 16;
		src2 += 16;
	    }
	    
	    
	    dest = (unsigned char *)decodeState->cr->data +
		   (crow * half_row) + ccol;
	    dest1 = (unsigned char *)decodeState->cb->data +
		    (crow * half_row) + ccol;
	    src1 = forw_cr;
	    src2 = back_cr;
	    src1a = forw_cb;
	    src2a = back_cb;
	    
	    for (rr = 0; rr < 8; rr++) {
		dest[0] = (int) (src1[0] + src2[0]) >> 1;
		dest[1] = (int) (src1[1] + src2[1]) >> 1;
		dest[2] = (int) (src1[2] + src2[2]) >> 1;
		dest[3] = (int) (src1[3] + src2[3]) >> 1;
		dest[4] = (int) (src1[4] + src2[4]) >> 1;
		dest[5] = (int) (src1[5] + src2[5]) >> 1;
		dest[6] = (int) (src1[6] + src2[6]) >> 1;
		dest[7] = (int) (src1[7] + src2[7]) >> 1;
		dest += half_row;
		src1 += 8;
		src2 += 8;
	
		dest1[0] = (int) (src1a[0] + src2a[0]) >> 1;
		dest1[1] = (int) (src1a[1] + src2a[1]) >> 1;
		dest1[2] = (int) (src1a[2] + src2a[2]) >> 1;
		dest1[3] = (int) (src1a[3] + src2a[3]) >> 1;
		dest1[4] = (int) (src1a[4] + src2a[4]) >> 1;
		dest1[5] = (int) (src1a[5] + src2a[5]) >> 1;
		dest1[6] = (int) (src1a[6] + src2a[6]) >> 1;
		dest1[7] = (int) (src1a[7] + src2a[7]) >> 1;
		dest1 += half_row;
		src1a += 8;
		src2a += 8;
	    }
	}
    }
}

/*****************************************************************
 *
 * GetSoftwareMpegDevice
 *
 * Parameters:
 *     objPtr - pointer to mpeg play object.
 *
 * Return:
 *     Pointer to new software device.
 *
 * Action:
 *     Creates device structure and fills it.
 *
 *****************************************************************/
MPEGDevice *
GetSoftwareMpegDevice (MPEGPlay *objPtr) 
{
    MPEGDevice *new_dev;

    new_dev = (MPEGDevice *) ckalloc(sizeof(MPEGDevice));

    new_dev->deviceData = NULL;
    new_dev->asyncDecodeFlag = 0;
    new_dev->decodeProc = MPEG_SoftDev_Decode;
    new_dev->refreshProc = MPEG_SoftDev_Refresh;
    new_dev->displayProc = MPEG_SoftDev_Display;
    new_dev->initProc = MPEG_SoftDev_Init;
    new_dev->reinitProc = MPEG_SoftDev_Reinit;
    new_dev->shutdownProc = MPEG_SoftDev_Shutdown;
    new_dev->freeFrameProc = MPEG_SoftDev_FreeFrame;
    new_dev->configProc = MPEG_SoftDev_Config;
    new_dev->nextFrameProc = MPEG_SoftDev_NextFrame; 
    new_dev->errorProc = MPEG_SoftDev_Error;

    return new_dev;
}

/*********************************************************************
 *
 * MPEG_SoftDev_Decode
 *
 * Parameters:
 *     framePtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_SoftDev_Decode(MPEGFrame *framePtr)
{
    MPEGPlay *objPtr;
    MPEGDevice *devicePtr;
    MPEGSoftwareDevice *softDevPtr;
    MPEGSoftwareDecodeState *decodeState;
    int do_mblocks, i;
    double start_decode_time, end_decode_time;
    double idle_end_time, idle_time;
    unsigned int nextData;
    int new_remaining, num_done;
    int scale;
    int w,h;

    w = framePtr->frameHeader.width;
    h = framePtr->frameHeader.height;

    if (framePtr == NULL) return 0;
    
    objPtr = framePtr->objPtr;
    if (objPtr == NULL) return 0;
    
    devicePtr = objPtr->devicePtr;
    if (devicePtr == NULL) return 0;

    softDevPtr = (MPEGSoftwareDevice *) devicePtr->deviceData;
    if (softDevPtr == NULL) return 0;

    /*
     * Check device client data in frame ptr.
     * NULL indicates new frame not in progress.
     *      Allocate structures and fill them.
     *      Parse off picture header.
     *      Set device's asynchDecode flag to 1.
     * Calculate how many macroblocks to be decoded.
     * Decode them.
     * If frame is completely decoded:
     *    Dither the picture.
     *    Set device's asynchDecode flag to 0.
     */
    
    if (framePtr->deviceData == NULL) {
	framePtr->deviceData = (ClientData) ckalloc(sizeof(MPEGSoftwareDecodeState));
	decodeState = (MPEGSoftwareDecodeState *) framePtr->deviceData;

	decodeState->mb_width = ((framePtr->frameHeader.width + 15) / 16);
	decodeState->mb_height = ((framePtr->frameHeader.height + 15) / 16);
	decodeState->num_mblocks = (decodeState->mb_width *
				    decodeState->mb_height);

	decodeState->mblocks_remaining = decodeState->num_mblocks;
	
	decodeState->luminance = MakeBuffer(objPtr->interp, 
					    decodeState->num_mblocks*16*16, NULL);
	decodeState->cr = MakeBuffer(objPtr->interp, 
				     decodeState->num_mblocks*8*8, NULL);
	decodeState->cb = MakeBuffer(objPtr->interp, 
				     decodeState->num_mblocks*8*8, NULL);

	
	if (softDevPtr->ditherRec->depth == 8) scale = 1;
	else scale = 4;

	if (softDevPtr->shared_mem_flag) {
	    decodeState->display = MakeSharedBuffer(objPtr->interp, 
						    decodeState->num_mblocks*16*16*scale,
						    NULL);
	    if (decodeState->display == NULL) {

		decodeState->display = MakeBuffer(objPtr->interp, 
						  decodeState->num_mblocks*16*16*scale, NULL);
	    }
	} else {
	    decodeState->display = MakeBuffer(objPtr->interp, 
					      decodeState->num_mblocks*16*16*scale, NULL);
	}

	decodeState->curData = (unsigned char *) framePtr->data;
	decodeState->offset = 0;
	decodeState->errorFlag = 0;
	decodeState->idle_start_time = 0.0;

	bp_Start(decodeState->curData, decodeState->offset);
	NextStartCode();
	bp_Peek(32, &nextData);
	
	if (nextData != MPEG_PICTURE_START_CODE) {
	    decodeState->errorFlag = 1;
	    devicePtr->asyncDecodeFlag = 0;
	    decodeState->idle_start_time = 0.0;
	    return DECODE_ERROR;
	}

	if (ParsePicture(framePtr) != PARSE_OK) {
	    decodeState->errorFlag = 1;
	    devicePtr->asyncDecodeFlag = 0;
	    decodeState->idle_start_time = 0.0;
	    return DECODE_ERROR;
	}	
	devicePtr->asyncDecodeFlag = 1;
    } else {
	decodeState = (MPEGSoftwareDecodeState *) framePtr->deviceData;

	if (decodeState->idle_start_time > 0.0) {
	    idle_end_time = Tcm_ReadSysClock();
	    idle_time = idle_end_time - decodeState->idle_start_time;
	    if (softDevPtr->secPerIdle != 0.0) {
		softDevPtr->secPerIdle = ((0.80 * softDevPtr->secPerIdle) +
					  (0.20 * idle_time));
	    } else {
		softDevPtr->secPerIdle = idle_time;
	    }
	}
	bp_Start(decodeState->curData, decodeState->offset);
    }

    if (softDevPtr->secPerMblock != 0.0)
      do_mblocks = (0.010 / ((double) softDevPtr->secPerMblock));
    else do_mblocks = 200;

    if (do_mblocks <= 0) do_mblocks = 1;

    if (do_mblocks > decodeState->mblocks_remaining) {
	do_mblocks = decodeState->mblocks_remaining;
    }

    start_decode_time = Tcm_ReadSysClock();

    for (i=0; i<do_mblocks; i++) {
	
	if (decodeState->mblock.mb_address == decodeState->num_mblocks-1) 
	  break;

	/*
	 * Check for start code. If so, check for slice start.
	 */

	bp_Peek(23, &nextData);

	if (nextData == 0) {

	    /*
	     * Found a start code. Must be slice.
	     */

	    NextStartCode();
	    bp_Peek(32, &nextData);

	    if ((nextData < SLICE_MIN_START_CODE) || (nextData > SLICE_MAX_START_CODE)) {
		/*
		 * Unexpected start code.
		 */
		
		decodeState->errorFlag = 1;
		devicePtr->asyncDecodeFlag = 0;
		decodeState->idle_start_time = 0.0;
		return DECODE_ERROR;
	    }
	    
	    if (ParseSlice(framePtr) != PARSE_OK) {
		decodeState->errorFlag = 1;
		devicePtr->asyncDecodeFlag = 0;
		decodeState->idle_start_time = 0.0;
		return DECODE_ERROR;
	    }
	}
	
	if (ParseMacroBlock(framePtr) != PARSE_OK) {
	    decodeState->errorFlag = 1;
	    devicePtr->asyncDecodeFlag = 0;
	    decodeState->idle_start_time = 0.0;
	    return DECODE_ERROR;
	}
    }

    end_decode_time = Tcm_ReadSysClock();

    new_remaining = decodeState->num_mblocks -
      decodeState->mblock.mb_address - 1;

    num_done = decodeState->mblocks_remaining - new_remaining;
    if (num_done <= 0) {
	num_done = 1;
    }

    decodeState->mblocks_remaining = new_remaining;

    if (num_done > 0) {
	if (softDevPtr->secPerMblock > 0.0) {
	    softDevPtr->secPerMblock = ((0.80 * softDevPtr->secPerMblock) +
					(0.20 * ((end_decode_time - start_decode_time) /
						 ((double) num_done))));
	} else {
	    softDevPtr->secPerMblock = ((end_decode_time - start_decode_time) /
					((double) num_done));
	}
    }

    if (decodeState->mblocks_remaining == 0) {
	double start_dither_time, end_dither_time;

	/*
	 * Dither picture now.
	 */

	start_dither_time = Tcm_ReadSysClock();
	
	if (softDevPtr->ditherProc != NULL) {
	  if (softDevPtr->ditherRec->depth <= 8) {
	    softDevPtr->ditherProc(softDevPtr->ditherRec,
				   0, 0, 
				   decodeState->mb_width*16, 
				   decodeState->mb_height*16, 
				   decodeState->luminance->data,
				   decodeState->cr->data,
				   decodeState->cb->data,
				   decodeState->display->data,
				   decodeState->mb_width*16,
				   decodeState->mb_width*8,
				   decodeState->mb_width*8,
				   decodeState->mb_width*16,
				   decodeState->mb_width*16, 
				   decodeState->mb_height*16);
#ifdef _WINDOWS
	  } else if (softDevPtr->ditherRec->depth == 24) {
	    softDevPtr->ditherProc(softDevPtr->ditherRec,
				   0, 0, 
				   decodeState->mb_width*16, 
				   decodeState->mb_height*16, 
				   decodeState->luminance->data,
				   decodeState->cr->data,
				   decodeState->cb->data,
				   decodeState->display->data,
				   decodeState->mb_width*16,
				   decodeState->mb_width*8,
				   decodeState->mb_width*8,
				   (((decodeState->mb_width*16*3) + 3) / 4) * 4,
				   decodeState->mb_width*16, 
				   decodeState->mb_height*16);
	    
#endif

	  } else if (softDevPtr->ditherRec->depth >= 24) {
	    softDevPtr->ditherProc(softDevPtr->ditherRec,
				   0, 0, 
				   decodeState->mb_width*16, 
				   decodeState->mb_height*16, 
				   decodeState->luminance->data,
				   decodeState->cr->data,
				   decodeState->cb->data,
				   decodeState->display->data,
				   decodeState->mb_width*16,
				   decodeState->mb_width*8,
				   decodeState->mb_width*8,
				   decodeState->mb_width*16*4,
				   decodeState->mb_width*16, 
				   decodeState->mb_height*16);
	  }
	}
	end_dither_time = Tcm_ReadSysClock();

	if (softDevPtr->secPerDither != 0.0) {
	    softDevPtr->secPerDither = ((0.80 * softDevPtr->secPerDither) + 
					(0.20 * (end_dither_time - start_dither_time)));
	} else {
	    softDevPtr->secPerDither = end_dither_time - start_dither_time;
	}

	devicePtr->asyncDecodeFlag = 0;
    }     
    if (devicePtr->asyncDecodeFlag) {
	decodeState->idle_start_time = Tcm_ReadSysClock();
	bp_End (&(decodeState->curData), &(decodeState->offset));
    } else {
	decodeState->idle_start_time = 0.0;
    }

    return 0;
}


#ifdef HAVE_SHARED_MEMORY
/*
 *--------------------------------------------------------------
 *
 * LookForShmEvent --
 *
 *	This little helper funciton is used in combination with
 *	the XIfEvent call below to look for ShmCompletion events.
 *
 * Results:
 *	1 if event is a ShmCompletion event, 0 otherwise.
 *
 * Side effects:
 *	None.
 *
 * Parameters:
 *
 *--------------------------------------------------------------
 */
static int
LookForShmEvent (Display *display, XEvent *e, char *arg)
{
    MPEGPlay *objPtr;
    MPEGSoftwareDevice *softDevPtr;
    XAnyEvent *eventPtr = (XAnyEvent *)e;

    objPtr = (MPEGPlay *) arg;
    softDevPtr = (MPEGSoftwareDevice *) objPtr->devicePtr->deviceData;
    return ((eventPtr->type == softDevPtr->ShmEvent) && 
	    (eventPtr->window == objPtr->xid));
}
#endif

/*
 *--------------------------------------------------------------
 *
 * WaitForCompletion --
 *
 *	Blocks waiting for a XShmCompletion event from the X server.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	May block
 *
 *--------------------------------------------------------------
 */
static void
WaitForCompletion (MPEGPlay *objPtr)
{
    MPEGSoftwareDevice *softDevPtr =
	(MPEGSoftwareDevice *)  objPtr->devicePtr->deviceData;

#ifdef HAVE_SHARED_MEMORY
    XEvent shmEvent;

    if (!objPtr->xid) {
	softDevPtr->waiting = 0;
	return;
    }
    if (softDevPtr->waiting) {
	XIfEvent(Tcm_Display(), &shmEvent, LookForShmEvent, (char *) objPtr);
	softDevPtr->waiting = 0;
    }
#else
    softDevPtr->waiting = 0;
#endif

}



#ifdef HAVE_SHARED_MEMORY
/*
 *--------------------------------------------------------------
 *
 * ProcessShmEvent --
 *
 *	Called when a shmCompletion event is received.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */
static int
ProcessShmEvent (ClientData clientData, XEvent *e)
{
    MPEGPlay *objPtr = (MPEGPlay *)clientData;
    MPEGSoftwareDevice *softDevPtr = 
      (MPEGSoftwareDevice *) objPtr->devicePtr->deviceData;
    XAnyEvent *shmEventPtr = (XAnyEvent *)e;

    if (!softDevPtr->waiting) {
	return 0;
    } else if (!objPtr->xid) {
	softDevPtr->waiting = 0;
    } else if ((softDevPtr->waiting) &&
        (shmEventPtr->type == softDevPtr->ShmEvent) && 
	(shmEventPtr->window == objPtr->xid)) {
	softDevPtr->waiting = 0;
	return 1;
    }
    return 0;
}
#endif

/*********************************************************************
 *
 * MPEG_SoftDev_Refresh
 *
 * Parameters:
 *     objPtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_SoftDev_Refresh(MPEGPlay *objPtr)
{
    MPEGDevice *devicePtr;
    MPEGSoftwareDevice *softDevPtr;
    int w, h, xoff, yoff;

    devicePtr = objPtr->devicePtr;
    if (devicePtr == NULL) return 0;

    softDevPtr = (MPEGSoftwareDevice *) devicePtr->deviceData;
    if (softDevPtr == NULL) return 0;

    if (softDevPtr->image == NULL) return 0;
    if (softDevPtr->last_display == NULL) return 0;

    /*
     * Wait for completion.
     */

    WaitForCompletion(objPtr);

    
    /*
     * Calculate X and Y offset for centering, etc.
     */

    w = objPtr->width;
    h = objPtr->height;
    xoff = objPtr->xpos;
    yoff = objPtr->ypos;

    if (w == 0) {
	w = softDevPtr->last_width;
	xoff = 0;
    } else if (w != softDevPtr->last_width) {
	if (xoff == -1) {
	    if (w > softDevPtr->last_width) {
		xoff = (w - softDevPtr->last_width - 1 ) / 2;
	    }
	} else {
	    w = softDevPtr->last_width;
	}
    }

    if (h == 0) {
	h = softDevPtr->last_height;
	yoff = 0;
    } else if (h != softDevPtr->last_height) {
	if (yoff == -1) {
	    if (h > softDevPtr->last_height) {
		yoff = (h - softDevPtr->last_height - 1 ) / 2;
	    }
	} else {
	    h = softDevPtr->last_height;
	}
    }

    if (softDevPtr->shared_mem_flag) {
	
#ifdef HAVE_SHARED_MEMORY
	if (softDevPtr->last_display->shmId != -1) {
	    memcpy(&(softDevPtr->last_shminfo), &(softDevPtr->last_display->shminfo),
		   (sizeof(XShmSegmentInfo)));
	    
	    XShmPutImage(Tcm_Display(), objPtr->xid, softDevPtr->gc, 
			 softDevPtr->image, 0, 0, xoff, yoff, 
			 softDevPtr->last_width, 
			 softDevPtr->last_height, 1);
	    softDevPtr->waiting = 1;
	} else {
	    XPutImage(Tcm_Display(), objPtr->xid, softDevPtr->gc,
		      softDevPtr->image, 0, 0, xoff, yoff,
		      softDevPtr->last_width,
		      softDevPtr->last_height);
	    softDevPtr->waiting = 0;
	}
#endif
    } else {
#ifndef _WINDOWS
	XPutImage(Tcm_Display(), objPtr->xid, softDevPtr->gc,
		  softDevPtr->image, 0, 0, xoff, yoff,
		  softDevPtr->last_width, 
		  softDevPtr->last_height);
#else
	{
	    HPALETTE pal, palPrev;
	    
	    pal = XGetWindowsPalette(Tcm_Display(), 
				     XDefaultColormap(Tcm_Display(), 
						      DefaultScreen(Tcm_Display())));
	    palPrev = SelectPalette(tcm_DitherHDC, pal, FALSE);
	    
	    SetDIBitsToDevice(tcm_DitherHDC, xoff, yoff,
			      softDevPtr->last_width,softDevPtr->last_height,
			      0, 0, 0, softDevPtr->last_height,
			      softDevPtr->image_data,
			      softDevPtr->image,
			      ((softDevPtr->ditherRec->depth == 8) ?
			       DIB_PAL_COLORS : DIB_RGB_COLORS));
	    SelectPalette(tcm_DitherHDC, palPrev, FALSE);
	}

#endif
	softDevPtr->waiting = 0;
    }
    return 0;
}


/*********************************************************************
 *
 * MPEG_SoftDev_Display
 *
 * Parameters:
 *     framePtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int 
MPEG_SoftDev_Display(MPEGFrame *framePtr)
{
    MPEGSoftwareDevice *softDevPtr;
    MPEGDevice *devicePtr;
    MPEGSoftwareDecodeState *decodeState;
    MPEGPlay *objPtr;
    int w, h, xoff, yoff;
    int imWidth, imHeight;

    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    softDevPtr = (MPEGSoftwareDevice *) devicePtr->deviceData;
    decodeState = (MPEGSoftwareDecodeState *) framePtr->deviceData;

    imWidth = framePtr->frameHeader.width;
    imHeight = framePtr->frameHeader.height;

#if 0
    {
      char logStr[256];
      double now;

      now = LTS_Value(objPtr->lts);
      sprintf(logStr, "Displaying: start: %f now: %f end: %f index: %d type: %d", framePtr->logStart, 
	      now, framePtr->logEnd, framePtr->frameHeader.index, framePtr->frameHeader.type);
      Tcm_Log(logStr);
    }
#endif

    WaitForCompletion(objPtr);

    /* 
     * Check to see if we need to reallocate the image.
     */

    if ((decodeState->mb_width * 16 != softDevPtr->last_width) ||
	(decodeState->mb_height * 16 != softDevPtr->last_height) ||
	(softDevPtr->image == NULL)) {
      
      if (softDevPtr->image != NULL) {
#ifndef _WINDOWS
	softDevPtr->image->data = NULL;
	XDestroyImage(softDevPtr->image);
#else
	softDevPtr->image_data = NULL;
	free((char *) softDevPtr->image);
#endif
      }

      if (softDevPtr->shared_mem_flag) {
#ifdef HAVE_SHARED_MEMORY
	softDevPtr->image = XShmCreateImage(Tcm_Display(), softDevPtr->ditherRec->visual, 
						  softDevPtr->ditherRec->depth, 
						  ZPixmap, NULL, &(softDevPtr->last_shminfo),
						  (decodeState->mb_width*16),
						  (decodeState->mb_height*16));
#endif
      } else {
#ifndef _WINDOWS
	softDevPtr->image = XCreateImage(Tcm_Display(), softDevPtr->ditherRec->visual, 
					  softDevPtr->ditherRec->depth, 
					  ZPixmap, 0, NULL,
					  (decodeState->mb_width * 16),
					  (decodeState->mb_height * 16), 
					  8, 0);
#else
	int i;
	short *pPalIndex;
	int allocSize;

	allocSize = sizeof(BITMAPINFO);
	if (softDevPtr->ditherRec->depth == 8) {
	    allocSize += 256 * sizeof(short);
	}

	softDevPtr->image = (BITMAPINFO *) malloc(allocSize);

	softDevPtr->image->bmiHeader.biSize = (LONG)sizeof(BITMAPINFOHEADER);
	softDevPtr->image->bmiHeader.biWidth = imWidth;
	softDevPtr->image->bmiHeader.biHeight = imHeight;
	softDevPtr->image->bmiHeader.biPlanes = 1;
	softDevPtr->image->bmiHeader.biBitCount = softDevPtr->ditherRec->depth;
	softDevPtr->image->bmiHeader.biCompression = 0L;
	softDevPtr->image->bmiHeader.biSizeImage = 0L;
	softDevPtr->image->bmiHeader.biXPelsPerMeter = 0L;
	softDevPtr->image->bmiHeader.biYPelsPerMeter = 0L;
	softDevPtr->image->bmiHeader.biClrUsed = 0L;
	softDevPtr->image->bmiHeader.biClrImportant = 0L;
	pPalIndex = (short *) softDevPtr->image->bmiColors;
	if (softDevPtr->ditherRec->depth == 8) {
	    for (i = 0; i < 256; i ++) {
		pPalIndex[i] = i;
	    }
	}
	softDevPtr->image_data = NULL;
#endif
      }

      softDevPtr->last_width = decodeState->mb_width * 16;
      softDevPtr->last_height = decodeState->mb_height * 16;      
    }

    if (softDevPtr->last_display != NULL) {
      DetachBuffer(objPtr->interp, softDevPtr->last_display, NULL);
    }

    softDevPtr->last_display = decodeState->display;
    AttachBuffer(objPtr->interp, softDevPtr->last_display, NULL);

#ifndef _WINDOWS
    softDevPtr->image->data = decodeState->display->data;
#else
    softDevPtr->image_data = decodeState->display->data;
#endif

    /*
     * Calculate X and Y offset for centering, etc.
     */

    w = objPtr->width;
    h = objPtr->height;
    xoff = objPtr->xpos;
    yoff = objPtr->ypos;

    if (w == 0) {
      w = softDevPtr->last_width;
      xoff = 0;
    } else if (w != softDevPtr->last_width) {
      if (xoff == -1) {
	if (w > softDevPtr->last_width) {
	  xoff = (w - softDevPtr->last_width - 1 ) / 2;
	}
      } else {
	w = softDevPtr->last_width;
      }
    }

    if (h == 0) {
      h = softDevPtr->last_height;
      yoff = 0;
    } else if (h != softDevPtr->last_height) {
      if (yoff == -1) {
	if (h > softDevPtr->last_height) {
	  yoff = (h - softDevPtr->last_height - 1 ) / 2;
	}
      } else {
	h = softDevPtr->last_height;
      }
    }

    if (softDevPtr->shared_mem_flag) {
      
#ifdef HAVE_SHARED_MEMORY
      if (softDevPtr->last_display->shmId != -1) {
	memcpy(&(softDevPtr->last_shminfo), &(softDevPtr->last_display->shminfo),
	       (sizeof(XShmSegmentInfo)));

	XShmPutImage(Tcm_Display(), objPtr->xid, softDevPtr->gc, 
		     softDevPtr->image, 0, 0, xoff, yoff, 
		     softDevPtr->last_width, 
		     softDevPtr->last_height, 1);
	softDevPtr->waiting = 1;
      } else {
	XPutImage(Tcm_Display(), objPtr->xid, softDevPtr->gc,
		  softDevPtr->image, 0, 0, xoff, yoff,
		  softDevPtr->last_width, 
		  softDevPtr->last_height);
	softDevPtr->waiting = 0;
      }
#endif
    } else {
#ifndef _WINDOWS
      XPutImage(Tcm_Display(), objPtr->xid, softDevPtr->gc,
		softDevPtr->image, 0, 0, xoff, yoff,
		softDevPtr->last_width, 
		softDevPtr->last_height);
#else
      {
	  HPALETTE pal, palPrev;
	  
	  pal = XGetWindowsPalette(Tcm_Display(), 
				   XDefaultColormap(Tcm_Display(), 
						    DefaultScreen(Tcm_Display())));
	  palPrev = SelectPalette(tcm_DitherHDC, pal, FALSE);
	  
	  SetDIBitsToDevice(tcm_DitherHDC, xoff, yoff, 
			    softDevPtr->last_width, 
			    softDevPtr->last_height,
			    0, 0, 0, softDevPtr->last_height,
			    softDevPtr->image_data,
			    softDevPtr->image,
			    ((softDevPtr->ditherRec->depth == 8) ?
			     DIB_PAL_COLORS : DIB_RGB_COLORS));
	  SelectPalette(tcm_DitherHDC, palPrev, FALSE);
      }
      
#endif
      softDevPtr->waiting = 0;
    }
    return 0;
}


/*
 *------------------------------------------------------------
 *
 * NoXShm
 *
 *	Registered temporarily as a Tk error handler so that
 *	if we can't use the XShm extension (for example if the
 *	display is remote) the hasShm flag is unset.
 *
 * Results:
 *	Always returns 0 so the program doesn't exit
 *
 * Side Effects:
 *	Unsets the shared_mem_flag flag
 *
 *------------------------------------------------------------
 */
static int
NoXShm(ClientData cd, XErrorEvent *evt)
{
    MPEGSoftwareDevice *swDevice = (MPEGSoftwareDevice *) cd;
    
    swDevice->shared_mem_flag = 0;
    return 0;
}


/*********************************************************************
 *
 * MPEG_SoftDev_Init
 *
 * Parameters:
 *     objPtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_SoftDev_Init(MPEGPlay *objPtr)
{
    MPEGDevice *devicePtr;
    MPEGSoftwareDevice *softDevPtr;
    DitherParam ditherParam;
#ifdef HAVE_SHARED_MEMORY
    XShmSegmentInfo si;
#endif

    init_tables();

    if (objPtr == NULL) return 0;
   
    devicePtr = objPtr->devicePtr;
    if (devicePtr == NULL) return 0;

    softDevPtr = (MPEGSoftwareDevice *) devicePtr->deviceData;
    if (softDevPtr == NULL) {
	softDevPtr = (MPEGSoftwareDevice *) ckalloc(sizeof(MPEGSoftwareDevice));
	devicePtr->deviceData = (ClientData) softDevPtr;
    }

    softDevPtr->secPerMblock = 0.0;
    softDevPtr->secPerDither = 0.0;
    softDevPtr->secPerIdle = 0.0;
    softDevPtr->ditherProc = NULL;
    softDevPtr->ditherRec = NULL;
    softDevPtr->image = NULL;
#ifdef _WINDOWS
    softDevPtr->image_data = NULL;
#else
    softDevPtr->gc = XCreateGC(Tcm_Display(), objPtr->xid, 0, 0);
#endif
    softDevPtr->last_display = NULL;
    softDevPtr->last_width = softDevPtr->last_height = 0;
    /*
     * Initialize dither stuff here.
     */

    ditherParam.dType = CHOOSE_BEST;
    ditherParam.minColors = 8;
    ditherParam.maxColors = 256;
    ditherParam.pFormat = YCRCB411;

    softDevPtr->ditherRec = InitDither(ditherParam,
				       Tcm_Display(),
				       objPtr->xid,
				       &(softDevPtr->ditherProc));
    if (softDevPtr->ditherRec == NULL) {
      softDevPtr->ditherProc = NULL;
      return DITHER_INIT_ERROR;
    }

#ifdef HAVE_SHARED_MEMORY
    if (cmXShmQueryExtension(Tcm_Display())) {
	softDevPtr->shared_mem_flag = 1;
	softDevPtr->ShmEvent = XShmGetEventBase(Tcm_Display())+ShmCompletion;
	Tk_CreateGenericHandler(ProcessShmEvent, (ClientData) objPtr);
    } else {
      softDevPtr->shared_mem_flag = 0;
    }
#else
    softDevPtr->shared_mem_flag = 0;
#endif
    softDevPtr->waiting = 0;
    return 0;
}


/*********************************************************************
 *
 * MPEG_SoftDev_Reinit
 *
 * Parameters:
 *     objPtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_SoftDev_Reinit(MPEGPlay *objPtr)
{
    MPEGDevice *devicePtr;
    MPEGSoftwareDevice *softDevPtr;

    devicePtr = objPtr->devicePtr;
    if (devicePtr == NULL) return 0;

    softDevPtr = (MPEGSoftwareDevice *) devicePtr->deviceData;
    if (softDevPtr == NULL) return 0;

    softDevPtr->secPerMblock = 0.0;
    softDevPtr->secPerDither = 0.0;
    softDevPtr->secPerIdle = 0.0;
    
#ifndef _WINDOWS
    if (softDevPtr->gc != NULL) {
	XFreeGC(Tcm_Display(), softDevPtr->gc);
	softDevPtr->gc = NULL;
    }
    softDevPtr->gc = XCreateGC(Tcm_Display(), objPtr->xid, 0, 0);
#endif
    softDevPtr->waiting = 0;

    return 0;
}


/*********************************************************************
 *
 * MPEG_SoftDev_Shutdown
 *
 * Parameters:
 *     objPtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_SoftDev_Shutdown(MPEGPlay *objPtr) 
{
    MPEGSoftwareDevice *softDevPtr;

    if (objPtr == NULL) return 0;
    if (objPtr->devicePtr == NULL) return 0;

    softDevPtr = (MPEGSoftwareDevice *) objPtr->devicePtr->deviceData;

    if (softDevPtr != NULL) {
	if (softDevPtr->ditherRec != NULL) {
	    FreeDither(softDevPtr->ditherRec);
	    softDevPtr->ditherRec = NULL;
	    softDevPtr->ditherProc = NULL;
	}

#ifndef _WINDOWS
	if (softDevPtr->gc != None) {
	    XFreeGC(Tcm_Display(), softDevPtr->gc);
	}
#endif
	
	if (softDevPtr->last_display != NULL) {
	    DetachBuffer(objPtr->interp, softDevPtr->last_display, NULL);
	    softDevPtr->last_display = NULL;
	}

	if (softDevPtr->image != NULL) {
#ifndef _WINDOWS
	    softDevPtr->image->data = NULL;
	    XDestroyImage(softDevPtr->image);
#else
	    softDevPtr->image_data = NULL;
	    free((char *) softDevPtr->image);
#endif
	    softDevPtr->image = NULL;
	}

#ifdef HAVE_SHARED_MEMORY
	Tk_DeleteGenericHandler(ProcessShmEvent, (ClientData)objPtr);
#endif

	ckfree((char *) softDevPtr);
    }
    objPtr->devicePtr->deviceData = NULL;
    return 0;
}


/*********************************************************************
 *
 * MPEG_SoftDev_FreeFrame
 *
 * Parameters:
 *     framePtr
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_SoftDev_FreeFrame(MPEGFrame *framePtr)
{
    MPEGPlay *objPtr;
    MPEGSoftwareDecodeState *decodeState;

    objPtr = framePtr->objPtr;
    decodeState = (MPEGSoftwareDecodeState *) framePtr->deviceData;

    if (framePtr->deviceData != NULL) {
	DetachBuffer(objPtr->interp, decodeState->luminance, NULL);
	DetachBuffer(objPtr->interp, decodeState->cr, NULL);
	DetachBuffer(objPtr->interp, decodeState->cb, NULL);
	DetachBuffer(objPtr->interp, decodeState->display, NULL);
    
	ckfree((char *) decodeState);
	framePtr->deviceData = NULL;
    }

    return 0;
}


/*********************************************************************
 *
 * MPEG_SoftDev_Config
 *
 * Parameters:
 *     objPtr
 *     option
 *     value
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int 
MPEG_SoftDev_Config(MPEGPlay *objPtr, char *option, char *value)
{
    return 0;
}

/*********************************************************************
 *
 * MPEG_SoftDev_NextFrame
 *
 * Parameters:
 *     objPtr
 *     frameList, 
 *     chosenFrame
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static int
MPEG_SoftDev_NextFrame(MPEGPlay *objPtr, MPEGFrame *frameList, MPEGFrame **chosenFrame)
{
    MPEGFrame *framePtr, *latestFrame;
    double speed, now, end, decode;

    MPEGSoftwareDevice *softDevPtr = 
      (MPEGSoftwareDevice *) objPtr->devicePtr->deviceData;

    if (frameList == NULL) {
	*chosenFrame = NULL;
	return 1;
    }

    speed = LTS_Speed(objPtr->lts);
    now = LTS_Value(objPtr->lts);

    latestFrame = frameList;

    for (framePtr = frameList; framePtr != NULL; framePtr = framePtr->next) {
	end = (speed < 0.0) ? framePtr->logStart : framePtr->logEnd;

	decode = now + speed*((softDevPtr->secPerMblock*
			       (framePtr->frameHeader.width*
				framePtr->frameHeader.height/(16*16))) +
			      softDevPtr->secPerDither);
	if (decode <= end) {
	    *chosenFrame = framePtr;
	    return 0;
	}
	
	if (speed < 0.0) {
	    if (framePtr->logEnd < latestFrame->logEnd) {
		latestFrame = framePtr;
	    }
	} else {
	    if (framePtr->logEnd > latestFrame->logEnd) {
		latestFrame = framePtr;
	    }
	}
    }

    if (speed > 0.0) {
	if (latestFrame->logEnd < now) {
	    *chosenFrame = NULL;
	    return 1;
	}
    } else if (speed < 0.0) {
	if (latestFrame->logEnd > now) {
	    *chosenFrame = NULL;
	    return 1;
	}
    }

    *chosenFrame = latestFrame;
    return 0;
}


/*********************************************************************
 *
 * MPEG_SoftDev_Error
 *
 * Parameters:
 *     objPtr
 *     errNum
 *
 * Results:
 *
 * Actions:
 *
 *********************************************************************/
static char *dummyStr="Soft Dev dummy string";

static char *
MPEG_SoftDev_Error(MPEGPlay *objPtr, int errNum) 
{
    printf("In Soft Dev error: %d\n", errNum);
    return dummyStr;
}


/*
 *--------------------------------------------------------------
 *
 * ParsePicture --
 *
 *      Parses picture header. 
 *
 * Results:
 *      Values from picture header put into decode state structure.
 *
 * Side effects:
 *      Bit stream irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

static int
ParsePicture(MPEGFrame *framePtr)
{
    MPEGPlay *objPtr;
    MPEGDevice *devicePtr;
    MPEGSoftwareDevice *softDevPtr;
    MPEGSoftwareDecodeState *decodeState;

    unsigned int data;

    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    softDevPtr = (MPEGSoftwareDevice *) devicePtr->deviceData;
    decodeState = (MPEGSoftwareDecodeState *) framePtr->deviceData;


    /* Flush header start code. */
    
    bp_Flush(32);

    /* Parse off temporal reference. */
    bp_Flush(10);

    /* Parse of picture type. */
    decodeState->picture.type = framePtr->frameHeader.type;
    bp_Flush(3);

    /* Parse off vbv buffer delay value. */
    bp_Flush(16);

    /* If P or B type frame... */

    if ((decodeState->picture.type == MPEG_P_FRAME) ||
	(decodeState->picture.type == MPEG_B_FRAME)) {

	/* Parse off forward vector full pixel flag. */
	bp_Get(1, &data);
	
	if (data)
	  decodeState->picture.full_pel_forw_vector = 1;
	else
	  decodeState->picture.full_pel_forw_vector = 0;

	/* Parse of forw_r_code. */
	bp_Get(3, &data);

	/* Decode forw_r_code into forw_r_size and forw_f. */

	decodeState->picture.forw_r_size = data - 1;
	decodeState->picture.forw_f = (1 << decodeState->picture.forw_r_size);
    }
    /* If B type frame... */

    if (decodeState->picture.type == MPEG_B_FRAME) {

	/* Parse off back vector full pixel flag. */
	bp_Get(1, &data);

	if (data)
	  decodeState->picture.full_pel_back_vector = 1;
	else
	  decodeState->picture.full_pel_back_vector = 0;

	/* Parse off back_r_code. */
	bp_Get(3, &data);

	/* Decode back_r_code into back_r_size and back_f. */

	decodeState->picture.back_r_size = data - 1;
	decodeState->picture.back_f = (1 << decodeState->picture.back_r_size);
    }
    /* Get extra bit picture info. */

    ParseExtraBitInfo();

    /* Goto next start code. */
    NextStartCode();

    /* If start code is extension start code, parse off extension data. */
    
    bp_Peek(32, &data);
    if (data == MPEG_EXTENSION_START_CODE) {
	
	bp_Flush(32);
	NextStartCode();
    }

    /* If start code is user start code, parse off user data. */

    bp_Peek(32, &data);
    if (data == MPEG_USER_START_CODE) {
	
	bp_Flush(32);
	NextStartCode();
    }

    decodeState->mblock.mb_address = -1;
    decodeState->mblock.past_mb_addr = -1;

    return PARSE_OK;
}


/*
 *--------------------------------------------------------------
 *
 * ParseSlice --
 *
 *      Parses off slice header.
 *
 * Results:
 *      Values found in slice header put into decode state.
 *
 * Side effects:
 *      Bit stream irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

static int
ParseSlice(MPEGFrame *framePtr)
{
    unsigned int data;
    MPEGPlay *objPtr;
    MPEGDevice *devicePtr;
    MPEGSoftwareDevice *softDevPtr;
    MPEGSoftwareDecodeState *decodeState;

    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    softDevPtr = (MPEGSoftwareDevice *) devicePtr->deviceData;
    decodeState = (MPEGSoftwareDecodeState *) framePtr->deviceData;

    /* Flush slice start code. */

    bp_Flush(24);

    /* Parse off slice vertical position. */

    bp_Get(8, &data);
    decodeState->slice.vert_pos = data;
    
    /* Parse off quantization scale. */
    
    bp_Get(5, &data);
    decodeState->slice.quant_scale = data;
    
    /* Parse off extra bit slice info. */

    ParseExtraBitInfo();

    /* Reset past intrablock address. */

    decodeState->mblock.past_intra_addr = -2;

    /* Reset previous recon motion vectors. */

    decodeState->mblock.recon_right_for_prev = 0;
    decodeState->mblock.recon_down_for_prev = 0;
    decodeState->mblock.recon_right_back_prev = 0;
    decodeState->mblock.recon_down_back_prev = 0;

    /* Reset macroblock address. */

    decodeState->mblock.mb_address = ((decodeState->slice.vert_pos - 1) *
				      decodeState->mb_width) - 1;

    /* Reset past dct dc y, cr, and cb values. */

    decodeState->block.dct_dc_y_past = 1024 << 3;
    decodeState->block.dct_dc_cr_past = 1024 << 3;
    decodeState->block.dct_dc_cb_past = 1024 << 3;

    return PARSE_OK;
}


/*
 *--------------------------------------------------------------
 *
 * ParseMacroBlock --
 *
 *      Parseoff macroblock. Reconstructs DCT values. Applies
 *      inverse DCT, reconstructs motion vectors, calculates and
 *      set pixel values for macroblock in current display buffer.
 *
 * Results:
 *      Here's where everything really happens. Welcome to the
 *      heart of darkness.
 *
 * Side effects:
 *      Bit stream irreversibly parsed off.
 *
 *--------------------------------------------------------------
 */

static int
ParseMacroBlock(MPEGFrame *framePtr)
{
    int addr_incr;
    unsigned int data;
    int mask, i, recon_right_for, recon_down_for, recon_right_back,
    recon_down_back;
    int zero_block_flag;
    int mb_quant, mb_motion_forw, mb_motion_back, mb_pattern;
    MPEGPlay *objPtr;
    MPEGDevice *devicePtr;
    MPEGSoftwareDevice *softDevPtr;
    MPEGSoftwareDecodeState *decodeState;
    
    objPtr = framePtr->objPtr;
    devicePtr = objPtr->devicePtr;
    softDevPtr = (MPEGSoftwareDevice *) devicePtr->deviceData;
    decodeState = (MPEGSoftwareDecodeState *) framePtr->deviceData;
    
    /*
     * Parse off macroblock address increment and add to macroblock address.
     */
    do {
	DecodeMBAddrInc(addr_incr);
	if (addr_incr == MACRO_BLOCK_ESCAPE) {
	    decodeState->mblock.mb_address += 33;
	    addr_incr = MACRO_BLOCK_STUFFING;
	}
    } while (addr_incr == MACRO_BLOCK_STUFFING);
    decodeState->mblock.mb_address += addr_incr;

    if (decodeState->mblock.mb_address > (decodeState->mb_height *
					  decodeState->mb_width - 1))
      return MB_DECODE_ERROR;
    
    /*
     * If macroblocks have been skipped, process skipped macroblocks.
     */

    if (decodeState->mblock.mb_address - decodeState->mblock.past_mb_addr > 1) {
	if (decodeState->picture.type == MPEG_P_FRAME)
	  ProcessSkippedPFrameMBlocks(framePtr);
	else if (decodeState->picture.type== MPEG_B_FRAME)
	  ProcessSkippedBFrameMBlocks(framePtr);
    }

    /* Set past macroblock address to current macroblock address. */
    decodeState->mblock.past_mb_addr = decodeState->mblock.mb_address;

    /* Based on picture type decode macroblock type. */
    switch (decodeState->picture.type) {
      case MPEG_I_FRAME:

	DecodeMBTypeI(mb_quant, mb_motion_forw, mb_motion_back, mb_pattern,
		      decodeState->mblock.mb_intra);
    break;

      case MPEG_P_FRAME:
	DecodeMBTypeP(mb_quant, mb_motion_forw, mb_motion_back, mb_pattern,
		      decodeState->mblock.mb_intra);
	break;

      case MPEG_B_FRAME:
	DecodeMBTypeB(mb_quant, mb_motion_forw, mb_motion_back, mb_pattern,
		      decodeState->mblock.mb_intra);
	break;
    }

    /* If quantization flag set, parse off new quantization scale. */

    if (mb_quant == 1) {
	bp_Get(5, &data);
	decodeState->slice.quant_scale = data;
    }
    /* If forward motion vectors exist... */
    if (mb_motion_forw == 1) {

	/* Parse off and decode horizontal forward motion vector. */
	DecodeMotionVectors(decodeState->mblock.motion_h_forw_code);

	/* If horiz. forward r data exists, parse off. */

	if ((decodeState->picture.forw_f != 1) &&
	    (decodeState->mblock.motion_h_forw_code != 0)) {
	    bp_Get(decodeState->picture.forw_r_size, &data);
	    decodeState->mblock.motion_h_forw_r = data;
	}

	/* Parse off and decode vertical forward motion vector. */
	DecodeMotionVectors(decodeState->mblock.motion_v_forw_code);

	/* If vert. forw. r data exists, parse off. */

	if ((decodeState->picture.forw_f != 1) &&
	    (decodeState->mblock.motion_v_forw_code != 0)) {
	    bp_Get(decodeState->picture.forw_r_size, &data);
	    decodeState->mblock.motion_v_forw_r = data;
	}
    }
    /* If back motion vectors exist... */
    if (mb_motion_back == 1) {

	/* Parse off and decode horiz. back motion vector. */
	DecodeMotionVectors(decodeState->mblock.motion_h_back_code);

	/* If horiz. back r data exists, parse off. */

	if ((decodeState->picture.back_f != 1) &&
	    (decodeState->mblock.motion_h_back_code != 0)) {
	    bp_Get(decodeState->picture.back_r_size, &data);
	    decodeState->mblock.motion_h_back_r = data;
	}
	/* Parse off and decode vert. back motion vector. */
	DecodeMotionVectors(decodeState->mblock.motion_v_back_code);

	/* If vert. back r data exists, parse off. */
	
	if ((decodeState->picture.back_f != 1) &&
	    (decodeState->mblock.motion_v_back_code != 0)) {
	    bp_Get(decodeState->picture.back_r_size, &data);
	    decodeState->mblock.motion_v_back_r = data;
	}
    }

    /* If mblock pattern flag set, parse and decode CBP (code block pattern). */
    if (mb_pattern == 1) {
	DecodeCBP(decodeState->mblock.cbp);
    }
    /* Otherwise, set CBP to zero. */
    else
      decodeState->mblock.cbp = 0;

    /* Reconstruct motion vectors depending on picture type. */
    if (decodeState->picture.type == MPEG_P_FRAME) {

	/*
	 * If no forw motion vectors, reset previous and current vectors to 0.
	 */

	if (!mb_motion_forw) {
	    recon_right_for = 0;
	    recon_down_for = 0;
	    decodeState->mblock.recon_right_for_prev = 0;
	    decodeState->mblock.recon_down_for_prev = 0;
	}
	/*
	 * Otherwise, compute new forw motion vectors. Reset previous vectors to
	 * current vectors.
	 */

	else {
	    ComputeForwVector(decodeState, &recon_right_for, &recon_down_for);
	}
    }
    if (decodeState->picture.type == MPEG_B_FRAME) {
	
	/* Reset prev. and current vectors to zero if mblock is intracoded. */
	
	if (decodeState->mblock.mb_intra) {
	    decodeState->mblock.recon_right_for_prev = 0;
	    decodeState->mblock.recon_down_for_prev = 0;
	    decodeState->mblock.recon_right_back_prev = 0;
	    decodeState->mblock.recon_down_back_prev = 0;
	} else {
	    
	    /* If no forw vectors, current vectors equal prev. vectors. */
	    
	    if (!mb_motion_forw) {
		recon_right_for = decodeState->mblock.recon_right_for_prev;
		recon_down_for = decodeState->mblock.recon_down_for_prev;
	    }
	    /*
	     * Otherwise compute forw. vectors. Reset prev vectors to new values.
	     */
	    
	    else {
		ComputeForwVector(decodeState, &recon_right_for, &recon_down_for);
	    }

	    /* If no back vectors, set back vectors to prev back vectors. */

	    if (!mb_motion_back) {
		recon_right_back = decodeState->mblock.recon_right_back_prev;
		recon_down_back = decodeState->mblock.recon_down_back_prev;
	    }
	    /* Otherwise compute new vectors and reset prev. back vectors. */

	    else {
		ComputeBackVector(decodeState, &recon_right_back, &recon_down_back);
	    }
	    
	    /*
	     * Store vector existance flags in structure for possible skipped
	     * macroblocks to follow.
	     */
	    
	    decodeState->mblock.bpict_past_forw = mb_motion_forw;
	    decodeState->mblock.bpict_past_back = mb_motion_back;
	}
    }

    for (mask = 32, i = 0; i < 6; mask >>= 1, i++) {
	
	/* If block exists... */
	if ((decodeState->mblock.mb_intra) || (decodeState->mblock.cbp & mask)) {
	    zero_block_flag = 0;
	    ParseReconBlock(framePtr, i);
	} else {
	    zero_block_flag = 1;
	}
	
	/* If macroblock is intra coded... */
	if (decodeState->mblock.mb_intra) {
	    ReconIMBlock(framePtr, i);
	} else if (mb_motion_forw && mb_motion_back) {
	    ReconBiMBlock(framePtr, i, recon_right_for, recon_down_for,
			  recon_right_back, recon_down_back, zero_block_flag);
	} else if (mb_motion_forw || (decodeState->picture.type == MPEG_P_FRAME)) {
	    ReconPMBlock(framePtr, i, recon_right_for, recon_down_for,
			 zero_block_flag);
	} else if (mb_motion_back) {
	    ReconBMBlock(framePtr, i, recon_right_back, recon_down_back,
			 zero_block_flag);
	}
    }

    /* If D Type picture, flush marker bit. */
    if (decodeState->picture.type == MPEG_D_FRAME) {
	bp_Flush(1);
    }

    /* If macroblock was intracoded, set macroblock past intra address. */
    if (decodeState->mblock.mb_intra)
      decodeState->mblock.past_intra_addr =
	decodeState->mblock.mb_address;

    return PARSE_OK;
}

#else /* MPEG_SW_DEVICE */

/*****************************************************************
 *
 * GetSoftwareMpegDevice
 *
 * Parameters:
 *     objPtr - pointer to mpeg play object.
 *
 * Return:
 *     Pointer to new software device.
 *
 * Action:
 *     Stub for compiling without software device.
 *
 *****************************************************************/
MPEGDevice *
GetSoftwareMpegDevice (MPEGPlay *objPtr) 
{

    return NULL;
}

#endif /* MPEG_SW_DEVICE */
